/* Eirik's Miscellaneous Modified NetBeans Libraries.
   Written or modified (as noted) by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. */

/* This file is based in part on:
 *   org.netbeans.modules.editor.impl.ActionsList
 *   org.openide.util.Utilities
 * The complete license headers for the original files can be found in the LICENSE_ActionsList.txt
 * and LICENSE_openideUtilities.txt files, respectively.
 */
package com.sieuferd.upstream.actions;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.upstream.UpstreamUtil;
import java.awt.Component;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;
import org.openide.cookies.InstanceCookie;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataFolder;
import org.openide.loaders.DataObject;
import org.openide.util.ContextAwareAction;
import org.openide.util.Exceptions;
import org.openide.util.Lookup;
import org.openide.util.Utilities;
import org.openide.util.actions.ActionPresenterProvider;
import org.openide.util.actions.Presenter;

public final class ActionUtil {
  private static final Logger LOG = Logger.getLogger(ActionUtil.class.getName());

  private ActionUtil() { }

  /**
   * Retrieve an ordered map of FileObject/Object pairs registered in the system filesystem under
   * the given path. Separators will be assigned null values. Duplicate successive values will be
   * skipped, as will leading and trailing separators.
   */
  public static Map<FileObject,Object> getRegisteredObjects(String path) {
    Preconditions.checkNotNull(path);
    final Map<FileObject,Object> ret = Maps.newLinkedHashMap();
    final List<FileObject> fileObjects = FileUtil.getOrder(
        Arrays.asList(FileUtil.getConfigFile(path).getChildren()), true);
    FileObject previousKey = null;
    Object previousValue = null;
    for (FileObject fileObject : fileObjects) {
      final Object item;
      try {
        item = ActionUtil.convertFileObject(fileObject);
      } catch (Exception e) {
        LOG.log(Level.WARNING, "Failed to convert folder item to object", e);
        continue;
      }
      // Skip duplicate items and leading separators.
      if (!Utilities.compareObjects(previousValue, item)) {
        ret.put(fileObject, item);
        previousKey = fileObject;
        previousValue = item;
      }
    }
    // Remove any trailing separator.
    if (previousKey != null && previousValue == null)
      ret.remove(previousKey);
    return Collections.unmodifiableMap(ret);
  }

  private static Object convertFileObject(FileObject item)
      throws IOException, ClassNotFoundException {
    // A simplified version of the logic in ActionsList.
    final DataObject dob = DataObject.find(item);
    final InstanceCookie ic = dob.getLookup().lookup(InstanceCookie.class);
    if (ic != null) {
      if (isSeparator(ic)) {
        return null;
      } else {
        Object ret = ic.instanceCreate();
        return (ret instanceof JSeparator) ? null : ret;
      }
    } else if (dob instanceof DataFolder) {
      return dob;
    } else {
      return dob.getName();
    }
  }

  // From ActionsList.
  private static boolean isSeparator(InstanceCookie o) throws IOException, ClassNotFoundException {
    return (o instanceof InstanceCookie.Of && ((InstanceCookie.Of) o).instanceOf(JSeparator.class))
        || JSeparator.class.isAssignableFrom(o.instanceClass());
  }

  /**
   * Version of {@link Utilities#actionsToPopup(Action[], Lookup)} that supports icons in the
   * returned popup.
   */
  public static JPopupMenu actionsToPopup(Action[] actions, Lookup context, boolean withIcons) {
    /* Copied verbatim from org.openide.util.Utilities.actionsToPopup, with logic added to support
    the withIcons parameter. */

    // keeps actions for which was menu item created already (do not add them twice)
    final Set<Action> counted = new HashSet<>();
    // components to be added (separators are null)
    final List<Component> components = new ArrayList<>();

    for (Action action : actions) {
      if (action != null && counted.add(action)) {
        // switch to replacement action if there is some
        if (action instanceof ContextAwareAction) {
          Action contextAwareAction =
              ((ContextAwareAction) action).createContextAwareInstance(context);
          if (contextAwareAction == null) {
            Logger.getLogger(ActionUtil.class.getName()).log(Level.WARNING,
                "createContextAwareInstance(context) returns null. That is illegal!" +
                " action={0}, context={1}", new Object[]{ action, context });
          } else {
            action = contextAwareAction;
          }
        }

        final JMenuItem item;
        if (action instanceof Presenter.Popup) {
          item = ((Presenter.Popup) action).getPopupPresenter();
          if (item == null) {
            Logger.getLogger(ActionUtil.class.getName()).log(Level.WARNING,
                "findContextMenuImpl, getPopupPresenter returning null for {0}", action);
            continue;
          }
        } else {
          item = ActionPresenterProvider.getDefault().createPopupPresenter(action);
        }

        for (Component c : ActionPresenterProvider.getDefault().convertComponents(item)) {
          if (c instanceof JSeparator) {
            components.add(null);
          } else {
            // This logic was added.
            if (withIcons && (c instanceof JMenuItem)) {
              Object icon = null;
              if (UpstreamUtil.USE_CUSTOM_LOADIMAGEICON) {
                Object iconBase = action.getValue("iconBase");
                if (iconBase instanceof String)
                  icon = UpstreamUtil.loadImageIcon((String) iconBase);
              }
              if (icon == null)
                icon = action.getValue(Action.SMALL_ICON);
              if (icon instanceof Icon)
                ((JMenuItem) c).setIcon((Icon) icon);
            }

            components.add(c);
          }
        }
      } else {
        components.add(null);
      }
    }

    // Now create actual menu. Strip adjacent, leading, and trailing separators.
    final JPopupMenu menu = ActionPresenterProvider.getDefault().createEmptyPopup();
    boolean nonempty = false; // has anything been added yet?
    boolean pendingSep = false; // should there be a separator before any following item?
    for (Component c : components) {
      try {
        if (c == null) {
          pendingSep = nonempty;
        } else {
          nonempty = true;
          if (pendingSep) {
            pendingSep = false;
            menu.addSeparator();
          }
          menu.add(c);
        }
      } catch (RuntimeException ex) {
        Exceptions.attachMessage(ex, "Current component: " + c); // NOI18N
        Exceptions.attachMessage(ex, "List of components: " + components); // NOI18N
        Exceptions.attachMessage(ex, "List of actions: " + Arrays.asList(actions)); // NOI18N
        Exceptions.printStackTrace(ex);
      }
    }
    return menu;
  }
}
