/* Eirik's Miscellaneous Modified NetBeans Libraries.
   Written or modified (as noted) by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. */

/* This file is based in part on:
 *   org.netbeans.editor.Utilities (with listed author Miloslav Metelka)
 * The complete license header for the original file can be found in the LICENSE_editorUtilities.txt
 * file.
 */
package com.sieuferd.upstream.localeditor;

import com.sieuferd.uiutil.api.CaretInfoProvider;
import com.sieuferd.uiutil.api.KeyStrokeForwardable;
import com.sieuferd.uiutil.api.CommonUIUtil;
import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Insets;
import java.awt.KeyboardFocusManager;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseWheelEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JViewport;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.border.Border;
import javax.swing.border.CompoundBorder;
import javax.swing.border.MatteBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.plaf.TextUI;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;
import javax.swing.text.Document;
import javax.swing.text.EditorKit;
import javax.swing.text.View;
import org.netbeans.api.editor.settings.SimpleValueNames;
import org.netbeans.editor.BaseTextUI;
import org.netbeans.editor.EditorUI;
import org.netbeans.editor.Utilities;
import org.netbeans.lib.editor.util.swing.DocumentListenerPriority;
import org.netbeans.lib.editor.util.swing.DocumentUtilities;

/* TODO[pd]: File a bug about the selection area not being properly repainted when a selection is
             made across two lines and the caret is moved, or find a workaround. It might only
             happen when CaretRowHighlighting is excluded, since the caret row highlighting change
             would force a repaint anyway. */
/**
 * An editor component suitable for use in a dialog box or toolbar. It builds on the logic used by
 * {@link org.netbeans.editor.Utilities#createSingleLineEditor(String)}, but is more flexible.
 * Features:
 *
 * <ul>
 *   <li>Full-fledged NetBeans editor; supports syntax coloring, code completion, etc.</li>
 *   <li>The editor can be a specific number of rows tall, and manages its own internal scroll
 *       pane.</li>
 *   <li>Includes a number of tweaks to improve how the scroll viewport moves with the caret.</li>
 *   <li>Automatically activates line-wrapping in multi-row mode.</li>
 *   <li>Deactivates highlighting of the line on which the caret is located.</li>
 *   <li>Implementation avoids L&amp;F-specific tweaks. Component is confirmed to have correct size
 *       on MacOS/Aqua and Windows.</li>
 * </ul>
 *
 * <p>Some other relevant tweaks, such as hiding the text limit line, can be done in the FontsColors
 * layer settings on the MIME type level.
 */
public final class LocalEditor extends JPanel implements KeyStrokeForwardable {
  private static final Logger LOG = Logger.getLogger(LocalEditor.class.getName());
  private final LocalEditorPane editorPane = new LocalEditorPane();
  private final JScrollPane scrollPane = new JScrollPane(editorPane,
      JScrollPane.VERTICAL_SCROLLBAR_NEVER, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
  private final PropertyChangeListener lineHeightListener = (PropertyChangeEvent evt) -> {
    SwingUtilities.invokeLater(() -> {
      editorPane.invalidate();
      LocalEditor.this.revalidate();
    });
  };
  private final ComponentListener viewportComponentListener = new ComponentAdapter() {
    @Override
    public void componentResized(ComponentEvent e) {
      /* This might not be needed, since adjustViewportPosition() is already called when the
      document is changed, but do it to be safe. */
      adjustViewportPosition();
    }
  };
  private boolean wrapping;

  public LocalEditor() {
    this(null);
  }

  public LocalEditor(@Nullable Border customBorder) {
    initComponents(customBorder);
  }

  // Used to forward keystrokes from the formula bar to the CellWidget.
  /**
   * Set a {@code KeyStrokeForwardable} to forward keystrokes to, <em>from</em> the editor.
   */
  public void setKeyStrokeForwardable(@Nullable KeyStrokeForwardable keyStrokeForwardable) {
    editorPane.setKeyStrokeForwardable(keyStrokeForwardable);
  }

  public @Nullable KeyStrokeForwardable getKeyStrokeForwardable() {
    return editorPane.getKeyStrokeForwardable();
  }

  /* Maintaining the usedByCloneableEditor property seems necessary to avoid losing code completion
  behavior when switching between cloned TopComponent containers containing the LocalEditor. See
  http://forums.netbeans.org/topic52812.html . */
  /**
   * Get the underlying {@code JEditorPane}. The client may call
   * {@link JEditorPane#setEditorKit(EditorKit)} and optionally
   * {@link JEditorPane#setDocument(Document)} to further configure the editor. For editors that
   * will be registered in NetBeans' editor registry, including editors using NetBeans-based editor
   * kits, the client should also set the {@code usedByCloneableEditor} client property to true
   * whenever the editor is in use and to false otherwise. The returned component implements
   * {@link CaretInfoProvider}.
   */
  public JEditorPane getEditorPane() {
    return editorPane;
  }

  private void resetViewportPosition() {
    scrollPane.getViewport().setViewPosition(new Point(0, 0));
  }

  // Used to forward keystrokes from the CellWidget to the inplace editor.
  /** Forward a keystroke <em>to</em> the editor. */
  @Override
  public boolean forwardUnhandledKeyStroke(KeyStroke ks, KeyEvent evt, boolean pressed) {
    return editorPane.processKeyBinding(ks, evt, JComponent.WHEN_FOCUSED, pressed);
  }

  /** @param rows may be null to size the editor vertically on demand */
  public void setVisibleRows(@Nullable Integer rows) {
    CommonUIUtil.checkOnEDT();
    /* I considered setting the verticalScrollBarPolicy to show vertical scrollbars whenever more
    than one row is being displayed, however this causes layout bugs on MacOS. Besides, I will
    eventually want the vertical scrollbar to line up with the button affordance for expanding the
    editor. So the best way to introduce a vertical scrollbar will probably be to use an external
    JScrollBar component that is not directly associated with the scrollPane. */

    editorPane.setRows(rows);
    revalidate();
  }

  public @Nullable Integer getVisibleRows() {
    return editorPane.getRows();
  }

  public void setWrapping(final boolean wrapping) {
    if (this.wrapping != wrapping) {
      this.wrapping = wrapping;
      updateWrapping();
    }
  }

  public boolean isWrapping() {
    return wrapping;
  }

  private void updateWrapping() {
    // This is one of a few settings that can actually be set on a per-component basis.
    editorPane.putClientProperty(SimpleValueNames.TEXT_LINE_WRAP, isWrapping() ? "words" : "none");
    /* Allow the editor to resize itself before we apply viewport positioning tweaks. I'm using
    validate() rather than revalidate() here, as the former is synchronous. */
    invalidate();
    validate();
    scrollToCursor();
  }

  private void scrollToCursor() {
    try {
      final TextUI ui = editorPane.getUI();
      final Caret caret = editorPane.getCaret();
      if (ui == null || caret == null)
        return;
      final Rectangle dotRect = ui.modelToView(editorPane, caret.getDot());
      final Rectangle markRect = ui.modelToView(editorPane, caret.getMark());
      // EditorUI.modelToView may sometimes return null, for instance during initialization.
      /* Scroll to the mark first to have a better chance of getting both the selection and the
      cursor dot in the viewport. */
      if (markRect != null)
        editorPane.scrollRectToVisible(markRect);
      if (dotRect != null)
        editorPane.scrollRectToVisible(dotRect);
    } catch (BadLocationException e) {
      LOG.log(Level.WARNING, "Unexpected exception while finding cursor view position", e);
    }
    adjustViewportPosition();
  }

  private void initComponents(@Nullable Border customBorder) {
    CommonUIUtil.checkOnEDT();

    // Avoid trapping scroll wheel events in the editor's own scroll pane.
    scrollPane.setWheelScrollingEnabled(false);
    scrollPane.addMouseWheelListener((MouseWheelEvent e) -> {
      /* Unconsumed mouse events do not 'bubble' up by themselves in Swing, so send it up in the
      component hierarchy manually. */
      Container parent = scrollPane.getParent();
      if (parent != null)
        parent.dispatchEvent(e);
    });

    // Don't let the scroll pane handle any keyboard inputs by itself.
    scrollPane.setInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, new InputMap());

    /* This does not seem to do much; line wrapping works fine without it, and it is still necessary
    do apply the scrolling adjustment in manageViewPositionListener. But if it does what I think it
    does, namely avoiding the behavior mentioned in
    https://netbeans.org/bugzilla/show_bug.cgi?id=241652 , then it might prevent bugs that I have
    yet to discover. */
    editorPane.putClientProperty("document-view-accurate-span", true);

    setVisibleRows(1);
    updateWrapping();

    /* I do _not_ use the AsTextField client property here, as it is too over-reaching in its
    modifications to editor behavior. Notably, it does not permit multi-line editing. Instead, I
    have gone through all the places in the NetBeans source code that access the AsTextField
    property and implemented only the specific tweaks I have found relevant.

      editorPane.putClientProperty("AsTextField", Boolean.TRUE)
    */

    editorPane.addFocusListener(focusListener);
    updateHighlightExcludes();

    /* The bug on https://netbeans.org/bugzilla/show_bug.cgi?id=110237 seems to suggest that it
    might be necessary to call org.netbeans.modules.editor.lib2.EditorApiPackageAccessor.register
    reflectively in order to make code completion work propertly; however, code completion seems to
    work fine without it when I have tried it. I think the explicit registration is only needed when
    using a JTextField; when using a JEditorPane, I think setEditorKit takes care of the
    registration through BaseTextUI.installUI. */

    editorPane.addPropertyChangeListener((PropertyChangeEvent evt) -> {
      final String propName = evt.getPropertyName();
      if ("editorKit".equals(propName)) {
        editorPane.invalidate();
        LocalEditor.this.revalidate();
        updateWrapping();
      } else if ("document".equals(propName)) {
        final Document oldDoc = (Document) evt.getOldValue();
        if (oldDoc != null) {
          DocumentUtilities.removeDocumentListener(oldDoc, documentListener,
              DocumentListenerPriority.AFTER_CARET_UPDATE);
        }
        final Document newDoc = (Document) evt.getNewValue();
        if (newDoc != null) {
          DocumentUtilities.addDocumentListener(newDoc, documentListener,
              DocumentListenerPriority.AFTER_CARET_UPDATE);
        }
        resetViewportPosition();
      } else if ("caret".equals(propName)) {
        final Caret oldCaret = (Caret) evt.getOldValue();
        if (oldCaret != null)
          oldCaret.removeChangeListener(caretChangeListener);
        final Caret newCaret = (Caret) evt.getNewValue();
        if (newCaret != null)
          newCaret.addChangeListener(caretChangeListener);
      } else if ("UI".equals(propName)) {
        /* Use the same logic to get to the EditorUI as in Utilities.getEditorUI. For a given
        instance of BaseTextUI, getEditorUI() will always return the same object */
        final TextUI oldUI = (TextUI) evt.getOldValue();
        if (oldUI instanceof BaseTextUI) {
          ((BaseTextUI) oldUI).getEditorUI().removePropertyChangeListener(
              EditorUI.LINE_HEIGHT_CHANGED_PROP, lineHeightListener);
        }
        final TextUI newUI = (TextUI) evt.getNewValue();
        if (newUI instanceof BaseTextUI) {
          ((BaseTextUI) newUI).getEditorUI().addPropertyChangeListener(
              EditorUI.LINE_HEIGHT_CHANGED_PROP, lineHeightListener);
        }
      }
    });
    // Initial attachments.
    DocumentUtilities.addDocumentListener(editorPane.getDocument(), documentListener,
        DocumentListenerPriority.AFTER_CARET_UPDATE);
    editorPane.getCaret().addChangeListener(caretChangeListener);
    {
      final EditorUI editorUI = Utilities.getEditorUI(editorPane);
      if (editorUI != null)
        editorUI.addPropertyChangeListener(EditorUI.LINE_HEIGHT_CHANGED_PROP, lineHeightListener);
    }

    {
      final JTextField referenceTextField = new JTextField("M");
      /* This is probably irrelevant for the formula bar, but keep it in here since LocalEditor is
      supposed to be a general-purpose component. */
      editorPane.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
          referenceTextField.getFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS));
      editorPane.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
          referenceTextField.getFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS));
    }

    {
      /* Remove all margins from the editor component, instead rendering them in the enclosing
      JScrollPane. This ensures that the viewport height matches the height of the desired number of
      visible rows exactly. */

      if (customBorder != null)
        scrollPane.setBorder(customBorder);
      final Insets noMargin = new Insets(0, 0, 0, 0);
      editorPane.setMargin(noMargin);
      final Insets margin = UIManager.getInsets("TextField.margin");
      if (margin != null && !margin.equals(noMargin)) {
        // TODO[pd]: Keep the color of the inner border in sync with the color of the editorPane.
        scrollPane.setBorder(new CompoundBorder(scrollPane.getBorder(),
            new MatteBorder(margin, editorPane.getBackground())));
      }
    }
    setLayout(new BorderLayout());
    add(scrollPane, BorderLayout.CENTER);

    scrollPane.getViewport().addComponentListener(viewportComponentListener);
  }

  private static final String[] HIGHLIGHT_EXCLUDES_FOCUS = new String[] {
    "^org\\.netbeans\\.modules\\.editor\\.lib2\\.highlighting\\.CaretRowHighlighting$",
    "^org-netbeans-modules-editor-bracesmatching-BracesMatchHighlighting$"
  };
  private static final String[] HIGHLIGHT_EXCLUDES_NOFOCUS = new String[] {
    "^org\\.netbeans\\.modules\\.editor\\.lib2\\.highlighting\\.CaretRowHighlighting$",
    "^org-netbeans-modules-editor-bracesmatching-BracesMatchHighlighting$"
  };

  private void updateHighlightExcludes() {
    /* See "Filtering layers used for JTextComponent" from
    http://bits.netbeans.org/dev/javadoc/org-netbeans-modules-editor-lib2/index.html?org/netbeans/spi/editor/highlighting/package-summary.html .
    Another example of using this property can be seen in
    org.netbeans.modules.mercurial.ui.annotate.TooltipWindow . */
    editorPane.putClientProperty("HighlightsLayerExcludes",
        editorPane.isFocusOwner() ? HIGHLIGHT_EXCLUDES_FOCUS : HIGHLIGHT_EXCLUDES_NOFOCUS);
  }

  private final FocusListener focusListener = new FocusListener() {
    @Override
    public void focusGained(FocusEvent e) {
      updateHighlightExcludes();
    }

    @Override
    public void focusLost(FocusEvent e) {
      updateHighlightExcludes();
    }
  };

  // TODO[pd]: Remove the following workaround when the mentioned bug is resolved.
  /* Workaround for an editor bug where scrolling due to a caret position change happens before the
  editor has had time to update its preferred size. To reproduce the bug, use setVisibleRows(2) and
  add new lines in the editor. This bug is not normally exposed in the NetBeans editor because a
  "virtual space" equal to a third of the viewport's height is added to the preferred size of the
  editor, and this space is normally taller than a single line. If the virtual space is disabled
  using DISABLE_END_VIRTUAL_SPACE, the bug exhibits itself when lines are added at the bottom of an
  edited file; the viewport will be one step behind the cursor. See
  https://netbeans.org/bugzilla/show_bug.cgi?id=241898 . */
  private final ChangeListener caretChangeListener = (ChangeEvent e) -> {
    SwingUtilities.invokeLater(this::scrollToCursor);
  };

  private final DocumentListener documentListener = new DocumentListener() {
    @Override
    public void insertUpdate(DocumentEvent e) {
      changed();
    }

    @Override
    public void removeUpdate(DocumentEvent e) {
      changed();
    }

    @Override
    public void changedUpdate(DocumentEvent e) {
      changed();
    }

    private void changed() {
      firePropertyChange("preferredSize", null, getPreferredSize());
      /* Push the viewport adjustment to the end of the event queue to avoid being "one step behind"
      values we depend on to calculate the new viewport position. Doing this fixes real observable
      bugs in the original code, including characters going past the viewport when typing characters
      at the end of the string and the viewport leaving space for one extra character when deleting
      characters at the end of the string. DocumentListenerPriority.AFTER_CARET_UPDATE seems not to
      be late enough. */
      SwingUtilities.invokeLater(LocalEditor.this::adjustViewportPosition);
    }
  };

  private void adjustViewportPosition() {
    final JViewport viewport = scrollPane.getViewport();
    final Point viewPosition = viewport.getViewPosition();
    if (isWrapping()) {
      viewport.setViewPosition(new Point(0, viewPosition.y));
    } else {
      /* Scroll left to the end of the longest line. This is a hack to make JEditorPane scrolling
      behave more like that of a JTextField, most notably when the viewport is being scrolled all
      the way to the end of a long line and a character is deleted. It also ensures that a viewport
      that is enlarged to the full width of the text gets scrolled to actually display all of the
      text. */
      if (viewPosition.x > 0) {
        final int textWidth = editorPane.getAccuratePreferredSpan(View.X_AXIS);
        final int viewportWidth = viewport.getExtentSize().width;
        if (textWidth < (viewPosition.x + viewportWidth)) {
          // Align the end of the viewport with the end of the string.
          viewPosition.x = Math.max(textWidth - viewportWidth, 0);
          viewport.setViewPosition(viewPosition);
        }
      }
    }
  }
}
