/* Eirik's Miscellaneous Modified NetBeans Libraries.
   Written or modified (as noted) by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. */

package com.sieuferd.upstream.localeditor;

import com.google.common.base.Preconditions;
import com.sieuferd.uiutil.api.CaretInfoProvider;
import com.sieuferd.uiutil.api.KeyStrokeForwardable;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.KeyStroke;
import javax.swing.text.AbstractDocument;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import javax.swing.text.Element;
import javax.swing.text.Position.Bias;
import javax.swing.text.StyledDocument;
import javax.swing.text.View;
import org.netbeans.editor.EditorUI;
import org.netbeans.editor.Utilities;
import org.openide.text.NbDocument;

final class LocalEditorPane extends JEditorPane implements CaretInfoProvider {
  private static final Logger LOG = Logger.getLogger(LocalEditorPane.class.getName());
  private static final int WIDTH_CALCULATION_CUTOFF_LINES = 30;
  private @Nullable Integer rows = 1;
  private @Nullable KeyStrokeForwardable keyStrokeForwardable;

  public LocalEditorPane() {
  }

  @Override
  public Dimension getPreferredScrollableViewportSize() {
    // Assume no insets, as margins and borders need to be handled in the enclosing JScrollPane.
    return new Dimension(
        getAccuratePreferredSpan(View.X_AXIS),
        rows != null ? rows * getRowHeight() : getAccuratePreferredSpan(View.Y_AXIS));
  }

  private int getRowHeight() {
    final EditorUI editorUI = Utilities.getEditorUI(this);
    final int height;
    if (editorUI != null) {
      final int lineHeight = editorUI.getLineHeight();
      height = lineHeight > 1 ? lineHeight :
          // Same hack as used in Utilities.adjustScrollPaneSize.
          (editorUI.getLineAscent() * 4) / 3;
    } else {
      // Similar to JTextArea.getPreferredSize()/getRowHeight().
      height = getFontMetrics(getFont()).getHeight();
    }
    return height;
  }

  public void setRows(@Nullable Integer rows) {
    Preconditions.checkArgument(rows == null || rows >= 1);
    if (!Objects.equals(this.rows, rows)) {
      this.rows = rows;
      invalidate();
    }
  }

  public @Nullable Integer getRows() {
    return rows;
  }

  public void setKeyStrokeForwardable(@Nullable KeyStrokeForwardable keyStrokeForwardable) {
    this.keyStrokeForwardable = keyStrokeForwardable;
  }

  public @Nullable KeyStrokeForwardable getKeyStrokeForwardable() {
    return keyStrokeForwardable;
  }

  @Override
  protected boolean processKeyBinding(KeyStroke ks, KeyEvent evt, int condition, boolean pressed) {
    Preconditions.checkNotNull(ks);
    Preconditions.checkNotNull(evt);
    final boolean ret = super.processKeyBinding(ks, evt, condition, pressed);
    if (!ret && condition == JComponent.WHEN_FOCUSED && keyStrokeForwardable != null) {
      return keyStrokeForwardable.forwardUnhandledKeyStroke(ks, evt, pressed);
    } else {
      return ret;
    }
  }

  // TODO[pd]: Consider if optimizations are necessary to avoid calling this method too often.
  /**
   * Calculate the preferred size of the document, in pixels. This method is more accurate than the
   * implementation provided by NetBeans' {@code EditorKit} implementation, which pads the size in
   * the vertical direction with a fraction of the viewport size, and which only maintains an
   * ever-growing maximum line width in the horizontal direction.
   *
   * @param axis must be either {@link View#X_AXIS} or {@link View#Y_AXIS}
   */
  final int getAccuratePreferredSpan(final int axis) {
    Preconditions.checkArgument(axis == View.X_AXIS || axis == View.Y_AXIS);
    final Document doc = getDocument();
    final Integer retRef[] = new Integer[] { null };
    if ((doc instanceof StyledDocument) && (doc instanceof AbstractDocument)) {
      ((AbstractDocument) doc).render(() -> {
        /* See org.netbeans.modules.mercurial.ui.annotate.TooltipWindow for an example of how to
        iterate over lines.  */
        /* "[T]he default root element of the document has child elements corresponding to all text
        lines; and if using StyledDocument, [...] "paragraph" elements again correspond to text
        lines."
        http://bits.netbeans.org/dev/javadoc/org-openide-text/org/openide/text/doc-files/api.html
        */
        final Element rootElement = NbDocument.findLineRootElement((StyledDocument) doc);
        final int lineCount = rootElement.getElementCount();
        int ret = 0;
        try {
          if (axis == View.X_AXIS) {
            /* TODO[pd]: If the bug below gets fixed, it might be possible to avoid the manual
                         width calculation here:
                         https://netbeans.org/bugzilla/show_bug.cgi?id=241652 */
            if (lineCount > WIDTH_CALCULATION_CUTOFF_LINES)
              return;
            for (int i = 0; i < lineCount; i++) {
              final Element lineElement = rootElement.getElement(i);
              // Unsure if this would work with RTL content.
              /* Subtract one from getEndOffset to avoid the offset representing the line break,
              since it actually lies on the next line if said next line exists. */
              final Rectangle textRect =
                  getUI().modelToView(LocalEditorPane.this, lineElement.getEndOffset() - 1);
              // I once observed modelToView to return null during initialization.
              if (textRect == null)
                return;
              ret = Math.max(ret, textRect.x + textRect.width);
            }
          } else {
            if (lineCount > 0) {
              final Element lastLineElement = rootElement.getElement(lineCount - 1);
              /* Subtract one from getEndOffset to avoid the offset representing the line break
              (see above). */
              for (int offset = lastLineElement.getStartOffset();
                  offset <= lastLineElement.getEndOffset() - 1;
                  offset++)
              {
                final Rectangle textRect = getUI().modelToView(LocalEditorPane.this, offset);
                // See above.
                if (textRect == null)
                  return;
                ret = Math.max(ret, textRect.y + textRect.height);
              }
            }
          }
        } catch (BadLocationException e) {
          LOG.log(Level.WARNING, "Unexpected exception while calculating document span", e);
          return;
        }
        retRef[0] = ret;
      });
    }
    if (retRef[0] == null) {
      // Fall back on the EditorKit's implementation.
      final Runnable runnable = () -> {
        retRef[0] = (int) Math.min(Integer.MAX_VALUE, Math.ceil(
            getUI().getRootView(LocalEditorPane.this).getPreferredSpan(axis)));
      };
      if (doc instanceof AbstractDocument) {
        ((AbstractDocument) doc).render(runnable);
      } else {
        runnable.run();
      }
    }
    return retRef[0];
  }

  @Override
  public boolean isTextCaretAtEdge(final boolean top) {
    final Document doc = getDocument();
    final boolean retRef[] = new boolean[] { false };
    if ((doc instanceof StyledDocument) && (doc instanceof AbstractDocument)) {
      ((AbstractDocument) doc).render(() -> {
        // This implementation works both with wrapping enabled and disabled.
        final int caretPos = getCaretPosition();
        try {
          // Not sure the bias is relevant here, but set it in the conservative direction anyway.
          final Rectangle caret = getUI().modelToView(LocalEditorPane.this, caretPos,
              top ? Bias.Forward : Bias.Backward);
          if (top) {
            final Rectangle start =
                getUI().modelToView(LocalEditorPane.this, 0, Bias.Backward);
            retRef[0] = caret.y <= start.y;
          } else {
            final Rectangle end =
                getUI().modelToView(LocalEditorPane.this, doc.getLength(), Bias.Forward);
            retRef[0] = caret.y >= end.y;
          }
        } catch (BadLocationException e) {
          LOG.log(Level.WARNING, "Unexpected error while calculating visual caret position", e);
        }
      });
    }
    return retRef[0];
  }
}
