/* Eirik's Miscellaneous Modified NetBeans Libraries.
   Written or modified (as noted) by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. */

/* This file is based in part on:
 *   org.netbeans.modules.editor.codegen.PopupUtil (with listed author Petr Hrebejk/phrebejk)
 *     as of 2014-10-10
 * The complete license header for the original file can be found in the LICENSE_PopupUtil.txt file.
 */
package com.sieuferd.upstream.popup;

import com.sieuferd.uiutil.api.PopupLike;
import com.google.common.base.Preconditions;
import com.sieuferd.uiutil.api.CommonUIUtil;
import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.AWTEventListener;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowStateListener;
import javax.accessibility.AccessibleContext;
import javax.annotation.Nullable;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JRootPane;
import javax.swing.JWindow;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import org.openide.util.Utilities;
import org.openide.windows.WindowManager;

/* This class is a refactored and modified version of PopupUtil, a class that seems to have been
copied around the NetBeans codebase quite a number of times. Here are all the files named
"PopupUtil.java" per a checkout from February 14, 2014:

     cnd.utils/src/org/netbeans/modules/cnd/utils/ui/PopupUtil.java
     csl.api/src/org/netbeans/modules/csl/editor/hyperlink/PopupUtil.java
     editor/src/org/netbeans/modules/editor/codegen/PopupUtil.java
     gototest/src/org/netbeans/modules/gototest/PopupUtil.java
     java.editor/src/org/netbeans/modules/java/editor/overridden/PopupUtil.java
     php.editor/src/org/netbeans/modules/php/editor/actions/PopupUtil.java
     vmd.io/src/org/netbeans/modules/vmd/api/io/PopupUtil.java
     web.beans/src/org/netbeans/modules/web/beans/navigation/PopupUtil.java

The files above contain nearly identical logic, but org.netbeans.modules.editor.codegen.PopupUtil
seems to be the most complete one (containing all the logic found in each of the other files). */

public class WindowPopupLike extends PopupLike {
  private static final String POPUP_NAME = "popupComponent"; // NOI18N
  private static final String OK_KEY = "OKKey"; // NOI18N
  private static final String CANCEL_KEY = "CancelKey"; // NOI18N
  private static final KeyStroke ESC_KEY_STROKE = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0);
  private static final KeyStroke ENTER_KEY_STROKE = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);

  private final Action OK_ACTION = new CloseAction(false);
  private final Action CANCEL_ACTION = new CloseAction(true);
  // TODO[cc]: Make these anonymous classes to save some space.
  private final AWTEventListener hideAWTListener = new HideAWTListener();
  private final WindowStateListener hideWindowStateListener = new HideWindowStateListener();
  private final ComponentListener hideComponentListener = new HideComponentListener();
  private final WindowListener popupClosedItselfListener = new WindowAdapter() {
    @Override
    public void windowClosing(WindowEvent e) {
      /* Ensure that the proper popup event is fired in case the window is closed using the corner
      close button (for decorated windows). */
      cancel();
    }
  };

  private final JComponent content;
  private final boolean undecorated;
  private final @Nullable Dimension size;
  // TODO[pd]: Consider reusing the same window if the owner hasn't changed.
  private @Nullable PopupWithOwner windowWithOwner;

  // Keep these together, since they will always be present or absent together.
  private static final class PopupWithOwner {
    final Window popup;
    final Window owner;

    public PopupWithOwner(Window popup, Window owner) {
      Preconditions.checkNotNull(popup);
      Preconditions.checkNotNull(owner);
      this.popup = popup;
      this.owner = owner;
    }
  }

  public WindowPopupLike(JComponent content, @Nullable Dimension size, boolean undecorated) {
    CommonUIUtil.checkOnEDT();
    Preconditions.checkNotNull(content);
    this.content = content;
    this.size = size;
    this.undecorated = undecorated;
  }

  @Override
  public void show(Component invoker, Point pos) {
    if (windowWithOwner != null)
      cancel();
    Preconditions.checkNotNull(invoker);
    Window owner = CommonUIUtil.getWindowForComponent(invoker);
    if (owner == null)
      owner = WindowManager.getDefault().getMainWindow();
    if (owner == null)
      throw new NullPointerException();
    final Window popup;
    final Container contentPane;
    final JRootPane rootPane;

    if (undecorated) {
      /* Using a JWindow rather than an undecorated JFrame or JDialog prevents the owner from
      changing its appearance when the popup has the focus (at least on MacOS/Aqua). Note that the
      JWindow needs to be assigned an owner in order to be focusable; see the Javadoc for
      Window.setFocusableWindowState. */
      final JWindow toPopup = new JWindow(owner);
      popup = toPopup;
      contentPane = toPopup.getContentPane();
      rootPane = toPopup.getRootPane();
    } else {
      // TODO[pd]: Consider using JFrame instead.
      final JDialog toPopup = new JDialog(owner);
      popup = toPopup;
      contentPane = toPopup.getContentPane();
      rootPane = toPopup.getRootPane();
      if (CommonUIUtil.isAqua())
        rootPane.putClientProperty("Window.style", "small");
    }
    this.windowWithOwner = new PopupWithOwner(popup, owner);
    popup.setName(POPUP_NAME);

    final InputMap im = rootPane.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
    im.put(ENTER_KEY_STROKE, OK_KEY);
    im.put(ESC_KEY_STROKE, CANCEL_KEY);
    final ActionMap am = rootPane.getActionMap();
    am.put(OK_KEY, OK_ACTION);
    am.put(CANCEL_KEY, CANCEL_ACTION);

    {
      final AccessibleContext accessibleContext = content.getAccessibleContext();
      if (accessibleContext != null) {
        final String accesibleName = accessibleContext.getAccessibleName();
        if (accesibleName != null && !accesibleName.isEmpty())
          popup.getAccessibleContext().setAccessibleName(accesibleName);
        final String accessibleDescription = accessibleContext.getAccessibleDescription();
        if (accessibleDescription != null && !accessibleDescription.isEmpty())
          popup.getAccessibleContext().setAccessibleDescription(accessibleDescription);
      }
    }

    contentPane.setLayout(new BorderLayout());
    contentPane.add(content, BorderLayout.CENTER);

    /* NOT using PopupFactory:
        1. on linux, creates mediumweight popup that doesn't refresh behind visible glasspane
        2. on mac, needs an owner frame otherwise hiding tooltip also hides the popup. (linux
           requires no owner frame to force heavyweight)
        3. the created window is not focusable window
    */

    Toolkit.getDefaultToolkit().addAWTEventListener(hideAWTListener, AWTEvent.MOUSE_EVENT_MASK);

    /* Leave both focusableWindowState and focusable at their default settings (true and false,
    respectively). The "focusable" property must be off to avoid the Window itself getting the focus
    during tab cycling, while the "focusableWindowState" must be on to allow components contained
    in the window to receive focus. */

    owner.addWindowStateListener(hideWindowStateListener);
    owner.addComponentListener(hideComponentListener);

    // Try to reset state before adjusting size, for more deterministic behavior in case of bugs.
    popup.invalidate();
    popup.setSize(0, 0);
    if (size == null) {
      popup.pack();
    } else {
      popup.invalidate();
      popup.setSize(size);
      popup.validate();
    }

    final Dimension actualSize = popup.getSize();
    final Point screenPos = convertPointToScreen(invoker, pos);
    final Rectangle screenBounds = getUsableScreenBounds(invoker, screenPos);
    // For testing:
    // screenBounds = new Rectangle(invoker.getLocationOnScreen(), new Dimension(800, 600));
    screenPos.x = Math.min(screenPos.x, screenBounds.x + screenBounds.width - actualSize.width);
    screenPos.y = Math.min(screenPos.y, screenBounds.y + screenBounds.height - actualSize.height);
    screenPos.x = Math.max(screenPos.x, screenBounds.x);
    screenPos.y = Math.max(screenPos.y, screenBounds.y);

    popup.setLocation(screenPos);

    popup.setVisible(true);
    popup.addWindowListener(popupClosedItselfListener);
    firePopupMenuEvent(PopupEventType.VISIBLE);
    /* Perform the focus requests after other events have been processed, otherwise
    requestFocusInWindow may fail (at least when called on intermediate JPanel containers). */
    SwingUtilities.invokeLater(() -> {
      content.requestFocus();
      content.requestFocusInWindow();
    });
  }

  private static Point convertPointToScreen(Component invoker, Point componentPos) {
    final Point ret = new Point(componentPos);
    SwingUtilities.convertPointToScreen(ret, invoker);
    return ret;
  }

  private static Rectangle getUsableScreenBounds(@Nullable Component invoker, Point screenPos) {
    GraphicsConfiguration gc = getPointGraphicsConfiguration(screenPos);
    if (gc == null && invoker != null)
      gc = invoker.getGraphicsConfiguration();
    return gc == null ? Utilities.getUsableScreenBounds() : Utilities.getUsableScreenBounds(gc);
  }

  private static @Nullable GraphicsConfiguration getPointGraphicsConfiguration(Point screenPos) {
    // See similar logic in JPopupMenu.getCurrentGraphicsConfiguration(Point)
    for (GraphicsDevice graphicsDevice :
        GraphicsEnvironment.getLocalGraphicsEnvironment().getScreenDevices())
    {
      if (graphicsDevice.getType() == GraphicsDevice.TYPE_RASTER_SCREEN) {
        final GraphicsConfiguration gc = graphicsDevice.getDefaultConfiguration();
        if (gc.getBounds().contains(screenPos))
          return gc;
      }
    }
    return null;
  }

  @Override
  public void cancel() {
    hide(true);
  }

  private void hide(boolean canceled) {
    CommonUIUtil.checkOnEDT();
    if (windowWithOwner == null)
      return;
    // Do this after for checking for the no-op above.
    if (canceled)
      firePopupMenuEvent(PopupEventType.CANCELED);
    final Window popup = windowWithOwner.popup;
    final Window owner = windowWithOwner.owner;
    windowWithOwner = null;
    /* The invisible event should always be fired at end, including when canceled. (As observed for
    JPopup.) */
    firePopupMenuEvent(PopupEventType.INVISIBLE);
    Toolkit.getDefaultToolkit().removeAWTEventListener(hideAWTListener);
    popup.removeWindowListener(popupClosedItselfListener);
    popup.setVisible(false);
    content.getParent().remove(content);
    popup.dispose();
    owner.removeWindowStateListener(hideWindowStateListener);
    owner.removeComponentListener(hideComponentListener);
  }

  @Override
  protected void startListening() {
    // No action required.
  }

  @Override
  protected void stopListening() {
    // No action required.
  }

  private class HideAWTListener implements AWTEventListener {
    @Override
    public void eventDispatched(AWTEvent evt) {
      if (evt instanceof MouseEvent) {
        MouseEvent mv = (MouseEvent) evt;
        if (mv.getID() == MouseEvent.MOUSE_PRESSED) {
          // NetBeans bug #118828 (source could be a java.awt.TrayIcon).
          if (!(evt.getSource() instanceof Component)) {
            cancel();
            return;
          }

          Component comp = (Component) evt.getSource();
          // Cancel if clicked outside the popup.
          if (SwingUtilities.getAncestorNamed(POPUP_NAME, comp) == null)
            cancel();
        }
      }
    }
  }

  private class HideWindowStateListener implements WindowStateListener {
    @Override
    public void windowStateChanged(WindowEvent evt) {
      int oldState = evt.getOldState();
      int newState = evt.getNewState();
      if (((oldState & Frame.ICONIFIED) == 0)
          && ((newState & Frame.ICONIFIED) == Frame.ICONIFIED))
      {
        /* A previous comment suggested we could show the popup again if the window is restored.
        I think it's fine, and indeed more sensible, to just close it and forget it. */
        cancel();
      }
    }
  }

  private class HideComponentListener extends ComponentAdapter {
    @Override
    public void componentResized(ComponentEvent evt) {
      cancel();
    }

    @Override
    public void componentMoved(ComponentEvent evt) {
      cancel();
    }
  }

  private class CloseAction extends AbstractAction {
    private final boolean canceled;

    public CloseAction(boolean canceled) {
      this.canceled = canceled;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
      hide(canceled);
    }
  }
}
