/* Eirik's Miscellaneous Modified NetBeans Libraries.
   Written or modified (as noted) by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. I elect to include
   this software in this distribution under the GPL Version 2 (with Classpath Exception) license. */

/* This file is based in part on:
 *   org.netbeans.modules.editor.NbEditorToolBar
 *   org.openide.awt.Actions
 * The complete license headers for the original files can be found below (identical for both
 * files).
 */
/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2010 Oracle and/or its affiliates. All rights reserved.
 *
 * Oracle and Java are registered trademarks of Oracle and/or its affiliates.
 * Other names may be trademarks of their respective owners.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2007 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package com.sieuferd.upstream.toolbar;

import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;
import com.sieuferd.upstream.actions.ActionUtil;
import java.awt.Component;
import java.awt.Insets;
import java.awt.Toolkit;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.PreferenceChangeEvent;
import java.util.prefs.PreferenceChangeListener;
import java.util.prefs.Preferences;
import javax.accessibility.Accessible;
import javax.annotation.Nullable;
import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import javax.swing.UIManager;
import javax.swing.border.LineBorder;
import javax.swing.plaf.ToolBarUI;
import org.netbeans.api.editor.mimelookup.MimeLookup;
import org.netbeans.api.editor.mimelookup.MimePath;
import org.openide.awt.AcceleratorBinding;
import org.openide.awt.Actions;
import org.openide.awt.ToolbarWithOverflow;
import org.openide.filesystems.FileChangeAdapter;
import org.openide.filesystems.FileChangeListener;
import org.openide.filesystems.FileEvent;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.ContextAwareAction;
import org.openide.util.Lookup;
import org.openide.util.Mutex;
import org.openide.util.NbBundle;
import org.openide.util.WeakListeners;
import org.openide.util.actions.Presenter;

// TODO[pd]: See if I need to reintroduce the bugfix for bug #69642.
/**
 * A toolbar automatically populated with actions from a folder in the system filesystem, with a
 * custom lookup used for ContextAwareAction actions. The toolbar is intended to look and feel as
 * much as possible like the toolbar used in NetBeans' text editor (NbEditorToolBar), but does not
 * assume the presence of text-related concepts such as JTextComponent or EditorKit. For instance,
 * action buttons are assigned tooltips with keyboard shortcuts which are updated if the keyboard
 * shortcuts change, without relying on the Editor Settings API.
 *
 * Based on org.netbeans.modules.editor.NbEditorToolBar (author Miloslav Metelka).
 */
public class LocalContextToolbar extends ToolbarWithOverflow {
  private static final Logger LOG = Logger.getLogger(LocalContextToolbar.class.getName());

  /* This property is from org.netbeans.api.editor.settings.SimpleValueNames.TOOLBAR_VISIBLE_PROP,
  but like in org.netbeans.core.multiview.MultiViewPeer, we avoid depending on the editor settings
  API here. */
  private static final String TOOLBAR_VISIBLE_PROP = "toolbarVisible";
  private static final Insets BUTTON_INSETS = new Insets(2, 1, 0, 1);
  private FileChangeListener toolbarPathListener;
  /** Strong references to weak property change listeners used to keep action button tooltips
  updated. */
  private final Set<PropertyChangeListener> tooltipListenerRefs = Sets.newLinkedHashSet();

  private final Reference<JComponent> componentRef;
  private boolean presentersAdded;
  private boolean addListener = true;

  @Nullable // Handle the null case since MultiViewPeer does so.
  private final Preferences preferences;
  private final PreferenceChangeListener prefsTracker = (PreferenceChangeEvent evt) -> {
    String settingName = evt == null ? null : evt.getKey();
    if (settingName == null || TOOLBAR_VISIBLE_PROP.equals(settingName))
      refreshToolbarButtons();
  };

  private final Lookup contextLookup;
  private final String toolbarPath;

  private static boolean isXPTheme() {
    return Boolean.TRUE.equals(
        Toolkit.getDefaultToolkit().getDesktopProperty("win.xpstyle.themeActive"));
  }

  /**
   * Create a new LocalContextToolbar. For each object in the system filesystem folder toolbarPath,
   * the following types of objects will be recognized in order and used to produce a Component to
   * add to the toolbar:
   *
   * <ol>
   *   <li>null or JSeparator: Adds a separator.</li>
   *   <li>String: Use an Action from the component's ActionMap with the specified key.</li>
   *   <li>ContextAwareAction: Use the action returned by
   *       createContextAwareInstance(contextLookup).</li>
   *   <li>Presenter.Toolbar: Use the Component returned by getToolbarPresenter().</li>
   *   <li>Component: Use the Component.</li>
   *   <li>Action: Create and use a JButton bound to the Action, with some tweaks including making
   *       sourceComponent the source of each ActionEvent.</li>
   *   <li>AbstractButton: Apply some appropriate styling and use the AbstractButton.</li>
   * </ol>
   *
   * @param sourceComponent component to be used as a source for action events, and to get an
   *        ActionMap from when resolving String values in the toolbarPath folder
   * @param contextLookup the lookup to use for context-aware actions
   * @param toolbarPath path to a folder in the system filesystem holding actions to be used to
   *        populate the toolbar
   */
  public LocalContextToolbar(JComponent sourceComponent, Lookup contextLookup, String toolbarPath) {
    Preconditions.checkNotNull(sourceComponent);
    Preconditions.checkNotNull(contextLookup);
    this.componentRef = new WeakReference<>(sourceComponent);
    this.contextLookup = contextLookup;
    this.toolbarPath = toolbarPath;
    // Finish member initialization before adding any listeners.
    // Use MimePath.EMPTY like in MultiViewPeer.
    preferences = MimeLookup.getLookup(MimePath.EMPTY).lookup(Preferences.class);

    /* Based on similar code in org.netbeans.core.multiview.TabsComponent and others. Only deal with
    properties that are logically internal to the toolbar rather than those that relate to how the
    toolbar blends in with its adjacent components. */
    setFocusable(true);
    if ("Windows".equals(UIManager.getLookAndFeel().getID()) && !isXPTheme())
      setRollover(true);

    installToolbarPathListener();
    if (preferences != null) {
      preferences.addPreferenceChangeListener(
          WeakListeners.create(PreferenceChangeListener.class, prefsTracker, preferences));
    }
    refreshToolbarButtons();
  }

  @Override
  public void addNotify() {
    super.addNotify();
    // In GTK L&F the border of the toolbar looks raised. It does not help to set null border
    // nor setting EmptyBorder helps. Current solution is to set one-pixel LineBorder
    // which overwrites the "raising line".
    if ("GTK".equals(UIManager.getLookAndFeel().getID()))
      setBorder(new LineBorder(getBackground(), 1));
  }

  private void installToolbarPathListener() {
    toolbarPathListener = new FileChangeAdapter() {
      public @Override
      void fileChanged(FileEvent fe) {
        Mutex.EVENT.writeAccess(() -> {
          if (isToolbarVisible()) {
            checkPresentersRemoved();
            checkPresentersAdded();
          }
        });
      }
    };

    FileObject toolbarFolder = FileUtil.getConfigFile(toolbarPath);
    if (toolbarFolder != null) {
      toolbarFolder.addFileChangeListener(
          FileUtil.weakFileChangeListener(toolbarPathListener, toolbarFolder));
    }
  }

  @Override
  public String getUIClassID() {
    //For GTK and Aqua look and feels, we provide a custom toolbar UI -
    //but we cannot override this globally or it will cause problems for
    //the form editor & other things
    if (UIManager.get("Nb.Toolbar.ui") != null) { //NOI18N
      return "Nb.Toolbar.ui"; //NOI18N
    } else {
      return super.getUIClassID();
    }
  }

  @Override
  public String getName() {
    //Required for Aqua L&F toolbar UI
    return "editorToolbar"; // NOI18N
  }

  @Override
  public void setUI(ToolBarUI ui) {
    addListener = false;
    super.setUI(ui);
    addListener = true;
  }

  @Override
  public final synchronized void addMouseListener(MouseListener l) {
    if (addListener) {
      super.addMouseListener(l);
    }
  }

  @Override
  public synchronized void addMouseMotionListener(MouseMotionListener l) {
    if (addListener) {
      super.addMouseMotionListener(l);
    }
  }

  private boolean isToolbarVisible() {
    return preferences == null || preferences.getBoolean(TOOLBAR_VISIBLE_PROP, true);
  }

  private void refreshToolbarButtons() {
    final boolean visible = isToolbarVisible();
    Mutex.EVENT.writeAccess(() -> {
      if (visible) {
        checkPresentersAdded();
      } else {
        checkPresentersRemoved();
      }
      setVisible(visible);
    });
  }

  private void checkPresentersAdded() {
    if (!presentersAdded) {
      presentersAdded = true;
      addPresenters();
    }
  }

  private void checkPresentersRemoved() {
    presentersAdded = false;
    tooltipListenerRefs.clear();
    removeAll();
  }

  /**
   * Add the presenters (usually buttons) for the contents of the toolbar.
   */
  private void addPresenters() {
    JComponent c = componentRef.get();
    if (c == null)
      return;
    for (final Entry<FileObject,Object> entry :
        ActionUtil.getRegisteredObjects(toolbarPath).entrySet())
    {
      final FileObject fileObject = entry.getKey();
      final Object item = entry.getValue();

      if (item == null) {
        addSeparator();
        continue;
      }

      final Action wrappedAction;
      final Component customPresenter;

      {
        final Action action;
        if (item instanceof String) {
          action = c.getActionMap().get(item);
          if (action != null) {
            LOG.log(Level.FINE, "Item {0} converted to a component action {1}",
                new Object[]{item, s2s(action)}); //NOI18N
          }
        } else if (item instanceof ContextAwareAction) {
          action = ((ContextAwareAction) item).createContextAwareInstance(contextLookup);
          LOG.log(Level.FINE, "Item {0} converted to a context-aware Action {1}",
              new Object[]{s2s(item), s2s(action)}); //NOI18N
        } else if (item instanceof Action) {
          action = (Action) item;
        } else {
          action = null;
        }
        if (action == null) {
          LOG.log(Level.FINE, "Failed to resolve action for item {0}", new Object[]{item});
          continue;
        }
        // TODO[cc]: Consider filing a bug for UndoRedoAction instead.
        /* Workaround for the fact that org.openide.actions.UndoRedoAction will not initialize its
        icon until isEnabled() has been called. */
        action.isEnabled();

        if (action instanceof Presenter.Toolbar) {
          customPresenter = ((Presenter.Toolbar) action).getToolbarPresenter();
          LOG.log(Level.FINE, "Item {0} converted to a Presenter.Toolbar {1}",
              new Object[]{s2s(item), s2s(customPresenter)}); //NOI18N
        } else if (item instanceof Component) {
          customPresenter = (Component) item;
        } else {
          customPresenter = null;
        }

        /* TODO[pd]: If needed, consider doing something like this:

                 if (customPresenter instanceof JComponent)
                   ((JComponent) presenter).putClientProperty(JComponent.class, c);
        */

        wrappedAction = new PropertyNotifyWrapperAction(new WrapperAction(componentRef, action));
        /* TODO[pd]: File a bug suggesting that AcceleratorBinding.setAccelerator() and
                     Actions.connect() could be made to support multi-key bindings in tooltips, by
                     setting the AbstractEditorAction.MULTI_ACCELERATOR_LIST_KEY or an equivalent
                     property on the action. */
        AcceleratorBinding.setAccelerator(wrappedAction, fileObject);
      }
      final Component presenter;
      if (customPresenter == null) {
        presenter = add(wrappedAction);
      } else {
        presenter = customPresenter;
        add(presenter);
        LOG.log(Level.FINE, "Adding custom presenter component {0}", s2s(presenter));
      }

      if (presenter instanceof AbstractButton) {
        connectTooltip((AbstractButton) presenter, wrappedAction);
        processButton((AbstractButton) presenter);
        LOG.log(Level.FINE, "Special treatment for button {0}", s2s(presenter)); //NOI18N
      }
    }
  }

  // Based on org.openide.awt.Actions.findKey(Action)
  private static String findKey(Action action) {
    if (action == null)
      return null;
    KeyStroke accelerator = (KeyStroke) action.getValue(Action.ACCELERATOR_KEY);
    if (accelerator == null)
      return null;
    return ToolbarUtil.getDisplayTextForAccelerator(accelerator);
  }

  // Based on org.openide.awt.Actions.ButtonBridge.updateState
  private static void updateTooltip(AbstractButton button, Action action, String changedProperty) {
    if ((changedProperty == null)
        || changedProperty.equals(Action.ACCELERATOR_KEY)
        || (changedProperty.equals(Action.NAME) &&
            action.getValue(Action.SHORT_DESCRIPTION) == null)
        || changedProperty.equals(Action.SHORT_DESCRIPTION))
    {
      String tip = findKey(action);
      String toolTip = (String) action.getValue(Action.SHORT_DESCRIPTION);
      if (toolTip == null) {
        toolTip = (String) action.getValue(Action.NAME);
        toolTip = (toolTip == null) ? "" : Actions.cutAmpersand(toolTip);
      }
      if ((tip == null) || tip.isEmpty()) { // NOI18N
        button.setToolTipText(toolTip);
      } else {
        button.setToolTipText(NbBundle.getMessage(Actions.class, "FMT_ButtonHint", toolTip, tip));
      }
    }
    if (button instanceof Accessible &&
        ((changedProperty == null) || changedProperty.equals(Action.NAME)))
    {
      button.getAccessibleContext().setAccessibleName((String) action.getValue(Action.NAME));
    }
  }

  private void connectTooltip(final AbstractButton button, final Action action) {
    /* The most comprehensive way to connect a button to an action would be to do
    Actions.connect(button, action), but we only want to set the tooltip, since the button may be
    a custom presenter that shouldn't be meddled with more than necessary. */
    PropertyChangeListener pcl = (PropertyChangeEvent evt) -> {
      updateTooltip(button, action, evt.getPropertyName());
    };
    action.addPropertyChangeListener(WeakListeners.propertyChange(pcl, action));
    tooltipListenerRefs.add(pcl);
    updateTooltip(button, action, null);
  }

  private void processButton(AbstractButton button) {
    button.setMargin(BUTTON_INSETS);
    ToolbarUtil.setBorderOnHover(button);
    //fix of issue #69642. Focus shouldn't stay in toolbar
    button.setFocusable(false);
  }

  private static Object s2s(Object o) {
    if (LOG.isLoggable(Level.FINE)) {
      return o == null ? "null" : o.getClass().getName() + "@" +
          Integer.toHexString(System.identityHashCode(o)); //NOI18N
    } else {
      return o;
    }
  }
}
