/* Eirik's Miscellaneous Modified NetBeans Libraries.
   Written or modified (as noted) by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. */

package com.sieuferd.upstream.toolbar;

import com.google.common.collect.Sets;
import com.sieuferd.uiutil.api.CommonUIUtil;
import java.awt.event.ActionEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Set;
import javax.swing.Action;
import org.openide.util.Mutex;
import org.openide.util.WeakListeners;

// TODO[pd]: Get rid of this once bug #234245 is fixed.
/**
 * Action wrapper that makes sure to notify property change listeners about property changes even if
 * the delegate forgets to. Workaround for
 * <a href="https://netbeans.org/bugzilla/show_bug.cgi?id=234245">bug #234245</a>.
 */
class PropertyNotifyWrapperAction implements Action {
  private final Action delegate;
  private final PropertyChangeSupport support;
  private final Set<String> notifiedProperties = Sets.newLinkedHashSet();
  private final PropertyChangeListener listener = (PropertyChangeEvent evt) -> {
    CommonUIUtil.checkOnEDT();
    notifiedProperties.add(evt.getPropertyName());
  };

  PropertyNotifyWrapperAction(Action delegate) {
    CommonUIUtil.checkOnEDT();
    this.delegate = delegate;
    this.support = new PropertyChangeSupport(this);
    delegate.addPropertyChangeListener(WeakListeners.propertyChange(listener, delegate));
  }

  @Override
  public Object getValue(String key) {
    CommonUIUtil.checkOnEDT();
    return delegate.getValue(key);
  }

  @Override
  public void putValue(String key, Object value) {
    CommonUIUtil.checkOnEDT();
    Object oldValue = delegate.getValue(key);
    notifiedProperties.clear();
    try {
      delegate.putValue(key, value);
      if (!notifiedProperties.contains(null) && !notifiedProperties.contains(key))
        support.firePropertyChange(key, oldValue, value);
    } finally {
      notifiedProperties.clear();
    }
  }

  @Override
  public void setEnabled(boolean b) {
    CommonUIUtil.checkOnEDT();
    delegate.setEnabled(b);
  }

  @Override
  public boolean isEnabled() {
    CommonUIUtil.checkOnEDT();
    return delegate.isEnabled();
  }

  @Override
  public void addPropertyChangeListener(PropertyChangeListener listener) {
    CommonUIUtil.checkOnEDT();
    delegate.addPropertyChangeListener(listener);
    support.addPropertyChangeListener(listener);
  }

  @Override
  public void removePropertyChangeListener(final PropertyChangeListener listener) {
    // Support weak listener removal, which may happen from a thread other than the EDT.
    Mutex.EVENT.readAccess(() -> {
      support.removePropertyChangeListener(listener);
      delegate.removePropertyChangeListener(listener);
    });
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    delegate.actionPerformed(e);
  }

  @Override
  public String toString() {
    return delegate.toString();
  }
}
