/* Eirik's Miscellaneous Swing Utilities.
   Written by Eirik Bakke 2013-2016, ebakke@alum.mit.edu. */

package com.sieuferd.uiutil.api;

import com.google.common.base.Preconditions;
import java.awt.Component;
import java.awt.Point;
import javax.annotation.Nonnull;
import javax.swing.JPopupMenu;
import javax.swing.event.EventListenerList;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;

public abstract class PopupLike {
  private final EventListenerList listenerList = new EventListenerList();
  protected enum PopupEventType { VISIBLE, INVISIBLE, CANCELED };

  public final void addPopupMenuListener(PopupMenuListener l) {
    if (listenerList.getListenerCount() == 0)
      startListening();
    listenerList.add(PopupMenuListener.class,l);
  }

  public final void removePopupMenuListener(PopupMenuListener l) {
    listenerList.remove(PopupMenuListener.class,l);
    if (listenerList.getListenerCount() == 0)
      stopListening();
  }

  protected void firePopupMenuEvent(PopupEventType type) {
    // Follow the same pattern as in JPopupMenu.firePopupXXX.
    final Object[] listeners = listenerList.getListenerList();
    PopupMenuEvent e = null;
    for (int i = listeners.length - 2; i >= 0; i -= 2) {
      if (listeners[i] == PopupMenuListener.class) {
        if (e == null)
          e = new PopupMenuEvent(this);
        final PopupMenuListener listener = (PopupMenuListener) listeners[i + 1];
        if (type == PopupEventType.VISIBLE) {
          listener.popupMenuWillBecomeVisible(e);
        } else if (type == PopupEventType.INVISIBLE) {
          listener.popupMenuWillBecomeInvisible(e);
        } else if (type == PopupEventType.CANCELED) {
          listener.popupMenuCanceled(e);
        }
      }
    }
  }

  protected abstract void startListening();
  protected abstract void stopListening();
  /* @param pos the position to show the popup at, in the {@code invoker}'s coordinate space */
  public abstract void show(@Nonnull Component invoker, Point pos);
  /** Cancel and hide the popup. */
  public abstract void cancel();

  public static PopupLike fromJPopupMenu(JPopupMenu popupMenu) {
    return new JPopupLike(popupMenu);
  }

  private static final class JPopupLike extends PopupLike {
    private final JPopupMenu delegate;
    private final PopupMenuListener listener = new PopupMenuListener() {
      @Override
      public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
        firePopupMenuEvent(PopupEventType.VISIBLE);
      }

      @Override
      public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
        firePopupMenuEvent(PopupEventType.INVISIBLE);
      }

      @Override
      public void popupMenuCanceled(PopupMenuEvent e) {
        firePopupMenuEvent(PopupEventType.CANCELED);
      }
    };

    public JPopupLike(JPopupMenu menu) {
      Preconditions.checkNotNull(menu);
      this.delegate = menu;
    }

    @Override
    protected void startListening() {
      delegate.addPopupMenuListener(listener);
    }

    @Override
    protected void stopListening() {
      delegate.removePopupMenuListener(listener);
    }

    @Override
    public void show(Component invoker, Point pos) {
      Preconditions.checkNotNull(invoker);
      delegate.show(invoker, pos.x, pos.y);
    }

    @Override
    public void cancel() {
      delegate.setVisible(false);
    }
  }
}
