// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.builders;

import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.impl.TextCell;
import com.google.common.base.Preconditions;
import com.sieuferd.cell.styles.TextFitPolicy;
import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.cell.styles.LineBreakPolicy;
import com.sieuferd.cell.styles.Text;
import com.sieuferd.cell.styles.Text.SimpleText;
import com.sieuferd.cell.styles.TextMeasurement;
import com.sieuferd.cell.styles.TextMeasurer;
import com.sieuferd.cell.styles.TextStyle;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewPropertySource;
import java.awt.Color;
import java.awt.Font;
import javax.annotation.Nullable;

public class TextCellBuilder extends CellBuilder<TextCellBuilder> {
  private static final Text TEXT_X = new SimpleText("X");
  private final Font font;
  private final Color color;
  private final TextMeasurer textMeasurer;
  private HorizontalAlignment horizontalAlignment;
  private @Nullable Text text = null;
  private boolean vertical = false;
  private boolean justified = false;
  private TextFitPolicy textFitPolicy = TextFitPolicy.PARAGRAPH;
  private LineBreakPolicy lineBreakPolicy = LineBreakPolicy.DEFAULT_PARAGRAPH;

  // TODO[cc]: Don't take all these arguments.
  public TextCellBuilder(RenderContext renderContext, TextMeasurer textMeasurer,
      ViewPropertySource viewPropertySource, Font font, Color color,
      HorizontalAlignment horizontalAlignment, Pixels width)
  {
    super(renderContext, viewPropertySource);
    Preconditions.checkNotNull(font);
    Preconditions.checkNotNull(color);
    Preconditions.checkNotNull(horizontalAlignment);
    Preconditions.checkNotNull(textMeasurer);
    this.textMeasurer = textMeasurer;
    this.font = font;
    this.color = color;
    this.horizontalAlignment = horizontalAlignment;
    minimumDimensions(new Coordinates(width, Pixels.ZERO));
  }

  public TextCellBuilder textFitPolicy(TextFitPolicy textFitPolicy) {
    Preconditions.checkNotNull(textFitPolicy);
    this.textFitPolicy = textFitPolicy;
    return this;
  }

  public TextCellBuilder lineBreakPolicy(LineBreakPolicy lineBreakPolicy) {
    Preconditions.checkNotNull(lineBreakPolicy);
    this.lineBreakPolicy = lineBreakPolicy;
    return this;
  }

  public TextCellBuilder vertical(boolean vertical) {
    this.vertical = vertical;
    return this;
  }

  public TextCellBuilder vertical(ViewProperty<Boolean> vertical) {
    this.vertical = viewPropertySource.get(vertical);
    return this;
  }

  public TextCellBuilder content(Text text) {
    Preconditions.checkNotNull(text);
    this.text = text;
    return this;
  }

  public TextCellBuilder content(String str) {
    Preconditions.checkNotNull(str);
    this.text = new SimpleText(str);
    return this;
  }

  public TextCellBuilder content(ViewProperty<String> str) {
    Preconditions.checkNotNull(str);
    this.text = new SimpleText(viewPropertySource.get(str));
    return this;
  }

  public TextCellBuilder justified(ViewProperty<Boolean> justified) {
    this.justified = viewPropertySource.get(justified);
    return this;
  }

  public TextCellBuilder justified(boolean justified) {
    this.justified = justified;
    return this;
  }

  public TextCellBuilder horizontalAlignment(HorizontalAlignment horizontalAlignment) {
    Preconditions.checkNotNull(horizontalAlignment);
    this.horizontalAlignment = horizontalAlignment;
    return this;
  }

  @Override
  public Cell build() {
    if (text == null)
      throw new IllegalStateException("Content not set");
    final Pixels width = minimumDimensions.getX();
    final TextStyle textStyle = new TextStyle(font, lineBreakPolicy);
    final TextMeasurement measurement;
    /* TODO[pd]: Avoid the "\n" condition, instead fitting vertically as well. Or keep the
                 condition, but still fit taller-than-usual lines vertically. */
    if (textFitPolicy == TextFitPolicy.PARAGRAPH ||
        (textFitPolicy == TextFitPolicy.FIT_LINE && text.getString().contains("\n")))
    {
      /* TODO[pd]: This does not properly measure fitted text. Fix this. I should just have the
                   TextCell store its ParagraphLayout instead. */
      measurement = textMeasurer.measure(
          textStyle, text,
          textFitPolicy == TextFitPolicy.FIT_LINE ? null : minimumDimensions.getX(),
          false, vertical);
    } else {
      // Use constant parameters to aid memoization.
      measurement = textMeasurer.measure(textStyle, TEXT_X, null, true, vertical);
    }
    final Pixels measuredHeight = measurement.getDimensions().getY();
    final Coordinates measuredDimensions = new Coordinates(width, measuredHeight);
    return new TextCell(renderContext, measuredDimensions.max(minimumDimensions),
        measurement.getFirstBaselineOffset(), textStyle, color, text, vertical, textFitPolicy,
        horizontalAlignment, justified);
  }

  @Override
  public TextCellBuilder getThis() {
    return this;
  }
}
