// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.impl;

import com.google.common.base.Function;
import com.google.common.base.Functions;
import com.sieuferd.cell.support.BiasedLocation;
import com.sieuferd.cell.support.CellParam;
import com.sieuferd.cell.support.Direction;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.styles.BorderStyle;
import com.sieuferd.cell.support.Canvas;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Location;
import com.sieuferd.cell.support.CursorStyle;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.CellPaintParameters;
import com.sieuferd.cell.support.CellPaintParameters.Layer;
import com.sieuferd.cell.support.CellRectangles;
import com.sieuferd.cell.support.HitParameters;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.Views;
import com.sieuferd.uiutil.UIUtil;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Rectangle2D;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.Map;
import java.util.Map.Entry;
import javax.annotation.Nullable;

public final class BorderCell extends Cell {
  private static final boolean TEST_DROPDOWN_BUTTONS = false;
  private static final boolean EXTEND_DROPDOWN_BUTTONS_VERTICALLY = false;
  private static final boolean DEBUG_FROZEN_VALUES = false;
  // I decided I'd rather want frozen values to stay baseline-aligned.
  private static final boolean TOP_ALIGN_FROZEN_VALUES = false;
  private final Cell childCell;
  private final Coordinates dimensions;
  private final Color backgroundColor;
  private final BorderStyle borderStyle;
  private final BiasedLocation biasedLocation;
  private final boolean justified;

  /** @param biasedLocation must have getCursorBias() == CursorBias.NONE if containsSelectable()
             would be true */
  public BorderCell(RenderContext renderContext, Coordinates minimumStackableDimensions,
      Cell childCell, BorderStyle borderStyle, Color backgroundColor, BiasedLocation biasedLocation,
      boolean justified)
  {
    super(renderContext);
    Preconditions.checkNotNull(minimumStackableDimensions);
    Preconditions.checkNotNull(childCell);
    Preconditions.checkNotNull(borderStyle);
    Preconditions.checkNotNull(backgroundColor);
    Preconditions.checkNotNull(biasedLocation);
    final boolean opaque = backgroundColor.getAlpha() == 255;
    /* I considered requiring all cells with non-zero borders to have an opaque background color
    defined as well, but decided against it for now. In the future I might reconsider this. */
    if (biasedLocation.getCursorLocation() != null && !opaque)
      throw new IllegalArgumentException("Cursor-selectable cell must be opaque");
    this.borderStyle = borderStyle;
    this.backgroundColor = backgroundColor;
    this.childCell = childCell;
    this.dimensions = childCell.getDimensions()
        .plus(borderStyle.topLeftBorderDimensions(renderContext))
        .max(minimumStackableDimensions);

    if (biasedLocation.getCursorLocation() != null && childCell.containsBiasedLocation()) {
      throw new IllegalArgumentException("Overlapping location " +
          biasedLocation.getCursorLocation());
    }
    if (biasedLocation.getCursorBias() != Direction.NONE && childCell.containsBiasedLocation())
      throw new IllegalArgumentException("Selectable cell cannot have a local cursor bias");
    this.biasedLocation = biasedLocation;
    this.justified = justified;
  }

  @Override
  public Coordinates getDimensions() {
    return dimensions;
  }

  @Override
  public Coordinates getBottomRightOverprintLimit(RenderContext renderContext) {
    return borderStyle.bottomRightBorderDimensions(renderContext)
        .max(childCell.getBottomRightOverprintLimit(renderContext));
  }

  // TODO[pd]: Test and verify this.
  @Override
  public Pixels getPaginatedHeight(RenderContext renderContext, Pixels startOffset,
      Pixels preferredMax, Pixels paperHeight) {
    checkRenderContext(renderContext);
    checkRenderContext(renderContext);
    Preconditions.checkArgument(preferredMax.isGreaterThan(Pixels.ZERO));
    Preconditions.checkArgument(paperHeight.isGreaterThan(Pixels.ZERO));
    Preconditions.checkArgument(!startOffset.isNegative());
    final Pixels borderTopHeight = borderStyle.topLeftBorderDimensions(renderContext).getY();
    final Pixels borderOverlap = borderTopHeight.minus(startOffset).max(Pixels.ZERO);
    final Pixels childOffset = startOffset.minus(borderTopHeight).max(Pixels.ZERO);
    final Pixels childPreferredMax = preferredMax.minus(borderOverlap);
    final Pixels childHeight;
    if (childPreferredMax.isGreaterThan(Pixels.ZERO)) {
      childHeight = childCell.getPaginatedHeight(renderContext, childOffset,
          childPreferredMax, paperHeight);
    } else {
      childHeight = Pixels.ZERO;
    }
    return childHeight.plus(borderOverlap);
  }

  /** @param highlights may yield null for non-highlighted paths */
  private static @Nullable Entry<Path,Color> getEnclosingHighlight(
      Function<Path,Color> highlights, Path path)
  {
    final Color ret = highlights.apply(path);
    if (ret != null)
      return new SimpleImmutableEntry<>(path, ret);
    return path.isRoot() ? null : getEnclosingHighlight(highlights, path.getParent());
  }

  private static boolean correlationMatches(
      Location correlationLocation, Location highlightableLocation,
      Path enclosingHighlightPath)
  {
    Preconditions.checkNotNull(correlationLocation);
    Preconditions.checkNotNull(highlightableLocation);
    Preconditions.checkNotNull(enclosingHighlightPath);
    Preconditions.checkArgument(
        enclosingHighlightPath.isEqualToOrAncestorOf(highlightableLocation.getPath()));
    final RelationPath targetIdentifyingPath =
        correlationLocation.getIdentifyingPath() == null ? null :
        Views.findLowestCommonRelationAncestor(
        ImmutableSet.<RelationPath>of(correlationLocation.getIdentifyingPath(),
            enclosingHighlightPath.isRelation() ?
            enclosingHighlightPath.getRelation() : enclosingHighlightPath.getParent()));
    return correlationLocation.matchesAt(highlightableLocation, targetIdentifyingPath);
  }

  /**
   * @param highlights may yield null for non-highlighted paths
   * @param correlationLocation may be non-null to produce a stronger highlight for value locations
   *        in the same correlation context
   */
  private Color getHighlightColor(Function<Path,Color> highlights,
      final @Nullable Location correlationLocation)
  {
    final Location highlightableLocation = biasedLocation.getLocation();
    if (highlightableLocation == null)
      return UIUtil.COLOR_NONE;
    final Path highlightablePath = highlightableLocation.getPath();
    final @Nullable Color color = highlights.apply(highlightablePath);
    if (!highlightableLocation.isValue() && color != null &&
        biasedLocation.getCursorBias() == Direction.NONE)
    {
      // Label location for exact path.
      return UIUtil.opacity(color, CursorStyle.HARD_HIGHLIGHT_OPACITY);
    }
    final @Nullable Entry<Path,Color> enclosingHighlight =
        getEnclosingHighlight(highlights, highlightablePath);
    if (highlightableLocation.isValue() && correlationLocation != null &&
        enclosingHighlight != null &&
        correlationMatches(correlationLocation, highlightableLocation, enclosingHighlight.getKey()))
    {
      // Value location in the same correlation context as correlationLocation.
      return UIUtil.opacity(enclosingHighlight.getValue(), CursorStyle.MEDIUM_HIGHLIGHT_OPACITY);
    } else if (enclosingHighlight != null) {
      return UIUtil.opacity(enclosingHighlight.getValue(), correlationLocation == null
          ? CursorStyle.STANDARD_HIGHLIGHT_OPACITY : CursorStyle.SOFT_HIGHLIGHT_OPACITY);
    } else {
      return UIUtil.COLOR_NONE;
    }
  }

  /* TODO[pd]: Preprocess the paths from getPathHighlights and getSelectedPaths() so that the
               selection highlight color can be found in O(1) time. */
  private Color getHighlightColor(final CellParam param) {
    final Color pathHighlightColor;
    {
      final Map<Path,Color> pathHighlights = param.getPathHighlights().getHighlights();
      if (pathHighlights.isEmpty()) {
        // Small optimization for the common case.
        pathHighlightColor = UIUtil.COLOR_NONE;
      } else {
        final @Nullable Location cursorLocation = param.getSelection().getCursorLocation();
        final @Nullable Location editLocation = param.getPathHighlights().getEditLocation();
        final @Nullable Location correlationLocation;
        if (cursorLocation == null || editLocation == null) {
          correlationLocation = null;
        } else {
          final RelationPath editIdentifyingPath = editLocation.getIdentifyingPath();
          final RelationPath cursorIdentifyingPath = cursorLocation.getIdentifyingPath();
          /* TODO[pd]: Don't restrict to editIdentifyingPath if editing the formula from the formula
                       bar rather than an inplace editor. */
          if (editIdentifyingPath == null ||
              cursorIdentifyingPath != null &&
              editIdentifyingPath.isEqualToOrAncestorOf(cursorIdentifyingPath))
          {
            correlationLocation =
                cursorLocation.getAncestorWithIdentifyingPath(editIdentifyingPath);
          } else {
            correlationLocation = editLocation;
          }
        }
        pathHighlightColor =
            getHighlightColor(Functions.forMap(pathHighlights, null), correlationLocation);
      }
    }
    final Color selectionHighlightColor = getHighlightColor((Path input) -> {
      if (param.getSelection().getSelectedPaths().contains(input)) {
        final CellPaintParameters pp = param.getPaintParameters();
        /* The style is irrelevant if we are not painting, or if we are in reference selection
        mode (which does not use multiple selection). */
        return (pp != null && pp.getFocusMode() == FocusMode.FOCUSED) ?
            CursorStyle.SELECTION_COLOR : CursorStyle.SELECTION_COLOR_NOFOCUS;
      } else {
        return null;
      }
    }, null);
    return UIUtil.blendOver(selectionHighlightColor, pathHighlightColor);
  }

  private void paintBackground(
      Canvas canvas, Coordinates position, Coordinates dimensions, Color color)
  {
    Preconditions.checkNotNull(canvas);
    Preconditions.checkNotNull(position);
    Preconditions.checkNotNull(dimensions);
    Preconditions.checkNotNull(color);
    if (!(dimensions.getX().isZero() || dimensions.getY().isZero()) && color.getAlpha() > 0) {
      final Graphics2D g = canvas.getGraphics();
      g.setColor(canvas.getRenderContext().safeColor(color));
      g.fill(new Rectangle2D.Double(
          position.getX().getDouble(), position.getY().getDouble(),
          dimensions.getX().getDouble(), dimensions.getY().getDouble()));
    }
  }

  @Override
  public void traverse(
      CellParam param, Coordinates position, Coordinates justifiedDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      @Nullable HitParameters hitParameters)
  {
    final RenderContext rcx = param.getRenderContext();
    checkRenderContext(rcx);
    final Coordinates effectiveDimensions = dimensions.max(justifiedDimensions);
    final Coordinates topLeft = borderStyle.topLeftBorderDimensions(rcx);
    final Coordinates bottomRight = borderStyle.bottomRightBorderDimensions(rcx);

    if (hitParameters != null && param.getSelection().getCursorLocation() != null)
      throw new IllegalArgumentException("Can't do hit traversal with already-defined selection");
    final @Nullable HitParameters childHitParameters = calculateHit(param, hitParameters);
    /* Allow the cursor rectangle to be found either from the selection model or during the same
    traversal that detected the Location. */
    final @Nullable Location selectedCursorLocation =
          (hitParameters != null && childHitParameters == null)
        ? hitParameters.getResult().getCursorLocation()
        : param.getSelection().getCursorLocation();

    final @Nullable Location hereCursorLocation = biasedLocation.getCursorLocation();
    final boolean atCursor = hereCursorLocation != null && selectedCursorLocation != null &&
        selectedCursorLocation.equalsLocation(hereCursorLocation);

    final Coordinates frozenValueOffset;
    if (overlayMode && backgroundColor.getAlpha() == 255 &&
        headerPoint.getY().isGreaterThan(position.getY()) &&
        position.getY().plus(effectiveDimensions.getY()).isGreaterThan(headerPoint.getY()))
    {
      frozenValueOffset = new Coordinates(Pixels.ZERO, headerPoint.minus(position).getY()
          .min(effectiveDimensions.minus(childCell.getDimensions().plus(topLeft)).getY()));
    } else {
      frozenValueOffset = Coordinates.ZERO;
    }
    final boolean visible =
        atCursor || !overlayMode || !frozenValueOffset.isEqualTo(Coordinates.ZERO);
    final @Nullable CellPaintParameters pp = visible ? param.getPaintParameters() : null;
    final @Nullable Shape oldClip = (pp == null) ? null : pp.getCanvas().getGraphics().getClip();
    if (pp != null && overlayMode) {
      /* Clip both to avoid the cursor overlapping with frozen headers and to avoid frozen values
      running into the content above. */
      if (headerPoint.getY().isGreaterThan(position.getY())) {
        /* Don't be tempted to limit the clipping area to the cell's unjustified height; baseline
        alignment may cause content to shift further down. */
        pp.getCanvas().getGraphics().clip(new Rectangle2D.Double(
            UIUtil.NEGATIVE_INFINITY_CLIP_POSITION, headerPoint.getY().getDouble(),
            Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY));
      }
      if (headerPoint.getX().isGreaterThan(position.getX())) {
        pp.getCanvas().getGraphics().clip(new Rectangle2D.Double(
            headerPoint.getX().getDouble(), UIUtil.NEGATIVE_INFINITY_CLIP_POSITION,
            Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY));
      }
    }

    final Coordinates backgroundOff, backgroundDim;
    if (rcx.isAlphaPermitted()) {
      // Draw background underneath borders in case borders are semi-transparent.
      backgroundOff = Coordinates.ZERO;
      backgroundDim = effectiveDimensions.plus(bottomRight);
    } else {
      /* Otherwise, avoid overlapping the background with the borders. This avoids visual artifacts
      when the cell is printed via the Java Printing API to a PDF printer such as that shipped with
      Adobe Acrobat and then viewed in Acrobat or Acrobat Reader. */
      backgroundOff = topLeft;
      backgroundDim = effectiveDimensions.minus(topLeft);
    }

    /* TODO[pd]: If isAlphaPermitted and backgroundColor is null or translucent, pre-blend the
                 border color to be non-transparent based on a white background to avoid darker
                 borders where borders overlap. Or consider disallowing having a BorderCell without
                 a backgroundColor if a border is defined. */

    /* TODO[cc]: Avoid dealing with the cursor logic when there is not actually any cursor to
                 paint or find a rectangle for. */
    final Color highlightColor = getHighlightColor(param);

    final CursorStyle cursorStyle;
    if (pp != null && pp.getFocusMode() == FocusMode.FOCUSED) {
      cursorStyle = CursorStyle.CURSOR_BLUE;
    } else if (pp != null && pp.getFocusMode() == FocusMode.REFERENCE_SELECTION) {
      /* Use a translucent cursor that resembles the non-focused cursor when blended over a white
      background, but which blends well over path highlighting colors as well. */
      cursorStyle = CursorStyle.CURSOR_GRAY_TRANSLUCENT;
    } else {
      cursorStyle = CursorStyle.CURSOR_GRAY;
    }

    Coordinates cursorTopLeft = cursorStyle.getBorderStyle().topLeftBorderDimensions(rcx);
    Coordinates cursorBottomRight = cursorStyle.getBorderStyle().bottomRightBorderDimensions(rcx);
    Coordinates cursorExpandTopLeft = cursorTopLeft.minus(
        cursorStyle.getOverlapStyle().topLeftBorderDimensions(rcx));
    Coordinates cursorExpandBottomRight = cursorBottomRight.minus(
        cursorStyle.getOverlapStyle().bottomRightBorderDimensions(rcx));
    Coordinates cursorPos = position.plus(topLeft)
        .minus(cursorExpandTopLeft);
    Coordinates cursorDim = effectiveDimensions
        .minus(topLeft)
        .minus(cursorBottomRight)
        .plus(cursorExpandTopLeft)
        .plus(cursorExpandBottomRight);

    if (pp != null) {
      if (pp.includesLayer(Layer.BACKGROUND)) {
        final Color useBackgroundColor = DEBUG_FROZEN_VALUES && overlayMode
            ? UIUtil.blendOver(new Color(255, 255, 0, 64), backgroundColor)
            : backgroundColor;
        paintBackground(pp.getCanvas(), position.plus(backgroundOff), backgroundDim,
            UIUtil.blendOver(highlightColor, useBackgroundColor));
      } else if (pp.includesLayer(Layer.CURSOR)) {
        if (atCursor)
          cursorStyle.getBorderStyle().paintBorder(pp.getCanvas(), cursorPos, cursorDim);
      }
    }
    final Coordinates childPosition = position.plus(topLeft).plus(frozenValueOffset);
    final Coordinates childJustifiedDimensions =
          (justifiedDimensions.isEqualTo(Coordinates.ZERO) && !justified)
        ? Coordinates.ZERO
        : effectiveDimensions.minus(topLeft).minus(frozenValueOffset);
    final Pixels childBaselineOffset = baselineOffset == null ? null :
        baselineOffset.minus(topLeft.getY()).minus(
        TOP_ALIGN_FROZEN_VALUES ? frozenValueOffset.getY() : Pixels.ZERO).max(Pixels.ZERO);
    childCell.traverse(param, childPosition, childJustifiedDimensions,
        childBaselineOffset, headerPoint, overlayMode && !visible,
        childHitParameters);

    if (pp != null) {
      if (pp.includesLayer(Layer.BACKGROUND)) {
        borderStyle.paintBorder(pp.getCanvas(), position, effectiveDimensions);
      } else if (pp.includesLayer(Layer.TOP)) {
        if (TEST_DROPDOWN_BUTTONS && hereCursorLocation != null && !hereCursorLocation.isValue()) {
          final int BORDER = UIUtil.DROPDOWN_BUTTON_BORDER_WIDTH_PIXELS;
          // TODO[pd]: Consider scaling this up with the zoom level.
          final Pixels MAXIMUM_WIDTH = Pixels.create(16);
          // TODO[pd]: Figure out how to avoid extreme justification for relation labels.
          /* Use a similar calculation as in TextCell, to avoid placing the button way over to the
          right in the last column of a justified table. Allow the extra space to be used to avoid
          overlapping the child content and the button, however. */
          final Coordinates textChildDimensions = getDimensions()
              .max(new Coordinates(
                  justifiedDimensions.getX().min(getDimensions().getX().plus(MAXIMUM_WIDTH)),
                  justifiedDimensions.getY()));
          final PixelRectangle availableRect = new PixelRectangle(
              position.plus(topLeft).minus(new Coordinates(BORDER, BORDER)),
              textChildDimensions.minus(topLeft).plus(new Coordinates(BORDER * 2, BORDER * 2)));
          final Coordinates buttonDim;
          if (EXTEND_DROPDOWN_BUTTONS_VERTICALLY) {
            buttonDim = new Coordinates(
                availableRect.getWidth().divideBy(rcx, 3.0).min(MAXIMUM_WIDTH),
                availableRect.getHeight());
          } else {
            final Pixels size =
                availableRect.getWidth().min(availableRect.getHeight()).min(MAXIMUM_WIDTH);
            buttonDim = new Coordinates(size, size);
          }
          final Coordinates buttonPos =
              availableRect.getPosition().plus(availableRect.getDimensions().minus(buttonDim));
          UIUtil.drawDropdownButton(pp.getCanvas().getGraphics(),
              new PixelRectangle(buttonPos, buttonDim).getRectangle2D());
        }
      }
    }

    final @Nullable CellRectangles cellRectangles = param.getCellRectangles();
    if (cellRectangles != null) {
      if (atCursor) {
        cellRectangles.setCursorRectangle(
            new PixelRectangle(position.plus(topLeft), effectiveDimensions.minus(topLeft)));
      }
      if (atCursor || highlightColor.getAlpha() > 0) {
        cellRectangles.addRepaintRectangle(new PixelRectangle(cursorPos,
            cursorDim.plus(cursorStyle.getBorderStyle().bottomRightBorderDimensions(rcx))));
      }
    }

    if (pp != null)
      pp.getCanvas().getGraphics().setClip(oldClip);
  }

  @Override
  public final boolean containsBiasedLocation() {
    return (!biasedLocation.isNone()) || childCell.containsBiasedLocation();
  }

  /**
   * @return hit parameters to be passed to child cell
   */
  private @Nullable HitParameters calculateHit(CellParam param, @Nullable HitParameters hitParam) {
    Preconditions.checkNotNull(param);
    if (hitParam == null)
      return null;
    final Location cursorLocation = biasedLocation.getCursorLocation();
    final @Nullable Location avoid = param.getAvoidHit();
    if (cursorLocation != null) {
      if (avoid == null || !cursorLocation.equalsLocation(avoid)) {
        hitParam.setResult(biasedLocation);
      } else {
        hitParam.setResult(new BiasedLocation(hitParam.getCursorBias(), null));
      }
      return null;
    } else {
      if (biasedLocation.getCursorBias() != Direction.NONE) {
        hitParam.setResult(biasedLocation);
        return null;
      } else {
        // Delegate to child cell.
        return hitParam;
      }
    }
  }

  @Override
  public String toString() {
    return "[" + childCell + "]";
  }

  @Override
  public Pixels getMinimumBaselineOffset(RenderContext renderContext) {
    return childCell.getMinimumBaselineOffset(renderContext)
        .plus(borderStyle.topLeftBorderDimensions(renderContext).getY());
  }
}
