// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.impl;

import com.sieuferd.cell.support.CellParam;
import com.sieuferd.cell.support.Direction;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.cell.support.Cell;
import java.awt.Graphics2D;
import java.awt.Shape;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.CellPaintParameters;
import com.sieuferd.cell.support.CellPaintParameters.Layer;
import com.sieuferd.cell.support.CellRectangles;
import com.sieuferd.cell.support.CellRectangles.HeaderRectangle;
import com.sieuferd.cell.support.HitParameters;
import com.sieuferd.uiutil.UIUtil;
import java.awt.geom.Rectangle2D;
import javax.annotation.Nullable;

public final class HeaderedCell extends Cell {
  private final Cell headerCell;
  private final Cell contentCell;
  private final ListCell concatenatedCell;
  private final Direction direction;

  public HeaderedCell(
      RenderContext renderContext, Cell headerCell, Cell contentCell, boolean vertical)
  {
    super(renderContext);
    Preconditions.checkNotNull(headerCell);
    Preconditions.checkNotNull(contentCell);
    this.headerCell = headerCell;
    this.contentCell = contentCell;
    this.concatenatedCell = new ListCell(renderContext, Coordinates.ZERO, vertical,
        ImmutableList.of(headerCell, contentCell));
    this.direction = vertical ? Direction.DOWN : Direction.RIGHT;
  }

  @Override
  public Coordinates getDimensions() {
    return concatenatedCell.getDimensions();
  }

  @Override
  public Coordinates getBottomRightOverprintLimit(RenderContext renderContext) {
    return concatenatedCell.getBottomRightOverprintLimit(renderContext);
  }

  @Override
  public boolean containsBiasedLocation() {
    return concatenatedCell.containsBiasedLocation();
  }

  @Override
  public void traverse(
      CellParam param, Coordinates position, Coordinates justifiedDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      @Nullable HitParameters hitParameters)
  {
    final @Nullable CellRectangles cellRectangles = param.getCellRectangles();
    final boolean cursorFoundBefore = cursorFound(cellRectangles);
    final Coordinates effectiveDimensions = getDimensions().max(justifiedDimensions);
    final Coordinates headerPosition;
    if (!overlayMode) {
      headerPosition = position;
    } else {
      final Pixels minimumHeaderPosition = direction.flowDir(position);
      final Pixels maximumHeaderPosition = minimumHeaderPosition
          .plus(direction.flowDir(effectiveDimensions)
          .minus(direction.flowDir(headerCell.getDimensions())));
      headerPosition = direction.newCoordinates(
          direction.flowDir(headerPoint)
            .max(minimumHeaderPosition).min(maximumHeaderPosition),
          direction.stillDir(position));
    }
    final Coordinates headerOffset = headerPosition.minus(position);
    final boolean overlayHeader = !headerOffset.isEqualTo(Coordinates.ZERO);
    final HitParameters headerHitParameters, contentHitParameters;
    if (!overlayHeader && hitParameters != null) {
      // Delegate the cursor bias handling to ListCell.
      concatenatedCell.traverse(CellParam.createHit(param.getClipBounds(), param.getRenderContext(),
          param.getAvoidHit()), position, justifiedDimensions, baselineOffset, headerPoint,
          overlayMode, hitParameters);
      if (cellRectangles == null && param.getPaintParameters() == null)
        return; // No further traversal necessary.
      headerHitParameters = null;
      contentHitParameters = null;
    } else {
      final boolean hitInHeader = direction.flowDir(headerPosition.plus(headerCell.getDimensions()))
          .isGreaterThan(Pixels.ZERO);
      contentHitParameters = hitInHeader ? null : hitParameters;
      headerHitParameters  = hitInHeader ? hitParameters : null;
    }

    final @Nullable CellPaintParameters pp = param.getPaintParameters();
    final @Nullable Shape oldClip = (pp == null) ? null : pp.getCanvas().getGraphics().getClip();
    final Coordinates contentOffset =
        direction.newCoordinates(headerCell.getDimensions(), Coordinates.ZERO);

    /* Unlike in a ListCell, we paint the header after the content (which may include nested headers
    in overlayMode). This makes the header's bottom border, which is normally overprinted by the
    content, visible. */
    {
      final Coordinates contentPosition = position.plus(contentOffset);
      final Coordinates contentJustifiedDimensions =
          justifiedDimensions.equals(Coordinates.ZERO) ? Coordinates.ZERO :
          direction.newCoordinates(
            justifiedDimensions.minus(headerCell.getDimensions()),
            justifiedDimensions);
      final Pixels contentBaselineOffset = direction.isHorizontal() ? baselineOffset : null;
      final Coordinates contentHeaderPoint = headerPoint.plus(contentOffset);
      contentCell.traverse(param, contentPosition, contentJustifiedDimensions,
          contentBaselineOffset, contentHeaderPoint, overlayMode, contentHitParameters);
    }

    if (overlayHeader && pp != null &&
        pp.includesLayer(Layer.BACKGROUND) &&
        pp.getCanvas().getRenderContext().getPaintUIShadows() &&
        pp.getCanvas().getRenderContext().isAlphaPermitted())
    {
      final Graphics2D g = pp.getCanvas().getGraphics();
      final Rectangle2D headerRect =
          new PixelRectangle(headerPosition, headerCell.getDimensions()).getRectangle2D();
      g.clip(new PixelRectangle(headerPosition, direction.newCoordinates(
          effectiveDimensions.minus(headerOffset), effectiveDimensions)).getRectangle2D());
      UIUtil.drawShadowBorder(pp.getCanvas().getGraphics(), headerRect, true);
      g.setClip(oldClip);
    }

    /* Clip nested frozen headers so that they appear to slide under their parent-level frozen
    headers rather than overlap them. Due to the current layering approach, this is mainly relevant
    when the cursor is in the nested frozen header. Specifically, clip away anything above (or to
    the left of, for a horizontally oriented HeaderedCell) the frozen header point. Only do this
    once the sliding actually starts, to avoid clipping the (potentially overprinted) cursor. */
    if (pp != null &&
        direction.flowDir(headerPoint.minus(headerPosition)).isGreaterThan(Pixels.ZERO))
    {
      Graphics2D g = pp.getCanvas().getGraphics();
      final double NI = UIUtil.NEGATIVE_INFINITY_CLIP_POSITION;
      g.clip(new Rectangle2D.Double(
          direction.getHorizontalBiasRight() != 0 ? headerPoint.getX().getDouble() : NI,
          direction.getVerticalBiasDown()    != 0 ? headerPoint.getY().getDouble() : NI,
          Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY));
    }

    {
      final boolean cursorFoundBeforeHeader = cursorFound(cellRectangles);
      final Coordinates headerJustifiedDimensions =
          direction.newCoordinates(Pixels.ZERO, direction.stillDir(justifiedDimensions));
      /* Apply the baseline offset even when the header is overlaid, to ensure that the header does
      not change appearance in the transition from non-overlaid to overlaid mode. */
      final Pixels headerBaselineOffset = baselineOffset;
      if (overlayHeader && pp != null && pp.includesLayer(Layer.FROZEN_HEADER_BACKGROUND)) {
        // Change the layer of the traversal to force the header to paint.
        headerCell.traverse(CellParam.createPaint(pp.getCanvas(), param.getSelection(),
              param.getPathHighlights(), pp.getFocusMode(), ImmutableSet.of(Layer.BACKGROUND)),
              headerPosition, headerJustifiedDimensions, headerBaselineOffset, headerPoint, false,
              headerHitParameters);
      } else if (overlayHeader && pp != null && pp.includesLayer(Layer.BACKGROUND)) {
        // No action.
        if (param.getCellRectangles() != null)
          throw new UnsupportedOperationException("Assumed it was safe not to traverse here");
      } else {
        headerCell.traverse(param, headerPosition, headerJustifiedDimensions, headerBaselineOffset,
              headerPoint, overlayMode && !overlayHeader, headerHitParameters);
      }
      if (!cursorFoundBeforeHeader && cursorFound(cellRectangles) && overlayHeader)
        cellRectangles.markCursorInOverlayHeader();
    }

    if (pp != null)
      pp.getCanvas().getGraphics().setClip(oldClip);

    // Add the header rectangle only if the cursor was in the header or content somewhere.
    if (!cursorFoundBefore && cursorFound(cellRectangles)) {
      /* Note that unlike ListCell, HeaderedCell should _not_ try to optimize away the traversal of
      header or content cells based on clip bounds, since we need to find all header rectangles for
      a given found cursor rectangle. */
      cellRectangles.addHeaderRectangle(new HeaderRectangle(
          new PixelRectangle(headerPosition, headerCell.getDimensions()),
          direction, overlayHeader));
    }
  }

  private static boolean cursorFound(@Nullable CellRectangles cellRectangles) {
    return cellRectangles != null && cellRectangles.getCursorRectangle() != null;
  }

  @Override
  public Pixels getPaginatedHeight(RenderContext renderContext, Pixels startOffset,
      Pixels preferredMax, Pixels paperHeight)
  {
    return concatenatedCell.getPaginatedHeight(
        renderContext, startOffset, preferredMax, paperHeight);
  }

  @Override
  public Pixels getMinimumBaselineOffset(RenderContext renderContext) {
    return concatenatedCell.getMinimumBaselineOffset(renderContext);
  }

  @Override
  public String toString() {
    return headerCell + (direction.isVertical() ? "\\H\\" : "|H|") + contentCell;
  }
}
