// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.impl;

import com.google.common.base.Preconditions;
import com.sieuferd.cell.styles.LineStyle;
import com.sieuferd.cell.support.Axis;
import com.sieuferd.cell.support.Canvas;
import com.sieuferd.cell.support.Direction;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.Points;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.CellParam;
import com.sieuferd.cell.support.CellPaintParameters;
import com.sieuferd.cell.support.CellPaintParameters.Layer;
import com.sieuferd.cell.support.HitParameters;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.util.ObjectReference;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Rectangle2D;
import javax.annotation.Nullable;

public final class HorizontalBarCell extends EmptyCell {
  private static final LineStyle ZERO_MARKER_LINE_STYLE =
      new LineStyle(new Color(0, 0, 0, 255), new Points(1.0));
  // Overflow logic depends on a non-zero margin.
  private static final Points MINIMUM_MARGIN_WIDTH = new Points(1.0);
  private final Font sizingFont;
  private final Color negativeColor;
  private final Color positiveColor;
  private final Axis axis;
  private final double value;

  /**
   * @param value may be NaN to indicate a missing data point; infinity values are permitted as
   *        valid data points
   */
  public HorizontalBarCell(
      RenderContext renderContext, Coordinates dimensions, Direction localCursorBias,
      Font sizingFont, Color negativeColor, Color positiveColor, Axis axis, double value)
  {
    super(renderContext, dimensions, localCursorBias);
    Preconditions.checkNotNull(sizingFont);
    Preconditions.checkNotNull(negativeColor);
    Preconditions.checkNotNull(positiveColor);
    Preconditions.checkNotNull(axis);
    this.sizingFont = sizingFont;
    this.negativeColor = negativeColor;
    this.positiveColor = positiveColor;
    this.axis = axis;
    this.value = value;
  }

  private @Nullable Pixels getPixelPosition(RenderContext rcx, double forValue, Pixels cellWidth,
      Pixels marginWidth, boolean constrainEdges, ObjectReference<Double> fractionalPixelWidth)
  {
    final Pixels rangeWidth = cellWidth.minus(marginWidth.times(2));
    if (!rangeWidth.isGreaterThan(Pixels.ZERO))
      return null;
    double relativePosition = axis.getRelativePosition(forValue);
    if (Double.isInfinite(relativePosition) && constrainEdges)
      return relativePosition < 0 ? Pixels.ZERO : cellWidth;
    if (!Double.isFinite(relativePosition))
      return null;
    Pixels ret = rangeWidth.times(rcx, relativePosition).plus(marginWidth);
    fractionalPixelWidth.setValue(rangeWidth.getDouble() * relativePosition);
    if (ret.isNegative())
      return constrainEdges ? Pixels.ZERO : null;
    if (ret.isGreaterThan(cellWidth))
      return constrainEdges ? cellWidth : null;
    return ret;
  }

  @Override
  public void traverse(
      CellParam param, Coordinates pos, Coordinates justifiedDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      @Nullable HitParameters hitParameters)
  {
    super.traverse(
        param, pos, justifiedDimensions, baselineOffset, headerPoint, overlayMode, hitParameters);
    final CellPaintParameters pp = param.getPaintParameters();
    if (overlayMode || pp == null || !pp.includesLayer(Layer.TOP) ||
        Double.isNaN(value))
    {
      return;
    }
    final Canvas canvas = pp.getCanvas();
    final RenderContext rcx = canvas.getRenderContext();
    final Graphics2D g = canvas.getGraphics();
    final Coordinates effDim = getDimensions().max(justifiedDimensions);

    /* TODO[pd]: Avoid recomputing these font metrics every time. (Here and another call to
                 scaledFont in processText.) */
    final Font scaledFont = rcx.scaledFont(sizingFont);
    final float descent = scaledFont.getLineMetrics("g", rcx.getFontRenderContext()).getDescent();
    final Pixels marginWidth;
    {
      Pixels toMarginWidth = Pixels.create(descent);
      if (rcx.getRoundToPixels() != null)
        toMarginWidth = toMarginWidth.roundUp(rcx.getRoundToPixels());
      toMarginWidth = toMarginWidth.max(Pixels.create(rcx, MINIMUM_MARGIN_WIDTH));
      marginWidth = toMarginWidth;
    }
    if (!marginWidth.isGreaterThan(Pixels.ZERO))
      throw new AssertionError();
    // Use a flat-topped letter like "H" to determine the font's cap height.
    final Pixels barHeight = Pixels.create(rcx, scaledFont.createGlyphVector(
        rcx.getFontRenderContext(), "H").getVisualBounds().getHeight());

    /* TODO[pd]: Calcluate a baseline offset from sizingFont. Then use it here and return it from
                 getMinimumBaselineOffset. */
    /* This cell should always be adjacent to another cell with a well-defined baseline offset, so
    the null baselineOffset case shouldn't really happen. */
    final Pixels barYposAbs = pos.getY().plus(baselineOffset == null
          ? marginWidth
          : baselineOffset.minus(barHeight));
    final Pixels zeroMarkerOvershoot = Pixels.create(rcx, descent / 2.0);
    final Pixels zeroMarkerYPosAbs = barYposAbs.minus(zeroMarkerOvershoot);
    final Pixels zeroMarkerHeight = barHeight.plus(zeroMarkerOvershoot.times(2));

    final Pixels zeroXPosRel = getPixelPosition(
        rcx, 0.0, effDim.getX(), marginWidth, false, ObjectReference.create());
    final ObjectReference<Double> fractionalPixelWidth = ObjectReference.create();
    final Pixels valueXPosRel = getPixelPosition(
        rcx, value, effDim.getX(), marginWidth, true, fractionalPixelWidth);

    if (zeroXPosRel != null && valueXPosRel != null) {
      g.setColor(canvas.getRenderContext().safeColor(value < 0.0 ? negativeColor : positiveColor));
      /* The zig zag clipping effect is applied as soon as the bar starts infringing on the
      horizontal margin by any visible amount. */
      if (value < 0) {
        drawBar(g, rcx, false, marginWidth.isGreaterThan(valueXPosRel)
              ? marginWidth : Pixels.ZERO,
            pos.getX().plus(valueXPosRel),
            barYposAbs,
            zeroXPosRel.minus(valueXPosRel),
            barHeight,
            fractionalPixelWidth.getValue());
      } else if (value > 0) {
        drawBar(g, rcx, true, valueXPosRel.isGreaterThan(effDim.getX().minus(marginWidth))
              ? marginWidth : Pixels.ZERO,
            pos.getX().plus(zeroXPosRel),
            barYposAbs,
            valueXPosRel.minus(zeroXPosRel),
            barHeight,
            fractionalPixelWidth.getValue());
      }
      if (axis.start < 0.0) {
        // Center the marker on the zero point.
        Pixels markerOffset = ZERO_MARKER_LINE_STYLE.measure(rcx, true).divideBy(rcx, 2.0);
        ZERO_MARKER_LINE_STYLE.paint(canvas, true, new Coordinates(
            pos.getX().plus(zeroXPosRel).minus(markerOffset),
            zeroMarkerYPosAbs), zeroMarkerHeight);
      }
    }
  }

  /**
   * @param zigZagWidth may be zero to indicate no zigzag pattern
   */
  private void drawBar(Graphics2D g, RenderContext rcx, boolean positive, Pixels zigZagWidth,
      Pixels posX, Pixels posY, Pixels dimX, Pixels dimY, double fractionalPixelWidth)
  {
    // TODO[pd]: Make sure the zigZag doesn't overshoot the cell.
    final double[] ZIGZAG_PATTERN = new double[] { 0.0, 0.6, -0.4, 1.2, 0.3 };
    final double zigHeight = dimY.getDouble() / (ZIGZAG_PATTERN.length - 1);
    final double zigWidth = Math.min(dimX.getDouble() / 2.0, zigZagWidth.getDouble());

    Color color = positive ? positiveColor : negativeColor;
    if (rcx.isAlphaPermitted() &&
        rcx.getRoundToPixels() != null && !dimX.isGreaterThan(rcx.getRoundToPixels()))
    {
      color = UIUtil.opacity(color, Math.max(0f, Math.min(1f, (float) fractionalPixelWidth)));
      dimX = rcx.getRoundToPixels();
    }
    g.setColor(rcx.safeColor(color));
    if (zigZagWidth.isZero()) {
      g.fill(new Rectangle2D.Double(
          posX.getDouble(), posY.getDouble(), dimX.getDouble(), dimY.getDouble()));
    } else {
      final Path2D.Double p = new Path2D.Double();
      p.moveTo(0, 0);
      for (int i = 0; i < ZIGZAG_PATTERN.length; i++)
        p.lineTo(dimX.getDouble() - zigWidth * ZIGZAG_PATTERN[i], i * zigHeight);
      p.lineTo(0, dimY.getDouble());
      if (!positive) {
        p.transform(AffineTransform.getScaleInstance(-1, 1.0));
        p.transform(AffineTransform.getTranslateInstance(dimX.getDouble(), 0.0));
      }
      p.transform(AffineTransform.getTranslateInstance(posX.getDouble(), posY.getDouble()));
      g.fill(p);
    }
  }
}
