// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.impl;

import com.sieuferd.cell.support.BiasedLocation;
import com.sieuferd.cell.support.CellParam;
import com.sieuferd.cell.support.Direction;
import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.sieuferd.cell.support.Cell;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.CursorStyle;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.CellPaintParameters;
import com.sieuferd.cell.support.CellPaintParameters.Layer;
import com.sieuferd.cell.support.HitParameters;
import com.sieuferd.uiutil.UIUtil;
import javax.annotation.Nullable;

public final class ListCell extends Cell {
  private static final boolean DEBUG_CLIPPING = false;
  /**
   * Should the first cell in a vertical stack of more than one cell be baseline-aligned? I've
   * changed my mind on this several times. Setting this to true looks better in the full table
   * course catalog example with a large-font title.
   */
  private static final boolean BASELINE_ALIGN_FIRST_CELL_IN_VERTICAL_STACK = true;
  private final Direction direction;
  private final List<Cell> childCells;
  /** List of strictly inreasing positions of child cells. The data in this field is redundant,
  since it can be calculated from the contents of childCells, but it is stored here to allow quick
  binary search lookup. */
  private final List<Pixels> childCellPositions;
  private final Coordinates dimensions;
  /**
   * The overprint limit. For simplicity, this stores the maximum of any child's
   * getBottomRightOverprintLimit(), which is then used for the overall ListCell's
   * getBottomRightOverprintLimit() as well. In the still direction, this is actually the right
   * thing to do in any case, since the ListCell may be justified at paint time.
   */
  private final Coordinates bottomRightOverprintLimit;
  private final boolean containsSelectable;
  private final Pixels minimumBaselineOffset;

  public ListCell(RenderContext renderContext, Coordinates minimumDimensions,
      boolean vertical, List<Cell> childCells)
  {
    super(renderContext);
    this.direction = vertical ? Direction.DOWN : Direction.RIGHT;
    ArrayList<Cell> toChildCells = new ArrayList<>(childCells.size());
    ArrayList<Pixels> toChildCellPositions = new ArrayList<>(childCells.size());
    Pixels flowPos = Pixels.ZERO;
    Pixels stillWidth = Pixels.ZERO;
    Coordinates toBottomRightOverprintLimit = Coordinates.ZERO;
    boolean toContainsSelectable = false;
    Pixels toMinimumBaselineOffset = Pixels.ZERO;
    for (Cell childCell : childCells) {
      toContainsSelectable |= childCell.containsBiasedLocation();
      stillWidth = stillWidth.max(direction.stillDir(childCell.getDimensions()));
      toBottomRightOverprintLimit =
          toBottomRightOverprintLimit.max(childCell.getBottomRightOverprintLimit(renderContext));
      if (!vertical || flowPos.isZero()) {
        toMinimumBaselineOffset =
            toMinimumBaselineOffset.max(childCell.getMinimumBaselineOffset(renderContext));
      }
      {
        final int lastIndex = toChildCellPositions.size() - 1;
        if (lastIndex >= 0 && toChildCellPositions.get(lastIndex).equals(flowPos)) {
          /* Ensure sensible behavor in the presence of struts of zero length in the flow direction,
          in particular so that no two cells are at the same starting position. Make sure to keep
          the last cell even if it has a zero length, as it may still be expanded during
          justification. */
          toChildCells.set(lastIndex, childCell);
        } else {
          toChildCells.add(childCell);
          toChildCellPositions.add(flowPos);
        }
      }
      flowPos = flowPos.plus(direction.flowDir(childCell.getDimensions()));
    }
    this.containsSelectable = toContainsSelectable;
    toChildCells.trimToSize();
    toChildCellPositions.trimToSize();
    this.childCells = Collections.unmodifiableList(toChildCells);
    this.childCellPositions = Collections.unmodifiableList(toChildCellPositions);
    this.dimensions = direction.newCoordinates(flowPos, stillWidth).max(minimumDimensions);
    this.bottomRightOverprintLimit = toBottomRightOverprintLimit;
    this.minimumBaselineOffset = toMinimumBaselineOffset;
  }

  @Override
  public Coordinates getDimensions() {
    return dimensions;
  }

  @Override
  public Coordinates getBottomRightOverprintLimit(RenderContext renderContext) {
    return bottomRightOverprintLimit;
  }

  private int getFirstCellIndex(Pixels flowDir) {
    int ret;
    int bsRes = Collections.binarySearch(childCellPositions, flowDir);
    if (bsRes >= 0) {
      ret = bsRes;
    } else {
      // Reverse the odd binarySearch() return value convention.
      int insertionPoint = -(bsRes + 1);
      /* We want to start with the last cell _before_ the insertion point, so subtract one. But
      there are (childCells.size() + 1) possible insertion points, and flowDir may be negative
      or childCells.size() may be zero, so we need to eliminate the -1 case. */
      ret = Math.max(0, insertionPoint - 1);
    }
    return ret;
  }

  @Override
  public void traverse(
      CellParam param, Coordinates position, Coordinates justifiedDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      @Nullable HitParameters hitParameters)
  {
    final RenderContext renderContext = param.getRenderContext();
    final Coordinates effectiveDimensions = getDimensions().max(justifiedDimensions);

    if (hitParameters != null) {
      final BiasedLocation biasedLocation = findBiasedLocation(
          param, position, justifiedDimensions, effectiveDimensions, baselineOffset,
          headerPoint, overlayMode, hitParameters.getCursorBias());
      hitParameters.setResult(biasedLocation);
    }

    if (param.getCellRectangles() == null && param.getPaintParameters() == null)
      return; // No further traversal necessary.

    final Pixels clipBegin;
    final Pixels clipEnd;

    {
      final PixelRectangle clip = param.getClipBounds();
      final Coordinates clipTL = clip.getPosition();
      final Coordinates clipBR = clipTL.plus(clip.getDimensions());
      /* Account for cursor overprinting, which goes in all four directions, as well as regular
      bottom-right overprinting. */
      clipBegin = direction.flowDir(clipTL.minus(position)).minus(
          direction.flowDir(bottomRightOverprintLimit).max(direction.flowDir(
          CursorStyle.CURSOR_DIMENSIONS.bottomRightBorderDimensions(renderContext))));
      clipEnd = direction.flowDir(clipBR.minus(position)).plus(direction.flowDir(
          CursorStyle.CURSOR_DIMENSIONS.topLeftBorderDimensions(renderContext)));
    }
    for (int i = getFirstCellIndex(clipBegin); i < childCells.size(); i++) {
      // Stop once past the clipping rectangle.
      if (childCellPositions.get(i).isGreaterThan(clipEnd))
        break;

      traverseChildCell(param, position, justifiedDimensions, effectiveDimensions,
          baselineOffset, headerPoint, overlayMode, i, null);
    }
  }

  private BiasedLocation findBiasedLocation(CellParam param, Coordinates position,
      Coordinates justifiedDimensions, Coordinates effectiveDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      Direction cursorBias)
  {
    // See the Javadoc for HitParameters.setResult.
    final Coordinates positionInCellCoordinateSystem = Coordinates.ZERO.minus(position);
    /* TODO[pd]: Consider skipping cells with zero height in the flow direction. This case can be
                 tested with an empty table with all headers hidden. */
    Integer incr = null;
    for (int i = getFirstCellIndex(direction.flowDir(positionInCellCoordinateSystem));
         i >= 0 && i < childCells.size();
         i += incr)
    {
      final BiasedLocation ret;
      {
        final HitParameters childHitParam = new HitParameters(cursorBias);
        traverseChildCell(param, position, justifiedDimensions, effectiveDimensions, baselineOffset,
            headerPoint, overlayMode, i, childHitParam);
        ret = childHitParam.getResult();
      }
      if (ret.getCursorLocation() != null)
        return ret;
      if (incr == null) {
        // TODO[pd]: See if there can be situations allowing inifinite bias loops.
        Direction useCursorBias =
            (ret.getCursorBias() != Direction.NONE) ? ret.getCursorBias() : cursorBias;
        incr = (direction.isVertical() ?
            useCursorBias.getVerticalBiasDown() :
            useCursorBias.getHorizontalBiasRight());
      }
      if (incr == 0)
        return BiasedLocation.NONE;
    }
    return BiasedLocation.NONE;
  }

  /**
   * Traverse a single child cell. All parameters except those prefixed "child" relate to the parent
   * ListCell.
   */
  private void traverseChildCell(CellParam param, Coordinates position,
      Coordinates justifiedDimensions, Coordinates effectiveDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      final int childIndex, @Nullable HitParameters childHitParameters)
  {
    final boolean lastOne = (childIndex == childCells.size() - 1);
    final Cell childCell = childCells.get(childIndex);
    final Pixels childCellPos = childCellPositions.get(childIndex);

    final Coordinates absPos = direction.addFlow(position, childCellPos);

    final Coordinates childJustifiedDimensions;
    if (!justifiedDimensions.isEqualTo(Coordinates.ZERO)) {
      childJustifiedDimensions = direction.newCoordinates(
          lastOne ? direction.flowDir(effectiveDimensions).minus(childCellPos) : Pixels.ZERO,
          direction.stillDir(effectiveDimensions));
    } else {
      childJustifiedDimensions = Coordinates.ZERO;
    }
    final Pixels childBaselineOffset = (direction.isVertical() &&
        (BASELINE_ALIGN_FIRST_CELL_IN_VERTICAL_STACK ? (childIndex > 0) : (childCells.size() > 1)))
        ? null
        : minimumBaselineOffset.max(baselineOffset);
    childCell.traverse(param, absPos, childJustifiedDimensions, childBaselineOffset,
        headerPoint, overlayMode, childHitParameters);

    final CellPaintParameters pp = param.getPaintParameters();
    if (DEBUG_CLIPPING &&
        pp != null && pp.includesLayer(Layer.BACKGROUND) && !overlayMode)
    {
      Graphics2D g = pp.getCanvas().getGraphics();
      g.setColor(UIUtil.blendOver(Color.getHSBColor(
          (float) (1.0 * Math.random()), 1f, 1), new Color(255, 255, 255, 0), 0.1f));
      g.fill(new Rectangle2D.Double(absPos.getX().getDouble(), absPos.getY().getDouble(),
          childCell.getDimensions().getX().getDouble(),
          childCell.getDimensions().getY().getDouble()));
    }
  }

  @Override
  public String toString() {
    return Joiner.on(direction.isVertical() ? "\\\\" : "|").join(childCells);
  }

  @Override
  public Pixels getPaginatedHeight(RenderContext renderContext,
      Pixels startOffset, Pixels preferredMax, Pixels paperHeight) {
    checkRenderContext(renderContext);
    Preconditions.checkArgument(preferredMax.isGreaterThan(Pixels.ZERO));
    Preconditions.checkArgument(paperHeight.isGreaterThan(Pixels.ZERO));
    Preconditions.checkArgument(!startOffset.isNegative());
    /* This function is not direction-agnostic. Do not refer to flowDir/stillDir
    here. Mixing them will cause confusion or bugs. */
    final Pixels preferredEnd = startOffset.plus(preferredMax);
    Pixels ret = Pixels.ZERO;
    if (direction.isVertical()) {
      for (int i = getFirstCellIndex(startOffset); i < childCells.size(); i++) {
        Cell childCell = childCells.get(i);
        Pixels curY = childCellPositions.get(i);
        Pixels remaining = preferredMax.minus(curY.minus(startOffset));
        Pixels childFullHeight = childCell.getDimensions().getY();
        Pixels fullEndY = curY.plus(childFullHeight);
        /* If the child does not fit on this page, but it would fit on the next page without
        internal breaking, force a page break here. */
        if (fullEndY.isGreaterThan(preferredEnd) && !childFullHeight.isGreaterThan(paperHeight))
          break;
        Pixels brokenEndY = curY.plus(
            childCell.getPaginatedHeight(renderContext, Pixels.ZERO, remaining, paperHeight));
        /* This wouldn't fit on this page even if internally broken as much as possible, so force a
        page break. */
        if (brokenEndY.isGreaterThan(preferredEnd))
          break;
        ret = brokenEndY.minus(startOffset);
      }
    } else {
      for (Cell childCell : childCells) {
        Pixels res =
            childCell.getPaginatedHeight(renderContext, startOffset, preferredMax, paperHeight);
        if (!res.isGreaterThan(preferredMax))
          ret = ret.max(res);
      }
    }
    if (!ret.isGreaterThan(Pixels.ZERO))
      ret = getDimensions().getY().minus(startOffset);
    return ret;
  }

  @Override
  public boolean containsBiasedLocation() {
    return containsSelectable;
  }

  @Override
  public Pixels getMinimumBaselineOffset(RenderContext renderContext) {
    return minimumBaselineOffset;
  }
}
