// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.impl;

import com.sieuferd.cell.styles.TextFitPolicy;
import com.google.common.base.Preconditions;
import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.cell.styles.Text;
import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.support.BiasedLocation;
import com.sieuferd.cell.support.CellParam;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import com.sieuferd.cell.styles.TextStyle;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.CellPaintParameters;
import com.sieuferd.cell.support.CellPaintParameters.Layer;
import com.sieuferd.cell.support.HitParameters;
import java.awt.Shape;
import javax.annotation.Nullable;

public final class TextCell extends Cell {
  private static final boolean DEBUG_BOUNDS = false;
  private final TextStyle textStyle;
  private final Color color;
  private final Text text;
  private final boolean vertical;
  private final Coordinates dimensions;
  private final Pixels firstBaselineOffset;
  private final TextFitPolicy textFitPolicy;
  private final HorizontalAlignment horizontalAlignment;
  private final boolean justified;

  public TextCell(RenderContext renderContext, Coordinates dimensions, Pixels firstBaselineOffset,
      TextStyle textStyle, Color color, Text text, boolean vertical, TextFitPolicy textFitPolicy,
      HorizontalAlignment horizontalAlignment, boolean justified)
  {
    super(renderContext);
    Preconditions.checkNotNull(dimensions);
    Preconditions.checkNotNull(firstBaselineOffset);
    Preconditions.checkNotNull(textStyle);
    Preconditions.checkNotNull(color);
    Preconditions.checkNotNull(text);
    this.textStyle = textStyle;
    this.color = color;
    this.text = text;
    this.vertical = vertical;
    this.dimensions = dimensions;
    this.firstBaselineOffset = firstBaselineOffset;
    this.textFitPolicy = textFitPolicy;
    this.horizontalAlignment = horizontalAlignment;
    this.justified = justified;
  }

  @Override
  public Coordinates getDimensions() {
    return dimensions;
  }

  @Override
  public Coordinates getBottomRightOverprintLimit(RenderContext renderContext) {
    return new Coordinates(Pixels.ZERO, getVerticalOverprintLimit());
  }

  private Pixels getVerticalOverprintLimit() {
    /* Allow descenders to extend beyond the vertical end of the cell. The amount to allow is
    somewhat arbitrary, but use the baseline offset, since it is likely to be as large as any
    descender, and since it is properly sized according to the font in use. */
    return firstBaselineOffset;
  }

  @Override
  public void traverse(
      CellParam param, Coordinates position, Coordinates justifiedDimensions,
      @Nullable Pixels baselineOffset, Coordinates headerPoint, boolean overlayMode,
      @Nullable HitParameters hitParameters)
  {
    if (hitParameters != null)
      hitParameters.setResult(BiasedLocation.NONE);
    final CellPaintParameters pp = param.getPaintParameters();
    if (overlayMode || pp == null || !pp.includesLayer(Layer.TOP))
      return; // No action.
    /* Don't justify the text box in the horizontal direction, as it might cause text breaking to
    change from what the TextCell was originally measured for. */
    final Coordinates effectiveDimensions = getDimensions().max(new Coordinates(
        Pixels.ZERO, justifiedDimensions.getY()));
    final Graphics2D g = pp.getCanvas().getGraphics();
    if (DEBUG_BOUNDS) {
      g.setColor(new Color(128, 255, 128, 50));
      g.fill(new Rectangle2D.Double(
          position.getX().getDouble(), position.getY().getDouble(),
          effectiveDimensions.getX().getDouble(), effectiveDimensions.getY().getDouble()));
    }
    final @Nullable Shape oldClip = g.getClip();
    {
      g.clip(new PixelRectangle(position, new Coordinates(
          effectiveDimensions.getX(),
          effectiveDimensions.getY().plus(getVerticalOverprintLimit()))
      ).getRectangle2D());
    }
    textStyle.paint(pp.getCanvas(), text, new PixelRectangle(position, effectiveDimensions),
        baselineOffset, vertical, textFitPolicy, color, horizontalAlignment, justified);
    g.setClip(oldClip);
  }

  @Override
  public String toString() {
    return text.getString().replace('\n', ' ');
  }

  @Override
  public Pixels getPaginatedHeight(RenderContext renderContext,
      Pixels startOffset, Pixels preferredMax, Pixels paperHeight)
  {
    checkRenderContext(renderContext);
    Pixels fullHeight = getDimensions().getY().minus(startOffset);
    if (vertical || !startOffset.plus(fullHeight).isGreaterThan(preferredMax))
      return fullHeight;
    Pixels ret = textStyle.getHorizontalLineBreakBefore(renderContext, text,
        startOffset.plus(preferredMax), getDimensions()).minus(startOffset);
    return ret.isGreaterThan(startOffset) ? ret : fullHeight;
  }

  @Override
  public boolean containsBiasedLocation() {
    return false;
  }

  @Override
  public Pixels getMinimumBaselineOffset(RenderContext renderContext) {
    return firstBaselineOffset;
  }
}
