// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.styles;

import com.google.common.base.Preconditions;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.Points;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import com.sieuferd.cell.support.Canvas;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.math.BigDecimal;
import javax.annotation.Nullable;

public class LineStyle {
  public static final boolean DEBUG_FORCE_STROKES = false;
  /* TODO[pd]: If I'm not going to do coherent dashes anyway, consider automatically adjusting the
               dash period so that the beginning and the end of the line is always a drawn dash. */
  /* TODO[pd]: Coherent dashes can't quite be made to work with the current API, because it would
               require the position relative to the entire layout, rather than relative to the
               current viewport, to be known. Arguably, it also looks bad to start a dashed border
               with an open space, so maybe I shouldn't be trying to do this in the first place. */
  public static final boolean DEBUG_COHERENT_DASHES = false;
  public static final LineStyle NONE = new LineStyle(null, Points.ZERO);
  public static final LineStyle DEBUG1 = new LineStyle(Color.RED, new Points(5));
  public static final LineStyle DEBUG2 = new LineStyle(Color.BLUE, new Points(2));
  public static final LineStyle DEBUG3 = new LineStyle(Color.GREEN, new Points(7));
  public static final LineStyle DEBUG4 = new LineStyle(Color.ORANGE, new Points(4));

  private final @Nullable Color color;
  private final Points thickness;
  private final @Nullable Points dashPeriod;
  private final BigDecimal dashDutyCycle;

  public LineStyle(@Nullable Color color, Points thickness,
      @Nullable Points dashPeriod, BigDecimal dashDutyCycle)
  {
    Preconditions.checkNotNull(thickness);
    Preconditions.checkArgument(dashPeriod == null || dashPeriod.getValue() > 0);
    Preconditions.checkArgument(dashDutyCycle.equals(BigDecimal.ONE) == (dashPeriod == null));
    Preconditions.checkArgument(dashDutyCycle.compareTo(BigDecimal.ZERO) > 0);
    Preconditions.checkArgument(dashDutyCycle.compareTo(BigDecimal.ONE) <= 0);
    this.color = color;
    this.thickness = thickness;
    this.dashPeriod = dashPeriod;
    this.dashDutyCycle = dashDutyCycle;
  }

  public LineStyle(@Nullable Color color, Points thickness) {
    this(color, thickness, null, BigDecimal.ONE);
  }

  @Override
  public String toString() {
    return "LineStyle(" + color + ", " + thickness + ")";
  }

  public @Nullable Color getColor() {
    return color;
  }

  public Points getThickness() {
    return thickness;
  }

  public Pixels measure(RenderContext rcx, boolean vertical) {
    return paintHelper(rcx, vertical, null, null, null);
  }

  public Pixels paint(Canvas canvas, boolean vertical, Coordinates pos, Pixels length) {
    return paintHelper(canvas.getRenderContext(), vertical, canvas.getGraphics(), pos, length);
  }

  private Pixels paintHelper(
      RenderContext rcx, boolean vertical, Graphics2D g, Coordinates pos, Pixels length)
  {
    if (color == null)
      return Pixels.create(rcx, thickness);
    if (thickness.isZero())
      return Pixels.ZERO;

    // If alpha channel can not be used, simulate by making the line thinner.
    double targetThicknessPoints = thickness.getValue() *
        (rcx.isAlphaPermitted() ? 1.0 : (color.getAlpha() / 255.0));
    final Pixels useThickness = Pixels.create(rcx, new Points(targetThicknessPoints));

    if (g != null) {
      final Color useColor;
      /* If the alpha channel can be used, use it to compensate for any thickness overshoot due to
      roundoff error. */
      if (rcx.isAlphaPermitted()) {
        double overShoot = useThickness.getDouble() / (targetThicknessPoints * rcx.getScaling());
        double newAlpha  = color.getAlpha() / overShoot;
        useColor = new Color(color.getRed(), color.getGreen(), color.getBlue(),
            Math.min(Math.max((int) newAlpha, 0), 255));
      } else {
        useColor = color;
      }
      if (useColor.getAlpha() > 0 && !useThickness.isZero()) {
        final double lengthD = length.getDouble();
        final double useThicknessD = useThickness.getDouble();
        g.setColor(rcx.safeColor(useColor));
        if (dashPeriod != null || DEBUG_FORCE_STROKES) {
          final double adjX = pos.getX().getDouble() + (vertical ? (useThicknessD / 2) : 0);
          final double adjY = pos.getY().getDouble() + (vertical ? 0 : (useThicknessD / 2));
          final Stroke oldStroke = g.getStroke();
          final Stroke stroke;
          if (dashPeriod == null) {
            stroke = new BasicStroke((float) useThicknessD,
                BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0);
          } else {
            final Pixels onPixels = Pixels.create(rcx, new Points(
                dashPeriod.getValue() * dashDutyCycle.doubleValue()));
            final Pixels offPixels = Pixels.create(rcx, new Points(
                dashPeriod.getValue() * (1 - dashDutyCycle.doubleValue())));
            float dashPattern[] = new float[] {
              (float) onPixels.getDouble(),
              (float) offPixels.getDouble(),
            };
            /* Set the phase to keep dashed lines coherent even if painted in multiple separate
            but adjacent segments. */
            final double along = DEBUG_COHERENT_DASHES
                ? (vertical ? pos.getY() : pos.getX()).getDouble()
                : 0;
            final double useDashPeriod = onPixels.plus(offPixels).getDouble();
            final float phase = (float) (along - Math.floor(along / useDashPeriod) * useDashPeriod);
            stroke = new BasicStroke((float) useThicknessD,
                BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0, dashPattern, phase);
          }
          g.setStroke(stroke);
          try {
            /* Confirmed to render correctly with RenderingHints.VALUE_STROKE_PURE (for non-dashed
            lines, should yield exactly the same result as the Rectangle2D-based method below). */
            g.draw(new Line2D.Double(adjX, adjY,
                adjX + (vertical ? 0 : lengthD),
                adjY + (vertical ? lengthD : 0)
            ));
          } finally {
            g.setStroke(oldStroke);
          }
        } else {
          /* "For horizontal or vertical 1-pixel wide line fills, it is generally better to use
          fillRect rather than drawLine."
          http://java.sun.com/products/java-media/2D/reference/faqs/index.html */
          /* Use g.fill(Shape) instead of fillRect to be able to specify a rectangle coordinates as
          doubles. */
          g.fill(new Rectangle2D.Double(pos.getX().getDouble(), pos.getY().getDouble(),
              (vertical ? useThicknessD : lengthD      ),
              (vertical ? lengthD       : useThicknessD)));
        }
      }
    }
    return useThickness;
  }
}
