// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.styles;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.util.StringUtil;
import java.awt.Font;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.text.BreakIterator;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

/**
 * Container for text that has been broken into lines and paragraphs, and which is ready to be
 * painted.
 */
final class ParagraphLayout {
  private static final Logger LOG = Logger.getLogger(ParagraphLayout.class.getName());
  private static final double MINIMUM_SCALE_OVERFLOW_PIXELS = 0.1;
  /* Even empty lines should be counted for baseline alignment purposes, to be consistent with the
  fact that even empty cells have an ascent equal to a single line. This design ensures tidy
  vertical spacing, as well as visual stability in face of edited values. */
  private static final boolean COUNT_EMPTY_BASELINES = true;

  private final ImmutableList<TextLayout> lines;
  private final ImmutableList<Double> baselineOffsets;
  private final double width;
  private final double height;
  private final double requiredScaling;
  private final String debugString;

  private ParagraphLayout(
      List<TextLayout> lines, double width, double height, double requiredScaling,
      String debugString)
  {
    Preconditions.checkNotNull(lines);
    Preconditions.checkArgument(width >= 0.0);
    Preconditions.checkArgument(height >= 0.0);
    Preconditions.checkArgument(requiredScaling >= 0.0);
    Preconditions.checkNotNull(debugString);
    this.lines = ImmutableList.copyOf(lines);
    this.width = width;
    this.height = height;
    this.requiredScaling = requiredScaling;
    this.debugString = debugString;
    List<Double> toBaselineOffsets = new ArrayList<>(lines.size());
    double curY = 0.0;
    for (TextLayout line : lines) {
      if (!COUNT_EMPTY_BASELINES && line.getVisibleAdvance() == 0) {
        continue;
      }
      curY += requiredScaling * line.getAscent();
      toBaselineOffsets.add(curY);
      curY += requiredScaling * (line.getDescent() + line.getLeading());
    }
    baselineOffsets = ImmutableList.copyOf(toBaselineOffsets);
  }

  /**
   * Get the {@code TextLayout}s representing each line of the paragraph layout. Clients are
   * prohibited from modifying the returned objects.
   */
  public List<TextLayout> getLines() {
    return lines;
  }

  public List<Double> getBaselineOffsets() {
    return baselineOffsets;
  }

  public double getWidth() {
    return width;
  }

  public double getHeight() {
    return height;
  }

  /**
   * Get the scaling required when painting lines returned by {@link #getLines()} such that the
   * final dimensions are those returned by {@link #getWidth()}/{@link #getHeight()}.
   */
  public double getRequiredScaling() {
    return requiredScaling;
  }

  public ParagraphLayout scaled(double scaling) {
    Preconditions.checkArgument(scaling >= 0);
    return new ParagraphLayout(
        lines, width * scaling, height * scaling, this.requiredScaling * scaling, debugString);
  }

  private static ParagraphLayout createInternal(
      RenderContext rcx, Text text, Font font, boolean justified, LineBreakPolicy lineBreakPolicy,
      double lineLength, boolean firstLineOnly)
  {
    Preconditions.checkNotNull(rcx);
    Preconditions.checkNotNull(text);
    Preconditions.checkNotNull(font);
    Preconditions.checkNotNull(lineBreakPolicy);
    Preconditions.checkArgument(lineLength >= 0);
    double useLineLength = Math.max(
        /* This epsilon will be added back at the end; forgetting to subtract it here once caused a
        rare bug in the vertical header decision logic. */
        lineLength - Pixels.EPSILON.getDouble(),
        // Functions like getJustifiedLayout() below may require a strictly positive width.
        Pixels.EPSILON.getDouble());

    final String s;
    final AttributedString as;
    {
      final String originalString = text.getString();
      // TODO[pd]: Don't call scaledFont over and over again.
      final Font scaledFont = rcx.scaledFont(font);
      final Map<TextAttribute,Object> originalAttributes =
          Maps.newLinkedHashMap(scaledFont.getAttributes());
      /* Oddly enough, TextAttribute.SIZE seems not to be defined if the size is 12 (presumably a
      default value somewhere). The following returns null on Java 1.8.0_72 for MacOS:

          new Font("Tahoma", Font.PLAIN, 12).getAttributes().get(TextAttribute.SIZE)
      */
      if (originalAttributes.get(TextAttribute.SIZE) == null)
        originalAttributes.put(TextAttribute.SIZE, (Float) scaledFont.getSize2D());
      if (originalString.isEmpty()) {
        /* AttributedString/LineBreakMeasurer don't work well for empty strings, and we always wish
        to produce at least one line. */
        s = " ";
        as = new AttributedString(s, originalAttributes);
      } else {
        s = originalString;
        as = text.createAttributedString(originalAttributes, rcx);
      }
      if (s.length() != as.getIterator().getEndIndex())
        throw new AssertionError();
    }

    BreakIterator breakIter = lineBreakPolicy.createBreakIterator();
    AttributedCharacterIterator aci = as.getIterator();
    /* TODO[pd]: Consider synchronizing access to LineBreakMeasurer and friends, in case they have
                 global state that is not thread-safe. I've seen some very rare errors that could
                 indicate problems with thread-safety, e.g. this NPE:

                   java.lang.NullPointerException
                     at sun.font.ExtendedTextSourceLabel.createCharinfo(ExtendedTextSourceLabel.java:715)

                 On the other hand, various JDK bug reports seem to indicate that these APIs
                 are supposed to work from any thread, e.g.:

                   http://bugs.sun.com/view_bug.do?bug_id=6611637
    */
    LineBreakMeasurer lbm = new LineBreakMeasurer(aci, breakIter, rcx.getFontRenderContext());

    double actWidth = 0.0f, actHeight = 0.0f;
    double curY = 0.0f;
    int nextNewline = 0;
    List<TextLayout> lines = Lists.newArrayList();
    lbm.setPosition(aci.getBeginIndex());
    final int endIndex = aci.getEndIndex();
    while (lbm.getPosition() < endIndex) {
      // Account for any forced line breaks that might exist in the string to be displayed.
      if (nextNewline <= lbm.getPosition()) {
        for (nextNewline = lbm.getPosition() + 1; nextNewline < endIndex; nextNewline++) {
          if (isLinebreakAtIndex(s, nextNewline))
            break;
        }
      }
      final boolean noBadBreaks = true;
      TextLayout line = lbm.nextLayout((float) useLineLength, nextNewline, noBadBreaks);
      if (line == null) {
        int offsetLimit = Math.min(nextNewline, breakIter.following(lbm.getPosition()));
        line = lbm.nextLayout(Float.POSITIVE_INFINITY, offsetLimit, true);
        if (line == null)
          throw new AssertionError();
      }
      actWidth = Math.max(actWidth, line.getVisibleAdvance());
      if (justified && useLineLength < Double.POSITIVE_INFINITY) {
        // Don't justify the last line of a paragraph.
        if (lbm.getPosition() < endIndex && !isLinebreakAtIndex(s, lbm.getPosition()))
          line = line.getJustifiedLayout((float) useLineLength);
      }
      curY += line.getAscent() + line.getDescent();
      actHeight = curY;
      lines.add(line);
      curY += line.getLeading();
      if (firstLineOnly)
        break;
    }
    /* TODO[pd]: Include some logic to use a larger epsilon when dealing with very large numbers, or
                 verify that it is unnecessary. */
    /* Underreporting the required line length by any amount, such as that caused by roundoff error
    in float operations, may cause incorrect line breaks. Therefore, add a small epsilon to the
    calculated line width. */
    return new ParagraphLayout(lines, actWidth + Pixels.EPSILON.getDouble(), actHeight, 1.0, s);
  }

  private static boolean isLinebreakAtIndex(String s, int index) {
    final char c = s.charAt(index);
    /* Recognize either Windows (\r\n), Unix (\n), or pre-Unix Macintosh line endings (\r). The
    latter may seem like a relic of the past, but Excel for Mac 2011 generates them on CSV
    export, so they may occur for instance if such CSV files are imported to PostgreSQL via
    the COPY command. Be careful to only recognize a single line break for the "\r\n"
    combination. */
    return (c == '\r' || c == '\n' && (index == 0 || s.charAt(index - 1) != '\r'));
  }

  public static ParagraphLayout create(
      RenderContext rcx, Text text, Font font, boolean justified, LineBreakPolicy lineBreakPolicy,
      double breakLineLength, double fitLineLength, boolean firstLineOnly)
  {
    Preconditions.checkNotNull(rcx);
    Preconditions.checkNotNull(text);
    Preconditions.checkNotNull(font);
    Preconditions.checkNotNull(lineBreakPolicy);
    Preconditions.checkArgument(breakLineLength >= 0.0);
    ParagraphLayout ret =
        createInternal(rcx, text, font, justified, lineBreakPolicy, breakLineLength, firstLineOnly);
    if (ret.width > fitLineLength) {
      /* If the paragraph needs to be wider due to a long word that doesn't fit on a single line,
      redo the entire layout with the new width. */
      if (ret.width > breakLineLength)
        ret = createInternal(rcx, text, font, justified, lineBreakPolicy, ret.width, firstLineOnly);
      // Fit any resulting paragraph to the specified line length.
      if (ret.width - MINIMUM_SCALE_OVERFLOW_PIXELS > fitLineLength)
        ret = ret.scaled(fitLineLength / ret.width);
      return ret;
    } else {
      return ret;
    }
  }

  @Override
  public String toString() {
    return "PL(" + StringUtil.debugQuote(debugString) + ")";
  }
}
