// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.styles;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import java.util.HashMap;
import java.util.Map;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;

public abstract class TextMeasurer {
  /** @param prefMaxWidth can be null to indicate no break limit. The indicated width is
             horizontal regardless of the value of {@code vertical}. */
  public abstract TextMeasurement measure(TextStyle textStyle, Text text,
      @Nullable Pixels prefMaxWidth, boolean allowWider, boolean vertical);

  public static final class DefaultTextMeasurer extends TextMeasurer {
    private final RenderContext renderContext;

    public DefaultTextMeasurer(RenderContext renderContext) {
      Preconditions.checkNotNull(renderContext);
      this.renderContext = renderContext;
    }

    @Override
    public TextMeasurement measure(
        TextStyle textStyle, Text text, Pixels prefMaxWidth, boolean allowWider, boolean vertical)
    {
      return textStyle.measure(renderContext, text, prefMaxWidth, allowWider, vertical);
    }
  }

  @NotThreadSafe
  public static final class MemoizedTextMeasurer extends TextMeasurer {
    private final TextMeasurer delegate;
    private final Map<TextMeasureArguments,TextMeasurement> memo = new HashMap<>(10000);

    public MemoizedTextMeasurer(TextMeasurer delegate) {
      Preconditions.checkNotNull(delegate);
      this.delegate = delegate;
    }

    public TextMeasurer getDelegate() {
      return delegate;
    }

    @Override
    public TextMeasurement measure(TextStyle textStyle, Text text, @Nullable Pixels prefMaxWidth,
        boolean allowWider, boolean vertical)
    {
      TextMeasureArguments args =
          new TextMeasureArguments(textStyle, text, prefMaxWidth, allowWider, vertical);
      TextMeasurement ret = memo.get(args);
      if (ret == null) {
        ret = delegate.measure(textStyle, text, prefMaxWidth, allowWider, vertical);
        memo.put(args, ret);
      }
      return ret;
    }
  }

  private static final class TextMeasureArguments {
    final int hashCode;
    final TextStyle textStyle;
    final Text text;
    final @Nullable Pixels prefMaxWidth;
    final boolean allowWider;
    final boolean vertical;

    TextMeasureArguments(TextStyle textStyle, Text text, @Nullable Pixels prefMaxWidth,
        boolean allowWider, boolean vertical)
    {
      Preconditions.checkNotNull(textStyle);
      Preconditions.checkNotNull(text);
      this.textStyle = textStyle;
      this.text = text;
      this.prefMaxWidth = prefMaxWidth;
      this.allowWider = allowWider;
      this.vertical = vertical;
      this.hashCode = Objects.hashCode(textStyle, text, prefMaxWidth, allowWider, vertical);
    }

    @Override
    public int hashCode() {
      return hashCode;
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof TextMeasureArguments))
        return false;
      TextMeasureArguments other = (TextMeasureArguments) obj;
      return hashCode == other.hashCode
          && allowWider == other.allowWider
          && vertical == other.vertical
          && Objects.equal(textStyle, other.textStyle)
          && Objects.equal(text, other.text)
          && Objects.equal(prefMaxWidth, other.prefMaxWidth);
    }
  }
}
