// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.styles;

import com.google.common.base.Preconditions;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import com.sieuferd.cell.support.Canvas;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import java.util.Objects;
import javax.annotation.Nullable;

public final class TextStyle {
  private final Font font;
  private final LineBreakPolicy lineBreakPolicy;

  /* TODO[cc]: Consider taking individual font parameters instead of a Font object, so the size can
               be supplied in a Points object. */
  public TextStyle(Font font, LineBreakPolicy lineBreakPolicy) {
    Preconditions.checkNotNull(font);
    Preconditions.checkNotNull(lineBreakPolicy);
    this.font = font;
    this.lineBreakPolicy = lineBreakPolicy;
  }

  @Override
  public int hashCode() {
    return Objects.hash(font, lineBreakPolicy);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof TextStyle))
      return false;
    TextStyle other = (TextStyle) obj;
    return this.font.equals(other.font)
        && this.lineBreakPolicy.equals(other.lineBreakPolicy);
  }

  private static final class TextPaintParameters {
    final Graphics2D g;
    final Coordinates pos;
    final Coordinates availableDimensions;
    final @Nullable Pixels baselineOffset;
    final Color color;
    final HorizontalAlignment horizontalAlignment;
    final boolean justified;

    TextPaintParameters(Graphics2D g, Coordinates pos, Coordinates availableDimensions,
        @Nullable Pixels baselineOffset, Color color, HorizontalAlignment horizontalAlignment,
        boolean justified)
    {
      Preconditions.checkNotNull(g);
      Preconditions.checkNotNull(pos);
      Preconditions.checkNotNull(availableDimensions);
      Preconditions.checkNotNull(color);
      Preconditions.checkNotNull(horizontalAlignment);
      this.g = g;
      this.pos = pos;
      this.availableDimensions = availableDimensions;
      this.baselineOffset = baselineOffset;
      this.color = color;
      this.horizontalAlignment = horizontalAlignment;
      this.justified = justified;
    }
  }

  // TODO[pd]: Consider fitting almost-fitting lines in the TextFitPolicy.OVERFLOW_LINE case.
  // TODO[cc]: Try simplifying the arguments to processText and ParagraphLayout.create.
  /**
   * @param lineLength can be null to indicate no break limit
   */
  private TextMeasurement processText(RenderContext rcx, Text text, @Nullable Pixels lineLength,
      boolean allowWider, boolean vertical, TextFitPolicy textFitPolicy,
      @Nullable TextPaintParameters paintParameters)
  {
    Preconditions.checkNotNull(rcx);
    Preconditions.checkNotNull(text);
    Preconditions.checkArgument(lineLength == null || !lineLength.isNegative());

    final boolean justified = (paintParameters == null) ? false : paintParameters.justified;
    final double useLineLength;
    final LineBreakPolicy useLineBreakPolicy;
    final boolean firstLineOnly;
    if (lineLength == null || textFitPolicy == TextFitPolicy.FIT_LINE) {
      useLineLength = Double.POSITIVE_INFINITY;
      useLineBreakPolicy = lineBreakPolicy;
      firstLineOnly = false;
    } else if (textFitPolicy == TextFitPolicy.OVERFLOW_LINE) {
      /* Continue processing the line a few characters beyond the clipping point. This will make for
      an abrupt visual cutoff in the middle of the last visible character on the line, which is a
      good signal to the user that this line is in fact overflowing. */
      useLineLength = lineLength.getDouble() + rcx.scaledFont(font).getSize2D() * 4;
      useLineBreakPolicy = LineBreakPolicy.CHARACTER_POLICY;
      firstLineOnly = true;
    } else {
      useLineLength = lineLength.getDouble();
      useLineBreakPolicy = lineBreakPolicy;
      firstLineOnly = false;
    }
    final double fitLength =
        (lineLength == null || allowWider || textFitPolicy == TextFitPolicy.OVERFLOW_LINE)
        ? Double.POSITIVE_INFINITY : lineLength.getDouble();
    final ParagraphLayout layout = ParagraphLayout.create(
        rcx, text, font, justified, useLineBreakPolicy, useLineLength, fitLength, firstLineOnly);

    return processLayout(rcx, vertical, paintParameters, layout);
  }

  private TextMeasurement processLayout(RenderContext rcx, boolean vertical,
      @Nullable TextPaintParameters paintParameters, ParagraphLayout layout)
  {
    Preconditions.checkNotNull(rcx);
    final double DEGREES = vertical ? 90.0 : 0.0;
    final double RADIANS = (DEGREES / 360.0) * 2.0 * Math.PI;

    AffineTransform transform = AffineTransform.getRotateInstance(RADIANS);
    // TODO[cc]: Find a simpler way to do this.
    Point2D.Double rp1 = new Point2D.Double();
    Point2D.Double rp2 = new Point2D.Double();
    Point2D.Double rp3 = new Point2D.Double();
    Point2D.Double rp4 = new Point2D.Double();
    transform.transform(new Point2D.Double(0.0              , 0.0               ), rp1);
    transform.transform(new Point2D.Double(0.0              , layout.getHeight()), rp2);
    transform.transform(new Point2D.Double(layout.getWidth(), layout.getHeight()), rp3);
    transform.transform(new Point2D.Double(layout.getWidth(), 0.0               ), rp4);
    Pixels minX = Pixels.min(rcx, rp1.x, rp2.x, rp3.x, rp4.x);
    Pixels minY = Pixels.min(rcx, rp1.y, rp2.y, rp3.y, rp4.y);
    Pixels maxX = Pixels.max(rcx, rp1.x, rp2.x, rp3.x, rp4.x);
    Pixels maxY = Pixels.max(rcx, rp1.y, rp2.y, rp3.y, rp4.y);
    // Since lines.width/height are already scaled, apply the scaling _after_ using them above.
    transform.scale(layout.getRequiredScaling(), layout.getRequiredScaling());

    Pixels retWidth  = maxX.minus(minX);
    Pixels retHeight = maxY.minus(minY);
    if (paintParameters != null) {
      Graphics2D g = paintParameters.g;
      AffineTransform oldTransform = g.getTransform();
      g.translate(paintParameters.pos.getX().getDouble(), paintParameters.pos.getY().getDouble());
      // Horizontally align the paragraph's effective bounding box after rotation.
      g.translate(getHorizontalAlignmentOffset(paintParameters.horizontalAlignment,
          paintParameters.availableDimensions.getX().getDouble(), retWidth.getDouble()), 0);

      // Baseline alignment.
      {
        /** The amount of excess vertical space available for baseline adjustment. */
        final double extraSpaceY =
            paintParameters.availableDimensions.getY().getDouble() - layout.getHeight();
        double baselineAdjustment = 0.0;
        if (!vertical && paintParameters.baselineOffset != null) {
          final double baselineOffset = paintParameters.baselineOffset.getDouble();
          boolean noSpace = false;
          /* Prefer aligning on the first line of a paragraph. An alternative approach would be to
          try to minimize the baseline adjustment by traversing layoutBaselineOffets in reverse
          order. After some visual experiments, I prefered the former one. */
          for (double layoutBaselineOffset : layout.getBaselineOffsets()) {
            double baselineAdjustmentCandidate = baselineOffset - layoutBaselineOffset;
            if (baselineAdjustmentCandidate >= 0) {
              if (baselineAdjustmentCandidate <= extraSpaceY) {
                baselineAdjustment = baselineAdjustmentCandidate;
                noSpace = false;
                break;
              } else {
                noSpace = true;
              }
            }
          }
          /* If there is not enough space for proper baseline alignment, center the text vertically
          instead. To test this, put a single line of a large font next to two lines of a small
          font. */
          if (noSpace)
            baselineAdjustment = Pixels.create(rcx, extraSpaceY).divideBy(rcx, 2.0).getDouble();
        }
        g.translate(0, baselineAdjustment);
      }
      g.translate(-minX.getDouble(), -minY.getDouble());
      g.transform(transform);
      g.setColor(rcx.safeColor(paintParameters.color));
      double curY = 0.0f;
      /* Use layout.width rather than lineLength here, as the inner bounding box should not change
      as a result of the alignment. */
      final double rightMargin = layout.getWidth() / layout.getRequiredScaling();
      for (TextLayout line : layout.getLines()) {
        curY += line.getAscent();
        /* For horizontal or near-horizontal paragraphs, align individual lines in the paragraph.
        Contrast with the alignment of the entire paragraph box that was done earlier, see above. */
        final double xOffset = (DEGREES < 45)
            ? getHorizontalAlignmentOffset(
                paintParameters.horizontalAlignment, rightMargin, line.getVisibleAdvance())
            : 0.0;
        line.draw(g, (float) xOffset, (float) curY);
        curY += line.getDescent() + line.getLeading();
      }
      g.setTransform(oldTransform);
    }
    final Pixels firstBaselineOffset = (vertical || layout.getBaselineOffsets().isEmpty())
        ? Pixels.ZERO
        : Pixels.create(rcx, layout.getBaselineOffsets().get(0));
    return new TextMeasurement(new Coordinates(retWidth, retHeight), firstBaselineOffset);
  }

  private static double getHorizontalAlignmentOffset(
      HorizontalAlignment horizontalAlignment, double rightMargin, double actualWidth)
  {
    if (horizontalAlignment.equals(HorizontalAlignment.RIGHT)) {
      return rightMargin - actualWidth;
    } else if (horizontalAlignment.equals(HorizontalAlignment.CENTER)) {
      return (rightMargin - actualWidth) / 2.0;
    } else {
      return 0.0;
    }
  }

  private Pixels getLinePadding(RenderContext rcx) {
    /* TODO[pd]: Avoid recomputing this every time. (Here and another call to scaledFont in
                 processText.) */
    return Pixels.create(rcx,
        rcx.scaledFont(font).getLineMetrics("x", rcx.getFontRenderContext()).getDescent());
  }

  // TODO[cc]: Get the padding logic out of this class, or clean it up.
  private Coordinates paddingOff(RenderContext rcx, boolean vertical) {
    /* Padding is needed only in the text flow direction. I considered making the padding a constant
    number of points to better align the beginning of lines in vertical stacks of text, but it looks
    better to have the padding follow the text size (otherwise cells look too cramped at large font
    sizes). */
    if (vertical) {
      return new Coordinates(Pixels.ZERO, getLinePadding(rcx));
    } else {
      return new Coordinates(getLinePadding(rcx), Pixels.ZERO);
    }
  }

  private Coordinates paddingDim(RenderContext rcx, boolean vertical) {
    if (vertical) {
      return new Coordinates(Pixels.ZERO, getLinePadding(rcx).times(2));
    } else {
      return new Coordinates(getLinePadding(rcx).times(2), Pixels.ZERO);
    }
  }

  /**
   * Paint the provided text in the style defined by this {@code TextStyle} object and other
   * parameters. Some characters may extend outside the specified rectangle; the client should set
   * an appropriate clipping area on the supplied canvas before calling this method.
   */
  public void paint(Canvas canvas, Text text, PixelRectangle rectangle,
      @Nullable Pixels baselineOffset, boolean vertical, TextFitPolicy textFitPolicy, Color color,
      HorizontalAlignment horizontalAlignment, boolean justified)
  {
    Preconditions.checkNotNull(canvas);
    Preconditions.checkNotNull(text);
    Preconditions.checkNotNull(rectangle);
    Preconditions.checkNotNull(color);

    RenderContext rcx = canvas.getRenderContext();
    Coordinates paddingDimensions = paddingDim(rcx, vertical);
    Coordinates availableDimensions =
        rectangle.getDimensions().minus(paddingDimensions).max(Coordinates.ZERO);
    Pixels useLength = vertical ? availableDimensions.getY() : availableDimensions.getX();
    Coordinates usePos = rectangle.getPosition().plus(paddingOff(rcx, vertical));
    TextPaintParameters param = new TextPaintParameters(canvas.getGraphics(), usePos,
        availableDimensions, baselineOffset, color, horizontalAlignment, justified);
    /* It would make sense to assert that the actual dimensions are within the supplied dimensions,
    but in practice the measured dimensions may be slightly larger. In the horizontal direction, a
    larger break width may actually lead to an even (slightly) wider text layout. The assertion
    would also fail in the vertical direction in the case of extremely small font sizes. */
    Coordinates actualDimensions = processText(
        canvas.getRenderContext(), text, useLength, false, vertical, textFitPolicy, param)
        .getDimensions()
        .plus(paddingDimensions);
  }

  public Pixels getHorizontalLineBreakBefore(
      RenderContext rcx, Text text, Pixels before, Coordinates dimensions)
  {
    Pixels useLength = Pixels.ZERO.max(dimensions.getX().minus(paddingDim(rcx, false).getX()));
    /* TODO[pd]: Take the logic in processText into account here. The best thing to do is probably
                 to store the ParagraphLayout, or the parameters required to create it, in the
                 TextCell, and move this logic there. */
    // For horizontal lines, there is no vertical padding, so no compensation for this is necessary.
    ParagraphLayout lines = ParagraphLayout.create(rcx, text, font, false, lineBreakPolicy,
        useLength.getDouble(), useLength.getDouble(), false);
    double curY = 0.0f;
    double ret  = 0.0f;
    /* TODO[cc]: Consider consolidating with the logic that computes
                 ParagraphLayout.baselineOffsets. */
    final double scaling = lines.getRequiredScaling();
    for (TextLayout line : lines.getLines()) {
      curY += scaling * line.getAscent();
      // curY is now at baseline.
      curY += scaling * (line.getDescent() + line.getLeading());
      if (curY < before.getDouble())
        ret = curY;
    }
    return Pixels.create(rcx, ret);
  }

  /**
   * @param prefMaxWidth can be null to indicate no break limit. The indicated width is
   *        horizontal regardless of the value of {@code vertical}.
   * @param allowWider is ignored if {@code vertical} is true
   */
  public TextMeasurement measure(RenderContext rcx, Text text, @Nullable Pixels prefMaxWidth,
      boolean allowWider, boolean vertical)
  {
    TextMeasurement ret;
    if (vertical && prefMaxWidth != null) {
      // Optimize for least vertical space while keeping horizontal width within prefMaxWidth.
      long largestRejectedLineWidth = 0;
      long smallestAcceptableLineWidth = (long) Math.ceil(ParagraphLayout.create(
          rcx, text, font, false, lineBreakPolicy,
          Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY, false)
          .getWidth());
      while (Math.abs(largestRejectedLineWidth - smallestAcceptableLineWidth) > 1) {
        long candidateWidth = (largestRejectedLineWidth + smallestAcceptableLineWidth) / 2;
        TextMeasurement res = processText(rcx, text, Pixels.create(candidateWidth), true, vertical,
            TextFitPolicy.PARAGRAPH, null);
        if (res.getDimensions().getX().isGreaterThan(prefMaxWidth)) {
          largestRejectedLineWidth = candidateWidth;
        } else {
          smallestAcceptableLineWidth = candidateWidth;
        }
      }
      ret = processText(rcx, text, Pixels.create(smallestAcceptableLineWidth), true, vertical,
          TextFitPolicy.PARAGRAPH, null);
    } else {
      Pixels lineLength = (prefMaxWidth == null || vertical) ? null :
          Pixels.ZERO.max(prefMaxWidth.minus(paddingDim(rcx, vertical).getX()));
      ret = processText(
          rcx, text, lineLength, vertical || allowWider, vertical, TextFitPolicy.PARAGRAPH, null);
    }
    ret = new TextMeasurement(
        ret.getDimensions().plus(paddingDim(rcx, vertical)),
        ret.getFirstBaselineOffset().plus(paddingOff(rcx, vertical).getY()));
    return ret;
  }
}
