// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.util.Collection;
import java.util.List;
import com.sieuferd.cell.styles.LineStyle;
import com.sieuferd.cell.styles.PathHighlights;
import com.sieuferd.cell.support.CellPaintParameters.Layer;
import java.util.EnumSet;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

// TODO[cc]: Consider renaming to "Box".
@Immutable
public abstract class Cell {
  public static final boolean DEBUG_PAGINATION = false;
  // TODO[cc]: Get rid of this.
  public static final boolean DEBUG_CURSOR = false;

  /** Only for assertion checking. */
  private final RenderContext renderContext;
  /** For debugging use only. */
  public volatile String debugIdentifier;

  protected Cell(RenderContext renderContext) {
    Preconditions.checkNotNull(renderContext);
    this.renderContext = renderContext;
  }

  public final void printDebug(String msg) {
    if (debugIdentifier != null)
      System.out.println(debugIdentifier + ": " + msg);
  }

  /* Because the superclass constructor may need to be called before subclasses can calculate the
  dimensions, have subclasses implement getDimensions(). */
  public abstract Coordinates getDimensions();

  public abstract Coordinates getBottomRightOverprintLimit(RenderContext renderContext);

  public final Coordinates getDimensionsWithOverprint(RenderContext renderContext) {
    checkRenderContext(renderContext);
    return getDimensions().plus(getBottomRightOverprintLimit(renderContext));
  }

  /** For assertion checking only. */
  public abstract boolean containsBiasedLocation();

  public final void checkRenderContext(RenderContext renderContext) {
    if (!this.renderContext.equals(renderContext))
      throw new IllegalArgumentException("Mismatched RenderContext");
  }

  public void paint(Canvas canvas, Coordinates position, Selection selection,
      PathHighlights pathHighlights, @Nullable Coordinates frozenHeaderPoint, FocusMode focusMode,
      boolean cursorPainted)
  {
    checkRenderContext(canvas.getRenderContext());
    if (frozenHeaderPoint == null && !cursorPainted) {
      // The BACKGROUND and TOP layers can be painted in one traversal in this case.
      traverse(CellParam.createPaint(
          canvas, selection, pathHighlights, focusMode, EnumSet.of(Layer.BACKGROUND, Layer.TOP)),
          position, Coordinates.ZERO, null, Coordinates.ZERO, false, null);
    } else {
      for (Layer layer : Layer.values()) {
        if (layer == Layer.CURSOR && !cursorPainted)
          continue;
        traverse(CellParam.createPaint(
            canvas, selection, pathHighlights, focusMode, ImmutableSet.of(layer)),
            position, Coordinates.ZERO, null,
            frozenHeaderPoint == null ? Coordinates.ZERO : frozenHeaderPoint,
            frozenHeaderPoint != null, null);
        if (DEBUG_CURSOR && layer == Layer.BACKGROUND) {
          Graphics2D g = canvas.getGraphics();
          g.setColor(new Color(255, 255, 255, 128));
          Rectangle bounds = g.getClipBounds();
          if (bounds != null)
            g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);
        }
      }
    }

    if (DEBUG_CURSOR) {
      Graphics2D g = canvas.getGraphics();
      PixelRectangle bounds = canvas.getClipBounds();
      PixelRectangle rect = findRectangles(
          renderContext, selection.getCursorLocation(), bounds, null).getCursorRectangle();
      if (rect != null) {
        g.setColor(new Color(255, 0, 0, 255));
        g.draw(rect.getRectangle2D());
      }
    }
    if (DEBUG_PAGINATION) {
      final LineStyle PAGE_BREAKS = new LineStyle(new Color(0, 0, 255, 255), new Points(1.0));
      Pixels lastY = Pixels.ZERO;
      for (Pixels curY : paginate(canvas.getRenderContext(),
          Pixels.create(canvas.getRenderContext(), new Points(8 * 72.0))))
      {
        PAGE_BREAKS.paint(canvas, false, position.plus(new Coordinates(Pixels.ZERO, curY)),
            Pixels.create(canvas.getRenderContext(),
            canvas.getGraphics().getClipBounds().getWidth()));
        lastY = curY;
      }
      final LineStyle LAST_PAGE_BREAK = new LineStyle(new Color(255, 0, 0, 255), new Points(1.0));
      LAST_PAGE_BREAK.paint(canvas, false, position.plus(new Coordinates(Pixels.ZERO, lastY)),
          Pixels.create(canvas.getRenderContext(),
          canvas.getGraphics().getClipBounds().getWidth()));
    }
  }

  public Collection<PixelRectangle> findDirtyRegions(RenderContext renderContext,
      PixelRectangle clipBounds, Selection selection)
  {
    CellParam param = CellParam.createFindRectangles(
        clipBounds, renderContext, selection, PathHighlights.NONE);
    traverse(param, Coordinates.ZERO, Coordinates.ZERO, null, Coordinates.ZERO, false, null);
    return param.getCellRectangles().getRepaintRecangles();
  }

  public final @Nullable Location getClosestCursorLocationAt(RenderContext renderContext,
      Coordinates position, Direction cursorBias, @Nullable Location avoid,
      @Nullable Coordinates frozenHeaderPoint)
  {
    Preconditions.checkNotNull(renderContext);
    Preconditions.checkNotNull(position);
    Preconditions.checkNotNull(cursorBias);
    /* Shift the coordinates system of the cell traversal such that the cursor search position is at
    coordinates (0,0); see the Javadoc for HitParameters.setResult. */
    final Coordinates shiftedPosition = Coordinates.ZERO.minus(position);
    final Coordinates shiftedHeaderPoint =
        (frozenHeaderPoint == null ? Coordinates.ZERO : frozenHeaderPoint).plus(shiftedPosition);
    final CellParam cellParam = CellParam.createHit(
        new PixelRectangle(shiftedPosition, getDimensionsWithOverprint(renderContext)),
        renderContext, avoid);
    final HitParameters hitParam = new HitParameters(cursorBias);
    traverse(cellParam, shiftedPosition, Coordinates.ZERO, null, shiftedHeaderPoint,
        frozenHeaderPoint != null, hitParam);
    BiasedLocation ret = hitParam.getResult();
    return ret == null ? null : ret.getCursorLocation();
  }

  /**
   * @param clipBounds restricts the search space to the specified bounds; if null or very large,
   *        the search may be expensive
   */
  public CellRectangles findRectangles(RenderContext renderContext,
      @Nullable Location cursorLocation, @Nullable PixelRectangle clipBounds,
      @Nullable Coordinates frozenHeaderPoint)
  {
    if (cursorLocation == null)
      return new CellRectangles();
    final Selection selection = new Selection(cursorLocation, ImmutableSet.of());
    final PixelRectangle useClipBounds = (clipBounds != null) ? clipBounds :
        new PixelRectangle(new Rectangle(0, 0, Integer.MAX_VALUE, Integer.MAX_VALUE));
    final CellParam param = CellParam.createFindRectangles(
        useClipBounds, renderContext, selection, PathHighlights.NONE);
    traverse(param, Coordinates.ZERO, Coordinates.ZERO, null,
        frozenHeaderPoint == null ? Coordinates.ZERO : frozenHeaderPoint, frozenHeaderPoint != null,
        null);
    final CellRectangles ret = param.getCellRectangles();
    if (ret == null)
      throw new AssertionError();
    return ret;
  }

  /** Returns a list of offsets, each corresponding to the end of a printed page. */
  public final ImmutableList<Pixels> paginate(RenderContext renderContext, Pixels pageHeight) {
    checkRenderContext(renderContext);
    List<Pixels> ret = Lists.newArrayList();
    Pixels curY = Pixels.ZERO;
    while (getDimensions().getY().isGreaterThan(curY)) {
      Pixels h = getPaginatedHeight(renderContext, curY, pageHeight, pageHeight).min(pageHeight);
      Preconditions.checkState(h.isGreaterThan(Pixels.ZERO));
      curY = curY.plus(h);
      ret.add(curY);
    }
    return ImmutableList.copyOf(ret);
  }

  // TODO[cc]: Consider hiding.
  // TODO[cc]: Consider changing justifiedDimensions spec to use null to mean no justification.
  /* TODO[cc]: Consider changing headerPoint/overlayMode back to just a nullable frozenHeaderPoint,
               if I don't end up using header points outside of overlayMode. */
  /**
   * @param justifiedDimensions minimum size for justification. If equal to
   *        Coordinates.ZERO, do not justify.
   * @param overlayMode if true, painting and rectangle finding will be done for frozen headers
   *        and values instead of for the regular static layout.
   * @param hitParameters if set, {@code param.getSelection().getCursorLocation()} should be null
   */
  public abstract void traverse(CellParam param, Coordinates position,
      Coordinates justifiedDimensions, @Nullable Pixels baselineOffset,
      Coordinates headerPoint, boolean overlayMode, @Nullable HitParameters hitParameters);

  // TODO[cc]: Consider hiding.
  public abstract Pixels getPaginatedHeight(RenderContext renderContext,
      Pixels startOffset, Pixels preferredMax, Pixels paperHeight);

  public abstract Pixels getMinimumBaselineOffset(RenderContext renderContext);

  public static enum FocusMode {
    FOCUSED,
    UNFOCUSED,
    REFERENCE_SELECTION
  }
}
