// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.sieuferd.cell.impl.BorderCell;
import com.sieuferd.cell.impl.HeaderedCell;
import java.util.Collection;
import java.util.List;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import javax.annotation.concurrent.NotThreadSafe;

/**
 * Collection of data found during a rectangle-finding cell traversal.
 *
 * <p>The term <em>header</em> refers to headers defined by a {@link HeaderedCell}, not frozen
 * values facilitated by {@link BorderCell}.
 */
@NotThreadSafe
public class CellRectangles {
  private final List<PixelRectangle> repaintRecangles = Lists.newArrayList();
  private final List<HeaderRectangle> headerRectangles = Lists.newArrayList();
  private @Nullable PixelRectangle cursorRectangle;
  private boolean cursorInOverlayHeader;

  /**
   * Get the rectangles guaranteed to cover the painted areas of selected or highlighted cells.
   */
  public Collection<PixelRectangle> getRepaintRecangles() {
    return ImmutableList.copyOf(repaintRecangles);
  }

  /**
   * Get data about all header rectangles that are ancestors of the found cursor rectangle.
   */
  public Collection<HeaderRectangle> getHeaderRectangles() {
    return ImmutableList.copyOf(headerRectangles);
  }

  public void addRepaintRectangle(PixelRectangle rect) {
    repaintRecangles.add(rect);
  }

  public void addHeaderRectangle(HeaderRectangle headerRectangle) {
    headerRectangles.add(headerRectangle);
  }

  public void setCursorRectangle(PixelRectangle cursorRectangle) {
    this.cursorRectangle = cursorRectangle;
  }

  public @Nullable PixelRectangle getCursorRectangle() {
    return cursorRectangle;
  }

  public void markCursorInOverlayHeader() {
    cursorInOverlayHeader = true;
  }

  /**
   * Indicate if the cursor was found in a frozen overlay header.
   */
  public boolean isCursorInOverlayHeader() {
    return cursorInOverlayHeader;
  }

  @Immutable
  public static final class HeaderRectangle {
    public final PixelRectangle rectangle;
    public final Direction direction;
    public final boolean overlay;

    public HeaderRectangle(PixelRectangle rectangle, Direction direction, boolean overlay) {
      Preconditions.checkNotNull(rectangle);
      Preconditions.checkNotNull(direction);
      this.rectangle = rectangle;
      this.direction = direction;
      this.overlay = overlay;
    }
  }
}
