// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import com.google.common.base.Preconditions;

// TODO[cc]: Change the flow/still terminology to major/minor, as in the Swing EditorKit/View API.
/* TODO[cc]: Generalize or remove methods that only apply in the orthogonal or RIGHT/DOWN case, or
             limit the class itself to these cases. */
public enum Direction {
  NONE      ( 0,  0),
  UP        ( 0, -1),
  RIGHT     (+1,  0),
  DOWN      ( 0, +1),
  LEFT      (-1,  0),
  UP_LEFT   (-1, -1),
  DOWN_RIGHT(+1, +1);

  /** -1, 0, or 1. */
  private final int horizontalBiasRight;
  /** -1, 0, or 1. */
  private final int verticalBiasDown;

  private Direction(int horizontalBiasRight, int verticalBiasDown) {
    this.horizontalBiasRight = horizontalBiasRight;
    this.verticalBiasDown = verticalBiasDown;
  }

  public static Direction fromSigns(int horizontalBiasRight, int verticalBiasDown) {
    Preconditions.checkArgument(horizontalBiasRight >= -1 && horizontalBiasRight <= 1);
    Preconditions.checkArgument(verticalBiasDown >= -1 && verticalBiasDown <= 1);
    for (Direction bias : Direction.values()) {
      if (bias.getHorizontalBiasRight() == horizontalBiasRight &&
          bias.getVerticalBiasDown() == verticalBiasDown)
      {
        return bias;
      }
    }
    throw new AssertionError("Expected all signs to be enumerated");
  }

  public Direction inverse() {
    return fromSigns(-getHorizontalBiasRight(), -getVerticalBiasDown());
  }

  /** @return -1, 0, or 1. */
  public int getHorizontalBiasRight() {
    return horizontalBiasRight;
  }

  /** @return -1, 0, or 1. */
  public int getVerticalBiasDown() {
    return verticalBiasDown;
  }

  private void checkOrthogonal() {
    if (this != UP && this != RIGHT && this != DOWN && this != LEFT)
      throw new IllegalStateException("Operation only applicable to orthogonal directions");
  }

  public boolean isVertical() {
    return verticalBiasDown != 0;
  }

  public boolean isHorizontal() {
    return horizontalBiasRight != 0;
  }

  public Pixels flowDir(Coordinates c) {
    checkOrthogonal();
    return (isHorizontal()) ? c.getX() : c.getY();
  }

  public Pixels stillDir(Coordinates c) {
    checkOrthogonal();
    return (isHorizontal()) ? c.getY() : c.getX();
  }

  public Coordinates addFlow(Coordinates point, Pixels offset) {
    checkOrthogonal();
    return new Coordinates(
        point.getX().plus(offset.times(getHorizontalBiasRight())),
        point.getY().plus(offset.times(getVerticalBiasDown())));
  }

  public Coordinates newCoordinates(Pixels flowDir, Pixels stillDir) {
    checkOrthogonal();
    return (isHorizontal())
        ? new Coordinates(flowDir, stillDir)
        : new Coordinates(stillDir, flowDir);
  }

  public Coordinates newCoordinates(Coordinates useFlowDir, Coordinates useStillDir) {
    checkOrthogonal();
    return newCoordinates(flowDir(useFlowDir), stillDir(useStillDir));
  }

  public Coordinates multiply(Coordinates coordinates) {
    return new Coordinates(
        coordinates.getX().times(horizontalBiasRight),
        coordinates.getY().times(verticalBiasDown));
  }

  public Coordinates getBiasPoint(RenderContext rcx, PixelRectangle onRectangle) {
    // TODO[pd]: Deal with the case of an empty rectangle, on which no point technically lies.
    // Ensure that the point lies on the rectangle in all cases.
    Coordinates openIntervalDimensions =
        onRectangle.getDimensions().minus(new Coordinates(Pixels.EPSILON, Pixels.EPSILON))
        .max(Coordinates.ZERO);
    return onRectangle.getPosition().plus(new Coordinates(
        openIntervalDimensions.getX().times(rcx, (getHorizontalBiasRight() + 1.0) / 2.0),
        openIntervalDimensions.getY().times(rcx, (getVerticalBiasDown()    + 1.0) / 2.0)));
  }
}
