// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.expressions.TupleConstant;
import java.util.List;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/**
 * A location in a layout where the cell cursor or a selection highlight may be present.
 *
 * <p>A cursor location at a table label for a primitive field is currently distinct from cursor
 * locations at corresponding primitive field labels in an outline view; this is intentional, since
 * there can't be a one-to-one mapping in this case.
 */
@Immutable
public final class Location {
  private final Data data;
  private final Path path;
  private final @Nullable RelationPath identifyingPath;
  private final Tuple identifyingTuple;
  private final boolean value;
  private final ImmutableList<TupleConstant> tupleConstraints;

  /** @param identifyingPath can be null when identifyingTuple is the root parent tuple */
  public Location(Data data, Path path, @Nullable RelationPath identifyingPath,
      Tuple identifyingTuple, boolean value, List<TupleConstant> tupleConstraints)
  {
    Preconditions.checkNotNull(data);
    Preconditions.checkNotNull(path);
    Preconditions.checkNotNull(identifyingTuple);
    Preconditions.checkNotNull(tupleConstraints);
    Preconditions.checkArgument(
        identifyingPath == null || identifyingPath.isEqualToOrAncestorOf(path));
    this.data = data;
    this.path = path;
    this.identifyingPath = identifyingPath;
    this.identifyingTuple = identifyingTuple;
    this.value = value;
    this.tupleConstraints = ImmutableList.copyOf(tupleConstraints);
    identifyingTuple.sanityCheck(data, identifyingPath);
  }

  public Data getData() {
    return data;
  }

  public Path getPath() {
    return path;
  }

  public @Nullable RelationPath getIdentifyingPath() {
    return identifyingPath;
  }

  public Tuple getIdentifyingTuple() {
    return identifyingTuple;
  }

  public boolean isValue() {
    return value;
  }

  public List<TupleConstant> getTupleConstraints() {
    return tupleConstraints;
  }

  /* TODO[pd]: Consider just storing the parent instead of having this convoluted logic. Or get
               rid of this somehow. */
  private Location getParent() {
    Preconditions.checkState(!path.isRoot());
    Preconditions.checkState(identifyingPath != null);
    final RelationPath newIdentifyingPath;
    final Tuple newIdentifyingTuple;
    if (path.equals(identifyingPath)) {
      newIdentifyingPath = identifyingPath.isRoot() ? null : identifyingPath.getParent();
      newIdentifyingTuple = data.getParentTuple(identifyingPath, identifyingTuple);
    } else {
      newIdentifyingPath = identifyingPath;
      newIdentifyingTuple = identifyingTuple;
    }
    return new Location(data, path.getParent(), newIdentifyingPath, newIdentifyingTuple,
        value, tupleConstraints);
  }

  public @Nullable Location getAncestorWithIdentifyingPath(
      @Nullable RelationPath targetIdentifyingPath)
  {
    if (Objects.equals(identifyingPath, targetIdentifyingPath))
      return this;
    if (identifyingPath == null || path.isRoot())
      return null;
    return getParent().getAncestorWithIdentifyingPath(targetIdentifyingPath);
  }

  public boolean matchesAt(Location other, @Nullable RelationPath targetIdentifyingPath) {
    Preconditions.checkNotNull(other);
    if (other.value != value || !other.data.equals(data))
      return false;
    final Location thisCL = this.getAncestorWithIdentifyingPath(targetIdentifyingPath);
    if (thisCL == null)
      return false;
    final Location otherCL = other.getAncestorWithIdentifyingPath(targetIdentifyingPath);
    if (otherCL == null)
      return false;
    return thisCL.identifyingTuple.equalsTuple(otherCL.identifyingTuple);
  }

  public boolean equalsLocation(Location other) {
    Preconditions.checkNotNull(other);
    if (!other.data.equals(data) || other.value != value || !other.path.equals(path) ||
        !Objects.equals(other.identifyingPath, identifyingPath) ||
        !other.tupleConstraints.equals(tupleConstraints))
    {
      return false;
    }
    return other.identifyingTuple.equalsTuple(identifyingTuple);
  }

  @Override
  public int hashCode() {
    return Objects.hash(
        data, value, path, identifyingPath, identifyingTuple, tupleConstraints);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof Location))
      return false;
    return equalsLocation((Location) obj);
  }

  // TODO[cc]: Avoid nullable cursor locations.
  public static @Nullable Path maybePath(@Nullable Location location) {
    return location == null ? null : location.getPath();
  }

  public static boolean equals(@Nullable Location cl1, @Nullable Location cl2) {
    if (cl1 == cl2) {
      return true;
    } else if (cl1 == null || cl2 == null) {
      return false;
    } else {
      return cl1.equalsLocation(cl2);
    }
  }

  @Override
  public String toString() {
    return (value ? "value" : "label") + " at " + path + " identified by " + identifyingTuple +
        (tupleConstraints.isEmpty() ? "" : (" with tuple constraints " + tupleConstraints));
  }
}
