// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import javax.annotation.concurrent.Immutable;

@Immutable
public final class PixelRectangle {
  private final Coordinates position;
  private final Coordinates dimensions;

  public PixelRectangle(Coordinates position, Coordinates dimensions) {
    Preconditions.checkNotNull(position);
    Preconditions.checkNotNull(dimensions);
    Preconditions.checkArgument(!dimensions.getX().isNegative());
    Preconditions.checkArgument(!dimensions.getY().isNegative());
    this.position = position;
    this.dimensions = dimensions;
  }

  public PixelRectangle(Rectangle rectangle) {
    this(new Coordinates(Pixels.create(rectangle.x    ), Pixels.create(rectangle.y     )),
         new Coordinates(Pixels.create(rectangle.width), Pixels.create(rectangle.height)));
  }

  @Override
  public String toString() {
    return position + "-" + position.plus(dimensions);
  }

  public Coordinates getPosition() {
    return position;
  }

  public Coordinates getDimensions() {
    return dimensions;
  }

  public Pixels getX() {
    return position.getX();
  }

  public Pixels getY() {
    return position.getY();
  }

  public Pixels getWidth() {
    return dimensions.getX();
  }

  public Pixels getHeight() {
    return dimensions.getY();
  }

  public boolean contains(Coordinates point) {
    if (position.getX().isGreaterThan(point.getX()))
      return false;
    if (position.getY().isGreaterThan(point.getY()))
      return false;
    if (!position.getX().plus(dimensions.getX()).isGreaterThan(point.getX()))
      return false;
    if (!position.getY().plus(dimensions.getY()).isGreaterThan(point.getY()))
      return false;
    return true;
  }

  public Rectangle2D getRectangle2D() {
    return new Rectangle2D.Double(getX().getDouble(), getY().getDouble(),
        getWidth().getDouble(), getHeight().getDouble());
  }

  public Rectangle getBoundingAWTRectangle() {
    Coordinates end = position.plus(dimensions);
    Pixels xRounded = position.getX().roundDown(Pixels.ONE);
    Pixels yRounded = position.getY().roundDown(Pixels.ONE);
    Pixels xEndRounded = end.getX().roundUp(Pixels.ONE);
    Pixels yEndRounded = end.getY().roundUp(Pixels.ONE);
    Pixels wRounded = xEndRounded.minus(xRounded);
    Pixels hRounded = yEndRounded.minus(yRounded);
    return new Rectangle(
        xRounded.getInteger(), yRounded.getInteger(), wRounded.getInteger(), hRounded.getInteger());
  }

  public boolean isEqualTo(PixelRectangle other) {
    return other.getPosition().isEqualTo(getPosition())
        && other.getDimensions().isEqualTo(getDimensions());
  }

  @Override
  public boolean equals(Object obj) {
    if (!(obj instanceof PixelRectangle))
      return false;
    if (obj == this)
      return true;
    return isEqualTo((PixelRectangle) obj);
  }

  @Override
  public int hashCode() {
    return Objects.hashCode(getPosition(), getDimensions());
  }
}
