// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/**
 * A distance measured in pixels. Values are stored and manipulated using fixed-point (exact)
 * arithmetic.
 */
@Immutable
public final class Pixels implements Comparable<Pixels> {
  /* This constant should be a power of two to ensure a one-to-one mapping with float/double.
  Subpixel rendering defects were observed when a power of ten was used and graphics scaling was in
  effect (though with a previous scaling approach). However, the constant should not be so high that
  its reciprocal would vanish if added to a float in the normal range. Setting the constant to
  1024*1024 caused a line breaking bug for this reason. The latter bug was observed before this
  codebase was changed from using float to using double, but parts of the Swing library may still
  rely on floats, so it's better to be safe. */
  private static final long DENOMINATOR = 512;
  public static final Pixels ZERO = Pixels.create(0);
  public static final Pixels ONE = Pixels.create(1);
  public static final Pixels HALF = new Pixels(DENOMINATOR / 2);
  /** The smallest increment supported by this scalar representation. The value can be assumed to be
  much smaller than a single whole pixel, but much larger than typical roundoff errors encountered
  with arithmetic operations on floats. */
  public static final Pixels EPSILON = new Pixels(1);
  private final long numerator;

  private Pixels(long numerator) {
    this.numerator = numerator;
  }

  public static Pixels create(double value) {
    long numerator = Math.round(value * DENOMINATOR);
    // Never convert non-zero values to zero, as this may cause division-by-zero errors.
    if (numerator == 0 && value != 0)
      numerator = (value > 0) ? 1 : -1;
    return new Pixels(numerator);
  }

  public static Pixels create(RenderContext rcx, double value) {
    if (!Double.isFinite(value))
      throw new ArithmeticException();
    if (value == 0) {
      return Pixels.ZERO;
    } else {
      Pixels ret = create(value);
      if (rcx.getRoundToPixels() != null)
        ret = ret.roundBiased(rcx.getRoundToPixels());
      return ret;
    }
  }

  public static Pixels create(RenderContext rcx, Points value) {
    return create(rcx, value.getValue() * rcx.getScaling());
  }

  public static Pixels create(long value) {
    return new Pixels(value * DENOMINATOR);
  }

  public double getDouble() {
    return numerator / (double) DENOMINATOR;
  }

  public int getInteger() {
    final double d = Math.round(getDouble());
    if (d > Integer.MAX_VALUE)
      return Integer.MAX_VALUE;
    if (d < Integer.MIN_VALUE)
      return Integer.MIN_VALUE;
    return (int) d;
  }

  /** Return the number of inches represented by this pixel value if one pixel is assumed to be
  rendered as one typographical point. This corresponds to the actual distance on paper resulting
  from using this pixel value with the Java Printing API. */
  public double toInchesAfterScaling() {
    return getDouble() / 72.0;
  }

  /** Return the number of typographical points represented by this pixel value in the specified
  RenderContext. Depending on rounding and the RenderContext supplied,
  {@code Pixels.create(rcx, new Point(someDouble)).toPoints(rcx).getValue()} will not necessarily
  be equal to the double {@code someDouble}, though they have the same units and conceptually
  represent the same distance. */
  public Points toPoints(RenderContext rcx) {
    return new Points(getDouble() / rcx.getScaling());
  }

  private Pixels roundDirected(RenderContext rcx, boolean up) {
    if (rcx.getRoundToPixels() == null) {
      return this;
    } else {
      return up ? roundUp(rcx.getRoundToPixels()) : roundDown(rcx.getRoundToPixels());
    }
  }

  public static Pixels max(RenderContext rcx, double oneValue, double... otherValues) {
    double ret = oneValue;
    for (Double value : otherValues)
      ret = Math.max(ret, value);
    return new Pixels((long) Math.ceil(ret * DENOMINATOR)).roundDirected(rcx, true);
  }

  public static Pixels min(RenderContext rcx, double oneValue, double... otherValues) {
    double ret = oneValue;
    for (Double value : otherValues)
      ret = Math.min(ret, value);
    return new Pixels((long) Math.floor(ret * DENOMINATOR)).roundDirected(rcx, false);
  }

  public Pixels roundUp(Pixels divisor) {
    long n = numerator;
    long d = divisor.numerator;
    return (n % d == 0) ? this : new Pixels((n / d + 1) * d);
  }

  public Pixels roundDown(Pixels divisor) {
    long n = numerator;
    long d = divisor.numerator;
    return (n % d == 0) ? this : new Pixels((n / d) * d);
  }

  public Pixels round(Pixels divisor) {
    long n = numerator;
    long d = divisor.numerator;
    return (n % d == 0) ? this : new Pixels(n > 0
        ? ((n + d / 2) / d) * d
        : -(((-n) + d / 2) / d) * d);
  }

  public Pixels roundBiased(Pixels divisor) {
    Pixels ret = round(divisor);
    if (ret.isZero() && !isZero())
      ret = isNegative() ? divisor.times(-1) : divisor;
    return ret;
  }

  public Pixels max(@Nullable Pixels other) {
    return other == null || this.isGreaterThan(other) ? this : other;
  }

  public Pixels min(@Nullable Pixels other) {
    return other != null && this.isGreaterThan(other) ? other : this;
  }

  public boolean isZero() {
    return numerator == 0;
  }

  public static Pixels sum(Pixels... terms) {
    long ret = 0;
    for (Pixels term : terms)
      ret += term.numerator;
    return new Pixels(ret);
  }

  public Pixels plus(Pixels term) {
    return new Pixels(numerator + term.numerator);
  }

  public Pixels times(int factor) {
    return new Pixels(numerator * factor);
  }

  public Pixels times(RenderContext rcx, double factor) {
    if (!Double.isFinite(factor))
      throw new ArithmeticException();
    Pixels ret = new Pixels(Math.round(numerator * factor));
    if (rcx.getRoundToPixels() != null)
      ret = ret.round(rcx.getRoundToPixels());
    return ret;
  }

  public Pixels minus(Pixels subtrahend) {
    return new Pixels(numerator - subtrahend.numerator);
  }

  public Pixels abs() {
    return new Pixels(Math.abs(numerator));
  }

  /** Guaranteed to round down to whatever precision is supported by the specified RenderContext. */
  public Pixels divideBy(RenderContext rcx, double divisor) {
    if (divisor == 0.0 || Double.isNaN(divisor))
      throw new ArithmeticException();
    return new Pixels((long) (numerator / divisor)).roundDirected(rcx, false);
  }

  public boolean isEqualTo(Pixels other) {
    return numerator == other.numerator;
  }

  public boolean isNegative() {
    return ZERO.isGreaterThan(this);
  }

  /* TODO[cc]: Get rid of this in favor of the idiom described in the JavaDoc for
               BigInteger.compareTo(). */
  public boolean isGreaterThan(Pixels other) {
    return compareTo(other) > 0;
  }

  @Override
  public int compareTo(Pixels other) {
    if        (numerator > other.numerator) {
      return 1;
    } else if (numerator < other.numerator) {
      return -1;
    } else {
      return 0;
    }
  }

  @Override
  public boolean equals(Object obj) {
    if (!(obj instanceof Pixels))
      return false;
    if (obj == this)
      return true;
    return isEqualTo((Pixels) obj);
  }

  @Override
  public int hashCode() {
    return ((Long) numerator).hashCode();
  }

  @Override
  public String toString() {
    return (numerator / (double) DENOMINATOR) + "px";
  }
}
