// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.cell.support;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import java.awt.Color;
import java.awt.Font;
import java.awt.font.FontRenderContext;
import javax.annotation.Nullable;

/** Assumed to be immutable in practice (fontRenderContext should not be changed externally). */
public final class RenderContext {
  private final FontRenderContext fontRenderContext;
  private final double scaling;
  private final @Nullable Pixels roundToPixels;
  private final boolean alphaPermitted;
  private final boolean paintUIShadows;

  public RenderContext(FontRenderContext fontRenderContext, double scaling,
      @Nullable Pixels roundToPixels, boolean alphaPermitted, boolean paintUIShadows)
  {
    Preconditions.checkNotNull(fontRenderContext);
    Preconditions.checkArgument(scaling > 0.0);
    Preconditions.checkArgument(Double.isFinite(scaling));
    this.fontRenderContext  = fontRenderContext;
    this.scaling            = scaling;
    this.roundToPixels      = roundToPixels;
    this.alphaPermitted     = alphaPermitted;
    this.paintUIShadows     = paintUIShadows;
  }

  public FontRenderContext getFontRenderContext() {
    return fontRenderContext;
  }

  public double getScaling() {
    return scaling;
  }

  public @Nullable Pixels getRoundToPixels() {
    return roundToPixels;
  }

  public boolean isAlphaPermitted() {
    return alphaPermitted;
  }

  /**
   * Indicate if UI shadows be painted. Such shadows are considered part of the on-screen user
   * interface rather than part of the cell layout, and should look consistent across the overall
   * Swing GUI. Thus, their physical dimensions are <em>not</em> affected by the {@code scaling}
   * property. Furthermore, such shadows should not be included for instance during painting for
   * print.
   */
  public boolean getPaintUIShadows() {
    return paintUIShadows;
  }

  @Override
  @SuppressWarnings("StringConcatenationInsideStringBufferAppend")
  public String toString() {
    StringBuilder ret = new StringBuilder();
    ret.append("RenderContext\n");
    ret.append("  FontRenderContext\n");
    ret.append("    antiAliasingHint     : " + fontRenderContext.getAntiAliasingHint() + "\n");
    ret.append("    fractionalMetricsHint: " + fontRenderContext.getFractionalMetricsHint() + "\n");
    ret.append("    transform            : " + fontRenderContext.getTransform() + "\n");
    ret.append("  scaling       : " + scaling + "\n");
    ret.append("  roundToPixels : " + roundToPixels + "\n");
    ret.append("  alphaPermitted: " + alphaPermitted + "\n");
    ret.append("  paintUIShadows: " + paintUIShadows + "\n");
    return ret.toString();
  }

  // TODO[pd]: Consider removing.
  public Color safeColor(Color mayHaveAlpha) {
    if (isAlphaPermitted() || mayHaveAlpha.getAlpha() == 255) {
      return mayHaveAlpha;
    } else {
      // TODO[pd]: Blend with white.
      return new Color(mayHaveAlpha.getRed(), mayHaveAlpha.getGreen(), mayHaveAlpha.getBlue());
    }
  }

  public Font scaledFont(Font font) {
    /* On Windows, fonts rendered on screen look dramatically better, in particular with respect to
    kerning and anti-aliasing, if the font size is an integral number of pixels. This rounding will
    be negated in cases where the text is scaled down to fit in a cell, but in this case the exact
    starting size is not essential anyway, so the rounding at this stage won't hurt. */
    Pixels scaledAndRoundedFontSize = Pixels.create(this, font.getSize2D() * getScaling());
    /* TODO[pd]: Make sure deriveFont is safe.
                 See https://netbeans.org/bugzilla/show_bug.cgi?id=49973 . */
    return font.deriveFont((float) scaledAndRoundedFontSize.getDouble());
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || obj.getClass() != RenderContext.class)
      return false;
    if (obj == this)
      return true;
    RenderContext rhs = (RenderContext) obj;
    return Objects.equal(getScaling()          , rhs.getScaling())
        && Objects.equal(getRoundToPixels()    , rhs.getRoundToPixels())
        && Objects.equal(isAlphaPermitted()    , rhs.isAlphaPermitted())
        && Objects.equal(getFontRenderContext(), rhs.getFontRenderContext());
  }

  @Override
  public int hashCode() {
    return Objects.hashCode(
        getScaling(), getRoundToPixels(), isAlphaPermitted(), getFontRenderContext());
  }
}
