// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.ObjectReference;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;

public abstract class MutableView extends ObservableView {
  /* The restriction that write transactions may not be entered from within read transactions allows
  EDTMutableView to run all read transactions from the calling thread. */
  /** Run the supplied Runnable synchonously with read/write transactional access to the View. The
  View is guaranteed not to be changed externally during the transaction. Transactions are
  reentrant, and read-only transactions may be entered from within a read/write transaction (but not
  the other way around). All view change events generated will be fired in batch after the outer
  transaction ends, when the view is in its final state. The {@code Runnable} is <em>not</em>
  guaranteed to be run on the same thread as the caller. Behavior and caveats are otherwise similar
  to those of {@link View#performReadTransaction(Runnable)}. */
  public abstract void performWriteTransaction(Runnable runnable);

  /** @throws IllegalArgumentException if value is null or of the wrong type */
  public abstract <T> void set(final ViewProperty<T> property, final Path path, final T value)
      throws IllegalArgumentException;

  /**
   * Remove a path from the View, including any child paths. The root path may not be removed.
   */
  public abstract void removePath(final Path path);
  /**
   * Add the specified new path. The path's parent must exist in the View. Default properties will
   * be set on the new path, but will not trigger individual change events.
   */
  public abstract void addPath(Path path);

  public final Path addPath(RelationPath parent, boolean relation) {
    Path ret = Path.create(parent, relation);
    addPath(ret);
    return ret;
  }

  public final RelationPath addRelationPath(RelationPath parent) {
    return addPath(parent, true).getRelation();
  }

  public final PrimitivePath addPrimitivePath(RelationPath parent) {
    return addPath(parent, false).getPrimitive();
  }

  public final void copyFrom(final View source) {
    copyFrom(source, Path.ROOT, Path.ROOT, null, null);
  }

  public final void copyFrom(final View source, final Path path,
      @Nullable final Set<ViewProperty<?>> properties, @Nullable final Set<Path> paths)
  {
    copyFrom(source, path, path, properties, paths);
  }

  public final PathMap copyFrom(final View source, final Path srcPath, final Path destPath) {
    return copyFrom(source, srcPath, destPath, null, null);
  }

  /* TODO[pd]: Consider automatically calling Views.rewritePathReferences after a copy operation on
               the same view with different source and destination paths. */
  /**
   * Copy applicable properties and child paths from a source View to this View.
   *
   * @param srcPath the source path at which to start copying properties
   * @param destPath the destination path at which to start copying properties
   * @param properties the properties to copy, or null to copy all properties
   * @param paths the source paths to include, or null to include all source paths
   */
  public final PathMap copyFrom(final View source, final Path srcPath, final Path destPath,
      Set<ViewProperty<?>> properties, Set<Path> paths)
  {
    Preconditions.checkArgument(source != this);
    /* Make immutable copies of arguments here, since the write transaction may happen in a
    different thread. */
    final ImmutableSet<ViewProperty<?>> useProperties =
        properties == null ? null : ImmutableSet.copyOf(properties);
    final ImmutableSet<Path> pathsWithAncestors = (paths == null) ? null :
        ImmutableSet.copyOf(Views.getPathsWithAncestors(paths));
    final ObjectReference<PathMap> ret = ObjectReference.create();
    performWriteTransaction(() -> {
      source.performReadTransaction(() -> {
        final Map<Path,Path> toPathMap = srcPath.equals(destPath) ? null : Maps.newLinkedHashMap();
        copyFromHelper(source, srcPath, destPath, toPathMap, useProperties, pathsWithAncestors);
        ret.setValue(toPathMap == null ? PathMap.EMPTY : new PathMap(toPathMap));
      });
    });
    return ret.getValue();
  }

  private void copyFromHelper(
      final View source, final Path srcPath, final Path destPath,
      final @Nullable Map<Path,Path> populatePathMap,
      final @Nullable ImmutableSet<ViewProperty<?>> properties,
      final @Nullable ImmutableSet<Path> pathsWithAncestors)
  {
    Preconditions.checkNotNull(source);
    Preconditions.checkNotNull(srcPath);
    Preconditions.checkNotNull(destPath);
    Preconditions.checkArgument(srcPath.isRelation() == destPath.isRelation());

    if (pathsWithAncestors != null && !pathsWithAncestors.contains(srcPath))
      return;
    if (populatePathMap != null && !srcPath.equals(destPath))
      MapUtil.putChecked(populatePathMap, srcPath, destPath);
    {
      final Set<ViewProperty<?>> definedProperties = source.getDefinedProperties(srcPath);
      if (properties == null) {
        for (ViewProperty<?> property : definedProperties) {
          if (!SProp.CHILD_PATHS.equals(property))
            setFrom(property, destPath, source, srcPath);
        }
      } else {
        for (ViewProperty<?> property : properties) {
          if (definedProperties.contains(property) && !SProp.CHILD_PATHS.equals(property))
            setFrom(property, destPath, source, srcPath);
        }
      }
    }
    if (srcPath.isRelation()) {
      final Set<Path> existingPaths = ImmutableSet.copyOf(getChildPaths(destPath.getRelation()));
      final PathList sourcePaths = source.getChildPaths(srcPath.getRelation());
      final boolean childPathsPropertyIncluded =
          properties == null || properties.contains(SProp.CHILD_PATHS);
      for (Path srcChildPath : sourcePaths) {
        final boolean processChild = (childPathsPropertyIncluded ||
            pathsWithAncestors == null || pathsWithAncestors.contains(srcChildPath));
        if (processChild) {
          final Path destChildPath;
          if (srcPath.equals(destPath)) {
            destChildPath = srcChildPath;
          } else {
            destChildPath = Path.create(destPath.getRelation(), srcChildPath.isRelation());
            destChildPath.updateDebugStringRepresentation(srcChildPath);
          }
          // Don't use View.exists, as we want to be sure the path is under the right parent.
          final boolean alreadyExists = existingPaths.contains(destChildPath);
          if (!alreadyExists && childPathsPropertyIncluded)
            addPath(destChildPath);
          if (alreadyExists || childPathsPropertyIncluded) {
            copyFromHelper(source, srcChildPath, destChildPath, populatePathMap,
                properties, pathsWithAncestors);
          }
        }
      }
      if (childPathsPropertyIncluded) {
        List<Path> newPathOrder = Lists.newArrayList(sourcePaths);
        for (Path superfluousPath :
            Sets.difference(existingPaths, ImmutableSet.copyOf(sourcePaths)))
        {
          removePath(superfluousPath);
        }
        if (srcPath.equals(destPath))
          set(SProp.CHILD_PATHS, destPath, new PathList(newPathOrder));
      }
    }
  }

  public final <T> void setFromDefault(ViewProperty<T> property, Path path) {
    if (property.appliesTo(path))
      set(property, path, getDefaults().get(property));
  }

  public final <T> void setFrom(ViewProperty<T> property, Path path, ViewPropertySource source) {
    if (property.appliesTo(path))
      set(property, path, source.get(property));
  }

  public final <T> void setFrom(ViewProperty<T> property, Path path, View source, Path sourcePath) {
    Preconditions.checkArgument(sourcePath.isRelation() == path.isRelation());
    if (property.appliesTo(sourcePath) && property.appliesTo(path))
      set(property, path, source.get(property, sourcePath));
  }
}
