// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.util.ObjectReference;
import com.sieuferd.util.StringUtil;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/** Immutable (except for a debugging string field) and opaque. */
@Immutable
public abstract class Path extends TreeStructure<Path,RelationPath,PrimitivePath> {
  public static final RelationPath ROOT = new RelationPath(null);
  /** Null for the root node */
  private final @Nullable RelationPath parent;
  /**
   * Since paths can be shared across different views, this representation should not include
   * information about parent paths, and should always be null for the root path. Used for debugging
   * purposes only.
   */
  private volatile @Nullable String debugStringRepresentation;

  Path(RelationPath parent) {
    this.parent = parent;
  }

  public RelationPath getParent() {
    if (isRoot())
      throw new IllegalStateException("Can't get the parent of the root path");
    return parent;
  }

  // TODO[cc]: Consider having these helper methods work on RelationPath only.

  public final boolean isRoot() {
    return Path.ROOT.equals(this);
  }

  public final boolean isParentOf(Path candidate) {
    Preconditions.checkNotNull(candidate);
    return isRelation() && !candidate.isRoot() && candidate.getParent().equals(this);
  }

  public final boolean isStrictAncestorOf(Path candidate) {
    Preconditions.checkNotNull(candidate);
    return candidate.isRoot() ? false : isEqualToOrAncestorOf(candidate.getParent());
  }

  public final boolean isEqualToOrAncestorOf(Path candidate) {
    Preconditions.checkNotNull(candidate);
    return equals(candidate) || isStrictAncestorOf(candidate);
  }

  // Uses referential equality.

  /**
   * Set a string representation to be returned by toString() based on any label stored in the
   * supplied view at the time when this method is called. Used for debugging purposes only.
   */
  public void updateDebugStringRepresentation(final View view) {
    if (isRoot())
      return;
    String label = StringUtil.trimmedAndEmptyIfNull(view.getOrNull(CProp.LABEL, Path.this));
    if (!label.isEmpty())
      debugStringRepresentation = label;
  }

  /**
   * Set the debug string representation for this path based on another path of the same type.
   */
  public void updateDebugStringRepresentation(Path path) {
    if (isRoot())
      return;
    Preconditions.checkArgument(isRelation() == path.isRelation());
    if (path.debugStringRepresentation != null)
      this.debugStringRepresentation = path.debugStringRepresentation;
  }

  public final String toString(final View view) {
    Preconditions.checkNotNull(view);
    final ObjectReference<String> ret = ObjectReference.create();
    view.performReadTransaction(() -> {
      String localLabel = view.exists(Path.this) && view.defines(CProp.LABEL, Path.this)
          ? view.get(CProp.LABEL, Path.this) : Path.this.toString();
      String prefix = isRoot() ? "@/" : getParent().toString(view);
      String postfix = isRelation() ? "/" : "";
      ret.setValue(prefix + localLabel + postfix);
    });
    return ret.getValue();
  }

  private String getGenericLabel() {
    return isRoot() ? "ROOT" :
        ((isRelation() ? "RP" : "PP") + ":" + Integer.toHexString(hashCode()));
  }

  // TODO[cc]: Consider using parent.toString() to return a recursive representation.
  @Override
  public String toString() {
    final String debugStringReprSnapshot = this.debugStringRepresentation;
    if (debugStringReprSnapshot != null)
      return debugStringReprSnapshot;
    return getGenericLabel();
  }

  @Override
  protected final Class<RelationPath> getRelationClass() {
    return RelationPath.class;
  }

  @Override
  protected final Class<PrimitivePath> getPrimitiveClass() {
    return PrimitivePath.class;
  }

  public static Path create(RelationPath parent, boolean relation) {
    return relation ? RelationPath.create(parent) : PrimitivePath.create(parent);
  }

  public static final class RelationPath extends Path {
    private RelationPath(RelationPath parent) {
      super(parent);
    }

    public static RelationPath create(RelationPath parent) {
      Preconditions.checkNotNull(parent);
      return new RelationPath(parent);
    }
  }

  public static final class PrimitivePath extends Path {
    private PrimitivePath(RelationPath parent) {
      super(parent);
    }

    public static PrimitivePath create(RelationPath parent) {
      Preconditions.checkNotNull(parent);
      return new PrimitivePath(parent);
    }
  }
}
