// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.Path.RelationPath;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.Locale;
import javax.annotation.concurrent.Immutable;

/* ExpressionContext is a better name, but ANTLR uses the word "context" to mean something else, and
in fact also generates a class called "ExpressionContext". */
/**
 * Context for parsing and generating string representations of property values
 * (STringREPresentationCOntext), including expressions.
 */
@Immutable
public class Strepco {
  public static final ImmutableSet<ViewProperty<?>> STREPCO_PROPERTIES =
        ImmutableSet.<ViewProperty<?>>builder()
        .add(CProp.LABEL)
        .add(CProp.VISIBLE)
        .build();

  private final ImmutableView view;
  private final RelationPath path;
  private final ZoneId timeZoneId;
  private final Locale locale;
  private final boolean strictParsing;

  public Strepco(
      View view, RelationPath path, ZoneId timeZoneId, Locale locale, boolean strictParsing)
  {
    view.checkPathExists(path);
    Preconditions.checkNotNull(timeZoneId);
    Preconditions.checkNotNull(locale);
    this.view = ImmutableView.copyOf(view, STREPCO_PROPERTIES, null);
    this.path = path;
    this.timeZoneId = timeZoneId;
    this.locale = locale;
    this.strictParsing = strictParsing;
  }

  public static Strepco createNormalized(View view) {
    return Strepco.createNormalized(view, Path.ROOT);
  }

  // TODO[cc]: Get rid of this. It is currently used only in tests.
  public static Strepco createNormalized(View view, RelationPath path) {
    return new Strepco(view, path, ZoneOffset.UTC, Locale.US, true);
  }

  public static Strepco create(View view, Path path) {
    /* Don't use path-level locales or time zones here; they would not be well-defined for relation
    paths, and formulas should probably use a consistent format across paths anyway. */
    return new Strepco(
        view, Views.getContextPath(path), ZoneId.systemDefault(), Locale.getDefault(), false);
  }

  public ImmutableView getView() {
    return view;
  }

  public RelationPath getPath() {
    return path;
  }

  public ZoneId getTimeZoneId() {
    return timeZoneId;
  }

  public Locale getLocale() {
    return locale;
  }

  public boolean isStrictParsing() {
    return strictParsing;
  }
}
