// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.Path.RelationPath;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

// TODO[cc]: Consider renaming to TupleHandle.
/**
 * Handle identifying a tuple at a particular {@link RelationPath} for a particular {@link Data}
 * instance.
 */
@Immutable
public final class Tuple {
  /** For assertion checking only. */
  private final Data data;
  // TODO[cc]: Consider exposing this. Then get rid of CursorLocation.identifyingPath.
  /** For assertion checking only. Is null for the root parent tuple. */
  private final @Nullable RelationPath identifyingPath;
  private final int tupleIndex;

  /**
   * @param identifyingPath is null for the root parent tuple, in which case tupleIndex must be 0;
   *        identifyingPath must be a plural path in data's view
   */
  private Tuple(Data data, @Nullable RelationPath identifyingPath, int tupleIndex) {
    Preconditions.checkNotNull(data);
    Preconditions.checkArgument(tupleIndex >= 0);
    Preconditions.checkArgument(identifyingPath != null || tupleIndex == 0);
    this.data = data;
    this.identifyingPath = identifyingPath;
    this.tupleIndex = tupleIndex;
  }

  public static Tuple create(Data data, RelationPath identifyingPath, int tupleIndex) {
    Preconditions.checkNotNull(identifyingPath);
    return new Tuple(data, identifyingPath, tupleIndex);
  }

  /* TODO[cc]: Avoid the nonsensical "root parent" terminology; call this the "root identifying"
               tuple or something like that. */
  public static Tuple createRootParent(Data data) {
    return new Tuple(data, null, 0);
  }

  // TODO[cc]: Always use in conjuction with getTupleIndex.
  /**
   * @param expectedIdentifyingPath can be null for the root parent tuple
   */
  public void sanityCheck(Data expectedData, @Nullable RelationPath expectedIdentifyingPath) {
    Preconditions.checkArgument(Objects.equals(data, expectedData));
    Preconditions.checkArgument(Objects.equals(identifyingPath, expectedIdentifyingPath));
  }

  public boolean equalsTuple(Tuple other) {
    sanityCheck(other.data, other.identifyingPath);
    return tupleIndex == other.tupleIndex;
  }

  @Override
  public int hashCode() {
    return Objects.hash(data, identifyingPath, tupleIndex);
  }

  /**
   * When possible, use {@link #equalsTuple(Tuple)} instead.
   */
  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof Tuple))
      return false;
    Tuple other = (Tuple) obj;
    if (!Objects.equals(this.data, other.data) ||
        !Objects.equals(this.identifyingPath, other.identifyingPath))
    {
      return false;
    }
    return equalsTuple(other);
  }

  public int getTupleIndex() {
    return tupleIndex;
  }

  @Override
  public String toString() {
    return identifyingPath == null ? "(root)" : (identifyingPath + "#" + tupleIndex);
  }
}
