// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.Path.RelationPath;
import java.util.Iterator;
import java.util.NoSuchElementException;

public abstract class TupleInterval implements Iterable<Tuple> {
  /**
   * Indicates whether this interval is complete at this level (non-recursive definition of
   * completeness).
   */
  public abstract boolean isComplete();
  public abstract int size();
  public abstract int getFirstTupleIndex();
  public abstract Tuple get(int index);

  public boolean isEmpty() {
    return size() == 0;
  }

  @Override
  public Iterator<Tuple> iterator() {
    return new Iterator<Tuple>() {
      private int i = 0;

      @Override
      public boolean hasNext() {
        return i < size();
      }

      @Override
      public Tuple next() {
        if (!hasNext())
          throw new NoSuchElementException();
        return get(i++);
      }

      @Override
      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
  }

  // TODO[cc]: Can I get rid of this now that the branch path concept is gone?
  public static final class SingletonTupleInterval extends TupleInterval {
    private final Tuple tuple;

    public SingletonTupleInterval(Tuple tuple) {
      Preconditions.checkNotNull(tuple);
      this.tuple = tuple;
    }

    @Override
    public boolean isComplete() {
      return true;
    }

    @Override
    public int getFirstTupleIndex() {
      return tuple.getTupleIndex();
    }

    @Override
    public int size() {
      return 1;
    }

    @Override
    public Tuple get(int index) {
      Preconditions.checkArgument(index == 0);
      return tuple;
    }
  }

  public static final class PluralTupleInterval extends TupleInterval {
    private final int firstTupleIndex;
    private final int size;
    private final boolean complete;
    private final Data data;
    private final RelationPath identifyingPath;

    public PluralTupleInterval(
        Data data, RelationPath identifyingPath, int firstTupleIndex, int size, boolean complete)
    {
      Preconditions.checkNotNull(data);
      Preconditions.checkNotNull(identifyingPath);
      Preconditions.checkArgument(firstTupleIndex >= 0);
      Preconditions.checkArgument(size >= 0);
      this.data = data;
      this.identifyingPath = identifyingPath;
      this.firstTupleIndex = firstTupleIndex;
      this.size = size;
      this.complete = complete;
    }

    @Override
    public boolean isComplete() {
      return complete;
    }

    @Override
    public int getFirstTupleIndex() {
      return firstTupleIndex;
    }

    @Override
    public int size() {
      return size;
    }

    @Override
    public Tuple get(int localIndex) {
      Preconditions.checkArgument(localIndex >= 0);
      Preconditions.checkArgument(localIndex < size());
      return Tuple.create(data, identifyingPath, firstTupleIndex + localIndex);
    }
  }
}
