// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.TreeMultimap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.SortedSet;
import javax.annotation.concurrent.Immutable;

@Immutable
public final class ViewDefaults implements ViewPropertySource {
  public static final ViewDefaults EMPTY = builder().build();

  private final ImmutableSortedMap<ViewProperty<?>,Object> defaults;
  private final ImmutableSetMultimap<ViewPropertySet,ViewProperty<?>> propertiesBySet;

  private ViewDefaults(Map<ViewProperty<?>,Object> defaults) {
    this.defaults = ImmutableSortedMap.copyOf(defaults);

    Multimap<ViewPropertySet,ViewProperty<?>> toPropertiesBySet = TreeMultimap.create();
    for (ViewProperty<?> prop : this.defaults.keySet())
      toPropertiesBySet.put(prop.getPropertySet(), prop);
    propertiesBySet = ImmutableSetMultimap.copyOf(toPropertiesBySet);
  }

  public ViewDefaults createSubset(Set<ViewProperty<?>> properties) {
    Map<ViewProperty<?>,Object> ret = Maps.newLinkedHashMap();
    for (ViewProperty<?> property : properties)
      ret.put(property, get(property));
    return new ViewDefaults(ret);
  }

  public SortedSet<ViewProperty<?>> getProperties() {
    return defaults.keySet();
  }

  public Multimap<ViewPropertySet,ViewProperty<?>> getSortedPropertiesBySet() {
    return propertiesBySet;
  }

  @Override
  public <T> T get(ViewProperty<T> property) {
    Preconditions.checkNotNull(property);
    @SuppressWarnings("unchecked")
    T ret = (T) defaults.get(property);
    if (ret == null)
      throw new NoSuchElementException("No property " + property + " defined as a default");
    return ret;
  }

  public void applyToUndefined(MutableView view) {
    applyToUndefined(view, Path.ROOT);
  }

  // TODO[cc]: Make this private.
  public <T> void apply(MutableView view, Path path, ViewProperty<T> viewProperty) {
    view.set(viewProperty, path, get(viewProperty));
  }

  public void applyToUndefined(MutableView view, Path path) {
    for (Entry<ViewProperty<?>,Object> entry : defaults.entrySet()) {
      if (entry.getKey().appliesTo(path) && !view.defines(entry.getKey(), path))
        apply(view, path, entry.getKey());
    }
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation()))
        applyToUndefined(view, childPath);
    }
  }

  public static Builder builder() {
    return new Builder();
  }

  public static final class Builder {
    private final Map<ViewProperty<?>,Object> defaults = Maps.newLinkedHashMap();

    private Builder() {
    }

    public <T> Builder set(ViewProperty<T> property, T value) {
      Preconditions.checkNotNull(property);
      Preconditions.checkNotNull(value);
      defaults.put(property, value);
      return this;
    }

    private <T> void setFromOther(ViewDefaults other, ViewProperty<T> prop) {
      set(prop, other.get(prop));
    }

    public Builder setAll(ViewDefaults other) {
      for (ViewProperty<?> prop : other.getProperties())
        setFromOther(other, prop);
      return this;
    }

    public ViewDefaults build() {
      return new ViewDefaults(defaults);
    }
  }
}
