// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import java.util.MissingResourceException;
import java.util.concurrent.atomic.AtomicInteger;
import javax.annotation.Nullable;
import org.openide.util.NbBundle;

/* I considered taking taking a Path type parameter as well in order to make the type system enforce
access to stylesheet properties using only the correct kinds of path nodes, but it complicates
client code with generics annotations, and it is not too useful in the context of the relaxed type
safety of the TreeStructure idiom (isRelation()/getRelation()/getPrimitive() rather than using a
visitor pattern). */
/** Immutable/constant. Uses referential equality. */
public final class ViewProperty<T> implements Comparable<ViewProperty<T>> {
  private static final String VIEW_PROPERTY_PREFIX = "viewproperty.";
  private static final AtomicInteger displayIndexCounter = new AtomicInteger();
  private final ViewPropertySet propertySet;
  private final String storageName;
  private final String displayName;
  private final AppliesTo appliesTo;
  private final ViewPropertyType<T> type;
  private final int displayIndex;
  private final boolean userReadOnly;

  @NbBundle.Messages({
    "ENU_AppliesTo_NONE=Applies to no fields.",
    "ENU_AppliesTo_ANY=Applies to all fields.",
    "ENU_AppliesTo_ROOT_ONLY=Applies to the root field only.",
    "ENU_AppliesTo_RELATION_ONLY=Applies to relation fields only.",
    "ENU_AppliesTo_PRIMITIVE_ONLY=Applies to primitive fields only."
  })
  /* TODO[cc]: Find a better name for this, e.g. PathType. Then move the presentation strings
               elsewhere (e.g. in FieldNode). */
  public static enum AppliesTo {
    NONE(Bundle.ENU_AppliesTo_NONE()),
    ANY(Bundle.ENU_AppliesTo_ANY()),
    ROOT_ONLY(Bundle.ENU_AppliesTo_ROOT_ONLY()),
    RELATION_ONLY(Bundle.ENU_AppliesTo_RELATION_ONLY()),
    PRIMITIVE_ONLY(Bundle.ENU_AppliesTo_PRIMITIVE_ONLY());

    private final String presentationString;

    private AppliesTo(String presentationString) {
      this.presentationString = presentationString;
    }

    public String getPresentationString() {
      return presentationString;
    }

    public boolean appliesTo(Path path) {
      Preconditions.checkNotNull(path);
      if (this == AppliesTo.NONE)
        return false;
      if (this == AppliesTo.PRIMITIVE_ONLY && path.isRelation())
        return false;
      if (this == AppliesTo.RELATION_ONLY && !path.isRelation())
        return false;
      if (this == AppliesTo.ROOT_ONLY && !path.isRoot())
        return false;
      return true;
    }
  }

  private ViewProperty(ViewPropertySet propertySet, String storageName,
      AppliesTo appliesTo, ViewPropertyType<T> type, boolean userReadOnly)
  {
    Preconditions.checkNotNull(propertySet);
    Preconditions.checkNotNull(storageName);
    Preconditions.checkNotNull(appliesTo);
    Preconditions.checkNotNull(type);
    this.displayIndex = displayIndexCounter.getAndIncrement();
    this.propertySet = propertySet;
    this.storageName = storageName;
    this.appliesTo = appliesTo;
    this.type = type;
    this.userReadOnly = userReadOnly;
    final String displayNameKey = "TIT_ViewProperty." + storageName;
    displayName = NbBundle.getMessage(propertySet.getClass(), displayNameKey);
    if (displayName.isEmpty()) {
      throw new MissingResourceException("Empty ViewProperty name",
          propertySet.getClass().getName(), displayNameKey);
    }
  }

  @Override
  public int compareTo(ViewProperty<T> other) {
    if (other == this)
      return 0;
    int setComp = getPropertySet().compareTo(other.getPropertySet());
    return setComp != 0 ? setComp : Integer.compare(displayIndex, other.displayIndex);
  }

  static <T> ViewProperty<T> create(ViewPropertySet propertySet,
      String storageName, AppliesTo appliesTo, ViewPropertyType<T> type, boolean userReadOnly)
  {
    return new ViewProperty<>(propertySet, storageName, appliesTo, type, userReadOnly);
  }

  public ViewPropertySet getPropertySet() {
    return propertySet;
  }

  public ViewPropertyType<T> getType() {
    return type;
  }

  public String getDisplayName() {
    return displayName;
  }

  public @Nullable String getShortDescription() {
    return getAppliesTo().getPresentationString();
  }

  public String getStorageName() {
    return storageName;
  }

  public String getBeanPropertyName() {
    return VIEW_PROPERTY_PREFIX + getStorageName();
  }

  public AppliesTo getAppliesTo() {
    return appliesTo;
  }

  @Override
  public String toString() {
    return getStorageName();
  }

  public boolean appliesTo(Path path) {
    return getAppliesTo().appliesTo(path);
  }

  public boolean isUserReadOnly() {
    return userReadOnly;
  }
}
