// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSortedSet;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.util.MapUtil;
import java.util.Set;
import java.util.SortedSet;
import javax.annotation.Nullable;

/** A set of view properties. Subclasses should follow the same pattern as CProp. */
public abstract class ViewPropertySet implements Comparable<ViewPropertySet> {
  private final String name;
  private final String displayName;
  private final String shortDescription;
  private final int displayIndex;
  /** For consistency checking only. */
  private final Set<String> storageName = Sets.newLinkedHashSet();
  private final Set<ViewProperty<?>> properties = Sets.newTreeSet();
  private @Nullable ImmutableSortedSet<ViewProperty<?>> immutablePropertiesCache;

  public ViewPropertySet(
      String name, String displayName, @Nullable String shortDescription, int displayIndex)
  {
    Preconditions.checkNotNull(name);
    Preconditions.checkNotNull(displayName);
    this.name = name;
    this.displayName = displayName;
    this.shortDescription = shortDescription;
    this.displayIndex = displayIndex;
  }

  @Override
  public final int compareTo(ViewPropertySet other) {
    if (other == this)
      return 0;
    return Integer.compare(displayIndex, other.displayIndex);
  }

  final synchronized void registerProperty(ViewProperty<?> prop) {
    Preconditions.checkArgument(prop.getPropertySet() == this);
    MapUtil.addChecked(storageName, prop.getStorageName());
    MapUtil.addChecked(properties, prop);
    immutablePropertiesCache = null;
  }

  protected static final <T> ViewProperty<T> addProperty(ViewPropertySet propertySet,
      String storageName, AppliesTo appliesTo, ViewPropertyType<T> type, boolean userReadOnly)
  {
    final ViewProperty<T> ret =
        ViewProperty.create(propertySet, storageName, appliesTo, type, userReadOnly);
    propertySet.registerProperty(ret);
    return ret;
  }

  /** Get the programmatic name of this property set. */
  public final String getName() {
    return name;
  }

  /** Get the human-readable name of this property set. */
  public final String getDisplayName() {
    return displayName;
  }

  public final @Nullable String getShortDescription() {
    return shortDescription;
  }

  @Override
  public final String toString() {
    return getName();
  }

  public final synchronized SortedSet<ViewProperty<?>> getViewProperties() {
    if (immutablePropertiesCache == null)
      immutablePropertiesCache = ImmutableSortedSet.copyOf(properties);
    return immutablePropertiesCache;
  }
}
