// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.collect.ImmutableBiMap;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.proptypes.OptionalBigDecimal;
import com.sieuferd.datamodel.proptypes.OptionalInteger;
import com.sieuferd.datamodel.proptypes.Position;
import com.sieuferd.util.MapUtil;
import java.math.BigDecimal;
import java.util.Locale;
import java.util.Map;

public final class ViewPropertyTypes {
  private ViewPropertyTypes() { }

  /* I considered using lookups to register view property types, but at least for now I don't want
  the "common" module to have a NetBeans dependency. */

  public interface EnumMapping {
    public String getStorageName();
  }

  public static <E extends Enum<E> & EnumMapping> ViewPropertyType<E>
      createEnumMappingType(Class<E> clazz)
  {
    final Map<String,E> ret = Maps.newLinkedHashMap();
    for (E value : clazz.getEnumConstants())
      MapUtil.putChecked(ret, value.getStorageName(), value);
    return createSimpleMappingType(clazz, ImmutableBiMap.copyOf(ret));
  }

  public static <T> ViewPropertyType<T> createNonpersistableType(Class<T> clazz) {
    return new ViewPropertyType<T>(clazz) { };
  }

  // TODO[cc]: Replace uses with createEnumMappingType where sensible.
  public static <T> ViewPropertyType<T> createSimpleMappingType(
      Class<T> clazz, final ImmutableBiMap<String,T> values)
  {
    return new ViewPropertyType<T>(clazz) {
      @Override
      public String valueToString(Strepco strepco, T value) {
        return values.inverse().get(value);
      }

      @Override
      public T stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
        T ret = values.get(str.trim().toLowerCase(Locale.ROOT));
        if (ret == null)
          throw new IllegalArgumentException("Expected one of " + values.keySet());
        return ret;
      }
    };
  }

  public static final ViewPropertyType<Integer> INTEGER =
      new ViewPropertyType<Integer>(Integer.class)
  {
    @Override
    public String valueToString(Strepco strepco, Integer value) {
      return Integer.toString(value);
    }

    @Override
    public Integer stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
      // Let any NumberFormatException propagate.
      return Integer.parseInt(str);
    }
  };

  public static final ViewPropertyType<Double> DOUBLE = new ViewPropertyType<Double>(Double.class) {
    @Override
    public String valueToString(Strepco strepco, Double value) {
      return Double.toString(value);
    }

    @Override
    public Double stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
      // Let any NumberFormatException propagate.
      return Double.parseDouble(str);
    }
  };

  public static final ViewPropertyType<BigDecimal> BIG_DECIMAL =
      new ViewPropertyType<BigDecimal>(BigDecimal.class)
  {
    @Override
    public String valueToString(Strepco strepco, BigDecimal value) {
      return value.toString();
    }

    @Override
    public BigDecimal stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
      // Let any NumberFormatException propagate.
      return new BigDecimal(str);
    }
  };

  public static final ViewPropertyType<Boolean> BOOLEAN =
      createSimpleMappingType(Boolean.class, ImmutableBiMap.of("true", true, "false", false));

  public static final ViewPropertyType<String> STRING = new ViewPropertyType<String>(String.class) {
    @Override
    public String valueToString(Strepco strepco, String value) {
      return value;
    }

    @Override
    public String stringToValue(Strepco strepco, String str) {
      return str;
    }
  };

  public static final ViewPropertyType<Position> POSITION =
      new ViewPropertyType<Position>(Position.class)
  {
    @Override
    public String valueToString(Strepco strepco, Position value) {
      BigDecimal position = value.getPosition();
      return (position == null) ? "" : position.toPlainString();
    }

    @Override
    public Position stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
      // Let any NumberFormatException propagate.
      return Position.create(str.trim().isEmpty() ? null : new BigDecimal(str));
    }
  };

  public static final ViewPropertyType<OptionalInteger> OPTIONAL_INTEGER =
      new ViewPropertyType<OptionalInteger>(OptionalInteger.class)
  {
    @Override
    public String valueToString(Strepco strepco, OptionalInteger value) {
      if (!value.isPresent())
        return "";
      return Integer.toString(value.get());
    }

    @Override
    public OptionalInteger stringToValue(Strepco strepco, String str)
        throws IllegalArgumentException
    {
      return str.trim().isEmpty() ?
          OptionalInteger.absent() : OptionalInteger.of(Integer.parseInt(str));
    }
  };

  public static final ViewPropertyType<OptionalBigDecimal> OPTIONAL_BIG_DECIMAL =
      new ViewPropertyType<OptionalBigDecimal>(OptionalBigDecimal.class)
  {
    @Override
    public String valueToString(Strepco strepco, OptionalBigDecimal value) {
      if (!value.isPresent())
        return "";
      return value.get().toString();
    }

    @Override
    public OptionalBigDecimal stringToValue(Strepco strepco, String str)
        throws IllegalArgumentException
    {
      return str.trim().isEmpty() ?
          OptionalBigDecimal.absent() : OptionalBigDecimal.of(new BigDecimal(str));
    }
  };
}
