// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.internal.ConcurrentMutableView;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import javax.annotation.Nullable;

// In the style of com.google.common.collect.Sets/Maps, etc.

public final class Views {
  private Views() {
  }

  /** Create a thread-safe, mutable View. */
  public static MutableView newMutable() {
    return new ConcurrentMutableView(ViewDefaults.EMPTY);
  }

  /** Create a thread-safe, mutable View with the given defaults. */
  public static MutableView newMutable(ViewDefaults defaults) {
    return new ConcurrentMutableView(defaults);
  }

  /** Create a thread-safe, mutable View copied from an existing view. */
  public static MutableView newMutable(View source) {
    return newMutable(source, ViewDefaults.EMPTY);
  }

  /** Create a thread-safe, mutable View copied from an existing view, with the given defaults. */
  public static MutableView newMutable(View source, ViewDefaults defaults) {
    ConcurrentMutableView ret = new ConcurrentMutableView(defaults);
    ret.copyFrom(source);
    return ret;
  }

  public static String toLongString(final View view) {
    final StringBuilder ret = new StringBuilder();
    view.performReadTransaction(() -> {
      toLongString(view, ret, Path.ROOT);
    });
    return ret.toString();
  }

  private static void toLongString(View view, StringBuilder ret, Path path) {
    ret.append(path.toString(view)).append("\n");
    for (ViewProperty<?> prop : view.getDefinedProperties(path))
      ret.append("  ").append(prop).append(": ").append(view.get(prop, path)).append("\n");
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation()))
        toLongString(view, ret, childPath);
    }
  }

  public static void rewritePathReferences(MutableView view, Path path, PathMap pathMap) {
    for (ViewProperty<?> property : view.getDefinedProperties(path))
      rewritePathReferences(view, path, pathMap, property);
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation()))
        rewritePathReferences(view, childPath, pathMap);
    }
  }

  private static <T> void rewritePathReferences(
      MutableView view, Path path, PathMap pathMap, ViewProperty<T> property)
  {
    view.set(property, path,
        property.getType().rewritePathReferences(pathMap, view.get(property, path)));
  }

  /* TODO[cc]: Now that a View object is no longer required in order to get the parent of a path,
               consider moving the path-related utility methods below to a different utility
               class. */

  private static Iterable<Path> traversePathToRoot(final Path path) {
    Preconditions.checkNotNull(path);
    return () -> new Iterator<Path>() {
      private @Nullable Path currentPath = path;

      @Override
      public boolean hasNext() {
        return currentPath != null;
      }

      @Override
      public Path next() {
        if (currentPath == null)
          throw new NoSuchElementException();
        Path ret = currentPath;
        currentPath = currentPath.isRoot() ? null : currentPath.getParent();
        return ret;
      }

      @Override
      public void remove() {
        throw new UnsupportedOperationException();
      }
    };
  }

  private static Path findLowestCommonAncestor(Path node1, Path node2) {
    final Set<Path> node1Ancestors = Sets.newLinkedHashSet(traversePathToRoot(node1));
    for (Path path : traversePathToRoot(node2)) {
      if (node1Ancestors.contains(path))
        return path;
    }
    throw new AssertionError("All nodes should have a common ancestor in the root node");
  }

  /**
   * Find the lowest common ancestor of the given paths. For the purposes of this method, a path is
   * considered a valid ancestor of itself.
   */
  public static Path findLowestCommonAncestor(Set<Path> paths) {
    Preconditions.checkNotNull(paths);
    // Could implement something more efficient here, but view trees are seldom very tall.
    Path ret = null;
    for (Path path : paths) {
      if (ret == null) {
        ret = path;
      } else {
        ret = findLowestCommonAncestor(ret, path);
      }
    }
    return ret == null ? Path.ROOT : ret;
  }

  public static RelationPath findLowestCommonRelationAncestor(Set<RelationPath> paths) {
    return findLowestCommonAncestor(ImmutableSet.<Path>copyOf(paths)).getRelation();
  }

  public static Set<RelationPath> getRelationPathsWithAncestors(Set<RelationPath> paths) {
    return Sets.union(paths, getStrictAncestors(paths));
  }

  public static <P extends Path> Set<Path> getPathsWithAncestors(Set<P> paths) {
    Set<Path> ret = Sets.newLinkedHashSet();
    for (Path path : paths) {
      populateStrictAncestors(ret, path);
      ret.add(path);
    }
    return Collections.unmodifiableSet(ret);
  }

  public static <P extends Path> Set<RelationPath> getStrictAncestors(Set<P> paths) {
    Set<RelationPath> ret = Sets.newLinkedHashSet();
    for (Path path : paths)
      populateStrictAncestors(ret, path);
    return Collections.unmodifiableSet(ret);
  }

  private static void populateStrictAncestors(Set<? super RelationPath> populate, Path path) {
    if (path.isRoot())
      return;
    RelationPath parentPath = path.getParent();
    if (populate.add(parentPath))
      populateStrictAncestors(populate, parentPath);
  }

  /**
   * @param paths paths that may or may not exist in {@code view}
   * @return paths from {@code paths} that exists in {@code view}, in view order (preorder
   *         traversal)
   */
  public static <T extends Path> List<T> getPathsInViewOrder(View view, Iterable<T> paths) {
    // Hack to help type checking.
    Map<Path,T> pathsMap = Maps.newHashMap();
    for (T path : paths)
      pathsMap.put(path, path);
    List<T> ret = Lists.newArrayList();
    collectPathsInViewOrder(view, Path.ROOT, pathsMap, ret);
    return ret;
  }

  private static <T extends Path> void collectPathsInViewOrder(
      View view, Path path, Map<Path,T> pathsMap, List<T> toDisplayedPathsInOrder)
  {
    T tPath = pathsMap.get(path);
    if (tPath != null)
      toDisplayedPathsInOrder.add(tPath);
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation()))
        collectPathsInViewOrder(view, childPath, pathsMap, toDisplayedPathsInOrder);
    }
  }

  public static <T extends Path> Set<T> visiblePaths(final View view, Set<T> paths) {
    Set<T> ret = Sets.newLinkedHashSet();
    for (T path : paths) {
      if (view.get(CProp.VISIBLE, path))
        ret.add(path);
    }
    return Collections.unmodifiableSet(ret);
  }

  public static <T extends Path> Set<T> existingPaths(final View view, Set<T> paths) {
    Set<T> ret = Sets.newLinkedHashSet();
    for (T path : paths) {
      if (view.exists(path))
        ret.add(path);
    }
    return Collections.unmodifiableSet(ret);
  }

  public static boolean hasVisiblePrimitiveChildren(View view, RelationPath parent) {
    for (Path childPath : view.getChildPaths(parent)) {
      if (!childPath.isRelation() && view.get(CProp.VISIBLE, childPath))
        return true;
    }
    return false;
  }

  public static Set<Path> getVisibleChildren(View view, RelationPath parent) {
    Set<Path> ret = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(parent)) {
      if (view.get(CProp.VISIBLE, childPath))
        ret.add(childPath);
    }
    return Collections.unmodifiableSet(ret);
  }

  public static Set<Path> getPathsBetweenInclusive(View view, Path path1, Path path2) {
    final Path lowestCommonAncestor = findLowestCommonAncestor(path1, path2);
    if (path1.equals(lowestCommonAncestor) || path2.equals(lowestCommonAncestor) ||
        !lowestCommonAncestor.isRelation())
    {
      return ImmutableSet.of(lowestCommonAncestor);
    }
    int endpointsFound = 0;
    final Set<Path> ret = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(lowestCommonAncestor.getRelation())) {
      final boolean matchesEndpoint =
          (childPath.isEqualToOrAncestorOf(path1) || childPath.isEqualToOrAncestorOf(path2));
      if (matchesEndpoint)
        endpointsFound++;
      if (endpointsFound == 1 || (endpointsFound == 2 && matchesEndpoint))
        ret.add(childPath);
    }
    return Collections.unmodifiableSet(ret);
  }

  public static boolean isAncestorOrEqualInSet(Set<Path> ancestorCandidates, Path path) {
    for (Path candidateAncestorPath : ancestorCandidates) {
      if (candidateAncestorPath.isEqualToOrAncestorOf(path))
        return true;
    }
    return false;
  }

  public static RelationPath getContextPath(Path path) {
    return path.isRelation() ? path.getRelation() : path.getParent();
  }
}
