// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.Path;
import java.math.BigDecimal;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalTime;
import java.util.List;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/** AST for an expression or subexpression. */
@Immutable
public abstract class Expression {
  public abstract <R,E extends Exception> R accept(ExpressionVisitor<R,E> visitor) throws E;

  /**
   * Holds a constant value that can be represented unambiguously by an expression. Not all
   * constants can be represented by a single literal; for instance, a negative number is
   * represented by a unary minus operator on a positive number literal.
   */
  public static final class ConstantExpression extends Expression {
    private static final ImmutableSet<Class<?>> SUPPORTED_TYPES = ImmutableSet.<Class<?>>of(
        Boolean.class, BigDecimal.class, String.class,
        LocalDate.class, LocalTime.class, Instant.class);

    public static final ConstantExpression TRUE = new ConstantExpression(true);
    public static final ConstantExpression FALSE = new ConstantExpression(false);
    public static final ConstantExpression NULL = new ConstantExpression(null);

    private final @Nullable Object value;

    /** @param value must be immutable, and must be of a supported type */
    private ConstantExpression(@Nullable Object value) {
      Preconditions.checkArgument(value == null || SUPPORTED_TYPES.contains(value.getClass()));
      this.value = value;
    }

    public static ConstantExpression create(long value) {
      return create(new BigDecimal(value));
    }

    public static ConstantExpression create(boolean value) {
      return value ? TRUE : FALSE;
    }

    public static ConstantExpression create(LocalDate value) {
      Preconditions.checkNotNull(value);
      return new ConstantExpression(value);
    }

    public static ConstantExpression create(LocalTime value) {
      Preconditions.checkNotNull(value);
      return new ConstantExpression(value);
    }

    public static ConstantExpression create(Instant value) {
      Preconditions.checkNotNull(value);
      return new ConstantExpression(value);
    }

    /* Store all number literals in BigDecimal form; this preserves both the exact precision of the
    number that was written in the expression and the number of decimal points that was used in the
    original string representation (e.g. so "5.00" will be regenerated as "5.00" rather than "5.0"
    or "5" if the expression is turned back into string form). */
    public static ConstantExpression create(BigDecimal value) {
      Preconditions.checkNotNull(value);
      return new ConstantExpression(value);
    }

    public static ConstantExpression create(String value) {
      Preconditions.checkNotNull(value);
      return new ConstantExpression(value);
    }

    /** @return null if the value is NaN or Infinity */
    private static @Nullable ConstantExpression create(double value) {
      Preconditions.checkNotNull(value);
      if (!Double.isFinite(value))
        return null;
      /* Don't try to distinguish positive and negative zero, since the two would be defined as
      equal, and since the BigDecimal representation doesn't distinguish them either. */
      return new ConstantExpression(new BigDecimal(value));
    }

    /**
     * @return null if the value cannot be represented by a ConstantExpression
     */
    public static @Nullable ConstantExpression fromObject(@Nullable Object value) {
      if (value == null) {
        return NULL;
      } else if (value instanceof Boolean) {
        return create((Boolean) value);
      } else if (value instanceof BigDecimal) {
        return create((BigDecimal) value);
      } else if (value instanceof String) {
        return create((String) value);
      } else if (value instanceof Double) {
        return create((Double) value);
      } else if (value instanceof Float) {
        return create((double) ((Float) value));
      } else if (value instanceof Long) {
        return create((Long) value);
      } else if (value instanceof Integer) {
        return create((long) ((Integer) value));
      } else if (value instanceof Short) {
        return create((long) ((Short) value));
      } else if (value instanceof Byte) {
        return create((long) ((Byte) value));
      } else if (value instanceof Character) {
        return create(((Character) value).toString());
      } else if (value instanceof LocalDate) {
        return create((LocalDate) value);
      } else if (value instanceof LocalTime) {
        return create((LocalTime) value);
      } else if (value instanceof Instant) {
        return create((Instant) value);
      } else {
        return null;
      }
    }

    /** @return is always immutable */
    public @Nullable Object getValue() {
      return value;
    }

    @Override
    public String toString() {
      return "Constant(" + value + ")";
    }

    @Override
    public <R,E extends Exception> R accept(ExpressionVisitor<R,E> visitor) throws E {
      return visitor.visit(this);
    }

    @Override
    public int hashCode() {
      return Objects.hash(value);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof ConstantExpression))
        return false;
      ConstantExpression other = (ConstantExpression) obj;
      return Objects.equals(this.value, other.value);
    }
  }

  public static final class FunctionExpression extends Expression {
    private final Function function;
    private final ImmutableList<Expression> arguments;

    // TODO[cc]: Report the problem below as a cobertura bug.
    /* Putting these constants directly under the Expression superclass seems to trip up the
    cobertura test coverage reporter with an ExceptionInInitializerError, probably due to a circular
    classloading dependency that is not properly handled. The problem was present both on
    jdk1.7.0_51 and jdk1.8.0_20, and was solved in both cases by moving the constants into the
    subclass. */
    public static final FunctionExpression SYNTAX_ERROR = new FunctionExpression(
        Function.SYNTAX_ERROR, ImmutableList.of());
    public static final FunctionExpression BROKEN_REFERENCE = new FunctionExpression(
        Function.BROKEN_REFERENCE, ImmutableList.of());

    public FunctionExpression(Function function, List<Expression> arguments) {
      Preconditions.checkNotNull(function);
      Preconditions.checkNotNull(arguments);
      Preconditions.checkArgument(function.isArgumentCountCorrect(arguments.size()));
      this.function = function;
      this.arguments = ImmutableList.copyOf(arguments);
      for (Expression argument : this.arguments)
        Preconditions.checkNotNull(argument);
    }

    public Function getFunction() {
      return function;
    }

    public List<Expression> getArguments() {
      return arguments;
    }

    @Override
    public String toString() {
      if (function.equals(Function.PARENTHESES)) {
        return "(" + arguments.get(0) + ")";
      } else if (function.equals(Function.BROKEN_REFERENCE)) {
        return "#BROKEN_REFERENCE!";
      } else if (function.equals(Function.SYNTAX_ERROR)) {
        return "#SYNTAX_ERROR!";
      } else {
        return function + "(" + Joiner.on(", ").join(arguments) + ")";
      }
    }

    @Override
    public <R,E extends Exception> R accept(ExpressionVisitor<R,E> visitor) throws E {
      return visitor.visit(this);
    }

    @Override
    public int hashCode() {
      return Objects.hash(function, arguments);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof FunctionExpression))
        return false;
      FunctionExpression other = (FunctionExpression) obj;
      return Objects.equals(this.function, other.function)
          && Objects.equals(this.arguments, other.arguments);
    }
  }

  public static final class PathExpression extends Expression {
    private final Path path;

    public PathExpression(Path path) {
      Preconditions.checkNotNull(path);
      this.path = path;
    }

    public Path getPath() {
      return path;
    }

    @Override
    public String toString() {
      return path.toString();
    }

    @Override
    public <R,E extends Exception> R accept(ExpressionVisitor<R,E> visitor) throws E {
      return visitor.visit(this);
    }

    @Override
    public int hashCode() {
      return path.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof PathExpression))
        return false;
      PathExpression other = (PathExpression) obj;
      return path.equals(other.path);
    }
  }
}
