// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import java.util.List;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

@Immutable
public abstract class ExpressionErrorReport {
  // TODO[cc]: Store start and end character offsets instead.
  private final int line;
  private final int charPositionInLine;

  public ExpressionErrorReport(int line, int charPositionInLine) {
    Preconditions.checkArgument(line >= 0);
    Preconditions.checkArgument(charPositionInLine >= 0);
    this.line = line;
    this.charPositionInLine = charPositionInLine;
  }

  public int getLine() {
    return line;
  }

  public int getCharPositionInLine() {
    return charPositionInLine;
  }

  public static final class ExpressionParseException extends Exception {
    private final ImmutableList<ExpressionErrorReport> reports;

    public ExpressionParseException(List<ExpressionErrorReport> reports) {
      super(buildMessage(reports));
      this.reports = ImmutableList.copyOf(reports);
      for (ExpressionErrorReport report : this.reports)
        Preconditions.checkNotNull(report);
    }

    private static String buildMessage(List<ExpressionErrorReport> reports) {
      StringBuilder ret = new StringBuilder();
      ret.append("ExpressionParseException with errors:\n");
      for (ExpressionErrorReport report : reports)
        ret.append("  ").append(report.toString()).append("\n");
      return ret.toString();
    }

    public ImmutableList<ExpressionErrorReport> getReports() {
      return reports;
    }
  }

  /** Indicates a problem looking up a path referenced by an otherwise well-formed relative path
  reference expression. */
  public static final class PathErrorReport extends ExpressionErrorReport {
    private final PathReferenceException pathReferenceException;

    public PathErrorReport(
        int line, int charPositionInLine, PathReferenceException pathReferenceException)
    {
      super(line, charPositionInLine);
      Preconditions.checkNotNull(pathReferenceException);
      this.pathReferenceException = pathReferenceException;
    }

    public PathReferenceException getPathReferenceException() {
      return pathReferenceException;
    }

    @Override
    public String toString() {
      return "Line " + getLine() + ":" + getCharPositionInLine() + ": " + pathReferenceException;
    }
  }

  public static final class SyntaxErrorReport extends ExpressionErrorReport {
    private final boolean parsePhase;
    private final String message;
    private final @Nullable Object offendingSymbol;
    private final @Nullable Exception cause;

    /**
     * @param parsePhase true if the error was reported during the ANTLR parse phase, false if the
     *        error was reported during the parse tree traversal phase
     */
    public SyntaxErrorReport(int line, int charPositionInLine, boolean parsePhase, String message,
        @Nullable Exception cause, @Nullable Object offendingSymbol)
    {
      super(line, charPositionInLine);
      Preconditions.checkNotNull(message);
      this.parsePhase = parsePhase;
      this.message = message;
      this.offendingSymbol = offendingSymbol;
      this.cause = cause;
    }

    public boolean isParsePhase() {
      return parsePhase;
    }

    public String getMessage() {
      return message;
    }

    public @Nullable Object getOffendingSymbol() {
      return offendingSymbol;
    }

    public @Nullable Exception getCause() {
      return cause;
    }

    @Override
    public String toString() {
      return "Line " + getLine() + ":" + getCharPositionInLine() +
          (parsePhase ? " (parse phase)" : "") + " at " + offendingSymbol + ": " + message;
    }
  }
}
