// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.sieuferd.datamodel.Strepco;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableBiMap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.expressions.Expression.*;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.ExpressionParseException;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.PathErrorReport;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.SyntaxErrorReport;
import com.sieuferd.datamodel.expressions.ExpressionVisitor.VoidExpressionVisitor;
import com.sieuferd.datamodel.expressions.parser.ANTLR_ExpressionLexer;
import com.sieuferd.datamodel.expressions.parser.ANTLR_ExpressionParser;
import com.sieuferd.datamodel.expressions.parser.ANTLR_ExpressionParser.*;
import com.sieuferd.datamodel.expressions.parser.ANTLR_ExpressionVisitor;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.ObjectReference;
import java.math.BigDecimal;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAccessor;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import javax.annotation.Nullable;
import org.antlr.v4.runtime.ANTLRErrorListener;
import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.BaseErrorListener;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.Lexer;
import org.antlr.v4.runtime.RecognitionException;
import org.antlr.v4.runtime.Recognizer;
import org.antlr.v4.runtime.RuleContext;
import org.antlr.v4.runtime.Token;
import org.antlr.v4.runtime.TokenStream;
import org.antlr.v4.runtime.misc.Interval;
import org.antlr.v4.runtime.tree.AbstractParseTreeVisitor;

public final class Expressions {
  private Expressions() { }

  private static final ImmutableBiMap<String,Boolean> BOOLEAN_LITERALS =
      ImmutableBiMap.of("true", true, "false", false);

  private static final ImmutableBiMap<String,Function> ERROR_LITERALS =
      ImmutableBiMap.<String,Function>builder()
      .put("ref", Function.BROKEN_REFERENCE)
      .put("huh", Function.SYNTAX_ERROR)
      .build();

  /* In the following maps, when multiple tokens map to the same Function, the first
  listing denotes the preferred form that will be used when generating expression strings from
  ASTs. */

  private static final ImmutableMap<String,Function> UNARY_OPERATORS =
      ImmutableMap.<String,Function>builder()
      .put("+", Function.UNARY_PLUS)
      .put("-", Function.UNARY_MINUS)
      .put("not", Function.LOGICAL_NOT) // Preferred form.
      .put("!", Function.LOGICAL_NOT)
      .build();

  private static final ImmutableMap<String,Function> BINARY_OPERATORS =
      ImmutableMap.<String,Function>builder()
      .put("*", Function.MULTIPLY)
      .put("**", Function.POWER)
      .put("/", Function.DIVIDE)
      .put("%", Function.MODULO)
      .put("+", Function.PLUS)
      .put("-", Function.MINUS)
      .put("is", Function.IS)
      .put("is not", Function.IS_NOT)
      .put(">=", Function.GREATER_OR_EQUAL)
      .put("<=", Function.LESS_OR_EQUAL)
      .put(">", Function.GREATER)
      .put("<", Function.LESS)
      .put("=", Function.EQUAL) // Preferred form.
      .put("==", Function.EQUAL)
      .put("<>", Function.NOT_EQUAL) // Preferred form (used in Excel and ANSI SQL).
      .put("!=", Function.NOT_EQUAL)
      .put("and", Function.LOGICAL_AND) // Preferred form.
      .put("&&", Function.LOGICAL_AND)
      .put("or", Function.LOGICAL_OR) // Preferred form.
      .put("||", Function.LOGICAL_OR)
      .build();

  private static final ImmutableMap<String,Function> NAMED_BUILTIN_FUNCTIONS =
      ImmutableMap.<String,Function>builder()
      .put("sum", Function.SUM)
      .put("avg", Function.AVERAGE)
      .put("stdev", Function.STDEV)
      .put("max", Function.MAX)
      .put("min", Function.MIN)
      .put("count", Function.COUNT)
      .put("if", Function.CONDITIONAL)
      .put("ln", Function.LN)
      .put("concat", Function.STRING_CONCAT)
      .put("lower", Function.STRING_LOWER)
      .put("upper", Function.STRING_UPPER)
      .put("trim", Function.STRING_TRIM)
      .put("len", Function.STRING_LENGTH)
      .put("contains", Function.STRING_CONTAINS)
      .put("startswith", Function.STRING_STARTSWITH)
      .put("endswith", Function.STRING_ENDSWITH)
      .put("left", Function.STRING_LEFT)
      .put("pos", Function.STRING_POSITION)
      .put("year", Function.TEMPORAL_YEAR)
      .put("month", Function.TEMPORAL_MONTH)
      .put("day", Function.TEMPORAL_DAY)
      .put("hour", Function.TEMPORAL_HOUR)
      .put("minute", Function.TEMPORAL_MINUTE)
      .put("second", Function.TEMPORAL_SECOND)
      .put("normalize_whitespace", Function.STRING_NORMALIZE_WHITESPACE)
      .put("group_concat", Function.GROUP_CONCAT)
      .put("cuml_sum", Function.CUML_SUM)
      .build();

  private static final ImmutableMap<Function,String> ERROR_LITERALS_INVERSE =
      invertOperatorList(ERROR_LITERALS);

  private static final ImmutableMap<Function,String> UNARY_OPERATORS_INVERSE =
      invertOperatorList(UNARY_OPERATORS);

  private static final ImmutableMap<Function,String> BINARY_OPERATORS_INVERSE =
      invertOperatorList(BINARY_OPERATORS);

  private static final ImmutableMap<Function,String> NAMED_BUILTIN_FUNCTIONS_INVERSE =
      invertOperatorList(NAMED_BUILTIN_FUNCTIONS);

  /** Keys are ANTLR token types. */
  private static final ImmutableMap<Integer,ExpressionTokenType> TOKEN_TYPES = buildTokenTypes();

  private static ImmutableMap<Function,String> invertOperatorList(
      Map<String,Function> operatorList)
  {
    Map<Function,String> ret = Maps.newLinkedHashMap();
    for (Entry<String,Function> entry : operatorList.entrySet()) {
      Function function = entry.getValue();
      // Use the first listed string representation as the preferred form of each operator.
      if (!ret.containsKey(function))
        ret.put(function, entry.getKey());
    }
    return ImmutableMap.copyOf(ret);
  }

  private static ImmutableMap<Integer,ExpressionTokenType> buildTokenTypes() {
    final Map<Integer,ExpressionTokenType> ret = Maps.newLinkedHashMap();
    for (ExpressionTokenType type : ExpressionTokenType.values()) {
      final Integer antlrTokenType = type.getANTLRTokenType();
      if (antlrTokenType != null)
        MapUtil.putChecked(ret, antlrTokenType, type);
    }
    return ImmutableMap.copyOf(ret);
  }

  private static String standardizedText(Token token) {
    /* Use token.getInputStream().getText instead of token.getText, as the latter calls
    CharStream.size(), which may not be implemented for all streams. */
    return token.getInputStream()
        .getText(Interval.of(token.getStartIndex(), token.getStopIndex()))
        .toLowerCase(Locale.ROOT);
  }

  public static String generateString(Strepco strepco, Expression expression) {
    return generateString(strepco, expression, false);
  }

  public static String generateString(
      final Strepco strepco, final Expression expression, final boolean formulaStyle)
  {
    final String ret = (formulaStyle ? "=" : "") +
        (formulaStyle && expression.equals(ConstantExpression.NULL) ? "" :
        generateStringInternal(strepco, expression, false));
    // Test reversibility.
    final List<ExpressionErrorReport> expressionErrorReports = Lists.newArrayList();
    final Expression reversed = parseInternal(strepco, ret, expressionErrorReports);
    if (!expressionErrorReports.isEmpty()) {
      throw new AssertionError(
          "Got unexpected error reports while re-parsing: " + expressionErrorReports);
    }
    final View contextView = strepco.getView();
    if (!canonicalize(contextView, expression).equals(canonicalize(contextView, reversed))) {
      // TODO[cc]: Log this instead of using System.err.
      System.err.println("Original AST              : " + expression);
      System.err.println("Original AST Canonicalized: " + canonicalize(contextView, expression));
      System.err.println("Generated String          : " + ret);
      System.err.println("Reversed AST              : " + reversed);
      System.err.println("Reversed AST Canonicalized: " + canonicalize(contextView, reversed));
      throw new AssertionError();
    }
    return ret;
  }

  private static String generateStringInternal(
      final Strepco strepco, final Expression expression,
      final boolean parenthesizationTrial)
  {
    final StringBuilder ret = new StringBuilder();
    expression.accept(new VoidExpressionVisitor<RuntimeException>() {
      private void acceptArgumentMaybeParenthesized(FunctionExpression exp, int argIndex) {
        final Expression mainArgument = exp.getArguments().get(argIndex);
        final boolean needsParentheses;

        if (parenthesizationTrial && expression == exp) {
          needsParentheses = false;
        } else {
          final List<Expression> trialArguments = Lists.newArrayList();
          int i = 0;
          for (Expression originalArgument : exp.getArguments()) {
            final Expression trialArgument;
            if (i++ == argIndex) {
              /* Only the first level of the argument AST is needed to determine whether parentheses
              are necessary or not, so prune to avoid a bad run-time complexity. */
              trialArgument = rewrite(strepco.getView(), originalArgument, false, true);
            } else {
              // Any terminal will do here.
              trialArgument = FunctionExpression.BROKEN_REFERENCE;
            }
            trialArguments.add(trialArgument);
          }
          final FunctionExpression trialExpression =
              new FunctionExpression(exp.getFunction(), trialArguments);
          final String trialString = generateStringInternal(strepco, trialExpression, true);
          final List<ExpressionErrorReport> errorReports = Lists.newArrayList();
          final Expression trialParse = parseInternal(strepco, trialString, errorReports);
          if (!errorReports.isEmpty())
            throw new AssertionError("Got error reports while re-parsing \"" + trialString + "\"");
          /* Canonicalize before comparison to avoid unnecessary parentheses around compound
          constants, e.g. in "2 + -3". */
          needsParentheses =
              !canonicalize(null, trialParse).equals(canonicalize(null, trialExpression));
        }
        if (needsParentheses)
          ret.append("(");
        mainArgument.accept(this);
        if (needsParentheses)
          ret.append(")");
      }

      @Override
      public void voidVisit(ConstantExpression exp) {
        Object value = exp.getValue();
        if (value == null) {
          ret.append("null");
        } else if (value instanceof Boolean) {
          ret.append(MapUtil.getChecked(BOOLEAN_LITERALS.inverse(), (Boolean) value));
        } else if (value instanceof BigDecimal) {
          /* For positive values, the format produced by BigDecimal.toString() is always parseable
          as either an integer or floating point literal, both which are represented as a
          BigDecimal. Negative values will be parsed back as a literal enclosed in a unary minus. */
          ret.append(((BigDecimal) value).toString());
        } else if (value instanceof String) {
          ret.append(generateStringLiteral((String) value, false));
        } else if (value instanceof LocalDate) {
          ret.append("{");
          ret.append(DateTimeFormatter.ISO_LOCAL_DATE.format((LocalDate) value));
          ret.append("}");
        } else if (value instanceof LocalTime) {
          ret.append("{");
          ret.append(DateTimeFormatter.ISO_LOCAL_TIME.format((LocalTime) value));
          ret.append("}");
        } else if (value instanceof Instant) {
          ret.append("{");
          ret.append(TemporalFormats.ISO_LOCAL_DATE_TIME_WITH_SPACE.format(
              ((Instant) value).atZone(strepco.getTimeZoneId()).toLocalDateTime()));
          ret.append("}");
        } else {
          throw new AssertionError("Unexpected constant value of type " + value.getClass());
        }
      }

      @Override
      public void voidVisit(FunctionExpression exp) {
        // TODO[cc]: Use FunctionTemplate for this.
        if (exp.getFunction().equals(Function.PARENTHESES)) {
          ret.append("(");
          exp.getArguments().get(0).accept(this);
          ret.append(")");
          return;
        }
        {
          final String operator = ERROR_LITERALS_INVERSE.get(exp.getFunction());
          if (operator != null) {
            ret.append("#").append(operator).append("!");
            return;
          }
        }
        {
          final String operator = UNARY_OPERATORS_INVERSE.get(exp.getFunction());
          if (operator != null) {
            ret.append(operator);
            if (Character.isLetter(operator.charAt(operator.length() - 1)))
              ret.append(" ");
            acceptArgumentMaybeParenthesized(exp, 0);
            return;
          }
        }
        {
          final String operator = BINARY_OPERATORS_INVERSE.get(exp.getFunction());
          if (operator != null) {
            acceptArgumentMaybeParenthesized(exp, 0);
            ret.append(" ").append(operator).append(" ");
            acceptArgumentMaybeParenthesized(exp, 1);
            return;
          }
        }
        {
          String functionName = exp.getFunction().getName();
          if (functionName == null)
            functionName = NAMED_BUILTIN_FUNCTIONS_INVERSE.get(exp.getFunction());
          if (functionName != null) {
            ret.append(functionName).append("(");
            int i = 0;
            for (Expression argument : exp.getArguments()) {
              if (i++ > 0)
                ret.append(", ");
              argument.accept(this);
            }
            ret.append(")");
            return;
          }
        }
        throw new AssertionError();
      }

      @Override
      public void voidVisit(PathExpression exp) {
        if (!strepco.getView().exists(exp.getPath())) {
          FunctionExpression.BROKEN_REFERENCE.accept(this);
        } else {
          ret.append(PathReferences.getTokenStringFromPath(strepco, exp.getPath()));
        }
      }
    });
    return ret.toString();
  }

  /**
   * Remove AST nodes representing parentheses, fold constant patterns that might have been
   * generated from a single original ConstantExpression, and, if markBrokenReferencesView is not
   * null, replace references to non-existent paths with Expression.BROKEN_REFERENCE.
   */
  public static Expression canonicalize(@Nullable View markBrokenReferencesView, Expression e) {
    return foldCompoundConstants(rewrite(markBrokenReferencesView, e, true, false));
  }

  /**
   * Fold compound expressions generated by @code{generateStringInternal}/
   * @code{visit(ConstantExpression)} back into a single ConstantExpression.
   */
  private static Expression foldCompoundConstants(Expression exp) {
    return exp.accept(new ExpressionVisitor<Expression,RuntimeException>() {
      @Override
      public Expression visit(ConstantExpression exp) {
        return exp;
      }

      @Override
      public Expression visit(FunctionExpression exp) {
        final List<Expression> foldedArguments = Lists.newArrayList();
        for (Expression arg : exp.getArguments())
          foldedArguments.add(foldCompoundConstants(arg));
        FunctionExpression exp2 =
            new FunctionExpression(exp.getFunction(), foldedArguments);

        if (exp2.getFunction().equals(Function.UNARY_MINUS) &&
            exp2.getArguments().get(0) instanceof ConstantExpression &&
            ((ConstantExpression) exp2.getArguments().get(0)).getValue() instanceof BigDecimal)
        {
          return ConstantExpression.create(
              ((BigDecimal) ((ConstantExpression) exp2.getArguments().get(0)).getValue()).negate());
        } else {
          return exp2;
        }
      }

      @Override
      public Expression visit(PathExpression exp) {
        return exp;
      }
    });
  }

  private static Expression rewrite(final @Nullable View markBrokenReferencesView, Expression exp,
      final boolean eliminateParentheses, final boolean prune)
  {
    return exp.accept(new ExpressionVisitor<Expression,RuntimeException>() {
      @Override
      public Expression visit(ConstantExpression exp) {
        return exp;
      }

      @Override
      public Expression visit(FunctionExpression exp) {
        final Function function = exp.getFunction();
        final List<Expression> canonicalizedArguments = Lists.newArrayList();
        for (Expression arg : exp.getArguments()) {
          canonicalizedArguments.add(prune
              ? FunctionExpression.BROKEN_REFERENCE
              : rewrite(markBrokenReferencesView, arg, eliminateParentheses, prune));
        }
        return eliminateParentheses && function.equals(Function.PARENTHESES)
            ? canonicalizedArguments.get(0)
            : new FunctionExpression(function, canonicalizedArguments);
      }

      @Override
      public Expression visit(PathExpression exp) {
        return markBrokenReferencesView != null && !markBrokenReferencesView.exists(exp.getPath())
            ? FunctionExpression.BROKEN_REFERENCE : exp;
      }
    });
  }

  /**
   * Parse a string as an expression in a non-lenient manner, throwing an
   * {@link ExpressionParseException} if there are any errors.
   */
  public static Expression parse(Strepco strepco, String expressionString)
      throws ExpressionParseException
  {
    final List<ExpressionErrorReport> errorReports = Lists.newArrayList();
    Expression ret = parse(strepco, expressionString, errorReports);
    if (!errorReports.isEmpty())
      throw new ExpressionParseException(errorReports);
    return ret;
  }

  /**
   * Parse a string as an expression with graceful error handling. Errors are added to the supplied
   * {@code errorReports} list, possibly resulting in ErrorExpression nodes being included in the
   * returned AST.
   *
   * @param retErrorReports will be populated with error reports, if any
   */
  public static Expression parse(Strepco strepco,
      String expressionString, final List<ExpressionErrorReport> retErrorReports)
  {
    final Expression ret = parseInternal(strepco, expressionString, retErrorReports);
    // Test reversibility.
    generateString(strepco, ret);
    generateString(strepco, canonicalize(strepco.getView(), ret));
    return ret;
  }

  /* TODO[cc]: Consider hiding the ANTLR-specific data types from this API, instead providing a
               dedicated class that allows iteration returning ExpressionTokenType objects, and
               which hides the ANTLRErrorListener details. Such a class would also be an appropriate
               place to incorporate the extra logic for splitting up path reference tokens,
               currently residing in the ExpressionLexer class in the Project module. */
  /**
   * Get an ANTLR lexer suitable for tokenizing the expression language. It is the responsibility of
   * the client to register any relevant error listeners.
   */
  public static org.antlr.v4.runtime.Lexer getANTLRLexer(CharStream charStream) {
    Preconditions.checkNotNull(charStream);
    final org.antlr.v4.runtime.Lexer ret = new ANTLR_ExpressionLexer(charStream);
    // Remove the default listener, which prints errors to the console.
    ret.removeErrorListeners();
    return ret;
  }

  /**
   * @param token a token produced by a lexer retrieved via {@link #getANTLRLexer(CharStream)}
   * @return the token type
   */
  public static ExpressionTokenType getTokenType(Token token) {
    Preconditions.checkNotNull(token);
    final int antlrTokenType = token.getType();
    {
      ExpressionTokenType ret = TOKEN_TYPES.get(antlrTokenType);
      if (ret != null)
        return ret;
    }
    final String text = standardizedText(token);
    if (UNARY_OPERATORS.containsKey(text) || BINARY_OPERATORS.containsKey(text) ||
        text.equals("?") || text.equals(":"))
    {
      return ExpressionTokenType.OPERATOR;
    } else {
      return ExpressionTokenType.OTHER;
    }
  }

  /**
   * @param errorReportList will be added to when the listener is notified of syntax errors
   */
  static ANTLRErrorListener createErrorListener(final List<ExpressionErrorReport> errorReportList) {
    return new BaseErrorListener() {
      @Override
      public void syntaxError(Recognizer<?,?> recognizer, @Nullable Object offendingSymbol,
          int line, int charPositionInLine, String message,
          @Nullable RecognitionException recognitionException)
      {
        Preconditions.checkNotNull(message);
        errorReportList.add(new SyntaxErrorReport(
            line, charPositionInLine, true, message, recognitionException, offendingSymbol));
      }
    };
  }

  private static boolean isEmptyWithoutErrors(String expressionString) {
    /* Permit an empty expression, or a lone equals sign (OP_ASSIGN), as a shorthand for a null
    literal. Don't build this into the grammar, as it can interfere with error handling by serving
    as an "always possible" option. */
    final Lexer lexer = getANTLRLexer(new ANTLRInputStream(expressionString));
    final List<ExpressionErrorReport> errors = Lists.newArrayList();
    lexer.addErrorListener(Expressions.createErrorListener(errors));
    boolean seenEqualsSign = false;
    Token token;
    while ((token = lexer.nextToken()).getType() != Lexer.EOF) {
      if (token.getChannel() == Token.HIDDEN_CHANNEL)
        continue;
      if (!seenEqualsSign && token.getType() == ANTLR_ExpressionLexer.OP_ASSIGN) {
        seenEqualsSign = true;
      } else {
        return false;
      }
    }
    return errors.isEmpty();
  }

  private static Expression parseInternal(Strepco strepco,
      String expressionString, final List<ExpressionErrorReport> retErrorReports)
  {
    Preconditions.checkNotNull(retErrorReports);
    if (isEmptyWithoutErrors(expressionString))
      return ConstantExpression.NULL;
    final ANTLRErrorListener errorListener = createErrorListener(retErrorReports);
    final CharStream charStream = new ANTLRInputStream(expressionString);
    final org.antlr.v4.runtime.Lexer lexer = getANTLRLexer(charStream);
    lexer.addErrorListener(errorListener);
    final TokenStream tokenStream = new CommonTokenStream(lexer);
    final ANTLR_ExpressionParser parser = new ANTLR_ExpressionParser(tokenStream);
    parser.removeErrorListeners();
    parser.addErrorListener(errorListener);

    /* TODO[pd]: Eliminate ambiguities in the grammar so that I can make use of
                 DiagnosticErrorListener. */
    /* The DiagnosticErrorListener is not too useful here since I'm relying on ANTLR4 resolving my
    intentionally ambiguous expression grammar by means of operator precedence parsing. */
    // parser.addErrorListener(new DiagnosticErrorListener(true));

    // No exception is expected here, since I'm using the default non-bailing error handler.
    final BaseExpressionContext context = parser.baseExpression();
    return context.accept(new ASTBuildingExpressionVisitor(strepco, retErrorReports));
  }

  public static @Nullable String readOneStringLiteral(
      String stringToTokenize, ObjectReference<Integer> retTokenLength)
  {
    final Lexer lexer = Expressions.getANTLRLexer(new ANTLRInputStream(stringToTokenize));
    final List<ExpressionErrorReport> errors = Lists.newArrayList();
    lexer.addErrorListener(Expressions.createErrorListener(errors));
    final Token token = lexer.nextToken();
    if (!errors.isEmpty())
      return null;
    final String text = token.getText();
    if (token.getType() == ANTLR_ExpressionLexer.StringLiteral) {
      retTokenLength.setValue(text.length());
      return parseStringLiteral(text);
    } else if (token.getType() == ANTLR_ExpressionLexer.UnterminatedStringLiteral) {
      retTokenLength.setValue(text.length());
      return parseStringLiteral(text + text.charAt(0));
    } else {
      retTokenLength.setValue(0);
      return null;
    }
  }

  private static @Nullable String parseStringLiteral(String escaped) {
    final char quoteChar = escaped.charAt(0);
    if (escaped.charAt(escaped.length() - 1) != quoteChar)
      throw new AssertionError();
    final String core = escaped.substring(1, escaped.length() - 1);
    String modifiedCore;
    if (quoteChar == '\"') {
      /* The sequence "\"\"" can not occur in a valid JSON string token, so it's safe to accept as
      an alternative way to escape quotes ("double double quotes" is how double quotes are escaped
      in Excel). */
      modifiedCore = core.replace("\"\"", "\\\"");
    } else if (quoteChar == '\'') {
      /* The sequence "''" can not occur in valid variants of JSON string tokens that use single
      quotes. */
      modifiedCore = core.replace("''", "'");
      if (!modifiedCore.contains("\\\""))
        modifiedCore = modifiedCore.replace("\"", "\\\"");
    } else {
      throw new AssertionError();
    }
    /* Since the logic for allowing additional kinds of escape sequences is a bit complex, don't
    rely on the lexer to catch all errors. Though currently, we never return null from this
    method. */
    return StringEscaper.unescape(modifiedCore);
  }

  public static String generateStringLiteral(String unescaped, boolean forJSON) {
    String ret = "\"" + StringEscaper.escape(unescaped, forJSON) + "\"";
    if (!unescaped.equals(parseStringLiteral(ret)))
      throw new AssertionError();
    return ret;
  }

  private static final class ASTBuildingExpressionVisitor
      extends AbstractParseTreeVisitor<Expression> implements ANTLR_ExpressionVisitor<Expression>
  {
    private final Strepco strepco;
    private final List<ExpressionErrorReport> errorReports;

    /** @param errorReports a list to be populated with error reports, if any */
    ASTBuildingExpressionVisitor(Strepco strepco, List<ExpressionErrorReport> errorReports) {
      Preconditions.checkNotNull(strepco);
      Preconditions.checkNotNull(errorReports);
      this.strepco = strepco;
      this.errorReports = errorReports;
    }

    @Override
    protected Expression defaultResult() {
      /* This default may be needed in cases where child nodes that are normally expected are absent
      due to ANTLR error recovery. It is also called by the default implementation of
      visitErrorNode. I've never actually seen visitErrorNode be called, but if it ever is, it will
      be properly handled by means of this default. */
      return FunctionExpression.SYNTAX_ERROR;
    }

    /**
     * Parse the specified AST node as a subexpression, providing a default value if ctx is null, as
     * may be the case in certain error conditions.
     */
    private Expression recurse(@Nullable RuleContext ctx) {
      if (ctx == null)
        return defaultResult();
      return ctx.accept(this);
    }

    @Override
    public Expression visitBaseExpression(BaseExpressionContext ctx) {
      return recurse(ctx.expression());
    }

    @Override
    public Expression visitLiteralExpression(LiteralExpressionContext ctx) {
      return recurse(ctx.literal());
    }

    @Override
    public Expression visitNumberLiteral(NumberLiteralContext ctx) {
      Token token = ctx.NumberLiteral().getSymbol();
      try {
        return ConstantExpression.create(new BigDecimal(token.getText()));
      } catch (NumberFormatException e) {
        errorReports.add(new SyntaxErrorReport(token.getLine(), token.getCharPositionInLine(),
            false, "Malformed number literal", e, token));
        return FunctionExpression.SYNTAX_ERROR;
      }
    }

    @Override
    public Expression visitStringLiteral(StringLiteralContext ctx) {
      final Token token = ctx.StringLiteral().getSymbol();
      final String unescaped = parseStringLiteral(token.getText());
      if (unescaped == null) {
        errorReports.add(new SyntaxErrorReport(token.getLine(), token.getCharPositionInLine(),
            false, "Malformed string literal", null, token));
        return FunctionExpression.SYNTAX_ERROR;
      } else {
        return ConstantExpression.create(unescaped);
      }
    }

    @Override
    public Expression visitTemporalLiteral(TemporalLiteralContext ctx) {
      final Token token = ctx.TemporalLiteral().getSymbol();
      final String text = token.getText();
      if (!text.startsWith("{") || !text.endsWith("}"))
        throw new AssertionError();
      final String core = text.substring(1, text.length() - 1);
      TemporalAccessor ta = TemporalFormats.parseTemporal(
          core.trim().replaceAll("\\s+", " "),
          strepco.getLocale(), strepco.getTimeZoneId());
      if (ta instanceof LocalDate) {
        return ConstantExpression.create((LocalDate) ta);
      } else if (ta instanceof LocalTime) {
        return ConstantExpression.create((LocalTime) ta);
      } else if (ta instanceof Instant) {
        return ConstantExpression.create((Instant) ta);
      } else {
        errorReports.add(new SyntaxErrorReport(token.getLine(), token.getCharPositionInLine(),
            false, "Unrecognized date or time format", null, token));
        return FunctionExpression.SYNTAX_ERROR;
      }
    }

    @Override
    public Expression visitBooleanLiteral(BooleanLiteralContext ctx) {
      return ConstantExpression.create(MapUtil.getChecked(BOOLEAN_LITERALS,
          standardizedText(ctx.BooleanLiteral().getSymbol())));
    }

    @Override
    public Expression visitNullLiteral(NullLiteralContext ctx) {
      return ConstantExpression.NULL;
    }

    @Override
    public Expression visitErrorLiteral(ErrorLiteralContext ctx) {
      final Token token = ctx.ErrorLiteral().getSymbol();
      final String text = standardizedText(token);
      final String identifier = text.substring(1, text.length() - 1);
      final Function errorFunction = ERROR_LITERALS.get(identifier);
      if (errorFunction == null) {
        errorReports.add(new SyntaxErrorReport(
            token.getLine(), token.getCharPositionInLine(),
            false, "Unknown error literal type", null, token));
        return FunctionExpression.SYNTAX_ERROR;
      } else {
        return new FunctionExpression(errorFunction, ImmutableList.of());
      }
    }

    @Override
    public Expression visitUnaryOperatorExpression(UnaryOperatorExpressionContext ctx) {
      return new FunctionExpression(
          MapUtil.getChecked(UNARY_OPERATORS, standardizedText(ctx.op)),
          ImmutableList.of(recurse(ctx.a)));
    }

    @Override
    public Expression visitBinaryOperatorExpression(BinaryOperatorExpressionContext ctx)
    {
      final String operatorText =
          (standardizedText(ctx.op) + (ctx.op2 == null ? "" : (" " + standardizedText(ctx.op2))));
      return new FunctionExpression(
          MapUtil.getChecked(BINARY_OPERATORS, operatorText),
          ImmutableList.of(recurse(ctx.aLeft), recurse(ctx.aRight)));
    }

    @Override
    public Expression visitPathExpression(PathExpressionContext ctx) {
      final Token token = ctx.PathReference().getSymbol();
      try {
        return new PathExpression(PathReferences.getPathFromTokenString(strepco, token.getText()));
      } catch (PathReferenceException e) {
        errorReports.add(new PathErrorReport(token.getLine(), token.getCharPositionInLine(), e));
        return FunctionExpression.BROKEN_REFERENCE;
      }
    }

    @Override
    public Expression visitConditionalExpression(ConditionalExpressionContext ctx) {
      /* The error input "true?" has been observed to make ctx.ifFalse null; this case is handled
      gracefully by the recurse() function. */
      return new FunctionExpression(Function.CONDITIONAL, ImmutableList.of(
          recurse(ctx.cond), recurse(ctx.ifTrue), recurse(ctx.ifFalse)));
    }

    @Override
    public Expression visitFunctionExpression(FunctionExpressionContext ctx) {
      final List<Expression> arguments = Lists.newArrayList();
      for (ExpressionContext expCtx : ctx.expression())
        arguments.add(recurse(expCtx));
      final String name = standardizedText(ctx.Identifier().getSymbol());
      Function function = NAMED_BUILTIN_FUNCTIONS.get(name);
      if (function == null)
        function = Function.create(name);
      if (!function.isArgumentCountCorrect(arguments.size()))
        return FunctionExpression.SYNTAX_ERROR;
      return new FunctionExpression((function), arguments);
    }

    @Override
    public Expression visitParenExpression(ParenExpressionContext ctx) {
      return new FunctionExpression(
          Function.PARENTHESES, ImmutableList.of(recurse(ctx.expression())));
    }
  }

  public static Expression rewritePathReferences(final PathMap pathMap, Expression input) {
    return input.accept(new ExpressionVisitor<Expression,RuntimeException>() {
      @Override
      public Expression visit(ConstantExpression exp) {
        return exp;
      }

      @Override
      public Expression visit(FunctionExpression exp) {
        final List<Expression> arguments = Lists.newArrayList();
        for (Expression argument : exp.getArguments())
          arguments.add(rewritePathReferences(pathMap, argument));
        return new FunctionExpression(exp.getFunction(), arguments);
      }

      @Override
      public Expression visit(PathExpression exp) {
        return new PathExpression(pathMap.getNewPath(exp.getPath()));
      }
    });
  }

  public static Expression and(Iterable<Expression> expressions) {
    return fold(Function.LOGICAL_AND, ConstantExpression.TRUE, expressions);
  }

  public static Expression and(Expression... expressions) {
    return and(Arrays.asList(expressions));
  }

  public static Expression or(Iterable<Expression> expressions) {
    return fold(Function.LOGICAL_OR, ConstantExpression.FALSE, expressions);
  }

  public static Expression or(Expression... expressions) {
    return or(Arrays.asList(expressions));
  }

  public static List<Expression> unfoldAnd(Expression exp) {
    return unfold(Function.LOGICAL_AND, ConstantExpression.TRUE, exp);
  }

  public static List<Expression> unfoldOr(Expression exp) {
    return unfold(Function.LOGICAL_OR, ConstantExpression.FALSE, exp);
  }

  private static Expression fold(
      Function function, Expression identity, Iterable<Expression> operands)
  {
    Expression ret = identity;
    for (Expression operand : operands) {
      if (!operand.equals(identity)) {
        ret = ret.equals(identity)
            ? operand : new FunctionExpression(function, ImmutableList.of(ret, operand));
      }
    }
    return ret;
  }

  private static List<Expression> unfold(
      final Function function, Expression identity, Expression expression)
  {
    if (expression.equals(identity))
      return ImmutableList.of();
    if (expression instanceof FunctionExpression) {
      FunctionExpression fexp = (FunctionExpression) expression;
      if (fexp.getFunction().equals(function)) {
        List<Expression> ret = Lists.newArrayList();
        for (Expression argument : fexp.getArguments())
          ret.addAll(unfold(function, identity, argument));
        return ret;
      } else if (fexp.getFunction().equals(Function.PARENTHESES)) {
        return unfold(function, identity, fexp.getArguments().get(0));
      }
    }
    return ImmutableList.of(expression);
  }
}
