// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import java.util.Locale;
import java.util.Objects;
import java.util.regex.Pattern;
import javax.annotation.Nullable;

/**
 * A function that can be used in front-end {@link com.sieuferd.datamodel.expressions.Expression}s.
 */
public final class Function {
  // VALID_NAME_REGEX must be defined before other constants, since it's used by createBuiltin.
  // No regex anchors needed here.
  private static final Pattern VALID_NAME_REGEX = Pattern.compile("[_a-z][_a-z0-9]*");

  /* Marker for a subexpression enclosed in parentheses. Keeping track of parentheses allows
  expressions to be stored as ASTs while still being able to faithfully recreate them during user
  interactions. */
  public static final Function PARENTHESES = createBuiltin("PARENTHESES", 1);

  /**
   * On some databases, the default collation of certain data types may not necessarily constitute a
   * total order on values in the domain of said type. For instance, MySQL performs case-insensitive
   * ordering on string values by default, and may collapse text values that are equal when compared
   * in a case-insensitive manner during SELECT DISTINCT. This function should apply whatever cast
   * or similar annotation that might be necessary on the underlying database in order to ensure a
   * total order on the supplied argument.
   */
  public static final Function TOTALLY_ORDERED_COLLATION =
      createBuiltin("TOTALLY_ORDERED_COLLATION", 1);
  // Markers for error tokens.
  public static final Function BROKEN_REFERENCE = createBuiltin("BROKEN_REFERENCE", 0);
  public static final Function SYNTAX_ERROR = createBuiltin("SYNTAX_ERROR", 0);
  // Unary operators.
  public static final Function UNARY_PLUS = createBuiltin("UNARY_PLUS", 1);
  public static final Function UNARY_MINUS = createBuiltin("UNARY_MINUS", 1);
  public static final Function LOGICAL_NOT = createBuiltin("LOGICAL_NOT", 1);
  // Binary operators.
  /**
   * Arithmetic multiplication. If either argument is null, the result is null, even if the other
   * argument is zero.
   */
  public static final Function MULTIPLY = createBuiltin("MULTIPLY", 2);
  /* The power function is defined to return either a finite number or NULL given that the both
  arguments are either finite numbers or NULL. The semantics were worked out by carefully reducing
  the "special cases" list in the Javadoc for Math.pow(). They also match the error conditions
  seen in the source code for Postgres' power function implementation; see the dpow function in
  http://doxygen.postgresql.org/float_8c_source.html . */
  /**
   * Mathematical power function. The result is NULL if the first argument is negative and the
   * second argument is not an integer, or if the first argument is zero and the second argument is
   * negative.
   */
  public static final Function POWER = createBuiltin("POWER", 2);
  /** Division by zero should yield NULL, not a floating-point NaN or Infinity. */
  public static final Function DIVIDE = createBuiltin("DIVIDE", 2);
  public static final Function MODULO = createBuiltin("MODULO", 2);
  public static final Function PLUS = createBuiltin("PLUS", 2);
  public static final Function MINUS = createBuiltin("MINUS", 2);

  /* Greater than/less than operate on numbers only. */
  public static final Function GREATER_OR_EQUAL = createBuiltin("GREATER_OR_EQUAL", 2);
  public static final Function LESS_OR_EQUAL = createBuiltin("LESS_OR_EQUAL", 2);
  public static final Function GREATER = createBuiltin("GREATER", 2);
  public static final Function LESS = createBuiltin("LESS", 2);

  /* EQUAL, NOT_EQUAL, IS, and IS_NOT use exact matches, even when the input values might come from
  string columns with case-insensitive collations. */
  public static final Function EQUAL = createBuiltin("EQUAL", 2);
  public static final Function NOT_EQUAL = createBuiltin("NOT_EQUAL", 2);

  public static final Function IS = createBuiltin("IS", 2);
  public static final Function IS_NOT = createBuiltin("IS_NOT", 2);
  public static final Function LOGICAL_AND = createBuiltin("LOGICAL_AND", 2);
  public static final Function LOGICAL_OR = createBuiltin("LOGICAL_OR", 2);
  /* I consider making (null ? x : x) return x, but this is probably a too obscure case to justify
  the added complexity. Besides, it would make the operator useless for return types without a
  well-defined equality operator (e.g. the XML type). */
  /**
   * Excel-style IF function. Also used to realize the C-style ternary operator. Returns the second
   * or third argument if the first argument is true or false, respectively. Returns NULL if the
   * first argument is NULL. The not-selected second or third argument may be NULL without affecting
   * the return value.
   *
   * <p>The second and third argument must be of the same type, or null.
   */
  public static final Function CONDITIONAL = createBuiltin("CONDITIONAL", 3);

  /**
   * Natural logarithm. Returns NULL if the argument is zero or negative.
   */
  public static final Function LN = createBuiltin("LN", 1);

  // String functions.
  /** Accepts arguments of any type; null values are translated to empty strings. */
  public static final Function STRING_CONCAT = createBuiltin("STRING_CONCAT", null);
  public static final Function STRING_LOWER = createBuiltin("STRING_LOWER", 1);
  public static final Function STRING_UPPER = createBuiltin("STRING_UPPER", 1);
  /** Removes leading and trailing spaces. Other kinds of whitespace characters are not removed. */
  public static final Function STRING_TRIM = createBuiltin("STRING_TRIM", 1);
  /** Get the length of a string. Incudes any leading or trailing whitespace characters. */
  public static final Function STRING_LENGTH = createBuiltin("STRING_LENGTH", 1);
  /** Returns true iff the first string contains the second. Case-sensitive. */
  public static final Function STRING_CONTAINS = createBuiltin("STRING_CONTAINS", 2);
  /** Returns true iff the first string starts with the second. Case-sensitive. */
  public static final Function STRING_STARTSWITH = createBuiltin("STRING_STARTSWITH", 2);
  /** Returns true iff the first string ends with the second. Case-sensitive. */
  public static final Function STRING_ENDSWITH = createBuiltin("STRING_ENDSWITH", 2);
  public static final Function STRING_LEFT = createBuiltin("STRING_LEFT", 2);

  public static final Function TEMPORAL_YEAR = createBuiltin("TEMPORAL_YEAR", 1);
  public static final Function TEMPORAL_MONTH = createBuiltin("TEMPORAL_MONTH", 1);
  public static final Function TEMPORAL_DAY = createBuiltin("TEMPORAL_DAY", 1);
  public static final Function TEMPORAL_HOUR = createBuiltin("TEMPORAL_HOUR", 1);
  public static final Function TEMPORAL_MINUTE = createBuiltin("TEMPORAL_MINUTE", 1);
  public static final Function TEMPORAL_SECOND = createBuiltin("TEMPORAL_SECOND", 1);

  /**
   * Return the one-based position of the first occurence of the second string argument in the
   * first, or NULL if there is no match. Case-sensitive.
   */
  public static final Function STRING_POSITION = createBuiltin("STRING_POSITION", 2);
  public static final Function STRING_NORMALIZE_WHITESPACE =
      createBuiltin("STRING_NORMALIZE_WHITESPACE", 1);
  // Aggregate function.
  /**
   * The first argument is the separator string. The second argument may be of any type; null
   * values in the group are translated to empty strings.
   */
  public static final Function GROUP_CONCAT = createBuiltin("GROUP_CONCAT", 2);

  // Standard aggregate functions.

  /* I considered including a boolean EXISTS(R) function, but the equivalent COUNT(R) > 0 is easy to
  use instead, so there's little reason to increase the number of aggregate functions just to create
  a shortcut. */
  /* For a discussion of NULL-related SUM behavior in SQL, see SQLite's TOTAL function
  ( http://www.sqlite.org/lang_aggfunc.html ) and
  http://stackoverflow.com/questions/6524352/aggregate-of-an-empty-result-set . */
  /**
   * Computes the total of all values, ignoring null values. If there are no values to total, the
   * result is zero (not null). The latter is an intentional deviation from the behavior of SQL's
   * SUM function.
   */
  public static final Function SUM = createBuiltin("SUM", 1);
  public static final Function AVERAGE = createBuiltin("AVERAGE", 1);
  /** Unbiased (sample) standard deviation. */
  public static final Function STDEV = createBuiltin("STDEV", 1);
  /** Population standard deviation. */
  public static final Function STDEVP = createBuiltin("STDEVP", 1);
  public static final Function MAX = createBuiltin("MAX", 1);
  public static final Function MIN = createBuiltin("MIN", 1);
  /** Counts non-null values. */
  public static final Function COUNT = createBuiltin("COUNT", 1);

  // For testing window functions.
  public static final Function CUML_SUM = createBuiltin("CUML_SUM", 3);

  // TODO[cc]: Make this a property of each Function instead.
  public static final ImmutableSet<Function> AGGREGATE_FUNCTIONS = ImmutableSet.of(
      SUM, AVERAGE, STDEV, MAX, MIN, COUNT, GROUP_CONCAT
  );

  /** Null for variadic functions and non-builtins. */
  private final Integer argumentCount;
  private final String name;
  private final boolean builtin;

  private Function(String name, Integer argumentCount, boolean builtin) {
    Preconditions.checkArgument(isValidFunctionName(name));
    this.name = name;
    this.argumentCount = argumentCount;
    this.builtin = builtin;
  }

  private static Function createBuiltin(String name, Integer argumentCount) {
    return new Function(name.toLowerCase(Locale.ROOT), argumentCount, true);
  }

  /**
   * @param name must be of the format {@code "[_a-z][_a-z0-9]*"}
   */
  public static Function create(String name) {
    return new Function(name, null, false);
  }

  public boolean isArgumentCountCorrect(int candidateArgumentCount) {
    return argumentCount == null || candidateArgumentCount == argumentCount;
  }

  public @Nullable Integer getArgumentCount() {
    return argumentCount;
  }

  public static boolean isValidFunctionName(String name) {
    Preconditions.checkNotNull(name);
    return VALID_NAME_REGEX.matcher(name).matches();
  }

  /** @return can be null to indicate a built-in function */
  public @Nullable String getName() {
    return builtin ? null : name;
  }

  public boolean isBuiltin() {
    return builtin;
  }

  @Override
  public String toString() {
    return builtin ? name.toUpperCase(Locale.ROOT) : name;
  }

  @Override
  public int hashCode() {
    return Objects.hash(name, builtin);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof Function))
      return false;
    Function other = (Function) obj;
    return this.builtin == other.builtin
        && this.name.equals(other.name);
  }
}
