// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.google.common.base.CharMatcher;
import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.expressions.PathReference.Segment.LabelSegment;
import com.sieuferd.datamodel.expressions.PathReference.Segment.OrdinalSegment;
import com.sieuferd.datamodel.expressions.PathReferenceException.LabelNotAtPathException;
import com.sieuferd.datamodel.expressions.PathReferenceException.LabelNotUniqueException;
import com.sieuferd.datamodel.expressions.PathReferenceException.OrdinalOutOfRangeException;
import java.util.List;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

 /**
  * Representation of a path reference token. While path references do have a bit of internal
  * structure to them, only the path reference as a whole corresponds to an actual
  * {@link Expression}, and path references need to be considered as a single unit for syntax
  * highlighting purposes. The internal segments of a path reference are also tokenized in a
  * different way than the rest of the expression language. For these reasons, a path reference is
  * considered a single token, and path references are parsed manually here rather than via the
  * ANTLR grammar for the expression language.
  */
@Immutable
final class PathReference {
  private final String initialLabel;
  private final ImmutableList<Segment> descendantSegments;

  /** @param initialLabel may be empty, but not null */
  PathReference(String initialLabel, List<Segment> descendantSegments) {
    Preconditions.checkArgument(!containsIllegalLabelCharacter(initialLabel));
    this.initialLabel = initialLabel;
    this.descendantSegments = ImmutableList.copyOf(descendantSegments);
  }

  public static boolean containsIllegalLabelCharacter(String label) {
    Preconditions.checkNotNull(label);
    return CharMatcher.anyOf(PathReferences.ILLEGAL_CHARACTERS_LABEL).matchesAnyOf(label);
  }

  public PathReference appendDescendantSegment(Segment segment) {
    List<Segment> newDescendantSegments = Lists.newArrayList(getDescendantSegments());
    newDescendantSegments.add(segment);
    return new PathReference(getInitialLabel(), newDescendantSegments);
  }

  public String getInitialLabel() {
    return initialLabel;
  }

  public List<Segment> getDescendantSegments() {
    return descendantSegments;
  }

  public String toTokenString() {
    return "[" + initialLabel + Joiner.on("").join(descendantSegments) + "]";
  }

  @Override
  public String toString() {
    return "PathReference('" + initialLabel + "', " + descendantSegments + ")";
  }

  /**
   * @param tokenString must be parseable as a valid PathReference, although non-numeric ordinals
   *        will be handled here (correct prior tokenization should ensure that the contract is
   *        satisfied)
   */
  public static @Nonnull PathReference parseTokenString(String tokenString) {
    // Expect and otherwise ignore enclosing brackets, for symmetry with toTokenString.
    Preconditions.checkArgument(tokenString.startsWith("[") && tokenString.endsWith("]"));
    final String s = tokenString.substring(1, tokenString.length() - 1);
    Preconditions.checkArgument(
        CharMatcher.anyOf(PathReferences.ILLEGAL_CHARACTERS_CORE).matchesNoneOf(s));
    final List<Segment> descendantSegments = Lists.newArrayList();
    String initialLabel = "";
    Integer lastMarkerPosition = null;
    /* Iterate one character past the end as a trick to allow the last segment to be extracted with
    the same logic as the earlier ones. */
    for (int i = 0; i < s.length() + 1; i++) {
      if (i == s.length() ||
          s.charAt(i) == PathReferences.LABEL_MARKER ||
          s.charAt(i) == PathReferences.ORDINAL_MARKER)
      {
        if (lastMarkerPosition == null) {
          initialLabel = s.substring(0, i);
        } else {
          final char segmentMarker = s.charAt(lastMarkerPosition);
          final String segmentContent = s.substring(lastMarkerPosition + 1, i);
          final Segment segment;
          if (segmentMarker == PathReferences.LABEL_MARKER) {
            // Ignore extra label markers (like UNIX does with extra path separators).
            segment = segmentContent.isEmpty() ? null : new LabelSegment(segmentContent);
          } else if (segmentMarker == PathReferences.ORDINAL_MARKER) {
            Long ordinal;
            try {
              ordinal = Long.valueOf(segmentContent);
            } catch (NumberFormatException e) {
              ordinal = null;
            }
            segment = new OrdinalSegment(ordinal);
          } else {
            throw new AssertionError();
          }
          if (segment != null)
            descendantSegments.add(segment);
        }
        lastMarkerPosition = i;
      }
    }
    return new PathReference(initialLabel, descendantSegments);
  }

  /* Don't use Strepco here, as the context path used for instance in toPath refers to a per-Segment
  context internal to a PathReference rather than the context that is used at the Expression
  level. */
  static abstract class Segment {
    public static Segment fromPath(View view, Path path) {
      Preconditions.checkArgument(!path.isRoot());
      @Nullable String label = view.getOrNull(CProp.LABEL, path);
      @Nullable Long ordinal = null;
      long i = 1;
      for (Path siblingPath : view.getChildPaths(path.getParent())) {
        if (path.equals(siblingPath)) {
          ordinal = i;
        } else if (label != null) {
          String siblingLabel = view.getOrNull(CProp.LABEL, siblingPath);
          if (siblingLabel != null && siblingLabel.equals(label))
            label = null;
        }
        i++;
      }
      if (ordinal == null)
        throw new AssertionError();
      return (label == null || label.isEmpty() ||
              PathReference.containsIllegalLabelCharacter(label))
          ? new OrdinalSegment(ordinal)
          : new LabelSegment(label);
    }

    public abstract Path toPath(View view, RelationPath contextPath) throws PathReferenceException;

    static class LabelSegment extends Segment {
      private final String label;

      /** @param label may not be empty */
      LabelSegment(String label) {
        Preconditions.checkArgument(!label.isEmpty());
        Preconditions.checkArgument(!PathReference.containsIllegalLabelCharacter(label));
        this.label = label;
      }

      @Override
      public Path toPath(View view, RelationPath contextPath) throws PathReferenceException {
        final List<Path> found = Lists.newArrayList();
        if (!label.isEmpty()) {
          for (Path path : view.getChildPaths(contextPath)) {
            if (label.equals(view.getOrNull(CProp.LABEL, path)))
              found.add(path);
          }
        }
        if (found.isEmpty()) {
          throw new LabelNotAtPathException(view, contextPath, this);
        } else if (found.size() > 1) {
          throw new LabelNotUniqueException(view, contextPath, this);
        } else {
          return found.get(0);
        }
      }

      @Override
      public String toString() {
        return PathReferences.LABEL_MARKER + label;
      }
    }

    static class OrdinalSegment extends Segment {
      private final @Nullable Long ordinal;

      /** @param ordinal may be null to indicate a malformed ordinal */
      OrdinalSegment(@Nullable Long ordinal) {
        this.ordinal = ordinal;
      }

      @Override
      public Path toPath(View view, RelationPath contextPath) throws PathReferenceException {
        PathList childPaths = view.getChildPaths(contextPath);
        if (ordinal == null || ordinal <= 0 || ordinal > childPaths.size())
          throw new OrdinalOutOfRangeException(view, contextPath, this);
        return childPaths.get((int) (ordinal - 1L));
      }

      @Override
      public String toString() {
        return PathReferences.ORDINAL_MARKER + "" + ordinal;
      }
    }
  }
}
