// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.sieuferd.datamodel.Strepco;
import com.google.common.base.Preconditions;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Ordering;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.expressions.PathReference.Segment;
import com.sieuferd.datamodel.expressions.PathReferenceException.LabelNotInScopeException;
import com.sieuferd.datamodel.expressions.PathReferenceException.PrimitivePathDereferencedException;
import com.sieuferd.util.MapUtil;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import javax.annotation.Nullable;

/* TODO[pd]: Have every View store a bidirectional multimap between label and path, and use it to
             avoid indexing the entire view on every path/expression conversion. */
public final class PathReferences {
  public static final char LABEL_MARKER = '\\';
  public static final char ORDINAL_MARKER = '@';
  /** Characters that are illegal between the enclosing brackets of the path reference token. */
  public static final String ILLEGAL_CHARACTERS_CORE = "[]\n";
  public static final String ILLEGAL_CHARACTERS_LABEL =
      ILLEGAL_CHARACTERS_CORE + LABEL_MARKER + ORDINAL_MARKER;

  private PathReferences() { }

  public static Path getPathFromTokenString(Strepco strepco, String tokenString)
      throws PathReferenceException
  {
    /* Special case: Don't allow an empty token string to be used to refer to the root path, as it
    can cause an annoying intermittent highlight to appear while the user is typing a formula in the
    the formula bar (for users, like the author, that like to type "[foo]" by first typing "[]",
    then left-arrowing the caret into the brackets, and finally typing "foo"). To refer to a root
    path with an empty label string, or with a label string containing disallowed characters,
    use "[\]". */
    if (tokenString.equals("[]"))
      throw new LabelNotInScopeException(strepco, "");
    return getPathFromReference(strepco, PathReference.parseTokenString(tokenString));
  }

  public static String getTokenStringFromPath(Strepco strepco, Path path) {
    String ret = createPathReference(strepco, path).toTokenString();
    // Special case; see getPathFromTokenString.
    if (ret.equals("[]"))
      return "[\\]";
    return ret;
  }

  private static PathReference createPathReference(Strepco strepco, Path path) {
    final View view = strepco.getView();
    Preconditions.checkArgument(view.exists(path));
    PathReference ret = createPathReferenceFromContext(
        HashBiMap.create(buildContext(strepco)).inverse(), view, path);
    try {
      if (!getPathFromReference(strepco, ret).equals(path))
        throw new AssertionError("Produced non-reversible path reference");
    } catch (PathReferenceException e) {
      throw new AssertionError("Problem reversing produced path reference", e);
    }
    return ret;
  }

  private static PathReference createPathReferenceFromContext(
      Map<Path,String> inverseContext, View view, Path path)
  {
    final String label = inverseContext.get(path);
    if (label != null && !label.isEmpty() && !PathReference.containsIllegalLabelCharacter(label)) {
      return new PathReference(label, ImmutableList.of());
    } else if (path.isRoot()) {
      return new PathReference("", ImmutableList.of());
    } else {
      return createPathReferenceFromContext(inverseContext, view, path.getParent())
          .appendDescendantSegment(Segment.fromPath(view, path));
    }
  }

  private static Path getPathFromReference(Strepco strepco, PathReference reference)
      throws PathReferenceException
  {
    final String initialLabel = reference.getInitialLabel();
    Path ret;
    if (initialLabel.isEmpty()) {
      ret = Path.ROOT;
    } else {
      ret = buildContext(strepco).get(initialLabel);
      if (ret == null)
        throw new LabelNotInScopeException(strepco, initialLabel);
    }
    final View view = strepco.getView();
    for (Segment segment : reference.getDescendantSegments()) {
      if (ret.isRelation()) {
        final RelationPath previousPath = ret.getRelation();
        ret = segment.toPath(view, previousPath);
      } else {
        throw new PrimitivePathDereferencedException(view, ret.getPrimitive(), segment);
      }
    }
    return ret;
  }

  private static String getLabel(View view, Path path) {
    final String ret = view.getOrNull(CProp.LABEL, path);
    // TODO[cc]: Get an "official" default value.
    return ret == null ? "" : ret;
  }

  private static boolean isVisible(View view, Path path) {
    final Boolean ret = view.getOrNull(CProp.VISIBLE, path);
    // TODO[cc]: Get an "official" default value.
    return ret == null ? true : ret;
  }

  private static Map<String,Path> buildContext(Strepco strepco) {
    final View view = strepco.getView();
    final RelationPath contextPath = strepco.getPath();
    /* The context path and its direct visible children take precedence over other paths in the
    context. Take care to detect ambiguities between sibling labels, however. */
    final Multimap<String,Path> localMap = ArrayListMultimap.create();
    for (Path childPath : view.getChildPaths(contextPath.getRelation())) {
      if (isVisible(view, childPath))
        localMap.put(getLabel(view, childPath), childPath);
    }
    // TODO[pd]: Add the local context as an overlay over the global map, and cache the latter.
    final Map<String,Path> ret = Maps.newLinkedHashMap(buildGlobalContext(view, Path.ROOT));
    for (Entry<String,Collection<Path>> entry : localMap.asMap().entrySet()) {
      ret.remove(entry.getKey());
      if (entry.getValue().size() == 1)
        MapUtil.putChecked(ret, entry.getKey(), entry.getValue().iterator().next());
    }
    ret.put(getLabel(view, contextPath), contextPath);
    return Collections.unmodifiableMap(ret);
  }

  private static Map<String,Path> buildGlobalContext(final View view, RelationPath path) {
    final Multimap<String,Path> allPaths = ArrayListMultimap.create();
    collectAllPaths(view, path, allPaths);
    final Map<String,Path> ret = Maps.newLinkedHashMap();
    for (Entry<String,Collection<Path>> entry : allPaths.asMap().entrySet()) {
      // In case of ambiguity, prefer visible, then shallower paths.
      final Path bestPath = max(entry.getValue(), new Ordering<Path>() {
        @Override
        public int compare(Path p1, Path p2) {
          return (isVisibleToRoot(view, p1) ? 1 : 0) - (isVisibleToRoot(view, p2) ? 1 : 0);
        }
      }.compound(new Ordering<Path>() {
        @Override
        public int compare(Path p1, Path p2) {
          return -(getPathDepth(p1) - getPathDepth(p2));
        }
      }));
      if (bestPath != null)
        MapUtil.putChecked(ret, entry.getKey(), bestPath);
    }
    return Collections.unmodifiableMap(ret);
  }

  private static boolean isVisibleToRoot(View view, Path path) {
    if (isVisible(view, path)) {
      return path.isRoot() || isVisibleToRoot(view, path.getParent());
    } else {
      return false;
    }
  }

  private static int getPathDepth(Path path) {
    return path.isRoot() ? 0 : (getPathDepth(path.getParent()) + 1);
  }

  private static @Nullable <T> T max(
      Collection<? extends T> collection, Comparator<? super T> comparator)
  {
    final Iterator<? extends T> iter = collection.iterator();
    T ret = null;
    boolean tie = false;
    while (iter.hasNext()) {
      final T next = iter.next();
      final int compRes = (ret == null) ? 1 : comparator.compare(next, ret);
      if (compRes > 0) {
        ret = next;
        tie = false;
      } else if (compRes == 0) {
        tie = true;
      }
    }
    return tie ? null : ret;
  }

  private static void collectAllPaths(View view, Path path, Multimap<String,Path> ret) {
    ret.put(getLabel(view, path), path);
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation()))
        collectAllPaths(view, childPath, ret);
    }
  }
}
