// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.expressions.Expression.FunctionExpression;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import javax.annotation.concurrent.Immutable;

/**
 * A set of tuples representable as an {@link Expression}. The iteration order over tuples and
 * paths is preserved as originally supplied, but is not significant for comparison with
 * {@code equals}.
 */
@Immutable
public final class RelationConstant {
  public static final RelationConstant UNIT = new RelationConstant(
      ImmutableSet.of(), ImmutableSet.of(TupleConstant.UNIT));
  private final ImmutableSet<PrimitivePath> paths;
  private final ImmutableSet<TupleConstant> tuples;

  public RelationConstant(Set<PrimitivePath> paths, Set<TupleConstant> tuples) {
    for (TupleConstant tuple : tuples)
      tuple.checkSamePathSet(paths);
    this.paths = tuples.isEmpty() ? ImmutableSet.of(): ImmutableSet.copyOf(paths);
    /* Keep the tuple set even if the set of paths is empty, as it may or may not contain a unit
    tuple. */
    this.tuples = ImmutableSet.copyOf(tuples);
  }

  public Expression toExpression() {
    List<Expression> tupleExpressions = Lists.newArrayList();
    for (TupleConstant tuple : tuples) {
      List<Expression> equiConstraints = Lists.newArrayList();
      for (PrimitivePath path : getPaths()) {
        equiConstraints.add(new FunctionExpression(Function.IS, ImmutableList.of(
            new PathExpression(path), tuple.getValue(path))));
      }
      tupleExpressions.add(Expressions.and(equiConstraints));
    }
    final Expression ret = Expressions.or(tupleExpressions);
    RelationConstant reverse = fromExpression(ret);
    if (!equals(reverse)) {
      throw new AssertionError(
          "Failed to re-parse generated relation constant; got " + reverse + ", expected " + this +
          " (generated expression was " + ret + ")");
    }
    return ret;
  }

  public static RelationConstant fromExpression(Expression expression)
      throws IllegalArgumentException
  {
    final Expression coreExp = Expressions.canonicalize(null, expression);
    final Set<Set<PrimitivePath>> pathSets = Sets.newLinkedHashSet();
    final Set<TupleConstant> tuples = Sets.newLinkedHashSet();
    for (Expression tupleExpression : Expressions.unfoldOr(coreExp)) {
      Map<PrimitivePath,ConstantExpression> tupleMap = Maps.newLinkedHashMap();
      for (Expression equiConstraint : Expressions.unfoldAnd(tupleExpression)) {
        if (equiConstraint instanceof FunctionExpression) {
          FunctionExpression fexp = (FunctionExpression) equiConstraint;
          if (fexp.getFunction().equals(Function.IS)) {
            Expression left = fexp.getArguments().get(0);
            Expression right = fexp.getArguments().get(1);
            if ((left instanceof PathExpression) && (right instanceof ConstantExpression)) {
              Path path = ((PathExpression) left).getPath();
              if (path.isRelation())
                throw new IllegalArgumentException("Path " + path + " is not primitive");
              tupleMap.put(path.getPrimitive(), (ConstantExpression) right);
              continue;
            }
          }
        }
        throw new IllegalArgumentException(
            "Tuple constraint '" + equiConstraint + "' is not of the form <path> is <value>");
      }
      pathSets.add(tupleMap.keySet());
      tuples.add(new TupleConstant(tupleMap));
    }
    if (pathSets.size() > 1) {
      throw new IllegalArgumentException(
          "Tuple constraints must all constrain the same set of paths (found " + pathSets + ")");
    }
    return new RelationConstant(
        pathSets.isEmpty() ? ImmutableSet.of() : pathSets.iterator().next(),
        tuples);
  }

  // TODO[cc]: Consider having clients enforce this before calling the constructor instead.
  /** @return will be empty if the constant contains no tuples */
  public Set<PrimitivePath> getPaths() {
    return paths;
  }

  public Set<TupleConstant> getTuples() {
    return tuples;
  }

  @Override
  public int hashCode() {
    return Objects.hash(paths, tuples);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof RelationConstant))
      return false;
    RelationConstant other = (RelationConstant) obj;
    return Objects.equals(this.getPaths(), other.getPaths())
        && Objects.equals(this.getTuples(), other.getTuples());
  }

  @Override
  public String toString() {
    StringBuilder ret = new StringBuilder();
    if (getTuples().isEmpty()) {
      ret.append("empty relation");
    } else if (getPaths().isEmpty()){
      ret.append("relation with unit tuple");
    } else {
      ret.append("relation with tuples:\n");
      for (PrimitivePath path : getPaths())
        ret.append("\t").append(path);
      ret.append("\n");
      for (TupleConstant tuple : getTuples()) {
        for (PrimitivePath path : getPaths())
          ret.append("\t").append(tuple.getValue(path));
        ret.append("\n");
      }
    }
    return ret.toString();
  }
}
