// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import java.util.Locale;

final class StringEscaper {
  private StringEscaper() { }

  /**
   * Unescape a JSON string literal. This parser accepts some C-style backslash escapes in addition
   * to those supported by JSON, and is also lenient wrt. errors in backslash escapes.
   *
   * @param escaped the escaped string, without quotes
   * @return the unescaped string
   */
  public static String unescape(String escaped) {
    final StringBuilder ret = new StringBuilder(escaped.length());
    for (int i = 0; i < escaped.length(); i++) {
      final char c = escaped.charAt(i);
      if (c == '\\') {
        i++;
        if (i >= escaped.length())
          break;
        final char nc = escaped.charAt(i);
        switch (nc) {
          case 'b':
            ret.append('\b');
          break; case 'f':
            ret.append('\f');
          break; case 'n':
            ret.append('\n');
          break; case 'r':
            ret.append('\r');
          break; case 't':
            ret.append('\t');
          break; case 'a': // Supported in C, but not JSON. Be forgiving and include it anyway.
            ret.append('\u0007');
          break; case 'v': // See above.
            ret.append('\u000b');
          break; case '0': // See above. Actually just a common case of C-style octal literals.
            ret.append('\u0000');
          break; case 'u': {
            final StringBuilder hexString = new StringBuilder(4);
            for (int hi = i + 1; hi < escaped.length() && hexString.length() < 4; hi++) {
              final char hc = Character.toLowerCase(escaped.charAt(hi));
              if (!(hc >= '0' && hc <= '9' || hc >= 'a' && hc <= 'f'))
                break;
              hexString.append(hc);
            }
            if (hexString.length() > 0)
              ret.append((char) Integer.parseInt(hexString.toString(), 16));
            i += hexString.length();
          }
          break; default:
            /* Handles JSON's double quote, backslash, and forward slash escapes, and permits any
            other character to be escaped without causing an error as well. */
            ret.append(nc);
        }
      } else {
        ret.append(c);
      }
    }
    return ret.toString();
  }

  /**
   * Escape a string using JSON string literal rules (RFC 7159).
   *
   * @param protectEndTags if true, escape any forward slash that is preceded by a less-than symbol
   * @return escaped string, without quotes
   */
  public static String escape(String unescaped, boolean protectEndTags) {
    // See https://tools.ietf.org/html/rfc7159 section 7.
    final int LENGTH_ESTIMATE = (int) (1.3 * unescaped.length());
    final StringBuilder ret = new StringBuilder(LENGTH_ESTIMATE);
    boolean escapeForwardSlash = false;
    for (int i = 0; i < unescaped.length(); i++) {
      final char c = unescaped.charAt(i);
      if (c == '\"' || c == '\\' || (escapeForwardSlash && c == '/')) {
        ret.append('\\');
        ret.append(c);
      } else if (c == '\b') {
        ret.append("\\b");
      } else if (c == '\f') {
        ret.append("\\f");
      } else if (c == '\n') {
        ret.append("\\n");
      } else if (c == '\r') {
        ret.append("\\r");
      } else if (c == '\t') {
        ret.append("\\t");
      } else if (Character.isISOControl(c) || c != ' ' && Character.isSpaceChar(c)) {
        ret.append("\\u");
        /* Use uppercase hex digits to differentiate from the initial 'u', and to better visually
        align the top of alpha digits with decimal digits. */
        String unpaddedHex = Integer.toHexString(c).toUpperCase(Locale.US);
        ret.append("0000", 0, 4 - unpaddedHex.length());
        ret.append(unpaddedHex);
      } else {
        ret.append(c);
      }
      /* Escape a forward slash if it's preceded by a less-than sign. See
      http://stackoverflow.com/questions/4264877/why-is-the-slash-an-escapable-character-in-json .*/
      escapeForwardSlash = (protectEndTags && c == '<');
    }
    return ret.toString();
  }
}
