// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.util.MapUtil;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.OffsetDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.chrono.IsoChronology;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.format.DateTimeParseException;
import java.time.format.ResolverStyle;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalAccessor;
import java.time.temporal.TemporalQuery;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;

final class TemporalFormats {
  private TemporalFormats() { }

  /**
   * Like {@link DateTimeFormatter#ISO_LOCAL_DATE_TIME}, but with a space instead of a 'T' character
   * separating the date and the time portion. This is the standard way we render timestamps in
   * expressions.
   */
  public static final DateTimeFormatter ISO_LOCAL_DATE_TIME_WITH_SPACE =
      new DateTimeFormatterBuilder()
      .parseCaseInsensitive()
      .append(DateTimeFormatter.ISO_LOCAL_DATE)
      .appendLiteral(' ') // Use a space instead of a 'T' here.
      .append(DateTimeFormatter.ISO_LOCAL_TIME)
      .toFormatter()
      .withResolverStyle(ResolverStyle.STRICT)
      .withChronology(IsoChronology.INSTANCE);
  /**
   * Like {@link DateTimeFormatter#ISO_OFFSET_DATE_TIME}, but with a space instead of a 'T'
   * character separating the date and the time portion.
   */
  private static final DateTimeFormatter ISO_OFFSET_DATE_TIME_WITH_SPACE =
      new DateTimeFormatterBuilder()
      .parseCaseInsensitive()
      .append(ISO_LOCAL_DATE_TIME_WITH_SPACE)
      .appendOffsetId()
      .toFormatter()
      .withResolverStyle(ResolverStyle.STRICT)
      .withChronology(IsoChronology.INSTANCE);
  /**
   * Like {@link DateTimeFormatter#BASIC_ISO_DATE}, but without the optional time zone offset.
   */
  private static final DateTimeFormatter BASIC_ISO_DATE_WITHOUT_OFFSET =
      new DateTimeFormatterBuilder()
      .parseCaseInsensitive()
      .appendValue(ChronoField.YEAR, 4)
      .appendValue(ChronoField.MONTH_OF_YEAR, 2)
      .appendValue(ChronoField.DAY_OF_MONTH, 2)
      .toFormatter()
      .withResolverStyle(ResolverStyle.STRICT)
      .withChronology(IsoChronology.INSTANCE);

  /** @return either a {@link LocalTime}, a {@link LocalDate}, an {@link Instant}, or null */
  public static @Nullable TemporalAccessor parseTemporal(String s, Locale locale, ZoneId zoneId) {
    return MapUtil.getLoneElementOrNull(parseTemporalSet(s, locale, zoneId));
  }

  private static Set<TemporalAccessor> parseTemporalSet(String s, Locale locale, ZoneId zoneId) {
    Map<DateTimeFormatter,TemporalQuery<TemporalAccessor>> formatters = Maps.newLinkedHashMap();

    formatters.put(DateTimeFormatter.ISO_LOCAL_DATE, LocalDate::from);
    formatters.put(DateTimeFormatter.ISO_LOCAL_TIME, LocalTime::from);
    formatters.put(ISO_LOCAL_DATE_TIME_WITH_SPACE, LocalDateTime::from);

    formatters.put(DateTimeFormatter.ISO_LOCAL_DATE_TIME, LocalDateTime::from);
    formatters.put(DateTimeFormatter.ISO_OFFSET_DATE_TIME, OffsetDateTime::from);
    formatters.put(DateTimeFormatter.RFC_1123_DATE_TIME, OffsetDateTime::from);
    formatters.put(BASIC_ISO_DATE_WITHOUT_OFFSET, LocalDate::from);
    formatters.put(ISO_OFFSET_DATE_TIME_WITH_SPACE, OffsetDateTime::from);

    /*
      Other DateTimeFormatter ISO formats not included, and the rationale for omitting them:
        ISO_DATE
        ISO_OFFSET_DATE
        ISO_TIME
        ISO_OFFSET_TIME
          These include time zone offsets that are not meaningful when only a date or a time is
          present in isolation.
        ISO_WEEK_DATE
          Too obscure, and different countries define the first week of the year differently (even
          though the ISO format picks a particular definition).
        ISO_ORDINAL_DATE
          Too obscure; more likely to be used by accident. (Also includes offset.)
        ISO_ZONED_DATE_TIME
          Includes a time zone specification which would be just be ignored in favor of the time
          zone offset. Allowing the time zone specification to be present suggests at least that
          some error checking is done in it.
        ISO_DATE_TIME
          Covered by ISO_LOCAL_DATE_TIME and ISO_ZONED_DATE_TIME.
        ISO_INSTANT
          Covered by ISO_OFFSET_DATE_TIME.
    */

    final Set<TemporalAccessor> ret = Sets.newLinkedHashSet();
    for (Entry<DateTimeFormatter,TemporalQuery<TemporalAccessor>> entry : formatters.entrySet()) {
      try {
        TemporalAccessor res = entry.getKey().withLocale(locale).parse(s, entry.getValue());
        if (res instanceof LocalDateTime) {
          res = ((LocalDateTime) res).atZone(zoneId).toInstant();
        } else if (res instanceof ZonedDateTime) {
          res = ((ZonedDateTime) res).toInstant();
        } else if (res instanceof OffsetDateTime) {
          res = ((OffsetDateTime) res).toInstant();
        }
        ret.add(res);
      } catch (DateTimeParseException e) {
        // No action.
      }
    }
    return Collections.unmodifiableSet(ret);
  }
}
