// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
/* This file is based in part on:
 *   https://github.com/antlr/grammars-v4/blob/master/java/Java.g4
 *   as retrieved on 2014-01-02.
 * The complete license header for the original file can be found in the LICENSE_Javag4.txt file.
 */
// See also http://docs.oracle.com/javase/specs/jls/se7/html/jls-18.html
// See also http://openjdk.java.net/projects/compiler-grammar/antlrworks/Java.g (ANTLR3 example)

// Prefix the grammar name with "ANTLR" to avoid naming conflicts with my own classes.
grammar ANTLR_Expression;

// Expressions

baseExpression
// Accept an optional equals before a formula, as Excel users may have it in muscle memory.
    :   '='? expression EOF
    ;

literal
    :   NumberLiteral # numberLiteral
    |   StringLiteral # stringLiteral
    |   TemporalLiteral # temporalLiteral
    |   BooleanLiteral # booleanLiteral
    |   NullLiteral # nullLiteral
    |   ErrorLiteral # errorLiteral
    ;

expression
// Accept missing parenthesis at endwithout error (here and in functionExpression).
    :   '(' expression (')' | EOF) # parenExpression
    |   literal # literalExpression
    |   PathReference # pathExpression
    |   Identifier '(' (expression (',' expression)*)? (')' | EOF) # functionExpression

/* Don't use the caret as an operator, as it has different meanings in different languages. This is
the OCaml-style power operator. */
/* "Starting in ANTLR 4.2, the assoc option should be placed on the alternative itself." See
https://github.com/antlr/antlr4/issues/488 . See
https://github.com/antlr/grammars-v4/blob/106c7183c34aa187367b780f6b63b2cdccc9ed0d/r/R.g4 for
an operator precedence example that includes a right-associative power function. */
    |   <assoc=right> aLeft=expression op='**' aRight=expression # binaryOperatorExpression
    |   op=('+'|'-') a=expression # unaryOperatorExpression
/* Logical NOT binds tighter in Java than in SQL; we follow the Java convention. */
    |   op=('!' | KEYW_NOT) a=expression # unaryOperatorExpression
    |   aLeft=expression op=('*'|'/'|'%') aRight=expression # binaryOperatorExpression
    |   aLeft=expression op=('+'|'-') aRight=expression # binaryOperatorExpression
/* Precedence for IS/IS NOT is between the arithmetic operators and the other comparison operators,
like in PostgreSQL: http://www.postgresql.org/docs/8.0/interactive/sql-syntax.html */
    |   aLeft=expression op=KEYW_IS (op2=KEYW_NOT)? aRight=expression # binaryOperatorExpression
    |   aLeft=expression op=('<=' | '>=' | '>' | '<') aRight=expression # binaryOperatorExpression
    |   aLeft=expression op=('==' | '=' | '!=' | '<>') aRight=expression # binaryOperatorExpression
    |   aLeft=expression op=('&&' | KEYW_AND) aRight=expression # binaryOperatorExpression
    |   aLeft=expression op=('||' | KEYW_OR) aRight=expression # binaryOperatorExpression
// The ternary operator is accepted, but never generated.
    |   cond=expression '?' ifTrue=expression ':' ifFalse=expression # conditionalExpression
    ;

// Literals

NumberLiteral
    :   IntegerLiteral
    |   FloatingPointLiteral
    ;

StringLiteral
    :   '\"' DoubleQuotedStringLiteralCore '\"'
    |   '\'' SingleQuotedStringLiteralCore '\''
    ;

// For improved syntax highlighting. This definition must appear after that of StringLiteral.
UnterminatedStringLiteral
    :   '\"' DoubleQuotedStringLiteralCore
    |   '\'' SingleQuotedStringLiteralCore
    ;

TemporalLiteral
    :   '{' TemporalLiteralCore '}'
    ;

// For improved syntax highlighting. This definition must appear after that of TemporalLiteral.
UnterminatedTemporalLiteral
    :   '{' TemporalLiteralCore
    ;

fragment IntegerLiteral
    :   '0'
    |   NonZeroDigit Digits?
    ;

fragment Digits
    :   Digit Digit*
    ;

fragment Digit
    :   '0'
    |   NonZeroDigit
    ;

fragment NonZeroDigit
    :   [1-9]
    ;

/* There is no literal for negative numbers; negative numbers are produced by combining a unary
minus with a positive number literal. If necessary, Expression.canonicalize can be used to fold this
combination into a single ConstantExpression. There are also no literals for NaN or infinity, as
these special values are not used in the language. */
fragment FloatingPointLiteral
    :   Digits '.' Digits? ExponentPart?
    |   '.' Digits ExponentPart?
    |   Digits ExponentPart
    ;

fragment ExponentPart
    :   [eE] [+-]? Digits
    ;

BooleanLiteral
    :   KEYW_TRUE
    |   KEYW_FALSE
    ;

// Excel-style error literal (e.g. "#ref!" for broken references).
ErrorLiteral
    :   '#' Identifier '!'
    ;

NullLiteral
    :   KEYW_NULL
    ;

fragment TemporalLiteralCore
    :   (~[\n\r{}])*
    ;

fragment DoubleQuotedStringLiteralCore
    :   (~["\n\r\\] | ('"' '"') | BackslashEscapeSequence)*
    ;

fragment SingleQuotedStringLiteralCore
    :   (~['\n\r\\] | ('\'' '\'') | BackslashEscapeSequence)*
    ;

// See https://github.com/antlr/grammars-v4/blob/master/json/JSON.g4

fragment BackslashEscapeSequence
    :   '\\' (["\\/bfnrtav0] | ('u' HexDigit HexDigit HexDigit HexDigit))
    ;

fragment HexDigit
    : [0-9a-fA-F]
    ;

PathReference
    :   '[' PathReferenceCore ']'
    ;

// For improved syntax highlighting. This definition must appear after that of PathReference.
UnterminatedPathReference
    :   '[' PathReferenceCore
    ;

/* It's best to disallow newlines inside path reference tokens. This ensures that syntax
highlighting of an UnterminatedPathReference does not include any terminating newline that might be
inserted by the editor infrastructure. */
fragment PathReferenceCore
    :   (~('\n' | '[' | ']'))*
    ;

// Keywords

/* This seems to be the correct way to get case-insensitive keywords; see
"The Definitive ANTLR 4 Reference" by Terence Parr, p. 96. */

KEYW_AND        : [Aa][Nn][Dd];
KEYW_OR         : [Oo][Rr];
KEYW_NOT        : [Nn][Oo][Tt];
KEYW_IS         : [Ii][Ss];
KEYW_NULL       : [Nn][Uu][Ll][Ll];
KEYW_TRUE       : [Tt][Rr][Uu][Ee];
KEYW_FALSE      : [Ff][Aa][Ll][Ss][Ee];

// This definition must appear after all keyword definitions.
Identifier
    :   [_a-zA-Z] [_a-zA-Z0-9]*
    ;

// Operators

/* While certain operators do not have defined meanings (e.g. bitwise operators), keep the lexer
aware of them to improve error behavior, for instance to ensure that '>>' is lexed as a single token
rather than two. */

OP_ASSIGN       : '=';
OP_GT           : '>';
OP_LT           : '<';
OP_BANG         : '!';
OP_TILDE        : '~';
OP_QUESTION     : '?';
OP_COLON        : ':';
OP_EQUAL        : '==';
OP_LE           : '<=';
OP_GE           : '>=';
OP_NEQ_BANG     : '!=';
OP_NEQ_BRACKET  : '<>';
OP_AND          : '&&';
OP_OR           : '||';
OP_ADD          : '+';
OP_SUB          : '-';
OP_MUL          : '*';
OP_POW          : '**';
OP_DIV          : '/';
OP_BITAND       : '&';
OP_BITOR        : '|';
OP_CARET        : '^';
OP_MOD          : '%';
OP_SHIFT_RIGHT  : '>>';
OP_SHIFT_LEFT   : '<<';

/* Do not treat these as separate tokens, as it should be valid to use two unary minus or plus
operators after each other without a separator. Expressions.generateStringInternal depends on this.

OP_INC          : '++';
OP_DEC          : '--';
*/

/* Named separators. Separators only need to be explicitly listed here if their corresponding
constant in ANTLR_ExpressionLexer will be accessed directly (e.g. from ExpressionTokenType). */

SEP_LPAREN      : '(';
SEP_RPAREN      : ')';
SEP_COMMA       : ',';

/* Send whitespace to the hidden channel rather than skipping it altogether, as it needs to have a
token assigned for syntax highlighting purposes. */
WS  :  [ \t\r\n\u000C ]+ -> channel(HIDDEN)
    ;

// This definition must appear after all other definitions.
// See http://stackoverflow.com/questions/24724194/antlr-help-on-lexing-errors-for-a-custom-grammar-example
UNKNOWN_CHAR : . ;
