// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.internal;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.NodeTypeException;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewProperty;
import java.util.Map;
import com.sieuferd.util.MapUtil;
import java.util.NoSuchElementException;
import java.util.Set;
import javax.annotation.concurrent.NotThreadSafe;

/** Mutable. Not thread-safe in the absence of external synchronization, and does not support view
listeners; used only as a delegate. */
@NotThreadSafe
public final class BaseView extends MutableView {
  private final Map<Path,ViewBranch> branches = Maps.newLinkedHashMap(
      ImmutableMap.<Path,ViewBranch>of(Path.ROOT, new ViewBranch(true)));
  private final ViewDefaults defaults;

  public BaseView(ViewDefaults defaults) {
    Preconditions.checkNotNull(defaults);
    Preconditions.checkArgument(!defaults.getProperties().contains(SProp.CHILD_PATHS));
    this.defaults = defaults;
  }

  private ViewBranch getBranch(Path path) {
    Preconditions.checkNotNull(path);
    ViewBranch ret = branches.get(path);
    if (ret == null)
      throw new IllegalArgumentException("Path " + path + " not reachable from this view");
    return ret;
  }

  @Override
  public void performReadTransaction(Runnable runnable) {
    runnable.run();
  }

  @Override
  public void performWriteTransaction(Runnable runnable) {
    runnable.run();
  }

  @Override
  public void addViewListener(Path pathToListenTo, ViewListener listener) {
    throw new UnsupportedOperationException();
  }

  @Override
  public void removeViewListener(ViewListener listener) {
    throw new UnsupportedOperationException();
  }

  @Override
  public boolean exists(Path path) {
    Preconditions.checkNotNull(path);
    return branches.containsKey(path);
  }

  @Override
  public void addPath(Path path) {
    Preconditions.checkNotNull(path);
    Preconditions.checkArgument(!path.isRoot());
    final RelationPath parent = path.getParent();
    if (branches.containsKey(path))
      throw new IllegalStateException("The path " + path.toString(this) + " already exists");
    // The order of the following operations prevents partial mutation and corruption during errors.
    ViewBranch parentBranch = getBranch(parent);
    MapUtil.putChecked(branches, path, new ViewBranch(path.isRelation()));
    parentBranch.addChildPath(path);
    getDefaults().applyToUndefined(this, path);
  }

  /**
   * This implementation requires child paths to be removed by the client prior to path removal.
   */
  @Override
  public void removePath(Path path) {
    Preconditions.checkNotNull(path);
    Preconditions.checkArgument(!path.isRoot());
    if (path.isRelation() && !getChildPaths(path.getRelation()).isEmpty()) {
      throw new UnsupportedOperationException(
          "Path " + path.toString(this) + " still has children; cannot remove");
    }
    // The order of the following operations ensures that failures do not cause corruption.
    RelationPath parent = path.getParent();
    ViewBranch parentBranch = getBranch(parent);
    parentBranch.removeChildPath(path);
    MapUtil.removeChecked(branches, path);
  }

  private void checkAppliesTo(ViewProperty<?> property, Path path) {
    Preconditions.checkNotNull(property);
    Preconditions.checkNotNull(path);
    if (!property.appliesTo(path)) {
      throw new NodeTypeException(
          "Property " + property + " does not apply to " + path.toString(this) +
          " (restricted to " + property.getAppliesTo() + ")");
    }
  }

  @Override
  public <T> T get(ViewProperty<T> property, Path path) {
    checkAppliesTo(property, path);
    T ret = getBranch(path).getProperty(property);
    if (ret == null) {
      throw new NoSuchElementException(
          "No property " + property + " at path " + path.toString(this));
    }
    return ret;
  }

  @Override
  public Set<ViewProperty<?>> getDefinedProperties(Path path) {
    return getBranch(path).getDefinedProperties();
  }

  @Override
  public boolean defines(ViewProperty<?> property, Path path) {
    checkAppliesTo(property, path);
    return getBranch(path).definesProperty(property);
  }

  @Override
  public <T> void set(ViewProperty<T> property, Path path, T value) throws IllegalArgumentException
  {
    checkAppliesTo(property, path);
    if (property.equals(SProp.CHILD_PATHS)) {
      if (!((PathList) value).equalsUnordered(get(SProp.CHILD_PATHS, path))) {
        throw new IllegalArgumentException("Do not set the " + SProp.CHILD_PATHS +
            " property to add or remove nodes; use addPath()/removePath()");
      }
    }
    if (value == null)
      throw new IllegalArgumentException("Null values are not supported for view properties");
    if (!property.getType().getValueClass().isAssignableFrom(value.getClass())) {
      throw new IllegalArgumentException("Can't use object " + value + " as value for property "
          + property + "; " + property.getType() + " required");
    }
    getBranch(path).setProperty(property, value);
    if (property.equals(CProp.LABEL))
      path.updateDebugStringRepresentation(this);
  }

  @Override
  public String toString() {
    return "View(" + Path.ROOT.toString(this) + ")@" + Integer.toHexString(hashCode());
  }

  @Override
  public ViewDefaults getDefaults() {
    return defaults;
  }
}
