// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.internal;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewProperty;
import java.util.Set;
import javax.annotation.concurrent.ThreadSafe;

@ThreadSafe
public final class ConcurrentMutableView extends MutableView {
  private final BaseView delegate;
  private final ViewTransactionManager transactionManager = new ViewTransactionManager(this);

  public ConcurrentMutableView(ViewDefaults defaults) {
    this.delegate = new BaseView(defaults);
    defaults.applyToUndefined(ConcurrentMutableView.this, Path.ROOT);
  }

  @Override
  public void performReadTransaction(Runnable runnable) {
    transactionManager.performTransaction(runnable, true);
  }

  @Override
  public void performWriteTransaction(Runnable runnable) {
    transactionManager.performTransaction(runnable, false);
  }

  @Override
  public synchronized boolean exists(Path path) {
    return delegate.exists(path);
  }

  @Override
  public synchronized <T> T get(ViewProperty<T> property, Path path) {
    return delegate.get(property, path);
  }

  @Override
  public synchronized Set<ViewProperty<?>> getDefinedProperties(Path path) {
    return delegate.getDefinedProperties(path);
  }

  @Override
  public synchronized boolean defines(ViewProperty<?> property, Path path) {
    return delegate.defines(property, path);
  }

  @Override
  public void addPath(final Path path) {
    Preconditions.checkNotNull(path);
    Preconditions.checkArgument(!path.isRoot());
    performWriteTransaction(() -> {
      RelationPath parent = path.getParent();
      final PathList oldValue = delegate.getChildPaths(parent);
      delegate.addPath(path);
      transactionManager.registerEvent(SProp.CHILD_PATHS, parent, oldValue);
    });
  }

  @Override
  public void removePath(final Path path) {
    Preconditions.checkArgument(!path.isRoot());
    performWriteTransaction(() -> {
      removePathsRecursively(path);
    });
  }

  private void removePathsRecursively(final Path path) {
    Preconditions.checkArgument(!path.isRoot());
    Preconditions.checkArgument(exists(path));
    if (path.isRelation()) {
      for (Path childPath : getChildPaths(path.getRelation()))
        removePathsRecursively(childPath);
    }
    final RelationPath parent = path.getParent();
    PathList oldValue = delegate.getChildPaths(parent);
    delegate.removePath(path);
    transactionManager.registerEvent(SProp.CHILD_PATHS, parent, oldValue);
  }

  @Override
  public <T> void set(final ViewProperty<T> property, final Path path, final T value) {
    Preconditions.checkNotNull(property);
    Preconditions.checkNotNull(path);
    Preconditions.checkNotNull(value);
    performWriteTransaction(() -> {
      final T oldValue = delegate.defines(property, path) ? delegate.get(property, path) : null;
      delegate.set(property, path, value);
      transactionManager.registerEvent(property, path, oldValue);
    });
  }

  @Override
  public synchronized void addViewListener(Path pathToListenTo, ViewListener listener) {
    /* While ViewTransactionManager.add/removeViewListener() are themselves thread-safe,
    MutableView.add/removeViewListener() still need to be synchronized on MutableView.this in order
    to stay out of MutableView transactions. */
    transactionManager.getListeners().addViewListener(pathToListenTo, listener);
  }

  @Override
  public synchronized void removeViewListener(ViewListener listener) {
    transactionManager.getListeners().removeViewListener(listener);
  }

  @Override
  public synchronized String toString() {
    return delegate.toString();
  }

  @Override
  public ViewDefaults getDefaults() {
    return delegate.getDefaults();
  }
}
