// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.internal;

import com.google.common.base.Optional;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.util.MapUtil;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

@ThreadSafe
public final class ViewListenerMap {
  private final Multimap<Optional<Path>,ViewListener> listeners = LinkedHashMultimap.create();
  /** Reverse map of listeners. */
  private final Map<ViewListener,Optional<Path>> listenerPaths = Maps.newLinkedHashMap();
  private final Set<ViewListenerMapSnapshot> snapshots = Sets.newLinkedHashSet();

  /**
   * Each listener may only be added once.
   *
   * @param pathToListenTo can be null to listen to all paths.
   */
  public synchronized void addViewListener(@Nullable Path pathToListenTo, ViewListener listener) {
    Preconditions.checkNotNull(listener);
    if (listenerPaths.containsKey(listener))
      throw new IllegalStateException("Each listener may only be added once");
    ensureSnapshotsImmutable();
    Optional<Path> optionalPath = Optional.fromNullable(pathToListenTo);
    MapUtil.putChecked(listeners, optionalPath, listener);
    listenerPaths.put(listener, optionalPath);
    if (listeners.size() != listenerPaths.size())
      throw new AssertionError();
  }

  public synchronized void removeViewListener(ViewListener listener) {
    Preconditions.checkNotNull(listener);
    Optional<Path> path = listenerPaths.remove(listener);
    if (path == null)
      throw new IllegalStateException("Tried to remove non-existent ViewListener");
    ensureSnapshotsImmutable();
    MapUtil.removeChecked(listeners, path, listener);
    if (listeners.size() != listenerPaths.size())
      throw new AssertionError();
  }

  private void ensureSnapshotsImmutable() {
    for (ViewListenerMapSnapshot snapshot : snapshots)
      snapshot.ensureImmutable();
  }

  /** The client must call release() on the returned snapshot once it is no longer needed. */
  public synchronized ViewListenerMapSnapshot createSnapshot() {
    ViewListenerMapSnapshot ret = new ViewListenerMapSnapshot(listeners);
    snapshots.add(ret);
    return ret;
  }

  private void releaseSnapshot(ViewListenerMapSnapshot snapshot) {
    MapUtil.removeChecked(snapshots, snapshot);
  }

  public final class ViewListenerMapSnapshot {
    private Multimap<Optional<Path>,ViewListener> listeners;
    private boolean released;

    /** The client is required to call ensureImmutable() on the created object prior to making any
    modifications to the supplied map. */
    private ViewListenerMapSnapshot(Multimap<Optional<Path>,ViewListener> listeners) {
      this.listeners = listeners;
    }

    private void ensureImmutable() {
      // ImmutableSetMultimap.copyOf() will only actually incur an array copy the first time.
      listeners = ImmutableSetMultimap.copyOf(listeners);
    }

    public void release() {
      synchronized (ViewListenerMap.this) {
        released = true;
        ViewListenerMap.this.releaseSnapshot(this);
      }
    }

    public Iterable<ViewListener> getApplicableListeners(Path path) {
      synchronized (ViewListenerMap.this) {
        if (released)
          throw new IllegalStateException("Snapshot released; cannot get listeners");
        /* TODO[pd]: Consider returning in reverse order, for firing like in the convention seen in
                     javax.swing.event.EventListenerList. */
        List<ViewListener> ret = Lists.newArrayList();
        ret.addAll(listeners.get(Optional.of(path)));
        ret.addAll(listeners.get(Optional.<Path>absent()));
        return Collections.unmodifiableList(ret);
      }
    }
  }
}
