// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.internal;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewPropertyChangedEvent;
import com.sieuferd.datamodel.ViewTransactionCompletedEvent;
import com.sieuferd.datamodel.internal.ViewListenerMap.ViewListenerMapSnapshot;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

@ThreadSafe
final class ViewTransactionManager {
  private static final Logger LOG = Logger.getLogger(ViewTransactionManager.class.getName());
  private final ConcurrentMutableView view;
  private final ViewListenerMap listeners = new ViewListenerMap();
  /** Null iff transactionLevel == 0. */
  private @Nullable TransactionState state;
  private int level = 0;

  ViewTransactionManager(ConcurrentMutableView view) {
    this.view = view;
  }

  /** Transactions are reentrant. */
  public void performTransaction(Runnable runnable, boolean readOnly) {
    TransactionState stateToFire = null;
    Exception propagatedException = null;
    synchronized (view) {
      // Increase transaction level
      if (level == 0) {
        Preconditions.checkState(state == null);
        state = new TransactionState(listeners, readOnly);
      } else {
        Preconditions.checkState(state != null);
        if (state.readOnly && !readOnly) {
          throw new IllegalStateException(
              "Cannot initiate write transaction from within read-only transaction");
        }
      }
      level++;
      // Perform the operation, which may include nested transactions.
      try {
        runnable.run();
      } catch (Exception e) {
        propagatedException = e;
        if (!state.propertyChanges.isEmpty()) {
          LOG.log(Level.SEVERE, "View transaction threw exception after modification; " +
              "view will remain partially updated, but should be internally uncorrupted", e);
        }
      }
      // Decrease transaction level
      if (level <= 0)
        throw new IllegalStateException("Transaction begin/end mismatch");
      Preconditions.checkState(state != null);
      level--;
      if (level == 0) {
        /* Make a copy here just to be absolutely sure we're thread-safe, since we're going to be
        accessing stateToFire after we release the lock on view. */
        stateToFire = new TransactionState(state);
        state = null;
      }
    }
    if (stateToFire != null) {
      try {
        Set<ViewListener> transactionCompletedListeners = Sets.newLinkedHashSet();
        for (PropertyChange<?> propertyChange : stateToFire.propertyChanges.values()) {
          ViewPropertyChangedEvent<?> evt = createViewPropertyChangedEvent(propertyChange);
          if (evt == null)
            continue;
          for (ViewListener listener : stateToFire.listeners.getApplicableListeners(evt.getPath()))
          {
            listener.propertyChanged(evt);
            transactionCompletedListeners.add(listener);
          }
        }
        if (!transactionCompletedListeners.isEmpty()) {
          ViewTransactionCompletedEvent evt = new ViewTransactionCompletedEvent(view);
          for (ViewListener listener : transactionCompletedListeners)
            listener.transactionCompleted(evt);
        }
      } finally {
        stateToFire.listeners.release();
      }
    }
    if (propagatedException != null)
      throw new RuntimeException("Got exception during view transaction", propagatedException);
  }

  private @Nullable <T> ViewPropertyChangedEvent<?> createViewPropertyChangedEvent(
      PropertyChange<T> propertyChange)
  {
    T newValue = view.getOrNull(propertyChange.key.property, propertyChange.key.path);
    if (newValue == null || newValue.equals(propertyChange.oldValue))
      return null;
    return new ViewPropertyChangedEvent<>(view,
        propertyChange.key.property, propertyChange.key.path, propertyChange.oldValue, newValue);
  }

  /** May only be called from within a transaction. */
  public <T> void registerEvent(ViewProperty<T> prop, Path path, T oldValue) {
    synchronized (view) {
      Preconditions.checkState(level > 0);
      Preconditions.checkNotNull(state);
      if (state.readOnly)
        throw new IllegalStateException("Can't mutate View from read-only transaction");
      PropertyChangeKey<T> key = new PropertyChangeKey<>(prop, path);
      if (!state.propertyChanges.containsKey(key))
        state.propertyChanges.put(key, new PropertyChange<>(key, oldValue));
    }
  }

  public boolean isInTransaction() {
    synchronized (view) {
      return level > 0;
    }
  }

  public ViewListenerMap getListeners() {
    return listeners;
  }

  private static final class TransactionState {
    final ViewListenerMapSnapshot listeners;
    final boolean readOnly;
    final Map<PropertyChangeKey<?>,PropertyChange<?>> propertyChanges;

    TransactionState(ViewListenerMap listeners, boolean readOnly) {
      this.listeners = listeners.createSnapshot();
      this.readOnly = readOnly;
      this.propertyChanges = Maps.newLinkedHashMap();
    }

    TransactionState(TransactionState source) {
      this.listeners = source.listeners;
      this.readOnly = source.readOnly;
      this.propertyChanges = ImmutableMap.copyOf(source.propertyChanges);
    }
  }

  private static final class PropertyChange<T> {
    final PropertyChangeKey<T> key;
    final @Nullable T oldValue;

    PropertyChange(PropertyChangeKey<T> key, @Nullable T oldValue) {
      Preconditions.checkNotNull(key);
      this.key = key;
      this.oldValue = oldValue;
    }
  }

  private static final class PropertyChangeKey<T> {
    final ViewProperty<T> property;
    final Path path;

    PropertyChangeKey(ViewProperty<T> property, Path path) {
      Preconditions.checkNotNull(property);
      Preconditions.checkNotNull(path);
      this.property = property;
      this.path = path;
    }

    @Override
    public int hashCode() {
      return Objects.hash(property, path);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof PropertyChangeKey))
        return false;
      PropertyChangeKey<?> other = (PropertyChangeKey<?>) obj;
      return this.property.equals(other.property)
          && this.path.equals(other.path);
    }
  }
}
