// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.datamodel.expressions;

import com.sieuferd.datamodel.Strepco;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.ExpressionParseException;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.PathErrorReport;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.SyntaxErrorReport;
import com.sieuferd.util.ObjectReference;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.List;
import java.util.Locale;
import javax.annotation.Nullable;
import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.Lexer;
import org.antlr.v4.runtime.Token;
import org.junit.Assert;
import org.junit.Test;

public class ExpressionTest {
  private final ImmutableView EMPTY_VIEW;
  private final Strepco EMPTY_VIEW_STREPCO;

  public ExpressionTest() {
    MutableView toEmptyView = Views.newMutable();
    toEmptyView.set(CProp.LABEL, Path.ROOT, "Root");
    EMPTY_VIEW = ImmutableView.copyOf(toEmptyView);
    EMPTY_VIEW_STREPCO = Strepco.createNormalized(EMPTY_VIEW);
  }

  private void checkExactStringGeneration(String stringToParse)
      throws ExpressionParseException
  {
    checkStringGeneration(stringToParse, stringToParse);
  }

  private void checkStringGeneration(String stringToParse, String expectedString)
      throws ExpressionParseException
  {
    Expression exp = Expressions.canonicalize(EMPTY_VIEW,
        Expressions.parse(EMPTY_VIEW_STREPCO, stringToParse));
    Assert.assertEquals(expectedString, Expressions.generateString(EMPTY_VIEW_STREPCO, exp));
  }

  private void checkLocalizedStringGeneration(
      ZoneId timeZoneId, Locale locale, String stringToParse, String expectedString)
      throws ExpressionParseException
  {
    Expression exp = Expressions.canonicalize(EMPTY_VIEW,
        Expressions.parse(EMPTY_VIEW_STREPCO, stringToParse));
    Assert.assertEquals(expectedString, Expressions.generateString(
        new Strepco(EMPTY_VIEW, Path.ROOT, timeZoneId, locale, true), exp));
  }

  private void compareLocalizedInput(
      ZoneId timeZoneId, Locale locale, String stringToParse, String expectedASTStringRep)
      throws ExpressionParseException
  {
    Expression exp = Expressions.parse(
        new Strepco(EMPTY_VIEW, Path.ROOT, timeZoneId, locale, true), stringToParse);
    Expression canonicalizedExp = Expressions.canonicalize(EMPTY_VIEW, exp);
    Expressions.generateString(EMPTY_VIEW_STREPCO, exp);
    Expressions.generateString(EMPTY_VIEW_STREPCO, canonicalizedExp);
    Assert.assertEquals(expectedASTStringRep, exp.toString());
  }

  private void compareASTRepresentation(String stringToParse, String expectedASTStringRep)
      throws ExpressionParseException
  {
    compareASTRepresentation(stringToParse, expectedASTStringRep, false);
  }

  private void compareASTRepresentation(
      String stringToParse, String expectedASTStringRep, boolean canonicalized)
      throws ExpressionParseException
  {
    Expression exp = Expressions.parse(EMPTY_VIEW_STREPCO, stringToParse);
    Expression canonicalizedExp = Expressions.canonicalize(EMPTY_VIEW, exp);
    Expressions.generateString(EMPTY_VIEW_STREPCO, exp);
    Expressions.generateString(EMPTY_VIEW_STREPCO, canonicalizedExp);
    Assert.assertEquals(expectedASTStringRep, (canonicalized ? canonicalizedExp : exp).toString());
  }

  private void compareErrorASTRepresentation(
      String stringToParse, String expectedASTStringRep,
      final boolean expectSyntaxError, final boolean expectPathError)
  {
    compareErrorASTRepresentation(
        EMPTY_VIEW, stringToParse, expectedASTStringRep, expectSyntaxError, expectPathError);
  }

  private void compareErrorASTRepresentation(View view,
      String stringToParse, String expectedASTStringRep,
      final boolean expectSyntaxError, final boolean expectPathError)
  {
    List<ExpressionErrorReport> errorReports = Lists.newArrayList();
    Strepco strepco = Strepco.createNormalized(view);
    Expression exp = Expressions.parse(strepco, stringToParse, errorReports);
    Expressions.generateString(strepco, exp);
    Expressions.generateString(strepco, Expressions.canonicalize(view, exp));
    Assert.assertEquals(expectedASTStringRep, exp.toString());
    boolean containsSyntaxError = false;
    boolean containsPathError = false;
    for (ExpressionErrorReport errorReport : errorReports) {
      if (errorReport instanceof SyntaxErrorReport)
        containsSyntaxError = true;
      if (errorReport instanceof PathErrorReport)
        containsPathError = true;
    }
    Assert.assertEquals(expectSyntaxError, containsSyntaxError);
    Assert.assertEquals(expectPathError, containsPathError);
  }

  private void compareTokenTypes(String stringToTokenize,
      List<ExpressionTokenType> expectedTokenTypes, boolean expectLexerErrors)
  {
    final Lexer lexer = Expressions.getANTLRLexer(new ANTLRInputStream(stringToTokenize));
    final List<ExpressionErrorReport> errors = Lists.newArrayList();
    lexer.addErrorListener(Expressions.createErrorListener(errors));
    final List<ExpressionTokenType> actualTokenTypes = Lists.newArrayList();
    Token token;
    while ((token = lexer.nextToken()).getType() != Lexer.EOF)
      actualTokenTypes.add(Expressions.getTokenType(token));
    if (!expectLexerErrors && !errors.isEmpty())
      throw new AssertionError("Expected no lexer errors, got: " + errors);
    if (expectLexerErrors && errors.isEmpty())
      throw new AssertionError("Expected lexer errors, got none");
    Assert.assertEquals(expectedTokenTypes, actualTokenTypes);
  }

  @Test
  public void testInitialEquals() throws ExpressionParseException {
    compareASTRepresentation("5.4", "Constant(5.4)");
    compareASTRepresentation("=5.4", "Constant(5.4)");
    compareASTRepresentation("550 + 4", "PLUS(Constant(550), Constant(4))");
    compareASTRepresentation("=550 + 4", "PLUS(Constant(550), Constant(4))");
    compareASTRepresentation("550 = 4", "EQUAL(Constant(550), Constant(4))");
    compareASTRepresentation("=550 = 4", "EQUAL(Constant(550), Constant(4))");
    compareASTRepresentation(" = 550 + 4", "PLUS(Constant(550), Constant(4))");
  }

  @Test
  public void testLoneEquals() throws ExpressionParseException {
    compareASTRepresentation("=null", "Constant(null)");
    compareASTRepresentation(" =null", "Constant(null)");
    compareASTRepresentation(" null ", "Constant(null)");

    compareASTRepresentation("", "Constant(null)");
    compareASTRepresentation(" ", "Constant(null)");
    compareASTRepresentation(" ", "Constant(null)");

    compareASTRepresentation("=", "Constant(null)");
    compareASTRepresentation("= ", "Constant(null)");
    compareASTRepresentation(" = ", "Constant(null)");
  }

  private void testJsonLiteral(String expected, String toParse, String explanation)
      throws ExpressionParseException
  {
    compareASTRepresentation(toParse, "Constant(" + expected + ")");
  }

  @Test
  public void testLiterals() throws ExpressionParseException {
    compareASTRepresentation("5.4", "Constant(5.4)");
    compareASTRepresentation("5e3", "Constant(5E+3)");
    compareASTRepresentation("5.42e3", "Constant(5.42E+3)");
    compareASTRepresentation("5.42E3", "Constant(5.42E+3)");
    compareASTRepresentation("5.420E+3", "Constant(5420)");
    compareASTRepresentation("5.42000E+3", "Constant(5420.00)");
    compareASTRepresentation("5420.0", "Constant(5420.0)");
    compareASTRepresentation("5420.0000", "Constant(5420.0000)");
    compareASTRepresentation("5.42e-3", "Constant(0.00542)");
    compareASTRepresentation("  5.4  ", "Constant(5.4)");
    compareASTRepresentation("  5.4000  ", "Constant(5.4000)");
    compareASTRepresentation("  5e3  ", "Constant(5E+3)");
    compareASTRepresentation("5", "Constant(5)");
    compareASTRepresentation("0", "Constant(0)");
    compareASTRepresentation("0.0", "Constant(0.0)");
    compareASTRepresentation("0.", "Constant(0)");
    compareASTRepresentation("550.", "Constant(550)");
    compareASTRepresentation("550.0", "Constant(550.0)");
    compareASTRepresentation("550", "Constant(550)");
    compareASTRepresentation("null", "Constant(null)");
    compareASTRepresentation("true", "Constant(true)");
    compareASTRepresentation("false", "Constant(false)");
    compareASTRepresentation("#ref!", "#BROKEN_REFERENCE!");
    compareASTRepresentation("#huh!", "#SYNTAX_ERROR!");
    compareASTRepresentation("#REF!", "#BROKEN_REFERENCE!");
    compareASTRepresentation("#HUH!", "#SYNTAX_ERROR!");
    compareASTRepresentation("#rEf!", "#BROKEN_REFERENCE!");
    compareASTRepresentation("#HuH!", "#SYNTAX_ERROR!");

    // Double-quoted strings
    compareASTRepresentation("\"\"", "Constant()");
    compareASTRepresentation("\"t\"", "Constant(t)");
    compareASTRepresentation("\"test\"", "Constant(test)");
    compareASTRepresentation(" \"test\" ", "Constant(test)");
    compareASTRepresentation(" \"t'est\" ", "Constant(t'est)");
    compareASTRepresentation(" \"te\\\"st\" ", "Constant(te\"st)");
    compareASTRepresentation(" \"te\"\"st\" ", "Constant(te\"st)");
    compareASTRepresentation(" \"te\"\"\" ", "Constant(te\")");
    compareASTRepresentation(" \"te\"\"\"\"\" ", "Constant(te\"\")");
    compareASTRepresentation("\"a\"\"sdf\"", "Constant(a\"sdf)");

    // Tests from https://github.com/bestiejs/json3/blob/master/test/test_json3.js
    testJsonLiteral("value", "\"value\"", "Double-quoted string literal");
    testJsonLiteral("", "\"\"", "Empty string literal");
    testJsonLiteral("\u2028", "\"\\u2028\"", "String containing an escaped Unicode line separator");
    testJsonLiteral("\u2029", "\"\\u2029\"", "String containing an escaped Unicode paragraph separator");
    testJsonLiteral("\ud834\udf06", "\"\ud834\udf06\"", "String containing an unescaped Unicode surrogate pair");
    testJsonLiteral("\u0001", "\"\\u0001\"", "String containing an escaped ASCII control character");
    testJsonLiteral("\b", "\"\\b\"", "String containing an escaped backspace");
    testJsonLiteral("\f", "\"\\f\"", "String containing an escaped form feed");
    testJsonLiteral("\n", "\"\\n\"", "String containing an escaped line feed");
    testJsonLiteral("\r", "\"\\r\"", "String containing an escaped carriage return");
    testJsonLiteral("\t", "\"\\t\"", "String containing an escaped tab");
    testJsonLiteral("hello/world", "\"hello\\/world\"", "String containing an escaped solidus");
    testJsonLiteral("hello\\world", "\"hello\\\\world\"", "String containing an escaped reverse solidus");
    testJsonLiteral("hello\"world", "\"hello\\\"world\"", "String containing an escaped double-quote character");
    /* TODO[te]: Do the control character tests from
                 https://github.com/bestiejs/json3/blob/master/test/test_json3.js */

    compareASTRepresentation("\"\'\"", "Constant(')");
    compareASTRepresentation("'\"'", "Constant(\")");
    compareASTRepresentation("'\"\"\"'", "Constant(\"\"\")");
    compareASTRepresentation("\"This is a \"\"test\"\".\"", "Constant(This is a \"test\".)");
    compareASTRepresentation("\"This is a \\\"test\\\".\"", "Constant(This is a \"test\".)");
    compareASTRepresentation("\"This is a \\\"test\"\".\"", "Constant(This is a \"test\".)");
    compareASTRepresentation("\"This is a \\\\\\\"test\"\".\"", "Constant(This is a \\\"test\".)");

    // Single-quoted strings
    compareASTRepresentation("\'\'", "Constant()");
    compareASTRepresentation("\'t\'", "Constant(t)");
    compareASTRepresentation("\'test\'", "Constant(test)");
    compareASTRepresentation(" \'test\' ", "Constant(test)");
    compareASTRepresentation(" \'t''est\' ", "Constant(t'est)");
    compareASTRepresentation(" \'te\"\"st\' ", "Constant(te\"\"st)");
    compareASTRepresentation(" \'te\"\"\' ", "Constant(te\"\")");
    compareASTRepresentation(" \'te\"\"\"\"\' ", "Constant(te\"\"\"\")");
    compareASTRepresentation("\'a\"\"sdf\'", "Constant(a\"\"sdf)");

    compareASTRepresentation("\'a\\nsdf\'", "Constant(a\nsdf)");
    compareASTRepresentation("\'a\"sdf\'", "Constant(a\"sdf)");
    compareASTRepresentation("\'a\"\"sdf\'", "Constant(a\"\"sdf)");
    compareASTRepresentation("\'a\"\"s''df\'", "Constant(a\"\"s'df)");

    // Newline characters in string literal tokens are no longer permitted.
    //compareASTRepresentation("\"te\nst\"", "Constant(te\nst)");
    //compareASTRepresentation("\'te\nst\'", "Constant(te\nst)");

    /* Make sure we don't have a bug like https://hibernate.atlassian.net/browse/HHH-241 . The
    string is "Chartest «ÆØÅÉÈÑÜæøåéèñü¢~§'» máx", but spell it out in unicode literals to ensure
    build/test stability even if the compiler's character encoding setting is wrong. */
    final String complicatedUnicodeString =
        "Chartest \u00ab\u00c6\u00d8\u00c5\u00c9\u00c8\u00d1\u00dc\u00e6\u00f8\u00e5\u00e9\u00e8" +
        "\u00f1\u00fc\u00a2\u007e\u00a7\u0027\u00bb\u0020\u006d\u00e1\u0078";
    compareASTRepresentation(
        "\"" + complicatedUnicodeString + "\"", "Constant(" + complicatedUnicodeString + ")");
    // Fun with Unicode code points.
    // See http://stackoverflow.com/questions/1029897/comparing-a-char-to-a-code-point
    compareASTRepresentation(
        "\"\uD835\uDD0A\"", "Constant(\uD835\uDD0A)");
    compareASTRepresentation(
        "\"\uD835\uDD0A\"\"\"", "Constant(\uD835\uDD0A\")");
  }

  @Test
  public void testTemporalLiterals() throws ExpressionParseException {
    compareASTRepresentation("{2015-05-08}", "Constant(2015-05-08)");
    compareASTRepresentation("{20150508}", "Constant(2015-05-08)");

    compareASTRepresentation("{20111203}", "Constant(2011-12-03)");
    compareASTRepresentation("{2007-12-03}", "Constant(2007-12-03)");
    compareASTRepresentation("{10:15}", "Constant(10:15)");
    compareASTRepresentation("{10:15:14}", "Constant(10:15:14)");
    compareASTRepresentation("{10:15:14.1234}", "Constant(10:15:14.123400)");
    compareASTRepresentation("{2007-12-03 10:15}", "Constant(2007-12-03T10:15:00Z)");
    compareASTRepresentation("{2007-12-03T10:15}", "Constant(2007-12-03T10:15:00Z)");
    compareASTRepresentation("{2007-12-03 10:15:30}", "Constant(2007-12-03T10:15:30Z)");
    compareASTRepresentation("{2007-12-03T10:15:30}", "Constant(2007-12-03T10:15:30Z)");
    compareASTRepresentation("{2007-12-03 10:15:30.1234}", "Constant(2007-12-03T10:15:30.123400Z)");
    compareASTRepresentation("{2007-12-03T10:15:30.1234}", "Constant(2007-12-03T10:15:30.123400Z)");
    compareASTRepresentation("{2007-12-03T10:15:30.1234+03:00}", "Constant(2007-12-03T07:15:30.123400Z)"); // ISO_ZONED_DATE_TIME/ISO_OFFSET_DATE_TIME
    compareASTRepresentation("{2007-12-03 10:15:30.1234+03:00}", "Constant(2007-12-03T07:15:30.123400Z)"); // ISO_OFFSET_DATE_TIME_WITH_SPACE
    compareASTRepresentation("{2011-12-03T10:15:30Z}", "Constant(2011-12-03T10:15:30Z)"); // ISO_INSTANT
    compareASTRepresentation("{2011-12-03 10:15:30Z}", "Constant(2011-12-03T10:15:30Z)"); // ISO_INSTANT
    compareASTRepresentation("{Tue, 3 Jun 2008 11:05:30 GMT}", "Constant(2008-06-03T11:05:30Z)"); // RFC_1123_DATE_TIME

    // Test trimming.
    compareASTRepresentation("{ 2007-12-03 }", "Constant(2007-12-03)");
    compareASTRepresentation("{ 10:15 }", "Constant(10:15)");
    compareASTRepresentation("{ 10:15:14 }", "Constant(10:15:14)");
    compareASTRepresentation("{ 10:15:14.1234 }", "Constant(10:15:14.123400)");
    compareASTRepresentation("{ 2007-12-03 10:15 }", "Constant(2007-12-03T10:15:00Z)");
    // Test collapsing of duplicate spaces.
    compareASTRepresentation("{ 2007-12-03  10:15 }", "Constant(2007-12-03T10:15:00Z)");
    compareASTRepresentation("{ 2007-12-03   10:15 }", "Constant(2007-12-03T10:15:00Z)");

    // Formats unsupported by design.
    compareErrorASTRepresentation("{2011-12-03T10:15:30+01:00[Europe/Paris]}", "#SYNTAX_ERROR!", true, false); // ISO_ZONED_DATE_TIME
    compareErrorASTRepresentation("{2011-12-03T10:15:30+07:00[Europe/Paris]}", "#SYNTAX_ERROR!", true, false); // ISO_ZONED_DATE_TIME
    compareErrorASTRepresentation("{2007-12-03T10:15:30.1234+00:00[UTC]}", "#SYNTAX_ERROR!", true, false); // ISO_ZONED_DATE_TIME
    compareErrorASTRepresentation("{2011-12-03+01:00}", "#SYNTAX_ERROR!", true, false); // ISO_OFFSET_DATE
    compareErrorASTRepresentation("{10:15+01:00}", "#SYNTAX_ERROR!", true, false); // ISO_OFFSET_TIME

    // Other errors.
    compareErrorASTRepresentation("{}", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("{  }", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("{asdfasdfa}", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("{10:15:14.1234567891}", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("{2007-12-03 10:15:14.1234567891}", "#SYNTAX_ERROR!", true, false);

    // Generation.
    checkExactStringGeneration("{2007-12-03}");
    checkExactStringGeneration("{2007-12-03 10:15:00}");
    checkStringGeneration("{10:15}", "{10:15:00}");
    checkExactStringGeneration("{10:15:00}");
    // Make sure only necessary fractional digits are generated.
    checkExactStringGeneration("{10:15:14}");
    checkStringGeneration("{10:15:14.10}", "{10:15:14.1}");
    checkExactStringGeneration("{10:15:14.1}");
    checkExactStringGeneration("{10:15:14.12}");
    checkExactStringGeneration("{10:15:14.123}");
    checkExactStringGeneration("{10:15:14.1234}");
    checkExactStringGeneration("{10:15:14.12345}");
    checkExactStringGeneration("{10:15:14.123456}");
    checkExactStringGeneration("{10:15:14.1234567}");
    checkExactStringGeneration("{10:15:14.12345678}");
    checkExactStringGeneration("{10:15:14.123456789}");

    // Check for correct time zone interpretation.
    Locale locale = Locale.US;
    /* New York at summer time (set it to a constant value to avoid having the unit tests depend on
    the local time zone database). */
    ZoneId zoneId = ZoneOffset.ofHours(-4);
    // These two print the same value: ("-04:00")
    if (false) {
      System.out.println(zoneId);
      System.out.println(ZoneId.of("America/New_York").getRules().getOffset(
          LocalDateTime.parse("2015-07-08T10:00:00")));
    }

    compareLocalizedInput(zoneId, locale, "{2015-05-08}", "Constant(2015-05-08)"); // Not affected.
    compareLocalizedInput(zoneId, locale, "{10:15:14.1234}", "Constant(10:15:14.123400)"); // Not affected.
    compareLocalizedInput(zoneId, locale, "{2007-12-03 10:15}", "Constant(2007-12-03T14:15:00Z)"); // Affected.
    compareLocalizedInput(zoneId, locale, "{2007-12-03T10:15:30.1234+03:00}", "Constant(2007-12-03T07:15:30.123400Z)"); // Not affected.
    compareLocalizedInput(zoneId, locale, "{2007-12-03 10:15:30.1234+03:00}", "Constant(2007-12-03T07:15:30.123400Z)"); // Not affected.

    checkLocalizedStringGeneration(zoneId, locale, "{2015-05-08}", "{2015-05-08}"); // Not affected by time zone.
    checkLocalizedStringGeneration(zoneId, locale, "{10:15:14.1234}", "{10:15:14.1234}"); // Not affected by time zone.
    compareASTRepresentation("{2007-12-03 14:15}", "Constant(2007-12-03T14:15:00Z)");
    checkLocalizedStringGeneration(zoneId, locale,
        "{2007-12-03 14:15:00}", // Input: UTC.
        "{2007-12-03 10:15:00}" // Generated: New York time.
    );
    compareASTRepresentation("{2007-12-03T10:15:30.1234+03:00}", "Constant(2007-12-03T07:15:30.123400Z)");
    checkLocalizedStringGeneration(zoneId, locale, "{2007-12-03T10:15:30.1234+03:00}", "{2007-12-03 03:15:30.1234}");
    checkLocalizedStringGeneration(zoneId, locale, "{2007-12-03 10:15:30.1234+03:00}", "{2007-12-03 03:15:30.1234}");
  }

  @Test
  public void testCompoundConstants() throws ExpressionParseException {
    // First check that the "canonicalized" option works as expected.
    compareASTRepresentation("5.4", "Constant(5.4)", false);
    compareASTRepresentation("5.4", "Constant(5.4)", true);
    compareASTRepresentation("(5.4)", "(Constant(5.4))", false);
    compareASTRepresentation("(5.4)", "Constant(5.4)", true);

    // Make sure the canonicalizer isn't doing anything crazy.
    compareASTRepresentation("0 / 5", "DIVIDE(Constant(0), Constant(5))", false);
    compareASTRepresentation("0 / 5", "DIVIDE(Constant(0), Constant(5))", true);
    compareASTRepresentation("5 / 0", "DIVIDE(Constant(5), Constant(0))", false);
    compareASTRepresentation("5 / 0", "DIVIDE(Constant(5), Constant(0))", true);

    /* Now check for proper canonicalization of constants that may not be directly representable by
    a literal alone. */
    compareASTRepresentation("-5.4", "UNARY_MINUS(Constant(5.4))", false);
    compareASTRepresentation("-5.4", "Constant(-5.4)", true);

    /* Test bug that once resulted from two concatenated unary minus signs being interpreted as a
    decrement token. Double-minus and double-plus are no longer considered separate tokens. */
    compareASTRepresentation("- -5.4", "UNARY_MINUS(UNARY_MINUS(Constant(5.4)))", false);
    compareASTRepresentation("--5.4", "UNARY_MINUS(UNARY_MINUS(Constant(5.4)))", false);
    compareASTRepresentation("- -5.4", "Constant(5.4)", true);
    compareASTRepresentation("--5.4", "Constant(5.4)", true);
    compareASTRepresentation("+ +5.4", "UNARY_PLUS(UNARY_PLUS(Constant(5.4)))", false);
    compareASTRepresentation("++5.4", "UNARY_PLUS(UNARY_PLUS(Constant(5.4)))", false);
    // Canonicalization does currently not do anything about the unary plus operator.
    // compareASTRepresentation("+ +5.4", "Constant(5.4)", true);
    // compareASTRepresentation("++5.4", "Constant(5.4)", true);

    /* These used to be translated to floating-point NaN or infinity values, but I have since
    changed the language design to not actually provide any representation of floating point
    specials. */
    compareASTRepresentation("0 / 0", "DIVIDE(Constant(0), Constant(0))", false);
    compareASTRepresentation("0 / 0", "DIVIDE(Constant(0), Constant(0))", true);
    //compareASTRepresentation("0 / 0", "Constant(NaN)", true);
    compareASTRepresentation("1 / 0", "DIVIDE(Constant(1), Constant(0))", false);
    compareASTRepresentation("1 / 0", "DIVIDE(Constant(1), Constant(0))", true);
    //compareASTRepresentation("1 / 0", "Constant(Infinity)", true);
    compareASTRepresentation("-1 / 0", "DIVIDE(UNARY_MINUS(Constant(1)), Constant(0))", false);
    compareASTRepresentation("-1 / 0", "DIVIDE(Constant(-1), Constant(0))", true);
    //compareASTRepresentation("-1 / 0", "Constant(-Infinity)", true);

    // Compound constants wrapped in other expressions.
    compareASTRepresentation("5 == -5.4", "EQUAL(Constant(5), UNARY_MINUS(Constant(5.4)))", false);
    compareASTRepresentation("5 == -5.4", "EQUAL(Constant(5), Constant(-5.4))", true);
    compareASTRepresentation("5 == 0 / 0",
        "EQUAL(Constant(5), DIVIDE(Constant(0), Constant(0)))", false);

    // No special handling of floating point specials anymore; see above.
    compareASTRepresentation("5 == 0 / 0",
        "EQUAL(Constant(5), DIVIDE(Constant(0), Constant(0)))", true);
    //compareASTRepresentation("5 == 0 / 0", "EQUAL(Constant(5), Constant(NaN))", true);
    compareASTRepresentation("5 == 1 / 0",
        "EQUAL(Constant(5), DIVIDE(Constant(1), Constant(0)))", false);
    compareASTRepresentation("5 == 1 / 0",
        "EQUAL(Constant(5), DIVIDE(Constant(1), Constant(0)))", true);
    //compareASTRepresentation("5 == 1 / 0", "EQUAL(Constant(5), Constant(Infinity))", true);

    // String generation.
    checkExactStringGeneration("0 / 0");
    checkExactStringGeneration("1 / 0");
    checkExactStringGeneration("-1 / 0");
    checkExactStringGeneration("5 + 0 / 0 + 5");
    checkExactStringGeneration("not (0 / 0)");
    checkExactStringGeneration("not 0 / 0");
    checkStringGeneration("(not 0) / 0", "not 0 / 0");
  }

  @Test
  public void testMissingEndParenthesis() throws ExpressionParseException {
    compareASTRepresentation("5", "Constant(5)");
    compareASTRepresentation("(5)", "(Constant(5))");
    compareASTRepresentation("((5))", "((Constant(5)))");
    // Missing end parenthesis.
    compareASTRepresentation("(5", "(Constant(5))");
    compareASTRepresentation("((5", "((Constant(5)))");
    compareASTRepresentation("((5)", "((Constant(5)))");
    compareASTRepresentation("(5 + 4", "(PLUS(Constant(5), Constant(4)))");
    // Missing end parenthesis in function.
    compareASTRepresentation("COUNT(5", "COUNT(Constant(5))");
    compareASTRepresentation("COUNT((5", "COUNT((Constant(5)))");
    compareASTRepresentation("COUNT((5)", "COUNT((Constant(5)))");
    compareASTRepresentation("COUNT(5 + 4", "COUNT(PLUS(Constant(5), Constant(4)))");
    compareASTRepresentation("(COUNT(5 + 4", "(COUNT(PLUS(Constant(5), Constant(4))))");
    compareASTRepresentation("COUNT((5 + 4", "COUNT((PLUS(Constant(5), Constant(4))))");
    compareASTRepresentation("CONCAT(5, 3", "STRING_CONCAT(Constant(5), Constant(3))");
    compareASTRepresentation("CONCAT(5, (3", "STRING_CONCAT(Constant(5), (Constant(3)))");
    compareASTRepresentation("CONCAT(5, (3)", "STRING_CONCAT(Constant(5), (Constant(3)))");
    compareErrorASTRepresentation(
        "CONCAT((5 + 4,", "STRING_CONCAT((PLUS(Constant(5), Constant(4))))", true, false);
    /* Other recovered ASTs could be valid here; this is a syntax error. The exact recovered AST
    sometimes changes as ANTLR versions are bumped. */
    compareErrorASTRepresentation("CONCAT((5 + 4, 3",
        "STRING_CONCAT((PLUS(Constant(5), Constant(4))), Constant(3))", true, false);
  }

  @Test
  public void testLiteralsInExpressions() throws ExpressionParseException {
    compareASTRepresentation("550 == 4", "EQUAL(Constant(550), Constant(4))");
    compareASTRepresentation("550 = 4", "EQUAL(Constant(550), Constant(4))");
    compareASTRepresentation("550 != 4", "NOT_EQUAL(Constant(550), Constant(4))");
    compareASTRepresentation("550 <> 4", "NOT_EQUAL(Constant(550), Constant(4))");
    compareASTRepresentation("550 + 4", "PLUS(Constant(550), Constant(4))");
    compareASTRepresentation("550 + 4.0", "PLUS(Constant(550), Constant(4.0))");
    compareASTRepresentation("550 + \"asdf\"", "PLUS(Constant(550), Constant(asdf))");
    compareASTRepresentation("\"asdf2\" + \"asdf\"", "PLUS(Constant(asdf2), Constant(asdf))");
    compareASTRepresentation("\"asdf2\" + \"asdf\"", "PLUS(Constant(asdf2), Constant(asdf))");
    compareASTRepresentation("5e2 + 235233222 + \"asdf\"",
        "PLUS(PLUS(Constant(5E+2), Constant(235233222)), Constant(asdf))");
  }

  @Test
  public void testOverlappingTokens() throws ExpressionParseException {
    // Make sure there are no problems with tokens containing similar characters as each other.
    compareASTRepresentation("5 <= 4", "LESS_OR_EQUAL(Constant(5), Constant(4))");
    compareASTRepresentation("5 == 4", "EQUAL(Constant(5), Constant(4))");
    compareASTRepresentation("5 < 4", "LESS(Constant(5), Constant(4))");
  }

  @Test
  public void testAutoParenthesization() throws ExpressionParseException {
    checkExactStringGeneration("2");
    checkExactStringGeneration("2 + 3");
    /* This test coincidentally tests interaction between compound constants and
    auto-parenthesization. */
    checkExactStringGeneration("2 + -3");
    checkStringGeneration("2 + (-3)", "2 + -3");
    checkStringGeneration("2 + -3 ** 4", "2 + -3 ** 4");
    checkStringGeneration("2 + -(3 ** 4)", "2 + -3 ** 4");
    checkStringGeneration("(-3) ** 4", "(-3) ** 4");
    checkStringGeneration("-(3 ** 4)", "-3 ** 4");

    checkStringGeneration("2 + 3 * 4", "2 + 3 * 4");
    checkStringGeneration("(2 + 3) * 4", "(2 + 3) * 4");
    checkStringGeneration("2 + (3 * 4)", "2 + 3 * 4");

    // Left-associativity
    checkStringGeneration("2 + 3 + 4 + 5", "2 + 3 + 4 + 5");
    checkStringGeneration("(2 + 3) + 4 + 5", "2 + 3 + 4 + 5");
    checkStringGeneration("2 + 3 + (4 + 5)", "2 + 3 + (4 + 5)");

    // Right-associativity
    checkStringGeneration("2 ** 3 ** 4 ** 5", "2 ** 3 ** 4 ** 5");
    checkStringGeneration("(2 ** 3) ** 4 ** 5", "(2 ** 3) ** 4 ** 5");
    checkStringGeneration("2 ** 3 ** (4 ** 5)", "2 ** 3 ** 4 ** 5");

    // No ambiguity in any of these; parentheses disappear.
    checkStringGeneration("not -3", "not -3");
    checkStringGeneration("not (-3)", "not -3");
    checkStringGeneration("-(not 3)", "-not 3");
  }

  @Test
  public void testMiscStringGeneration() throws ExpressionParseException {
    checkStringGeneration("5 + (2 ? 3 : 4)", "5 + if(2, 3, 4)");
    checkExactStringGeneration("5 + if(2, 3, 4)");
    checkStringGeneration("5 + 2 ? 3 : 4", "if(5 + 2, 3, 4)");
  }

  @Test
  public void testStringLiteralGeneration() throws ExpressionParseException {
    checkStringGeneration("\"hello\"", "\"hello\"");
    checkStringGeneration("\"hel\tlo\"", "\"hel\\tlo\"");
    checkStringGeneration("\"hel\\tlo\"", "\"hel\\tlo\"");
    checkStringGeneration("\"\u2709\"", "\"\u2709\"");
    checkStringGeneration("\"\u270a\"", "\"\u270a\"");
    checkStringGeneration("\"\u270A\"", "\"\u270a\"");
    final String complicatedUnicodeString =
        "Chartest \u00ab\u00c6\u00d8\u00c5\u00c9\u00c8\u00d1\u00dc\u00e6\u00f8\u00e5\u00e9\u00e8" +
        "\u00f1\u00fc\u00a2\u007e\u00a7\u0027\u00bb\u0020\u006d\u00e1\u0078";
    checkStringGeneration(
        "\"" + complicatedUnicodeString + "\"",
        "\"" + complicatedUnicodeString + "\"");
    final String longCodePoint = "\uD835\uDD0A";
    checkStringGeneration(
        "\"" + longCodePoint + "\"",
        "\"" + longCodePoint + "\"");
    checkStringGeneration("\"Abc\\tDef\"", "\"Abc\\tDef\"");
    checkStringGeneration("\"Abc\tDef\"", "\"Abc\\tDef\"");

    checkStringGeneration("\"Abc\\bDef\"", "\"Abc\\bDef\"");
    checkStringGeneration("\"Abc\\fDef\"", "\"Abc\\fDef\"");
    checkStringGeneration("\"Abc\\nDef\"", "\"Abc\\nDef\"");
    checkStringGeneration("\"Abc\\rDef\"", "\"Abc\\rDef\"");
    checkStringGeneration("\"Abc\\\"Def\"", "\"Abc\\\"Def\"");

    // Escape codes supported in C but not JSON.
    checkStringGeneration("\"Abc\\aDef\"", "\"Abc\\u0007Def\"");
    checkStringGeneration("\"Abc\\vDef\"", "\"Abc\\u000BDef\"");
    checkStringGeneration("\"Abc\\0Def\"", "\"Abc\\u0000Def\"");

    // Might be supported in the future.
    // checkStringGeneration("\"Abc\\'Def\"", "\"Abc'Def\"");
    checkStringGeneration("\"Abc\\/Def\"", "\"Abc/Def\"");
    checkStringGeneration("\"Abc\\u005CDef\"", "\"Abc\\\\Def\"");
    /* Control character that must be encoded as a hex code escape. Includes the hex digit "A" to
    check capitalization. */
    checkStringGeneration("\"Abc\\u008ADef\"", "\"Abc\\u008ADef\"");

    /* JSON generators will sometime escape the forward slash if preceded by a less-than symbol.
    Make sure this is _not_ done for our own string literal generator. */
    checkStringGeneration("\"a/b/c\"", "\"a/b/c\"");
    checkStringGeneration("\"</script>\"", "\"</script>\"");
  }

  @Test
  public void testPrecedence() throws ExpressionParseException {
    // Left-associative division operator.
    compareASTRepresentation("5 + 4 / 3 / 2",
        "PLUS(Constant(5), DIVIDE(DIVIDE(Constant(4), Constant(3)), Constant(2)))");
    // Right-associative power operator.
    compareASTRepresentation("5 + 4 ** 3 ** 2",
        "PLUS(Constant(5), POWER(Constant(4), POWER(Constant(3), Constant(2))))");

    compareASTRepresentation("5/4 * 2",
        "MULTIPLY(DIVIDE(Constant(5), Constant(4)), Constant(2))");
    compareASTRepresentation("5 / 4 / 2",
        "DIVIDE(DIVIDE(Constant(5), Constant(4)), Constant(2))");
    compareASTRepresentation("5 / 4/2",
        "DIVIDE(DIVIDE(Constant(5), Constant(4)), Constant(2))");

    // Power has higher precedence than even unary operators; compare with plus and multiply.
    compareASTRepresentation("-2 + 3",
        "PLUS(UNARY_MINUS(Constant(2)), Constant(3))");
    compareASTRepresentation("-2 * 3",
        "MULTIPLY(UNARY_MINUS(Constant(2)), Constant(3))");
    compareASTRepresentation("-2 ** 3",
        "UNARY_MINUS(POWER(Constant(2), Constant(3)))");

    // Standard stuff
    compareASTRepresentation("5 + 4", "PLUS(Constant(5), Constant(4))");
    compareASTRepresentation("5 + 4 * 8", "PLUS(Constant(5), MULTIPLY(Constant(4), Constant(8)))");
    compareASTRepresentation("4 * 8 + 5", "PLUS(MULTIPLY(Constant(4), Constant(8)), Constant(5))");
    compareASTRepresentation("(5 + 4)", "(PLUS(Constant(5), Constant(4)))");
    compareASTRepresentation("5 + (4 * 8)",
        "PLUS(Constant(5), (MULTIPLY(Constant(4), Constant(8))))");
    compareASTRepresentation("(5 + 4) * 8",
        "MULTIPLY((PLUS(Constant(5), Constant(4))), Constant(8))");
    compareASTRepresentation("2 + 5 * (2)",
        "PLUS(Constant(2), MULTIPLY(Constant(5), (Constant(2))))");

    compareASTRepresentation("5 || 4",
        "LOGICAL_OR(Constant(5), Constant(4))");
    compareASTRepresentation("5 || 4 && 8",
        "LOGICAL_OR(Constant(5), LOGICAL_AND(Constant(4), Constant(8)))");
    compareASTRepresentation("4 && 8 || 5",
        "LOGICAL_OR(LOGICAL_AND(Constant(4), Constant(8)), Constant(5))");
    compareASTRepresentation("(5 || 4)",
        "(LOGICAL_OR(Constant(5), Constant(4)))");
    compareASTRepresentation("5 || (4 && 8)",
        "LOGICAL_OR(Constant(5), (LOGICAL_AND(Constant(4), Constant(8))))");
    compareASTRepresentation("(5 || 4) && 8",
        "LOGICAL_AND((LOGICAL_OR(Constant(5), Constant(4))), Constant(8))");
  }

  @Test
  public void testIs() throws ExpressionParseException {
    compareASTRepresentation("5 is 2",
        "IS(Constant(5), Constant(2))");
    compareASTRepresentation("5 is not 2",
        "IS_NOT(Constant(5), Constant(2))");
    compareASTRepresentation("5 is (not 2)",
        "IS(Constant(5), (LOGICAL_NOT(Constant(2))))");
  }

  @Test
  public void testKeywordOperatorsAndCaseInsensitivity() throws ExpressionParseException {
    compareASTRepresentation("not 5 or 4",
        "LOGICAL_OR(LOGICAL_NOT(Constant(5)), Constant(4))");
    compareASTRepresentation("nOT 5 or 4",
        "LOGICAL_OR(LOGICAL_NOT(Constant(5)), Constant(4))");
    compareASTRepresentation("5 or 4",
        "LOGICAL_OR(Constant(5), Constant(4))");
    compareASTRepresentation("5 or 4 AND 8",
        "LOGICAL_OR(Constant(5), LOGICAL_AND(Constant(4), Constant(8)))");
    compareASTRepresentation("4 and 8 or 5",
        "LOGICAL_OR(LOGICAL_AND(Constant(4), Constant(8)), Constant(5))");
    compareASTRepresentation("(5 oR 4)",
        "(LOGICAL_OR(Constant(5), Constant(4)))");
    compareASTRepresentation("5 Or (4 AND 8)",
        "LOGICAL_OR(Constant(5), (LOGICAL_AND(Constant(4), Constant(8))))");
    compareASTRepresentation("(5 || 4) aNd 8",
        "LOGICAL_AND((LOGICAL_OR(Constant(5), Constant(4))), Constant(8))");
  }

  @Test
  public void testFunctions() throws ExpressionParseException {
    compareASTRepresentation("somefun(1,2,3)", "somefun(Constant(1), Constant(2), Constant(3))");
    compareASTRepresentation("somefun(1,2)", "somefun(Constant(1), Constant(2))");
    compareASTRepresentation("somefun(1)", "somefun(Constant(1))");
    compareASTRepresentation("somefun()", "somefun()");
    compareASTRepresentation("(somefun())", "(somefun())");
    compareASTRepresentation("5 + (somefun())", "PLUS(Constant(5), (somefun()))");
    compareASTRepresentation("5 + somefun()", "PLUS(Constant(5), somefun())");
    compareASTRepresentation("5 + somefun() - (8)",
        "MINUS(PLUS(Constant(5), somefun()), (Constant(8)))");
    compareASTRepresentation("somefun(somefun2(1))", "somefun(somefun2(Constant(1)))");

    // Named built-ins.
    compareASTRepresentation("avg(1)", "AVERAGE(Constant(1))");
    compareASTRepresentation("count(2)", "COUNT(Constant(2))");
    compareASTRepresentation("group_concat('j', 1)", "GROUP_CONCAT(Constant(j), Constant(1))");
  }

  @Test
  public void testFunctionCaseInsensitivity() throws ExpressionParseException {
    compareASTRepresentation("someFun(1,2,3)", "somefun(Constant(1), Constant(2), Constant(3))");
    compareASTRepresentation("SOMEFUN(1,2)", "somefun(Constant(1), Constant(2))");
    compareASTRepresentation("__someFun24(1)", "__somefun24(Constant(1))");
  }

  @Test
  public void testPrecedencePotpourri() throws ExpressionParseException {
    {
      final String expression = "8 + 4 * 2 / asdf(2,3,null,4 + 8) + true && false + 5 + 5.0";
      final String namedFunctionAST =
          "asdf(Constant(2), Constant(3), Constant(null), PLUS(Constant(4), Constant(8)))";
      final String andLeftExpressionAST =
          "PLUS(PLUS(Constant(8), DIVIDE(MULTIPLY(Constant(4), Constant(2)), " + namedFunctionAST +
          ")), Constant(true))";
      final String andRightExpressionAST =
          "PLUS(PLUS(Constant(false), Constant(5)), Constant(5.0))";
      compareASTRepresentation(expression,
          "LOGICAL_AND(" + andLeftExpressionAST + ", " + andRightExpressionAST + ")");
    }
    compareASTRepresentation("5 + 4 * 8 == 3 > 4 * 2",
        "EQUAL(PLUS(Constant(5), MULTIPLY(Constant(4), Constant(8))), " +
        "GREATER(Constant(3), MULTIPLY(Constant(4), Constant(2))))");
    // Logical NOT binds tighter in Java than in SQL; we follow the Java convention.
    compareASTRepresentation("not 5 = 3", "EQUAL(LOGICAL_NOT(Constant(5)), Constant(3))");
    compareASTRepresentation("not 5 <> 3", "NOT_EQUAL(LOGICAL_NOT(Constant(5)), Constant(3))");
    compareASTRepresentation("not 5 and 3", "LOGICAL_AND(LOGICAL_NOT(Constant(5)), Constant(3))");
    compareASTRepresentation("not 5 or 3", "LOGICAL_OR(LOGICAL_NOT(Constant(5)), Constant(3))");
  }

  private static void compareReference(View view, String referenceExpression, Path expectedPath)
      throws ExpressionParseException
  {
    Expression exp = Expressions.parse(Strepco.createNormalized(view), referenceExpression);
    if (!(exp instanceof PathExpression))
      throw new AssertionError("Expected PathExpression, got " + exp);
    Assert.assertTrue(exp instanceof PathExpression);
    PathExpression pathExpression = (PathExpression) exp;
    Assert.assertEquals(expectedPath, pathExpression.getPath());
  }

  @Test
  public void testRootPathExpressions() throws ExpressionParseException {
    MutableView view = Views.newMutable();
    view.set(CProp.LABEL, Path.ROOT, "");
    compareReference(view, "[\\]", Path.ROOT);
    view.set(CProp.LABEL, Path.ROOT, "Root");
    compareReference(view, "[Root]", Path.ROOT);
    Assert.assertEquals("PLUS(" + Path.ROOT.toString() + ", Constant(4))",
        Expressions.parse(Strepco.createNormalized(view), "[Root] + 4").toString());
    view.set(CProp.LABEL, Path.ROOT, "Root[foo]");
    compareReference(view, "[\\]", Path.ROOT);
    view.set(CProp.LABEL, Path.ROOT, "[Foo]");
    compareReference(view, "[\\]", Path.ROOT);

    view.set(CProp.LABEL, Path.ROOT, "Foo \"Test\" Jaa");
    compareReference(view, "[Foo \"Test\" Jaa]", Path.ROOT);
    view.set(CProp.LABEL, Path.ROOT, "fooroot");
    compareReference(view, "[fooroot]", Path.ROOT);
    /* A previous version of the grammar used the slash as a path label separator, which resulted in
    an ambiguity with the division operator. This test will serve as a reminder not to reintroduce
    this design mistake. */
    compareASTRepresentation("[Root] / [Root]", "DIVIDE(" + Path.ROOT + ", " + Path.ROOT + ")");
  }

  @Test
  public void testDeepPathExpressions() throws ExpressionParseException {
    MutableView view = Views.newMutable();
    view.set(CProp.LABEL, Path.ROOT, "Root");
    PrimitivePath rA = view.addPrimitivePath(Path.ROOT);
    PrimitivePath rB = view.addPrimitivePath(Path.ROOT);
    PrimitivePath rC = view.addPrimitivePath(Path.ROOT);
    PrimitivePath rD = view.addPrimitivePath(Path.ROOT);
    view.set(CProp.LABEL, rA, "Child1");
    view.set(CProp.LABEL, rB, "Child2");
    view.set(CProp.LABEL, rC, "ChildAmbig");
    view.set(CProp.LABEL, rD, "ChildAmbig");
    RelationPath rE = view.addRelationPath(Path.ROOT);
    view.set(CProp.LABEL, rE, "Child4");
    PrimitivePath rEA = view.addPrimitivePath(rE);
    PrimitivePath rEB = view.addPrimitivePath(rE);
    PrimitivePath rEC = view.addPrimitivePath(rE);
    PrimitivePath rED = view.addPrimitivePath(rE);
    RelationPath rEE = view.addRelationPath(rE);
    RelationPath rEF = view.addRelationPath(rE);
    view.set(CProp.LABEL, rEA, "GrandSonAmbig");
    view.set(CProp.LABEL, rEB, "GrandSonAmbig");
    view.set(CProp.LABEL, rEC, "GrandSon3");
    view.set(CProp.LABEL, rED, "GrandSon4");
    view.set(CProp.LABEL, rEE, "GrandSon5");
    view.set(CProp.LABEL, rEF, "GrandSonAmbig");
    PrimitivePath rEFA = view.addPrimitivePath(rEF);
    PrimitivePath rEFB = view.addPrimitivePath(rEF);
    view.set(CProp.LABEL, rEFA, "GreatGrandSonAmbig");
    view.set(CProp.LABEL, rEFB, "GreatGrandSonAmbig");
    PrimitivePath rEFC = view.addPrimitivePath(rEF);
    PrimitivePath rEFD = view.addPrimitivePath(rEF);
    PrimitivePath rEFE = view.addPrimitivePath(rEF);
    view.set(CProp.LABEL, rEFC, "GreatGrandSon3");
    view.set(CProp.LABEL, rEFD, "GreatGrandSo\nn4[");
    view.set(CProp.LABEL, rEFE, "");

    compareReference(view, "[Root]", Path.ROOT);
    compareReference(view, "[Root\\Child1]", rA);
    compareReference(view, "[Root\\Child2]", rB);
    compareReference(view, "[Root@3]", rC);
    compareReference(view, "[Root@4]", rD);
    compareReference(view, "[Child1]", rA);
    compareReference(view, "[Child2]", rB);

    compareReference(view, "[Root\\Child4]", rE);
    compareReference(view, "[Child4]", rE);

    compareReference(view, "[Root\\Child4@1]", rEA);
    compareReference(view, "[Child4@1]", rEA);
    compareReference(view, "[Root\\Child4@2]", rEB);
    compareReference(view, "[Child4@2]", rEB);
    compareReference(view, "[Root\\Child4\\GrandSon3]", rEC);
    compareReference(view, "[Child4\\GrandSon3]", rEC);
    compareReference(view, "[Root\\Child4\\GrandSon4]", rED);
    compareReference(view, "[Child4\\GrandSon4]", rED);
    compareReference(view, "[Root\\Child4\\GrandSon5]", rEE);
    compareReference(view, "[Child4\\GrandSon5]", rEE);
    compareReference(view, "[GrandSon5]", rEE);
    compareReference(view, "[Root\\Child4@6]", rEF);
    compareReference(view, "[Child4@6]", rEF);
    compareReference(view, "[Root\\Child4@6@1]", rEFA);
    compareReference(view, "[Root\\Child4@6@2]", rEFB);
    compareReference(view, "[Root\\Child4@6@3]", rEFC);
    compareReference(view, "[Root\\Child4@6@4]", rEFD);
    compareReference(view, "[Root\\Child4@6@5]", rEFE);

    // Check ordinal boundaries and illegal ordinals.
    compareErrorASTRepresentation(
        "[@0]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[@999999999999999999999999999999999999999999999999]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[@9999]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[Root\\Child4@6@-5]", "#BROKEN_REFERENCE!", false, true);
      compareErrorASTRepresentation(
        "[Root\\Child4@6@asdf]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[Root\\Child4@6@0]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[Root\\Child4@6@5]", "#BROKEN_REFERENCE!", false, true);

    compareReference(view, "[Root\\Child4@6\\GreatGrandSon3]", rEFC);

    /* Extraneous label separators should be ignored. (In a previous version of the path reference
    system, empty labels would be legal in label segments, but should not actually match any paths,
    even if an empty label were present.) */
    compareReference(view, "[Root\\Child4@6\\]", rEF);
    compareReference(view, "[Root\\Child4@1\\]", rEA);
    compareReference(view, "[Child4@1\\]", rEA);
    compareReference(view, "[Root\\\\Child4@2\\]", rEB);
    compareReference(view, "[Child4@2\\]", rEB);
    compareReference(view, "[\\Child4\\\\\\GrandSon3]", rEC);
    compareReference(view, "[Child4\\\\GrandSon3\\\\]", rEC);
    compareReference(view, "[Root\\\\\\Child4\\GrandSon4]", rED);
    compareReference(view, "[Child4\\\\GrandSon4]", rED);
    compareReference(view, "[Root\\\\\\\\Child4\\\\GrandSon5\\]", rEE);
    compareReference(view, "[\\]", Path.ROOT);
    compareReference(view, "[\\\\]", Path.ROOT);
    compareReference(view, "[@3]", rC);
    compareReference(view, "[@4]", rD);
    compareReference(view, "[\\@3]", rC);
    compareReference(view, "[\\\\@4]", rD);

    // Newlines and open brackets used to be permitted in label segments, but not anymore.
    compareErrorASTRepresentation(
        "[Root\\Child4@6\\GreatGrandSon4[]", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation(
        "[Root\\Child4@6\\GreatGrandSo\nn4]", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation(
        "[Root\\Child4@6\\GreatGrandSo\nn4[]", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation(
        "[Ro[ot\\Child4@6", "#SYNTAX_ERROR!", true, false);
  }

  @Test
  public void testLiteralErrors() throws ExpressionParseException {
    // No errors.
    compareErrorASTRepresentation("5 + 3", "PLUS(Constant(5), Constant(3))", false, false);
    /* These used to produce errors, but no longer do now that BigDecimal is used to represent
    numeric literals. Keep the tests here to verify that there are no errors. */
    compareErrorASTRepresentation(
        "5000000000000000000000 + 3", "PLUS(Constant(5000000000000000000000), Constant(3))",
        false, false);
    compareErrorASTRepresentation(
        "3 + 5000000000000000000000", "PLUS(Constant(3), Constant(5000000000000000000000))",
        false, false);
    // Error literal for unknown error code.
    compareErrorASTRepresentation("#asdf!", "#SYNTAX_ERROR!", true, false);
    // Incorrect String literal quoting (lexer-level error).
    compareErrorASTRepresentation("'asdf", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("='asdf", "#SYNTAX_ERROR!", true, false);

    compareErrorASTRepresentation("\"\"\"", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("'''", "#SYNTAX_ERROR!", true, false);
    compareErrorASTRepresentation("'\\u000q'", "#SYNTAX_ERROR!", true, false);
    // Not currently allowed, but might be allowed by a lenient parser in the future.
    compareErrorASTRepresentation("\"\\'\"", "#SYNTAX_ERROR!", true, false);
    /* Don't accept unescaped double quotes if the string contains any unix-escaped quotes.
    Actually, this is now allowed. */
    // compareErrorASTRepresentation("\'a\\\"\"sdf\'", "#SYNTAX_ERROR!", true, false);
    compareASTRepresentation("\'a\\\"\"sdf\'", "Constant(a\"\"sdf)");
  }

  @Test
  public void testPathErrors() {
    // Try the old syntax.
    compareErrorASTRepresentation(
        "[Roo3t]\\[Foo]#3", "#BROKEN_REFERENCE!", true, true);
    // This becomes ROOT after parse error recovery (not really a requirement, and might change).
    compareErrorASTRepresentation(
        "[Root]\\[Foo]#3", Path.ROOT.toString(), true, false);

    // Ordinal path segment.
    compareErrorASTRepresentation(
        "[Root@3]", "#BROKEN_REFERENCE!", false, true);
    /* This used to be a syntax error due to integer overflow, but is now a regular broken
    reference. */
    compareErrorASTRepresentation(
        "[Root@5000000000000000000000]", "#BROKEN_REFERENCE!", false, true);
    // An empty initial segment is interpreted as a root reference.
    compareErrorASTRepresentation(
        "[@2]", "#BROKEN_REFERENCE!", false, true);
    // Extraneous label separators are ignored.
    compareErrorASTRepresentation(
        "[\\]", Path.ROOT.toString(), false, false);
    compareErrorASTRepresentation(
        "[\\\\]", Path.ROOT.toString(), false, false);
    // Empty brackets are never allowed to match a path (special case).
    compareErrorASTRepresentation(
        "[]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[Root@5000000000000000000000]", "#BROKEN_REFERENCE!", false, true);

    compareErrorASTRepresentation(
        "[Foobar@5000000000000000000000]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[Foobar@5]", "#BROKEN_REFERENCE!", false, true);

    // Simple reference error (first try one with no error)
    compareErrorASTRepresentation("[Root]", Path.ROOT.toString(), false, false);
    compareErrorASTRepresentation("[Foobar]", "#BROKEN_REFERENCE!", false, true);
    compareErrorASTRepresentation(
        "[Foobar] + 5", "PLUS(#BROKEN_REFERENCE!, Constant(5))", false, true);
  }

  @Test
  public void testParseErrorsWithRecovery() {
    /* This test might break if ANTLR changes its specific recovery strategy. The main point is to
    test that we get some expression back and that an error is reported. */
    compareErrorASTRepresentation("asdf", "asdf()", true, false);
    compareErrorASTRepresentation("()", "(#SYNTAX_ERROR!)", true, false);
    compareErrorASTRepresentation(
        "3 + 5 2", "PLUS(Constant(3), Constant(5))", true, false);
    compareErrorASTRepresentation(
        "2 3 + 5", "PLUS(Constant(2), Constant(5))", true, false);
    /* Cases where ASTBuildingExpressionVisitor.accept would return the value from
    ASTBuildingExpressionVisitor.defaultResult (with some analogous non-error cases as well). */
    compareErrorASTRepresentation("somefunc(5,2,6 9,12)",
        "somefunc(Constant(5), Constant(2), Constant(6), Constant(12))", true, false);
    compareErrorASTRepresentation("somefunc(5,2,,12)",
        "somefunc(Constant(5), Constant(2), #SYNTAX_ERROR!, Constant(12))", true, false);
    compareErrorASTRepresentation("5 + (2 ? 3 : 4)",
        "PLUS(Constant(5), (CONDITIONAL(Constant(2), Constant(3), Constant(4))))", false, false);
    compareErrorASTRepresentation("5 + (2 ? : 4)",
        "PLUS(Constant(5), (CONDITIONAL(Constant(2), #SYNTAX_ERROR!, Constant(4))))", true, false);
    compareErrorASTRepresentation("5 + (2 ? #huh! : 4)",
        "PLUS(Constant(5), (CONDITIONAL(Constant(2), #SYNTAX_ERROR!, Constant(4))))", false, false);
    compareErrorASTRepresentation(
        "2 + (5 +)", "PLUS(Constant(2), (PLUS(Constant(5), #SYNTAX_ERROR!)))", true, false);
    // Empty input. Used to be an error; now permitted as a shorthand for a null literal.
    if (false) {
      compareErrorASTRepresentation("", "#SYNTAX_ERROR!", true, false);
      compareErrorASTRepresentation(" ", "#SYNTAX_ERROR!", true, false);
    }

    // No longer a syntax error; the open parenthesis will be automatically closed at EOF.
    compareErrorASTRepresentation("2 + 5 * (2",
        "PLUS(Constant(2), MULTIPLY(Constant(5), (Constant(2))))", false, false);
  }

  @Test
  public void testConditionalNPEBug() {
    compareErrorASTRepresentation("2 +", "PLUS(Constant(2), #SYNTAX_ERROR!)", true, false);
    // Test for a bug that once caused a NullPointerException in response to the input "true?".
    compareErrorASTRepresentation("true?",
        "CONDITIONAL(Constant(true), #SYNTAX_ERROR!, #SYNTAX_ERROR!)", true, false);
  }

  @Test
  public void testTokenTypes() {
    /* Note: Due to the presence of the UNKNOWN_CHAR lexer definition, the lexer itself should not
             be expected to report errors in any of the following tests. */
    compareTokenTypes("\n", ImmutableList.of(
        ExpressionTokenType.WHITESPACE), false);
    compareTokenTypes("  \n", ImmutableList.of(
        ExpressionTokenType.WHITESPACE), false);
    compareTokenTypes("2 + 5 [foo]", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.PATH_REFERENCE
        ), false);
    compareTokenTypes("2+5[foo]", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.PATH_REFERENCE
        ), false);
    compareTokenTypes("2 + 5 [foo 2", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.UNTERMINATED_PATH_REFERENCE
        ), false);
    compareTokenTypes("5 [foo\n2", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.UNTERMINATED_PATH_REFERENCE,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER
        ), false);
    compareTokenTypes("5 [foo[2]'f'", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.UNTERMINATED_PATH_REFERENCE,
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.LITERAL_STRING
        ), false);
    // The lexer's current error handling strategy simply ignores unrecognized characters.
    compareTokenTypes(" 2 + 5 ] +5", ImmutableList.of(
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OTHER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.LITERAL_NUMBER
        ), false);
    compareTokenTypes(
        "=if([name_first] is not null, concat('Something: ', 123.45 * [id]), [name_last])",
        ImmutableList.of(
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.IDENTIFIER,
          ExpressionTokenType.PARENTHESIS_LEFT,
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR, // is
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR, // not
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NULL,
          ExpressionTokenType.ARGUMENT_SEPARATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.IDENTIFIER,
          ExpressionTokenType.PARENTHESIS_LEFT,
          ExpressionTokenType.LITERAL_STRING,
          ExpressionTokenType.ARGUMENT_SEPARATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.PARENTHESIS_RIGHT,
          ExpressionTokenType.ARGUMENT_SEPARATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.PARENTHESIS_RIGHT
        ), false);
    compareTokenTypes(
        "blah ) ( #huh! 0.0    #ref! and 'string'>>'asdf' ",
        ImmutableList.of(
          ExpressionTokenType.IDENTIFIER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.PARENTHESIS_RIGHT,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.PARENTHESIS_LEFT,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_ERROR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_ERROR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_STRING,
          /* ">>" should be a single unknown token, since we define lexer rules for unused operators
          that are nevertheless common in other languages. */
          ExpressionTokenType.OTHER,
          ExpressionTokenType.LITERAL_STRING,
          ExpressionTokenType.WHITESPACE
        ), false);
    compareTokenTypes("][test]", ImmutableList.of(
          ExpressionTokenType.OTHER,
          ExpressionTokenType.PATH_REFERENCE
        ), false);
    compareTokenTypes("5+'unterminated string", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.UNTERMINATED_LITERAL_STRING
        ), false);
    compareTokenTypes("5+'unterm\"inated st''ring", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.UNTERMINATED_LITERAL_STRING
        ), false);
    compareTokenTypes("5+\"unterminated string", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.UNTERMINATED_LITERAL_STRING
        ), false);
    compareTokenTypes("5+\"unterm'inated st\"\"ring", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.UNTERMINATED_LITERAL_STRING
        ), false);
    compareTokenTypes("5+[ref1]/[ref2", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.UNTERMINATED_PATH_REFERENCE
        ), false);
    compareTokenTypes("test]]][test]'asdfasdf'", ImmutableList.of(
          ExpressionTokenType.IDENTIFIER,
          ExpressionTokenType.OTHER,
          ExpressionTokenType.OTHER,
          ExpressionTokenType.OTHER,
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.LITERAL_STRING
        ), false);
    compareTokenTypes("5+{unterminated temporal", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("5+{terminated temporal}", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("5+{}", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("5+{f}", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("+{f}+", ImmutableList.of(
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.LITERAL_TEMPORAL,
          ExpressionTokenType.OPERATOR
        ), false);
    compareTokenTypes("{}", ImmutableList.of(
          ExpressionTokenType.LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("{", ImmutableList.of(
          ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("{     ", ImmutableList.of(
          ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("{  }   ", ImmutableList.of(
          ExpressionTokenType.LITERAL_TEMPORAL,
          ExpressionTokenType.WHITESPACE
        ), false);
    compareTokenTypes("{{", ImmutableList.of(
          ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL,
          ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("[hello {whoo} yeah] {hello [whoo] yeah}", ImmutableList.of(
          ExpressionTokenType.PATH_REFERENCE,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.LITERAL_TEMPORAL
        ), false);
    compareTokenTypes("{hello [whoo] yeah} [hello {whoo} yeah]", ImmutableList.of(
          ExpressionTokenType.LITERAL_TEMPORAL,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.PATH_REFERENCE
        ), false);
    compareTokenTypes("5 + }1", ImmutableList.of(
          ExpressionTokenType.LITERAL_NUMBER,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OPERATOR,
          ExpressionTokenType.WHITESPACE,
          ExpressionTokenType.OTHER,
          ExpressionTokenType.LITERAL_NUMBER
        ), false);
  }

  private void expectReadOneStringLiteral(String toTokenize, int expectedLength,
      @Nullable String expectedUnescaped)
  {
    final ObjectReference<Integer> actualLength = ObjectReference.create();
    String actualUnescaped = Expressions.readOneStringLiteral(toTokenize, actualLength);
    Assert.assertEquals(expectedLength, (int) actualLength.getValue());
    Assert.assertEquals(expectedUnescaped, actualUnescaped);
  }

  @Test
  public void testReadOneStringLiteral() {
    expectReadOneStringLiteral("hello", 0, null);
    expectReadOneStringLiteral("'hello'", 7, "hello");
    expectReadOneStringLiteral("   'hello'", 0, null);
    expectReadOneStringLiteral("'hello", 6, "hello");
    expectReadOneStringLiteral("'hel\\nlo'", 9, "hel\nlo");
    expectReadOneStringLiteral("'hel\nlo'", 4, "hel");
    expectReadOneStringLiteral("'hello' yeah", 7, "hello");
    expectReadOneStringLiteral("'' yeah", 2, "");
    expectReadOneStringLiteral("\"\" yeah", 2, "");
    expectReadOneStringLiteral("5\"\\\"\" yeah", 0, null);
    expectReadOneStringLiteral("'", 1, "");
  }
}
