// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.cell.styles.LineStyle;
import com.sieuferd.cell.support.Points;
import com.sieuferd.datamodel.proptypes.OptionalBigDecimal;
import com.sieuferd.datamodel.proptypes.OptionalInteger;
import java.awt.Color;
import java.awt.Font;
import java.math.BigDecimal;

public final class DefaultStyleSheets {
  private DefaultStyleSheets() { }
  /* Spreadsheet cell heights, including top border:
     * This system with the settings below, 96dpi: 20px
     * This system with the settings below, 72dpi: 15px
     * Microsoft Excel 2010 for Windows, 100%: 20px
     * Microsoft Excel 2010 for Windows, 75%: 15px
     * Microsoft Excel 2011 for Mac, 100%: 15px
     * Microsoft Excel 2011 for Mac, Retina display, 100%: 30px
     * Google Docs Spreadsheets (Chrome for Windows, 100%): 17px

     Cell cursors widths and overlap with area inside cell borders (all zoom levels):
     * Microsoft Excel 2010 for Windows: 3px wide, 1px internal/1px external overlap.
     * Microsoft Excel 2011 for Mac: 2px wide, 1px internal/0px external overlap.
     * Microsoft Excel 2011 for Mac, Retina display: 4px wide, 2px internal/0px external overlap.
   */

  /* TODO[cc]: Avoid using LineStyle and BorderStyle objects with null colors as a means to store a
               set of border dimensions. */

  public static final ViewDefaults COLOR = buildColorStyleSheet();
  public static final ViewDefaults GRAYSCALE = buildGrayscaleStyleSheet();
  public static final ViewDefaults DEBUG = buildDebugStyleSheet();
  public static final ViewDefaults DATA_COLLECTION = buildDataCollectionStyleSheet();

  private static ViewDefaults buildColorStyleSheet() {
    final Color COLOR_SEP         = new Color(  0,   0,   0,  80);
    final Color COLOR_SEP_LIGHTER = new Color(  0,   0,   0,  20);

    ViewDefaults.Builder ret =  ViewDefaults.builder();
    ret.set(LBProp.COLOR_VALUE,               new Color(255, 232, 185, 255));
    ret.set(LBProp.COLOR_VALUE_ALTERNATIVE,     new Color(255, 237, 204, 255));
    //ret.setGeneral(SSProp.COLOR_LABEL,               new Color(222, 228, 241, 255));
    // Translucent version of the above color (same when drawn over white background).
    ret.set(LBProp.COLOR_LABEL,               new Color(219, 225, 240, 230));

    ret.set(LBProp.OUTLINE_COLUMNS, true);
    ret.set(LBProp.TABLE_JUSTIFY, true);
    ret.set(LBProp.INTERNAL_TABLE_HEADER, false);

    Font font = new Font("Tahoma", Font.PLAIN, 11);
    Font fontBold = new Font("Tahoma", Font.BOLD, 11);
    ret.set(LBProp.RELATION_LABEL_FONT, fontBold);
    ret.set(LBProp.PRIMITIVE_LABEL_FONT, font);
    ret.set(LBProp.CROSSTAB_HEADING_FONT, new Font("Tahoma", Font.ITALIC, 11));
    ret.set(LBProp.VALUE_FONT, font);
    ret.set(LBProp.LABEL_TEXT_COLOR, Color.BLACK);
    ret.set(LBProp.VALUE_TEXT_COLOR, Color.BLACK);

    ret.set(LBProp.BULLET_FONT, font);
    ret.set(LBProp.BULLET_COLOR, Color.BLACK);
    ret.set(LBProp.BULLET_STRING, "\u2022"); // Alternative: "\u25CF" (Black Circle).

    ret.set(LBProp.ELLIPSIS_FONT, font);
    ret.set(LBProp.ELLIPSIS_COLOR, Color.BLACK);
    ret.set(LBProp.ELLIPSIS_STRING, "\u2026");

    final Points LINE_WIDTH = new Points(1.0);
    ret.set(LBProp.SEP_STYLE_TABLE_FIELD,     new LineStyle(COLOR_SEP_LIGHTER, LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW,       new LineStyle(COLOR_SEP        , LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW_MINOR, new LineStyle(COLOR_SEP_LIGHTER, LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_HEADER,    new LineStyle(COLOR_SEP        , LINE_WIDTH));
    ret.set(LBProp.OUTLINE_LABEL_BORDER,      new LineStyle(COLOR_SEP_LIGHTER, LINE_WIDTH));
    ret.set(LBProp.VALUE_BORDER,              new LineStyle(COLOR_SEP_LIGHTER, LINE_WIDTH));

    ret.set(LHProp.MAX_COLUMNS, 3);

    ret.set(LHProp.TABLE_MAX_PRIMITIVE_WIDTH, new Points(300));
    ret.set(LBProp.TABLE_NEST_SPACE_SIDE,     new Points(10));
    ret.set(LHProp.OUTLINE_MIN_VALUE_WIDTH,   new Points(100)); // 50 for width-collage figure.
    ret.set(LHProp.OUTLINE_MAX_LABEL_WIDTH,   new Points(80));
    ret.set(LHProp.OUTLINE_VALUE_SNAP_WIDTH,  new Points(30));
    ret.set(LBProp.OUTLINE_INDENT_WIDTH,      new Points(15));
    ret.set(LBProp.OUTLINE_TUPLE_SPACE,       new Points(7));
    ret.set(LBProp.OUTLINE_EMPTY_TABLE_SPACE, new Points(9));
    ret.set(LHProp.COLUMN_MIN_WIDTH,          new Points(260));

    ret.set(LBProp.MAX_INCOMPLETE_TUPLES, 1);

    ret.set(LMProp.OVERRIDE_MEASUREMENTS, false);
    ret.set(LMProp.SAMPLE_SIZE, 0);
    ret.set(LMProp.VARIABLE_WIDTH, false);
    ret.set(LMProp.AVERAGE_CARDINALITY, 3.0);
    // 0.9 inches, the default cell width of Excel for Mac 2011.
    ret.set(LMProp.VALUE_WIDTH, new Points(0.9 * 72));
    ret.set(LMProp.CONFIDENCE_WIDTH, new Points(0.9 * 72));
    ret.set(LMProp.DECIMAL_PLACES, OptionalInteger.absent());
    ret.set(LMProp.GROUP_DIGITS, false);
    ret.set(LMProp.TEMPORAL_PRECISION, TemporalPrecision.MINUTE);
    ret.set(LMProp.NUMERIC_MIN, OptionalBigDecimal.absent());
    ret.set(LMProp.NUMERIC_MAX, OptionalBigDecimal.absent());
    ret.set(LMProp.CROSSTAB_COLUMNS, CrosstabColumns.UNIT);
    ret.set(LBProp.CROSSTAB, false);
    ret.set(LBProp.VISUALIZATION, VisualizationType.NONE);
    ret.set(LBProp.VISUALIZATION_WIDTH, new Points(2 * 0.9 * 72));
    ret.set(LBProp.TEMPORAL_STYLE, TemporalStyle.ISO);
    ret.set(LBProp.DISPLAY_LOCALE, "");
    ret.set(LBProp.DISPLAY_TIME_ZONE, "");
    ret.set(LBProp.HORIZONTAL_ALIGNMENT, HorizontalAlignment.LEFT);
    ret.set(LBProp.VALUE_JUSTIFIED, false);

    return ret.build();
  }

  private static ViewDefaults buildDebugStyleSheet() {
    final Color COLOR_SEP             = new Color(  0,   0,   0,  128);
    final Color COLOR_SEP_LIGHTER     = new Color(  0,   0,   0,  128);

    ViewDefaults.Builder ret = ViewDefaults.builder();
    ret.setAll(buildColorStyleSheet());
    ret.set(LBProp.OUTLINE_LABEL_BORDER,      new LineStyle(COLOR_SEP_LIGHTER, new Points(10)));
    ret.set(LBProp.VALUE_BORDER,              new LineStyle(COLOR_SEP_LIGHTER, new Points(15)));
    ret.set(LBProp.SEP_STYLE_TABLE_FIELD,     new LineStyle(COLOR_SEP_LIGHTER, new Points(2)));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW,       new LineStyle(COLOR_SEP        , new Points(3)));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW_MINOR, new LineStyle(
        new Color(175, 163, 140, 255), new Points(5)));
    ret.set(LBProp.SEP_STYLE_TABLE_HEADER,    new LineStyle(COLOR_SEP        , new Points(7)));
    return ret.build();
  }

  private static ViewDefaults buildGrayscaleStyleSheet() {
    //final Color COLOR_SEP         = new Color(203, 203, 203, 255);
    //final Color COLOR_SEP_LIGHTER = new Color(233, 233, 233, 255);

    //final Color COLOR_SEP         = new Color(227, 227, 227, 255); // 29 alpha black equivalent
    //final Color COLOR_SEP_LIGHTER = new Color(240, 240, 240, 255); // 16 alpha black equivalent
    final Color COLOR_SEP         = new Color(0, 0, 0, 45); // Old value: 29
    final Color COLOR_SEP_LIGHTER = new Color(0, 0, 0, 18); // Old value: 16

    ViewDefaults.Builder ret = ViewDefaults.builder();
    ret.set(LBProp.COLOR_VALUE,               new Color(255, 255, 255, 255));
    //ret.set(LBProp.COLOR_VALUE_ALTERNATIVE,     new Color(250, 250, 250, 255)); // Old value
    ret.set(LBProp.COLOR_VALUE_ALTERNATIVE,     new Color(249, 249, 249, 255));
    //ret.set(LBProp.COLOR_VALUE,               new Color(0, 0, 0, 0));
    //ret.set(LBProp.COLOR_VALUE_ALTERNATE,     new Color(0, 0, 0, 4));

    //ret.setGeneral(SSProp.COLOR_LABEL,               new Color(220, 220, 220, 255));
    // Translucent version of the above color (same when drawn over white background).
    //ret.set(LBProp.COLOR_LABEL,               new Color(216, 216, 216, 230));

    // ret.set(LBProp.COLOR_LABEL, new Color(235, 235, 235, 255)); // Old value
    ret.set(LBProp.COLOR_LABEL, new Color(234, 234, 234, 255));

    ret.set(LBProp.OUTLINE_COLUMNS, true);
    ret.set(LBProp.TABLE_JUSTIFY, true);
    ret.set(LBProp.INTERNAL_TABLE_HEADER, false);

    // TODO[pd]: Bundle a specific default font for portability.
    //Font font = UIManager.getFont("TextField.font");
    Font font = new Font("Tahoma", Font.PLAIN, 11);
    Font fontBold = new Font("Tahoma", Font.BOLD, 11);
    ret.set(LBProp.RELATION_LABEL_FONT, fontBold);
    ret.set(LBProp.PRIMITIVE_LABEL_FONT, font);
    ret.set(LBProp.CROSSTAB_HEADING_FONT, new Font("Tahoma", Font.ITALIC, 11));
    ret.set(LBProp.VALUE_FONT, font);
    ret.set(LBProp.LABEL_TEXT_COLOR, Color.BLACK);
    ret.set(LBProp.VALUE_TEXT_COLOR, Color.BLACK);

    ret.set(LBProp.BULLET_FONT, font);
    ret.set(LBProp.BULLET_COLOR, Color.BLACK);
    ret.set(LBProp.BULLET_STRING, "\u2022"); // Alternative: "\u25CF" (Black Circle).

    /* Tahoma does not have the ellipsis character on Windows; use Dialog instead. This will
    eventually be replaced by a dedicated icon in Sieuf-Dingbats anyway. */
    ret.set(LBProp.ELLIPSIS_FONT, new Font("Dialog", Font.PLAIN, 11));
    ret.set(LBProp.ELLIPSIS_COLOR, Color.DARK_GRAY);
    // Midline horizontal ellipsis.
    ret.set(LBProp.ELLIPSIS_STRING, "\u22ef");
    // Regular ellipsis.
    //ret.set(LBProp.ELLIPSIS_STRING, "\u2026");

    final Points LINE_WIDTH = new Points(1.0);
    ret.set(LBProp.SEP_STYLE_TABLE_FIELD,     new LineStyle(COLOR_SEP, LINE_WIDTH));
    // For testing dashed borders.
    if (false)
      ret.set(LBProp.SEP_STYLE_TABLE_FIELD,     new LineStyle(Color.BLUE, LINE_WIDTH, new Points(14.0), new BigDecimal("0.6")));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW,       new LineStyle(COLOR_SEP, LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW_MINOR, new LineStyle(COLOR_SEP_LIGHTER, LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_HEADER,    new LineStyle(COLOR_SEP, LINE_WIDTH));
    ret.set(LBProp.OUTLINE_LABEL_BORDER,      new LineStyle(COLOR_SEP, LINE_WIDTH));
    ret.set(LBProp.VALUE_BORDER,              new LineStyle(COLOR_SEP, LINE_WIDTH));

    ret.set(LHProp.MAX_COLUMNS, 3);

    ret.set(LHProp.TABLE_MAX_PRIMITIVE_WIDTH, new Points(300));
    ret.set(LBProp.TABLE_NEST_SPACE_SIDE,     new Points(10));
    ret.set(LHProp.OUTLINE_MIN_VALUE_WIDTH,   new Points(50));
    ret.set(LHProp.OUTLINE_MAX_LABEL_WIDTH,   new Points(80));
    ret.set(LHProp.OUTLINE_VALUE_SNAP_WIDTH,  new Points(30));
    ret.set(LBProp.OUTLINE_INDENT_WIDTH,      new Points(15));
    ret.set(LBProp.OUTLINE_TUPLE_SPACE,       new Points(7));
    ret.set(LBProp.OUTLINE_EMPTY_TABLE_SPACE, new Points(9));
    ret.set(LHProp.COLUMN_MIN_WIDTH,          new Points(230));

    ret.set(LBProp.MAX_INCOMPLETE_TUPLES, 1);

    ret.set(LMProp.OVERRIDE_MEASUREMENTS, false);
    ret.set(LMProp.SAMPLE_SIZE, 0);
    ret.set(LMProp.VARIABLE_WIDTH, false);
    ret.set(LMProp.AVERAGE_CARDINALITY, 3.0);
    // 0.9 inches, the default cell width of Excel for Mac 2011.
    ret.set(LMProp.VALUE_WIDTH, new Points(0.9 * 72));
    ret.set(LMProp.CONFIDENCE_WIDTH, new Points(0.9 * 72));
    ret.set(LMProp.DECIMAL_PLACES, OptionalInteger.absent());
    ret.set(LMProp.GROUP_DIGITS, false);
    ret.set(LMProp.TEMPORAL_PRECISION, TemporalPrecision.MINUTE);
    ret.set(LMProp.NUMERIC_MIN, OptionalBigDecimal.absent());
    ret.set(LMProp.NUMERIC_MAX, OptionalBigDecimal.absent());
    ret.set(LMProp.CROSSTAB_COLUMNS, CrosstabColumns.UNIT);
    ret.set(LBProp.CROSSTAB, false);
    ret.set(LBProp.VISUALIZATION, VisualizationType.NONE);
    ret.set(LBProp.VISUALIZATION_WIDTH, new Points(2 * 0.9 * 72));
    ret.set(LBProp.TEMPORAL_STYLE, TemporalStyle.ISO);
    ret.set(LBProp.DISPLAY_LOCALE, "");
    ret.set(LBProp.DISPLAY_TIME_ZONE, "");
    ret.set(LBProp.HORIZONTAL_ALIGNMENT, HorizontalAlignment.LEFT);
    ret.set(LBProp.VALUE_JUSTIFIED, false);

    return ret.build();
  }

  private static ViewDefaults buildDataCollectionStyleSheet() {
    final Color COLOR_SEP         = new Color(  0,   0,   0,  80);
    final Color COLOR_SEP_LIGHTER = new Color(  0,   0,   0,  20);

    ViewDefaults.Builder ret = ViewDefaults.builder();
    ret.set(LBProp.COLOR_VALUE,               new Color(255, 232, 185, 255));
    ret.set(LBProp.COLOR_VALUE_ALTERNATIVE,     new Color(255, 237, 204, 255));
    ret.set(LBProp.COLOR_LABEL,               new Color(222, 228, 241, 255));

    ret.set(LBProp.OUTLINE_COLUMNS, true);
    ret.set(LBProp.TABLE_JUSTIFY, true);
    ret.set(LBProp.INTERNAL_TABLE_HEADER, false);

    Font font = new Font("Tahoma", Font.PLAIN, 11);
    ret.set(LBProp.RELATION_LABEL_FONT, font);
    ret.set(LBProp.PRIMITIVE_LABEL_FONT, font);
    ret.set(LBProp.CROSSTAB_HEADING_FONT, new Font("Tahoma", Font.ITALIC, 11));
    ret.set(LBProp.VALUE_FONT, font);
    ret.set(LBProp.LABEL_TEXT_COLOR, Color.BLACK);
    ret.set(LBProp.VALUE_TEXT_COLOR, Color.BLACK);

    ret.set(LBProp.BULLET_FONT, font);
    ret.set(LBProp.BULLET_COLOR, Color.BLACK);
    ret.set(LBProp.BULLET_STRING, "\u2022"); // Alternative: "\u25CF" (Black Circle).

    ret.set(LBProp.ELLIPSIS_FONT, font);
    ret.set(LBProp.ELLIPSIS_COLOR, Color.BLACK);
    ret.set(LBProp.ELLIPSIS_STRING, "\u2026");

    // TODO[pd]: Make lines black and wider.
    final Points LINE_WIDTH_LIGHTER = new Points(0.75);
    final Points LINE_WIDTH = new Points(0.75);
    ret.set(LBProp.VALUE_BORDER,              new LineStyle(Color.BLACK, LINE_WIDTH));
    ret.set(LBProp.OUTLINE_LABEL_BORDER,      new LineStyle(Color.BLACK, LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_FIELD,     new LineStyle(Color.BLACK, LINE_WIDTH_LIGHTER));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW,       new LineStyle(Color.BLACK, LINE_WIDTH));
    ret.set(LBProp.SEP_STYLE_TABLE_ROW_MINOR, new LineStyle(Color.BLACK, LINE_WIDTH_LIGHTER));
    ret.set(LBProp.SEP_STYLE_TABLE_HEADER,    new LineStyle(Color.BLACK, LINE_WIDTH));

    ret.set(LHProp.TABLE_MAX_PRIMITIVE_WIDTH, new Points(300)); // 200 for paper figures.
    ret.set(LBProp.TABLE_NEST_SPACE_SIDE,     new Points(10));
    ret.set(LHProp.OUTLINE_MIN_VALUE_WIDTH,   new Points(100));
    ret.set(LHProp.OUTLINE_MAX_LABEL_WIDTH,   new Points(70));
    ret.set(LHProp.OUTLINE_VALUE_SNAP_WIDTH,  new Points(30));
    ret.set(LBProp.OUTLINE_INDENT_WIDTH,      new Points(15));
    ret.set(LBProp.OUTLINE_TUPLE_SPACE,       new Points(7));
    ret.set(LBProp.OUTLINE_EMPTY_TABLE_SPACE, new Points(9));
    ret.set(LHProp.COLUMN_MIN_WIDTH,          new Points(230));

    ret.set(LHProp.MAX_COLUMNS, Integer.MAX_VALUE);

    ret.set(LBProp.MAX_INCOMPLETE_TUPLES, 1);

    ret.set(LMProp.OVERRIDE_MEASUREMENTS, false);
    ret.set(LMProp.SAMPLE_SIZE, 0);
    ret.set(LMProp.VARIABLE_WIDTH, false);
    ret.set(LMProp.AVERAGE_CARDINALITY, 3.0);
    // 0.9 inches, the default cell width of Excel for Mac 2011.
    ret.set(LMProp.VALUE_WIDTH, new Points(0.9 * 72));
    ret.set(LMProp.CONFIDENCE_WIDTH, new Points(0.9 * 72));
    ret.set(LMProp.DECIMAL_PLACES, OptionalInteger.absent());
    ret.set(LMProp.GROUP_DIGITS, false);
    ret.set(LMProp.TEMPORAL_PRECISION, TemporalPrecision.MINUTE);
    ret.set(LMProp.NUMERIC_MIN, OptionalBigDecimal.absent());
    ret.set(LMProp.NUMERIC_MAX, OptionalBigDecimal.absent());
    ret.set(LMProp.CROSSTAB_COLUMNS, CrosstabColumns.UNIT);
    ret.set(LBProp.CROSSTAB, false);
    ret.set(LBProp.VISUALIZATION, VisualizationType.NONE);
    ret.set(LBProp.VISUALIZATION_WIDTH, new Points(2 * 0.9 * 72));
    ret.set(LBProp.TEMPORAL_STYLE, TemporalStyle.ISO);
    ret.set(LBProp.DISPLAY_LOCALE, "");
    ret.set(LBProp.DISPLAY_TIME_ZONE, "");
    ret.set(LBProp.HORIZONTAL_ALIGNMENT, HorizontalAlignment.LEFT);
    ret.set(LBProp.VALUE_JUSTIFIED, false);

    return ret.build();
  }
}
