// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.google.common.base.Preconditions;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalAccessor;
import java.time.temporal.TemporalField;
import java.time.temporal.TemporalQuery;
import java.time.temporal.UnsupportedTemporalTypeException;

/**
 * Wrapper for temporal values, used to trim away fields according to a specified desired precision.
 * Allows optional sections in localized time formats to be enabled or disabled on demand.
 */
final class FilteredTemporalAccessor implements TemporalAccessor {
  private final TemporalAccessor delegate;
  private final TemporalPrecision precision;

  public FilteredTemporalAccessor(TemporalAccessor delegate, TemporalPrecision precision) {
    Preconditions.checkNotNull(delegate);
    Preconditions.checkNotNull(precision);
    this.delegate = delegate;
    this.precision = precision;
  }

  @Override
  public boolean isSupported(TemporalField field) {
    return precisionIncludes(field) && delegate.isSupported(field);
  }

  private boolean precisionIncludes(TemporalField field) {
    /* Some of the long localized formats seem to require NANO_OF_SECOND even if it is not
    ultimately displayed. */
    if (field.equals(ChronoField.SECOND_OF_MINUTE) || field.equals(ChronoField.NANO_OF_SECOND)) {
      return precision.isIncludingSeconds();
    } else {
      return true;
    }
  }

  @Override
  public <R> R query(TemporalQuery<R> query) {
    /* Some localized formats require this method to be properly implemented. Filtering fields here
    would be more difficult. It's not the end of the world if some odd localization manages to
    bypass the filter by using this method; the temporal precision is only a best-effort setting for
    localized styles. */
    return delegate.query(query);
  }

  @Override
  public long getLong(TemporalField field) throws UnsupportedTemporalTypeException {
    // Throw any exceptions from the delegate first.
    final long ret = delegate.getLong(field);
    if (!precisionIncludes(field)) {
      throw new UnsupportedTemporalTypeException(
          "Attempted to retrieve temporal field " + field + " of filtered precision " + precision);
    }
    return ret;
  }
}
