// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.sieuferd.layout.internal.OutlineField;
import com.sieuferd.layout.internal.LayoutContext;
import com.sieuferd.layout.internal.PartitionAlgorithm;
import com.sieuferd.layout.internal.Column;
import com.sieuferd.datamodel.View;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Ordering;
import com.sieuferd.cell.builders.BorderCellBuilder;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.layout.internal.LayoutContext.RelationLayoutContext;
import com.sieuferd.layout.internal.PartitionAlgorithm.ScalarOperations;
import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.support.Direction;
import com.sieuferd.cell.builders.ListCellBuilder;
import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.cell.styles.TextFitPolicy;
import com.sieuferd.cell.styles.LineStyle;
import com.sieuferd.cell.styles.Text;
import com.sieuferd.cell.styles.Text.SimpleText;
import com.sieuferd.cell.support.Axis;
import com.sieuferd.cell.support.CursorStyle;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.expressions.TupleConstant;
import com.sieuferd.layout.internal.LayoutContext.PrimitiveLayoutContext;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.util.MapUtil;
import java.awt.Color;
import java.util.Collections;
import javax.annotation.Nullable;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;

// TODO[cc]: Remove experimental code ("if (false)")
/* TODO[cc]: Consider handling all of empty relation placeholder cells, ellipsis cells, headers, and
             filters by means of placeholder traversals rather than by the current ad hoc means. */
public final class LayoutGenerator {
  /**
   * If true, the last column of a justified table will absorb any remaining space in the table if
   * it is strictly left-justified, _after_ extra space has first been allocated according to the
   * regular rules. Currently turned off to avoid showing popup affordance buttons way off to the
   * right of the screen, and because very long paragraphs are undesirable.
   */
  private static final boolean FORCE_LAST_COLUMN_JUSTIFICATION = false;
  private static final int MAXIMUM_COLUMN_LEVELS = 4;
  private static final boolean HIDE_SIBLINGLESS_LABELS = false;
  public static final ImmutableSet<ViewProperty<?>> LAYOUT_PROPERTIES =
      ImmutableSet.<ViewProperty<?>>builder()
      .add(SProp.CHILD_PATHS)
      .add(CProp.LABEL)
      .add(CProp.VISIBLE)
      .add(CProp.PLURAL)
      .addAll(LAProp.INSTANCE.getViewProperties())
      .addAll(LBProp.INSTANCE.getViewProperties())
      .addAll(LHProp.INSTANCE.getViewProperties())
      .addAll(LMProp.INSTANCE.getViewProperties())
      .build();

  private LayoutGenerator() {
  }

  /* ******************************************************************************************** */
  /** Sets LEProp.EFF_OUTLINE_LABEL_WIDTH and LAProp.EFF_LABEL_VISIBLE.
  @param labelWidthStats includes label left borders, null for the root context */
  private static void measureLabels(
      LayoutContext context, final @Nullable SummaryStatistics labelWidthStats)
  {
    Preconditions.checkArgument(!context.isValue());
    Preconditions.checkArgument(labelWidthStats == null || context.getParent() != null);
    if (context.isRelation()) {
      SummaryStatistics childLabelWidthStats = new SummaryStatistics();
      for (LayoutContext childContext : context.getRelation().traverseFields())
        measureLabels(childContext, childLabelWidthStats);
      if (context.pendingAuto(LEProp.EFF_OUTLINE_LABEL_WIDTH)) {
        // TODO[pd]: Set a minimum width here.
        context.setP(LEProp.EFF_OUTLINE_LABEL_WIDTH, Double.isNaN(childLabelWidthStats.getMax())
            ? Pixels.ZERO
            : Pixels.create(context.getRenderContext(), childLabelWidthStats.getMax()));
      }
    }
    if (context.pendingAuto(LEProp.EFF_LABEL_VISIBLE, LAProp.LABEL_VISIBLE)) {
      context.setP(LEProp.EFF_LABEL_VISIBLE,
          HIDE_SIBLINGLESS_LABELS ? context.hasSiblings() : true);
    }
    if (labelWidthStats != null && context.getP(LEProp.EFF_LABEL_VISIBLE)) {
      Pixels borderWidth = context.getP(LBProp.OUTLINE_LABEL_BORDER)
          .measure(context.getRenderContext(), true);
      final Pixels outlineLabelWidth = context.measureLabel(
          context.getParent().getSSPixels(LHProp.OUTLINE_MAX_LABEL_WIDTH).minus(borderWidth),
          false, false).getX().plus(borderWidth);
      labelWidthStats.addValue(outlineLabelWidth.getDouble());
    }
  }
  /* ******************************************************************************************** */

  private static void autoPartitionColumns(Map<LayoutContext,OutlineField> cells, int noColumns) {
    List<List<Entry<LayoutContext,OutlineField>>> partitionedColumns =
        PartitionAlgorithm.doPartition(ImmutableList.copyOf(cells.entrySet()), noColumns,
        new ScalarOperations<Entry<LayoutContext,OutlineField>,Pixels>()
    {
      @Override
      public boolean greaterThan(Pixels v1, Pixels v2) {
        return v1.isGreaterThan(v2);
      }

      @Override
      public Pixels plus(Pixels v1, Pixels v2) {
        return v1.plus(v2);
      }

      @Override
      public Pixels getZero() {
        return Pixels.ZERO;
      }

      @Override
      public Pixels getValue(Entry<LayoutContext,OutlineField> wrapper) {
        return wrapper.getValue().getCell(Pixels.ZERO).getDimensions().getY();
      }
    });
    for (List<Entry<LayoutContext,OutlineField>> column : partitionedColumns) {
      int i = 0;
      for (Entry<LayoutContext,OutlineField> columnFields : column)
        columnFields.getKey().setP(LEProp.EFF_START_NEW_COLUMN, i++ == 0);
    }
  }

  private static int getPreferredNumberOfColumns(RelationLayoutContext context, Pixels width) {
    if (!context.getP(LBProp.OUTLINE_COLUMNS))
      return 1;
    /* It makes sense to define the COLUMN_MIN_WIDTH parameter to include any horizontal space
    that might be taken up by a label, since child cells may be rendered with varying label
    styles. */
    /* Since the number of columns is decided purely based on available horizontal space, the layout
    may overflow horizontally if each column takes up more than its alloted space (e.g. in the case
    of an extremely long label). I considered this acceptable. */
    return Math.max(1, Math.min(context.getP(LHProp.MAX_COLUMNS),
        (int) Math.floor(width.getDouble() / context.getSSPixels(LHProp.COLUMN_MIN_WIDTH)
        .getDouble())));
  }

  private static Pixels getColumnWidth(RenderContext renderContext, Pixels width, int noColumns) {
    return width.divideBy(renderContext, noColumns);
  }

  // TODO[cc]: Consider removing if only used in one place in the end.
  private static boolean wouldUseTable(
      RelationLayoutContext rcx, Pixels labelWidthWithBorder, Pixels width)
      throws InterruptedException
  {
    if (!rcx.pendingAuto(LEProp.EFF_USE_TABLE)) {
      return rcx.getP(LEProp.EFF_USE_TABLE);
    } else {
      if (parentWasTable(rcx))
        return true;
      if (!rcx.getRelation().isLayoutPlural())
        return false;

      // TODO[pd]: Account for width of separator between label and value.
      final Pixels widthAvailableForTable = width.minus(labelWidthWithBorder);
      final Pixels minimumTableWidth;
      {
        RelationLayoutContext speculativeContext = rcx.speculativeTraversal();
        speculativeContext.setP(LEProp.EFF_USE_TABLE, true);
        minimumTableWidth = createTableHeader(speculativeContext).getDimensions().getX()
            .plus(speculativeContext.getP(LBProp.VALUE_BORDER)
            .measure(rcx.getRenderContext(), true));
      }
      return !minimumTableWidth.isGreaterThan(widthAvailableForTable);
    }
  }

  private static OutlineField createOutlineField(LayoutContext context, Pixels width)
      throws InterruptedException
  {
    final boolean labelVisible = context.getP(LEProp.EFF_LABEL_VISIBLE);
    final Pixels labelWidthInclBorder = context.getParent().getP(LEProp.EFF_OUTLINE_LABEL_WIDTH);
    if (context.isRelation() && context.pendingAuto(LEProp.EFF_USE_TABLE, LAProp.USE_TABLE)) {
      context.setP(LEProp.EFF_USE_TABLE, wouldUseTable(context.getRelation(),
          labelVisible ? labelWidthInclBorder : Pixels.ZERO, width));
    }
    final Pixels labelBorderWidth =
        context.getP(LBProp.OUTLINE_LABEL_BORDER).measure(context.getRenderContext(), true);
    final Pixels labelWidthExclBorder = labelWidthInclBorder.minus(labelBorderWidth);
    if ((context.isRelation() && !context.getP(LEProp.EFF_USE_TABLE))) {
      /* TODO[pd]: If labelVisible is false, have the sublayout use this level's outline label width
                   instead. This is not entirely trivial, however; see if I can find some smart
                   solution. */
      // Label above
      Cell contentCell = createSublayout(context, width,
          labelVisible ? context.getSSPixels(LBProp.OUTLINE_INDENT_WIDTH) : Pixels.ZERO,
          true);
      // Widen labels over outline layouts.
      final Pixels widenedLabelWidthExclBorder =
          labelWidthExclBorder.max(width.minus(labelBorderWidth));
      return new OutlineField(context, contentCell, widenedLabelWidthExclBorder, false);
    } else {
      // Label adjacent
      Cell contentCell = createSublayout(context,
          width.minus(labelVisible ? labelWidthInclBorder : Pixels.ZERO), Pixels.ZERO, true);
      return new OutlineField(context, contentCell, labelWidthExclBorder, true);
    }
  }

  private static List<List<LayoutContext>> getPartitionedColumnContexts(
      Collection<LayoutContext> childContexts)
  {
    List<List<LayoutContext>> partitionedColumns = Lists.newArrayList();
    List<LayoutContext> currentColumn = Lists.newArrayList();
    for (LayoutContext childContext : childContexts) {
      if (childContext.getP(LEProp.EFF_START_NEW_COLUMN) && !currentColumn.isEmpty()) {
        partitionedColumns.add(currentColumn);
        currentColumn = Lists.newArrayList();
      }
      currentColumn.add(childContext);
    }
    if (!currentColumn.isEmpty())
      partitionedColumns.add(currentColumn);
    return partitionedColumns;
  }

  private static Cell createOutlineColumnSet(
      RelationLayoutContext context, Collection<LayoutContext> fieldContexts, Pixels width)
      throws InterruptedException
  {
    List<List<LayoutContext>> partitionedColumnContexts = null;
    {
      boolean partitioningPending = true;
      for (LayoutContext childContext : fieldContexts) {
        if (childContext.pendingAuto(LEProp.EFF_START_NEW_COLUMN, LAProp.START_NEW_COLUMN)) {
          childContext.setP(LEProp.EFF_START_NEW_COLUMN, false);
        } else {
          partitioningPending = false;
        }
      }
      if (!partitioningPending)
        partitionedColumnContexts = getPartitionedColumnContexts(fieldContexts);
    }
    final int numColumns;
    if (partitionedColumnContexts == null) {
      numColumns = Math.min(fieldContexts.size(), getPreferredNumberOfColumns(context, width));
    } else {
      numColumns = partitionedColumnContexts.size();
    }
    final Pixels columnWidth = getColumnWidth(context.getRenderContext(), width, numColumns);
    final Pixels roundingError = width.minus(columnWidth.times(numColumns));
    if (roundingError.isNegative())
      throw new AssertionError();
    /* Let the last column absorb any rounding error, to make the right-hand side of the layout line
    up properly. */
    final Set<LayoutContext> roundingErrorAbsorbers =
        (partitionedColumnContexts == null || partitionedColumnContexts.isEmpty())
        ? ImmutableSet.of()
        : ImmutableSet.copyOf(partitionedColumnContexts.get(partitionedColumnContexts.size() - 1));
    final Map<LayoutContext,OutlineField> fieldCellMap = Maps.newLinkedHashMap();
    for (LayoutContext childContext : fieldContexts) {
      MapUtil.putChecked(fieldCellMap, childContext, createOutlineField(childContext,
          columnWidth.plus(
          roundingErrorAbsorbers.contains(childContext) ? roundingError : Pixels.ZERO)));
    }

    if (partitionedColumnContexts == null) {
      autoPartitionColumns(fieldCellMap, numColumns);
      partitionedColumnContexts = getPartitionedColumnContexts(fieldContexts);
    }

    final List<Column> columns = Lists.newArrayList();
    Pixels maxColumnHeight = Pixels.ZERO;
    for (List<LayoutContext> columnContext : partitionedColumnContexts) {
      final Column column = new Column(context, columnWidth);
      for (LayoutContext childContext : columnContext)
        column.addField(MapUtil.getChecked(fieldCellMap, childContext));
      maxColumnHeight = maxColumnHeight.max(column.getUnpaddedHeight());
      columns.add(column);
    }
    final ListCellBuilder ret = context.horizontalListCellBuilder();
    for (Column column : columns)
      ret.addCell(column.getPaddedCell(maxColumnHeight));
    /* Ensure that the width of the returned Cell remains the same with or without the rounding
    error adjustment, since said adjustment is not applied during a placeholder pass. Actually
    applying the adjustment during the placeholder pass would require building the layouts of the
    last column twice, leading to exponential runtime behavior. */
    ret.minimumDimensions(new Coordinates(width, Pixels.ZERO));
    return ret.build();
  }

  /** Placeholder context only. */
  private static boolean containsPluralOutline(LayoutContext context) {
    if (context.isRelation()) {
      if (context.getRelation().isLayoutPlural() && !context.getP(LEProp.EFF_USE_TABLE))
        return true;
      for (LayoutContext childContext : context.getRelation().traverseFields()) {
        if (containsPluralOutline(childContext))
          return true;
      }
    }
    return false;
  }

  /** Placeholder context only. */
  private static boolean hasMoreLevelsThan(LayoutContext context, int levels) {
    if (levels < 1)
      return true;
    if (context.isRelation()) {
      for (LayoutContext field : context.getRelation().traverseFields()) {
        if (hasMoreLevelsThan(field, levels - 1))
          return true;
      }
    }
    return false;
  }

  private static Cell createOutlineTupleLayout(
      RelationLayoutContext context, Tuple tuple, Pixels width)
      throws InterruptedException
  {
    if (!context.isValue()) {
      final int noColumns = getPreferredNumberOfColumns(context, width);
      final Pixels columnWidth = getColumnWidth(context.getRenderContext(), width, noColumns);

      boolean previousWasExcluded = true;
      for (LayoutContext childContext : context.traverseFields(tuple)) {
        if (childContext.pendingAuto(
            LEProp.EFF_OUTLINE_COLUMN_EXCLUDE, LAProp.OUTLINE_COLUMN_EXCLUDE))
        {
          final boolean excluded;
          if (noColumns > 1) {
            // TODO[pd]: Find a more elegant way to avoid exponential runtime here.
            if (hasMoreLevelsThan(childContext, MAXIMUM_COLUMN_LEVELS)) {
              excluded = true;
            } else {
              LayoutContext speculativeTraversal = childContext.speculativeTraversal();
              Cell speculativeField =
                  createOutlineField(speculativeTraversal, columnWidth).getCell(Pixels.ZERO);
              if (speculativeField.getDimensions().getX().isGreaterThan(columnWidth)) {
                excluded = true;
              } else {
                excluded = containsPluralOutline(speculativeTraversal);
              }
            }
          } else {
            excluded = false;
          }
          childContext.setP(LEProp.EFF_OUTLINE_COLUMN_EXCLUDE, excluded);
        }
        final boolean excluded = childContext.getP(LEProp.EFF_OUTLINE_COLUMN_EXCLUDE);
        if (childContext.pendingAuto(LEProp.EFF_START_NEW_COLUMN_SET, LAProp.START_NEW_COLUMN_SET))
          childContext.setP(LEProp.EFF_START_NEW_COLUMN_SET, excluded || previousWasExcluded);
        previousWasExcluded = excluded;
      }
    }

    List<List<LayoutContext>> columnSetFieldLists = Lists.newArrayList();
    {
      // TODO[cc]: Make helper function.
      List<LayoutContext> fieldsForCurrentColumnSet = Lists.newArrayList();
      for (LayoutContext childContext : context.traverseFields(tuple)) {
        if (childContext.getP(LEProp.EFF_START_NEW_COLUMN_SET) &&
            !fieldsForCurrentColumnSet.isEmpty())
        {
          columnSetFieldLists.add(fieldsForCurrentColumnSet);
          fieldsForCurrentColumnSet = Lists.newArrayList();
        }
        fieldsForCurrentColumnSet.add(childContext);
      }
      if (!fieldsForCurrentColumnSet.isEmpty())
        columnSetFieldLists.add(fieldsForCurrentColumnSet);
    }

    ListCellBuilder ret = context.verticalListCellBuilder();
    for (List<LayoutContext> columnSetFieldList : columnSetFieldLists)
      ret.addCell(createOutlineColumnSet(context, columnSetFieldList, width));
    return ret.build();
  }

  private static List<PrimitiveLayoutContext> getJustifiableTableContexts(LayoutContext context) {
    List<PrimitiveLayoutContext> ret = Lists.newArrayList();
    collectJustifiableColumnContexts(context, ret);
    return Collections.unmodifiableList(ret);
  }

  private static void collectJustifiableColumnContexts(
      LayoutContext context, List<PrimitiveLayoutContext> contexts)
  {
    if (context.isRelation()) {
      if (!context.getP(LEProp.EFF_USE_TABLE) || context.getP(LBProp.CROSSTAB))
        return;
      for (LayoutContext childContext : context.getRelation().traverseFields())
        collectJustifiableColumnContexts(childContext, contexts);
    } else {
      contexts.add(context.getPrimitive());
    }
  }

  private static @Nullable Pixels getMaximumTableColumnWidth(PrimitiveLayoutContext context) {
    Pixels valueWidth = context.getSSPixels(LMProp.VALUE_WIDTH);
    if (context.getP(LBProp.VISUALIZATION) == VisualizationType.HORIZONTAL_BARS) {
      return null;
    } else if (context.getP(LMProp.VARIABLE_WIDTH)) {
      return valueWidth
          .max(context.getSSPixels(LHProp.TABLE_MAX_PRIMITIVE_WIDTH))
          .max(context.getSSPixels(LMProp.CONFIDENCE_WIDTH));
    } else {
      return valueWidth;
    }
  }

  private static void justifyTable(RelationLayoutContext context, Pixels width)
      throws InterruptedException
  {
    // TODO[cc]: Get the type system to enforce this.
    if (context.isValue())
      throw new IllegalStateException();
    Map<LayoutContext,Pixels> horizontalizableCosts = Maps.newLinkedHashMap();
    Pixels tightTableHeaderWidth = createTableRelationSublayout(context.placeholderTraversal(),
        true, horizontalizableCosts).getDimensions().getX();
    Pixels extraSpaceAvailable = width.minus(tightTableHeaderWidth);
    if (!extraSpaceAvailable.isGreaterThan(Pixels.ZERO))
      return;

    /* First horizontalize any vertical labels, in order of least cost first. The greedy order
    maximizes the number of labels that can be horizontalized. */

    final List<Entry<LayoutContext,Pixels>> sortedCosts =
        Ordering.from((Entry<LayoutContext,Pixels> e1, Entry<LayoutContext,Pixels> e2) ->
        e1.getValue().compareTo(e2.getValue())).sortedCopy(horizontalizableCosts.entrySet());

    for (Entry<LayoutContext,Pixels> horizontalizable : sortedCosts) {
      if (!horizontalizable.getValue().isGreaterThan(extraSpaceAvailable)) {
        if (horizontalizable.getKey().getP(LAProp.USE_VERTICAL_TABLE_HEADER) == YesNoAuto.AUTO) {
          horizontalizable.getKey().setP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER, false);
          extraSpaceAvailable = extraSpaceAvailable.minus(horizontalizable.getValue());
        }
      }
    }

    tightTableHeaderWidth = createTableRelationSublayout(context.placeholderTraversal(), true, null)
        .getDimensions().getX();
    extraSpaceAvailable = width.minus(tightTableHeaderWidth);

    /* Because justification may be limited by maximum column widths, it may be necessary to
    distribute remaining space over several iterations. */
    while (extraSpaceAvailable.isGreaterThan(Pixels.ZERO)) {
      Pixels previousExtraSpaceAvailable = extraSpaceAvailable;
      extraSpaceAvailable = distributeTableSpace(context, extraSpaceAvailable);
      if (extraSpaceAvailable.equals(previousExtraSpaceAvailable))
        break;
    }
    // Give any remaining space to the last column if it is strictly left-justified.
    if (FORCE_LAST_COLUMN_JUSTIFICATION && extraSpaceAvailable.isGreaterThan(Pixels.ZERO)) {
      List<PrimitiveLayoutContext> tableColumnContexts = getJustifiableTableContexts(context);
      if (!tableColumnContexts.isEmpty()) {
        PrimitiveLayoutContext lastColumnContext =
            tableColumnContexts.get(tableColumnContexts.size() - 1);
        if (lastColumnContext.isStrictlyLeftJustified()) {
          lastColumnContext.setP(LEProp.EFF_TABLE_COLUMN_WIDTH,
              lastColumnContext.getP(LEProp.EFF_TABLE_COLUMN_WIDTH).plus(extraSpaceAvailable));
        }
      }
    }
  }

  /**
   * Redistribute available space among variable-width columns in proportion with their value
   * widths. Never make columns narrower.
   *
   * @return the amount of space to distribute
   */
  private static Pixels distributeTableSpace(
      final RelationLayoutContext context, final Pixels extraSpaceAvailable)
  {
    // First gather and total weights.
    final Map<PrimitiveLayoutContext,Double> expandables = Maps.newLinkedHashMap();
    // Technically in the unit of points.
    double sumWeights = 0.0;
    Pixels totalSpaceAvailable = extraSpaceAvailable;
    for (PrimitiveLayoutContext columnContext : getJustifiableTableContexts(context)) {
      final @Nullable Pixels maximumWidth = getMaximumTableColumnWidth(columnContext);
      final Pixels existingWidth = columnContext.getP(LEProp.EFF_TABLE_COLUMN_WIDTH);
      if (maximumWidth == null || maximumWidth.isGreaterThan(existingWidth)) {
        final double weight = columnContext.getP(LMProp.VALUE_WIDTH).getValue() +
            (columnContext.getP(LBProp.VISUALIZATION) == VisualizationType.HORIZONTAL_BARS
            ? columnContext.getP(LBProp.VISUALIZATION_WIDTH).getValue() : 0.0);
        if (weight > 0.0) {
          sumWeights += weight;
          totalSpaceAvailable = totalSpaceAvailable.plus(existingWidth);
          MapUtil.putChecked(expandables, columnContext, weight);
        }
      }
    }
    // Then distribute the extra space proportionally, up to applicable limits.
    Pixels remainingExtraSpaceAvailable = extraSpaceAvailable;
    for (Entry<PrimitiveLayoutContext,Double> expandable : expandables.entrySet()) {
      final PrimitiveLayoutContext columnContext = expandable.getKey();
      final double fraction = expandable.getValue() / sumWeights;
      final Pixels existingWidth = columnContext.getP(LEProp.EFF_TABLE_COLUMN_WIDTH);
      /* Use Pixels.max(), but limited by remainingExtraSpaceAvailable, to ensure that every pixel
      and no more is used up even in the face of rounding errors. */
      final Pixels expandBy =
          Pixels.max(columnContext.getRenderContext(), totalSpaceAvailable.getDouble() * fraction)
          .min(getMaximumTableColumnWidth(columnContext))
          .minus(existingWidth)
          .min(remainingExtraSpaceAvailable)
          .max(Pixels.ZERO);
      // TODO[cc]: Remove this condition and fix the error that results.
      if (expandBy.isGreaterThan(Pixels.ZERO))
        columnContext.setP(LEProp.EFF_TABLE_COLUMN_WIDTH, existingWidth.plus(expandBy));
      remainingExtraSpaceAvailable = remainingExtraSpaceAvailable.minus(expandBy);
    }
    return remainingExtraSpaceAvailable;
  }

  // TODO[cc]: Add nullable annotations to horizontalizeCosts elsewhere in this file.
  private static Pixels computeHeaderWidth(LayoutContext context,
      @Nullable Map<LayoutContext,Pixels> horizontalizeCosts, Pixels givenMinimum)
  {
    if (!context.getP(LEProp.EFF_LABEL_VISIBLE)) {
      if (context.pendingAuto(
          LEProp.EFF_USE_VERTICAL_TABLE_HEADER, LAProp.USE_VERTICAL_TABLE_HEADER))
      {
        context.setP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER, false);
      }
      return Pixels.ZERO;
    }
    Pixels ret;
    Pixels labelWidthHorz = null;

    if (!context.pendingAuto(
        LEProp.EFF_USE_VERTICAL_TABLE_HEADER, LAProp.USE_VERTICAL_TABLE_HEADER))
    {
      ret = context.measureLabel(
          givenMinimum, context.getP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER), true).getX();
    } else {
      labelWidthHorz = context.measureLabel(givenMinimum, false, true).getX();
      context.setP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER, false);
      ret = labelWidthHorz;
      if (!context.isRelation() && labelWidthHorz.isGreaterThan(givenMinimum)) {
        Pixels labelWidthVert = context.measureLabel(labelWidthHorz, true, true).getX();
        if (labelWidthHorz.isGreaterThan(labelWidthVert)) {
          context.setP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER, true);
          ret = labelWidthVert;
        }
      }
    }
    if (horizontalizeCosts != null && context.getP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER)) {
      if (labelWidthHorz == null)
        labelWidthHorz = context.measureLabel(givenMinimum, false, true).getX();
      // ret holds the vertical width, since USE_VERTICAL_TABLE_HEADER was enabled.
      MapUtil.putChecked(horizontalizeCosts, context,
          labelWidthHorz.max(givenMinimum).minus(
          ret.max(givenMinimum)).max(Pixels.ZERO));
    }
    return ret;
  }

  // TODO[pd]: Avoid remeasuring vertical labels over and over again (fix by creating headers once).
  private static Cell createTableField(LayoutContext context, boolean headerMode,
      @Nullable Map<LayoutContext,Pixels> horizontalizeCosts, LineStyle borderLeft,
      LineStyle borderTop)
      throws InterruptedException
  {
    if (headerMode && context.isValue())
      throw new IllegalStateException();
    /* I used to omit the margins if the relation label was hidden, but later decided against it;
    the marging serves a useful purpose of separating tuples from sibling relations whether the
    label is shown or not. */
    final boolean nestSpaceSide = context.isRelation() && context.getRelation().isLayoutPlural();
    Cell bottom;

    // Assign main cell content.
    if (headerMode) {
      if (context.isRelation()) {
        // By default, never use an outline sublayout inside a table sublayout.
        if (context.pendingAuto(LEProp.EFF_USE_TABLE, LAProp.USE_TABLE))
          context.setP(LEProp.EFF_USE_TABLE, true);
        if (context.getP(LEProp.EFF_USE_TABLE)) {
          bottom = createTableRelationSublayout(context.getRelation(), true, horizontalizeCosts);
        } else {
          bottom = context.emptyCellBuilder()
              .cursorBias(Direction.UP)
              .minimumWidth(createSublayout(context, Pixels.ZERO, Pixels.ZERO, false)
                  .getDimensions().getX())
              .build();
        }
      } else {
        Pixels minimumWidth =
            context.getSSPixels(LMProp.VALUE_WIDTH)
            .min(context.getSSPixels(LHProp.TABLE_MAX_PRIMITIVE_WIDTH))
            .plus(context.getP(LBProp.VISUALIZATION) == VisualizationType.HORIZONTAL_BARS
                ? context.getSSPixels(LBProp.VISUALIZATION_WIDTH)
                : Pixels.ZERO);
        /* TODO[cc]: This is a weird place to set a minimum width for a table column. Move these
                 constraints closer to where VALUE_WIDTH is used. */
        bottom = context.emptyCellBuilder()
            .cursorBias(Direction.UP)
            .minimumWidth(minimumWidth)
            .build();
      }
      Pixels minimumBottomWidth = bottom.getDimensions().getX();
      if (!context.pendingAuto(LEProp.EFF_TABLE_COLUMN_WIDTH))
        minimumBottomWidth = minimumBottomWidth.max(context.getP(LEProp.EFF_TABLE_COLUMN_WIDTH));

      // TODO[pd]: See if I should avoid recomputing this every time.
      Pixels headerWidth =
          computeHeaderWidth(context, horizontalizeCosts, bottom.getDimensions().getX());
      // TODO[cc]: Clean up.
      minimumBottomWidth = minimumBottomWidth.max(headerWidth.minus(nestSpaceSide
          ? context.getSSPixels(LBProp.TABLE_NEST_SPACE_SIDE).times(2)
          : Pixels.ZERO));

      // TODO[cc]: Avoid this.
      /* Set the column width even if the property was already defined, in case it needs to be
      enlarged. This may be necessary due to justification, or because a partially customized
      stylesheet has been used. */
      if (context.pendingAuto(LEProp.EFF_TABLE_COLUMN_WIDTH) ||
          minimumBottomWidth.isGreaterThan(context.getP(LEProp.EFF_TABLE_COLUMN_WIDTH)))
      {
        context.setP(LEProp.EFF_TABLE_COLUMN_WIDTH, minimumBottomWidth);
      }

      if (context.isRelation() &&
          minimumBottomWidth.isGreaterThan(bottom.getDimensions().getX()))
      {
        // TODO[pd]: Ensure no bad runtime behavior here.
        justifyTable(context.getRelation(), minimumBottomWidth);
      }
    } else {
      Preconditions.checkState(!context.pendingAuto(LEProp.EFF_TABLE_COLUMN_WIDTH));
      bottom = createSublayout(context, context.getP(LEProp.EFF_TABLE_COLUMN_WIDTH),
          Pixels.ZERO, false);
    }
    bottom = context.newBigger(bottom,
        new Coordinates(context.getP(LEProp.EFF_TABLE_COLUMN_WIDTH), Pixels.ZERO));

    // Append margins around nested content.
    if (nestSpaceSide) {
      Cell emptyCell = context.emptyCellBuilder().build();
      BorderCellBuilder builder = (headerMode
          ? context.selectableBorderCellBuilder(emptyCell)
          : context.valueSelectableBorderCellBuilder(emptyCell));
      builder.minimumWidth(context.getSSPixels(LBProp.TABLE_NEST_SPACE_SIDE));
      /* I considered making tableNestSpaceSideLeft/tableNestSpaceSideRight the color of the first
      and last visible child column, respectively, but this might communicate the wrong
      structure. */
      final Cell tableNestSpaceSideLeft = builder.cursorBias(Direction.RIGHT).build();
      final Cell tableNestSpaceSideRight = builder.cursorBias(Direction.LEFT).build();
      bottom = context.horizontalListCellBuilder()
          .addCells(tableNestSpaceSideLeft, bottom, tableNestSpaceSideRight)
          .build();
    }

    final Cell ret;
    if (headerMode) {
      // Add main header above content.
      Pixels fieldWidth = bottom.getDimensions().getX();
      Cell masterHeader;
      if (context.getP(LEProp.EFF_LABEL_VISIBLE)) {
        final boolean vertical = context.getP(LEProp.EFF_USE_VERTICAL_TABLE_HEADER);
        final boolean horizontalBar = !context.isRelation() &&
            context.getP(LBProp.VISUALIZATION) == VisualizationType.HORIZONTAL_BARS;
        masterHeader = context.textCellBuilder(
            LBProp.LABEL_TEXT_COLOR, context.getLabelFontProperty(), fieldWidth, horizontalBar)
            .content(context.getStyledLabel(!(vertical || horizontalBar)))
            .vertical(vertical)
            .minimumDimensions(new Coordinates(fieldWidth, Pixels.ZERO))
            .build();
      } else {
        masterHeader = context.emptyCellBuilder()
            .minimumDimensions(new Coordinates(fieldWidth, Pixels.ZERO))
            .build();
      }
      Cell selectableMasterHeader = context.selectableBorderCellBuilder(masterHeader)
          .backgroundColor(LBProp.COLOR_LABEL)
          .build();
      // In the primitive case, bottom will only be a placeholder (see above).
      if (bottom.getDimensions().getY().isZero() || context.getP(LBProp.INTERNAL_TABLE_HEADER)) {
        /* Even if a master header's label is hidden, there may still be a non-empty space between
        adjactent columns. It is desirable to be able to select this space as if the label was
        there. If we include the empty bottom portion of the header, however, selection requests
        will go to bottom rather than selectableMasterHeader, because bottom comes last and thus is
        the one to be enlarged by justification. So only include selectableMasterHeader in this
        case. */
        ret = context.borderCellBuilder(selectableMasterHeader)
            .minimumDimensions(bottom.getDimensions())
            .build();
      } else if (selectableMasterHeader.getDimensions().getY().isZero()) {
        ret = context.borderCellBuilder(bottom)
            .minimumDimensions(selectableMasterHeader.getDimensions())
            .build();
      } else {
        ret = context.verticalListCellBuilder()
          .addCell(selectableMasterHeader)
          .addCell(context.borderCellBuilder(bottom)
              // TODO[pd]: Consider making this a separate propery.
              // Internal horizontal header separator.
              .borderTop(LBProp.SEP_STYLE_TABLE_FIELD)
              .build())
          .build();
      }
    } else {
      ret = bottom;
    }
    // Set the background color to improve the appearance of borders.
    final Color backgroundColor = headerMode
        ? context.getP(LBProp.COLOR_LABEL)
        : getValueColor(context);
    return context.borderCellBuilder(ret)
        .backgroundColor(backgroundColor)
        .borderTop(borderTop)
        .borderLeft(borderLeft)
        .build();
  }

  private static @Nullable Cell getHorizontalBarCell(
      PrimitiveLayoutContext context, Pixels horizontalBarWidth)
  {
    if (context.getP(LBProp.VISUALIZATION) != VisualizationType.HORIZONTAL_BARS)
      return null;
    if (!horizontalBarWidth.isGreaterThan(Pixels.ZERO))
      return null;
    return context.emptyCellBuilder()
        .minimumWidth(horizontalBarWidth)
        .buildHorizontalBar(context.getP(LBProp.VALUE_FONT),
            CursorStyle.NEGATIVE_BAR_VISUALIZATION_COLOR, CursorStyle.BLUE_CURSOR_COLOR,
            context.getAxis(true), context.getDoubleValue());
  }

  private static Color getValueColor(LayoutContext context) {
    final Color nonHeatmapColor = context.getP(
        context.isAlternativeValueColor() ? LBProp.COLOR_VALUE_ALTERNATIVE : LBProp.COLOR_VALUE);
    if (context.isRelation() || context.getP(LBProp.VISUALIZATION) != VisualizationType.COLOR_RANGE)
      return nonHeatmapColor;
    final Axis axis = context.getPrimitive().getAxis(false);
    final double valueD = context.getPrimitive().getDoubleValue();
    final double position = Math.min(1.0, Math.max(0.0, axis.getRelativePosition(valueD)));
    if (Double.isNaN(valueD))
      return UIUtil.desaturate(nonHeatmapColor);
    final double minHue =   0.0 / 360.0; // Red
    final double maxHue = 120.0 / 360.0; // Green
    return Color.getHSBColor((float) (position * (maxHue - minHue) + minHue), 0.3f, 1.0f);
  }

  private static Cell createTableRow(
      RelationLayoutContext context, Tuple tuple, boolean header, boolean alternativeRowColor,
      LineStyle borderTop, @Nullable Map<LayoutContext,Pixels> horizontalizeCosts)
      throws InterruptedException
  {
    if (header && context.isValue())
      throw new IllegalStateException();
    ListCellBuilder ret = context.horizontalListCellBuilder();
    final LineStyle fieldSeparatorStyle = context.getP(LBProp.SEP_STYLE_TABLE_FIELD);
    int i = 0;
    for (LayoutContext childContext : context.traverseFields(tuple, alternativeRowColor)) {
      if (!childContext.isRelation() && childContext.getPrimitive().isInCrosstabHeader())
        continue;
      final boolean columnSeparatorBefore = (i++ > 0);
      ret.addCell(createTableField(childContext, header, horizontalizeCosts,
          columnSeparatorBefore ? fieldSeparatorStyle : LineStyle.NONE, borderTop));
    }
    return ret.build();
  }

  private static LineStyle getRowSeparatorStyle(
      RelationLayoutContext context, boolean headerMode, int rowIndex)
  {
    if (parentWasTable(context)) {
      return (rowIndex > 0) ? context.getP(LBProp.SEP_STYLE_TABLE_ROW_MINOR) : LineStyle.NONE;
    } else {
      if (rowIndex == 0) {
        return context.getP(headerMode ? LBProp.SEP_STYLE_TABLE_HEADER : LBProp.VALUE_BORDER);
      } else {
        return context.getP(LBProp.SEP_STYLE_TABLE_ROW);
      }
    }
  }

  private static Cell createTableRelationSublayout(RelationLayoutContext context,
      boolean headerMode, @Nullable Map<LayoutContext,Pixels> horizontalizeCosts)
      throws InterruptedException
  {
    if (headerMode && context.isValue())
      throw new IllegalStateException();

    if (context.pendingAuto(LEProp.EFF_ALTERNATE_ROW_COLORS, LAProp.ALTERNATE_ROW_COLORS)) {
      LayoutContext currentContext = context;
      boolean aParentWasAlternating = false;
      while (true) {
        currentContext = currentContext.getParent();
        if (currentContext == null || !currentContext.getP(LEProp.EFF_USE_TABLE))
          break;
        if (currentContext.getP(LEProp.EFF_ALTERNATE_ROW_COLORS))
          aParentWasAlternating = true;
      }
      context.setP(LEProp.EFF_ALTERNATE_ROW_COLORS,
          context.getRelation().isLayoutPlural() && !aParentWasAlternating);
    }
    if (context.getP(LMProp.CROSSTAB_COLUMNS).getPaths().isEmpty()) {
      return createNonCrosstabLayout(context, headerMode, horizontalizeCosts);
    } else {
      return createCrosstabLayout(context, headerMode, horizontalizeCosts);
    }
  }

  private static Cell createNonCrosstabLayout(RelationLayoutContext context,
      boolean headerMode, @Nullable Map<LayoutContext,Pixels> horizontalizeCosts)
      throws InterruptedException
  {
    ListCellBuilder ret = context.verticalListCellBuilder();
    boolean childAlternativeRowColor = context.isAlternativeValueColor();
    int i = 0;
    final TupleInterval tuples = context.getTuples(true);
    if (headerMode || !context.traverseFields().isEmpty()) {
      for (Tuple tuple : tuples) {
        if (Thread.interrupted())
          throw new InterruptedException();
        final LineStyle rowSeparatorStyle = getRowSeparatorStyle(context, headerMode, i++);
        Cell row = createTableRow(context, tuple, headerMode, childAlternativeRowColor,
            rowSeparatorStyle, horizontalizeCosts);
        ret.addCell(row);
        if (context.getP(LEProp.EFF_ALTERNATE_ROW_COLORS))
          childAlternativeRowColor = !childAlternativeRowColor;
      }
    } else {
      /* Create a placeholder cell for relations with no visible fields. This ensures that the
      vertical size of the layout is still roughly proportional to the number of tuples in the
      result, which is important for making infinite scrolling feel right. It might be more logical
      to create placeholders at the child tuple levels, but create it at the relation level since
      there is currently no way to create a LayoutContext or CursorLocation that descends into a
      child tuple without also descending into a specific child path. */
      ret.addCell(context.selectableBorderCellBuilder(context.emptyCellBuilder()
          .minimumHeight(context.getSSPixels(LBProp.OUTLINE_EMPTY_TABLE_SPACE).times(tuples.size()))
          .build())
          .backgroundColor(getValueColor(context))
          .build());
    }
    if (!tuples.isComplete() || tuples.isEmpty()) {
      final Cell fillCell;
      if (tuples.isComplete()) {
        /* For some reason, making this zero pixels tall causes extraneous row borders to appear.
        Enforcing a minimum height is good in any case, though, especially now that fill cells are
        selectable. */
        fillCell = context.emptyCellBuilder()
            .minimumHeight(context.getSSPixels(LBProp.OUTLINE_EMPTY_TABLE_SPACE))
            .build();
      } else {
        final String ellipsisString = context.getP(LBProp.ELLIPSIS_STRING);
        final Pixels ellipsisWidth = context.measureText(
            LBProp.ELLIPSIS_FONT, new SimpleText(ellipsisString), Pixels.ZERO, true, false).getX();
        fillCell = context.textCellBuilder(
            LBProp.ELLIPSIS_COLOR, LBProp.ELLIPSIS_FONT, ellipsisWidth, false)
            .content(ellipsisString)
            .build();
      }
      // TODO[cc]: Avoid duplication of row creation logic here.
      final LineStyle borderTop = getRowSeparatorStyle(context, headerMode, i++);
      /* Allow fill cells to be selected. This might be used in the future to allow an "add row"
      affordance to be reached even from empty subrelations. */
      final Cell borderedFillCell = context.selectableBorderCellBuilder(fillCell)
          .backgroundColor(childAlternativeRowColor
              ? LBProp.COLOR_VALUE_ALTERNATIVE : LBProp.COLOR_VALUE)
          .borderTop(borderTop)
          .build();
      ret.addCell(borderedFillCell);
    }
    return ret.build();
  }

  private static Cell createCrosstabLayout(RelationLayoutContext context,
      boolean headerMode, @Nullable Map<LayoutContext,Pixels> horizontalizeCosts)
      throws InterruptedException
  {
    final LineStyle rowSeparatorStyle = getRowSeparatorStyle(context, headerMode, 0);
    final Pixels minimumWidth;
    {
      RelationLayoutContext placeholderTraversal = context.placeholderTraversal();
      Tuple placeholderTuple = placeholderTraversal.getTuples(false).iterator().next();
      Cell fakeRow = createTableRow(placeholderTraversal, placeholderTuple, true,
          false, LineStyle.NONE, null);
      minimumWidth = fakeRow.getDimensions().getX();
    }
    final ListCellBuilder ret = context.horizontalListCellBuilder();
    int i = 0;
    for (Entry<TupleConstant,Tuple> entry : context.getCrosstabColumnTuples().entrySet()) {
      final TupleConstant tupleConstant = entry.getKey();
      Tuple tuple = entry.getValue();
      final Cell tupleCell;
      final RelationLayoutContext useContext;
      if (tuple == null && headerMode) {
        useContext = context.withAdditionalTupleConstraint(tupleConstant);
        tuple = useContext.placeholderTraversal().getTuples(false).iterator().next();
      } else {
        useContext = context;
      }
      if (tuple == null) {
        tupleCell = context.emptyCellBuilder().minimumWidth(minimumWidth).build();
      } else {
        final Cell row = createTableRow(useContext, tuple, headerMode,
            context.isAlternativeValueColor(), rowSeparatorStyle, null);
        tupleCell = headerMode ? createCrosstabHeader(useContext, tupleConstant, i == 0, row) : row;
      }
      ret.addCell(context.borderCellBuilder(tupleCell)
          .borderLeft(i == 0 ? LineStyle.NONE : context.getP(LBProp.SEP_STYLE_TABLE_FIELD))
          .build());
      i++;
    }
    return ret.build();
  }

  private static Cell createCrosstabHeader(RelationLayoutContext crosstabColumnContext,
      TupleConstant tupleConstraint, boolean firstColumn, Cell regularHeader)
  {
    ListCellBuilder vertical = crosstabColumnContext.verticalListCellBuilder();
    Set<PrimitivePath> headerPaths = crosstabColumnContext.getP(LMProp.CROSSTAB_COLUMNS).getPaths();
    for (LayoutContext fieldContext : crosstabColumnContext.traverseFields()) {
      if (headerPaths.contains(fieldContext.getPath()) && fieldContext.getP(CProp.VISIBLE)) {
        PrimitiveLayoutContext constraintContext = fieldContext.getPrimitive();
        PrimitivePath constraintPath = constraintContext.getPath();
        final String valueStr = constraintContext.getPrimitiveValueFormatter().getFormattedValue(
            new PrimitiveValue(tupleConstraint.getValue(constraintPath).getValue()));
        final Text headerText = firstColumn
            ? constraintContext.getLabelFormatter().getStyledLabel(
              constraintContext.getView(), constraintPath, false, valueStr)
            : new SimpleText(valueStr);
        Cell textCell = constraintContext.textCellBuilder(LBProp.LABEL_TEXT_COLOR,
            LBProp.CROSSTAB_HEADING_FONT, regularHeader.getDimensions().getX(), true)
            .content(headerText)
            .textFitPolicy(TextFitPolicy.FIT_LINE)
            .horizontalAlignment(HorizontalAlignment.CENTER)
            .build();
        vertical.addCell(constraintContext
            .valueSelectableBorderCellBuilder(textCell)
            .backgroundColor(LBProp.COLOR_LABEL)
            .build());
      }
    }
    vertical.addCell(crosstabColumnContext.borderCellBuilder(regularHeader)
        .borderTop(crosstabColumnContext.getP(LBProp.SEP_STYLE_TABLE_FIELD))
        .build());
    return vertical.build();
  }

  private static Cell addBulletCell(
      RelationLayoutContext context, Cell bulletCell, Cell contentCell)
  {
    /* Expand and justify the bullet cell to ensure it has extra space available for baseline
    alignment. */
    // Don't enable highlights for the bullet, so use a regular borderCellBuilder.
    final Cell biasedAndJustifiedBulletCell = context.borderCellBuilder(bulletCell)
        .cursorBias(Direction.RIGHT)
        .justify()
        .minimumDimensions(new Coordinates(Pixels.ZERO, contentCell.getDimensions().getY()))
        .build();
    return context.horizontalListCellBuilder()
        .addCells(biasedAndJustifiedBulletCell, contentCell)
        .build();
  }

  private static Cell createOutlineRelationSublayout(
      RelationLayoutContext context, Pixels width, Cell indentCell) throws InterruptedException
  {
    ListCellBuilder ret = context.verticalListCellBuilder();
    final Cell bulletCell;
    if (context.getRelation().isLayoutPlural()) {
      bulletCell = context.textCellBuilder(LBProp.BULLET_COLOR, LBProp.BULLET_FONT,
          context.getSSPixels(LBProp.OUTLINE_INDENT_WIDTH), true)
          .content(LBProp.BULLET_STRING)
          .build();
    } else {
      bulletCell = indentCell;
    }
    Pixels remainingWidth = width.minus(bulletCell.getDimensions().getX());
    final Cell tupleSpaceAfter = context.emptyCellBuilder()
        .cursorBias(Direction.UP)
        .minimumHeight(context.getSSPixels(LBProp.OUTLINE_TUPLE_SPACE))
        .build();
    final boolean inTable = parentWasTable(context);
    if (inTable) {
      /* For outline layouts embedded in table cells, make sure there is some space before the
      entire layout, not just after each tuple. */
      ret.addCell(context.selectableBorderCellBuilder(context.emptyCellBuilder().build())
        .cursorBias(Direction.DOWN)
        .minimumHeight(context.getSSPixels(LBProp.OUTLINE_TUPLE_SPACE))
        .build());
    }
    final TupleInterval tuples = context.getTuples(true);
    for (Tuple tuple : tuples) {
      if (Thread.interrupted())
        throw new InterruptedException();
      Cell tupleCell = createOutlineTupleLayout(context, tuple, remainingWidth);
      ret.addCell(addBulletCell(context, bulletCell, tupleCell));
      if (context.getRelation().isLayoutPlural() || inTable)
        ret.addCell(tupleSpaceAfter);
    }
    if (!tuples.isComplete()) {
      Cell ellipsisCell = context.textCellBuilder(
          LBProp.ELLIPSIS_COLOR, LBProp.ELLIPSIS_FONT, remainingWidth, true)
          .content(LBProp.ELLIPSIS_STRING)
          .build();
      // Don't enable highlights for outline-level ellipsis, so use a regular borderCellBuilder.
      final Cell biasedEllipsisCell = context.borderCellBuilder(ellipsisCell)
          .cursorBias(Direction.UP)
          .build();
      ret.addCell(addBulletCell(context, bulletCell, biasedEllipsisCell));
      if (context.getRelation().isLayoutPlural() || inTable)
        ret.addCell(tupleSpaceAfter);
    }
    return ret.build();
  }

  // TODO[cc]: Get rid of or move.
  private static boolean parentWasTable(LayoutContext context) {
    return !context.getPath().isRoot() &&
        context.getView().get(LEProp.EFF_USE_TABLE, context.getPath().getParent());
  }

  private static Cell createTableHeader(RelationLayoutContext context) throws InterruptedException {
    return createTableRelationSublayout(context.placeholderTraversal(), true, null);
  }

  // TODO[cc]: Avoid this weird condition.
  /** If outlineLevel is true, adorn with borders and, if relation and USE_TABLE, a header. */
  private static Cell createSublayout(LayoutContext context, Pixels width,
      Pixels requiredIndentationWidth, boolean outlineLevel) throws InterruptedException
  {
    final Cell indentCell = context.emptyCellBuilder()
        .minimumWidth(requiredIndentationWidth)
        .build();
    /* To test for correct uses of width vs. widthAfterIndentation, exaggerate
    OUTLINE_INDENT_WIDTH in the stylesheet. */
    final Pixels widthAfterIndentation = width.minus(indentCell.getDimensions().getX());
    final Pixels widthAfterBorder = widthAfterIndentation.minus(
        context.getP(LBProp.VALUE_BORDER).measure(context.getRenderContext(), true));
    if (context.isRelation()) {
      /* Will only be the case for the root level and inner tables; if createSublayout is called
      recursively, USE_TABLE will already have been set in createOutlineFieldLayout. */
      if (context.pendingAuto(LEProp.EFF_USE_TABLE, LAProp.USE_TABLE)) {
        Preconditions.checkState(outlineLevel || parentWasTable(context));
        context.setP(LEProp.EFF_USE_TABLE,
            wouldUseTable(context.getRelation(), Pixels.ZERO, widthAfterIndentation));
      }
      Cell relationContent;
      final Cell relationHeader;
      if (context.getP(LEProp.EFF_USE_TABLE)) {
        // TODO[pd]: Have header creation happen only once.
        if (!context.isValue() && outlineLevel && context.getP(LBProp.TABLE_JUSTIFY))
          justifyTable(context.getRelation(), widthAfterBorder);

        /* TODO[pd]: Allow internal headers for plural nested tables. In that case, there needs to
                     be margins above and below in addition to on the sides. */
        // Create header first to ensure correct autosetting of stylesheet properties.
        relationHeader = outlineLevel || context.getP(LBProp.INTERNAL_TABLE_HEADER)
            ? createTableHeader(context.getRelation()) : null;
        relationContent = createTableRelationSublayout(context.getRelation(), false, null);
      } else {
        relationHeader = null;
        relationContent = createOutlineRelationSublayout(context.getRelation(), width, indentCell);
      }
      if (!context.isValue() && context.getRelation().isLayoutPlural()) {
        relationContent = context.newBigger(relationContent, new Coordinates(Pixels.ZERO,
            Pixels.create(context.getRenderContext(),
              relationContent.getDimensions().getY().getDouble() *
              context.getP(LMProp.AVERAGE_CARDINALITY))));
      }
      Cell ret;
      if (relationHeader == null) {
        ret = relationContent;
      } else {
        Cell decoratedRelationHeader, decoratedRelationContent;
        /* If the table is contained in an outline column system, absorbing of rounding error in the
        last column may cause the actual desired width of the table to be a few pixels wider than
        the width it was formally justified for. This strut expands the table by these remaining few
        pixels, without bothering to distribute them carefully between columns like justifyTable
        does. It also finally widens the table in cases where justifyTable must give up, as when no
        variable-width primitive fields are visible. The relationHeader/relationContent width part
        of the adjustment serves mainly to deal with the case of a single ellipsis row in the
        content area (there might be better ways to handle this case, such as generating the
        ellipsis row at the right width immediately). */
        final Pixels borderWidth = context.getP(LBProp.VALUE_BORDER)
            .measure(context.getRenderContext(), true);
        final Pixels justifiedWidth =
            (context.getP(LBProp.TABLE_JUSTIFY) ? width : Pixels.ZERO)
            .max(relationHeader.getDimensions().getX().plus(borderWidth))
            .max(relationContent.getDimensions().getX().plus(borderWidth));
        if (relationHeader.getDimensions().getY().isZero()) {
          // May occur in the case of hidden headers.
          decoratedRelationHeader = context.borderCellBuilder(relationHeader)
              .justify()
              .minimumWidth(justifiedWidth)
              .build();
          decoratedRelationContent = context.borderCellBuilder(relationContent)
              .borders(LBProp.VALUE_BORDER)
              .backgroundColor(getValueColor(context))
              .justify()
              .minimumWidth(justifiedWidth)
              .build();
        } else {
          // Top borders are handled by getRowSeparatorStyle.
          /* Use a VALUE_BORDER here rather than an OUTLINE_LABEL_BORDER, since the header must be
          aligned with the content. */
          decoratedRelationHeader = context.borderCellBuilder(relationHeader)
              .borderRight(LBProp.VALUE_BORDER)
              .borderBottom(LBProp.VALUE_BORDER) // For the benefit of frozen headers.
              .borderLeft(LBProp.VALUE_BORDER)
              .backgroundColor(LBProp.COLOR_LABEL)
              .justify()
              .minimumWidth(justifiedWidth)
              .build();
          decoratedRelationContent = context.borderCellBuilder(relationContent)
              .borderRight(LBProp.VALUE_BORDER)
              .borderBottom(LBProp.VALUE_BORDER)
              .borderLeft(LBProp.VALUE_BORDER)
              .backgroundColor(LBProp.COLOR_VALUE)
              .justify()
              .minimumWidth(justifiedWidth)
              .build();
        }
        ret = context.headeredCellBuilder(true, decoratedRelationHeader, decoratedRelationContent)
            .build();
      }
      if (context.getP(LEProp.EFF_USE_TABLE))
        return context.horizontalListCellBuilder().addCells(indentCell, ret).build();
      return ret;
    } else {
      Pixels maxContentWidth = outlineLevel ? widthAfterBorder : widthAfterIndentation;
      if (outlineLevel)
        maxContentWidth = maxContentWidth.max(context.getSSPixels(LHProp.OUTLINE_MIN_VALUE_WIDTH));
      // Does not include border.
      Pixels contentWidth;
      if (context.getP(LMProp.VARIABLE_WIDTH) ||
          context.getP(LBProp.VISUALIZATION) == VisualizationType.HORIZONTAL_BARS)
      {
        contentWidth = maxContentWidth;
      } else {
        contentWidth = context.getSSPixels(LMProp.VALUE_WIDTH);
        if (outlineLevel) {
          /* TODO[cc]: Handle out-of-range View properties either more broadly or in a more general
                       way than through one-off adjustments like the below EPSILON limit. */
          contentWidth = contentWidth
              .max(context.getSSPixels(LHProp.OUTLINE_MIN_VALUE_WIDTH))
              .roundUp(context.getSSPixels(LHProp.OUTLINE_VALUE_SNAP_WIDTH).max(Pixels.EPSILON))
              .max(context.getSSPixels(LHProp.OUTLINE_VALUE_SNAP_WIDTH));
        }
        contentWidth = contentWidth.min(maxContentWidth);
      }
      final Cell contentCell;
      if (context.isValue()) {
        if (parentWasTable(context)) {
          // TODO[cc]: Don't bother with the contentWidth calculation above in this case. Clean up.
          contentWidth = context.getP(LEProp.EFF_TABLE_COLUMN_WIDTH);
        }
        if (!context.isProjected()) {
          final String ellipsisString = context.getP(LBProp.ELLIPSIS_STRING);
          contentCell = context.textCellBuilder(
              LBProp.ELLIPSIS_COLOR, LBProp.ELLIPSIS_FONT, contentWidth, false)
              .content(ellipsisString)
              .minimumDimensions(new Coordinates(contentWidth, Pixels.ZERO))
              .build();
        } else {
          final TextFitPolicy textFitPolicy = context.getP(LMProp.VARIABLE_WIDTH)
              ? TextFitPolicy.PARAGRAPH : TextFitPolicy.FIT_LINE;
          final Pixels valueWidth = context.getSSPixels(LMProp.VALUE_WIDTH);
          final @Nullable Cell horizontalBar =
              getHorizontalBarCell(context.getPrimitive(), contentWidth.minus(valueWidth));
          if (horizontalBar != null) {
            contentCell = context.horizontalListCellBuilder()
                .addCell(context.getPrimitive().buildValueCell(valueWidth, textFitPolicy))
                .addCell(horizontalBar)
                .build();
          } else {
            contentCell = context.getPrimitive().buildValueCell(contentWidth, textFitPolicy);
          }
        }
      } else {
        // TODO[cc]: Consider using PlaceholderData.valueIsPlaceholder().
        // TODO[cc]: Make this less ugly.
        Pixels oneLineHeight =
            context.measureText(LBProp.VALUE_FONT, Text.EMPTY, contentWidth, false, false).getY();
        /* Since contentWidth is derived from VALUE_WIDTH, I don't think fraction below will blow up
        under normal circumstances. The case where both the numerator and the denominator is zero is
        handled specially. */
        double factor = Math.max(1.0,
            contentWidth.getDouble() <= 0.0 ? 1.0 :
            (context.getSSPixels(LMProp.VALUE_WIDTH).getDouble() / contentWidth.getDouble()));
        Pixels height =
            Pixels.create(context.getRenderContext(), oneLineHeight.getDouble() * factor);
        contentCell = context.emptyCellBuilder()
            .cursorBias(Direction.UP)
            .minimumDimensions(new Coordinates(contentWidth, height))
            .build();
      }
      final Cell borderedContentCell;
      if (outlineLevel) {
        borderedContentCell = context
            .valueSelectableBorderCellBuilder(contentCell)
            .backgroundColor(getValueColor(context))
            .borders(LBProp.VALUE_BORDER)
            .build();
      } else {
        borderedContentCell = context
            .valueSelectableBorderCellBuilder(contentCell)
            .backgroundColor(getValueColor(context))
            .build();
      }
      return context.horizontalListCellBuilder()
          .addCells(indentCell, borderedContentCell)
          .build();
    }
  }

  private static void debugSetProperties(LayoutContext context) {
    Preconditions.checkArgument(!context.isValue());
    if (context.toString().equals("@/Course Offerings/Sections/Meetings/End Time")) {
      context.setP(LBProp.COLOR_LABEL, new Color(0, 255, 0));
      context.setP(LBProp.COLOR_VALUE, new Color(196, 255, 196));
      context.setP(LBProp.COLOR_VALUE_ALTERNATIVE, new Color(127, 255, 127));
    }
    if (context.isRelation()) {
      for (LayoutContext field : context.getRelation().traverseFields())
        debugSetProperties(field);
    }
  }

  private static Coordinates getBorderOverprint(LayoutContext context, View view) {
    Preconditions.checkArgument(!context.isValue());
    Coordinates ret = new Coordinates(
        context.getP(LBProp.VALUE_BORDER).measure(context.getRenderContext(), true),
        context.getP(LBProp.VALUE_BORDER).measure(context.getRenderContext(), false));
    if (context.isRelation()) {
      for (LayoutContext field : context.getRelation().traverseFields())
        ret = ret.max(getBorderOverprint(field, view));
    }
    return ret;
  }

  /* ******************************************************************************************** */

  public static Cell autoStyle(
      Data data, MutableView view, RenderContext renderContext, Pixels width,
      LabelFormatter labelFormatter)
      throws InterruptedException
  {
    LayoutContext placeholderContext =
        LayoutContext.startTraversal(data, renderContext, view, false, labelFormatter);
    Pixels layoutWidth = width.minus(getBorderOverprint(placeholderContext, view).getX());
    measureLabels(placeholderContext, null);
    if (false)
      debugSetProperties(placeholderContext);
    return createSublayout(placeholderContext, layoutWidth, Pixels.ZERO, true);
  }

  public static Cell createLayout(
      Data data, ImmutableView view, RenderContext renderContext, Pixels width,
      LabelFormatter labelFormatter)
      throws InterruptedException
  {
    LayoutContext context =
        LayoutContext.startTraversal(data, renderContext, view, true, labelFormatter);
    Cell ret = createSublayout(context,
        width.minus(getBorderOverprint(context.placeholderTraversal(), view).getX()), Pixels.ZERO,
        true);
    // TODO[cc]: Consider finding a better way to add this label.
    final Pixels preferredLabelWidth = ret.getDimensions().getX().minus(
        context.getP(LBProp.OUTLINE_LABEL_BORDER).measure(context.getRenderContext(), true));
    final Pixels minimumLabelWidth = context.measureLabel(preferredLabelWidth, false, false).getX();
    return new OutlineField(context, ret, preferredLabelWidth.max(minimumLabelWidth), false)
        .getCell(Pixels.ZERO);
  }
}
