// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.google.common.base.Preconditions;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.Points;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.proptypes.OptionalBigDecimal;
import com.sieuferd.datamodel.proptypes.OptionalInteger;
import com.sieuferd.layout.PathStats.ImmutablePathStats;
import com.sieuferd.layout.PathStats.MutablePathStats;
import com.sieuferd.layout.internal.LayoutContext;
import com.sieuferd.util.StringUtil;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.ObjectReference;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.math.BigDecimal;
import java.util.Map;
import javax.annotation.Nullable;

// TODO[pd]: Consider calling Thread.interrupted() less frequently, in case it is expensive.
// TODO[pd]: Consider measuring only parts of long fields, extrapolating based on character length.
public class LayoutMeasurer {
  private LayoutMeasurer() { }

  public static final ImmutableSet<ViewProperty<?>> MEASUREMENT_INPUTS =
      ImmutableSet.<ViewProperty<?>>of(
      SProp.CHILD_PATHS, LMProp.SAMPLE_SIZE, LBProp.VALUE_FONT, LHProp.TABLE_MAX_PRIMITIVE_WIDTH,
      LBProp.TEMPORAL_STYLE, LBProp.DISPLAY_LOCALE, LBProp.DISPLAY_TIME_ZONE,
      LMProp.OVERRIDE_MEASUREMENTS, LMProp.TEMPORAL_PRECISION);
  private static final int MINIMUM_VARCHAR_CHARS = 10;
  private static final double CONFIDENCE_STANDARD_DEVIATIONS = 3.0;
  private static final double GENERIC_RCX_SAFETY_MARGIN = 0.10;
  private static final int TARGET_SAMPLE_SIZE = 50;
  private static final int SMALL_SAMPLE_SIZE = 15;
  private static final double MINIMUM_SAMPLE_SIZE_INCREASE_FACTOR = 1.75;
  private static final int MAXIMUM_SIGNIFICANT_DIGITS = 4;
  /** Rationale: for numbers smaller than this, it would be better to use different units. */
  private static final int MAXIMUM_DECIMAL_PLACES = 6;

  private static RenderContext createStandardizedRenderContext() {
    // Use the same render context as when printing to the Adobe PDF Printer on Windows at 1200dpi.
    final double PIXELS_PER_INCH = 1200; // DPI
    final double POINTS_PER_INCH = 72.0;
    final double PIXELS_PER_POINT = PIXELS_PER_INCH / POINTS_PER_INCH;
    FontRenderContext frc = new FontRenderContext(
        new AffineTransform(PIXELS_PER_POINT, 0.0, 0.0, PIXELS_PER_POINT, 0.0, 0.0),
        false, true);
    return new RenderContext(frc, 1.0, null, false, false);
  }

  /**
   * Reads MEASUREMENT_INPUTS and sets LMProp properties on view. Potentially long-running
   * operation.
   */
  public static void measureData(Data data, MutableView view, @Nullable RenderContext renderContext)
      throws InterruptedException
  {
    ThreadUtil.checkNotOnEDT();

    final LayoutContext dataTraversal = LayoutContext.startTraversal(data,
        (renderContext == null) ? createStandardizedRenderContext() : renderContext,
        view, true, LabelFormatter.SIMPLE);
    final LayoutContext viewTraversal = dataTraversal.placeholderTraversal();

    // Values are new sample sizes.
    final ImmutableMap<PrimitivePath,Integer> primitivesToMeasure;
    {
      // TODO[pd]: Consider just storing a set of hash codes instead.
      Multimap<PrimitivePath,Object> uniqueValues = HashMultimap.create();
      MutablePathStats<PrimitivePath> toDecimalPlaceStats = MutablePathStats.create();
      MutablePathStats<PrimitivePath> toDecimalValueStats = MutablePathStats.create();
      measureFirstPass(dataTraversal, uniqueValues, toDecimalPlaceStats, toDecimalValueStats);
      Map<PrimitivePath,Integer> toPrimitivesToMeasure = Maps.newLinkedHashMap();
      determinePrimitivesToMeasure(viewTraversal, uniqueValues, toPrimitivesToMeasure);
      primitivesToMeasure = ImmutableMap.copyOf(toPrimitivesToMeasure);
      setNumerics(viewTraversal,
          toDecimalPlaceStats.freeze(), toDecimalValueStats.freeze(), primitivesToMeasure);
    }

    if (!primitivesToMeasure.isEmpty()) {
      final ImmutablePathStats<PrimitivePath> valueWidthStats;
      {
        MutablePathStats<PrimitivePath> toValueWidthStats = MutablePathStats.create();
        MutablePathStats<PrimitivePath> toCharacterCountStats = MutablePathStats.create();
        MutablePathStats<RelationPath> toCardinalityStats = MutablePathStats.create();
        MutablePathStats<RelationPath> toIncompleteCardinalityStats = MutablePathStats.create();
        measureValueStatistics(dataTraversal, toValueWidthStats, toCharacterCountStats,
            toCardinalityStats, toIncompleteCardinalityStats, primitivesToMeasure);
        valueWidthStats = toValueWidthStats.freeze();
        setMeasuredProperties(viewTraversal, valueWidthStats,
            toCharacterCountStats.freeze(), toCardinalityStats.freeze(),
            toIncompleteCardinalityStats.freeze(), primitivesToMeasure);
      }

      {
        MutablePathStats<PrimitivePath> toValueWidthStatsNBB = MutablePathStats.create();
        measureValueNBBStatistics(dataTraversal, toValueWidthStatsNBB, primitivesToMeasure);
        double useSafetyMargin = (renderContext == null) ? GENERIC_RCX_SAFETY_MARGIN : 0.0;
        applyNBBAdjustments(viewTraversal, valueWidthStats, toValueWidthStatsNBB.freeze(),
            primitivesToMeasure, useSafetyMargin);
      }
    }
  }

  private static void measureFirstPass(final LayoutContext context,
      final Multimap<PrimitivePath,Object> uniqueValues,
      final MutablePathStats<PrimitivePath> decimalPlaceStats,
      final MutablePathStats<PrimitivePath> decimalValueStats)
      throws InterruptedException
  {
    if (context.isRelation()) {
      for (Tuple tuple : context.getRelation().getTuples(false)) {
        if (Thread.interrupted())
          throw new InterruptedException();
        for (LayoutContext fieldContext : context.getRelation().traverseFields(tuple))
          measureFirstPass(fieldContext, uniqueValues, decimalPlaceStats, decimalValueStats);
      }
    } else {
      final PrimitiveValue value = context.getPrimitive().getValue();
      final Object valueObj = value.getContent();
      if (!isValueEmpty(value))
        uniqueValues.put(context.getPrimitive().getPath(), valueObj);
      final @Nullable BigDecimal bd = (valueObj == null) ? null :
          PrimitiveValueFormatter.normalizeNumber(valueObj);
      if (bd != null) {
        decimalValueStats.addValue(context.getPrimitive().getPath(), bd);

        // TODO[pd]: Consider collecting decimal place statistics in a separate pass.
        /* Collect decimal place statistics in the first pass, since it should be cheap. Only bother
        doing it if the sample size is less than the target, though. */
        if (context.getP(LMProp.SAMPLE_SIZE) < TARGET_SAMPLE_SIZE) {
          /* Don't read any special significance into the number trailing zeros, as the BigDecimal
          object may come from a variety of sources (e.g. JDBC drivers) each with their own
          potentially buggy policy. Round to MAXIMUM_DECIMAL_PLACES before calling
          stripTrailingZeros to deal with decimal values that have been converted to binary floating
          point and back, e.g. 2.60000000000000008881. */
          final BigDecimal adjBD = bd
              .setScale(MAXIMUM_DECIMAL_PLACES, PrimitiveValueFormatter.ROUNDING_MODE)
              .stripTrailingZeros();
          final int existingSignificantDigits = adjBD.precision();
          final int existingDecimalPlaces = adjBD.scale();
          final int desiredSignificantDigits =
              Math.min(MAXIMUM_SIGNIFICANT_DIGITS, existingSignificantDigits);
          /* Allow desiredDecimalPlaces to be negative while collecting statistics; this would mean
          rounding to show no more than the desired number of significant digits. A policy of
          avoiding negative values can be implemented in setDecimalPlaces. */
          final int desiredDecimalPlaces =
              existingDecimalPlaces - (existingSignificantDigits - desiredSignificantDigits);
          decimalPlaceStats.addValue(context.getPrimitive().getPath(), desiredDecimalPlaces);
        }
      }
    }
  }

  private static boolean isValueEmpty(PrimitiveValue value) {
    Preconditions.checkNotNull(value);
    final @Nullable Object valueObj = value.getContent();
    return valueObj == null ||
        valueObj instanceof String && StringUtil.isEmptyOrSpaceOnly((String) valueObj) ||
        valueObj instanceof Character && Character.isWhitespace((Character) valueObj);
  }

  private static void determinePrimitivesToMeasure(
      LayoutContext context, Multimap<PrimitivePath,Object> uniqueValues,
      Map<PrimitivePath,Integer> primitivesToMeasure)
  {
    Preconditions.checkArgument(!context.isValue());
    if (context.isRelation()) {
      for (LayoutContext fieldContext : context.getRelation().traverseFields())
        determinePrimitivesToMeasure(fieldContext, uniqueValues, primitivesToMeasure);
    } else {
      final PrimitivePath path = context.getPrimitive().getPath();
      final int oldSampleSize = context.getP(LMProp.SAMPLE_SIZE);
      final int newSampleSize = uniqueValues.get(path).size();
      if (!context.getP(LMProp.OVERRIDE_MEASUREMENTS) &&
          oldSampleSize < TARGET_SAMPLE_SIZE && (
          oldSampleSize < SMALL_SAMPLE_SIZE && newSampleSize > oldSampleSize ||
          newSampleSize > oldSampleSize * MINIMUM_SAMPLE_SIZE_INCREASE_FACTOR))
      {
        primitivesToMeasure.put(path, newSampleSize);
      }
    }
  }

  private static void setNumerics(
      LayoutContext context, ImmutablePathStats<PrimitivePath> decimalPlaceStats,
      ImmutablePathStats<PrimitivePath> decimalValueStats,
      ImmutableMap<PrimitivePath,Integer> primitivesToMeasure)
  {
    Preconditions.checkArgument(!context.isValue());
    if (context.isRelation()) {
      for (LayoutContext fieldContext : context.getRelation().traverseFields())
        setNumerics(fieldContext, decimalPlaceStats, decimalValueStats, primitivesToMeasure);
    } else {
      final PrimitivePath path = context.getPrimitive().getPath();
      final OptionalBigDecimal valueMax =
          OptionalBigDecimal.ofNullable(decimalValueStats.getMaxBD(path));
      final OptionalBigDecimal valueMin =
          OptionalBigDecimal.ofNullable(decimalValueStats.getMinBD(path));
      if (!context.getP(LMProp.OVERRIDE_MEASUREMENTS)) {
        // Always set, even if path is not in primitivesToMeasure.
        context.setP(LMProp.NUMERIC_MAX, valueMax);
        context.setP(LMProp.NUMERIC_MIN, valueMin);
      }
      if (primitivesToMeasure.containsKey(path)) {
        final double statMax = decimalPlaceStats.getMax(path);
        final OptionalInteger decimalPlaces;
        final boolean groupDigits;
        if (Double.isNaN(statMax)) {
          decimalPlaces = OptionalInteger.absent();
          groupDigits = false;
        } else {
          decimalPlaces = OptionalInteger.of(
              Math.max(0, Math.min(MAXIMUM_DECIMAL_PLACES, (int) Math.round(statMax))));
          /* Avoid grouping digits in columns containing only integers of four digits or less. This
          avoids the particular undesired case of inserting a thousands separator in year numbers.
          There is also a similar policy in the SI Brouchure, Section 5.3.4: "when there are only
          four digits before or after the decimal marker, it is customary not to use a space to
          isolate a single digit"
          ( http://www.bipm.org/en/publications/si-brochure/section5-3-4.html ). */
          groupDigits =
              decimalPlaces.isPresent() && decimalPlaces.get() > 0 ||
              valueMax.isPresent() && valueMax.get().compareTo(BigDecimal.valueOf(10000)) >= 0 ||
              valueMin.isPresent() && valueMin.get().compareTo(BigDecimal.valueOf(-10000)) <= 0;
        }
        context.setP(LMProp.DECIMAL_PLACES, decimalPlaces);
        context.setP(LMProp.GROUP_DIGITS, groupDigits);
      }
    }
  }

  /* TODO[pd]: Consider caching stats per unique string value to take advantage of repeated
               identical values. */
  /**
   * @param widthStats in points
   */
  private static void measureValueStatistics(final LayoutContext context,
      final MutablePathStats<PrimitivePath> widthStats,
      final MutablePathStats<PrimitivePath> characterCountStats,
      final MutablePathStats<RelationPath> cardinalityStats,
      final MutablePathStats<RelationPath> incompleteCardinalityStats,
      final ImmutableMap<PrimitivePath,Integer> primitivesToMeasure)
      throws InterruptedException
  {
    Preconditions.checkArgument(context.isValue());
    if (context.isRelation()) {
      int tupleCount = 0;
      final TupleInterval tuples = context.getRelation().getTuples(false);
      for (Tuple tuple : tuples) {
        if (Thread.interrupted())
          throw new InterruptedException();
        for (LayoutContext fieldContext : context.getRelation().traverseFields(tuple)) {
          measureValueStatistics(fieldContext, widthStats, characterCountStats,
              cardinalityStats, incompleteCardinalityStats, primitivesToMeasure);
        }
        tupleCount++;
      }
      (tuples.isComplete() ? cardinalityStats : incompleteCardinalityStats)
          .addValue(context.getPath().getRelation(), tupleCount);
    } else {
      final PrimitivePath path = context.getPrimitive().getPath();
      if (primitivesToMeasure.containsKey(path)) {
        if (!isValueEmpty(context.getPrimitive().getValue())) {
          final ObjectReference<String> formattedValueRef = ObjectReference.create();
          Pixels width =
              context.getPrimitive().measureValue(null, false, false, formattedValueRef).getX();
          characterCountStats.addValue(path, formattedValueRef.getValue().length());
          widthStats.addValue(path, width.toPoints(context.getRenderContext()).getValue());
        }
      }
    }
  }

  /**
   * @param widthStats in points
   */
  private static void setMeasuredProperties(final LayoutContext context,
      final ImmutablePathStats<PrimitivePath> widthStats,
      final ImmutablePathStats<PrimitivePath> characterCountStats,
      final ImmutablePathStats<RelationPath> cardinalityStats,
      final ImmutablePathStats<RelationPath> incompleteCardinalityStats,
      final ImmutableMap<PrimitivePath,Integer> primitivesToMeasure)
  {
    Preconditions.checkArgument(!context.isValue());
    if (context.isRelation()) {
      for (LayoutContext fieldContext : context.getRelation().traverseFields()) {
        setMeasuredProperties(fieldContext, widthStats, characterCountStats, cardinalityStats,
            incompleteCardinalityStats, primitivesToMeasure);
      }
      {
        final RelationPath path = context.getRelation().getPath();
        double mean = cardinalityStats.getMean(path);
        double N = cardinalityStats.getN(path);
        if (Double.isNaN(mean)) {
          mean = incompleteCardinalityStats.getMean(path);
          N = incompleteCardinalityStats.getN(path);
        }
        final int oldSampleSize = context.getP(LMProp.SAMPLE_SIZE);
        final int newSampleSize = (int) Math.min(Integer.MAX_VALUE, N);
        if (!Double.isNaN(mean) &&
            oldSampleSize < TARGET_SAMPLE_SIZE &&
            newSampleSize > oldSampleSize)
        {
          context.setP(LMProp.AVERAGE_CARDINALITY, mean);
          context.setP(LMProp.SAMPLE_SIZE, newSampleSize);
        }
      }
    } else {
      final PrimitivePath path = context.getPrimitive().getPath();
      final Integer sampleSize = primitivesToMeasure.get(path);
      if (sampleSize != null) {
        Boolean variableWidth = null;
        final double widthMax = widthStats.getMax(path);
        final double widthMean = widthStats.getMean(path);
        final double widthStDev = widthStats.getUnbiasedStandardDeviation(path);
        if (context.getP(LMProp.DECIMAL_PLACES).isPresent()) {
          variableWidth = false;
        } else {
          final double charMax = characterCountStats.getMax(path);
          if (charMax >= MINIMUM_VARCHAR_CHARS)
            variableWidth = true;
          if (charMax > 0.0 && characterCountStats.getVariance(path) == 0.0 &&
              widthMax < context.getP(LHProp.TABLE_MAX_PRIMITIVE_WIDTH).getValue())
          {
            variableWidth = false;
          }
        }
        final double widthPoints;
        final double confidencePoints;
        if (Double.isNaN(widthMax) || widthMax == 0.0) {
          // Special case for no recorded values.
          if (variableWidth == null)
            variableWidth = true;
          // TODO[pd]: Consider if zero is the correct default.
          widthPoints = 0.0;
          confidencePoints = 0.0;
        } else {
          if (variableWidth == null)
            variableWidth = false;
          widthPoints = variableWidth ? widthMean : widthMax;
          confidencePoints = Double.isFinite(widthStDev)
              ? widthMean + CONFIDENCE_STANDARD_DEVIATIONS * widthStDev
              : widthMax;
        }
        context.setP(LMProp.VARIABLE_WIDTH, variableWidth);
        context.setP(LMProp.VALUE_WIDTH, new Points(widthPoints));
        context.setP(LMProp.CONFIDENCE_WIDTH, new Points(confidencePoints));
        context.setP(LMProp.SAMPLE_SIZE, sampleSize);
      }
    }
  }

  private static void measureValueNBBStatistics(final LayoutContext context,
      final MutablePathStats<PrimitivePath> widthStatsNBB,
      final ImmutableMap<PrimitivePath,Integer> primitivesToMeasure)
      throws InterruptedException
  {
    Preconditions.checkArgument(context.isValue());
    if (context.isRelation()) {
      for (Tuple tuple : context.getRelation().getTuples(false)) {
        if (Thread.interrupted())
          throw new InterruptedException();
        for (LayoutContext fieldContext : context.getRelation().traverseFields(tuple))
          measureValueNBBStatistics(fieldContext, widthStatsNBB, primitivesToMeasure);
      }
    } else {
      final PrimitivePath path = context.getPrimitive().getPath();
      /* The TABLE_MAX_PRIMITIVE_WIDTH condition is an optimization to ensure that we don't spend
      time trying to find the width of the longest word in a field that is going to have the maximum
      allowable table column width anyway. */
      if (primitivesToMeasure.containsKey(path) &&
          context.getP(LMProp.VARIABLE_WIDTH) &&
          context.getSSPixels(LHProp.TABLE_MAX_PRIMITIVE_WIDTH).isGreaterThan(
              context.getSSPixels(LMProp.VALUE_WIDTH)))
      {
        if (!isValueEmpty(context.getPrimitive().getValue())) {
          Pixels widthNBB =
              context.getPrimitive().measureValue(Pixels.ZERO, true, false, null).getX();
          widthStatsNBB.addValue(path, widthNBB.toPoints(context.getRenderContext()).getValue());
        }
      }
    }
  }

  /**
   * @param widthStats in points
   * @param widthStatsNBB in points
   */
  private static void applyNBBAdjustments(final LayoutContext context,
      final ImmutablePathStats<PrimitivePath> widthStats,
      final ImmutablePathStats<PrimitivePath> widthStatsNBB,
      final ImmutableMap<PrimitivePath,Integer> primitivesToMeasure,
      final double safetyMargin)
  {
    Preconditions.checkArgument(!context.isValue());
    if (context.isRelation()) {
      for (LayoutContext fieldContext : context.getRelation().traverseFields()) {
        applyNBBAdjustments(
            fieldContext, widthStats, widthStatsNBB, primitivesToMeasure, safetyMargin);
      }
    } else {
      final PrimitivePath path = context.getPrimitive().getPath();
      if (primitivesToMeasure.containsKey(context.getPrimitive().getPath())) {
        final double confidenceWidthPoints, noBadBreaksWidthPoints;
        {
          /* Use the confidence width as an upper bound on the potential impact of rare values with
          long non-breakable widths. */
          confidenceWidthPoints = context.getP(LMProp.CONFIDENCE_WIDTH).getValue();
          noBadBreaksWidthPoints = widthStatsNBB.getMax(path);
        }
        context.setP(LMProp.VALUE_WIDTH, new Points((1.0 + safetyMargin) * Math.max(
            context.getP(LMProp.VALUE_WIDTH).getValue(),
            /* This adjustment is useful in particular for avoiding bad breaks in table columns when
            table justification is not applied (e.g. in wider-than-viewport table layouts). */
            Double.isNaN(noBadBreaksWidthPoints) ? 0.0 :
            Math.min(noBadBreaksWidthPoints, confidenceWidthPoints)
        )));
      }
    }
  }
}
