// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.sieuferd.cell.styles.HorizontalAlignment;
import com.google.common.collect.ImmutableBiMap;
import com.sieuferd.cell.support.Points;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.ViewPropertyType;
import com.sieuferd.datamodel.ViewPropertyTypes;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.ExpressionParseException;
import com.sieuferd.datamodel.expressions.Expressions;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public final class LayoutPropertyTypes {
  // Package-visible for testing.
  static final Pattern DECIMAL_WITH_UNITS_PATTERN = Pattern.compile("^(.*[0-9]) *([A-Za-z]*) *$");

  private LayoutPropertyTypes() { }

  public static final ViewPropertyType<Points> POINTS = new ViewPropertyType<Points>(Points.class) {
    @Override
    public String valueToString(Strepco strepco, Points value) {
      return formatDouble(value.getValue()) + "pt";
    }

    private String formatDouble(double value) {
      // TODO[cc]: Consider doing something more elegant.
      long longCast = (long) value;
      return (value == longCast) ? Long.toString(longCast) : Double.toString(value);
    }

    @Override
    public Points stringToValue(Strepco strepco, String str)
        throws IllegalArgumentException
    {
      Matcher matcher = DECIMAL_WITH_UNITS_PATTERN.matcher(str);
      if (!matcher.matches())
        throw new IllegalArgumentException("Expected a number followed by an optional unit (pt)");
      String unit = matcher.group(2).toLowerCase(Locale.ROOT);
      if (!(unit.isEmpty() || unit.equals("pt"))) {
        throw new IllegalArgumentException(
            "Unsupported unit '" + unit + "'; please enter a value in points (pt)");
      }
      // Let any NumberFormatException propagate.
      double value = Double.parseDouble(matcher.group(1));
      if (!Double.isFinite(value))
        throw new IllegalArgumentException("Finite number required, got " + value);
      return new Points(value);
    }
  };

  public static final ViewPropertyType<HorizontalAlignment> HORIZONTAL_ALIGNMENT =
      ViewPropertyTypes.createSimpleMappingType(HorizontalAlignment.class, ImmutableBiMap.of(
      "left", HorizontalAlignment.LEFT,
      "center", HorizontalAlignment.CENTER,
      "right", HorizontalAlignment.RIGHT));

  public static final ViewPropertyType<YesNoAuto> YES_NO_AUTO =
      ViewPropertyTypes.createSimpleMappingType(YesNoAuto.class, ImmutableBiMap.of(
      "true", YesNoAuto.YES, "false", YesNoAuto.NO, "auto", YesNoAuto.AUTO));

  public static final ViewPropertyType<TemporalStyle> TEMPORAL_STYLE =
      ViewPropertyTypes.createEnumMappingType(TemporalStyle.class);

  public static final ViewPropertyType<TemporalPrecision> TEMPORAL_PRECISION =
      ViewPropertyTypes.createEnumMappingType(TemporalPrecision.class);

  public static final ViewPropertyType<VisualizationType> VISUALIZATION_TYPE =
      ViewPropertyTypes.createEnumMappingType(VisualizationType.class);

  public static final ViewPropertyType<CrosstabColumns> CROSSTAB_COLUMNS =
      new ViewPropertyType<CrosstabColumns>(CrosstabColumns.class)
  {
    @Override
    public String valueToString(Strepco strepco, CrosstabColumns value) {
      return Expressions.generateString(strepco, value.toExpression());
    }

    @Override
    public CrosstabColumns stringToValue(Strepco strepco, String str)
        throws IllegalArgumentException
    {
      Expression expression;
      try {
        expression = Expressions.parse(strepco, str);
      } catch (ExpressionParseException e) {
        throw new IllegalArgumentException("Failed to parse expression", e);
      }
      return CrosstabColumns.fromExpression(expression);
    }

    @Override
    public CrosstabColumns rewritePathReferences(PathMap pathMap, CrosstabColumns input) {
      return CrosstabColumns.fromExpression(
          Expressions.rewritePathReferences(pathMap, input.toExpression()));
    }
  };
}
