// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.Path;
import com.sieuferd.layout.PathStats.MutablePathStats;
import java.math.BigDecimal;
import java.util.Map;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import org.apache.commons.math3.stat.descriptive.SummaryStatistics;

public abstract class PathStats<P extends Path> {
  private final SummaryStatistics emptyStatistics = new SummaryStatistics();
  protected final Map<P,SummaryStatistics> map;
  protected boolean frozen;

  protected PathStats(Map<P,SummaryStatistics> map, boolean frozen) {
    Preconditions.checkNotNull(map);
    this.map = map;
    this.frozen = frozen;
  }

  /** Should be called by subclasses only. */
  protected final SummaryStatistics getStatistics(P path) {
    SummaryStatistics ret = map.get(path);
    if (ret == null) {
      if (frozen) {
        return emptyStatistics;
      } else {
        ret = new SummaryStatistics();
        map.put(path, ret);
      }
    }
    return ret;
  }

  public double getMax(P path) {
    return getStatistics(path).getMax();
  }

  public double getMin(P path) {
    return getStatistics(path).getMin();
  }

  public @Nullable BigDecimal getMaxBD(P path) {
    double ret = getMax(path);
    return Double.isFinite(ret) ? BigDecimal.valueOf(ret) : null;
  }

  public @Nullable BigDecimal getMinBD(P path) {
    double ret = getMin(path);
    return Double.isFinite(ret) ? BigDecimal.valueOf(ret) : null;
  }

  public long getN(P path) {
    return getStatistics(path).getN();
  }

  public double getMean(P path) {
    return getStatistics(path).getMean();
  }

  public double getStandardDeviation(P path) {
    return getStatistics(path).getStandardDeviation();
  }

  public double getVariance(P path) {
    return getStatistics(path).getVariance();
  }

  // TODO[cc]: Consider putting this in a utility class, and using it from ProfileDriver.
  public double getUnbiasedStandardDeviation(P path) {
    final long N = getN(path);
    return getStandardDeviation(path) * Math.sqrt(N / (N - 1.0));
  }

  static class MutablePathStats<P extends Path> extends PathStats<P> {
    private MutablePathStats() {
      super(Maps.newLinkedHashMap(), false);
    }

    public static <P extends Path> MutablePathStats<P> create() {
      return new MutablePathStats<>();
    }

    public void addValue(P path, long value) {
      addValue(path, BigDecimal.valueOf(value));
    }

    public void addValue(P path, BigDecimal value) {
      addValue(path, value.doubleValue());
    }

    public void addValue(P path, double value) {
      Preconditions.checkState(!frozen);
      getStatistics(path).addValue(value);
    }

    public ImmutablePathStats<P> freeze() {
      frozen = true;
      return new ImmutablePathStats<>(map);
    }
  }

  @Immutable
  static class ImmutablePathStats<P extends Path> extends PathStats<P> {
    /** @param map must not be modified after this constructor has been called */
    private ImmutablePathStats(Map<P,SummaryStatistics> map) {
      super(map, true);
    }
  }
}
