// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.ViewPropertySource;
import com.sieuferd.datamodel.proptypes.OptionalInteger;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.AttributedCharacterIterator;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.time.DateTimeException;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.temporal.TemporalAccessor;
import java.util.Arrays;
import java.util.Locale;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;

/**
 * Formatter for primitive values. Logically immutable, but may not be thread-safe due to
 * implementation details of underlying formatting libraries.
 */
@NotThreadSafe
public class PrimitiveValueFormatter {
  /**
   * Quoth the Javadoc: "this is the rounding mode commonly taught at school." Any other rounding
   * mode might baffle the observant user. The statistical bias of the rounding mode is irrelevant
   * for presentation purposes anyway.
   */
  public static final RoundingMode ROUNDING_MODE = RoundingMode.HALF_UP;

  private final Locale locale;
  private final OptionalInteger fractionalDigits;
  private final boolean groupDigits;
  private final TemporalPrecision temporalPrecision;
  private final TemporalStyle temporalStyle;
  private final ZoneId timeZoneId;

  private PrimitiveValueFormatter(Locale locale, OptionalInteger fractionalDigits,
      boolean groupDigits, TemporalPrecision temporalPrecision, TemporalStyle temporalStyle,
      ZoneId timeZoneId)
  {
    Preconditions.checkNotNull(locale);
    Preconditions.checkNotNull(fractionalDigits);
    Preconditions.checkNotNull(temporalPrecision);
    Preconditions.checkNotNull(temporalStyle);
    Preconditions.checkNotNull(timeZoneId);
    this.locale = locale;
    this.fractionalDigits = fractionalDigits;
    this.groupDigits = groupDigits;
    this.temporalPrecision = temporalPrecision;
    this.temporalStyle = temporalStyle;
    this.timeZoneId = timeZoneId;
  }

  /* TODO[cc]: Initialize formatter objects in the constructor, and have each call site construct
               only one PrimitiveValueFormatter per path. */
  public static PrimitiveValueFormatter fromViewPropertySource(ViewPropertySource vps) {
    final String localeLanguageTag = vps.get(LBProp.DISPLAY_LOCALE);
    Locale locale = localeLanguageTag.isEmpty() ? null : Locale.forLanguageTag(localeLanguageTag);
    // TODO[pd]: Avoid duplication between here and LayoutContext.
    if (locale == null)
      locale = Locale.getDefault(Locale.Category.FORMAT);
    final String timeZoneIdString = vps.get(LBProp.DISPLAY_TIME_ZONE);
    ZoneId timeZoneId = null;
    if (!timeZoneIdString.isEmpty()) {
      try {
        timeZoneId = ZoneId.of(timeZoneIdString);
      } catch (DateTimeException e) {
        // No action.
      }
    }
    if (timeZoneId == null)
      timeZoneId = ZoneId.systemDefault();
    return new PrimitiveValueFormatter(
        locale,
        vps.get(LMProp.DECIMAL_PLACES),
        vps.get(LMProp.GROUP_DIGITS),
        vps.get(LMProp.TEMPORAL_PRECISION),
        vps.get(LBProp.TEMPORAL_STYLE),
        timeZoneId);
  }

  private static @Nullable BigDecimal normalizeDouble(double value) {
    return !Double.isFinite(value) ? null : new BigDecimal(value);
  }

  private static BigDecimal normalizeLong(long value) {
    return new BigDecimal(value);
  }

  // TODO[cc]: Consider combining this logic with that of ConstantExpression.
  public static @Nullable BigDecimal normalizeNumber(@Nonnull Object value) {
    Preconditions.checkNotNull(value);
    if (value instanceof BigDecimal) {
      return (BigDecimal) value;
    } else if (value instanceof Double) {
      return normalizeDouble((Double) value);
    } else if (value instanceof Float) {
      return normalizeDouble((Float) value);
    } else if (value instanceof Long) {
      return normalizeLong((Long) value);
    } else if (value instanceof Integer) {
      return normalizeLong((Integer) value);
    } else if (value instanceof Short) {
      return normalizeLong((Short) value);
    } else if (value instanceof Byte) {
      return normalizeLong((Byte) value);
    } else {
      return null;
    }
  }

  public @Nonnull String getFormattedValue(PrimitiveValue value) {
    Preconditions.checkNotNull(fractionalDigits);
    Preconditions.checkNotNull(temporalStyle);
    Preconditions.checkNotNull(locale);
    Preconditions.checkNotNull(timeZoneId);
    Preconditions.checkNotNull(value);
    final Object content = value.getContent();
    if (content == null)
      return "";
    if (fractionalDigits.isPresent()) {
      final BigDecimal bd = normalizeNumber(content);
      if (bd != null) {
        // TODO[cc]: Store the actual DecimalFormat in a ThreadLocal.
        NumberFormat nf = NumberFormat.getNumberInstance(locale);
        /* The format methods in NumberFormat do not handle BigDecimal correctly, but the
        specific implementations in DecimalFormat do, so only use the formatter if it is in fact an
        instance of DecimalFormat. */
        if (nf instanceof DecimalFormat) {
          nf.setMinimumFractionDigits(fractionalDigits.get());
          nf.setMaximumFractionDigits(fractionalDigits.get());
          nf.setGroupingUsed(groupDigits);
          // DecimalFormat defaults to RoundingMode.HALF_EVEN (not what we want).
          // Casts are here to make sure the right Javadoc shows up in the IDE.
          ((DecimalFormat) nf).setRoundingMode(ROUNDING_MODE);
          AttributedCharacterIterator aci = ((DecimalFormat) nf).formatToCharacterIterator(bd);
          StringBuilder sb = new StringBuilder();
          for (int i = aci.getBeginIndex(); i < aci.getEndIndex(); i++)
            sb.append(aci.setIndex(i));
          return sb.toString();
        } else {
          return bd.setScale(fractionalDigits.get(), ROUNDING_MODE).toPlainString();
        }
      }
    }
    if (content instanceof LocalTime) {
      return formatTemporal(TemporalFormatterProvider.TYPE_LOCALTIME, (LocalTime) content);
    } else if (content instanceof LocalDate) {
      return formatTemporal(TemporalFormatterProvider.TYPE_LOCALDATE, (LocalDate) content);
    } else if (content instanceof Instant) {
      /* Always show timestamps without time zone information, since including it might erroneously
      suggest that time zone information is stored in the database, and since every value in a
      column would just show the same time zone. */
      return formatTemporal(TemporalFormatterProvider.TYPE_LOCALDATETIME,
          ZonedDateTime.ofInstant((Instant) content, timeZoneId).toLocalDateTime());
    }
    {
      String ret = formatIfArray(content);
      if (ret != null)
        return ret;
    }
    /* The trim is important for LayoutMeasurer's use case, which involves counting characters in
    the formatted string. */
    return content.toString().trim();
  }

  public static @Nullable String formatIfArray(Object obj) {
    Preconditions.checkNotNull(obj);
    if (obj instanceof double[]) {
      return Arrays.toString((double[]) obj);
    } else if (obj instanceof float[]) {
      return Arrays.toString((float[]) obj);
    } else if (obj instanceof boolean[]) {
      return Arrays.toString((boolean[]) obj);
    } else if (obj instanceof long[]) {
      return Arrays.toString((long[]) obj);
    } else if (obj instanceof int[]) {
      return Arrays.toString((int[]) obj);
    } else if (obj instanceof short[]) {
      return Arrays.toString((short[]) obj);
    } else if (obj instanceof byte[]) {
      return formatByteArray((byte[]) obj);
    } else if (obj instanceof char[]) {
      return new String((char[]) obj).trim();
    } else if (obj instanceof Object[]) {
      // TODO[pd]: Consider recursing here.
      return Arrays.toString((Object[]) obj);
    } else {
      return null;
    }
  }

  private static final char[] HEX_DIGITS = "0123456789ABCDEF".toCharArray();

  private static String formatByteArray(byte[] arr) {
    final char ret[] = new char[arr.length * 3 - 1];
    int j = 0;
    for (int i = 0; i < arr.length; i++) {
      // Careful with signed bytes.
      final int b = arr[i] & 0xff;
      ret[j++] = HEX_DIGITS[b >>> 4];
      ret[j++] = HEX_DIGITS[b & 0x0f];
      if (i != arr.length - 1)
        ret[j++] = ' ';
    }
    return new String(ret);
  }

  private String formatTemporal(TemporalFormatterProvider type, TemporalAccessor accessor) {
    return temporalStyle.getFormatter(type, locale, temporalPrecision)
        .format(new FilteredTemporalAccessor(accessor, temporalPrecision));
  }
}
