// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.google.common.base.Preconditions;
import java.time.DateTimeException;
import java.time.LocalTime;
import java.time.ZonedDateTime;
import java.time.chrono.IsoChronology;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.format.FormatStyle;
import java.time.format.ResolverStyle;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalAccessor;
import java.util.Locale;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

@Immutable
abstract class TemporalFormatterProvider {
  public static final TemporalFormatterProvider TYPE_LOCALDATE = new LocalDateFormatterProvider();
  public static final TemporalFormatterProvider TYPE_LOCALTIME = new LocalTimeFormatterProvider();
  public static final TemporalFormatterProvider TYPE_ZONEDDATETIME =
      new DateTimeFormatterProvider(true);
  public static final TemporalFormatterProvider TYPE_LOCALDATETIME =
      new DateTimeFormatterProvider(false);

  private final @Nullable TemporalAccessor fallbackTestExample;

  /**
   * @param fallbackTestExample an example temporal value to be used while detecting an appropriate
   *        localized format; null if such detection is not necessary
  */
  TemporalFormatterProvider(@Nullable TemporalAccessor fallbackTestExample) {
    this.fallbackTestExample = fallbackTestExample;
  }

  abstract DateTimeFormatter getRawISOFormatter(TemporalPrecision precision);
  abstract DateTimeFormatter getRawLocalizedFormatter(FormatStyle dateStyle, FormatStyle timeStyle);

  public final DateTimeFormatter getISOFormatter(TemporalPrecision precision) {
    Preconditions.checkNotNull(precision);
    return getRawISOFormatter(precision)
        /* The locale should not matter for the ISO formats, but set an explicit constant locale to
        be completely sure that the format never changes. */
        .withLocale(Locale.US);
  }

  public final DateTimeFormatter getLocalizedFormatter(
      final FormatStyle formatStyle, final Locale locale, TemporalPrecision precision)
  {
    Preconditions.checkNotNull(precision);
    Preconditions.checkNotNull(locale);
    Preconditions.checkNotNull(formatStyle);
    if (fallbackTestExample == null) {
      return getRawLocalizedFormatter(formatStyle, formatStyle).withLocale(locale);
    } else {
      final FilteredTemporalAccessor filteredAccessor =
          new FilteredTemporalAccessor(fallbackTestExample, precision);
      final FormatStyle allFormatStyles[] = FormatStyle.values();
      /* Fall back using progressively shorter time formats (increasing FormatStyle indices) than
      the one originally specified. */
      for (int i = formatStyle.ordinal(); i < allFormatStyles.length; i++) {
        final FormatStyle useDateStyle = formatStyle;
        final FormatStyle useTimeStyle = allFormatStyles[i];
        final DateTimeFormatter ret =
            getRawLocalizedFormatter(useDateStyle, useTimeStyle).withLocale(locale);
        try {
          // Ignore return value; just check for exceptions.
          ret.format(filteredAccessor);
        } catch (DateTimeException e) {
          // Try the next format instead.
          continue;
        }
        return ret;
      }
      return getISOFormatter(precision);
    }
  }

  /* ============================================================================================ */

  private static final class LocalDateFormatterProvider extends TemporalFormatterProvider {
    public LocalDateFormatterProvider() {
      super(null);
    }

    @Override
    public DateTimeFormatter getRawISOFormatter(TemporalPrecision precision) {
      return DateTimeFormatter.ISO_LOCAL_DATE;
    }

    @Override
    DateTimeFormatter getRawLocalizedFormatter(FormatStyle dateStyle, FormatStyle timeStyle) {
      return DateTimeFormatter.ofLocalizedDate(dateStyle);
    }
  }

  /* ============================================================================================ */

  private static final class LocalTimeFormatterProvider extends TemporalFormatterProvider {
    public LocalTimeFormatterProvider() {
      /* Use fallback detection: Some of the long localized formats include a mandatory time zone
      field, which is not available on LocalTime values. */
      super(LocalTime.parse("10:15:14.1234"));
    }

    @Override
    public DateTimeFormatter getRawISOFormatter(TemporalPrecision precision) {
      /* Use the same settings and format as DateTimeFormatter.ISO_LOCAL_TIME, except with a
      specific forced setting for the number of decimals shown for fractions of a second. Rounding
      the NANO_OF_SECOND field is not enough, as trailing zeroes would be omitted. */
      DateTimeFormatterBuilder ret = new DateTimeFormatterBuilder()
          .appendValue(ChronoField.HOUR_OF_DAY, 2)
          .appendLiteral(':')
          .appendValue(ChronoField.MINUTE_OF_HOUR, 2)
          .optionalStart()
          .appendLiteral(':')
          .appendValue(ChronoField.SECOND_OF_MINUTE, 2);
      final int fractionDigits = precision.getSecondFractionDigits();
      if (fractionDigits > 0) {
        ret = ret
          .optionalStart()
          .appendFraction(ChronoField.NANO_OF_SECOND, fractionDigits, fractionDigits, true);
      }
      return ret.toFormatter()
          .withResolverStyle(ResolverStyle.STRICT);
    }

    @Override
    DateTimeFormatter getRawLocalizedFormatter(FormatStyle dateStyle, FormatStyle timeStyle) {
      return DateTimeFormatter.ofLocalizedTime(timeStyle);
    }
  }

  /* ============================================================================================ */

  private static final class DateTimeFormatterProvider extends TemporalFormatterProvider {
    public DateTimeFormatterProvider(boolean zoned) {
      /* Use fallback detection: Some of the long localized formats include a mandatory
      second-of-minute field, which may not be available due to the selected precision. The same
      goes for mandatory time zone fields in the !zoned case. */
      super(getFallbackTestExample(zoned));
    }

    private static TemporalAccessor getFallbackTestExample(boolean zoned) {
      ZonedDateTime zonedDateTime = ZonedDateTime.parse("2007-12-03T10:15:30.1234+00:00[UTC]");
      return zoned ? zonedDateTime : zonedDateTime.toLocalDateTime();
    }

    @Override
    public DateTimeFormatter getRawISOFormatter(TemporalPrecision precision) {
      if (precision == TemporalPrecision.DATE)
        return TYPE_LOCALDATE.getISOFormatter(precision);
      /* Use the same settings and format as DateTimeFormatter.ISO_LOCAL_DATE_TIME, except with a
      space instead of the ugly 'T' character between the date and the time. This can still be
      considered standards-conformant:

        "By mutual agreement of the partners in information interchange, the character [T] may be
        omitted in applications where there is no risk of confusing a date and time of day
        representation with others defined in this International Standard."
        (ISO 8601, section 4.3.2)

      Similar to com.sieuferd.datamodel.expressions.TemporalFormats.ISO_LOCAL_DATE_TIME_WITH_SPACE,
      but using the customized version of ISO_LOCAL_TIME with a forced number of fractional digits.
      */
      return new DateTimeFormatterBuilder()
          .parseCaseInsensitive()
          .append(TYPE_LOCALDATE.getISOFormatter(precision))
          .appendLiteral(' ') // Use a space instead of a 'T' here.
          .append(TYPE_LOCALTIME.getISOFormatter(precision))
          .toFormatter()
          .withResolverStyle(ResolverStyle.STRICT)
          .withChronology(IsoChronology.INSTANCE);
    }

    @Override
    DateTimeFormatter getRawLocalizedFormatter(FormatStyle dateStyle, FormatStyle timeStyle) {
      return DateTimeFormatter.ofLocalizedDateTime(dateStyle, timeStyle);
    }
  }
}
