// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout;

import com.google.common.base.Preconditions;
import com.sieuferd.cell.support.Canvas;
import com.sieuferd.cell.styles.LineStyle;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.Points;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import com.sieuferd.cell.styles.PathHighlights;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.font.FontRenderContext;
import java.awt.geom.Rectangle2D;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.util.Map;
import javax.print.attribute.PrintRequestAttributeSet;
import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.support.Cell.FocusMode;
import com.sieuferd.cell.support.Selection;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.datamodel.Views;
import com.sieuferd.uiutil.UIUtil;

public class ViewResultPrinter implements Printable {
  private final ViewResult result;
  private final PrintRequestAttributeSet printAttributes;
  private final double scaling;
  private final boolean fitWidth;
  private Cell layout;
  private RenderContext renderContext;
  private ImmutableList<Pixels> pagination;
  private Coordinates pageDimensions;
  private double adjustedScaling;

  public ViewResultPrinter(ViewResult result, PrintRequestAttributeSet printAttributes,
      double scaling, boolean fitWidth)
  {
    Preconditions.checkNotNull(result);
    Preconditions.checkNotNull(printAttributes);
    Preconditions.checkArgument(scaling > 0.0);
    this.result = result;
    // TODO[cc]: See if this one can be defensively copied.
    this.printAttributes = printAttributes;
    this.scaling = scaling;
    this.fitWidth = fitWidth;
  }

  // Note: This function has state-changing side-effects.
  private void createLayout(FontRenderContext frc, PageFormat pageFormat) {
    renderContext = new RenderContext(frc, adjustedScaling, null, false, false);
    pageDimensions = new Coordinates(
        Pixels.create(renderContext, pageFormat.getImageableWidth()),
        Pixels.create(renderContext, pageFormat.getImageableHeight()));
    {
      MutableView view = Views.newMutable(result.getView());
      try {
        LayoutMeasurer.measureData(result.getData(), view, renderContext);
        LayoutGenerator.autoStyle(result.getData(), view, renderContext, pageDimensions.getX(),
            LabelFormatter.SIMPLE);
        layout = LayoutGenerator.createLayout(
            result.getData(), ImmutableView.copyOf(view), renderContext, pageDimensions.getX(),
            LabelFormatter.SIMPLE);
      } catch (InterruptedException e) {
        // TODO[cc]: Use the data flow API in this code, then get rid of this unexpected case.
        throw new RuntimeException("Unexpected thread interruption", e);
      }
    }
    pagination = layout.paginate(renderContext, pageDimensions.getY());
  }

  @Override
  public int print(Graphics graphics, PageFormat pageFormat, int pageIndex)
      throws PrinterException
  {
    /*
    System.out.println("Rendering using pageFormat "
        + pageFormat.getWidth() / 72.0 + ", " + pageFormat.getHeight() / 72.0 + " / "
        + pageFormat.getImageableWidth() / 72.0 + ", " + pageFormat.getImageableHeight() / 72.0);
    */
    Graphics2D g = (Graphics2D) graphics;

    Map<Object,Object> renderingHints = Maps.newLinkedHashMap();
    renderingHints.put(RenderingHints.KEY_FRACTIONALMETRICS,
        RenderingHints.VALUE_FRACTIONALMETRICS_ON);
    renderingHints.put(RenderingHints.KEY_STROKE_CONTROL,
        RenderingHints.VALUE_STROKE_PURE);
    g.addRenderingHints(renderingHints);

    g.translate(pageFormat.getImageableX(), pageFormat.getImageableY());

    if (layout == null) {
      adjustedScaling = scaling;
      createLayout(g.getFontRenderContext(), pageFormat);
      if (fitWidth) {
        adjustedScaling = scaling * pageFormat.getImageableWidth() /
            (layout.getDimensionsWithOverprint(renderContext).getX().getDouble() * 1.05);
        createLayout(g.getFontRenderContext(), pageFormat);
      }
    }

    if (!renderContext.getFontRenderContext().equals(g.getFontRenderContext())) {
      System.out.println("---------------------");
      System.out.println("FRC changed:");
      System.out.println("Original FRC: " +
          UIUtil.frcToString(renderContext.getFontRenderContext()));
      System.out.println("Current  FRC: " +
          UIUtil.frcToString(g.getFontRenderContext()));
      System.out.println("---------------------");
    }

    int ret;
    if (pageIndex < pagination.size()) {
      ret = Printable.PAGE_EXISTS;
      Pixels pageStart = (pageIndex == 0) ? Pixels.ZERO : pagination.get(pageIndex - 1);
      Pixels pageEnd   = pagination.get(pageIndex);
      Coordinates offset = new Coordinates(Pixels.ZERO, Pixels.ZERO.minus(pageStart));
      Graphics2D clippedGraphics = (Graphics2D) g.create();
      try {
        // TODO[pd]: Should I rather use clipRect here?
        clippedGraphics.setClip(new Rectangle2D.Double(0, 0,
            pageDimensions.getX().getDouble(), pageEnd.minus(pageStart).getDouble()));
        if (Cell.DEBUG_PAGINATION) {
          /* Draw a very thin red line at the page break mark that can be examined in Acrobat
          Reader. */
          final LineStyle PAGE_BREAKS = new LineStyle(new Color(255, 0, 0, 255), new Points(0.1));
          PAGE_BREAKS.paint(new Canvas(g, renderContext), false,
              new Coordinates(Pixels.ZERO, pageEnd.minus(pageStart)), pageDimensions.getX());
        }
        /* Since offset is negative, need to set the frozen header point to offset rather than zero
        to avoid frozen headers. */
        layout.paint(new Canvas(clippedGraphics, renderContext), offset, Selection.NONE,
            PathHighlights.NONE, null, FocusMode.UNFOCUSED, false);
      } finally {
        clippedGraphics.dispose();
      }
    } else {
      ret = Printable.NO_SUCH_PAGE;
    }
    return ret;
  }
}
