// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout.internal;

import com.sieuferd.cell.support.Location;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.Points;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.styles.TextStyle;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import java.util.Collection;
import java.util.List;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.TreeStructure;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.layout.internal.LayoutContext.PrimitiveLayoutContext;
import com.sieuferd.layout.internal.LayoutContext.RelationLayoutContext;
import com.sieuferd.cell.builders.BorderCellBuilder;
import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.builders.EmptyCellBuilder;
import com.sieuferd.cell.builders.HeaderedCellBuilder;
import com.sieuferd.cell.builders.ListCellBuilder;
import com.sieuferd.cell.builders.TextCellBuilder;
import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.cell.styles.LineBreakPolicy;
import com.sieuferd.cell.styles.Text;
import com.sieuferd.cell.styles.Text.SimpleText;
import com.sieuferd.cell.styles.TextFitPolicy;
import com.sieuferd.cell.styles.TextMeasurer;
import com.sieuferd.cell.styles.TextMeasurer.DefaultTextMeasurer;
import com.sieuferd.cell.styles.TextMeasurer.MemoizedTextMeasurer;
import com.sieuferd.cell.support.Axis;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.TupleInterval.PluralTupleInterval;
import com.sieuferd.datamodel.TupleInterval.SingletonTupleInterval;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewPropertySource;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.TupleConstant;
import com.sieuferd.datamodel.proptypes.OptionalBigDecimal;
import com.sieuferd.layout.YesNoAuto;
import com.sieuferd.layout.LBProp;
import com.sieuferd.layout.LMProp;
import com.sieuferd.layout.LabelFormatter;
import com.sieuferd.layout.PrimitiveValueFormatter;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.ObjectReference;
import java.awt.Color;
import java.awt.Font;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

// TODO[cc]: Get rid of the old "Traversal" terminology.
public abstract class LayoutContext
    extends TreeStructure<LayoutContext,RelationLayoutContext,PrimitiveLayoutContext>
{
  public abstract LayoutContext speculativeTraversal();
  public abstract LayoutContext placeholderTraversal();

  private static final PrimitiveValue PLACEHOLDER_VALUE = new PrimitiveValue("");

  static final class Globals {
    final RenderContext renderContext;
    final MemoizedTextMeasurer textMeasurer;
    final View view;
    final LabelFormatter labelFormatter;

    Globals(RenderContext renderContext, MemoizedTextMeasurer textMeasurer, View view,
        LabelFormatter labelFormatter)
    {
      Preconditions.checkNotNull(renderContext);
      Preconditions.checkNotNull(textMeasurer);
      Preconditions.checkNotNull(view);
      Preconditions.checkNotNull(labelFormatter);
      this.renderContext = renderContext;
      this.textMeasurer = textMeasurer;
      this.view = view;
      this.labelFormatter = labelFormatter;
    }
  }

  private final Globals globals;
  private final @Nullable RelationLayoutContext parent;
  private final Location location;
  private final boolean alternativeValueColor;

  @Override
  public String toString() {
    return getPath().toString(getView()) +
        (getTupleConstraints().isEmpty() ? "" : (" TC(" + getTupleConstraints() + ")"));
  }

  public boolean isValue() {
    return location.isValue();
  }

  public Path getPath() {
    return location.getPath();
  }

  @Override
  protected final Class<RelationLayoutContext> getRelationClass() {
    return RelationLayoutContext.class;
  }

  @Override
  protected final Class<PrimitiveLayoutContext> getPrimitiveClass() {
    return PrimitiveLayoutContext.class;
  }

  public RenderContext getRenderContext() {
    return globals.renderContext;
  }

  public LabelFormatter getLabelFormatter() {
    return globals.labelFormatter;
  }

  private LayoutContext(Globals globals, @Nullable RelationLayoutContext parent,
      Location location, boolean alternativeValueColor)
  {
    Preconditions.checkNotNull(globals);
    Preconditions.checkNotNull(location);
    this.globals = globals;
    this.parent = parent;
    this.location = location;
    this.alternativeValueColor = alternativeValueColor;
  }

  @Nullable RelationPath getIdentifyingPath() {
    return location.getIdentifyingPath();
  }

  Tuple getIdentifyingTuple() {
    return location.getIdentifyingTuple();
  }

  List<TupleConstant> getTupleConstraints() {
    return location.getTupleConstraints();
  }

  MemoizedTextMeasurer getTextMeasurer() {
    return globals.textMeasurer;
  }

  Globals getGlobals() {
    return globals;
  }

  Location getLocation() {
    return location;
  }

  final Data getData() {
    return location.getData();
  }

  public static LayoutContext startTraversal(
      Data data, RenderContext renderContext, View view, boolean value,
      LabelFormatter labelFormatter)
  {
    final Globals globals = new Globals(renderContext,
        new MemoizedTextMeasurer(new DefaultTextMeasurer(renderContext)), view, labelFormatter);
    final Location location = new Location(
        data, Path.ROOT, null, Tuple.createRootParent(data), value, ImmutableList.of());
    return new RelationLayoutContext(globals, null, location, false);
  }

  public @Nullable RelationLayoutContext getParent() {
    return parent;
  }

  public boolean hasSiblings() {
    return parent != null && parent.traverseFields().size() > 1;
  }

  public Pixels getSSPixels(ViewProperty<Points> property) {
    return Pixels.create(getRenderContext(), getP(property));
  }

  public View getView() {
    return globals.view;
  }

  ViewPropertySource getViewPropertySource() {
    return getView().getViewPropertySource(getPath());
  }

  public boolean isAlternativeValueColor() {
    return alternativeValueColor;
  }

  public Coordinates measureLabel(Pixels prefMaxWidth, boolean vertical, boolean mayRightAlign) {
    /* Always include badges when measuring labels, otherwise ugly line breaks may result. While
    this policy may cause layout decisions to change as a result of changing badges, in practice the
    changes are localized and less disruptive than the vertical shifts that may result from line
    breaks in labels. */
    return measureText(getLabelFontProperty(),
        getStyledLabel(mayRightAlign && !vertical), prefMaxWidth, true, vertical);
  }

  public ViewProperty<Font> getLabelFontProperty() {
    return isRelation() ? LBProp.RELATION_LABEL_FONT : LBProp.PRIMITIVE_LABEL_FONT;
  }

  /**
   * Measure the dimensions of a string using a memoized text measurer. For value strings, use
   * {@link PrimitiveValue#measureValue(Object)} instead.
   */
  public Coordinates measureText(ViewProperty<Font> fontProp, Text text, Pixels prefMaxWidth,
      boolean allowWider, boolean vertical)
  {
    TextStyle textStyle = new TextStyle(getP(fontProp), getLineBreakPolicy(false));
    return getTextMeasurer()
        .measure(textStyle, text, prefMaxWidth, allowWider, vertical).getDimensions();
  }

  LineBreakPolicy getLineBreakPolicy(boolean value) {
    if (!value || isRelation()) {
      return LineBreakPolicy.DEFAULT_PARAGRAPH;
    } else {
      /* TODO[pd]: Get the BreakIterator from a cached PrimitiveValueFormatter. This will get rid of
                   the duplicated logic below, which is copied from the latter. */
      final String localeLanguageTag = getP(LBProp.DISPLAY_LOCALE);
      Locale locale = localeLanguageTag.isEmpty() ? null :
          Locale.forLanguageTag(localeLanguageTag);
      if (locale == null)
        locale = Locale.getDefault(Locale.Category.FORMAT);
      return LineBreakPolicy.createParagraphPolicy(locale);
    }
  }

  public Text getStyledLabel(boolean mayRightAlign) {
    final boolean rightAligned =
        mayRightAlign && getEffectiveHorizontalAlignment(false) == HorizontalAlignment.RIGHT;
    return getLabelFormatter().getStyledLabel(getView(), getPath(), rightAligned);
  }

  public <T> T getP(ViewProperty<T> property) {
    return getView().get(property, getPath());
  }

  public <T> void setP(ViewProperty<T> property, T value) {
    if (isValue())
      throw new IllegalStateException("Attempted to set property " + property + " in value pass");
    final View viewToMutate = getView();
    if (!(viewToMutate instanceof MutableView))
      throw new IllegalStateException("Cannot update read-only view");
    ((MutableView) viewToMutate).set(property, getPath(), value);
  }

  /* TODO[cc]: Clean up the pendingAuto() system once I'm done distinguishing input layout
               properties from effective layout properties. */
  public boolean pendingAuto(ViewProperty<Boolean> effective, ViewProperty<YesNoAuto> input) {
    boolean ret = !getView().defines(effective, getPath());
    if (ret && isValue())
      throw new IllegalStateException("Property " + effective + " not set in placeholder pass");
    if (ret) {
      final Boolean nullableBoolean = getView().get(input, getPath()).getValue();
      if (nullableBoolean != null) {
        setP(effective, nullableBoolean);
        return false;
      }
    }
    return ret;
  }

  public boolean pendingAuto(ViewProperty<?> property) {
    boolean ret = !getView().defines(property, getPath());
    if (ret && isValue())
      throw new IllegalStateException("Property " + property + " not set in placeholder pass");
    return ret;
  }

  public Cell newBigger(Cell childCell, Coordinates minimumDimensions) {
    return verticalListCellBuilder()
        .addCell(childCell)
        .minimumDimensions(minimumDimensions).build();
  }

  public ListCellBuilder verticalListCellBuilder() {
    return new ListCellBuilder(getRenderContext(), getViewPropertySource(), true);
  }

  public ListCellBuilder horizontalListCellBuilder() {
    return new ListCellBuilder(getRenderContext(), getViewPropertySource(), false);
  }

  /**
   * Get a {@code TextCellBuilder} using a memoized text measurer. For value strings, use
   * {@link PrimitiveLayoutContext#buildValueCell(Pixels,TextFitPolicy) instead}
   */
  public TextCellBuilder textCellBuilder(ViewProperty<Color> colorProperty,
      ViewProperty<Font> fontProperty, Pixels width, boolean forceLeftAligned)
  {
    return new TextCellBuilder(getRenderContext(), getTextMeasurer(), getViewPropertySource(),
        getP(fontProperty), getP(colorProperty), getEffectiveHorizontalAlignment(forceLeftAligned),
        width)
        .lineBreakPolicy(getLineBreakPolicy(false));
  }

  HorizontalAlignment getEffectiveHorizontalAlignment(boolean forceLeft) {
    if (forceLeft)
      return HorizontalAlignment.LEFT;
    /* I considered doing something special to make sure right-aligned numbers align even for the
    rare case of fonts that define different advance widths for each various digits (e.g. the font
    "Birch Std"). On further investigation, however, neither (Windows) Excel 2013 or Excel for
    Mac 2011 does this, so I won't bother either. */
    return !isRelation() && getP(LMProp.DECIMAL_PLACES).isPresent()
        ? HorizontalAlignment.RIGHT
        : getP(LBProp.HORIZONTAL_ALIGNMENT);
  }

  public EmptyCellBuilder emptyCellBuilder() {
    return new EmptyCellBuilder(getRenderContext(), getViewPropertySource());
  }

  /* TODO[cc]: Simplify the case where the border cell just contains an empty cell. Maybe get rid of
           EmptyCellBuilder altogether. */
  public BorderCellBuilder borderCellBuilder(Cell childCell) {
    return new BorderCellBuilder(getRenderContext(), getViewPropertySource(), childCell);
  }

  public boolean isProjected() {
    return getData().isProjected(getPath());
  }

  public BorderCellBuilder selectableBorderCellBuilder(Cell childCell) {
    return new BorderCellBuilder(getRenderContext(), getViewPropertySource(), childCell)
        .location(location);
  }

  /** Equivalent to {@link #selectableBorderCellBuilder(Cell)} except that the selectable
  CursorLocation is forced into value mode even if this LayoutContext is a placeholder traversal.
  This distinction is only relevant for selection in schema-only layout mode. */
  public BorderCellBuilder valueSelectableBorderCellBuilder(Cell childCell) {
    final Location useLocation = isValue() ? location : new Location(
        getData(), getPath(), getIdentifyingPath(), getIdentifyingTuple(), true,
        getTupleConstraints());
    return new BorderCellBuilder(getRenderContext(), getViewPropertySource(), childCell)
        .location(useLocation);
  }

  public HeaderedCellBuilder headeredCellBuilder(
      boolean vertical, Cell headerCell, Cell contentCell)
  {
    return new HeaderedCellBuilder(
        getRenderContext(), getViewPropertySource(), vertical, headerCell, contentCell);
  }

  /* ******************************************************************************************** */
  public static class RelationLayoutContext extends LayoutContext {
    private RelationLayoutContext(Globals globals, @Nullable RelationLayoutContext parent,
        Location location, boolean alternativeValueColor)
    {
      super(globals, parent, location, alternativeValueColor);
    }

    public boolean isLayoutPlural() {
      return getView().get(CProp.PLURAL, getPath());
    }

    public RelationLayoutContext withAdditionalTupleConstraint(TupleConstant constraint) {
      List<TupleConstant> newTupleConstraints = ImmutableList.<TupleConstant>builder()
          .addAll(getTupleConstraints())
          .add(constraint)
          .build();
      Location newCursorLocation = new Location(getData(), getPath(),
          getIdentifyingPath(), getIdentifyingTuple(), isValue(), newTupleConstraints);
      return new RelationLayoutContext(
          getGlobals(), this, newCursorLocation, isAlternativeValueColor());
    }

    private @Nullable TupleConstant createCrosstabTupleConstant(Tuple tuple) {
      Preconditions.checkNotNull(tuple);
      Preconditions.checkState(isValue());
      final Map<PrimitivePath,ConstantExpression> ret = Maps.newLinkedHashMap();
      for (PrimitivePath constraintPath : getP(LMProp.CROSSTAB_COLUMNS).getPaths()) {
        if (!getPath().isParentOf(constraintPath) || !getData().isProjected(constraintPath))
          return null;
        PrimitiveValue value = getData().getTupleField(tuple, constraintPath);
        ConstantExpression constantExpression = ConstantExpression.fromObject(value.getContent());
        if (constantExpression == null)
          return null;
        MapUtil.putChecked(ret, constraintPath, constantExpression);
      }
      return new TupleConstant(ret);
    }

    /**
     * Get an ordered map of child tuples matched to crosstab columns configured for the current
     * relation.
     *
     * @return may contain null values to indicate child tuples not occurring, or occurring multiple
     *         times, in the current (parent) tuple
     */
    public Map<TupleConstant,Tuple> getCrosstabColumnTuples() {
      final Map<TupleConstant,Tuple> mappedTuples = Maps.newLinkedHashMap();
      final Set<TupleConstant> nonUniqueTuples = Sets.newLinkedHashSet();
      if (isValue()) {
        for (Tuple tuple : getTuples(true)) {
          final TupleConstant tupleConstant = createCrosstabTupleConstant(tuple);
          if (tupleConstant != null) {
            if (mappedTuples.put(tupleConstant, tuple) != null) {
              // At most one match is permitted.
              nonUniqueTuples.add(tupleConstant);
            }
          }
        }
      }
      final Map<TupleConstant,Tuple> ret = Maps.newLinkedHashMap();
      for (TupleConstant constant : getP(LMProp.CROSSTAB_COLUMNS).getTuples()) {
        MapUtil.putChecked(ret, constant,
            nonUniqueTuples.contains(constant) ? null : mappedTuples.get(constant));
      }
      return Collections.unmodifiableMap(ret);
    }

    public Collection<LayoutContext> traverseFields(Tuple tuple, boolean alternativeValueColor) {
      List<LayoutContext> ret = Lists.newArrayList();
      /* Plural paths in value traversals are identified by tuples at those paths. Non-plural paths,
      and paths traversed in schema-only mode, are identified by tuples at the same path as their
      parent. The correctness of this logic is verified by the call to Tuple.checkPath() in the
      LayoutContext constructor. */
      /* TODO[pd]: Instead of doing isProjected() here, I might want to do a schema-only traversal
                   into non-projected children. But I need to make sure this doesn't trigger padding
                   of primitive fields. Or actually, maybe that would be useful? */
      final RelationPath childIdentifyingPath =
          (isValue() && isProjected()) ? getPath() : getIdentifyingPath();
      for (Path childPath : getView().getChildPaths(getPath())) {
        if (Boolean.FALSE.equals(getView().getOrNull(CProp.VISIBLE, childPath)))
          continue;
        final Location childLocation =
            new Location(getData(), childPath, childIdentifyingPath, tuple, isValue(),
            getTupleConstraints());
        final LayoutContext childContext = childPath.isRelation()
            ? new RelationLayoutContext(getGlobals(), this, childLocation, alternativeValueColor)
            : new PrimitiveLayoutContext(getGlobals(), this, childLocation, alternativeValueColor);
        ret.add(childContext);
      }
      return ret;
    }

    public Collection<LayoutContext> traverseFields(Tuple tuple) {
      return traverseFields(tuple, isAlternativeValueColor());
    }

    public Collection<LayoutContext> traverseFields() {
      if (isValue()) {
        return placeholderTraversal().traverseFields();
      } else {
        return traverseFields(getIdentifyingTuple(), isAlternativeValueColor());
      }
    }

    /** @return true iff the specified path is or contains a visible plural relation */
    private boolean hasVisiblePluralDescendant(RelationPath path) {
      if (!getView().get(CProp.VISIBLE, path))
        return false;
      if (getView().get(CProp.PLURAL, path))
        return true;
      for (Path childPath : getView().getChildPaths(path)) {
        if (childPath.isRelation() && hasVisiblePluralDescendant(childPath.getRelation()))
          return true;
      }
      return false;
    }

    private boolean hasParallelVisiblePlurals() {
      boolean seenPlural = false;
      for (Path childPath : getView().getChildPaths(getPath())) {
        if (childPath.isRelation() && hasVisiblePluralDescendant(childPath.getRelation())) {
          if (seenPlural)
            return true;
          seenPlural = true;
        }
      }
      return false;
    }

    public TupleInterval getTuples(boolean truncateIncompleteTuples) {
      if (isValue()) {
        if (isProjected()) {
          final TupleInterval ret = getData().getRelationTuples(getPath(), getIdentifyingTuple());
          /* Avoid calling isLayoutPlural if truncateIncompleteTuples is false, to avoid having the
          layout measurer depend on CProp.PLURAL. */
          if (!truncateIncompleteTuples)
            return ret;
          /* If a stale result is used, the data may contain plural relations at paths that we are
          now rendering as non-plural. */
          final boolean plural = isLayoutPlural() || ret.size() > 1;
          /* Only truncate results at plural relations, to be visually consistent with the idea that
          plural relations correspond to an array of objects while a non-plural relation corresponds
          to just a single object. For non-plural relations, truncation can happen in plural child
          relations instead. */
          if (!plural)
            return ret;
          /* For the special case where there is only one child relation that is visibly plural,
          always allow one incomplete tuple to be displayed at the end, since it is unlikely to
          cause an unbalanced layout. This is the common "single-branch grouping" case, which is
          often the only kind of grouping supported in other tools (e.g. pivot tables). */
          final int bonusTuple = hasParallelVisiblePlurals() ? 0 : 1;
          final int truncatedSize = Math.max(
              getData().getCompleteTupleCount(getPath()) - ret.getFirstTupleIndex() + bonusTuple,
              getView().get(LBProp.MAX_INCOMPLETE_TUPLES, Path.ROOT));
          if (truncatedSize >= ret.size()) {
            return ret;
          } else {
            return new PluralTupleInterval(
                getData(), getPath(), ret.getFirstTupleIndex(), truncatedSize, false);
          }
        } else {
          return new PluralTupleInterval(getData(), getPath(), 0, 0, false);
        }
      } else {
        return new SingletonTupleInterval(getIdentifyingTuple());
      }
    }

    @Override
    public RelationPath getPath() {
      return super.getPath().getRelation();
    }

    @Override
    public RelationLayoutContext placeholderTraversal() {
      Location newLocation = new Location(getData(), getPath(), getIdentifyingPath(),
          getIdentifyingTuple(), false, getTupleConstraints());
      return new RelationLayoutContext(getGlobals(),
          getParent() == null ? null : getParent().placeholderTraversal(),
          newLocation, isAlternativeValueColor());
    }

    @Override
    public RelationLayoutContext speculativeTraversal() {
      RelationLayoutContext speculativeParent = getParent() == null ? null :
          getParent().speculativeTraversal();
      Globals newGlobals = new Globals(getRenderContext(), getTextMeasurer(),
          Views.newMutable(getView()), getLabelFormatter());
      return new RelationLayoutContext(
          newGlobals, speculativeParent, getLocation(), isAlternativeValueColor());
    }
  }

  /* ******************************************************************************************** */
  public static class PrimitiveLayoutContext extends LayoutContext {
    private PrimitiveLayoutContext(Globals globals, @Nullable RelationLayoutContext parent,
        Location location, boolean alternativeValueColor)
    {
      super(globals, parent, location, alternativeValueColor);
    }

    @Override
    public PrimitivePath getPath() {
      return super.getPath().getPrimitive();
    }

    public boolean isInCrosstabHeader() {
      return getView().get(LMProp.CROSSTAB_COLUMNS, getPath().getParent())
          .getPaths().contains(getPath());
    }

    @Override
    public PrimitiveLayoutContext placeholderTraversal() {
      Location newLocation = new Location(getData(), getPath(), getIdentifyingPath(),
          getIdentifyingTuple(), false, getTupleConstraints());
      return new PrimitiveLayoutContext(getGlobals(),
          getParent() == null ? null : getParent().placeholderTraversal(),
          newLocation, isAlternativeValueColor());
    }

    @Override
    public PrimitiveLayoutContext speculativeTraversal() {
      RelationLayoutContext speculativeParent = getParent() == null ? null :
          getParent().speculativeTraversal();
      Globals newGlobals = new Globals(
          getRenderContext(), getTextMeasurer(), Views.newMutable(getView()),
          getLabelFormatter());
      return new PrimitiveLayoutContext(
          newGlobals, speculativeParent, getLocation(), isAlternativeValueColor());
    }

    public boolean isStrictlyLeftJustified() {
      return getEffectiveHorizontalAlignment(false) == HorizontalAlignment.LEFT &&
          !getEffectiveValueJustification();
    }

    public boolean getEffectiveValueJustification() {
      return getP(LMProp.DECIMAL_PLACES).isPresent()
          ? false
          : getP(LBProp.VALUE_JUSTIFIED);
    }

    public PrimitiveValue getValue() {
      return isValue() && isProjected()
          ? getData().getTupleField(getIdentifyingTuple(), getPath()) : PLACEHOLDER_VALUE;
    }

    public double getDoubleValue() {
      final @Nullable Object valueContent = getValue().getContent();
      if (!(valueContent instanceof Number))
        return Double.NaN;
      return ((Number) valueContent).doubleValue();
    }

    public PrimitiveValueFormatter getPrimitiveValueFormatter() {
      // TODO[pd]: Cache the PrimitiveValueFormatter in the globals.
      return PrimitiveValueFormatter.fromViewPropertySource(
          getView().getViewPropertySource(getPath()));
    }

    public String getFormattedValue() {
      return getPrimitiveValueFormatter().getFormattedValue(getValue());
    }

    public @Nonnull Axis getAxis(boolean forceIncludeZero) {
      final OptionalBigDecimal numericMin = getP(LMProp.NUMERIC_MIN);
      final OptionalBigDecimal numericMax = getP(LMProp.NUMERIC_MAX);
      final double notPresent = forceIncludeZero ? 0.0 : Double.NaN;
      double minD = numericMin.isPresent() ? numericMin.get().doubleValue() : notPresent;
      double maxD = numericMax.isPresent() ? numericMax.get().doubleValue() : notPresent;
      if (forceIncludeZero) {
        minD = Math.min(minD, 0.0);
        maxD = Math.max(maxD, 0.0);
      }
      return Axis.create(minD, maxD);
    }

    public Cell buildValueCell(Pixels width, TextFitPolicy textFitPolicy) {
      final String str = getFormattedValue();
      final boolean memoized = str.isEmpty() || textFitPolicy != TextFitPolicy.PARAGRAPH;
      TextMeasurer useMeasurer = memoized ? getTextMeasurer() : getTextMeasurer().getDelegate();
      return new TextCellBuilder(getRenderContext(), useMeasurer, getViewPropertySource(),
          getP(LBProp.VALUE_FONT), getP(LBProp.VALUE_TEXT_COLOR),
          getEffectiveHorizontalAlignment(false), width)
          .lineBreakPolicy(getLineBreakPolicy(true))
          .textFitPolicy(textFitPolicy)
          .content(str)
          .justified(getEffectiveValueJustification())
          .minimumDimensions(new Coordinates(width, Pixels.ZERO))
          .build();
    }

    /**
     * @param retFormattedString if not empty, will be set to the formatted string value
     */
    public Coordinates measureValue(Pixels prefMaxWidth, boolean allowWider, boolean vertical,
        @Nullable ObjectReference<String> retFormattedString)
    {
      String str = getFormattedValue();
      if (retFormattedString != null)
        retFormattedString.setValue(str);
      final boolean memoized = str.isEmpty();
      TextMeasurer useMeasurer = memoized ? getTextMeasurer() : getTextMeasurer().getDelegate();
      TextStyle textStyle = new TextStyle(getP(LBProp.VALUE_FONT), getLineBreakPolicy(true));
      return useMeasurer.measure(
          textStyle, new SimpleText(str), prefMaxWidth, allowWider, vertical).getDimensions();
    }
  }
}
