// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.layout.internal;

import com.google.common.base.Objects;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import java.util.List;
import java.util.Map;

public class PartitionAlgorithm {
  private PartitionAlgorithm() { }

  public static interface ScalarOperations<W,V> {
    public V getValue(W wrapper);
    public boolean greaterThan(V v1, V v2);
    public V plus(V v1, V v2);
    public V getZero();
  }

  private static class PartitionParameters {
    private final int startIndex;
    private final int noPartitions;

    PartitionParameters(int startIndex, int noPartitions) {
      this.startIndex   = startIndex;
      this.noPartitions = noPartitions;
    }

    public int getStartIndex() {
      return startIndex;
    }

    public int getNoPartitions() {
      return noPartitions;
    }

    @Override
    public int hashCode() {
      return Objects.hashCode(startIndex, noPartitions);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof PartitionParameters))
        return false;
      PartitionParameters rhs = (PartitionParameters) obj;
      return Objects.equal(startIndex  , rhs.startIndex)
          && Objects.equal(noPartitions, rhs.noPartitions);
    }
  }

  private static class PartitionResult<W,V> {
    private final V highestSum;
    private final ImmutableList<List<W>> partitions;

    PartitionResult(V highestSum, List<List<W>> partitions) {
      this.highestSum = highestSum;
      this.partitions = ImmutableList.copyOf(partitions);
    }

    public V getHighestSum() {
      return highestSum;
    }

    public List<List<W>> getPartitions() {
      return partitions;
    }
  }

  private static <W,V> PartitionResult<W,V> doPartitionMemoized(
      Map<PartitionParameters,PartitionResult<W,V>> pastResults, List<W> sequence,
      PartitionParameters parameters, ScalarOperations<W,V> accumulator)
  {
    PartitionResult<W,V> ret = pastResults.get(parameters);
    if (ret == null) {
      ret = doPartition(pastResults, sequence, parameters, accumulator);
      pastResults.put(parameters, ret);
    }
    return ret;
  }

  /* The current subsequence is passed as the original sequence plus a start index rather than as an
  actual sublist, since the start index needs to be part of the hashable PartitionParameters. */
  private static <W,V> PartitionResult<W,V> doPartition(
      Map<PartitionParameters,PartitionResult<W,V>> pastResults, List<W> sequence,
      PartitionParameters parameters, ScalarOperations<W,V> accumulator)
  {
    if (parameters.getNoPartitions() < 0 ||
        parameters.getStartIndex() > sequence.size() || parameters.getStartIndex() < 0 ||
        (parameters.getNoPartitions() == 0 && parameters.getStartIndex() < sequence.size()))
    {
      throw new IllegalArgumentException();
    }
    // TODO[cc]: See if this method can be simplified and the number of base cases reduced.
    // This base case is not used in the recursion, but is allowed in case it is used by the client.
    if (parameters.getNoPartitions() == 0)
      return new PartitionResult<>(accumulator.getZero(), Lists.newArrayList());
    V runningSum = accumulator.getZero();
    PartitionResult<W,V> bestResult = null;
    for (int i = parameters.getStartIndex(); i < sequence.size(); i++) {
      runningSum = accumulator.plus(runningSum, accumulator.getValue(sequence.get(i)));
      if (parameters.getNoPartitions() == 1)
        continue;
      PartitionParameters subproblem =
          new PartitionParameters(i + 1, parameters.getNoPartitions() - 1);
      PartitionResult<W,V> subproblemResult =
          doPartitionMemoized(pastResults, sequence, subproblem, accumulator);
      V highestSum = accumulator.greaterThan(runningSum, subproblemResult.getHighestSum()) ?
          runningSum : subproblemResult.getHighestSum();
      /* Using greater-than instead of greater-than-or-equals will make the earlier partitions
      larger in the case of a tie. */
      if (bestResult == null || !accumulator.greaterThan(highestSum, bestResult.getHighestSum())) {
        List<List<W>> newPartitions = Lists.newArrayList(subproblemResult.getPartitions());
        newPartitions.add(0, sequence.subList(parameters.getStartIndex(),
            subproblem.getStartIndex()));
        bestResult = new PartitionResult<>(highestSum, newPartitions);
      }
    }
    if (parameters.getNoPartitions() == 1) {
      // First recursion base case: Only one partition left to assign.
      List<List<W>> partitions = Lists.newArrayList();
      partitions.add(sequence.subList(parameters.getStartIndex(), sequence.size()));
      return new PartitionResult<>(runningSum, partitions);
    } else if (bestResult == null) {
      // Second recursion base case: Reached end of sequence, make remaining partitions empty.
      List<List<W>> partitions = Lists.newArrayList();
      for (int i = 0; i < parameters.getNoPartitions(); i++)
        partitions.add(Lists.newArrayList());
      return new PartitionResult<>(accumulator.getZero(), partitions);
    } else {
      return bestResult;
    }
  }

  public static <W,V> List<List<W>> doPartition(List<W> sequence, int noPartitions,
      ScalarOperations<W,V> accumulator)
  {
    PartitionParameters params = new PartitionParameters(0, noPartitions);
    Map<PartitionParameters,PartitionResult<W,V>> pastResults = Maps.newLinkedHashMap();
    PartitionResult<W,V> res = doPartitionMemoized(pastResults, sequence, params, accumulator);
    return res.getPartitions();
  }
}
