// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.query;

import com.sieuferd.datamodel.expressions.TupleConstant;
import com.sieuferd.datamodel.expressions.RelationConstant;
import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.expressions.Function;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.FunctionExpression;
import com.sieuferd.datamodel.expressions.Expressions;
import java.util.Objects;
import java.util.Set;
import javax.annotation.concurrent.Immutable;

@Immutable
public final class Filter {
  public static final Filter UNCONSTRAINED =
      new Filter(new RelationConstant(ImmutableSet.of(), ImmutableSet.of()), true);

  private final RelationConstant relationConstant;
  private final boolean excluding;

  private Filter(RelationConstant relationConstant, boolean excluding) {
    Preconditions.checkNotNull(relationConstant);
    Preconditions.checkArgument(!(!excluding && relationConstant.equals(RelationConstant.UNIT)));
    this.relationConstant = relationConstant;
    this.excluding = excluding;
  }

  public static Filter create(RelationConstant relationConstant, boolean excluding) {
    if (!excluding && relationConstant.equals(RelationConstant.UNIT))
      return UNCONSTRAINED;
    return new Filter(relationConstant, excluding);
  }

  public final boolean isConstrained() {
    return !equals(UNCONSTRAINED);
  }

  public Expression toExpression() {
    Expression ret = relationConstant.toExpression();
    // TODO[te]: Add a unit test for this case.
    /* The format uses the presence of the "not" operator to distinguish between, say, "excluding
    none" (not false) and "including unit tuple" (true), so don't try to do any simplifications at
    this stage. */
    if (excluding)
      ret = new FunctionExpression(Function.LOGICAL_NOT, ImmutableList.of(ret));
    Filter reverse = fromExpression(ret);
    if (!equals(reverse)) {
      throw new AssertionError(
          "Failed to re-parse generated filter expression; got " + reverse + ", expected " + this +
          " (generated expression was " + ret + ")");
    }
    return ret;
  }

  public static Filter fromExpression(Expression expression) throws IllegalArgumentException {
    Expression coreExp = Expressions.canonicalize(null, expression);
    boolean excluding = false;
    if (coreExp instanceof FunctionExpression) {
      FunctionExpression fexp = (FunctionExpression) coreExp;
      if (fexp.getFunction().equals(Function.LOGICAL_NOT)) {
        coreExp = fexp.getArguments().get(0);
        excluding = true;
      }
    }
    return Filter.create(RelationConstant.fromExpression(coreExp), excluding);
  }

  /** @return will be empty if the filter contains no tuples */
  public Set<PrimitivePath> getPaths() {
    return relationConstant.getPaths();
  }

  public Set<TupleConstant> getTuples() {
    return relationConstant.getTuples();
  }

  public boolean isExcluding() {
    return excluding;
  }

  @Override
  public int hashCode() {
    return Objects.hash(relationConstant, excluding);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof Filter))
      return false;
    Filter other = (Filter) obj;
    return this.isExcluding() == other.isExcluding()
        && Objects.equals(this.relationConstant, other.relationConstant);
  }

  @Override
  public String toString() {
    return (excluding ? "Filter excluding " : "Filter including ") + relationConstant.toString();
  }
}
