// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.query;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Position;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.ExpressionVisitor.ExpressionVisitorAdapter;
import java.math.BigDecimal;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

public final class Queries {
  private Queries() { }

  public static boolean isInPrimaryKey(View view, PrimitivePath path) {
    return view.exists(path) &&
        !view.get(QProp.SQL_COLUMN_NAME, path).isEmpty() && view.get(QProp.IN_PRIMARY_KEY, path);
  }

  public static void showHeadingPathsOnly(MutableView view, RelationPath path) {
    for (Path childPath : view.getChildPaths(path)) {
      view.set(CProp.VISIBLE, childPath,
          !childPath.isRelation() && view.get(CProp.HEADING, childPath));
    }
  }

  /* ==== Sort ================================================================================== */

  public static void sort(MutableView view, Path path, boolean descending, boolean afterPrevious) {
    if (path.isRoot())
      return;
    BigDecimal maxPosition = null;
    boolean alreadySortedOn = false;
    final RelationPath parentPath = path.getParent();
    for (Path siblingPath : view.getChildPaths(parentPath)) {
      if (afterPrevious) {
        BigDecimal position = view.get(QProp.ORDER_RANK, siblingPath).getPosition();
        if (siblingPath.equals(path) && position != null)
          alreadySortedOn = true;
        maxPosition = (maxPosition == null ||
            (position != null && position.compareTo(maxPosition) > 0)) ? position : maxPosition;
      } else {
        view.set(QProp.ORDER_RANK, siblingPath, Position.NONE);
      }
    }
    /* Technically, sorting again on the same field should not have an effect at all, but it's
    probably more useful to allow the action to be used to change the ascending/descending state
    in this case instead. */
    if (!alreadySortedOn) {
      view.set(QProp.ORDER_RANK, path, Position.create(
          maxPosition == null ? BigDecimal.ONE : maxPosition.add(BigDecimal.ONE)));
    }
    view.set(QProp.ORDER_DESCENDING, path, descending);
    view.set(QProp.ORDER_NULLS_FIRST, path, false);
    if (!view.get(CProp.PLURAL, parentPath)) {
      if (afterPrevious) {
        if (view.get(QProp.ORDER_RANK, parentPath).getPosition() == null)
          sort(view, parentPath, false, true);
      } else {
        sort(view, parentPath, false, false);
      }
    }
  }

  /* ==== Move ================================================================================== */

  public static void move(MutableView view, Iterable<Path> paths, boolean forward) {
    RelationPath commonParentPath = null;
    Set<Path> pathsToMove = Sets.newLinkedHashSet();
    for (Path path : paths) {
      if (!view.exists(path))
        continue;
      if (path.isRoot())
        return;
      RelationPath parentPath = path.getParent();
      if (commonParentPath != null && !commonParentPath.equals(parentPath))
        return;
      commonParentPath = parentPath;
      pathsToMove.add(path);
    }
    if (commonParentPath == null)
      return;
    final List<Path> siblings = Lists.newArrayList(view.getChildPaths(commonParentPath));
    movePaths(view, ImmutableSet.copyOf(pathsToMove), siblings, true, forward);
    view.set(SProp.CHILD_PATHS, commonParentPath, new PathList(siblings));
  }

  /** @param reorderable the list to be mutated */
  private static void movePaths(
      View view, ImmutableSet<Path> pathsToMove, List<Path> reorderable,
      boolean skipInvisible, boolean forward)
  {
    if (!forward)
      Collections.reverse(reorderable);
    movePathsForward(view, pathsToMove, reorderable, skipInvisible);
    if (!forward)
      Collections.reverse(reorderable);
  }

  /** @param reorderable the list to be mutated */
  private static void movePathsForward(
      View view, ImmutableSet<Path> pathsToMove, List<Path> reorderable, boolean skipInvisible)
  {
    final List<Path> pathsToMoveInOrder = Lists.newArrayList();
    final int lastPathIndex;
    final boolean contiguous;
    /* Remove the pathsToMove from reorderable and determine
    lastPathIndex/contiguous/pathsToMoveInOrder. */
    {
      final Set<Path> pathsRemaining = Sets.newLinkedHashSet(pathsToMove);
      boolean seenFirstPath = false;
      boolean toContiguous = true;
      int i = 0;
      final Iterator<Path> iter = reorderable.iterator();
      while (iter.hasNext()) {
        final Path path = iter.next();
        if (pathsRemaining.remove(path)) {
          seenFirstPath = true;
          pathsToMoveInOrder.add(path);
          iter.remove();
        } else {
          if (seenFirstPath)
            toContiguous = false;
          i++;
        }
        if (pathsRemaining.isEmpty())
          break;
      }
      lastPathIndex = i;
      contiguous = toContiguous;
    }
    final int insertionPoint;
    if (contiguous) {
      /* Move pathsToMove past the first following path, or the first visible following path if
      skipInvisible is false. If no such following path can be found, insert the paths at most one
      space forward from their original position. */
      int toInsertionPoint = Math.min(reorderable.size(), lastPathIndex + 1);
      for (int i = lastPathIndex; i < reorderable.size(); i++) {
        if (!skipInvisible || view.get(CProp.VISIBLE, reorderable.get(i))) {
          toInsertionPoint = i + 1;
          break;
        }
      }
      insertionPoint = toInsertionPoint;
    } else {
      // Don't move the paths past any succeeding paths; just move them together.
      insertionPoint = lastPathIndex;
    }
    reorderable.addAll(insertionPoint, pathsToMoveInOrder);
  }

  /* ==== Group ================================================================================= */

  private static void project(MutableView view, ImmutableSet<Path> groupByPaths,
      ImmutableSet<Path> groupByPathsAndAncestors, RelationPath path)
  {
    if (!groupByPaths.contains(path)) {
      view.set(QProp.INCLUDE_PRIMARY_KEY, path, false);
      for (Path childPath : view.getChildPaths(path)) {
        final boolean projected = groupByPathsAndAncestors.contains(childPath);
        view.set(CProp.VISIBLE, childPath, projected);
        if (projected) {
          if (childPath.isRelation())
            project(view, groupByPaths, groupByPathsAndAncestors, childPath.getRelation());
        } else {
          view.set(QProp.ORDER_RANK, childPath, Position.NONE);
          // Leave it up to the view updater to remove redundant non-visible paths.
        }
      }
    }
  }

  public static void group(MutableView view, ImmutableSet<Path> groupByPaths) {
    final ImmutableView originalView = ImmutableView.copyOf(view);
    final RelationPath groupRelation;
    final ImmutableSet<RelationPath> groupByRelations;
    {
      final Set<RelationPath> parentPaths = Sets.newLinkedHashSet();
      for (Path groupByPath : groupByPaths) {
        if (!groupByPath.isRoot())
          parentPaths.add(groupByPath.getParent());
      }
      groupRelation = Views.findLowestCommonRelationAncestor(parentPaths);
      project(view, groupByPaths, ImmutableSet.copyOf(Views.getPathsWithAncestors(groupByPaths)),
          groupRelation);
      groupByRelations = ImmutableSet.copyOf(Sets.difference(
          Views.getStrictAncestors(ImmutableSet.<Path>copyOf(groupByPaths)),
          Views.getStrictAncestors(ImmutableSet.<Path>of(groupRelation))));
    }

    final RelationPath nestedRelationPath = view.addRelationPath(groupRelation);
    final PathMap pathMap = view.copyFrom(originalView, groupRelation, nestedRelationPath);
    Views.rewritePathReferences(view, nestedRelationPath, pathMap);

    for (RelationPath groupByRelation : groupByRelations) {
      for (Path childPath : view.getChildPaths(pathMap.getNewPath(groupByRelation))) {
        if (!childPath.isRelation()) {
          final PrimitivePath originalPath = pathMap.getOldPath(childPath.getPrimitive());
          /* I believe it is correct to use QProp.IN_PRIMARY_KEY here rather than
          CProp.IN_TUPLE_KEY. Both would be correct immediately after the operation has completed,
          but using QProp.IN_PRIMARY_KEY ensures that the grouped-by columns can subsequently be
          modified through hiding and showing of the relevant fields without breaking the group-by
          semantics. Furthermore, primary key fields are less likely to be nullable, which makes the
          semantics of the resulting joins more straightforward (JOINED_ON joins do not match on
          nulls). */
          view.set(QProp.JOINED_ON, childPath, isInPrimaryKey(view, originalPath) ?
              OptionalPrimitivePath.of(originalPath) : OptionalPrimitivePath.absent());
        }
        view.set(CProp.VISIBLE, childPath,
            view.get(CProp.VISIBLE, childPath) &&
            !groupByPaths.contains(pathMap.getOldPath(childPath)));
      }
    }

    view.set(CProp.LABEL, groupRelation, view.get(CProp.LABEL, groupRelation) + " (group)");
  }

  /* ==== Miscellaneous ========================================================================= */

  private static boolean hasFilterEnabled(View view, RelationPath path) {
    if (view.get(QProp.FILTER, path).isConstrained())
      return true;
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation()) {
        if (view.get(QProp.EXISTENTIAL_FILTER, childPath))
          return true;
      } else {
        if (view.get(QProp.FILTER, childPath).isConstrained())
          return true;
      }
    }
    return false;
  }

  private static final boolean PROPAGATE_EXISTENTIAL_FILTER_THROUGH_PLURAL_RELATIONS = true;

  private static <T> void propagateExistentialFilter(
      MutableView view, RelationPath path, boolean value)
  {
    // Don't clear existential filters for relations that still have filters active.
    if (!value && hasFilterEnabled(view, path))
      return;
    view.set(QProp.EXISTENTIAL_FILTER, path, value);
    if (path.isRoot())
      return;
    if (!PROPAGATE_EXISTENTIAL_FILTER_THROUGH_PLURAL_RELATIONS && view.get(CProp.PLURAL, path))
      return;
    propagateExistentialFilter(view, path.getParent(), value);
  }

  public static <T> void propagateExistentialFilter(MutableView view, Path path, boolean value) {
    propagateExistentialFilter(view, Views.getContextPath(path), value);
  }

  public static boolean hasSignificantSort(View view, Path path) {
    if (path.isRoot() || view.get(QProp.ORDER_RANK, path).getPosition() == null)
      return false;
    /* A sort is only significant in plural relations, since there need to be more than one tuple
    to reorder. And while sorting on hidden fields can additionally have the effect of forcing the
    field to be projected in the result, this effect is also only significant in plural relations,
    where the field is not necessarily fully identified by already-projected ones. */
    return view.get(CProp.PLURAL, path.getParent()) || hasSignificantSort(view, path.getParent());
  }

  public static boolean doesPathInfluenceParent(View view, Path path) {
    if (path.isRoot())
      return false;
    final boolean sorted = hasSignificantSort(view, path);
    final boolean filtered = path.isRelation()
        ? view.get(QProp.EXISTENTIAL_FILTER, path)
        : view.get(QProp.FILTER, path).isConstrained();
    return sorted || filtered;
  }

  public static Set<Path> getUsedPaths(View view) {
    Set<Path> ret = Sets.newLinkedHashSet();
    collectUsedPaths(view, Path.ROOT, ret);
    return Collections.unmodifiableSet(ret);
  }

  private static void collectUsedPaths(View view, Path knownUsedPath, Set<Path> ret) {
    if (!view.exists(knownUsedPath))
      return;
    if (!ret.add(knownUsedPath))
      return;
    if (!knownUsedPath.isRoot())
      collectUsedPaths(view, knownUsedPath.getParent(), ret);
    if (knownUsedPath.isRelation()) {
      for (Path childPath : view.getChildPaths(knownUsedPath.getRelation())) {
        if (view.get(CProp.VISIBLE, childPath) ||
            (!childPath.isRelation() && isInPrimaryKey(view, childPath.getPrimitive())) ||
            doesPathInfluenceParent(view, childPath))
        {
          collectUsedPaths(view, childPath, ret);
        }
      }
    } else {
      PrimitivePath joinedOnPath = view.get(QProp.JOINED_ON, knownUsedPath).getNullable();
      if (joinedOnPath != null)
        collectUsedPaths(view, joinedOnPath, ret);
      for (Path referencedPath :
          getDirectlyReferencedPaths(view.get(QProp.FORMULA_EXPRESSION, knownUsedPath)))
      {
        collectUsedPaths(view, referencedPath, ret);
      }
    }
    for (Path filterPath : view.get(QProp.FILTER, knownUsedPath).getPaths())
      collectUsedPaths(view, filterPath, ret);
  }

  public static Set<Path> getDirectlyReferencedPaths(Expression expression) {
    final Set<Path> ret = Sets.newLinkedHashSet();
    expression.accept(new ExpressionVisitorAdapter<RuntimeException>() {
      @Override
      public void voidVisit(PathExpression exp) {
        ret.add(exp.getPath());
      }
    });
    return Collections.unmodifiableSet(ret);
  }

  public static boolean shouldProject(View view, Path path) {
    if (view.get(CProp.VISIBLE, path)) {
      return true;
    } else if (path.isRelation()) {
      return path.isRoot();
    } else if (view.get(QProp.ORDER_RANK, path).getPosition() != null) {
      return true;
    } else {
      final RelationPath parentPath = path.getParent();
      final boolean includePrimaryKey = view.get(QProp.INCLUDE_PRIMARY_KEY, parentPath);
      return (includePrimaryKey && Queries.isInPrimaryKey(view, path.getPrimitive()));
    }
  }

  /** @param path does not have to be plural */
  public static Set<PrimitivePath> getImmediateProjectedChildren(View view, RelationPath path) {
    Set<PrimitivePath> ret = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(path)) {
      if (!childPath.isRelation() && shouldProject(view, childPath))
        ret.add(childPath.getPrimitive());
    }
    return Collections.unmodifiableSet(ret);
  }

  public static Set<PrimitivePath> getProjectedToRoot(View view, RelationPath targetPath) {
    Set<PrimitivePath> ret = Sets.newLinkedHashSet();
    for (RelationPath ancestorPath :
        Views.getRelationPathsWithAncestors(ImmutableSet.of(targetPath)))
    {
      ret.addAll(getImmediateProjectedChildren(view, ancestorPath));
    }
    return Collections.unmodifiableSet(ret);
  }

  // TODO[cc]: Avoid duplicating the logic here and in PathReferences.isVisibleToRoot.
  public static boolean isVisibleToRoot(View view, Path path) {
    if (!view.get(CProp.VISIBLE, path)) {
      return false;
    } else if (path.isRoot()) {
      return true;
    } else {
      return isVisibleToRoot(view, path.getParent());
    }
  }
}
