// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.query;

import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.ViewPropertyType;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.ExpressionParseException;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.Expressions;

public final class QueryPropertyTypes {
  private QueryPropertyTypes() { }

  public static final ViewPropertyType<InstanceDefinition> INSTANCE_DEFINITION_TYPE =
      new ViewPropertyType<InstanceDefinition>(InstanceDefinition.class)
  {
    @Override
    public String valueToString(Strepco strepco, InstanceDefinition value) {
      if (value.getTable() == null) {
        return "";
      } else {
        TableIdentifier ti = value.getTable();
        if (!(ti instanceof SQLTable))
          throw new UnsupportedOperationException("Unsupported TableIdentifier");
        return ((SQLTable) ti).generateStringRepresentation();
      }
    }

    @Override
    public InstanceDefinition stringToValue(Strepco strepco, String str)
        throws IllegalArgumentException
    {
      if (str.trim().isEmpty()) {
        return new InstanceDefinition(null);
      } else {
        SQLTable table = SQLTable.createFromStringRepresentation(str);
        if (table == null)
          throw new IllegalArgumentException("Malformed table identifier");
        return new InstanceDefinition(table);
      }
    }
  };

  public static final ViewPropertyType<Expression> EXPRESSION_TYPE =
      new ViewPropertyType<Expression>(Expression.class)
  {
    @Override
    public String valueToString(Strepco strepco, Expression value) {
      return Expressions.generateString(strepco, value);
    }

    @Override
    public Expression stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
      try {
        return Expressions.parse(strepco, str);
      } catch (ExpressionParseException e) {
        throw new IllegalArgumentException("Failed to parse expression", e);
      }
    }

    @Override
    public Expression rewritePathReferences(final PathMap pathMap, Expression input) {
      return Expressions.rewritePathReferences(pathMap, input);
    }
  };

  public static final ViewPropertyType<OptionalPrimitivePath> OPTIONAL_PRIMITIVE_PATH_TYPE =
      new ViewPropertyType<OptionalPrimitivePath>(OptionalPrimitivePath.class)
  {
    @Override
    public String valueToString(Strepco strepco, OptionalPrimitivePath value) {
      return !value.isPresent() ? "" :
          Expressions.generateString(strepco, new PathExpression(value.getPath()));
    }

    @Override
    public OptionalPrimitivePath stringToValue(Strepco strepco, String str)
        throws IllegalArgumentException
    {
      if (str.trim().isEmpty()) {
        return OptionalPrimitivePath.absent();
      } else {
        Expression exp;
        try {
          exp = Expressions.parse(strepco, str);
        } catch (ExpressionParseException e) {
          throw new IllegalArgumentException("Failed to parse path reference", e);
        }
        if (!(exp instanceof PathExpression))
          throw new IllegalArgumentException("Value must be a path reference expression");
        Expression.PathExpression ret = (Expression.PathExpression) exp;
        if (ret.getPath().isRelation())
          throw new IllegalArgumentException("Value must reference primitive paths only");
        return OptionalPrimitivePath.of(ret.getPath().getPrimitive());
      }
    }

    @Override
    public OptionalPrimitivePath rewritePathReferences(
        PathMap pathMap, OptionalPrimitivePath input)
    {
      return input.isPresent()
          ? OptionalPrimitivePath.of(pathMap.getNewPath(input.getPath()))
          : OptionalPrimitivePath.absent();
    }
  };

  public static final ViewPropertyType<Filter> FILTER_TYPE =
      new ViewPropertyType<Filter>(Filter.class)
  {
    @Override
    public String valueToString(Strepco strepco, Filter value) {
      return Expressions.generateString(strepco, value.toExpression());
    }

    @Override
    public Filter stringToValue(Strepco strepco, String str) throws IllegalArgumentException {
      Expression expression;
      try {
        expression = Expressions.parse(strepco, str);
      } catch (ExpressionParseException e) {
        throw new IllegalArgumentException("Failed to parse expression", e);
      }
      return Filter.fromExpression(expression);
    }

    @Override
    public Filter rewritePathReferences(PathMap pathMap, Filter input) {
      return Filter.fromExpression(
          Expressions.rewritePathReferences(pathMap, input.toExpression()));
    }
  };
}
