// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.query;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

// TODO[cc]: Move this back into the Database module, or get rid of TableIdentifier.
@Immutable
public final class SQLTable implements TableIdentifier {
  private final SQLDatabase database;
  private final String tableName;

  private SQLTable(SQLDatabase database, String tableName) {
    Preconditions.checkNotNull(database);
    Preconditions.checkNotNull(tableName);
    Preconditions.checkArgument(!tableName.isEmpty());
    this.database = database;
    this.tableName = tableName;
  }

  /* TODO[cc]: Move generateStringRepresentation/createFromStringRepresentation into
               QueryPropertyTypes. */

  public String generateStringRepresentation() {
    try {
      /* Some kind of escaping is needed in order to unambiguously represent the catalog, schema,
      and table names that make up a fully qualified table identifier. Using the URL encoding
      guarantees that the slash character ('/') will never be present in the output, so I can use it
      as a separator and easily decode the result using a simple split. */
      String ret =
          URLEncoder.encode(database.getPrimaryNamePart(), "UTF-8") + "/" +
          URLEncoder.encode(database.getSecondaryNamePart(), "UTF-8") + "/" +
          URLEncoder.encode(getTableName(), "UTF-8");
      if (!equals(createFromStringRepresentation(ret)))
        throw new AssertionError("Encoding not reversible");
      return ret;
    } catch (UnsupportedEncodingException e) {
      throw new RuntimeException("Unexpected exception", e);
    }
  }

  /** @return null iff format is unacceptable */
  public static @Nullable SQLTable createFromStringRepresentation(String str) {
    String[] split = str.split("/");
    if (split.length != 3)
      return null;
    try {
      String primaryDBNamePart = URLDecoder.decode(split[0], "UTF-8").trim();
      String secondaryDBNamePart = URLDecoder.decode(split[1], "UTF-8").trim();
      String tableName = URLDecoder.decode(split[2], "UTF-8").trim();
      if (tableName.isEmpty() || (primaryDBNamePart.isEmpty() && !secondaryDBNamePart.isEmpty()))
        return null;
      return SQLTable.create(new SQLDatabase(primaryDBNamePart, secondaryDBNamePart), tableName);
    } catch (UnsupportedEncodingException e) {
      throw new RuntimeException("Unexpected exception", e);
    }
  }

  /** @param tableName will be trimmed. */
  public static SQLTable create(SQLDatabase database, @Nonnull String tableName) {
    Preconditions.checkNotNull(tableName);
    return new SQLTable(database, tableName.trim());
  }

  public SQLDatabase getDatabase() {
    return database;
  }

  /** Will never be empty. */
  public String getTableName() {
    return tableName;
  }

  @Override
  @SuppressWarnings("AccessingNonPublicFieldOfAnotherObject")
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof SQLTable))
      return false;
    if (obj == this)
      return true;
    SQLTable rhs = (SQLTable) obj;
    return Objects.equal(database , rhs.database)
        && Objects.equal(tableName, rhs.tableName);
  }

  @Override
  public int hashCode() {
    return Objects.hashCode(database, tableName);
  }

  @Override
  public String toString() {
    return generateStringRepresentation();
  }
}
