// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicReference;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import org.openide.util.RequestProcessor;
import org.openide.util.RequestProcessor.Task;

/**
 * A value computed computed asynchronously from an input value. Output values are not computed
 * unless at least one listener is registered.
 *
 * Properties:
 * <ul>
 *   <li>No out-of-date values will ever be returned.</li>
 *   <li>Multiple subsequent input changes may be coalesced into a single {@code ChangeEvent} being
 *       fired.</li>
 *   <li>When no listeners are registered, the implementation will not keep strong references to
 *       previously calculated results.</li>
 *   <li>After {@link ValueProcessor#process(Object) ValueProcessor.process} has returned,
 *       {@link ValueProcessor#clearState()} will be called if the {@code AsynchronousFlow} no
 *       longer has any listeners.</li>
 * </ul>
 */
final class AsynchronousFlow<I,O> extends Flow<O> {
  private static final Logger LOG = Logger.getLogger(AsynchronousFlow.class.getName());
  /* TODO[pd]: Technically, Runtime.getRuntime().availableProcessors() may change during the
               lifetime of the JVM. Consider finding a way to refresh the throughput
               occasionally. */
  private static final RequestProcessor RP_INT = new RequestProcessor(
      "Interruptible Value Processor", Runtime.getRuntime().availableProcessors(), true, true);
  /* For processors that don't support thread interruption, we still would like to be able to cancel
  tasks that haven't been started yet. Keep a separate RequestProcessor for this purpose. */
  private static final RequestProcessor RP_NOINT = new RequestProcessor(
      "Non-interruptible Value Processor", Runtime.getRuntime().availableProcessors(), false, true);
  /**
   * For debugging only; used to wait until all tasks have completed. Used both as a counter and a
   * monitor lock.
   */
  private static final AtomicInteger DEBUG_TASKS_IN_PROGRESS = new AtomicInteger();
  /**
   * For debugging only; ensures that tests fail in the case of exceptions thrown from within the
   * request processor thread.
   */
  private static final AtomicReference<RuntimeException> DEBUG_STORED_EXCEPTION =
      new AtomicReference<>();

  private final Flow<I> input;
  private final ValueProcessor<I,O> processor;
  private final int timeToWaitMillis;
  private int processesInProgress;
  private boolean hasListeners;
  private @Nullable Task ongoingTask;
  /** The last evaluated input for the current value or for an ongoing process. */
  private @Nullable Snapshot<?> lastEvaluatedInput;

  /**
   * @param processor must be asynchronous
   */
  AsynchronousFlow(Flow<I> input, int timeToWaitMillis, ValueProcessor<I,O> processor) {
    super("Asynchronous");
    Preconditions.checkNotNull(input);
    Preconditions.checkArgument(processor.isAsynchronous());
    Preconditions.checkArgument(timeToWaitMillis >= 0);
    this.input = input;
    this.processor = processor;
    this.timeToWaitMillis = timeToWaitMillis;
  }

  /** Block until all processes have terminated. For use by unit tests only. */
  static void awaitTermination() {
    RuntimeException storedException = DEBUG_STORED_EXCEPTION.get();
    if (storedException != null)
      throw new RuntimeException("Processor thread threw exception", storedException);
    synchronized (DEBUG_TASKS_IN_PROGRESS) {
      while (DEBUG_TASKS_IN_PROGRESS.get() > 0) {
        try {
          DEBUG_TASKS_IN_PROGRESS.wait();
        } catch (InterruptedException e) {
          throw new AssertionError(e);
        }
      }
    }
  }

  @Override
  protected void firstListenerAdded() {
    Preconditions.checkState(Thread.holdsLock(this));
    input.addChangeListener(propagatingChangeListener);
    hasListeners = true;
  }

  @Override
  protected void lastListenerRemoved() {
    Preconditions.checkState(Thread.holdsLock(this));
    input.removeChangeListener(propagatingChangeListener);
    lastEvaluatedInput = null;
    hasListeners = false;
    cancelOngoingTask();
    clearStateWhenAppropriate();
  }

  /**
   * Call {@code process.clearState()} if there are no listeners attached and no processes currently
   * going on. The caller must hold a lock on this object. Should never block.
   */
  private void clearStateWhenAppropriate() {
    Preconditions.checkState(Thread.holdsLock(this));
    if (!hasListeners && processesInProgress == 0)
      processor.clearState();
  }

  private void cancelOngoingTask() {
    Preconditions.checkState(Thread.holdsLock(this));
    if (ongoingTask != null) {
      ongoingTask.cancel();
      ongoingTask = null;
    }
  }

  @Override
  protected void update() {
    Preconditions.checkState(Thread.holdsLock(this));
    final Snapshot<I> currentEvaluatedInput = input.getSnapshot();
    if (Objects.equal(currentEvaluatedInput, lastEvaluatedInput)) {
      // No update needed.
      return;
    }
    cancelOngoingTask();
    lastEvaluatedInput = currentEvaluatedInput;
    if (currentEvaluatedInput.getValue() == null) {
      setInternal(null);
      return;
    }
    // Invalidate the output until a new value has been computed.
    setInternal(null);
    RequestProcessor requestProcessor = processor.isInterruptible() ? RP_INT : RP_NOINT;
    ongoingTask = requestProcessor.post(() -> {
      synchronized (DEBUG_TASKS_IN_PROGRESS) {
        DEBUG_TASKS_IN_PROGRESS.incrementAndGet();
      }
      try {
        performAsynchronousProcess(currentEvaluatedInput);
      } catch (RuntimeException e) {
        LOG.log(Level.SEVERE, "Asynchronous process failed", e);
        DEBUG_STORED_EXCEPTION.compareAndSet(null, e);
      } finally {
        synchronized (DEBUG_TASKS_IN_PROGRESS) {
          DEBUG_TASKS_IN_PROGRESS.decrementAndGet();
          DEBUG_TASKS_IN_PROGRESS.notifyAll();
        }
      }
    }, timeToWaitMillis);
  }

  private void performAsynchronousProcess(Snapshot<I> currentEvaluatedInput) {
    Preconditions.checkState(!Thread.holdsLock(this));
    Preconditions.checkNotNull(currentEvaluatedInput.getValue());
    final O output;
    try {
      LOG.log(Level.FINE, "Processor {0} started", processor);
      output = calculateOutput(currentEvaluatedInput.getValue());
      if (output == null) {
        LOG.log(Level.FINE, "Processor {0} completed (returned null)", processor);
      } else {
        LOG.log(Level.FINE, "Processor {0} completed", processor);
      }
    } catch (InterruptedException e) {
      LOG.log(Level.FINE, "Processor {0} was gracefully interrupted", processor);
      // No action; leave output as null.
      return;
    }
    final boolean inputStillValid;
    synchronized (this) {
      inputStillValid = Objects.equal(currentEvaluatedInput, lastEvaluatedInput);
      if (inputStillValid) {
        setInternal(output);
        /* Resetting this reference is not critical for correctness, but allow it to be
        garbage-collected anyway. */
        ongoingTask = null;
      } else {
        LOG.log(Level.INFO, "Input changed while processor {0} was running; " +
            "deferring to later update (which may or may not already have completed)", processor);
      }
    }
    if (inputStillValid)
      propagateChanges();
  }

  private O calculateOutput(I input) throws InterruptedException {
    Preconditions.checkNotNull(input);
    synchronized (this) {
      processesInProgress++;
    }
    try {
      return processor.process(input);
    } finally {
      synchronized (this) {
        processesInProgress--;
        clearStateWhenAppropriate();
      }
    }
  }

  @Override
  protected Set<Flow<?>> getDependencies() {
    return ImmutableSet.<Flow<?>>of(input);
  }
}
