// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;

/**
 * A value computed synchronously from zero or more Flow inputs.
 *
 * <p>Properties:
 * <ul>
 *   <li>Whenever a new value needs to be produced, all input values are retrieved as if they were
 *       read at a single point in time.</li>
 *   <li>When no listeners are registered, the implementation will not keep strong references to
 *       previously calculated results.</li>
 *   <li>When at least one listener is registered, all input values are listened to.</li>
 * </ul>
 * </p>
 */
final class CompositeFlow<O> extends Flow<O> {
  private static final Logger LOG = Logger.getLogger(CompositeFlow.class.getName());
  private final ImmutableList<Flow<?>> inputs;
  private final CompositeFactory<O> compositeFactory;
  /** The last evaluated inputs for the current value or for an ongoing process. */
  private @Nullable List<Snapshot<?>> evaluatedInputs;

  /**
   * Construct a {@code CompositeFlow} with an arbitrary number of inputs. The output value type
   * and number of {@code inputs} must correspond to those expected by the {@code compositeFactory}.
   */
  CompositeFlow(Flow<?>[] inputs, CompositeFactory<O> compositeFactory) {
    super("Composite");
    Preconditions.checkNotNull(compositeFactory);
    Preconditions.checkNotNull(inputs);
    this.inputs = ImmutableList.copyOf(inputs);
    this.compositeFactory = compositeFactory;
  }

  @Override
  protected void firstListenerAdded() {
    Preconditions.checkState(Thread.holdsLock(this));
    for (Flow<?> input : inputs)
      input.addChangeListener(propagatingChangeListener);
  }

  @Override
  protected void lastListenerRemoved() {
    Preconditions.checkState(Thread.holdsLock(this));
    for (Flow<?> input : inputs)
      input.removeChangeListener(propagatingChangeListener);
    evaluatedInputs = null;
    compositeFactory.clearState();
  }

  /**
   * Get a time-consistent set of snapshots from the inputs. It is not required that the caller hold
   * a lock on this object.
   */
  private List<Snapshot<?>> getConsistentInputSnapshots() {
    final int WARN_ITERATION_COUNT = 100;
    /* Seed the snapshot-finding loop with the previous input snapshot to ensure that getSnapshot()
    is called only a single time for each input in the case where inputs have not changed. This
    avoids exponential runtime behavior for cascaded composite flows. */
    List<Snapshot<?>> currentSnapshots = evaluatedInputs;
    List<Snapshot<?>> previousSnapshots = null;
    /* Read all inputs until we get two identical lists of snapshots in a row. Getting two identical
    lists of snapshots in a row means that no values were changed during the time between the two
    series of reads. This in turn means that we have a point-in-time consistent snapshot of the
    input snapshots. */
    int i = 0;
    do {
      if (currentSnapshots != null && previousSnapshots != null && i <= WARN_ITERATION_COUNT) {
        if (i < WARN_ITERATION_COUNT) {
          if (LOG.isLoggable(Level.FINE)) {
            LOG.log(Level.FINE,
                "Input snapshot for composite {0} disturbed; re-reading (got {1} then {2})",
                new Object[] { compositeFactory, previousSnapshots, currentSnapshots });
          }
        } else {
          LOG.log(Level.WARNING, "Failed to grab consistent input snapshots after " +
              WARN_ITERATION_COUNT + " iterations", new RuntimeException());
        }
        i++;
      }
      previousSnapshots = currentSnapshots;
      currentSnapshots = Lists.newArrayListWithCapacity(inputs.size());
      for (Flow<?> input : inputs)
        currentSnapshots.add(input.getSnapshot());
    } while (previousSnapshots == null || !Objects.equal(currentSnapshots, previousSnapshots));
    return currentSnapshots;
  }

  @Override
  protected void update() {
    Preconditions.checkState(Thread.holdsLock(this));
    final List<Snapshot<?>> currentEvaluatedInputs = getConsistentInputSnapshots();
    if (Objects.equal(currentEvaluatedInputs, evaluatedInputs)) {
      // No update needed.
      return;
    }
    evaluatedInputs = currentEvaluatedInputs;
    for (Snapshot<?> snapshot : currentEvaluatedInputs) {
      if (snapshot.getValue() == null) {
        setInternal(null);
        return;
      }
    }
    List<Object> inputObjects = Lists.newArrayListWithCapacity(inputs.size());
    for (Snapshot<?> snapshot : currentEvaluatedInputs)
      inputObjects.add(snapshot.getValue());
    O output = null;
    try {
      output = compositeFactory.construct(inputObjects.toArray(new Object[inputObjects.size()]));
    } finally {
      setInternal(output);
    }
  }

  @Override
  protected Set<Flow<?>> getDependencies() {
    return ImmutableSet.copyOf(inputs);
  }
}
