// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.util.MapUtil;
import java.util.Objects;
import java.util.Set;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/* TODO[cc]: This class was once called "DynamicValue". Rename any remaining identifiers that still
             use the word "value" to refer to this class and its descendants (e.g.
             ValueProcessor). */
/**
 * Model for a value reference that may change over time. Values themselves may be mutable or
 * immutable, but only changes in the actual reference will trigger firing of change events.
 * Implementations must be thread-safe and deadlock-free.
 *
 * <p>This API has been designed to allow the construction of data flow DAGs for which output values
 * can always be guaranteed to be consistent with input values (or {@code null} until an output has
 * been computed), without risk of deadlock. Consistency and deadlock avoidance is achieved as
 * follows:</p>
 *
 * <ul>
 *   <li>A {@code Flow} object holding a lock on itself may indirectly acquire the lock on other
 *       {@code Flow} objects by calling methods on the latter, but only in the input direction.
 *       {@code Flow} objects that are used as outputs are inaccessable to their input {@code Flow}
 *       objects except through change event listeners, and these listeners can only be called once
 *       the source {@code Flow} has released the lock on itself.</li>
 *   <li>Because change events are fired only after the source {@code Flow} releases the lock on
 *       itself, and because change events may be coalesced, the client is not guaranteed to be able
 *       to track every new value of the output. However, because output values are wrapped in a
 *       {@link Snapshot} object that is recreated on each change, it is possible for clients to
 *       notice output transients even when the final output value is the same as when the client
 *       last read it. This makes it possible to derive point-in-time snapshots of multiple
 *       simultaneous {@code Flow} outputs.</li>
 * </ul>
 *
 * @param <T> the type of the value represented
 */
@ThreadSafe
public abstract class Flow<T> {
  private final ChangeEvent changeEvent = new ChangeEvent(this);
  private final Set<ChangeListener> changeListeners = Sets.newLinkedHashSet();
  private final String description;
  private @Nonnull Snapshot<T> snapshot = Snapshot.create();
  private boolean changeEventDue = false;
  /**
   * A ChangeListener that simply calls {@link #propagateChanges()}. Provided as a convenience for
   * subclasses.
   */
  protected final ChangeListener propagatingChangeListener = (ChangeEvent e) -> {
    propagateChanges();
  };

  protected Flow(String description) {
    Preconditions.checkNotNull(description);
    this.description = description;
  }

  /**
   * Add a {@code ChangeListener}. Change listeners should never block.
   */
  public final synchronized void addChangeListener(ChangeListener listener) {
    final boolean wasEmpty = changeListeners.isEmpty();
    MapUtil.addChecked(changeListeners, listener);
    if (wasEmpty) {
      firstListenerAdded();
      update();
      /* Adding a change listener only guarantees that events will be fired for changes occuring
      _after_ this method has returned, so there is no need to fire events for changes happening
      immediately as a result of the first listener being added. */
      changeEventDue = false;
    }
  }

  /**
   * Remove a previously added {@code ChangeListener}. There is no guarantee that the listener will
   * not continue to receive additional events for some time after this method has returned.
   */
  public final synchronized void removeChangeListener(final ChangeListener listener) {
    /* Once upon a time, it used to be illegal to call propagateChanges() if no listeners were
    registered. Thus, it was important to call lastListenerRemoved() _before_ the listener was
    removed, to give the subclass time to unregister all of its own listeners before it became
    illegal to call propagateChanges(). This might no longer be critical, but the original order of
    operations remains the same for now. */
    final boolean isLastOne = changeListeners.equals(ImmutableSet.of(listener));
    if (isLastOne) {
      lastListenerRemoved();
      snapshot = snapshot.update(null);
      // Since there are no listeners, no change events need to be fired.
      changeEventDue = false;
    }
    MapUtil.removeChecked(changeListeners, listener);
    Preconditions.checkState(!isLastOne || changeListeners.isEmpty());
  }

  /**
   * Called after the first listener is added, while still holding a lock on this object. The
   * {@link #update()} method should <em>not</em> be called from this method. After
   * {@link #firstListenerAdded()} has returned, {@link #update()} will be called once.
   */
  protected void firstListenerAdded() {
  }

  /**
   * Called just before the last listener is removed, while holding a lock on this object. The
   * {@link #update()} method should <em>not</em> be called from this method.
   */
  protected void lastListenerRemoved() {
  }

  /**
   * Update output and fire change events if appropriate. May not be called while holding a lock on
   * this object. It is OK to call this method even if no listeners are currently registered.
   */
  protected final void propagateChanges() {
    Preconditions.checkState(!Thread.holdsLock(this));
    ImmutableSet<ChangeListener> changeListenersSnapshot;
    synchronized (this) {
      if (changeListeners.isEmpty())
        return;
      update();
      if (!changeEventDue)
        return;
      changeEventDue = false;
      /* Make a copy of the listener list before calling any listeners, and before relinquishing the
      lock, since the listeners could themselves modify the list, for instance to unregister
      themselves in the same operation. */
      changeListenersSnapshot = ImmutableSet.copyOf(changeListeners);
    }
    for (ChangeListener changeListener : changeListenersSnapshot)
      changeListener.stateChanged(changeEvent);
  }

  /**
   * Set the output value and schedule a change event to be fired the next time
   * {@link #propagateChanges()} is called. The output snapshot will always be updated, regardless
   * of whether the new value is different from the old one or not. Should be called by subclasses
   * only.
   */
  protected synchronized void setInternal(@Nullable T newValue) {
    if (changeListeners.isEmpty())
      return;
    snapshot = snapshot.update(newValue);
    changeEventDue = true;
  }

  /**
   * Same as {@link #setInternal(Object)}, except only set the value if it differs from the current
   * one (as per {@code Object.equals}).
   */
  protected synchronized void setInternalIfDifferent(@Nullable T newValue) {
    if (!Objects.equals(snapshot.getValue(), newValue))
      setInternal(newValue);
  }

  /**
   * Get the current value. Shorthand for {@code getSnapshot().getValue()}.
   *
   * @see #getSnapshot()
   * @return the value; null if unavailable or not yet computed
   */
  public final @Nullable T get() {
    return getSnapshot().getValue();
  }

  /**
   * Get the current value snapshot. At least one listener must be registered for the value to be
   * computed. Calling this method is still permitted when no listeners are registered; this design
   * ensures that it is always valid to call {@code get()} from {@code ChangeListener.stateChanged},
   * which may be called for some time even after {@link #removeChangeListener(ChangeListener)} has
   * returned.
   */
  public final synchronized @Nonnull Snapshot<T> getSnapshot() {
    if (!changeListeners.isEmpty())
      update();
    return snapshot;
  }

  /**
   * Ensure that the current output is up to date. Implementations should call
   * {@link #setInternal(Object)} to change the output snapshot if and only if the semantics of the
   * value require it. Need not be called directly by subclasses or clients. Will only be called
   * while at least one listener is registered, while holding a lock on this object.
   */
  protected abstract void update();

  @Override
  public final String toString() {
    return description;
  }

  final String getDescription() {
    return description;
  }

  /**
   * Get the input dependencies of this {@code Flow}, if any. Used only for debugging.
   */
  protected Set<Flow<?>> getDependencies() {
    return ImmutableSet.of();
  }
}
