// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.google.common.base.Preconditions;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.escape.Escaper;
import com.google.common.escape.Escapers;
import com.sieuferd.flow.ProcessorTemplates.Function2;
import com.sieuferd.flow.ProcessorTemplates.Function3;
import com.sieuferd.flow.ProcessorTemplates.Function4;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate2;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate3;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate4;
import com.sieuferd.flow.Refreshable.RefreshableAndFlow;
import com.sieuferd.util.UnitObject;
import java.io.PrintStream;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import javax.annotation.Nullable;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public final class Flows {
  private Flows() { }

  public static <T> SettableFlow<T> newSettable() {
    return newSettable(null);
  }

  public static <T> SettableFlow<T> newSettable(@Nullable T initialValue) {
    SettableFlow<T> ret = new SettableFlow<>();
    ret.set(initialValue);
    return ret;
  }

  public static <T> Flow<T> newConstant(@Nullable T value) {
    return newIdentity(newSettable(value));
  }

  /**
   * @return a Flow that outputs its input without modification; useful for hiding specific Flow
   *         implementations from clients
   */
  public static <O> Flow<O> newIdentity(Flow<O> input) {
    return newProcessed(input, input2 -> input2);
  }

  private static <I,O> Flow<O> newProcessed(
      Flow<I> input, int timeToWaitMillis, final ValueProcessor<I,O> processor)
  {
    Preconditions.checkArgument(processor.isAsynchronous() || timeToWaitMillis == 0);
    if (processor.isAsynchronous()) {
      return new AsynchronousFlow<>(input, timeToWaitMillis, processor);
    } else {
      // Just run the processor from a singleton CompositeFactory.
      return newComposite(new Flow<?>[] { input }, new CompositeFactory<O>() {
        @Override
        protected O construct(Object[] args) {
          @SuppressWarnings("unchecked")
          final I input = (I) args[0];
          try {
            return processor.process(input);
          } catch (InterruptedException e) {
            throw new RuntimeException("Unexpected thread interruption", e);
          }
        }

        @Override
        void clearState() {
          processor.clearState();
        }
      });
    }
  }

  public static <I,O> Flow<O> newProcessed(Flow<I> input, ValueProcessor<I,O> processor) {
    return newProcessed(input, 0, processor);
  }

  /**
   * @param function a function to be run synchronously
   */
  public static <O,I> Flow<O> newProcessed(Flow<I> input, Function<I,O> function) {
    return newProcessed(input, 0, new ValueProcessor<I,O>() {
      @Override
      public O process(I input) {
        return function.apply(input);
      }
    });
  }

  public static <O,I0,I1> Flow<O> newComposite2(
      Flow<I0> i0, Flow<I1> i1, Function2<O,I0,I1> f)
  {
    return new ProcessorTemplate2<O,I0,I1>() {
      @Override
      protected O process(I0 arg0, I1 arg1) {
        return f.apply(arg0, arg1);
      }
    }.getFlow(i0, i1);
  }

  public static <O,I0,I1,I2> Flow<O> newComposite3(
      Flow<I0> i0, Flow<I1> i1, Flow<I2> i2, Function3<O,I0,I1,I2> f)
  {
    return new ProcessorTemplate3<O,I0,I1,I2>() {
      @Override
      protected O process(I0 arg0, I1 arg1, I2 arg2) {
        return f.apply(arg0, arg1, arg2);
      }
    }.getFlow(i0, i1, i2);
  }

  public static <O,I0,I1,I2,I3> Flow<O> newComposite4(
      Flow<I0> i0, Flow<I1> i1, Flow<I2> i2, Flow<I3> i3, Function4<O,I0,I1,I2,I3> f)
  {
    return new ProcessorTemplate4<O,I0,I1,I2,I3>() {
      @Override
      protected O process(I0 arg0, I1 arg1, I2 arg2, I3 arg3) {
        return f.apply(arg0, arg1, arg2, arg3);
      }
    }.getFlow(i0, i1, i2, i3);
  }

  // TODO[cc]: Get rid of this.
  public static <O> Flow<O> newComposite(Flow<?>[] inputs, CompositeFactory<O> inputFactory) {
    return new CompositeFlow<>(inputs, inputFactory);
  }

  public static <I,O> Flow<O> newIndirect(Flow<I> input, IndirectProcessor<I,O> processor) {
    return new IndirectFlow<>(input, processor);
  }

  /**
   * @param function a function to be run synchronously
   */
  public static <I,O> Flow<O> newIndirect(Flow<I> input, Function<I,Flow<O>> function) {
    return new IndirectFlow<>(input, new IndirectProcessor<I,O>() {
      @Override
      public Flow<O> process(I input) {
        return function.apply(input);
      }
    });
  }

  // TODO[pd]: Don't delay propagation events if they are more than delayMillis apart.
  public static <T> Flow<T> newCoalescing(int delayMillis, Flow<T> input) {
    if (delayMillis == 0)
      return input;
    /* An alternative implementation would be to simply wait for delayMillis before returning from
    ValueProcessor.process, but relying on RequestProcessor.post's timeToWait argument seems more
    likely to guarantee that no expensive operations are done. */
    return newProcessed(input, delayMillis, new ValueProcessor<T,T>() {
      @Override
      public ValueProcessor.Synchronicity getSynchronicity() {
        return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
      }

      @Override
      public T process(T input) {
        return input;
      }
    });
  }

  public static <T> RefreshableAndFlow<T> newRefreshableAndFlow(Flow<T> input) {
    return new RefreshableAndFlow<>(input);
  }

  /**
   * A flow that is connected to its input only when {@code switchFlow} is set. Can be used as an
   * explicitly managed alternative to {@code newWeak(Flow)}.
   */
  public static <T> Flow<T> newValve(Flow<UnitObject> switchFlow, final Flow<T> input) {
    return Flows.newIndirect(switchFlow, switchValue -> input);
  }

  /**
   * A flow that produces the last non-null input value available.
   */
  public static <T> Flow<T> newSticky(Flow<T> input, @Nullable T initialValue) {
    return new StickyFlow<>(input, initialValue);
  }

  /**
   * A flow that produces a default value if the input value is null.
   */
  public static <T> Flow<T> newDefault(Flow<T> input, @Nullable T defaultValue) {
    return new DefaultFlow<>(input, defaultValue);
  }

  /**
   * A flow containing the value true if the input values is not null, otherwise false.
   */
  public static <T> Flow<Boolean> newBoolean(Flow<T> input) {
    return newDefault(newProcessed(input, input2 -> true), false);
  }

  public static <T> Flow<T> attachSideEffectDependency(
      boolean waitForSideEffect, Flow<UnitObject> sideEffectFlow, Flow<T> inputFlow)
  {
    final Flow<UnitObject> useSideEffect = waitForSideEffect ? sideEffectFlow :
        Flows.newSticky(sideEffectFlow, UnitObject.INSTANCE);
    return newComposite2(inputFlow, useSideEffect, (input, sideEffect) -> input);
  }

  /* While the GraphViz documentation claims that "In quoted strings in DOT, the only escaped
  character is double-quote" ( http://www.graphviz.org/doc/info/lang.html ), I've found this to be
  wrong; backslashes and newlines do need to be escaped as well. */
  private static final Escaper GV_ESCAPER = Escapers.builder()
      .addEscape('\\', "\\\\")
      .addEscape('\"', "\\\"")
      .addEscape('\n', "\\n")
      .build();

  private static String gvStringLiteral(String s) {
    return "\"" + GV_ESCAPER.escape(s) + "\"";
  }

  private static void collectDependencies(
      Multimap<Flow<?>,Flow<?>> ret, Flow<?> target)
  {
    if (ret.containsKey(target))
      return;
    final Set<Flow<?>> dependencies = target.getDependencies();
    ret.putAll(target, dependencies);
    for (Flow<?> dependency : dependencies)
      collectDependencies(ret, dependency);
  }

  private static String getFlowLabel(BiMap<Flow<?>,String> typeLabels, Flow<?> flow) {
    // TODO[pd]: Rewrite this. Flows no longer store their types at runtime.
    /* If multiple values have the same value type, distinguish the values by adding a postfix to
    the value type part of the label, regardless of whether differences in flow type might make the
    value label as a whole unique. */
    String typeLabel = typeLabels.get(flow);
    if (typeLabel == null) {
      int postfix = 1;
      do {
        typeLabel = ("#" + postfix);
        postfix++;
      } while ((typeLabels.inverse().containsKey(typeLabel)));
      typeLabels.put(flow, typeLabel);
    }
    return flow.getDescription() + "\n<" + typeLabel + ">";
  }

  /**
   * Print {@code Flow} dependencies in a format suitable for producing a dependency graph using
   * <a href="http://www.graphviz.org">GraphViz</a>.
   */
  public static void printGraphVizDependencyChart(PrintStream out, Flow<?> goal) {
    final Multimap<Flow<?>,Flow<?>> dependencies = LinkedHashMultimap.create();
    collectDependencies(dependencies, goal);
    final BiMap<Flow<?>,String> typeLabels = HashBiMap.create();
    out.println(
        "// ===== Data Flow Diagram (auto-generated in GraphViz DOT format) =============");
    out.println("// Example usage: dot deps.gv -Tpdf > deps.pdf");
    out.println("digraph {");
    for (Map.Entry<Flow<?>,Flow<?>> entry : dependencies.entries()) {
      /* Make the arrows point in the direction of the data flow rather than in the direction of the
      dependencies. */
      out.println("  " +
          gvStringLiteral(getFlowLabel(typeLabels, entry.getValue())) + " -> " +
          gvStringLiteral(getFlowLabel(typeLabels, entry.getKey())) + ";");
    }
    out.println("}");
    out.println(
        "// =============================================================================");
  }

  private static final ChangeListener DUMMY_LISTENER = (ChangeEvent e) -> { };

  /**
   * Momentarily attach a listener to a flow to read its immediate value. This method does
   * <em>not</em> block to wait for asynchronously computed values.
   *
   * The preferred method of getting values from a flow is for clients to listen for changes on the
   * flow and retrieve values (using {@link Flow#get()}) in response to such changes. Thus, uses of
   * this method should be rare, and mostly restricted to cases where the value of a flow needs to
   * be retrieved in response to external events.
   */
  public static final @Nullable <T> T getMomentary(Flow<T> flow) {
    flow.addChangeListener(DUMMY_LISTENER);
    try {
      return flow.get();
    } finally {
      flow.removeChangeListener(DUMMY_LISTENER);
    }
  }
}
