// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.google.common.base.Preconditions;
import java.util.Set;
import javax.annotation.Nullable;

/**
 * A value derived from a {@code Flow} which is itself produced by another {@code Flow}. The output
 * value of the {@code IndirectFlow} is non-null iff the outer {@code Flow} has produced a non-null
 * {@code Flow} which in turn has produced a non-null output value. The currently exposed interface
 * only allows inputs from a value created in the {@code IndirectFlow} constructor using an
 * {@code IndirectProcessor}. This interface eliminates cases where clients would have to deal with
 * multiple levels of generic type arguments.
 */
final class IndirectFlow<O> extends Flow<O> {
  private final Flow<Flow<O>> outerFlow;

  /** Null iff there are no listeners registered on the {@code IndirectFlow}. */
  private @Nullable Snapshot<Flow<O>> outerSnapshot;
  private @Nullable Snapshot<O> innerSnapshot;

  /**
   * Create an {@code IndirectFlow} based on an {@code IndirectProcessor}.
   */
  <I> IndirectFlow(Flow<I> input, IndirectProcessor<I,O> processor) {
    this(processor.isAsynchronous() ? "IndirectAsynchronous" : "IndirectSynchronous",
        Flows.newProcessed(input, processor));
  }

  /**
   * Create an {@code IndirectFlow} based on an arbitrary input {@code Flow}. Package-visible for
   * testing only.
   */
  <I> IndirectFlow(String description, Flow<Flow<O>> outerFlow) {
    super(description);
    Preconditions.checkNotNull(outerFlow);
    this.outerFlow = outerFlow;
  }

  /**
   * @return true iff the snapshot had changed
   */
  private boolean setOuterSnapshot(final @Nullable Snapshot<Flow<O>> newOuterSnapshot) {
    Preconditions.checkState(Thread.holdsLock(this));
    if (outerSnapshot == newOuterSnapshot)
      return false;
    if (outerSnapshot != null && outerSnapshot.getValue() != null)
      outerSnapshot.getValue().removeChangeListener(propagatingChangeListener);
    outerSnapshot = newOuterSnapshot;
    if (outerSnapshot != null && outerSnapshot.getValue() != null)
      outerSnapshot.getValue().addChangeListener(propagatingChangeListener);
    return true;
  }

  @Override
  protected void firstListenerAdded() {
    outerFlow.addChangeListener(propagatingChangeListener);
  }

  @Override
  protected void lastListenerRemoved() {
    outerFlow.removeChangeListener(propagatingChangeListener);
    setOuterSnapshot(null);
    innerSnapshot = null;
  }

  @Override
  protected void update() {
    Preconditions.checkState(Thread.holdsLock(this));
    final boolean outerSnapshotModified = setOuterSnapshot(outerFlow.getSnapshot());
    final Snapshot<O> newInnerSnapshot =
        (outerSnapshot == null || outerSnapshot.getValue() == null)
        ? null : outerSnapshot.getValue().getSnapshot();
    final boolean modified = outerSnapshotModified || innerSnapshot != newInnerSnapshot;
    innerSnapshot = newInnerSnapshot;
    if (modified)
      setInternal(innerSnapshot == null ? null : innerSnapshot.getValue());
  }

  @Override
  protected Set<Flow<?>> getDependencies() {
    /* No need to include the dependencies of the inner flow, since the latter are generated from
    the inputs of the outer one anyway. */
    return outerFlow.getDependencies();
  }
}
