// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import javax.annotation.Nullable;

/**
 * Abstract base class for {@code Flow} implementations that wrap some kind of observable and
 * thread-safe model object. The model value will only be retrieved once the first listener has been
 * registered on the {@code Flow}, and henceforth only when the model has signified a change, as
 * indicated via {@link #markModelUpdated()}. The output of the {@code Flow} will only change once
 * the model returns a different object (as per {@code Object.equals}).
 */
public abstract class ModelAdapterFlow<T> extends Flow<T> {
  private boolean pendingModelUpdate;

  public ModelAdapterFlow(String description) {
    super(description);
  }

  /** May be called from any thread. */
  protected abstract void addModelListener();
  /** May be called from any thread. */
  protected abstract void removeModelListener();
  /** May be called from any thread. */
  protected abstract @Nullable T getModelValue();

  /** May be called from any thread. */
  protected final void markModelUpdated() {
    final boolean oldPendingModelUpdate;
    synchronized (ModelAdapterFlow.this) {
      oldPendingModelUpdate = pendingModelUpdate;
      pendingModelUpdate = true;
    }
    /* Some models may trigger their own listeners during getModelValue(); avoid re-propagating in
    this case. */
    if (!oldPendingModelUpdate)
      propagateChanges();
  }

  @Override
  protected final void firstListenerAdded() {
    pendingModelUpdate = true;
    addModelListener();
  }

  @Override
  protected final void lastListenerRemoved() {
    removeModelListener();
    pendingModelUpdate = false;
  }

  @Override
  protected final void update() {
    if (pendingModelUpdate) {
      setInternalIfDifferent(getModelValue());
      pendingModelUpdate = false;
    }
  }
}
