// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import java.util.Objects;
import javax.annotation.Nullable;

/**
 * A {@link Flow} for which the content will be explicitly set by the client.
 */
public final class SettableFlow<T> extends Flow<T> {
  private final Flow<T> readOnly;
  private @Nullable T value;

  @SuppressWarnings("LeakingThisInConstructor")
  SettableFlow() {
    super("Settable");
    readOnly = Flows.newIdentity(this);
  }

  public Flow<T> getReadOnly() {
    return readOnly;
  }

  @Override
  protected void update() {
    setInternalIfDifferent(value);
  }

  public @Nullable T getWithoutListener() {
    return value;
  }

  /**
   * Set the value. The output will only change if the value is different from the previous value
   * per {@code Object.equals}).
   *
   * @param value the new value
   */
  public void set(@Nullable T value) {
    synchronized (this) {
      /* Test for equality here even though setInternalIfDifferent could have been used to do the
      sme, to ensure that the object returned by getWithoutListener is identicial to that served in
      the flow's output (by object identity). */
      if (!Objects.equals(this.value, value))
        this.value = value;
    }
    propagateChanges();
  }
}
