// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

/* TODO[pd]: Consider adding support for exceptions. In that case, do not permit process() to return
             null. On the other hand, the current approach might actually be more appropriate, as it
             ensures each exception is only reported once (by the processor itself). */

/**
 * Encapsulation of a potentially long-running process that transforms one value to another.
 * Implementations must be thread-safe; this is trivially achievable in the case where the class
 * keeps or accesses no state of its own.
 *
 * <p>Interruptability support is provided via {@code Thread.interrupt()}. More elaborate
 * interruption models can be supported by having the ValueProcessor maintain its own additional
 * worker thread and wait on its completion. For instance, a ValueProcessor that performs
 * interruptible SQL queries could start a separate thread for the query, wait for its completion in
 * the thread where {@link #process(Object) process} was called, and call
 * {@link java.sql.Statement#cancel() Statement.cancel()} if said wait throws an
 * {@code InterruptedException}.</p>
 *
 * @param <I> the input value type.
 * @param <O> the output value type.
 */
@ThreadSafe
public abstract class ValueProcessor<I,O> {
  /**
   * Calculate an output value from an input value. Long-running operations may safely be embodied
   * in this method if the {@code ValueProcessor} was marked as asynchronous. There is no
   * requirement that the process be purely functional; that is, it is permissible for the result to
   * depend on external state and for the process to cause side-effects if desired.
   *
   * <p>It is valid either to throw an {@code InterruptedException} or to return a value (e.g. null)
   * in case the thread is interrupted; any return value will be ignored.</p>
   *
   * @param input the input value to use
   * @return the resulting output
   */
  public abstract @Nullable O process(I input) throws InterruptedException;

  public final boolean isAsynchronous() {
    return getSynchronicity() != Synchronicity.SYNCHRONOUS;
  }

  public final boolean isInterruptible() {
    return getSynchronicity() == Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
  }

  /**
   * May be overridden by subclasses. Synchronous mode, the default, should be used if
   * {@link #process(Object) process} is non-blocking. Interruptible asynchronous mode should be
   * used if {@link #process(Object) process} may block but supports interruption via
   * {@link Thread#interrupt()}.
   */
  public Synchronicity getSynchronicity() {
    return Synchronicity.SYNCHRONOUS;
  }

  /**
   * Called to provide a hint that caches or other state implemented by the {@code ValueProcessor}
   * may be cleared. Guaranteed not to be called while {@link #process(Object)} is running in any
   * other thread. Should never block.
   */
  public void clearState() {
  }

  @Override
  public String toString() {
    return "Processor";
  }

  public static enum Synchronicity {
    SYNCHRONOUS,
    ASYNCHRONOUS,
    ASYNCHRONOUS_INTERRUPTABLE
  }
}
