// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.util;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import java.util.List;
import java.util.Map;
import javax.annotation.Nullable;
import org.openide.xml.XMLUtil;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;

public final class MyXMLUtil {
  private MyXMLUtil() { }

  public static Element findMandatoryElement(Element parent, String tagName, String namespace)
      throws SieuferdXMLException
  {
    Element ret = XMLUtil.findElement(parent, tagName, namespace);
    if (ret == null)
      throw new SieuferdXMLException("Expected <" + tagName + "> element");
    return ret;
  }

  public static boolean matchesTag(Element element, String tagName, String namespaceURI) {
    return Objects.equal(namespaceURI, element.getNamespaceURI()) &&
        tagName.equals(element.getTagName());
  }

  public static @Nullable String getTextNullable(Element parent) throws SieuferdXMLException {
    String ret = XMLUtil.findText(parent);
    return (ret == null || ret.isEmpty()) ? null : ret;
  }

  public static String getText(Element parent) throws SieuferdXMLException {
    String ret = getTextNullable(parent);
    if (ret == null) {
      throw new SieuferdXMLException(
          "Expected non-empty string in <" + parent.getTagName() + "> element");
    }
    return ret;
  }

  public static String getNonEmptyAttribute(Element element, String attributeName)
      throws SieuferdXMLException
  {
    String ret = element.getAttribute(attributeName);
    if (ret.isEmpty()) {
      throw new SieuferdXMLException("Expected non-empty attribute '" + attributeName +
          "' in <" + element.getTagName() + "> element");
    }
    return ret;
  }

  /** Attribute must be present, but may be empty. */
  public static String getMandatoryAttribute(Element element, String attributeName)
      throws SieuferdXMLException
  {
    Attr attr = element.getAttributeNode(attributeName);
    if (attr == null) {
      throw new SieuferdXMLException("Expected attribute '" + attributeName +
          "' in <" + element.getTagName() + "> element");
    }
    return attr.getValue();
  }

  /** Attribute must be present, but may be empty, in which case null is returned. */
  public static String getMandatoryAttributeOrNull(Element element, String attributeName)
      throws SieuferdXMLException
  {
    String ret = getMandatoryAttribute(element, attributeName);
    return ret.isEmpty() ? null : ret;
  }

  public static <T> T getMandatoryMappedAttribute(
      Element element, String attributeName, Map<String,T> map)
      throws SieuferdXMLException
  {
    String input = getMandatoryAttribute(element, attributeName);
    if (!map.containsKey(input)) {
      throw new SieuferdXMLException("Expected one of " + map.keySet() + " for attribute '" +
          attributeName + "' in element <" + element.getTagName() + ">");
    }
    T ret = map.get(input);
    Preconditions.checkNotNull(ret);
    return ret;
  }

  private static final ImmutableMap<String,Boolean> BOOLEAN_MAP = ImmutableMap.of(
    "true", true,
    "false", false);

  public static boolean getMandatoryBooleanAttribute(Element element, String attributeName)
      throws SieuferdXMLException
  {
    return getMandatoryMappedAttribute(element, attributeName, BOOLEAN_MAP);
  }

  public static List<Element> findChildElementsByTagName(
      Element parent, String tagName, String namespace)
  {
    List<Element> ret = Lists.newArrayList();
    for (Element element : XMLUtil.findSubElements(parent)) {
      if (element.getTagName().equals(tagName) && element.getNamespaceURI().equals(namespace))
        ret.add(element);
    }
    return ret;
  }
}
