// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.util;

import com.google.common.base.Preconditions;
import com.google.common.escape.Escaper;
import com.google.common.escape.Escapers;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import javax.annotation.Nullable;

public final class StringUtil {
  /* TODO[pd]: Handle more cases here. Preferably just replace with a library function, e.g.
               https://github.com/google/guava/issues/1620 (not yet supported in Guava). */
  private static final Escaper JAVA_STRING_ESCAPER = Escapers.builder()
      .addEscape('\\', "\\\\")
      .addEscape('\"', "\\\"")
      .addEscape('\n', "\\n")
      .addEscape('\t', "\\t")
      .build();

  private StringUtil() { }

  /**
   * Return a Java-style string literal. Not currently foolproof; mostly useful for test code.
   */
  public static String quoteJavaLike(String s) {
    return "\"" + JAVA_STRING_ESCAPER.escape(s) + "\"";
  }

  public static String debugQuote(String s) {
    return quoteJavaLike(s);
  }

  // TODO[cc]: Consider defining a special string type that is guaranteed to be trimmed.
  public static String trimmedAndEmptyIfNull(@Nullable String s) {
    return s == null ? "" : s.trim();
  }

  public static String urlEncode(String stringToEncode) {
    try {
      return URLEncoder.encode(stringToEncode, "UTF-8");
    } catch (UnsupportedEncodingException e) {
      throw new AssertionError(e);
    }
  }

  public static String urlDecode(String stringToDecode) {
    try {
      return URLDecoder.decode(stringToDecode, "UTF-8");
    } catch (UnsupportedEncodingException e) {
      throw new AssertionError(e);
    }
  }

  public static String titleCaseEnglish(String str) {
    StringBuilder ret = new StringBuilder(str.length());
    boolean capitalizeNext = true;
    for (int i = 0; i < str.length(); i++) {
      final char c = str.charAt(i);
      ret.append(capitalizeNext ? Character.toUpperCase(c) : Character.toLowerCase(c));
      capitalizeNext = Character.isSpaceChar(c);
    }
    return ret.toString();
  }

  // TODO[cc]: Replace this with String.join once I switch to Java 8.
  public static String join(String separator, Iterable<String> strings) {
    StringBuilder ret = new StringBuilder();
    boolean first = true;
    for (String string : strings) {
      if (first) {
        first = false;
      } else {
        ret.append(separator);
      }
      ret.append(string);
    }
    return ret.toString();
  }

  public static boolean isEmptyOrSpaceOnly(String str) {
    final int N = str.length();
    for (int i = 0; i < N;) {
      final int codepoint = str.codePointAt(i);
      if (!Character.isSpaceChar(codepoint))
        return false;
      i += Character.charCount(codepoint);
    }
    return true;
  }

  public static final void checkArgumentTrimmed(String str) {
    Preconditions.checkNotNull(str);
    Preconditions.checkArgument(str.trim().equals(str));
  }
}
