// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.google.common.base.Optional;
import com.google.common.collect.Lists;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.junit.Assert;

class AnticipatingChangeListener<T> implements ChangeListener {
  private final Flow<T> flow;
  private final List<T> allAnticipatedValues = Lists.newArrayList();
  private final BlockingQueue<Optional<T>> queue = new LinkedBlockingQueue<>();
  private Object previousValue = null;
  private boolean ignoreExtraNulls = false;

  AnticipatingChangeListener(Flow<T> flow) {
    this.flow = flow;
  }

  public synchronized void setIgnoreExtraNulls(boolean ignoreExtraNulls) {
    this.ignoreExtraNulls = ignoreExtraNulls;
  }

  public synchronized void anticipate(T anticipatedValue) {
    try {
      queue.put(Optional.fromNullable(anticipatedValue));
      allAnticipatedValues.add(anticipatedValue);
    } catch (InterruptedException e) {
      throw new AssertionError(e);
    }
  }

  public synchronized void waitForChanges() {
    while (!queue.isEmpty()) {
      try {
        wait();
      } catch (InterruptedException e) {
        throw new AssertionError(e);
      }
    }
  }

  @Override
  public synchronized void stateChanged(ChangeEvent evt) {
    Assert.assertEquals(flow, evt.getSource());
    final T actual = flow.get();
    Optional<T> peek = queue.peek();
    if (ignoreExtraNulls && actual == null && (peek == null || peek.isPresent()))
      return;
    final boolean noChange = (previousValue == actual);
    previousValue = actual;
    if (noChange)
      return;
    final T anticipated;
    try {
      anticipated = queue.remove().orNull();
    } catch (NoSuchElementException e) {
      throw new NoSuchElementException("Got unanticipated change event with value " + actual +
          " (anticipated sequence was: " + allAnticipatedValues + ")");
    }
    Assert.assertEquals(anticipated, actual);
    notifyAll();
  }
}
