// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.flow;

import com.sieuferd.util.Blocker;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Objects;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.junit.Test;
import static org.junit.Assert.*;

public class FlowTest {
  private static class TestComposite {
    final int a, b;

    TestComposite(int a, int b) {
      this.a = a;
      this.b = b;
    }

    @Override
    public String toString() {
      return "TestComposite(" + a + ", " + b + ")";
    }
  }

  private static class DummyChangeListener implements ChangeListener {
    @Override
    public void stateChanged(ChangeEvent e) {
    }
  }

  private static class TestCompositeFactory extends CompositeFactory<TestComposite> {
    @Override
    protected TestComposite construct(Object[] args) {
      return new TestComposite((Integer) args[0], (Integer) args[1]);
    }
  }

  private static class InterruptibleProcessor extends ValueProcessor<Integer,Integer> {
    public static final int BLOCKING_VALUE = 100;
    private final Blocker blocker = new Blocker();

    InterruptibleProcessor() {
      super();
    }

    @Override
    public ValueProcessor.Synchronicity getSynchronicity() {
      return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
    }

    /** Wait until the processor is actually in the blocking state (meaning the RequestProcessor in
    ValueProcessor has actually started the task). */
    public void waitForBlock() {
      blocker.blockUntilReleased();
    }

    @Override
    @SuppressWarnings("SleepWhileInLoop")
    public Integer process(Integer input) throws InterruptedException {
      if (Objects.equals(input, BLOCKING_VALUE)) {
        blocker.releaseBlockingThread();
        while (true) {
          Thread.sleep(1000);
        }
      }
      return 2 * input;
    }
  }

  private static class IdentityProcessor<T> extends ValueProcessor<T,T> {
    private final boolean asynchronous;

    IdentityProcessor(Class<T> valueClass, boolean asynchronous) {
      super();
      this.asynchronous = asynchronous;
    }

    @Override
    public ValueProcessor.Synchronicity getSynchronicity() {
      return asynchronous ? Synchronicity.ASYNCHRONOUS : Synchronicity.SYNCHRONOUS;
    }

    @Override
    public T process(T input) throws InterruptedException {
      return input;
    }
  }

  private static class SumProcessor extends ValueProcessor<TestComposite,Integer> {
    private final boolean asynchronous;

    SumProcessor(boolean asynchronous) {
      super();
      this.asynchronous = asynchronous;
    }

    @Override
    public ValueProcessor.Synchronicity getSynchronicity() {
      return asynchronous ? Synchronicity.ASYNCHRONOUS : Synchronicity.SYNCHRONOUS;
    }

    @Override
    public Integer process(TestComposite input) throws InterruptedException {
      return input.a + input.b;
    }
  }

  /* For deterministic testing behavior, first specify all anticipated values, then mutate inputs,
  then wait for changes, then assert values. */

  @Test
  public void testSettableValueInitiallySet() {
    SettableFlow<Integer> a = Flows.newSettable();
    AnticipatingChangeListener<Integer> listener = new AnticipatingChangeListener<>(a);
    a.set(5);
    a.addChangeListener(listener);
    assertEquals(5, (long) a.get());
    a.removeChangeListener(listener);
    a.addChangeListener(listener);
    assertEquals(5, (long) a.get());
    listener.anticipate(8);
    a.set(8);
    listener.waitForChanges();
    assertEquals(8, (long) a.get());
    a.removeChangeListener(listener);
  }

  private static <T> Flow<T> wrapInComposite(Flow<T> inputFlow) {
    return new CompositeFlow<>(new Flow<?>[] { inputFlow }, new CompositeFactory<T>() {
      @Override
      @SuppressWarnings("unchecked")
      protected T construct(Object[] args) {
        return (T) args[0];
      }
    });
  }

  @Test
  public void testExponentialCompositeBug() {
    testSettableValue(true);
  }

  @Test
  public void testSettableValue() {
    testSettableValue(false);
  }

  private void testSettableValue(boolean testExponentialComposite) {
    final SettableFlow<Integer> sf = Flows.newSettable();
    Flow<Integer> a = sf;
    if (testExponentialComposite) {
      for (int i = 0; i < 1000; i++)
        a = wrapInComposite(a);
    }
    AnticipatingChangeListener<Integer> listener = new AnticipatingChangeListener<>(a);
    a.addChangeListener(listener);
    assertEquals(null, a.get());
    listener.anticipate(5);
    listener.anticipate(6);
    listener.anticipate(7);
    sf.set(5);
    assertEquals(5, (int) a.get());
    sf.set(6);
    assertEquals(6, (int) a.get());
    sf.set(7);
    assertEquals(7, (int) a.get());
    listener.waitForChanges();
    assertEquals(7, (int) a.get());
    a.removeChangeListener(listener);
  }

  @Test
  public void testConsistentPropagation() {
    final SettableFlow<Integer> sf = Flows.newSettable();
    Flow<Integer> m0 = wrapInComposite(sf);
    Flow<Integer> m1 = wrapInComposite(sf);
    Flow<Integer> m2 = wrapInComposite(wrapInComposite(m1));
    Flow<Integer> a = new CompositeFlow<>(new Flow<?>[] { m0, m2, m1 },
        new CompositeFactory<Integer>()
    {
      @Override
      @SuppressWarnings("unchecked")
      protected Integer construct(Object[] args) {
        int m0v = (Integer) args[0];
        int m1v = (Integer) args[1];
        int m2v = (Integer) args[2];
        if (m0v != m1v || m1v != m2v)
          throw new AssertionError("Got inconsistent inputs " + Arrays.toString(args));
        return m0v;
      }
    });;
    AnticipatingChangeListener<Integer> listener = new AnticipatingChangeListener<>(a);
    a.addChangeListener(listener);
    assertEquals(null, a.get());
    listener.anticipate(5);
    listener.anticipate(6);
    listener.anticipate(7);
    sf.set(5);
    assertEquals(5, (int) a.get());
    sf.set(6);
    assertEquals(6, (int) a.get());
    sf.set(7);
    assertEquals(7, (int) a.get());
    listener.waitForChanges();
    assertEquals(7, (int) a.get());
    a.removeChangeListener(listener);
  }


  @Test
  public void testSynchronous() {
    testFlow(false, false);
  }

  @Test
  public void testAsyncronous() {
    // Run many times to maximize chances of exercising race conditions.
    for (int i = 0; i < 100; i++) {
      testFlow(false, true);
      testFlow(true, true);
      testFlow(true, false);
    }
  }

  private void testFlow(boolean asyncIdent, boolean asyncSum) {
    SettableFlow<Integer> a = Flows.newSettable();
    SettableFlow<Integer> b = Flows.newSettable();
    {
      ChangeListener dummyListener = new DummyChangeListener();
      a.addChangeListener(dummyListener);
      b.addChangeListener(dummyListener);
      assertEquals(null, a.get());
      assertEquals(null, b.get());
      a.removeChangeListener(dummyListener);
      b.removeChangeListener(dummyListener);
    }
    Flow<Integer> a2 =
        Flows.newProcessed(a, new IdentityProcessor<>(Integer.class, asyncIdent));
    Flow<TestComposite> sumInput =
        Flows.newComposite(new Flow<?>[] {a, b}, new TestCompositeFactory());
    Flow<Integer> sum = Flows.newProcessed(sumInput, new SumProcessor(asyncSum));

    AnticipatingChangeListener<Integer> aListener = new AnticipatingChangeListener<>(a);
    AnticipatingChangeListener<Integer> bListener = new AnticipatingChangeListener<>(b);
    AnticipatingChangeListener<Integer> a2Listener = new AnticipatingChangeListener<>(a2);
    AnticipatingChangeListener<Integer> sumListener = new AnticipatingChangeListener<>(sum);

    a2Listener.setIgnoreExtraNulls(asyncIdent);
    sumListener.setIgnoreExtraNulls(asyncSum);

    a.addChangeListener(aListener);
    b.addChangeListener(bListener);
    a2.addChangeListener(a2Listener);
    sum.addChangeListener(sumListener);

    assertEquals(null, a.get());
    assertEquals(null, b.get());
    assertEquals(null, a2.get());
    assertEquals(null, sum.get());

    aListener.anticipate(3);
    a2Listener.anticipate(3);
    a.set(3);
    aListener.waitForChanges();
    a2Listener.waitForChanges();
    assertEquals(3, (long) a.get());
    assertEquals(null, b.get());
    assertEquals(3, (long) a2.get());
    assertEquals(null, sum.get());

    bListener.anticipate(5);
    sumListener.anticipate(8);
    b.set(5);
    sumListener.waitForChanges();
    bListener.waitForChanges();
    assertEquals(3, (long) a.get());
    assertEquals(5, (long) b.get());
    assertEquals(3, (long) a2.get());
    assertEquals(8, (long) sum.get());

    aListener.anticipate(null);
    a2Listener.anticipate(null);
    sumListener.anticipate(null);
    a.set(null);
    aListener.waitForChanges();
    a2Listener.waitForChanges();
    /* TODO[pd]: A hang was once observed here, after which I made an adjustment to FlowTest. See if
                 it happens again. */
    sumListener.waitForChanges();
    assertEquals(null, a.get());
    assertEquals(5, (long) b.get());
    assertEquals(null, a2.get());
    assertEquals(null, sum.get());

    bListener.anticipate(null);
    b.set(null);
    bListener.waitForChanges();

    aListener.anticipate(7);
    bListener.anticipate(9);
    a2Listener.anticipate(7);
    sumListener.anticipate(16);
    a.set(7);
    b.set(9);
    aListener.waitForChanges();
    bListener.waitForChanges();
    a2Listener.waitForChanges();
    sumListener.waitForChanges();
    assertEquals(7, (long) a.get());
    assertEquals(9, (long) b.get());
    assertEquals(7, (long) a2.get());
    assertEquals(16, (long) sum.get());

    a2.removeChangeListener(a2Listener);
    //assertEquals(null, a2.get());
    aListener.anticipate(42);
    sumListener.anticipate(51);
    a.set(42);
    aListener.waitForChanges();
    sumListener.waitForChanges();
    assertEquals(42, (long) a.get());
    assertEquals(9, (long) b.get());
    assertEquals(51, (long) sum.get());

    a.removeChangeListener(aListener);
    b.removeChangeListener(bListener);
    sum.removeChangeListener(sumListener);

    AsynchronousFlow.awaitTermination();
  }

  @Test
  public void testInterruptible() {
    for (int i = 0; i < 100; i++)
      testInterruptibleImpl();
  }

  private void testInterruptibleImpl() {
    SettableFlow<Integer> a = Flows.newSettable();
    InterruptibleProcessor ip = new InterruptibleProcessor();
    Flow<Integer> a2 = Flows.newProcessed(a, ip);
    AnticipatingChangeListener<Integer> aListener = new AnticipatingChangeListener<>(a);
    AnticipatingChangeListener<Integer> a2Listener = new AnticipatingChangeListener<>(a2);
    a2Listener.setIgnoreExtraNulls(true);
    a.addChangeListener(aListener);
    a2.addChangeListener(a2Listener);

    aListener.anticipate(5);
    a2Listener.anticipate(10);
    a.set(5);
    aListener.waitForChanges();
    a2Listener.waitForChanges();

    aListener.anticipate(8);
    a2Listener.anticipate(16);
    a.set(8);
    aListener.waitForChanges();
    a2Listener.waitForChanges();

    aListener.anticipate(InterruptibleProcessor.BLOCKING_VALUE);
    a2Listener.anticipate(null);
    a.set(InterruptibleProcessor.BLOCKING_VALUE);
    aListener.waitForChanges();
    a2Listener.waitForChanges();

    /* To ensure that we are actually testing interruptability and not just the RequestProcessor's
    ability to cancel a not-yet-started task, wait for the blocking loop in the processor to
    actually have started. */
    ip.waitForBlock();

    aListener.anticipate(12);
    a2Listener.anticipate(24);
    a.set(12);
    aListener.waitForChanges();
    a2Listener.waitForChanges();

    a.removeChangeListener(aListener);
    a2.removeChangeListener(a2Listener);

    AsynchronousFlow.awaitTermination();
  }

  @Test
  public void testSimpleCorrectTypes() {
    SettableFlow<Number> a = Flows.newSettable();
    ChangeListener listener = new DummyChangeListener();
    a.addChangeListener(listener);
    a.set(null);
    assertEquals(null, a.get());
    a.set(BigDecimal.ONE);
    assertEquals(BigDecimal.ONE, a.get());
    a.set(5.0);
    assertEquals(5.0, a.get());
    a.removeChangeListener(listener);
  }

  //@Test(expected = FlowTypeException.class)
  // Note that the suppressed warning would actually have detected the error in this case.
  @SuppressWarnings("unchecked")
  public void testSimpleIncorrectType() {
    SettableFlow<Number> a = Flows.newSettable();
    a.set(BigDecimal.ONE);
    ChangeListener listener = new DummyChangeListener();
    a.addChangeListener(listener);
    assertEquals(BigDecimal.ONE, a.get());
    a.removeChangeListener(listener);
    // Would have failed back when Flow's did their own type checking. No longer will.
    ((SettableFlow) a).set("badvalue");
  }

  @Test
  public void testCompositeCorrectTypes() {
    final SettableFlow<BigDecimal> a = Flows.newSettable();
    a.set(BigDecimal.ZERO);
    // Exact type as composite input.
    Flows.newComposite(new Flow<?>[] { a }, new CompositeFactory<String>() {
      @Override
      protected String construct(Object[] args) {
        throw new UnsupportedOperationException();
      }
    });
    // Supertype as composite input.
    Flows.newComposite(new Flow<?>[] { a }, new CompositeFactory<String>() {
      @Override
      protected String construct(Object[] args) {
        throw new UnsupportedOperationException();
      }
    });
  }

  //@Test(expected = FlowTypeException.class)
  public void testCompositeIncorrectType() {
    SettableFlow<Number> a = Flows.newSettable();
    // Would have failed back when Flow's did their own type checking. No longer will.
    // Number is not assignable to BigDecimal.
    Flows.newComposite(new Flow<?>[] { a }, new CompositeFactory<String>() {
      @Override
      protected String construct(Object[] args) {
        throw new UnsupportedOperationException();
      }
    });
  }

  @Test
  public void testIndirectExactType() {
    SettableFlow<String> sv = Flows.newSettable();
    Flow<Number> iv = Flows.newIndirect(sv,
        input -> { throw new UnsupportedOperationException(); });
    Flows.newProcessed(iv, input -> { throw new UnsupportedOperationException(); });
  }

  @Test
  public void testIndirectValue() {
    SettableFlow<Number> inner = Flows.newSettable();
    SettableFlow<Flow<Number>> outer = new SettableFlow<>();
    Flow<Number> indirect = new IndirectFlow<>("TestIndirect", outer);
    ChangeListener innerL = new DummyChangeListener();
    ChangeListener outerL = new DummyChangeListener();
    AnticipatingChangeListener<Number> indirectL = new AnticipatingChangeListener<>(indirect);

    inner.addChangeListener(innerL);
    outer.addChangeListener(outerL);
    indirect.addChangeListener(indirectL);

    assertEquals(null, inner.get());
    assertEquals(null, outer.get());
    assertEquals(null, indirect.get());

    inner.set(5);
    assertEquals(null, indirect.get());

    indirectL.anticipate(5);
    outer.set(inner);
    assertEquals(5, indirect.get());

    indirectL.anticipate(6);
    inner.set(6);
    assertEquals(6, indirect.get());

    indirectL.anticipate(null);
    inner.set(null);
    assertEquals(null, indirect.get());

    indirectL.anticipate(7);
    inner.set(7);
    assertEquals(7, indirect.get());

    indirectL.anticipate(null);
    outer.set(null);
    assertEquals(null, indirect.get());

    indirectL.anticipate(7);
    outer.set(inner);
    assertEquals(7, indirect.get());

    indirectL.anticipate(null);
    outer.set(null);
    assertEquals(null, indirect.get());

    indirectL.anticipate(8);
    inner.set(8);
    outer.set(inner);
    assertEquals(8, indirect.get());

    SettableFlow<Number> inner2 = Flows.newSettable();
    inner2.set(9);
    indirectL.anticipate(9);
    outer.set(inner2);
    assertEquals(9, indirect.get());

    indirectL.anticipate(10);
    inner.set(10);
    outer.set(inner);
    assertEquals(10, indirect.get());

    inner.removeChangeListener(innerL);
    outer.removeChangeListener(outerL);
    indirect.removeChangeListener(indirectL);
  }
}
