// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.viewset;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.ObservableView;
import java.util.Set;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.undo.UndoableEdit;

// TODO[cc]: Find a better name.

/**
 * A {@link com.sieuferd.datamodel.View} implementation that permits modifications only through a
 * dedicated interface. Undo/Redo support is provided through the same convention as for
 * {@link javax.swing.text.Document}.
 */
public abstract class GatedView extends ObservableView {
  private final Object undoableEditListenersLock = new Object();
  private final Set<UndoableEditListener> undoableEditListeners = Sets.newLinkedHashSet();

  // TODO[cc]: There is no longer a need to do performWriteOperation everywhere; avoid it.
  // TODO[cc]: Consider removing the separate view transaction concept.
  /**
   * @param operationPresentationName will have any mnemonics automatically removed
   */
  public abstract void mutate(
      String operationPresentationName, boolean incrementUndoVersion, ViewOperation operation);

  public final void addUndoableEditListener(UndoableEditListener listener) {
    synchronized (undoableEditListenersLock) {
      undoableEditListeners.add(listener);
    }
  }

  public final void removeUndoableEditListener(UndoableEditListener listener) {
    synchronized (undoableEditListenersLock) {
      undoableEditListeners.remove(listener);
    }
  }

  protected void fireUndoableEditHappened(UndoableEdit edit) {
    Preconditions.checkNotNull(edit);
    Preconditions.checkState(!Thread.holdsLock(this));
    final Set<UndoableEditListener> listenersSnapshot;
    synchronized (undoableEditListenersLock) {
      listenersSnapshot = ImmutableSet.copyOf(undoableEditListeners);
    }
    final UndoableEditEvent evt = new UndoableEditEvent(this, edit);
    for (UndoableEditListener listener : listenersSnapshot)
      listener.undoableEditHappened(evt);
  }

  public interface ViewOperation {
    /**
     * Perform a mutating operation on {@code view}. This method must not block. No guarantees are
     * made as to what thread this method may be called on. Implementations should not hold on to
     * {@code view} after the method has returned. The method could in theory be called more than
     * once, for instance if the operation needs to be repeated on a newer version of the view
     * retrieved from a remote server as part of conflict resolution during collaborative editing.
     * Operations should be deterministic, and should always succeed, even if relevant paths no
     * longer exist.
     */
    public void performOperation(MutableView view);
  }
}
