// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.viewset;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.flow.Flow;
import java.util.Set;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.uiutil.flow.UIFlows;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

@ThreadSafe
public class SieuferdProjectConfiguration {
  public static final String PROP_PROJECT_NAME = "projectName";
  public static final String PROP_DATABASE_CONNECTION_NAME = "databaseConnectionName";
  public static final String PROP_MANAGED_DATABASES = "managedDatabases";
  public static final ImmutableSet<String> PRIMARY_VIEW_SET_PROPERTIES =
      ImmutableSet.of(PROP_DATABASE_CONNECTION_NAME, PROP_MANAGED_DATABASES);
  private final PropertyChangeSupport pcs = new PropertyChangeSupport(this);
  private @Nullable String databaseConnectionName;
  private String projectName = "";
  private ImmutableSet<SQLDatabase> managedDatabases = ImmutableSet.of();

  public SieuferdProjectConfiguration() {
  }

  public void copyFrom(SieuferdProjectConfiguration other) {
    setProjectName(other.getProjectName());
    setDatabaseConnectionName(other.getDatabaseConnectionName());
    setManagedDatabases(other.getManagedDatabases());
  }

  public synchronized void addPropertyChangeListener(PropertyChangeListener listener) {
    pcs.addPropertyChangeListener(listener);
  }

  public synchronized void removePropertyChangeListener(PropertyChangeListener listener) {
    pcs.removePropertyChangeListener(listener);
  }

  public synchronized String getProjectName() {
    return projectName;
  }

  public void setProjectName(String projectName) {
    Preconditions.checkNotNull(projectName);
    final String oldValue;
    synchronized (this) {
      oldValue = this.projectName;
      this.projectName = projectName;
    }
    pcs.firePropertyChange(PROP_PROJECT_NAME, oldValue, projectName);
  }

  public synchronized @Nullable String getDatabaseConnectionName() {
    return databaseConnectionName;
  }

  public void setDatabaseConnectionName(@Nullable String databaseConnectionName) {
    final String oldValue;
    synchronized (this) {
      oldValue = this.databaseConnectionName;
      this.databaseConnectionName = databaseConnectionName;
    }
    if (oldValue != null || databaseConnectionName != null)
      pcs.firePropertyChange(PROP_DATABASE_CONNECTION_NAME, oldValue, databaseConnectionName);
  }

  public synchronized Set<SQLDatabase> getManagedDatabases() {
    return managedDatabases;
  }

  public void setManagedDatabases(Set<SQLDatabase> managedDatabases) {
    Preconditions.checkNotNull(managedDatabases);
    final ImmutableSet<SQLDatabase> oldValue;
    final ImmutableSet<SQLDatabase> newValue = ImmutableSet.copyOf(managedDatabases);
    synchronized (this) {
      oldValue = this.managedDatabases;
      this.managedDatabases = newValue;
    }
    pcs.firePropertyChange(PROP_MANAGED_DATABASES, oldValue, newValue);
  }

  // TODO[cc]: Consider just storing this in a SettableFlow instead.
  public Flow<String> getDatabaseConnectionNameFlow() {
    return UIFlows.fromPropertyBean(PROP_DATABASE_CONNECTION_NAME,
        listener -> { SieuferdProjectConfiguration.this.addPropertyChangeListener(listener); },
        listener -> { SieuferdProjectConfiguration.this.removePropertyChangeListener(listener); },
        () -> SieuferdProjectConfiguration.this.getDatabaseConnectionName());
  }
}
