// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.viewset;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.List;
import javax.xml.XMLConstants;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import com.sieuferd.datamodel.Path;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.util.MyXMLUtil;
import com.sieuferd.util.SieuferdXMLException;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.StringUtil;
import java.io.OutputStream;
import java.util.Map;
import java.util.Set;
import org.netbeans.api.annotations.common.StaticResource;
import org.openide.xml.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public final class SieuferdProjectXML {
  @StaticResource
  private static final String RESOURCE_PROJECT_SCHEMA =
      "com/sieuferd/viewset/resources/sieuferdproject.xsd";
  @StaticResource
  private static final String RESOURCE_VIEW_SCHEMA =
      "com/sieuferd/viewset/resources/sieuferdview.xsd";
  private static final String PROJECT_NS = "xmlns://sieuferd.com/sieuferdproject";
  private static final String VIEW_NS = "xmlns://sieuferd.com/sieuferdview";

  private SieuferdProjectXML() { }

  public static SieuferdProjectConfiguration readSieuferdProjectXML(
      InputStream inputStream, String systemID)
      throws IOException, SieuferdXMLException
  {
    return parseSieuferdProjectXML(
        readValidatedDocument(inputStream, systemID, RESOURCE_PROJECT_SCHEMA));
  }

  public static View readSieuferdViewXML(
      InputStream inputStream, String systemID, Set<ViewProperty<?>> recognizedProperties)
      throws IOException, SieuferdXMLException
  {
    return parseSieuferdViewXML(
        readValidatedDocument(inputStream, systemID, RESOURCE_VIEW_SCHEMA), recognizedProperties);
  }

  private static Document readValidatedDocument(
      InputStream inputStream, String systemID, String schemaResource)
      throws IOException, SieuferdXMLException
  {
    Preconditions.checkNotNull(inputStream);
    try (InputStream is = new BufferedInputStream(inputStream)) {
      InputSource src = new InputSource(is);
      src.setSystemId(systemID);
      final Schema schema;
      URL schemaURL = SieuferdProjectXML.class.getClassLoader().getResource(schemaResource);
      if (schemaURL == null)
        throw new IOException("Could not find XML schema resource");
      try {
        schema = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI).newSchema(schemaURL);
      } catch (SAXException e) {
        throw new SieuferdXMLException("Error parsing XML schema", e);
      }
      try {
        Document document = XMLUtil.parse(src, false, true, XMLUtil.defaultErrorHandler(), null);
        XMLUtil.validate(document.getDocumentElement(), schema);
        return document;
      } catch (SAXException e) {
        throw new SieuferdXMLException("Error parsing Sieuferd XML document at " + systemID, e);
      }
    }
  }

  private static View parseSieuferdViewXML(
      Document document, Set<ViewProperty<?>> recognizedProperties) throws SieuferdXMLException
  {
    Element rootElement = document.getDocumentElement();
    if (!VIEW_NS.equals(rootElement.getNamespaceURI()))
      throw new SieuferdXMLException("Incorrect namespace");
    if (!"sieuferdview".equals(rootElement.getTagName()))
      throw new SieuferdXMLException("Incorrect root tag name");
    final ImmutableMap<String,ViewProperty<?>> propMap;
    {
      Map<String,ViewProperty<?>> toPropMap = Maps.newLinkedHashMap();
      for (ViewProperty<?> prop : recognizedProperties)
        MapUtil.putChecked(toPropMap, prop.getStorageName(), prop);
      propMap = ImmutableMap.copyOf(toPropMap);
    }
    Element rootRelation = MyXMLUtil.findMandatoryElement(rootElement, "relation", VIEW_NS);
    MutableView ret = Views.newMutable();
    parseViewPath(Strepco.createNormalized(ret), ret, rootRelation, Path.ROOT, propMap, true);
    // Rerun with a new Strepco, since STREPCO_PROPERTIES will now have been populated.
    parseViewPath(Strepco.createNormalized(ret), ret, rootRelation, Path.ROOT, propMap, false);
    return ImmutableView.copyOf(ret);
  }

  private static <T> void setPropertyFromElement(
      Strepco strepco, MutableView view, Path path, ViewProperty<T> prop, Element element)
      throws SieuferdXMLException
  {
    final T value;
    String str = MyXMLUtil.getMandatoryAttribute(element, "value");
    try {
      value = prop.getType().stringToValue(strepco, str);
    } catch (IllegalArgumentException e) {
      throw new SieuferdXMLException("Malformed value \"" + str + "\" for property " + prop, e);
    } catch (UnsupportedOperationException e) {
      throw new SieuferdXMLException(
          "Property type for property " + prop + " does not support deserialization");
    }
    if (view.defines(prop, path)) {
      throw new SieuferdXMLException("Property " + prop + " already defined at path " +
          path.toString(view));
    }
    view.set(prop, path, value);
  }

  private static void parseViewPath(Strepco strepco, MutableView view, Element element, Path path,
      Map<String,ViewProperty<?>> propMap, boolean firstPass)
      throws SieuferdXMLException
  {
    for (Element propElement : MyXMLUtil.findChildElementsByTagName(element, "property", VIEW_NS)) {
      String key = MyXMLUtil.getMandatoryAttribute(propElement, "key");
      ViewProperty<?> prop = propMap.get(key);
      if (prop == null)
        throw new SieuferdXMLException("Unrecognized property " + key);
      if (firstPass == Strepco.STREPCO_PROPERTIES.contains(prop))
        setPropertyFromElement(strepco, view, path, prop, propElement);
    }
    if (path.isRelation()) {
      final List<Path> paths =
          firstPass ? null : ImmutableList.copyOf(view.getChildPaths(path.getRelation()));
      int pathIndex = 0;
      for (Element childElement : XMLUtil.findSubElements(element)) {
        final boolean relation;
        if (MyXMLUtil.matchesTag(childElement, "relation", VIEW_NS)) {
          relation = true;
        } else if (MyXMLUtil.matchesTag(childElement, "primitive", VIEW_NS)) {
          relation = false;
        } else {
          continue;
        }
        final Path childPath = (paths == null)
            ? view.addPath(path.getRelation(), relation)
            : paths.get(pathIndex++);
        parseViewPath(strepco, view, childElement, childPath, propMap, firstPass);
      }
    }
  }

  public static void writeSieuferdProjectXML(
      SieuferdProjectConfiguration configuration, OutputStream outputStream) throws IOException
  {
    XMLUtil.write(generateSieuferdProjectXML(configuration), outputStream, "UTF-8");
  }

  public static void writeSieuferdViewXML(
      View view, ViewDefaults defaults, OutputStream outputStream) throws IOException
  {
    XMLUtil.write(generateSieuferdViewXML(view, defaults), outputStream, "UTF-8");
  }

  /* TODO[cc]: Since several XMLUtil methods throw an IllegalArgumentException, consider doing the
               same in these methods instead of using SieuferdXMLException. */
  private static SieuferdProjectConfiguration parseSieuferdProjectXML(Document document)
      throws SieuferdXMLException
  {
    Element rootElement = document.getDocumentElement();
    if (!PROJECT_NS.equals(rootElement.getNamespaceURI()))
      throw new SieuferdXMLException("Incorrect namespace");
    if (!"sieuferdproject".equals(rootElement.getTagName()))
      throw new SieuferdXMLException("Incorrect root tag name");
    final String projectName = MyXMLUtil.getText(
        MyXMLUtil.findMandatoryElement(rootElement, "name", PROJECT_NS));
    SieuferdProjectConfiguration ret = new SieuferdProjectConfiguration();
    ret.setProjectName(projectName);
    ret.setDatabaseConnectionName(MyXMLUtil.getTextNullable(MyXMLUtil.findMandatoryElement(
        rootElement, "connection", PROJECT_NS)));
    Set<SQLDatabase> databases = Sets.newLinkedHashSet();
    List<Element> databaseElements = MyXMLUtil.findChildElementsByTagName(
        MyXMLUtil.findMandatoryElement(rootElement, "managed_databases", PROJECT_NS),
        "database", PROJECT_NS);
    for (Element databaseElement : databaseElements) {
      String primaryNamePart = StringUtil.trimmedAndEmptyIfNull(
          MyXMLUtil.getMandatoryAttributeOrNull(databaseElement, "name_primary"));
      String secondaryNamePart = StringUtil.trimmedAndEmptyIfNull(
          MyXMLUtil.getMandatoryAttributeOrNull(databaseElement, "name_secondary"));
      if (primaryNamePart.isEmpty() && !secondaryNamePart.isEmpty()) {
        throw new SieuferdXMLException(
            "No primary name given for database '" + secondaryNamePart + "'");
      }
      databases.add(new SQLDatabase(primaryNamePart, secondaryNamePart));
    }
    ret.setManagedDatabases(databases);
    return ret;
  }

  private static Document generateSieuferdProjectXML(SieuferdProjectConfiguration configuration) {
    Document ret = XMLUtil.createDocument("sieuferdproject", PROJECT_NS, null, null);
    return ret;
  }

  private static Document generateSieuferdViewXML(final View view, final ViewDefaults defaults) {
    final Document ret = XMLUtil.createDocument("sieuferdview", VIEW_NS, null, null);
    view.performReadTransaction(() -> {
      ret.getDocumentElement().appendChild(
          generateViewPath(ret, Strepco.createNormalized(view), view, defaults, Path.ROOT));
    });
    return ret;
  }

  private static <T> void appendProperty(Element parent, Strepco strepco, View view,
      ViewDefaults defaults, ViewProperty<T> prop, Path path)
  {
    T value = view.get(prop, path);
    if (defaults.getProperties().contains(prop) && value.equals(defaults.get(prop)))
      return;
    Element ret = parent.getOwnerDocument().createElement("property");
    ret.setAttribute("key", prop.getStorageName());
    String str = prop.getType().valueToString(strepco, value);
    if (str == null) {
      throw new UnsupportedOperationException(
          "Failed to find string representation for value " + value);
    }
    ret.setAttribute("value", str);
    parent.appendChild(ret);
  }

  private static Element generateViewPath(
      Document document, Strepco strepco, View view, ViewDefaults defaults, Path path)
  {
    Element ret = document.createElementNS(VIEW_NS, path.isRelation() ? "relation" : "primitive");
    for (ViewProperty<?> prop : defaults.getSortedPropertiesBySet().values()) {
      if (!prop.appliesTo(path))
        continue;
      if (SProp.INSTANCE.getViewProperties().contains(prop))
        continue;
      appendProperty(ret, strepco, view, defaults, prop, path);
    }
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation())) {
        if (!Boolean.TRUE.equals(view.getOrNull(CProp.AFFORDANCE, childPath)))
          ret.appendChild(generateViewPath(document, strepco, view, defaults, childPath));
      }
    }
    return ret;
  }
}
