// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.viewset;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewDefaults;
import java.util.Set;
import javax.annotation.concurrent.ThreadSafe;

/** Model for storage of a set of views. Implementations must be thread-safe. */
@ThreadSafe
public abstract class ViewSet {
  private final ViewDefaults defaults;
  private final Set<ViewSetListener> listeners = Sets.newLinkedHashSet();

  public ViewSet(ViewDefaults defaults) {
    Preconditions.checkNotNull(defaults);
    this.defaults = defaults;
  }

  public final ViewDefaults getDefaults() {
    return defaults;
  }

  public final synchronized void addViewSetListener(ViewSetListener listener) {
    listeners.add(listener);
  }

  public final synchronized void removeViewSetListener(ViewSetListener listener) {
    listeners.remove(listener);
  }

  protected final synchronized ImmutableSet<ViewSetListener> getListenersSnapshot() {
    return ImmutableSet.copyOf(listeners);
  }

  /**
   * @param listeners set of listeners retrieved via {@link #getListenersSnapshot()} during the same
   *        transaction (synchronization on {@code this}) in which the change was made
   * @param added true if added, false if removed
   */
  protected final void fireViewSetChanged(
      ImmutableSet<ViewSetListener> listeners, GatedView view, boolean added)
  {
    Preconditions.checkState(!Thread.holdsLock(this));
    ViewSetEvent evt = new ViewSetEvent(this, view, added);
    for (ViewSetListener listener : listeners)
      listener.viewSetChanged(evt);
  }

  // TODO[cc]: Consider allowing getViews/createAndAddView to throw IOException.

  /** May incur I/O; may not be called on the EDT. */
  public abstract ImmutableSet<GatedView> getViews();
  /** May incur I/O; may not be called on the EDT. */
  public abstract GatedView createAndAddView(View source);
  /** May incur I/O; may not be called on the EDT. */
  public abstract void removeView(GatedView view);
}
