// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.updater;

import java.awt.EventQueue;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.netbeans.api.annotations.common.StaticResource;
import org.netbeans.api.autoupdate.InstallSupport;
import org.netbeans.api.autoupdate.InstallSupport.Installer;
import org.netbeans.api.autoupdate.InstallSupport.Validator;
import org.netbeans.api.autoupdate.OperationException;
import org.netbeans.api.autoupdate.OperationSupport.Restarter;
import org.netbeans.api.autoupdate.UpdateElement;
import org.netbeans.api.autoupdate.UpdateManager;
import org.netbeans.api.autoupdate.UpdateUnit;
import org.netbeans.api.autoupdate.UpdateUnitProvider;
import org.netbeans.api.autoupdate.UpdateUnitProviderFactory;
import org.openide.modules.ModuleInstall;
import org.openide.util.Exceptions;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.Utilities;

// TODO[pd]: Consider forcing exit if update is unsuccessful.
/* TODO[pd]: See if there is a way to make use of org.netbeans.updater.UpdaterFrame to show download
             progress. */
/* TODO[pd]: Consider using a custom main class instead of a module installer. When a module
             installer is used, all modules have to be loaded before we can check for updates, even
             if we are going to end up restarting anyway. */
public class UpdaterInstaller extends ModuleInstall {
  public static final String AUTOUPDATE_ENABLE_SYSPROP = "sieuferd.autoupdate";
  public static final String CUSTOM_UPDATE_CENTER_URL_SYSPROP = "sieuferd.updateurl";
  public static final String CUSTOM_UPDATE_CENTER_PATH_SYSPROP = "sieuferd.updatepath";
  // Must correspond to the name of the update center defined in the branding module's layer.xml.
  public static final String DEFAULT_UPDATE_CENTER_NAME = "sieuferd_update_center_default";
  public static final String CUSTOM_UPDATE_CENTER_NAME = "sieuferd_update_center_custom";
  public static final String UPDATE_CENTER_CATEGORY = "Sieuferd";
  // TODO[cc]: Avoid duplicating the icon here and in the branding module.
  @StaticResource
  public static final String UPDATE_CENTER_ICON_BASE = "com/sieuferd/updater/uc_icon.png";
  /** May be incremented to test autoupdate of the updater module. */
  private static final int UPDATER_VERSION = 1;
  private static final Logger LOG = Logger.getLogger(UpdaterInstaller.class.getName());

  // TODO[cc]: Consider using annotations in this module as well.
  /** @return may be null */
  private static UpdateUnitProvider configureActiveUpdateProvider() {
    final boolean autoUpdateEnabled =
        Boolean.valueOf(System.getProperty(AUTOUPDATE_ENABLE_SYSPROP, "true"));
    final URL customUpdateCenterURL;
    {
      String urlProp = System.getProperty(CUSTOM_UPDATE_CENTER_URL_SYSPROP, "").trim();
      // Don't trim this one; technically, paths may contain spaces.
      String pathProp = System.getProperty(CUSTOM_UPDATE_CENTER_PATH_SYSPROP, "");
      if (urlProp.isEmpty()) {
        if (pathProp.isEmpty()) {
          customUpdateCenterURL = null;
        } else {
          LOG.log(Level.INFO, "Got custom update center path ''{0}''", pathProp);
          try {
            customUpdateCenterURL = Utilities.toURI(new File(pathProp).getAbsoluteFile()).toURL();
          } catch (MalformedURLException e) {
            LOG.log(Level.SEVERE, "Could not convert update center path to URL", e);
            return null;
          }
        }
      } else {
        if (!pathProp.isEmpty()) {
          LOG.log(Level.SEVERE,
              "Cannot specify both an update center URL and an update center path");
          return null;
        }
        try {
          customUpdateCenterURL = new URL(urlProp);
        } catch (MalformedURLException e) {
          LOG.log(Level.SEVERE, "Malformed custom update center URL", e);
          return null;
        }
      }
      if (customUpdateCenterURL != null)
        LOG.log(Level.INFO, "Effective custom update center URL is ''{0}''", customUpdateCenterURL);
    }

    UpdateUnitProvider ret = null;
    final UpdateUnitProviderFactory providerFactory = UpdateUnitProviderFactory.getDefault();
    for (UpdateUnitProvider provider :
        new ArrayList<>(providerFactory.getUpdateUnitProviders(false)))
    {
      if (CUSTOM_UPDATE_CENTER_NAME.equals(provider.getName())) {
        URL providerURL = provider.getProviderURL();
        if (!autoUpdateEnabled || providerURL == null || customUpdateCenterURL == null ||
            !providerURL.toExternalForm().equals(customUpdateCenterURL.toExternalForm()))
        {
          providerFactory.remove(provider);
        }
      } else {
        boolean use = autoUpdateEnabled && customUpdateCenterURL == null &&
            DEFAULT_UPDATE_CENTER_NAME.equals(provider.getName());
        provider.setEnable(use);
        if (use)
          ret = provider;
      }
    }
    if (!autoUpdateEnabled) {
      LOG.info("Automatic updates disabled");
      return null;
    }
    if (ret == null) {
      if (customUpdateCenterURL == null) {
        LOG.severe("No update center defined");
        return null;
      } else {
        ret = providerFactory.create(CUSTOM_UPDATE_CENTER_NAME,
            NbBundle.getMessage(UpdaterInstaller.class, "LBL_sieuferd_update_center_custom"),
            customUpdateCenterURL, UPDATE_CENTER_ICON_BASE, UPDATE_CENTER_CATEGORY);
        ret.setEnable(true);
      }
    }
    return ret;
  }

  // TODO[pd]: Show errors to user instead of just logging them.
  private static void ensureModulesInstalled() {
    LOG.info("Checking for updates");
    UpdateUnitProvider provider = configureActiveUpdateProvider();
    if (provider == null)
      return;
    try {
      provider.refresh(null, true);
    } catch (IOException e) {
      LOG.log(Level.WARNING, "Failed to refresh update provider", e);
    }
    // TODO[pd]: Remove unneeded modules.

    // TODO[pd]: Get rid of this.
    LOG.info("Available UpdateUnits:");
    for (UpdateUnit updateUnit : provider.getUpdateUnits()) {
      LOG.log(Level.INFO, "  {0}/type={1}/installed={2}/available={3}", new Object[] {
          updateUnit, updateUnit.getType(), updateUnit.getInstalled(),
          updateUnit.getAvailableUpdates() });
    }

    final OperationCollector operations = new OperationCollector();
    /* TODO[pd]: Require a single KIT_MODULE instead of including everything that might be available
                 in the update center. Note that doing
                 provider.getUpdateUnits(UpdateManager.TYPE.values()) gives a "Cannot mix types
                 MODULE and KIT_MODULE into once [sic] list" error, while doing
                 provider.getUpdateUnits() only returns kit modules, unlike
                 UpdateManager.getDefault().getUpdateUnits(). */
    for (UpdateUnit updateUnit : UpdateManager.getDefault().getUpdateUnits()) {
      List<UpdateElement> versions = updateUnit.getAvailableUpdates();
      if (!versions.isEmpty()) {
        /* Higher specification versions are listed first;
        see org.netbeans.modules.autoupdate.services.UpdateUnitImpl.identifyUpdates. */
        operations.addUpdateElement(versions.get(0));
        versions.get(0);
      }
    }
    final InstallSupport support = operations.getInstallSupport();
    if (support == null)
      return;
    try {
      LOG.info("Downloading");
      // TODO[pd]: Consider if isGlobal should be set to false instead.
      Validator v = support.doDownload(null, true, true);
      LOG.info("Validating");
      Installer i = support.doValidate(v, null);
      LOG.info("Installing");
      final Restarter r = support.doInstall(i, null);
      if (r != null) {
        /* Do the restart on a separate thread to avoid deadlock, since doRestart will block and
        may not actually perform the restart until the module installer has completed. */
        RequestProcessor.getDefault().post(() -> {
          try {
            LOG.info("Restarting");
            support.doRestart(r, null);
            LOG.info("doRestart returned");
          } catch (OperationException e) {
            Exceptions.printStackTrace(e);
          }
        });
      }
    } catch (OperationException e) {
      LOG.log(Level.SEVERE, "Process failed", e);
    }
  }

  /* Overriding validate() instead would be too early; we need the autoupdate system to be loaded
  first. */
  @Override
  public void restored() {
    LOG.log(Level.INFO, "Sieuferd updater module version {0}", UPDATER_VERSION);
    // TODO[pd]: Consider registering a warm-up task instead.
    // Wait until the splash screen is gone.
    EventQueue.invokeLater(UpdaterInstaller::ensureModulesInstalled);
  }
}
