// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.TreeStructure;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.project.FieldContext.PrimitiveFieldContext;
import com.sieuferd.project.FieldContext.RelationFieldContext;
import com.sieuferd.project.viewfile.ViewDataObject;
import com.sieuferd.viewset.GatedView;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

/** Shallowly immutable. */
public abstract class FieldContext
    extends TreeStructure<FieldContext,RelationFieldContext,PrimitiveFieldContext>
{
  private final ViewContext viewContext;
  private final Path path;

  private FieldContext(final ViewContext viewContext, final Path path) {
    Preconditions.checkNotNull(viewContext);
    Preconditions.checkNotNull(path);
    this.viewContext = viewContext;
    this.path = path;
  }

  public static FieldContext create(final ViewContext viewContext, final Path path) {
    return path.isRelation()
        ? new RelationFieldContext(viewContext, path.getRelation())
        : new PrimitiveFieldContext(viewContext, path.getPrimitive());
  }

  public boolean exists() {
    return getView().exists(path);
  }

  public ViewContext getViewContext() {
    return viewContext;
  }

  public GatedView getView() {
    return viewContext.getView();
  }

  public SieuferdProject getProject() {
    return viewContext.getProject();
  }

  public ViewDataObject getViewDataObject() {
    return viewContext.getDataObject();
  }

  public <T> T get(ViewProperty<T> property) {
    return getView().get(property, path);
  }

  public @Nullable <T> T getOrNull(ViewProperty<T> property) {
    return getView().getOrNull(property, path);
  }

  public @Nonnull <T> T getOrDefault(ViewProperty<T> property) {
    return getView().getOrDefault(property, path);
  }

  // TODO[cc]: Consider encapsulating getPath to avoid errors when the path has been removed.

  public Path getPath() {
    return path;
  }

  public boolean isRoot() {
    return path.isRoot();
  }

  @Override
  public String toString() {
    return path.toString(viewContext.getView());
  }

  /* Implementing equals() and hashCode() makes FieldContext work better as a key for ChildFactory.
  For instance, it avoids recreating nodes unnecessarily, in turn avoiding expanded nodes
  spontaneously collapsing when siblings are added or removed. */

  @Override
  public final int hashCode() {
    return Objects.hashCode(viewContext, path);
  }

  @Override
  public final boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof FieldContext))
      return false;
    FieldContext other = (FieldContext) obj;
    return Objects.equal(this.viewContext, other.viewContext)
        && Objects.equal(this.path, other.path);
  }

  @Override
  protected Class<RelationFieldContext> getRelationClass() {
    return RelationFieldContext.class;
  }

  @Override
  protected Class<PrimitiveFieldContext> getPrimitiveClass() {
    return PrimitiveFieldContext.class;
  }

  public static final class RelationFieldContext extends FieldContext {
    public RelationFieldContext(final ViewContext viewContext, final RelationPath path) {
      super(viewContext, path);
    }

    @Override
    public RelationPath getPath() {
      return super.getPath().getRelation();
    }
  }

  public static final class PrimitiveFieldContext extends FieldContext {
    public PrimitiveFieldContext(final ViewContext viewContext, final PrimitivePath path) {
      super(viewContext, path);
    }

    @Override
    public PrimitivePath getPath() {
      return super.getPath().getPrimitive();
    }
  }
}
