// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewPropertyChangedEvent;
import com.sieuferd.datamodel.ViewTransactionCompletedEvent;
import com.sieuferd.datamodel.Views;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.ObjectReference;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.viewset.GatedView;
import java.awt.EventQueue;
import java.lang.reflect.Proxy;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.concurrent.atomic.AtomicLong;
import javax.annotation.concurrent.Immutable;
import javax.swing.UIManager;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;
import org.openide.awt.Actions;
import org.openide.util.Mutex;

// TODO[cc]: Move classes in this package elsewhere.

/* TODO[cc]: Document the fact that this view will only change and deliver events on the EDT. This
             is a specific implementation class, and so might not be the appropriate place for this
             (whereas the superclass is not supposed to be Swing-dependent). */
public final class GatedViewImpl extends GatedView {
  private static final AtomicLong DEBUG_NEXT_VIEW_ID = new AtomicLong();
  private final long debugViewID;
  private final AtomicLong debugRevision = new AtomicLong();
  /** Set to 0 to disable. */
  private static final int DEBUG_DUMP_LISTENERS_INTERVAL = 400;
  private final MutableView delegate;
  private final Map<ViewListener,ViewListener> listenerMap = Maps.newLinkedHashMap();
  // TODO[pd]: Consider minimizing the amount of memory needed to store each edit.
  private final Map<ViewVersion,ImmutableView> viewSnapshots = new WeakHashMap<>();
  private ViewVersion currentVersion;
  /** Will only be accessed from the EDT. Used for assertion checking only. */
  private boolean mutationInProgress;

  public GatedViewImpl(MutableView template) {
    Preconditions.checkNotNull(template);
    this.debugViewID = DEBUG_NEXT_VIEW_ID.getAndIncrement();
    this.delegate = Views.newMutable(template, template.getDefaults());
    UIUtil.instrumentation("NewView viewID=" + debugViewID + ", label='" +
        this.delegate.getOrNull(CProp.LABEL, Path.ROOT) + "'");
    updateViewSnapshot(true);
  }

  private synchronized void updateViewSnapshot(boolean newVersion) {
    if (newVersion)
      currentVersion = new ViewVersion();
    viewSnapshots.put(currentVersion, ImmutableView.copyOf(this.delegate));
  }

  private synchronized ViewVersion getCurrentViewVersion() {
    return currentVersion;
  }

  /** @return true iff successful */
  private boolean compareAndSetViewSnapshot(
      final ViewVersion priorVersion, final ViewVersion targetVersion)
  {
    final ObjectReference<Boolean> ret = ObjectReference.create();
    Mutex.EVENT.writeAccess((Mutex.Action<Object>) () -> {
      final ImmutableView targetView = viewSnapshots.get(targetVersion);
      if (getCurrentViewVersion() == priorVersion && targetView != null) {
        synchronized (GatedViewImpl.this) {
          currentVersion = targetVersion;
        }
        delegate.copyFrom(targetView);
        ret.setValue(true);
      } else {
        ret.setValue(false);
      }
      return null;
    });
    return ret.getValue();
  }

  /* TODO[pd]: Consider including pluralized nouns in operation presentation names, e.g.
           "Hide Field/Fields" instead of just "Hide". */
  @Override
  public void mutate(final String operationPresentationName, final boolean incrementUndoVersion,
      final ViewOperation operation)
  {
    if (incrementUndoVersion) {
      UIUtil.instrumentation("Mutate viewID=" + debugViewID + ", revision=" +
          debugRevision.getAndIncrement() + ", operationName='" + operationPresentationName + "'");
    }
    Mutex.EVENT.writeAccess((Mutex.Action<Object>) () -> {
      final ViewVersion versionBefore = getCurrentViewVersion();
      delegate.performWriteTransaction(() -> {
        try (final TransientView transientView = new TransientView(delegate)) {
          mutationInProgress = true;
          operation.performOperation(transientView);
        } finally {
          mutationInProgress = false;
        }
      });
      updateViewSnapshot(incrementUndoVersion);
      /* Making changes to the view without incrementing the undo version is not quite the same
      as the "insignificant edits" concept from Swing's UndoManager, because even an insignificant
      edit will invalidate subsequent undoable edits, making redo impossible. Hence, I only create
      edit events for edits that should really show up in the Undo/Redo timeline. */
      if (incrementUndoVersion) {
        fireUndoableEditHappened(new UndoableViewEdit(
            Actions.cutAmpersand(operationPresentationName),
            versionBefore, getCurrentViewVersion()));
      }
      return null;
    });
  }

  // Wrap the listeners to ensure that the correct source object is used in events.
  private ViewListener wrapListener(final ViewListener original) {
    return new ViewListener() {
      @Override
      public <T> void propertyChanged(ViewPropertyChangedEvent<T> evt) {
        ThreadUtil.checkOnEDT();
        if (isListenerStillListening(original)) {
          original.propertyChanged(new ViewPropertyChangedEvent<>(GatedViewImpl.this,
              evt.getProperty(), evt.getPath(), evt.getOldValue(), evt.getNewValue()));
        }
      }

      @Override
      public void transactionCompleted(ViewTransactionCompletedEvent evt) {
        ThreadUtil.checkOnEDT();
        if (isListenerStillListening(original)) {
          original.transactionCompleted(
              new ViewTransactionCompletedEvent(GatedViewImpl.this));
        }
      }
    };
  }

  // TODO[cc]: Do I need this?
  private synchronized boolean isListenerStillListening(ViewListener originalListener) {
    return listenerMap.containsKey(originalListener);
  }

  private synchronized void debugPrintListeners() {
    System.out.println(this + " has " + listenerMap.size() + " listeners:");
    Map<Object,Integer> summary = Maps.newLinkedHashMap();
    for (ViewListener viewListener : listenerMap.keySet()) {
      final String key = Proxy.isProxyClass(viewListener.getClass())
          ? Proxy.getInvocationHandler(viewListener).toString()
          : viewListener.getClass().toString();
      Integer count = summary.get(key);
      count = (count == null ? 0 : count) + 1;
      summary.put(key, count);
    }
    for (Map.Entry<Object,Integer> entry : summary.entrySet())
      System.out.println("\t" + entry.getValue() + "\t" + entry.getKey());
  }

  private void checkNotCalledFromMutatingThread() {
    Preconditions.checkState(!(EventQueue.isDispatchThread() && mutationInProgress),
        "Don't access the GatedView from within a ViewOperation; " +
        "use the supplied MutableView instead");
  }

  @Override
  public synchronized void addViewListener(Path pathToListenTo, ViewListener listener) {
    checkNotCalledFromMutatingThread();
    ViewListener wrappedListener = wrapListener(listener);
    MapUtil.putChecked(listenerMap, listener, wrappedListener);
    delegate.addViewListener(pathToListenTo, wrappedListener);
    if (DEBUG_DUMP_LISTENERS_INTERVAL > 0 &&
        listenerMap.size() % DEBUG_DUMP_LISTENERS_INTERVAL == 0)
    {
      debugPrintListeners();
    }
  }

  @Override
  public synchronized void removeViewListener(final ViewListener listener) {
    checkNotCalledFromMutatingThread();
    ViewListener wrappedListener = MapUtil.removeChecked(listenerMap, listener);
    delegate.removeViewListener(wrappedListener);
  }

  @Override
  public boolean exists(Path path) {
    checkNotCalledFromMutatingThread();
    return delegate.exists(path);
  }

  @Override
  public <T> T get(ViewProperty<T> property, Path path) {
    checkNotCalledFromMutatingThread();
    return delegate.get(property, path);
  }

  @Override
  public Set<ViewProperty<?>> getDefinedProperties(Path path) {
    checkNotCalledFromMutatingThread();
    return delegate.getDefinedProperties(path);
  }

  @Override
  public boolean defines(ViewProperty<?> property, Path path) {
    checkNotCalledFromMutatingThread();
    return delegate.defines(property, path);
  }

  @Override
  public ViewDefaults getDefaults() {
    checkNotCalledFromMutatingThread();
    return delegate.getDefaults();
  }

  @Override
  public void performReadTransaction(Runnable runnable) {
    checkNotCalledFromMutatingThread();
    delegate.performReadTransaction(runnable);
  }

  @Override
  public String toString() {
    return delegate.toString();
  }

  /** Used for referential keys denoting a particular version of a View. */
  private static class ViewVersion { }

  @Immutable
  private class UndoableViewEdit implements UndoableEdit {
    private final String presentationName;
    private final ViewVersion versionBefore;
    private final ViewVersion versionAfter;

    UndoableViewEdit(
        String presentationName, ViewVersion versionBefore, ViewVersion versionAfter)
    {
      Preconditions.checkNotNull(presentationName);
      Preconditions.checkNotNull(versionBefore);
      Preconditions.checkNotNull(versionAfter);
      this.presentationName = presentationName;
      this.versionBefore = versionBefore;
      this.versionAfter = versionAfter;
    }

    @Override
    public boolean canUndo() {
      return getCurrentViewVersion() == versionAfter;
    }

    @Override
    public boolean canRedo() {
      return getCurrentViewVersion() == versionBefore;
    }

    @Override
    public void undo() throws CannotUndoException {
      if (!compareAndSetViewSnapshot(versionAfter, versionBefore))
        throw new CannotUndoException();
      UIUtil.instrumentation("Undo viewID=" + debugViewID + ", revision=" +
          debugRevision.getAndIncrement());
    }

    @Override
    public void redo() throws CannotRedoException {
      if (!compareAndSetViewSnapshot(versionBefore, versionAfter))
        throw new CannotUndoException();
      UIUtil.instrumentation("Redo viewID=" + debugViewID + ", revision=" +
          debugRevision.getAndIncrement());
    }

    @Override
    public void die() {
      // No action.
    }

    @Override
    public boolean addEdit(UndoableEdit anEdit) {
      return false;
    }

    @Override
    public boolean replaceEdit(UndoableEdit anEdit) {
      return false;
    }

    @Override
    public boolean isSignificant() {
      // TODO[pd]: Consider making use of this.
      return true;
    }

    @Override
    public String getPresentationName() {
      return presentationName;
    }

    @Override
    public String getUndoPresentationName() {
      /* Do the same thing as in AbstractUndoableEdit, which is what gives the best behavior when
      used with org.openide.actions.UndoRedoAction anyway (see
      https://netbeans.org/bugzilla/show_bug.cgi?id=40823 ). */
      return (UIManager.getString("AbstractUndoableEdit.undoText") + " " + getPresentationName())
          .trim();
    }

    @Override
    public String getRedoPresentationName() {
      // See above.
      return (UIManager.getString("AbstractUndoableEdit.redoText") + " " + getPresentationName())
          .trim();
    }
  }
}
