// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLTaskRunner;
import com.sieuferd.database.SQLTask;
import com.sieuferd.database.SQLTask.UserInformedConnectionException;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.Stopwatch;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import org.netbeans.api.db.explorer.ConnectionListener;
import org.netbeans.api.db.explorer.ConnectionManager;
import org.netbeans.api.db.explorer.DatabaseConnection;
import org.netbeans.api.db.explorer.DatabaseException;
import org.openide.util.Mutex;
import org.openide.util.Mutex.Action;

// TODO[cc]: Consider moving this class into a different package.
// TODO[pd]: Move query interruption logic into this class.
public final class NBSQLTaskRunner extends SQLTaskRunner {
  private static final Logger LOG = Logger.getLogger(NBSQLTaskRunner.class.getName());
  private static final ConnectionManager CM = ConnectionManager.getDefault();
  /** Synchronized on itself. */
  private static final Map<String,Object> connectionLocks = Maps.newLinkedHashMap();
  private static final ConnectionListener connectionListener = new ConnectionListener() {
    @Override
    public void connectionsChanged() {
      // A lot of work just to try to avoid a memory leak in connectionLocks, but oh well...
      synchronized (connectionLocks) {
        final Set<String> allConnectionNames = Sets.newLinkedHashSet();
        for (DatabaseConnection connection : CM.getConnections())
          allConnectionNames.add(connection.getName());
        for (String connectionName : ImmutableSet.copyOf(connectionLocks.keySet())) {
          if (!allConnectionNames.contains(connectionName))
            connectionLocks.remove(connectionName);
        }
        if (connectionLocks.isEmpty())
          CM.removeConnectionListener(connectionListener);
      }
    }
  };
  private final @Nullable String connectionName;

  // TODO[cc]: Go over all calls to this and see if I should provide a shortcut.
  /** @param connectionName may be null to indicate that no connection is configured */
  public static SQLTaskRunner get(@Nullable String connectionName) {
    return new NBSQLTaskRunner(connectionName);
  }

  private Object getConnectionLock() {
    synchronized (connectionLocks) {
      Object ret = connectionLocks.get(connectionName);
      if (ret == null) {
        ret = new Object();
        // Only store lock objects for connections that are actually still in the ConnectionManager.
        if (CM.getConnection(connectionName) != null) {
          connectionLocks.put(connectionName, ret);
          if (connectionLocks.size() == 1)
            CM.addConnectionListener(connectionListener);
        }
      }
      return ret;
    }
  }

  private NBSQLTaskRunner(@Nullable String connectionName) {
    this.connectionName = connectionName;
  }

  @Override
  public <R> R runTask(
      boolean showDialogOnConnect, boolean testConnection, SQLTask<R> task)
      throws SQLException, UserInformedConnectionException
  {
    ThreadUtil.checkNotOnEDT();
    if (connectionName == null)
      throw new UserInformedConnectionException("No connection configured");
    /* The connection lock is tied to the DatabaseConnection (via its connection name) rather than
    the underlying JDBC Connection so that the call to getLiveJDBCConnection can be done from within
    the mutex. Acquiring the lock after calling getLiveJDBCConnection once lead to a bug where the
    retrieved connection could be closed by another thread right afterwards. */
    synchronized (getConnectionLock()) {
      final Connection jdbcConnection;
      try {
        jdbcConnection = getLiveJDBCConnection(connectionName, showDialogOnConnect, testConnection);
      } catch (DatabaseException e) {
        throw new SQLException(e);
      }
      final SQLConnection sqlConnection = new SQLConnection(jdbcConnection, true);
      try {
        LOG.log(Level.INFO, "Running SQLTask ''{0}''", task);
        Stopwatch stopwatch = new Stopwatch("SQLTask '" + task.toString() + "'");
        R ret = task.run(sqlConnection);
        stopwatch.stop();
        return ret;
      } catch (Exception e) {
        if (e instanceof SQLException)
          sqlConnection.reportSQLException((SQLException) e);
        throw e;
      } finally {
        sqlConnection.markClosed();
        if (sqlConnection.isInvalidated()) {
          DatabaseConnection databaseConnection = CM.getConnection(connectionName);
          if (databaseConnection != null) {
            LOG.info("Disconnecting invalidated database connection");
            CM.disconnect(databaseConnection);
          }
        }
      }
    }
  }

  private static final int CONNECTION_ISVALID_TIMEOUT = 10;

  private static @Nullable Connection getValidatedJDBCConnection(
      DatabaseConnection databaseConnection)
  {
    Connection conn = databaseConnection.getJDBCConnection();
    try {
      /* NetBeans bug #221602 does not apply here, as we are not trying to run isClosed()
      concurrently with other queries. */
      if (conn == null || conn.isClosed())
        return conn;
      try {
        /* The documentation for c3p0 ( http://www.mchange.com/projects/c3p0 ) places quite an
        emphasis on using an efficient test query. Use Connection.isValid() if the driver supports
        it (available since JDBC4). Same logic as
        com.mchange.v2.c3p0.impl.DefaultConnectionTester.ThreadLocalQuerylessTestRunner.findTestRunner */
        conn.getClass().getDeclaredMethod("isValid", new Class<?>[] { Integer.TYPE });
        // TODO[pd]: Should I use a zero (infinite) timeout here like c3p0 does?
        return conn.isValid(CONNECTION_ISVALID_TIMEOUT) ? conn : null;
      } catch (NoSuchMethodException | SecurityException e) {
        // Use NetBeans' testing implementation.
        return databaseConnection.getJDBCConnection(true);
      }
    } catch (SQLException e) {
      return null;
    }
  }

  /**
   * Get a working SQLConnection if present for the given connection name, possibly prompting the
   * user to connect in a modal dialog if showDialog is true. Only one dialog will be shown at the
   * time, and only one dialog will be shown for each simultaneous call to this method for the same
   * connection. All calls to this method will block until the dialog for the connection, if any,
   * has exited. The returned SQLConnection may safely be closed, but this is not required. This
   * method may not be called from the EDT.
   *
   * @param databaseConnectionName the name of the database connection
   * @param showDialog whether to show a connection dialog to the user if the connection exists but is
   *        not live.
   * @param testConnection whether the connection should be validated before being returned (may take
   *        more time)
   * @throws UserInformedConnectionException if the action failed in a way already communicated to the
   *         user (e.g. due to the user canceling the task), or if showDialog is false and there
   *         currently exists no live connection.
   */
  private static @Nonnull Connection getLiveJDBCConnection(
      String databaseConnectionName, boolean showDialog, boolean testConnection)
      throws DatabaseException, UserInformedConnectionException
  {
    ThreadUtil.checkNotOnEDT();
    Preconditions.checkNotNull(databaseConnectionName);
    /* "The [Database] API is thread-safe."
    See http://bits.netbeans.org/7.3/javadoc/org-netbeans-modules-db/architecture-summary.html . */
    final DatabaseConnection databaseConnection = CM.getConnection(databaseConnectionName);
    if (databaseConnection == null) {
      if (showDialog) {
        throw new DatabaseException("Unknown database connection '" + databaseConnectionName + "'");
      } else {
        throw new UserInformedConnectionException(
            "Unknown database connection '" + databaseConnectionName + "'");
      }
    }
    /* Not sure if this will ever happen, since I already got the connection from
    ConnectionManager.getConnection(). */
    if (!Sets.newHashSet(CM.getConnections()).contains(databaseConnection))
      CM.addConnection(databaseConnection);
    // TODO[pd]: Can a connection be deleted by a different thread right here?
    Connection jdbcConnection;
    if (testConnection) {
      jdbcConnection = getValidatedJDBCConnection(databaseConnection);
      if (jdbcConnection == null)
        CM.disconnect(databaseConnection);
    } else {
      jdbcConnection = databaseConnection.getJDBCConnection();
    }
    if (jdbcConnection == null) {
      waitForConnectionDialog(CM, databaseConnection, showDialog);
      /* Don't try to detect whether a dialog was shown or not before calling getJDBCConnection;
      this led to a race condition bug in the past. */
      /* Don't test the connection again here, but rely on getJDBCConnection to return null if the
      user pressed the cancel button. */
      jdbcConnection = databaseConnection.getJDBCConnection();
    }
    if (jdbcConnection == null)
      throw new UserInformedConnectionException("User canceled connection");
    return jdbcConnection;
  }

  /** Mutex to ensure that only one dialog is shown at a time, regardless of connection. */
  private static final Mutex globalConnectionDialogMutex = new Mutex();
  /* TODO[cc]: I might be able to get rid of this now that I am enclosing calls to
               getLiveJDBCConnection in a DatabaseConnection-specific mutex anyway. */
  /** Holds one Mutex for each DatabaseConnection that currently has a connection dialog open or is
  waiting to display one. Access should be synchronized on the object itself. */
  private static final Map<DatabaseConnection,Mutex> specificConnectionDialogMutexes =
      Maps.newLinkedHashMap();

  /**
   * Wait for any open connection dialogs for the specified connection to close. If showDialog is
   * true and there is currently no connection dialog showing for the specified connection, show the
   * dialog first.
   */
  private static void waitForConnectionDialog(
      final ConnectionManager cm, final DatabaseConnection connection, final boolean showDialog)
  {
    ThreadUtil.checkNotOnEDT();
    final Mutex specificConnectionDialogMutex;
    final boolean showDialogFromThisThread;
    synchronized (specificConnectionDialogMutexes) {
      Mutex toMutex = specificConnectionDialogMutexes.get(connection);
      showDialogFromThisThread = (toMutex == null && showDialog);
      if (showDialogFromThisThread) {
        toMutex = new Mutex();
        specificConnectionDialogMutexes.put(connection, toMutex);
      }
      specificConnectionDialogMutex = toMutex;
    }
    // No one is showing a dialog for this connection, including ourselves.
    if (specificConnectionDialogMutex == null) {
      LOG.log(Level.INFO, "No thread is showing connection dialog");
      return;
    }
    final long timeBefore = System.nanoTime();
    try {
      /* Wait for the mutex even if not showing the dialog from this thread, to make sure we are
      waiting for _any_ dialog being shown for the specified connection that might be open at the
      moment. */
      specificConnectionDialogMutex.writeAccess((Action<Void>) () -> {
        if (showDialogFromThisThread) {
          /* First try to connect without a dialog. Note that the NetBeans Database API seems to
          save the password in memory even when the user has not opted to store it with the
          "Remember Password" option. This is true even when the connection is temporarily
          closed. */
          try {
            if (cm.connect(connection)) {
              LOG.info("Silent reconnect succeeded");
              return null;
            } else {
              LOG.info("Silent reconnect not attempted");
            }
          } catch (DatabaseException e) {
            LOG.log(Level.INFO, "Silent reconnect failed with exception: {0}", e.getMessage());
            /* Ignore; I've observed ConnectionManager.connect to try to connect even when no
            password has been stored. There is no way to check whether a password is stored or
            not, because DatabaseConnection.getPassword() returns the empty string rather than
            null when no password is stored, and the empty string is in fact a valid password. */
          }
          /* Wait on the globalConnectionDialogMutex only once we have acquired the
          specificConnectionDialogMutex, since we want other threads to wait for the latter if we
          are in line to show a dialog for the connection they are interested in. */
          globalConnectionDialogMutex.writeAccess((Action<Void>) () -> {
            /* Calling ConnectionManager.showConnectionDialog() from the EDT ensures blocking
            behavior. */
            Mutex.EVENT.writeAccess((Action<Void>) () -> {
              cm.showConnectionDialog(connection);
              return null;
            });
            return null;
          });
        }
        return null;
      });
    } finally {
      final long timeAfter = System.nanoTime();
      if (showDialogFromThisThread) {
        synchronized (specificConnectionDialogMutexes) {
          specificConnectionDialogMutexes.remove(connection);
        }
      }
      if (showDialog) {
        LOG.log(Level.INFO, "Completed waitForConnectionDialog with dialog from {0} after {1}us",
            new Object[] { showDialogFromThisThread ? "this thread": "other thread",
            (timeAfter - timeBefore) / 1000L});
      }
    }
  }
}
