// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.project.viewfile.VSProp;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewAdapter;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewTransactionCompletedEvent;
import com.sieuferd.datamodel.Views;
import com.sieuferd.layout.DefaultStyleSheets;
import com.sieuferd.layout.LAProp;
import com.sieuferd.project.viewfile.ViewIdentifier;
import com.sieuferd.project.viewfile.ViewURLComponents;
import com.sieuferd.query.QProp;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.Stopwatch;
import com.sieuferd.util.StringUtil;
import com.sieuferd.viewset.GatedView;
import com.sieuferd.viewset.SieuferdProjectXML;
import com.sieuferd.viewset.ViewSet;
import com.sieuferd.viewset.ViewSetListener;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileUtil;
import org.openide.util.RequestProcessor;

// TODO[cc]: Clean up the interactions between NBViewSet, ViewContext, and ViewURLMapper.
public class NBViewSet extends ViewSet {
  /* TODO[cc]: Clean up all the current dependencies on this. Get the list of properties with defaults
               from getViewSet().getDefaults() instead. */
  public static final ViewDefaults VIEW_DEFAULTS = ViewDefaults.builder()
      .setAll(CProp.DEFAULTS)
      .setAll(QProp.DEFAULTS)
      .setAll(LAProp.DEFAULTS)
      .setAll(DefaultStyleSheets.GRAYSCALE)
      .build();
  private static final Logger LOG = Logger.getLogger(NBViewSet.class.getName());
  private static final String VIEWS_DIRECTORY = "views";
  private static final RequestProcessor RP_REFRESH_DERIVATIVES =
      new RequestProcessor("Refresh Derivatives", 1);
  private final RequestProcessor.Task refreshDerivativesTask =
      RP_REFRESH_DERIVATIVES.create(new Runnable()
  {
    @Override
    public void run() {
      LOG.info("Refreshing derivatives");
      UIUtil.instrumentation("RefreshDerivatives");
      Stopwatch stopwatch = new Stopwatch("Refreshing derivatives");
      ViewUpdater.refreshDerivatives(project);
      stopwatch.stop();
      UIUtil.instrumentation("RefreshedDerivatives");
    }
  });
  private final SieuferdProject project;
  private final FileSystem viewFileSystem = FileUtil.createMemoryFileSystem();
  private final Object loadPersistedViewsLock = new Object();
  /** Guarded by {@code loadPersistedViewsLock}. */
  private boolean persistedViewsLoaded = false;
  /** Guarded by {@code this}. */
  private final Map<ViewIdentifier,GatedView> views = Maps.newLinkedHashMap();
  private final ViewListener refreshDerivativesViewListener = new ViewAdapter() {
    @Override
    public void transactionCompleted(ViewTransactionCompletedEvent evt) {
      scheduleRefreshDerivatives();
    }
  };

  private void scheduleRefreshDerivatives() {
    refreshDerivativesTask.schedule(200);
  }

  /** The constructor will not block, and is safe to call from the EDT. */
  public NBViewSet(SieuferdProject project) {
    super(VIEW_DEFAULTS);
    Preconditions.checkNotNull(project);
    this.project = project;
  }

  /** Load all persisted views if such loading has not already been initiated, and block until
  loading has completed. May not be called from the EDT. */
  private void ensurePersistedViewsLoaded() {
    ThreadUtil.checkNotOnEDT();
    synchronized (loadPersistedViewsLock) {
      if (persistedViewsLoaded)
        return;
      persistedViewsLoaded = true;
      FileObject viewsDirectory = project.getProjectDirectory().getFileObject(VIEWS_DIRECTORY);
      if (viewsDirectory == null)
        return;
      final Enumeration<? extends FileObject> queryFiles = viewsDirectory.getData(true);
      while (queryFiles.hasMoreElements()) {
        final FileObject fileObject = queryFiles.nextElement();
        try {
          createAndAddView(SieuferdProjectXML.readSieuferdViewXML(
              fileObject.getInputStream(), fileObject.toURI().toString(),
              getDefaults().getProperties()), ViewIdentifier.fromURI(fileObject.toURI()));
        } catch (IOException e) {
          LOG.log(Level.WARNING, "Error loading view file", e);
        }
      }
    }
  }

  private final Object fileCreationLock = new Object();

  /**
   * This method may be called from any thread, including the EDT.
   */
  public @Nullable FileObject getFileObjectForViewIdentifier(ViewIdentifier viewIdentifier) {
    String extension = "view";
    String fileName = "view-" + StringUtil.urlEncode(viewIdentifier.toString());
    synchronized (fileCreationLock) {
      FileObject ret = viewFileSystem.getRoot().getFileObject(fileName, extension);
      if (ret == null) {
        try {
          ret = viewFileSystem.getRoot().createData(fileName, extension);
          new ViewURLComponents(
              project.getProjectDirectory().toURI(), viewIdentifier).prepareFileObject(ret);
        } catch (IOException e) {
          throw new AssertionError("Unexpected IOException on in-memory file system", e);
        }
      }
      return ret;
    }
  }

  /**
   * May not be called from the EDT.
   */
  public @Nullable GatedView loadViewByIdentifier(ViewIdentifier identifier) throws IOException {
    ensurePersistedViewsLoaded();
    final GatedView ret;
    synchronized (this) {
      ret = views.get(identifier);
    }
    // TODO[cc]: This is a bit ugly. Clean up.
    if (ret == null) {
      FileObject fo = getFileObjectForViewIdentifier(identifier);
      if (fo != null)
        fo.delete();
    }
    return ret;
  }

  // TODO[cc]: Consider returning a ViewContext.
  @Override
  public GatedView createAndAddView(View source) {
    ThreadUtil.checkNotOnEDT();
    return createAndAddView(source, ViewIdentifier.newTransient());
  }

  private GatedView createAndAddView(View source, ViewIdentifier viewIdentifier) {
    ThreadUtil.checkNotOnEDT();
    MutableView viewTemplate = Views.newMutable(source, getDefaults());
    viewTemplate.set(VSProp.VIEW_IDENTIFIER, Path.ROOT, viewIdentifier);
    GatedView ret = new GatedViewImpl(Views.newMutable(viewTemplate, getDefaults()));
    final ImmutableSet<ViewSetListener> listenersSnapshot;
    synchronized (this) {
      MapUtil.putChecked(views, viewIdentifier, ret);
      listenersSnapshot = getListenersSnapshot();
    }
    fireViewSetChanged(listenersSnapshot, ret, true);
    ret.addViewListener(null, refreshDerivativesViewListener);
    scheduleRefreshDerivatives();
    return ret;
  }

  @Override
  public ImmutableSet<GatedView> getViews() {
    ThreadUtil.checkNotOnEDT();
    ensurePersistedViewsLoaded();
    synchronized (this) {
      return ImmutableSet.copyOf(views.values());
    }
  }

  @Override
  public void removeView(GatedView view) {
    ThreadUtil.checkNotOnEDT();
    try {
      final ViewContext viewContext = ViewContext.create(project, view);
      final ImmutableSet<ViewSetListener> listenersSnapshot;
      // TODO[pd]: An error in MapUtil.removeChecked was once observed here.
      view.removeViewListener(refreshDerivativesViewListener);
      synchronized (this) {
        MapUtil.removeChecked(views, view.get(VSProp.VIEW_IDENTIFIER, Path.ROOT));
        listenersSnapshot = getListenersSnapshot();
      }
      fireViewSetChanged(listenersSnapshot, view, false);
      viewContext.getDataObject().delete();
    } catch (IOException e) {
      LOG.log(Level.WARNING, "Problem removing view", e);
    }
  }
}
