// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.project.util.QueryLabelFormatter;
import com.sieuferd.query.QProp;
import com.sieuferd.util.ObjectReference;
import java.awt.Font;
import java.awt.FontFormatException;
import java.awt.Image;
import java.io.IOException;
import java.io.InputStream;
import javax.swing.Icon;
import javax.swing.UIManager;
import org.netbeans.api.annotations.common.StaticResource;
import org.openide.util.ImageUtilities;

public final class SieuferdIcons {
  private static final String ICON_FOLDER = "com/sieuferd/project/icons/";
  @StaticResource
  public static final String ICON_WARNING = ICON_FOLDER + "warning.png";
  @StaticResource
  private static final String ICON_DEFAULT_FOLDER = ICON_FOLDER + "folder.png";
  @StaticResource
  private static final String ICON_DEFAULT_FOLDER_OPEN = ICON_FOLDER + "folder_open.png";
  @StaticResource
  private static final String ICON_RELATION_JOINED_NOINST = ICON_FOLDER + "relation_joined_noinst2.png";
  @StaticResource
  private static final String ICON_RELATION_JOINED_NOPK = ICON_FOLDER + "relation_joined_nopk2.png";
  @StaticResource
  private static final String ICON_RELATION_JOINED = ICON_FOLDER + "relation_joined2.png";
  @StaticResource
  private static final String ICON_RELATION_NESTED_NOINST = ICON_FOLDER + "relation_nested_noinst2.png";
  @StaticResource
  private static final String ICON_RELATION_NESTED_NOPK = ICON_FOLDER + "relation_nested_nopk2.png";
  @StaticResource
  private static final String ICON_RELATION_NESTED = ICON_FOLDER + "relation_nested2.png";
  @StaticResource
  private static final String ICON_PRIMITIVE_COLUMN = ICON_FOLDER + "primitive_column.png";
  @StaticResource
  private static final String ICON_PRIMITIVE_FORMULA = ICON_FOLDER + "primitive_formula.png";
  @StaticResource
  private static final String ICON_NEW_PRIMITIVE_FORMULA = ICON_FOLDER + "new_primitive_formula.png";
  @StaticResource
  private static final String ICON_NEW_PRIMITIVE_COLUMN = ICON_FOLDER + "new_primitive_column.png";
  @StaticResource
  private static final String ICON_NEW_RELATION_NESTED = ICON_FOLDER + "new_relation_nested.png";
  @StaticResource
  private static final String ICON_RELATION_BACKDROP = ICON_FOLDER + "relation_backdrop.png";
  @StaticResource
  private static final String ICON_DATA_TABLE = ICON_FOLDER + "data_table.png";
  @StaticResource
  private static final String ICON_PERSPECTIVE = ICON_FOLDER + "perspective2.png";
  @StaticResource
  private static final String ICON_FONT_FILE = "com/sieuferd/project/icons/sieuf_dingbats.ttf";

  private static final class LazyIconFontHolder {
    private static final Font ICON_FONT = readIconFont();

    private static Font readIconFont() {
      try {
        final InputStream fontStream =
            QueryLabelFormatter.class.getClassLoader().getResourceAsStream(ICON_FONT_FILE);
        if (fontStream == null)
          throw new IOException("Failed to get badge font resource");
        return Font.createFont(Font.TRUETYPE_FONT, fontStream);
      } catch (IOException | FontFormatException e) {
        throw new RuntimeException("Failed to load badge font", e);
      }
    }
  }

  private SieuferdIcons() { }

  /** Properties to which changes affect the node icon for the path in question. */
  public static final ImmutableSet<ViewProperty<?>> ICON_PROPERTIES =
      ImmutableSet.<ViewProperty<?>>builder()
      .add(QProp.PRIMARY)
      .add(QProp.INSTANTIATED_TABLE)
      .add(QProp.INCLUDE_PRIMARY_KEY)
      .add(QProp.SQL_COLUMN_NAME)
      .add(CProp.PLURAL)
      .add(CProp.AFFORDANCE)
      .build();

  /**
   * Get the font that contains Sieuferd-specific icon dingbats. Should be safe to call from any
   * thread.
   */
  public static Font getIconFont() {
    return LazyIconFontHolder.ICON_FONT;
  }

  public static Image getViewIcon(final View view) {
    return ImageUtilities.loadImage(
        view.get(QProp.PRIMARY, Path.ROOT) ? ICON_DATA_TABLE : ICON_PERSPECTIVE);
  }

  public static Image getPathIcon(final View view, final Path path) {
    final ObjectReference<Image> ret = ObjectReference.create();
    view.performReadTransaction(() -> {
      ret.setValue(getPathIconHelper(view, path));
    });
    return ret.getValue();
  }

  private static Image getPathIconHelper(View view, Path path) {
    // Changes in this method may require ICON_PROPERTIES to be updated.
    final String name;
    // TODO[cc]: Factor out common logic with FieldContext.
    final boolean exists = view.exists(path);
    final boolean isFormula = !path.isRelation() && view.get(QProp.SQL_COLUMN_NAME, path).isEmpty();
    final boolean affordance = exists && view.get(CProp.AFFORDANCE, path);
    if (!exists) {
      // TODO[cc]: Consider using getOrDefault instead (but then I'd have to use ObservableView).
      if (path.isRelation()) {
        name = ICON_RELATION_NESTED;
      } else {
        name = ICON_PRIMITIVE_COLUMN;
      }
    } else if (affordance) {
      if (path.isRelation()) {
        name = ICON_NEW_RELATION_NESTED;
      } else {
        name = isFormula ? ICON_NEW_PRIMITIVE_FORMULA : ICON_NEW_PRIMITIVE_COLUMN;
      }
    } else if (path.isRelation()) {
      final boolean nested = view.get(CProp.PLURAL, path);
      final boolean noinst = view.get(QProp.INSTANTIATED_TABLE, path).getTable() == null;
      if (noinst) {
        if (nested) {
          name = ICON_RELATION_NESTED_NOINST;
        } else {
          name = ICON_RELATION_JOINED_NOINST;
        }
      } else {
        final boolean nopk = !view.get(QProp.INCLUDE_PRIMARY_KEY, path);
        if (nopk) {
          if (nested) {
            name = ICON_RELATION_NESTED_NOPK;
          } else {
            name = ICON_RELATION_JOINED_NOPK;
          }
        } else {
          if (nested) {
            name = ICON_RELATION_NESTED;
          } else {
            name = ICON_RELATION_JOINED;
          }
        }
      }
    } else {
      name = isFormula ? ICON_PRIMITIVE_FORMULA : ICON_PRIMITIVE_COLUMN;
    }
    Image ret = ImageUtilities.loadImage(name, false);
    if (!exists)
      ret = ImageUtilities.createDisabledImage(ret);
    if (path.isRelation() && !path.isRoot() && !affordance) {
      final Image backdrop = ImageUtilities.loadImage(ICON_RELATION_BACKDROP, false);
      ret = ImageUtilities.mergeImages(backdrop, ret, 0, 0);
    }
    return ret;
  }

  /* Based on org.netbeans.modules.maven.spi.nodes.NodeUtils.getTreeFolderIcon. Similar logic exists
  in org.openide.loaders.DataFolder.FolderNode. */

  /**
   * Find an icon suitable for use with folder-like nodes, possibly customized for the current
   * L&amp;F.
   *
   * @param opened wheter an icon should be returned for an opened or closed folder
   */
  public static Image getTreeFolderIcon(boolean opened) {
    Image ret = (Image) UIManager.get(
        opened ? "Nb.Explorer.Folder.openedIcon" : "Nb.Explorer.Folder.icon");
    if (ret == null) {
      Icon baseIcon = UIManager.getIcon(opened ? "Tree.openIcon" : "Tree.closedIcon");
      if (baseIcon != null) {
        ret = ImageUtilities.icon2Image(baseIcon);
      } else {
        ret = ImageUtilities.loadImage(opened ? ICON_DEFAULT_FOLDER_OPEN : ICON_DEFAULT_FOLDER);
      }
    }
    Preconditions.checkNotNull(ret);
    return ret;
  }
}
