// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.sieuferd.util.SieuferdXMLException;
import com.sieuferd.viewset.SieuferdProjectConfiguration;
import com.sieuferd.viewset.SieuferdProjectXML;
import com.sieuferd.project.datasource.SieuferdProjectCustomizerProvider;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import org.netbeans.api.project.Project;
import org.netbeans.spi.project.ProjectState;
import org.openide.filesystems.FileObject;
import org.openide.util.Lookup;
import org.openide.util.lookup.Lookups;

@ThreadSafe
public final class SieuferdProject implements Project {
  private static final Logger LOG = Logger.getLogger(SieuferdProject.class.getName());
  private static final String SIEUFERD_XML = "sieuferd.xml";
  private final FileObject projectDirectory;
  // TODO[pd]: Is ProjectState thread-safe?
  private final ProjectState projectState;
  private final SieuferdProjectConfiguration configuration = new SieuferdProjectConfiguration();
  private final Lookup lookup;
  private final NBViewSet viewSet;

  private SieuferdProject(FileObject projectDirectory, ProjectState projectState) {
    this.projectDirectory = projectDirectory;
    this.projectState = projectState;
    // See https://platform.netbeans.org/tutorials/nbm-povray-3.html .
    this.lookup = Lookups.fixed(new Object[] {
      this,
      projectState,
      new SieuferdProjectInformation(this),
      new SieuferdProjectLogicalView(this),
      new SieuferdProjectCustomizerProvider(this)
    });
    this.viewSet = new NBViewSet(this);
  }

  private static @Nullable FileObject getProjectXMLFile(FileObject projectDirectory) {
    return projectDirectory.getFileObject(SIEUFERD_XML);
  }

  public static boolean isProject(FileObject projectDirectory) {
    return getProjectXMLFile(projectDirectory) != null;
  }

  // TODO[pd]: Consider not parsing the project XML file on the EDT.
  /* My experiments indicate that this method is called from the EDT during the splash screen phase
  of the platform startup (if the project was previously opened), and when a project is loaded as a
  result of opening a DataObject file from a project that's not currently open. When the user opens
  a project through the "Open Project..." action, loadProject() is called in a separate thread. In
  the latter case, the project's node does not appear until the method completes, but the GUI
  remains responsive. */
  public static SieuferdProject loadProject(FileObject projectDirectory, ProjectState projectState)
      throws IOException
  {
    FileObject projectXMLFile = getProjectXMLFile(projectDirectory);
    /* It's essential to return null here rather than throwing an IOException; see the interface
    Javadoc. If an IOException is thrown instead, problems may arise during startup as editors are
    opened for queries before the project has been loaded. I'm not exactly sure why the latter
    happens, but it seems that as long as I return null here, the platform always waits until the
    project has been loaded before opening editors for files belonging to the project (confirmed by
    introducing artificial delays in this method). */
    if (projectXMLFile == null)
      return null;
    try {
      SieuferdProject ret = new SieuferdProject(projectDirectory, projectState);
      ret.getConfiguration().copyFrom(SieuferdProjectXML.readSieuferdProjectXML(
          projectXMLFile.getInputStream(), projectXMLFile.toURI().toString()));
      return ret;
    } catch (SieuferdXMLException e) {
      /* TODO[pd]: Consider catching IOException in general as well, or on the contrary, not
                   catching SieuferdXMLException either. */
      LOG.log(Level.WARNING, "Error loading Sieuferd project",  e);
      return null;
    }
  }

  public NBViewSet getViewSet() {
    return viewSet;
  }

  @Override
  public FileObject getProjectDirectory() {
    return projectDirectory;
  }

  @Override
  public Lookup getLookup() {
    return lookup;
  }

  public SieuferdProjectConfiguration getConfiguration() {
    return configuration;
  }
}
