// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project;

import com.sieuferd.project.viewfile.VSProp;
import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.Path;
import com.sieuferd.project.viewfile.ViewDataObject;
import com.sieuferd.project.viewfile.ViewURLComponents;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.viewset.GatedView;
import java.io.IOException;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.URLMapper;

/** Shallowly immutable. */
public final class ViewContext {
  private final SieuferdProject project;
  private final GatedView view;
  // It's OK to keep a DataObject around in a Node; after all, DataNode does so.
  private final ViewDataObject dataObject;

  /** Create a ViewContext for a view known to belong to a project. May block and should not be
  called from the EDT. */
  public static ViewContext create(SieuferdProject project, GatedView view) throws IOException {
    ThreadUtil.checkNotOnEDT();
    FileObject fileObject = URLMapper.findFileObject(new ViewURLComponents(
        project.getProjectDirectory().toURI(),
        view.get(VSProp.VIEW_IDENTIFIER, Path.ROOT)).toURL());
    if (fileObject == null)
      throw new IOException("Could not find FileObject for view " + view);
    ViewDataObject viewDataObject = ViewDataObject.forFileObject(fileObject);
    return new ViewContext(project, view, viewDataObject);
  }

  /** Create a ViewContext for a ViewDataObject. May block and should not be called from the EDT. */
  public static ViewContext create(ViewDataObject viewDataObject) throws IOException {
    ThreadUtil.checkNotOnEDT();
    ViewURLComponents viewURLComponents =
        ViewURLComponents.fromPreparedFileObject(viewDataObject.getPrimaryFile());
    if (viewURLComponents == null)
      throw new IOException("Failed to get view URL from ViewDataObject");
    SieuferdProject project = viewURLComponents.getProject();
    if (project == null)
      throw new IOException("Failed to get SiefuerdProject from ViewDataObject");
    GatedView view =
        project.getViewSet().loadViewByIdentifier(viewURLComponents.getViewIdentifier());
    if (view == null)
      throw new IOException("Could not find view " + viewURLComponents.getViewIdentifier());
    return new ViewContext(project, view, viewDataObject);
  }

  private ViewContext(SieuferdProject project, GatedView view, ViewDataObject dataObject) {
    Preconditions.checkNotNull(project);
    Preconditions.checkNotNull(view);
    Preconditions.checkNotNull(dataObject);
    this.project = project;
    this.view = view;
    this.dataObject = dataObject;
  }

  public SieuferdProject getProject() {
    return project;
  }

  public GatedView getView() {
    return view;
  }

  public ViewDataObject getDataObject() {
    return dataObject;
  }

  @Override
  public String toString() {
    return view.toString();
  }

  @Override
  public int hashCode() {
    return Objects.hashCode(project, view, dataObject);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof ViewContext))
      return false;
    ViewContext other = (ViewContext) obj;
    return Objects.equal(this.project, other.project)
        && Objects.equal(this.view, other.view)
        && Objects.equal(this.dataObject, other.dataObject);
  }
}
