// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.google.common.collect.Maps;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.proptypes.Position;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.query.QProp;
import com.sieuferd.util.MapUtil;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.math.BigDecimal;
import java.util.List;
import java.util.Map;
import javax.annotation.Nullable;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionRegistration;
import org.openide.util.NbBundle;

@ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.ClearSortingAction")
@ActionRegistration(displayName = "#CTL_ClearSortingAction")
@ActionReference(path = "Sieuferd/FieldContextMenu", position = 2500)
@NbBundle.Messages("CTL_ClearSortingAction=Clear Sorting")
public final class ClearSortingAction implements ActionListener {
  private final List<FieldContext> fieldContexts;

  public ClearSortingAction(List<FieldContext> fieldContexts) {
    this.fieldContexts = fieldContexts;
  }

  private void clearSortingForPath(MutableView view, Path path, Position originalPosition) {
    final @Nullable BigDecimal originalPositionBD = originalPosition.getPosition();
    if (path.isRelation() && originalPositionBD == null) {
      /* When clearing sorting on a relation that is not sorted-on, interpret the clear action as an
      intention to clear the sorted-on state of its children. */
      for (Path childPath : view.getChildPaths(path.getRelation()))
        view.set(QProp.ORDER_RANK, childPath, Position.NONE);
    } else {
      for (Path siblingPath : view.getChildPaths(path.getParent())) {
        final BigDecimal siblingPositionBD = view.get(QProp.ORDER_RANK, siblingPath).getPosition();
        if (siblingPositionBD != null &&
            (originalPositionBD == null || siblingPositionBD.compareTo(originalPositionBD) >= 0))
        {
          view.set(QProp.ORDER_RANK, siblingPath, Position.NONE);
        }
      }
    }
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    final FieldSelection fieldSelection =
        FieldSelection.fromFieldContexts(fieldContexts, AppliesTo.ANY);
    if (fieldSelection == null)
      return;
    fieldSelection.getViewContext().getView().mutate(Bundle.CTL_ClearSortingAction(), true,
        (MutableView view) ->
    {
      /* The effect of clearSortingForPath depends on the original order positions of the selected
      paths; store these positions before doing any mutations to avoid unintended behavior when the
      action is invoked with a multiple selection. */
      final Map<Path,Position> originalPositions = Maps.newLinkedHashMap();
      for (Path path : fieldSelection.getSelectedPaths())
        MapUtil.putChecked(originalPositions, path, view.get(QProp.ORDER_RANK, path));
      for (Path path : fieldSelection.getSelectedPaths())
        clearSortingForPath(view, path, MapUtil.getChecked(originalPositions, path));
    });
  }
}
