// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.Views;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.project.util.FieldSelection;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import java.util.Set;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionRegistration;
import org.openide.util.NbBundle.Messages;

@ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.HideAction")
@ActionRegistration(displayName = "#CTL_HideAction")
@ActionReference(path = "Sieuferd/FieldContextMenu", position = 1200)
@Messages("CTL_HideAction=Hide")
public final class HideAction implements ActionListener {
  private final ImmutableList<FieldContext> fieldContexts;

  public HideAction(List<FieldContext> fieldContexts) {
    this.fieldContexts = ImmutableList.copyOf(fieldContexts);
  }

  public static String getDisplayName() {
    return Bundle.CTL_HideAction();
  }

  private static boolean setContainsStrictAncestor(Set<Path> set, Path path) {
    if (path.isRoot()) {
      return false;
    } else {
      RelationPath parent = path.getParent();
      return set.contains(parent) || setContainsStrictAncestor(set, parent);
    }
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    final FieldSelection fieldSelection =
        FieldSelection.fromFieldContexts(fieldContexts, AppliesTo.ANY);
    if (fieldSelection == null)
      return;
    final Set<Path> usePaths =
        Sets.difference(fieldSelection.getSelectedPaths(), ImmutableSet.of(Path.ROOT));
    if (usePaths.isEmpty())
      return;
    fieldSelection.getViewContext().getView().mutate(Bundle.CTL_HideAction(), true,
        (MutableView view) ->
    {
      Set<Path> usePaths2 = Sets.newLinkedHashSet(Views.existingPaths(view, usePaths));
      /* Usability tweak: If the user hides all of a relation's visible children in a single
      operation, hide the common parent relation instead (and apply the same logic again recursively
      to the root. This is only done from the HideAction, not from ViewUpdateCoalescer, since doing
      it in the latter would make FieldsPanel checkboxes appear to behave oddly. */
      /* TODO[cc]: Do this recursively instead of iteratively to make it easier to see that there
      can't be an infinite loop. */
      boolean changed;
      do {
        Set<RelationPath> parentRelations = Sets.newLinkedHashSet();
        for (Path path : usePaths2) {
          if (!path.isRoot())
            parentRelations.add(path.getParent());
        }
        changed = false;
        for (RelationPath parentRelation : parentRelations) {
          final Set<Path> visibleChildren = Views.getVisibleChildren(view, parentRelation);
          if (!parentRelation.isRoot() && usePaths2.containsAll(visibleChildren)) {
            usePaths2.removeAll(visibleChildren);
            usePaths2.add(parentRelation);
            changed = true;
          }
        }
      } while (changed);
      for (Path path : usePaths2) {
        if (!setContainsStrictAncestor(usePaths2, path))
          ViewUpdater.userSetProperty(view, CProp.VISIBLE, path, false);
      }
    });
  }
}
