// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLTask;
import com.sieuferd.database.SQLTaskRunner;
import com.sieuferd.importer.AccessImporter;
import com.sieuferd.project.NBSQLTaskRunner;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.UnitObject;
import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.Action;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;
import org.netbeans.api.project.Project;
import org.netbeans.spi.project.ui.support.ProjectActionPerformer;
import org.netbeans.spi.project.ui.support.ProjectSensitiveActions;
import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import static org.openide.NotifyDescriptor.OK_OPTION;
import org.openide.awt.ActionID;
import org.openide.awt.ActionRegistration;
import org.openide.util.NbBundle.Messages;
import org.openide.util.RequestProcessor;
import org.openide.windows.IOProvider;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;
import org.openide.windows.WindowManager;

public final class ImportAccessAction {
  private static final Logger LOG = Logger.getLogger(ImportAccessAction.class.getName());
  private ImportAccessAction() { }

  @ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.ImportAccessAction")
  @ActionRegistration(displayName = "#CTL_ImportAccessAction")
  @Messages("CTL_ImportAccessAction=Import from Microsoft Access...")
  public static Action createImportAccessAction() {
    /* ProjectSensitiveActions are "always considered synchronous"; see
    https://netbeans.org/projects/platform/lists/dev/archive/2005-04/message/124 . */
    return ProjectSensitiveActions.projectSensitiveAction(new ProjectActionPerformer() {
      @Override
      public boolean enable(Project project) {
        return (project instanceof SieuferdProject);
      }

      @Override
      public void perform(final Project project) {
        if (project instanceof SieuferdProject) {
          SwingUtilities.invokeLater(() -> {
            doImport((SieuferdProject) project);
          });
        }
      }
    }, Bundle.CTL_ImportAccessAction(), null);
  }

  private static @Nullable File defaultFileChooserDirectory = null;

  private static final RequestProcessor RP_IMPORT = new RequestProcessor("Import Access Database");

  private static void doImport(final SieuferdProject project) {
    ThreadUtil.checkOnEDT();
    Set<SQLDatabase> databases = project.getConfiguration().getManagedDatabases();
    if (databases.size() != 1) {
      final NotifyDescriptor nd = new NotifyDescriptor(
          "To import, first ensure that exactly one \"Managed Schema\"\n" +
          "is selected in the project properties.", "Import from Microsoft Access",
          JOptionPane.DEFAULT_OPTION, NotifyDescriptor.INFORMATION_MESSAGE,
          new Object[] { OK_OPTION }, null);
      DialogDisplayer.getDefault().notify(nd);
      return;
    }
    final SQLDatabase outputDatabase = databases.iterator().next();
    JFileChooser fileChooser = new JFileChooser(defaultFileChooserDirectory);
    // See com.healthmarketscience.jackcess.Database.FileFormat.
    FileFilter fileFilter = new FileNameExtensionFilter(
        "Microsoft Access Databases", "mdb", "accdb", "mny");
    fileChooser.setAcceptAllFileFilterUsed(true);
    fileChooser.addChoosableFileFilter(fileFilter);
    fileChooser.setFileFilter(fileFilter);
    if (fileChooser.showOpenDialog(WindowManager.getDefault().getMainWindow()) !=
        JFileChooser.APPROVE_OPTION)
    {
      return;
    }
    final File inputFile = fileChooser.getSelectedFile();
    defaultFileChooserDirectory = inputFile.getParentFile();

    final InputOutput io = IOProvider.getDefault().getIO("Microsoft Access Import", false);
    final OutputWriter out = io.getOut();
    try {
      out.reset();
    } catch (IOException e) {
      LOG.log(Level.WARNING, "Failed to clear the import output window", e);
    }

    final NotifyDescriptor nd = new NotifyDescriptor(
          "<html>Tables in the following file will be imported:<br><br>" +
              SwingString.toSwingHTML(inputFile.getPath()) + "<br><br>" +
          "<b>Existing tables in database \"" +
              SwingString.toSwingHTML(outputDatabase.toString()) + "\" will be dropped!</b><br><br>"
              + "OK to proceed?</html>",
          "Import from Microsoft Access",
          JOptionPane.OK_CANCEL_OPTION, NotifyDescriptor.QUESTION_MESSAGE, null, null);
    if (!DialogDisplayer.getDefault().notify(nd).equals(NotifyDescriptor.OK_OPTION))
      return;

    io.select();
    out.println(
        "Importing to database \"" + outputDatabase + "\" from file \"" + inputFile + "\"...");

    RP_IMPORT.post(() -> {
      final SQLTaskRunner sqlTaskRunner =
          NBSQLTaskRunner.get(project.getConfiguration().getDatabaseConnectionName());
      try {
        sqlTaskRunner.runTask(true, true, new SQLTask<UnitObject>("Import from Microsoft Access") {
          @Override
          public UnitObject run(SQLConnection sqlConnection) throws SQLException {
            try {
              AccessImporter.importFile(inputFile, sqlConnection, outputDatabase, out);
            } catch (IOException e) {
              e.printStackTrace(out);
            }
            return UnitObject.INSTANCE;
          }
        });
      } catch (SQLException | SQLTask.UserInformedConnectionException e) {
        e.printStackTrace(out);
      }
      ViewUpdater.refreshPrimaryViews(project);
      io.getOut().println("Operation completed.");
      io.getOut().close();
    });
  }
}
