// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.sieuferd.project.util.ViewPropertyAction;
import com.sieuferd.project.util.SelectionSnapshot;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.ViewContext;
import com.sieuferd.upstream.UpstreamUtil;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import javax.swing.Action;
import org.netbeans.api.annotations.common.StaticResource;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionRegistration;
import org.openide.awt.DynamicMenuContent;
import org.openide.util.NbBundle;

public abstract class InsertCalculatedFieldAction implements ActionListener {
  private final FieldContext fieldContext;
  private final boolean after;

  public InsertCalculatedFieldAction(FieldContext fieldContext, boolean after) {
    Preconditions.checkNotNull(fieldContext);
    this.fieldContext = fieldContext;
    this.after = after;
  }

  @NbBundle.Messages("CTL_InsertCalculatedFieldAction=Insert Calculated Field")
  @Override
  public void actionPerformed(ActionEvent e) {
    if (fieldContext.isRoot())
      return;
    // Use the same display string for Undo/Redo purposes whether in before or after mode.
    fieldContext.getView().mutate(Bundle.CTL_InsertCalculatedFieldAction(), true,
        (MutableView view) ->
    {
      final RelationPath parentPath = fieldContext.getPath().getParent();
      final List<Path> orderedPaths = Lists.newArrayList(view.getChildPaths(parentPath));
      final PrimitivePath newPath = addCalculatedField(view, parentPath);
      final int selectedPathIndex = orderedPaths.indexOf(fieldContext.getPath());
      final int insertIndex = selectedPathIndex < 0
          ? (after ? orderedPaths.size() : 0)
          : (after ? (selectedPathIndex + 1) : selectedPathIndex);
      orderedPaths.add(insertIndex, newPath);
      view.set(SProp.CHILD_PATHS, parentPath, new PathList(orderedPaths));
    });
  }

  private static PrimitivePath addCalculatedField(MutableView view, RelationPath parentPath) {
    return view.addPrimitivePath(parentPath);
  }

  @StaticResource
  private static final String ICON_FORMULA = "com/sieuferd/project/icons/badge_formula.png";

  @ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.InsertCalculatedFieldBeforeAction")
  @ActionRegistration(displayName = "#CTL_InsertCalculatedFieldBeforeAction",
      iconBase = ICON_FORMULA)
  @ActionReference(path = "Sieuferd/FieldContextMenu", position = 5300)
  @NbBundle.Messages("CTL_InsertCalculatedFieldBeforeAction=Insert Calculated Field Before")
  public static final class InsertCalculatedFieldBeforeAction extends InsertCalculatedFieldAction {
    public InsertCalculatedFieldBeforeAction(FieldContext fieldContext) {
      super(fieldContext, false);
    }
  }

  @ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.InsertCalculatedFieldAfterAction")
  @ActionRegistration(displayName = "#CTL_InsertCalculatedFieldAfterAction",
      iconBase = ICON_FORMULA)
  @ActionReference(path = "Sieuferd/FieldContextMenu", position = 5400)
  @NbBundle.Messages("CTL_InsertCalculatedFieldAfterAction=Insert Calculated Field After")
  public static final class InsertCalculatedFieldAfterAction extends InsertCalculatedFieldAction {
    public InsertCalculatedFieldAfterAction(FieldContext fieldContext) {
      super(fieldContext, true);
    }
  }

  @ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.AddCalculatedFieldAction")
  @ActionRegistration(displayName = "#CTL_AddCalculatedFieldAction", lazy = false)
  @ActionReference(path = "Sieuferd/FieldContextMenu", position = 5500)
  @NbBundle.Messages("CTL_AddCalculatedFieldAction=Add Calculated Field as Child")
  public static final class AddCalculatedFieldAction extends ViewPropertyAction {
    public AddCalculatedFieldAction() {
      super(Bundle.CTL_AddCalculatedFieldAction(), null, AppliesTo.RELATION_ONLY,
          ImmutableSet.<ViewProperty<?>>of(SProp.CHILD_PATHS), true);
      putValue(DynamicMenuContent.HIDE_WHEN_DISABLED, true);
      putValue(Action.SMALL_ICON, UpstreamUtil.loadImageIcon(ICON_FORMULA));
      ensureBindingsInitialized();
    }

    @Override
    public ViewPropertyAction create() {
      return new AddCalculatedFieldAction();
    }

    @Override
    public void actionPerformed(
        ActionEvent e, final ViewContext viewContext, final ImmutableSet<Path> selectedPaths)
    {
      viewContext.getView().mutate(Bundle.CTL_AddCalculatedFieldAction(), true,
          (MutableView view) ->
      {
        for (Path path : selectedPaths) {
          if (path.isRelation())
            addCalculatedField(view, path.getRelation());
        }
      });
    }

    @Override
    protected boolean isEnabledFor(SelectionSnapshot snapshot) {
      for (Path path : snapshot.getSelectedPaths()) {
        if (!path.isRelation() ||
            !snapshot.getViewSnapshot().get(SProp.CHILD_PATHS, path).isEmpty())
        {
          return false;
        }
      }
      return true;
    }
  }
}
