// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.Views;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.actions.join.Constraints;
import com.sieuferd.project.actions.join.JoinDialogPanel;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.query.OptionalPrimitivePath;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.util.ObjectReference;
import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import javax.annotation.Nullable;
import org.openide.DialogDescriptor;
import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.awt.ActionRegistration;
import org.openide.util.Mutex;
import org.openide.util.NbBundle.Messages;

@ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.JoinAction")
@ActionRegistration(displayName = "#CTL_JoinAction",
    iconBase = "com/sieuferd/project/icons/badge_join.png")
@ActionReferences({
  @ActionReference(path = "Sieuferd/FieldContextMenu", position = 5100, separatorBefore = 5000),
  @ActionReference(path = "Keymaps/NetBeans", name = "D-J")
})
@Messages("CTL_JoinAction=Custom Join...")
public class JoinAction implements ActionListener {
  private static final boolean JOINED_RELATIONS_AT_END = true;
  private final List<FieldContext> fieldContexts;

  public JoinAction(List<FieldContext> fieldContexts) {
    this.fieldContexts = fieldContexts;
  }

  /** @param fieldContext can be null for debugging */
  @Messages("TIT_JoinDialog=Custom Join")
  private static void showDialog(final @Nullable List<FieldContext> fieldContexts) {
    final @Nullable FieldSelection fieldSelection;
    if (fieldContexts == null) {
      fieldSelection = null;
    } else {
      fieldSelection = FieldSelection.fromFieldContexts(fieldContexts, AppliesTo.ANY);
      if (fieldSelection == null)
        return;
    }
    Mutex.EVENT.postWriteRequest(() -> {
      final JoinDialogPanel joinDialogPanel = new JoinDialogPanel();
      if (fieldSelection != null)
        joinDialogPanel.setFieldSelection(fieldSelection);
      DialogDescriptor dialogDescriptor =
          new DialogDescriptor(joinDialogPanel, Bundle.TIT_JoinDialog());
      dialogDescriptor.setClosingOptions(
          new Object[] { NotifyDescriptor.CANCEL_OPTION, NotifyDescriptor.CLOSED_OPTION});
      final ObjectReference<Dialog> dialogRef = ObjectReference.create();
      dialogDescriptor.setButtonListener((ActionEvent evt) -> {
        if (NotifyDescriptor.OK_OPTION.equals(evt.getSource())) {
          /* Get the constraints before closing the dialog, since closing it can upset its
          selection state. */
          Constraints constraints = joinDialogPanel.getConstraints();
          if (constraints != null)
            performJoin(constraints);
          dialogRef.getValue().dispose();
        }
      });
      final Dialog dialog = DialogDisplayer.getDefault().createDialog(dialogDescriptor);
      dialogRef.setValue(dialog);
      dialog.setVisible(true);
      dialog.dispose();
    });
  }

  private static void performJoin(final Constraints constraints) {
    Preconditions.checkNotNull(constraints);
    final ImmutableView foreignView =
        ImmutableView.copyOf(constraints.getForeignFieldContext().getView());
    constraints.getBaseFieldContext().getView().mutate(Bundle.TIT_JoinDialog(), true,
        (MutableView view) ->
    {
      final RelationPath basePath = constraints.getBaseFieldContext().getPath();
      final RelationPath joinPath = view.addRelationPath(basePath);
      final PathMap pathMap = view.copyFrom(
          foreignView, constraints.getForeignFieldContext().getPath(), joinPath);
      Views.rewritePathReferences(view, joinPath, pathMap);
      Queries.showHeadingPathsOnly(view, joinPath);
      PrimitivePath lastJoinedOn = null;
      for (Path childPath : view.getChildPaths(joinPath)) {
        if (childPath.isRelation())
          continue;
        final PrimitivePath originalPath = pathMap.getOldPath(childPath.getPrimitive());
        PrimitivePath joinedOn = constraints.getPathPairs().get(originalPath);
        if (joinedOn != null && !joinedOn.getParent().equals(basePath))
          joinedOn = null;
        view.set(QProp.JOINED_ON, childPath, joinedOn == null ?
            OptionalPrimitivePath.absent() : OptionalPrimitivePath.of(joinedOn));
        if (joinedOn != null) {
          lastJoinedOn = joinedOn;
          if (view.get(CProp.VISIBLE, joinedOn) || view.get(CProp.VISIBLE, childPath)) {
            /* The value of the field is identical in the child and the parent, so display it in
            the parent only. */
            view.set(CProp.VISIBLE, joinedOn, true);
            view.set(CProp.VISIBLE, childPath, false);
          }
        }
      }
      if (!JOINED_RELATIONS_AT_END && lastJoinedOn != null) {
        List<Path> siblings = Lists.newArrayList(view.getChildPaths(basePath));
        int index = siblings.indexOf(lastJoinedOn);
        if (index < 0)
          throw new AssertionError();
        if (!siblings.remove(joinPath))
          throw new AssertionError();
        siblings.add(index + 1, joinPath);
        view.set(SProp.CHILD_PATHS, basePath, new PathList(siblings));
      }
      /* Any existential filter previously set was meant for the parent relation in the foreign
      relation, so it makes the most sense to deactivate it by default here. */
      view.set(QProp.EXISTENTIAL_FILTER, joinPath, false);
    });
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    UIUtil.instrumentation("JoinAction");
    showDialog(fieldContexts);
  }

  public static final void main(String args[]) {
    showDialog(null);
  }
}
