// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.query.Queries;
import com.sieuferd.uiutil.UIUtil;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;
import javax.annotation.Nullable;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import org.openide.awt.ActionID;
import org.openide.awt.ActionRegistration;
import org.openide.awt.Actions;
import org.openide.util.ContextAwareAction;
import org.openide.util.Lookup;
import org.openide.util.NbBundle.Messages;
import org.openide.util.Utilities;

/* These actions would be cumbersome to use from the context menu, and so are not exposed there.
They are appropriate for use by keyboard, or from the FieldsAction dialog box. */
public abstract class MoveAction implements ActionListener {
  private final ImmutableSet<FieldContext> fieldContexts;
  private final boolean forward;

  public MoveAction(List<FieldContext> fieldContexts, boolean forward) {
    this.fieldContexts = ImmutableSet.copyOf(fieldContexts);
    this.forward = forward;
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    final FieldSelection fieldSelection =
        FieldSelection.fromFieldContexts(fieldContexts, AppliesTo.ANY);
    if (fieldSelection == null)
      return;
    fieldSelection.getViewContext().getView().mutate(
        forward ? Bundle.CTL_MoveForwardAction() : Bundle.CTL_MoveBackwardAction(),
        true, (MutableView view) -> {
          Queries.move(view, fieldSelection.getSelectedPaths(), forward);
        }
    );
  }


  public static void connectMoveAction(final JComponent acceleratorComponent, final Lookup lookup,
      JButton button, boolean forward)
  {
    connectSieuferdAction(acceleratorComponent, lookup,
        "com.sieuferd.actionids." + (forward ? "MoveForwardAction" : "MoveBackwardAction"),
        button,
        forward ? "SD-DOWN" : "SD-UP");
  }

  private static void connectSieuferdAction(
      final JComponent acceleratorComponent, final Lookup lookup,
      final String actionID, JButton button, @Nullable String defaultAccelerator)
  {
    Action action = Actions.forID("Sieuferd", actionID);
    if (action instanceof ContextAwareAction)
      action = ((ContextAwareAction) action).createContextAwareInstance(lookup);
    KeyStroke accelerator;
    {
      Object toAccelerator = action.getValue(Action.ACCELERATOR_KEY);
      if (toAccelerator instanceof KeyStroke) {
        accelerator = (KeyStroke) toAccelerator;
      } else {
        accelerator = defaultAccelerator == null ? null : Utilities.stringToKey(defaultAccelerator);
        if (accelerator != null)
          action.putValue(Action.ACCELERATOR_KEY, accelerator);
      }
    }
    Actions.connect(button, action);
    if (accelerator != null) {
      UIUtil.getSafeInputMap(acceleratorComponent, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT)
          .put(accelerator, actionID);
      UIUtil.getSafeActionMap(acceleratorComponent).put(actionID, action);
    }
  }

  @ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.MoveBackwardAction")
  @ActionRegistration(
      displayName = "#CTL_MoveBackwardAction",
      iconBase = "com/sieuferd/project/icons/up.png")
  @Messages("CTL_MoveBackwardAction=Move Backward")
  public static final class MoveBackwardAction extends MoveAction {
    public MoveBackwardAction(List<FieldContext> fieldContexts) {
      super(fieldContexts, false);
    }
  }

  @ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.MoveForwardAction")
  @ActionRegistration(
      displayName = "#CTL_MoveForwardAction",
      iconBase = "com/sieuferd/project/icons/down.png")
  @Messages("CTL_MoveForwardAction=Move Forward")
  public static final class MoveForwardAction extends MoveAction {
    public MoveForwardAction(List<FieldContext> fieldContexts) {
      super(fieldContexts, true);
    }
  }
}
