// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.datamodel.proptypes.Position;
import com.sieuferd.project.ViewContext;
import com.sieuferd.query.Filter;
import com.sieuferd.query.InstanceDefinition;
import com.sieuferd.query.OptionalPrimitivePath;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.query.SQLTable;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.StringUtil;
import com.sieuferd.viewset.GatedView;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionRegistration;
import org.openide.util.RequestProcessor;
import org.openide.windows.IOProvider;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;
import org.openide.windows.TopComponent;
import org.openide.windows.WindowManager;

@ThreadSafe
final class TestCaseBuilder implements AutoCloseable {
  private static final ViewDefaults TEST_VIEW_DEFAULTS = ViewDefaults.builder()
      .setAll(CProp.DEFAULTS)
      .setAll(QProp.DEFAULTS)
      .build();
  private static final Set<ViewProperty<?>> OTHER_EXCLUDED_PROPERTIES =
      ImmutableSet.of(CProp.HEADING, CProp.INHERIT_STYLES, CProp.AFFORDANCE);
  private static final Map<GatedView,TestCaseBuilder> BUILDERS = new WeakHashMap<>();
  private final OutputWriter output;
  private final ViewContext viewContext;
  private @Nullable ImmutableView currentView;
  private final Map<Path,String> pathVariableNames = Maps.newLinkedHashMap();

  public TestCaseBuilder(ViewContext viewContext) {
    InputOutput io = IOProvider.getDefault().getIO("Generated Test Cases", false);
    try {
      io.getIn().close();
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
    this.output = io.getOut();
    this.viewContext = viewContext;
  }

  @Override
  public void close() {
    try {
      output.reset();
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
    output.close();
  }

  public synchronized void emit() {
    ThreadUtil.checkNotOnEDT();
    if (currentView == null) {
      output.println("MutableView view = createView();");
      MutableView toCurrentView = Views.newMutable(TEST_VIEW_DEFAULTS);
      toCurrentView.set(CProp.LABEL, Path.ROOT, "root");
      currentView = ImmutableView.copyOf(toCurrentView);
    }
    ImmutableView newCurrentView = ImmutableView.copyOf(viewContext.getView(),
        TEST_VIEW_DEFAULTS.getProperties(), null);
    generateViewUpdateCode(currentView, newCurrentView);
    // TODO[pd]: Consider executing the query and filling in the expected result.
    output.println("expectResult(view, true, \"[]\");");
    this.currentView = newCurrentView;
    output.flush();
  }

  private void generateViewUpdateCode(ImmutableView viewBefore, ImmutableView viewAfter) {
    Preconditions.checkNotNull(viewBefore);
    Preconditions.checkNotNull(viewAfter);
    Preconditions.checkNotNull(output);
    output.println("// ----- Test Case ------------------------------------------------");
    final Set<Path> addedPaths = Sets.newLinkedHashSet();
    final Set<Path> pendingFormulaPaths = Sets.newLinkedHashSet();
    final Set<Path> pendingFilterPaths = Sets.newLinkedHashSet();
    // TODO[pd]: Set labels first?
    copyFrom(viewAfter, viewBefore, Path.ROOT, ImmutableSet.of(),
        addedPaths, pendingFormulaPaths, pendingFilterPaths);
    for (Path pendingPath : pendingFormulaPaths) {
      setProperty(QProp.FORMULA_EXPRESSION, viewAfter, viewBefore, pendingPath,
          ImmutableSet.copyOf(addedPaths), ImmutableSet.of());
    }
    for (Path pendingPath : pendingFilterPaths) {
      setProperty(QProp.FILTER, viewAfter, viewBefore, pendingPath,
          ImmutableSet.copyOf(addedPaths), ImmutableSet.of());
    }
  }

  private void copyFrom(final ImmutableView source, final ImmutableView dest, final Path path,
      final ImmutableSet<ViewProperty<?>> alreadySet,
      final Set<Path> addedPaths, final Set<Path> pendingFormulaPaths,
      final Set<Path> pendingFilterPaths)
  {
    Preconditions.checkNotNull(source);
    Preconditions.checkNotNull(dest);
    Preconditions.checkNotNull(path);
    Preconditions.checkNotNull(alreadySet);

    if (!pathVariableNames.containsKey(Path.ROOT))
      createPathVariableName(source, Path.ROOT);

    {
      final Set<ViewProperty<?>> definedProperties = source.getDefinedProperties(path);
      for (ViewProperty<?> property : definedProperties) {
        if (!SProp.CHILD_PATHS.equals(property) &&
            !alreadySet.contains(property) &&
            !OTHER_EXCLUDED_PROPERTIES.contains(property))
        {
          final Set<Path> pendingPathSet;
          if (QProp.FILTER.equals(property)) {
            pendingPathSet = pendingFilterPaths;
          } else if (QProp.FORMULA_EXPRESSION.equals(property)) {
            pendingPathSet = pendingFormulaPaths;
          } else {
            pendingPathSet = ImmutableSet.of();
          }
          setProperty(
              property, source, dest, path, ImmutableSet.copyOf(addedPaths), pendingPathSet);
        }
      }
    }
    if (path.isRelation()) {
      final Set<Path> existingPaths = dest.exists(path)
          ? ImmutableSet.copyOf(dest.getChildPaths(path.getRelation()))
          : ImmutableSet.of();
      final PathList sourcePaths = source.getChildPaths(path.getRelation());
      final List<Path> destPathOrder = Lists.newArrayList(existingPaths);
      for (Path childPath : sourcePaths) {
        final Set<ViewProperty<?>> childAlreadySet = Sets.newLinkedHashSet();
        final boolean alreadyExists = existingPaths.contains(childPath);
        if (!alreadyExists) {
          createPathVariableName(source, childPath);
          final String label = source.get(CProp.LABEL, childPath);
          if (childPath.isRelation()) {
            output.println("RelationPath " + getPathCode(childPath) + " = addRelation(view, " +
                getPathCode(path) + ", " + StringUtil.quoteJavaLike(label) + ");");
            childAlreadySet.add(CProp.LABEL);
          } else if (source.get(QProp.SQL_COLUMN_NAME, childPath).isEmpty()) {
            String expressionStringCode = getExpressionStringCode(
                source, dest, addedPaths, pendingFormulaPaths,
                source.get(QProp.FORMULA_EXPRESSION, childPath), childPath, false);
            if (expressionStringCode == null)
              expressionStringCode = "\"null\"";
            output.println("PrimitivePath " + getPathCode(childPath) +
                " = addFormulaColumn(view, " + getPathCode(path) + ", " +
                StringUtil.quoteJavaLike(label) + ", " + expressionStringCode + ");");
            childAlreadySet.add(CProp.LABEL);
            childAlreadySet.add(QProp.FORMULA_EXPRESSION);
          } else {
            final String sqlColumnName = source.get(QProp.SQL_COLUMN_NAME, childPath);
            output.println("PrimitivePath " + getPathCode(childPath) + " = addSQLColumn(view, " +
                getPathCode(path) + ", " + StringUtil.quoteJavaLike(sqlColumnName) + ");");
            childAlreadySet.add(QProp.SQL_COLUMN_NAME);
            if (sqlColumnName.equals(label))
              childAlreadySet.add(CProp.LABEL);
          }
          destPathOrder.add(childPath);
          addedPaths.add(childPath);
        }
        copyFrom(source, dest, childPath, ImmutableSet.copyOf(childAlreadySet),
            addedPaths, pendingFormulaPaths, pendingFilterPaths);
      }
      for (Path superfluousPath :
          Sets.difference(existingPaths, ImmutableSet.copyOf(sourcePaths)))
      {
        output.println("view.removePath(" + getPathCode(superfluousPath) + ");");
        destPathOrder.remove(superfluousPath);
        addedPaths.remove(superfluousPath);
      }
      if (path.equals(path)) {
        if (dest.exists(path) && !destPathOrder.equals(Lists.newArrayList(sourcePaths))) {
          List<String> pathCodes = Lists.newArrayList();
          for (Path orderPath : sourcePaths)
            pathCodes.add(getPathCode(orderPath));
          output.println("view.set(SProp.CHILD_PATHS, " + getPathCode(path) +
              ", new PathList(ImmutableList.of(" + Joiner.on(", ").join(pathCodes) + ")));");
        }
      }
    }
  }

  private <T> void setProperty(
      ViewProperty<T> property, ImmutableView source, ImmutableView dest, Path path,
      ImmutableSet<Path> addedPaths, Set<Path> pendingPaths)
  {
    final T value = source.get(property, path);
    if (!property.appliesTo(path) ||
        value.equals(dest.exists(path)
          ? dest.get(property, path) : TEST_VIEW_DEFAULTS.get(property)))
    {
      return;
    }
    final String pathCode = getPathCode(path);
    if (QProp.INSTANTIATED_TABLE.equals(property)) {
      InstanceDefinition instanceDefinition = (InstanceDefinition) value;
      final SQLTable table = (SQLTable) instanceDefinition.getTable();
      output.println("setInstantiatedTable(view, " + pathCode + ", " +
          (table == null ? "null" : StringUtil.quoteJavaLike(table.getTableName())) + ");");
    } else if (QProp.FORMULA_EXPRESSION.equals(property)) {
      String expressionStringCode = getExpressionStringCode(source, dest, addedPaths, pendingPaths,
          (Expression) value, path, false);
      if (expressionStringCode != null)
        output.println("setFormula(view, " + pathCode + ", " + expressionStringCode + ");");
    } else if (QProp.FILTER.equals(property)) {
      Filter filter = (Filter) value;
      // setFilter is the only helper that uses a path-local Strepco.
      String expressionStringCode = getExpressionStringCode(source, dest, addedPaths, pendingPaths,
          filter.toExpression(), path, true);
      if (expressionStringCode != null)
        output.println("setFilter(view, " + pathCode + ", " + expressionStringCode + ");");
    } else if (!property.isUserReadOnly()) {
      String propCode = getPropertyCode(property);
      String valueCode = getValueCode(value);
      output.println("view.set(" + propCode + ", " + pathCode + ", " + valueCode + ");");
    }
  }

  private static String getPropertyCode(ViewProperty<?> property) {
    return property.getPropertySet().getClass().getSimpleName() + "." + property.toString();
  }

  private String getValueCode(Object value) {
    Preconditions.checkNotNull(value);
    if (value instanceof String) {
      return StringUtil.quoteJavaLike((String) value);
    } else if (value instanceof Boolean) {
      return Boolean.TRUE.equals(value) ? "true" : "false";
    } else if (value instanceof Position) {
      final @Nullable BigDecimal position = ((Position) value).getPosition();
      return position == null
          ? "Position.NONE"
          : ("Position.create(new BigDecimal(" +
            StringUtil.quoteJavaLike(position.toString()) + "))");
    } else if (value instanceof OptionalPrimitivePath) {
      final OptionalPrimitivePath opp = (OptionalPrimitivePath) value;
      return opp.isPresent()
          ? "OptionalPrimitivePath.of(" + getPathCode(opp.getPath()) + ")"
          : "OptionalPrimitivePath.absent()";
    } else {
      throw new UnsupportedOperationException(
          "Can't encode property value of type " + value.getClass());
    }
  }

  private static String capitalizeFirst(String s) {
    if (s.isEmpty()) {
      return s;
    } else {
      return s.substring(0, 1).toUpperCase(Locale.US) + s.substring(1, s.length());
    }
  }

  private void createPathVariableNameInternal(Path path, String baseName) {
    for (int i = 1; ; i++) {
      final String fullName = baseName + (i == 1 ? "" : i);
      if (!pathVariableNames.containsValue(fullName)) {
        MapUtil.putChecked(pathVariableNames, path, fullName);
        break;
      }
    }
  }

  private static String getUnprefixed(String s, String prefix) {
    return s.startsWith(prefix) ? s.substring(prefix.length()) : s;
  }

  private static String getUnpostfixed(String s, String postfix) {
    return s.endsWith(postfix) ? s.substring(0, s.length() - postfix.length()) : s;
  }

  private void createPathVariableName(ImmutableView source, Path path) {
    Preconditions.checkState(!pathVariableNames.containsKey(path));
    if (path.isRoot()) {
      /* Give the root table a name as well, for the purposes of its column name, even though we
      will use Path.ROOT to refer to its path. The label is usually just "root" for the root, so
      base the name on the instantiated table instead. */
      final SQLTable sqlTable = (SQLTable) source.get(QProp.INSTANTIATED_TABLE, path).getTable();
      String tabName;
      if (sqlTable == null) {
        tabName = "U";
      } else {
        tabName = getUnprefixed(sqlTable.getTableName(), "tab");
      }
      createPathVariableNameInternal(path, "tab" + capitalizeFirst(tabName));
      return;
    }
    if (path.isRelation()) {
      final String tabName = getUnprefixed(source.get(CProp.LABEL, path), "tab");
      createPathVariableNameInternal(path, "tab" + capitalizeFirst(tabName));
    } else {
      final String tabName =
          getUnprefixed(MapUtil.getChecked(pathVariableNames, path.getParent()), "tab");
      String colName = source.get(CProp.LABEL, path);
      colName = getUnprefixed(colName, "pk_");
      colName = getUnpostfixed(colName, "_n");
      colName = getUnprefixed(colName, "col");
      createPathVariableNameInternal(path, "col" + capitalizeFirst(tabName) + "_" + colName);
    }
  }

  private @Nullable String getExpressionStringCode(
      ImmutableView source, ImmutableView dest, Set<Path> addedPaths, Set<Path> pendingPaths,
      Expression expression, Path path, boolean localizedStrepco)
  {
    for (Path referencedPath : Queries.getDirectlyReferencedPaths(expression)) {
      if (!dest.exists(referencedPath) && !addedPaths.contains(referencedPath)) {
        pendingPaths.add(path);
        return null;
      }
    }
    Preconditions.checkNotNull(source);
    Preconditions.checkNotNull(expression);
    Preconditions.checkNotNull(path);
    Strepco strepco = localizedStrepco
        ? Strepco.createNormalized(source, Views.getContextPath(path))
        : Strepco.createNormalized(source);
    return StringUtil.quoteJavaLike(Expressions.generateString(strepco, expression));
  }

  private String getPathCode(Path path) {
    if (path.isRoot())
      return "Path.ROOT";
    return MapUtil.getChecked(pathVariableNames, path);
  }

  static synchronized void closeForViewContext(ViewContext viewContext) {
    TestCaseBuilder b = BUILDERS.remove(viewContext.getView());
    if (b != null)
      b.close();
  }

  static synchronized TestCaseBuilder getForViewContext(ViewContext viewContext) {
    final GatedView gatedView = viewContext.getView();
    TestCaseBuilder ret = BUILDERS.get(gatedView);
    if (ret == null) {
      ret = new TestCaseBuilder(viewContext);
      MapUtil.putChecked(BUILDERS, gatedView, ret);
    }
    return ret;
  }

  @ActionID(category = "Sieuferd/Debug",
      id = "com.sieuferd.actionids.debug.ClearTestCaseBuilderAction")
  @ActionRegistration(displayName = "Test Case Builder: Clear State")
  @ActionReference(path = "Keymaps/NetBeans", name = "CS-R")
  public static final class ClearTestCaseBuilderAction implements ActionListener {
    private final ViewContext viewContext;

    public ClearTestCaseBuilderAction(ViewContext viewContext) {
      Preconditions.checkNotNull(viewContext);
      this.viewContext = viewContext;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
      TestCaseBuilder.closeForViewContext(viewContext);
    }
  }

  @ActionID(category = "Sieuferd/Debug",
      id = "com.sieuferd.actionids.EmitTestCaseAction")
  @ActionRegistration(displayName = "Test Case Builder: Emit Test Case")
  @ActionReference(path = "Keymaps/NetBeans", name = "CS-T")
  public static final class EmitTestCaseAction implements ActionListener {
    private static final RequestProcessor RP = new RequestProcessor("Test Case Generator", 1);
    private final ViewContext viewContext;

    public EmitTestCaseAction(ViewContext viewContext) {
      Preconditions.checkNotNull(viewContext);
      this.viewContext = viewContext;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
      final TopComponent tc = WindowManager.getDefault().findTopComponent("output");
      if (tc != null) {
        tc.open();
        tc.requestActive();
      }
      RP.post(TestCaseBuilder.getForViewContext(viewContext)::emit);
    }
  }
}
