// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.project.util.SelectionSnapshot;
import com.sieuferd.project.util.ViewPropertyAction;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.Set;
import javax.annotation.Nullable;
import javax.swing.Action;
import org.netbeans.api.annotations.common.StaticResource;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionRegistration;
import org.openide.awt.DynamicMenuContent;
import org.openide.util.NbBundle;

@ActionID(category = "Sieuferd", id = "com.sieuferd.actionids.UnhideInfluencersAction")
@ActionRegistration(displayName = "#CTL_UnhideInfluencersAction_Referenced", lazy = false)
@ActionReference(path = "Sieuferd/FieldContextMenu", position = 1400)
@NbBundle.Messages({
  "CTL_UnhideInfluencersAction_Referenced=Unhide Referenced",
  "CTL_UnhideInfluencersAction_Sorted=Unhide Sorted",
  "CTL_UnhideInfluencersAction_Filtered=Unhide Filtered",
  "CTL_UnhideInfluencersAction_SortedFiltered=Unhide Sorted/Filtered"
})
public final class UnhideInfluencersAction extends ViewPropertyAction {
  @StaticResource
  private static final String ICON_HIDE = "com/sieuferd/project/icons/badge_hide.png";
  private static final ImmutableMap<Set<Type>,String> DISPLAY_NAMES =
      ImmutableMap.<Set<Type>,String>builder()
      .put(ImmutableSet.of(Type.REFERENCED), Bundle.CTL_UnhideInfluencersAction_Referenced())
      .put(ImmutableSet.of(Type.SORTED), Bundle.CTL_UnhideInfluencersAction_Sorted())
      .put(ImmutableSet.of(Type.FILTERED), Bundle.CTL_UnhideInfluencersAction_Filtered())
      .put(ImmutableSet.of(Type.SORTED, Type.FILTERED),
          Bundle.CTL_UnhideInfluencersAction_SortedFiltered())
      .build();

  private enum Type {
    REFERENCED, SORTED, FILTERED
  }

  public UnhideInfluencersAction() {
    super(Bundle.CTL_UnhideInfluencersAction_Referenced(), ICON_HIDE, AppliesTo.ANY,
        ImmutableSet.<ViewProperty<?>>builder()
            .addAll(QueryUtil.QUERY_PROPERTIES)
            .add(CProp.PLURAL)
            .build(),
        false);
    putValue(DynamicMenuContent.HIDE_WHEN_DISABLED, true);
    ensureBindingsInitialized();
  }

  @Override
  protected String getDisplayNameFor(SelectionSnapshot selectionSnapshot) {
    final Set<Type> types = Sets.newLinkedHashSet();
    for (Path selectedPath : selectionSnapshot.getSelectedPaths())
      getHiddenInfluencers(selectionSnapshot.getViewSnapshot(), selectedPath, types);
    String ret = DISPLAY_NAMES.get(types);
    return ret != null ? ret : Bundle.CTL_UnhideInfluencersAction_Referenced();
  }

  @Override
  public ViewPropertyAction create() {
    return new UnhideInfluencersAction();
  }

  @Override
  public void actionPerformed(
      ActionEvent e, ViewContext viewContext, final ImmutableSet<Path> selectedPaths)
  {
    viewContext.getView().mutate((String) getValue(Action.NAME), true, (MutableView view) -> {
      for (Path selectedPath : selectedPaths) {
        if (view.exists(selectedPath)) {
          for (Path influencerPath : getHiddenInfluencers(view, selectedPath, null))
            ViewUpdater.userUnhide(view, influencerPath);
        }
      }
    });
  }

  @Override
  public boolean isEnabledFor(SelectionSnapshot snapshot) {
    for (Path path : snapshot.getSelectedPaths()) {
      if (isEnabledFor(snapshot.getViewSnapshot(), path))
        return true;
    }
    return false;
  }

  public static boolean isEnabledFor(View view, Path path) {
    return !getHiddenInfluencers(view, path, null).isEmpty();
  }

  private static Set<Path> getHiddenInfluencers(View view, Path path, @Nullable Set<Type> retTypes)
  {
    final Set<Path> ret = Sets.newLinkedHashSet();
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation())) {
        if (!view.get(CProp.VISIBLE, childPath)) {
          /* Duplicates the logic in Queries.doesPathInfluenceParent, but here I need to distinguish
          sorting and filtering for the purposes of populating retTypes to set the action's display
          name properly. */
          final boolean sorted = Queries.hasSignificantSort(view, childPath);
          final boolean filtered = childPath.isRelation()
              ? view.get(QProp.EXISTENTIAL_FILTER, childPath)
              : view.get(QProp.FILTER, childPath).isConstrained();
          if (sorted || filtered)
            ret.add(childPath);
          if (retTypes != null && sorted)
            retTypes.add(Type.SORTED);
          if (retTypes != null && filtered)
            retTypes.add(Type.FILTERED);
        }
      }
    } else {
      for (Path refPath :
          Queries.getDirectlyReferencedPaths(view.get(QProp.FORMULA_EXPRESSION, path)))
      {
        if (view.exists(refPath) && !Queries.isVisibleToRoot(view, refPath)) {
          ret.add(refPath);
          if (retTypes != null)
            retTypes.add(Type.REFERENCED);
        }
      }
    }
    return Collections.unmodifiableSet(ret);
  }
}
