// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions.join;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.uiutil.flow.LookupResultSnapshot;
import com.sieuferd.project.FieldContext.PrimitiveFieldContext;
import com.sieuferd.project.FieldContext.RelationFieldContext;
import com.sieuferd.util.MapUtil;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.concurrent.Immutable;

// TODO[cc]: Consider renaming to JoinParameters.
@Immutable
public class Constraints {
  private final RelationFieldContext baseFieldContext;
  private final RelationFieldContext foreignFieldContext;
  /** Keys are foreign table fields, values are base table fields. */
  private final ImmutableMap<PrimitivePath,PrimitivePath> pathPairs;

  /** @param pathPairs map where keys are foreign table fields and values are base table fields */
  public Constraints(
      RelationFieldContext baseFieldContext, RelationFieldContext foreignFieldContext,
      Map<PrimitivePath,PrimitivePath> pathPairs)
  {
    Preconditions.checkNotNull(baseFieldContext);
    Preconditions.checkNotNull(foreignFieldContext);
    this.baseFieldContext = baseFieldContext;
    this.foreignFieldContext = foreignFieldContext;
    this.pathPairs = ImmutableMap.copyOf(pathPairs);
  }

  public RelationFieldContext getBaseFieldContext() {
    return baseFieldContext;
  }

  public RelationFieldContext getForeignFieldContext() {
    return foreignFieldContext;
  }

  /** @return map where keys are foreign table fields and values are base table fields */
  public Map<PrimitivePath,PrimitivePath> getPathPairs() {
    return pathPairs;
  }

  @Override
  public String toString() {
    return "ConstraintsSelection(" +
        baseFieldContext + ", " + foreignFieldContext + ", " + pathPairs + ")";
  }

  private static List<PrimitivePath> getSelectedChildPathsInOrder(
      final RelationFieldContext parentContext,
      final LookupResultSnapshot<PrimitiveFieldContext> selection)
  {
    final Set<Path> selectedPaths = Sets.newLinkedHashSet();
    for (PrimitiveFieldContext fieldContext : selection.allInstances())
      selectedPaths.add(fieldContext.getPath());
    final List<PrimitivePath> ret = Lists.newArrayList();
    parentContext.getView().performReadTransaction(() -> {
      for (Path childPath : parentContext.getView().getChildPaths(parentContext.getPath())) {
        if (!childPath.isRelation() && selectedPaths.contains(childPath))
          ret.add(childPath.getPrimitive());
      }
    });
    return Collections.unmodifiableList(ret);
  }

  public static Constraints create(RelationFieldContext baseRelationContext,
      RelationFieldContext foreignRelationContext,
      LookupResultSnapshot<PrimitiveFieldContext> baseFieldsContexts,
      LookupResultSnapshot<PrimitiveFieldContext> foreignFieldsContexts)
  {
    final List<PrimitivePath> baseFieldsInOrder =
        getSelectedChildPathsInOrder(baseRelationContext, baseFieldsContexts);
    final List<PrimitivePath> foreignFieldsInOrder =
        getSelectedChildPathsInOrder(foreignRelationContext, foreignFieldsContexts);

    final Map<PrimitivePath,PrimitivePath> constraints = Maps.newLinkedHashMap();
    for (int i = 0; i < Math.min(baseFieldsInOrder.size(), foreignFieldsInOrder.size()); i++)
      MapUtil.putChecked(constraints, foreignFieldsInOrder.get(i), baseFieldsInOrder.get(i));

    return new Constraints(
        baseRelationContext, foreignRelationContext, constraints);
  }
}
