// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.actions.join;

import com.sieuferd.uiutil.ExplorerManagerProviderPanel;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.Views;
import com.sieuferd.flow.Flow;
import com.sieuferd.uiutil.flow.LookupResultSnapshot;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.flow.Flows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.FieldContext.PrimitiveFieldContext;
import com.sieuferd.project.FieldContext.RelationFieldContext;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.actions.join.ArrowWidgetModel.Arrow;
import com.sieuferd.project.navigator.FieldsPanel;
import com.sieuferd.project.nodes.FieldNode;
import com.sieuferd.project.nodes.FieldNodePathFinder;
import com.sieuferd.project.nodes.NodeSettings;
import com.sieuferd.project.nodes.SieuferdProjectNode;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.uiutil.EnterFilterInputMap;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.upstream.toolbar.ToolbarUtil;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import java.awt.CardLayout;
import java.awt.Point;
import java.awt.Rectangle;
import java.beans.BeanInfo;
import java.beans.PropertyVetoException;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.JPanel;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.openide.explorer.view.OutlineView;
import org.openide.nodes.Node;
import org.openide.util.ChangeSupport;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.RequestProcessor;

public class JoinDialogPanel extends JPanel {
  private static final RequestProcessor RP = new RequestProcessor("Join Dialog UI Node Operations");
  private static final NodeSettings PRIMITIVE_NODE_SETTINGS =
      new NodeSettings(AppliesTo.PRIMITIVE_ONLY, true, true, false, false, null, false);
  private static final Logger LOG = Logger.getLogger(JoinDialogPanel.class.getName());
  private final SettableFlow<RelationFieldContext> baseTableSettable = Flows.newSettable();
  private final Flow<RelationFieldContext> foreignTableFlow;
  private final Flow<Constraints> selectedConstraintsFlow;
  /** Keys are foreign table fields, values are base table fields. */
  private final Map<PrimitivePath,PrimitivePath> savedPathPairs = Maps.newLinkedHashMap();
  private final ArrowWidgetModel arrowWidgetModel = new ArrowWidgetModel() {
    @Override
    public void addChangeListener(ChangeListener changeListener) {
      arrowWidgetChangeSupport.addChangeListener(changeListener);
    }

    @Override
    public void removeChangeListener(ChangeListener changeListener) {
      arrowWidgetChangeSupport.removeChangeListener(changeListener);
    }

    @Override
    public List<Arrow> getArrows() {
      return calculateArrows();
    }
  };
  private final ChangeSupport arrowWidgetChangeSupport = new ChangeSupport(arrowWidgetModel);
  private final ChangeListener fieldListsViewportListener = (ChangeEvent e) -> {
    applyOversizeViewports();
    arrowWidgetChangeSupport.fireChange();
  };

  public JoinDialogPanel() {
    ThreadUtil.checkOnEDT();
    initComponents();

    EnterFilterInputMap.configure(baseFieldsListView.getList());
    EnterFilterInputMap.configure(foreignFieldsListView.getList());
    EnterFilterInputMap.configure(foreignTableOutlineView.getOutline());

    UIUtil.ensureSelectionOnFocusGained(baseFieldsListView.getList());
    UIUtil.ensureSelectionOnFocusGained(foreignFieldsListView.getList());
    UIUtil.ensureSelectionOnFocusGained(foreignTableOutlineView.getOutline());
    baseFieldsListView.getViewport().addChangeListener(fieldListsViewportListener);
    foreignFieldsListView.getViewport().addChangeListener(fieldListsViewportListener);
    ToolbarUtil.setBorderOnHover(addButton);
    ToolbarUtil.setBorderOnHover(removeButton);
    FieldsPanel.configureFieldsOutlineView(foreignTableOutlineView, false);
    foreignTableOutlineView.getOutline().setRootVisible(false);
    foreignTableOutlineView.getOutline().setTableHeader(null);
    foreignTableOutlineView.getOutline().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    /* Don't disable horizontal scollbars; this caused individual outline labels to be broken,
    creating visual artifacts.

    foreignTableOutlineView.setTreeHorizontalScrollBarPolicy(
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    */
    foreignTableOutlineView.setDragSource(false);
    foreignTableOutlineView.setDropTarget(false);
    foreignTableOutlineView.getOutline().setDragEnabled(false);

    Flow<LookupResultSnapshot<PrimitiveFieldContext>> baseFieldsValue =
        UIFlows.newLookup(baseFieldsExplorerPanel.getLookup(),
        new Lookup.Template<>(PrimitiveFieldContext.class));
    Flow<LookupResultSnapshot<PrimitiveFieldContext>> foreignFieldsValue =
        UIFlows.newLookup(foreignFieldsExplorerPanel.getLookup(),
        new Lookup.Template<>(PrimitiveFieldContext.class));

    foreignTableFlow =
        UIFlows.newSingularLookup(foreignTableExplorerPanel.getLookup(), RelationFieldContext.class);
    UIFlows.bindEDT(foreignTableFlow, (RelationFieldContext value) -> {
      Node rootNode = (value == null) ? Node.EMPTY :
          new FieldNode(value, PRIMITIVE_NODE_SETTINGS);
      ((CardLayout) foreignFieldsExplorerPanel.getLayout()).show(foreignFieldsExplorerPanel,
          value == null ? "noForeignTableSelected" : "foreignTableSelected");
      foreignFieldsExplorerPanel.getExplorerManager().setRootContext(rootNode);
      applyOversizeViewports();
      foreignFieldsListView.scrollToBeginning();
    });
    selectedConstraintsFlow = Flows.newComposite4(
        baseTableSettable, foreignTableFlow, baseFieldsValue, foreignFieldsValue,
        (baseRelationContext, foreignRelationContext, baseFieldsContexts, foreignFieldsContexts) ->
        Constraints.create(
        baseRelationContext, foreignRelationContext, baseFieldsContexts, foreignFieldsContexts));
    UIFlows.bindEDT(selectedConstraintsFlow, (Constraints value) -> {
      arrowWidgetChangeSupport.fireChange();
      addButton.setEnabled(value != null && !value.getPathPairs().isEmpty());
      removeButton.setEnabled(value != null && !value.getPathPairs().isEmpty());
    });
  }

  public @Nullable Constraints getConstraints() {
    Constraints selectedConstraints = selectedConstraintsFlow.get();
    if (selectedConstraints == null)
      return null;
    Map<PrimitivePath,PrimitivePath> pathPairs = Maps.newLinkedHashMap(savedPathPairs);
    pathPairs.putAll(selectedConstraints.getPathPairs());
    return new Constraints(selectedConstraints.getBaseFieldContext(),
        selectedConstraints.getForeignFieldContext(), pathPairs);
  }

  private List<Arrow> calculateArrows() {
    final Map<PrimitivePath,PrimitivePath> allPathPairs;
    final Map<PrimitivePath,PrimitivePath> selectedPathPairs;
    {
      Constraints allConstraints = getConstraints();
      allPathPairs = allConstraints == null ? ImmutableMap.of() : allConstraints.getPathPairs();
      Constraints selectedConstraints = selectedConstraintsFlow.get();
      selectedPathPairs = selectedConstraints == null ?
          ImmutableMap.of() : selectedConstraints.getPathPairs();
    }
    List<Arrow> ret = Lists.newArrayList();
    for (Entry<PrimitivePath,PrimitivePath> entry : allPathPairs.entrySet()) {
      Integer y1 = calculateFieldYCoordinate(
          baseFieldsExplorerPanel, baseFieldsListView, entry.getValue());
      Integer y2 = calculateFieldYCoordinate(
          foreignFieldsExplorerPanel, foreignFieldsListView, entry.getKey());
      if (y1 == null || y2 == null)
        continue;
      ret.add(new Arrow(y1, y2, entry.getValue().equals(selectedPathPairs.get(entry.getKey()))));
    }
    return Collections.unmodifiableList(ret);
  }

  private @Nullable Integer calculateFieldYCoordinate(
      ExplorerManagerProviderPanel panel, ListViewEx listView, PrimitivePath path)
  {
    Node rootNode = panel.getExplorerManager().getRootContext();
    int rowIndex = 0;
    for (Node childNode : rootNode.getChildren().getNodes(true)) {
      final FieldContext fieldContext = childNode.getLookup().lookup(FieldContext.class);
      if (fieldContext != null && fieldContext.getPath().equals(path)) {
        final Rectangle cellBounds = listView.getCellBounds(rowIndex, rowIndex);
        if (cellBounds == null)
          return null;
        final Point convertedPosition =
            SwingUtilities.convertPoint(listView, cellBounds.x, cellBounds.y, arrowWidget);
        return convertedPosition.y + cellBounds.height / 2;
      }
      rowIndex++;
    }
    return null;
  }

  private static void setRootAndExpand(
      final ExplorerManagerProviderPanel panel, final OutlineView ov, SieuferdProject project)
  {
    final Node rootNode = new SieuferdProjectNode(project,
        new NodeSettings(AppliesTo.RELATION_ONLY, true, true, false, false, null, false));
    panel.getExplorerManager().setRootContext(rootNode);
    RP.post(() -> {
      final List<Node> firstLevelNodes =
          ImmutableList.copyOf(rootNode.getChildren().getNodes(true));
      SwingUtilities.invokeLater(() -> {
        for (Node node : firstLevelNodes)
          ov.expandNode(node);
      });
    });
  }

  public void setFieldSelection(final FieldSelection fieldSelection) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkNotNull(fieldSelection);
    final ViewContext viewContext = fieldSelection.getViewContext();
    setRootAndExpand(foreignTableExplorerPanel, foreignTableOutlineView, viewContext.getProject());
    final Path lowestCommonAncestor =
        Views.findLowestCommonAncestor(fieldSelection.getSelectedPaths());
    final RelationPath baseRelationPath = lowestCommonAncestor.isRelation()
        ? lowestCommonAncestor.getRelation() : lowestCommonAncestor.getParent();
    final RelationFieldContext baseRelationContext =
        FieldContext.create(viewContext, baseRelationPath).getRelation();
    baseTableSettable.set(baseRelationContext);
    // TODO[cc]: Consider deduplicating logic with root setting for foreignTableFieldsPanel.
    Node baseTableNode = new FieldNode(baseRelationContext, PRIMITIVE_NODE_SETTINGS);
    baseFieldsExplorerPanel.getExplorerManager().setRootContext(baseTableNode);
    baseTableNodeLabel.setText(SwingString.escapeComplete(baseTableNode.getDisplayName()));
    baseTableNodeLabel.setIcon(
        ImageUtilities.image2Icon(baseTableNode.getIcon(BeanInfo.ICON_COLOR_16x16)));
    applyOversizeViewports();
    baseFieldsListView.scrollToBeginning();
    final List<Node> nodesToSelect = Lists.newArrayList();
    for (Path path : fieldSelection.getSelectedPaths()) {
      if (!path.isRelation() && baseRelationPath.isParentOf(path)) {
        Node nodeToSelect =
            FieldNodePathFinder.findNode(baseTableNode, FieldContext.create(viewContext, path));
        if (nodeToSelect != null)
          nodesToSelect.add(nodeToSelect);
      }
    }
    try {
      baseFieldsExplorerPanel.getExplorerManager().setSelectedNodes(
          nodesToSelect.toArray(new Node[nodesToSelect.size()]));
    } catch (PropertyVetoException e) {
      LOG.log(Level.WARNING, "Unexpected exception while selecting base field nodes", e);
    }
  }

  private void addSavedConstraints(Constraints constraints) {
    savedPathPairs.putAll(constraints.getPathPairs());
    arrowWidgetChangeSupport.fireChange();
    clearFieldSelections();
  }

  private void removeSavedConstraints(Constraints constraints) {
    for (Entry<PrimitivePath,PrimitivePath> entry : constraints.getPathPairs().entrySet()) {
      if (entry.getValue().equals(savedPathPairs.get(entry.getKey())))
        savedPathPairs.remove(entry.getKey());
    }
    arrowWidgetChangeSupport.fireChange();
    clearFieldSelections();
  }

  private void clearFieldSelections() {
    try {
      baseFieldsExplorerPanel.getExplorerManager().setSelectedNodes(new Node[] { });
      foreignFieldsExplorerPanel.getExplorerManager().setSelectedNodes(new Node[] { });
    } catch (PropertyVetoException e) {
      LOG.log(Level.WARNING, "Unexpected exception while resetting field node selections", e);
    }
  }

  private void applyOversizeViewports() {
    // TODO[cc]: Find a less hacky way of doing this.
    /* Make sure that every field can be scrolled across the entire length of the field list; this
    can make constraints visualized by the ArrowWidget easier to see or distinguish. */
    baseFieldsListView.applyOversizeViewport();
    foreignFieldsListView.applyOversizeViewport();
  }

  /**
   * This method is called from within the constructor to initialize the form. WARNING: Do NOT
   * modify this code. The content of this method is always regenerated by the Form Editor.
   */
  @SuppressWarnings("unchecked")
  // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
  private void initComponents() {
    java.awt.GridBagConstraints gridBagConstraints;

    baseTableLabel = new javax.swing.JLabel();
    baseTablePanel = new javax.swing.JPanel();
    baseTableNodeLabel = new javax.swing.JLabel();
    constraintPanel = new javax.swing.JPanel();
    baseFieldsOuterPanel = new javax.swing.JPanel();
    baseFieldsLabel = new javax.swing.JLabel();
    baseFieldsExplorerPanel = new com.sieuferd.uiutil.ExplorerManagerProviderPanel();
    baseFieldsListView = new com.sieuferd.project.actions.join.ListViewEx();
    constraintLabel = new javax.swing.JLabel();
    constraintsLayeredPane = new javax.swing.JLayeredPane();
    arrowWidget = new com.sieuferd.project.actions.join.ArrowWidget();
    buttonPanel = new javax.swing.JPanel();
    addButton = new javax.swing.JButton();
    removeButton = new javax.swing.JButton();
    foreignFieldsOuterPanel = new javax.swing.JPanel();
    foreignFieldsLabel = new javax.swing.JLabel();
    foreignFieldsExplorerPanel = new com.sieuferd.uiutil.ExplorerManagerProviderPanel();
    noForeignTableSelectedLabel = new javax.swing.JLabel();
    foreignFieldsListView = new com.sieuferd.project.actions.join.ListViewEx();
    foreignTableLabel = new javax.swing.JLabel();
    foreignTableExplorerPanel = new com.sieuferd.uiutil.ExplorerManagerProviderPanel();
    foreignTableOutlineView = new org.openide.explorer.view.OutlineView();
    infoLabel = new javax.swing.JLabel();

    setLayout(new java.awt.GridBagLayout());

    baseTableLabel.setLabelFor(baseTableNodeLabel);
    org.openide.awt.Mnemonics.setLocalizedText(baseTableLabel, org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.baseTableLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(10, 10, 5, 0);
    add(baseTableLabel, gridBagConstraints);

    baseTablePanel.setBackground(java.awt.Color.white);
    baseTablePanel.setBorder(javax.swing.BorderFactory.createEtchedBorder());
    baseTablePanel.setLayout(new java.awt.BorderLayout());

    org.openide.awt.Mnemonics.setLocalizedText(baseTableNodeLabel, " "); // NOI18N
    baseTableNodeLabel.setBorder(javax.swing.BorderFactory.createEmptyBorder(2, 2, 2, 2));
    baseTablePanel.add(baseTableNodeLabel, java.awt.BorderLayout.CENTER);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(0, 10, 5, 0);
    add(baseTablePanel, gridBagConstraints);

    constraintPanel.setLayout(new java.awt.GridBagLayout());

    baseFieldsOuterPanel.setPreferredSize(new java.awt.Dimension(235, 50));
    baseFieldsOuterPanel.setLayout(new java.awt.GridBagLayout());

    baseFieldsLabel.setLabelFor(baseFieldsExplorerPanel);
    org.openide.awt.Mnemonics.setLocalizedText(baseFieldsLabel, org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.baseFieldsLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(0, 0, 5, 0);
    baseFieldsOuterPanel.add(baseFieldsLabel, gridBagConstraints);

    baseFieldsExplorerPanel.setBorder(javax.swing.BorderFactory.createEtchedBorder());
    baseFieldsExplorerPanel.setLayout(new java.awt.BorderLayout());

    baseFieldsListView.setHorizontalScrollBarPolicy(javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    baseFieldsListView.setVerticalScrollBarPolicy(javax.swing.ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    baseFieldsListView.setTraversalAllowed(false);
    baseFieldsExplorerPanel.add(baseFieldsListView, java.awt.BorderLayout.CENTER);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    baseFieldsOuterPanel.add(baseFieldsExplorerPanel, gridBagConstraints);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    constraintPanel.add(baseFieldsOuterPanel, gridBagConstraints);

    constraintLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
    org.openide.awt.Mnemonics.setLocalizedText(constraintLabel, "="); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(0, 0, 5, 0);
    constraintPanel.add(constraintLabel, gridBagConstraints);

    constraintsLayeredPane.setLayout(new java.awt.GridBagLayout());

    arrowWidget.setModel(arrowWidgetModel);
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = 2;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.ipadx = 50;
    gridBagConstraints.weighty = 1.0;
    constraintsLayeredPane.add(arrowWidget, gridBagConstraints);

    buttonPanel.setOpaque(false);
    buttonPanel.setLayout(new java.awt.GridBagLayout());

    addButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/com/sieuferd/project/icons/plus.png"))); // NOI18N
    addButton.setToolTipText(org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.addButton.toolTipText")); // NOI18N
    addButton.setFocusable(false);
    addButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
    addButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        addButtonActionPerformed(evt);
      }
    });
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.SOUTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 5);
    buttonPanel.add(addButton, gridBagConstraints);

    removeButton.setIcon(new javax.swing.ImageIcon(getClass().getResource("/com/sieuferd/project/icons/minus.png"))); // NOI18N
    removeButton.setToolTipText(org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.removeButton.toolTipText")); // NOI18N
    removeButton.setFocusable(false);
    removeButton.setMargin(new java.awt.Insets(0, 0, 0, 0));
    removeButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        removeButtonActionPerformed(evt);
      }
    });
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 5);
    buttonPanel.add(removeButton, gridBagConstraints);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    constraintsLayeredPane.add(buttonPanel, gridBagConstraints);
    constraintsLayeredPane.setLayer(buttonPanel, javax.swing.JLayeredPane.PALETTE_LAYER);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weighty = 1.0;
    constraintPanel.add(constraintsLayeredPane, gridBagConstraints);

    foreignFieldsOuterPanel.setPreferredSize(new java.awt.Dimension(235, 50));
    foreignFieldsOuterPanel.setLayout(new java.awt.GridBagLayout());

    foreignFieldsLabel.setLabelFor(foreignFieldsExplorerPanel);
    org.openide.awt.Mnemonics.setLocalizedText(foreignFieldsLabel, org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.foreignFieldsLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(0, 0, 5, 0);
    foreignFieldsOuterPanel.add(foreignFieldsLabel, gridBagConstraints);

    foreignFieldsExplorerPanel.setBorder(javax.swing.BorderFactory.createEtchedBorder());
    foreignFieldsExplorerPanel.setLayout(new java.awt.CardLayout());

    noForeignTableSelectedLabel.setBackground(java.awt.Color.white);
    noForeignTableSelectedLabel.setForeground(java.awt.SystemColor.controlShadow);
    noForeignTableSelectedLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
    org.openide.awt.Mnemonics.setLocalizedText(noForeignTableSelectedLabel, org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.noForeignTableSelectedLabel.text")); // NOI18N
    noForeignTableSelectedLabel.setOpaque(true);
    foreignFieldsExplorerPanel.add(noForeignTableSelectedLabel, "noForeignTableSelected");

    foreignFieldsListView.setHorizontalScrollBarPolicy(javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    foreignFieldsListView.setVerticalScrollBarPolicy(javax.swing.ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    foreignFieldsListView.setTraversalAllowed(false);
    foreignFieldsExplorerPanel.add(foreignFieldsListView, "foreignTableSelected");

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    foreignFieldsOuterPanel.add(foreignFieldsExplorerPanel, gridBagConstraints);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    constraintPanel.add(foreignFieldsOuterPanel, gridBagConstraints);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 2;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 2.0;
    gridBagConstraints.weighty = 1.0;
    gridBagConstraints.insets = new java.awt.Insets(0, 10, 0, 0);
    add(constraintPanel, gridBagConstraints);

    foreignTableLabel.setLabelFor(foreignTableOutlineView);
    org.openide.awt.Mnemonics.setLocalizedText(foreignTableLabel, org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.foreignTableLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(10, 5, 5, 10);
    add(foreignTableLabel, gridBagConstraints);

    foreignTableExplorerPanel.setBorder(javax.swing.BorderFactory.createEtchedBorder());
    foreignTableExplorerPanel.setPreferredSize(new java.awt.Dimension(235, 350));
    foreignTableExplorerPanel.setLayout(new java.awt.BorderLayout());

    foreignTableOutlineView.setDefaultActionAllowed(false);
    foreignTableExplorerPanel.add(foreignTableOutlineView, java.awt.BorderLayout.CENTER);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = 2;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 10);
    add(foreignTableExplorerPanel, gridBagConstraints);

    org.openide.awt.Mnemonics.setLocalizedText(infoLabel, org.openide.util.NbBundle.getMessage(JoinDialogPanel.class, "JoinDialogPanel.infoLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 3;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.insets = new java.awt.Insets(10, 10, 10, 10);
    add(infoLabel, gridBagConstraints);
  }// </editor-fold>//GEN-END:initComponents

  private void addButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_addButtonActionPerformed
    Constraints selectedConstraints = selectedConstraintsFlow.get();
    if (selectedConstraints != null)
      addSavedConstraints(selectedConstraints);
  }//GEN-LAST:event_addButtonActionPerformed

  private void removeButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_removeButtonActionPerformed
    Constraints selectedConstraints = selectedConstraintsFlow.get();
    if (selectedConstraints != null)
      removeSavedConstraints(selectedConstraints);
  }//GEN-LAST:event_removeButtonActionPerformed


  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JButton addButton;
  private com.sieuferd.project.actions.join.ArrowWidget arrowWidget;
  private com.sieuferd.uiutil.ExplorerManagerProviderPanel baseFieldsExplorerPanel;
  private javax.swing.JLabel baseFieldsLabel;
  private com.sieuferd.project.actions.join.ListViewEx baseFieldsListView;
  private javax.swing.JPanel baseFieldsOuterPanel;
  private javax.swing.JLabel baseTableLabel;
  private javax.swing.JLabel baseTableNodeLabel;
  private javax.swing.JPanel baseTablePanel;
  private javax.swing.JPanel buttonPanel;
  private javax.swing.JLabel constraintLabel;
  private javax.swing.JPanel constraintPanel;
  private javax.swing.JLayeredPane constraintsLayeredPane;
  private com.sieuferd.uiutil.ExplorerManagerProviderPanel foreignFieldsExplorerPanel;
  private javax.swing.JLabel foreignFieldsLabel;
  private com.sieuferd.project.actions.join.ListViewEx foreignFieldsListView;
  private javax.swing.JPanel foreignFieldsOuterPanel;
  private com.sieuferd.uiutil.ExplorerManagerProviderPanel foreignTableExplorerPanel;
  private javax.swing.JLabel foreignTableLabel;
  private org.openide.explorer.view.OutlineView foreignTableOutlineView;
  private javax.swing.JLabel infoLabel;
  private javax.swing.JLabel noForeignTableSelectedLabel;
  private javax.swing.JButton removeButton;
  // End of variables declaration//GEN-END:variables
}
