// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.datasource;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import java.awt.BorderLayout;
import java.sql.SQLException;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLTask;
import com.sieuferd.database.SQLTask.UserInformedConnectionException;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.project.NBSQLTaskRunner;
import com.sieuferd.project.datasource.DatabaseTableModel.DatabaseTableColumn;
import com.sieuferd.query.SQLTable;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.TypedComboBoxCellRenderer;
import com.sieuferd.uiutil.TypedComboBoxModel;
import com.sieuferd.util.StringUtil;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.viewset.SieuferdProjectConfiguration;
import java.awt.Dimension;
import java.util.List;
import java.util.Map.Entry;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableColumn;
import org.netbeans.api.db.explorer.ConnectionListener;
import org.netbeans.api.db.explorer.ConnectionManager;
import org.netbeans.api.db.explorer.DatabaseConnection;
import org.netbeans.api.db.explorer.DatabaseException;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import org.openide.util.Mutex;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.WeakListeners;

public class DataSourceCustomizer extends JPanel {
  private static final Logger LOG = Logger.getLogger(DataSourceCustomizer.class.getName());
  private static final RequestProcessor DATABASE_RP =
      new RequestProcessor(DataSourceCustomizer.class);
  private final SieuferdProjectConfiguration configuration;
  // See org.netbeans.modules.apisupport.project.ui.platform.PlatformComponentFactory
  /* Could have used DatabaseExplorerUIs.connect(), but I discovered it too late, and I need the
  custom "missing connection" message. */
  /** Holds programmatic names of database connections. */
  private final TypedComboBoxModel<String> connectionComboBoxModel =
      new TypedComboBoxModel<>(String.class);
  private final DatabaseTableModel databaseTableModel;
  private final ConnectionManager connectionManager = ConnectionManager.getDefault();
  private final ConnectionListener connectionListener = this::updateConnectionList;
  private final TableModelListener databaseTableListener = new TableModelListener() {
    @Override
    public void tableChanged(TableModelEvent e) {
      configuration.setManagedDatabases(databaseTableModel.getManagedDatabases());
    }
  };
  private final ListDataListener connectionListListener = new ListDataListener() {
      @Override
      public void intervalAdded(ListDataEvent e) {
      }

      @Override
      public void intervalRemoved(ListDataEvent e) {
      }

      @Override
      public void contentsChanged(ListDataEvent e) {
        configuration.setDatabaseConnectionName(connectionComboBoxModel.getSelectedItem());
      }
    };

  /** May be called from any thread. */
  private void updateConnectionList() {
    Mutex.EVENT.writeAccess(() -> {
      ImmutableSet<String> oldConnections = connectionComboBoxModel.getExternalItems();
      List<String> toConnections = Lists.newArrayList();
      for (DatabaseConnection connection : connectionManager.getConnections())
        toConnections.add(connection.getName());
      connectionComboBoxModel.setExternalItems(toConnections);
      toConnections.removeAll(oldConnections);
      if (toConnections.contains(connectionComboBoxModel.getSelectedItem())) {
        // A selected but previously removed database was added.
        refreshDatabaseList();
      }
    });
  }

  // TODO[cc]: Avoid this suppression by fixing TypedComboBoxCellRenderer.
  @SuppressWarnings("unchecked")
  @NbBundle.Messages({
    "# {0} - programmatic JDBC connection name",
    "TXT_MissingConnectionFmt=Missing connection: {0}"})
  private void configureConnectionChoice() {
    connectionChoice.setRenderer(new TypedComboBoxCellRenderer<>(connectionComboBoxModel,
        Bundle.TXT_MissingConnectionFmt("{0}"), (String input) -> {
          /* Changes to the connection display name are not immediately reflected in the combobox.
          Fixing this would require access to the implementation class
          org.netbeans.modules.db.explorer.DatabaseConnection in a private package (adding a property
          change listener on "displayName"). Not really a problem. */
          DatabaseConnection connection = connectionManager.getConnection(input);
          return (connection == null) ? input : connection.getDisplayName();
    }));
  }

  public DataSourceCustomizer(final SieuferdProjectConfiguration configuration) {
    Preconditions.checkNotNull(configuration);
    this.configuration = configuration;
    initComponents();
    {
      /* Hack to make the add and remove buttons equal width. Using a GridLayout for the buttons
      would be a better way to achieve this, but this would not preserve baseline alignment with
      the dialect label. */
      Dimension addRemoveButtonPadding = new Dimension(Math.max(
          connectionAddButton.getPreferredSize().width,
          connectionRemoveButton.getPreferredSize().width), 0);
      connectionAddPadding.setPreferredSize(addRemoveButtonPadding);
      connectionRemovePadding.setPreferredSize(addRemoveButtonPadding);
    }
    configureConnectionChoice();
    databaseTableModel = DatabaseTableUtil.configureDatabaseTable(databaseTable);
    setDatabaseNameTerms(null);
    databaseTableModel.setManagedDatabases(configuration.getManagedDatabases());
    connectionComboBoxModel.setInitialSelection(configuration.getDatabaseConnectionName());
    updateConnectionList();
    connectionManager.addConnectionListener(WeakListeners.create(ConnectionListener.class,
        connectionListener, connectionManager));
    refreshRemoveButtonState();
    refreshDatabaseList();
    databaseTableModel.addTableModelListener(WeakListeners.create(TableModelListener.class,
        databaseTableListener, databaseTableModel));
    connectionComboBoxModel.addListDataListener(WeakListeners.create(ListDataListener.class,
        connectionListListener, connectionComboBoxModel));
    /* Postpone making the automatic combo box selection until the dialog is showing, since the
    selection can potentially trigger the connection dialog and database list retrieval progress
    bar. */
    SwingUtilities.invokeLater(() -> {
      connectionComboBoxModel.setSelectedItem(configuration.getDatabaseConnectionName());
    });
  }

  private void refreshRemoveButtonState() {
    ThreadUtil.checkOnEDT();
    connectionRemoveButton.setEnabled(connectionComboBoxModel.getSelectedItem() != null);
  }

  @NbBundle.Messages({
    "# {0} - display name of detected database dialect",
    "MSG_DetectedDialect=Detected dialect {0}"
  })
  private void setDialectString(@Nullable BackendDialect dialect) {
    ThreadUtil.checkOnEDT();
    dialectLabel.setText(
        dialect == null ? "" : Bundle.MSG_DetectedDialect(dialect.getDisplayName()));
  }

  private void setDatabaseNameTerms(@Nullable BackendDialect dialect) {
    ThreadUtil.checkOnEDT();
    String primaryTerm   = dialect == null ? "" : dialect.getDatabaseNameTerm(true).trim();
    String secondaryTerm = dialect == null ? "" : dialect.getDatabaseNameTerm(false).trim();
    if (primaryTerm.isEmpty()) {
      primaryTerm = "catalog";
      secondaryTerm = "schema";
    }
    setDatabaseNameTerm(primaryTerm, true);
    setDatabaseNameTerm(secondaryTerm, false);
  }

  @NbBundle.Messages({
    "# {0} - trimmed and title-cased JDBC term for the kind of database name in question",
    "MSG_DatabaseTableColumn_DatabaseName={0} Name"
  })
  private void setDatabaseNameTerm(String term, boolean primary) {
    String presentationName = term.isEmpty() ? "" : (SwingString.escapeComplete(
        Bundle.MSG_DatabaseTableColumn_DatabaseName(StringUtil.titleCaseEnglish(term))));
    DatabaseTableColumn column =
        primary ? DatabaseTableColumn.PRIMARY_NAME : DatabaseTableColumn.SECONDARY_NAME;
    TableColumn tableColumn = databaseTable.getColumnModel().getColumn(column.ordinal());
    tableColumn.setHeaderValue(presentationName);
    /* A nudge seems necessary to get the JTable to update its headers consistently (repaint() is
    not sufficient). */
    final int originalWidth = tableColumn.getWidth();
    tableColumn.setWidth(originalWidth + 1);
    tableColumn.setWidth(originalWidth);
  }

  @NbBundle.Messages("MSG_RetrievingDatabases=Retrieving schema list...")
  private void refreshDatabaseList() {
    ThreadUtil.checkOnEDT();
    setDialectString(null);
    final String databaseConnectionName = connectionComboBoxModel.getSelectedItem();
    if (databaseConnectionName == null ||
        ConnectionManager.getDefault().getConnection(databaseConnectionName) == null)
    {
      databaseTableModel.setDiscoveredDatabases(ImmutableMap.of());
    } else {
      final ProgressHandle progress = ProgressHandle.createHandle(Bundle.MSG_RetrievingDatabases());
      final JComponent progressComponent = ProgressHandleFactory.createProgressComponent(progress);
      /* Cosmetic MacOS tweak. In its current implementation,
      ProgressHandleFactory.createProgressComponent() returns a JProgressBar, which makes this
      work. See http://developer.apple.com/library/mac/#technotes/tn2007/tn2196.html . */
      progressComponent.putClientProperty("JProgressBar.style", "circular");
      // TODO[cc]: Use the dataflow API to manage the UI state here.
      /* Don't wait until the RequestProcessor job has been submitted; the progress indication
      should start immediately. */
      Mutex.EVENT.writeAccess(() -> {
        progress.start();
        // TODO[cc]: Consider using a CardLayout instead.
        progressPanel.add(progressComponent, BorderLayout.EAST);
        progressPanel.revalidate();
        databaseTableModel.setDiscoveredDatabases(null);
      });
      DATABASE_RP.submit(() -> {
        RetrievedMetadata toRetrievedMetadata = null;
        /* Ensure that temporary progress-indicating UI changes are reversed by using a try-finally
        pattern from within the request processor. */
        try {
          toRetrievedMetadata = NBSQLTaskRunner.get(databaseConnectionName).runTask(true, true,
              new SQLTask<RetrievedMetadata>("Retrieve database and table list")
              {
                @Override
                public RetrievedMetadata run(SQLConnection sqlConnection) throws SQLException {
                  final Map<SQLDatabase,Integer> discoveredDatabases = Maps.newLinkedHashMap();
                  for (Entry<SQLDatabase,Set<SQLTable>> entry :
                      sqlConnection.getDatabasesAndTables().entrySet())
                  {
                    discoveredDatabases.put(entry.getKey(), entry.getValue().size());
                  }
                  return new RetrievedMetadata(
                      sqlConnection.getBackendDialect(), discoveredDatabases);
                }
              });
        } catch (UserInformedConnectionException e) {
          return;
        } catch (SQLException e) {
          LOG.log(Level.WARNING, "Error retrieving database list", e);
        } finally {
          final RetrievedMetadata retrievedMetadata = toRetrievedMetadata;
          Mutex.EVENT.writeAccess(() -> {
            if (retrievedMetadata == null) {
              setDialectString(null);
              databaseTableModel.setDiscoveredDatabases(ImmutableMap.of());
            } else {
              setDialectString(retrievedMetadata.dialect);
              // Don't modify the column header names unless an explicit new dialect is known.
              setDatabaseNameTerms(retrievedMetadata.dialect);
              databaseTableModel.setDiscoveredDatabases(retrievedMetadata.discoveredDatabases);
            }
            progressPanel.remove(progressComponent);
            progressPanel.revalidate();
            progress.finish();
          });
        }
      });
    }
  }

  @Immutable
  private static class RetrievedMetadata {
    final BackendDialect dialect;
    final Map<SQLDatabase,Integer> discoveredDatabases;

    public RetrievedMetadata(BackendDialect dialect, Map<SQLDatabase,Integer> discoveredDatabases)
    {
      Preconditions.checkNotNull(dialect);
      Preconditions.checkNotNull(discoveredDatabases);
      this.dialect = dialect;
      this.discoveredDatabases = ImmutableMap.copyOf(discoveredDatabases);
    }
  }

  /**
   * This method is called from within the constructor to initialize the form. WARNING: Do NOT
   * modify this code. The content of this method is always regenerated by the Form Editor.
   */
  @SuppressWarnings({"unchecked", "rawtypes"})
  // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
  private void initComponents() {
    java.awt.GridBagConstraints gridBagConstraints;

    connectionChoiceLabel = new javax.swing.JLabel();
    connectionChoice = new javax.swing.JComboBox<>();
    addRemoveButtonPanel = new javax.swing.JPanel();
    dialectLabel = new javax.swing.JLabel();
    connectionAddButton = new javax.swing.JButton();
    connectionRemoveButton = new javax.swing.JButton();
    connectionAddPadding = new javax.swing.JPanel();
    connectionRemovePadding = new javax.swing.JPanel();
    progressPanel = new javax.swing.JPanel();
    databaseTableLabel = new javax.swing.JLabel();
    databaseTableScrollPane = new javax.swing.JScrollPane();
    databaseTable = new javax.swing.JTable();

    setLayout(new java.awt.GridBagLayout());

    connectionChoiceLabel.setLabelFor(connectionChoice);
    org.openide.awt.Mnemonics.setLocalizedText(connectionChoiceLabel, org.openide.util.NbBundle.getMessage(DataSourceCustomizer.class, "DataSourceCustomizer.connectionChoiceLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    gridBagConstraints.insets = new java.awt.Insets(0, 0, 0, 6);
    add(connectionChoiceLabel, gridBagConstraints);

    connectionChoice.setModel(connectionComboBoxModel);
    connectionChoice.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        connectionChoiceActionPerformed(evt);
      }
    });
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    gridBagConstraints.weightx = 1.0;
    add(connectionChoice, gridBagConstraints);

    addRemoveButtonPanel.setOpaque(false);
    addRemoveButtonPanel.setLayout(new java.awt.GridBagLayout());

    org.openide.awt.Mnemonics.setLocalizedText(dialectLabel, "Detected dialect LipsumSQL"); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE_LEADING;
    gridBagConstraints.weightx = 1.0;
    addRemoveButtonPanel.add(dialectLabel, gridBagConstraints);

    org.openide.awt.Mnemonics.setLocalizedText(connectionAddButton, org.openide.util.NbBundle.getMessage(DataSourceCustomizer.class, "DataSourceCustomizer.connectionAddButton.text")); // NOI18N
    connectionAddButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        connectionAddButtonActionPerformed(evt);
      }
    });
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    addRemoveButtonPanel.add(connectionAddButton, gridBagConstraints);

    org.openide.awt.Mnemonics.setLocalizedText(connectionRemoveButton, org.openide.util.NbBundle.getMessage(DataSourceCustomizer.class, "DataSourceCustomizer.connectionRemoveButton.text")); // NOI18N
    connectionRemoveButton.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        connectionRemoveButtonActionPerformed(evt);
      }
    });
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    addRemoveButtonPanel.add(connectionRemoveButton, gridBagConstraints);
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    addRemoveButtonPanel.add(connectionAddPadding, gridBagConstraints);
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    addRemoveButtonPanel.add(connectionRemovePadding, gridBagConstraints);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE_TRAILING;
    gridBagConstraints.insets = new java.awt.Insets(6, 24, 0, 0);
    add(addRemoveButtonPanel, gridBagConstraints);

    progressPanel.setOpaque(false);
    progressPanel.setLayout(new java.awt.BorderLayout());
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 2;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.ipadx = 100;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE_TRAILING;
    add(progressPanel, gridBagConstraints);

    databaseTableLabel.setLabelFor(databaseTable);
    org.openide.awt.Mnemonics.setLocalizedText(databaseTableLabel, org.openide.util.NbBundle.getMessage(DataSourceCustomizer.class, "DataSourceCustomizer.databaseTableLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 2;
    gridBagConstraints.gridwidth = 2;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    gridBagConstraints.insets = new java.awt.Insets(6, 0, 6, 0);
    add(databaseTableLabel, gridBagConstraints);

    databaseTable.setModel(new javax.swing.table.DefaultTableModel(
      new Object [][] {
        {null, null, null, null},
        {null, null, null, null},
        {null, null, null, null},
        {null, null, null, null}
      },
      new String [] {
        "Title 1", "Title 2", "Title 3", "Title 4"
      }
    ));
    databaseTable.setFillsViewportHeight(true);
    databaseTableScrollPane.setViewportView(databaseTable);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 3;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.NORTHWEST;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    add(databaseTableScrollPane, gridBagConstraints);
  }// </editor-fold>//GEN-END:initComponents

  @NbBundle.Messages({
    "# {0} - Programmatic or displayable database connection name",
    "MSG_ConnectionRemoveConfirm=" +
      "<html>The following stored database connection will be deleted:<br>{0}</html>",
    "TIT_ConnectionRemoveConfirm=Remove Database Connection"
  })
  private void connectionRemoveButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_connectionRemoveButtonActionPerformed
    String connectionName = connectionComboBoxModel.getSelectedItem();
    if (connectionName == null)
      return;
    final DatabaseConnection connection =
        ConnectionManager.getDefault().getConnection(connectionName);
    String displayName = (connection == null) ? connectionName : connection.getDisplayName();
    NotifyDescriptor confirmRemove = new NotifyDescriptor.Confirmation(
        Bundle.MSG_ConnectionRemoveConfirm(SwingString.toSwingHTML(displayName)),
        Bundle.TIT_ConnectionRemoveConfirm(),
        NotifyDescriptor.OK_CANCEL_OPTION, NotifyDescriptor.QUESTION_MESSAGE);
    if (DialogDisplayer.getDefault().notify(confirmRemove) != NotifyDescriptor.OK_OPTION)
      return;
    if (connectionName.equals(connectionComboBoxModel.getInitialSelection()))
      connectionComboBoxModel.setInitialSelection(null);
    connectionComboBoxModel.setSelectedItem(null);
    if (connection != null) {
      DATABASE_RP.submit(() -> {
        try {
          ConnectionManager.getDefault().removeConnection(connection);
        } catch (DatabaseException e) {
          LOG.log(Level.WARNING, "Error deleting database connection", e);
        }
      });
    }
  }//GEN-LAST:event_connectionRemoveButtonActionPerformed

  private void connectionAddButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_connectionAddButtonActionPerformed
    ConnectionManager cm = ConnectionManager.getDefault();
    Set<DatabaseConnection> connectionsBefore = Sets.newHashSet(cm.getConnections());
    DatabaseConnection connection = cm.showAddConnectionDialogFromEventThread(null);
    if (connection == null) {
      /* As of NetBeans 7.2.1, showAddConnectionDialogFromEventThread() only ever actually returns
      null; see the implementation of
      ConnectUsingDriverAction.NewConnectionDialogDisplayer.showDialog().
      This is a hack to still try to find the added connection, if any. */
      Set<DatabaseConnection> connectionsAfter = Sets.newHashSet(cm.getConnections());
      connectionsAfter.removeAll(connectionsBefore);
      if (connectionsAfter.size() == 1)
        connection = ImmutableList.copyOf(connectionsAfter).get(0);
    }
    if (connection != null)
      connectionComboBoxModel.setSelectedItem(connection.getName());
  }//GEN-LAST:event_connectionAddButtonActionPerformed

  private void connectionChoiceActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_connectionChoiceActionPerformed
    refreshRemoveButtonState();
    refreshDatabaseList();
  }//GEN-LAST:event_connectionChoiceActionPerformed

  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JPanel addRemoveButtonPanel;
  private javax.swing.JButton connectionAddButton;
  private javax.swing.JPanel connectionAddPadding;
  private javax.swing.JComboBox<String> connectionChoice;
  private javax.swing.JLabel connectionChoiceLabel;
  private javax.swing.JButton connectionRemoveButton;
  private javax.swing.JPanel connectionRemovePadding;
  private javax.swing.JTable databaseTable;
  private javax.swing.JLabel databaseTableLabel;
  private javax.swing.JScrollPane databaseTableScrollPane;
  private javax.swing.JLabel dialectLabel;
  private javax.swing.JPanel progressPanel;
  // End of variables declaration//GEN-END:variables
}
