// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.datasource;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Ordering;
import com.google.common.collect.Sets;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import javax.swing.table.AbstractTableModel;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.uiutil.SwingString;
import javax.annotation.Nullable;
import org.openide.util.NbBundle.Messages;

/** Note: This "model" intentionally contains a fair bit of presentation logic; this way I can avoid
creating a custom cell renderer. */
@Messages({
  "MSG_DatabaseTable_MissingDatabase=Not found",
  "MSG_DatabaseTableColumn_TableCount=Tables"
})
public class DatabaseTableModel extends AbstractTableModel {
  public static enum DatabaseTableColumn {
    MANAGED(Boolean.class),
    // Remaining columns use the default String renderer, which supports Swing HTML formatting.
    PRIMARY_NAME(String.class),
    SECONDARY_NAME(String.class),
    TABLE_COUNT(String.class),
    MISSING(String.class);

    public final Class<?> type;

    private DatabaseTableColumn(Class<?> type) {
      this.type = type;
    }
  };

  private final int tableCountColumnWidth;
  private ImmutableList<SQLDatabase> rows = ImmutableList.of();
  /** Value denotes the number of tables in the database. */
  private @Nullable ImmutableMap<SQLDatabase,Integer> discoveredDatabases = null;
  private Set<SQLDatabase> managedDatabases = Sets.newHashSet();

  public DatabaseTableModel(int tableCountColumnWidth) {
    this.tableCountColumnWidth = tableCountColumnWidth;
  }

  public void setManagedDatabases(Collection<SQLDatabase> managedDatabases) {
    this.managedDatabases = Sets.newLinkedHashSet(managedDatabases);
    updateRowList();
  }

  /**
   * @param discoveredDatabases can be null to avoid the "not found" message. Values in the map
   * denote table counts.
   */
  public void setDiscoveredDatabases(@Nullable Map<SQLDatabase,Integer> discoveredDatabases) {
    this.discoveredDatabases =
        (discoveredDatabases == null) ? null : ImmutableMap.copyOf(discoveredDatabases);
    updateRowList();
  }

  private boolean isMissing(SQLDatabase database) {
    return (discoveredDatabases != null &&
        !discoveredDatabases.containsKey(database));
  }

  private boolean hasTables(SQLDatabase database) {
    Integer tableCount = (discoveredDatabases == null) ? null : discoveredDatabases.get(database);
    return tableCount != null && tableCount > 0;
  }

  private final Ordering<SQLDatabase> databaseOrdering = new Ordering<SQLDatabase>() {
    @Override
    public int compare(SQLDatabase left, SQLDatabase right) {
      // List missing databases first.
      return (isMissing(right) ? 1 : 0) - (isMissing(left) ? 1 : 0);
    }
  }.compound(new Ordering<SQLDatabase>() {
    @Override
    public int compare(SQLDatabase left, SQLDatabase right) {
      /* List databases with more than zero tables first. In the MIT Data Warehouse example, there
      were 24000 databases (one for each user), with only 10 interesting ones containing any
      tables. */
      return (hasTables(right) ? 1 : 0) - (hasTables(left) ? 1 : 0);
    }
  }.compound(new Ordering<SQLDatabase>() {
    @Override
    public int compare(SQLDatabase left, SQLDatabase right) {
      return left.getPrimaryNamePart().compareTo(right.getPrimaryNamePart());
    }
  }.compound(new Ordering<SQLDatabase>() {
    @Override
    public int compare(SQLDatabase left, SQLDatabase right) {
      return left.getSecondaryNamePart().compareTo(right.getSecondaryNamePart());
    }
  })));

  private void updateRowList() {
    Set<SQLDatabase> allDatabases = Sets.newLinkedHashSet();
    if (discoveredDatabases != null)
      allDatabases.addAll(discoveredDatabases.keySet());
    allDatabases.addAll(managedDatabases);
    rows = databaseOrdering.immutableSortedCopy(allDatabases);
    fireTableDataChanged();
  }

  public Set<SQLDatabase> getManagedDatabases() {
    return Sets.newLinkedHashSet(managedDatabases);
  }

  @Override
  public boolean isCellEditable(int rowIndex, int columnIndex) {
    return columnIndex == DatabaseTableColumn.MANAGED.ordinal();
  }

  @Override
  public int getRowCount() {
    return rows.size();
  }

  @Override
  public int getColumnCount() {
    return DatabaseTableColumn.values().length;
  }

  @Override
  public Class<?> getColumnClass(int columnIndex) {
    return DatabaseTableColumn.values()[columnIndex].type;
  }

  @Override
  public String getColumnName(int columnIndex) {
    if (columnIndex == DatabaseTableColumn.TABLE_COUNT.ordinal()) {
      return
        "<html>" +
        tableCountColumnStyle() +
        "<p>" + Bundle.MSG_DatabaseTableColumn_TableCount() + "</p>" +
        "</html>";
    } else {
      return "";
    }
  }

  private String tableCountColumnStyle() {
    return
      "<style type=\"text/css\">" +
        "body { width: " + Integer.toString(tableCountColumnWidth) + "px; } " +
        "p { text-align: right; } " +
      "</style>";
  }

  @Override
  public Object getValueAt(int rowIndex, int columnIndex) {
    SQLDatabase database = rows.get(rowIndex);
    DatabaseTableColumn column = DatabaseTableColumn.values()[columnIndex];
    switch (column) {
      case MANAGED:
        return managedDatabases.contains(database);
      case PRIMARY_NAME:
        return SwingString.escapeComplete(database.getPrimaryNamePart());
      case SECONDARY_NAME:
        return SwingString.escapeComplete(database.getSecondaryNamePart());
      case TABLE_COUNT: {
        Integer tableCount = discoveredDatabases == null ? null : discoveredDatabases.get(database);
        return tableCount == null ? "" :
            "<html>" +
              tableCountColumnStyle() +
              "<p>" + Integer.toString(tableCount) + "</p>" +
            "</html>";
      }
      case MISSING: {
        if (isMissing(database)) {
          return "<html><font color=\"" + SwingString.htmlColor("nb.errorForeground") + "\">" +
              SwingString.toSwingHTML(Bundle.MSG_DatabaseTable_MissingDatabase()) +
              "</font></html>";
        } else {
          return "";
        }
      }
      default:
        throw new AssertionError();
    }
  }

  @Override
  public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
    if (columnIndex != DatabaseTableColumn.MANAGED.ordinal())
      throw new IllegalArgumentException("Non-editable column index " + columnIndex);
    SQLDatabase database = rows.get(rowIndex);
    if ((Boolean) aValue) {
      managedDatabases.add(database);
    } else {
      managedDatabases.remove(database);
    }
    fireTableCellUpdated(rowIndex, columnIndex);
  }
}
