// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.derivatives;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.query.QProp;
import com.sieuferd.util.MapUtil;
import java.util.Collections;
import java.util.Map;
import java.util.Objects;
import javax.annotation.Nullable;

/* TODO[pd]: Consider replacing this with logic that copies entire relations paths with appropriate
             path rewrites. */
/** Identifier for fields with inheritable properties. */
abstract class FieldKey {
  abstract @Nullable Path createField(MutableView targetView, RelationPath parentPath);

  private static @Nullable FieldKey getFieldKey(View view, Path path, boolean includeRelations) {
    if (path.isRelation()) {
      if (!includeRelations)
        return null;
      Relationship relationship =
          Relationships.getRelationshipForExistingPath(view, path.getRelation());
      return (relationship == null) ? null : new RelationshipFieldKey(relationship);
    } else {
      String sqlColumnName = view.get(QProp.SQL_COLUMN_NAME, path);
      if (sqlColumnName.isEmpty()) {
        return null;
      } else {
        return new SQLColumnFieldKey(sqlColumnName);
      }
    }
  }

  /* TODO[pd]: Get rid of includeRelations; delete them afterwards instead, or just don't descend
               into them. */
  static Map<Path,FieldKey> getFieldKeys(View view, RelationPath path, boolean includeRelations) {
    Map<Path,FieldKey> ret = Maps.newLinkedHashMap();
    for (Path childPath : view.getChildPaths(path)) {
      final FieldKey field = getFieldKey(view, childPath, includeRelations);
      if (field != null)
        MapUtil.putChecked(ret, childPath, field);
    }
    return Collections.unmodifiableMap(ret);
  }

  private static final class RelationshipFieldKey extends FieldKey {
    final Relationship relationship;

    RelationshipFieldKey(Relationship relationship) {
      Preconditions.checkNotNull(relationship);
      this.relationship = relationship;
    }

    @Override
    Path createField(MutableView view, RelationPath parentPath) {
      return Relationships.addPathForRelationship(view, parentPath, relationship, false, false);
    }

    @Override
    public int hashCode() {
      // Mix in a prime number, different for each subclass, for good measure.
      return Objects.hash(13469, relationship);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof RelationshipFieldKey))
        return false;
      RelationshipFieldKey other = (RelationshipFieldKey) obj;
      return this.relationship.equals(other.relationship);
    }
  }

  private static final class SQLColumnFieldKey extends FieldKey {
    final String sqlColumnName;

    SQLColumnFieldKey(String sqlColumnName) {
      Preconditions.checkArgument(!sqlColumnName.isEmpty());
      this.sqlColumnName = sqlColumnName;
    }

    @Override
    Path createField(MutableView view, RelationPath parentPath) {
      Path.PrimitivePath ret = view.addPrimitivePath(parentPath);
      view.set(QProp.SQL_COLUMN_NAME, ret, sqlColumnName);
      return ret;
    }

    @Override
    public int hashCode() {
      // See above.
      return Objects.hash(23531, sqlColumnName);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof SQLColumnFieldKey))
        return false;
      SQLColumnFieldKey other = (SQLColumnFieldKey) obj;
      return this.sqlColumnName.equals(other.sqlColumnName);
    }
  }
}
