// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.derivatives;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import com.sieuferd.query.SQLTable;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

@Immutable
final class Relationship {
  private final SQLTable parentTable;
  private final SQLTable childTable;
  private final ImmutableMap<String,String> childToParentColumnMapping;

  /** @param name may be empty, but not null */
  Relationship(
      SQLTable parentTable, SQLTable childTable, Map<String,String> childToParentColumnMapping)
  {
    Preconditions.checkNotNull(parentTable);
    Preconditions.checkNotNull(childTable);
    this.parentTable = parentTable;
    this.childTable = childTable;
    this.childToParentColumnMapping = ImmutableMap.copyOf(childToParentColumnMapping);
  }

  public SQLTable getParentTable() {
    return parentTable;
  }

  public SQLTable getChildTable() {
    return childTable;
  }

  public Map<String,String> getChildToParentColumnMapping() {
    return childToParentColumnMapping;
  }

  public @Nullable Relationship reverse() {
    Map<String,String> newChildToParentColumnMapping = Maps.newLinkedHashMap();
    for (Entry<String,String> entry : getChildToParentColumnMapping().entrySet()) {
      final String newValue = entry.getKey();
      final String existingValue = newChildToParentColumnMapping.put(entry.getValue(), newValue);
      /* Because each field in the child relation can only be joined on a single field in the
      parent relation, not all relationships are reversible. */
      if (existingValue != null && !existingValue.equals(newValue))
        return null;
    }
    return new Relationship(getChildTable(), getParentTable(), newChildToParentColumnMapping);
  }

  public Relationship tablePairOnly() {
    return new Relationship(getParentTable(), getChildTable(), ImmutableMap.of());
  }

  @Override
  public String toString() {
    return "Relationship(" +
        parentTable + ", " + childTable + ", " + childToParentColumnMapping + ")";
  }

  @Override
  public int hashCode() {
    return Objects.hash(parentTable, childTable, childToParentColumnMapping);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof Relationship))
      return false;
    final Relationship other = (Relationship) obj;
    return Objects.equals(this.parentTable, other.parentTable)
        && Objects.equals(this.childTable, other.childTable)
        && Objects.equals(this.childToParentColumnMapping, other.childToParentColumnMapping);
  }
}
