// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.derivatives;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.database.ForeignKeyConstraint;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.View;
import com.sieuferd.query.InstanceDefinition;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import com.sieuferd.query.OptionalPrimitivePath;
import com.sieuferd.query.QProp;
import com.sieuferd.query.SQLTable;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;

final class Relationships {
  private static final Logger LOG = Logger.getLogger(Relationships.class.getName());
  private Relationships() { }

  static @Nullable RelationPath addPathForRelationship(MutableView view, RelationPath parentPath,
      Relationship relationship, boolean insertAfterLastJoinPath, boolean hideParentJoinPath)
  {
    TableIdentifier parentTable =
        view.get(QProp.INSTANTIATED_TABLE, parentPath).getTable();
    if (!relationship.getParentTable().equals(
        view.get(QProp.INSTANTIATED_TABLE, parentPath).getTable()))
    {
      LOG.log(Level.WARNING, "Tried to add relationship {0} under wrong parent table {1}",
          new Object[] { relationship, parentTable });
      return null;
    }
    // TODO[pd]: Consider avoiding doing this tallying for every field.
    final Map<String,PrimitivePath> pathsBySQLColumnName = Maps.newLinkedHashMap();
    for (Path childPath : view.getChildPaths(parentPath)) {
      if (!childPath.isRelation()) {
        String sqlColumnName = view.get(QProp.SQL_COLUMN_NAME, childPath);
        if (!sqlColumnName.isEmpty() && !pathsBySQLColumnName.containsKey(sqlColumnName))
          pathsBySQLColumnName.put(sqlColumnName, childPath.getPrimitive());
      }
    }
    final Map<String,PrimitivePath> baseColumnPathsByForeignColumnNames =
        Maps.newLinkedHashMap();
    for (Entry<String,String> entry : relationship.getChildToParentColumnMapping().entrySet()) {
      baseColumnPathsByForeignColumnNames.put(
          entry.getKey(), pathsBySQLColumnName.get(entry.getValue()));
    }
    if (baseColumnPathsByForeignColumnNames.containsValue(null))
      return null;
    // Read original order before adding the new path.
    final List<Path> childPathOrder = insertAfterLastJoinPath ?
        Lists.newArrayList(view.get(SProp.CHILD_PATHS, Path.ROOT)) : null;
    final RelationPath ret = view.addRelationPath(parentPath);
    view.set(QProp.INSTANTIATED_TABLE, ret,
        new InstanceDefinition(relationship.getChildTable()));
    int insertIndex = -1;
    for (Entry<String,PrimitivePath> entry : baseColumnPathsByForeignColumnNames.entrySet()) {
      final String foreignColumnName = entry.getKey();
      final PrimitivePath parentJoinPath = entry.getValue();
      final PrimitivePath childJoinPath = view.addPrimitivePath(ret);
      view.set(CProp.VISIBLE, childJoinPath, true);
      view.set(CProp.INHERIT_STYLES, childJoinPath, true);
      view.set(QProp.SQL_COLUMN_NAME, childJoinPath, foreignColumnName);
      view.set(QProp.JOINED_ON, childJoinPath, OptionalPrimitivePath.of(parentJoinPath));
      if (hideParentJoinPath)
        view.set(CProp.VISIBLE, parentJoinPath, false);
      if (childPathOrder != null)
        insertIndex = Math.max(insertIndex, childPathOrder.indexOf(parentJoinPath));
    }
    /* Place the relation for the foreign key relationship after the last primitive path involved in
    the join. */
    if (childPathOrder != null) {
      childPathOrder.add(insertIndex < 0 ? childPathOrder.size() : (insertIndex + 1), ret);
      view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(childPathOrder));
    }
    return ret;
  }

  /* TODO[pd]: Consider taking an index of the parent to avoid quadratic behavior when calling this
               method once per path. */
  // TODO[cc]: Move into ViewUpdater.
  static @Nullable Relationship getRelationshipForExistingPath(View view, RelationPath joinPath) {
    if (joinPath.isRoot())
      return null;
    TableIdentifier baseTable = view.get(QProp.INSTANTIATED_TABLE, joinPath.getParent()).getTable();
    TableIdentifier foreignTable = view.get(QProp.INSTANTIATED_TABLE, joinPath).getTable();
    if (!(baseTable instanceof SQLTable) || !(foreignTable instanceof SQLTable))
      return null;
    Map<String,String> columnNamePairs = Maps.newLinkedHashMap();
    for (Path childPath : view.getChildPaths(joinPath)) {
      if (childPath.isRelation())
        continue;
      OptionalPrimitivePath joinedOn = view.get(QProp.JOINED_ON, childPath);
      if (joinedOn.isPresent()) {
        PrimitivePath joinedOnPath = joinedOn.getPath();
        if (!joinedOnPath.getParent().isParentOf(joinPath) || !view.exists(joinedOnPath))
          return null;
        String foreignColumnName = view.get(QProp.SQL_COLUMN_NAME, childPath);
        String baseColumnName = view.get(QProp.SQL_COLUMN_NAME, joinedOnPath);
        if (foreignColumnName.isEmpty() || baseColumnName.isEmpty())
          return null;
        columnNamePairs.put(foreignColumnName, baseColumnName);
      }
    }
    return new Relationship((SQLTable) baseTable, (SQLTable) foreignTable, columnNamePairs);
  }

  static Map<Relationship,String> relationshipsFromForeignKeyConstraints(
      Set<ForeignKeyConstraint> fkcs, boolean forward)
  {
    final Map<Relationship,String> ret = Maps.newLinkedHashMap();
    for (ForeignKeyConstraint fkc : fkcs) {
      Map<String,String> childToParentColumnMapping = Maps.newLinkedHashMap();
      boolean invalid = false;
      for (ForeignKeyConstraint.ColumnConstraint columnConstraint : fkc.getColumnConstraints()) {
        String parentColumnName = forward ?
            columnConstraint.getForeignKeyColumnName() : columnConstraint.getPrimaryKeyColumnName();
        String childColumnName = forward ?
            columnConstraint.getPrimaryKeyColumnName() : columnConstraint.getForeignKeyColumnName();
        /* Each field in the child relation can only be joined on a single field in the parent
        relation; skip any foreign key constraints that would violate this rule (should be rare). */
        if (childToParentColumnMapping.put(childColumnName, parentColumnName) != null) {
          invalid = true;
          break;
        }
      }
      if (!invalid) {
        /* I previously used fkc.getName() as the relationship name, which turned out to be a
        horrible choice for usability. Just use the table name. Don't bother trying to singularize
        the table name; it wouldn't work across languages. */
        SQLTable parentTable = forward ? fkc.getForeignKeyTable() : fkc.getPrimaryKeyTable();
        SQLTable childTable = forward ? fkc.getPrimaryKeyTable() : fkc.getForeignKeyTable();
        ret.put(new Relationship(parentTable, childTable, childToParentColumnMapping),
            childTable.getTableName());
      }
    }
    return Collections.unmodifiableMap(ret);
  }
}
