// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.derivatives;

import com.google.common.base.Preconditions;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import com.sieuferd.query.QProp;
import com.sieuferd.util.StringUtil;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;
import org.openide.util.NbBundle;

// TODO[cc]: Rename to ViewParam.
final class ViewGlobals {
  private final Map<TableIdentifier,View> primaryViews = Maps.newLinkedHashMap();
  private final Set<Relationship> relationships = Sets.newLinkedHashSet();
  private final Multimap<TableIdentifier,Relationship> relationshipsByParentTable =
      ArrayListMultimap.create();
  private final Multimap<Relationship,Relationship> relationshipsByTablePair =
      ArrayListMultimap.create();
  private final Map<Relationship,String> primaryRelationshipLabels = Maps.newLinkedHashMap();
  private final Multimap<Relationship,String> nonPrimaryLabelsByRelationship =
      ArrayListMultimap.create();

  public void addView(View view) {
    Preconditions.checkNotNull(view);
    if (view.get(QProp.PRIMARY, Path.ROOT)) {
      TableIdentifier tableIdentifier = view.get(QProp.INSTANTIATED_TABLE, Path.ROOT).getTable();
      /* TODO[pd]: Avoid doing the copy here (but consider that it may be necessary to access the
                   primary view while it is being modified from the same thread). */
      if (tableIdentifier != null)
        primaryViews.put(tableIdentifier, ImmutableView.copyOf(view));
    }
    addRelationships(view, Path.ROOT);
  }

  private void addRelationships(View view, RelationPath path) {
    final TableIdentifier baseTable = view.get(QProp.INSTANTIATED_TABLE, path).getTable();
    final boolean isPrimaryRoot = path.isRoot() && view.get(QProp.PRIMARY, Path.ROOT);
    for (Path childPath : view.getChildPaths(path)) {
      /* Avoid adding relationships for non-visible paths, as suggested relationships would become
      self-sustaining. */
      if (childPath.isRelation() && (view.get(CProp.VISIBLE, childPath) || isPrimaryRoot)) {
        if (baseTable != null && view.get(QProp.INCLUDE_PRIMARY_KEY, childPath)) {
          final Relationship relationship =
              Relationships.getRelationshipForExistingPath(view, childPath.getRelation());
          if (relationship != null) {
            String label = view.get(CProp.LABEL, childPath);
            if (path.isRoot() && view.get(QProp.PRIMARY, Path.ROOT)) {
              primaryRelationshipLabels.put(relationship, label);
            } else {
              nonPrimaryLabelsByRelationship.put(relationship, label);
            }
            addRelationship(relationship);
            Relationship reverseRelationship = relationship.reverse();
            if (reverseRelationship != null)
              addRelationship(reverseRelationship);
          }
        }
        addRelationships(view, childPath.getRelation());
      }
    }
  }

  private void addRelationship(Relationship relationship) {
    if (relationships.add(relationship)) {
      relationshipsByParentTable.put(relationship.getParentTable(), relationship);
      relationshipsByTablePair.put(relationship.tablePairOnly(), relationship);
    }
  }

  public @Nullable View getPrimaryView(TableIdentifier tableIdentifier) {
    return primaryViews.get(tableIdentifier);
  }

  public Set<Relationship> getRelationships(TableIdentifier parentTable) {
    return ImmutableSet.copyOf(relationshipsByParentTable.get(parentTable));
  }

  public String generateRelationshipLabel(Relationship relationship) {
    return generateRelationshipLabel(relationship, false, true);
  }

  // TODO[cc]: Consider removing.
  public String generateRelationshipLabel(Relationship relationship, boolean canonical) {
    return generateRelationshipLabel(relationship, canonical, true);
  }

  @NbBundle.Messages({
    "# {0} - label of the primary view of the child table of the relationship to be labeled",
    "# {1} - generated label of the reverse relationship",
    "DEF_ReverseRelationshipLabel={0} via {1}"
  })
  private String generateRelationshipLabel(
      Relationship relationship, boolean canonical, boolean allowReverse)
  {
    // TODO[pd]: Consider prioritizing visible relationships over non-visible ones.
    final String primaryRelationshipLabel = primaryRelationshipLabels.get(relationship);
    if (primaryRelationshipLabel != null)
      return primaryRelationshipLabel;
    final View childTablePrimaryView = primaryViews.get(relationship.getChildTable());
    final String childTableLabel = (childTablePrimaryView == null)
        ? relationship.getChildTable().getTableName() // Unexpected case
        : childTablePrimaryView.get(CProp.LABEL, Path.ROOT);
    Relationship reverseRelationship = relationship.reverse();
    String reverseLabel = (!allowReverse || reverseRelationship == null) ? null :
        Bundle.DEF_ReverseRelationshipLabel(
        childTableLabel, generateRelationshipLabel(reverseRelationship, canonical, false));
    final Set<String> nonPrimaryLabels;
    if (canonical) {
      nonPrimaryLabels = ImmutableSet.of();
    } else {
      nonPrimaryLabels =
          Sets.newLinkedHashSet(nonPrimaryLabelsByRelationship.get(relationship));
      nonPrimaryLabels.remove(childTableLabel);
      nonPrimaryLabels.remove(reverseLabel);
    }
    String combinedNonPrimaryLabels = StringUtil.join("/", nonPrimaryLabels);
    if (nonPrimaryLabels.size() == 1) {
      return nonPrimaryLabels.iterator().next();
    } else if (relationshipsByTablePair.get(relationship.tablePairOnly()).size() <= 1) {
      return childTableLabel;
    } else if (reverseLabel != null) {
      return reverseLabel;
    } else if (nonPrimaryLabels.size() > 0) {
      return combinedNonPrimaryLabels;
    } else {
      return childTableLabel;
    }
  }
}