// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.derivatives;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.database.ForeignKeyConstraint;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLTableInfo;
import com.sieuferd.database.SQLTask;
import com.sieuferd.database.SQLTask.UserInformedConnectionException;
import com.sieuferd.database.SQLTaskRunner;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.ViewPropertySource;
import com.sieuferd.datamodel.Views;
import com.sieuferd.layout.YesNoAuto;
import com.sieuferd.evaluator.DataProperties;
import com.sieuferd.importer.ImportUtil;
import com.sieuferd.layout.LAProp;
import com.sieuferd.layout.LBProp;
import com.sieuferd.layout.LHProp;
import com.sieuferd.layout.LMProp;
import com.sieuferd.layout.LayoutMeasurer;
import com.sieuferd.project.GatedViewImpl;
import com.sieuferd.project.NBSQLTaskRunner;
import com.sieuferd.project.NBViewSet;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.ViewContext;
import com.sieuferd.query.InstanceDefinition;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.query.SQLTable;
import com.sieuferd.util.MapUtil;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.viewset.GatedView;
import java.awt.EventQueue;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;

public final class ViewUpdater {
  private static final Logger LOG = Logger.getLogger(ViewUpdater.class.getName());
  private static final boolean INCLUDE_PLURAL_REVERSE_RELATIONSHIPS = false;
  private static final boolean DEBUG_DISABLE_SECONDARY_VIEW_DERIVATIVES = false;
  /* Keeping foreign key relationships next to their last join column seemed like a good idea at
  first, but becomes awkward once plural relationships are expanded further from the foreign key. I
  still have not quite decided on which setting to use here. */
  private static final boolean FOREIGN_KEY_RELATIONS_AT_END = false;
  public static final ImmutableSet<ViewProperty<?>> INHERITABLE_PROPERTIES =
      ImmutableSet.<ViewProperty<?>>builder()
      .add(CProp.LABEL)
      .add(CProp.HEADING)
      .addAll(LBProp.INSTANCE.getViewProperties())
      .addAll(LHProp.INSTANCE.getViewProperties())
      .build();
  private static final ImmutableSet<ViewProperty<?>> CHILD_INHERITABLE_PROPERTIES =
      createInheritablePropertySet(false);
  private static final ImmutableSet<ViewProperty<?>> SELF_INHERITABLE_PROPERTIES =
      createInheritablePropertySet(true);

  // TODO[cc]: Consider moving this elsewhere.
  public static <T> void userSetProperty(
      MutableView view, ViewProperty<T> prop, Path path, T value)
  {
    Preconditions.checkArgument(!prop.isUserReadOnly());
    final T oldValue = view.get(prop, path);
    if (!oldValue.equals(value)) {
      if (prop.equals(CProp.VISIBLE) && Boolean.TRUE.equals(value) && !path.isRelation() &&
          !Views.hasVisiblePrimitiveChildren(view, path.getParent()))
      {
        // First primitive to be made visible in this relation.
        // Don't propagate the setting of INCLUDE_PRIMARY_KEY in this case (here and below).
        view.set(QProp.INCLUDE_PRIMARY_KEY, path.getParent(), true);
      }
      view.set(prop, path, value);
      if (prop.equals(QProp.FORMULA_EXPRESSION) ||
          prop.equals(LMProp.OVERRIDE_MEASUREMENTS) && Boolean.FALSE.equals(value) ||
          LayoutMeasurer.MEASUREMENT_INPUTS.contains(prop) &&
          !prop.getPropertySet().equals(LMProp.INSTANCE))
      {
        view.set(LMProp.SAMPLE_SIZE, path, 0);
        view.set(LMProp.OVERRIDE_MEASUREMENTS, path, false);
      }
      if (!prop.equals(LMProp.OVERRIDE_MEASUREMENTS) &&
          !prop.equals(LMProp.SAMPLE_SIZE) &&
          LMProp.INSTANCE.getViewProperties().contains(prop))
      {
        view.set(LMProp.OVERRIDE_MEASUREMENTS, path, true);
      }
      if (ViewUpdater.INHERITABLE_PROPERTIES.contains(prop))
        view.set(CProp.INHERIT_STYLES, path, false);
      if (prop.equals(CProp.VISIBLE) && Boolean.FALSE.equals(value) && !path.isRelation() &&
          !Views.hasVisiblePrimitiveChildren(view, path.getParent()))
      {
        // Last primitive to be hidden in this relation.
        view.set(QProp.INCLUDE_PRIMARY_KEY, path.getParent(), false);
      }
    }
  }

  public static void userUnhide(MutableView view, Path path) {
    ViewUpdater.userSetProperty(view, CProp.VISIBLE, path, true);
    if (!path.isRoot())
      userUnhide(view, path.getParent());
  }

  private static boolean isSelfInheritable(ViewProperty<?> prop) {
    return prop.getAppliesTo().equals(AppliesTo.RELATION_ONLY);
  }

  private static ImmutableSet<ViewProperty<?>> createInheritablePropertySet(boolean self) {
    Set<ViewProperty<?>> ret = Sets.newLinkedHashSet();
    for (ViewProperty<?> prop : INHERITABLE_PROPERTIES) {
      if (isSelfInheritable(prop) == self)
        ret.add(prop);
    }
    return ImmutableSet.copyOf(ret);
  }

  // TODO[pd]: Use a connection-specific request processor instead.
  private static final RequestProcessor RP_REFRESH_PRIMARY =
      new RequestProcessor("Refresh Primary Views", 1);
  private ViewUpdater() { }

  /** Refresh a specific primary view. May not be called on the EDT. */
  public static void refreshPrimaryView(final ViewContext viewContext) {
    ThreadUtil.checkNotOnEDT();
    TableIdentifier ti = viewContext.getView().get(QProp.INSTANTIATED_TABLE, Path.ROOT).getTable();
    if (!(ti instanceof SQLTable))
      return;
    SQLTaskRunner taskRunner = NBSQLTaskRunner.get(
        viewContext.getProject().getConfiguration().getDatabaseConnectionName());
    View template = createPrimaryViewTemplate(taskRunner, (SQLTable) ti);
    if (template != null)
      applyPrimaryViewUpdate(template, viewContext.getView());
  }

  /**
   * Refresh the list of primary views, and refresh all individual primary views. May not be called
   * on the EDT.
   */
  public static void refreshPrimaryViews(final SieuferdProject project) {
    if (EventQueue.isDispatchThread()) {
      RP_REFRESH_PRIMARY.post(() -> {
        refreshPrimaryViewsSynchronous(project);
      });
    } else {
      refreshPrimaryViewsSynchronous(project);
    }
  }

  private static void refreshPrimaryViewsSynchronous(final SieuferdProject project) {
    ThreadUtil.checkNotOnEDT();

    /* Collect existing primary views by table. Delete any view that claims to be a primary view
    but which instantiates no table, or which instantiates a table for which we already have found
    a primary view. Future user interface improvements should prevent such situations from arising
    in the first place, but it is still good to have the logic here as long as the model classes do
    not provide stricter guarantees. */
    final ImmutableMap<TableIdentifier,GatedView> existingPrimaryViews;
    {
      final Map<TableIdentifier,GatedView> toExistingPrimaryViews = Maps.newLinkedHashMap();
      for (final GatedView view : project.getViewSet().getViews()) {
        if (view.get(QProp.PRIMARY, Path.ROOT)) {
          TableIdentifier table = view.get(QProp.INSTANTIATED_TABLE, Path.ROOT).getTable();
          if (table == null || toExistingPrimaryViews.containsKey(table)) {
            project.getViewSet().removeView(view);
          } else {
            MapUtil.putChecked(toExistingPrimaryViews, table, view);
          }
        }
      }
      existingPrimaryViews = ImmutableMap.copyOf(toExistingPrimaryViews);
    }

    final TableListing tableListing = getManagedTables(true, project);

    /* Delete existing primary views for which the instantiated table no longer exists in the
    database. However, be careful not to delete views simply because there was an error fetching the
    table list. Such errors might occur, for instance, if the user switches between JDBC connections
    with different privileges. */
    for (Entry<TableIdentifier,GatedView> entry : existingPrimaryViews.entrySet()) {
      if (!(entry.getKey() instanceof SQLTable))
        continue;
      SQLTable table = (SQLTable) entry.getKey();
      if (!tableListing.tables.contains(table) &&
          !tableListing.databasesWithTableListingErrors.contains(table.getDatabase()))
      {
        project.getViewSet().removeView(entry.getValue());
      }
    }

    // Create new primary views and update existing ones.
    final SQLTaskRunner taskRunner =
        NBSQLTaskRunner.get(project.getConfiguration().getDatabaseConnectionName());
    for (final SQLTable table : tableListing.tables) {
      final GatedView existingView = existingPrimaryViews.get(table);
      final GatedView targetView;
      if (existingView != null) {
         targetView = existingView;
      } else {
        /* It would make sense to just use the template returned by createPrimaryViewTemplate here,
        but in the future I'm hoping to decouple the initial creation of a primary view from its
        actual population with columns and foreign keys. */
        // TODO[cc]: Avoid accessing ViewSet.getDefaults() here.
        MutableView template = Views.newMutable(project.getViewSet().getDefaults());
        template.set(QProp.PRIMARY, Path.ROOT, true);
        template.set(CProp.EXPANDED, Path.ROOT, true);
        template.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, true);
        /* TODO[pd]: Once root labels are made inheritable, setting the label here will be
                     unnecessary. */
        template.set(CProp.LABEL, Path.ROOT,
            ImportUtil.decodeHumanIdentifier(table.getTableName()));
        template.set(LAProp.USE_TABLE, Path.ROOT, YesNoAuto.YES);
        template.set(QProp.INSTANTIATED_TABLE, Path.ROOT, new InstanceDefinition(table));
        targetView = new GatedViewImpl(template);
      }
      View template = createPrimaryViewTemplate(taskRunner, table);
      if (template != null) {
        applyPrimaryViewUpdate(template, targetView);
        if (existingView == null)
          project.getViewSet().createAndAddView(targetView);
      }
    }
  }

  private static @Nullable View createPrimaryViewTemplate(
      SQLTaskRunner taskRunner, final SQLTable sqlTable)
  {
    Preconditions.checkNotNull(taskRunner);
    Preconditions.checkNotNull(sqlTable);
    try {
      return taskRunner.runTask(true, false, new SQLTask<View>("Get metadata for table " + sqlTable)
      {
        @Override
        public View run(SQLConnection connection) throws SQLException {
          final SQLTableInfo tableInfo = connection.getTableInfo(sqlTable);
          final Set<String> columns = Sets.newLinkedHashSet();
          final Set<String> primaryKeyColumns = Sets.newLinkedHashSet();
          final Set<String> hiddenColumns = Sets.newLinkedHashSet();
          for (String columnName : tableInfo.getColumnNames()) {
            columns.add(columnName);
            if (tableInfo.isPrimaryKey(columnName))
              primaryKeyColumns.add(columnName);
            if (connection.getBackendDialect().isRowIdentifierPseudoColumn(columnName))
              hiddenColumns.add(columnName);
          }
          Set<ForeignKeyConstraint> foreignKeyConstraints;
          try {
            foreignKeyConstraints = connection.getForeignKeyConstraints(sqlTable, true);
          } catch (SQLException e) {
            connection.reportSQLException(e);
            LOG.log(Level.WARNING,
                "Failed to retrieve foreign key relationships for table " + sqlTable, e);
            // Don't treat this as a fatal error.
            foreignKeyConstraints = ImmutableSet.of();
          }
          return createPrimaryViewTemplate(
              sqlTable, columns, primaryKeyColumns, hiddenColumns, foreignKeyConstraints);
        }
      });
    } catch (SQLException e) {
      LOG.log(Level.WARNING, "Failed to retrieve list of columns for table " + sqlTable, e);
      return null;
    } catch (UserInformedConnectionException e) {
      return null;
    }
  }

  @NbBundle.Messages("CTL_RefreshPrimaryView=Refresh Table Metadata")
  private static void applyPrimaryViewUpdate(View template, GatedView target) {
    Preconditions.checkNotNull(template);
    Preconditions.checkNotNull(target);
    final ViewGlobals globals = new ViewGlobals();
    globals.addView(template);
    target.mutate(Bundle.CTL_RefreshPrimaryView(), false, (MutableView view) -> {
      view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, true);
      updateDerivedView(globals, view, Path.ROOT, true, Queries.getUsedPaths(view));
    });
  }

  private static View createPrimaryViewTemplate(TableIdentifier table, Set<String> columns,
      Set<String> primaryKeyColumns, Set<String> hiddenColumns,
      Set<ForeignKeyConstraint> foreignKeyConstraints)
  {
    Preconditions.checkNotNull(table);
    Preconditions.checkNotNull(columns);
    Preconditions.checkNotNull(foreignKeyConstraints);
    final Map<Relationship,String> foreignKeyRelationships =
        Relationships.relationshipsFromForeignKeyConstraints(foreignKeyConstraints, true);
    // TODO[cc]: Avoid dependency on NBViewSet (here and elsewhere).
    final MutableView ret = Views.newMutable(NBViewSet.VIEW_DEFAULTS);
    ret.set(QProp.PRIMARY, Path.ROOT, true);
    ret.set(QProp.INSTANTIATED_TABLE, Path.ROOT, new InstanceDefinition(table));
    ret.set(CProp.EXPANDED, Path.ROOT, true);
    ret.set(CProp.LABEL, Path.ROOT, (table instanceof SQLTable) ?
        ImportUtil.decodeHumanIdentifier(((SQLTable) table).getTableName()) : "");
    for (String columnName : columns) {
      PrimitivePath childPath = ret.addPrimitivePath(Path.ROOT);
      ret.set(CProp.LABEL, childPath, ImportUtil.decodeHumanIdentifier(columnName));
      ret.set(CProp.VISIBLE, childPath, !hiddenColumns.contains(columnName));
      ret.set(QProp.SQL_COLUMN_NAME, childPath, columnName);
      ret.set(QProp.IN_PRIMARY_KEY, childPath, primaryKeyColumns.contains(columnName));
    }
    final Set<Path> foreignKeyPaths =
        Sets.newLinkedHashSetWithExpectedSize(foreignKeyRelationships.size());
    for (Entry<Relationship,String> entry : foreignKeyRelationships.entrySet()) {
      /* Keep insertAfterLastJoinPath true here even if FOREIGN_KEY_RELATIONS_AT_END is enabled, as
      the foreign key relations should have a well-defined order between themselves if placed at the
      end below. */
      RelationPath childPath =
          Relationships.addPathForRelationship(ret, Path.ROOT, entry.getKey(), true, true);
      if (childPath != null) {
        ret.set(CProp.LABEL, childPath, entry.getValue());
        foreignKeyPaths.add(childPath);
      }
    }
    if (FOREIGN_KEY_RELATIONS_AT_END) {
      Set<Path> newChildOrder = Sets.newLinkedHashSet(ret.get(SProp.CHILD_PATHS, Path.ROOT));
      newChildOrder.removeAll(foreignKeyPaths);
      newChildOrder.addAll(foreignKeyPaths);
      ret.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(newChildOrder));
    }
    setHeadingDefaults(ret, Path.ROOT);
    return ret;
  }

  private static boolean identifierMatchesKeywords(
      Locale locale, String identifier, Set<String> keywords)
  {
    final Set<String> tokens = Sets.newLinkedHashSet();
    for (String token : splitIdentifier(identifier))
      tokens.add(token.toLowerCase(locale));
    for (String keyword : keywords) {
      if (tokens.contains(keyword.toLowerCase(locale)))
        return true;
    }
    return false;
  }

  // Package-private for testing.
  /**
   * Split a string by non-alphanumeric characters and camel casing.
   */
  static List<String> splitIdentifier(String identifier) {
    final List<String> ret = Lists.newArrayList();
    StringBuilder currentTokenSB = null;
    for (int i = 0; i < identifier.length(); i++) {
      final char c = identifier.charAt(i);
      final boolean includableCharacter = Character.isLetterOrDigit(c);
      final boolean newToken = !includableCharacter ||
          i > 0 && Character.isLowerCase(identifier.charAt(i - 1)) && Character.isUpperCase(c);
      if (newToken && currentTokenSB != null) {
        ret.add(currentTokenSB.toString());
        currentTokenSB = null;
      }
      if (includableCharacter) {
        if (currentTokenSB == null)
          currentTokenSB = new StringBuilder();
        currentTokenSB.append(identifier.charAt(i));
      }
    }
    if (currentTokenSB != null)
      ret.add(currentTokenSB.toString());
    return Collections.unmodifiableList(ret);
  }

  private static <T> Set<T> getFirstItemsSet(Iterable<T> iterable, int count) {
    final Set<T> ret = Sets.newLinkedHashSet();
    for (T value : iterable) {
      if (ret.size() >= count)
        break;
      ret.add(value);
    }
    return Collections.unmodifiableSet(ret);
  }

  private static Set<PrimitivePath> getHeadingDefaults(View view, RelationPath path) {
    final Locale LOCALE = Locale.US;
    final int MAX_HEADING_FIELDS = 4;
    /* TODO[pd]: Consider including "code" as well. Though prefer "title"/"name" over "code" or
                 "label" if both exist. */
    /* In the Semantic Web (RDF) community, "label" is the standard name for a property to be used
    as a resource's human-readable name. */
    final ImmutableSet<String> LIKELY_HEADING_KEYWORDS = ImmutableSet.of("title", "name", "label");
    final ImmutableSet<String> LIKELY_TECHNICAL_KEYWORDS = ImmutableSet.of("id", "ident", "key");
    final Set<PrimitivePath> eligibleHeadingKeywordMatchPaths = Sets.newLinkedHashSet();
    final Set<PrimitivePath> eligibleNonTechnicalPaths = Sets.newLinkedHashSet();
    final Set<PrimitivePath> primaryKeyPaths = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation())
        continue;
      final String columnName =
          ImportUtil.decodeHumanIdentifier(view.get(QProp.SQL_COLUMN_NAME, childPath));
      if (columnName.isEmpty())
        continue;
      if (view.get(CProp.VISIBLE, childPath)) {
        if (identifierMatchesKeywords(LOCALE, columnName, LIKELY_HEADING_KEYWORDS))
          eligibleHeadingKeywordMatchPaths.add(childPath.getPrimitive());
        if (!identifierMatchesKeywords(LOCALE, columnName, LIKELY_TECHNICAL_KEYWORDS))
          eligibleNonTechnicalPaths.add(childPath.getPrimitive());
      }
      if (Queries.isInPrimaryKey(view, childPath.getPrimitive()))
        primaryKeyPaths.add(childPath.getPrimitive());
    }
    if (!eligibleHeadingKeywordMatchPaths.isEmpty()) {
      return getFirstItemsSet(eligibleHeadingKeywordMatchPaths, MAX_HEADING_FIELDS);
    } else if (!eligibleNonTechnicalPaths.isEmpty()) {
      return getFirstItemsSet(eligibleNonTechnicalPaths, MAX_HEADING_FIELDS);
    } else {
      return primaryKeyPaths;
    }
  }

  private static void setHeadingDefaults(MutableView view, RelationPath path) {
    final Set<PrimitivePath> headingPaths = getHeadingDefaults(view, path);
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation())
        continue;
      view.set(CProp.HEADING, childPath, headingPaths.contains(childPath.getPrimitive()));
    }
  }

  @Immutable
  public static final class TableListing {
    public final @Nullable String connectionName;
    public final ImmutableSet<SQLDatabase> managedDatabases;
    public final ImmutableSet<SQLTable> tables;
    public final ImmutableSet<SQLDatabase> databasesWithTableListingErrors;

    public TableListing(@Nullable String connectionName, Set<SQLDatabase> managedDatabases,
        Set<SQLTable> tables, Set<SQLDatabase> databasesWithTableListingErrors)
    {
      this.connectionName = connectionName;
      this.managedDatabases = ImmutableSet.copyOf(managedDatabases);
      this.tables = ImmutableSet.copyOf(tables);
      this.databasesWithTableListingErrors = ImmutableSet.copyOf(databasesWithTableListingErrors);
    }
  }

  public static TableListing getManagedTables(boolean showDialogOnConnect, SieuferdProject project)
  {
    Set<SQLTable> tables = Sets.newLinkedHashSet();
    Set<SQLDatabase> databasesWithTableListingErrors = Sets.newLinkedHashSet();
    final String connectionName = project.getConfiguration().getDatabaseConnectionName();
    final Set<SQLDatabase> managedDatabases = project.getConfiguration().getManagedDatabases();
    final SQLTaskRunner taskRunner = NBSQLTaskRunner.get(connectionName);
    boolean first = true;
    for (final SQLDatabase database : managedDatabases) {
      try {
        tables.addAll(taskRunner.runTask(first && showDialogOnConnect, first,
            new SQLTask<Set<SQLTable>>("Retrieve table list for database " + database)
        {
          @Override
          public Set<SQLTable> run(SQLConnection sqlConnection) throws SQLException {
            return sqlConnection.getTables(database);
          }
        }));
      } catch (UserInformedConnectionException e) {
        databasesWithTableListingErrors.add(database);
      } catch (SQLException e) {
        LOG.log(Level.WARNING, "Problem retrieving table list for database {0}: {1}",
            new Object[] { database, e.getMessage() });
        databasesWithTableListingErrors.add(database);
      }
      first = false;
    }
    return new TableListing(
        connectionName, managedDatabases, tables, databasesWithTableListingErrors);
  }

  // TODO[cc]: Rename this and related private methods. Reword the message below accordingly.
  @NbBundle.Messages("CTL_RefreshDerivatives=Refresh Derived Properties")
  public static void refreshDerivatives(SieuferdProject project) {
    final ViewGlobals globals = new ViewGlobals();
    for (final GatedView view : project.getViewSet().getViews()) {
      // We're not on the EDT here, so make sure to read the view in an explicit read transaction.
      view.performReadTransaction(() -> {
        globals.addView(view);
      });
    }
    for (final GatedView gatedView : project.getViewSet().getViews()) {
      gatedView.mutate(Bundle.CTL_RefreshDerivatives(), false, (MutableView view) -> {
        final Set<Path> usedPaths = Queries.getUsedPaths(view);
        // TODO[te]: Always run view updates twice, ensuring that the update is stable.
        if (view.get(QProp.PRIMARY, Path.ROOT)) {
          /* Avoid running updateDerivedView on the root path, as primary views should be derived
          from template views that are only available during refreshPrimaryView. */
          for (Path childPath : view.getChildPaths(Path.ROOT)) {
            if (childPath.isRelation())
              updateDerivedView(globals, view, childPath.getRelation(), false, usedPaths);
          }
        } else if (!DEBUG_DISABLE_SECONDARY_VIEW_DERIVATIVES) {
          updateDerivedView(globals, view, Path.ROOT, true, usedPaths);
        }
        DataProperties.configureDataProperties(view);
        removeReverseRelationships(
            view, Path.ROOT, null, INCLUDE_PLURAL_REVERSE_RELATIONSHIPS, usedPaths);
      });
    }
  }

  /**
   * Maintain child path lists, inheritable properties, {@link CProp#CANONICAL_LABEL}, and
   * {@link QProp#IN_PRIMARY_KEY} at {@code path} and its descendants based on primary views and
   * global relationships.
   *
   * @param setCanonicalLabel whether {@link CProp#CANONICAL_LABEL} should be set for {@code path}
   */
  private static void updateDerivedView(
      ViewGlobals globals, MutableView view, RelationPath path, boolean setCanonicalLabel,
      Set<Path> usedPaths)
  {
    // Go one level up so that newly expanded nodes will always be immediately ready.
    final boolean allowChildRelations = path.isRoot() ||
        (view.get(CProp.VISIBLE, path.getParent()) || view.get(CProp.EXPANDED, path.getParent()));
    final Set<Relationship> remainingGlobalRelationships;
    final View primaryView;
    {
      final TableIdentifier tableIdentifier = view.get(QProp.INSTANTIATED_TABLE, path).getTable();
      primaryView = (tableIdentifier == null) ? null : globals.getPrimaryView(tableIdentifier);
      remainingGlobalRelationships =
          (tableIdentifier == null || !allowChildRelations || view.get(QProp.PRIMARY, Path.ROOT))
          ? Sets.newLinkedHashSet()
          : Sets.newLinkedHashSet(globals.getRelationships(tableIdentifier));
    }
    if (setCanonicalLabel) {
      if (primaryView == null) {
        view.setFromDefault(CProp.CANONICAL_LABEL, path);
      } else {
        view.set(CProp.CANONICAL_LABEL, path, primaryView.get(CProp.LABEL, Path.ROOT));
      }
    }
    if (primaryView != null) {
      setInheritedProperties(
          view, path, SELF_INHERITABLE_PROPERTIES, primaryView.getViewPropertySource(Path.ROOT));
    }
    final PathMap pathMap = (primaryView == null) ? PathMap.EMPTY :
          processSourcePaths(view, path, primaryView, Path.ROOT, allowChildRelations);

    // Now process child paths.

    final boolean primaryRoot = path.isRoot() && view.get(QProp.PRIMARY, Path.ROOT);
    for (Path childPath : view.getChildPaths(path)) {
      final Path primaryViewChildPath = pathMap.getOldPathIfMapped(childPath);
      final String canonicalLabel;
      if (childPath.isRelation()) {
        Relationship relationship =
            Relationships.getRelationshipForExistingPath(view, childPath.getRelation());
        if (relationship == null) {
          canonicalLabel = null;
        } else {
          remainingGlobalRelationships.remove(relationship);
          canonicalLabel = globals.generateRelationshipLabel(relationship, true);
        }
      } else {
        canonicalLabel = (primaryView == null || primaryViewChildPath == null) ? null :
            primaryView.get(CProp.LABEL, primaryViewChildPath);
      }

      if (primaryView != null && primaryViewChildPath == null) {
        final boolean okToRemove;
        if (primaryRoot) {
          final boolean isFormula =
              !childPath.isRelation() && view.get(QProp.SQL_COLUMN_NAME, childPath).isEmpty();
          okToRemove = !isFormula;
        } else {
          final boolean isSQLColumn =
              !childPath.isRelation() && !view.get(QProp.SQL_COLUMN_NAME, childPath).isEmpty();
          okToRemove = isSQLColumn && !usedPaths.contains(childPath);
        }
        if (okToRemove)
          view.removePath(childPath);
      }
      if (view.exists(childPath)) {
        if (canonicalLabel == null) {
          view.setFromDefault(CProp.CANONICAL_LABEL, childPath);
        } else {
          view.set(CProp.CANONICAL_LABEL, childPath, canonicalLabel);
        }
        if (primaryView != null && primaryViewChildPath != null) {
          ViewPropertySource propertySource =
              primaryView.getViewPropertySource(primaryViewChildPath);
          setInheritedProperties(view, childPath, CHILD_INHERITABLE_PROPERTIES, propertySource);
          view.setFrom(QProp.IN_PRIMARY_KEY, childPath, propertySource);
        } else {
          view.setFrom(QProp.IN_PRIMARY_KEY, childPath, view.getDefaults());
        }
        if (childPath.isRelation()) {
          updateDerivedView(
              globals, view, childPath.getRelation(), canonicalLabel == null, usedPaths);
        }
      }
    }

    for (Relationship relationship : remainingGlobalRelationships) {
      RelationPath childPath =
          Relationships.addPathForRelationship(view, path, relationship, false, false);
      if (childPath != null) {
        view.set(CProp.VISIBLE, childPath, false);
        view.set(CProp.LABEL, childPath,
            globals.generateRelationshipLabel(relationship, false));
        view.set(CProp.CANONICAL_LABEL, childPath,
            globals.generateRelationshipLabel(relationship, true));
        updateDerivedView(globals, view, childPath, false, usedPaths);
      }
    }

    if (!view.get(CProp.EXPANDED, path) && !usedPaths.contains(path))
      Queries.showHeadingPathsOnly(view, path);
  }

  private static void setInheritedProperties(
      MutableView view, Path path, Set<ViewProperty<?>> properties, ViewPropertySource source)
  {
    if (view.get(CProp.INHERIT_STYLES, path)) {
      for (ViewProperty<?> prop : properties)
        view.setFrom(prop, path, source);
    }
  }

  /**
   * Match child paths in {@code view} to child paths in {@code sourceView}, creating new paths in
   * the former as appropriate.
   */
  private static PathMap processSourcePaths(
      MutableView view, RelationPath path, View sourceView, RelationPath sourcePath,
      boolean createRelations)
  {
    final Map<Path,Path> ret = Maps.newLinkedHashMap();
    final List<Path> childPathOrder = Lists.newArrayList(view.get(SProp.CHILD_PATHS, path));
    Path previousSibling = null;
    final Map<FieldKey,Path> existingPathsByFieldKey =
        MapUtil.invert(FieldKey.getFieldKeys(view, path, true));
    final Map<Path,FieldKey> sourcePathsByFieldKey =
        FieldKey.getFieldKeys(sourceView, sourcePath, createRelations);
    final boolean primaryRoot = path.isRoot() && view.get(QProp.PRIMARY, Path.ROOT);
    for (Entry<Path,FieldKey> entry : sourcePathsByFieldKey.entrySet()) {
      final FieldKey fieldKey = entry.getValue();
      final Path sourceChildPath = entry.getKey();
      Path childPath = existingPathsByFieldKey.get(fieldKey);
      if (childPath == null) {
        /* TODO[pd]: Consider copying the source field with appropriate path transformations
                     instead. */
        childPath = fieldKey.createField(view, path);
        if (childPath != null) {
          /* Preserve the order of existing fields in the view while inserting new ones after the
          fields they ought to follow according to the order in sourceView. */
          final int insertPosition;
          if (previousSibling == null) {
            insertPosition = 0;
          } else {
            final int previousSiblingPosition = childPathOrder.indexOf(previousSibling);
            if (previousSiblingPosition < 0)
              throw new AssertionError();
            insertPosition = previousSiblingPosition + 1;
          }
          childPathOrder.add(insertPosition, childPath);
          view.set(CProp.INHERIT_STYLES, childPath, true);
          view.set(CProp.VISIBLE, childPath,
               primaryRoot ? sourceView.get(CProp.VISIBLE, sourceChildPath) : false);
        }
      }
      if (childPath != null) {
        previousSibling = childPath;
        ret.put(sourceChildPath, childPath);
      }
    }
    view.set(SProp.CHILD_PATHS, path, new PathList(childPathOrder));
    return new PathMap(ret);
  }

  private static void removeReverseRelationships(MutableView view, RelationPath path,
      @Nullable Relationship relationship, boolean savePluralRelationships, Set<Path> usedPaths)
  {
    final Relationship reverseRelationship =
        (relationship == null) ? null : relationship.reverse();
    for (Path childPath : view.getChildPaths(path.getRelation())) {
      if (childPath.isRelation()) {
        Relationship childRelationship =
            Relationships.getRelationshipForExistingPath(view, childPath.getRelation());
        if (!(savePluralRelationships && view.get(CProp.PLURAL, childPath)) &&
            !usedPaths.contains(childPath) &&
            Objects.equals(childRelationship, reverseRelationship))
        {
          view.removePath(childPath);
        } else {
          removeReverseRelationships(
              view, childPath.getRelation(), childRelationship, savePluralRelationships, usedPaths);
        }
      }
    }
  }
}
