// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.filter;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Position;
import com.sieuferd.evaluator.DataProperties;
import com.sieuferd.evaluator.FunctionalDependencies;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.evaluator.OrderedPath;
import com.sieuferd.project.NBViewSet;
import com.sieuferd.query.Filter;
import com.sieuferd.datamodel.expressions.TupleConstant;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.FunctionExpression;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.datamodel.expressions.RelationConstant;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.ObjectReference;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

@Immutable
public final class FilterOptionsQuery {
  private static final Logger LOG = Logger.getLogger(FilterOptionsQuery.class.getName());
  private static final Set<ViewProperty<?>> FILTER_OPTIONS_PROPERTIES = QueryUtil.QUERY_PROPERTIES;
  private static final ViewDefaults VIEW_DEFAULTS = NBViewSet.VIEW_DEFAULTS.createSubset(
      Sets.difference(FILTER_OPTIONS_PROPERTIES, ImmutableSet.of(SProp.CHILD_PATHS)));
  private final Path filterPath;
  private final ImmutableSet<PrimitivePath> constrainedPaths;
  private final ImmutableSet<PrimitivePath> displayedPaths;
  // Kept for O(1) lookup by index; otherwise redundant with displayedPaths.
  private final ImmutableList<PrimitivePath> displayedPathsInOrder;
  private final PathMap dataPaths;
  private final PathMap containerPaths;
  private final RelationPath containerRootPath;
  private final ImmutableList<OrderedPath> preservedOrderedPaths;
  private final ImmutableView sourceView;
  // Keeping this allows us to keep the containerPaths PathMap constant.
  private final ImmutableView optionsViewTemplate;

  /** @param displayedPaths must be fully identified by {@code constrainedPaths} in {@code view} */
  private FilterOptionsQuery(ImmutableView view, Path filterPath,
      Set<PrimitivePath> constrainedPaths, Set<PrimitivePath> displayedPaths,
      List<OrderedPath> preservedOrderedPaths)
  {
    Preconditions.checkNotNull(filterPath);
    LOG.log(Level.INFO, "Created filter with constrained paths {0}, displayed paths {1}",
        new Object[] { constrainedPaths, displayedPaths });
    this.filterPath = filterPath;
    this.constrainedPaths = ImmutableSet.copyOf(constrainedPaths);
    this.displayedPaths = ImmutableSet.copyOf(Views.getPathsInViewOrder(view, displayedPaths));
    this.displayedPathsInOrder = ImmutableList.copyOf(this.displayedPaths);
    this.sourceView = ImmutableView.copyOf(view, FILTER_OPTIONS_PROPERTIES, null);
    this.preservedOrderedPaths = ImmutableList.copyOf(preservedOrderedPaths);
    final Map<PrimitivePath,PrimitivePath> toDataPaths = Maps.newLinkedHashMap();
    for (PrimitivePath sourcePath : getSourcePathsToMap())
      MapUtil.putChecked(toDataPaths, sourcePath, PrimitivePath.create(Path.ROOT));
    this.dataPaths = new PathMap(toDataPaths);

    {
      final MutableView ovt = Views.newMutable(VIEW_DEFAULTS);
      this.containerRootPath = ovt.addRelationPath(Path.ROOT);
      this.containerPaths = ovt.copyFrom(
          sourceView, Path.ROOT, containerRootPath, FILTER_OPTIONS_PROPERTIES, null);
      // Set labels for debugging purposes.
      ovt.set(CProp.LABEL, Path.ROOT,
          "Filter Options View (" + filterPath.toString(sourceView) + ")");
      ovt.set(QProp.ORDER_RANK, containerRootPath, Position.NONE);
      Views.rewritePathReferences(ovt, containerRootPath, containerPaths);
      ovt.set(CProp.VISIBLE, containerRootPath, false);
      for (RelationPath path : Views.getRelationPathsWithAncestors(
          ImmutableSet.of(Views.getContextPath(containerPaths.getNewPath(filterPath)))))
      {
        if (!path.isRoot())
          ovt.set(QProp.EXISTENTIAL_FILTER, path, true);
      }
      this.optionsViewTemplate = ImmutableView.copyOf(ovt);
    }
  }

  private Set<PrimitivePath> getSourcePathsToMap() {
    Set<PrimitivePath> ret = Sets.newLinkedHashSet();
    ret.addAll(this.constrainedPaths);
    ret.addAll(this.displayedPathsInOrder);
    for (OrderedPath orderedPath : this.preservedOrderedPaths)
      ret.add(orderedPath.getPath());
    return Collections.unmodifiableSet(ret);
  }

  /** @param fullyIdentifiedPaths may have additional dependents added to it */
  private static void addDisplayedPaths(View view, FunctionalDependencies functionalDependencies,
      Set<PrimitivePath> fullyIdentifiedPaths, Path path, Set<PrimitivePath> toDisplayedPaths)
  {
    if (path.isRelation()) {
      for (Path childPath : view.getChildPaths(path.getRelation())) {
        // The starting path is always included, so apply these conditions on children only.
        if (view.get(CProp.VISIBLE, childPath)) {
          addDisplayedPaths(
              view, functionalDependencies, fullyIdentifiedPaths, childPath, toDisplayedPaths);
        }
      }
    } else {
      if (functionalDependencies.isIdentifiedBy(view, path.getPrimitive(), fullyIdentifiedPaths))
        toDisplayedPaths.add(path.getPrimitive());
    }
  }

  public static FilterOptionsQuery create(
      ImmutableView view, Path filterPath, Set<PrimitivePath> constrainedPaths,
      Set<Path> selectedPaths)
  {
    Preconditions.checkNotNull(view);
    final FunctionalDependencies functionalDependencies = new FunctionalDependencies(view);
    final Set<PrimitivePath> displayedPaths = Sets.newLinkedHashSet();
    final Set<PrimitivePath> fullyIdentifiedPaths = Sets.newLinkedHashSet(constrainedPaths);
    for (Path path : selectedPaths)
      addDisplayedPaths(view, functionalDependencies, fullyIdentifiedPaths, path, displayedPaths);
    if (displayedPaths.isEmpty()) {
      /* Handle the case where the filter was opened on a set of fields disjoint from the set of
      constrained paths. */
      addDisplayedPaths(
          view, functionalDependencies, fullyIdentifiedPaths, filterPath, displayedPaths);
    }
    final List<OrderedPath> preservedOrderedPaths;
    if (filterPath.isRelation()) {
      preservedOrderedPaths =
          QueryUtil.getOrderedPaths(view, filterPath.getRelation(), functionalDependencies, false);
    } else if (!view.get(QProp.ORDER_RANK, filterPath).equals(Position.NONE)) {
      preservedOrderedPaths = ImmutableList.of(new OrderedPath(filterPath.getPrimitive(),
          view.get(QProp.ORDER_DESCENDING, filterPath),
          view.get(QProp.ORDER_NULLS_FIRST, filterPath)));
    } else {
      preservedOrderedPaths = ImmutableList.of();
    }
    return new FilterOptionsQuery(
        view, filterPath, constrainedPaths, displayedPaths, preservedOrderedPaths);
  }

  public Path getFilterPath() {
    return filterPath;
  }

  public Set<PrimitivePath> getConstrainedPaths() {
    return constrainedPaths;
  }

  public Set<PrimitivePath> getDisplayedPaths() {
    return displayedPaths;
  }

  public ImmutableList<PrimitivePath> getDisplayedPathsInOrder() {
    return displayedPathsInOrder;
  }

  public @Nullable PrimitiveValue getValue(Data data, Tuple tuple, PrimitivePath displayedPath) {
    final PrimitivePath path = dataPaths.getNewPath(displayedPath);
    return data.isProjected(path) ? data.getTupleField(tuple, path) : null;
  }

  /**
   * Should cast value to a lowercase string that does not contain tabs. May also do other
   * whitespace cleanup.
   */
  private static Expression normalizedSearchString(Expression exp, boolean normalizeWhitespace) {
    final Expression ret =
        fun(Function.STRING_LOWER,
        fun(Function.STRING_CONCAT, exp));
    return normalizeWhitespace ? fun(Function.STRING_NORMALIZE_WHITESPACE, ret) : ret;
  }

  private static Expression fun(Function function, Expression... exp) {
    return new FunctionExpression(function, Arrays.asList(exp));
  }

  public static List<String> getSearchTerms(String searchString) {
    final List<String> ret = Lists.newArrayList();
    int termStart = 0;
    int i = 0;
    while (i < searchString.length()) {
      final char c = searchString.charAt(i);
      if ((c == '\"' || c == '\'') && termStart == i) {
        final ObjectReference<Integer> tokenLength = ObjectReference.create();
        String unescaped =
            Expressions.readOneStringLiteral(searchString.substring(i), tokenLength);
        if (unescaped != null) {
          if (!unescaped.isEmpty())
            ret.add(unescaped);
          i += tokenLength.getValue();
          termStart = i;
          continue;
        }
      }
      if (Character.isWhitespace(c)) {
        if (termStart < i)
          ret.add(searchString.substring(termStart, i));
        termStart = i + 1;
      }
      i++;
    }
    if (termStart < i)
      ret.add(searchString.substring(termStart, i));
    return Collections.unmodifiableList(ret);
  }

  public ImmutableView getFilterOptionsView(String searchString, boolean applyExistingFilter) {
    ObjectReference<PrimitivePath> toSearchFilterPath = ObjectReference.create();
    ObjectReference<PrimitivePath> toSearchOrderPath = ObjectReference.create();
    ObjectReference<PrimitivePath> toHaystackPath = ObjectReference.create();
    // TODO[cc]: Just inline createFilterOptionsView.
    ImmutableView filterOptionsView =
        createFilterOptionsView(toSearchFilterPath, toSearchOrderPath, toHaystackPath);
    final PrimitivePath haystackPath = toHaystackPath.getValue();
    final PrimitivePath searchFilterPath = toSearchFilterPath.getValue();
    final PrimitivePath searchOrderPath = toSearchOrderPath.getValue();

    MutableView ret = Views.newMutable(filterOptionsView);
    List<String> searchTerms = getSearchTerms(searchString);
    Expression haystackExpression = new PathExpression(haystackPath);

    if (!applyExistingFilter)
      ret.set(QProp.FILTER, containerPaths.getNewPath(filterPath), Filter.UNCONSTRAINED);

    if (searchTerms.isEmpty()) {
      /* A small optimization for the common case where no query is specified (though the database
      might still end up calculating the haystack string). */
      ret.set(QProp.FORMULA_EXPRESSION, searchFilterPath, ConstantExpression.TRUE);
    } else {
      // Include the result if each search term is contained in some field.
      List<Expression> allSearchTermsFound = Lists.newArrayList();
      for (String searchTerm : searchTerms) {
        allSearchTermsFound.add(new FunctionExpression(Function.STRING_CONTAINS,
            ImmutableList.of(haystackExpression,
            normalizedSearchString(ConstantExpression.create(searchTerm), false))));
      }
      ret.set(QProp.FORMULA_EXPRESSION, searchFilterPath, Expressions.and(allSearchTermsFound));
    }

    /* If the raw query string is contained at the start of any field, order those tuples first,
    while also prioritizing earlier fields. */
    {
      Expression needleExpression = fun(Function.STRING_CONCAT,
          ConstantExpression.create("\t"),
          normalizedSearchString(ConstantExpression.create(searchString), true));
      ret.set(QProp.ORDER_NULLS_FIRST, searchOrderPath, false);
      ret.set(QProp.FORMULA_EXPRESSION, searchOrderPath,
          fun(Function.STRING_POSITION, haystackExpression, needleExpression));
    }
    return ImmutableView.copyOf(ret);
  }

  private static Filter createPathIsTrueFilter(PrimitivePath path) {
    return Filter.create(new RelationConstant(ImmutableSet.of(path),
        ImmutableSet.of(new TupleConstant(ImmutableMap.of(path, ConstantExpression.TRUE)))), false);
  }

  private ImmutableView createFilterOptionsView(ObjectReference<PrimitivePath> searchFilterPath,
      ObjectReference<PrimitivePath> searchOrderPath, ObjectReference<PrimitivePath> haystackPath)
  {
    final MutableView ret = Views.newMutable(optionsViewTemplate, VIEW_DEFAULTS);
    for (PrimitivePath sourcePath : getSourcePathsToMap()) {
      PrimitivePath containerPath = containerPaths.getNewPath(sourcePath);
      PrimitivePath dataPath = dataPaths.getNewPathIfMapped(sourcePath);
      if (dataPath == null)
        throw new AssertionError();
      ret.addPath(dataPath);
      ret.set(CProp.LABEL, dataPath, "fp_data/" + sourcePath.toString(sourceView));
      ret.set(QProp.FORMULA_EXPRESSION, dataPath, new PathExpression(containerPath));
      ret.set(CProp.VISIBLE, dataPath,
          constrainedPaths.contains(sourcePath) || displayedPathsInOrder.contains(sourcePath));
    }
    searchFilterPath.setValue(ret.addPrimitivePath(Path.ROOT));
    ret.set(CProp.LABEL, searchFilterPath.getValue(), "fp_searchFilter");
    searchOrderPath.setValue(ret.addPrimitivePath(Path.ROOT));
    ret.set(CProp.LABEL, searchOrderPath.getValue(), "fp_searchOrder");
    ret.set(QProp.FILTER, searchFilterPath.getValue(),
        createPathIsTrueFilter(searchFilterPath.getValue()));

    DataProperties.configureDataProperties(ret);
    Queries.sort(ret, searchOrderPath.getValue(), false, false);
    final Set<PrimitivePath> displayedDataPathsInSortOrder = Sets.newLinkedHashSet();
    // First sort according to the explicitly defined sort order.
    for (OrderedPath orderedPath : preservedOrderedPaths) {
      PrimitivePath dataPath = dataPaths.getNewPathIfMapped(orderedPath.getPath());
      if (dataPath == null)
        throw new AssertionError("Expected all preserved sort paths to be mapped");
      displayedDataPathsInSortOrder.add(dataPath);
      Queries.sort(ret, dataPath, orderedPath.isDescending(), true);
      /* TODO[pd]: Consider having Queries.sort take an OrderedPath as an argument, or
                   taking a nullsFirst argument as well. */
      ret.set(QProp.ORDER_NULLS_FIRST, dataPath, orderedPath.isNullsFirst());
    }
    // Sort any remaining displayed paths in display order.
    for (PrimitivePath path : displayedPathsInOrder) {
      PrimitivePath dataPath = dataPaths.getNewPath(path);
      if (!displayedDataPathsInSortOrder.contains(dataPath)) {
        displayedDataPathsInSortOrder.add(dataPath);
        Queries.sort(ret, dataPath, false, true);
        // Order null values first to make them stand out in the filter options list.
        ret.set(QProp.ORDER_NULLS_FIRST, dataPath, true);
      }
    }

    {
      haystackPath.setValue(ret.addPrimitivePath(containerRootPath));
      ret.set(CProp.LABEL, haystackPath.getValue(), "fp_haystack");
      List<Expression> concatArgs = Lists.newArrayList();
      Expression leadingSeparatorExpression = ConstantExpression.create("\t");
      for (PrimitivePath path : displayedDataPathsInSortOrder) {
        concatArgs.add(leadingSeparatorExpression);
        PrimitivePath originalPath = dataPaths.getOldPathIfMapped(path);
        if (originalPath == null)
          throw new AssertionError();
        // Normalizing the string to be searched will also remove any tab characters.
        concatArgs.add(normalizedSearchString(
            new PathExpression(containerPaths.getNewPath(originalPath)), true));
      }
      ret.set(QProp.FORMULA_EXPRESSION, haystackPath.getValue(),
          new FunctionExpression(Function.STRING_CONCAT, concatArgs));
    }
    return ImmutableView.copyOf(ret);
  }

  // TODO[cc]: Avoid exposing this, or rename.
  public PathMap getContainerPaths() {
    return containerPaths;
  }
}
