// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.filter;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.LinkedListMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.database.dialect.JDBCObjectWrapper;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.ViewFlow;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.Flows;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate2;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.flow.ValueProcessor.Synchronicity;
import com.sieuferd.layout.LBProp;
import com.sieuferd.layout.LHProp;
import com.sieuferd.layout.LMProp;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.result.EvaluationInput;
import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.project.result.QueryProcessor;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.query.Filter;
import com.sieuferd.query.QProp;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.uiutil.EnterFilterInputMap;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.flow.UIModelFlows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.uiutil.table.ColorOverrideRenderer;
import com.sieuferd.uiutil.table.SpanningTableCellRenderer;
import com.sieuferd.uiutil.api.TableColors;
import com.sieuferd.util.StringUtil;
import com.sieuferd.util.UnitObject;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.BoundedRangeModel;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ChangeEvent;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import org.openide.util.NbBundle;

@NbBundle.Messages({
  // The wording here should remain consistent with the wording used in FilterTupleTableModel.
  "# {0} - the number of values included by the filter",
  "MSG_FilterDescription_Including=" +
      "Including {0, choice,0#no values|1#one value|2#{0} values}",
  "# {0} - the number of values excluded by the filter",
  "MSG_FilterDescription_Excluding=" +
      "Including {0, choice,0#all values|1#all except one value|2#all except {0} values}",
})
public final class FilterPanel extends JPanel {
  private static final Logger LOG = Logger.getLogger(FilterPanel.class.getName());
  /* I've kept the infinite scrolling logic here separate from that of ViewResultTopComponent, as
  the latter will likely become more specialized and complicated in the future. */
  private static final int INITIAL_OPTIONS_QUERY_LIMIT = 500;
  private static final double OPTIONS_QUERY_LIMIT_INCREASE_FACTOR = 2.0;
  // For consistency, I just always display the header.
  private static final boolean HIDE_HEADER_FOR_SINGLE_COLUMN = false;
  private static final int SEARCH_FIELD_COALESCE_MS = 200;
  /* I considered using the MacOS-style search box, but decided against it, as the "Search" label is
  a little more explicit, and has to be included on Windows anyway. */
  private static final boolean USE_AQUA_SEARCH_BOX = false;
  private final SettableFlow<UnitObject> openedSettable = Flows.newSettable();
  private final SettableFlow<FieldSelection> fieldSelectionSettable = Flows.newSettable();
  private final Flow<String> searchStringFlow;
  private final SettableFlow<Integer> queryLimitSettable = Flows.newSettable();
  private @Nullable FilterTupleTableModel tableModel = null;
  private ImmutableData optionsData = QueryUtil.EMPTY_INCOMPLETE_DATA;
  private boolean scrollToTopDue = false;

  public FilterPanel() {
    ThreadUtil.checkOnEDT();
    initComponents();
    searchStringFlow = Flows.newCoalescing(SEARCH_FIELD_COALESCE_MS,
        UIModelFlows.document(searchField.getDocument()));
    // TODO[pd]: Consider allowing this, modifying the sort order to always be left-to-right.
    tupleTable.getTableHeader().setReorderingAllowed(false);
    tupleTable.setFillsViewportHeight(true);
    tupleTable.setShowVerticalLines(false);
    tupleTable.setShowHorizontalLines(false);
    tupleTable.setIntercellSpacing(new Dimension(0, 0));
    tupleTable.setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
    tupleTable.getSelectionModel().setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    if (USE_AQUA_SEARCH_BOX && UIUtil.isAqua()) {
      // TODO[pd]: Tweak the text field's padding a little bit in this case.
      searchField.putClientProperty("JTextField.variant", "search");
      remove(searchLabel);
    }
    UIUtil.setSimpleFocusTraversalKeys(searchField);
    UIUtil.setSimpleFocusTraversalKeys(tupleTable);
    UIUtil.configureBooleanTableValueShortcuts(tupleTable, 0);
    TableColors.registerFocusRepaint(tupleTable);

    /* Listen on the vertical scrollbar rather than the viewport; I had some trouble with the
    latter, getting outdated position values. Be sure to also query the scrollbar in
    updateQueryLimit. */
    tupleScrollPane.getVerticalScrollBar().getModel().addChangeListener((ChangeEvent evt) -> {
      updateQueryLimit();
    });

    searchField.addKeyListener(new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        final int keyCode = e.getKeyCode();
        /* Move focus to to tuple table on down arrow press. Consume both up and down arrows for
        consistency. */
        if (e.getModifiersEx() == 0 && (keyCode == KeyEvent.VK_DOWN || keyCode == KeyEvent.VK_UP)) {
          if (keyCode == KeyEvent.VK_DOWN)
            tupleTable.requestFocusInWindow();
          e.consume();
        }
      }
    });
    UIUtil.ensureSelectionOnFocusGained(tupleTable);
    EnterFilterInputMap.configure(tupleTable);

    final Flow<FilterParameters> filterParametersFlow = Flows.newValve(openedSettable,
        Flows.newProcessed(fieldSelectionSettable,
        /* TODO[pd]: Make this properly context sensitive. If the view changes in a way that would
                     change the filter parameters, the filter parameters should be updated
                     appropriately. */
        fieldSelection -> FilterParameters.fromFieldSelection(fieldSelection)));
    final Flow<Filter> filterFlow =
        Flows.newIndirect(filterParametersFlow, parameters ->
        Flows.newProcessed(new ViewFlow(parameters.getViewContext().getView(),
        ImmutableSet.of(QProp.FILTER), ImmutableSet.of(parameters.getFilterPath())),
        view -> view.getOrNull(QProp.FILTER, parameters.getFilterPath())));
    final Flow<ViewContext> viewContextFlow = Flows.newProcessed(filterParametersFlow,
        filterParameters -> filterParameters.getViewContext());
    final Flow<String> databaseConnectionName = Flows.newIndirect(viewContextFlow,
        viewContext -> viewContext.getProject().getConfiguration().getDatabaseConnectionNameFlow());
    final Flow<EvaluationInput> optionsEvaluationInput = Flows.newComposite4(
        filterParametersFlow, databaseConnectionName, searchStringFlow, queryLimitSettable,
        (filterParameters, connectionName, searchString, queryLimit) -> new EvaluationInput(
        connectionName, filterParameters.getFilterOptionsView(searchString), queryLimit));

    // Add warnings for options tuples that cannot be used as filters for various reasons.
    final Flow<ViewResult> optionsOutput =
        new ProcessorTemplate2<ViewResult,ViewResult,FilterParameters>()
    {
      @Override
      public Synchronicity getSynchronicity() {
        return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
      }

      @Override
      protected ViewResult process(ViewResult result, FilterParameters parameters)
          throws InterruptedException
      {
        final Data data = result.getData();
        final Multimap<PrimitivePath,String> problemClassNames = LinkedHashMultimap.create();
        final Set<PrimitivePath> floatingProblemPaths = Sets.newLinkedHashSet();
        for (Tuple tuple : data.getRelationTuples(Path.ROOT, Tuple.createRootParent(data))) {
          if (Thread.interrupted())
            throw new InterruptedException();
          for (PrimitivePath path : parameters.getConstrainedPaths()) {
            PrimitiveValue value = parameters.getValue(data, tuple, path);
            if (value == null)
              continue;
            Object content = value.getContent();
            if (ConstantExpression.fromObject(content) == null) {
              if (content instanceof Double || content instanceof Float) {
                floatingProblemPaths.add(path);
              } else {
                final Class<?> problemClass;
                if (content instanceof JDBCObjectWrapper) {
                  problemClass = ((JDBCObjectWrapper) content).getDelegateClass();
                } else {
                  problemClass = content.getClass();
                }
                problemClassNames.put(path, problemClass.getCanonicalName());
              }
            }
          }
        }
        if (!floatingProblemPaths.isEmpty() || !problemClassNames.isEmpty()) {
          final Multimap<Path,Warning> warnings = LinkedListMultimap.create(result.getWarnings());
          for (PrimitivePath path : floatingProblemPaths) {
            parameters.addWarning(warnings, path,
                new Warning("Filtering not supported on floating-point Infinity or NaN values"));
          }
          for (PrimitivePath path : problemClassNames.keySet()) {
            parameters.addWarning(warnings, path,
                new Warning("Filtering not supported on this data type: " +
                StringUtil.join(", ", problemClassNames.get(path))));
          }
          return new ViewResult(result.getLimit(), result.getView(), result.getData(), warnings);
        } else {
          return result;
        }
      }
    }.getFlow(Flows.newProcessed(
        optionsEvaluationInput, new QueryProcessor("Filter")), filterParametersFlow);

    /* Enclose the sticky flow in a valve to avoid storing results when the filter is closed. The
    latter behavior caused old warnings to flash before the user's eye for a moment before new
    filter options were computed. */
    final Flow<ViewResult> stickyOptionsOutput =
        Flows.newValve(openedSettable,
        Flows.newSticky(optionsOutput, QueryProcessor.createEmptyOutput()));
    UIFlows.bindEDT(filterParametersFlow, (FilterParameters param) -> {
      scrollToTopDue = true;
      final boolean headerVisible;
      if (param == null) {
        tupleTable.setModel(new DefaultTableModel());
        tableModel = null;
        headerVisible = false;
        optionsData = QueryUtil.EMPTY_INCOMPLETE_DATA;
        queryLimitSettable.set(null);
        updateData();
        updateHighlightStrings();
      } else {
        final List<PrimitivePath> displayedPaths = param.getDisplayedPathsInOrder();
        boolean usePaddingColumn = true;
        for (int i = 0; i < displayedPaths.size(); i++) {
          final PrimitivePath path = displayedPaths.get(i);
          if ((i == displayedPaths.size() - 1)
              ? param.getEffectiveHorizontalAlignment(path) == HorizontalAlignment.LEFT
              : param.getStyleProperty(LMProp.VARIABLE_WIDTH, path))
          {
            usePaddingColumn = false;
            break;
          }
        }
        tableModel = new FilterTupleTableModel(param, usePaddingColumn);
        tupleTable.setModel(tableModel);
        final TableColumnModel columnModel = tupleTable.getColumnModel();
        final TableColumn checkboxColumn = columnModel.getColumn(0);
        {
          int checkboxWidth = new JCheckBox().getPreferredSize().width;
          checkboxColumn.setMaxWidth(checkboxWidth);
          checkboxColumn.setMinWidth(checkboxWidth);
        }
        queryLimitSettable.set(INITIAL_OPTIONS_QUERY_LIMIT);
        updateData();
        updateHighlightStrings();
        final double fontSize = tupleTable.getFont().getSize2D();
        TableColors tableColors = TableColors.getInstance();
        checkboxColumn.setCellRenderer(new ColorOverrideRenderer(
            tupleTable.getDefaultRenderer(Boolean.class), tableColors));
        TableCellRenderer cellRenderer = new SpanningTableCellRenderer(true, tableColors);
        for (int i = 0; i < displayedPaths.size(); i++) {
          final PrimitivePath path = displayedPaths.get(i);
          final double emWidth =
              Math.min(param.getStyleProperty(LMProp.VALUE_WIDTH, path).getValue(),
                       param.getStyleProperty(LHProp.TABLE_MAX_PRIMITIVE_WIDTH, path).getValue())
              / param.getStyleProperty(LBProp.VALUE_FONT, path).getSize2D();
          final int width = (int) Math.ceil(fontSize * emWidth);
          final TableColumn pathColumn = columnModel.getColumn(i + 1);
          if (!param.getStyleProperty(LMProp.VARIABLE_WIDTH, path)) {
            /* Extra margin added by SpanningTableCellRenderer in the case of right-aligned
            values. */
            final int extraMargin =
                (param.getEffectiveHorizontalAlignment(path) == HorizontalAlignment.RIGHT
                ? SpanningTableCellRenderer.RIGHT_MARGIN_PX : 0);
            // Ensure we don't truncate the values here.
            final int minWidth = Math.round(width * 1.05f) + extraMargin;
            pathColumn.setPreferredWidth(minWidth);
            pathColumn.setMinWidth(minWidth);
            if (i + 1 < columnModel.getColumnCount() - 1)
              pathColumn.setMaxWidth(Math.round(width * 1.5f) + extraMargin);
          } else {
            pathColumn.setPreferredWidth(width);
          }
          pathColumn.setCellRenderer(cellRenderer);
        }
        if (usePaddingColumn) {
          final TableColumn paddingColumn = columnModel.getColumn(columnModel.getColumnCount() - 1);
          paddingColumn.setCellRenderer(cellRenderer);
          paddingColumn.setMinWidth(0);
          paddingColumn.setPreferredWidth(0);
        }
        headerVisible = HIDE_HEADER_FOR_SINGLE_COLUMN ? (displayedPaths.size() > 1) : true;
      }
      tupleTable.getTableHeader().setVisible(headerVisible);
      tupleTable.getTableHeader().setPreferredSize(headerVisible ? null : new Dimension(0, 0));
    });
    UIFlows.bindEDT(optionsOutput, (ViewResult viewResult) -> {
      progressSpinner.setRunning(viewResult == null);
      if (viewResult != null) {
        optionsData = viewResult.getData();
        updateData();
      }
    });
    final Flow<String> messageFlow = Flows.newComposite3(
        filterFlow, filterParametersFlow, stickyOptionsOutput,
        (Filter filter, FilterParameters filterParameters, ViewResult result) ->
    {
      StringBuilder ret = new StringBuilder();
      ret.append("<html>\n");

      final int tupleCount = filter.getTuples().size();
      final String description = filter.isExcluding()
          ? Bundle.MSG_FilterDescription_Excluding(tupleCount)
          : Bundle.MSG_FilterDescription_Including(tupleCount);
      ret.append(SwingString.toSwingHTML(description)).append("<br>\n");

      if (LOG.isLoggable(Level.INFO) && !result.getWarnings().isEmpty()) {
        StringBuilder msg = new StringBuilder();
        msg.append("Filter options evaluation warnings:\n");
        for (Entry<Path,Warning> entry : result.getWarnings().entries()) {
          msg.append("  ").append(entry.getKey()).append(": ")
              .append(entry.getValue().getDisplayMessage()).append("\n");
        }
        LOG.info(msg.toString());
      }
      ret.append(filterParameters.generateWarningMessage(result).getForConcatenation());

      ret.append("</html>");
      return ret.toString();
    });
    UIFlows.bindEDT(messageFlow, (String message) -> {
      messageLabel.setText(message == null ? "" : message);
      UIUtil.forceSwingHtmlSizeComputation(messageLabel);
    });
    UIFlows.bindEDT(searchStringFlow, (String query) -> {
      scrollToTopDue = true;
      queryLimitSettable.set(INITIAL_OPTIONS_QUERY_LIMIT);
      // TODO[cc]: Try to do a proper data flow here instead.
      updateHighlightStrings();
    });
  }

  // Not yet used.
  private void invert() {
    if (tableModel != null)
      tableModel.invert();
  }

  private void updateQueryLimit() {
    if (tableModel == null || !optionsData.isIncomplete())
      return;
    final BoundedRangeModel scrollModel = tupleScrollPane.getVerticalScrollBar().getModel();
    if (scrollModel.getValueIsAdjusting())
      return;
    int bottomRowIndex = tupleTable.rowAtPoint(new Point(0,
        scrollModel.getValue() + scrollModel.getExtent()));
    if (bottomRowIndex < 0)
      bottomRowIndex = tableModel.getRowCount() - 1;
    final int visibleTupleCount =
        Math.max(0, bottomRowIndex + 1 - tableModel.getNonDataRowCount());
    final int retrievedTupleCount = optionsData.getCompleteTupleCount(Path.ROOT);
    if (visibleTupleCount >= retrievedTupleCount) {
      queryLimitSettable.set(Math.max(INITIAL_OPTIONS_QUERY_LIMIT, (int) Math.min(Integer.MAX_VALUE,
          Math.round(retrievedTupleCount * OPTIONS_QUERY_LIMIT_INCREASE_FACTOR))));
    }
  }

  private void updateHighlightStrings() {
    ThreadUtil.checkOnEDT();
    String query = searchField.getText();
    Set<String> highlightStrings = Sets.newLinkedHashSet(FilterOptionsQuery.getSearchTerms(query));
    if (!StringUtil.isEmptyOrSpaceOnly(query))
      highlightStrings.add(query.trim());
    if (tableModel != null)
      tableModel.setHighlightStrings(highlightStrings);
  }

  private void updateData() {
    ThreadUtil.checkOnEDT();
    if (tableModel != null) {
      tableModel.setData(optionsData);
      if (scrollToTopDue)
        tupleScrollPane.getViewport().setViewPosition(new Point(0, 0));
      scrollToTopDue = false;
      updateQueryLimit();
    }
  }

  public void setOpened(boolean opened) {
    if (opened && openedSettable.get() == null)
      searchField.setText("");
    openedSettable.set(opened ? UnitObject.INSTANCE : null);
  }

  public void setFieldContexts(Set<FieldContext> fieldContexts) {
    ThreadUtil.checkOnEDT();
    fieldSelectionSettable.set(FieldSelection.fromFieldContexts(fieldContexts, AppliesTo.ANY));
  }

  // Implement requestFocusInWindow and requestFocus, like in FieldsPanel.

  @Override
  public boolean requestFocusInWindow() {
    if (!isFocusable())
      return false;
    return searchField.requestFocusInWindow();
  }

  @Override
  public void requestFocus() {
    if (!isFocusable())
      return;
    searchField.requestFocus();
  }

  /**
   * This method is called from within the constructor to initialize the form. WARNING: Do NOT
   * modify this code. The content of this method is always regenerated by the Form Editor.
   */
  // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
  private void initComponents() {
    java.awt.GridBagConstraints gridBagConstraints;

    searchLabel = new javax.swing.JLabel();
    searchField = new javax.swing.JTextField();
    tupleScrollPane = new javax.swing.JScrollPane();
    tupleTable = new javax.swing.JTable();
    messageLabel = new javax.swing.JLabel();
    progressSpinner = new com.sieuferd.uiutil.ProgressSpinner();

    setLayout(new java.awt.GridBagLayout());

    searchLabel.setLabelFor(searchField);
    org.openide.awt.Mnemonics.setLocalizedText(searchLabel, org.openide.util.NbBundle.getMessage(FilterPanel.class, "FilterPanel.searchLabel.text")); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    gridBagConstraints.insets = new java.awt.Insets(2, 5, 0, 5);
    add(searchLabel, gridBagConstraints);
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 1;
    gridBagConstraints.gridy = 0;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.anchor = java.awt.GridBagConstraints.BASELINE;
    gridBagConstraints.insets = new java.awt.Insets(2, 0, 0, 2);
    add(searchField, gridBagConstraints);

    tupleScrollPane.setBorder(javax.swing.BorderFactory.createMatteBorder(1, 0, 1, 0, javax.swing.UIManager.getDefaults().getColor("Separator.shadow")));
    tupleScrollPane.setVerticalScrollBarPolicy(javax.swing.ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);

    tupleTable.setModel(new javax.swing.table.DefaultTableModel(
      new Object [][] {
        {null, null, null, null},
        {null, null, null, null},
        {null, null, null, null},
        {null, null, null, null}
      },
      new String [] {
        "Title 1", "Title 2", "Title 3", "Title 4"
      }
    ));
    tupleTable.setFillsViewportHeight(true);
    tupleScrollPane.setViewportView(tupleTable);

    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 1;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.weighty = 1.0;
    gridBagConstraints.insets = new java.awt.Insets(2, 0, 2, 0);
    add(tupleScrollPane, gridBagConstraints);

    org.openide.awt.Mnemonics.setLocalizedText(messageLabel, "<html>Filter description goes here.</html>"); // NOI18N
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 0;
    gridBagConstraints.gridy = 2;
    gridBagConstraints.gridwidth = 2;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.weightx = 1.0;
    gridBagConstraints.insets = new java.awt.Insets(0, 5, 3, 5);
    add(messageLabel, gridBagConstraints);
    gridBagConstraints = new java.awt.GridBagConstraints();
    gridBagConstraints.gridx = 2;
    gridBagConstraints.gridy = 2;
    gridBagConstraints.gridwidth = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.gridheight = java.awt.GridBagConstraints.REMAINDER;
    gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
    gridBagConstraints.insets = new java.awt.Insets(0, 0, 3, 5);
    add(progressSpinner, gridBagConstraints);
  }// </editor-fold>//GEN-END:initComponents

  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JLabel messageLabel;
  private com.sieuferd.uiutil.ProgressSpinner progressSpinner;
  private javax.swing.JTextField searchField;
  private javax.swing.JLabel searchLabel;
  private javax.swing.JScrollPane tupleScrollPane;
  private javax.swing.JTable tupleTable;
  // End of variables declaration//GEN-END:variables
}
