// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.filter;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.cell.styles.HorizontalAlignment;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.FunctionalDependencies;
import com.sieuferd.layout.LBProp;
import com.sieuferd.layout.LHProp;
import com.sieuferd.layout.LMProp;
import com.sieuferd.layout.PrimitiveValueFormatter;
import com.sieuferd.project.ViewContext;
import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.project.result.ViewResultTopComponent;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.query.Filter;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.uiutil.SwingString;
import java.util.AbstractMap.SimpleEntry;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import org.openide.util.NbBundle.Messages;

@Messages("LBL_FilterWarningGeneratedQueryLabel=generated query")
@Immutable
final class FilterParameters {
  private static final boolean DEBUG_GENERATE_FILTER_VIEWS = false;
  private static final Set<ViewProperty<?>> FILTER_STYLE_PROPERTIES =
      ImmutableSet.<ViewProperty<?>>builder()
      .add(CProp.LABEL)
      .add(LBProp.VALUE_FONT)
      .add(LBProp.DISPLAY_LOCALE)
      .add(LBProp.DISPLAY_TIME_ZONE)
      .add(LBProp.TEMPORAL_STYLE)
      .add(LBProp.HORIZONTAL_ALIGNMENT)
      .add(LMProp.VALUE_WIDTH)
      .add(LMProp.VARIABLE_WIDTH)
      .add(LMProp.DECIMAL_PLACES)
      .add(LMProp.GROUP_DIGITS)
      .add(LMProp.TEMPORAL_PRECISION)
      .add(LHProp.TABLE_MAX_PRIMITIVE_WIDTH)
      .build();
  private final FilterOptionsQuery filterOptionsQuery;
  private final ViewContext viewContext;
  private final ImmutableView styleView;

  /** @param displayedPaths must be fully identified by {@code constrainedPaths} in {@code view} */
  private FilterParameters(final ViewContext viewContext, ImmutableView view, Path filterPath,
      Set<PrimitivePath> constrainedPaths, Set<Path> selectedPaths)
  {
    Preconditions.checkNotNull(viewContext);
    Preconditions.checkNotNull(filterPath);
    this.viewContext = viewContext;
    this.styleView = ImmutableView.copyOf(view, FILTER_STYLE_PROPERTIES, null);
    this.filterOptionsQuery =
        FilterOptionsQuery.create(view, filterPath, constrainedPaths, selectedPaths);

    if (DEBUG_GENERATE_FILTER_VIEWS) {
      new Thread(() -> {
        viewContext.getProject().getViewSet().createAndAddView(getFilterOptionsView(""));
      }).start();
    }
  }

  public static FilterParameters fromFieldSelection(FieldSelection fieldSelection) {
    Preconditions.checkNotNull(fieldSelection);
    final ImmutableView view = ImmutableView.copyOf(fieldSelection.getViewContext().getView());
    /* I considered going to the plural ancestor here. However, this might lead to counter-intuitive
    results, e.g. when filtering "Instructor Assignments" instead of "Instructors" (in the course
    catalog example). */
    final Path filterPath = Views.findLowestCommonAncestor(fieldSelection.getSelectedPaths());
    final FunctionalDependencies functionalDependencies = new FunctionalDependencies(view);
    final ImmutableSet<PrimitivePath> constrainedPaths;
    final Filter existingFilter = view.get(QProp.FILTER, filterPath);
    if (!existingFilter.getTuples().isEmpty()) {
      constrainedPaths = ImmutableSet.copyOf(existingFilter.getPaths());
    } else {
      if (!filterPath.isRelation()) {
        /* For primitive filters, always constrain the single primitive path even if it currently
        might contain a constant formula whose key would be the empty set. */
        constrainedPaths = ImmutableSet.of(filterPath.getPrimitive());
      } else {
        final Set<PrimitivePath> toConstrainedPaths = Sets.newLinkedHashSet();
        for (Path path : fieldSelection.getSelectedPaths()) {
          if (path.isRelation()) {
            toConstrainedPaths.addAll(
                Queries.getImmediateProjectedChildren(view, path.getRelation()));
          } else {
            toConstrainedPaths.add(path.getPrimitive());
          }
        }
        // Reduce the set of constrained paths to a key subset only.
        constrainedPaths = ImmutableSet.copyOf(
            functionalDependencies.getKey(view, ImmutableSet.copyOf(toConstrainedPaths)));
      }
    }
    return new FilterParameters(fieldSelection.getViewContext(), view, filterPath, constrainedPaths,
        fieldSelection.getSelectedPaths());
  }

  public ViewContext getViewContext() {
    return viewContext;
  }

  public Path getFilterPath() {
    return filterOptionsQuery.getFilterPath();
  }

  public Set<PrimitivePath> getConstrainedPaths() {
    return filterOptionsQuery.getConstrainedPaths();
  }

  public Set<PrimitivePath> getDisplayedPaths() {
    return filterOptionsQuery.getDisplayedPaths();
  }

  public List<PrimitivePath> getDisplayedPathsInOrder() {
    return filterOptionsQuery.getDisplayedPathsInOrder();
  }

  public @Nullable PrimitiveValue getValue(Data data, Tuple tuple, PrimitivePath displayedPath) {
    return filterOptionsQuery.getValue(data, tuple, displayedPath);
  }

  public PrimitiveValueFormatter getPrimitiveValueFormatter(Path displayedPath) {
    return PrimitiveValueFormatter.fromViewPropertySource(
        styleView.getViewPropertySource(displayedPath));
  }

  public <T> T getStyleProperty(ViewProperty<T> prop, Path displayedPath) {
    Preconditions.checkArgument(FILTER_STYLE_PROPERTIES.contains(prop));
    return styleView.get(prop, displayedPath);
  }

  public HorizontalAlignment getEffectiveHorizontalAlignment(PrimitivePath path) {
    return getStyleProperty(LMProp.DECIMAL_PLACES, path).isPresent()
        ? HorizontalAlignment.RIGHT
        : getStyleProperty(LBProp.HORIZONTAL_ALIGNMENT, path);
  }

  public ImmutableView getFilterOptionsView(String query) {
    return filterOptionsQuery.getFilterOptionsView(query, false);
  }

  // TODO[cc]: Get rid of this.
  public void addWarning(Multimap<Path,Warning> warnings, PrimitivePath path, Warning warning) {
    warnings.put(filterOptionsQuery.getContainerPaths().getNewPath(path), warning);
  }

  public SwingString generateWarningMessage(ViewResult result) {
    final List<Entry<SwingString,Warning>> labelsAndWarnings = Lists.newArrayList();
    final Strepco strepco = Strepco.create(styleView, filterOptionsQuery.getFilterPath());
    for (Entry<Path,Warning> entry : result.getWarnings().entries()) {
      final Path dataPath = entry.getKey();
      final Warning warning = entry.getValue();
      /* Don't include "error in referenced field"-type warnings, since warnings from the fields in
      question will be included anyway. */
      if (warning.getType() == Warning.Type.ERROR_IN_REFERENCED_FIELD)
        continue;
      final Path sourceViewPath =
          filterOptionsQuery.getContainerPaths().getOldPathIfMapped(dataPath);
      final SwingString label = (sourceViewPath == null)
          ? SwingString.styled("<i>(" +
              SwingString.toSwingHTML(Bundle.LBL_FilterWarningGeneratedQueryLabel()) + ")</i>")
          : ViewResultTopComponent.getPathDisplayLabel(strepco, sourceViewPath);
      labelsAndWarnings.add(new SimpleEntry<>(label, warning));
    }
    return ViewResultTopComponent.generateWarningMessage(labelsAndWarnings, true);
  }
}
