// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.filter;

import com.sieuferd.uiutil.table.StyledTableString;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.query.Filter;
import com.sieuferd.datamodel.expressions.TupleConstant;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.RelationConstant;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.MapUtil;
import java.awt.Color;
import java.awt.font.TextAttribute;
import java.text.AttributedString;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import javax.annotation.Nullable;
import javax.swing.SwingUtilities;
import javax.swing.table.AbstractTableModel;
import org.openide.util.NbBundle.Messages;

/* TODO[cc]: Create a table model that delegates to a set of other "unioned" table models, to avoid
             awkward row index offsets everywhere. */
@Messages({
  "CTL_ModifyFilter=Modify Filter",
  /* TODO[cc]: Rename to indicate that this is a Swing HTML value (and establish a convention for
               such bundle keys. */
  "LBL_FilterTupleTable_IncludeAll=(Include All)"
})
final class FilterTupleTableModel extends AbstractTableModel {
  private final FilterParameters parameters;
  private TupleInterval tupleInterval;
  private ImmutableData data;
  private ImmutableSet<String> highlightStrings = ImmutableSet.of();
  private boolean pendingViewUpdate = false;
  /* Keep our own copy of the filter state here rather than using the Filter object in the View as
  the model, since the FilterTupleTableModel is only compatible with one set of FilterParameters at
  a time. */
  private final Set<TupleConstant> activeFilterTuples;
  private final boolean usePaddingColumn;
  private boolean excluding = false;
  /**
   * A usability tweak for filters initially in the common "Include All" configuration. The quick
   * check mode makes the "Include All" item checked while keeping the filter tuples unchecked,
   * allowing the user to go directly to inclusion mode with one item selected in a single click.
   */
  private boolean quickCheckMode = true;

  FilterTupleTableModel(FilterParameters filterParameters, boolean usePaddingColumn) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkNotNull(filterParameters);
    this.parameters = filterParameters;
    this.usePaddingColumn = usePaddingColumn;
    this.data = QueryUtil.EMPTY_INCOMPLETE_DATA;
    this.tupleInterval = data.getRelationTuples(Path.ROOT, Tuple.createRootParent(data));
    /* TODO[cc]: Get the initial filter from the client. The client will need to ensure that the
                 FilterParameters are appropriate for the existing Filter value. */
    Filter initialFilter = parameters.getViewContext().getView()
        .get(QProp.FILTER, parameters.getFilterPath());
    excluding = initialFilter.isExcluding();
    activeFilterTuples = Sets.newLinkedHashSet(initialFilter.getTuples());
    quickCheckMode = excluding && activeFilterTuples.isEmpty();
  }

  private void setFilter() {
    pendingViewUpdate = true;
    SwingUtilities.invokeLater(() -> {
      if (!pendingViewUpdate)
        return;
      pendingViewUpdate = false;
      final Filter filter = Filter.create(
          new RelationConstant(parameters.getConstrainedPaths(), activeFilterTuples), excluding);
      parameters.getViewContext().getView().mutate(Bundle.CTL_ModifyFilter(), true,
          (MutableView view) ->
      {
        if (view.exists(parameters.getFilterPath())) {
          view.set(QProp.FILTER, parameters.getFilterPath(), filter);
          Queries.propagateExistentialFilter(
              view, parameters.getFilterPath(), filter.isConstrained());
        }
      });
    });
  }

  public void invert() {
    excluding = !excluding;
    quickCheckMode = false;
    fireTableRowsUpdated(0, getRowCount() - 1);
    setFilter();
  }

  public void setData(ImmutableData data) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkNotNull(data);

    final TupleInterval oldTupleInterval = this.tupleInterval;
    final Data oldData = this.data;
    final int oldRowCount = getRowCount();

    this.data = data;
    this.tupleInterval = data.getRelationTuples(Path.ROOT, Tuple.createRootParent(data));

    /* To avoid resetting the selection during infinite scrolling, avoid calling
    fireTableDataChanged for the common case where all old tuples--or rather, their constrained path
    keys--remain intact. */
    boolean oldTuplesPreserved;
    if (tupleInterval.size() < oldTupleInterval.size()) {
      oldTuplesPreserved = false;
    } else {
      oldTuplesPreserved = true;
      /* Technically things that scale with the size of result sets should be done off the EDT, but
      I think this should be quick enough to never cause a perceptible delay. */
      for (int i = 0; i < oldTupleInterval.size(); i++) {
        // No need to treat null values specially here; just compare them normally.
        if (!Objects.equals(createFilterTuple(oldData, oldTupleInterval, i), createFilterTuple(i)))
        {
          oldTuplesPreserved = false;
          break;
        }
      }
    }
    if (oldTuplesPreserved) {
      /* If display paths are different from constrained paths, then display values may still need
      updating. */
      fireTableRowsUpdated(1, oldRowCount - 1);
      fireTableRowsInserted(oldRowCount, getRowCount() - 1);
    } else {
      fireTableDataChanged();
    }
  }

  public void setHighlightStrings(Set<String> highlightStrings) {
    ThreadUtil.checkOnEDT();
    this.highlightStrings = ImmutableSet.copyOf(highlightStrings);
    fireTableRowsUpdated(1, getRowCount() - 1);
  }

  public int getNonDataRowCount() {
    // Just the "(Select All)" row.
    return 1;
  }

  @Override
  public int getRowCount() {
    return tupleInterval.size() + getNonDataRowCount();
  }

  @Override
  public int getColumnCount() {
    int pathCount = parameters.getDisplayedPathsInOrder().size();
    // Checkbox column, path columns, plus a padding column, if used.
    return 1 + pathCount + (usePaddingColumn ? 1 : 0);
  }

  @Override
  public Class<?> getColumnClass(int columnIndex) {
    return (columnIndex == 0) ? Boolean.class : String.class;
  }

  @Override
  public boolean isCellEditable(int rowIndex, int columnIndex) {
    if (columnIndex != 0) {
      return false;
    } else {
      if (rowIndex == 0) {
        return true;
      } else {
        return createFilterTuple(rowIndex - 1) != null;
      }
    }
  }

  @Override
  public String getColumnName(int columnIndex) {
    if (columnIndex == 0 || columnIndex - 1 >= parameters.getDisplayedPathsInOrder().size()) {
      return "";
    } else {
      return parameters.getStyleProperty(
          CProp.LABEL, parameters.getDisplayedPathsInOrder().get(columnIndex - 1));
    }
  }

  private @Nullable TupleConstant createFilterTuple(int dataTupleIndex) {
    return createFilterTuple(data, tupleInterval, dataTupleIndex);
  }

  private @Nullable TupleConstant createFilterTuple(
      Data useData, TupleInterval useTupleInterval, int dataTupleIndex)
  {
    Map<PrimitivePath,ConstantExpression> ret = Maps.newLinkedHashMap();
    for (PrimitivePath path : parameters.getConstrainedPaths()) {
      PrimitiveValue value =
          parameters.getValue(useData, useTupleInterval.get(dataTupleIndex), path);
      if (value == null)
        throw new AssertionError("Expected all constrained paths to be projected");
      ConstantExpression constantExpression = ConstantExpression.fromObject(value.getContent());
      if (constantExpression == null)
        return null;
      MapUtil.putChecked(ret, path, constantExpression);
    }
    return new TupleConstant(ret);
  }

  @Override
  public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
    if (columnIndex != 0)
      throw new AssertionError();
    final boolean wasQuickCheckMode = quickCheckMode;
    if (wasQuickCheckMode) {
      quickCheckMode = false;
      if (rowIndex != 0)
        excluding = false;
      // Shouldn't be necessary, but will help me catch logic bugs.
      fireTableRowsUpdated(0, getRowCount() - 1);
    }
    if (rowIndex == 0) {
      /* I considered having a separate special row representing the "excluding" state (e.g. "future
      values not in this list"), but it would be distracting to users who just expect a regular
      filter UI, and it would encourage a dangerous kind of filter inversion that makes the WHERE
      clause include every single item in the list while keeping the current query result exactly
      the same. */
      /* If the "Include All" item is clicked in quick check mode, keep it checked while turning off
      the quick check mode, to allow the exclusion mode to be reached without first hiding all
      items. */
      excluding = wasQuickCheckMode || Boolean.TRUE.equals(aValue);
      activeFilterTuples.clear();
      fireTableRowsUpdated(0, getRowCount() - 1);
      setFilter();
    } else {
      TupleConstant filterTuple = createFilterTuple(rowIndex - 1);
      if (filterTuple != null) {
        if (excluding != Boolean.TRUE.equals(aValue)) {
          activeFilterTuples.add(filterTuple);
        } else {
          activeFilterTuples.remove(filterTuple);
        }
        fireTableRowsUpdated(0, 0);
        fireTableRowsUpdated(rowIndex, rowIndex);
        setFilter();
      }
    }
  }

  @Override
  public @Nullable Object getValueAt(int rowIndex, int columnIndex) {
    if (rowIndex == 0) {
      if (columnIndex == 0) {
        return quickCheckMode ? null : excluding && activeFilterTuples.isEmpty();
      } else if (columnIndex == 1) {
        return Bundle.LBL_FilterTupleTable_IncludeAll();
      } else {
        return null;
      }
    } else {
      if (columnIndex == 0) {
        if (quickCheckMode)
          return false;
        return excluding != activeFilterTuples.contains(createFilterTuple(rowIndex - 1));
      } else {
        final List<PrimitivePath> displayedPaths = parameters.getDisplayedPathsInOrder();
        if (columnIndex - 1 >= displayedPaths.size())
          return null;
        final PrimitivePath path = displayedPaths.get(columnIndex - 1);
        final PrimitiveValue ret =
            parameters.getValue(data, tupleInterval.get(rowIndex - 1), path);
        return (ret == null) ? metaSwingString("...").getComplete() : getValueObject(ret, path);
      }
    }
  }

  private @Nullable Object getValueObject(PrimitiveValue value, PrimitivePath path) {
    Preconditions.checkNotNull(value);
    /* Only use null/empty indications when a single path is displayed. In the case of multiple
    displayed paths, we'd rather use the empty cells to display overflowing values from previous
    cells (using SpanningTableCellRenderer). */
    if (parameters.getDisplayedPaths().size() == 1) {
      if (value.isNull())
        return metaSwingString("null").getComplete();
      if ("".equals(value.getContent()))
        return metaSwingString("empty").getComplete();
    }
    final String formattedValue = SwingString.normalizePresentationWhitespace(
        parameters.getPrimitiveValueFormatter(path).getFormattedValue(value));
    if (formattedValue.trim().isEmpty()) {
      // Allow SpanningTableCellRenderer to overflow the left neighbor's value into this one.
      return null;
    }
    AttributedString as = null;
    for (String highlightString : highlightStrings) {
      if (highlightString.isEmpty())
        continue;
      int pos = formattedValue.toLowerCase().indexOf(
          SwingString.normalizePresentationWhitespace(highlightString).toLowerCase());
      if (pos >= 0) {
        if (as == null)
          as = new AttributedString(formattedValue);
        as.addAttribute(
            TextAttribute.BACKGROUND, Color.YELLOW, pos, pos + highlightString.length());
      }
    }
    final SwingString plain = SwingString.plain(formattedValue);
    return new StyledTableString(plain,
        as == null ? plain : SwingString.attributed(as),
        parameters.getEffectiveHorizontalAlignment(path).getSwingConstant());
  }

  private static SwingString metaSwingString(String str) {
    return SwingString.styled(
        "<font color=\"" + SwingString.htmlColor("controlShadow") + "\">" +
        SwingString.toSwingHTML(str) +
        "</font>");
  }
}
