// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.project.formulabar.types.NoneType;
import com.google.common.collect.Sets;
import com.sieuferd.cell.styles.PathHighlights;
import com.sieuferd.cell.support.Location;
import com.sieuferd.cell.support.SelectionModel;
import com.sieuferd.datamodel.Path;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.Flows;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.project.formulabar.ReferenceInserter.InsertionParameters;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.PathReferences;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.ObjectReference;
import com.sieuferd.util.UnitObject;
import com.sieuferd.widget.CellWidget;
import com.sieuferd.widget.CursorEditInputMap;
import com.sieuferd.widget.InteractionProvider;
import com.sieuferd.widget.InteractionProvider.Acceptance;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.Caret;
import javax.swing.text.Document;
import javax.swing.text.EditorKit;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyledDocument;
import javax.swing.undo.UndoManager;
import org.netbeans.api.lexer.TokenHierarchy;
import org.netbeans.api.lexer.TokenSequence;
import org.openide.text.CloneableEditorSupport;
import org.openide.text.NbDocument;
import org.openide.util.Lookup;

/* TODO[cc]: Consider moving this and other classes into a separate cursoreditors package. Do this
             for the types subpackage as well. */
public final class CursorEditManager {
  private static final Logger LOG = Logger.getLogger(CursorEditManager.class.getName());
  private static final boolean USE_CUSTOM_PLAIN_EDITOR_KIT = false;
  // From o.n.editor.BaseDocument.
  private static final String UNDO_MANAGER_PROP = "undo-manager";
  // TODO[cc]: Set up a proper data data flow instead of relying on SettableFlow everywhere.
  private final SettableFlow<CursorEditTarget> targetFlow = Flows.newSettable();
  private final SettableFlow<ExpressionHighlighter> referenceHighlighterFlow = Flows.newSettable();
  private final SettableFlow<UnitObject> referenceSelectionActiveFlow = Flows.newSettable();
  private final Flow<SelectionModel> referenceSelectionModelFlow =
      Flows.newProcessed(referenceSelectionActiveFlow, active -> new SelectionModel());
  /* Make this flow update on every change in CursorLocation rather than every change in Path, so
  that a new path reference can be inserted more easily if the cursor is already at the desired
  path (by selecting another cell the same path). */
  private final Flow<Location> referenceSelectionPathFlow = Flows.newIndirect(
      referenceSelectionModelFlow, selectionModel ->
      UIFlows.fromPropertyBean(SelectionModel.PROP_SELECTION,
        listener -> { selectionModel.addPropertyChangeListener(listener); },
        listener -> { selectionModel.removePropertyChangeListener(listener); },
        () -> selectionModel.getSelection().getCursorLocation()));
  private final Flow<PathHighlights> highlightsFlow =
      Flows.newIndirect(referenceHighlighterFlow,
      expressionHighlighter -> expressionHighlighter.getHighlightsFlow());
  private final FocusListener focusListener = new FocusListener() {
    @Override
    public void focusGained(FocusEvent e) {
      final JEditorPane pane = (JEditorPane) e.getSource();
      lastFocusedEditorPanes.remove(pane);
      lastFocusedEditorPanes.add(0, pane);
      updateReferenceSelectionActive();
    }

    @Override
    public void focusLost(FocusEvent e) {
      if (!inAcceptEdit) {
        final JEditorPane pane = (JEditorPane) e.getSource();
        final Caret caret = pane.getCaret();
        // Remove any text selection without changing the viewport or caret position.
        caret.setDot(caret.getDot());
      }
      updateReferenceSelectionActive();
    }
  };
  private final MouseListener editorMouseListener = new MouseAdapter() {
    @Override
    public void mousePressed(MouseEvent e) {
      if (!(e.getSource() instanceof JEditorPane))
        throw new AssertionError();
      /* Switch the editor to extended input mode if the user clicks it (e.g. to change the caret
      position). */
      CursorEditInputMap.configureCursorEditor(cellWidget, (JComponent) e.getSource(), true);
    }
  };
  private final CellWidget cellWidget;
  private final ImmutableMap<CursorEditType,DocumentKit> documentKits;
  private final Set<CursorEditorProvider> editorProviders = Sets.newLinkedHashSet();
  // TODO[cc]: Consider storing editKey and editStrepco together in a single object.
  /** Don't expose this to clients. */
  private @Nonnull CursorEditKey editKey = NoneType.EMPTY_EDIT_KEY;
  /** The Strepco used when generating {@code originalValueStringAfterNormalizations}, if any. */
  private @Nullable Strepco editStrepco = null;
  private String originalValueStringAfterNormalizations = "";
  private boolean inAcceptEdit;
  /**
   * List of all editor panes, with the most recently focused ones listed first. Don't use
   * NetBeans' {@code EditorRegistry}, as {@code EditorRegistry.register} is not public, and we
   * don't want to assume NetBeans-only editor kits.
   */
  private final List<JEditorPane> lastFocusedEditorPanes = Lists.newArrayList();
  private final Map<CursorEditorProvider,JEditorPane> panesByProvider = Maps.newLinkedHashMap();

  public CursorEditManager(CellWidget cellWidget) {
    Preconditions.checkNotNull(cellWidget);
    this.cellWidget = cellWidget;
    {
      final Map<CursorEditType,DocumentKit> toDocumentKits = Maps.newLinkedHashMap();
      final List<CursorEditType> types =
          Lists.newArrayList(Lookup.getDefault().lookupAll(CursorEditType.class));
      types.add(NoneType.EMPTY_EDIT_KEY.getType());
      for (CursorEditType type : types) {
        final String contentType = type.getMimeType();
        final EditorKit kit;
        if (USE_CUSTOM_PLAIN_EDITOR_KIT && type.getUsePlainEditor()) {
          kit = new SimpleEditorKit();
        } else {
          kit = CloneableEditorSupport.getEditorKit(contentType);
        }
        if (kit.getClass().getName().startsWith(
            "org.openide.text.CloneableEditorSupport$PlainEditorKit"))
        {
          /* PlainEditorKit has a line wrapping bug related to
          https://netbeans.org/bugzilla/show_bug.cgi?id=241652 , and so should be avoided. */
          LOG.warning("Got a PlainEditorKit; probably missing a runtime dependency on the " +
              "org-netbeans-modules-editor-plain module");
        }
        final Document document;
        document = kit.createDefaultDocument();
        if (type.getMimeType().equals(ExpressionLanguageConfig.MIME_TYPE))
          document.putProperty(ExpressionHighlighter.class, new ExpressionHighlighter(document));
        if (USE_CUSTOM_PLAIN_EDITOR_KIT && type.getUsePlainEditor() &&
            document.getProperty(UNDO_MANAGER_PROP) == null)
        {
          document.putProperty(UNDO_MANAGER_PROP, new UndoManager());
        }
        MapUtil.putChecked(toDocumentKits, type, new DocumentKit(kit, document));
      }
      this.documentKits = ImmutableMap.copyOf(toDocumentKits);
    }
    loadTargetValue();
    UIFlows.bindEDT(referenceSelectionPathFlow, (@Nullable Location cursorLocation) -> {
      CursorEditTarget target = targetFlow.getWithoutListener();
      /* Avoid calling insertPathReference for the initial setting of the reference selection
      cursor, as this would force the editor into non-extended input mode. */
      if (cursorLocation != null && target != null &&
          !cursorLocation.equalsLocation(target.getCursorLocation()))
      {
        insertPathReference(target, cursorLocation.getPath());
      }
    });
    // TODO[cc]: Get rid of updateHighlighterTarget and its callees.
    UIFlows.bindEDT(referenceSelectionModelFlow, (@Nullable SelectionModel value) -> {
      updateHighlighterTarget();
    });
    UIFlows.bindEDT(referenceHighlighterFlow, (@Nullable ExpressionHighlighter value) -> {
      updateHighlighterTarget();
    });
    UIFlows.bindEDT(targetFlow, (@Nullable CursorEditTarget target) -> {
      final ObjectReference<String> originalValueStringRef = ObjectReference.create();
      final CursorEditKey newEditKey = createEditKey(target, originalValueStringRef);
      /* Always reload the value if not actively editing, in case the Strepco has changed (e.g. due
      to changed labels that need to be reflected in formula references.) */
      if (!isEditingActive() || !editKey.equals(newEditKey)) {
        cellWidget.getUI().closeInplaceEditor();
        setEditKey(newEditKey, originalValueStringRef.getValue(),
            (target == null) ? null : target.getStrepco());
      }
      updateHighlighterTarget();
    });
  }

  /** @see InteractionProvider#cancelEdit() */
  public void cancelEdit() {
    cellWidget.getUI().closeInplaceEditor();
    loadTargetValue();
    updateReferenceSelectionActive();
  }

  private Document getDocument() {
    return MapUtil.getChecked(documentKits, editKey.getType()).document;
  }

  /** @see InteractionProvider#acceptEdit() */
  public Acceptance acceptEdit() {
    final Document document = getDocument();
    Preconditions.checkState(document != null);
    final String oldValueString = originalValueStringAfterNormalizations;
    final String newValueString = UIUtil.getDocumentText(document);
    // This condition avoids calling setValue for read-only edit keys.
    if (oldValueString.equals(newValueString)) {
      // Still need to recompute the target value in case the Strepco changed.
      loadTargetValue();
      return Acceptance.ACCEPTED_IMMEDIATELY;
    }
    if (editKey.isReadOnly())
      throw new IllegalStateException("Cannot modify value for read-only cursor edit key");
    inAcceptEdit = true;
    updateReferenceSelectionActive();
    try {
      final Acceptance ret;
      final CursorEditTarget target = targetFlow.getWithoutListener();
      if (target == null) {
        ret = Acceptance.REJECTED;
      } else {
        if (editStrepco == null)
          throw new AssertionError();
        ret = editKey.setValue(target.getViewContext(), editStrepco, newValueString);
      }
      if (ret == Acceptance.REJECTED) {
        if (!lastFocusedEditorPanes.isEmpty()) {
          JEditorPane pane = lastFocusedEditorPanes.iterator().next();
          pane.requestFocusInWindow();
        }
      } else {
        /* Restore the contents of the editor to the generated form. This will happen by itself if
        the model is updated and propagates its changes, but not if the model decides that no actual
        change has happened. For instance, modifying the whitespace arrangement in an existing
        formula does not result in a change in the model, but it is desirable to restore the formula
        to its generated form anyway when the edit is accepted. A more important case occurs when an
        incorrect formula was accepted with proposed changes, but the proposed changes turn out to
        make the formula the same as its previous value. In this case, the loadTargetValue() call
        is essential to avoid the formula bar getting stuck with the incorrect string
        representation. */
        loadTargetValue();
      }
      return ret;
    } finally {
      inAcceptEdit = false;
      updateReferenceSelectionActive();
    }
  }

  private boolean isEditingActive() {
    if (targetFlow.getWithoutListener() == null)
      return false;
    if (inAcceptEdit)
      return true;
    final String oldValueString = originalValueStringAfterNormalizations;
    final String newValueString = UIUtil.getDocumentText(getDocument());
    /* This condition avoids being too dependent on focus state, which may change momentarily after
    closing acceptance dialogs. */
    if (!oldValueString.equals(newValueString))
      return true;
    for (JEditorPane pane : panesByProvider.values()) {
      if (pane.isFocusOwner())
        return true;
    }
    return false;
  }

  private void updateReferenceSelectionActive() {
    ThreadUtil.checkOnEDT();
    final boolean newValue = isEditingActive() && isExpressionType();
    referenceSelectionActiveFlow.set(newValue ? UnitObject.INSTANCE : null);
  }

  // TODO[cc]: Get rid of this.
  private void updateHighlighterTarget() {
    ThreadUtil.checkOnEDT();
    ExpressionHighlighter highlighter = referenceHighlighterFlow.getWithoutListener();
    if (highlighter != null) {
      CursorEditTarget target = Flows.getMomentary(referenceSelectionModelFlow) == null ? null :
          targetFlow.getWithoutListener();
      highlighter.setTarget(target == null ? null : target.getCursorLocation(), editStrepco);
    }
  }

  /** @return the reference highlights requested by the currently focused cursor editor */
  public Flow<PathHighlights> getReferenceHighlightsFlow() {
    return highlightsFlow;
  }

  public Flow<SelectionModel> getReferenceSelectionModelFlow() {
    return referenceSelectionModelFlow;
  }

  public void setTarget(final @Nullable CursorEditTarget target) {
    targetFlow.set(target);
  }

  private boolean isExpressionType() {
    return ExpressionLanguageConfig.MIME_TYPE.equals(editKey.getType().getMimeType());
  }

  private @Nonnull CursorEditKey createEditKey(
      @Nullable CursorEditTarget target, ObjectReference<String> originalValueString)
  {
    ThreadUtil.checkOnEDT();
    if (target != null) {
      for (CursorEditType type : documentKits.keySet()) {
        final CursorEditKey ret = type.createEditKey(target, originalValueString);
        if (ret != null)
          return ret;
      }
    }
    originalValueString.setValue("");
    return NoneType.EMPTY_EDIT_KEY;
  }

  private void setEditKey(
      CursorEditKey editKey, String originalValueString, @Nullable Strepco strepco)
  {
    Preconditions.checkNotNull(editKey);
    this.editKey = editKey;
    this.editStrepco = strepco;
    // Update the highlighter's Strepco.
    updateHighlighterTarget();
    initializeEditorProviders(originalValueString);
  }

  private void loadTargetValue() {
    ThreadUtil.checkOnEDT();
    final CursorEditTarget target = targetFlow.getWithoutListener();
    final ObjectReference<String> originalValueStringRef = ObjectReference.create();
    final CursorEditKey newEditKey = createEditKey(target, originalValueStringRef);
    setEditKey(newEditKey, originalValueStringRef.getValue(),
        target == null ? null : target.getStrepco());
  }

  private void initializeEditorProviders(String originalValueString) {
    final Document document = getDocument();
    UIUtil.setDocumentText(document, originalValueString);
    /* Don't try to detect subsequent user changes to the document by comparing directly with
    originalValueString, since the document model itself may perform normalizations such as
    converting '\r\n' to '\n'. Instead, compare strings generated by the document model. */
    originalValueStringAfterNormalizations = UIUtil.getDocumentText(document);

    // Avoid carrying undo/redo state along cached editors.
    {
      Object undoManager = document.getProperty(UNDO_MANAGER_PROP);
      if (undoManager instanceof UndoManager)
        ((UndoManager) undoManager).discardAllEdits();
    }

    // TODO[cc]: Consider delegating this to the CursorEditType.
    referenceHighlighterFlow.set(
        (ExpressionHighlighter) document.getProperty(ExpressionHighlighter.class));

    for (CursorEditorProvider provider : editorProviders)
      updateEditorProvider(provider);
  }

  // TODO[cc]: Clean this up.
  private void updateEditorProvider(CursorEditorProvider provider) {
    final JEditorPane oldPane = panesByProvider.remove(provider);
    final JEditorPane pane = provider.update(editKey.getType());
    if (pane != oldPane) {
      if (oldPane != null)
        unregisterPane(oldPane);
      if (pane != null)
        registerPane(pane);
    }
    if (pane == null)
      return;
    MapUtil.putChecked(panesByProvider, provider, pane);
    updatePane(pane);
  }

  private void updatePane(JEditorPane pane) {
    /* TODO[pd]: Make sure the JEditorPane maintains a sane preferred width, as the default
                 implementation only ever increases it (might already be fixed at this point). */
    final Document document = getDocument();
    /* We avoid changing the Document/EditorKit on every selection change, since these operations
    may be expensive, or be associated with resource leaks. */
    if (document != pane.getDocument())
      throw new AssertionError("CursorEditorProvider.update contract violated");
    pane.setEditable(!editKey.isReadOnly());
    /* Ensure no selection and a proper viewport position. Individual editors need to implement
    "select all", "move caret to end", or "clear" logic once editing actually starts. */
    pane.getCaret().setDot(0);
  }

  public void registerEditorProvider(CursorEditorProvider provider) {
    ThreadUtil.checkOnEDT();
    MapUtil.addChecked(editorProviders, provider);
    // Eagerly configure each type of editor.
    for (Entry<CursorEditType,DocumentKit> entry : documentKits.entrySet()) {
      final CursorEditType type = entry.getKey();
      final JEditorPane pane = provider.update(type);
      if (pane != null) {
        // Required for NetBeans-based editor kits; see the Javadoc for LocalEditor.getEditorPane.
        pane.putClientProperty("usedByCloneableEditor", Boolean.TRUE);
        pane.setEditorKit((EditorKit) entry.getValue().kit.clone());
        pane.setDocument(entry.getValue().document);
      }
    }
    updateEditorProvider(provider);
    updateReferenceSelectionActive();
  }

  public void unregisterEditorProvider(CursorEditorProvider provider) {
    ThreadUtil.checkOnEDT();
    // Perform cleanup similar to that done in org.openide.text.CloneableEditor.componentClosed.
    for (CursorEditType type : documentKits.keySet()) {
      final JEditorPane pane = provider.update(type);
      if (pane != null) {
        pane.setEditorKit(null);
        pane.putClientProperty("usedByCloneableEditor", Boolean.FALSE);
      }
    }
    MapUtil.removeChecked(editorProviders, provider);
    // TODO[cc]: Avoid this asymmetry.
    JEditorPane pane = panesByProvider.remove(provider);
    if (pane != null)
      unregisterPane(pane);
    updateReferenceSelectionActive();
  }

  private void registerPane(JEditorPane pane) {
    Preconditions.checkNotNull(pane);
    if (lastFocusedEditorPanes.contains(pane))
      throw new AssertionError();
    lastFocusedEditorPanes.add(pane);
    pane.addFocusListener(focusListener);
    pane.addMouseListener(editorMouseListener);
  }

  private void unregisterPane(JEditorPane pane) {
    Preconditions.checkNotNull(pane);
    pane.removeMouseListener(editorMouseListener);
    pane.removeFocusListener(focusListener);
    if (!lastFocusedEditorPanes.contains(pane))
      throw new AssertionError();
    lastFocusedEditorPanes.remove(pane);
  }

  private void insertPathReference(final CursorEditTarget target, final Path path) {
    Preconditions.checkNotNull(target);
    Preconditions.checkNotNull(path);
    final Document document = getDocument();
    if (!(document instanceof StyledDocument))
      return;
    if (lastFocusedEditorPanes.isEmpty())
      return;
    final JEditorPane editorPane = lastFocusedEditorPanes.get(0);
    /* Turn extended input mode off so that arrow keys can be used to move the reference selection
    cursor. */
    CursorEditInputMap.configureCursorEditor(cellWidget, editorPane, false);
    NbDocument.runAtomic((StyledDocument) document, () -> {
      if (lastFocusedEditorPanes.isEmpty())
        return;
      if (editorPane.getDocument() != document)
        return;
      final Strepco strepco = editStrepco;
      final Path formulaPath = target.getCursorLocation().getPath();
      if (formulaPath.isRelation())
        throw new AssertionError();
      if (formulaPath.equals(path) || !strepco.getView().exists(path))
        return;
      final TokenHierarchy<Document> tokenHierarchy = TokenHierarchy.get(document);
      final TokenSequence<ExpressionTokenId> tokenSequence =
          tokenHierarchy.tokenSequence(ExpressionLanguageConfig.LANGUAGE);
      if (tokenSequence == null || !tokenSequence.isValid()) {
        LOG.info("Got invalid token sequence");
        return;
      }

      final Caret caret = editorPane.getCaret();
      final InsertionParameters parameters = ReferenceInserter.calculateInsertionParameters(
          tokenSequence, caret.getMark(), caret.getDot());

      final String insertString =
          parameters.insertPreceeding + PathReferences.getTokenStringFromPath(strepco, path);
      try {
        final int docLength = document.getLength();
        /* The final whitespace token seems to go one character beyond the actual document
        length. */
        final int insertStart = Math.min(parameters.interval.start, docLength);
        final int insertEnd = Math.min(parameters.interval.end, docLength);
        final int replaceLength = insertEnd - insertStart;
        final String existingText = document.getText(insertStart, replaceLength);
        /* Avoid disturbing the Document object when the cursor is just moved between different
        cells on the same path, as this would cause an annoying flicker in the path reference
        token highlighting. It is useful to move the reference selection cursor between different
        cells on the same path, because correlation context highlighting will show which values
        are used to compute the result in each cell of the formula path. */
        if (!existingText.equals(insertString)) {
          if (replaceLength > 0)
            document.remove(insertStart, replaceLength);
          document.insertString(insertStart, insertString, SimpleAttributeSet.EMPTY);
          caret.setDot(insertStart + insertString.length());
        }
      } catch (BadLocationException e) {
        LOG.log(Level.WARNING, "Unexpected error while inserting path reference", e);
      }
    });
  }

  private static final class DocumentKit {
    /** This kit should be cloned before being used in editors. */
    final EditorKit kit;
    final Document document;

    public DocumentKit(EditorKit kit, Document document) {
      Preconditions.checkNotNull(kit);
      Preconditions.checkNotNull(document);
      this.kit = kit;
      this.document = document;
    }
  }
}
