// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.collect.ImmutableSet;
import com.sieuferd.upstream.localeditor.ColoringsWriter;
import java.awt.Color;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.UIManager;
import javax.swing.text.AttributeSet;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;
import org.netbeans.api.editor.mimelookup.MimePath;
import org.netbeans.api.editor.settings.FontColorNames;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileSystem;
import org.openide.filesystems.FileUtil;
import org.openide.filesystems.MultiFileSystem;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;
import org.openide.xml.XMLUtil;

/* This class sets various L&F-dependent editor settings for the formula bar. While the Editor
Settings API does not expose an official API for modifying settings (only for reading them), the
XML format used to store settings is stable enough that settings files can be generated directly
without risk of regressions. See ColoringsWriter.

See http://wiki.netbeans.org/DevFaqDynamicSystemFilesystem for documentation about injecting files
into the system filesystem at runtime. Repository.LayerProvider is not relevant here, as it assumes
that injected XML layers are available as URL-reachable resources; neither are the static methods in
the example. */
@ServiceProvider(service = FileSystem.class)
public class EditorSettingsInjector extends MultiFileSystem {
  private static final Logger LOG = Logger.getLogger(EditorSettingsInjector.class.getName());

  public EditorSettingsInjector() {
    setPropagateMasks(true);
    final FileSystem delegate = FileUtil.createMemoryFileSystem();
    try {
      addSettings(delegate, DataValueLanguageConfig.MIME_TYPE, getDefaultTokenSettings(), false);
      addSettings(delegate, ExpressionLanguageConfig.MIME_TYPE, getDefaultTokenSettings(), false);
      // TODO[cc]: Consider putting this part in the branding module instead.
      /* Since highlighting settings are global rather than specific to a particular MIME type, only
      set them if the module is being used as part of a standalone application. */
      if ("sieuferd".equals(NbBundle.getBranding()))
        addSettings(delegate, "", getHighlightSettings(), true);
    } catch (IOException e) {
      LOG.log(Level.WARNING, "Failed to inject editor settings for formula bar", e);
    }
    setDelegates(delegate);
  }

  private void addSettings(
    FileSystem fileSystem, String mimeType, Set<AttributeSet> settings, boolean highlight)
    throws IOException
  {
    final String filename = "com-sieuferd-project-formulabar-" +
        (highlight ? "highlight" : "token") + "-colorings-injected";
    /* The injected settings go directly under FontsColors/NetBeans, not under "Defaults". See the
    comment in the branding module's layer.xml. */
    final FileObject fo = FileUtil.createFolder(fileSystem.getRoot(),
        "Editors/" + MimePath.get(mimeType).getPath() + "/FontsColors/NetBeans")
        .createData(filename, "xml");
    // Not strictly necessary, but try to make any problems more deterministic.
    fo.setAttribute("position", 100);
    fo.setAttribute(ColoringsWriter.FA_TYPE,
        highlight ? ColoringsWriter.FAV_HIGHLIGHT : ColoringsWriter.FAV_TOKEN);
    try (OutputStream os = fo.getOutputStream()) {
      XMLUtil.write(ColoringsWriter.getDocument(settings), os, "UTF-8");
    }
  }

  private static Set<AttributeSet> getHighlightSettings() {
    /* See org.netbeans.modules.editor.lib2.highlighting.CaretBasedBlockHighlighting.setAttrs() for
    analogous logic in the handling of the AsTextField client property. */
    final MutableAttributeSet ret = new SimpleAttributeSet();
    ret.addAttribute(StyleConstants.NameAttribute, FontColorNames.SELECTION_COLORING);
    Color foreground = null, background = null;
    if (!GraphicsEnvironment.isHeadless()) {
      foreground = UIManager.getColor("TextField.selectionForeground");
      background = UIManager.getColor("TextField.selectionBackground");
    }
    if (foreground != null)
      ret.addAttribute(StyleConstants.Foreground, foreground);
    if (background != null)
      ret.addAttribute(StyleConstants.Background, background);
    return ImmutableSet.<AttributeSet>of(ret);
  }

  private static Set<AttributeSet> getDefaultTokenSettings() {
    /* See getHardcodedDefaultColoring() in
    org.netbeans.modules.editor.settings.storage.fontscolors.CompositeFCS for similar logic. */
    final MutableAttributeSet ret = new SimpleAttributeSet();
    ret.addAttribute(StyleConstants.NameAttribute, FontColorNames.DEFAULT_COLORING);

    String fontFamily = null;
    Integer fontSize = null;
    Integer fontStyle = null;
    Color foreground = null;
    Color background = null;

    if (!GraphicsEnvironment.isHeadless()) {
      final Font textFieldFont = UIManager.getFont("TextField.font");
      if (textFieldFont != null) {
        fontFamily = textFieldFont.getFamily();
        fontSize = textFieldFont.getSize();
        fontStyle = textFieldFont.getStyle();
      }
      foreground = UIManager.getColor("TextField.foreground");
      background = UIManager.getColor("TextField.background");
      final Integer customFontSize = (Integer) UIManager.get("customFontSize");
      if (customFontSize != null)
        fontSize = customFontSize;
    }

    if (fontFamily != null)
      ret.addAttribute(StyleConstants.FontFamily, fontFamily);
    if (fontSize != null)
      ret.addAttribute(StyleConstants.FontSize, fontSize);
    if (fontStyle != null) {
      ret.addAttribute(StyleConstants.Bold, (fontStyle & Font.BOLD) != 0);
      ret.addAttribute(StyleConstants.Italic, (fontStyle & Font.ITALIC) != 0);
    }
    if (foreground != null)
      ret.addAttribute(StyleConstants.Foreground, foreground);
    if (background != null)
      ret.addAttribute(StyleConstants.Background, background);
    return ImmutableSet.<AttributeSet>of(ret);
  }
}
