// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.Preconditions;
import com.sieuferd.uiutil.ThreadUtil;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import org.netbeans.api.editor.mimelookup.MimePath;
import org.netbeans.api.editor.mimelookup.MimeRegistration;
import org.netbeans.api.lexer.Token;
import org.netbeans.api.lexer.TokenHierarchy;
import org.netbeans.api.lexer.TokenSequence;
import org.netbeans.spi.editor.typinghooks.DeletedTextInterceptor;

public class ExpressionDeletedTextInterceptor implements DeletedTextInterceptor {
  private static final Logger LOG =
      Logger.getLogger(ExpressionDeletedTextInterceptor.class.getName());
  private @Nullable ToRemove toRemove;

  private static final class ToRemove {
    final int startPos;
    final int length;
    final String expectedString;

    public ToRemove(int startPos, int length, String expectedString) {
      Preconditions.checkArgument(startPos >= 0);
      Preconditions.checkArgument(length > 0);
      Preconditions.checkArgument(expectedString.length() == length);
      this.startPos = startPos;
      this.length = length;
      this.expectedString = expectedString;
    }
  }

  @Override
  public boolean beforeRemove(final Context context) {
    ThreadUtil.checkOnEDT();
    /* Calculate the range to remove before the first character has actually been removed, so we can
    know the true tokenization of the original text. */
    toRemove = null;
    final boolean backward = context.isBackwardDelete();
    if (!(backward ? "]" : "[").equals(context.getText()))
      return false;
    final Document doc = context.getDocument();
    doc.render(() -> {
      final TokenHierarchy<Document> tokenHierarchy = TokenHierarchy.get(doc);
      final TokenSequence<ExpressionTokenId> tokenSequence =
          tokenHierarchy.tokenSequence(ExpressionLanguageConfig.LANGUAGE);
      // Offset by one to account for initial open/close bracket.
      tokenSequence.move(context.getOffset() + (backward ? -1 : +1));

      while (backward ? tokenSequence.movePrevious() : tokenSequence.moveNext()) {
        final Token<ExpressionTokenId> token = tokenSequence.token();
        final ExpressionTokenId id = token.id();
        if (id.equals(backward
            ? ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET
            : ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET))
        {
          final int startPos, endPos, length;
          final String expectedText;
          try {
            if (backward) {
              startPos = tokenSequence.offset();
              endPos = context.getOffset() - 1;
              length = endPos - startPos;
              expectedText = doc.getText(startPos, length);
            } else {
              startPos = context.getOffset();
              endPos = tokenSequence.offset() + token.length() - 1;
              length = endPos - startPos;
              expectedText = doc.getText(startPos + 1, length);
            }
            toRemove = new ToRemove(startPos, length, expectedText);
          } catch (BadLocationException e) {
            LOG.log(Level.WARNING, "Unexpected error while removing path reference", e);
          }
          break;
        } else if (
            !id.equals(ExpressionTokenId.PATH_REFERENCE_CORE) &&
            !id.equals(ExpressionTokenId.WHITESPACE))
        {
          break;
        }
      }
    });
    return false;
  }

  @Override
  public void remove(Context context) throws BadLocationException {
    ThreadUtil.checkOnEDT();
    if (toRemove == null)
      return;
    try {
      final Document doc = context.getDocument();
      /* Only remove the interval if its text remained unchanged since the time when we calculated
      its bounds. */
      try {
        if (!doc.getText(toRemove.startPos, toRemove.length).equals(toRemove.expectedString))
          return;
      } catch (BadLocationException e) {
        // Ignore; this could conceivably happen due to the possible above-mentioned race condition.
        return;
      }
      doc.remove(toRemove.startPos, toRemove.length);
    } finally {
      toRemove = null;
    }
  }

  @Override
  public void afterRemove(Context context) throws BadLocationException {
    toRemove = null;
  }

  @Override
  public void cancelled(Context context) {
    toRemove = null;
  }

  @MimeRegistration(
      mimeType = ExpressionLanguageConfig.MIME_TYPE,
      service = DeletedTextInterceptor.Factory.class)
  public static class Factory implements DeletedTextInterceptor.Factory {
    @Override
    public DeletedTextInterceptor createDeletedTextInterceptor(MimePath mimePath) {
      return new ExpressionDeletedTextInterceptor();
    }
  }
}
