// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.cell.styles.PathHighlights;
import com.sieuferd.cell.support.Location;
import com.sieuferd.datamodel.Path;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.Flows;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.PathReferenceException;
import com.sieuferd.datamodel.expressions.PathReferences;
import com.sieuferd.query.Queries;
import com.sieuferd.util.MapUtil;
import java.awt.Color;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import javax.swing.SwingUtilities;
import javax.swing.text.Document;
import org.netbeans.api.lexer.Token;
import org.netbeans.api.lexer.TokenHierarchy;
import org.netbeans.api.lexer.TokenHierarchyEvent;
import org.netbeans.api.lexer.TokenHierarchyListener;
import org.netbeans.api.lexer.TokenSequence;
import org.netbeans.spi.editor.highlighting.support.OffsetsBag;
import org.openide.util.WeakListeners;

// TODO[pd]: Put the highlighting in its own thread, as is idiomatic.

// Idiomatic highlighter classes do not extend from any specific API class or interface.
@ThreadSafe
final class ExpressionHighlighter {
  private final Document document;
  private final SettableFlow<PathHighlights> highlightsFlow = Flows.newSettable();
  private final OffsetsBag offsetsBag;
  private final TokenHierarchy<Document> tokenHierarchy;
  private final TokenHierarchyListener tokenHierarchyListener = (TokenHierarchyEvent evt) -> {
    // Give the OffsetsBag time to do its own updating before we update the highlights.
    SwingUtilities.invokeLater(this::updateHighlights);
  };
  private @Nullable Location cursorLocation;
  private @Nullable Strepco strepco;

  public ExpressionHighlighter(Document document) {
    Preconditions.checkNotNull(document);
    this.document = document;
    this.offsetsBag = new OffsetsBag(document);
    tokenHierarchy = TokenHierarchy.get(document);
    tokenHierarchy.addTokenHierarchyListener(
        WeakListeners.create(TokenHierarchyListener.class, tokenHierarchyListener, tokenHierarchy));
    // Initial update.
    updateHighlights();
  }

  public void setTarget(@Nullable Location cursorLocation, @Nullable Strepco strepco) {
    synchronized (this) {
      if (Objects.equals(this.cursorLocation, cursorLocation) &&
          Objects.equals(this.strepco, strepco))
      {
        return;
      }
      this.cursorLocation = cursorLocation;
      this.strepco = strepco;
    }
    updateHighlights();
  }

  /** May be called from any thread. */
  private void updateHighlights() {
    final @Nullable Location useCursorLocation;
    final @Nullable Strepco useStrepco;
    synchronized (this) {
      useCursorLocation = this.cursorLocation;
      useStrepco = this.strepco;
    }
    document.render(() -> {
      if (useCursorLocation == null || useStrepco == null) {
        highlightsFlow.set(PathHighlights.NONE);
        offsetsBag.clear();
        return;
      }
      final TokenSequence<ExpressionTokenId> tokenSequence =
          tokenHierarchy.tokenSequence(ExpressionLanguageConfig.LANGUAGE);
      if (tokenSequence == null || !tokenSequence.isValid()) {
        // Not sure when this would happen, but don't reset any existing highlights in this case.
        return;
      }
      final OffsetsBag newOffsetsBag = new OffsetsBag(document);
      final Map<Path,ReferenceHighlightStyle> highlightStyles = Maps.newLinkedHashMap();
      // Always highlight the field containing the formula being edited, in a neutral gray.
      highlightStyles.put(useCursorLocation.getPath(), ReferenceHighlightStyle.NEUTRAL);
      String pathReferenceText = null;
      Integer pathReferenceStartOffset = null;
      int colorIndex = 0;
      while (tokenSequence.moveNext()) {
        final Token<ExpressionTokenId> token = tokenSequence.token();
        final ExpressionTokenId id = token.id();
        final int offset = tokenSequence.offset();
        if (id.equals(ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET)) {
          pathReferenceStartOffset = tokenSequence.offset();
          /* Need to set this to support the case of empty brackets. (Though under current rules,
          these will never refer to a valid path; see
          com.sieuferd.datamodel.expressions.PathReferences ). */
          pathReferenceText = "";
        } else if (id.equals(ExpressionTokenId.PATH_REFERENCE_CORE)) {
          pathReferenceText = token.text().toString();
        } else if (id.equals(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET) &&
            pathReferenceStartOffset != null)
        {
          Path foundPath = null;
          if (pathReferenceText != null) {
            try {
              foundPath = PathReferences.getPathFromTokenString(
                  strepco, "[" + pathReferenceText + "]");
            } catch (PathReferenceException e) {
              // No action.
            }
          }
          if (foundPath != null) {
            ReferenceHighlightStyle style = highlightStyles.get(foundPath);
            if (style == null) {
              if (Queries.isVisibleToRoot(strepco.getView(), foundPath)) {
                style = ReferenceHighlightStyle.getForIndex(colorIndex++);
              } else {
                style = ReferenceHighlightStyle.NEUTRAL;
              }
              highlightStyles.put(foundPath, style);
            }
            style.addHighlight(newOffsetsBag, pathReferenceStartOffset, offset + 1);
          }
          pathReferenceStartOffset = null;
          pathReferenceText = null;
        }
      }
      // Make all changes atomically.
      /* Update highlights while still holding the read lock on the Document, as required by
      http://bits.netbeans.org/dev/javadoc/org-netbeans-modules-editor-lib2/org/netbeans/spi/editor/highlighting/package-summary.html
      ("Locking and Document changes"). */
      offsetsBag.setHighlights(newOffsetsBag);
      final Map<Path,Color> highlightColors = Maps.newLinkedHashMap();
      for (Entry<Path,ReferenceHighlightStyle> entry : highlightStyles.entrySet())
        MapUtil.putChecked(highlightColors, entry.getKey(), entry.getValue().getColor());
      highlightsFlow.set(new PathHighlights(highlightColors, useCursorLocation));
    });
  }

  public Flow<PathHighlights> getHighlightsFlow() {
    return highlightsFlow.getReadOnly();
  }

  public OffsetsBag getOffsetsBag() {
    return offsetsBag;
  }
}
