// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.CharMatcher;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.sieuferd.datamodel.expressions.ExpressionTokenType;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.datamodel.expressions.PathReferences;
import com.sieuferd.util.MapUtil;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.misc.Interval;
import org.netbeans.api.lexer.PartType;
import org.netbeans.spi.lexer.LexerInput;
import org.netbeans.spi.lexer.LexerRestartInfo;

// TODO[te]: Add the kinds of unit tests suggested by the Javadoc for org.netbeans.spi.lexer.Lexer.
// TODO[te]: Add unit tests for error conditions such as "test]]][test]'asdfasdf'".
@NotThreadSafe
public class ExpressionLexer implements org.netbeans.spi.lexer.Lexer<ExpressionTokenId> {
  private static final int STATE_IN_PATH_REFERENCE = 2;
  private final LexerRestartInfo<ExpressionTokenId> info;
  private final org.antlr.v4.runtime.Lexer antlrLexer;
  private final CharMatcher pathReferenceTerminatorMatcher =
      CharMatcher.anyOf(PathReferences.ILLEGAL_CHARACTERS_CORE);
  private boolean inPathReference;

  // Don't include path reference tokens; they need to be handled specially by the lexer.
  private static final ImmutableMap<ExpressionTokenType,ExpressionTokenId> TOKEN_ID_BY_TYPE =
      ImmutableMap.<ExpressionTokenType,ExpressionTokenId>builder()
      .put(ExpressionTokenType.IDENTIFIER, ExpressionTokenId.IDENTIFIER)
      .put(ExpressionTokenType.ARGUMENT_SEPARATOR, ExpressionTokenId.ARGUMENT_SEPARATOR)
      .put(ExpressionTokenType.PARENTHESIS_LEFT, ExpressionTokenId.PARENTHESIS_LEFT)
      .put(ExpressionTokenType.PARENTHESIS_RIGHT, ExpressionTokenId.PARENTHESIS_RIGHT)
      .put(ExpressionTokenType.LITERAL_NUMBER, ExpressionTokenId.LITERAL_NUMBER)
      .put(ExpressionTokenType.LITERAL_STRING, ExpressionTokenId.LITERAL_STRING)
      .put(ExpressionTokenType.LITERAL_TEMPORAL, ExpressionTokenId.LITERAL_TEMPORAL)
      .put(ExpressionTokenType.LITERAL_BOOLEAN, ExpressionTokenId.LITERAL_BOOLEAN)
      .put(ExpressionTokenType.LITERAL_ERROR, ExpressionTokenId.LITERAL_ERROR)
      .put(ExpressionTokenType.LITERAL_NULL, ExpressionTokenId.LITERAL_NULL)
      .put(ExpressionTokenType.OPERATOR, ExpressionTokenId.OPERATOR)
      .put(ExpressionTokenType.WHITESPACE, ExpressionTokenId.WHITESPACE)
      .put(ExpressionTokenType.UNTERMINATED_LITERAL_STRING, ExpressionTokenId.LITERAL_STRING)
      .put(ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL, ExpressionTokenId.LITERAL_TEMPORAL)
      .put(ExpressionTokenType.OTHER, ExpressionTokenId.OTHER)
      .build();

  /** Package-visible for testing. */
  static ExpressionTokenId tokenIdFromType(ExpressionTokenType tokenType) {
    return MapUtil.getChecked(TOKEN_ID_BY_TYPE, tokenType);
  }

  public ExpressionLexer(LexerRestartInfo<ExpressionTokenId> info) {
    Preconditions.checkNotNull(info);
    this.info = info;
    CharStream charStream = new LexerInputCharStream(info.input(), "formula_expression");
    // No error listener needs to be added here.
    this.antlrLexer = Expressions.getANTLRLexer(charStream);
    if (info.state() == null) {
      inPathReference = false;
    } else {
      Preconditions.checkArgument(info.state().equals(STATE_IN_PATH_REFERENCE));
      inPathReference = true;
    }
  }

  @Override
  public @Nullable org.netbeans.api.lexer.Token<ExpressionTokenId> nextToken() {
    final LexerInput lexerInput = info.input();
    if (inPathReference) {
      // Do our own tokenization in this case.
      int c;
      while ((c = lexerInput.read()) != LexerInput.EOF) {
        if (pathReferenceTerminatorMatcher.matches((char) c))
          break;
      }
      final int readLengthEOF = lexerInput.readLengthEOF();
      if (readLengthEOF == 0) {
        throw new AssertionError();
      } else if (readLengthEOF == 1) {
        // Only the terminating character was read.
        inPathReference = false;
        if (c == LexerInput.EOF) {
          // Incomplete path reference terminated by EOF.
          return null;
        } else if (c == ']') {
          // Closing bracket of a complete path reference.
          return info.tokenFactory().createToken(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET);
        } else {
          /* Incomplete path reference terminated by an illegal character; back up and let the ANTLR
          lexer handle said character. */
          lexerInput.backup(1);
        }
      } else {
        /* Additional characters were read before the terminating character (possibly EOF); produce
        a core token and handle the terminating character in the next pass. */
        lexerInput.backup(1);
        return info.tokenFactory().createToken(ExpressionTokenId.PATH_REFERENCE_CORE);
      }
    }
    org.antlr.v4.runtime.Token antlrToken = antlrLexer.nextToken();
    final int readLength = lexerInput.readLength();
    final int antlrTokenLength = (antlrToken.getType() == org.antlr.v4.runtime.Lexer.EOF)
        ? 0 : Interval.of(antlrToken.getStartIndex(), antlrToken.getStopIndex()).length();
    final ExpressionTokenType tokenType;
    if (antlrTokenLength == readLength) {
      if (readLength == 0)
        return null;
      tokenType = Expressions.getTokenType(antlrToken);
      if (tokenType == ExpressionTokenType.PATH_REFERENCE ||
          tokenType == ExpressionTokenType.UNTERMINATED_PATH_REFERENCE)
      {
        /* Consume only the open bracket for now, lexing the rest of the path reference manually.
        By the next time we call antlrLexer.nextToken(), lexerInput will be be back to where it was
        before we backed up. */
        lexerInput.backup(readLength - 1);
        inPathReference = true;
        return info.tokenFactory().createToken(ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET);
      }
    } else {
      /* ANTLR's lexer may skip characters due to errors, but we need to generate a token for the
      entire length of characters read; see http://wiki.netbeans.org/Antlr_Notes . If skipped
      characters need to be part of the returned token, then don't use ANTLR's token type, as it
      applies only to the last part of said returned token. But since we include the UNKNOWN_CHAR
      token as a catch-all in our lexer definition, this case should not happen under normal
      circumstances. */
      tokenType = ExpressionTokenType.OTHER;
    }
    // TODO[cc]: Decide what I want to do with PartType. Maybe I should get rid of it.
    /* I'm not sure the PartType matters here since the complete and incomplete versions of tokens
    are already distinguished by distinct ExpressionTokenType entries. Set it in any case. Look for
    usages of Token.partType() before changing this. */
    return info.tokenFactory().createToken(tokenIdFromType(tokenType), readLength,
        (tokenType == ExpressionTokenType.UNTERMINATED_LITERAL_STRING ||
         tokenType == ExpressionTokenType.UNTERMINATED_LITERAL_TEMPORAL)
        ? PartType.START : PartType.COMPLETE);
  }

  @Override
  public Object state() {
    final int antlrLexerState = antlrLexer.getState();
    if (antlrLexerState != -1) {
      /* If this error triggers, it might be necessary to include the ANTLR lexer's state in the
      NetBeans Lexer state. */
      throw new AssertionError("Unexpected ANTLR lexer state: " + antlrLexerState);
    }
    return inPathReference ? STATE_IN_PATH_REFERENCE : null;
  }

  @Override
  public void release() {
    // No action.
  }
}
