// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.sieuferd.cell.support.CursorStyle;
import com.sieuferd.uiutil.UIUtil;
import java.awt.Color;
import javax.swing.text.AttributeSet;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;
import org.netbeans.api.editor.settings.AttributesUtilities;
import org.netbeans.api.editor.settings.EditorStyleConstants;
import org.netbeans.spi.editor.highlighting.support.OffsetsBag;

final class ReferenceHighlightStyle {
  // Uses the same technique as org.netbeans.modules.csl.editor.InstantRenamePerformer.

  // TODO[cc]: Move all color-related constants to a single place.
  public static final ReferenceHighlightStyle NEUTRAL =
      new ReferenceHighlightStyle(CursorStyle.SELECTION_COLOR_NOFOCUS);
  /* The palette below contains colors with different hues but roughly equal luminance and
  saturation, with some adjustments, based on the HUSL color model ( http://www.boronine.com/husl ).
  Luminance and saturation are not too useful as distinguishing variables, since highlights are
  already used at two different opacity levels (main labels vs. other cells in a field), and since
  black text needs to remain readable on top of a highlighted background. Keeping them similar also
  helps ensure an aesthetically pleasing palette.

  The colors in the palette should be easily distinguishable by people with normal color vision. For
  people with deuteran, protan, or tritan colorblindness (8% of the male population), the first four
  colors in the palette should remain distinguishable. Earlier entries in the palette are expected
  to be used more frequently, since most formulas would be expected only to have a few
  references. */
  private static final ImmutableList<ReferenceHighlightStyle> STYLES = ImmutableList.of(
    new ReferenceHighlightStyle(new Color(241, 149, 76)),
    new ReferenceHighlightStyle(new Color(111, 219, 160)),
    new ReferenceHighlightStyle(new Color(171, 200, 244)),
    new ReferenceHighlightStyle(new Color(221, 236, 102)),
    new ReferenceHighlightStyle(new Color(248, 140, 144)),
    new ReferenceHighlightStyle(new Color(68, 219, 216)),
    new ReferenceHighlightStyle(new Color(231, 181, 244)),
    new ReferenceHighlightStyle(new Color(223, 198, 89))
  );
  private final Color color;
  private final AttributeSet attributesLeft;
  private final AttributeSet attributesRight;
  private final AttributeSet attributesMiddle;
  private final AttributeSet attributesAll;

  ReferenceHighlightStyle(Color color) {
    Preconditions.checkNotNull(color);
    this.color = color;

    final Color borderColor = color;
    // TODO[pd]: Consider using the alpha channel instead.
    final Color backgroundColor =
        UIUtil.blendOver(borderColor, Color.WHITE, CursorStyle.BACKGROUND_TOKEN_HIGHLIGHT_OPACITY);
    attributesLeft = createAttributes(borderColor, backgroundColor, true, false);
    attributesRight = createAttributes(borderColor, backgroundColor, false, true);
    attributesMiddle = createAttributes(borderColor, backgroundColor, false, false);
    attributesAll = createAttributes(borderColor, backgroundColor, true, true);
  }

  public static ReferenceHighlightStyle getForIndex(int index) {
    return STYLES.get(index % STYLES.size());
  }

  private static AttributeSet createAttributes(
      Color borderColor, Color backgroundColor, boolean left, boolean right)
  {
    final MutableAttributeSet ret = new SimpleAttributeSet();
    ret.addAttribute(StyleConstants.Background, backgroundColor);
    ret.addAttribute(EditorStyleConstants.TopBorderLineColor, borderColor);
    ret.addAttribute(EditorStyleConstants.BottomBorderLineColor, borderColor);
    if (left)
      ret.addAttribute(EditorStyleConstants.LeftBorderLineColor, borderColor);
    if (right)
      ret.addAttribute(EditorStyleConstants.RightBorderLineColor, borderColor);
    return AttributesUtilities.createImmutable(ret);
  }

  public Color getColor() {
    return color;
  }

  public void addHighlight(OffsetsBag offsetsBag, int startOffset, int endOffset) {
    Preconditions.checkArgument(endOffset >= startOffset);
    final int length = endOffset - startOffset;
    if (length == 1) {
      /* Since path references have a minimum of two characters (the enclosing [brackets]), this
      case will normally not occur, but include it for completeness. */
      offsetsBag.addHighlight(startOffset, endOffset, attributesAll);
    } else if (length > 1) {
      offsetsBag.addHighlight(startOffset, startOffset + 1, attributesLeft);
      offsetsBag.addHighlight(endOffset - 1, endOffset, attributesRight);
      if (length > 2)
        offsetsBag.addHighlight(startOffset + 1, endOffset - 1, attributesMiddle);
    }
  };
}
