// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import java.util.Set;
import java.util.logging.Logger;
import org.netbeans.api.lexer.PartType;
import org.netbeans.api.lexer.Token;
import org.netbeans.api.lexer.TokenSequence;

public final class ReferenceInserter {
  private static final Logger LOG = Logger.getLogger(ReferenceInserter.class.getName());
  private ReferenceInserter() { }

  private static int extendInsertionPointForward(
      TokenSequence<ExpressionTokenId> tokenSequence, int at)
  {
    tokenSequence.move(at);
    if (!tokenSequence.moveNext())
      return Integer.MAX_VALUE;
    final Token<ExpressionTokenId> token = tokenSequence.token();
    final ExpressionTokenId id = token.id();
    if (id.equals(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET) ||
        id.equals(ExpressionTokenId.PARENTHESIS_LEFT))
    {
      return tokenSequence.offset() + token.length();
    } else if (id.equals(ExpressionTokenId.PATH_REFERENCE_CORE) ||
        id.equals(ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET))
    {
      while (tokenSequence.moveNext()) {
        Token<ExpressionTokenId> nextToken = tokenSequence.token();
        if (nextToken.id().equals(ExpressionTokenId.PATH_REFERENCE_CORE)) {
          // No action.
        } else if (nextToken.id().equals(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET)) {
          return tokenSequence.offset() + nextToken.length();
        }
      }
      return Integer.MAX_VALUE;
    } else if (id.equals(ExpressionTokenId.WHITESPACE) || tokenSequence.offset() == at) {
      return at;
    } else if (id.equals(ExpressionTokenId.IDENTIFIER)) {
      final int identifierEnd = tokenSequence.offset() + token.length();
      while (tokenSequence.moveNext()) {
        Token<ExpressionTokenId> nextToken = tokenSequence.token();
        if (nextToken.id().equals(ExpressionTokenId.WHITESPACE)) {
          // No action.
        } else if (nextToken.id().equals(ExpressionTokenId.PARENTHESIS_LEFT)) {
          return tokenSequence.offset() + nextToken.length();
        } else {
          return identifierEnd;
        }
      }
      final Token<ExpressionTokenId> nextToken =
          tokenSequence.moveNext() ? tokenSequence.token() : null;
      if (nextToken != null &&
          nextToken.id().equals(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET))
      {
        return tokenSequence.offset() + nextToken.length();
      } else {
        return identifierEnd;
      }
    } else {
      return tokenSequence.offset() + token.length();
    }
  }

  private static int extendInsertionPointBackward(
      TokenSequence<ExpressionTokenId> tokenSequence, int at)
  {
    tokenSequence.move(at);
    /* The movePrevious condition is not currently necessary for correct operation of
    calculateInsertionParameters, but it makes extendInsertionPointBackward behave more
    logically. */
    if (!tokenSequence.movePrevious() ||
        (tokenSequence.token().id() != ExpressionTokenId.PATH_REFERENCE_CORE &&
         tokenSequence.token().id() != ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET))
    {
      tokenSequence.move(at);
      if (!tokenSequence.moveNext())
        return at;
    }
    final Token<ExpressionTokenId> token = tokenSequence.token();
    final ExpressionTokenId id = token.id();
    if (id.equals(ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET)) {
      return tokenSequence.offset();
    } else if (id.equals(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET) ||
        id.equals(ExpressionTokenId.PATH_REFERENCE_CORE))
    {
      while (tokenSequence.movePrevious()) {
        Token<ExpressionTokenId> previousToken = tokenSequence.token();
        if (previousToken.id().equals(ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET)) {
          return tokenSequence.offset();
        } else if (previousToken.id().equals(ExpressionTokenId.PATH_REFERENCE_CORE)) {
          // No action.
        } else {
          return tokenSequence.offset();
        }
      }
      return 0;
    } else if (id.equals(ExpressionTokenId.WHITESPACE) ||
        (tokenSequence.offset() + token.length()) == at)
    {
      return at;
    } else {
      return tokenSequence.offset();
    }
  }

  // Package-visible for testing.
  static int extendInsertionPoint(
      TokenSequence<ExpressionTokenId> tokenSequence, int at, boolean forward)
  {
    final int ret;
    if (forward) {
      ret = extendInsertionPointForward(tokenSequence, at);
      if (ret < at) {
        LOG.warning("extendInsertionPointForward moved backward");
        return at;
      }
    } else {
      ret = extendInsertionPointBackward(tokenSequence, at);
      if (ret > at) {
        LOG.warning("extendInsertionPointBackward moved forward");
        return at;
      }
    }
    return ret;
  }

  private static final Set<ExpressionTokenId> VALID_PRECEDING_TOKEN_IDS = ImmutableSet.of(
      ExpressionTokenId.PARENTHESIS_LEFT,
      ExpressionTokenId.ARGUMENT_SEPARATOR,
      ExpressionTokenId.OPERATOR
  );

  private static final Set<ExpressionTokenId> REPLACEABLE_PRECEDING_TOKEN_IDS = ImmutableSet.of(
      ExpressionTokenId.LITERAL_ERROR,
      ExpressionTokenId.LITERAL_NULL
  );

  private static boolean shouldWhitespaceFollowToken(Token<ExpressionTokenId> token) {
    return
        token.id() == ExpressionTokenId.ARGUMENT_SEPARATOR ||
        /* TODO[cc]: Find a way to make the tests use the CharSequence implementation that does not
                     implement equals. */
        /* Never do token.text().equals(); it will work in tests, but break in production due to
        different implementations of the underlying CharSequence! */
        /* The "!" operator is the only one that is never followed by a space. Unary "+" and "-"
        can also be binary operators, which should be followed by a space. */
        token.id() == ExpressionTokenId.OPERATOR && !token.text().toString().equals("!");
  }

  public static InsertionParameters calculateInsertionParameters(
      TokenSequence<ExpressionTokenId> tokenSequence, int mark, int dot)
  {
    int insertStart = Math.min(mark, dot);
    int insertEnd = Math.max(mark, dot);

    insertEnd = extendInsertionPoint(tokenSequence, insertEnd, true);
    insertStart = (mark == dot) ? insertEnd :
        extendInsertionPoint(tokenSequence, insertStart, false);

    String insertPreceeding = "";
    tokenSequence.move(insertStart);
    while (tokenSequence.movePrevious()) {
      final Token<ExpressionTokenId> token = tokenSequence.token();
      final ExpressionTokenId id = token.id();
      if (!id.equals(ExpressionTokenId.WHITESPACE)) {
        if (id.equals(ExpressionTokenId.IDENTIFIER)) {
          insertPreceeding = "(";
        } else if (
            id.equals(ExpressionTokenId.PATH_REFERENCE_OPEN_BRACKET) ||
            id.equals(ExpressionTokenId.PATH_REFERENCE_CLOSE_BRACKET) ||
            id.equals(ExpressionTokenId.PATH_REFERENCE_CORE))
        {
          insertStart = extendInsertionPoint(tokenSequence, tokenSequence.offset(), false);
        } else if (REPLACEABLE_PRECEDING_TOKEN_IDS.contains(id)) {
          insertStart = tokenSequence.offset();
        } else if (id.equals(ExpressionTokenId.LITERAL_STRING) &&
            token.partType() == PartType.START)
        {
          // Terminate unterminated string literals.
          insertPreceeding = token.text().charAt(0) + ", ";
        } else if (id.equals(ExpressionTokenId.LITERAL_TEMPORAL) &&
            token.partType() == PartType.START)
        {
          // Terminate unterminated temporal literals.
          insertPreceeding =  "}, ";
        } else if (VALID_PRECEDING_TOKEN_IDS.contains(id)) {
          final boolean previousWasWhitespace;
          if (insertStart == 0) {
            previousWasWhitespace = false;
          } else {
            tokenSequence.move(insertStart - 1);
            previousWasWhitespace = tokenSequence.moveNext() &&
                tokenSequence.token().id() == ExpressionTokenId.WHITESPACE;
          }
          if (!previousWasWhitespace && shouldWhitespaceFollowToken(token) &&
              !(token.text().toString().equals("=") && tokenSequence.offset() == 0))
          {
            insertPreceeding = " ";
          }
        } else {
          insertPreceeding = ", ";
        }
        break;
      }
    }
    return new InsertionParameters(new Interval(insertStart, insertEnd), insertPreceeding);
  }

  public static final class Interval {
    public final int start;
    public final int end;

    public Interval(int start, int end) {
      Preconditions.checkArgument(start >= 0);
      Preconditions.checkArgument(end >= start);
      this.start = start;
      this.end = end;
    }

    public boolean equalsInterval(Interval other) {
      return start == other.start
          && end == other.end;
    }
  }

  public static final class InsertionParameters {
    public Interval interval;
    public String insertPreceeding;

    public InsertionParameters(Interval interval, String insertPreceeding) {
      Preconditions.checkNotNull(interval);
      Preconditions.checkNotNull(insertPreceeding);
      this.interval = interval;
      this.insertPreceeding = insertPreceeding;
    }
  }
}
