// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar.types;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.project.formulabar.ExpressionLanguageConfig;
import com.sieuferd.project.formulabar.CursorEditKey;
import com.sieuferd.project.formulabar.CursorEditType;
import com.sieuferd.project.formulabar.CursorEditTarget;
import com.sieuferd.query.QProp;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.project.ViewContext;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.ObjectReference;
import com.sieuferd.widget.InteractionProvider.Acceptance;
import java.util.List;
import javax.annotation.Nullable;
import org.openide.DialogDisplayer;
import org.openide.NotifyDescriptor;
import org.openide.util.ImageUtilities;
import org.openide.util.NbBundle;
import org.openide.util.NbBundle.Messages;
import org.openide.util.lookup.ServiceProvider;

@ServiceProvider(service = CursorEditType.class)
public final class FormulaType extends CursorEditType {
  public FormulaType() {
    super(ImageUtilities.loadImageIcon(ICON_FORMULA, false),
        ExpressionLanguageConfig.MIME_TYPE, false);
  }

  public static String getActionDisplayName() {
    return Bundle.CTL_EditFormula();
  }

  @Messages("CTL_EditFormula=Edit Formula")
  @Override
  public CursorEditKey createEditKey(
      CursorEditTarget target, ObjectReference<String> originalValueString)
  {
    final ImmutableView view = target.getResult().getView();
    final Path path = target.getCursorLocation().getPath();
    if (!target.isValue() || path.isRelation() || !view.get(QProp.SQL_COLUMN_NAME, path).isEmpty())
      return null;
    originalValueString.setValue(Expressions.generateString(target.getStrepco(),
          target.getResult().getView().get(QProp.FORMULA_EXPRESSION, path), true));
    return new FormulaEditKey(path.getPrimitive());
  }

 @NbBundle.Messages({
    "# {0} - Proposed corrected formula",
    // Similar to Excel's analogous dialog.
    "MSG_ErrorInExpression=<html>" +
      "The formula you entered had one or more errors.<br>" +
      "Do you want to accept the correction proposed below?<br><br>{0}</html>",
    "TIT_ErrorInExpression=Error in Formula"
  })
  public static @Nullable Expression parseUserExpression(
      Strepco strepco, String str, ObjectReference<Boolean> retDialogShown)
  {
    ThreadUtil.checkOnEDT();
    final List<ExpressionErrorReport> errors = Lists.newArrayList();
    final Expression exp = Expressions.parse(strepco, str, errors);
    if (errors.isEmpty()) {
      retDialogShown.setValue(false);
    } else {
      retDialogShown.setValue(true);
      final String proposedCorrectedFormula = Expressions.generateString(strepco, exp, true);
      final NotifyDescriptor nd = new NotifyDescriptor(Bundle.MSG_ErrorInExpression(
          SwingString.toSwingHTML(proposedCorrectedFormula)), Bundle.TIT_ErrorInExpression(),
          NotifyDescriptor.YES_NO_OPTION, NotifyDescriptor.INFORMATION_MESSAGE, null, null);
      if (!NotifyDescriptor.YES_OPTION.equals(DialogDisplayer.getDefault().notify(nd)))
        return null;
    }
    return exp;
  }

  private final class FormulaEditKey extends CursorEditKey {
    private final PrimitivePath path;

    public FormulaEditKey(PrimitivePath path) {
      super(FormulaType.this, false);
      Preconditions.checkNotNull(path);
      this.path = path;
    }

    @Override
    public Acceptance setValue(ViewContext viewContext, Strepco strepco, String newValueString) {
      /* Parse the expression back using the same context view as was used to generate it in the
      first place. */
      final ObjectReference<Boolean> dialogShown = ObjectReference.create();
      final Expression exp = parseUserExpression(strepco, newValueString, dialogShown);
      if (exp == null) {
        return Acceptance.REJECTED;
      } else {
        viewContext.getView().mutate(Bundle.CTL_EditFormula(), true, (MutableView view) ->
        {
          if (view.exists(path))
            ViewUpdater.userSetProperty(view, QProp.FORMULA_EXPRESSION, path.getPrimitive(), exp);
        });
        /* Always accept, even if the label could not be set because the path no longer exists; this
        seems to be a reasonable thing to do if the path was actually deleted. */
        return dialogShown.getValue() ?
            Acceptance.ACCEPTED_WITH_INTERACTION : Acceptance.ACCEPTED_IMMEDIATELY;
      }
    }

    @Override
    public int hashCode() {
      return path.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
      if (!(obj instanceof FormulaEditKey))
        return false;
      final FormulaEditKey other = (FormulaEditKey) obj;
      return this.path.equals(other.path);
    }
  }
}
