// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.navigator;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.ViewAdapter;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.ViewPropertyChangedEvent;
import com.sieuferd.datamodel.Views;
import com.sieuferd.project.nodes.FieldNode;
import java.util.List;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.nodes.FieldNodePathFinder;
import com.sieuferd.uiutil.ExplorerManagerProviderPanel;
import com.sieuferd.project.nodes.NodeSettings;
import com.sieuferd.project.util.FieldSelection;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.table.ColorOverrideRenderer;
import com.sieuferd.uiutil.api.TableColors;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.swing.JTable;
import javax.swing.JViewport;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;
import javax.swing.table.TableColumn;
import org.netbeans.swing.outline.Outline;
import org.openide.explorer.ExplorerManager;
import org.openide.explorer.view.CheckableNode;
import org.openide.explorer.view.OutlineView;
import org.openide.nodes.Node;
import org.openide.util.Mutex;
import org.openide.util.NbBundle;

public final class FieldsPanel extends ExplorerManagerProviderPanel {
  private static final boolean EXPAND_NODE_ON_VISIBLE = true;
  private static final Logger LOG = Logger.getLogger(FieldsPanel.class.getName());
  private @Nullable FieldContext rootFieldContext;
  private final ViewListener viewListener = new ViewAdapter() {
    @Override
    public <T> void propertyChanged(ViewPropertyChangedEvent<T> evt) {
      if (rootFieldContext == null)
        return;
      if (!evt.getSource().exists(rootFieldContext.getPath())) {
        Mutex.EVENT.writeAccess(() -> {
          selectFieldContexts(ImmutableSet.of(), false);
        });
      }
      if (EXPAND_NODE_ON_VISIBLE && evt.getProperty().equals(CProp.VISIBLE)) {
        // TODO[cc]: Do I already have this logic somewhere else?
        final Node selectedNodes[] = getExplorerManager().getSelectedNodes();
        if (selectedNodes.length == 1) {
          final Node selectedNode = selectedNodes[0];
          if (!selectedNode.isLeaf()) {
            final FieldContext fieldContext = selectedNode.getLookup().lookup(FieldContext.class);
            if (fieldContext != null && evt.getPath().equals(fieldContext.getPath())) {
              if (fieldContext.get(CProp.VISIBLE)) {
                fieldsView.expandNode(selectedNode);
              } else {
                fieldsView.collapseNode(selectedNode);
              }
            }
          }
        }
      }
      /* Workaround for an OutlineView repaint bug. When the OutlineView has a row selected and a
      property on a different row changes, the property on the unselected row gets painted with the
      background color of a selected row. */
      fieldsView.getOutline().repaint();
    }
  };

  private void setRootFieldContext(@Nullable FieldContext fieldContext) {
    if (rootFieldContext != null)
      rootFieldContext.getView().removeViewListener(viewListener);
    rootFieldContext = fieldContext;
    if (rootFieldContext != null)
      rootFieldContext.getView().addViewListener(null, viewListener);
  }

  public @Nullable FieldContext getRootFieldContext() {
    return rootFieldContext;
  }

  /**
   * Update the FieldsPanel to show and select the provided {@code fieldContexts}. The selection is
   * only changed at the moment during which this method is called, and may henceforth be modified
   * by the user. Setting the selection is only guaranteed to have an effect while the FieldsPanel
   * is displayable.
   */
  public void selectFieldContexts(Set<FieldContext> fieldContexts, boolean navigatorMode) {
    ThreadUtil.checkOnEDT();
    final @Nullable FieldContext newRootFieldContext;
    boolean rootVisible = navigatorMode;
    {
      final FieldSelection fieldSelection =
          FieldSelection.fromFieldContexts(fieldContexts, AppliesTo.ANY);
      if (fieldSelection == null) {
        newRootFieldContext = null;
      } else {
        final Set<RelationPath> rootPathCandidates = Sets.newLinkedHashSet();
        final ViewContext viewContext = fieldSelection.getViewContext();
        for (Path path : fieldSelection.getSelectedPaths()) {
          /* If navigatorMode is true, go one level up from the actual selection. This way, moving
          between siblings fields never changes the displayed root node, and there is always a
          relation consistently defined as the display root whether the user selects a primitive
          node or a relation node. */
          rootPathCandidates.add((navigatorMode || !path.isRelation())
              ? (path.isRoot() ? Path.ROOT : path.getParent())
              : path.getRelation());
        }
        RelationPath rootNodePath = Views.findLowestCommonRelationAncestor(rootPathCandidates);
        if (navigatorMode) {
          /* Use the plural ancestor, or the node itself if already plural, as the root. Don't try
          to refresh the navigator if CProp.PLURAL is changed, as it would be more confusing than
          helpful. Instead, the navigator will only change its contents when a new selection is made
          or the navigator root node is deleted. */
          while (!rootNodePath.isRoot() &&
              !viewContext.getView().getOrDefault(CProp.PLURAL, rootNodePath))
          {
            rootNodePath = rootNodePath.getParent();
          }
        } else {
          while (!rootNodePath.isRoot() &&
              !Views.hasVisiblePrimitiveChildren(viewContext.getView(), rootNodePath.getParent()))
          {
            rootNodePath = rootNodePath.getParent();
            rootVisible = true;
          }
        }
        newRootFieldContext = FieldContext.create(viewContext, rootNodePath);
        rootVisible |= fieldSelection.getSelectedPaths().contains(rootNodePath);
      }
    }
    // Always recreate the root node to collapse expanded relations, except in navigator mode.
    if (!(navigatorMode && Objects.equal(rootFieldContext, newRootFieldContext))) {
      setRootFieldContext(newRootFieldContext);
      if (newRootFieldContext != null) {
        fieldsView.setEnabled(true);
        fieldsView.getOutline().setRootVisible(rootVisible);
        Node rootNode = new FieldNode(newRootFieldContext,
            new NodeSettings(AppliesTo.ANY, false, false, true, true, null, true));
        getExplorerManager().setRootContext(rootNode);

        /* Work around a bug where if the node to be selection changes from a non-leaf to a leaf
        node, the Outline does not get properly repainted. Could this be due to behavior being
        undefined if using a leaf node as the root? Doing
        fieldsView.getOutline().repaint() would also work here, but in case there's a deeper
        problem, I figured expandNode() has a higher probability of fixing any underlying broken
        state. Note that using JViewport.SIMPLE_SCROLL_MODE does not prevent this bug. */
        fieldsView.expandNode(rootNode);
      } else {
        fieldsView.setEnabled(false);
        try {
          getExplorerManager().setSelectedNodes(new Node[] {});
        } catch (PropertyVetoException e) {
          LOG.log(Level.WARNING, "Failed to unselect navigator nodes", e);
        }
        getExplorerManager().setRootContext(Node.EMPTY);
        fieldsView.getOutline().setRootVisible(false);
      }
      fieldsView.getViewport().setViewPosition(new Point(0, 0));
    }
    if (newRootFieldContext != null) {
      final Node nodesToSelect[] =
          FieldNodePathFinder.findNodes(getExplorerManager().getRootContext(), fieldContexts);
      /* Usability tweak: Expand any single selected (relation) node. This ensures that a user that
      clicks a relation label gets to see the available child fields in the Navigator. Do this
      _before_ selecting nodes, however, as the node expansion operation has been observed to mess
      up the selection (see the possibly related bug involving ETable.tableChanged() mentioned
      further below). Also avoid trying to expand non-leaf nodes, just to be safe. */
      if (nodesToSelect.length == 1 && !nodesToSelect[0].isLeaf())
        fieldsView.expandNode(nodesToSelect[0]);
      try {
        getExplorerManager().setSelectedNodes(nodesToSelect);
        /* There is a bug, likely in OutlineView, where the wrong node may end up being selected if
        a node needs to be expanded to make the node in question visible. An example can be seen in
        the "04 InfoVis Full Table" example if the "Number" field is selected and then the
        "(Department/)Code" field--the navigator ends up expanding the "Department" node to reveal
        the "Code" field, but the selection stays at "Number". The workaround is to simply repeat
        the operation once. */
        getExplorerManager().setSelectedNodes(nodesToSelect);
      } catch (PropertyVetoException e) {
        LOG.log(Level.WARNING, "Failed to select navigator nodes", e);
      }
    }
  }

  @NbBundle.Messages("CTL_Fields=Fields")
  public static void configureFieldsOutlineView(
      OutlineView fieldsView, boolean allowColumnSelection)
  {
    /* Do this before doing other configurations, as the call to OutlineView.addPropertyColumn will
    trigger a call to ETable.createDefaultColumnsFromModel, which will reset any custom
    renderers. */
    if (allowColumnSelection) {
      ViewPropertyTableColumnSelector.configureColumns(fieldsView);
    } else {
      fieldsView.getOutline().setColumnHidingAllowed(false);
    }

    UIUtil.setOutlineColumnsLabel(fieldsView.getOutline(), Bundle.CTL_Fields());
    // Permit in-place renaming.
    fieldsView.setDefaultActionAllowed(false);
    fieldsView.getOutline().setShowHorizontalLines(false);
    fieldsView.getOutline().setShowVerticalLines(false);
    fieldsView.getOutline().setIntercellSpacing(new Dimension(0, 0));
    /* TODO[pd]: Put setFillsViewportHeight(true) back in once the bug below is fixed and released
                 (NPE when clicking in the empty area underneath the last node):
                 https://netbeans.org/bugzilla/show_bug.cgi?id=247882 */
    // fieldsView.getOutline().setFillsViewportHeight(true);
    fieldsView.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    fieldsView.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    fieldsView.setTreeHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    fieldsView.getOutline().setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
    // Field order matters, so avoid confusion by turning off sorting.
    fieldsView.setTreeSortable(false);
    fieldsView.getOutline().setRowSorter(null);
    /* Don't enable unless I can figure out how to prevent the order to be reset when the user
    enters the column selector dialog and presses Cancel. It's probably best to keep this off in any
    case. */
    fieldsView.getOutline().getTableHeader().setReorderingAllowed(false);
    /* Improve drag handling. This avoids a bug where drag-to-select interferes with
    drag-to-reorder, but unfortunately introduces a less than ideal dragging behavior which by
    design requires cells to be selected before they can be reordered by dragging. */
    fieldsView.getOutline().setDragEnabled(true);
    /*
    Below are some other hacks which do not _work_ properly. The first one causes various bugs when
    displaying more than one column, and makes node expansion not work properly. Besides, it uses
    a private API which is subject to change in the future:

      fieldsView.getOutline().putClientProperty("Table.isFileList", true);

    This one makes drag-and-drop work properly, but unfortunately requires us to disable multiple
    selection, which is unacceptable:

      fieldsView.getOutline().getSelectionModel()
          .setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
      fieldsView.getOutline().getColumnModel().getSelectionModel()
          .setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    */

    /* TODO[pd]: See if this workaround only needs to be applied on MacOS, or whether it can now be
                 removed entirely. */
    /* On Java 6 on MacOS, there's a repaint bug in OutlineView or one of its underlying components
    that causes the background to be a slightly different shade of gray on newly scrolled-to areas.
    Using the dumb scroll mode avoids this. */
    fieldsView.getViewport().setScrollMode(JViewport.SIMPLE_SCROLL_MODE);

    /* Avoid hijacking the standard focus traversal keys (for when the OutlineView is used in a
    dialog). */
    UIUtil.setSimpleFocusTraversalKeys(fieldsView.getOutline());

    TreeCellRendererInjector injector = new TreeCellRendererInjector(fieldsView.getOutline());
    injector.inject();
    fieldsView.getOutline().addPropertyChangeListener(TreeCellRendererInjector.PROPERTY, injector);
  }

  private static final class TreeCellRendererInjector implements PropertyChangeListener {
    public static final String PROPERTY = "createdDefaultColumnsFromModel";
    private final Outline outline;

    public TreeCellRendererInjector(Outline outline) {
      Preconditions.checkNotNull(outline);
      this.outline = outline;
    }

    private void inject() {
      final TableColumn treeColumn = outline.getColumnModel().getColumn(0);
      /* Support SwingHtmlDisplayNameProvider, and apply the bugfixes from ColorOverrideRenderer to
      the tree column as well. */
      if (!(treeColumn.getCellRenderer() instanceof ColorOverrideRenderer)) {
        treeColumn.setCellRenderer(new ColorOverrideRenderer(new SwingStringOutlineCellRenderer(),
            TableColors.getInstance()));
      }
    }

    @Override
    public void propertyChange(PropertyChangeEvent evt) {
      if (PROPERTY.equals(evt.getPropertyName()) && evt.getNewValue() != null)
        inject();
    }
  }

  public FieldsPanel() {
    this(false);
  }

  public FieldsPanel(boolean allowColumnSelection) {
    ThreadUtil.checkOnEDT();
    initComponents();
    configureFieldsOutlineView(fieldsView, allowColumnSelection);
    fieldsView.getOutline().setRootVisible(false);

    /* TODO[cc]: Consider factoring out the common functionality between here and
                 UIUtil.configureBooleanTableValueShortcuts. */
    /* The space-on-multiple selection behavior on NetBeans does not do quite what I want (see
    http://netbeans.org/bugzilla/show_bug.cgi?id=243922 ), so override it here. This code does not
    seem to cause problems with the search or editing functionality of the Outline; space still
    works in the related embedded components. See similar logic in
    UIUtil.configureBooleanTableValueShortcuts. */
    fieldsView.getOutline().addKeyListener(new KeyAdapter() {
      private boolean alreadyPressed;

      @Override
      public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_SPACE && e.getModifiersEx() == 0) {
          if (!alreadyPressed) {
            final List<CheckableNode> checkableNodes = Lists.newArrayList();
            boolean allPreviouslyChecked = true;
            for (Node node : getExplorerManager().getSelectedNodes()) {
              CheckableNode checkableNode = node.getLookup().lookup(CheckableNode.class);
              if (checkableNode != null && checkableNode.isCheckEnabled()) {
                checkableNodes.add(checkableNode);
                allPreviouslyChecked &= Boolean.TRUE.equals(checkableNode.isSelected());
              }
            }
            for (CheckableNode checkableNode : checkableNodes)
              checkableNode.setSelected(!allPreviouslyChecked);
            alreadyPressed = true;
          }
          e.consume();
        }
      }

      @Override
      public void keyReleased(KeyEvent e) {
        alreadyPressed = false;
      }
    });

    /* There seems to be a bug in OutlineView where during certain structural changes to the node
    tree, selection is lost. I have traced this back to a call to Outline.tableChanged(), which
    calls its superclass method ETable.tableChanged(), which promises to keep the selection intact.
    ETable.tableChanged() in turn calls its superclass method JTable.tableChanged(), which warns
    that it intentionally will clear the selection, but ETable.tableChanged() then attempts to
    restore the selection. Something probably goes wrong at this point. */
    getExplorerManager().addPropertyChangeListener((PropertyChangeEvent evt) -> {
      if (evt.getPropertyName().equals(ExplorerManager.PROP_SELECTED_NODES)) {
        final Node[] oldValue = (Node[]) evt.getOldValue();
        final Node[] newValue = (Node[]) evt.getNewValue();
        if (oldValue.length > 0 && newValue.length == 0) {
          // Useful for debugging.
          if (false) {
            new Exception("Selected node change " +
                ImmutableList.copyOf(oldValue) + " became " +
                ImmutableList.copyOf(newValue)).printStackTrace();
          }
          SwingUtilities.invokeLater(() -> {
            if (getExplorerManager().getSelectedNodes().length == 0) {
              try {
                getExplorerManager().setSelectedNodes(oldValue);
              } catch (PropertyVetoException e) {
                LOG.log(Level.WARNING, "Failed to apply lost node selection bug workaround", e);
              }
            }
          });
        }
      }
    });
  }

  public OutlineView getOutlineView() {
    return fieldsView;
  }

  /* Like when implementing a TopComponent, we override requestFocusInWindow() and requestFocus() to
  ensure proper focus behavior for instance when the navigator component is activated using
  Ctrl+Tab. See similar code in org.netbeans.modules.csl.navigation.ClassMemberPanelUI. */

  @Override
  public boolean requestFocusInWindow() {
    ThreadUtil.checkOnEDT();
    if (!isFocusable())
      return false;
    return fieldsView.requestFocusInWindow();
  }

  @Override
  public void requestFocus() {
    ThreadUtil.checkOnEDT();
    if (!isFocusable())
      return;
    fieldsView.requestFocus();
  }

  /**
   * This method is called from within the constructor to initialize the form. WARNING: Do NOT
   * modify this code. The content of this method is always regenerated by the Form Editor.
   */
  // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
  private void initComponents() {

    fieldsView = new org.openide.explorer.view.OutlineView();

    setLayout(new java.awt.BorderLayout());
    add(fieldsView, java.awt.BorderLayout.CENTER);
  }// </editor-fold>//GEN-END:initComponents

  // Variables declaration - do not modify//GEN-BEGIN:variables
  private org.openide.explorer.view.OutlineView fieldsView;
  // End of variables declaration//GEN-END:variables
}
