// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.navigator;

import com.google.common.base.Preconditions;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.UIUtil;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Shape;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;
import org.netbeans.swing.outline.DefaultOutlineCellRenderer;
import org.openide.explorer.view.Visualizer;
import org.openide.nodes.Node;

final class SwingStringOutlineCellRenderer implements TableCellRenderer {
  private final DefaultOutlineCellRenderer delegate = new DefaultOutlineCellRenderer();
  private final JLabel label = new JLabel();

  @Override
  public Component getTableCellRendererComponent(
      JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column)
  {
    /* The component generated by DefaultOutlineCellRenderer will be from a special subclass of
    JLabel that uses NetBeans' own HTML renderer by default. The most portable way to change this
    behavior seems to be to let the DefaultOutlineCellRenderer configure its own JLabel but then
    copy over all the properties to a new, regular JLabel instead. */
    Component delegateComponent =
        delegate.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
    if (!(delegateComponent instanceof JLabel))
      return delegateComponent;
    SwingString displayName = null;
    {
      final Node node = (value == null) ? null : Visualizer.findNode(value);
      if (node != null) {
        final SwingStringDisplayNameProvider provider =
            node.getLookup().lookup(SwingStringDisplayNameProvider.class);
        if (provider != null)
          displayName = provider.getSwingStringDisplayName();
      }
    }
    if (displayName == null)
      return delegateComponent;

    label.setText(displayName.getComplete());
    /* Take all other JLabel properties that could be relevant to rendering from the JLabel
    configured by the delegate. */
    JLabel dLabel = (JLabel) delegateComponent;
    label.setBorder(new DelegateBorder(dLabel));
    label.setIcon(dLabel.getIcon());
    label.setDisabledIcon(dLabel.getDisabledIcon());
    label.setIconTextGap(dLabel.getIconTextGap());
    label.setFont(dLabel.getFont());
    label.setAlignmentX(dLabel.getAlignmentX());
    label.setAlignmentY(dLabel.getAlignmentY());
    label.setEnabled(dLabel.isEnabled());
    label.setHorizontalAlignment(dLabel.getHorizontalAlignment());
    label.setVerticalAlignment(dLabel.getVerticalAlignment());
    label.setHorizontalTextPosition(dLabel.getHorizontalTextPosition());
    label.setVerticalTextPosition(dLabel.getVerticalTextPosition());
    label.setSize(dLabel.getSize());
    label.setBackground(dLabel.getBackground());
    label.setForeground(dLabel.getForeground());
    label.setOpaque(dLabel.isOpaque());
    return label;
  }

  private static final class DelegateBorder implements Border {
    private final JLabel delegateLabel;

    public DelegateBorder(JLabel delegateLabel) {
      Preconditions.checkNotNull(delegateLabel);
      this.delegateLabel = delegateLabel;
    }

    @Override
    public void paintBorder(Component c, Graphics g, int x, int y, int width, int height) {
      Shape oldClip = g.getClip();
      try {
        /* TODO[pd]: Paint a proper lead selection border from ColorOverrideRenderer instead,
                     including on Windows. This way all columns can be made to include the
                     border. */
        /* TODO[cc]: File a bug about the OutlineView checkbox overpainting the lighter-blue
                     selection border on MacOS (purely cosmetic, not a repaint bug). Then get rid of
                     this workaround. */
        /* The lead selection rectangle looks bad on Aqua (overpainted by checkbox, dim color), and
        does not serve much of a purpose since MacOS does not allow pick selection by keyboard, so
        don't bother painting it. */
        if (UIUtil.isAqua())
          g.clipRect(x + 1, y + 1, width - 2, height - 2);
        delegateLabel.getBorder().paintBorder(delegateLabel, g, x, y, width, height);
      } finally {
        g.setClip(oldClip);
      }
    }

    @Override
    public Insets getBorderInsets(Component c) {
      return delegateLabel.getBorder().getBorderInsets(delegateLabel);
    }

    @Override
    public boolean isBorderOpaque() {
      return delegateLabel.getBorder().isBorderOpaque();
    }
  }
}
