// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.navigator;

import com.google.common.collect.ImmutableSet;
import com.sieuferd.flow.Flow;
import com.sieuferd.uiutil.flow.LookupResultSnapshot;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.flow.Flows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.result.ViewResultProvider;
import com.sieuferd.project.viewfile.ViewDataObject;
import java.util.Set;
import javax.swing.JComponent;
import javax.swing.SwingUtilities;
import org.netbeans.spi.navigator.NavigatorPanel;
import org.netbeans.spi.navigator.NavigatorPanelWithUndo;
import org.openide.awt.UndoRedo;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.Lookups;

// See org.netbeans.modules.beans.BeanPanel in the beans module.

@NbBundle.Messages({
  "LBL_Fields=Fields",
  "HINT_Fields=Shows the fields at the currently selected view node"
})
@NavigatorPanel.Registration(mimeType=ViewDataObject.MIME_TYPE,
    position=1,
    displayName="#LBL_Fields")
public final class ViewNavigatorPanel implements NavigatorPanelWithUndo {
  private FieldsPanel fieldsPanel;
  private final SettableFlow<Lookup> contextLookupSettable = Flows.newSettable();
  private final Flow<LookupResultSnapshot<FieldContext>> selectedResultValue =
      Flows.newIndirect(contextLookupSettable,
      lookup -> UIFlows.newLookup(lookup, new Lookup.Template<>(FieldContext.class)));
  private final UndoRedo undoRedo;

  public ViewNavigatorPanel() {
    UIFlows.bindEDT(contextLookupSettable, (Lookup value) -> {
      getFieldsPanel().setActionsActivated(value != null);
    });
    UIFlows.bindEDT(selectedResultValue, (LookupResultSnapshot<FieldContext> value) -> {
      if (value != null) {
        Set<FieldContext> selected = ImmutableSet.copyOf(value.allInstances());
        if (!selected.isEmpty())
          getFieldsPanel().selectFieldContexts(selected, true);
      }
    });
    undoRedo = ViewResultProvider.createUndoRedo(
        Flows.newIndirect(contextLookupSettable, lookup ->
        Flows.newIndirect(UIFlows.newSingularLookup(lookup, ViewDataObject.class), viewDataObject ->
        UIFlows.newSingularLookup(viewDataObject.getLookup(), ViewResultProvider.class)))
    );
  }

  @Override
  public String getDisplayName() {
    return Bundle.LBL_Fields();
  }

  @Override
  public String getDisplayHint() {
    return Bundle.HINT_Fields();
  }

  @Override
  public JComponent getComponent() {
    return getFieldsPanel();
  }

  private synchronized FieldsPanel getFieldsPanel() {
    if (fieldsPanel == null) {
      fieldsPanel = new FieldsPanel(true);
      /* If the navigator panel is initially focusable, it will steal the focus the first time it is
      shown, for instance when a relevant node is selected with the mouse or at startup in a
      different TopComponent. Furthermore, the focus would be stolen in an odd manner that didn't
      actually mark the navigator as the active TopComponent in the window system (with a blue focus
      indicator on its tab). Note that this focus stealing has also been observed to happen with the
      NetBeans IDE's own Java members navigator. To avoid the focus stealing, we postpone making the
      navigator panel focusable until the end of the event queue, and respect isFocusable() in our
      overridden implementations of requestFocusInWindow() and requestFocus(). */
      fieldsPanel.setFocusable(false);
      SwingUtilities.invokeLater(() -> {
        fieldsPanel.setFocusable(true);
      });
    }
    return fieldsPanel;
  }

  @Override
  public void panelActivated(Lookup lookup) {
    contextLookupSettable.set(lookup);
  }

  @Override
  public void panelDeactivated() {
    contextLookupSettable.set(null);
  }

  @Override
  public Lookup getLookup() {
    /* Avoid exposing DataObject or FileObject objects from the Navigator, as it seems to trigger a
    resource leak when "Synchronize Editor with Views" is enabled. */
    return Lookups.exclude(getFieldsPanel().getLookup(), DataObject.class, FileObject.class);
  }

  @Override
  public UndoRedo getUndoRedo() {
    return undoRedo;
  }
}
