// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.navigator;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.project.NBViewSet;
import com.sieuferd.util.MapUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.table.ColorOverrideRenderer;
import com.sieuferd.uiutil.api.TableColors;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import org.netbeans.swing.etable.ETableColumnModel;
import org.netbeans.swing.etable.TableColumnSelector;
import org.openide.explorer.view.OutlineView;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.openide.util.NbPreferences;
import org.openide.util.RequestProcessor;

final class ViewPropertyTableColumnSelector extends TableColumnSelector {
  /** These properties are already represented by the first column of the OutlineView, which
  contains the node tree. */
  public static final List<ViewProperty<?>> TREE_COLUMN_PROPERTIES =
      ImmutableList.<ViewProperty<?>>of(CProp.LABEL, CProp.VISIBLE);
  private static final Logger LOG =
      Logger.getLogger(ViewPropertyTableColumnSelector.class.getName());
  private static final Preferences PREFERENCES =
      NbPreferences.forModule(FieldsPanel.class).node("NavigatorPreferences");
  private final OutlineView view;

  public ViewPropertyTableColumnSelector(OutlineView view) {
    Preconditions.checkNotNull(view);
    this.view = view;
  }

  @Override
  public String[] selectVisibleColumns(TreeNode root, String[] selected) {
    /* This method is only called if ColumnModel.setColumnHierarchyRoot() has been used. See
     org.netbeans.swing.etable.ColumnSelectionPanel.showColumnSelectionDialog() . */
    throw new UnsupportedOperationException();
  }

  @NbBundle.Messages("TIT_SelectColumns=Select Navigator Columns")
  @Override
  public String[] selectVisibleColumns(String[] available, String[] selected) {
    Map<String,Boolean> selectedMap = Maps.newLinkedHashMap();
    for (String key : available)
      selectedMap.put(key, false);
    for (String key : selected)
      selectedMap.put(key, true);
    // The tree column should always be first. Remove here and reinsert below before the other keys.
    selectedMap.remove(getColumnIdentifierForTree(view));
    Map<String,Boolean> originalSelectedMap = ImmutableMap.copyOf(selectedMap);
    JPanel panel = new ViewPropertyColumnSelectorPanel(selectedMap);
    /* Use getWindowForComponent() to center the dialog on the window instead of just the navigator
    component, since the column selection dialog can be large and it would be undesirable to cover
    the navigator while the user is trying to select columns to show. */
    int res = JOptionPane.showConfirmDialog(UIUtil.getWindowForComponent(view), panel,
        Bundle.TIT_SelectColumns(), JOptionPane.OK_CANCEL_OPTION);
    if (res == JOptionPane.OK_OPTION) {
      /* Persist the column selection across restarts. Don't use
      ETableColumnModel.read/writeSettings(), as they are too low-level and don't preserve
      invariants well. */
      for (ViewProperty<?> prop : NBViewSet.VIEW_DEFAULTS.getProperties()) {
        if (TREE_COLUMN_PROPERTIES.contains(prop))
          continue;
        final String cid = getColumnIdentifierForProperty(prop);
        final String pname = getVisibilityPreferenceKey(prop);
        if (Boolean.TRUE.equals(selectedMap.get(cid))) {
          PREFERENCES.putBoolean(pname, true);
        } else {
          PREFERENCES.remove(pname);
        }
      }
      RequestProcessor.getDefault().create(() -> {
        try {
          PREFERENCES.flush();
        } catch (BackingStoreException e) {
          LOG.log(Level.WARNING, "Failed to persist navigator column selections", e);
        }
      });
      UIUtil.updatePreferredOutlineColumnWidths(view.getOutline());
    }
    /* Both available and selected are supplied in alphabetical order, and the order of the returned
    identifiers is used to reorder columns, so I have to redo the ordering even if the user presses
    cancel. This is fine, since I don't want to allow column reordering anyway. */
    Map<String,Boolean> useSelectedMap =
        (res == JOptionPane.OK_OPTION) ? selectedMap : originalSelectedMap;
    List<String> ret = Lists.newArrayList();
    ret.add(getColumnIdentifierForTree(view));
    for (ViewProperty<?> prop : NBViewSet.VIEW_DEFAULTS.getProperties()) {
      String key = getColumnIdentifierForProperty(prop);
      if (Boolean.TRUE.equals(useSelectedMap.get(key)))
        ret.add(key);
    }
    return ret.toArray(new String[ret.size()]);
  }

  public static void configureColumns(OutlineView view) {
    checkUniqueColumnIdentifiers(view);
    final ETableColumnModel columnModel;
    {
      TableColumnModel toColumnModel = view.getOutline().getColumnModel();
      if (toColumnModel instanceof ETableColumnModel) {
        columnModel = (ETableColumnModel) toColumnModel;
      } else {
        LOG.warning("TableColumnModel for OutlineView was not an ETableColumnModel");
        columnModel = null;
      }
    }
    if (columnModel != null) {
      // Make all view properties available as initially hidden columns.
      for (ViewProperty<?> prop : NBViewSet.VIEW_DEFAULTS.getProperties()) {
        if (TREE_COLUMN_PROPERTIES.contains(prop))
          continue;
        view.addPropertyColumn(
            prop.getBeanPropertyName(), prop.getDisplayName(), prop.getShortDescription());
        TableColumn column =
            view.getOutline().getColumn(getColumnIdentifierForProperty(prop));
        columnModel.setColumnHidden(column,
            !ViewPropertyTableColumnSelector.PREFERENCES
            .getBoolean(getVisibilityPreferenceKey(prop), false));
      }
      // Workaround for https://netbeans.org/bugzilla/show_bug.cgi?id=238079 .
      view.getOutline().setDefaultRenderer(Node.Property.class, new ColorOverrideRenderer(
            view.getOutline().getDefaultRenderer(Node.Property.class), TableColors.getInstance()));
    }
    view.getOutline().setColumnSelector(new ViewPropertyTableColumnSelector(view));
    view.getOutline().setColumnHidingAllowed(true);
    view.getOutline().setShowVerticalLines(true);
    UIUtil.updatePreferredOutlineColumnWidths(view.getOutline());
  }

  private static String getVisibilityPreferenceKey(ViewProperty<?> property) {
    return "visible." + property.getStorageName();
  }

  static String getColumnIdentifierForProperty(ViewProperty<?> property) {
    return property.getDisplayName();
  }

  private static String getColumnIdentifierForTree(OutlineView outlineView) {
    return outlineView.getOutline().getModel().getColumnName(0);
  }

  // TODO[te]: Include this check in a unit test.
  private static void checkUniqueColumnIdentifiers(OutlineView outlineView) {
    Set<String> columnIdentifiers = Sets.newLinkedHashSet();
    MapUtil.addChecked(columnIdentifiers, getColumnIdentifierForTree(outlineView));
    for (ViewProperty<?> prop : NBViewSet.VIEW_DEFAULTS.getProperties())
      MapUtil.addChecked(columnIdentifiers, getColumnIdentifierForProperty(prop));
  }
}
