// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.nodes;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.ViewAdapter;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewListener;
import java.awt.Image;
import javax.swing.Action;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.datamodel.ViewPropertyChangedEvent;
import com.sieuferd.datamodel.ViewPropertySet;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.SieuferdIcons;
import com.sieuferd.project.util.QueryLabelFormatter;
import com.sieuferd.uiutil.SwingString;
import com.sieuferd.upstream.actions.ActionUtil;
import com.sieuferd.uiutil.ThreadUtil;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Map.Entry;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import org.netbeans.api.actions.Openable;
import org.openide.actions.CopyAction;
import org.openide.actions.CutAction;
import org.openide.actions.DeleteAction;
import org.openide.actions.MoveDownAction;
import org.openide.actions.MoveUpAction;
import org.openide.actions.NewAction;
import org.openide.actions.OpenAction;
import org.openide.actions.PasteAction;
import org.openide.actions.RenameAction;
import org.openide.actions.ReorderAction;
import org.openide.awt.AcceleratorBinding;
import org.openide.explorer.view.CheckableNode;
import org.openide.filesystems.FileObject;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Index;
import org.openide.nodes.Node;
import org.openide.nodes.Sheet;
import org.openide.util.NbBundle.Messages;
import org.openide.util.Utilities;
import org.openide.util.WeakListeners;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.NewType;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import com.sieuferd.project.navigator.SwingStringDisplayNameProvider;
import com.sieuferd.query.Queries;

/* TODO[pd]: Move away from the ability to listen to individual paths of a View. Just recalculate
             derived display data such as label badges and compare against the previous derived
             value before firing events from the Node. */
/* The constructor and the destroy() methods have been observed to be called from threads other than
the EDT, so I'm making sure to keep this class thread-safe. */
/** Node representing a view field. */
@ThreadSafe
public final class FieldNode extends AbstractNode {
  private static final Logger LOG = Logger.getLogger(FieldNode.class.getName());
  private static final boolean SHOW_APPLICABLE_PROPERTIES_ONLY = false;
  private static final boolean SHOW_RELATION_PROPERTIES_IN_PRIMITIVES = false;
  private static final boolean FORCE_NON_LEAF_AFFORDANCE_RELATIONS = false;
  private static final QueryLabelFormatter LABEL_BADGE_PROVIDER =
      QueryLabelFormatter.INSTANCE;
  private final FieldContext fieldContext;
  private final NodeSettings settings;
  private final ViewListener viewListenerRef;
  /** The Children object to be used whenever this node has at least one child node. When there are
  no actual child nodes, Children.LEAF is used instead. Can be null if this node can never have any
  children. */
  private final @Nullable FieldChildren nonLeafChildren;
  private final NewType[] newTypes;
  private final FieldNodeProperties customProperties;
  private final PropertyChangeListener customPropertiesListenerRef;

  public FieldNode(final FieldContext fieldContext, final NodeSettings settings) {
    this(fieldContext, settings, new InstanceContent());
  }

  @Messages("CTL_ReorderFields=Reorder Fields")
  private FieldNode(
      final FieldContext fieldContext, final NodeSettings settings, InstanceContent ic)
  {
    super(Children.LEAF, new AbstractLookup(ic));
    Preconditions.checkNotNull(fieldContext);
    Preconditions.checkNotNull(settings);
    Preconditions.checkNotNull(ic);
    this.settings = settings;
    this.fieldContext = fieldContext;
    this.nonLeafChildren =
        (fieldContext.isRelation() && settings.getIncludedPathTypes() != AppliesTo.NONE)
        ? new FieldChildren(fieldContext.getRelation(), settings) : null;
    this.newTypes = fieldContext.isRelation()
        ? new NewType[] { new CreateNodeNewType(true), new CreateNodeNewType(false) }
        : new NewType[] {};
    updateChildPaths();
    ic.add(fieldContext.getViewContext());
    ic.add(fieldContext);
    ic.add((SwingStringDisplayNameProvider) () -> FieldNode.this.getSwingStringDisplayName());
    /* Needed for the OpenAction. Don't include the entire lookup of the DataObject, as this would
    confuse RenameAction (e.g. asking the user to rename the underlying file instead of the root
    label). */
    Openable openable = fieldContext.getViewDataObject().getLookup().lookup(Openable.class);
    if (openable != null)
      ic.add(openable);
    // Needed for the Navigator.
    ic.add(fieldContext.getViewDataObject());
    ic.add(FieldNodePathFinder.INSTANCE);
    if (settings.isCheckable()) {
      ic.add(new CheckableNode() {
        @Override
        public boolean isCheckable() {
          return true;
        }

        @Override
        public boolean isCheckEnabled() {
          return fieldContext.exists() && !fieldContext.isRoot();
        }

        @Override
        public Boolean isSelected() {
          if (!fieldContext.exists())
            return null;
          boolean actualValue = fieldContext.getOrDefault(CProp.VISIBLE);
          if (!actualValue)
            return false;
          return Queries.isVisibleToRoot(fieldContext.getView(), fieldContext.getPath())
              ? true : null;
        }

        @Override
        public void setSelected(Boolean selected) {
          Preconditions.checkNotNull(selected);
          /* Override the behavior defined in o.n.swing.outline.Outline.checkAt for the
          "half-selected" case; in this case, turn selection on rather than off on first click. */
          if (isSelected() == null)
            selected = true;
          ViewUpdateCoalescer.INSTANCE.setProperty(fieldContext, CProp.VISIBLE, selected);
        }
      });
    }
    if (fieldContext.isRelation()) {
      final RelationPath contextPath = fieldContext.getRelation().getPath();
      ic.add(new Index.Support() {
        private FieldChildren allChildren;

        @Override
        public Node[] getNodes() {
          ThreadUtil.checkOnEDT();
          if (allChildren == null) {
            /* Always use the full list of child nodes for the Index, otherwise reorder() may be
            called with the wrong number of permutation elements. By returning the full list of
            child nodes here, the "Reorder Nodes..." dialog correctly shows all nodes rather than
            just the child nodes from the relation-only projects hierarchy. */
            if (nonLeafChildren != null &&
                settings.getIncludedPathTypes() == AppliesTo.ANY &&
                settings.getIncludeAffordanceNodes())
            {
              /* Move Up/Move Down seems to only work if I use the actual Children object from
              getChildren() here (presumably because the involved Node objects need to be
              identical). */
              allChildren = nonLeafChildren;
            } else {
              allChildren = new FieldChildren(fieldContext.getRelation(), settings);
              allChildren.setVisibleChildPaths(
                  Lists.newArrayList(fieldContext.getOrDefault(SProp.CHILD_PATHS)));
            }
          }
          // There is no recursion here, so doing getNodes(true) should be safe.
          return allChildren.getNodes(true);
        }

        @Override
        public int getNodesCount() {
          /* This method is not called very frequently, so to avoid any potential for bugs, just
          get the length from the actual list of nodes. */
          return getNodes().length;
        }

        @Override
        public void reorder(final int[] permuation) {
          /* TODO[pd]: Ignore the operation if the path set changed, or do a variation of the
                       operation that will always succeed. */
          if (permuation.length == 0)
            return;
          fieldContext.getView().mutate(Bundle.CTL_ReorderFields(), true, (MutableView view) -> {
            // See https://java.net/projects/nb-api-samples/sources/api-samples/content/versions/7.1/misc/DragDrop/CustomerViewer/src/org/customer/viewer/CustomerList.java?rev=529
            List<Path> oldList =
                ImmutableList.copyOf(view.getOrDefault(SProp.CHILD_PATHS, contextPath));
            final int N = oldList.size();
            Preconditions.checkArgument(permuation.length == N);
            final Path[] reordered = new Path[N];
            for (int i = 0; i < N; i++) {
              final int j = permuation[i];
              Preconditions.checkArgument(j >= 0 && j < N);
              if (reordered[j] != null) {
                throw new IllegalArgumentException(
                    "permutation may not include the same index twice");
              }
              reordered[j] = oldList.get(i);
            }
            /* Since N assignments have been made each to a new slot in the reordered array, there
            can be no null values in the array at this point. */
            view.set(SProp.CHILD_PATHS, contextPath, new PathList(ImmutableSet.copyOf(reordered)));
          });
        }
      });
    }
    // Initialize customProperties before viewListenerRef.
    if (settings.getCustomPropertiesProvider() == null) {
      customProperties = null;
      customPropertiesListenerRef = null;
    } else {
      customProperties =
          settings.getCustomPropertiesProvider().createFieldNodeProperties(fieldContext);
      customPropertiesListenerRef = (PropertyChangeEvent evt) -> {
        firePropertyChange(evt.getPropertyName(), evt.getOldValue(), evt.getNewValue());
      };
      customProperties.addPropertyChangeListener(
          WeakListeners.propertyChange(customPropertiesListenerRef, customProperties));
    }
    viewListenerRef = new ViewAdapter() {
      @Override
      public <T> void propertyChanged(ViewPropertyChangedEvent<T> evt) {
        final ViewProperty<T> property = evt.getProperty();
        if (evt.getPath().equals(fieldContext.getPath())) {
          if (!(property.getPropertySet() == SProp.INSTANCE) && customProperties == null) {
            firePropertyChange(
                property.getBeanPropertyName(), evt.getOldValue(), evt.getNewValue());
          }
          if (property.equals(SProp.CHILD_PATHS)) {
            updateChildPaths();
          } else if (property.equals(CProp.LABEL)) {
            FieldNode.super.setName(fieldContext.getOrDefault(CProp.LABEL));
          } else if (property.equals(CProp.AFFORDANCE) ||
              LABEL_BADGE_PROVIDER.getBadgeProperties().contains(property))
          {
            // Force getHtmlDisplayName() to be called again.
            fireDisplayNameChange(null, getDisplayName());
            // Affordance node status may affect the decision to use Children.LEAF or not.
            if (evt.getPath().isRelation() && property.equals(CProp.AFFORDANCE))
              updateChildPaths();
          }
          // Use fireIconChange() to update the CheckableNode for VISIBLE as well.
          if (property.equals(CProp.VISIBLE) || SieuferdIcons.ICON_PROPERTIES.contains(property))
            fireIconChange();
        } else if (property.equals(CProp.AFFORDANCE) &&
            fieldContext.getPath().isParentOf(evt.getPath()))
        {
          updateChildPaths();
        }
      }
    };
    /* Need to listen on all paths since the tristate checkable state of the node may depend on the
    visibility of ancestor paths. */
    fieldContext.getView().addViewListener(null,
        WeakListeners.create(ViewListener.class, viewListenerRef, fieldContext.getView()));
    // The label needs to be made the actual name of the node for RenameAction to work properly.
    super.setName(fieldContext.getOrDefault(CProp.LABEL));
  }

  private List<Path> getVisibleChildPaths() {
    if (!fieldContext.isRelation() || !fieldContext.exists())
      return ImmutableList.of();
    final List<Path> ret = Lists.newArrayList();
    for (Path childPath : fieldContext.getOrDefault(SProp.CHILD_PATHS)) {
      if (settings.getIncludedPathTypes().appliesTo(childPath) &&
          (settings.getIncludeAffordanceNodes() ||
          !Boolean.TRUE.equals(fieldContext.getView().getOrNull(CProp.AFFORDANCE, childPath))))
      {
        ret.add(childPath);
      }
    }
    return Collections.unmodifiableList(ret);
  }

  private void updateChildPaths() {
    final boolean leaf;
    if (nonLeafChildren == null) {
      leaf = true;
    } else if (FORCE_NON_LEAF_AFFORDANCE_RELATIONS &&
        settings.getIncludedPathTypes() == AppliesTo.ANY &&
        fieldContext.isRelation() && Boolean.TRUE.equals(fieldContext.getOrNull(CProp.AFFORDANCE)))
    {
      leaf = false;
    } else {
      final List<Path> visibleChildPaths = getVisibleChildPaths();
      leaf = visibleChildPaths.isEmpty();
      nonLeafChildren.setVisibleChildPaths(visibleChildPaths);
    }
    if (leaf != (getChildren() == Children.LEAF))
      setChildren(leaf ? Children.LEAF : nonLeafChildren);
  }

  @Override
  public boolean canCut() {
    /* Returning true from canCut() and canCopy() is required for drag-and-drop reordering to work.
    Still, drag-and-drop reordering within a single level of children will work without implementing
    getDropType()/createPasteTypes()/clipboardCut()/clipboardCopy(). */
    return !settings.getActionsDisabled() && fieldContext.exists();
  }

  @Override
  public boolean canCopy() {
    // This is the default from the superclass, but be explicit.
    return !settings.getActionsDisabled() && fieldContext.exists();
  }

  @Override
  public boolean canRename() {
    return !settings.getActionsDisabled() && fieldContext.exists();
  }

  @Override
  public boolean canDestroy() {
    return !settings.getActionsDisabled() && fieldContext.exists();
  }

  @Override
  public void destroy() throws IOException {
    // Don't call the superclass implementation (see its Javadoc).

    ViewUpdateCoalescer.INSTANCE.removePath(fieldContext);

    /* Don't call fireNodeDestroyed() here; instead, wait for the node to be removed from its
    parent once the underlying View model has been updated. This could potentially also be useful if
    other parts of the system would like to do things like vetoing the deletion of the node. See
    https://netbeans.org/bugzilla/show_bug.cgi?id=32250 for a discussion of fireNodeDestroyed(). */
  }

  @Override
  public void setName(String newName) {
    // Don't call the superclass method; it will be called from the ViewListener.
    ViewUpdateCoalescer.INSTANCE.setProperty(fieldContext, CProp.LABEL, newName);
  }

  private SwingString getSwingStringDisplayName() {
    if (!fieldContext.exists() ||
        Boolean.TRUE.equals(fieldContext.getOrNull(CProp.AFFORDANCE)))
    {
      return SwingString.styled(
          "<font color=\"" + SwingString.htmlColor("controlShadow")+ "\">" +
          SwingString.toSwingHTML(getDisplayName()) +
          "</font>");
    } else {
      if (settings.isLabelBadgeVisible()) {
        return LABEL_BADGE_PROVIDER
            .getLabelSwingString(fieldContext.getView(), fieldContext.getPath());
      } else {
        return SwingString.plain(getDisplayName());
      }
    }
  }

  @Override
  public Image getIcon(int type) {
    if (fieldContext.isRoot() &&
        (settings.getIncludedPathTypes() == AppliesTo.ROOT_ONLY ||
        settings.getIncludedPathTypes() == AppliesTo.RELATION_ONLY))
    {
      return SieuferdIcons.getViewIcon(fieldContext.getView());
    } else {
      return SieuferdIcons.getPathIcon(fieldContext.getView(), fieldContext.getPath());
    }
  }

  @Override
  public Image getOpenedIcon(int type) {
    return getIcon(type);
  }

  @Override
  public Action getPreferredAction() {
    if (settings.getActionsDisabled()) {
      return null;
    } else {
      return fieldContext.isRoot() ?
          SystemAction.get(OpenAction.class) : SystemAction.get(RenameAction.class);
    }
  }

  @Override
  @SuppressWarnings("DeadBranch")
  public Action[] getActions(boolean context) {
    if (settings.getActionsDisabled())
      return new Action[] {};
    if (context)
      return new Action[] {};
    if (settings.isUsingFullContextMenu()) {
      List<Action> ret = Lists.newArrayList();
      for (Entry<FileObject,Object> entry :
          ActionUtil.getRegisteredObjects("Sieuferd/FieldContextMenu").entrySet())
      {
        if (entry.getValue() instanceof Action) {
          Action action = (Action) entry.getValue();
          AcceleratorBinding.setAccelerator(action, entry.getKey());
          ret.add(action);
        } else if (entry.getValue() == null) {
          ret.add(null);
        }
      }
      return ret.toArray(new Action[ret.size()]);
    }
    List<Action> ret = Lists.newArrayList();
    // Note: Actions are added in the order they should be displayed to the user.
    if (fieldContext.isRoot()) {
      ret.add(SystemAction.get(OpenAction.class));
      ret.add(null);
    }
    if (fieldContext.isRelation()) {
      /* Put New and Reorder in the same group since they both relate to child nodes. Reorder and
      MoveUp/MoveDown are intentionally in different groups since only the former relates to child
      nodes. */
      ret.add(SystemAction.get(NewAction.class));
      ret.addAll(Utilities.actionsForPath("Actions/Sieuferd/RelationNodeContext"));
      ret.add(SystemAction.get(ReorderAction.class));
      ret.add(null);
    }
    // TODO[pd]: Consider implementing cut/copy/paste.
    if (false) {
      ret.add(SystemAction.get(CutAction.class));
      ret.add(SystemAction.get(CopyAction.class));
      if (fieldContext.isRelation())
        ret.add(SystemAction.get(PasteAction.class));
      ret.add(null);
    }
    ret.add(SystemAction.get(DeleteAction.class));
    ret.add(SystemAction.get(RenameAction.class));
    ret.add(null);
    if (getParentNode() != null && settings.getIncludedPathTypes() == AppliesTo.ANY) {
      ret.add(SystemAction.get(MoveUpAction.class));
      ret.add(SystemAction.get(MoveDownAction.class));
    }
    return ret.toArray(new Action[ret.size()]);
  }

  @Override
  protected Sheet createSheet() {
    if (customProperties != null)
      return customProperties.createSheet();
    if (!fieldContext.exists())
      return Sheet.createDefault();
    final Sheet ret = new Sheet();
    final ViewDefaults viewDefaults =
        fieldContext.getViewContext().getProject().getViewSet().getDefaults();
    final Multimap<ViewPropertySet,ViewProperty<?>> propertiesBySet =
        viewDefaults.getSortedPropertiesBySet();
    for (ViewPropertySet viewPropertySet : propertiesBySet.keySet()) {
      final Sheet.Set ps = new Sheet.Set();
      ps.setName(viewPropertySet.getName());
      ps.setDisplayName(viewPropertySet.getDisplayName());
      ps.setShortDescription(viewPropertySet.getShortDescription());
      ret.put(ps);
      for (final ViewProperty<?> prop : propertiesBySet.get(viewPropertySet)) {
        FieldContext useFieldContext = null;
        if (prop.appliesTo(fieldContext.getPath())) {
          useFieldContext = fieldContext;
        } else {
          if (SHOW_APPLICABLE_PROPERTIES_ONLY)
            continue;
          if (SHOW_RELATION_PROPERTIES_IN_PRIMITIVES && !fieldContext.isRelation()) {
            final RelationPath parentPath = fieldContext.getPath().getParent();
            if (prop.appliesTo(parentPath))
              useFieldContext = FieldContext.create(fieldContext.getViewContext(), parentPath);
          }
        }
        ps.put(ViewPropertySupport.create(
            useFieldContext, prop, viewDefaults, settings.getActionsDisabled()));
      }
    }
    return ret;
  }

  @Override
  public NewType[] getNewTypes() {
    return newTypes;
  }

  private final class CreateNodeNewType extends NewType {
    private final boolean relation;

    CreateNodeNewType(boolean relation) {
      this.relation = relation;
    }

    @Messages({
      "CTL_AddField=Add Field",
      /* TODO[cc]: Figure out the right message name prefix to use here (and others starting with
                   DEF_). */
      "DEF_NewField=New Field"
    })
    @Override
    public void create() throws IOException {
      fieldContext.getView().mutate(Bundle.CTL_AddField(), true, (MutableView view) -> {
        Path newPath = view.addPath(fieldContext.getPath().getRelation(), relation);
        view.set(CProp.LABEL, newPath, Bundle.DEF_NewField());
      });
    }

    @Messages({
      "CTL_NewRelationChildField=Relation Child Field",
      "CTL_NewPrimitiveChildField=Primitive Child Field"
    })
    @Override
    public String getName() {
      return relation ? Bundle.CTL_NewRelationChildField() : Bundle.CTL_NewPrimitiveChildField();
    }
  }
}
