// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.nodes;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.sieuferd.project.FieldContext;
import java.util.List;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import org.netbeans.spi.project.ui.PathFinder;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;

/**
 * Provides lookup of {@code FieldNode}s in a way that avoids traversing irrelevant paths. This
 * avoids potential infinite recursion problems. This PathFinder should also be efficient enough to
 * use on the EDT.
 */
@ThreadSafe
public final class FieldNodePathFinder implements PathFinder {
  public static final PathFinder INSTANCE = new FieldNodePathFinder();

  private FieldNodePathFinder() { }

  @Override
  public @Nullable Node findPath(Node root, Object target) {
    if (matchesDataOrFileObject(root, target)) {
      return root;
    } else if (target instanceof FieldContext) {
      return findFieldContext(root, (FieldContext) target);
    } else {
      return null;
    }
  }

  // TODO[cc]: Consider moving this elsewhere.
  public static boolean matchesDataOrFileObject(Node root, Object target) {
    if (target instanceof DataObject || target instanceof FileObject) {
      final DataObject dataObject = root.getLookup().lookup(DataObject.class);
      if (dataObject != null) {
        FileObject nodeFileObject = dataObject.getPrimaryFile();
        FileObject targetFileObject = (target instanceof DataObject)
            ? ((DataObject) target).getPrimaryFile() : (FileObject) target;
        if (nodeFileObject.equals(targetFileObject))
          return true;
      }
    }
    return false;
  }

  private static @Nullable Node findFieldContext(Node root, FieldContext targetFieldContext) {
    final FieldContext candidateFieldContext = root.getLookup().lookup(FieldContext.class);
    if (candidateFieldContext == null)
      return null; // Don't bother examining children in this case; probably the wrong type of node.
    if (candidateFieldContext.equals(targetFieldContext))
      return root;

    // Only traverse the correct view.
    if (!candidateFieldContext.getViewContext().equals(targetFieldContext.getViewContext()))
      return null;
    // Only recurse down the relevant path.
    if (!candidateFieldContext.getPath().isStrictAncestorOf(targetFieldContext.getPath()))
      return null;
    for (Node childNode : root.getChildren().getNodes(true)) {
      // Every FieldNode should have a PathFinder; don't recurse otherwise.
      final PathFinder childPathFinder = childNode.getLookup().lookup(PathFinder.class);
      if (childPathFinder != null) {
        Node ret = childPathFinder.findPath(childNode, targetFieldContext);
        if (ret != null)
          return ret;
      }
    }
    return null;
  }

  /**
   * Find the node for the specified {@code FieldContext}. May be called on any thread, including
   * the EDT.
   */
  public static @Nullable Node findNode(Node root, FieldContext fieldContext) {
    Preconditions.checkNotNull(root);
    Preconditions.checkNotNull(fieldContext);
    // Only accept FieldNodePathFinder here, as we know that it's safe to use on the EDT.
    final PathFinder pathFinder = root.getLookup().lookup(FieldNodePathFinder.class);
    return pathFinder == null ? null : pathFinder.findPath(root, fieldContext);
  }

  /**
   * Find the node(s) for the specified {@code FieldContext}. May be called on any thread, including
   * the EDT.
   */
  public static Node[] findNodes(Node root, Iterable<FieldContext> fieldContexts) {
    Preconditions.checkNotNull(root);
    Preconditions.checkNotNull(fieldContexts);
    // See above.
    final PathFinder pathFinder = root.getLookup().lookup(FieldNodePathFinder.class);
    if (pathFinder == null)
      return new Node[] {};
    List<Node> ret = Lists.newArrayList();
    for (FieldContext fieldContext : fieldContexts) {
      Node node = pathFinder.findPath(root, fieldContext);
      if (node != null)
        ret.add(node);
    }
    return ret.toArray(new Node[ret.size()]);
  }
}
