// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.nodes;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Views;
import com.sieuferd.layout.YesNoAuto;
import com.sieuferd.layout.LAProp;
import java.awt.Image;
import javax.swing.Action;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.SieuferdProjectInformation;
import com.sieuferd.viewset.SieuferdProjectConfiguration;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import org.netbeans.spi.project.ui.support.CommonProjectActions;
import org.openide.actions.NewAction;
import org.openide.actions.RenameAction;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.util.Mutex;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.WeakListeners;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.NewType;
import org.openide.util.lookup.Lookups;

// See https://platform.netbeans.org/tutorials/nbm-povray-3.html
public class SieuferdProjectNode extends AbstractNode {
  private static final RequestProcessor RP_ADD_VIEW = new RequestProcessor("Add View", 1);
  private final SieuferdProject project;
  private final PropertyChangeListener configurationListenerReference;
  private final NodeSettings settings;
  private final NewType[] newTypes = new NewType[] {
    new ViewNewType()
  };

  public SieuferdProjectNode(SieuferdProject project, NodeSettings settings) {
    super(Children.create(new SieuferdProjectChildFactory(project, settings), false),
        Lookups.singleton(project));
    Preconditions.checkNotNull(project);
    Preconditions.checkNotNull(settings);
    this.settings = settings;
    this.project = project;
    final SieuferdProjectConfiguration configuration = project.getConfiguration();
    super.setName(configuration.getProjectName());
    configurationListenerReference = (final PropertyChangeEvent evt) -> {
      Mutex.EVENT.writeAccess(() -> {
        if (evt.getPropertyName().equals(SieuferdProjectConfiguration.PROP_PROJECT_NAME))
          SieuferdProjectNode.super.setName(configuration.getProjectName());
      });
    };
    configuration.addPropertyChangeListener(
        WeakListeners.propertyChange(configurationListenerReference, configuration));
  }

  @Override
  public Image getIcon(int type) {
    return SieuferdProjectInformation.getIconImageStatic();
  }

  @Override
  public Image getOpenedIcon(int type) {
    return getIcon(type);
  }

  @Override
  public void setName(String name) {
    // Super implementation is called from configurationListenerReference.
    project.getConfiguration().setProjectName(name);
  }

  @Override
  public NewType[] getNewTypes() {
    return newTypes;
  }

  @Override
  public boolean canRename() {
    return !settings.getActionsDisabled();
  }

  @Override
  public Action[] getActions(boolean context) {
    if (settings.getActionsDisabled())
      return new Action[] {};
    // Don't include the actions for the folder itself, as they do not make much sense here.
    return new Action[] {
      SystemAction.get(NewAction.class),
      null,
      CommonProjectActions.closeProjectAction(),

      /* Use the system rename action rather than CommonProjectActions.renameProjectAction() to
      get the standard node renaming behavior rather than the project renaming behavior, as the
      latter is designed for actually moving the project directory rather than simply setting the
      display name. */
      SystemAction.get(RenameAction.class),

      // CommonProjectActions.moveProjectAction(),
      // CommonProjectActions.copyProjectAction(),
      // CommonProjectActions.deleteProjectAction(),
      CommonProjectActions.customizeProjectAction()
    };
  }

  @NbBundle.Messages("DEF_NewView=New Perspective")
  private final class ViewNewType extends NewType {
    @Override
    public void create() throws IOException {
      // This method is known to be run from the EDT, despite being allowed to throw an IOException.
      RP_ADD_VIEW.execute(() -> {
        MutableView view = Views.newMutable();
        view.set(CProp.LABEL, Path.ROOT, Bundle.DEF_NewView());
        view.set(LAProp.USE_TABLE, Path.ROOT, YesNoAuto.YES);
        project.getViewSet().createAndAddView(view);
      });
    }

    @NbBundle.Messages("CTL_NewView=Perspective")
    @Override
    public String getName() {
      return Bundle.CTL_NewView();
    }
  }
}
