// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.nodes;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ObservableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewListener;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewPropertyChangedEvent;
import com.sieuferd.datamodel.ViewTransactionCompletedEvent;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.FieldContext.RelationFieldContext;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.query.QProp;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.viewset.GatedView;
import com.sieuferd.viewset.SieuferdProjectConfiguration;
import com.sieuferd.viewset.ViewSet;
import com.sieuferd.viewset.ViewSetEvent;
import com.sieuferd.viewset.ViewSetListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;
import org.openide.nodes.ChildFactory;
import org.openide.nodes.Node;

public class ViewFolderChildFactory extends ChildFactory.Detachable<RelationFieldContext> {
  private static final Logger LOG = Logger.getLogger(ViewFolderChildFactory.class.getName());
  private static final ImmutableSet<ViewProperty<?>> REFRESHABLE_VIEW_PROPERTIES =
      ImmutableSet.<ViewProperty<?>>of(CProp.LABEL, QProp.PRIMARY);
  private final SieuferdProject project;
  private final ViewSet viewSet;
  private final NodeSettings nodeSettings;
  private final boolean primary;
  /** Views that have had viewSetListener attached to them. Synchronized on itself. */
  private final Set<ObservableView> viewsWithListeners = Sets.newLinkedHashSet();
  private final ViewSetListener viewSetListener = (ViewSetEvent evt) -> {
    /** Postpone the refresh to the end of the event queue even if we are currently on the EDT.
    Otherwise, org.openide.actions.PasteAction.NodeSelector.select() will hang on calling
    Children.getNodes(true) when a new view is added for as long as it takes to run createKeys(),
    defeating the purpose of making this ChildFactory asynchronous. */
    SwingUtilities.invokeLater(() -> {
      refresh(false);
    });
  };
  private final ViewListener viewListener = new ViewListener() {
    private final AtomicBoolean refreshRequired = new AtomicBoolean();

    @Override
    public <T> void propertyChanged(ViewPropertyChangedEvent<T> evt) {
      if (REFRESHABLE_VIEW_PROPERTIES.contains(evt.getProperty()))
        refreshRequired.set(true);
    }

    @Override
    public void transactionCompleted(ViewTransactionCompletedEvent evt) {
      ThreadUtil.checkOnEDT();
      if (refreshRequired.getAndSet(false)) {
        // Refresh in case the label change would change the sorted order of the children.
        refresh(false);
      }
    }
  };

  private final PropertyChangeListener configurationListener = (PropertyChangeEvent evt) -> {
    // TODO[pd]: Avoid triggering one refresh for each of the two properties.
    if (evt.getPropertyName() == null ||
        SieuferdProjectConfiguration.PRIMARY_VIEW_SET_PROPERTIES.contains(evt.getPropertyName()))
    {
      //refreshPrimaryViews();
    }
  };

  private void refreshPrimaryViews() {
    ViewUpdater.refreshPrimaryViews(project);
  }

  public ViewFolderChildFactory(
      SieuferdProject project, NodeSettings nodeSettings, boolean primary)
  {
    Preconditions.checkNotNull(nodeSettings);
    this.viewSet = project.getViewSet();
    this.project = project;
    this.primary = primary;
    this.nodeSettings = nodeSettings;
  }

  @Override
  protected void addNotify() {
    viewSet.addViewSetListener(viewSetListener);
    if (primary) {
      project.getConfiguration().addPropertyChangeListener(configurationListener);
      //refreshPrimaryViews();
    }
  }

  @Override
  protected void removeNotify() {
    if (primary)
      project.getConfiguration().removePropertyChangeListener(configurationListener);
    viewSet.removeViewSetListener(viewSetListener);
    removeAllViewListeners();
  }

  @Override
  protected Node createNodeForKey(RelationFieldContext key) {
    return new FieldNode(key, nodeSettings);
  }

  private void removeAllViewListeners() {
    synchronized (viewsWithListeners) {
      for (ObservableView view : viewsWithListeners)
        view.removeViewListener(viewListener);
      viewsWithListeners.clear();
    }
  }

  private void addViewListener(ObservableView view) {
    synchronized (viewsWithListeners) {
      if (viewsWithListeners.add(view))
        view.addViewListener(Path.ROOT, viewListener);
    }
  }

  @Override
  protected boolean createKeys(List<RelationFieldContext> toPopulate) {
    removeAllViewListeners();
    /* Don't operate directly on toPopulate; it causes a distracting flicker when expanded nodes are
    reordered. */
    final List<RelationFieldContext> ret = Lists.newArrayList();
    for (GatedView view : viewSet.getViews()) {
      addViewListener(view);
      if (Thread.interrupted())
        return true;
      try {
        if (view.get(QProp.PRIMARY, Path.ROOT) == primary) {
          ret.add(new RelationFieldContext(ViewContext.create(project, view), Path.ROOT));
        }
      } catch (IOException e) {
        LOG.log(Level.WARNING, "Failed to create FieldContext for View", e);
        // Continue despite the error.
      }
    }
    /* An earlier implementation added the fields directly to a TreeSet; however, this inadvertently
    collapsed elements with the same name. */
    /* I considered allowing the view order to be specified explicitly, but decided against it.
    Then I would have to provide an explicit option or action to alphabetize. Besides, file
    systems don't do this either. */
    Collections.sort(ret, (FieldContext o1, FieldContext o2) ->
        o1.get(CProp.LABEL).compareToIgnoreCase(o2.get(CProp.LABEL))
    );
    toPopulate.addAll(ret);
    return true;
  }
}
