// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.nodes;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.propertyeditors.GenericViewPropertyEditor;
import com.sieuferd.project.propertyeditors.ViewPropertyEditor;
import com.sieuferd.uiutil.SwingString;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorSupport;
import java.util.Objects;
import javax.annotation.Nullable;
import org.openide.nodes.PropertySupport;

final class ViewPropertySupport<T> extends PropertySupport<T> {
  private final ViewProperty<T> prop;
  private final T defaultValue;
  /** Null for a non-applicable property */
  private final @Nullable FieldContext fieldContext;

  /** @param fieldContext can be null for a non-applicable property */
  ViewPropertySupport(@Nullable FieldContext fieldContext, ViewProperty<T> prop, T defaultValue,
      boolean forceReadOnly)
  {
    super(prop.getBeanPropertyName(),
        prop.getType().getValueClass(), prop.getDisplayName(), prop.getShortDescription(),
        true, fieldContext != null && !forceReadOnly && !prop.isUserReadOnly());
    Preconditions.checkNotNull(prop);
    Preconditions.checkNotNull(defaultValue);
    this.prop = prop;
    this.defaultValue = defaultValue;
    this.fieldContext = fieldContext;
    if (fieldContext != null) {
      final Path path = fieldContext.getPath();
      setValue(ViewPropertyEditor.PROP_VIEW_CONTEXT, fieldContext.getViewContext());
      setValue(ViewPropertyEditor.PROP_CONTEXT_PATH, Views.getContextPath(path));
      setValue(ViewPropertyEditor.PROP_PROPERTY_TYPE, prop.getType());
    }
  }

  /** @param fieldContext can be null for a non-applicable property */
  public static <T> ViewPropertySupport<T> create(@Nullable FieldContext fieldContext,
      ViewProperty<T> prop, ViewDefaults defaults, boolean forceReadOnly)
  {
    return new ViewPropertySupport<>(fieldContext, prop, defaults.get(prop), forceReadOnly);
  }

  @Override
  public PropertyEditor getPropertyEditor() {
    if (fieldContext != null) {
      PropertyEditor ret = super.getPropertyEditor();
      if (ret != null || Enum.class.isAssignableFrom(prop.getType().getValueClass())) {
        return ret;
      } else {
        // TODO[pd]: Should I be caching this?
        return new GenericViewPropertyEditor<>();
      }
    } else {
      // Create a blank property editor.
      return new PropertyEditorSupport() {
        @Override
        public boolean isPaintable() {
          /* Set as paintable, but don't actually do any painting in paintValue. An alternative
          would have been to define getAsText to return an empty string, but when I tried that,
          boolean properties would still show up with checkboxes in both the Navigator's
          OutlineView and in the property sheet. */
          return true;
        }
      };
    }
  }

  @Override
  public boolean supportsDefaultValue() {
    return true;
  }

  @Override
  public boolean isDefaultValue() {
    return Objects.equals(getValue(), defaultValue);
  }

  @Override
  public void restoreDefaultValue() {
    setValue(defaultValue);
  }

  @Override
  public T getValue() {
    /* Using the default value rather than null or a non-readable value makes the property sheet UI
    work better during multiple selection that might include multiple node types. */
    if (fieldContext == null)
      return defaultValue;
    T ret = fieldContext.getOrNull(prop);
    return (ret == null) ? defaultValue : ret;
  }

  @Override
  public void setValue(T value) throws IllegalArgumentException {
    if (fieldContext == null)
      throw new IllegalArgumentException();
    /* The built-in property editor for an early access version of jdk1.7.0_40 was observed to call
    setValue() with null when the cancel button was clicked--although not quite consistently. Rely
    on MutableSet throwing an IllegalArgumentException in case a null value ever appears. */
    ViewUpdateCoalescer.INSTANCE.setProperty(fieldContext, prop, value);
  }

  @Override
  public String getHtmlDisplayName() {
    if (fieldContext == null) {
      return "<font color=\"!controlShadow\">" +
          SwingString.toSwingHTML(getDisplayName()) + "</font>";
    } else {
      return (isDefaultValue() || !canWrite()) ? null :
          ("<b>" + SwingString.toSwingHTML(getDisplayName()) + "</b>");
    }
  }
}