// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.nodes;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSetMultimap;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.actions.DeleteFieldAction;
import com.sieuferd.project.actions.HideAction;
import com.sieuferd.project.actions.UnhideAction;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.project.formulabar.types.FormulaType;
import com.sieuferd.project.formulabar.types.LabelType;
import com.sieuferd.query.QProp;
import com.sieuferd.util.MapUtil;
import java.util.Set;
import javax.annotation.concurrent.ThreadSafe;
import javax.swing.SwingUtilities;
import org.openide.util.NbBundle.Messages;
import org.openide.util.RequestProcessor;

@Messages({
  "# {0} - Display name of the edited property",
  "CTL_EditProperty=Edit {0}",
  "CTL_EditProperties=Edit Properites",
  "# {0} - Display name of the enabled boolean property",
  "CTL_EnableBooleanProperty=Enable {0}",
  "# {0} - Display name of the disabled boolean property",
  "CTL_DisableBooleanProperty=Disable {0}"
})
@ThreadSafe
final class ViewUpdateCoalescer {
  public static final ViewUpdateCoalescer INSTANCE = new ViewUpdateCoalescer();
  private static final RequestProcessor RP = new RequestProcessor("Remove View", 1);
  /** Synchronized on itself. */
  private final Multimap<ViewContext,Path> pendingDeleteOperations = LinkedHashMultimap.create();
  /** Synchronized on itself. */
  private final Multimap<ViewContext,SetPropertyOperation<?>> pendingSetPropertyOperations =
      LinkedHashMultimap.create();

  private ViewUpdateCoalescer() { }

  private static interface OperationType<T> {
    void performOperations(ViewContext viewContext, Iterable<T> operations);
  }

  private <T> void enqueueCoalescibleOperation(final ViewContext viewContext,
      final Multimap<ViewContext,T> toPendingOperationsMap, final T operation,
      final OperationType<T> operationType)
  {
    final ImmutableMultimap<ViewContext,T> previousPendingOperations;
    synchronized (toPendingOperationsMap) {
      toPendingOperationsMap.put(viewContext, operation);
      previousPendingOperations = ImmutableSetMultimap.copyOf(toPendingOperationsMap);
    }
    // Postpone operation until the end of the event queue.
    SwingUtilities.invokeLater(() -> {
      final ImmutableMultimap<ViewContext,T> pendingOperationsSnapshot;
      synchronized (toPendingOperationsMap) {
        /* If more operation have accumulated since this update was scheduled, wait for a later
        update, otherwise process all coalesced operations now. */
        if (toPendingOperationsMap.equals(previousPendingOperations)) {
          pendingOperationsSnapshot = ImmutableSetMultimap.copyOf(toPendingOperationsMap);
          toPendingOperationsMap.clear();
        } else {
          return;
        }
      }
      for (ViewContext viewContext1 : pendingOperationsSnapshot.keySet())
        operationType.performOperations(viewContext1, pendingOperationsSnapshot.get(viewContext1));
    });
  }

  public void removePath(FieldContext fieldContext) {
    enqueueCoalescibleOperation(fieldContext.getViewContext(), pendingDeleteOperations,
        fieldContext.getPath(), this::removePaths);
  }

  public <T> void setProperty(
      FieldContext fieldContext, final ViewProperty<T> property, final T value)
  {
    enqueueCoalescibleOperation(fieldContext.getViewContext(), pendingSetPropertyOperations,
        new SetPropertyOperation<>(fieldContext.getPath(), property, value), this::setProperties);
  }

  private void removePaths(final ViewContext viewContext, final Iterable<Path> paths) {
    final ImmutableSet<Path> usePaths = ImmutableSet.copyOf(paths);
    if (usePaths.contains(Path.ROOT)) {
      RP.post(() -> {
        viewContext.getProject().getViewSet().removeView(viewContext.getView());
      });
    } else {
      viewContext.getView().mutate(DeleteFieldAction.getDisplayName(), true, (MutableView view) -> {
        for (Path path : usePaths)
          view.removePath(path);
      });
    }
  }

  private void setProperties(
      final ViewContext viewContext, final Iterable<SetPropertyOperation<?>> operations)
  {
    final Object value;
    final ViewProperty<?> property;
    {
      final Set<Object> values = Sets.newLinkedHashSet();
      final Set<ViewProperty<?>> properties = Sets.newLinkedHashSet();
      for (SetPropertyOperation<?> operation : operations) {
        values.add(operation.value);
        properties.add(operation.property);
      }
      value = MapUtil.getLoneElementOrNull(values);
      property = MapUtil.getLoneElementOrNull(properties);
    }
    final String operationPresentationName;
    if (property == null) {
      operationPresentationName = Bundle.CTL_EditProperties();
    } else if (CProp.VISIBLE.equals(property) && value != null) {
      operationPresentationName = Boolean.TRUE.equals(value) ?
          UnhideAction.getDisplayName() : HideAction.getDisplayName();
    } else if (CProp.LABEL.equals(property)) {
      operationPresentationName = LabelType.getActionDisplayName();
    } else if (QProp.FORMULA_EXPRESSION.equals(property)) {
      operationPresentationName = FormulaType.getActionDisplayName();
    } else if (value != null && property.getType().getValueClass().equals(Boolean.class)) {
      operationPresentationName = Boolean.TRUE.equals(value)
          ? Bundle.CTL_EnableBooleanProperty(property.getDisplayName())
          : Bundle.CTL_DisableBooleanProperty(property.getDisplayName());
    } else {
      operationPresentationName = Bundle.CTL_EditProperty(property.getDisplayName());
    }
    viewContext.getView().mutate(operationPresentationName, true, (MutableView view) -> {
      for (SetPropertyOperation<?> operation : operations)
        operation.apply(view);
    });
  }

  private static final class SetPropertyOperation<T> {
    public final Path path;
    public final ViewProperty<T> property;
    public final T value;

    SetPropertyOperation(Path path, ViewProperty<T> property, T value) {
      Preconditions.checkArgument(property.appliesTo(path));
      Preconditions.checkNotNull(path);
      Preconditions.checkNotNull(property);
      Preconditions.checkNotNull(value);
      this.path = path;
      this.property = property;
      this.value = value;
    }

    public void apply(MutableView view) {
      if (property.equals(CProp.VISIBLE) && Boolean.TRUE.equals(value)) {
        ViewUpdater.userUnhide(view, path);
      } else {
        ViewUpdater.userSetProperty(view, property, path, value);
      }
    }
  }
}
