// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.propertyeditors;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.sieuferd.query.SQLTable;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.ViewContext;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.query.InstanceDefinition;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import java.util.List;
import java.util.WeakHashMap;
import org.openide.explorer.propertysheet.PropertyEnv;
import org.openide.nodes.PropertyEditorRegistration;

@PropertyEditorRegistration(targetType = InstanceDefinition.class)
public class InstanceDefinitionEditor extends ViewPropertyEditor<InstanceDefinition> {
  // TODO[pd]: File an RFE for an API to set a custom ComboBoxModel on a property editor.
  /** An empty or spaces-only string will always be interpreted as 'no instantiated table'. In the
  dropdown this option is encoded as a single space to serve as a hint to setAsText() as to whether
  the value was entered manually or selected from the dropdown. Picking the empty value from the
  dropdown may trigger a database connection dialog in the case that there is no existing live
  connection and there is no previous list of tables in the TableListCache. This is a bit of a hack,
  but there is currently no way to provide a more specialized ComboBoxModel for the built-in
  dropdown editor. And according to Tim Boudreau, "using a vanilla ComboBox in the property sheet is
  a recipe for all sorts of trouble", so I have to make do with getTags(). */
  private static final String EMPTY_TAG = " ";
  private static final WeakHashMap<SieuferdProject,TableListCache> tableListCaches =
      new WeakHashMap<>();

  private static TableListCache getTableListCache(SieuferdProject project) {
    ThreadUtil.checkOnEDT();
    TableListCache ret = tableListCaches.get(project);
    if (ret == null) {
      ret = new TableListCache();
      ret.postRefresh(project, false);
      tableListCaches.put(project, ret);
    }
    return ret;
  }

  @Override
  public void attachEnv(PropertyEnv env) {
    ThreadUtil.checkOnEDT();
    super.attachEnv(env);
    // Don't update table list on every attach, but make sure it has been done once for the project.
    ViewContext viewContext = getViewContext();
    if (viewContext != null)
      getTableListCache(viewContext.getProject());
  }

  @Override
  public String valueToText(InstanceDefinition value) {
    if (value == null)
      return "";
    TableIdentifier table = value.getTable();
    if (table == null || !(table instanceof SQLTable))
      return "";
    return ((SQLTable) table).generateStringRepresentation();
  }

  @Override
  public InstanceDefinition textToValue(String text) throws IllegalArgumentException {
    final ViewContext viewContext = getViewContext();
    final boolean refreshOnly;
    if (viewContext == null) {
      refreshOnly = false;
    } else {
      SieuferdProject project = viewContext.getProject();
      TableListCache cache = getTableListCache(project);
      refreshOnly = text.equals(EMPTY_TAG) && cache.getTables(project) == null;
      /* Use the action of setting the value as a hint to update the dropdown. It's inelegant, but
      it's the best I can do since by the time getTags() is called, I have to return a result
      immediately. */
      cache.postRefresh(project, refreshOnly);
    }
    if (refreshOnly) {
      return null;
    } else {
      final InstanceDefinition ret;
      if (text.trim().isEmpty()) {
        ret = new InstanceDefinition(null);
      } else {
        SQLTable table = SQLTable.createFromStringRepresentation(text);
        if (table == null) {
          throw new IllegalArgumentException(
              "Expected table format catalogName/schemaName/tableName");
        }
        ret = new InstanceDefinition(table);
      }
      return ret;
    }
  }

  @Override
  public String[] getTags() {
    ThreadUtil.checkOnEDT();
    /* Don't do a TableListCache.postRefresh() here, as getTags() may be called many times in rapid
    succession even for a single dropdown action. */
    List<String> tags = Lists.newArrayList();
    ViewContext viewContext = getViewContext();
    List<SQLTable> tables = viewContext == null ? ImmutableList.of()
        : getTableListCache(viewContext.getProject()).getTables(viewContext.getProject());
    tags.add(EMPTY_TAG);
    if (tables != null) {
      for (SQLTable table : tables)
        tags.add(table.generateStringRepresentation());
    }
    return tags.toArray(new String[tags.size()]);
  }
}
