// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.propertyeditors;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.query.SQLTable;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.project.derivatives.ViewUpdater;
import com.sieuferd.project.derivatives.ViewUpdater.TableListing;
import com.sieuferd.uiutil.ThreadUtil;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import javax.swing.SwingUtilities;
import org.openide.util.RequestProcessor;

/** All methods must be called from the EDT only. All methods are non-blocking. This class avoids
holding a strong reference to its associated SieuferdProject by requiring the project to be supplied
in method arguments instead. */
class TableListCache {
  private static final Logger LOG = Logger.getLogger(TableListCache.class.getName());
  private static final RequestProcessor RP =
      new RequestProcessor("InstanceDefinitionEditor table list loader");
  private @Nullable CachedContent content;
  private final AtomicBoolean isLoading = new AtomicBoolean(false);

  TableListCache() {
    ThreadUtil.checkOnEDT();
  }

  /**
   * Subsequent invocations will always return the same result until the client relinquishes the
   * EDT, provided the project configuration is not changed in the meantime.
   */
  public @Nullable ImmutableList<SQLTable> getTables(SieuferdProject project) {
    ThreadUtil.checkOnEDT();
    if (content != null && !content.configurationMatches(project))
      content = null;
    return content == null ? null : content.tables;
  }

  public void postRefresh(SieuferdProject project, boolean showConnectionDialog) {
    ThreadUtil.checkOnEDT();
    if (isLoading.getAndSet(true))
      return;
    RP.post(new Updater(project, showConnectionDialog));
  }

  private final class Updater implements Runnable {
    private final SieuferdProject project;
    private final boolean showConnectionDialog;

    Updater(SieuferdProject project, boolean showConnectionDialog) {
      Preconditions.checkNotNull(project);
      this.project = project;
      this.showConnectionDialog = showConnectionDialog;
    }

    @Override
    public void run() {
      final TableListing tableListing = ViewUpdater.getManagedTables(showConnectionDialog, project);
      SwingUtilities.invokeLater(() -> {
        // Keep existing cache content in the case of a complete failure.
        if (tableListing.databasesWithTableListingErrors.isEmpty() ||
            !tableListing.tables.isEmpty())
        {
          content = new CachedContent(tableListing.connectionName, tableListing.managedDatabases,
              ImmutableList.copyOf(tableListing.tables));
        }
        isLoading.set(false);
      });
    }
  }

  @Immutable
  private static final class CachedContent {
    final String connectionName;
    final ImmutableSet<SQLDatabase> managedDatabases;
    final ImmutableList<SQLTable> tables;

    CachedContent(String connectionName, Set<SQLDatabase> managedDatabases, List<SQLTable> tables) {
      Preconditions.checkNotNull(connectionName);
      Preconditions.checkNotNull(managedDatabases);
      Preconditions.checkNotNull(tables);
      this.connectionName = connectionName;
      this.managedDatabases = ImmutableSet.copyOf(managedDatabases);
      this.tables = ImmutableList.copyOf(tables);
    }

    boolean configurationMatches(SieuferdProject project) {
      return connectionName.equals(project.getConfiguration().getDatabaseConnectionName()) &&
          project.getConfiguration().getManagedDatabases().equals(managedDatabases);
    }
  }
}
