// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.propertyeditors;

import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.ViewPropertyType;
import com.sieuferd.project.ViewContext;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.uiutil.ThreadUtil;
import java.beans.PropertyEditorSupport;
import java.lang.ref.WeakReference;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import org.openide.explorer.propertysheet.ExPropertyEditor;
import org.openide.explorer.propertysheet.PropertyEnv;

public abstract class ViewPropertyEditor<T>
    extends PropertyEditorSupport implements ExPropertyEditor
{
  public static final String PROP_VIEW_CONTEXT = "sieuferdViewPropertyEditorViewContext";
  public static final String PROP_CONTEXT_PATH = "sieuferdViewPropertyEditorContextPath";
  public static final String PROP_PROPERTY_TYPE = "sieuferdViewPropertyEditorPropertyType";

  private WeakReference<ViewContext> viewContext = new WeakReference<>(null);
  private @Nullable RelationPath contextPath;
  private @Nullable ViewPropertyType<T> viewPropertyType;

  /** Overriding methods in subclasses must call the super method first. */
  @Override
  @SuppressWarnings("unchecked")
  public void attachEnv(PropertyEnv env) {
    ThreadUtil.checkOnEDT();
    {
      Object obj = env.getFeatureDescriptor().getValue(PROP_VIEW_CONTEXT);
      viewContext = new WeakReference<>(
          (obj instanceof ViewContext) ? (ViewContext) obj : null);
    }
    {
      Object obj = env.getFeatureDescriptor().getValue(PROP_PROPERTY_TYPE);
      viewPropertyType = (obj instanceof ViewPropertyType<?>) ? (ViewPropertyType<T>) obj : null;
    }
    if (viewContext != null) {
      Object obj = env.getFeatureDescriptor().getValue(PROP_CONTEXT_PATH);
      contextPath = (obj instanceof RelationPath) ? (RelationPath) obj : null;
    }
  }

  // TODO[cc]: Get rid of this.
  public @Nullable ViewContext getViewContext() {
    return viewContext.get();
  }

  public @Nullable Strepco getStrepco() {
    @Nullable ViewContext useViewContext = getViewContext();
    if (useViewContext == null)
      return null;
    RelationPath useContextPath = (contextPath == null) ? Path.ROOT : contextPath;
    return Strepco.create(useViewContext.getView(), useContextPath);
  }

  public @Nullable ViewPropertyType<T> getViewPropertyType() {
    return viewPropertyType;
  }

  @Override
  @SuppressWarnings("unchecked")
  public @Nullable T getValue() {
    Object ret = super.getValue();
    if (ret != null && viewPropertyType != null &&
        !viewPropertyType.getValueClass().isAssignableFrom(ret.getClass()))
    {
      return null;
    }
    return (T) ret;
  }

  /** @deprecated use setValueTyped instead */
  @Override
  @Deprecated
  public final void setValue(@Nullable Object value) {
    setValueInternal(value);
  }

  /** @param value may not be null */
  public final void setValueTyped(T value) {
    if (value == null)
      throw new IllegalArgumentException("Value may not be null");
    setValueInternal(value);
  }

  private void setValueInternal(@Nullable Object value) {
    if (value != null && viewPropertyType != null &&
        !viewPropertyType.getValueClass().isAssignableFrom(value.getClass()))
    {
      throw new IllegalArgumentException("Tried to set value of incorrect type");
    }
    super.setValue(value);
  }

  @Override
  public final void setAsText(String text) throws IllegalArgumentException {
    ThreadUtil.checkOnEDT();
    if (text == null)
      throw new IllegalArgumentException("Cannot set value from null text string");
    T value = textToValue(text);
    if (value != null)
      setValueTyped(value);
  }

  @Override
  public final String getAsText() {
    ThreadUtil.checkOnEDT();
    T value = getValue();
    if (value == null)
      return null;
    return valueToText(value);
  }

  /**
   * Should be overridden by subclasses if text input is to be supported.
   *
   * @return can be null if the value should not change
   */
  protected @Nullable T textToValue(@Nonnull String text) throws IllegalArgumentException {
    throw new IllegalArgumentException("Value parsing not supported for this property");
  }

  /**
   * Should be overridden by subclasses if text output is to be supported.
   *
   * @return can be null if the value can't be expressed as a string
   */
  protected @Nullable String valueToText(@Nonnull T value) {
    return null;
  }
}
