// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.result;

import com.sieuferd.datamodel.ViewResult;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.RelationConstant;
import com.sieuferd.datamodel.expressions.TupleConstant;
import com.sieuferd.evaluator.FunctionalDependencies;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate2;
import com.sieuferd.flow.ValueProcessor.Synchronicity;
import com.sieuferd.layout.CrosstabColumns;
import com.sieuferd.layout.LBProp;
import com.sieuferd.layout.LMProp;
import com.sieuferd.project.filter.FilterOptionsQuery;
import com.sieuferd.query.Queries;
import com.sieuferd.util.MapUtil;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

public class CrosstabMeasurer extends ProcessorTemplate2<ImmutableView,ImmutableView,String> {
  public static final CrosstabMeasurer INSTANCE = new CrosstabMeasurer();
  private static final int MAX_CROSSTAB_COLUMNS = 30;
  public static final Set<ViewProperty<?>> CROSSTAB_INPUTS = ImmutableSet.<ViewProperty<?>>builder()
      .addAll(QueryUtil.QUERY_PROPERTIES)
      .add(LBProp.CROSSTAB)
      .add(LMProp.CROSSTAB_COLUMNS)
      .build();
  public static final Set<ViewProperty<?>> CROSSTAB_OUTPUTS =
      ImmutableSet.<ViewProperty<?>>builder()
      .add(LMProp.CROSSTAB_COLUMNS)
      .build();

  @Override
  public Synchronicity getSynchronicity() {
    return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
  }

  /** @param ret will be populated with an entry for every visible relation path in {@code view} */
  private void computeCrosstabHeadingPaths(
      FunctionalDependencies fds, ImmutableView view, RelationPath path,
      ImmutableSet<PrimitivePath> fullyIdentifiedPaths,
      Map<RelationPath,ImmutableSet<PrimitivePath>> ret)
  {
    final ImmutableSet<PrimitivePath> childFullyIdentifiedPaths;
    final Set<PrimitivePath> localTupleKey;
    {
      final Set<PrimitivePath> localIdentifiedPaths = Sets.newLinkedHashSet(fullyIdentifiedPaths);
      localTupleKey = Sets.newLinkedHashSet(fds.getKey(view, localIdentifiedPaths,
          ImmutableSet.copyOf(Queries.getImmediateProjectedChildren(view, path))));
      childFullyIdentifiedPaths = ImmutableSet.copyOf(localIdentifiedPaths);
    }
    final Set<PrimitivePath> headingPaths;
    if (view.get(LBProp.CROSSTAB, path)) {
      headingPaths = Sets.newLinkedHashSet();
      for (Path childPath : view.getChildPaths(path)) {
        if (!childPath.isRelation()) {
          /* To ensure that each crosstab column has no more than one tuple for each tuple in the
          crosstab relation's parent relation, there must be one crosstab column per unique tuple in
          the projection of the crosstab relation's local tuple key. Given this, paths in the
          crosstab relation that are fully identified by the crosstab relation's local tuple key can
          always be made part of the crosstab heading. There is no benefit to putting such paths in
          the result area, as their values would be repeated for every tuple in the crosstab
          relation's parent relation. Thus we always assume that such paths will be part of the
          crosstab heading. We make an exception for constant formulas, which are always made part
          of the result area in anticipation of being edited into aggregate formulas. */
          if (Queries.shouldProject(view, childPath) &&
              fds.isIdentifiedBy(view, childPath.getPrimitive(), localTupleKey) &&
              !fds.isIdentifiedBy(view, childPath.getPrimitive(), Sets.newLinkedHashSet()))
          {
            headingPaths.add(childPath.getPrimitive());
          }
        }
      }
    } else {
      headingPaths = ImmutableSet.of();
    }
    MapUtil.putChecked(ret, path, ImmutableSet.copyOf(headingPaths));
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation() && view.get(CProp.VISIBLE, childPath)) {
        computeCrosstabHeadingPaths(
            fds, view, childPath.getRelation(), childFullyIdentifiedPaths, ret);
      }
    }
  }

  private static CrosstabColumns calculateCrosstabColumns(
      ImmutableView view, RelationPath path, QueryProcessor queryProcessor,
      String databaseConnection, ImmutableSet<PrimitivePath> headingPaths)
  {
    final FilterOptionsQuery query = FilterOptionsQuery.create(
        view, path, headingPaths, ImmutableSet.copyOf(headingPaths));
    final ImmutableView optionsView = query.getFilterOptionsView("", true);
    final ViewResult viewResult = queryProcessor.process(
        new EvaluationInput(databaseConnection, optionsView, MAX_CROSSTAB_COLUMNS));
    final ImmutableData data = viewResult.getData();
    final TupleInterval tupleInterval =
        data.getRelationTuples(Path.ROOT, Tuple.createRootParent(data));
    final Set<TupleConstant> tupleConstants = Sets.newLinkedHashSet();
    for (Tuple tuple : tupleInterval) {
      final Map<PrimitivePath,ConstantExpression> toTuple = Maps.newLinkedHashMap();
      boolean unencodeable = false;
      for (PrimitivePath headingPath : headingPaths) {
        PrimitiveValue primitiveValue = query.getValue(data, tuple, headingPath);
        if (primitiveValue == null)
          throw new AssertionError();
        ConstantExpression constantExpression =
            ConstantExpression.fromObject(primitiveValue.getContent());
        if (constantExpression == null) {
          unencodeable = true;
          break;
        }
        MapUtil.putChecked(toTuple, headingPath, constantExpression);
      }
      if (!unencodeable)
        tupleConstants.add(new TupleConstant(toTuple));
    }
    return new CrosstabColumns(new RelationConstant(headingPaths, tupleConstants));
  }

  @Override
  protected ImmutableView process(ImmutableView view, String databaseConnection)
      throws InterruptedException
  {
    final Map<RelationPath,ImmutableSet<PrimitivePath>> headings = Maps.newLinkedHashMap();
    computeCrosstabHeadingPaths(
        new FunctionalDependencies(view), view, Path.ROOT, ImmutableSet.of(), headings);
    final QueryProcessor queryProcessor = new QueryProcessor("Crosstab");
    final MutableView ret = Views.newMutable(view);
    for (Entry<RelationPath,ImmutableSet<PrimitivePath>> entry : headings.entrySet()) {
      if (Thread.interrupted())
        throw new InterruptedException();
      final RelationPath relationPath = entry.getKey();
      final ImmutableSet<PrimitivePath> headingPaths = entry.getValue();
      final CrosstabColumns crosstabColumns = headingPaths.isEmpty() ? CrosstabColumns.UNIT :
          calculateCrosstabColumns(
          view, relationPath, queryProcessor, databaseConnection, headingPaths);
      ret.set(LMProp.CROSSTAB_COLUMNS, relationPath, crosstabColumns);
    }
    return ImmutableView.copyOf(ret, CROSSTAB_OUTPUTS, null);
  }
}
