// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.result;

import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;
import java.io.IOException;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.concurrent.ThreadSafe;
import org.openide.util.NbBundle.Messages;
import org.openide.windows.IOProvider;
import org.openide.windows.InputOutput;
import org.openide.windows.OutputWriter;

/** Access methods for the "SQL Queries" output window, which is shared between view result
providers. */
@ThreadSafe
final class QueryLog {
  private static final Logger LOG = Logger.getLogger(QueryLog.class.getName());
  private static final Set<Object> queryLogAcquirers = Sets.newLinkedHashSet();
  private static InputOutput queryLog;

  private QueryLog() { }

  /**
   * Indicate that the specified acquirer would like to use the query log, for which an
   * {@code OutputWriter} is returned. Safe to call multiple times.
   */
  @Messages("TIT_SQLQueries=SQL Queries")
  public static synchronized
    OutputWriter acquireQueryLog(Object acquirer, boolean clear)
  {
    Preconditions.checkNotNull(acquirer);
    queryLogAcquirers.add(acquirer);
    if (queryLog == null)
      queryLog = IOProvider.getDefault().getIO(Bundle.TIT_SQLQueries(), false);
    // Don't do queryLog.select(), as that would force the Output window to be opened.
    if (clear) {
      try {
        /* There was once a bug where the output pane's contents became garbled when two threads
        were using it simultaneously; see https://netbeans.org/bugzilla/show_bug.cgi?id=242979 . It
        seems to have been fixed. */
        queryLog.getOut().reset();
      } catch (IOException e) {
        LOG.log(Level.WARNING, "Failed to clear the SQL Queries output window", e);
      }
    }
    return queryLog.getOut();
  }

  /**
   * Indicate that the specified acquirer is no longer using the query log, regardless of how many
   * times {@link acquireQueryLog(Object,boolean) might have been called before}. Safe to call
   * multiple times.
   */
  public static synchronized void releaseQueryLog(Object acquirer) {
    Preconditions.checkNotNull(acquirer);
    queryLogAcquirers.remove(acquirer);
    if (queryLogAcquirers.isEmpty() && queryLog != null) {
      queryLog.closeInputOutput();
      queryLog = null;
    }
  }
}
