// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.result;

import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.datamodel.ResultLimit;
import com.google.common.base.Preconditions;
import com.google.common.collect.LinkedListMultimap;
import com.google.common.collect.Multimap;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLTask;
import com.sieuferd.database.SQLTask.UserInformedConnectionException;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.evaluator.CancelHandler;
import com.sieuferd.evaluator.QueryEvaluator;
import com.sieuferd.flow.ValueProcessor;
import com.sieuferd.project.NBSQLTaskRunner;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.util.Stopwatch;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nonnull;
import org.openide.util.RequestProcessor;
import org.openide.util.RequestProcessor.Task;

public class QueryProcessor extends ValueProcessor<EvaluationInput,ViewResult> {
  private static final RequestProcessor RP = new RequestProcessor(
      "Query Processor", 50, false, true);
  private static final RequestProcessor RP_CANCELER = new RequestProcessor(
      "Query Canceler", 50, false, true);
  private static final AtomicLong DEBUG_NEXT_QUERY_ID = new AtomicLong();
  private static final Logger LOG = Logger.getLogger(QueryProcessor.class.getName());
  private final AtomicBoolean testConnection = new AtomicBoolean(true);
  private final String debugName;

  public QueryProcessor(String debugName) {
    super();
    Preconditions.checkNotNull(debugName);
    this.debugName = debugName;
  }

  @Override
  public ValueProcessor.Synchronicity getSynchronicity() {
    return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
  }

  @Override
  public ViewResult process(final EvaluationInput input) {
    final long debugQueryID = DEBUG_NEXT_QUERY_ID.getAndIncrement();
    final CancelHandler cancelHandler = new CancelHandler();
    final ViewResult[] ret = new ViewResult[] { null };
    Task task = RP.post(() -> {
      ret[0] = processHelper(input, cancelHandler, debugQueryID);
    });
    boolean interruptHandled = false;
    while (true) {
      try {
        task.waitFinished(0);
        break;
      } catch (InterruptedException e) {
        ThreadUtil.checkNotOnEDT();
        Thread.currentThread().interrupt();
      }
      if (Thread.interrupted()) {
        if (!interruptHandled) {
          interruptHandled = true;
          // Do this on a separate thread to avoid blocking on stale connections.
          RP_CANCELER.post(() -> {
            LOG.info("Attempting to cancel query");
            UIUtil.instrumentation(
                "CancelQuery processorName='" + debugName + "', queryID=" + debugQueryID);
            try {
              cancelHandler.cancel();
            } catch (SQLException e) {
              LOG.log(Level.WARNING, "Statement.cancel() threw exception", e);
            }
          });
        }
      }
    }
    if (ret[0] == null)
      throw new AssertionError("Unhandled exception in QueryProcessor");
    return ret[0];
  }

  public static ViewResult createEmptyOutput(Warning ... warnings) {
    Multimap<Path,Warning> warningMap = LinkedListMultimap.create();
    warningMap.putAll(Path.ROOT, Arrays.asList(warnings));
    return new ViewResult(ResultLimit.UNDEFINED, QueryUtil.EMPTY_DATA_VIEW,
        QueryUtil.EMPTY_INCOMPLETE_DATA, warningMap);
  }

  private @Nonnull ViewResult processHelper(
      final EvaluationInput input, final CancelHandler cancelHandler, final long debugQueryID)
  {
    try {
      return NBSQLTaskRunner.get(input.getDatabaseConnectionName())
          .runTask(true, testConnection.getAndSet(false),
          new SQLTask<ViewResult>("Process view query")
      {
        @Override
        public ViewResult run(SQLConnection sqlConnection) {
          final QueryEvaluator evaluator;
          final MutableView view = Views.newMutable(input.getQuerySnapshot());
          UIUtil.instrumentation(
                  "PrepareQuery processorName='" + debugName + "', queryID=" + debugQueryID);
          Stopwatch stopwatch1 = new Stopwatch("Query preparation");
          evaluator = QueryEvaluator.create(view, sqlConnection,
              QueryLog.acquireQueryLog(QueryProcessor.this, true), cancelHandler);
          stopwatch1.stop();
          final ImmutableData data;
          final Multimap<Path,Warning> warnings = LinkedListMultimap.create();
          try {
            LOG.info("Executing query");
            UIUtil.instrumentation(
                  "ExecuteQuery processorName='" + debugName + "', queryID=" + debugQueryID);
            Stopwatch stopwatch2 = new Stopwatch("Query execution");
            data = evaluator.evaluate(input.getLimit(), warnings);
            stopwatch2.stop();
          } finally {
            try {
              evaluator.close();
            } catch (SQLException e) {
              sqlConnection.reportSQLException(e);
              warnings.put(Path.ROOT, new Warning(
                  "Failed to clean up after query: " + e.getMessage()));
            }
          }
          UIUtil.instrumentation(
              "QueryComplete processorName='" + debugName + "', queryID=" + debugQueryID);
          return new ViewResult(ResultLimit.create(input.getQuerySnapshot(), input.getLimit()),
              input.getQuerySnapshot(), data, warnings);
        }
      });
    } catch (SQLException | UserInformedConnectionException e) {
      return createEmptyOutput(new Warning("Database connection problem: " + e.getMessage()));
    }
  }

  @Override
  public void clearState() {
    LOG.info("Released result");
    testConnection.set(true);
    QueryLog.releaseQueryLog(this);
  }
}
