// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.result;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.datamodel.ResultLimit;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewFlow;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate2;
import com.sieuferd.flow.Refreshable;
import com.sieuferd.flow.Refreshable.RefreshableAndFlow;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.flow.ValueProcessor;
import com.sieuferd.flow.Flows;
import com.sieuferd.flow.ProcessorTemplates.ProcessorTemplate3;
import com.sieuferd.flow.ValueProcessor.Synchronicity;
import com.sieuferd.layout.LMProp;
import com.sieuferd.layout.LayoutGenerator;
import com.sieuferd.layout.LayoutMeasurer;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.util.QueryLabelFormatter;
import com.sieuferd.uiutil.DelegatingUndoRedo;
import com.sieuferd.project.viewfile.ViewDataObject;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.util.Stopwatch;
import com.sieuferd.util.UnitObject;
import com.sieuferd.viewset.GatedView;
import java.io.IOException;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import org.openide.awt.UndoRedo;
import org.openide.nodes.Node.Cookie;
import org.openide.util.NbBundle.Messages;

/**
 * State and logic for retrieving and updating query results and structural layout properties for a
 * View. Each view has one {@code ViewResultProvider}, which may share query result state between
 * multiple cloned editors or other clients. Except where noted, methods in this class may only be
 * called from the EDT.
 */
@Messages("CTL_PerformLayoutMeasurements=Perform Layout Measurements")
public final class ViewResultProvider implements Cookie {
  private static final Logger LOG = Logger.getLogger(ViewResultProvider.class.getName());
  private static final int QUERY_LIMIT_INITIAL = 100;
  private static final double QUERY_LIMIT_INCREASE_FACTOR = 2.0;
  private final QueryProcessor queryProcessor = new QueryProcessor("View");
  private final SettableFlow<UnitObject> stopValveInput = Flows.newSettable(UnitObject.INSTANCE);
  private final Flow<ViewContext> viewContextFlow;
  private final Flow<UnitObject> progressFlow;
  private final Flow<ViewResult> viewResultFlow;
  private final Flow<UndoRedo> undoRedoFlow;
  private final Map<Object,Integer> minimumQueryLimit = Maps.newLinkedHashMap();
  private final Refreshable refreshable;
  private final Object minimumQueryLimitLock = new Object();
  private final Object measurementLock = new Object();
  // TODO[cc]: Consider creating a non-nullable version of SettableFlow.
  /** Content should never be null. */
  private final SettableFlow<ResultLimit> resultLimitSettable =
      Flows.newSettable(ResultLimit.UNDEFINED);
  private final Set<ViewProperty<?>> LAYOUT_AND_BADGE_PROPERTIES =
      Sets.union(Sets.union(LayoutGenerator.LAYOUT_PROPERTIES, QueryUtil.QUERY_PROPERTIES),
      /* TODO[cc]: Use QueryLabelFormatter.INSTANCE in one place only, then pass around and use a
                   generic LabelFormatter instance in places like this. */
      QueryLabelFormatter.INSTANCE.getBadgeProperties());

  public Flow<ViewContext> getViewContextFlow() {
    return viewContextFlow;
  }

  public Flow<UndoRedo> getUndoRedoFlow() {
    return undoRedoFlow;
  }

  /** @return a flow whose value will be null while evaluation is in progress */
  public Flow<UnitObject> getProgressFlow() {
    return progressFlow;
  }

  public Flow<ViewResult> getViewResultFlow() {
    return viewResultFlow;
  }

  public Refreshable getRefreshable() {
    return refreshable;
  }

  public void stop() {
    stopValveInput.set(null);
  }

  // TODO[cc]: Consider if there's a way to clean up this ugly API.
  /**
   * Set a minimum for the initial query limit that will be used whenever a new query is
   * encountered. Multiple clients may maintain independent minimums which can be increased or
   * decreased. When a client is no longer using the ViewResultProvider, it should set its limits to
   * zero. May be called from any thread.
   *
   * @param key an identifier that can subsequently be used to change the minimum; allows multiple
   *        clients (e.g. for cloned ViewResultTopComponent tabs) to maintain their own minimums
   * @param value the minimum query limit
   */
  public void setMinimumQueryLimit(Object key, int value) {
    Preconditions.checkNotNull(key);
    Preconditions.checkArgument(value >= 0);
    synchronized (minimumQueryLimitLock) {
      if (value == 0) {
        minimumQueryLimit.remove(key);
      } else {
        minimumQueryLimit.put(key, value);
      }
    }
  }

  private int getMinimumQueryLimit() {
    int ret = QUERY_LIMIT_INITIAL;
    synchronized (minimumQueryLimitLock) {
      for (Integer value : minimumQueryLimit.values())
        ret = Math.max(ret, value);
    }
    return ret;
  }

  public void ensureQueryLimitIncreased(ResultLimit displayedResultLimit) {
    ThreadUtil.checkOnEDT();
    /* Use getMinimumQueryLimit here as well (not just for initial limits), for cases where the
    standard increase factor might not be sufficient to fill the viewport in one go. */
    resultLimitSettable.set(ResultLimit.getIncreasedLimit(
        resultLimitSettable.getWithoutListener(), displayedResultLimit,
        QUERY_LIMIT_INCREASE_FACTOR, getMinimumQueryLimit()));
  }

  public ViewResultProvider(ViewDataObject viewDataObject) {
    viewContextFlow = Flows.newProcessed(Flows.newConstant(viewDataObject),
        new ValueProcessor<ViewDataObject,ViewContext>()
    {
      @Override
      public ValueProcessor.Synchronicity getSynchronicity() {
        return ValueProcessor.Synchronicity.ASYNCHRONOUS;
      }

      @Override
      public ViewContext process(ViewDataObject input) {
        try {
          return ViewContext.create(input);
        } catch (IOException e) {
          // Ignore; the underlying FileObject will be deleted by NBViewSet.
          return null;
        }
      }
    });

    final Flow<ImmutableView> layoutFlow = Flows.newIndirect(viewContextFlow,
        viewContext -> new ViewFlow(viewContext.getView(), LAYOUT_AND_BADGE_PROPERTIES, null));
    final Flow<ImmutableView> crosstabFlow = Flows.newIndirect(viewContextFlow,
        viewContext -> new ViewFlow(viewContext.getView(), CrosstabMeasurer.CROSSTAB_INPUTS, null)
    );
    final Flow<ImmutableView> queryFlow = Flows.newIndirect(viewContextFlow, viewContext ->
        new ViewFlow(viewContext.getView(), QueryUtil.QUERY_PROPERTIES, null));
    final Flow<String> databaseConnection = Flows.newIndirect(viewContextFlow, viewContext ->
        viewContext.getProject().getConfiguration().getDatabaseConnectionNameFlow());

    final Flow<Integer> effectiveLimitFlow = Flows.newComposite2(queryFlow, resultLimitSettable,
        (querySnapshot, resultLimit) ->
    {
      return resultLimit.appliesTo(querySnapshot)
          ? resultLimit.getLimit() : getMinimumQueryLimit();
    });
    final Flow<ViewResult> rawResultFlow;
    {
      Flow<EvaluationInput> inputBeforeRefresh = Flows.newComposite3(
          queryFlow, databaseConnection, effectiveLimitFlow,
          (view, dbc, limit) -> new EvaluationInput(dbc, view, limit));
      RefreshableAndFlow<EvaluationInput> refreshableAndFlow =
          Flows.newRefreshableAndFlow(inputBeforeRefresh);
      this.refreshable = refreshableAndFlow.getRefreshable();
      /* Un-stop queries in response to any change in the input, including changes caused by the
      nudge of the refresh button. */
      Flow<UnitObject> unstopSideEffectFlow = Flows.newProcessed(refreshableAndFlow.getFlow(),
          new ValueProcessor<EvaluationInput,UnitObject>()
      {
        @Override
        public Synchronicity getSynchronicity() {
          // Needed to avoid deadlock due to the flow-accessing side-effect of this processor.
          return Synchronicity.ASYNCHRONOUS;
        }

        @Override
        public UnitObject process(EvaluationInput input) {
          stopValveInput.set(UnitObject.INSTANCE);
          return UnitObject.INSTANCE;
        }
      });
      rawResultFlow = Flows.attachSideEffectDependency(true, unstopSideEffectFlow,
          Flows.newProcessed(
          Flows.newValve(stopValveInput, refreshableAndFlow.getFlow()), queryProcessor));
    }

    final Flow<ImmutableView> measuredFlow =
        new ProcessorTemplate2<ImmutableView,ViewResult,ImmutableView>()
    {
      @Override
      public Synchronicity getSynchronicity() {
        return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
      }

      @Override
      protected ImmutableView process(ViewResult viewResult, ImmutableView layoutView)
          throws InterruptedException
      {
        ThreadUtil.checkNotOnEDT();
        /* Don't have several invocations of LayoutMeasurer.measureData run simultaneously; they
        would be doing the same work. */
        synchronized (measurementLock) {
          /* Measure in a separate view without default properties set so that only measured paths
          end up having properties overwritten in the original view. */
          final MutableView measuredView = Views.newMutable();
          measuredView.copyFrom(layoutView, Path.ROOT, LayoutMeasurer.MEASUREMENT_INPUTS, null);
          Stopwatch stopwatch = new Stopwatch("Layout measurements");
          LayoutMeasurer.measureData(viewResult.getData(), measuredView, null);
          stopwatch.stop();
          return ImmutableView.copyOf(measuredView);
        }
      }
    }.getFlow(rawResultFlow, layoutFlow);

    final Flow<ImmutableView> calculatedCrosstabFlow =
        CrosstabMeasurer.INSTANCE.getFlow(crosstabFlow, databaseConnection);

    final Flow<UnitObject> writeBackSideEffectFlow =
        new ProcessorTemplate3<UnitObject,ViewContext,ImmutableView,ImmutableView>()
    {
      @Override
      public Synchronicity getSynchronicity() {
        /* This processor is a bit of a hack; it mutates the view, which may in turn cause another
        remeasurement, making the processor potentially interrupt itself. For this to work without
        deadlocks, the processor must be non-interruptable and asynchronous. */
        return Synchronicity.ASYNCHRONOUS;
      }

      @Override
      protected UnitObject process(ViewContext viewContext, final ImmutableView measuredView,
          final ImmutableView calculatedCrosstabView)
      {
        viewContext.getView().mutate(Bundle.CTL_PerformLayoutMeasurements(), false, view -> {
          view.copyFrom(measuredView, Path.ROOT, LMProp.INSTANCE.getViewProperties(), null);
          view.copyFrom(calculatedCrosstabView, Path.ROOT, CrosstabMeasurer.CROSSTAB_OUTPUTS, null);
        });
        return UnitObject.INSTANCE;
      }
    }.getFlow(viewContextFlow, measuredFlow, calculatedCrosstabFlow);

    /* Don't include a warning in the empty default result; the result will only be shown when the
    view is first opened, while the progress indicator is spinning. */
    final ViewResult defaultResult = QueryProcessor.createEmptyOutput();

    viewResultFlow = Flows.attachSideEffectDependency(false, writeBackSideEffectFlow,
        Flows.newComposite2(Flows.newSticky(rawResultFlow, defaultResult), layoutFlow,
        (dataResult, layoutView) -> new ViewResult(
            dataResult.getLimit(), layoutView, dataResult.getData(), dataResult.getWarnings())));

    this.progressFlow = Flows.newComposite3(Flows.newBoolean(rawResultFlow),
        Flows.newBoolean(stopValveInput), Flows.newBoolean(writeBackSideEffectFlow),
        (Boolean result, Boolean notStopped, Boolean wroteBack) ->
        (result && wroteBack || !notStopped) ? UnitObject.INSTANCE : null);

    undoRedoFlow = Flows.newProcessed(viewContextFlow, new ValueProcessor<ViewContext,UndoRedo>() {
      private final Object lock = new Object();
      private @Nullable GatedView view;
      private @Nullable UndoRedo.Manager manager;

      @Override
      public UndoRedo process(ViewContext input) {
        return setView(input.getView());
      }

      @Override
      public void clearState() {
        setView(null);
        LOG.info("Released undo/redo state");
      }

      private UndoRedo.Manager setView(@Nullable GatedView view) {
        synchronized (lock) {
          final GatedView oldView = this.view;
          if (oldView != null) {
            oldView.removeUndoableEditListener(manager);
            manager = null;
          }
          this.view = view;
          if (view != null) {
            manager = new UndoRedo.Manager();
            view.addUndoableEditListener(manager);
          }
          return this.manager;
        }
      }
    });
  }

  public static UndoRedo createUndoRedo(Flow<ViewResultProvider> viewResultProviderFlow) {
    final Flow<UndoRedo> delegateFlow = Flows.newIndirect(viewResultProviderFlow,
        viewResultProvider -> viewResultProvider.getUndoRedoFlow());
    return DelegatingUndoRedo.createFromFlow(delegateFlow);
  }
}
