// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.toolbox;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import com.sieuferd.flow.Flows;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.project.FieldContext;
import com.sieuferd.uiutil.api.PopupLike;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.flow.UIModelFlows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.upstream.popup.WindowPopupLike;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.UnitObject;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.SingleSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import org.netbeans.swing.tabcontrol.TabData;
import org.netbeans.swing.tabcontrol.TabDataModel;
import org.netbeans.swing.tabcontrol.TabbedContainer;

final class ToolboxPanel extends JPanel {
  private final ImmutableList<ToolboxTab> TABS =
      ImmutableList.of(new FieldsTabPanel(), new FilterTabPanel());
  private final SettableFlow<UnitObject> openedSettable = Flows.newSettable();
  private final TabbedContainer tabbedContainer;
  private @Nullable PopupLike popup;
  private @Nullable TabCycleKeyEventDispatcher tabCycleKeyEventDispatcher;

  ToolboxPanel() {
    /* Neither TYPE_EDITOR or TYPE_SLIDING are very useful here. TYPE_TOOLBAR shows buttons instead
    of tabs. */
    tabbedContainer = new TabbedContainer(null, TabbedContainer.TYPE_VIEW, new ToolboxWinsysInfo());
    /* I considered using vertical tabs to the left of the toolbox; this might require using a
    TabDisplayer rather than a TabbedContainer. I've decided against it, however; it's unfamiliar to
    users and it's just a tradeoff between horizontal and vertical space. */
    setLayout(new BorderLayout());
    add(tabbedContainer, BorderLayout.CENTER);
    final TabDataModel tabDataModel = tabbedContainer.getModel();
    {
      int i = 0;
      for (ToolboxTab tab : TABS)
        tabDataModel.addTab(i++, new TabData(tab, null, tab.getTitle(), null));
    }
    final SingleSelectionModel selectionModel = tabbedContainer.getSelectionModel();
    if (!TABS.isEmpty())
      selectionModel.setSelectedIndex(0);
    tabbedContainer.setActive(true);

    UIFlows.bindEDT(Flows.newValve(openedSettable, UIModelFlows.singleSelection(selectionModel)),
        (@Nullable Integer value) ->
    {
      for (int i = 0; i < TABS.size(); i++)
        TABS.get(i).setOpened(value != null && i == value);
      if (value != null)
        requestFocusInWindow();
    });
  }

  public PopupLike getPopup() {
    ThreadUtil.checkOnEDT();
    if (popup == null) {
      popup = new WindowPopupLike(this, new Dimension(350, 500), false); // (300, 280)
      popup.addPopupMenuListener(new PopupMenuListener() {
        @Override
        public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
          ToolboxPanel.this.setOpened(true);
        }

        @Override
        public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
          ToolboxPanel.this.setOpened(false);
        }

        @Override
        public void popupMenuCanceled(PopupMenuEvent e) {
          /* Probably not necessary, since popupMenuWillBecomeInvisible will be called, too, but
          doesn't hurt. */
          ToolboxPanel.this.setOpened(false);
        }
      });
    }
    return popup;
  }

  private void setOpened(boolean opened) {
    ThreadUtil.checkOnEDT();
    openedSettable.set(opened ? UnitObject.INSTANCE : null);
    if (opened && tabCycleKeyEventDispatcher == null) {
      KeyboardFocusManager fm = KeyboardFocusManager.getCurrentKeyboardFocusManager();
      tabCycleKeyEventDispatcher = new TabCycleKeyEventDispatcher(fm);
      fm.addKeyEventDispatcher(tabCycleKeyEventDispatcher);
    } else if (!opened && tabCycleKeyEventDispatcher != null) {
      tabCycleKeyEventDispatcher.focusManager.removeKeyEventDispatcher(tabCycleKeyEventDispatcher);
      tabCycleKeyEventDispatcher = null;
    }
  }

  public void setFieldContexts(final Set<FieldContext> fieldContexts) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkNotNull(fieldContexts);
    for (ToolboxTab tab : TABS)
      tab.setFieldContexts(fieldContexts);
  }

  private @Nullable Component getSelectedComponent() {
    final int idx = tabbedContainer.getSelectionModel().getSelectedIndex();
    return idx < 0 ? null : tabbedContainer.getModel().getTab(idx).getComponent();
  }

  // TODO[cc]: Simplify this.
  public void setActiveTab(Class<? extends ToolboxTab> tabClass) {
    Preconditions.checkNotNull(tabClass);
    final Map<Integer,ToolboxTab> matches = Maps.newLinkedHashMap();
    int i = 0;
    for (ToolboxTab tab : TABS) {
      if (tabClass.isAssignableFrom(tab.getClass()))
        MapUtil.putChecked(matches, i, tab);
      i++;
    }
    if (matches.size() != 1)
      throw new IllegalArgumentException("Expected to match one tab, matched " + matches);
    final int matchedIndex = matches.keySet().iterator().next();
    tabbedContainer.getSelectionModel().setSelectedIndex(matchedIndex);
  }

  @Override
  public final boolean requestFocusInWindow() {
    if (!isFocusable())
      return false;
    final Component defaultFocusComponent = getSelectedComponent();
    if (defaultFocusComponent == null) {
      return super.requestFocusInWindow();
    } else {
      return defaultFocusComponent.requestFocusInWindow();
    }
  }

  @Override
  public final void requestFocus() {
    if (!isFocusable())
      return;
    final Component defaultFocusComponent = getSelectedComponent();
    if (defaultFocusComponent == null) {
      super.requestFocus();
    } else {
      defaultFocusComponent.requestFocus();
    }
  }

  /* Handle Ctrl[+Shift]+Tab to cycle tabs through a KeyEventDispatcher rather than through actions,
  since trying to avoid conflicts with focus traversal keys via JComponent.setFocusTraversalKeys is
  tricky and error-prone. For instance, OutlineView/Outline/ETable actively frustrates such attempts
  by trying to restore previously modified focus traversal keys. */
  private class TabCycleKeyEventDispatcher implements KeyEventDispatcher {
    final KeyboardFocusManager focusManager;

    public TabCycleKeyEventDispatcher(KeyboardFocusManager focusManager) {
      Preconditions.checkNotNull(focusManager);
      this.focusManager = focusManager;
    }

    @Override
    public boolean dispatchKeyEvent(KeyEvent e) {
      final boolean tab = (e.getKeyCode() == KeyEvent.VK_TAB);
      final boolean pressed = (e.getID() == KeyEvent.KEY_PRESSED);
      final int modifiersEx = e.getModifiersEx();
      if (tab && modifiersEx == KeyEvent.CTRL_DOWN_MASK) {
        if (pressed)
          cycleTabs(true);
        return true;
      } else if (tab && modifiersEx == (KeyEvent.CTRL_DOWN_MASK | KeyEvent.SHIFT_DOWN_MASK)) {
        if (pressed)
          cycleTabs(false);
        return true;
      }
      return false;
    }

    private void cycleTabs(boolean forward) {
      final int increment = forward ? 1 : -1;
      final int N = TABS.size();
      if (N == 0)
        return;
      final SingleSelectionModel model = tabbedContainer.getSelectionModel();
      final int oldIndex = model.getSelectedIndex();
      final int newIndex = Math.floorMod(oldIndex + increment, N);
      model.setSelectedIndex(newIndex);
    }
  }

  public static final void main(String args[]) {
    SwingUtilities.invokeLater(() -> {
      UIUtil.setDefaultLookAndFeel(true);
      final Toolbox toolbox = Toolbox.getGlobalInstance();
      final JFrame frame = new JFrame("Test JFrame");
      frame.setSize(800, 600);
      frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
      frame.addKeyListener(new KeyAdapter() {
        @Override
        public void keyPressed(KeyEvent e) {
          Point p = MouseInfo.getPointerInfo().getLocation();
          Component invoker = frame.getRootPane();
          SwingUtilities.convertPointFromScreen(p, invoker);
          toolbox.getPopup().show(invoker, p);
        }
      });
      frame.setVisible(true);
    });
  }
}
