// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.util;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.Flows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.project.ViewContext;
import com.sieuferd.project.derivatives.ViewUpdater;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.util.Set;
import javax.annotation.Nullable;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JToggleButton;
import javax.swing.JToggleButton.ToggleButtonModel;
import org.openide.awt.Actions;
import org.openide.util.NbBundle.Messages;
import org.openide.util.actions.Presenter;

public final class BooleanViewPropertyAction extends ViewPropertyAction
    implements Presenter.Popup, Presenter.Menu, Presenter.Toolbar
{
  private final ViewProperty<Boolean> viewProperty;
  private final boolean inverted;
  private final boolean continueToPluralAncestor;
  private final Flow<Boolean> selectionFlow;

  public BooleanViewPropertyAction(final String displayName, final @Nullable String iconBase,
      final ViewProperty<Boolean> viewProperty, boolean inverted, boolean continueToPluralAncestor)
  {
    super(displayName, iconBase,
        continueToPluralAncestor ? AppliesTo.ANY : viewProperty.getAppliesTo(),
        ImmutableSet.<ViewProperty<?>>of(viewProperty), true);
    Preconditions.checkNotNull(viewProperty);
    this.viewProperty = viewProperty;
    this.inverted = inverted;
    this.continueToPluralAncestor = continueToPluralAncestor;
    this.selectionFlow = Flows.newProcessed(getSelectionSnapshotFlow(),
        selectionSnapshot -> getDisplayedState(
        selectionSnapshot.getViewSnapshot(), selectionSnapshot.getSelectedPaths()));
    ensureBindingsInitialized();
  }

  @Override
  public ViewPropertyAction create() {
    return new BooleanViewPropertyAction(
        getDefaultDisplayName(), getIconBase(), viewProperty, inverted, continueToPluralAncestor);
  }

  private boolean getDisplayedState(View view, Set<Path> selectedPaths) {
    for (Path path : selectedPaths) {
      Path effectivePath;
      if (!viewProperty.appliesTo(path)) {
        if (!continueToPluralAncestor || path.isRelation())
          throw new AssertionError();
        effectivePath = path.getParent();
      } else {
        effectivePath = path;
      }
      if (view.exists(path) && !(view.get(viewProperty, effectivePath) ^ inverted))
        return false;
    }
    return true;
  }

  @Messages({
    "# {0} - Display name of the disabled boolean property",
    "CTL_DisableBooleanProperty=Disable {0}"
  })
  @Override
  public void actionPerformed(
      ActionEvent e, ViewContext viewContext, final ImmutableSet<Path> selectedPaths)
  {
    /* Make a determination of whether to enable or disable the property before constructing the
    mutating operation, so that the latter always performs the same logical operation, and so that I
    can provide a sensible action description for undo/redo purposes. */
    final boolean newState = !Boolean.TRUE.equals(Flows.getMomentary(selectionFlow));
    viewContext.getView().mutate(newState ? getDefaultDisplayName() :
        Bundle.CTL_DisableBooleanProperty(getDefaultDisplayName()), true, (MutableView view) ->
    {
      final boolean value = newState ^ inverted;
      for (Path path : selectedPaths)
        setProperty(view, path, value);
    });
  }

  private void setProperty(MutableView view, Path path, boolean value) {
    if (viewProperty.appliesTo(path))
      ViewUpdater.userSetProperty(view, viewProperty, path, value);
    if (continueToPluralAncestor &&
        !path.isRoot() && !(path.isRelation() && view.get(CProp.PLURAL, path)))
    {
      setProperty(view, path.getParent(), value);
    }
  }

  private ToggleButtonModel createModel() {
    final ToggleButtonModel model = new ToggleButtonModel() {
      // Hold a strong reference to the weak Flow for as long as the model exists.
      private final Flow<Boolean> weakSelectedFlow = UIFlows.newWeak(selectionFlow);
      {
        UIFlows.bindEDT(weakSelectedFlow, (Boolean value) -> {
          setSelectedInternal(Boolean.TRUE.equals(value));
        });
      }

      private void setSelectedInternal(boolean b) {
        super.setSelected(b);
      }

      @Override
      public void setSelected(boolean b) {
        /* Ignore; the only actual selection changes will come from the underlying model, which I
        will in turn only modify in response to an ActionEvent. */
      }
    };
    return model;
  }

  @Override
  public boolean isEnabledFor(SelectionSnapshot selectionSnapshot) {
    return !viewProperty.isUserReadOnly();
  }

  /* Actions.connect, like Mnemonics.setLocalizedText, properly hides ampersand symbols from the
  user-visible action name. */

  @Override
  public JMenuItem getPopupPresenter() {
    JMenuItem ret = new JCheckBoxMenuItem();
    ret.setModel(createModel());
    Actions.connect(ret, this, true);
    return ret;
  }

  @Override
  public JMenuItem getMenuPresenter() {
    JMenuItem ret = new JCheckBoxMenuItem();
    ret.setModel(createModel());
    Actions.connect(ret, this, false);
    // TODO[pd]: Remove this workaround when the bug is fixed.
    // Workaround for https://netbeans.org/bugzilla/show_bug.cgi?id=243270 .
    ret.putClientProperty("ancestor", new JMenu("FakeMenu"));
    return ret;
  }

  @Override
  public Component getToolbarPresenter() {
    JToggleButton ret = new JToggleButton();
    ret.setModel(createModel());
    Actions.connect(ret, this);
    return ret;
  }
}
