// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.util;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.flow.Flow;
import com.sieuferd.uiutil.flow.LookupResultSnapshot;
import com.sieuferd.flow.Flows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.project.FieldContext;
import com.sieuferd.project.ViewContext;
import java.util.Set;
import javax.annotation.Nullable;
import org.openide.util.Lookup;

/** Shallowly immutable. */
public final class FieldSelection {
  private final ViewContext viewContext;
  private final ImmutableSet<Path> selectedPaths;

  /**
   * @param selectedPaths may not be empty
   */
  FieldSelection(ViewContext viewContext, Set<Path> selectedPaths) {
    Preconditions.checkNotNull(viewContext);
    Preconditions.checkNotNull(selectedPaths);
    Preconditions.checkArgument(!selectedPaths.isEmpty());
    this.viewContext = viewContext;
    this.selectedPaths = ImmutableSet.copyOf(selectedPaths);
  }

  public ViewContext getViewContext() {
    return viewContext;
  }

  public ImmutableSet<Path> getSelectedPaths() {
    return selectedPaths;
  }

  public static @Nullable FieldSelection fromFieldContexts(
      Iterable<? extends FieldContext> fieldContexts, final AppliesTo acceptablePathTypes)
  {
    ViewContext viewContext = null;
    Set<Path> selectedPaths = Sets.newLinkedHashSet();
    for (FieldContext fieldContext : fieldContexts) {
      if (!acceptablePathTypes.appliesTo(fieldContext.getPath()))
        return null;
      if (viewContext == null) {
        viewContext = fieldContext.getViewContext();
      } else {
        if (!viewContext.equals(fieldContext.getViewContext()))
          return null;
      }
      selectedPaths.add(fieldContext.getPath());
    }
    return viewContext == null ? null : new FieldSelection(viewContext, selectedPaths);
  }

  private static @Nullable FieldSelection fromLookupResultSnapshot(
      LookupResultSnapshot<FieldContext> lrs, final AppliesTo acceptablePathTypes)
  {
    return fromFieldContexts(lrs.allInstances(), acceptablePathTypes);
  }

  static Flow<FieldSelection> fromLookup(Lookup lookup, final AppliesTo acceptablePathTypes) {
    return Flows.newProcessed(UIFlows.newLookup(lookup, new Lookup.Template<>(FieldContext.class)),
        lrs -> fromLookupResultSnapshot(lrs, acceptablePathTypes));
  }

  @Override
  public String toString() {
    return "FieldSelection(" + viewContext + ", " + selectedPaths + ")";
  }
}
