// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.util;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.cell.styles.Text;
import com.sieuferd.cell.support.CursorStyle;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.layout.LabelFormatter;
import com.sieuferd.project.SieuferdIcons;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.project.actions.UnhideInfluencersAction;
import com.sieuferd.uiutil.SwingString;
import java.awt.Font;
import java.awt.font.TextAttribute;
import java.math.BigDecimal;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.text.BreakIterator;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.annotation.Nullable;

public final class QueryLabelFormatter extends LabelFormatter {
  private static final Logger LOG = Logger.getLogger(QueryLabelFormatter.class.getName());
  public static final QueryLabelFormatter INSTANCE = new QueryLabelFormatter();
  private static final boolean INCLUDE_FORMULA_PLURALITY_INDICATION = false;
  private static final boolean USE_SMALL_CAPS = true;
  /** Non-breaking space. */
  private static final String NBSP = "\u00a0";
  private static final Pattern ADDITIONAL_BREAKABLE_CHARACTERS_PATTERN = Pattern.compile("([_/]+)");
  // Use a font that is known to support subscript number characters on both MacOS and Windows.
  private static final Font BADGE_STD_FONT = new Font("Dialog", Font.PLAIN, 11);
  private static final Font BADGE_ICON_FONT =
      SieuferdIcons.getIconFont().deriveFont(BADGE_STD_FONT.getSize2D());
  private static final String ICON_FONT_SWING_HTML_TAG =
      "<font face=\"" + SwingString.toSwingHTML(BADGE_ICON_FONT.getFontName()) + "\">";

  private QueryLabelFormatter() {
    super(ImmutableSet.<ViewProperty<?>>builder()
        .addAll(Strepco.STREPCO_PROPERTIES)
        .add(CProp.PLURAL)
        .add(CProp.IN_TUPLE_KEY)
        .add(QProp.INCLUDE_PRIMARY_KEY)
        .add(QProp.FILTER)
        .add(QProp.EXISTENTIAL_FILTER)
        .add(QProp.ORDER_RANK)
        .add(QProp.ORDER_DESCENDING)
        .add(CProp.CANONICAL_LABEL)
        .add(QProp.JOINED_ON)
        .build());
  }

  @Override
  public Text getStyledLabel(View view, Path path, boolean rightAligned, String label) {
    final String badge = getBadge(view, path, false).trim();
    /* TODO[cc]: Consider changing the tabs into soft hyphen characters. They would have to
                 converted back into tabs before being fed to the LineBreakMeasurer, however. */
    /* To improve the display of technical label strings, allow line breaks on the underscore
    character. Keep the underscore at the beginning or the end of broken lines depending on
    horizontal alignment. This logic resides here rather than in TextStyle because the horizontal
    alignment would otherwise have to be taken into account when measuring text. This is also done
    for slashes, which might be used in human-readable labels. */
    label = ADDITIONAL_BREAKABLE_CHARACTERS_PATTERN
        .matcher(label).replaceAll(rightAligned ? "\t$1" : "$1\t");
    // TODO[pd]: Get rid of the ugly logic for disabling/enabling small caps.
    return new LabelText(label, badge, rightAligned,
        USE_SMALL_CAPS && label.equals(view.get(CProp.LABEL, path)));
  }

  private String getBadge(View view, Path path, boolean extended) {
    final StringBuilder ret = new StringBuilder();
    if (extended) {
      String canonicalLabel = view.get(CProp.CANONICAL_LABEL, path);
      if (!canonicalLabel.isEmpty() && !canonicalLabel.equals(view.get(CProp.LABEL, path)))
        ret.append("(").append(canonicalLabel).append(") ");
    }
    if (!path.isRelation() && view.get(QProp.SQL_COLUMN_NAME, path).isEmpty()) {
      // Formula "fx" icon bolder (Sieuf-Dingbats).
      ret.append("\ue617");
      if (INCLUDE_FORMULA_PLURALITY_INDICATION && view.get(CProp.IN_TUPLE_KEY, path)) {
        // Crow's foot one-to-many (Sieuf-Dingbats).
        ret.append("\ue602");
      }
    }
    if (path.isRelation() && view.get(CProp.PLURAL, path)) {
      // Crow's foot one-to-many (Sieuf-Dingbats).
      ret.append("\ue602");
    }
    if (path.isRelation() && !view.get(QProp.INCLUDE_PRIMARY_KEY, path)) {
      // Bracket symbol (Sieuf-Dingbats).
      ret.append("\ue607");
    }
    if (view.get(QProp.FILTER, path).isConstrained())
      // Filter symbol (Sieuf-Dingbats).
      ret.append("\ue603");
    if (!path.isRoot()) {
      if (path.isRelation() && view.get(QProp.EXISTENTIAL_FILTER, path)) {
        // Mathematical "exists" symbol (Sieuf-Dingbats).
        // ret.append("\ue604");
        // Up-left arrow (Sieuf-Dingbats).
        ret.append("\ue613");
      }
      final BigDecimal position = view.get(QProp.ORDER_RANK, path).getPosition();
      if (position != null && Queries.hasSignificantSort(view, path)) {
        final boolean descending = view.get(QProp.ORDER_DESCENDING, path);
        // Decreasing and increasing bars (Sieuf-Dingbats).
        //ret.append(descending ? "\ue60b" : "\ue60a");
        // Z-A and A-Z (Sieuf-Dingbats).
        //ret.append(descending ? "\ue609" : "\ue608");
        // Larger Z-A and A-Z (Sieuf-Dingbats).
        //ret.append(descending ? "\ue60d" : "\ue60c");
        // Larger decreasing and increasing bars (Sieuf-Dingbats).
        ret.append(descending ? "\ue60f" : "\ue60e");
        if (position.compareTo(BigDecimal.ONE) > 0) {
          try {
            ret.append(getSubscriptUnicodeNumber(position.intValueExact()));
          } catch (ArithmeticException e) {
            // Ignore.
          }
        }
      }
    }
    if (!path.isRelation()) {
      PrimitivePath joinedOn = view.get(QProp.JOINED_ON, path).getNullable();
      if (joinedOn != null) {
        if (extended && !ret.toString().endsWith(" "))
          ret.append(" ");
        //ret.append("=");
        // Join symbol (Sieuf-Dingbats).
        ret.append("\ue600");
        if (extended) {
          ret.append(Expressions.generateString(Strepco.create(view, path),
              new PathExpression(joinedOn)));
        }
      }
    }
    /* Only show this indication in non-extended mode, as the fields in question are not actually
    hidden in the extended views (Navigator, Fields toolbox tab). */
    if (!extended && UnhideInfluencersAction.isEnabledFor(view, path)) {
      // Dashed box symbol (Sieuf-Dingbats).
      ret.append("\ue605");
    }
    return ret.toString();
  }

  private static char getSubscriptUnicodeDigit(int d) {
    Preconditions.checkArgument(d >= 0 && d < 10);
    return (char) (0x2080 + d);
  }

  private static String getSubscriptUnicodeNumber(final int n) {
    if (n == 0)
      return Character.toString(getSubscriptUnicodeDigit(0));
    // Avoid the odd Math.abs overflow case by using a long here.
    long a = Math.abs((long) n);
    StringBuilder ret = new StringBuilder();
    while (a > 0) {
      ret.append(getSubscriptUnicodeDigit((int) (a % 10)));
      a /= 10;
    }
    if (n < 0)
      ret.append((char) 0x208b); // Subscript minus symbol
    return ret.reverse().toString();
  }

  private static boolean isBadgeIconCharacter(char c) {
    // TODO[pd]: Just check for the character range instead.
    return !Character.isWhitespace(c) && BADGE_ICON_FONT.canDisplay(c);
  }

  public SwingString getLabelSwingString(View view, Path path) {
    final String badge = getBadge(view, path, true).trim();
    StringBuilder ret = new StringBuilder();
    ret.append(SwingString.toSwingHTML(view.get(CProp.LABEL, path)));
    if (!badge.isEmpty()) {
      ret.append(" ");
      ret.append("<font color=\"").append(SwingString.htmlColor("controlShadow")).append("\">");
      boolean inBadgeFontTag = false;
      for (int i = 0; i < badge.length(); i++) {
        final char c = badge.charAt(i);
        final boolean badgeCharacter = isBadgeIconCharacter(c);
        if (inBadgeFontTag != badgeCharacter) {
          if (inBadgeFontTag)
            ret.append("</font>");
          if (badgeCharacter)
            ret.append(ICON_FONT_SWING_HTML_TAG);
          inBadgeFontTag = badgeCharacter;
        }
        ret.append(SwingString.toSwingHTML(Character.toString(c)));
      }
      if (inBadgeFontTag)
        ret.append("</font>");
      ret.append("</font>");
    }
    return SwingString.styled(ret.toString());
  }

  public static void main(String args[]) {
    for (int i = 0; i < 10; i++)
      System.out.print(getSubscriptUnicodeDigit(i));
    System.out.println();
    for (int i = -150; i < 150; i++) {
      System.out.println(getSubscriptUnicodeNumber(i));
    }
  }

  private static final class LabelText extends Text {
    private final String label;
    private final String badge;
    private final boolean rightAligned;
    private final boolean allowSmallCaps;

    public LabelText(String label, String badge, boolean rightAligned, boolean allowSmallCaps) {
      Preconditions.checkNotNull(label);
      Preconditions.checkNotNull(badge);
      Preconditions.checkNotNull(rightAligned);
      this.label = label;
      this.badge = badge;
      this.rightAligned = rightAligned;
      this.allowSmallCaps = allowSmallCaps;
    }

    private static void styleBadge(
        AttributedString as, String s, RenderContext rcx, int beginIndex, int endIndex)
    {
      as.addAttribute(TextAttribute.FONT, rcx.scaledFont(BADGE_STD_FONT), beginIndex, endIndex);
      as.addAttribute(TextAttribute.FOREGROUND,
          rcx.safeColor(CursorStyle.LABEL_BADGE_COLOR), beginIndex, endIndex);
      final Font scaledIconFont = rcx.scaledFont(BADGE_ICON_FONT);
      for (int i = beginIndex; i < endIndex; i++) {
        final char c = s.charAt(i);
        if (isBadgeIconCharacter(c))
          as.addAttribute(TextAttribute.FONT, scaledIconFont, i, i + 1);
      }
    }

    private void styleLabel(
        AttributedString as, String s, RenderContext rcx, int beginIndex, int endIndex)
    {
      Preconditions.checkNotNull(as);
      Preconditions.checkNotNull(s);
      Preconditions.checkNotNull(rcx);
      /* Always use the US locale here, as this logic exists to improve the display of technical
      identifiers, which are usually limited to ANSI characters anyway. */
      final Locale locale = Locale.US;
      if (allowSmallCaps && s.toUpperCase(locale).equals(s)) {
        final AttributedCharacterIterator aci = as.getIterator(null, beginIndex, endIndex);
        final BreakIterator breakIterator = BreakIterator.getLineInstance(locale);
        breakIterator.setText(aci);
        int breakStart = breakIterator.first();
        @Nullable Number startSize = (Number) aci.getAttribute(TextAttribute.SIZE);
        while (true) {
          final int breakEnd = breakIterator.next();
          if (breakEnd == BreakIterator.DONE)
            break;
          final @Nullable Number endSize = (Number) aci.getAttribute(TextAttribute.SIZE);
          /* Handle cases like "BUSINESS\t_AREA" (the tab being introduced because of
          right-alignment. */
          while (breakStart < breakEnd) {
            // Probably overkill, but do this properly in terms of codepoints and locales.
            String c = s.substring(
                breakStart, breakStart + Character.charCount(s.codePointAt(breakStart)));
            if (!c.toLowerCase(locale).equals(c))
              break;
            breakStart += c.length();
          }
          if (breakEnd - breakStart > 1 && startSize != null) {
            as.addAttribute(TextAttribute.SIZE, startSize.doubleValue() * 0.85,
                breakStart + 1, breakEnd);
          }
          breakStart = breakEnd;
          startSize = endSize;
        }
      }
    }


    @Override
    public String getString() {
      // TODO[cc]: Move this in with the logic in createAttributedString.
      if (badge.isEmpty()) {
        return label;
      } else if (label.isEmpty()) {
        // TODO[pd]: Handle this case in ParagraphLayout instead.
        // Include a NBSP in the label's font to ensure proper baseline alignment.
        return badge + NBSP;
      } else if (rightAligned) {
        return badge + NBSP + label;
      } else {
        return label + NBSP + badge;
      }
    }

    @Override
    public AttributedString createAttributedString(
        Map<TextAttribute,?> originalAttributes, RenderContext rcx)
    {
      Preconditions.checkState(!getString().isEmpty());
      String s = getString();
      AttributedString ret = new AttributedString(s, originalAttributes);
      if (badge.isEmpty()) {
        styleLabel(ret, s, rcx, 0, s.length());
      } else if (label.isEmpty()) {
        // badge + NBSP (see above)
        styleBadge(ret, s, rcx, 0, badge.length());
        styleLabel(ret, s, rcx, badge.length(), s.length());
      } else if (rightAligned) {
        // badge + NBSP + label
        styleBadge(ret, s, rcx, 0, badge.length());
        styleLabel(ret, s, rcx, badge.length() + 1, s.length());
      } else {
        // label + NBSP + badge
        styleLabel(ret, s, rcx, 0, label.length());
        styleBadge(ret, s, rcx, label.length() + 1, s.length());
      }
      return ret;
    }

    @Override
    public int hashCode() {
      return Objects.hashCode(label, badge, rightAligned);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof LabelText))
        return false;
      final LabelText other = (LabelText) obj;
      return this.label.equals(other.label)
          && this.badge.equals(other.badge)
          && this.rightAligned == other.rightAligned;
    }
  }
}
