// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.util;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.ViewProperty.AppliesTo;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.flow.Flows;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.project.ViewContext;
import com.sieuferd.upstream.UpstreamUtil;
import java.awt.event.ActionEvent;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;
import javax.swing.AbstractAction;
import javax.swing.Action;
import org.openide.util.ContextAwareAction;
import org.openide.util.Lookup;
import org.openide.util.Utilities;

@ThreadSafe
public abstract class ViewPropertyAction extends AbstractAction implements ContextAwareAction {
  private final SettableFlow<Lookup> lookupSettable = Flows.newSettable();
  private final Flow<SelectionSnapshot> selectionSnapshotFlow;
  private final AtomicBoolean contextSpecific = new AtomicBoolean();
  private final String defaultDisplayName;
  private final @Nullable String iconBase;
  private final AtomicBoolean initialized = new AtomicBoolean();

  /* TODO[cc]: Find a way to get rid of the horrible requirement for subclasses to call
               ensureBindingsInitialized. Calling it from SwingUtilities.invokeLater at the end of
               the superclass constructor does not cut it, since the latter may not be called from
               the EDT. */
  /**
   * Initialize the ViewPropertyAction. All non-abstract subclasses must be final and call
   * {@link #ensureBindingsInitialized()} at the end of their constructors.
   */
  public ViewPropertyAction(final String defaultDisplayName, final @Nullable String iconBase,
      final AppliesTo acceptablePathTypes, final Set<ViewProperty<?>> properties,
      final boolean trackSelectedPathOnly)
  {
    Preconditions.checkNotNull(defaultDisplayName);
    Preconditions.checkNotNull(acceptablePathTypes);
    Preconditions.checkNotNull(properties);
    this.defaultDisplayName = defaultDisplayName;
    this.iconBase = iconBase;
    lookupSettable.set(Utilities.actionsGlobalContext());
    if (iconBase != null)
      putValue(Action.SMALL_ICON, UpstreamUtil.loadImageIcon(iconBase));

    this.selectionSnapshotFlow = Flows.newIndirect(lookupSettable,
        lookup -> UIFlows.newWeak(SelectionSnapshot.fromLookup(
        lookup, acceptablePathTypes, properties, trackSelectedPathOnly)));
    setEnabled(false);
    putValue(Action.NAME, defaultDisplayName);
  }

  // Make setEnabled and putValue final to avoid warnings about calls from constructor.

  @Override
  public final void setEnabled(boolean newValue) {
    super.setEnabled(newValue);
  }

  @Override
  public final void putValue(String key, Object newValue) {
    super.putValue(key, newValue);
  }

  protected void ensureBindingsInitialized() {
    if (!initialized.getAndSet(true)) {
      UIFlows.bindEDT(selectionSnapshotFlow, (SelectionSnapshot value) -> {
        setEnabled(value != null && isEnabledFor(value));
        String dynamicDisplayName = value == null ? null : getDisplayNameFor(value);
        putValue(Action.NAME, dynamicDisplayName == null ? defaultDisplayName : dynamicDisplayName);
        updateActionState(value);
      });
    }
  }

  @Override
  public boolean isEnabled() {
    /* Try to avoid a flicker in enablement state that might occur if we waited for the invokeLater
    from the constructor to be processed. */
    ensureBindingsInitialized();
    return super.isEnabled();
  }

  @Override
  public Object getValue(String key) {
    // See above.
    ensureBindingsInitialized();
    return super.getValue(key);
  }

  public String getDefaultDisplayName() {
    return defaultDisplayName;
  }

  public @Nullable String getIconBase() {
    return iconBase;
  }

  public final Flow<SelectionSnapshot> getSelectionSnapshotFlow() {
    return selectionSnapshotFlow;
  }

  public final @Nullable SelectionSnapshot getSelectionSnapshot() {
    ensureBindingsInitialized();
    return selectionSnapshotFlow.get();
  }

  private void setLookup(Lookup lookup) {
    this.lookupSettable.set(lookup);
    contextSpecific.set(true);
  }

  public final boolean isContextSpecific() {
    return contextSpecific.get();
  }

  @Override
  public final void actionPerformed(ActionEvent e) {
    SelectionSnapshot selectionSnapshot = selectionSnapshotFlow.get();
    if (selectionSnapshot == null || !isEnabledFor(selectionSnapshot))
      return;
    actionPerformed(e, selectionSnapshot.getViewContext(),
        ImmutableSet.copyOf(selectionSnapshot.getSelectedPaths()));
  }

  @Override
  public final Action createContextAwareInstance(Lookup actionContext) {
    ViewPropertyAction ret = create();
    Preconditions.checkNotNull(ret);
    ret.setLookup(actionContext);
    /* TODO[pd]: Consider attaching a PropertyChangeListener on the original action to keep this up
                 to date (without interfering with updates from this instance). */
    // Copy over any properties set, for instance, by AcceleratorBinding.setAccelerator.
    Object keys[] = getKeys();
    if (keys != null) {
      for (Object key : keys) {
        if (key instanceof String)
          ret.putValue((String) key, getValue((String) key));
      }
    }
    return ret;
  }

  protected abstract ViewPropertyAction create();

  protected abstract void actionPerformed(
      ActionEvent e, ViewContext viewContext, ImmutableSet<Path> selectedPaths);

  /**
   * May be implemented by subclasses to generate a custom display name for the given
   * SelectionSnapshot.
   *
   * @param selectionSnapshot will never be null
   * @return can be null to use the default display name
   */
  protected @Nullable String getDisplayNameFor(SelectionSnapshot selectionSnapshot) {
    return null;
  }

  /**
   * May be implemented by subclasses to customize enablement for the given SelectionSnapshot.
   *
   * @param selectionSnapshot will never be null
   */
  protected boolean isEnabledFor(SelectionSnapshot selectionSnapshot) {
    return true;
  }

  /**
   * May be implemented by subclasses to update action state other than enablement or display name
   * in response to a change in the selectionSnapshot. Will only be called on the EDT.
   */
  protected void updateActionState(@Nullable SelectionSnapshot selectionSnapshot) {
    // Empty by default.
  }
}
