// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.viewfile;

import com.google.common.base.Preconditions;
import com.sieuferd.project.SieuferdProject;
import com.sieuferd.util.StringUtil;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLStreamHandler;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.annotation.Nullable;
import org.netbeans.api.project.FileOwnerQuery;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.URLMapper;
import org.openide.util.URLStreamHandlerRegistration;
import org.openide.util.lookup.ServiceProvider;

// TODO[cc]: Consider using an IOException convention instead of returning null everywhere.
public final class ViewURLComponents {
  private static final String PROTOCOL = "sieuferd";
  private static final String HOST = "views";
  private static final String FILE_OBJECT_ATTRIBUTE_NAME = "sieuferdViewURLComponents";
  private static final Pattern PATH_PATTERN = Pattern.compile("^/([^/]*)/([^/]*)$");
  private final URI projectDirectoryURI;
  private final ViewIdentifier viewIdentifier;

  public ViewURLComponents(URI projectDirectoryURI, ViewIdentifier viewIdentifier) {
    Preconditions.checkNotNull(projectDirectoryURI);
    Preconditions.checkNotNull(viewIdentifier);
    this.projectDirectoryURI = projectDirectoryURI.normalize();
    this.viewIdentifier = viewIdentifier;
  }

  @Override
  public String toString() {
    return "ViewURLComponents(\"" + getProjectURI() + "\", \"" + getViewIdentifier() + "\")";
  }

  public URI getProjectURI() {
    return projectDirectoryURI;
  }

  public ViewIdentifier getViewIdentifier() {
    return viewIdentifier;
  }

  private boolean isEqualTo(@Nullable ViewURLComponents other) {
    if (other == null)
      return false;
    return getProjectURI().equals(other.getProjectURI())
        && getViewIdentifier().equals(other.getViewIdentifier());
  }

  public static @Nullable ViewURLComponents fromPreparedFileObject(FileObject fo) {
    Object attr = fo.getAttribute(FILE_OBJECT_ATTRIBUTE_NAME);
    if (!(attr instanceof ViewURLComponents))
      return null;
    return (ViewURLComponents) attr;
  }

  public void prepareFileObject(FileObject fo) throws IOException {
    fo.setAttribute(FILE_OBJECT_ATTRIBUTE_NAME, this);
    fo.setAttribute("mimeType", ViewDataObject.MIME_TYPE);
    SieuferdProject project = getProject();
    if (project != null)
      FileOwnerQuery.markExternalOwner(fo, project, FileOwnerQuery.EXTERNAL_ALGORITHM_TRANSIENT);
  }

  public @Nullable SieuferdProject getProject() {
    Project ret = FileOwnerQuery.getOwner(getProjectURI());
    if (!(ret instanceof SieuferdProject))
      return null;
    return (SieuferdProject) ret;
  }

  public URL toURL() {
    final URL ret;
    try {
      // Use URI.toString() rather than URI.toASCIIString() to ensure a reversible transform.
      final String projectDirectoryURIString = projectDirectoryURI.toString();
      ret = new URL(null, PROTOCOL + "://" + HOST +
          "/" + StringUtil.urlEncode(projectDirectoryURIString) +
          "/" + StringUtil.urlEncode(viewIdentifier.toString()), new ViewURLStreamHandler());
    } catch (MalformedURLException e) {
      throw new AssertionError(e);
    }
    ViewURLComponents reverse = ViewURLComponents.fromURL(ret);
    if (!isEqualTo(reverse))
      throw new AssertionError("Original " + this + " differed from re-parsed " + reverse);
    return ret;
  }

  public static @Nullable ViewURLComponents fromURL(URL url) {
    if (!PROTOCOL.equals(url.getProtocol()))
      return null;
    if (!HOST.equals(url.getHost()))
      return null;
    Matcher matcher = PATH_PATTERN.matcher(url.getPath());
    if (!matcher.matches())
      return null;
    URI projectDirectoryURI;
    try {
      projectDirectoryURI = new URI(StringUtil.urlDecode(matcher.group(1)));
    } catch (URISyntaxException e) {
      return null;
    }
    ViewIdentifier viewIdentifier =
        ViewIdentifier.fromStringRepresentation(StringUtil.urlDecode(matcher.group(2)));
    return new ViewURLComponents(projectDirectoryURI, viewIdentifier);
  }

  /* ============================================================================================ */

  // Provide a position to ensure precedence over the URLMapper for MemoryFileSystem.
  @ServiceProvider(service=URLMapper.class, position = -100)
  public static class ViewURLMapper extends URLMapper {
    @Override
    public FileObject[] getFileObjects(URL url) {
      ViewURLComponents components = ViewURLComponents.fromURL(url);
      if (components == null)
        return null;
      SieuferdProject project = components.getProject();
      if (project == null)
        return null;
      FileObject ret =
          project.getViewSet().getFileObjectForViewIdentifier(components.getViewIdentifier());
      return (ret == null) ? null : new FileObject[] { ret };
    }

    @Override
    public URL getURL(FileObject fo, int type) {
      ViewURLComponents ret = ViewURLComponents.fromPreparedFileObject(fo);
      return (ret == null) ? null : ret.toURL();
    }
  }

  @URLStreamHandlerRegistration(protocol=PROTOCOL)
  public static final class ViewURLStreamHandler extends URLStreamHandler {
    @Override
    protected URLConnection openConnection(URL url) throws IOException {
      return new ViewURLConnection(url);
    }
  }

  private static final class ViewURLConnection extends URLConnection {
    ViewURLConnection(URL url) {
      super(url);
    }

    @Override
    public void connect() throws IOException {
      // No action.
    }
  }
}
