// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.project.formulabar;

import com.google.common.base.Preconditions;
import com.sieuferd.project.formulabar.ReferenceInserter.InsertionParameters;
import java.util.Objects;
import javax.annotation.concurrent.Immutable;
import javax.swing.text.AbstractDocument;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;
import static org.junit.Assert.assertEquals;
import org.junit.Test;
import org.netbeans.api.lexer.Language;
import org.netbeans.api.lexer.TokenHierarchy;
import org.netbeans.api.lexer.TokenSequence;

public class ReferenceInserterTest {
  private TokenSequence<ExpressionTokenId> getTokenSequence(String s) {
    final AbstractDocument d = new PlainDocument();
    d.putProperty(Language.class, ExpressionLanguageConfig.LANGUAGE);

    try {
      d.insertString(0, s, null);
    } catch (BadLocationException e) {
      throw new RuntimeException(e);
    }
    // Don't bother unlocking in the test code.
    d.readLock();
    final TokenHierarchy<?> h = TokenHierarchy.get(d);
    return h.tokenSequence(ExpressionLanguageConfig.LANGUAGE);
  }

  @Immutable
  private static final class TextWithCaret {
    final String text;
    final int caretStart;
    final int caretEnd;

    public TextWithCaret(String parseableString, int caretStart, int caretEnd) {
      this(parseableString, caretStart, caretEnd, true);
    }

    private TextWithCaret(String parseableString, int caretStart, int caretEnd, boolean check) {
      Preconditions.checkNotNull(parseableString);
      Preconditions.checkArgument(caretStart >= 0);
      Preconditions.checkArgument(caretEnd >= caretStart);
      Preconditions.checkArgument(caretEnd <= parseableString.length());
      this.text = parseableString;
      this.caretStart = caretStart;
      this.caretEnd = caretEnd;

      if (check && !equalsTextWithCaret(fromEncoded(toString())))
        throw new AssertionError();
    }

    public boolean equalsTextWithCaret(TextWithCaret other) {
      return text.equals(other.text)
          && caretStart == other.caretStart
          && caretEnd == other.caretEnd;
    }

    @Override
    public int hashCode() {
      return Objects.hash(text, caretStart, caretEnd);
    }

    @Override
    public boolean equals(Object obj) {
      if (!(obj instanceof TextWithCaret))
        return false;
      return equalsTextWithCaret((TextWithCaret) obj);
    }

    public static TextWithCaret fromEncoded(String encoded) {
      final StringBuilder psb = new StringBuilder();
      Integer caretStart = null;
      Integer caretEnd = null;
      for (char ec : encoded.toCharArray()) {
        if (ec == '<') {
          if (caretStart != null)
            throw new AssertionError();
          caretStart = psb.length();
        } else if (ec == '>') {
          if (caretEnd != null)
            throw new AssertionError();
          caretEnd = psb.length();
        } else {
          psb.append(ec);
        }
      }
      if (caretStart == null || caretEnd == null)
        throw new AssertionError();
      return new TextWithCaret(psb.toString(), caretStart, caretEnd, false);
    }

    @Override
    public String toString() {
      return
          text.substring(0, caretStart) + '<' +
          text.substring(caretStart, caretEnd) + '>' +
          text.substring(caretEnd);
    }
  }

  @Test
  public void testTestMethods() {
    // The Document adds a whitespace token at the end of every string.
    assertEquals(3, getTokenSequence("=9").tokenCount());
    assertEquals(8, getTokenSequence("dette er en test").tokenCount());
    TextWithCaret.fromEncoded("ajsdlka sdjfkl<>asdf");
    TextWithCaret.fromEncoded("<ajsdlka sdjfkl asdf>");
    TextWithCaret.fromEncoded("<ajsdlka sdjfkl asd>f");
    TextWithCaret.fromEncoded("a<jsdlka sdjfkl asd>f");
    TextWithCaret.fromEncoded("ajsdlka sdjfkl asdf<>");
    TextWithCaret.fromEncoded("<>ajsdlka sdjfkl asdf");
    TextWithCaret.fromEncoded("a<>jsdlka sdjfkl asdf");
    TextWithCaret.fromEncoded("ajsdlka sdjfkl asd<>f");
  }

  private void testBoundary(String before, String expectedAfter, boolean forward) {
    TextWithCaret beforeTWC = TextWithCaret.fromEncoded(before);
    TextWithCaret expectedAfterTWC = TextWithCaret.fromEncoded(expectedAfter);
    Preconditions.checkArgument(beforeTWC.caretStart == beforeTWC.caretEnd);
    final int caretPos = beforeTWC.caretStart;
    final int res = Math.min(beforeTWC.text.length(), ReferenceInserter.extendInsertionPoint(
        getTokenSequence(beforeTWC.text), caretPos, forward));
    assertEquals(expectedAfterTWC, new TextWithCaret(beforeTWC.text, res, res));
  }

  @Test
  public void testExtendToTokenBoundaryForward() {
    testBoundary("1234 56  7890<>", "1234 56  7890<>", true);
    testBoundary("1234 56  789<>0", "1234 56  7890<>", true);
    testBoundary("1234 56  78<>90", "1234 56  7890<>", true);
    testBoundary("1234 56  7<>890", "1234 56  7890<>", true);
    testBoundary("1234 56  <>7890", "1234 56  <>7890", true);
    testBoundary("1234 56 <> 7890", "1234 56 <> 7890", true);
    testBoundary("1234 56<>  7890", "1234 56<>  7890", true);
    testBoundary("1234 5<>6  7890", "1234 56<>  7890", true);
    testBoundary("1234 <>56  7890", "1234 <>56  7890", true);
    testBoundary("1234<> 56  7890", "1234<> 56  7890", true);
    testBoundary("123<>4 56  7890", "1234<> 56  7890", true);
    testBoundary("12<>34 56  7890", "1234<> 56  7890", true);
    testBoundary("1<>234 56  7890", "1234<> 56  7890", true);
    testBoundary("<>1234 56  7890", "<>1234 56  7890", true);

    testBoundary("[1234] [56]  [7890]<>", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56]  [7890<>]", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56]  [789<>0]", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56]  [78<>90]", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56]  [7<>890]", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56]  [<>7890]", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56]  <>[7890]", "[1234] [56]  [7890]<>", true);
    testBoundary("[1234] [56] <> [7890]", "[1234] [56] <> [7890]", true);
    testBoundary("[1234] [56]<>  [7890]", "[1234] [56]<>  [7890]", true);
    testBoundary("[1234] [56<>]  [7890]", "[1234] [56]<>  [7890]", true);
    testBoundary("[1234] [5<>6]  [7890]", "[1234] [56]<>  [7890]", true);
    testBoundary("[1234] [<>56]  [7890]", "[1234] [56]<>  [7890]", true);
    testBoundary("[1234] <>[56]  [7890]", "[1234] [56]<>  [7890]", true);
    testBoundary("[1234]<> [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("[1234<>] [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("[123<>4] [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("[12<>34] [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("[1<>234] [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("[<>1234] [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("<>[1234] [56]  [7890]", "[1234]<> [56]  [7890]", true);
    testBoundary("[1234] <>[56fa", "[1234] [56fa<>", true);

    testBoundary("[1234] [5] <> [7890]", "[1234] [5] <> [7890]", true);
    testBoundary("[1234] [5]<>  [7890]", "[1234] [5]<>  [7890]", true);
    testBoundary("[1234] [5<>]  [7890]", "[1234] [5]<>  [7890]", true);
    testBoundary("[1234] [<>5]  [7890]", "[1234] [5]<>  [7890]", true);
    testBoundary("[1234] <>[5]  [7890]", "[1234] [5]<>  [7890]", true);
    testBoundary("[1234]<> [5]  [7890]", "[1234]<> [5]  [7890]", true);

    testBoundary("[1234] 'fobar<>", "[1234] 'fobar<>", true);
    testBoundary("[1234] 'foba<>r", "[1234] 'fobar<>", true);
    testBoundary("[1234] 'fob<>ar", "[1234] 'fobar<>", true);
    testBoundary("[1234] 'f<>obar", "[1234] 'fobar<>", true);
    testBoundary("[1234] '<>fobar", "[1234] 'fobar<>", true);
    testBoundary("[1234] <>'fobar", "[1234] <>'fobar", true);
    testBoundary("[1234]<> 'fobar", "[1234]<> 'fobar", true);

    testBoundary("<>[fo", "[fo<>", true);
    testBoundary("[<>fo", "[fo<>", true);
    testBoundary("[f<>o", "[fo<>", true);
    testBoundary("[fo<>", "[fo<>", true);
    testBoundary("<>[f", "[f<>", true);
    testBoundary("[<>f", "[f<>", true);
    testBoundary("[f<>", "[f<>", true);
    testBoundary("<>[", "[<>", true);
    testBoundary("[<>", "[<>", true);
    testBoundary("a<>[fo", "a[fo<>", true);
    testBoundary("a[<>fo", "a[fo<>", true);
    testBoundary("a[f<>o", "a[fo<>", true);
    testBoundary("a[fo<>", "a[fo<>", true);
    testBoundary("a<>[f", "a[f<>", true);
    testBoundary("a[<>f", "a[f<>", true);
    testBoundary("a[f<>", "a[f<>", true);
    testBoundary("a<>[", "a[<>", true);
    testBoundary("a[<>", "a[<>", true);

    testBoundary("[1234] foo(567,<> 89", "[1234] foo(567,<> 89", true);
    testBoundary("[1234] foo(567<>, 89", "[1234] foo(567<>, 89", true);
    testBoundary("[1234] foo(56<>7, 89", "[1234] foo(567<>, 89", true);
    testBoundary("[1234] foo(5<>67, 89", "[1234] foo(567<>, 89", true);
    testBoundary("[1234] foo(<>567, 89", "[1234] foo(<>567, 89", true);
    testBoundary("[1234] foo<>(567, 89", "[1234] foo(<>567, 89", true);
    testBoundary("[1234] fo<>o(567, 89", "[1234] foo(<>567, 89", true);
    testBoundary("[1234] f<>oo(567, 89", "[1234] foo(<>567, 89", true);
    testBoundary("[1234] <>foo(567, 89", "[1234] <>foo(567, 89", true);
    testBoundary("[1234]<> foo(567, 89", "[1234]<> foo(567, 89", true);

    testBoundary("[1234] foo 567,<> 89", "[1234] foo 567,<> 89", true);
    testBoundary("[1234] foo 567<>, 89", "[1234] foo 567<>, 89", true);
    testBoundary("[1234] foo 56<>7, 89", "[1234] foo 567<>, 89", true);
    testBoundary("[1234] foo 5<>67, 89", "[1234] foo 567<>, 89", true);
    testBoundary("[1234] foo <>567, 89", "[1234] foo <>567, 89", true);
    testBoundary("[1234] foo<> 567, 89", "[1234] foo<> 567, 89", true);
    testBoundary("[1234] fo<>o 567, 89", "[1234] foo<> 567, 89", true);
    testBoundary("[1234] f<>oo 567, 89", "[1234] foo<> 567, 89", true);
    testBoundary("[1234] <>foo 567, 89", "[1234] <>foo 567, 89", true);
    testBoundary("[1234]<> foo 567, 89", "[1234]<> foo 567, 89", true);
  }

  @Test
  public void testExtendToTokenBoundaryBackward() {
    testBoundary("1234 56  7890<>", "1234 56  7890<>", false);
    testBoundary("1234 56  789<>0", "1234 56  <>7890", false);
    testBoundary("1234 56  78<>90", "1234 56  <>7890", false);
    testBoundary("1234 56  7<>890", "1234 56  <>7890", false);
    testBoundary("1234 56  <>7890", "1234 56  <>7890", false);
    testBoundary("1234 56 <> 7890", "1234 56 <> 7890", false);
    testBoundary("1234 56<>  7890", "1234 56<>  7890", false);
    testBoundary("1234 5<>6  7890", "1234 <>56  7890", false);
    testBoundary("1234 <>56  7890", "1234 <>56  7890", false);
    testBoundary("1234<> 56  7890", "1234<> 56  7890", false);
    testBoundary("123<>4 56  7890", "<>1234 56  7890", false);
    testBoundary("12<>34 56  7890", "<>1234 56  7890", false);
    testBoundary("1<>234 56  7890", "<>1234 56  7890", false);
    testBoundary("<>1234 56  7890", "<>1234 56  7890", false);

    testBoundary("[1234] [56]  [7890]<>", "[1234] [56]  [7890]<>", false);
    testBoundary("[1234] [56]  [7890<>]", "[1234] [56]  <>[7890]", false);
    testBoundary("[1234] [56]  [789<>0]", "[1234] [56]  <>[7890]", false);
    testBoundary("[1234] [56]  [78<>90]", "[1234] [56]  <>[7890]", false);
    testBoundary("[1234] [56]  [7<>890]", "[1234] [56]  <>[7890]", false);
    testBoundary("[1234] [56]  [<>7890]", "[1234] [56]  <>[7890]", false);
    testBoundary("[1234] [56]  <>[7890]", "[1234] [56]  <>[7890]", false);
    testBoundary("[1234] [56] <> [7890]", "[1234] [56] <> [7890]", false);
    testBoundary("[1234] [56]<>  [7890]", "[1234] [56]<>  [7890]", false);
    testBoundary("[1234] [56<>]  [7890]", "[1234] <>[56]  [7890]", false);
    testBoundary("[1234] [5<>6]  [7890]", "[1234] <>[56]  [7890]", false);
    testBoundary("[1234] [<>56]  [7890]", "[1234] <>[56]  [7890]", false);
    testBoundary("[1234] <>[56]  [7890]", "[1234] <>[56]  [7890]", false);
    testBoundary("[1234]<> [56]  [7890]", "[1234]<> [56]  [7890]", false);
    testBoundary("[1234<>] [56]  [7890]", "<>[1234] [56]  [7890]", false);
    testBoundary("[123<>4] [56]  [7890]", "<>[1234] [56]  [7890]", false);
    testBoundary("[12<>34] [56]  [7890]", "<>[1234] [56]  [7890]", false);
    testBoundary("[1<>234] [56]  [7890]", "<>[1234] [56]  [7890]", false);
    testBoundary("[<>1234] [56]  [7890]", "<>[1234] [56]  [7890]", false);
    testBoundary("<>[1234] [56]  [7890]", "<>[1234] [56]  [7890]", false);

    testBoundary("[1234] [5] <> [7890]", "[1234] [5] <> [7890]", false);
    testBoundary("[1234] [5]<>  [7890]", "[1234] [5]<>  [7890]", false);
    testBoundary("[1234] [5<>]  [7890]", "[1234] <>[5]  [7890]", false);
    testBoundary("[1234] [<>5]  [7890]", "[1234] <>[5]  [7890]", false);
    testBoundary("[1234] <>[5]  [7890]", "[1234] <>[5]  [7890]", false);
    testBoundary("[1234]<> [5]  [7890]", "[1234]<> [5]  [7890]", false);

    testBoundary("[1234] 'fobar<>", "[1234] 'fobar<>", false);
    testBoundary("[1234] 'foba<>r", "[1234] <>'fobar", false);
    testBoundary("[1234] 'fob<>ar", "[1234] <>'fobar", false);
    testBoundary("[1234] 'f<>obar", "[1234] <>'fobar", false);
    testBoundary("[1234] '<>fobar", "[1234] <>'fobar", false);
    testBoundary("[1234] <>'fobar", "[1234] <>'fobar", false);
    testBoundary("[1234]<> 'fobar", "[1234]<> 'fobar", false);

    testBoundary("<>[fo", "<>[fo", false);
    testBoundary("[<>fo", "<>[fo", false);
    testBoundary("[f<>o", "<>[fo", false);
    testBoundary("[fo<>", "<>[fo", false);
    testBoundary("<>[f", "<>[f", false);
    testBoundary("[<>f", "<>[f", false);
    testBoundary("[f<>", "<>[f", false);
    testBoundary("<>[", "<>[", false);
    testBoundary("[<>", "<>[", false);
    testBoundary("a<>[fo", "a<>[fo", false);
    testBoundary("a[<>fo", "a<>[fo", false);
    testBoundary("a[f<>o", "a<>[fo", false);
    testBoundary("a[fo<>", "a<>[fo", false);
    testBoundary("a<>[f", "a<>[f", false);
    testBoundary("a[<>f", "a<>[f", false);
    testBoundary("a[f<>", "a<>[f", false);
    testBoundary("a<>[", "a<>[", false);
    testBoundary("a[<>", "a<>[", false);
  }

  private void testInsertion(String before, String expectedAfter, String expectedInsertPreceeding) {
    System.out.println("Testing string " + before);
    TextWithCaret beforeTWC = TextWithCaret.fromEncoded(before);
    TextWithCaret expectedAfterTWC = TextWithCaret.fromEncoded(expectedAfter);
    TokenSequence<ExpressionTokenId> tokenSequence = getTokenSequence(beforeTWC.text);

    InsertionParameters res = ReferenceInserter.calculateInsertionParameters(
        tokenSequence, beforeTWC.caretStart, beforeTWC.caretEnd);

    assertEquals(expectedAfterTWC, new TextWithCaret(beforeTWC.text,
        res.interval.start, Math.min(res.interval.end, beforeTWC.text.length())));
    assertEquals(expectedInsertPreceeding, res.insertPreceeding);
  }

  @Test
  public void testInsertionParameterCalculation() {
    testInsertion("=<>", "=<>", "");
    testInsertion("=foo<>", "=foo<>", "(");
    testInsertion("='fooo<>", "='fooo<>", "', ");
    testInsertion("=\"fooo<>", "=\"fooo<>", "\", ");
    testInsertion("=523<>", "=523<>", ", ");
    testInsertion("=52<>3", "=523<>", ", ");
    testInsertion("=<>523", "=<>523", "");
    testInsertion("=null<>", "=<null>", "");
    testInsertion("=nul<l>", "=<null>", "");
    testInsertion("=#ref!<>", "=<#ref!>", "");
    testInsertion("='toba'<>", "='toba'<>", ", ");
    testInsertion("=foo<>bar", "=foobar<>", "(");
    testInsertion("=foo<b>ar", "=<foobar>", "");
    testInsertion("=23523 + 2<34 - 2>31 + 678", "=23523 + <234 - 231> + 678", "");
    testInsertion("=23523 <+ 234 - 2>31 + 678", "=23523 <+ 234 - 231> + 678", ", ");
    testInsertion("=sum<>", "=sum<>", "(");
    testInsertion("=sum(<>", "=sum(<>", "");
    testInsertion("=sum([hallo] + <>", "=sum([hallo] + <>", "");
    testInsertion("=sum([hallo]  <>", "=sum(<[hallo]  >", "");
    testInsertion("=sum([hallo] <>", "=sum(<[hallo] >", "");
    testInsertion("=sum([hallo]<>", "=sum(<[hallo]>", "");
    testInsertion("=sum([hallo<>]", "=sum(<[hallo]>", "");
    testInsertion("=sum([hall<>o]", "=sum(<[hallo]>", "");
    testInsertion("=sum([ha<>llo]", "=sum(<[hallo]>", "");
    testInsertion("=sum([h<>allo]", "=sum(<[hallo]>", "");
    testInsertion("=sum([<>hallo]", "=sum(<[hallo]>", "");
    testInsertion("=sum(<>[hallo]", "=sum(<[hallo]>", "");
    testInsertion("=sum([ha<ll>o]", "=sum(<[hallo]>", "");
    testInsertion("=sum([ha<llo]>", "=sum(<[hallo]>", "");
    testInsertion("=sum([<hallo]>", "=sum(<[hallo]>", "");
    testInsertion("=sum(<[hallo]>", "=sum(<[hallo]>", "");
    testInsertion("=sum<([hallo]>", "=sum<([hallo]>", "(");
    testInsertion("=su<m([hallo]>", "=<sum([hallo]>", "");
    testInsertion("=su<>m([hallo]", "=sum(<>[hallo]", "");

    testInsertion("=hello([<>", "=hello(<[>", "");
    testInsertion("=hello([f<>", "=hello(<[f>", "");
    testInsertion("=hello([fo<>", "=hello(<[fo>", "");
    testInsertion("=hello([foo<>", "=hello(<[foo>", "");
    testInsertion("=hello(<[>", "=hello(<[>", "");
    testInsertion("=hello([<f>", "=hello(<[f>", "");
    testInsertion("=hello([f<o>", "=hello(<[fo>", "");
    testInsertion("=hello([<foo>", "=hello(<[foo>", "");
    testInsertion("=hello(<[foo>", "=hello(<[foo>", "");
    testInsertion("=hello(<[fo>o", "=hello(<[foo>", "");
    testInsertion("=hello([<fo>o", "=hello(<[foo>", "");
    testInsertion("=hello([f<o>o", "=hello(<[foo>", "");
    testInsertion("=hello([<f>oo", "=hello(<[foo>", "");

    testInsertion("=hello(<[hello[foo>", "=hello(<[hello[foo>", "");
    testInsertion("=hello([<hello[foo>", "=hello(<[hello[foo>", "");
    testInsertion("=hello([h<ello[foo>", "=hello(<[hello[foo>", "");
    testInsertion("=hello([hel<lo[foo>", "=hello(<[hello[foo>", "");
    testInsertion("=hello([hell<o[foo>", "=hello(<[hello[foo>", "");
    testInsertion("=hello([hello<[foo>", "=hello(<[hello[foo>", "");
    testInsertion("=hello([hello[<foo>", "=hello(<[hello[foo>", "");

    testInsertion("=hello(<[hello\nfoo>", "=hello(<[hello\nfoo>", "");
    testInsertion("=hello([<hello\nfoo>", "=hello(<[hello\nfoo>", "");
    testInsertion("=hello([h<ello\nfoo>", "=hello(<[hello\nfoo>", "");
    testInsertion("=hello([hel<lo\nfoo>", "=hello(<[hello\nfoo>", "");
    testInsertion("=hello([hell<o\nfoo>", "=hello(<[hello\nfoo>", "");
    testInsertion("=hello([hello<\nfoo>", "=hello(<[hello\nfoo>", "");
    testInsertion("=hello([hello\n<foo>", "=hello(<[hello\nfoo>", "");

    // Auto-complete unterminated temporal literals.
    testInsertion("={<>", "={<>", "}, ");
    testInsertion("={fooo<>", "={fooo<>", "}, ");

    // Automatically inserted spaces.
    testInsertion("=sum([hallo] +<>", "=sum([hallo] +<>", " ");
    testInsertion("=5 +<>", "=5 +<>", " ");
    testInsertion("=5 =<>", "=5 =<>", " ");
    testInsertion("=<>", "=<>", "");
    testInsertion("=<>\n", "=<>\n", "");
    testInsertion("=<>\n", "=<>\n", "");
    testInsertion("=<> \n", "=<> \n", "");
    testInsertion("<>", "<>", "");
    testInsertion("=5 !<>", "=5 !<>", "");
    testInsertion("=5 not<>", "=5 not<>", " ");
    testInsertion("=5 and<>", "=5 and<>", " ");
    testInsertion("=5 ?<>", "=5 ?<>", " ");
    testInsertion("=5 :<>", "=5 :<>", " ");
    testInsertion("=5 ,<>", "=5 ,<>", " ");
    testInsertion("=5 (<>", "=5 (<>", "");
    testInsertion("=5 )<>", "=5 )<>", ", ");
  }
}
