// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.sieuferd.flow.Flow;
import com.sieuferd.uiutil.flow.UIFlows;
import java.util.Set;
import javax.annotation.Nullable;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import org.openide.awt.UndoRedo;

/**
 * An {@link UndoRedo} implementation that delegates to another settable {@code UndoRedo}.
 * Thread-safe if the delegate is thread-safe.
 */
public final class DelegatingUndoRedo implements UndoRedo {
  private final Set<ChangeListener> changeListeners = Sets.newLinkedHashSet();
  private final ChangeListener delegateChangeListener = (ChangeEvent e) -> {
    fireStateChanged();
  };
  private @Nullable UndoRedo delegate;

  public DelegatingUndoRedo() {
  }

  public void setDelegate(@Nullable UndoRedo delegate) {
    synchronized (this) {
      if (this.delegate == delegate)
        return;
      if (this.delegate != null)
        this.delegate.removeChangeListener(delegateChangeListener);
      this.delegate = delegate;
      if (this.delegate != null)
        this.delegate.addChangeListener(delegateChangeListener);
    }
    fireStateChanged();
  }

  @Override
  public synchronized void addChangeListener(ChangeListener listener) {
    changeListeners.add(listener);
  }

  @Override
  public synchronized void removeChangeListener(ChangeListener listener) {
    changeListeners.remove(listener);
  }

  private void fireStateChanged() {
    Preconditions.checkState(!Thread.holdsLock(this));
    final Set<ChangeListener> changeListenersSnapshot;
    synchronized (this) {
      changeListenersSnapshot = ImmutableSet.copyOf(changeListeners);
    }
    final ChangeEvent evt = new ChangeEvent(this);
    for (ChangeListener listener : changeListenersSnapshot)
      listener.stateChanged(evt);
  }

  private synchronized UndoRedo getEffectiveDelegate() {
    return delegate == null ? UndoRedo.NONE : delegate;
  }

  @Override
  public boolean canUndo() {
    return getEffectiveDelegate().canUndo();
  }

  @Override
  public boolean canRedo() {
    return getEffectiveDelegate().canRedo();
  }

  @Override
  public void undo() throws CannotUndoException {
    getEffectiveDelegate().undo();
  }

  @Override
  public void redo() throws CannotRedoException {
    getEffectiveDelegate().redo();
  }

  @Override
  public String getUndoPresentationName() {
    return getEffectiveDelegate().getUndoPresentationName();
  }

  @Override
  public String getRedoPresentationName() {
    return getEffectiveDelegate().getRedoPresentationName();
  }

  /**
   * Create an {@code UndoRedo} that delegates to whatever instance resides in {@code undoRedoFlow}.
   */
  public static UndoRedo createFromFlow(Flow<UndoRedo> undoRedoFlow) {
    final DelegatingUndoRedo ret = new DelegatingUndoRedo();
    UIFlows.bindEDT(undoRedoFlow, ret::setDelegate);
    return ret;
  }
}
