// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil;

import com.google.common.collect.ImmutableMap;
import java.awt.BasicStroke;
import java.awt.CardLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.LayoutManager;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.geom.Line2D;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.Timer;

/* TODO[pd]: Consider using this in other places where I currently use an indeterminate
             JProgressBar. */
public final class ProgressSpinner extends JPanel {
  private static final String CARD_INVISIBLE = "progressInvisible";
  private static final String CARD_VISIBLE = "progressVisible";
  private final JComponent impl;
  private final CardLayout cardLayout;
  private boolean running;

  public ProgressSpinner() {
    cardLayout = new CardLayout();
    setLayout(cardLayout);
    setOpaque(false);
    JPanel hiddenPanel = new JPanel();
    hiddenPanel.setOpaque(false);
    add(hiddenPanel, CARD_INVISIBLE);
    if (UIUtil.isAqua()) {
      impl = new JProgressBar();
      impl.putClientProperty("JProgressBar.style", "circular");
    } else {
      impl = new ProgressSpinnerImpl();
    }
    add(impl, CARD_VISIBLE);
    setRunning(true);
    setRunning(false);
  }

  @Override
  public void setLayout(LayoutManager layoutManager) {
    if (cardLayout == null || layoutManager == cardLayout)
      super.setLayout(layoutManager);
  }

  public void setRunning(boolean running) {
    // TODO[pd]: Consider automatically waiting for a short moment before showing the spinner.
    if (running != this.running) {
      this.running = running;
      if (impl instanceof ProgressSpinnerImpl)
        ((ProgressSpinnerImpl) impl).setRunning(running);
      if (impl instanceof JProgressBar)
        ((JProgressBar) impl).setIndeterminate(running);
      cardLayout.show(this, running ? CARD_VISIBLE : CARD_INVISIBLE);
    }
  }

  public boolean isRunning() {
    return running;
  }

  private static final class ProgressSpinnerImpl extends JComponent {
    private static final Dimension DIMENSIONS = new Dimension(20, 20);
    private static final float MIN_ALPHA = 0.16f;
    private static final float MAX_ALPHA = 0.89f;
    private static final int ARMS = 12;
    private static final float STROKE_WIDTH = 1.25f;
    private static final double INNER_RADIUS = 4;
    private static final double OUTER_RADIUS = 7;
    private static final double REVOLUTIONS_PER_SECOND = 2.0;
    private final Timer timer;
    private int darkestArm = 0;

    ProgressSpinnerImpl() {
      timer = new Timer((int) Math.round(1000.0 / (REVOLUTIONS_PER_SECOND * ARMS)),
          (ActionEvent e) ->
      {
        darkestArm = (darkestArm + 1) % ARMS;
        repaint();
      });
      timer.setRepeats(true);
    }

    public void setRunning(boolean running) {
      if (running) {
        timer.start();
      } else {
        timer.stop();
      }
    }

    @Override
    public Dimension getPreferredSize() {
      return DIMENSIONS;
    }

    @Override
    public Dimension getMinimumSize() {
      return DIMENSIONS;
    }

    /* "Swing programs should override paintComponent() instead of overriding paint()."
    http://www.oracle.com/technetwork/java/painting-140037.html#callbacks */
    @Override
    public void paintComponent(Graphics g0) {
      super.paintComponent(g0);
      Graphics2D g = UIUtil.createGraphicsWithRenderingHintsConfigured(g0, ImmutableMap.of());
      try {
        paintHelper(g);
      } finally {
        g.dispose();
      }
    }

    private void paintHelper(Graphics2D g) {
      g.setRenderingHint(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);
      g.setStroke(new BasicStroke(STROKE_WIDTH, BasicStroke.CAP_ROUND, BasicStroke.CAP_ROUND));
      g.translate(DIMENSIONS.width / 2.0, DIMENSIONS.height / 2.0);
      for (int i = 0; i < ARMS; i++) {
        g.setColor(new Color(0f, 0f, 0f, MIN_ALPHA + (MAX_ALPHA - MIN_ALPHA) *
            (float) Math.pow((((darkestArm + i) % ARMS) / (ARMS - 1.0f)), 3.0)));
        double angle = -(2 * Math.PI / ARMS) * i;
        double kY = Math.sin(angle);
        double kX = Math.cos(angle);
        g.draw(new Line2D.Double(
            INNER_RADIUS * kX, INNER_RADIUS * kY,
            OUTER_RADIUS * kX, OUTER_RADIUS * kY));
      }
    }
  }
}