// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import java.awt.Color;
import java.awt.font.TextAttribute;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.List;
import javax.annotation.concurrent.Immutable;
import javax.swing.UIManager;

/**
 * A presentation string intended to be used in Swing components.
 */
@Immutable
public abstract class SwingString {
  public static final SwingString EMPTY = new Plain("");
  private static final String SWING_HTML_PREFIX = "<html>";
  // TODO[pd]: Support more attributes as necessary.
  private static final ImmutableSet<AttributedCharacterIterator.Attribute>
      SUPPORTED_HTML_TEXT_ATTRIBUTES =
      ImmutableSet.of(TextAttribute.FOREGROUND, TextAttribute.BACKGROUND, TextAttribute.POSTURE);

  /**
   * Convert all whitespace characters to spaces, with line break characters being collapsed into
   * any previous whitespace characters.
   */
  public static String normalizePresentationWhitespace(String s) {
    Preconditions.checkNotNull(s);
    final char ret[] = new char[s.length()];
    int length = 0;
    boolean lastWasWhitespace = false;
    for (int i = 0; i < s.length(); i++) {
      final char c = s.charAt(i);
      if (Character.isWhitespace(c)) {
        /* Only insert whitespace for line breaks if there was not already a whitespace preceding
        them. */
        if (!lastWasWhitespace || c != '\n' && c != '\r')
          ret[length++] = ' ';
        lastWasWhitespace = true;
      } else {
        ret[length++] = c;
        lastWasWhitespace = false;
      }
    }
    return new String(ret, 0, length);
  }

  public boolean isEmpty() {
    return this == EMPTY;
  }

  /**
   * Displayed the supplied string to the user exactly, escaping any Swing HTML syntax if necessary.
   * Whitespace is automatically normalized using {@link #normalizePresentationWhitespace(String)}.
   */
  public static SwingString plain(String stringToBeEscaped) {
    return stringToBeEscaped.isEmpty() ? EMPTY : new Plain(stringToBeEscaped);
  }

  /**
   * Display the supplied Swing HTML-styled string.
   */
  public static SwingString styled(String swingHtmlStringWithoutHtmlTags) {
    return swingHtmlStringWithoutHtmlTags.isEmpty()
        ? EMPTY : new Styled(swingHtmlStringWithoutHtmlTags);
  }

  /**
   * Create a formatted Swing HTML string from an AttributedString. The original string should have
   * been normalized using {@link #normalizePresentationWhitespace(String)}.
   */
  public static SwingString attributed(AttributedString as) {
    Preconditions.checkNotNull(as);
    final AttributedCharacterIterator iter = as.getIterator();
    final int beginIndex = iter.getBeginIndex();
    final int endIndex = iter.getEndIndex();
    if (endIndex - beginIndex <= 0)
      return EMPTY;
    if (!SUPPORTED_HTML_TEXT_ATTRIBUTES.containsAll(iter.getAllAttributeKeys()))
      throw new IllegalArgumentException();
    final StringBuilder ret = new StringBuilder();
    iter.setIndex(beginIndex);
    final HTMLEscaper escaper = new HTMLEscaper();
    for (int i = beginIndex; i < endIndex; ) {
      final List<String> closingTags = Lists.newArrayList();
      final Color attForeground = (Color) iter.getAttribute(TextAttribute.FOREGROUND);
      final Color attBackground = (Color) iter.getAttribute(TextAttribute.BACKGROUND);
      if (attBackground != null || attForeground != null) {
        ret.append("<font");
        if (attForeground != null)
          ret.append(" color='").append(htmlColor(attForeground)).append("'");
        if (attBackground != null) {
          ret.append(" style='background-color: ")
              .append(htmlColor(attBackground)).append("'");
        }
        ret.append(">");
        closingTags.add("</font>");
      }
      final Float attPosture = (Float) iter.getAttribute(TextAttribute.POSTURE);
      if (attPosture != null && attPosture >= TextAttribute.POSTURE_REGULAR) {
        ret.append("<i>");
        closingTags.add("</i>");
      }
      final int runStart = i;
      final int runLimit = iter.getRunLimit();
      if (runLimit - i <= 0)
        throw new AssertionError();
      if (runLimit > endIndex)
        throw new AssertionError();
      // See o.n.modules.editor.HtmlPrintContainer.addLines.
      for (; i < runLimit; i++) {
        final char c = (i == runStart) ? iter.current() : iter.next();
        ret.append(escaper.escapeSwingHTML(c));
      }
      iter.next();
      for (String closingTag : Lists.reverse(closingTags))
        ret.append(closingTag);
    }
    return new Styled(ret.toString());
  }

  protected static final boolean hasSwingHtmlPrefix(String s) {
    return s.length() >= SWING_HTML_PREFIX.length() &&
        s.substring(0, SWING_HTML_PREFIX.length()).equalsIgnoreCase(SWING_HTML_PREFIX);
  }

  /**
   * Escape a presentation string to make it suitable for concatenation with formatted Swing HTML
   * strings. This includes whitespace normalization as described by
   * {@link #normalizePresentationWhitespace(String)}.
   *
   * Analogous to org.openide.xml.XMLUtil.toElementContent(), but skips unencodable characters
   * rather than throwing an exception, normalizes non-space whitespace characters as noted above,
   * and converts non-collapsed whitespace characters beyond the first to non-breakable HTML spaces.
   */
  public static String toSwingHTML(String s) {
    final String s2 = normalizePresentationWhitespace(s);
    final StringBuilder ret = new StringBuilder();
    final HTMLEscaper escaper = new HTMLEscaper();
    for (int i = 0; i < s2.length(); i++)
      ret.append(escaper.escapeSwingHTML(s2.charAt(i)));
    return ret.toString();
  }

  /**
   * Stateful HTML escaper for individual characters. The state is needed in order to preserve
   * single spaces as regular breakable spaces while converting subsequent adjacent ones to
   * non-breaking spaces. (Always converting spaces to non-breaking spaces once caused a bug in
   * cursor error tooltips.)
   */
  private static final class HTMLEscaper {
    /**
     * May the previous character have been a whitespace? Assume true for the beginning of any
     * string to be escaped, since the string may need to be concatenated with another, unknown
     * string.
     */
    boolean whitespace = true;

    public String escapeSwingHTML(char c) {
      final boolean previousWhitespace = whitespace;
      whitespace = Character.isWhitespace(c);
      if (whitespace) {
        return previousWhitespace ? "&nbsp;" : " ";
      } else if (c == '>') {
        // Don't bother restricting to "]]>"-sequences only.
        return "&gt;";
      } else if (c == '<') {
        return "&lt;";
      } else if (c == '&') {
        return "&amp;";
      } else if (((int) c) >= 32) {
        return Character.toString(c);
      } else {
        // Skip unencodable control characters.
        return "";
      }
    }
  }

  public static String escapeComplete(String plainPresentationString) {
    return SwingString.plain(plainPresentationString).getComplete();
  }

  public static String htmlColor(Color color) {
    // See http://stackoverflow.com/questions/6539879/how-to-convert-a-color-integer-to-a-hex-string-in-android
    return String.format("#%06X", 16777215 & color.getRGB());
  }

  public static String htmlColor(String uiManagerKey) {
    return htmlColor(UIManager.getColor(uiManagerKey));
  }

  /**
   * Get a properly escaped, and possibly styled, string, suitable for concatenation with other
   * escaped and styled Swing HTML strings. Does not include enclosing HTML tags.
   */
  public abstract String getForConcatenation();

  /**
   * Get a complete string for use in a Swing component, possibly including enclosing HTML tags and
   * escaping if needed.
   */
  public abstract String getComplete();

  @Override
  public final String toString() {
    // Include the SwingString class name to avoid accidental uses of toString going unnoticed.
    return "SwingString(\"" + getComplete() + "\")";
  }

  private static final class Plain extends SwingString {
    private final String stringToBeEscaped;

    Plain(String stringToBeEscaped) {
      Preconditions.checkNotNull(stringToBeEscaped);
      this.stringToBeEscaped = stringToBeEscaped;
    }

    @Override
    public String getForConcatenation() {
      return toSwingHTML(stringToBeEscaped);
    }

    @Override
    public String getComplete() {
      return hasSwingHtmlPrefix(stringToBeEscaped)
          ? ("<html>" + toSwingHTML(stringToBeEscaped) + "</html>")
          : normalizePresentationWhitespace(stringToBeEscaped);
    }
  }

  private static final class Styled extends SwingString {
    private final String swingHtmlStringWithoutHtmlTags;

    /**
     * @param swingHtmlStringWithoutHtmlTags A styled and escaped Swing HTML string, not including
     *        the enclosing HTML tags.
     */
    Styled(String swingHtmlStringWithoutHtmlTags) {
      Preconditions.checkArgument(!hasSwingHtmlPrefix(swingHtmlStringWithoutHtmlTags));
      this.swingHtmlStringWithoutHtmlTags = swingHtmlStringWithoutHtmlTags;
    }

    @Override
    public String getForConcatenation() {
      return swingHtmlStringWithoutHtmlTags;
    }

    @Override
    public String getComplete() {
      return "<html>" + swingHtmlStringWithoutHtmlTags + "</html>";
    }
  }
}
