// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.uiutil.api.CommonUIUtil;
import com.sieuferd.uiutil.table.NoopTableCellEditor;
import java.awt.AWTKeyStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.KeyboardFocusManager;
import java.awt.Paint;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.font.FontRenderContext;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.net.URL;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.ImageIcon;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import javax.swing.text.AbstractDocument;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;
import org.netbeans.swing.outline.DefaultOutlineModel;
import org.netbeans.swing.outline.Outline;
import org.netbeans.swing.plaf.Startup;
import org.openide.util.ImageUtilities;
import org.openide.util.Mutex;
import org.openide.util.NbBundle;
import org.openide.util.Utilities;

public final class UIUtil {
  private static final Logger LOG = Logger.getLogger(UIUtil.class.getName());
  public static final boolean USE_CUSTOM_LOADIMAGEICON = false;
  // TODO[cc]: Use this instead of null in more places.
  public static final Color COLOR_NONE = new Color(0, 0, 0, 0);
  /**
   * Value to be used for the X or Y position of a {@link java.awt.geom.Rectangle2D.Double} when the
   * rectangle is intended to be positioned at negative infinity in the relevant direction. The
   * corresponding width or height should be {@link Double#POSITIVE_INFINITY}. This is useful
   * primarily for {@code Rectangle2D} instances passed to
   * {@link Graphics2D#setClip(java.awt.Shape)}. It is not possible to use
   * {@link Double#NEGATIVE_INFINITY} for the rectangle position, as the resulting width or height
   * of the rectangle would be NaN.
   */
  public static final double NEGATIVE_INFINITY_CLIP_POSITION = (double) Long.MIN_VALUE;

  private UIUtil() {}

  public static void instrumentation(String msg) {
    System.out.println("INSTRUMENTATION(" + System.nanoTime() / 1000L + "us): " + msg);
  }

  /* "To make a component ignore a key that it normally responds to, you can use the special action
  name[d] 'none'". http://docs.oracle.com/javase/tutorial/uiswing/misc/keybinding.html */
  public static final String NO_ACTION = "none";

  public static void setDefaultLookAndFeel(final boolean useNetBeansLAF) {
    if (Utilities.isMac())
      System.setProperty("apple.laf.useScreenMenuBar", "true");
    /* Use Mutex.EVENT.readAccess(Action) instead of EventQueue.invokeLater to ensure that things
    like UIManager.getFont("controlFont") will work immediately after we return. */
    Mutex.EVENT.readAccess((Mutex.Action<Void>) () -> {
      if (useNetBeansLAF) {
        // Will call UIManager.setLookAndFeel in turn.
        /* org.netbeans.core.CoreBridgeImpl.initializePlaf takes care to call this only from the
        EDT, so let's do the same here. */
        Startup.run(null, 0, null, NbBundle.getBundle(Startup.class));
      } else {
        // See http://docs.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html .
        try {
          UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (
            ClassNotFoundException | IllegalAccessException |
                InstantiationException | UnsupportedLookAndFeelException e)
        {
          LOG.log(Level.WARNING, "Failed to set default L&F", e);
        }
      }
      return null;
    });
  }

  /**
   * Hack to avoid layout flickering after updating the HTML content of a Swing component like
   * JLabel.
   */
  public static void forceSwingHtmlSizeComputation(JComponent component) {
    /* From http://stackoverflow.com/questions/16227877/how-to-update-a-jcomponent-with-html-without-flickering :
    "The problem is that during layout's getPreferredSize(), a JLabel containing HTML does not know
    how much horizontal space is available before it needs to wrap lines. It just returns a wide
    size assuming that it does not need to wrap at all. Then, during paint, the JLabel’s internal
    FlowView will "repair" the layout and set its actual required height, which is cached and
    returned at the next layout iteration. The solution that I used is to call JLabel.paint, since
    there seems to be no other way to tell it to finish doing layout with the actual available
    width." */
    Graphics g = component.getGraphics();
    if (g != null) {
      Graphics g2 = g.create();
      try {
        // Don't actually cause any painting to happen.
        g2.setClip(0, 0, 0, 0);
        component.paint(g2);
      } finally {
        g2.dispose();
      }
    }
  }

  public static void updatePreferredOutlineColumnWidths(final Outline outline) {
    /* Hack to force invocation of ETable.updatePreferredWidths(), which is not public. See
    ETable.setModel(). Post to the end of the event queue since the resize may need to happen after
    the table has finished processing any current operations. */
    SwingUtilities.invokeLater(() -> {
      outline.setModel(outline.getModel());
    });
  }

  public static void setOutlineColumnsLabel(Outline outline, String label) {
    TableModel tableModel = outline.getModel();
    if (tableModel instanceof DefaultOutlineModel) {
      ((DefaultOutlineModel) tableModel).setNodesColumnLabel(label);
    } else {
      LOG.warning("TableModel for Outline was not a DefaultOutlineModel");
    }
  }

  public static Window getWindowForComponent(Component component) {
    return CommonUIUtil.getWindowForComponent(component);
  }

  /** @return the coordinates of {@code evt} in {@code component}'s coordinates. */
  public static Point convertMousePoint(Component component, MouseEvent evt) {
    // See org.openide.explorer.view.TreeView for an example use of SwingUtilities.convertPoint.
    return SwingUtilities.convertPoint(evt.getComponent(), evt.getPoint(), component);
  }

  /**
   * Determined whether the supplied {@code KeyEvent} is likely to correspond to a character
   * insertion event in the text editor kits provided by Swing and NetBeans.
   */
  public static boolean isCharacterInsertionEvent(KeyEvent evt) {
    Preconditions.checkNotNull(evt);
    /* Use logic similar to that found in both
    javax.swing.text.DefaultEditorKit.DefaultKeyTypedAction and
    org.netbeans.editor.BaseKit.DefaultKeyTypedAction. */

    // Only KEY_TYPED events are guaranteed to give a meaningful result from getKeyChar().
    if (evt.getID() != KeyEvent.KEY_TYPED)
      return false;
    /* See o.n.editor.BaseKit.isValidDefaultTypedAction and
    sun.awt.SunToolkit.isPrintableCharacterModifiersMask (about the latter, see
    http://hg.openjdk.java.net/jdk8/jdk8/jdk/rev/8bfdb214756c and
    http://hg.openjdk.java.net/jdk9/jdk9/jdk/file/tip/src/java.desktop/share/classes/sun/awt/SunToolkit.java ). */
    /* Note that character-producing Ctrl+Alt keystrokes on Windows (e.g. Ctrl+Alt+E for the Euro
    symbol) are converted to keystrokes using Alt Graph as as a modifier. */
    final int m = evt.getModifiers();
    if ((m & Event.CTRL_MASK) != 0)
      return false;
    if ((m & (Utilities.isMac() ? Event.META_MASK /* Command key */ : Event.ALT_MASK)) != 0)
      return false;
    // See o.n.editor.BaseKit.isValidDefaultTypedCommand.
    final char c = evt.getKeyChar();
    // Delete (0x7F) and newline (0x0B) do not qualify.
    return c != KeyEvent.CHAR_UNDEFINED && c >= 0x20 && c != 0x7F;
  }

  public static @Nullable Action getEnabledAction(
      JComponent component, KeyStroke keyStroke, int condition)
  {
    Preconditions.checkNotNull(component);
    Preconditions.checkNotNull(keyStroke);
    Object actionKey = component.getInputMap(condition).get(keyStroke);
    if (actionKey == null)
      return null;
    Action ret = component.getActionMap().get(actionKey);
    return ret != null && ret.isEnabled() ? ret : null;
  }

  public static String frcToString(FontRenderContext frc) {
    StringBuilder ret = new StringBuilder();
    ret.append("transform=").append(frc.getTransform()).append(", ");
    ret.append("aa=").append(frc.getAntiAliasingHint()).append(", ");
    ret.append("fractional=").append(frc.getFractionalMetricsHint());
    return ret.toString();
  }

  public static String inputMapConditionToString(int condition) {
    if (condition == JComponent.WHEN_FOCUSED) {
      return "WHEN_FOCUSED";
    } else if (condition == JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT) {
      return "WHEN_ANCESTOR_OF_FOCUSED_COMPONENT";
    } else if (condition == JComponent.WHEN_IN_FOCUSED_WINDOW) {
      return "WHEN_IN_FOCUSED_WINDOW";
    } else {
      throw new IllegalArgumentException("Illegal InputMap condition " + condition);
    }
  }

  /**
   * Create a {@code Graphics2D} object with rendering hints correctly configured, based on an
   * existing {@code Graphics} instance provided by Swing. The client should call
   * {@link Graphics#dispose()} on the returned object after using it.
   */
  public static Graphics2D createGraphicsWithRenderingHintsConfigured(
      Graphics basedOn, Map<Object,Object> additionalHints)
  {
    Graphics2D ret = (Graphics2D) basedOn.create();
    Object desktopHints = Toolkit.getDefaultToolkit().getDesktopProperty("awt.font.desktophints");
    Map<Object,Object> hints = Maps.newLinkedHashMap();
    if(desktopHints != null && desktopHints instanceof Map<?,?>)
      hints.putAll((Map<?,?>) desktopHints);
    // Required to get non-text antialiasing on Windows.
    if (!hints.containsKey(RenderingHints.KEY_ANTIALIASING))
      hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    hints.putAll(additionalHints);
    ret.addRenderingHints(hints);
    return ret;
  }

  private static String getDocumentTextHelper(Document document) {
    try {
      return document.getText(0, document.getLength());
    } catch (BadLocationException e) {
      throw new RuntimeException(
          "Unexpected race condition while getting text of Swing Document", e);
    }
  }

  /**
   * Get the text of a Swing {@link Document}. If the {@code Document} extends from
   * {@link javax.swing.text.AbstractDocument}, the contents will be read from within
   * {@link javax.swing.text.AbstractDocument#render(Runnable)}.
   */
  public static @Nonnull String getDocumentText(final Document document) {
    Preconditions.checkNotNull(document);
    if (document instanceof AbstractDocument) {
      final String ret[] = new String[] { null };
      ((AbstractDocument) document).render(() -> {
        ret[0] = getDocumentTextHelper(document);
      });
      return ret[0];
    } else {
      return getDocumentTextHelper(document);
    }
  }

  /** Set the text of a Swing {@link Document}. Note that this may generate an undoable edit. */
  public static void setDocumentText(final Document document, final String text) {
    Preconditions.checkNotNull(document);
    Preconditions.checkNotNull(text);
    try {
      // Same logic as in JTextField.setText(String).
      if (document instanceof AbstractDocument) {
        /* In theory, the length of the document could change between the call to getLength and
        replace, but since AbstractDocument.writeLock is protected, there seems to be no way to
        avoid this race condition. But neither JTextField.setText nor JEditorPane.setText handle
        this in any special manner either, so it seems not to matter. */
        ((AbstractDocument) document).replace(0, document.getLength(), text, null);
      } else {
        document.remove(0, document.getLength());
        document.insertString(0, text, null);
      }
    } catch (BadLocationException e) {
      throw new RuntimeException(
          "Unexpected race condition while setting text of Swing Document", e);
    }
  }

  public static Color opacity(Color c, float opacity) {
    Preconditions.checkArgument(opacity >= 0.0f && opacity <= 1.0f);
    final int newAlpha = Math.round(c.getAlpha() * opacity);
    return (newAlpha == 0) ?
        COLOR_NONE : new Color(c.getRed(), c.getGreen(), c.getBlue(), newAlpha);
  }

  public static Color blendOver(final Color src, final Color dest) {
    return blendOver(src, dest, 1.0f);
  }

  public static Color blendOver(final Color src, final Color dest, float extraSourceAlpha) {
    Preconditions.checkArgument(extraSourceAlpha >= 0.0f && extraSourceAlpha <= 1.0f);
    // Small optimization.
    if (src.getAlpha() == 0 || extraSourceAlpha == 0.0f)
      return dest;
    /* Based on the formulas from the Porter and Duff paper, as presented in the Javadoc for
    java.awt.AlphaComposite. */
    // "The 'extra' alpha component from the AlphaComposite instance"
    final float Aac = extraSourceAlpha;

    final float Csr[] = src.getRGBComponents(null);
    final float Cdr[] = dest.getRGBComponents(null);
    final float Asr = Csr[3];
    final float Adr = Cdr[3];

    final float As = Asr * Aac;
    final float Ad = Adr;

    // Use the functions for SRC_OVER.
    final float Fs = 1.0f;
    final float Fd = 1.0f - As;

    final float Ar = As * Fs + Ad * Fd;
    final float Adf = Ar;

    // Avoid division by zero below.
    if (Adf == 0.0f)
      return UIUtil.COLOR_NONE;

    final float Cdf[] = new float[3];
    for (int i = 0; i < 3; i++) {
      /* The zero check ensures correct behavior in the case of a zero-alpha Color with NaN color
      components. */
      final float Cs = (As == 0.0f) ? 0.0f : Csr[i] * As;
      final float Cd = (Ad == 0.0f) ? 0.0f : Cdr[i] * Ad;
      final float Cr = Cs * Fs + Cd * Fd;
      Cdf[i] = Cr / Ar;
    }
    return new Color(Cdf[0], Cdf[1], Cdf[2], Adf);
  }

  public static Color desaturate(Color c) {
    /* Constants from
    http://developer.bostjan-cigan.com/java-color-image-to-grayscale-conversion-algorithm .
    Technically the "luminosity" approach. */
    final int brightness = Math.max(0, Math.min(255,
        (int) Math.round((0.299 * c.getRed() + 0.587 * c.getGreen() + 0.114 * c.getBlue()))));
    return new Color(brightness, brightness, brightness, c.getAlpha());
  }

  public static String colorToString(Color c) {
    return Arrays.toString(c.getRGBComponents(null));
  }

  public static boolean isAqua() {
    return CommonUIUtil.isAqua();
  }

  /**
   * Return true iff they specified keyboard modifiers include the default accelerator on this
   * platform. This corresponds to 'D' in the format used by
   * {@link org.openide.util.Utilities#stringToKey(KeyStroke, boolean)}; "the Control key on most
   * platforms, the Command (meta) key on Macintosh".
   *
   * @param modifiers non-extended modifiers code
   */
  public static boolean modifierIncludesDefaultAccelerator(int modifiers) {
    // See org.openide.util.Utilities.addModifiersPortable.
    return Utilities.isMac()
        ? ((modifiers & KeyEvent.META_MASK) != 0) // Command key.
        : ((modifiers & KeyEvent.CTRL_MASK) != 0);
  }

  /**
   * Return true iff they specified keyboard modifiers include the alternative accelerator on this
   * platform. This corresponds to 'O' in the format used by
   * {@link org.openide.util.Utilities#stringToKey(KeyStroke, boolean)}; "the Alt key on most
   * platforms, the Ctrl key on Macintosh".
   *
   * @param modifiers non-extended modifiers code
   */
  public static boolean modifierIncludesAlternateAccelerator(int modifiers) {
    // See org.openide.util.Utilities.addModifiersPortable.
    return Utilities.isMac()
        ? ((modifiers & KeyEvent.CTRL_MASK) != 0)
        : ((modifiers & KeyEvent.ALT_MASK) != 0);
  }

  public static void setSimpleFocusTraversalKeys(JComponent component) {
    // Subset of KeyboardFocusManager.defaultFocusTraversalKeyStrokes.
    component.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
        ImmutableSet.of(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB, 0, false)));
    component.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
        ImmutableSet.of(AWTKeyStroke.getAWTKeyStroke(KeyEvent.VK_TAB,
        InputEvent.SHIFT_DOWN_MASK | InputEvent.SHIFT_MASK, false)));
    component.setFocusTraversalKeys(KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS,
        ImmutableSet.of());
    component.setFocusTraversalKeys(KeyboardFocusManager.DOWN_CYCLE_TRAVERSAL_KEYS,
        ImmutableSet.of());
  }

  public static void ensureSelectionOnFocusGained(JTable table) {
    ensureSelectionOnFocusGained(table, table.getSelectionModel());
  }

  public static void ensureSelectionOnFocusGained(JList<?> list) {
    ensureSelectionOnFocusGained(list, list.getSelectionModel());
  }

  private static void ensureSelectionOnFocusGained(Component component, ListSelectionModel model) {
    component.addFocusListener(new FocusAdapter() {
      @Override
      public void focusGained(FocusEvent e) {
        if (model.isSelectionEmpty() && e.getOppositeComponent() != null &&
            /* Only trigger a selection when the focus is coming from a different component in the
            same Window, as in the case of tabbing between components. When a Window is first made
            visible, it may be cleanest to show no selection. */
            SwingUtilities.isDescendingFrom(e.getOppositeComponent(),
            SwingUtilities.getRoot(component)))
        {
          model.setSelectionInterval(0, 0);
          model.setLeadSelectionIndex(0);
          model.setAnchorSelectionIndex(0);
        }
      }
    });
  }

  public static void configureBooleanTableValueShortcuts(
      final JTable table, final int booleanColumnModelIndex)
  {
    table.addPropertyChangeListener("model", (PropertyChangeEvent evt) -> {
      TableColumnModel tcm = table.getColumnModel();
      /* Don't allow the checkbox to appear as an editor; its state can now be set both via mouse
      and keyboard using the shortcuts installed below. Avoiding use of the checkbox as an editor
      fixes several bugs (wrong background color when clicked when originally unselected, not
      updated if installed and value changed by keyboard shortcut, unsupported tri-state mode). */
      if (tcm.getColumnCount() > booleanColumnModelIndex)
        tcm.getColumn(booleanColumnModelIndex).setCellEditor(new NoopTableCellEditor());
    });
    table.addKeyListener(new KeyAdapter() {
      private boolean alreadyPressed;

      @Override
      public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_SPACE && e.getModifiersEx() == 0) {
          /* Avoid repeated selection/deselection if the key is held down; at least on MacOS, a
          key held down will generate multiple keyPressed events followed by a single keyReleased
          event when the key is released. Note that we need to consume the event every time. */
          if (!alreadyPressed) {
            TableModel model = table.getModel();
            List<Integer> selectedRowModelIndicies = Lists.newArrayList();
            for (int selectedRowIndex : table.getSelectedRows()) {
              int selectedRowModelIndex = table.convertRowIndexToModel(selectedRowIndex);
              if (table.getModel().isCellEditable(selectedRowModelIndex, booleanColumnModelIndex))
                selectedRowModelIndicies.add(selectedRowModelIndex);
            }
            boolean allPreviouslyChecked = true;
            for (int selectedRowModelIndex : selectedRowModelIndicies) {
              allPreviouslyChecked &= Boolean.TRUE.equals(
                  model.getValueAt(selectedRowModelIndex, booleanColumnModelIndex));
            }
            for (int selectedRowModelIndex : selectedRowModelIndicies) {
              model.setValueAt(
                  !allPreviouslyChecked, selectedRowModelIndex, booleanColumnModelIndex);
            }
            alreadyPressed = true;
          }
          e.consume();
        }
      }

      @Override
      public void keyReleased(KeyEvent e) {
        alreadyPressed = false;
      }
    });
    table.addMouseListener(new MouseAdapter() {
      @Nullable TableCheckState stateAtPress;

      private @Nullable TableCheckState getState(MouseEvent e) {
        if ((e.getModifiersEx() &
            (InputEvent.ALT_DOWN_MASK | InputEvent.ALT_GRAPH_DOWN_MASK |
            InputEvent.META_DOWN_MASK |  InputEvent.CTRL_DOWN_MASK |
            InputEvent.SHIFT_DOWN_MASK)) != 0)
        {
          return null;
        }
        int rowAtPoint = table.rowAtPoint(
            SwingUtilities.convertPoint((Component) e.getSource(), e.getPoint(), table));
        if (rowAtPoint < 0)
          return null;
        int rowModelIndex = table.convertRowIndexToModel(rowAtPoint);
        Object value = table.getModel().getValueAt(rowModelIndex, booleanColumnModelIndex);
        if (value != null && !(value instanceof Boolean))
          return null;
        return new TableCheckState(rowModelIndex, !Boolean.FALSE.equals(value));
      }

      @Override
      public void mousePressed(MouseEvent e) {
        stateAtPress = getState(e);
        /* I considered changing the state upon press instead of release, but it felt a bit
        unnatural due to the mouse simultaneously changing the row selection. Change the state upon
        release instead, as would be the default for checkboxes. This also allows range selection by
        dragging to work normally. */
        /* I considered trying to get the checkbox to display an "armed" state during the press, but
        it might not be worth it, since the current behavior makes sense when considering that
        clicks may also be part of a range selection. */
      }

      @Override
      public void mouseReleased(MouseEvent e) {
        TableCheckState state = getState(e);
        if (stateAtPress != null && state != null && state.equalsState(stateAtPress) &&
            table.getModel().isCellEditable(stateAtPress.rowModelIndex, booleanColumnModelIndex))
        {
          table.getModel().setValueAt(
              !stateAtPress.value, stateAtPress.rowModelIndex, booleanColumnModelIndex);
        }
        stateAtPress = null;
      }
    });
  }

  private static final class TableCheckState {
    final int rowModelIndex;
    final boolean value;

    TableCheckState(int rowModelIndex, boolean value) {
      Preconditions.checkArgument(rowModelIndex >= 0);
      this.rowModelIndex = rowModelIndex;
      this.value = value;
    }

    public boolean equalsState(TableCheckState other) {
      return this.rowModelIndex == other.rowModelIndex
          && this.value == other.value;
    }

    @Override
    public String toString() {
      return "CheckState(" + rowModelIndex + ", " + value + ")";
    }
  }

  private static final AffineTransform HALF_PIXEL_TRANSFORM =
      AffineTransform.getScaleInstance(2.0, 2.0);

  public static boolean useHalfPixels(Graphics2D g) {
    ThreadUtil.checkOnEDT();
    /* This method successfully detects Retina display modes on Java 1.7.0_45 for MacOS X 10.8.5. In
    particular, this works on multi-monitor setups where a window may be dragged between monitors
    that have retina mode independently enabled or disabled. */
    return g.getFontRenderContext().getTransform().equals(HALF_PIXEL_TRANSFORM);
  }

  public static int getPixelsPerInch() {
    ThreadUtil.checkOnEDT();
    // See http://blogs.msdn.com/b/fontblog/archive/2005/11/08/490490.aspx .
    if (Utilities.isMac()) {
      /* Toolkit.getScreenResolution() returns weird values on MacOS with Java 1.7.0_40-ea-b36 (126
      in non-retina and 63 in retina mode). Using the traditional MacOS convention of 72 PPI ensures
      the same font scaling as for example in Excel for Mac 2011. The PPI should be the same both
      with and without retina, as retina is supported in Swing by means of "half-pixels" rather than
      by actual doubling of the pixel coordinates. */
      return 72;
    } else {
      /* On Windows, this will return 96, which correctly ensures the same font scaling as for
      example in Microsoft Excel 2013. */
      return Toolkit.getDefaultToolkit().getScreenResolution();
    }
  }

  private static final int DROP_SHADOW_RADIUS_PIXELS = 10;

  /**
   * Get a bounding box of the supplied rectangle plus its shadow border as would be drawn by
   * {@link #drawShadowBorder(Graphics2D,Rectangle,boolean)}.
   */
  public static final Rectangle getShadowBorderBounds(Rectangle2D rect) {
    final int BORDER_WIDTH = 1;
    return new Rectangle2D.Double(
        rect.getX() - BORDER_WIDTH,
        rect.getY() - BORDER_WIDTH,
        BORDER_WIDTH + rect.getWidth() + Math.max(DROP_SHADOW_RADIUS_PIXELS, BORDER_WIDTH),
        BORDER_WIDTH + rect.getHeight() + Math.max(DROP_SHADOW_RADIUS_PIXELS, BORDER_WIDTH))
        .getBounds();
  }

  /**
   * Draw a border and drop shadow around the specified rectangle (which may also be drawn on as a
   * side-effect).
   *
   * @param g will be painted on but not otherwise modified
   */
  public static final void drawShadowBorder(Graphics2D g, Rectangle2D rect, boolean strong) {
    final int R = DROP_SHADOW_RADIUS_PIXELS;
    // Draw drop shadow.
    final int stepsPerPixel = useHalfPixels(g) ? 2 : 1;
    final double rectX = rect.getX();
    final double rectY = rect.getY();
    final double rectWidth = rect.getWidth();
    final double rectHeight = rect.getHeight();
    final Color oldColor = g.getColor();
    for (int i = 0; i < R * stepsPerPixel; i++) {
      // Go from largest to smallest radius (r goes from R, inclusive, to 0, exclusive).
      final double r = R - i / (double) stepsPerPixel;
      /* Paint concentric rounded rectangles with opacities increasing in a non-linear fashion.
      Since subsequent rectangles are drawn on top of previous ones, the actual gradient follows the
      integral of the function below (or something like that). The function below was found by trial
      and error. */
      g.setColor(new Color(0f, 0f, 0f,
          (float) ((0.05 / stepsPerPixel) * Math.pow(((R - r + 1f) / R), 2.0))));
      g.fill(new RoundRectangle2D.Double(
          rectX, rectY, rectWidth + r, rectHeight + r, r * 2, r * 2));
    }
    // Draw border.
    g.setColor(strong ? new Color(120, 120, 120, 255) : new Color(0, 0, 0, 30));
    g.fill(new Rectangle2D.Double(rectX - 1, rectY - 1         , rectWidth + 2, 1)); // Top
    g.fill(new Rectangle2D.Double(rectX - 1, rectY + rectHeight, rectWidth + 2, 1)); // Bottom
    g.fill(new Rectangle2D.Double(rectX - 1        , rectY, 1, rectHeight)); // Left
    g.fill(new Rectangle2D.Double(rectX + rectWidth, rectY, 1, rectHeight)); // Right
    g.setColor(oldColor);
  }

  public static final int DROPDOWN_BUTTON_BORDER_WIDTH_PIXELS = 1;
  private static final boolean DROPDOWN_BUTTON_CENTER_ARROW_VERTICALLY = true;

  // Take a Rectangle2D rather than Rectangle to allow half-pixel alignment on Retina displays.
  /**
   * Draw a rectangular dropdown button, such as those displayed by Excel when filters are enabled
   * on a table.
   */
  public static final void drawDropdownButton(Graphics2D g, Rectangle2D rect) {
    g.setColor(new Color(172, 172, 172, 255));
    g.fill(rect);
    g.setColor(new Color(255, 255, 255, 255));
    final Paint oldPaint = g.getPaint();
    try {
      g.setPaint(new GradientPaint(
          new Point2D.Double(0, rect.getY()),
          new Color(255, 255, 255),
          new Point2D.Double(0, rect.getY() + rect.getHeight()),
          new Color(242, 242, 242)));
      final int BORDER_WIDTH = DROPDOWN_BUTTON_BORDER_WIDTH_PIXELS;
      g.fill(new Rectangle2D.Double(rect.getX() + BORDER_WIDTH, rect.getY() + BORDER_WIDTH,
          rect.getWidth() - 2 * BORDER_WIDTH, rect.getHeight() - 2 * BORDER_WIDTH));
      final double arrowScale = Math.min(16.0, Math.min(rect.getWidth(), rect.getHeight())) / 16.0;
      final double arrowWidth = 8.0 * arrowScale;
      final double arrowHeight = arrowWidth / 2.0;
      final double arrowOffsetY;
      if (DROPDOWN_BUTTON_CENTER_ARROW_VERTICALLY) {
        arrowOffsetY = rect.getHeight() / 2.0 - arrowHeight / 2.0;
      } else {
        arrowOffsetY = rect.getHeight() - 10.0 * arrowScale;
      }
      final double roundedArrowOffsetY;
      if (arrowScale == 1.0) {
        // Try to make the flat top of the arrowhead line up with device pixels.
        if (useHalfPixels(g)) {
          roundedArrowOffsetY = Math.round(arrowOffsetY * 2.0) / 2.0;
        } else {
          roundedArrowOffsetY = Math.round(arrowOffsetY);
        }
      } else {
        roundedArrowOffsetY = arrowOffsetY;
      }
      final double arrowTopY = rect.getY() + roundedArrowOffsetY;
      final double arrowMidX = rect.getX() + rect.getWidth() / 2.0;
      Path2D.Double arrowPath = new Path2D.Double();
      arrowPath.moveTo(arrowMidX - arrowWidth / 2.0, arrowTopY);
      arrowPath.lineTo(arrowMidX, arrowTopY + arrowHeight);
      arrowPath.lineTo(arrowMidX + arrowWidth / 2.0, arrowTopY);
      arrowPath.closePath();
      g.setPaint(new GradientPaint(
          new Point2D.Double(arrowMidX - arrowWidth / 2.0, arrowTopY),
          new Color(140, 140, 140),
          new Point2D.Double(arrowMidX, arrowTopY + arrowHeight),
          new Color(86, 86, 86)));
      g.fill(arrowPath);
    } finally {
      g.setPaint(oldPaint);
    }
  }

  /**
   * Get an {@code InputMap} for a component that is safe for modification, that is, not a globally
   * shared UI resource.
   */
  public static InputMap getSafeInputMap(JComponent component, int condition) {
    final InputMap existingMap = component.getInputMap(condition);
    if (existingMap instanceof SafeInputMap) {
      return existingMap;
    } else {
      final InputMap newMap = new SafeInputMap();
      newMap.setParent(existingMap);
      component.setInputMap(condition, newMap);
      return newMap;
    }
  }

  /**
   * Get an {@code ActionMap} for a component that is safe for modification, that is, not a globally
   * shared UI resource.
   */
  public static ActionMap getSafeActionMap(JComponent component) {
    final ActionMap existingMap = component.getActionMap();
    if (existingMap instanceof SafeActionMap) {
      return existingMap;
    } else {
      final ActionMap newMap = new SafeActionMap();
      newMap.setParent(existingMap);
      component.setActionMap(newMap);
      return newMap;
    }
  }

  private static final class SafeActionMap extends ActionMap { }
  private static final class SafeInputMap extends InputMap { }

  private static @Nullable DebugFocusOwnerListener debugCurrentFocusOwnerListener;

  private static final class DebugFocusOwnerListener implements PropertyChangeListener {
    final KeyboardFocusManager focusManager;

    public DebugFocusOwnerListener(KeyboardFocusManager focusManager) {
      Preconditions.checkNotNull(focusManager);
      this.focusManager = focusManager;
    }

    @Override
    public void propertyChange(PropertyChangeEvent evt) {
      LOG.log(Level.INFO, "New focusOwner: {0}", evt.getNewValue());
    }
  }

  public static final void debugLogFocusChanges(boolean enable) {
    ThreadUtil.checkOnEDT();
    if (enable && debugCurrentFocusOwnerListener == null) {
      KeyboardFocusManager fm = KeyboardFocusManager.getCurrentKeyboardFocusManager();
      debugCurrentFocusOwnerListener = new DebugFocusOwnerListener(fm);
      fm.addPropertyChangeListener("focusOwner", debugCurrentFocusOwnerListener);
    } else if (!enable && debugCurrentFocusOwnerListener != null) {
      debugCurrentFocusOwnerListener.focusManager.removePropertyChangeListener(
          "focusOwner", debugCurrentFocusOwnerListener);
      debugCurrentFocusOwnerListener = null;
    }
  }

  /**
   * Ensure that Swing's double buffer limits are large enough to buffer the specified component.
   * This prevents tearing during scrolling in JScrollPane for instance in the case where a large
   * external monitor is connected to a laptop while the application is running. Should be called
   * whenever the size of the component in question changes. Has no effect on settings and buffers
   * in the case where the existing limit is already large enough.
   */
  public static void ensureSufficientDoubleBufferSize(JComponent comp) {
    Preconditions.checkNotNull(comp);
    final Component root = SwingUtilities.getRoot(comp);
    if (root == null)
      return;
    final Rectangle rootBounds = root.getBounds();
    final RepaintManager rm = RepaintManager.currentManager(comp);
    final Dimension currentMax = rm.getDoubleBufferMaximumSize();
    if (currentMax.width < rootBounds.width || currentMax.height < rootBounds.height)
      rm.setDoubleBufferMaximumSize(null);
  }

  /* TODO[pd]: Remove this once ImageUtilities.loadImageIcon is updated to support retina icons. See
               https://netbeans.org/bugzilla/show_bug.cgi?id=239745 . */
  // Alternative to ImageUtilities.loadImageIcon that supports @2x icons.
  public static final ImageIcon loadImageIcon(String resource) {
    if (USE_CUSTOM_LOADIMAGEICON) {
      final ClassLoader loader = Thread.currentThread().getContextClassLoader();
      final URL url = loader.getResource(resource);
      if (url == null)
        return null;
      return new ImageIcon(url);
    } else {
      return ImageUtilities.loadImageIcon(resource, false);
    }
  }
}
