// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.flow;

import com.sieuferd.flow.Flow;
import javax.annotation.Nullable;
import javax.swing.SwingUtilities;
import org.openide.util.Mutex;

/**
 * Abstract base class for {@code Flow} implementations that wrap a model object that should only be
 * accessed from the Event Dispatch Thread (EDT). The output of the {@code Flow} will only change
 * once the model returns a different object (as per {@code Object.equals}).
 */
abstract class EDTAdapterFlow<T> extends Flow<T> {
  private volatile boolean getModelValueUnderway;

  EDTAdapterFlow(String description) {
    super(description);
  }

  /** Will be called on the EDT only. */
  protected abstract void addModelListener();
  /** Will be called on the EDT only. */
  protected abstract void removeModelListener();
  /** Will be called on the EDT only. */
  protected abstract @Nullable T getModelValue();

  /** May be called from any thread, including the EDT. */
  protected final void markModelUpdated() {
    /* Some models may trigger their own listeners during getModelValue(); avoid re-propagating in
    this case. */
    if (getModelValueUnderway)
      return;
    Mutex.EVENT.readAccess(() -> {
      final T newValue;
      getModelValueUnderway = true;
      try {
        newValue = getModelValue();
      } finally {
        getModelValueUnderway = false;
      }
      setInternalIfDifferent(newValue);
      propagateChanges();
    });
  }

  @Override
  protected final void firstListenerAdded() {
    // Don't execute right away even if we're already on the EDT; first relinquish the lock.
    SwingUtilities.invokeLater(() -> {
      addModelListener();
      markModelUpdated();
    });
  }

  @Override
  protected final void lastListenerRemoved() {
    // See above.
    SwingUtilities.invokeLater(this::removeModelListener);
  }

  @Override
  protected final void update() {
    // Take no action.
  }
}
