// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.flow;

import com.google.common.base.Supplier;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.Flows;
import java.beans.PropertyChangeListener;
import javax.annotation.Nullable;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.openide.util.Lookup;
import org.openide.util.Mutex;
import org.openide.util.WeakListeners;

public final class UIFlows {
  private UIFlows() { }

  /**
   * Create a {@code Flow} reflecting the contents of the supplied lookup when it contains
   * exactly one object (per {@code Object.equals}) of the supplied type. Otherwise similar to
   * {@link #newLookup(Lookup, Lookup.Template)}.
   */
  public static <T> Flow<T> newSingularLookup(Lookup lookup, final Class<T> type) {
    return Flows.newProcessed(newLookup(lookup, new Lookup.Template<>(type)),
        lrs -> {
        T ret = null;
        for (T instance : lrs.allInstances()) {
          if (ret != null && !instance.equals(ret)) {
            // Prohibit multiple non-equal entries.
            return null;
          } else {
            ret = instance;
          }
        }
        return ret;
      }
    );
  }

  /**
   * Create a Lookup-based {@code Flow}. Clients should maintain a strong reference to the returned
   * object, or use it as an input to a strongly referenced {@code Flow}.
   */
  public static <L> Flow<LookupResultSnapshot<L>> newLookup(
      Lookup lookup, Lookup.Template<L> template)
  {
    return new LookupFlow<>(lookup, template);
  }

  // TODO[cc]: Move this into Flows.
  /**
   * A flow referenced by its input through a weak listener only. Commonly, the input is some sort
   * of global object that would otherwise force the returned object to remain in memory through a
   * strong reference from said global object's {@code ChangeListener} list. The client should
   * maintain a strong reference to the returned {@code Flow} for as long as updates are required,
   * either directly or by using the returned {@code Flow} as an input to other strongly referenced
   * {@code Flow} instances. Note that the returned object will in fact maintain a strong reference
   * to the {@code input} object; the weak reference is in the other direction.
   */
  public static <T> Flow<T> newWeak(Flow<T> input) {
    return new WeakFlow<>(input);
  }

  public static <T> Flow<T> fromPropertyBean(String propertyName,
      VoidFunction<PropertyChangeListener> addListener,
      VoidFunction<PropertyChangeListener> removeListener, Supplier<T> getter)
  {
    return new PropertyBeanFlow<>(new PropertyBeanAccessor<T>(propertyName) {
      @Override
      public void addPropertyChangeListener(PropertyChangeListener listener) {
        addListener.apply(listener);
      }

      @Override
      public void removePropertyChangeListener(PropertyChangeListener listener) {
        removeListener.apply(listener);
      }

      @Override
      public T getPropertyValue() {
        return getter.get();
      }
    });
  }

  public static <T> Flow<T> fromChangeBean(
      VoidFunction<ChangeListener> addListener, VoidFunction<ChangeListener> removeListener,
      Supplier<T> getter)
  {
    return new ChangeBeanFlow<>(new ChangeBeanAccessor<T>() {
      @Override
      public void addChangeListener(ChangeListener listener) {
        addListener.apply(listener);
      }

      @Override
      public void removeChangeListener(ChangeListener listener) {
        removeListener.apply(listener);
      }

      @Override
      public T getValue() {
        return getter.get();
      }
    });
  }

  /**
   * Bind a {@code Flow} to an EDT-based update mechanism. The binding can not be removed, but when
   * the Flow depends on global objects, memory leaks can be avoided by using either
   * {@link #newWeak(Flow)} or {@link Flows#newValve(Flow, Flow)} at an appropriate point in the
   * data flow. This method may be called from any thread.
   */
  public static <T> void bindEDT(final Flow<T> input, final ValueBinding<T> binding) {
    bindEDT(input, false, binding);
  }

  /**
   * Bind a {@code Flow} to an EDT-based update mechanism. May be called from any thread.
   *
   * @param weak if true, the client must hold strongly onto the returned listener; if false, the
   *        client may remove the binding using {@code input.removeChangeListener} on the returned
   *        value
   */
  public static <T> ChangeListener bindEDT(
      final Flow<T> input, boolean weak, final ValueBinding<T> binding)
  {
    /* All ValueBinding implementations should be able to handle a null input; attempt one
    immediately to ensure fail-fast behavior if they aren't. */
    Mutex.EVENT.writeAccess(() -> {
      binding.update(null);
    });
    final ChangeListener ret = (ChangeEvent e) -> {
      Mutex.EVENT.writeAccess(() -> {
        binding.update(input.get());
      });
    };
    input.addChangeListener(weak ? WeakListeners.change(ret, input) : ret);
    // Perform the initial update after the listener has been attached.
    Mutex.EVENT.writeAccess(() -> {
      final T currentValue = input.get();
      // The null case was already handled.
      if (currentValue != null)
        binding.update(currentValue);
    });
    return ret;
  }

  public interface ValueBinding<T> {
    /**
     * Will only be called on the EDT. All implementations must handle a null value gracefully.
     */
    public void update(@Nullable T value);
  }
}
