// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.flow;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.Snapshot;
import java.util.Objects;
import java.util.Set;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.openide.util.WeakListeners;

final class WeakFlow<T> extends Flow<T> {
  private final Flow<T> input;
  private final ChangeListener weakListener;
  private final IdempotentRemoveAdapter<T> idempotentRemoveAdapter;
  private Snapshot<T> evaluatedInput;

  WeakFlow(Flow<T> input) {
    super("Weak");
    Preconditions.checkNotNull(input);
    this.input = input;
    this.idempotentRemoveAdapter = new IdempotentRemoveAdapter<>(input);
    this.weakListener = WeakListeners.change(propagatingChangeListener, idempotentRemoveAdapter);
  }

  @Override
  protected void firstListenerAdded() {
    idempotentRemoveAdapter.addChangeListener(weakListener);
  }

  @Override
  protected void lastListenerRemoved() {
    idempotentRemoveAdapter.removeChangeListener(weakListener);
    evaluatedInput = null;
  }

  @Override
  protected void update() {
    final Snapshot<T> currentEvaluatedInput = input.getSnapshot();
    if (Objects.equals(currentEvaluatedInput, evaluatedInput))
      return;
    evaluatedInput = currentEvaluatedInput;
    setInternal(currentEvaluatedInput.getValue());
  }

  @Override
  protected Set<Flow<?>> getDependencies() {
    return ImmutableSet.<Flow<?>>of(input);
  }

  private static class IdempotentRemoveAdapter<T> {
    private final Flow<T> flow;
    private final ChangeListener innerListener = new ChangeListener() {
      @Override
      public void stateChanged(ChangeEvent e) {
        synchronized (IdempotentRemoveAdapter.this) {
          if (outerListener != null)
            outerListener.stateChanged(new ChangeEvent(IdempotentRemoveAdapter.this));
        }
      }
    };
    private ChangeListener outerListener;

    IdempotentRemoveAdapter(Flow<T> flow) {
      Preconditions.checkNotNull(flow);
      this.flow = flow;
    }

    /** Only a single listener may be registered at a time. */
    public final synchronized void addChangeListener(ChangeListener listener) {
      Preconditions.checkNotNull(listener);
      Preconditions.checkState(outerListener == null);
      outerListener = listener;
      flow.addChangeListener(innerListener);
    }

    /** May be called without error even if the listener has already been removed. */
    public final synchronized void removeChangeListener(final ChangeListener listener) {
      Preconditions.checkNotNull(listener);
      Preconditions.checkArgument(outerListener == null || listener == outerListener);
      if (outerListener == null)
        return;
      flow.removeChangeListener(innerListener);
      outerListener = null;
    }
  }
}
