// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.scroll;

import com.google.common.base.Preconditions;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.font.FontRenderContext;
import java.awt.image.VolatileImage;
import javax.annotation.Nullable;

final class BackingStore {
  // TODO[cc]: Encapsulate these fields in a wrapper class, to avoid all the nullables.
  /**
   * Snapshot of the underlying OverlayScrollable at the location viewportPosOnScrollable. Areas
   * covered by the scrollable's child components are always considered dirty, as we have no
   * reliable way to detect repaints on them separately from repaints generated by scrolling itself.
   */
  private @Nullable VolatileImage viewportCache;
  private @Nullable FontRenderContext viewportCacheFRC;
  // Non-null means viewportCache is up-to-date.
  private @Nullable Point viewportCachePosOnScrollable;

  public BackingStore() {
  }

  /* Various tests for empty clip bounds are done just in case the scrollable or Graphics
  implementation could have bugs related to negative areas in clipping rectangles (negative areas
  are known to occur. */

  /**
   * @param vcg will be painted on, but not otherwise have its state modified; should be the
   *        graphics of the viewportCache
   */
  private void paintViewport(
      Graphics2D vcg, Painter painter, Rectangle clipRectOnViewport)
  {
    Preconditions.checkNotNull(vcg);
    Preconditions.checkNotNull(painter);
    Preconditions.checkNotNull(clipRectOnViewport);
    Preconditions.checkState(viewportCachePosOnScrollable != null);
    Graphics2D scratchG = (Graphics2D) vcg.create();
    try {
      scratchG.clipRect(clipRectOnViewport.x, clipRectOnViewport.y,
          clipRectOnViewport.width, clipRectOnViewport.height);
      scratchG.translate(-viewportCachePosOnScrollable.x, -viewportCachePosOnScrollable.y);
      Rectangle clip = scratchG.getClipBounds();
      if (!clip.isEmpty()) {
        /* TODO[pd]: Is clearRect a safe way to clear the image to alpha zero? Consider if the alpha
                     mode should always be Transparency.TRANSLUCENT in order to correctly handle
                     non-opaque scrollables. */
        // Clear the background in case the component supplying the painter is not opaque.
        scratchG.clearRect(clip.x, clip.y, clip.width, clip.height);
        painter.paint(scratchG);
      }
    } finally {
      scratchG.dispose();
    }
  }

  /**
   * @param vcg may be modified, but should not be disposed; should be the graphics of the
   *        viewportCache
   */
  private void blitAndRepaintExposedSides(
      Graphics2D vcg, Rectangle viewportRectOnScrollable, Painter painter)
  {
    Preconditions.checkNotNull(vcg);
    Preconditions.checkNotNull(viewportRectOnScrollable);
    Preconditions.checkNotNull(painter);
    final int viewportWidth = viewportRectOnScrollable.width;
    final int viewportHeight = viewportRectOnScrollable.height;
    final Point oldViewportPositionOnScrollable = viewportCachePosOnScrollable;
    viewportCachePosOnScrollable = viewportRectOnScrollable.getLocation();
    vcg.clipRect(0, 0, viewportWidth, viewportHeight);
    if (oldViewportPositionOnScrollable == null) {
      paintViewport(vcg, painter, new Rectangle(0, 0, viewportWidth, viewportHeight));
    } else {
      final int dx = viewportRectOnScrollable.x - oldViewportPositionOnScrollable.x;
      final int dy = viewportRectOnScrollable.y - oldViewportPositionOnScrollable.y;
      if (dx != 0 || dy != 0) {
        if (!vcg.getClipBounds().isEmpty())
          vcg.copyArea(0, 0, viewportWidth, viewportHeight, -dx, -dy);
      }
      if (dy > 0) {
        // Scroll down, paint under.
        paintViewport(vcg, painter, new Rectangle(0, viewportHeight - dy, viewportWidth, dy));
      } else if (dy < 0) {
        // Scroll up, paint over.
        paintViewport(vcg, painter, new Rectangle(0, 0, viewportWidth, -dy));
      }
      /* TODO[pd]: In the case of diagonal scrolling, cut off the corners already painted by the
                   horizontal logic above. On MacOS I haven't been able to generate diagonal
                   scrolling in the first place, though; doing a diagonal swipe on the trackpad
                   just generates a zig-zag horizontal/vertical scrolling pattern. */
      if (dx > 0) {
        // Scroll right, paint right.
        paintViewport(vcg, painter, new Rectangle(viewportWidth - dx, 0, dx, viewportHeight));
      } else if (dx < 0) {
        // Scroll left, paint left.
        paintViewport(vcg, painter, new Rectangle(0, 0, -dx, viewportHeight));
      }
    }
  }

  /**
   * Clear the viewport cache if it is incompatible with the supplied rendering parameters. This
   * method can be called to prevent subsequent invocations of
   * {@link #updateDirtyCacheArea(BackingStore.Painter,Rectangle)} from doing unnecessary paints on
   * an incompatible backing store.
   */
  public void clearViewportCacheIfIncompatible(
      Graphics2D g, int transparencyMode, Dimension viewportDimension)
  {
    if (viewportCache == null)
      return;
    /* On Java 1.8.0_92-b14/MacOS 10.9.5, I've observed
    VolatileImage.validate(GraphicsConfiguration) return VolatileImage.IMAGE_COMPATIBLE even though
    the FontRenderContext has a different transform due to a switch between retina and non-retina
    mode. So validate the FontRenderContext separately here. This is a good idea in any case, since
    a retina-aware scrollable component may well be sensitive to FontRenderContext changes. */
    boolean flush =
        viewportCache.getWidth() != viewportDimension.width ||
        viewportCache.getHeight() != viewportDimension.height ||
        !viewportCacheFRC.equals(g.getFontRenderContext());
    if (!flush) {
      final int validation = viewportCache.validate(g.getDeviceConfiguration());
      flush = (validation == VolatileImage.IMAGE_INCOMPATIBLE);
      if (validation != VolatileImage.IMAGE_OK)
        viewportCachePosOnScrollable = null;
    }
    if (flush) {
      viewportCache.flush();
      viewportCache = null;
      viewportCacheFRC = null;
      viewportCachePosOnScrollable = null;
    }
  }

  /**
   * Ensure that the {@code viewportCache} is set to a valid non-null image, and reset
   * {@code viewportPosOnScrollable} if the image contents was lost.
   */
  private void validateViewportCache(
      Graphics2D g, int transparencyMode, Dimension viewportDimension)
  {
    clearViewportCacheIfIncompatible(g, transparencyMode, viewportDimension);
    if (viewportCache == null) {
      if (viewportCachePosOnScrollable != null)
        throw new AssertionError();
      viewportCache = g.getDeviceConfiguration().createCompatibleVolatileImage(
          viewportDimension.width, viewportDimension.height, transparencyMode);
      viewportCacheFRC = g.getFontRenderContext();
    }
  }

  private Graphics2D createViewportCacheGraphics() {
    Preconditions.checkState(viewportCache != null);
    Graphics2D ret = viewportCache.createGraphics();
    // VolatileImage.createGraphics oddly enough does not set a clipping area by itself.
    ret.clipRect(0, 0, viewportCache.getWidth(), viewportCache.getHeight());
    return ret;
  }

  /**
   * Paint the specified view rectangle on the supplied graphics at the origin of the latter's
   * coordinate system. Painting operations may be cached in intermediate buffers.
   *
   * @param viewportG will be drawn on, but will not otherwise have its state modified
   * @param transparencyMode transparency mode to use for cached image buffers, as defined by
   *        {@link java.awt.Transparency}
   */
  public void paintCached(Graphics2D viewportG, int transparencyMode,
      Rectangle viewportRectOnScrollable, Painter scrollablePainter)
  {
    Preconditions.checkArgument(viewportRectOnScrollable.x >= 0);
    Preconditions.checkArgument(viewportRectOnScrollable.y >= 0);
    Preconditions.checkNotNull(viewportG);
    Preconditions.checkNotNull(scrollablePainter);
    if (viewportRectOnScrollable.isEmpty())
      return;
    /* I once experimented with an optimization that would make the viewportCache larger than the
    actual viewport in order to not have to do a repaint on every single scroll movement. It made
    scolling jerky, though, which is unacceptable even if the average FPS is higher. Besides, I
    would have had to add a requirement that OverlayScrollable.markDirty be called even for areas
    outside the viewport. And it could break optimizations in the scrollable component that use the
    viewport rectangle to optimize searches for dirty areas. */
    final Rectangle cacheViewportRectOnScrollable = viewportRectOnScrollable;
    final int cacheOffsetX = cacheViewportRectOnScrollable.x - viewportRectOnScrollable.x;
    final int cacheOffsetY = cacheViewportRectOnScrollable.y - viewportRectOnScrollable.y;
    do {
      validateViewportCache(viewportG, transparencyMode, cacheViewportRectOnScrollable.getSize());
      /* First update the viewportCache to contain an appropriately positioned image of the
      scrollable itself, not including the scrollable's child components. */
      final Graphics2D vcg = createViewportCacheGraphics();
      try {
        blitAndRepaintExposedSides(vcg, cacheViewportRectOnScrollable, scrollablePainter);
      } finally {
        vcg.dispose();
      }
      if (!viewportG.getClipBounds().isEmpty())
        viewportG.drawImage(viewportCache, cacheOffsetX, cacheOffsetY, null);
    } while (viewportCache.contentsLost());
  }

  public void updateDirtyCacheArea(Painter painter, Rectangle dirtyRectOnScrollable) {
    Preconditions.checkNotNull(painter);
    Preconditions.checkNotNull(dirtyRectOnScrollable);
    if (viewportCache == null ||
        viewportCachePosOnScrollable == null || viewportCache.contentsLost())
    {
      return;
    }
    Graphics2D vcg = createViewportCacheGraphics();
    try {
      paintViewport(vcg, painter, new Rectangle(
          dirtyRectOnScrollable.x - viewportCachePosOnScrollable.x,
          dirtyRectOnScrollable.y - viewportCachePosOnScrollable.y,
          dirtyRectOnScrollable.width, dirtyRectOnScrollable.height));
    } finally {
      vcg.dispose();
    }
  }

  @FunctionalInterface
  public interface Painter {
    /** @param g may be modified, but should not be disposed */
    public void paint(Graphics g);
  }
}
