// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.scroll;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

final class DirtyCoalescer {
  private final int maxDirtyRectangles;
  private final float minimumRelativeCostForNewRectangle;
  private final List<Rectangle> dirtyRectangles = Lists.newArrayList();

  public DirtyCoalescer(int maxDirtyRectangles, float minimumRelativeCostForNewRectangle) {
    Preconditions.checkArgument(maxDirtyRectangles >= 1);
    Preconditions.checkArgument(minimumRelativeCostForNewRectangle >= 0);
    this.maxDirtyRectangles = maxDirtyRectangles;
    this.minimumRelativeCostForNewRectangle = minimumRelativeCostForNewRectangle;
  }

  public void clear() {
    dirtyRectangles.clear();
  }

  public Collection<Rectangle> getCoalescedDirtyRectangles() {
    List<Rectangle> ret = new ArrayList<>(dirtyRectangles.size());
    for (Rectangle rect : dirtyRectangles)
      ret.add(new Rectangle(rect));
    return Collections.unmodifiableCollection(ret);
  }

  public void recoalesce() {
    Collection<Rectangle> oldRectangles = getCoalescedDirtyRectangles();
    clear();
    for (Rectangle oldRectangle : oldRectangles)
      markDirty(oldRectangle);
  }

  private static long getArea(Rectangle rect) {
    return rect.isEmpty() ? 0 : (((long) rect.width) * ((long) rect.height));
  }

  public void markDirty(Rectangle newRectangle) {
    if (newRectangle.isEmpty())
      return;
    final long newRectangleArea = getArea(newRectangle);
    long lowestUnionCost = Long.MAX_VALUE;
    int lowestUnionCostIndex = -1;
    Rectangle lowestUnionCostUnionRectangle = null;
    for (int i = 0; i < dirtyRectangles.size(); i++) {
      final Rectangle existingRectangle = dirtyRectangles.get(i);
      final long existingRectangleArea = getArea(existingRectangle);
      final Rectangle unionRectangle = existingRectangle.union(newRectangle);
      final long unionRectangleArea = getArea(unionRectangle);
      final long unionCost = unionRectangleArea - existingRectangleArea - newRectangleArea;
      if (unionCost <= lowestUnionCost) {
        lowestUnionCost = unionCost;
        lowestUnionCostIndex = i;
        lowestUnionCostUnionRectangle = unionRectangle;
      }
    }
    if (lowestUnionCostUnionRectangle == null ||
        dirtyRectangles.size() < maxDirtyRectangles &&
        lowestUnionCost >= newRectangleArea * minimumRelativeCostForNewRectangle)
    {
      dirtyRectangles.add(new Rectangle(newRectangle));
    } else {
      dirtyRectangles.set(lowestUnionCostIndex, lowestUnionCostUnionRectangle);
    }
  }
}
