// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.scroll;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Rectangle;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.swing.CellRendererPane;
import javax.swing.JComponent;
import javax.swing.JScrollPane;

/**
 * Implemented by {@code JComponent} implementations that need to overlay graphics that remains
 * static during scrolling in a {@link JScrollPane}. The {@code JScrollPane} must use
 * {@link OverlayViewport} as its viewport implementation.
 *
 * <p>The naming convention of this API uses the terms <em>scrollable</em> and <em>viewport</em> to
 * distinguish between the coordinate systems of the component being scrolled and that of the static
 * {@code OverlayViewport}. We avoid the term "view", as it is inconsistently used in the
 * {@code JScrollPane} Javadoc (e.g. {@code setView} vs. {@code setViewPosition}).
 *
 * <p>{@code Graphics} parameters are always in the scrollable's coordinate system.
 */
public interface OverlayScrollable {
  /**
   * The method {@link JComponent#paintComponent(Graphics)}, exposed to be public. The scrolling
   * algorithm will attempt to call this only when necessary, with the smallest possible clipping
   * rectangle set.
   */
  public void paintComponent(Graphics scrollableG);

  /**
   * Paint the graphics to be overlaid on the viewport. If the scrollable has child components, they
   * should be painted here. This method is called on every scroll movement. Thus, this method
   * should be cheap to run for scroll performance to remain good.
   *
   * <p>Note that painting of child components from this method typically requires that the client
   * temporarily disable double buffering on those components; see the source for
   * {@link CellRendererPane#paintComponent(Graphics,Component,Container,int,int,int,int,boolean)}.
   */
  public void paintOverlay(Graphics scrollableG);

  public void setDirtyScrollableObserver(@Nullable DirtyScrollableObserver observer);

  public interface DirtyScrollableObserver {
    /**
     * Must be called whenever an area that would be painted by
     * {@link OverlayScrollable#paintComponent(Graphics)} has changed. Dirty areas that lie entirely
     * outside the enclosing scroll pane's viewport need not be reported. Generally, the scrollable
     * component should call this method before every one of its own calls to
     * {@link Component#repaint()} and related methods (though not on {@code repaint} calls made
     * from outside the component itself). The scrollable must still call {@code repaint} to report
     * dirty areas to Swing in the usual way.
     *
     * <p>This interface is needed because the {@code OverlayViewport} would otherwise have no way
     * to separately track repaints caused by changes in the underlying scrollable and repaints
     * caused by the scrolling itself, which is needed for efficient scrolling.
     *
     * <p>Repaints on the scrollable's child components, if any, need not be reported here, as child
     * components are always repainted as part of {@link OverlayScrollable#paintOverlay(Graphics)}.
     */
    public void markScrollableDirty(@Nonnull Rectangle rectOnScrollable);

    /**
     * Indicate that the overlay should be repainted.
     */
    public void markOverlayDirty();
  }
}
