// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.scroll;

import com.google.common.base.Preconditions;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Transparency;
import java.lang.reflect.InvocationTargetException;
import javax.annotation.Nullable;
import javax.swing.JViewport;
import com.sieuferd.uiutil.scroll.OverlayScrollable.DirtyScrollableObserver;
import java.awt.Color;

/**
 * Viewport capable of overlaying additional graphics for a scrollable implementing
 * {@link OverlayScrollable}.
 */
public class OverlayViewport extends JViewport {
  private final BackingStore cache = new BackingStore();
  private final DirtyCoalescer dirtyCoalescer = new DirtyCoalescer(4, 1.0f);
  private final DirtyScrollableObserver dirtyScrollableObserver = new DirtyScrollableObserver() {
    @Override
    public void markScrollableDirty(Rectangle rectOnScrollable) {
      Preconditions.checkNotNull(rectOnScrollable);
      dirtyCoalescer.markDirty(rectOnScrollable);
    }

    @Override
    public void markOverlayDirty() {
      // Just repaint the entire viewport from the cache.
      repaint();
    }
  };

  public OverlayViewport() {
    /* Setting SIMPLE_SCROLL_MODE here allows me to implement a custom scroll implementation by
    overriding isPaintingOrigin(boolean), setViewPosition(Point), and paint(Graphics). */
    setScrollMode(JViewport.SIMPLE_SCROLL_MODE);
  }

  private @Nullable OverlayScrollable getOverlayScrollable() {
    Component ret = getView();
    return (ret instanceof OverlayScrollable) ? (OverlayScrollable) ret : null;
  }

  @Override
  public void setView(Component scrollable) {
    Preconditions.checkArgument(scrollable == null || scrollable instanceof OverlayScrollable);
    OverlayScrollable oldView = getOverlayScrollable();
    if (oldView != null)
      oldView.setDirtyScrollableObserver(null);
    super.setView(scrollable);
    OverlayScrollable newView = getOverlayScrollable();
    if (newView != null)
      newView.setDirtyScrollableObserver(dirtyScrollableObserver);
  }

  @Override
  public final void setScrollMode(int mode) {
    if (mode != JViewport.SIMPLE_SCROLL_MODE)
      throw new UnsupportedOperationException();
    super.setScrollMode(mode);
  }

  // See RepaintManager.volatileBufferType.
  private final int transparencyMode = isSwingBackbufferTranslucencySupported()
      ? Transparency.TRANSLUCENT : Transparency.OPAQUE;

  private static boolean isSwingBackbufferTranslucencySupported() {
    try {
      final Class<?> dtClass = Class.forName("sun.awt.SunToolkit");
      final Toolkit tk = Toolkit.getDefaultToolkit();
      if (dtClass.isAssignableFrom(tk.getClass())) {
        return Boolean.TRUE.equals(
            dtClass.getDeclaredMethod("isSwingBackbufferTranslucencySupported").invoke(tk));
      }
    } catch (ClassNotFoundException | NoSuchMethodException | SecurityException |
        IllegalAccessException | IllegalArgumentException | InvocationTargetException e)
    {
      // Fall through to return false.
    }
    return false;
  }

  @Override
  protected boolean isPaintingOrigin() {
    /* Must be true to ensure that the overlay is repainted whenever the scrollable's child
    components request repaints. */
    return true;
  }

  @Override
  public void setViewPosition(Point p) {
    /* Our current scroll algorithm does not need to do anything special in this method, so just
    keep the existing implementation. The blitting-related code in the superclass method never runs
    because we set SIMPLE_SCROLL_MODE before. */
    super.setViewPosition(p);
  }

  @Override
  public void paint(Graphics viewportG) {
    final OverlayScrollable os = getOverlayScrollable();
    if (os == null) {
      super.paint(viewportG);
      return;
    }
    final Point viewportPos = getViewPosition();
    final Rectangle viewportRectOnScrollable =
        new Rectangle(viewportPos.x, viewportPos.y, getWidth(), getHeight());
    final Graphics2D g2 = (Graphics2D) viewportG.create(0, 0, getWidth(), getHeight());
    cache.clearViewportCacheIfIncompatible(
        g2, transparencyMode, viewportRectOnScrollable.getSize());
    dirtyCoalescer.recoalesce();
    for (Rectangle rectOnScrollable : dirtyCoalescer.getCoalescedDirtyRectangles())
      cache.updateDirtyCacheArea(getOverlayScrollable()::paintComponent, rectOnScrollable);
    dirtyCoalescer.clear();
    // Clipping mostly for the benefit of paintOverlay, since paintCached will do its own clipping.
    try {
      if (isOpaque()) {
        Color oldColor = g2.getColor();
        g2.setColor(getBackground());
        g2.fillRect(0, 0, getWidth(), getHeight());
        g2.setColor(oldColor);
      }
      cache.paintCached(g2, transparencyMode, viewportRectOnScrollable, os::paintComponent);
      g2.translate(-viewportPos.x, -viewportPos.y);
      os.paintOverlay(g2);
    } finally {
      g2.dispose();
    }
  }
}
