// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.table;

import com.sieuferd.uiutil.SwingString;
import com.sieuferd.uiutil.api.TableColors;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Point;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;

/**
 * Table cell renderer that allows values from left-adjacent cells overflow into its own cell if its
 * own cell contains a null value.
 */
public class SpanningTableCellRenderer implements TableCellRenderer {
  public static final int RIGHT_MARGIN_PX = 3;
  private static final Border EMPTY_BORDER = BorderFactory.createEmptyBorder();
  private final boolean allowHtml;
  private final TableColors tableColors;
  private final JPanel wrappingPanel;
  private final JScrollPane scrollPane;
  private @Nullable Component delegate;

  public SpanningTableCellRenderer(boolean allowHtml, TableColors tableColors) {
    this.allowHtml = allowHtml;
    this.tableColors = tableColors;
    wrappingPanel = new JPanel(new BorderLayout());
    wrappingPanel.setOpaque(false);
    wrappingPanel.add(
        // Avoid overflow errors while still specifying an effectively infinite width.
        Box.createHorizontalStrut(Integer.MAX_VALUE / 32),
        BorderLayout.NORTH);
    /* Use the technique from
    http://stackoverflow.com/questions/16305023/jtable-complex-cell-renderer . */
    scrollPane =  new JScrollPane();
    scrollPane.setViewportView(wrappingPanel);
    scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_NEVER);
    scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
    scrollPane.setViewportBorder(EMPTY_BORDER);
    scrollPane.setOpaque(false);
    scrollPane.getViewport().setOpaque(false);
  }

  private @Nullable Object getDisplayableCellValue(
      JTable table, int rowIndex, int columnIndex, boolean requireSpanning)
  {
    if (columnIndex >= table.getColumnCount())
      return null;
    final Object value = table.getModel().getValueAt(
        table.convertRowIndexToModel(rowIndex), table.convertColumnIndexToModel(columnIndex));
    if (table.getCellRenderer(rowIndex, columnIndex) instanceof SpanningTableCellRenderer) {
      return value;
    } else {
      return requireSpanning ? null : (value == null ? "" : value);
    }
  }

  @Override
  public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
      boolean hasFocus, int rowIndex, int columnIndex)
  {
    Object displayValue = null;
    int displayColumnIndex = -1;
    int horizontalAlignment = SwingConstants.LEFT;
    for (int siblingColumnIndex = columnIndex; siblingColumnIndex >= 0; siblingColumnIndex--) {
      final Object siblingDisplayValue =
          getDisplayableCellValue(table, rowIndex, siblingColumnIndex, true);
      if (siblingDisplayValue != null) {
        if (siblingDisplayValue instanceof StyledTableString) {
          horizontalAlignment = ((StyledTableString) siblingDisplayValue).getHorizontalAlignment();
          if (horizontalAlignment != SwingConstants.LEFT && siblingColumnIndex != columnIndex) {
            horizontalAlignment = SwingConstants.LEFT;
            // Only left-aligned values may overflow from one column to the next.
            displayValue = null;
          } else {
            displayValue =
                ((StyledTableString) siblingDisplayValue).getString(isSelected).getComplete();
          }
        } else {
          displayValue = allowHtml ?
              siblingDisplayValue : SwingString.escapeComplete(siblingDisplayValue.toString());
        }
        displayColumnIndex = siblingColumnIndex;
        break;
      }
    }

    /* Don't paint the contents of another cell in this cell if this cell is selected and the other
    one is not. */
    if (displayValue != null && isSelected && !table.isCellSelected(rowIndex, displayColumnIndex))
      displayValue = null;

    // TODO[pd]: Consider paining a border in the hasFocus case.
    /* Use the hasFocus and isSelected of the cell in this physical location rather than of the cell
    whose contents is being painted.  */
    Color foregroundColor = tableColors.getForeground(table, isSelected);
    Color backgroundColor = tableColors.getBackground(table, isSelected);
    // TODO[pd]: Can I reuse newDelegate?
    /* I used to use HtmlRenderer.createLabel() here, but search highlighting requires styling of
    the text background, which is not supported by the latter. */
    JLabel newDelegate = new JLabel();
    newDelegate.setForeground(foregroundColor);
    newDelegate.setBackground(backgroundColor);
    newDelegate.setText(Objects.toString(displayValue, ""));
    newDelegate.setOpaque(true);
    newDelegate.setHorizontalAlignment(horizontalAlignment);
    newDelegate.setBorder(horizontalAlignment == SwingConstants.RIGHT
        ? BorderFactory.createMatteBorder(0, 0, 0, RIGHT_MARGIN_PX, backgroundColor)
        : EMPTY_BORDER);
    if (horizontalAlignment != SwingConstants.LEFT)
      return newDelegate;

    if (delegate != null)
      wrappingPanel.remove(delegate);
    delegate = newDelegate;
    wrappingPanel.add(delegate, BorderLayout.CENTER);

    scrollPane.setBorder(getDisplayableCellValue(table, rowIndex, columnIndex + 1, false) != null
        ? BorderFactory.createMatteBorder(0, 0, 0, RIGHT_MARGIN_PX, backgroundColor)
        : EMPTY_BORDER);

    Point thisCellLocation = table.getCellRect(rowIndex, columnIndex, false).getLocation();
    Point displayLocation = displayColumnIndex < 0 ? thisCellLocation :
        table.getCellRect(rowIndex, displayColumnIndex, false).getLocation();
    scrollPane.getViewport().setViewPosition(new Point(
        thisCellLocation.x - displayLocation.x, thisCellLocation.y - displayLocation.y));
    return scrollPane;
  }
}