// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.uiutil.table;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Sets;
import java.util.List;
import java.util.Set;
import javax.annotation.Nullable;
import javax.swing.event.EventListenerList;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;

// TODO[cc]: Make use of this. Will probably need some debugging (hasn't been tested before).

// Don't bother extending AbstractTableModel, as we need to override everything.
public final class UnionTableModel implements TableModel {
  private final ImmutableList<TableModel> delegates;
  private final EventListenerList listenerList = new EventListenerList();
  private final TableModelListener delegateListener = this::fireTableChanged;

  public UnionTableModel(List<TableModel> delegates) {
    this.delegates = ImmutableList.copyOf(delegates);
  }

  private void fireTableChanged(TableModelEvent evt) {
    // TODO[pd]: Recalculate row indices.
    TableModelEvent delegateEvt = new TableModelEvent(this,
        evt.getFirstRow(), evt.getLastRow(), evt.getColumn(), evt.getType());
    // As in AbstractTableModel.fireTableChanged().
    Object[] listeners = listenerList.getListenerList();
    for (int i = listeners.length - 2; i >= 0; i -= 2) {
      if (listeners[i] == TableModelListener.class)
        ((TableModelListener) listeners[i + 1]).tableChanged(delegateEvt);
    }
  }

  @Override
  public int getRowCount() {
    int ret = 0;
    for (TableModel delegate : delegates)
      ret += delegate.getRowCount();
    return ret;
  }

  @Override
  public int getColumnCount() {
    int ret = 0;
    for (TableModel delegate : delegates)
      ret = Math.max(ret, delegate.getColumnCount());
    return ret;
  }

  private @Nullable Object getEmptyCellValue(int columnIndex) {
    // TODO[pd]: Consider caching.
    Class<?> type = getColumnClass(columnIndex);
    if (type.equals(String.class)) {
      // Use null here to be compatible with SpanningTableCellRenderer.
      return null;
    } else if (type.equals(Boolean.class) || type.equals(Boolean.TYPE)) {
      return false;
    } else {
      return null;
    }
  }

  @Override
  public @Nullable Object getValueAt(int rowIndex, int columnIndex) {
    DelegateRowIndex dri = getDelegateRowIndex(rowIndex);
    return (columnIndex < dri.delegate.getColumnCount())
        ? dri.delegate.getValueAt(dri.rowIndex, columnIndex)
        : getEmptyCellValue(columnIndex);
  }

  @Override
  public String getColumnName(int columnIndex) {
    Set<String> ret = Sets.newLinkedHashSet();
    for (TableModel delegate : delegates) {
      if (columnIndex < delegate.getColumnCount())
        ret.add(delegate.getColumnName(columnIndex));
    }
    ret.remove("");
    if (ret.isEmpty())
      return "";
    if (ret.size() > 1) {
      throw new IllegalArgumentException("Delegate models do not agree on a name for column " +
          columnIndex + " (got names " + ret + ")");
    }
    return ret.iterator().next();
  }

  @Override
  public Class<?> getColumnClass(int columnIndex) {
    Set<Class<?>> ret = Sets.newLinkedHashSet();
    for (TableModel delegate : delegates) {
      if (columnIndex < delegate.getColumnCount())
        ret.add(delegate.getColumnClass(columnIndex));
    }
    if (ret.size() > 1) {
      throw new IllegalArgumentException("Delegate models do not agree on a type for column " +
          columnIndex + " (got types " + ret + ")");
    }
    if (ret.isEmpty())
      return Object.class;
    return ret.iterator().next();
  }

  @Override
  public boolean isCellEditable(int rowIndex, int columnIndex) {
    DelegateRowIndex dri = getDelegateRowIndex(rowIndex);
    return (columnIndex < dri.delegate.getColumnCount())
        ? dri.delegate.isCellEditable(dri.rowIndex, columnIndex)
        : false;
  }

  @Override
  public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
    DelegateRowIndex dri = getDelegateRowIndex(rowIndex);
    dri.delegate.setValueAt(aValue, dri.rowIndex, columnIndex);
  }

  @Override
  public void addTableModelListener(TableModelListener l) {
    if (listenerList.getListenerCount() == 0) {
      for (TableModel delegate : delegates)
        delegate.addTableModelListener(delegateListener);
    }
    listenerList.add(TableModelListener.class, l);
  }

  @Override
  public void removeTableModelListener(TableModelListener l) {
    listenerList.remove(TableModelListener.class, l);
    if (listenerList.getListenerCount() == 0) {
      for (TableModel delegate : delegates)
        delegate.removeTableModelListener(delegateListener);
    }
  }

  private DelegateRowIndex getDelegateRowIndex(int rowIndex) {
    int rowOffset = 0;
    for (TableModel delegate : delegates) {
      final int delegateRowCount = delegate.getRowCount();
      final int delegateRowIndex = rowIndex - rowOffset;
      if (delegateRowIndex < delegateRowCount)
        return new DelegateRowIndex(delegate, delegateRowIndex);
      rowOffset += delegateRowCount;
    }
    throw new IndexOutOfBoundsException();
  }

  private static final class DelegateRowIndex {
    final TableModel delegate;
    final int rowIndex;

    DelegateRowIndex(TableModel delegate, int index) {
      Preconditions.checkNotNull(delegate);
      Preconditions.checkArgument(index >= 0);
      this.delegate = delegate;
      this.rowIndex = index;
    }
  }
}
