// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.sieuferd.cell.support.Selection;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.HierarchyEvent;
import java.awt.event.HierarchyListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Rectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.ComponentUI;
import com.sieuferd.widget.PhysicalCursorMovementAction.CursorDistance;
import com.sieuferd.cell.support.Canvas;
import com.sieuferd.cell.support.Cell.FocusMode;
import com.sieuferd.cell.support.CellRectangles;
import com.sieuferd.cell.support.CellRectangles.HeaderRectangle;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Location;
import com.sieuferd.cell.support.Direction;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.Points;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.SelectionModel;
import com.sieuferd.cell.support.SelectionModel.SelectionModelPropertyChangeEvent;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.api.PopupLike;
import com.sieuferd.uiutil.scroll.OverlayViewport;
import com.sieuferd.widget.InteractionProvider.Acceptance;
import java.awt.Component;
import java.awt.KeyboardFocusManager;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.util.Map;
import java.util.logging.Level;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.swing.KeyStroke;
import javax.swing.Timer;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import org.netbeans.api.annotations.common.StaticResource;
import org.openide.util.Utilities;

/* TODO[pd]: Figure out why the focus indication of the inplace editor is not repainted properly
             when a cell is double-clicked to edit (only visible if preparing the editor with
             select-all). */
final class BasicCellWidgetUI extends CellWidgetUI {
  /* This class follows the pattern described in
  http://today.java.net/pub/a/today/2006/10/03/enhancing-swing-applications.html . While I don't
  ever anticipate implementing more than a single L&F for the CellWidget, using the same
  architecture as the standard Swing components makes it easier to compare and contrast my custom
  component implementation with the former. It also makes for a convenient separation between the
  JavaBean interface of the CellWidget JComponent and the rest of the logic. */

  @StaticResource
  private static final String RESOURCE_CELL_CURSOR =
      "com/sieuferd/widget/resources/cell_cursor.gif";
  private static final Logger LOG = Logger.getLogger(BasicCellWidgetUI.class.getName());

  private static final boolean DEBUG_USE_INPLACE_EDITOR_PREFERRED_SIZE = false;
  /* It's useful to be able to open inplace editors even for non-editable cells, for instance to be
  able to select a substring for copying to the clipboard, or to be able to see the entire contents
  of clipped or over-fitted cells (e.g. per TextFitPolicy.FIT_LINE/OVERFLOW_LINE). */
  private static final boolean ALLOW_READONLY_INPLACE_EDITOR = true;
  private static final boolean POSITION_TOOLTIPS_IN_VIEWPORT = true;
  private static final Pixels CURSOR_TOOLTIP_UNDER_LAYOUT_HEIGHT_THRESHOLD = Pixels.create(200);
  private static final int CURSOR_TOOLTIP_UPDATE_COALESCE_MS = 250;
  private static final Points TOOLTIP_CURSOR_MARGIN = new Points(4.0);
  /* The maximum horizontal distance from the start of a cell to a popup opened for that cell. A
  sensible value is a little more than a wide paragraph of text, e.g. a little more than the default
  for LHProp.TABLE_MAX_PRIMITIVE_WIDTH. */
  private static final Points MAX_HORIZONTAL_POPUP_OFFSET = new Points(350);
  /** The fraction by which a page scrolling operation is allowed to deviate from a full page
  movement in order to align the position of the old and new cursor rectangles. */
  private static final double PAGE_SCROLL_ALIGNMENT_SLACK = 0.1;
  /* I also experimented with making the infinite scroll padding always be 50% or 25% of the
  viewport height at any given time. This felt unnatural (there was too much or too little padding
  depending on the viewport size), and is non-standard and harder to do than a constant padding in
  any case. */
  private static final int INFINITE_SCROLL_PADDING_PIXELS = 150;
  /** Size limit for the cursor rectangle for use when automatically scrolling to the cursor, as a
  fraction of the viewport size. */
  private static final double CURSOR_SCROLL_VIEWPORT_FRACTION_LIMIT = 0.1;
  private static final ImmutableSet<String> REPAINTABLE_PROPERTY_CHANGES = ImmutableSet.of(
      // JComponent properties.
      "opaque", "background",
      /* CellWidget-specific properties. The PROP_RENDER_CONTEXT is not included, as it is only set
      in situations where a repaint would already have been scheduled by the toolkit, and we should
      avoid double repaints. */
      CellWidget.PROP_SCALING,
      CellWidget.PROP_FORCE_FRACTIONAL_METRICS,
      CellWidget.PROP_ALPHA_PERMITTED,
      CellWidget.PROP_DRAW_CELL_DEBUG_GRAPHICS,
      /* TODO[pd]: Be smarter about what to repaint for PROP_PATH_HIGHLIGHTS. This includes updating
                   Cell.findDirtyRegions to include path highlights. */
      CellWidget.PROP_PATH_HIGHLIGHTS);
  private final RectangleFinder rectangleFinder = new RectangleFinder();
  private final Timer updateCursorTooltipTimer = new Timer(0, new ActionListener() {
    @Override
    public void actionPerformed(ActionEvent e) {
      if (cellWidget != null)
        updateCursorTooltip();
    }
  });
  private long lastCursorUpdateTimeNanos = Long.MIN_VALUE;

  /* TODO[cc]: Encapsulate fields instead of having this weird convention. Add nullable annotations
               as appropriate. */
  /* Note: All fields should be null before installUI() or after uninstallUI() has been called. At
  all other times, no fields can be null except where noted. */

  private CellWidget cellWidget;
  private @Nullable JComponent installedCursorTooltip;
  private @Nullable InplaceEditor installedInplaceEditor;
  /** An enclosing JViewport, if any, otherwise the CellWidget itself. */
  private Container viewportContainer;
  /**
   * Keep track of viewport rectangle manually rather than querying the viewport every time it is
   * needed. This way, any bugs in the listener-related code will be readily apparent. Even though
   * the frozen header logic no longer relies on code being run in updateViewportRectangle, keeping
   * the latter method working and bug-free may be useful for future features.
   */
  private PixelRectangle viewportRectangle;
  private FocusMode focusMode;
  private @Nullable PopupLike currentlyShowingPopup;
  private @Nullable Selection rangeSelectStart;
  private boolean inPopup;
  private boolean inAcceptEdit;
  /** The cursor rectangle at the time magicCursorPoint was set. */
  private @Nullable PixelRectangle magicCursorPointRectangle;
  private @Nonnull Coordinates magicCursorPoint = Coordinates.ZERO;
  private @Nonnull Coordinates nonReferenceMagicCursorPoint = Coordinates.ZERO;

  BasicCellWidgetUI() {
    updateCursorTooltipTimer.setCoalesce(true);
    updateCursorTooltipTimer.setRepeats(false);
  }

  /** Could be called reflectively from UIDefaults.getUI(). */
  public static ComponentUI createUI(JComponent c) {
    return new BasicCellWidgetUI();
  }

  @Override
  public void installUI(JComponent c) {
    cellWidget = (CellWidget) c;
    cellWidget.setRenderContext(null);
    updateViewportContainer();
    updateViewportRectangle();
    updateFocusMode();
    installDefaults();
    installComponents();
    installListeners();
    installKeyboardActions();
    updateCursorTooltip();
  }

  @Override
  public void uninstallUI(JComponent c) {
    setCurrentlyShowingPopup(null);
    uninstallKeyboardActions();
    uninstallListeners();
    uninstallComponents();
    uninstallDefaults();
    viewportContainer = null;
    setViewportRectangle(null);
    focusMode = FocusMode.UNFOCUSED;
    cellWidget.setRenderContext(null);
    cellWidget = null;
    rangeSelectStart = null;
  }

  private void installDefaults() {
    cellWidget.setBackground(Color.WHITE);
    cellWidget.setOpaque(true);
    Cursor cellCursor = Toolkit.getDefaultToolkit().createCustomCursor(
        Toolkit.getDefaultToolkit().getImage(getClass().getClassLoader().getResource(
        RESOURCE_CELL_CURSOR)), new Point(14, 14), "Cell Cursor");
    cellWidget.setCursor(cellCursor);
  }

  private void uninstallDefaults() {
    cellWidget.setCursor(null);
  }

  private void installComponents() {
  }

  private void uninstallComponents() {
    setInstalledCursorTooltip(null);
    setInstalledInplaceEditor(null);
  }

  private void installListeners() {
    viewportContainer.addComponentListener(containerComponentListener);
    JScrollPane scrollPane = getEnclosingScrollPane();
    if (scrollPane != null) {
      scrollPane.getViewport().addChangeListener(viewportChangeListener);
      /* During mouse wheel scrolling, the viewport does not fire scroll events until the viewport
      is repainted. This is too late for frozen headers to be updated without appearing to first
      move ever so briefly. Luckily, the scrollbars fire the events _before_ the repaints in this
      case, so just listen to those as well with the same listener. */
      scrollPane.getVerticalScrollBar().getModel().addChangeListener(viewportChangeListener);
      scrollPane.getHorizontalScrollBar().getModel().addChangeListener(viewportChangeListener);
    }
    cellWidget.addHierarchyListener(hierarchyListener);
    cellWidget.addFocusListener(focusListener);
    cellWidget.addMouseListener(mouseListener);
    cellWidget.addMouseMotionListener(mouseMotionListener);
    cellWidget.addPropertyChangeListener(cellWidgetPropertyChangeListener);
    getEffectiveSelectionModel().addPropertyChangeListener(selectionModelListener);
  }

  private void uninstallListeners() {
    getEffectiveSelectionModel().removePropertyChangeListener(selectionModelListener);
    cellWidget.removePropertyChangeListener(cellWidgetPropertyChangeListener);
    cellWidget.removeMouseMotionListener(mouseMotionListener);
    cellWidget.removeMouseListener(mouseListener);
    cellWidget.removeFocusListener(focusListener);
    cellWidget.removeHierarchyListener(hierarchyListener);
    JScrollPane scrollPane = getEnclosingScrollPane();
    if (scrollPane != null) {
      scrollPane.getVerticalScrollBar().getModel().removeChangeListener(viewportChangeListener);
      scrollPane.getHorizontalScrollBar().getModel().removeChangeListener(viewportChangeListener);
      scrollPane.getViewport().removeChangeListener(viewportChangeListener);
    }
    viewportContainer.removeComponentListener(containerComponentListener);
  }

  private void installKeyboardActions() {
    SwingUtilities.replaceUIActionMap(cellWidget, CellWidgetActions.createActionMap());
    SwingUtilities.replaceUIInputMap(cellWidget, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT,
        createInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT));
  }

  private void uninstallKeyboardActions() {
    SwingUtilities.replaceUIActionMap(cellWidget, null);
    SwingUtilities.replaceUIInputMap(
        cellWidget, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, null);
  }

  // Based on BasicTableUI.getInputMap().
  private @Nullable InputMap createInputMap(int condition) {
    if (condition == JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT) {
      InputMap keyMap = (InputMap) UIManager.get("Table.ancestorInputMap");
      InputMap rtlKeyMap;

      if (cellWidget.getComponentOrientation().isLeftToRight() ||
          ((rtlKeyMap = (InputMap) UIManager.get("Table.ancestorInputMap.RightToLeft")) == null))
      {
        /* Include this case for completeness, although right-to-left orientation is not really
        supported yet. */
        return keyMap;
      } else {
        rtlKeyMap.setParent(keyMap);
        return rtlKeyMap;
      }
    }
    return null;
  }

  /* === Constant Listener Objects ============================================================== */
  /* Could use addNotify()/removeNotify() here instead, like JTable does, but using a
  HierarchyListener seems to work equally well. */
  private final HierarchyListener hierarchyListener = (HierarchyEvent e) -> {
    // Could be more specific here, but doing a full uninstall/install is easier.
    uninstallListeners();
    updateViewportContainer();
    installListeners();
    updateViewportRectangle();
    // TODO[pd]: Consider removing this.
    updateRenderContext();
  };
  private final ComponentListener containerComponentListener = new ComponentAdapter() {
    @Override
    public void componentResized(ComponentEvent evt) {
      UIUtil.ensureSufficientDoubleBufferSize(cellWidget);
      updateViewportRectangle();
    }
  };
  private final ChangeListener viewportChangeListener = (ChangeEvent ce) -> {
    updateViewportRectangle();
  };
  private final FocusListener focusListener = new FocusListener() {
    @Override
    public void focusGained(FocusEvent e) {
      ThreadUtil.checkOnEDT();
      /* This is mostly insurance in case the focus goes to the CellWidget in any other way than via
      calls to CellWidget.requestFocus or requestFocusInWindow (Javadocs warn of this
      possibility). */
      if (cellWidget.isFocusOwner() && installedInplaceEditor != null) {
        // TODO[pd]: Get rid of this logging statement once the focus logic is stable.
        LOG.info("Redirecting keyboard focus from CellWidget to inplace editor");
        installedInplaceEditor.getFocusComponent().requestFocusInWindow();
      }
      updateFocusMode();
    }

    @Override
    public void focusLost(FocusEvent e) {
      ThreadUtil.checkOnEDT();
      /* Unlike JTable, we don't end an inplace edit when the focus moves to a different component.
      This is the same behavior as Excel on both MacOS and Windows, and makes sense for instance
      when considering that the focus might have moved to the formula bar. */
      /* Use invokeLater to avoid a potential flicker in the focus indication while the focus is
      transferred to an inplace editor component. */
      SwingUtilities.invokeLater(BasicCellWidgetUI.this::updateFocusMode);
    }
  };
  private final PropertyChangeListener inplaceEditorSizeListener = (PropertyChangeEvent evt) -> {
    updateInplaceEditorBounds();
  };
  private final PropertyChangeListener selectionModelListener = (final PropertyChangeEvent evt) -> {
    Preconditions.checkArgument(evt.getPropertyName().equals(SelectionModel.PROP_SELECTION));
    final Selection oldSelection = (Selection) evt.getOldValue();
    final Selection newSelection = getEffectiveSelectionModel().getSelection();
    if (!Location.equals(
        oldSelection.getCursorLocation(), newSelection.getCursorLocation())) {
      // Coalesce cursor tooltip updates without delaying infrequent ones.
      long currentTimeNanos = System.nanoTime();
      long nextCursorTooltipUpdateTimeNanos =
          lastCursorUpdateTimeNanos + CURSOR_TOOLTIP_UPDATE_COALESCE_MS * 1000000L;
      lastCursorUpdateTimeNanos = currentTimeNanos;
      if (currentTimeNanos >= nextCursorTooltipUpdateTimeNanos) {
        /* Some scrolling operations update the viewport position after first updating the
        selection model; wait until this has been done before updating the cursor tooltip. */
        SwingUtilities.invokeLater(this::updateCursorTooltip);
      } else {
        long timeUntilCursorTooltipUpdate = nextCursorTooltipUpdateTimeNanos - currentTimeNanos;
        if (installedCursorTooltip != null) {
          setInstalledComponentBounds(
              installedCursorTooltip, installedCursorTooltip.getBounds(), null);
        }
        updateCursorTooltipTimer.setInitialDelay((int) (timeUntilCursorTooltipUpdate / 1000000L));
        updateCursorTooltipTimer.restart();
      }
      migrateMagicCursorPoint();
      /* By the time the selection has changed, it is too late to try to accept an ongoing edit,
      since the cursor will have moved away from its original location and the editor might have
      needed to show a validation dialog. On the other hand, it's not strictly necessary to cancel
      any ongoing edit either, and it's possible that the selection change was in itself triggered
      in the process of canceling an edit (when turning off reference selection mode). Thus, it's
      best not to touch the editing state at this point. */
    }
    final GeneratedCell generatedCell = getDisplayableGeneratedCell();
    if (generatedCell == null)
      return;
    // Repaint affected regions only.
    final List<PixelRectangle> dirtyRegions = Lists.newArrayList();
    dirtyRegions.addAll(generatedCell.getCell().findDirtyRegions(
        generatedCell.getRenderContext(), getClipBounds(),
        selectionDirtyDifference(oldSelection, newSelection)));
    dirtyRegions.addAll(generatedCell.getCell().findDirtyRegions(
        generatedCell.getRenderContext(), getClipBounds(),
        selectionDirtyDifference(newSelection, oldSelection)));
    /* TODO[pd]: If I need to support path highlights without correlation indication, only include
                 highlights with correlation indication here. */
    if (!cellWidget.getPathHighlights().getHighlights().isEmpty() && !Location.equals(
        oldSelection.getCursorLocation(), newSelection.getCursorLocation()))
    {
      dirtyRegions.addAll(generatedCell.getCell().findDirtyRegions(
          generatedCell.getRenderContext(), getClipBounds(),
          new Selection(null, cellWidget.getPathHighlights().getHighlights().keySet())));
    }
    if (!(evt instanceof SelectionModelPropertyChangeEvent) ||
        ((SelectionModelPropertyChangeEvent) evt).getScrollToCursor())
    {
      scrollToCursor();
    }
    if (cellWidget.getDrawCellDebugGraphics()) {
      repaintAll();
    } else {
      repaintDirty(dirtyRegions);
    }
  };
  private final PropertyChangeListener cellWidgetPropertyChangeListener =
      new PropertyChangeListener()
  {
    @Override
    public void propertyChange(PropertyChangeEvent evt) {
      final String propertyName = evt.getPropertyName();
      if (REPAINTABLE_PROPERTY_CHANGES.contains(propertyName))
        repaintAll();
      if (propertyName.equals(CellWidget.PROP_GENERATED_CELL)) {
        // TODO[cc]: Consider tearing apart GeneratedCell and then simplifying this logic.
        final boolean cellChange;
        if (evt.getOldValue() == null || evt.getNewValue() == null) {
          cellChange = true;
        } else {
          final GeneratedCell oldCell = (GeneratedCell) evt.getOldValue();
          final GeneratedCell newCell = (GeneratedCell) evt.getNewValue();
          cellChange = oldCell.getCell() != newCell.getCell();
        }
        if (cellChange || cellWidget.getDrawCellDebugGraphics())
          repaintAll();
        if (cellChange) {
          final boolean cursorWasInViewport = viewportRectangle.contains(magicCursorPoint);
          rectangleFinder.setGeneratedCell(cellWidget.getGeneratedCell());
          if (findCursorRectangle() == null) {
            Location cursorLocation = getDefaultCursorLocation();
            if (cursorLocation != null)
              getEffectiveSelectionModel().setCursorLocation(cursorLocation, false);
          }
          migrateMagicCursorPoint();
          if (inReferenceSelectionMode() &&
              findCursorRectangle(cellWidget.getSelectionModel().getCursorLocation()) == null)
          {
            // Migrate the regular selection cursor as well.
            Location cursorLocation = getDefaultCursorLocation(nonReferenceMagicCursorPoint);
            if (cursorLocation != null)
              cellWidget.getSelectionModel().setCursorLocation(cursorLocation, false);
          }
          if (cursorWasInViewport)
            scrollToCursor();
          updateCursorTooltip();
          updateInplaceEditorBounds();
          updateScrollables();
        }
      } else if (propertyName.equals(CellWidget.PROP_SELECTION_MODEL)) {
        if (!inReferenceSelectionMode()) {
          effectiveSelectionModelChanged(
              (SelectionModel) evt.getOldValue(), (SelectionModel) evt.getNewValue());
        }
      } else if (propertyName.equals(CellWidget.PROP_REFERENCE_SELECTION_MODEL)) {
        final SelectionModel oldValue = (SelectionModel) evt.getOldValue();
        final SelectionModel newValue = (SelectionModel) evt.getNewValue();
        effectiveSelectionModelChanged(
            oldValue != null ? oldValue : cellWidget.getSelectionModel(),
            newValue != null ? newValue : cellWidget.getSelectionModel());
        updateCursorTooltip();
        updateFocusMode();
      } else if (propertyName.equals(CellWidget.PROP_FREEZE_SCROLLABLES)) {
        updateScrollables();
      }
    }
  };
  private final MouseListener mouseListener = new MouseAdapter() {
    @Override
    public void mousePressed(MouseEvent evt) {
      ThreadUtil.checkOnEDT();
      /* Don't request focus in reference selection mode, as the user might be editing a formula
      using an external component like a formula bar. */
      if (!inReferenceSelectionMode())
        cellWidget.requestFocusInWindow();
      // Let maybePopup do its own select handling in the isPopupTrigger case.
      if (!evt.isPopupTrigger()) {
        boolean valid = mouseSelect(evt, false);
        if (valid && !inReferenceSelectionMode() && evt.getClickCount() == 2 &&
            !isPickSelect(evt) && !isRangeSelect(evt))
        {
          startInplaceEdit();
          return;
        }
      }
      maybePopup(evt);
    }

    @Override
    public void mouseReleased(MouseEvent evt) {
      ThreadUtil.checkOnEDT();
      maybePopup(evt);
    }

    /* Follow the pattern of org.openide.awt.MouseUtils.PopupMouseAdapter (but no need to actually
    depend on it). */
    private void maybePopup(MouseEvent evt) {
      if (evt.isPopupTrigger()) {
        /* Selection may be necessary here even if the event originated from mouseReleased(). For
        instance, on Windows, clicking the right mouse button on a new cell after the popup menu is
        already showing results in an additional call to mouseReleased but not to mousePressed. */
        boolean valid = mouseSelect(evt, false);
        if (valid && !inReferenceSelectionMode()) {
          final PopupLike popupMenu =
              cellWidget.getInteractionProvider().getMousePopup(evt.getModifiersEx());
          if (popupMenu != null)
            openPopup(popupMenu, new Coordinates(UIUtil.convertMousePoint(cellWidget, evt)));
        }
      }
    }
  };
  private final MouseMotionListener mouseMotionListener = new MouseMotionAdapter() {
    @Override
    public void mouseDragged(MouseEvent evt) {
      /* Dragging with any other button than the left one never changes the selection or scrolls the
      window. This is consistent with Excel's behavior when dragging with the right or middle
      buttons. Note that just doing "evt.getButton() != MouseEvent.BUTTON1" worked on MacOS but not
      on Windows. */
      final int modifiersEx = evt.getModifiersEx();
      if (evt.isPopupTrigger()
          || (modifiersEx & MouseEvent.BUTTON1_DOWN_MASK) == 0
          || (modifiersEx & MouseEvent.BUTTON2_DOWN_MASK) != 0
          || (modifiersEx & MouseEvent.BUTTON3_DOWN_MASK) != 0)
      {
        return;
      }
      mouseSelect(evt, true);
      /* To ensure proper autoscrolling behavior, scroll to the mouse position even though
      mouseSelect() also scrolls to newly selected cells, since a new cell might not always be
      selected even though the mouse is dragged beyond the visible area. Do this after mouseSelect
      to ensure the cursor appears to remain at the edge of the viewport at all times. Scroll to a
      box bigger than 1x1 pixels to avoid very slow scrolling. Note that mouseDragged events for
      dragging outside of component boundaries seem to be synthesized at constant 100ms intervals,
      thus no special timing code should be necessary here to ensure the same scroll speed on all
      computers. See JComponent.setAutoScrolls. */
      final int OUTSIDE_DRAG_PADDING_PX = 200;
      final Point point = UIUtil.convertMousePoint(cellWidget, evt);
      if (!viewportRectangle.contains(new Coordinates(point))) {
        Rectangle r = new Rectangle(
            point.x - OUTSIDE_DRAG_PADDING_PX, point.y - OUTSIDE_DRAG_PADDING_PX,
            OUTSIDE_DRAG_PADDING_PX * 2, OUTSIDE_DRAG_PADDING_PX * 2);
        setViewportPosition(getScrolledViewportPosition(new PixelRectangle(r), Coordinates.ZERO));
      }
    }
  };
  private final PopupMenuListener popupMenuListener = new PopupMenuListener() {
    @Override
    public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
      // No action.
    }

    @Override
    public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
      setCurrentlyShowingPopup(null);
    }

    @Override
    public void popupMenuCanceled(PopupMenuEvent e) {
      setCurrentlyShowingPopup(null);
    }
  };

  /* === Cursor Helper Methods ================================================================== */

  private Selection selectionDirtyDifference(Selection s1, Selection s2) {
    final Location cursorLocation;
    if (isOverlayEnabled()) {
      /* When overlays are enabled, the cursor is painted as part of the overlay, and so does not
      generate dirty areas on the scrollable itself. */
      cursorLocation = null;
    } else {
      cursorLocation = Location.equals(s1.getCursorLocation(), s2.getCursorLocation())
          ? null : s1.getCursorLocation();
    }
    return new Selection(
        cursorLocation, Sets.difference(s1.getSelectedPaths(), s2.getSelectedPaths()));
  }

  /* TODO[pd]: Consider whether this should return null or a CursorMovement with a null
               CursorLocation. */
  private @Nullable CursorMovement getMouseCursorMovement(MouseEvent evt) {
    /* Using getDisplayableGeneratedCell() ensures that cells cannot be selected blindly while the
    cell is not being displayed. */
    final GeneratedCell generatedCell = getDisplayableGeneratedCell();
    if (generatedCell == null)
      return null;
    /* On MacOS with Java 1.7.0_25, whether using Nimbus or the Aqua LAF, the coordinates returned
    by evt.getPoint() are off by one pixel in the Y direction, making a click on the position
    corresponding to coordinate (0,0) in the graphics context yield coordinate (0,1) in the
    MouseEvent. This happens even for a simple JPanel enclosed by a JFrame, and even when borders
    and insets are zero in all directions. Getting the screen coordinates and adjusting them for the
    component's location on screen at first seemed to yield the correct coordinates both on Windows
    and MacOS, but unfortunately it did not work properly in a MacOS-based multi-screen
    environment. Note that using SwingUtilities.convertPoint (via UIUtil.convertMousePoint) does not
    fix the bug mentioned here. It's not a significant problem, however, just potentially confusing
    during pixel-level debugging. */
    final Location cursorLocation = generatedCell.getCell().getClosestCursorLocationAt(
        generatedCell.getRenderContext(),
        new Coordinates(UIUtil.convertMousePoint(cellWidget, evt)), Direction.NONE, null,
        getViewportPosition());
    return cursorLocation == null ? null : new CursorMovement(cursorLocation, null, null);
  }

  private Selection calculateSelection(@Nullable Location cursorLocation,
      boolean popupOpening, boolean pickSelect, boolean rangeSelect)
  {
    final Selection oldSelection = getEffectiveSelectionModel().getSelection();
    final @Nullable Location oldCursorLocation = oldSelection.getCursorLocation();
    final Set<Path> pathSelection;

    if (popupOpening && cursorLocation != null &&
        Views.isAncestorOrEqualInSet(oldSelection.getSelectedPaths(), cursorLocation.getPath()))
    {
      /* Leave both the path selection and the cursor as is when opening the popup menu over a
      selected path. An exception is for the case when the cursor was previously outside of the path
      selection, as might happen if the user just de-selected a path using pick selection. In the
      latter case, the cursor is allowed to move into the path selection. */
      pathSelection = oldSelection.getSelectedPaths();
      if (oldCursorLocation == null || Views.isAncestorOrEqualInSet(
          oldSelection.getSelectedPaths(), oldCursorLocation.getPath()))
      {
        cursorLocation = oldSelection.getCursorLocation();
      }
    } else if (pickSelect) {
      pathSelection = Sets.newLinkedHashSet(oldSelection.getSelectedPaths());
      if (cursorLocation != null) {
        final Path cursorPath = cursorLocation.getPath();
        /* Usability tweak: allow clicking anywhere for the common operation of adding a path to
        the set of selected paths, but require aiming at the label to remove a path from the
        selection again. */
        if (pathSelection.contains(cursorPath) && !cursorLocation.isValue()) {
          /* I experimented with keeping the cursorLocation at the old location in this case, but it
          felt unnatural, and there would still be cases where the cursor would end up at a
          non-selected path after pick selection. */
          pathSelection.remove(cursorPath);
        } else {
          /* Usability tweak: if the cursor is already at a label when the user starts to
          pick-select, assume that the first path under the cursor should also be included in the
          selection. The alternative would be to always select a path when the cursor is placed in
          a label, but this is distracting, and might confuse users attempting the common operation
          of editing a label (having the whole path selected suggests a change that will impact all
          rows). */
          if (pathSelection.isEmpty() && oldCursorLocation != null && !oldCursorLocation.isValue())
            pathSelection.add(oldCursorLocation.getPath());
          pathSelection.add(cursorPath);
        }
      }
    } else if (rangeSelect) {
      final Path startPath = (rangeSelectStart == null) ? null :
          Location.maybePath(rangeSelectStart.getCursorLocation());
      final Path endPath = Location.maybePath(cursorLocation);
      final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
      final View view = (generatedCell == null) ? null : generatedCell.getResult().getView();
      if (view != null && startPath != null && endPath != null) {
        if (startPath.equals(endPath)) {
          pathSelection = rangeSelectStart.getSelectedPaths();
        } else {
          final Set<Path> pathsBetween =
              Views.visiblePaths(view, Views.getPathsBetweenInclusive(view, startPath, endPath));
          pathSelection = Sets.union(rangeSelectStart.getSelectedPaths(), pathsBetween);
        }
      } else {
        pathSelection = oldSelection.getSelectedPaths();
      }
    } else {
      pathSelection = ImmutableSet.of();
    }

    final Selection ret = new Selection(cursorLocation, pathSelection);
    if (!rangeSelect)
      rangeSelectStart = ret;
    return ret;
  }

  private static boolean isPickSelect(InputEvent evt) {
    /* On MacOS, the default accelerator seems to always be present in the modifiers when a popup
    menu is opened, even when the popup menu is opened with a right mouse click and no keyboard
    buttons held down. For simplicity, just disable pickSelect while opening a popup menu on all
    platforms. */
    return !isPopupTrigger(evt) && UIUtil.modifierIncludesDefaultAccelerator(evt.getModifiers());
  }

  private static boolean isRangeSelect(InputEvent evt) {
    /* Don't enable range select either when right-clicking, as Shift+Right Click may be used as a
    shortcut for other purposes (e.g. to open the toolbox). */
    return !isPopupTrigger(evt) && (evt.getModifiersEx() & KeyEvent.SHIFT_DOWN_MASK) != 0;
  }

  private static boolean isPopupTrigger(InputEvent evt) {
    return (evt instanceof MouseEvent) && ((MouseEvent) evt).isPopupTrigger();
  }

  /** @return true iff a valid selection was made at this mouse location */
  private boolean mouseSelect(MouseEvent evt, boolean drag) {
    CursorMovement cursorMovement = getMouseCursorMovement(evt);
    if (cursorMovement == null)
      return false;
    final boolean pickSelect = isPickSelect(evt);
    if (!(drag && pickSelect) &&
        /* Usability tweak: don't destroy the path selection if the user is clicking to close a
        popup, except when the user is opening a popup somewhere else. */
        (!inPopup || evt.isPopupTrigger()))
    {
      final boolean cursorMoved = !Location.equals(cursorMovement.getCursorLocation(),
          getEffectiveSelectionModel().getCursorLocation());
      if (cursorMoved &&
          !inReferenceSelectionMode() && acceptEdit() != Acceptance.ACCEPTED_IMMEDIATELY)
      {
        return false;
      }
      // Handle the case specified in the Javadoc for CellWidget.setReferenceSelectionModel.
      if (!cursorMoved && inReferenceSelectionMode())
        getEffectiveSelectionModel().setSelection(Selection.NONE, false);
      /* On Windows, a MOUSE_PRESSED with isPopupTrigger == false will precede the popup trigger;
      this event also needs to be counted as popupOpening for processSelection's purposes. */
      final boolean popupOpening = evt.isPopupTrigger() ||
          Utilities.isWindows() && ((evt.getModifiersEx() & MouseEvent.BUTTON3_DOWN_MASK) != 0);
      processSelection(cursorMovement, popupOpening, pickSelect, drag || isRangeSelect(evt));
      return true;
    }
    return false;
  }

  /** Repaint the cell cursor and any selected path highlights. */
  private void repaintSelection() {
    final GeneratedCell generatedCell = getDisplayableGeneratedCell();
    if (generatedCell == null)
      return;
    List<PixelRectangle> dirtyRegions = Lists.newArrayList();
    dirtyRegions.addAll(generatedCell.getCell().findDirtyRegions(
        generatedCell.getRenderContext(), getClipBounds(),
        getEffectiveSelectionModel().getSelection()));
    repaintDirty(dirtyRegions);
  }

  private void processSelection(
      CursorMovement movement, boolean popupOpening, boolean pickSelect, boolean rangeSelect)
  {
    final Selection selection;
    if (inReferenceSelectionMode()) {
      // Don't use multiple selection in reference selection mode.
      selection = new Selection(movement.getCursorLocation(), ImmutableSet.of());
    } else {
      selection = calculateSelection(
          movement.getCursorLocation(), popupOpening, pickSelect, rangeSelect);
    }
    final Coordinates viewportPosition = movement.getViewportPosition();
    getEffectiveSelectionModel().setSelection(selection, movement.getScrollToCursor());
    if (viewportPosition != null)
      setViewportPosition(viewportPosition);
    final Coordinates newMagicCursorPoint = movement.getMagicCursorPoint();
    if (newMagicCursorPoint != null)
      setMagicCursorPoint(newMagicCursorPoint);
  }

  @Override
  public void moveCursorPhysical(
      Direction direction, CursorDistance distance, boolean rangeSelect)
  {
    if (!inReferenceSelectionMode() && acceptEdit() != Acceptance.ACCEPTED_IMMEDIATELY)
      return;
    CursorMovement movement = getPhysicalCursorMovement(direction, distance);
    if (movement != null)
      processSelection(movement, false, false, rangeSelect);
  }

  @Override
  public void moveCursorLogical(Direction direction) {
    // Logical cursor movements should cause edits to be accepted even in reference selection mode.
    if (acceptEdit() != Acceptance.ACCEPTED_IMMEDIATELY)
      return;
    LOG.log(Level.INFO, "Move cursor logical {0}", direction);
  }

  private @Nullable CursorMovement getPhysicalCursorMovement(
      Direction direction, CursorDistance distance)
  {
    final GeneratedCell generatedCell = getDisplayableGeneratedCell();
    if (generatedCell == null)
      return null;
    clearDebugLabels();
    final SelectionModel selectionModel = getEffectiveSelectionModel();
    final RenderContext rcx = generatedCell.getRenderContext();
    final @Nullable PixelRectangle oldCursorRectangle = findCursorRectangle();
    if (oldCursorRectangle == null) {
      Location cursorLocation = getDefaultCursorLocation();
      return (cursorLocation == null) ? null : new CursorMovement(cursorLocation, null, null);
    }
    final Coordinates oldViewportPosition = getViewportPosition();
    final Coordinates oldMagicCursorPoint = magicCursorPoint;
    setDebugLabel("oldViewportPosition", oldViewportPosition);
    setDebugLabel("oldMagicCursorPoint", oldMagicCursorPoint);

    final @Nullable Location oldCursorLocation =
        getEffectiveSelectionModel().getSelection().getCursorLocation();

    // TODO[cc]: Factor out common code once I've implemented all the distances.
    if (distance == CursorDistance.ONE_CELL) {
      // TODO[pd]: Can't I just use the magic cursor point here?
      final Coordinates newCursorSearchPoint =
          direction.newCoordinates(oldCursorRectangle.getPosition(), oldMagicCursorPoint);
      @Nullable Location newCursorLocation =
          generatedCell.getCell().getClosestCursorLocationAt(
              rcx, newCursorSearchPoint, direction, selectionModel.getCursorLocation(), null);
      final Coordinates newMagicCursorPoint;
      if (newCursorLocation == null) {
        newMagicCursorPoint = direction.newCoordinates(
            direction.getBiasPoint(rcx, oldCursorRectangle), oldMagicCursorPoint);
        newCursorLocation = oldCursorLocation;
      } else {
        PixelRectangle newCursorRectangle = rectangleFinder.findRectangles(
            newCursorLocation, newCursorSearchPoint, null).getCursorRectangle();
        // TODO[pd]: What to do in this case?
        if (newCursorRectangle == null)
          throw new AssertionError();
        newMagicCursorPoint =
            direction.newCoordinates(newCursorRectangle.getPosition(), oldMagicCursorPoint);
      }
      return new CursorMovement(newCursorLocation, null, newMagicCursorPoint);
    } else if (distance == CursorDistance.ONE_PAGE) {
      final Coordinates slack = new Coordinates(
          getViewportDimensions().getX().times(rcx, PAGE_SCROLL_ALIGNMENT_SLACK),
          getViewportDimensions().getY().times(rcx, PAGE_SCROLL_ALIGNMENT_SLACK));

      // If old cursor was not in the viewport, pretend that it was.
      final Coordinates adjustedOldMagicCursorPoint, adjustedOldCursorPosition;
      {
        boolean oldMagicCursorPointInViewport = oldMagicCursorPoint
            .max(getViewportPosition())
            .min(getViewportPosition().plus(getViewportDimensions()))
            .minus(oldMagicCursorPoint)
            .equals(Coordinates.ZERO);
        adjustedOldMagicCursorPoint = oldMagicCursorPointInViewport
            ? oldMagicCursorPoint
            : direction.newCoordinates(oldViewportPosition, oldMagicCursorPoint);
        adjustedOldCursorPosition = oldMagicCursorPointInViewport
            ? oldCursorRectangle.getPosition()
            : oldViewportPosition;
      }
      setDebugLabel("adjustedOldMagicCursorPoint", adjustedOldMagicCursorPoint);

      // The starting point is to scroll exactly one page minus the slack distance.
      final Coordinates newViewportPositionBeforeAlignment = constrainViewportPosition(
          oldViewportPosition.plus(direction.multiply(getViewportDimensions().minus(slack))));
      // The magic cursor point should stand still on the screen (prior to alignment adjustment).
      final Coordinates newMagicCursorPoint = adjustedOldMagicCursorPoint.plus(
          newViewportPositionBeforeAlignment.minus(oldViewportPosition));

      @Nullable Location newCursorLocation = generatedCell.getCell()
          .getClosestCursorLocationAt(rcx, newMagicCursorPoint, direction, null, null);

      final Coordinates newViewportPosition;
      if (newCursorLocation == null) {
        newViewportPosition = newViewportPositionBeforeAlignment;
        newCursorLocation = oldCursorLocation;
      } else {
        final PixelRectangle newCursorRectangle = rectangleFinder.findRectangles(
            newCursorLocation, newMagicCursorPoint, null).getCursorRectangle();
        if (newCursorRectangle == null) {
          // Unexpected case.
          newViewportPosition = newViewportPositionBeforeAlignment;
        } else {
          final Coordinates limitedViewportAlignmentOffset;
          /* Adjust the viewport position to align the top left-hand corner of the old and the new
          selection if the required adjustment is less than the permitted slack. Don't do it unless
          the selection actually changed; otherwise the alignment adjustment will simply undo the
          scrolling if the distance was small enough (e.g. when using Page Up to scroll the last
          little bit to the top even though the topmost cell is already selected). */
          if (newCursorRectangle.equals(oldCursorRectangle)) {
            limitedViewportAlignmentOffset = Coordinates.ZERO;
          } else {
            final Coordinates alignOff = constrainViewportPosition(oldViewportPosition.plus(
                newCursorRectangle.getPosition().minus(adjustedOldCursorPosition)))
                .minus(newViewportPositionBeforeAlignment);
            limitedViewportAlignmentOffset = new Coordinates(
                alignOff.getX().abs().isGreaterThan(slack.getX()) ? Pixels.ZERO : alignOff.getX(),
                alignOff.getY().abs().isGreaterThan(slack.getY()) ? Pixels.ZERO : alignOff.getY());
          }
          newViewportPosition =
              newViewportPositionBeforeAlignment.plus(limitedViewportAlignmentOffset);
        }
      }
      return new CursorMovement(newCursorLocation, newViewportPosition, newMagicCursorPoint);
    } else if (distance == CursorDistance.TO_END) {
      final Coordinates biasPoint = direction.getBiasPoint(rcx,
          new PixelRectangle(Coordinates.ZERO, generatedCell.getActualDimensions()));
      final Coordinates newMagicCursorPoint =
          direction.newCoordinates(biasPoint, oldMagicCursorPoint);
      final Location newCursorLocation = generatedCell.getCell()
          .getClosestCursorLocationAt(generatedCell.getRenderContext(), newMagicCursorPoint,
          direction, null, null);
      final Coordinates newViewportPosition =
          direction.newCoordinates(biasPoint, getViewportPosition());
      return new CursorMovement(newCursorLocation, newViewportPosition, newMagicCursorPoint);
    } else {
      throw new AssertionError("Unexpected CursorDistance " + distance);
    }
  }

  private void migrateMagicCursorPoint() {
    final PixelRectangle cursorRect = findCursorRectangle();
    if (cursorRect == null)
      return;
    /* Preserve the existing magic cursor point in directions for which the pixel rectangle's
    position remains constant. Keep the new magic cursor point strictly within the bounds of the new
    pixel rectangle. */
    Pixels newMagicCursorPointX = cursorRect.getX();
    Pixels newMagicCursorPointY = cursorRect.getY();
    if (magicCursorPointRectangle != null) {
      if (magicCursorPointRectangle.getX().isEqualTo(cursorRect.getX())) {
        newMagicCursorPointX = newMagicCursorPointX
            .max(magicCursorPoint.getX())
            .min(cursorRect.getX().plus(cursorRect.getWidth()).minus(Pixels.EPSILON));
      }
      if (magicCursorPointRectangle.getY().isEqualTo(cursorRect.getY())) {
        newMagicCursorPointY = newMagicCursorPointY
            .max(magicCursorPoint.getY())
            .min(cursorRect.getY().plus(cursorRect.getHeight()).minus(Pixels.EPSILON));
      }
    }
    setMagicCursorPoint(new Coordinates(newMagicCursorPointX, newMagicCursorPointY));
  }

  private void setMagicCursorPoint(final Coordinates newMagicCursorPoint) {
    Preconditions.checkNotNull(newMagicCursorPoint);
    this.magicCursorPointRectangle = findCursorRectangle();
    this.magicCursorPoint = newMagicCursorPoint;
    if (!inReferenceSelectionMode())
      this.nonReferenceMagicCursorPoint = newMagicCursorPoint;
  }

  private @Nullable Location getDefaultCursorLocation() {
    return getDefaultCursorLocation(magicCursorPoint);
  }

  private @Nullable Location getDefaultCursorLocation(Coordinates cursorPoint) {
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    if (generatedCell == null)
      return null;
    Location cursorLocation = generatedCell.getCell().getClosestCursorLocationAt(
        generatedCell.getRenderContext(), cursorPoint, Direction.DOWN_RIGHT, null, null);
    if (cursorLocation == null && !cursorPoint.isEqualTo(Coordinates.ZERO)) {
      cursorLocation = generatedCell.getCell().getClosestCursorLocationAt(
          generatedCell.getRenderContext(), Coordinates.ZERO, Direction.DOWN_RIGHT, null, null);
    }
    return cursorLocation;
  }

  /* === Uncategorized Methods ================================================================== */

  private boolean inReferenceSelectionMode() {
    return cellWidget.getReferenceSelectionModel() != null;
  }

  private SelectionModel getEffectiveSelectionModel() {
    final SelectionModel referenceSelectionModel = cellWidget.getReferenceSelectionModel();
    return referenceSelectionModel != null ? referenceSelectionModel
        : cellWidget.getSelectionModel();
  }

  private void effectiveSelectionModelChanged(
      SelectionModel oldSelectionModel, SelectionModel newSelectionModel)
  {
    oldSelectionModel.removePropertyChangeListener(selectionModelListener);
    newSelectionModel.addPropertyChangeListener(selectionModelListener);
    selectionModelListener.propertyChange(new SelectionModelPropertyChangeEvent(cellWidget,
        oldSelectionModel.getSelection(), newSelectionModel.getSelection(), false));
  }

  private void clearDebugLabels() {
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    if (generatedCell != null) {
      generatedCell.getDebugLabels().clear();
      if (cellWidget.getDrawCellDebugGraphics())
        repaintAll();
    }
  }

  private void setDebugLabel(String label, @Nullable Coordinates coordinates) {
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    if (generatedCell != null) {
      if (coordinates == null) {
        generatedCell.getDebugLabels().remove(label);
      } else {
        generatedCell.getDebugLabels().put(label, coordinates);
      }
      if (cellWidget.getDrawCellDebugGraphics())
        repaintAll();
    }
  }

  private void repaintAll() {
    cellWidget.repaintScrollable(
        new Rectangle(0, 0, cellWidget.getWidth(), cellWidget.getHeight()));
    cellWidget.repaintOverlay();
  }

  private void repaintInstalledComponentBounds(Rectangle bounds) {
    cellWidget.repaintScrollable(UIUtil.getShadowBorderBounds(bounds));
    cellWidget.repaintOverlay();
  }

  /**
   * Repaint dirty regions as well as the overlay.
   */
  private void repaintDirty(List<PixelRectangle> dirtyRegionsInScrollable) {
    final Set<Rectangle> clippedDirtyRegions = Sets.newLinkedHashSet();
    // Not sure if this is needed anymore.
    for (PixelRectangle rectangle : dirtyRegionsInScrollable) {
      Rectangle clippedRectangle = rectangle.getBoundingAWTRectangle().intersection(
          getClipBounds().getBoundingAWTRectangle());
      if (!clippedRectangle.isEmpty())
        clippedDirtyRegions.add(clippedRectangle);
    }
    for (Rectangle rectangle : clippedDirtyRegions)
      cellWidget.repaintScrollable(rectangle);
    /* The overlay may need a repaint even when no dirty regions lie within the viewport, since
    overlays may move other parts of the layout (e.g. frozen headers) into view. */
    cellWidget.repaintOverlay();
  }

  private PixelRectangle getClipBounds() {
    return new PixelRectangle(cellWidget.getVisibleRect());
  }

  private void scrollToCursor() {
    final CellRectangles cellRectangles = findRectangles(
        getEffectiveSelectionModel().getCursorLocation(), null, getViewportPosition());
    final PixelRectangle cursorRect = cellRectangles.getCursorRectangle();
    if (cursorRect != null && !cellRectangles.isCursorInOverlayHeader()) {
      Coordinates overlayHeaderDimensions = Coordinates.ZERO;
      for (HeaderRectangle header : cellRectangles.getHeaderRectangles()) {
        if (!header.overlay)
          continue;
        PixelRectangle headerRect = header.rectangle;
        overlayHeaderDimensions = overlayHeaderDimensions.plus(
            header.direction.newCoordinates(headerRect.getDimensions(), Coordinates.ZERO));
      }
      final Rectangle cr = cursorRect.getBoundingAWTRectangle();
      final Point vd = getViewportDimensions().getPoint();
      setViewportPosition(getScrolledViewportPosition(new PixelRectangle(new Rectangle(
          cr.x, cr.y,
          Math.max(1, (int) Math.min(cr.width, vd.x * CURSOR_SCROLL_VIEWPORT_FRACTION_LIMIT)),
          Math.max(1, (int) Math.min(cr.height, vd.y * CURSOR_SCROLL_VIEWPORT_FRACTION_LIMIT))
      )), overlayHeaderDimensions));
    }
  }

  private @Nullable JScrollPane getEnclosingScrollPane() {
    if (viewportContainer instanceof JViewport) {
      Container ret = viewportContainer.getParent();
      if (ret instanceof JScrollPane)
        return (JScrollPane) ret;
    }
    return null;
  }

  private void updateViewportContainer() {
    // TODO[pd]: Use SwingUtilities.getUnwrappedParent() instead.
    viewportContainer = cellWidget.getParent();
    if (!(viewportContainer instanceof JViewport))
      viewportContainer = cellWidget;
  }

  private void setViewportPosition(Coordinates newPosition) {
    JScrollPane scrollPane = getEnclosingScrollPane();
    if (scrollPane != null) {
      final Coordinates constrained = constrainViewportPosition(newPosition);
      scrollPane.getViewport().setViewPosition(constrained.getPoint());
    }
  }

  private Coordinates constrainViewportPosition(Coordinates position) {
    JScrollPane scrollPane = getEnclosingScrollPane();
    if (scrollPane != null) {
      return position
          .min(new Coordinates(scrollPane.getViewport().getViewSize())
            .minus(new Coordinates(scrollPane.getViewport().getExtentSize())))
          .max(Coordinates.ZERO);
    } else {
      return Coordinates.ZERO;
    }
  }

  private Pixels getScrolledViewportPosition(
      Pixels originalViewportPosition, Pixels viewportLength,
      Pixels targetIntervalPosition, Pixels targetIntervalLength)
  {
    final Pixels targetIntervalEnd = targetIntervalPosition.plus(targetIntervalLength);
    /* Don't scroll if the interval to be made visible is longer than the viewport and the viewport
    is already fully contained in the interval. */
    {
      final Pixels viewportEnd = originalViewportPosition.plus(viewportLength);
      if (targetIntervalLength.isGreaterThan(viewportLength) &&
          !targetIntervalPosition.isGreaterThan(originalViewportPosition) &&
          !viewportEnd.isGreaterThan(targetIntervalEnd))
      {
        return originalViewportPosition;
      }
    }
    Pixels viewportPosition = originalViewportPosition;
    // Scroll forward if necessary.
    {
      final Pixels viewportEnd = viewportPosition.plus(viewportLength);
      viewportPosition = viewportPosition.plus(
          targetIntervalEnd.minus(viewportEnd).max(Pixels.ZERO));
    }
    // Scroll backwards if necessary.
    viewportPosition = viewportPosition.minus(
        viewportPosition.minus(targetIntervalPosition).max(Pixels.ZERO));
    return viewportPosition;
  }

  // TODO[cc]: Move this logic out into a utility class somewhere.
  private Coordinates getScrolledViewportPosition(
      PixelRectangle toVisible, Coordinates overlayHeaderDimensions)
  {
    final Coordinates vp = viewportRectangle.getPosition().plus(overlayHeaderDimensions);
    final Coordinates vd = viewportRectangle.getDimensions().minus(overlayHeaderDimensions);
    Coordinates ret = new Coordinates(
        getScrolledViewportPosition(vp.getX(), vd.getX(), toVisible.getX(), toVisible.getWidth()),
        getScrolledViewportPosition(vp.getY(), vd.getY(), toVisible.getY(), toVisible.getHeight()));
    return ret.minus(overlayHeaderDimensions);
  }

  private void setViewportRectangle(@Nullable PixelRectangle viewportRectangle) {
    this.viewportRectangle = viewportRectangle;
    /* Avoid propagating bogus zero dimensions that may exist before the component has been laid
    out. Any existing generated cell may become relevant again as soon as real dimensions are made
    available (observed when invoking "Maximize Window" on a TopComponent containing a CellWidget in
    NetBeans). */
    if (viewportRectangle == null || !getViewportDimensions().isEqualTo(Coordinates.ZERO))
      cellWidget.setViewportDimensions(viewportRectangle == null ? null : getViewportDimensions());
  }

  private Coordinates getViewportPosition() {
    return viewportRectangle.getPosition();
  }

  private Coordinates getViewportDimensions() {
    return viewportRectangle.getDimensions();
  }

  private void updateViewportRectangle() {
    final Coordinates viewportPosition;
    JScrollPane scrollPane = getEnclosingScrollPane();
    if (scrollPane != null) {
      /* Get the position from the scrollbars rather than from the JViewport; see the comment in
      installListeners(). */
      viewportPosition = new Coordinates(
          Pixels.create(scrollPane.getHorizontalScrollBar().getValue()),
          Pixels.create(scrollPane.getVerticalScrollBar().getValue()));
    } else {
      viewportPosition = Coordinates.ZERO;
    }
    /* A negative viewportContainer size was once observed when a CellWidget was initialized during
    NetBeans Platform startup (as part of opening a persisted MultiViewElement). */
    setViewportRectangle(new PixelRectangle(viewportPosition,
        new Coordinates(viewportContainer.getSize()).max(Coordinates.ZERO)));
    updateScrollables();
  }

  private void updateRenderContext(Graphics2D g) {
    ThreadUtil.checkOnEDT();
    cellWidget.setRenderContext(createRenderContext(g));
  }

  private void updateRenderContext() {
    Graphics g0 = cellWidget.getGraphics();
    // getGraphics() may return null when the component has not yet been made displayable.
    if (g0 != null) {
      Graphics2D g = createGraphicsWithRenderingHintsConfigured(g0);
      try {
        updateRenderContext(g);
      } finally {
        g.dispose();
      }
    }
  }

  private RenderContext createRenderContext(Graphics2D g) {
    Pixels roundToPixels;
    if (cellWidget.getForceFractionalMetrics()) {
      roundToPixels = null;
    } else {
      roundToPixels = UIUtil.useHalfPixels(g) ? Pixels.HALF : Pixels.ONE;
    }
    double contextScaling = cellWidget.getScaling() * (UIUtil.getPixelsPerInch() / 72.0);
    return new RenderContext(g.getFontRenderContext(), contextScaling,
        roundToPixels, cellWidget.isAlphaPermitted(), true);
  }

  private @Nullable GeneratedCell getDisplayableGeneratedCell() {
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    final RenderContext specifiedContext = cellWidget.getRenderContext();
    boolean displayable = generatedCell != null && specifiedContext != null &&
        specifiedContext.getFontRenderContext().equals(
        generatedCell.getRenderContext().getFontRenderContext());
    return displayable ? generatedCell : null;
  }

  private void setFocusMode(FocusMode focusMode) {
    if (this.focusMode != focusMode) {
      this.focusMode = focusMode;
      repaintSelection();
    }
  }

  private boolean cellWidgetOrDescendantHasFocus() {
    Component focusOwner = KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
    return (focusOwner == null) ? false : SwingUtilities.isDescendingFrom(focusOwner, cellWidget);
  }

  private void updateFocusMode() {
    final FocusMode newValue;
    if (inReferenceSelectionMode()) {
      newValue = FocusMode.REFERENCE_SELECTION;
    } else {
      /* Don't gray out the cell cursor just because the popup menu or inplace editor is technically
      the component to have the focus. */
      newValue =
          (cellWidgetOrDescendantHasFocus() || inPopup || inAcceptEdit)
          ? FocusMode.FOCUSED : FocusMode.UNFOCUSED;
    }
    setFocusMode(newValue);
  }

  private void setInPopup(boolean inPopup) {
    if (this.inPopup == inPopup)
      return;
    this.inPopup = inPopup;
    updateFocusMode();
    updateCursorTooltip();
  }

  private void setCurrentlyShowingPopup(final @Nullable PopupLike newValue) {
    PopupLike oldValue = this.currentlyShowingPopup;
    if (oldValue == newValue)
      return;
    if (oldValue != null)
      oldValue.removePopupMenuListener(popupMenuListener);
    this.currentlyShowingPopup = newValue;
    if (newValue != null)
      newValue.addPopupMenuListener(popupMenuListener);
    updateSelectionForPopup(newValue != null);
    if (newValue == null) {
      /* Do this at the end of the event queue; this avoids a flicker in the focus indication while
      the menu is closing, and gives mouseSelect a chance to check inPopup to determine appropriate
      selection behavior. */
      SwingUtilities.invokeLater(() -> {
        setInPopup(currentlyShowingPopup != null);
      });
    } else {
      setInPopup(true);
    }
  }

  /**
   * Adds the specified component to the component hierarchy. The component will initially be made
   * invisible, and may be shown using
   * {@link #setInstalledComponentBounds(Component,Rectangle,Rectangle)}.
   */
  private void installComponent(JComponent component) {
    Preconditions.checkState(component.getParent() == null);
    component.setVisible(false);
    cellWidget.add(component);
  }

  private void uninstallComponent(JComponent component) {
    Preconditions.checkState(component.getParent() == cellWidget);
    Rectangle oldBounds = component.getBounds();
    cellWidget.remove(component);
    repaintInstalledComponentBounds(oldBounds);
  }

  /**
   * Show, hide, move, or resize an installed component.
   *
   * @param bounds may be null to hide the component temporarily
   */
  private void setInstalledComponentBounds(
      Component component, @Nullable Rectangle oldBounds, @Nullable Rectangle bounds)
  {
    Preconditions.checkState(component.getParent() == cellWidget);
    if (bounds == null) {
      if (!component.isVisible())
        return; // No need to repaint again.
      component.setVisible(false);
    } else {
      component.setBounds(bounds);
      /* Not sure if I need the explicit validate()/repaint() calls here, but this is what
      JTable.editCellAt does. The explicit repaint() call seems unnecessary on MacOS (Nimbus and
      Aqua L&F), but it might be due to setVisible(true) triggering one. */
      component.validate();
      component.setVisible(true);
      component.repaint();
      repaintInstalledComponentBounds(bounds);
    }
    if (oldBounds != null)
      repaintInstalledComponentBounds(oldBounds);
  }

  private void setInstalledCursorTooltip(JComponent newValue) {
    final JComponent oldValue = this.installedCursorTooltip;
    if (oldValue == newValue)
      return;
    if (oldValue != null)
      uninstallComponent(oldValue);
    this.installedCursorTooltip = newValue;
    if (newValue != null)
      installComponent(newValue);
  }

  private void setInstalledInplaceEditor(@Nullable InplaceEditor newValue) {
    final InplaceEditor oldValue = this.installedInplaceEditor;
    if (oldValue == newValue)
      return;
    if (oldValue != null) {
      oldValue.setActive(false);
      oldValue.getFocusComponent().removeFocusListener(focusListener);
      oldValue.getComponent().removePropertyChangeListener(
          "preferredSize", inplaceEditorSizeListener);
      uninstallComponent(oldValue.getComponent());
    }
    this.installedInplaceEditor = newValue;
    if (newValue != null) {
      newValue.getFocusComponent().addFocusListener(focusListener);
      newValue.getComponent().addPropertyChangeListener(
          "preferredSize", inplaceEditorSizeListener);
      installComponent(newValue.getComponent());
      newValue.setActive(true);
    }
    /* Delay this update to the end of the event queue to prevent the tooltip from appearing for a
    brief moment on the edited cell before the cursor moves away (where applicable). */
    SwingUtilities.invokeLater(this::updateCursorTooltip);
  }

  private void updateCursorTooltip() {
    JComponent cursorTooltip =
        inReferenceSelectionMode() ? null : cellWidget.getInteractionProvider().getCursorTooltip();
    final Rectangle oldBounds =
        (installedCursorTooltip == null) ? null : installedCursorTooltip.getBounds();
    setInstalledCursorTooltip(cursorTooltip);
    if (cursorTooltip != null) {
      final Dimension tooltipSize;
      {
        /* Discard the previous bounds of the tooltip to ensure the behavior of the logic below is
        always repeatable. Not doing this made a past bug hard to track down. */
        cursorTooltip.invalidate();
        cursorTooltip.setSize(0, 0);
        /* See BorderLayout.layoutContainer for an example of how to use setSize, getPreferredSize,
        and setBounds to implement a layout policy. */
        /* TODO[pd]: Consider fixing a bug where the getPreferredSize() does not give the right
                     value when the L&F has just been changed. Not a practical problem in itself,
                     though. */
        /* The tooltip needs to have been both installed and validated before getPreferredSize()
        will give the correct result. */
        cursorTooltip.validate();
        /* Avoid super-long cursor tooltips. Note that Swing HTML labels only seem to break text
        after a certain minimum width; 100 pixels was too small in the limit below, but 300 is
        plenty. */
        final int width = Math.min(cursorTooltip.getPreferredSize().width,
            Math.max(300, viewportRectangle.getBoundingAWTRectangle().width / 2));
        /* Setting the height to zero here causes the wrong preferred height to be calculated
        (enough only for a single line of text). Setting it to an essentially infinite value works;
        though don't actually use Integer.MAX_VALUE, since the latter would be likely to trigger
        overflow bugs. */
        cursorTooltip.setSize(width, 10000);
        UIUtil.forceSwingHtmlSizeComputation(cursorTooltip);
        Dimension d = cursorTooltip.getPreferredSize();
        tooltipSize = new Dimension(width, d.height);
      }
      final Coordinates pos = (inPopup || installedInplaceEditor != null)
          ? null : getBestTooltipPosition(new Coordinates(tooltipSize));
      setInstalledComponentBounds(cursorTooltip, oldBounds,
          (pos == null) ? null : new Rectangle(pos.getPoint(), tooltipSize));
    }
  }

  /**
   * Accept any ongoing edit, whether inplace or external, and request focus back to the
   * {@code CellWidget} if successful. OK to call even if no edit is in progress.
   */
  private Acceptance acceptEdit() {
    Preconditions.checkState(!inAcceptEdit);
    final Acceptance ret;
    try {
      inAcceptEdit = true;
      ret = cellWidget.getInteractionProvider().acceptEdit();
      if (ret != Acceptance.REJECTED) {
        setInstalledInplaceEditor(null);
        // Unconditionally request focus here.
        cellWidget.requestFocusInWindow();
      }
    } finally {
      /* Don't reset this until setInstalledInplaceEditor(null) has been called, to avoid potential
      focus indication flicker, or in case the latter needs to see it in order to request focus back
      to the CellWidget (no longer the case, since acceptEdit() is now responsible for restoring the
      focus, but the logic might change again in the future). */
      inAcceptEdit = false;
    }
    return ret;
  }

  @Override
  public void closeInplaceEditor() {
    // See JTable.removeEditor().
    if (installedInplaceEditor == null)
      return;
    final Component focusOwner =
        KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
    final boolean hadFocus =
        focusOwner != null && SwingUtilities.isDescendingFrom(focusOwner, cellWidget);
    setInstalledInplaceEditor(null);
    if (hadFocus)
      cellWidget.requestFocusInWindow();
  }

  @Override
  public JComponent getFocusComponent() {
    return installedInplaceEditor != null ? installedInplaceEditor.getFocusComponent() : cellWidget;
  }

  private void updateInplaceEditorBounds() {
    if (installedInplaceEditor == null)
      return;
    final Rectangle oldBounds = installedInplaceEditor.getComponent().isVisible()
        ? installedInplaceEditor.getComponent().getBounds()
        : null;
    // Use the regular selection model here, not the reference selection model.
    final PixelRectangle cursorRectangle =
        findCursorRectangle(cellWidget.getSelectionModel().getCursorLocation());
    if (cursorRectangle == null) {
      /* Close the inplace editor. The edit will likely be lost once a new default selection is
      made. Don't attempt to accept the edit; it is better to lose a few seconds of the user's most
      recent typing work than to save incomplete changes without the user's knowledge. Also beware
      that the call to updateInplaceEditorBounds may itself be a result of an acceptEdit operation.
      In the latter case it would not be permissible to call acceptEdit, though it would have been
      superfluous in any case with the current CursorEditManager implementation. */
      closeInplaceEditor();
    } else {
      final Coordinates maxSize =
          // Allow the inplace editor to expand to the edge of the current viewport rectangle.
          viewportRectangle.getPosition()
          .plus(viewportRectangle.getDimensions())
          .minus(cursorRectangle.getPosition())
          /* The editor should always cover the entire cell, even if this would put the editor
          partially outside the viewport, since the user might scroll the CellWidget after inplace
          editing has started. */
          .max(cursorRectangle.getDimensions())
          // Make odd corner cases less likely to cause bad errors.
          .max(new Coordinates(Pixels.ONE, Pixels.ONE));
      JComponent comp = installedInplaceEditor.getComponent();
      /* Invalidate the component to recompute its preferred size, but do not make it invisible, as
      the latter can cause it to lose focus. */
      comp.invalidate();
      comp.setSize(0, 0);
      comp.validate();
      /* TODO[cc]: Consider getting the preferred size through a separate API. The fact that we're
                   using the preferred size based on an initial size different from the current size
                   suggests that preferredSize might not be the right property to listen on. */
      final Dimension preferredSize = installedInplaceEditor.getComponent().getPreferredSize();
      final Rectangle bounds;
      if (DEBUG_USE_INPLACE_EDITOR_PREFERRED_SIZE) {
        bounds = new Rectangle(cursorRectangle.getPosition().getPoint(), preferredSize);
      } else {
        Coordinates useSize = new Coordinates(preferredSize)
            .max(cursorRectangle.getDimensions())
            .min(maxSize);
        bounds = new Rectangle(cursorRectangle.getPosition().getPoint(), useSize.getDimension());
      }
      setInstalledComponentBounds(comp, oldBounds, bounds);
    }
  }

  /**
   * @param evt if not null, only start an inplace edit if
   *        {@link InplaceEditor#shouldStartEditing(KeyEvent)} indicates it.
   */
  private void startInplaceEdit(@Nullable KeyEvent evt) {
    if (installedInplaceEditor == null) {
      final InplaceEditor editor = cellWidget.getInteractionProvider().getInplaceEditor();
      if (editor == null)
        return;
      /* In JTable, processKeyBinding performs its own vetting of the incoming KeyEvent, ignoring
      any events other than those of type KEY_PRESSED, and ignoring presses of modifiers by
      themselves. In the course of fixing various problems with JTable's approach, such as dealing
      with interactions with keyboard shortcuts, I found it necessary to delegate action vetting
      logic to individual InplaceEditor implementations, and to allow KEY_TYPED events to be used
      instead of KEY_PRESSED events (in order to be guaranteed a meaningful return value from
      KeyEvent.getKeyChar()). Thus the JTable's action vetting logic becomes irrelevant here. */
      if ((!ALLOW_READONLY_INPLACE_EDITOR || evt != null) && !editor.shouldStartEditing(evt))
        return;
      if (findCursorRectangle() == null)
        return;
      setInstalledInplaceEditor(editor);
      CursorEditInputMap.configureCursorEditor(cellWidget, editor.getFocusComponent(), evt == null);
      editor.prepare(evt != null);
      updateInplaceEditorBounds();
    } else {
      if (evt == null) {
        /* Allow the inplace editor to be switched into extended input mode. I considered having F2
        toggle between extended and non-extended input mode like in Excel, but decided against it
        since there is no visible indication of the state, since few people are likely to be aware
        of the feature, and since users can get out of extended input mode during formula editing by
        pressing arrow up/down (see the use case mentioned in
        http://www.wizardofexcel.com/2011/06/21/f2-the-most-useful-key-in-excel ). */
        CursorEditInputMap.configureCursorEditor(
            cellWidget, installedInplaceEditor.getFocusComponent(), true);
      }
    }
    scrollToCursor();
    /* Do what JTable does when the surrendersFocusOnKeystroke property is enabled. If we don't let
    the CellWidget surrender focus, the inplace editor will fail to show a text cursor, which is
    unacceptable. The latter is the case in JTable as well (with surrendersFocusOnKeystroke in its
    default false setting), on both Windows and MacOS. Said behavior is also inconsistent, because a
    text cursor _will_ show up in the JTable even when surrendersFocusOnKeystroke is false if
    editing is initiated by double-clicking the cell rather than by typing. ETable, which extends
    from JTable, forces surrendersFocusOnKeystroke on. */
    // TODO[pd]: Consider if this should be a temporary focus request.
    installedInplaceEditor.getFocusComponent().requestFocusInWindow();
  }

  @Override
  public void startInplaceEdit() {
    startInplaceEdit(null);
  }

  @Override
  public boolean forwardUnhandledKeyStroke(KeyStroke ks, KeyEvent evt, boolean pressed) {
    // Analogous to JTable.processKeyBinding.
    startInplaceEdit(evt);
    return (installedInplaceEditor == null) ? false :
        installedInplaceEditor.forwardUnhandledKeyStroke(ks, evt, pressed);
  }

  private @Nullable Coordinates getBestTooltipPosition(Coordinates tooltipSize) {
    GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    if (generatedCell == null)
      return null;
    // TODO[pd]: Avoid misalignment due to border widths.
    PixelRectangle cursorRect = findCursorRectangle();
    if (cursorRect == null)
      return null;
    RenderContext rcx = generatedCell.getRenderContext();
    Pixels marginPixels = Pixels.create(rcx, TOOLTIP_CURSOR_MARGIN);
    Coordinates margin = new Coordinates(marginPixels, marginPixels);
    PixelRectangle marginRect = new PixelRectangle(
        cursorRect.getPosition().minus(margin),
        cursorRect.getDimensions().plus(margin).plus(margin));
    Coordinates layoutDimensions = generatedCell.getActualDimensions();
    Coordinates alignPos = cursorRect.getPosition()
        .min(layoutDimensions.minus(tooltipSize))
        .max(Coordinates.ZERO);
    List<Coordinates> cands = Lists.newArrayList();
    /* Below the layout, if the layout is small (e.g. due to an error leading to no data being
    produced). */
    if (CURSOR_TOOLTIP_UNDER_LAYOUT_HEIGHT_THRESHOLD.isGreaterThan(layoutDimensions.getY()))
      cands.add(new Coordinates(alignPos.getX(), layoutDimensions.getY().plus(marginPixels)));
    // Right, left, below, and above the cursor.
    cands.add(new Coordinates(marginRect.getX().plus(marginRect.getWidth()), alignPos.getY()));
    cands.add(new Coordinates(marginRect.getX().minus(tooltipSize.getX()), alignPos.getY()));
    cands.add(new Coordinates(alignPos.getX(), marginRect.getY().plus(marginRect.getHeight())));
    cands.add(new Coordinates(alignPos.getX(), marginRect.getY().minus(tooltipSize.getY())));
    Coordinates best = null;
    long bestArea = -1;
    /* TODO[cc]: Avoid using the AWT Rectangle class here; implement intersection for PixelRectangle
                 instead. */
    Rectangle layoutRectangle = POSITION_TOOLTIPS_IN_VIEWPORT
        ? viewportRectangle.getBoundingAWTRectangle()
        : new PixelRectangle(Coordinates.ZERO, layoutDimensions).getBoundingAWTRectangle();
    for (Coordinates candidate : cands) {
      Rectangle intersection = layoutRectangle.intersection(
          new PixelRectangle(candidate, tooltipSize).getBoundingAWTRectangle());
      long area = intersection.width * (long) intersection.height;
      // Use strictly greater than here to prefer earlier entries in cands.
      if (area > bestArea) {
        bestArea = area;
        best = candidate;
      }
    }
    return best;
  }

  /* === Painting Logic ========================================================================= */

  private boolean isOverlayEnabled() {
    return (viewportContainer instanceof OverlayViewport);
  }

  @Override
  public void paint(Graphics g0, JComponent c) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkArgument(c == cellWidget);
    Graphics2D g = createGraphicsWithRenderingHintsConfigured(g0);
    try {
      /* When overlays are enabled, the cursor cell is painted as part of the overlay, to avoid
      being obscured by frozen value cells in the cursor overprinting area. */
      paintGeneratedCell(g, false, !isOverlayEnabled());
    } finally {
      g.dispose();
    }
  }

  @Override
  public void paintOverlay(Graphics g0) {
    Graphics2D g = createGraphicsWithRenderingHintsConfigured(g0);
    try {
      paintGeneratedCell(g, true, true);
    } finally {
      g.dispose();
    }
    // TODO[pd]: Handle the shadow drawing logic in the !isOverlayEnabled() case.
    paintChildComponents((Graphics2D) g0);
  }

  /**
   * @param g0 will be painted on but not otherwise modified
   */
  private void paintChildComponents(Graphics2D g0) {
    paintChildComponent(g0, installedCursorTooltip, true);
    paintChildComponent(g0,
        installedInplaceEditor == null ? null : installedInplaceEditor.getComponent(), false);
  }

  /**
   * @param g0 will be painted on but not otherwise modified
   */
  private void paintChildComponent(
      Graphics2D g0, @Nullable JComponent component, boolean strongBorder)
  {
    Preconditions.checkNotNull(g0);
    /* Check getParent() in case the paint is done in the middle of the steps taken to remove the
    component. */
    if (component == null || !component.isVisible() || component.getParent() != cellWidget)
      return;
    final Rectangle cr = component.getBounds();
    if (!UIUtil.getShadowBorderBounds(cr).intersects(g0.getClipBounds(new Rectangle())))
      return;
    UIUtil.drawShadowBorder(g0, component.getBounds(), strongBorder);
    /* To safely be able to call JComponent.paint(Graphics) ourselves, we need to temporarily turn
    off double buffering, otherwise weird painting bugs may result from buffer sharing at the
    RepaintManager level (notably when the component is a JPanel with another component installed in
    it). Similar logic is found in CellRenderPane.paintComponent. */
    final boolean oldDoubleBuffered = component.isDoubleBuffered();
    component.setDoubleBuffered(false);
    // See JComponent.paintChildren.
    Graphics g = g0.create(cr.x, cr.y, cr.width, cr.height);
    try {
      g.setColor(component.getForeground());
      g.setFont(component.getFont());
      component.paint(g);
    } finally {
      g.dispose();
      component.setDoubleBuffered(oldDoubleBuffered);
    }
  }

  private void paintGeneratedCell(Graphics2D g, boolean frozenHeaderMode, boolean cursorPainted) {
    // Update specifications in case the FontRenderContext changed.
    updateRenderContext(g);

    /* Don't paint the component background; this is done in the default implementation of
    ComponentUI.update(Graphics,JComponent). */

    final GeneratedCell generatedCell = getDisplayableGeneratedCell();
    if (generatedCell != null) {
      generatedCell.getCell().paint(new Canvas(g, generatedCell.getRenderContext()),
          Coordinates.ZERO, getEffectiveSelectionModel().getSelection(),
          cellWidget.getPathHighlights(), frozenHeaderMode ? getViewportPosition() : null,
          focusMode, cursorPainted);
    }

    if (!frozenHeaderMode) {
      if (cellWidget.getDrawCellDebugGraphics() && generatedCell != null) {
        g.setFont(new Font("Dialog", Font.PLAIN, 12));
        for (Entry<String,Coordinates> entry : generatedCell.getDebugLabels().entrySet()) {
          Coordinates point = entry.getValue();
          g.setColor(Color.WHITE);
          g.drawString(entry.getKey(),
              point.getX().getInteger() + 6, point.getY().getInteger() + 7);
          g.setColor(Color.BLUE);
          g.fill(new Rectangle2D.Double(
              point.getX().getDouble() - 2.0, point.getY().getDouble() - 2.0,
              5.0, 5.0));
          g.drawString(entry.getKey(),
              point.getX().getInteger() + 5, point.getY().getInteger() + 6);
        }
        PixelRectangle rect = findCursorRectangle();
        g.setColor(Color.RED);
        if (rect != null)
          g.draw(rect.getRectangle2D());
        g.fill(new Rectangle2D.Double(
            magicCursorPoint.getX().getDouble() - 2.0, magicCursorPoint.getY().getDouble() - 2.0,
            5.0, 5.0));
      }
    }
  }

  private Graphics2D createGraphicsWithRenderingHintsConfigured(Graphics basedOn) {
    final Map<Object,Object> additionalHints = Maps.newLinkedHashMap();
    // Needed for correct rendering of dashed borders, which are drawn using strokes.
    additionalHints.put(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_PURE);
    if (cellWidget.getForceFractionalMetrics()) {
      additionalHints.put(
          RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
    }
    return UIUtil.createGraphicsWithRenderingHintsConfigured(basedOn, additionalHints);
  }

  /* === Scrollable Interface =================================================================== */
  private boolean scrollableTracksViewportWidth;
  private boolean scrollableTracksViewportHeight;

  private void updateScrollables() {
    if (cellWidget.getFreezeScrollables())
      return;
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    final boolean newScrollableTracksViewportWidth;
    final boolean newScrollableTracksViewportHeight;
    Coordinates newPreferredSize = new Coordinates(1, 1);

    if (generatedCell == null) {
      newScrollableTracksViewportWidth = true;
      newScrollableTracksViewportHeight = true;
    } else {
      /* Enabling scroll tracking when the layout is smaller than the viewport ensures that cursor
      tooltips and other graphics can still be painted anywhere in the viewport without clipping. */
      newScrollableTracksViewportWidth = generatedCell.getTracksViewportWidth() ||
          getViewportDimensions().getX().isGreaterThan(generatedCell.getActualDimensions().getX());
      newScrollableTracksViewportHeight = generatedCell.getTracksViewportHeight() ||
          getViewportDimensions().getY().isGreaterThan(generatedCell.getActualDimensions().getY());
      final Coordinates actualDimensions = generatedCell.getActualDimensions();
      final int infiniteScrollPadding =
          generatedCell.getResult().getData().isIncomplete() ? INFINITE_SCROLL_PADDING_PIXELS : 0;
      newPreferredSize = newPreferredSize.max(new Coordinates(
          actualDimensions.getX().roundUp(Pixels.ONE).getInteger(),
          actualDimensions.getY().roundUp(Pixels.ONE).getInteger() + infiniteScrollPadding));
    }
    /* Avoid the viewport forcibly scrolling back, in particular for the case where the complete
    result turns out to be shorter than the previous incomplete result plus the infinite scroll
    padding. It's probably best to do this in the vertical direction only, as forced scrolling in
    the horizontal direction might actually be desirable for instance when narrowing layout changes
    are made. Furthermore, only apply this adjustment if it corresponds to half the viewport height
    or less. */
    {
      final Pixels currentViewportEnd =
          viewportRectangle.getPosition().plus(viewportRectangle.getDimensions()).getY();
      if (!currentViewportEnd.minus(newPreferredSize.getY()).times(2)
          .isGreaterThan(viewportRectangle.getHeight()))
      {
        newPreferredSize = newPreferredSize.max(new Coordinates(Pixels.ZERO, currentViewportEnd));
      }
    }
    final Dimension newPreferredSizeDim = newPreferredSize.getDimension();

    final boolean modified = (
        !newPreferredSizeDim.equals(cellWidget.getPreferredSize()) ||
        newScrollableTracksViewportWidth != scrollableTracksViewportWidth ||
        newScrollableTracksViewportHeight != scrollableTracksViewportHeight);
    /* TODO[pd]: Override ComponentUI.getPreferredSize() instead of calling
                 cellWidget.setPreferredSize(), making sure to fire an appropriate property change
                 event. This will ensure correct behavior for clients that may attempt to override
                 the preferred size. */
    cellWidget.setPreferredSize(newPreferredSizeDim);
    scrollableTracksViewportWidth = newScrollableTracksViewportWidth;
    scrollableTracksViewportHeight = newScrollableTracksViewportHeight;
    /* getScrollableTracksViewportWidth/Height are queried by the the javax.swing.ScrollPaneLayout
    layout manager, so the right way to propagate changes is to request the layout to be updated
    using revalidate(). Note that revalidate() does not necessarily imply repaint(). */
    if (modified)
      cellWidget.revalidate();
  }

  @Override
  public Dimension getPreferredScrollableViewportSize() {
    // TODO[pd]: Consider implementing something analogous to JList.setVisibleRowCount().
    return cellWidget.getPreferredSize();
  }

  @Override
  public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
    return 16;
  }

  @Override
  public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) {
    return 100;
  }

  /* Implementing getScrollableTracksViewportWidth/Height prevents scrollbars from appearing briefly
  while preferred dimensions are being updated. I still set explicit preferred dimensions for cases
  where the minimum layout dimensions exceed the available space. */

  @Override
  public boolean getScrollableTracksViewportWidth() {
    return scrollableTracksViewportWidth;
  }

  @Override
  public boolean getScrollableTracksViewportHeight() {
    return scrollableTracksViewportHeight;
  }

  private @Nullable PixelRectangle findCursorRectangle() {
    return findCursorRectangle(getEffectiveSelectionModel().getCursorLocation());
  }

  private @Nullable PixelRectangle findCursorRectangle(@Nullable Location cursorLocation) {
    return rectangleFinder.findCursorRectangle(cursorLocation, getRectangleFindingStartPoint());
  }

  // TODO[cc]: See if the clipBounds argument is still relevant, otherwise, remove that case.
  /**
   * @param clipBounds the rectangle to search; if null, the entire cell will be searched, subject
   *        to some optimizations
   */
  private CellRectangles findRectangles(
      @Nullable Location cursorLocation, @Nullable PixelRectangle clipBounds,
      @Nullable Coordinates frozenHeaderPoint)
  {
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    if (generatedCell == null || cursorLocation == null)
      return new CellRectangles();
    if (clipBounds != null) {
      return generatedCell.getCell().findRectangles(
          generatedCell.getRenderContext(), cursorLocation, clipBounds, frozenHeaderPoint);
    } else {
      return rectangleFinder.findRectangles(cursorLocation,
          getRectangleFindingStartPoint(), frozenHeaderPoint);
    }
  }

  private Coordinates getRectangleFindingStartPoint() {
    final GeneratedCell generatedCell = cellWidget.getGeneratedCell();
    if (generatedCell == null)
      return Coordinates.ZERO;
    return viewportRectangle.getPosition().plus(new Coordinates(
        viewportRectangle.getWidth().divideBy(generatedCell.getRenderContext(), 2.0),
        viewportRectangle.getHeight().divideBy(generatedCell.getRenderContext(), 2.0)));
  }

  @Override
  public void openCursorPopup(PopupLike popupMenu) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkNotNull(popupMenu);
    Coordinates pos;
    final CellRectangles cellRectangles = findRectangles(
        getEffectiveSelectionModel().getCursorLocation(), null, getViewportPosition());
    final PixelRectangle cursorRect = cellRectangles.getCursorRectangle();
    if (cursorRect == null) {
      pos = new Coordinates(Pixels.create(15), Pixels.create(15));
    } else {
      /* Show the popup menu on the top-right hand side of the cell (before any constraints are
      applied). Limit this to a constant amount (scaled with zoom level) to avoid the toolbox ending
      up all the way to the right of the screen for very wide cells. */
      Pixels horizontalOffset = cursorRect.getWidth();
      GeneratedCell generatedCell = cellWidget.getGeneratedCell();
      if (generatedCell != null) {
        RenderContext rcx = generatedCell.getRenderContext();
        horizontalOffset = horizontalOffset.min(Pixels.create(rcx, MAX_HORIZONTAL_POPUP_OFFSET));
      }
      pos = new Coordinates(cursorRect.getX().plus(horizontalOffset), cursorRect.getY());
      /* Offset the vertical position of the menu upwards a bit, like Excel does, and add a small
      horizontal margin. For comparison, Google Docs spreadsheets currently just uses the upper
      left-hand corner of the cell, obscuring the cell with the popup menu when the menu is invoked
      with the keyboard. */
      pos = pos.plus(new Coordinates(5, -50));
    }
    /* Avoid showing the popup outside of the scroll pane area, even if the selected cell is
    off-screen. */
    pos = pos
        .max(getViewportPosition())
        .min(getViewportPosition().plus(getViewportDimensions()));
    // Try to avoid opening the popup over any headers relevant to the current cursor location.
    for (HeaderRectangle header : cellRectangles.getHeaderRectangles()) {
      PixelRectangle headerRect = header.rectangle;
      final Coordinates headerEnd = headerRect.getPosition().plus(headerRect.getDimensions());
      pos = pos.max(header.direction.newCoordinates(headerEnd, Coordinates.ZERO));
    }
    openPopup(popupMenu, pos);
  }

  private void openPopup(final PopupLike popupMenu, final Coordinates position) {
    ThreadUtil.checkOnEDT();
    Preconditions.checkNotNull(popupMenu);
    Preconditions.checkNotNull(position);
    if (installedInplaceEditor != null)
      return;
    if (currentlyShowingPopup != null) {
      /* Don't try to do this from setCurrentlyShowingPopup, as it also gets called from
      popupMenuWillBecomeInvisible, which is called on cancel(). */
      currentlyShowingPopup.cancel();
    }
    setCurrentlyShowingPopup(popupMenu);
    popupMenu.show(cellWidget, position.getPoint());
  }

  private void updateSelectionForPopup(boolean opened) {
    /* TODO[pd]: Once I want to allow actions that do not necessarily operate on a per-path basis,
                 the InteractionProvider should be able to request different popup selection
                 behavior. */
    final SelectionModel model = getEffectiveSelectionModel();
    final Selection selection = model.getSelection();
    final Location cursorLocation = selection.getCursorLocation();
    if (cursorLocation == null)
      return;
    final Path cursorPath = cursorLocation.getPath();
    final Set<Path> previouslySelectedPaths = selection.getSelectedPaths();
    if (opened) {
      /* Ensure that the path under the cursor is selected. I considered using the more liberal
      condition previouslySelectedPaths.contains(cursorPath), but it would ruin the selection in
      several cases where it needs to be preserved. */
      if (!Views.isAncestorOrEqualInSet(previouslySelectedPaths, cursorPath))
        model.setSelectedPaths(ImmutableSet.of(cursorPath));
    } else {
      /* For simplicity (to avoid relying on extra stored state), undo any single path selection
      regardless of whether it was automatically added when opening the popup or not. Don't
      require the selected path to be on the cursor; some unhiding operations may cause the cursor
      to move away from the original path selection. */
      if (previouslySelectedPaths.size() == 1)
        model.setSelectedPaths(ImmutableSet.of());
    }
  }
}
