// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.sieuferd.uiutil.scroll.OverlayScrollable;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.cell.styles.PathHighlights;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.cell.support.SelectionModel;
import com.sieuferd.uiutil.api.KeyStrokeForwardable;
import com.sieuferd.uiutil.ThreadUtil;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.KeyboardFocusManager;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import javax.swing.Scrollable;
import javax.swing.UIManager;

/**
 * Component for displaying and interacting with {@link com.sieuferd.cell.support.Cell} layouts.
 * EDT only.
 */
public final class CellWidget extends JComponent
    implements Scrollable, OverlayScrollable, KeyStrokeForwardable
{
  private static final String uiClassID = "CellWidgetUI";

  public static final String PROP_FORCE_FRACTIONAL_METRICS = "forceFractionalMetrics";
  public static final String PROP_ALPHA_PERMITTED = "alphaPermitted";
  public static final String PROP_DRAW_CELL_DEBUG_GRAPHICS = "drawCellDebugGraphics";
  public static final String PROP_SCALING = "scaling";
  public static final String PROP_SELECTION_MODEL = "selectionModel";
  public static final String PROP_REFERENCE_SELECTION_MODEL = "referenceSelectionModel";
  public static final String PROP_GENERATED_CELL = "generatedCell";
  public static final String PROP_RENDER_CONTEXT = "renderContext";
  public static final String PROP_VIEWPORT_DIMENSIONS = "viewportDimensions";
  public static final String PROP_INTERACTION_PROVIDER = "interactionProvider";
  public static final String PROP_FREEZE_SCROLLABLES = "freezeScrollables";
  public static final String PROP_PATH_HIGHLIGHTS = "pathHighlights";

  private @Nullable DirtyScrollableObserver dirtyScrollableObserver;
  private SelectionModel selectionModel = new SelectionModel();
  private @Nullable SelectionModel referenceSelectionModel;
  private @Nullable GeneratedCell generatedCell;
  private double scaling = 1.0;
  private boolean forceFractionalMetrics = false;
  private boolean alphaPermitted = true;
  private boolean drawCellDebugGraphics = false;
  private boolean freezeScrollables = false;
  private @Nullable RenderContext renderContext;
  private @Nullable Coordinates viewportDimensions;
  private InteractionProvider interactionProvider = InteractionProvider.createDefault();
  private PathHighlights pathHighlights = PathHighlights.NONE;

  /* === Constructor ============================================================================ */
  public CellWidget() {
    setLayout(null);

    setFocusable(true);
    /* Same as for JTable constructor. Clients may use setFocusTraversalKeysEnabled(false) to
    disable. */
    setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS,
        ImmutableSet.of(KeyStroke.getKeyStroke(KeyEvent.VK_TAB,
        InputEvent.CTRL_MASK)));
    setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS,
        ImmutableSet.of(KeyStroke.getKeyStroke(KeyEvent.VK_TAB,
        InputEvent.SHIFT_MASK | InputEvent.CTRL_MASK)));

    setAutoscrolls(true);
    /* Don't call updateUI() until the end; no non-nullable fields should be null once a this
    reference is exposed to the UI delegate. */
    updateUI();
  }

  /* === Look & Feel Boilerplate ================================================================ */
  @Override
  public String getUIClassID() {
    return uiClassID;
  }

  public void setUI(CellWidgetUI ui) {
    super.setUI(ui);
  }

  public CellWidgetUI getUI() {
    return (CellWidgetUI) ui;
  }

  @Override
  public void updateUI() {
    /* TODO[pd]: Remove any embedded components before the CellWidgetUI is replaced. Also, if the
                 enclosing JScrollPane is automatically configured with custom borders or similar
                 that depend on the L&F, update these settings here. See JTable.updateUI(). */
    if (UIManager.get(getUIClassID()) != null) {
      setUI((CellWidgetUI) UIManager.getUI(this));
    } else {
      setUI(new BasicCellWidgetUI());
    }
  }

  /* === Bean Accessors for Component Properties ================================================ */
  /* Follow the pattern in
     http://docs.oracle.com/javase/tutorial/javabeans/writing/properties.html . */
  public void setForceFractionalMetrics(boolean forceFractionalMetrics) {
    boolean oldValue = this.forceFractionalMetrics;
    this.forceFractionalMetrics = forceFractionalMetrics;
    firePropertyChange(PROP_FORCE_FRACTIONAL_METRICS, oldValue, forceFractionalMetrics);
  }

  public boolean getForceFractionalMetrics() {
    return forceFractionalMetrics;
  }

  public void setAlphaPermitted(boolean alphaPermitted) {
    boolean oldValue = this.alphaPermitted;
    this.alphaPermitted = alphaPermitted;
    firePropertyChange(PROP_ALPHA_PERMITTED, oldValue, alphaPermitted);
  }

  public boolean isAlphaPermitted() {
    return alphaPermitted;
  }

  public void setDrawCellDebugGraphics(boolean drawCellDebugGraphics) {
    boolean oldValue = this.drawCellDebugGraphics;
    this.drawCellDebugGraphics = drawCellDebugGraphics;
    firePropertyChange(PROP_DRAW_CELL_DEBUG_GRAPHICS, oldValue, drawCellDebugGraphics);
  }

  public boolean getDrawCellDebugGraphics() {
    return drawCellDebugGraphics;
  }

  public void setScaling(double scaling) {
    Preconditions.checkArgument(scaling > 0.0);
    Preconditions.checkArgument(Double.isFinite(scaling));
    double oldValue = this.scaling;
    this.scaling = scaling;
    firePropertyChange(PROP_SCALING, oldValue, scaling);
  }

  public double getScaling() {
    return scaling;
  }

  public void setGeneratedCell(@Nullable GeneratedCell generatedCell) {
    ThreadUtil.checkOnEDT();
    GeneratedCell oldValue = this.generatedCell;
    this.generatedCell = generatedCell;
    if (oldValue != null || generatedCell != null)
      firePropertyChange(PROP_GENERATED_CELL, oldValue, generatedCell);
  }

  public @Nullable GeneratedCell getGeneratedCell() {
    return generatedCell;
  }

  public void setSelectionModel(SelectionModel selectionModel) {
    Preconditions.checkNotNull(selectionModel);
    SelectionModel oldValue = this.selectionModel;
    this.selectionModel = selectionModel;
    firePropertyChange(PROP_SELECTION_MODEL, oldValue, selectionModel);
  }

  public SelectionModel getSelectionModel() {
    if (selectionModel == null)
      throw new AssertionError();
    return selectionModel;
  }

  /**
   * Set a selection model to be used in reference selection mode. The {@code CellWidget} will be in
   * this mode whenever the {@code referenceSelectionModel} is not null.
   *
   * Spreadsheet-style formula editors may rely on changes in the reference selection to signal
   * times at which a reference should be inserted or adjusted. A special case is when the user
   * clicks an already-selected cell; in this case, the reference selection model will have its
   * selection momentarily cleared and restored, to ensure change listeners are still notified.
   */
  public void setReferenceSelectionModel(@Nullable SelectionModel referenceSelectionModel) {
    SelectionModel oldValue = this.referenceSelectionModel;
    this.referenceSelectionModel = referenceSelectionModel;
    firePropertyChange(PROP_REFERENCE_SELECTION_MODEL, oldValue, referenceSelectionModel);
  }

  public @Nullable SelectionModel getReferenceSelectionModel() {
    return referenceSelectionModel;
  }

  /** Set the reported RenderContext. Should only be called by the UI delegate. */
  void setRenderContext(@Nullable RenderContext renderContext) {
    RenderContext oldValue = this.renderContext;
    this.renderContext = renderContext;
    if (oldValue != null || renderContext != null)
      firePropertyChange(PROP_RENDER_CONTEXT, oldValue, renderContext);
  }

  public @Nullable RenderContext getRenderContext() {
    return renderContext;
  }

  /**
   *  Set the reported viewport size. Should only be called by the UI delegate.
   */
  void setViewportDimensions(@Nullable Coordinates viewportDimensions) {
    Coordinates oldValue = this.viewportDimensions;
    this.viewportDimensions = viewportDimensions;
    if (oldValue != null || viewportDimensions != null)
      firePropertyChange(PROP_VIEWPORT_DIMENSIONS, oldValue, viewportDimensions);
  }

  public @Nullable Coordinates getViewportDimensions() {
    return viewportDimensions;
  }

  public void setInteractionProvider(@Nonnull InteractionProvider interactionProvider) {
    Preconditions.checkNotNull(interactionProvider);
    InteractionProvider oldValue = this.interactionProvider;
    this.interactionProvider = interactionProvider;
    firePropertyChange(PROP_INTERACTION_PROVIDER, oldValue, interactionProvider);
  }

  public @Nonnull InteractionProvider getInteractionProvider() {
    return interactionProvider;
  }

  public void setFreezeScrollables(boolean freezeScrollables) {
    boolean oldValue = this.freezeScrollables;
    this.freezeScrollables = freezeScrollables;
    if (oldValue != freezeScrollables)
      firePropertyChange(PROP_FREEZE_SCROLLABLES, oldValue, freezeScrollables);
  }

  public boolean getFreezeScrollables() {
    return freezeScrollables;
  }

  public void setPathHighlights(PathHighlights pathHighlights) {
    Preconditions.checkNotNull(pathHighlights);
    PathHighlights oldValue = this.pathHighlights;
    if (!oldValue.equals(pathHighlights)) {
      this.pathHighlights = pathHighlights;
      firePropertyChange(PROP_PATH_HIGHLIGHTS, oldValue, pathHighlights);
    }
  }

  public PathHighlights getPathHighlights() {
    return pathHighlights;
  }

  /* === Scrollable Interface (all delegated) =================================================== */
  @Override
  public Dimension getPreferredScrollableViewportSize() {
    return getUI().getPreferredScrollableViewportSize();
  }

  @Override
  public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
    return getUI().getScrollableUnitIncrement(visibleRect, orientation, direction);
  }

  @Override
  public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) {
    return getUI().getScrollableBlockIncrement(visibleRect, orientation, direction);
  }

  @Override
  public boolean getScrollableTracksViewportWidth() {
    return getUI().getScrollableTracksViewportWidth();
  }

  @Override
  public boolean getScrollableTracksViewportHeight() {
    return getUI().getScrollableTracksViewportHeight();
  }

  /* === OverlayScrollable Interface ============================================================ */

  // Make public.
  @Override
  public void paintComponent(Graphics scrollableG) {
    super.paintComponent(scrollableG);
  }

  @Override
  public void paintOverlay(Graphics scrollableG) {
    getUI().paintOverlay(scrollableG);
  }

  @Override
  public void setDirtyScrollableObserver(DirtyScrollableObserver observer) {
    this.dirtyScrollableObserver = observer;
  }

  /** Repaint the overlay, if any. Should only be called by the UI delegate. */
  void repaintOverlay() {
    dirtyScrollableObserver.markOverlayDirty();
  }

  /** Repaint and notify the ScrollableObserver. Should only be called by the UI delegate. */
  void repaintScrollable(Rectangle r) {
    if (dirtyScrollableObserver != null)
      dirtyScrollableObserver.markScrollableDirty(r);
    super.repaint(r);
  }

  /* === Other Overrideables  =================================================================== */

  @Override
  public boolean requestFocusInWindow() {
    if (!isFocusable())
      return false;
    final JComponent focusComponent = getUI().getFocusComponent();
    if (focusComponent == this) {
      return super.requestFocusInWindow();
    } else {
      return focusComponent.requestFocusInWindow();
    }
  }

  @Override
  public void requestFocus() {
    if (!isFocusable())
      return;
    final JComponent focusComponent = getUI().getFocusComponent();
    if (focusComponent == this) {
      super.requestFocus();
    } else {
      focusComponent.requestFocus();
    }
  }

  /**
   * Handle a keystroke that was otherwise unhandled by a formula bar or similar external component
   * to the {@code CellWidget}. They keystroke will be handled in the same way as a keystroke typed
   * in, but unhandled by, an inplace editor.
   */
  @Override
  public boolean forwardUnhandledKeyStroke(KeyStroke ks, KeyEvent evt, boolean pressed) {
    // To get the same behavior as for inplace editors, don't include JComponent.FOCUSED here.
    return super.processKeyBinding(ks, evt, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, pressed);
  }

  @Override
  protected boolean processKeyBinding(KeyStroke ks, KeyEvent evt, int condition, boolean pressed) {
    Preconditions.checkNotNull(ks);
    Preconditions.checkNotNull(evt);
    // Use the same approach as JTable.processKeyBinding().
    final boolean ret = super.processKeyBinding(ks, evt, condition, pressed);
    /* This method (processKeyBinding) is called first with condition=WHEN_FOCUSED and then again
    with WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, so I only need to check for a single condition value
    here. See JComponent.processKeyBindings(KeyEvent,boolean). */
    if (!ret && condition == JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT && isFocusOwner()) {
      return getUI().forwardUnhandledKeyStroke(ks, evt, pressed);
    } else {
      return ret;
    }
  }
}
