// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import javax.swing.Action;
import javax.swing.ActionMap;
import javax.swing.TransferHandler;
import javax.swing.plaf.ActionMapUIResource;
import com.sieuferd.widget.PhysicalCursorMovementAction.CursorDistance;
import com.sieuferd.cell.support.Direction;
import com.sieuferd.uiutil.api.PopupLike;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import javax.swing.AbstractAction;
import javax.swing.SwingUtilities;

/** Action constants from the javax.swing.plaf.basic.BasicTableUI class, to allow using the same
keyboard bindings for CellWidget as for the standard JTable. */
public final class CellWidgetActions {
  /* === Action Name Constants ========================================================== */
  public static final String CANCEL_EDITING = "cancel";
  public static final String SELECT_ALL = "selectAll";
  public static final String CLEAR_SELECTION = "clearSelection";
  public static final String START_EDITING = "startEditing";

  public static final String NEXT_ROW = "selectNextRow";
  public static final String NEXT_ROW_CELL = "selectNextRowCell";
  public static final String NEXT_ROW_EXTEND_SELECTION = "selectNextRowExtendSelection";
  public static final String NEXT_ROW_CHANGE_LEAD = "selectNextRowChangeLead";
  public static final String PREVIOUS_ROW = "selectPreviousRow";
  public static final String PREVIOUS_ROW_CELL = "selectPreviousRowCell";
  public static final String PREVIOUS_ROW_EXTEND_SELECTION = "selectPreviousRowExtendSelection";
  public static final String PREVIOUS_ROW_CHANGE_LEAD = "selectPreviousRowChangeLead";

  public static final String NEXT_COLUMN = "selectNextColumn";
  public static final String NEXT_COLUMN_CELL = "selectNextColumnCell";
  public static final String NEXT_COLUMN_EXTEND_SELECTION = "selectNextColumnExtendSelection";
  public static final String NEXT_COLUMN_CHANGE_LEAD = "selectNextColumnChangeLead";
  public static final String PREVIOUS_COLUMN = "selectPreviousColumn";
  public static final String PREVIOUS_COLUMN_CELL = "selectPreviousColumnCell";
  public static final String PREVIOUS_COLUMN_EXTEND_SELECTION =
      "selectPreviousColumnExtendSelection";
  public static final String PREVIOUS_COLUMN_CHANGE_LEAD = "selectPreviousColumnChangeLead";

  public static final String SCROLL_LEFT_CHANGE_SELECTION = "scrollLeftChangeSelection";
  public static final String SCROLL_LEFT_EXTEND_SELECTION = "scrollLeftExtendSelection";
  public static final String SCROLL_RIGHT_CHANGE_SELECTION = "scrollRightChangeSelection";
  public static final String SCROLL_RIGHT_EXTEND_SELECTION = "scrollRightExtendSelection";

  public static final String SCROLL_UP_CHANGE_SELECTION = "scrollUpChangeSelection";
  public static final String SCROLL_UP_EXTEND_SELECTION = "scrollUpExtendSelection";
  public static final String SCROLL_DOWN_CHANGE_SELECTION = "scrollDownChangeSelection";
  public static final String SCROLL_DOWN_EXTEND_SELECTION = "scrollDownExtendSelection";

  public static final String FIRST_COLUMN = "selectFirstColumn";
  public static final String FIRST_COLUMN_EXTEND_SELECTION = "selectFirstColumnExtendSelection";
  public static final String LAST_COLUMN = "selectLastColumn";
  public static final String LAST_COLUMN_EXTEND_SELECTION = "selectLastColumnExtendSelection";

  public static final String FIRST_ROW = "selectFirstRow";
  public static final String FIRST_ROW_EXTEND_SELECTION = "selectFirstRowExtendSelection";
  public static final String LAST_ROW = "selectLastRow";
  public static final String LAST_ROW_EXTEND_SELECTION = "selectLastRowExtendSelection";

  // Javadoc comments below also from BasicTableUI.

  /** Add the lead item to the selection without changing lead or anchor. */
  public static final String ADD_TO_SELECTION = "addToSelection";

  /** Toggle the selected state of the lead item and move the anchor to it. */
  public static final String TOGGLE_AND_ANCHOR = "toggleAndAnchor";

  /** Extend the selection to the lead item. */
  public static final String EXTEND_TO = "extendTo";

  /** Move the anchor to the lead and ensure only that item is selected. */
  public static final String MOVE_SELECTION_TO = "moveSelectionTo";

  /** Give focus to the header. */
  public static final String FOCUS_HEADER = "focusHeader";

  // Clipboard action names.
  public static final String CUT = (String) TransferHandler.getCutAction().getValue(Action.NAME);
  public static final String COPY = (String) TransferHandler.getCopyAction().getValue(Action.NAME);
  public static final String PASTE = (String) TransferHandler.getPasteAction().getValue(Action.NAME);

  // Other actions. These action names are _not_ from existing actions.
  public static final String OPEN_POPUP_MENU = "openPopupMenu";

  /* ============================================================================================ */

  private CellWidgetActions() { }

  private static void putAction(ActionMap map, Action action) {
    map.put(action.getValue(Action.NAME), action);
  }

  /** Based on {@code BasicTableUI.loadActionMap}. */
  static ActionMap createActionMap() {
    ActionMap ret = new ActionMapUIResource();

    putAction(ret, new CellWidgetAction(ADD_TO_SELECTION));
    putAction(ret, new CellWidgetAction(TOGGLE_AND_ANCHOR));
    putAction(ret, new CellWidgetAction(EXTEND_TO));
    putAction(ret, new CellWidgetAction(MOVE_SELECTION_TO));
    putAction(ret, new CellWidgetAction(FOCUS_HEADER));

    putAction(ret, new CellWidgetAction(SELECT_ALL));
    putAction(ret, new CellWidgetAction(CLEAR_SELECTION));

    putAction(ret, new CellWidgetAction(PREVIOUS_ROW_CHANGE_LEAD, 0, -1, false, false));
    putAction(ret, new CellWidgetAction(NEXT_ROW_CHANGE_LEAD, 0, 1, false, false));
    putAction(ret, new CellWidgetAction(PREVIOUS_COLUMN_CHANGE_LEAD, -1, 0, false, false));
    putAction(ret, new CellWidgetAction(NEXT_COLUMN_CHANGE_LEAD, 1, 0, false, false));

    // Keep everything in {Up, Down, Left, Right} order.

    // Logical movements.
    putAction(ret, new LogicalCursorMovementAction(PREVIOUS_ROW_CELL, Direction.UP));
    putAction(ret, new LogicalCursorMovementAction(NEXT_ROW_CELL, Direction.DOWN));
    putAction(ret, new LogicalCursorMovementAction(PREVIOUS_COLUMN_CELL, Direction.LEFT));
    putAction(ret, new LogicalCursorMovementAction(NEXT_COLUMN_CELL, Direction.RIGHT));

    // Physical movements, one cell.
    putAction(ret, new PhysicalCursorMovementAction(PREVIOUS_ROW, Direction.UP, CursorDistance.ONE_CELL, false));
    putAction(ret, new PhysicalCursorMovementAction(NEXT_ROW, Direction.DOWN, CursorDistance.ONE_CELL, false));
    putAction(ret, new PhysicalCursorMovementAction(PREVIOUS_COLUMN, Direction.LEFT, CursorDistance.ONE_CELL, false));
    putAction(ret, new PhysicalCursorMovementAction(NEXT_COLUMN, Direction.RIGHT, CursorDistance.ONE_CELL, false));

    // Physical movements, one page.
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_UP_CHANGE_SELECTION, Direction.UP, CursorDistance.ONE_PAGE, false));
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_DOWN_CHANGE_SELECTION, Direction.DOWN, CursorDistance.ONE_PAGE, false));
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_LEFT_CHANGE_SELECTION, Direction.LEFT, CursorDistance.ONE_PAGE, false));
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_RIGHT_CHANGE_SELECTION, Direction.RIGHT, CursorDistance.ONE_PAGE, false));

    // Physical movements, to end.
    putAction(ret, new PhysicalCursorMovementAction(FIRST_ROW, Direction.UP, CursorDistance.TO_END, false));
    putAction(ret, new PhysicalCursorMovementAction(LAST_ROW, Direction.DOWN, CursorDistance.TO_END, false));
    putAction(ret, new PhysicalCursorMovementAction(FIRST_COLUMN, Direction.LEFT, CursorDistance.TO_END, false));
    putAction(ret, new PhysicalCursorMovementAction(LAST_COLUMN, Direction.RIGHT, CursorDistance.TO_END, false));

    // Physical movements, one cell, extending selection.
    putAction(ret, new PhysicalCursorMovementAction(PREVIOUS_ROW_EXTEND_SELECTION, Direction.UP, CursorDistance.ONE_CELL, true));
    putAction(ret, new PhysicalCursorMovementAction(NEXT_ROW_EXTEND_SELECTION, Direction.DOWN, CursorDistance.ONE_CELL, true));
    putAction(ret, new PhysicalCursorMovementAction(PREVIOUS_COLUMN_EXTEND_SELECTION, Direction.LEFT, CursorDistance.ONE_CELL, true));
    putAction(ret, new PhysicalCursorMovementAction(NEXT_COLUMN_EXTEND_SELECTION, Direction.RIGHT, CursorDistance.ONE_CELL, true));

    // Physical movements, one page, extending selection.
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_UP_EXTEND_SELECTION, Direction.UP, CursorDistance.ONE_PAGE, true));
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_DOWN_EXTEND_SELECTION, Direction.DOWN, CursorDistance.ONE_PAGE, true));
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_LEFT_EXTEND_SELECTION, Direction.LEFT, CursorDistance.ONE_PAGE, true));
    putAction(ret, new PhysicalCursorMovementAction(SCROLL_RIGHT_EXTEND_SELECTION, Direction.RIGHT, CursorDistance.ONE_PAGE, true));

    // Physical movements, to end, extending selection.
    putAction(ret, new PhysicalCursorMovementAction(FIRST_ROW_EXTEND_SELECTION, Direction.UP, CursorDistance.TO_END, true));
    putAction(ret, new PhysicalCursorMovementAction(LAST_ROW_EXTEND_SELECTION, Direction.DOWN, CursorDistance.TO_END, true));
    putAction(ret, new PhysicalCursorMovementAction(FIRST_COLUMN_EXTEND_SELECTION, Direction.LEFT, CursorDistance.TO_END, true));
    putAction(ret, new PhysicalCursorMovementAction(LAST_COLUMN_EXTEND_SELECTION, Direction.RIGHT, CursorDistance.TO_END, true));

    ret.put(CUT, TransferHandler.getCutAction());
    ret.put(COPY, TransferHandler.getCopyAction());
    ret.put(PASTE, TransferHandler.getPasteAction());

    putAction(ret, new AbstractAction(OPEN_POPUP_MENU) {
      @Override
      public void actionPerformed(ActionEvent e) {
        CellWidget cellWidget = (CellWidget) e.getSource();
        PopupLike popup =
            cellWidget.getInteractionProvider().getMousePopup(MouseEvent.NOBUTTON);
        if (popup != null)
          cellWidget.getUI().openCursorPopup(popup);
      }
    });

    putAction(ret, new AbstractAction(START_EDITING) {
      @Override
      public void actionPerformed(ActionEvent e) {
        CellWidget cellWidget = (CellWidget) e.getSource();
        cellWidget.getUI().startInplaceEdit();
      }
    });

    putAction(ret, new AbstractAction(CANCEL_EDITING) {
      @Override
      public void actionPerformed(ActionEvent e) {
        CellWidget cellWidget = (CellWidget) e.getSource();
        /* Close the inplace editor and request focus in the CellWidget before officially canceling
        the edit, and postpone the latter with an invokeLater, to avoid a brief cursor flicker in
        the formula bar. */
        cellWidget.getUI().closeInplaceEditor();
        cellWidget.requestFocusInWindow();
        SwingUtilities.invokeLater(() -> {
          cellWidget.getInteractionProvider().cancelEdit();
        });
      }
    });

    return ret;
  }
}
