// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.sieuferd.cell.support.Cell;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.cell.support.RenderContext;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.datamodel.Views;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.ValueProcessor;
import com.sieuferd.flow.Flows;
import com.sieuferd.layout.LabelFormatter;
import com.sieuferd.layout.LayoutGenerator;
import com.sieuferd.layout.LayoutMeasurer;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.flow.UIFlows;
import com.sieuferd.util.Stopwatch;
import java.util.concurrent.atomic.AtomicLong;
import java.util.logging.Logger;

public final class CellWidgetUtil {
  private static final Logger LOG = Logger.getLogger(CellWidgetUtil.class.getName());
  private static final int VIEWPORT_WIDTH_COALESCING_MS = 200;
  private static final AtomicLong DEBUG_NEXT_LAYOUT_ID = new AtomicLong();

  private CellWidgetUtil() { }

  public static Flow<GeneratedCell> getCellGenerator(CellWidget widget, Flow<ViewResult> resultFlow,
      Flow<Boolean> schemaOnlyFlow, LabelFormatter labelFormatter, boolean measure)
  {
    return getCellGenerator(getViewportWidthFlow(widget, VIEWPORT_WIDTH_COALESCING_MS),
        getRenderContextFlow(widget), resultFlow, schemaOnlyFlow, labelFormatter, measure);
  }

  public static Flow<GeneratedCell> getCellGenerator(
      Flow<Pixels> viewportWidth, Flow<RenderContext> renderContext, Flow<ViewResult> resultFlow,
      Flow<Boolean> schemaOnlyFlow, final LabelFormatter labelFormatter,
      final boolean measure)
  {
    final Flow<LayoutInput> layoutInput = Flows.newComposite4(
        resultFlow, schemaOnlyFlow, viewportWidth, renderContext,
        (viewResult, schemaOnly, viewportWidth2, renderContext2) ->
        new LayoutInput(viewResult, schemaOnly, viewportWidth2, renderContext2));
    return Flows.newProcessed(layoutInput,
        new ValueProcessor<LayoutInput,GeneratedCell>()
    {
      @Override
      public ValueProcessor.Synchronicity getSynchronicity() {
        return Synchronicity.ASYNCHRONOUS_INTERRUPTABLE;
      }

      @Override
      public GeneratedCell process(LayoutInput input) throws InterruptedException {
        final long debugLayoutID = DEBUG_NEXT_LAYOUT_ID.getAndIncrement();
        final ViewResult result = input.getResult();
        Pixels viewportWidth = input.getViewportWidth();
        RenderContext renderContext = input.getRenderContext();
        MutableView mutableView = Views.newMutable(result.getView());
        if (measure)
          LayoutMeasurer.measureData(result.getData(), mutableView, renderContext);
        Cell layout;
        UIUtil.instrumentation("AutoStyle layoutID=" + debugLayoutID);
        Stopwatch stopwatch1 = new Stopwatch("Auto-style");
        layout = LayoutGenerator.autoStyle(result.getData(), mutableView, renderContext,
            viewportWidth, labelFormatter);
        stopwatch1.stop();
        UIUtil.instrumentation("CreateLayout layoutID=" + debugLayoutID);
        if (!input.isSchemaOnly()) {
          Stopwatch stopwatch2 = new Stopwatch("Layout creation");
          layout = LayoutGenerator.createLayout(result.getData(),
              ImmutableView.copyOf(mutableView), renderContext, viewportWidth,
              labelFormatter);
          stopwatch2.stop();
        }
        UIUtil.instrumentation("LayoutCreated layoutID=" + debugLayoutID);
        boolean tracksViewportWidth =
            (!layout.getDimensionsWithOverprint(renderContext).getX().isGreaterThan(viewportWidth));
        return new GeneratedCell(
            input.getResult(), layout, renderContext, tracksViewportWidth, false);
      }
    });
  }

  public static Flow<Pixels> getViewportWidthFlow(
      final CellWidget cellWidget, int coalesceMillis)
  {
    final String PROP_NAME = CellWidget.PROP_VIEWPORT_DIMENSIONS;
    return Flows.newCoalescing(coalesceMillis, UIFlows.fromPropertyBean(PROP_NAME,
        listener -> { cellWidget.addPropertyChangeListener(PROP_NAME, listener); },
        listener -> { cellWidget.removePropertyChangeListener(PROP_NAME, listener); },
        () -> {
          Coordinates viewportDimensions = cellWidget.getViewportDimensions();
          return viewportDimensions == null ? null : viewportDimensions.getX();
        }));
  }

  public static Flow<RenderContext> getRenderContextFlow(final CellWidget cellWidget) {
    final String PROP_NAME = CellWidget.PROP_RENDER_CONTEXT;
    return UIFlows.fromPropertyBean(PROP_NAME,
        listener -> { cellWidget.addPropertyChangeListener(PROP_NAME, listener); },
        listener -> { cellWidget.removePropertyChangeListener(PROP_NAME, listener); },
        () -> cellWidget.getRenderContext());
  }
}
