// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.cell.support.Direction;
import com.sieuferd.uiutil.api.CaretInfoProvider;
import com.sieuferd.uiutil.FilterInputMap;
import com.sieuferd.uiutil.ThreadUtil;
import java.awt.Toolkit;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.swing.Action;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.KeyStroke;
import javax.swing.text.DefaultEditorKit;

/**
 * An {@code InputMap} for use by cursor editors that needs to integrate with {@link CellWidget}.
 *
 * <p>A custom {@code InputMap} implementation was chosen over adding and removing mappings from
 * existing {@code InputMap}s, because JTextComponent's custom InputMap implementation does not
 * handle remove() correctly. Furthermore, the {@code FilterInputMap} approach works cleanly even if
 * the {@code CellWidget}'s {@code InputMap} is later changed.
 *
 * <p>When using this {@code InputMap} with a {@code JEditorPane}, the {@code InputMap} should be
 * set directly on the {@code JEditorPane}, not on a parent component. {@code JTable} takes a
 * different approach (using {@link JComponent#WHEN_ANCESTOR_OF_FOCUSED_COMPONENT} on the parent
 * {@code JTable} component), but generally has a much more complicated way of dealing with
 * component focus that is not needed here. See
 * {@link javax.swing.plaf.basic.BasicTableUI.Handler#keyTyped()}.
 */
public final class CursorEditInputMap extends FilterInputMap {
  /**
   * List of specific actions that the inplace editor should always be allowed to process, even if
   * they could be handled by the {@code CellWidget}. For now, the clipboard actions seem like the
   * only relevant ones. Actions like Redo are not handled directly by the {@code CellWidget}, and
   * so will not be excluded despite not appearing on this list.
   *
   * <p>Actions are identified by their value in the
   * CellWidget's (not the inplace editor's) InputMap.
   */
  private static final ImmutableSet<Object> NEVER_EXCLUDE_CWA = ImmutableSet.<Object>of(
      CellWidgetActions.CUT,
      CellWidgetActions.COPY,
      CellWidgetActions.PASTE
  );
  /**
   * List of specific actions that should always be handled by the {@code CellWidget}. This list
   * will only be used in extended mode, as set by {@link #setExtended(boolean)}.
   *
   * <p>Actions are identified by their value in the
   * CellWidget's (not the inplace editor's) InputMap.
   */
  private static final ImmutableSet<Object> ALWAYS_EXCLUDE_CWA = ImmutableSet.<Object>of(
      CellWidgetActions.CANCEL_EDITING,
      // Logical cursor movements ([Shift+]Tab/Enter).
      CellWidgetActions.PREVIOUS_ROW_CELL,
      CellWidgetActions.NEXT_ROW_CELL,
      CellWidgetActions.PREVIOUS_COLUMN_CELL,
      CellWidgetActions.NEXT_COLUMN_CELL
  );
  private final CellWidget cellWidget;
  private final JComponent editorComponent;

  private boolean extended;

  private CursorEditInputMap(CellWidget cellWidget, JComponent editorComponent) {
    Preconditions.checkNotNull(cellWidget);
    Preconditions.checkNotNull(editorComponent);
    this.cellWidget = cellWidget;
    this.editorComponent = editorComponent;
    installInsertBreakAction(this);
  }

  /**
   * Indicate whether the editor in question should operate in "extended" editing mode or not. In
   * extended editing mode, all except a select few actions are handled by the inplace editor, and
   * arrow keys, for instance, can be used to move the text caret within the cell. In non-extended,
   * mode, on the other hand, all except a select few actions are handled by the {@code CellWidget},
   * and arrow keys, for instance, will accept the edit and move the cell cursor.
   */
  public void setExtended(boolean extended) {
    ThreadUtil.checkOnEDT();
    this.extended = extended;
  }

  public boolean isExtended() {
    ThreadUtil.checkOnEDT();
    return extended;
  }

  private static void installInsertBreakAction(InputMap inputMap) {
    /* In-cell line breaks: Excel has different shortcuts on MacOS and Windows, whereas Google
    Docs Spreadsheets seems to accept any combination of Ctrl/Alt/Command modifiers held down
    while pressing Enter. Excel also seems to accept various other combinations. The "anything
    goes" approach seems reasonable, so add every combination of the relevant modifiers below. */

    for (int i = 0; i < 8; i++) {
      int modifiers = 0;
      // This will include the Command key on MacOS.
      if ((i & 1) != 0)
        modifiers |= Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
      /* Despite this, Alt+Enter does not seem to work on MacOS--though Excel for MacOS does not
      support this particular shortcut either. I suspect there might be legitimate reasons for
      this, e.g. related to non-US keyboard layouts. */
      if ((i & 2) != 0)
        modifiers |= InputEvent.ALT_DOWN_MASK;
      if ((i & 4) != 0)
        modifiers |= InputEvent.CTRL_DOWN_MASK;
      if (modifiers != 0) {
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, modifiers),
            DefaultEditorKit.insertBreakAction);
      }
    }
  }

  // TODO[cc]: Centralize calls to this.
  public static void configureCursorEditor(
      CellWidget cellWidget, JComponent component, boolean extended)
  {
    ThreadUtil.checkOnEDT();
    component.setFocusTraversalKeysEnabled(false);

    /* It's not necessary to include the WHEN_IN_FOCUSED_WINDOW condition here. Including it would
    require a more complicated version of FilterInputMap that extended from ComponentInputMap. */
    for (int condition :
        ImmutableList.of(JComponent.WHEN_FOCUSED, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT))
    {
      final InputMap existingInputMap = component.getInputMap(condition);
      final CursorEditInputMap cursorEditInputMap;
      if (existingInputMap instanceof CursorEditInputMap) {
        cursorEditInputMap = (CursorEditInputMap) existingInputMap;
      } else {
        cursorEditInputMap = new CursorEditInputMap(cellWidget, component);
        cursorEditInputMap.setParent(existingInputMap);
        component.setInputMap(condition, cursorEditInputMap);
      }
      cursorEditInputMap.setExtended(extended);
    }
  }

  @Override
  protected boolean excludesKeyStroke(KeyStroke keyStroke) {
    ThreadUtil.checkOnEDT();
    /* Always let the editor component handle keystrokes that correspond to characters; this
    condition exists mainly to preserve spacebar, which may be used by the CellWidget for actions
    related to multiple selection. In JTable, the corresponding action's enablement state is used
    to implement a similar exception for space (BasicTableUI.Actions.isEnabled). It seems better to
    work at the InputMap level than at the Action level. */
    if (keyStroke.getKeyChar() != KeyEvent.CHAR_UNDEFINED)
      return false;
    /* All relevant CellWidget actions should be in WHEN_ANCESTOR_OF_FOCUSED_COMPONENT, as this is
    the condition under which keystrokes unhandled by an inplace editor would propagate to the
    parent CellWidget. Don't check WHEN_FOCUSED, as this would cause the Enter key to behave
    inappropriately. */
    final Object cellWidgetActionValue =
        cellWidget.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT).get(keyStroke);
    final boolean ret = cellWidgetActionValue != null && (extended
        ? ALWAYS_EXCLUDE_CWA.contains(cellWidgetActionValue)
        : !NEVER_EXCLUDE_CWA.contains(cellWidgetActionValue));
    if (!ret && cellWidgetActionValue != null && editorComponent instanceof CaretInfoProvider) {
      final Action action = cellWidget.getActionMap().get(cellWidgetActionValue);
      if (action instanceof PhysicalCursorMovementAction) {
        final Direction direction = ((PhysicalCursorMovementAction) action).getDirection();
        final CaretInfoProvider caretInfoProvider = (CaretInfoProvider) editorComponent;
        if (direction == Direction.UP && caretInfoProvider.isTextCaretAtEdge(true) ||
            direction == Direction.DOWN && caretInfoProvider.isTextCaretAtEdge(false))
        {
          /* Exclude up/down cursor actions (e.g. arrow up/down, page up/down) when the text caret
          is on the first or last visual line, respectively. */
          return true;
        }
      }
    }
    return ret;
  }
}
