// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.sieuferd.uiutil.api.KeyStrokeForwardable;
import java.awt.event.KeyEvent;
import javax.annotation.Nullable;
import javax.swing.JComponent;
import javax.swing.KeyStroke;

/* TODO[pd]: Compare and contrast with javax.swing.table.TableCellEditor (and its super-interface
             CellEditor), and make sure I didn't miss anything. */
/**
 * Interface for providing a component to be inserted onto a {@link CellWidget} during editing of
 * cell contents. Analogous to {@link javax.swing.table.TableCellEditor} and its super-interface
 * {@link javax.swing.CellEditor}. Unlike {@code TableCellEditor}, this interface assumes that
 * knowledge of what cell is being edited is tracked separately, namely from the
 * {@code CellWidget}'s {@link com.sieuferd.cell.support.SelectionModel}. Furthermore, rather than
 * having the {@code CellWidget} attach {@code CellEditorListener}s to the {@code InplaceEditor},
 * all inplace editing actions (starting, accepting, or canceling an edit) must be initiated by or
 * invoked via the {@code CellWidget}.
 */
public interface InplaceEditor extends KeyStrokeForwardable {
  /**
   * Get the component to be installed when inplace editing is started. Must always return the same
   * component.
   */
  public JComponent getComponent();
  /**
   * Get the component that should be used for focus requests on this inplace editor. Must be the
   * same component as, or a descendant of, the component returned by {@link #getComponent()}. Must
   * always return the same component.
   */
  public JComponent getFocusComponent();

  /**
   * Determine whether an inplace editing session should be started.
   *
   * <p>Any provided {@code KeyEvent} should be examined to determine whether this event should
   * start an inplace edit. Relevant events are those which would start some kind of input in the
   * editor component if received under {@code JComponent.WHEN_FOCUSED}, and which at the same time
   * are unlikely to collide with global or local keyboard shortcuts. Text editors, as an example,
   * should accept events representing simple insertion of a character.
   *
   * <p>For read-only cells, this method should always return false. If {@code evt} is null, this
   * method should always return true except if the cell is read-only.
   *
   * <p>The editor component (the component returned by {@link #getComponent()}) should not be
   * assumed to have been installed at the time when this method is called. When the editor
   * component is installed, its {@code InputMap} will be modified to exclude any mappings that
   * exist in the containing {@code CellWidget}, after which all {@code KeyEvent}s will be passed
   * unconditionally to the editor component. Thus, {@code shouldStartEditing} does not need to
   * accept every input that might be meaningful once editing has started.
   *
   * <p>This method will be called for {@code KeyEvent}s of all types, that is, {@code KEY_PRESSED},
   * {@code KEY_TYPED}, and {@code KEY_RELEASED}. Each event will generate only one call to this
   * method.
   */
  public boolean shouldStartEditing(@Nullable KeyEvent evt);

  /**
   * Forward the specified keystroke to the component returned by {@link #getFocusComponent()}.
   */
  @Override
  public abstract boolean forwardUnhandledKeyStroke(KeyStroke ks, KeyEvent evt, boolean pressed);

  /**
   * Prepare the component for editing. This will be called <em>after</em> the editor has been made
   * active using {@link #setActive(boolean)}.
   *
   * @param clear whether the editor should start in a cleared state
   */
  public void prepare(boolean clear);

  /**
   * Inform the editor that it has been activated or deactivated. The deactivation of an individual
   * editor does not imply that the edit should be accepted or canceled; for this, there is
   * {@link InteractionProvider#acceptEdit()} and {@link InteractionProvider#cancelEdit()}.
   */
  public void setActive(boolean active);
}
