// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.sieuferd.uiutil.api.PopupLike;
import javax.annotation.Nullable;
import javax.swing.JComponent;

/**
 * Interface for providing various components that can be opened up in response to user actions in
 * the {@link CellWidget}.
 *
 * <p>There is no requirement that methods in this interface return the same object every time,
 * although the caller may provide optimizations for this case. Each {@code InteractionProvider}
 * must be used with only one {@code CellWidget}.
 */
public class InteractionProvider {
  protected InteractionProvider() { }

  public static final InteractionProvider createDefault() {
    // Create a new one every time since the spec technically requires this.
    return new InteractionProvider();
  }

  /* This interface is used in place of Swing's own component popup concept
  (JComponent.setComponentPopupMenu). The latter does not allow for the popup menu to be
  context-sensitive with respect to selections made in the component itself, and furthermore
  prevents mouse events that triggers the popup menu from first making any appropriate selection
  changes. Manual popup menu handling seems to be standard throughout the NetBeans source code, e.g.
  in org.openide.explorer.view.TreeView. */
  /**
   * Get a popup menu to be dispayed in response to a right-click or analogous mouse gesture. This
   * method should be called <em>after</em> the caller has processed any selection changes that
   * might result from the action triggering the popup menu.
   *
   */
  public @Nullable PopupLike getMousePopup(int modifiersEx) {
    return null;
  }

  /**
   * Get a component suitable for displaying next to the cell cursor for indication purposes.
   * The component will be sized and positioned by the caller, taking into account the component's
   * preferred size. This method should be called again every time the tooltip is showed for a new
   * {@code CursorLocation}.
   */
  public @Nullable JComponent getCursorTooltip() {
    return null;
  }

  /**
   * Get an InplaceEditor for the current {@code CursorLocation}. The component will be sized and
   * positioned by the caller, taking into account the component's preferred size. This method
   * should be called <em>after</em> the caller has processed any selection changes that might
   * result from the action triggering the opening of the editor.
   */
  public @Nullable InplaceEditor getInplaceEditor() {
    return null;
  }

  /**
   * Attempt to accept any edit currently made in the context of the {@code CellWidget}'s selection.
   * Such an edit may be made through an {@link InplaceEditor}, or an external component such as a
   * formula bar. The callee may accept the edit immediately, or bring up a dialog box to inform the
   * user of input validation errors or to suggest a correction. If a dialog box is displayed, the
   * return value should be either {@link Acceptance#REJECTED} or
   * {@link Acceptance#ACCEPTED_WITH_INTERACTION}.
   *
   * <p>If {@link Acceptance#REJECTED} is returned, the callee is responsible for transferring the
   * keyboard focus back to the appropriate editor.
   *
   * <p>If no edit is actually in progress, this method should return
   * {@link Acceptance#ACCEPTED_IMMEDIATELY}.
   */
  public Acceptance acceptEdit() {
    return Acceptance.ACCEPTED_IMMEDIATELY;
  }

  /**
   * Cancel any edit currently made in the context of the {@code CellWidget}'s selection. It is not
   * permissible to show any dialog box at this point. If no edit is actually in progress, no action
   * is needed.
   */
  public void cancelEdit() {
  }

  public static enum Acceptance {
    /** The input value was accepted immediately, or there was no inplace edit in progress. */
    ACCEPTED_IMMEDIATELY,
    /**
     * The input value was accepted after first consulting the user. If the acceptance process was
     * started due to another action, such as attempting to move the cursor, the original action
     * should be suppressed, since the user might have forgotten about their original intent by now.
     */
    ACCEPTED_WITH_INTERACTION,
    /** The input value was rejected, and editing should continue so the user can try again. */
    REJECTED
  }
}
