// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.widget;

import com.google.common.base.Preconditions;
import com.sieuferd.cell.support.CellRectangles;
import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.Location;
import com.sieuferd.cell.support.PixelRectangle;
import com.sieuferd.cell.support.Pixels;
import com.sieuferd.util.MapUtil;
import java.awt.Rectangle;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;
import javax.annotation.Nullable;

final class RectangleFinder {
  /* TODO[pd]: The cache probably gets wiped pretty quickly in reference selection mode, evicting
               the rectangle of the original cell cursor. Consider if this is worth doing something
               about. */
  // Only one or two cache slots would probably be in actual use here.
  private static final int SIZE = 10;
  private @Nullable GeneratedCell generatedCell;

  /**
   * LRU cache of known recent cursor rectangles. For the few cases where searching for a cursor
   * rectangle can take a bit of time, such as on large layouts where the cursor is far away from
   * the viewport, the cursor rectangle would usually have been computed previously in any case,
   * when the cursor was first moved to its current position. The cache takes advantage of this
   * without adding extra state or logic to BasicCellWidgetUI.
   *
   * <p>The cache only tracks cursor locations computed with a null {@code frozenHeaderPoint}.
   */
  private final Map<Location,Optional<PixelRectangle>> cache =
      new LinkedHashMap<>(SIZE, 0.75f, true);

  public void setGeneratedCell(@Nullable GeneratedCell generatedCell) {
    this.generatedCell = generatedCell;
    cache.clear();
  }

  public @Nullable PixelRectangle findCursorRectangle(
      @Nullable Location location, Coordinates approximateCoordinates)
  {
    Preconditions.checkNotNull(approximateCoordinates);
    if (location == null)
      return null;
    final Optional<PixelRectangle> ret = cache.get(location);
    if (ret != null)
      return ret.orElse(null);
    return findRectangles(location, approximateCoordinates, null).getCursorRectangle();
  }

  public CellRectangles findRectangles(@Nullable Location location,
      Coordinates approximateCoordinates, @Nullable Coordinates frozenHeaderPoint)
  {
    Preconditions.checkNotNull(approximateCoordinates);
    if (location == null || generatedCell == null)
      return new CellRectangles();
    CellRectangles ret =
        findRectanglesUncached(generatedCell, location, approximateCoordinates, frozenHeaderPoint);
    if (frozenHeaderPoint == null && !cache.containsKey(location)) {
      if (cache.size() == SIZE) {
        // Remove least recently used element (first in iteration order).
        Iterator<Map.Entry<Location, Optional<PixelRectangle>>> iterator =
            cache.entrySet().iterator();
        iterator.next();
        iterator.remove();
      }
      MapUtil.putChecked(cache, location, Optional.ofNullable(ret.getCursorRectangle()));
    }
    return ret;
  }

  private static CellRectangles findRectanglesUncached(
      GeneratedCell generatedCell, Location location, Coordinates approximateCoordinates,
      @Nullable Coordinates frozenHeaderPoint)
  {
    Preconditions.checkNotNull(generatedCell);
    Preconditions.checkNotNull(approximateCoordinates);
    Preconditions.checkNotNull(location);
    final Rectangle generatedCellBounds =
        new PixelRectangle(Coordinates.ZERO, generatedCell.getActualDimensions())
        .getBoundingAWTRectangle();
    if (generatedCellBounds.isEmpty())
      return new CellRectangles();
    /* In most cases, the cursor rectangle should be found during the first iteration. If the
    rectangle is far from the approximate position (e.g. due to cursor bias effects), more
    iterations may be necessary. */
    Coordinates quadrant;
    {
      final Pixels firstQuadrantSize =
          Pixels.create((long) (generatedCell.getRenderContext().getScaling() * 400.0))
          .max(Pixels.create(10));
      quadrant = new Coordinates(firstQuadrantSize, firstQuadrantSize);
    }
    PixelRectangle searchBounds;
    do {
      searchBounds = new PixelRectangle(
          approximateCoordinates.minus(quadrant), quadrant.plus(quadrant).plus(quadrant));
      CellRectangles ret = generatedCell.getCell().findRectangles(
          generatedCell.getRenderContext(), location, searchBounds, frozenHeaderPoint);
      if (ret.getCursorRectangle() != null)
        return ret;
      quadrant = new Coordinates(quadrant.getX().times(3), quadrant.getY().times(3));
    } while (!searchBounds.getBoundingAWTRectangle().contains(generatedCellBounds));
    return new CellRectangles();
  }
}
