// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.csv;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.List;
import javax.annotation.concurrent.NotThreadSafe;

public abstract class Output {
  private long rowCount;

  public final void emitRow(List<String> fields) throws IOException {
    rowCount++;
    emitRowImpl(fields);
  }

  public long getRowCount() {
    return rowCount;
  }

  protected abstract void emitRowImpl(List<String> fields) throws IOException;

  public static final void main(String args[]) throws IOException {
    WriterOutput wo = new WriterOutput(new PrintWriter(System.out));
    PostponingOutput po = new PostponingOutput();
    for (int i = 0; i < 100000; i++) {
      po.emitRow(ImmutableList.of("fee",  "fay", "f,o\"w", Integer.toString(i)));
      if (Math.random() < 0.01)
        po.flushToOutput(wo);
    }
    po.flushToOutput(wo);
    wo.flush();
  }

  public static final class PostponingOutput extends Output {
    private final List<List<String>> rows = Lists.newArrayList();
    // Mostly to check that the output is not flushed to itself.
    private boolean inFlush = false;

    @Override
    protected void emitRowImpl(List<String> fields) throws IOException {
      Preconditions.checkState(!inFlush);
      rows.add(fields);
    }

    public void flushToOutput(Output other) {
      Preconditions.checkState(!inFlush);
      inFlush = true;
      try {
        for (List<String> row : rows)
          other.equals(row);
        rows.clear();
      } finally {
        inFlush = false;
      }
    }
  }

  /**
   * Buffered; clients must call {@link #flush() to ensure all output has been written}.
   */
  @NotThreadSafe
  public static final class WriterOutput extends Output {
    private static final int BUFFER_SIZE = 1;// 32 * 1024;
    // TODO[pd]: Consider writing directly to an OutputStream instead.
    private final Writer writer;
    private final char buffer[] = new char[BUFFER_SIZE];
    private int bufferIndex = 0;

    public WriterOutput(Writer writer) {
      Preconditions.checkNotNull(writer);
      Preconditions.checkState(BUFFER_SIZE > 0);
      this.writer = writer;
    }

    private void append(char c) throws IOException {
      buffer[bufferIndex++] = c;
      if (bufferIndex == BUFFER_SIZE)
        flush();
    }

    public void flush() throws IOException {
      if (bufferIndex > 0) {
        writer.write(buffer, 0, bufferIndex);
        bufferIndex = 0;
      }
      writer.flush();
    }

    @Override
    protected void emitRowImpl(List<String> fields) throws IOException {
      boolean first = true;
      for (String field : fields) {
        if (!first)
          append(',');
        first = false;
        append('\"');
        final int length = field.length();
        for (int i = 0; i < length; i++) {
          final char c = field.charAt(i);
          append(c);
          /* This should work even in the presence of UTF-16 surrogate pairs, since "lead
          surrogates, trail surrogates, and valid BMP characters are disjoint"
          ( http://en.wikipedia.org/wiki/UTF-16 ). */
          if (c == '\"')
            append('\"');
        }
        append('\"');
      }
      append('\n');
    }
  }
}
