// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.harness;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.annotation.Nullable;
import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;

public class ConsoleDataCollector {
  private static final File DIR = new File("console_logs");
  private static final File FILE = new File(DIR, "all.txt");

  public static final void main(String args[]) throws IOException {
    //System.out.println("File\tmQT\tmLT\tQL\tcanc\tN\tdropped");
    for (File file : DIR.listFiles())
      processFile(file);
    //processFile(FILE);
  }

  private static void processFile(File file) throws IOException {
    DataCollectionState state = new DataCollectionState();
    try (BufferedReader br = new BufferedReader(new FileReader(file))) {
      String line;
      while ((line = br.readLine()) != null) {
        state.processLine(line);
      }
    }
    state.flushAndClear();
    System.out.print(file.getName().replace(".txt", "") + " & ");
    System.out.print(form(state.summary.queryTimeStats.getPercentile(50)) + " & ");
    System.out.print(form(state.summary.layoutTimeStats.getPercentile(50)) + " & ");
    /*System.out.print(
        form((100.0 * state.summary.queryLongerCount) / state.summary.countingQueries) + "\\% & ");*/
    /*System.out.print(state.summary.cancelCount + " & ");*/
    System.out.print(state.summary.countingQueries + "\n");
    /*System.out.print(state.summary.droppedCount + "\n");*/
  }

  private static String form(double value) {
    return new BigDecimal(value).setScale(0, RoundingMode.HALF_UP).toString();
  }

  private static final Pattern TOOK_PATTERN =
      Pattern.compile("INFO \\[com.sieuferd.util.Stopwatch\\]: .* took ([0-9][,0-9]*)ms");

  private static long parseTook(String line) {
    Matcher m = TOOK_PATTERN.matcher(line);
    if (!m.matches())
      throw new AssertionError();
    String group = m.group(1);
    return Long.parseLong(group.replace(",", ""));
  }

  private static final class SummaryState {
    final boolean useWOM = false;
    final DescriptiveStatistics layoutTimeStats = new DescriptiveStatistics();
    final DescriptiveStatistics queryTimeStats = new DescriptiveStatistics();
    int countingQueries;
    int cancelCount;
    int queryLongerCount;
    int droppedCount;

    public void addDataPoint(long queryTime, long layoutTime, long layoutTimeWOM) {
      addDataPoint(queryTime, useWOM ? layoutTimeWOM : layoutTime);
    }

    public void addDataPoint(long queryTime, long layoutTime) {
      countingQueries++;
      if (queryTime > layoutTime)
        queryLongerCount++;
      //System.out.println(queryTime + "\t" + layoutTime);
      queryTimeStats.addValue(queryTime);
      layoutTimeStats.addValue(layoutTime);
    }

    public void addDropped() {
      droppedCount++;
    }

    public void addCanceled() {
      cancelCount++;
    }
  }

  // All times in milliseconds.
  private static final class DataCollectionState {
    final SummaryState summary = new SummaryState();
    @Nullable Long queryPrepTime;
    @Nullable Long queryExecTime;
    long totalMeasureTime;
    long maxAutoLayoutTime;
    long maxLayoutCreateTime;

    public DataCollectionState() {
      flushAndClear();
    }

    public void flushAndClear() {
      if (queryPrepTime == null || queryExecTime == null ||
          maxAutoLayoutTime == Long.MIN_VALUE ||
          maxLayoutCreateTime == Long.MIN_VALUE)
      {
        // This case happens when filter are used, for instance.
        //System.out.println("(no layout!)");
        // Don't reset the measure time.
        queryPrepTime = null;
        queryExecTime = null;
        maxAutoLayoutTime = Long.MIN_VALUE;
        maxLayoutCreateTime = Long.MIN_VALUE;
        summary.addDropped();
        return;
      }

      long queryTime = queryExecTime + queryPrepTime;
      long layoutTime = maxAutoLayoutTime + maxLayoutCreateTime + totalMeasureTime;

      /*String queryTime = (queryExecTime == null) ? "" : ("" + (queryPrepTime + queryExecTime));
      String layoutTime =
          (maxAutoLayoutTime == Long.MIN_VALUE || maxLayoutCreateTime == Long.MIN_VALUE)
          ? "" : ("" + (maxAutoLayoutTime + maxLayoutCreateTime));*/
      summary.addDataPoint(queryTime, layoutTime, maxAutoLayoutTime + maxLayoutCreateTime);

      queryPrepTime = null;
      queryExecTime = null;
      totalMeasureTime = 0;
      maxAutoLayoutTime = Long.MIN_VALUE;
      maxLayoutCreateTime = Long.MIN_VALUE;
    }

    void processLine(String line) {
      if (line.contains("Query preparation took")) {
        flushAndClear();
        queryPrepTime = parseTook(line);
        //System.out.println("For line " + line + " ****** " + parseTook(line));
      } else if (line.contains("Query execution took")) {
        if (queryPrepTime == null)
          return;
        if (queryExecTime != null)
          throw new AssertionError();
        queryExecTime = parseTook(line);
      } else if (line.contains("Layout measurements took")) {
        totalMeasureTime += parseTook(line);
      } else if (line.contains("Auto-style took")) {
        maxAutoLayoutTime = Math.max(maxAutoLayoutTime, parseTook(line));
      } else if (line.contains("Layout creation took")) {
        maxLayoutCreateTime = Math.max(maxLayoutCreateTime, parseTook(line));
      } else if (line.contains("Attempting to cancel query")) {
        // Invalidate the next "Query execution took".
        queryPrepTime = null;
        //System.out.println("(canceled)");
        summary.addCanceled();
      }
    }
  }
}
