// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.harness;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.database.SQLColumnInfo;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLFragmentBuilder;
import com.sieuferd.database.SQLInserter;
import com.sieuferd.database.SQLTableInfo;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.query.SQLTable;
import com.sieuferd.util.MapUtil;
import java.sql.Clob;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class CopyDriver {
  private static final int INPUT_FETCH_SIZE = 1000;
  private static final int OUTPUT_BATCH_SIZE = 1000;

  public static final void main(String args[]) throws SQLException {
    try (
      SQLConnection input = new SQLConnection(DriverManager.getConnection(
          "jdbc:mysql://127.0.0.1:3306/princeton", "duser", "dpass"));
      SQLConnection output = new SQLConnection(DriverManager.getConnection(
          "jdbc:postgresql://127.0.0.1:5432/demo", "duser", "dpass")))
    {
      try {
        processDatabase(
            input, new SQLDatabase("database", ""),
            output, new SQLDatabase("public", ""));
      } catch (SQLException e) {
        SQLException chain = e;
        while ((chain = chain.getNextException()) != null) {
          System.err.println("Chained SQL exception: " + chain.getMessage());
        }
        throw e;
      }
    }
  }

  private static void processTable(SQLConnection input, SQLTable inputTable,
      SQLConnection output, SQLDatabase outputDatabase)
      throws SQLException
  {
    System.out.println(inputTable);
    SQLTable outputTable = SQLTable.create(outputDatabase, inputTable.getTableName());
    final List<String> columnList;
    {
      final Map<String,SQLType> createdColumns = Maps.newLinkedHashMap();
      final SQLTableInfo tableInfo = input.getTableInfo(inputTable);
      for (SQLColumnInfo columnInfo : tableInfo.getColumns()) {
        final SQLType type = columnInfo.getType();
        if (type.getTypeCode().getCode() == Types.ROWID)
          continue;
        MapUtil.putChecked(createdColumns, columnInfo.getName(), type);
        System.out.println("  " + columnInfo.getName() + ": type=" + type +
            " in=" + input.getBackendDialect().getTypeName(type) +
            " out=" + output.getBackendDialect().getTypeName(type));
        //System.out.println("  " + columnInfo);
      }
      output.createTable(outputTable, createdColumns, ImmutableSet.of(), false);
      columnList = ImmutableList.copyOf(createdColumns.keySet());
    }
    final String inputQueryString =
        buildQuery(input.getBackendDialect(), inputTable, columnList).getJDBCString();
    final Object[] row = new Object[columnList.size()];
    try (SQLInserter inserter = new SQLInserter(output, outputTable, columnList, OUTPUT_BATCH_SIZE))
    {
      input.getJDBCConnection().setAutoCommit(false);
      try (PreparedStatement inputStatement =
          input.getJDBCConnection().prepareStatement(inputQueryString))
      {
        inputStatement.setFetchSize(INPUT_FETCH_SIZE);
        try (ResultSet rs = inputStatement.executeQuery()) {
          while (rs.next()) {
            for (int colIndex = 0; colIndex < columnList.size(); colIndex++) {
              Object obj = rs.getObject(colIndex + 1);
              if (obj instanceof Clob) {
                Clob clob = (Clob) obj;
                try {
                  final int CLOB_LIMIT = 1024 * 256;
                  if (clob.length() > CLOB_LIMIT) {
                    System.out.println(
                        "Warning: Large clob truncated (" + CLOB_LIMIT + " characters total)");
                  }
                  obj = clob.getSubString(colIndex, (int) Math.min(CLOB_LIMIT, clob.length()));
                } finally {
                  clob.free();
                }
              }
              // Fall-through, not else-if.
              if (obj instanceof String) {
                /* Avoid the 'ERROR: invalid byte sequence for encoding "UTF8": 0x00' error on
                PostgreSQL. */
                obj = ((String) obj).replace("\u0000", " ");
              }
              row[colIndex] = obj;
            }
            inserter.insert(row);
          }
        }
      }
    }
  }

  private static SQLFragment buildQuery(
      BackendDialect dialect, SQLTable table, List<String> columnNames)
      throws SQLException
  {
    Preconditions.checkArgument(!columnNames.isEmpty());
    SQLFragmentBuilder ret = SQLFragment.builder();
    ret.appendJDBCString("SELECT ");
    List<SQLFragment> columnFragments = Lists.newArrayList();
    for (String columnName : columnNames)
      columnFragments.add(dialect.quoteIdentifier(columnName, true));
    ret.appendSQLFragment(SQLFragment.join(", ", columnFragments));
    ret.appendJDBCString(" FROM ");
    ret.appendSQLFragment(dialect.getTableIdentifier(table));
    return ret.build();
  }

  private static boolean isTableEmpty(SQLConnection conn, SQLTable table) throws SQLException {
    SQLFragmentBuilder qb = SQLFragment.builder();
    qb.appendJDBCString("SELECT 1 FROM ");
    qb.appendSQLFragment(conn.getBackendDialect().getTableIdentifier(table));
    // TODO[cc]: Use the dialect's limit decorator here.
    qb.appendJDBCString(" LIMIT 1");
    SQLFragment query = qb.build();
    try (PreparedStatement stmt =
        conn.getJDBCConnection().prepareStatement(query.getJDBCString())) {
      try (ResultSet rs = stmt.executeQuery()) {
        return !rs.next();
      }
    }
  }

  private static final ImmutableSet<String> SKIP = ImmutableSet.<String>builder()
      .add("SOME_TABLE")
      .build();

  private static void processDatabase(
      SQLConnection input, SQLDatabase inputDatabase,
      SQLConnection output, SQLDatabase outputDatabase)
      throws SQLException
  {
    // Delete existing tables in the output schema.
    Set<String> existingTables = Sets.newLinkedHashSet();
    for (SQLTable outputTable : output.getTables(outputDatabase)) {
      if (isTableEmpty(output, outputTable)) {
        output.drop(outputTable, false, true, true);
      } else {
        System.out.println("NOT EMPTY: " + outputTable);
        existingTables.add(outputTable.getTableName());
      }
    }
    for (SQLTable inputTable : input.getTables(inputDatabase)) {
      String tn = inputTable.getTableName();
      if (existingTables.contains(tn) || SKIP.contains(tn))
        continue;
      processTable(input, inputTable, output, outputDatabase);
    }
  }
}
