// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.harness;

import com.google.common.base.Preconditions;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.util.IterableUtil;
import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Collection;
import javax.annotation.Nullable;

public final class JSONPrinter {
  private static final String INDENT = "  ";
  private JSONPrinter() { }

  private static void writeIndentation(Writer writer, @Nullable Integer level) throws IOException {
    if (level != null) {
      writer.append('\n');
      for (int i = 0; i < level; i++)
        writer.append(INDENT);
    }
  }

  private static void writeStringAsJSON(Writer writer, String s) throws IOException {
    writer.append(Expressions.generateStringLiteral(s, true));
  }

  private static void writePrimitiveAsJSONLiteral(Writer writer, @Nullable Object javaObject)
      throws IOException
  {
    Preconditions.checkArgument(javaObject == null || !javaObject.getClass().isArray());
    Preconditions.checkArgument(!(javaObject instanceof Collection));
    if (javaObject == null) {
      writer.append("null");
    } else if (
           (javaObject instanceof Boolean)
        || (javaObject instanceof BigDecimal)
        || (javaObject instanceof BigInteger)
        || (javaObject instanceof Integer)
        || (javaObject instanceof Long)
        || (javaObject instanceof Short)
        || (javaObject instanceof Byte)
        || (javaObject instanceof Float) && Float.isFinite((Float) javaObject)
        || (javaObject instanceof Double) && Double.isFinite((Double) javaObject))
    {
      /* Don't try to strip trailing zeroes; for classes like BigInteger, they denote meaningful
      information that can be validly encoded in the JSON syntax. */
      writer.append(javaObject.toString());
    } else {
      /* Includes the case where javaObject is actually a String, as well as the NaN and Infinity
      cases of floating-point values. */
      writeStringAsJSON(writer, javaObject.toString());
    }
  }

  private static void writeTupleAsJSONObject(Writer writer, View view, Data data, RelationPath path,
      Tuple tuple, @Nullable Integer indentLevel)
      throws IOException
  {
    final boolean pretty = (indentLevel != null);
    writer.append('{');
    boolean first = true;
    for (Path childPath : view.getChildPaths(path)) {
      // TODO[pd]: In an initial pass over the view, throw an error if there are duplicate labels.
      if (data.isProjected(childPath)) {
        if (!first) {
          writer.append(',');
          if (pretty)
            writer.append(' ');
        }
        writeStringAsJSON(writer, view.get(CProp.LABEL, childPath));
        writer.append(':');
        if (pretty)
          writer.append(' ');
        writeValueAsJSON(writer, view, data, childPath, tuple, indentLevel);
        first = false;
      }
    }
    writer.append('}');
  }

  /**
   * @param indentLevel if not null, pretty print and use the specified indentation level for
   *        additional lines
   */
  private static void writeValueAsJSON(Writer writer, View view, Data data, Path path,
      Tuple parentTuple, @Nullable Integer indentLevel)
      throws IOException
  {
    Preconditions.checkArgument(data.isProjected(path));
    if (path.isRelation()) {
      final TupleInterval childTuples = data.getRelationTuples(path.getRelation(), parentTuple);
      final boolean useArray = view.get(CProp.PLURAL, path.getRelation()) ||
          IterableUtil.hasMoreThanOneItem(childTuples);
      final @Nullable Integer nextIndentLevel = (indentLevel == null) ? null :
          (useArray ? (indentLevel + 1) : indentLevel);
      if (useArray)
        writer.append('[');
      boolean first = true;
      for (Tuple childTuple : childTuples) {
        if (useArray) {
          if (!first)
            writer.append(',');
          writeIndentation(writer, nextIndentLevel);
        }
        writeTupleAsJSONObject(writer, view, data, path.getRelation(), childTuple, nextIndentLevel);
        first = false;
      }
      if (useArray) {
        if (!first)
          writeIndentation(writer, indentLevel);
        writer.append(']');
      } else if (first) {
        writePrimitiveAsJSONLiteral(null, writer);
      }
    } else {
      writePrimitiveAsJSONLiteral(writer,
          data.getTupleField(parentTuple, path.getPrimitive()).getContent());
    }
  }

  public static void writeJSON(Writer writer, View view, Data data, boolean pretty)
    throws IOException
  {
    writeValueAsJSON(writer, view, data, Path.ROOT, Tuple.createRootParent(data),
        pretty ? 0 : null);
  }

  public static String generateJSON(View view, Data data, boolean pretty) {
    StringWriter writer = new StringWriter();
    try {
      writeJSON(writer, view, data, pretty);
    } catch (IOException e) {
      throw new AssertionError("Did not expect IOException from a StringWriter", e);
    }
    return writer.toString();
  }
}
