// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.harness;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import com.sieuferd.cell.styles.PathHighlights;
import com.sieuferd.datamodel.Path;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.util.concurrent.ExecutionException;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.MediaPrintableArea;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.OrientationRequested;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.UIManager;
import com.sieuferd.layout.ViewResultPrinter;
import com.sieuferd.datamodel.ViewResult;
import com.sieuferd.flow.Flow;
import com.sieuferd.flow.SettableFlow;
import com.sieuferd.flow.Flows;
import com.sieuferd.project.util.QueryLabelFormatter;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.PathReferenceException;
import com.sieuferd.datamodel.expressions.PathReferences;
import com.sieuferd.uiutil.api.PopupLike;
import com.sieuferd.uiutil.ThreadUtil;
import com.sieuferd.uiutil.UIUtil;
import com.sieuferd.uiutil.flow.UIModelFlows;
import com.sieuferd.upstream.popup.WindowPopupLike;
import com.sieuferd.util.MapUtil;
import com.sieuferd.widget.CellWidgetActions;
import com.sieuferd.widget.CellWidgetUtil;
import com.sieuferd.widget.GeneratedCell;
import com.sieuferd.widget.InplaceEditor;
import com.sieuferd.widget.InteractionProvider;
import java.awt.Color;
import java.awt.event.KeyEvent;
import java.util.Map;
import java.util.logging.Logger;
import javax.annotation.Nullable;
import javax.print.attribute.PrintRequestAttribute;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JTextArea;
import javax.swing.JToggleButton.ToggleButtonModel;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import org.netbeans.api.annotations.common.StaticResource;
import org.openide.util.Exceptions;

@SuppressWarnings("serial")
public class WidgetFrame extends JFrame {
  private static final Logger LOG = Logger.getLogger(WidgetFrame.class.getName());
  private static final boolean TEST_CURSOR_TOOLTIPS = false;
  private static final boolean TEST_WINDOW_POPUP = true;
  private static final boolean TEST_PATH_HIGHLIGHTING = false;

  @StaticResource
  private static final String RESOURCE_ICON16 = "com/sieuferd/harness/resources/icon16.png";
  @StaticResource
  private static final String RESOURCE_ICON32 = "com/sieuferd/harness/resources/icon32.png";
  @StaticResource
  private static final String RESOURCE_ICON48 = "com/sieuferd/harness/resources/icon48.png";

  public static final double ZOOM_INCREMENT = 0.2;
  private static final double PRINT_SCALING = 0.7;
  private static final boolean PRINT_FIT_WIDTH = false;
  private final PrintRequestAttributeSet printAttributes;
  private final PrinterJob printerJob;
  private final SettableFlow<ViewResult> resultSettable = Flows.newSettable();
  private final ToggleButtonModel schemaOnlyButtonModel = new ToggleButtonModel();
  private final Flow<Boolean> schemaOnlyFlow = UIModelFlows.toggleButton(schemaOnlyButtonModel);
  private final Flow<GeneratedCell> generatedCellFlow;
  private final ChangeListener generatedCellListener = new ChangeListener() {
    @Override
    public void stateChanged(ChangeEvent e) {
      final GeneratedCell generatedCell = generatedCellFlow.get();
      if (generatedCell != null) {
        SwingUtilities.invokeLater(() -> {
          cellWidget.setGeneratedCell(generatedCell);
        });
      }
    }
  };

  private Image getImage(String path) {
    return Toolkit.getDefaultToolkit().getImage(getClass().getClassLoader().getResource(path));
  }

  private boolean addConstantAttribute(
      PrintRequestAttributeSet printAttributes, String className, String constantName)
  {
    try {
      Class<?> clazz = Class.forName(className);
      Object attribute = clazz.getDeclaredField(constantName).get(null);
      printAttributes.add((PrintRequestAttribute) attribute);
      return true;
    } catch (ClassNotFoundException | NoSuchFieldException | SecurityException |
        IllegalArgumentException | IllegalAccessException e)
    { }
    return false;
  }

  private JComponent createTestCursorTooltip() {
    final JLabel ret = new JLabel();
    ret.setText("<html>This is a small test,<br>a test of cursor tooltip components<br></html>");
    ret.setBorder(new EmptyBorder(2, 3, 2, 3));
    ret.setOpaque(true);
    ret.setBackground(java.awt.SystemColor.info);
    return ret;
  }

  /** EDT only. */
  public WidgetFrame() {
    initComponents();
    UIUtil.getSafeInputMap(cellWidget, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT).put(
        KeyStroke.getKeyStroke(KeyEvent.VK_F10, KeyEvent.SHIFT_DOWN_MASK),
        CellWidgetActions.OPEN_POPUP_MENU);

    final JComponent testCursorTooltip = TEST_CURSOR_TOOLTIPS ? createTestCursorTooltip() : null;
    final TestInplaceEditor inplaceEditor = new TestInplaceEditor();

    cellWidget.setInteractionProvider(new InteractionProvider() {
      @Override
      public PopupLike getMousePopup(int modifiersEx) {
        if (TEST_WINDOW_POPUP) {
          return new WindowPopupLike(new JTextArea("I am a text area!"), null, false);
        } else {
          return PopupLike.fromJPopupMenu(testPopupMenu);
        }
      }

      @Override
      public JComponent getCursorTooltip() {
        return testCursorTooltip;
      }

      @Override
      public InplaceEditor getInplaceEditor() {
        return inplaceEditor;
      }

      @Override
      public Acceptance acceptEdit() {
        ThreadUtil.checkOnEDT();
        if (!inplaceEditor.isActive())
          return Acceptance.ACCEPTED_IMMEDIATELY;
        if (inplaceEditor.getText().startsWith("ask")) {
          final Acceptance ret = (JOptionPane.OK_OPTION == JOptionPane.showConfirmDialog(
              null, "Accept edit?", "TestInplaceEditor", JOptionPane.OK_CANCEL_OPTION))
              ? Acceptance.ACCEPTED_WITH_INTERACTION : Acceptance.REJECTED;
          cellWidget.requestFocusInWindow();
          return ret;
        } else {
          System.out.println("Accepted edit");
          return Acceptance.ACCEPTED_IMMEDIATELY;
        }
      }
    });

    generatedCellFlow = CellWidgetUtil.getCellGenerator(cellWidget, resultSettable, schemaOnlyFlow,
        QueryLabelFormatter.INSTANCE, true);
    generatedCellFlow.addChangeListener(generatedCellListener);

    setIconImages(ImmutableList.of(
        getImage(RESOURCE_ICON16), getImage(RESOURCE_ICON32), getImage(RESOURCE_ICON48)));

    printerJob = PrinterJob.getPrinterJob();
    printerJob.setJobName("Nested Table Flow document");
    printAttributes = new HashPrintRequestAttributeSet();

    /* Using the native print dialog makes the "Save to PDF" option available on MacOS.
    DialogTypeSelection inconveniently exists in different locations depending on the Java
    version, so I use some reflection magic to find the right one without introducing a dependency
    on either. */
    if (!(
        addConstantAttribute(printAttributes,
        "javax.print.attribute.standard.DialogTypeSelection", "NATIVE") ||
        addConstantAttribute(printAttributes,
        "sun.print.DialogTypeSelection", "NATIVE")))
    {
      LOG.warning("Could not set DialogTypeSelection.NATIVE");
    }

    // TODO[pd]: Consider removing.
    printAttributes.add(MediaSizeName.NA_LETTER);
    printAttributes.add(OrientationRequested.PORTRAIT);
    // media-printable-area: (12.7,12.7)->(190.5,254.0)mm
    printAttributes.add(new MediaPrintableArea(0.5f, 0.5f, 7.5f, 10f, MediaPrintableArea.INCH));

    // TODO[pd]: Remove.
    setSize(1000, 1000);
    setLocation(10, 15);
    //setExtendedState(JFrame.MAXIMIZED_BOTH);
  }

  private void setLookAndFeel(String lafClassName) {
    try {
      UIManager.setLookAndFeel(lafClassName);
      SwingUtilities.updateComponentTreeUI(this);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public void setResult(@Nullable ViewResult result) {
    resultSettable.set(result);
    printMenuItem.setEnabled(result != null);
    if (TEST_PATH_HIGHLIGHTING && result != null) {
      Map<Path,Color> highlightMap = Maps.newLinkedHashMap();
      Strepco strepco = Strepco.createNormalized(result.getView());
      try {
        MapUtil.putChecked(highlightMap,
            PathReferences.getPathFromTokenString(strepco, "[Meetings\\Beg. Time]"),
            Color.BLUE);
        MapUtil.putChecked(highlightMap,
            PathReferences.getPathFromTokenString(strepco, "[Instructor Assignments]"),
            Color.ORANGE);
        MapUtil.putChecked(highlightMap,
            PathReferences.getPathFromTokenString(strepco, "[Middle]"),
            Color.RED);
      } catch (PathReferenceException e) {
        Exceptions.printStackTrace(e);
      }
      cellWidget.setPathHighlights(new PathHighlights(highlightMap, null));
    }
  }

  /** This method is called from within the constructor to
   * initialize the form.
   * WARNING: Do NOT modify this code. The content of this method is
   * always regenerated by the Form Editor.
   */
  @SuppressWarnings({"unchecked", "rawtypes"})
  // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
  private void initComponents() {
    bindingGroup = new org.jdesktop.beansbinding.BindingGroup();

    lafButtonGroup = new javax.swing.ButtonGroup();
    testPopupMenu = new javax.swing.JPopupMenu();
    popupMenuItem1 = new javax.swing.JMenuItem();
    popupMenuItem2 = new javax.swing.JMenuItem();
    popupMenuItem3 = new javax.swing.JMenuItem();
    scrollPane = new javax.swing.JScrollPane();
    cellWidget = new com.sieuferd.widget.CellWidget();
    focusTraversalTestButton = new javax.swing.JButton();
    menuBar = new javax.swing.JMenuBar();
    fileMenu = new javax.swing.JMenu();
    pageSetupMenuItem = new javax.swing.JMenuItem();
    printMenuItem = new javax.swing.JMenuItem();
    fileMenuSeparator = new javax.swing.JPopupMenu.Separator();
    exitMenuItem = new javax.swing.JMenuItem();
    viewMenu = new javax.swing.JMenu();
    schemaOnly = new javax.swing.JCheckBoxMenuItem();
    viewMenuSeparator1 = new javax.swing.JPopupMenu.Separator();
    alphaPermitted = new javax.swing.JCheckBoxMenuItem();
    forceFractional = new javax.swing.JCheckBoxMenuItem();
    drawCellDebugGraphics = new javax.swing.JCheckBoxMenuItem();
    viewMenuSeparator2 = new javax.swing.JPopupMenu.Separator();
    lafPlatform = new javax.swing.JRadioButtonMenuItem();
    lafMetal = new javax.swing.JRadioButtonMenuItem();
    viewMenuSeparator3 = new javax.swing.JPopupMenu.Separator();
    zoomInItem = new javax.swing.JMenuItem();
    zoomOutItem = new javax.swing.JMenuItem();
    viewMenuSeparator4 = new javax.swing.JPopupMenu.Separator();
    zoomPlatformDefault = new javax.swing.JMenuItem();
    zoomOneToOne = new javax.swing.JMenuItem();
    zoomTwoToOne = new javax.swing.JMenuItem();

    popupMenuItem1.setText("Action 1"); // NOI18N
    testPopupMenu.add(popupMenuItem1);

    popupMenuItem2.setText("Action 2"); // NOI18N
    testPopupMenu.add(popupMenuItem2);

    popupMenuItem3.setText("Action 3"); // NOI18N
    testPopupMenu.add(popupMenuItem3);

    setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
    setTitle("Nested Table Flow");

    scrollPane.setViewport(new com.sieuferd.uiutil.scroll.OverlayViewport());

    org.jdesktop.beansbinding.Binding binding = org.jdesktop.beansbinding.Bindings.createAutoBinding(org.jdesktop.beansbinding.AutoBinding.UpdateStrategy.READ_WRITE, alphaPermitted, org.jdesktop.beansbinding.ELProperty.create("${selected}"), cellWidget, org.jdesktop.beansbinding.BeanProperty.create("alphaPermitted"));
    bindingGroup.addBinding(binding);
    binding = org.jdesktop.beansbinding.Bindings.createAutoBinding(org.jdesktop.beansbinding.AutoBinding.UpdateStrategy.READ_WRITE, drawCellDebugGraphics, org.jdesktop.beansbinding.ELProperty.create("${selected}"), cellWidget, org.jdesktop.beansbinding.BeanProperty.create("drawCellDebugGraphics"));
    bindingGroup.addBinding(binding);
    binding = org.jdesktop.beansbinding.Bindings.createAutoBinding(org.jdesktop.beansbinding.AutoBinding.UpdateStrategy.READ_WRITE, forceFractional, org.jdesktop.beansbinding.ELProperty.create("${selected}"), cellWidget, org.jdesktop.beansbinding.BeanProperty.create("forceFractionalMetrics"));
    bindingGroup.addBinding(binding);

    scrollPane.setViewportView(cellWidget);

    getContentPane().add(scrollPane, java.awt.BorderLayout.CENTER);

    focusTraversalTestButton.setText("Focus Traversal Test"); // NOI18N
    getContentPane().add(focusTraversalTestButton, java.awt.BorderLayout.SOUTH);

    menuBar.setBorder(null);

    fileMenu.setMnemonic('F');
    fileMenu.setText("File");

    pageSetupMenuItem.setMnemonic('u');
    pageSetupMenuItem.setText("Page Setup...");
    pageSetupMenuItem.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        pageSetupMenuItemActionPerformed(evt);
      }
    });
    fileMenu.add(pageSetupMenuItem);

    printMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_P, java.awt.event.InputEvent.CTRL_MASK));
    printMenuItem.setMnemonic('P');
    printMenuItem.setText("Print...");
    printMenuItem.setEnabled(false);
    printMenuItem.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        printMenuItemActionPerformed(evt);
      }
    });
    fileMenu.add(printMenuItem);
    fileMenu.add(fileMenuSeparator);

    exitMenuItem.setMnemonic('x');
    exitMenuItem.setText("Exit");
    exitMenuItem.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        exitMenuItemActionPerformed(evt);
      }
    });
    fileMenu.add(exitMenuItem);

    menuBar.add(fileMenu);

    viewMenu.setMnemonic('V');
    viewMenu.setText("View");
    lafButtonGroup.add(viewMenu);

    schemaOnly.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_L, java.awt.event.InputEvent.CTRL_MASK));
    schemaOnly.setMnemonic('S');
    schemaOnly.setText("Schema-Only Layout");
    schemaOnly.setModel(schemaOnlyButtonModel);
    viewMenu.add(schemaOnly);
    viewMenu.add(viewMenuSeparator1);

    alphaPermitted.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_A, java.awt.event.InputEvent.ALT_MASK | java.awt.event.InputEvent.CTRL_MASK));
    alphaPermitted.setMnemonic('A');
    alphaPermitted.setSelected(true);
    alphaPermitted.setText("Permit Alpha Compositing");
    viewMenu.add(alphaPermitted);

    forceFractional.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F, java.awt.event.InputEvent.ALT_MASK | java.awt.event.InputEvent.CTRL_MASK));
    forceFractional.setMnemonic('F');
    forceFractional.setText("Force Fractional Metrics");
    viewMenu.add(forceFractional);

    drawCellDebugGraphics.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_D, java.awt.event.InputEvent.ALT_MASK | java.awt.event.InputEvent.CTRL_MASK));
    drawCellDebugGraphics.setMnemonic('D');
    drawCellDebugGraphics.setText("Draw Cell Debug Graphics");
    drawCellDebugGraphics.setDisplayedMnemonicIndex(10);
    viewMenu.add(drawCellDebugGraphics);
    viewMenu.add(viewMenuSeparator2);

    lafPlatform.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_P, java.awt.event.InputEvent.ALT_MASK | java.awt.event.InputEvent.CTRL_MASK));
    lafButtonGroup.add(lafPlatform);
    lafPlatform.setText("Platform Default LAF");
    lafPlatform.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        lafPlatformActionPerformed(evt);
      }
    });
    viewMenu.add(lafPlatform);

    lafMetal.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_M, java.awt.event.InputEvent.ALT_MASK | java.awt.event.InputEvent.CTRL_MASK));
    lafButtonGroup.add(lafMetal);
    lafMetal.setText("Metal LAF");
    lafMetal.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        lafMetalActionPerformed(evt);
      }
    });
    viewMenu.add(lafMetal);
    viewMenu.add(viewMenuSeparator3);

    zoomInItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_EQUALS, java.awt.event.InputEvent.CTRL_MASK));
    zoomInItem.setMnemonic('I');
    zoomInItem.setText("Zoom In");
    zoomInItem.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        zoomInItemActionPerformed(evt);
      }
    });
    viewMenu.add(zoomInItem);

    zoomOutItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_MINUS, java.awt.event.InputEvent.CTRL_MASK));
    zoomOutItem.setMnemonic('o');
    zoomOutItem.setText("Zoom Out");
    zoomOutItem.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        zoomOutItemActionPerformed(evt);
      }
    });
    viewMenu.add(zoomOutItem);
    viewMenu.add(viewMenuSeparator4);

    zoomPlatformDefault.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_0, java.awt.event.InputEvent.CTRL_MASK));
    zoomPlatformDefault.setMnemonic('D');
    zoomPlatformDefault.setText("Zoom Platform Default");
    zoomPlatformDefault.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        zoomPlatformDefaultActionPerformed(evt);
      }
    });
    viewMenu.add(zoomPlatformDefault);

    zoomOneToOne.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_1, java.awt.event.InputEvent.CTRL_MASK));
    zoomOneToOne.setMnemonic('1');
    zoomOneToOne.setText("Zoom 1:1 (MacOS default)");
    zoomOneToOne.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        zoomOneToOneActionPerformed(evt);
      }
    });
    viewMenu.add(zoomOneToOne);

    zoomTwoToOne.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_2, java.awt.event.InputEvent.CTRL_MASK));
    zoomTwoToOne.setText("Zoom 2:1 (MacOS Retina default)");
    zoomTwoToOne.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(java.awt.event.ActionEvent evt) {
        zoomTwoToOneActionPerformed(evt);
      }
    });
    viewMenu.add(zoomTwoToOne);

    menuBar.add(viewMenu);

    setJMenuBar(menuBar);

    bindingGroup.bind();

    pack();
  }// </editor-fold>//GEN-END:initComponents

  private void exitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitMenuItemActionPerformed
    dispose();
  }//GEN-LAST:event_exitMenuItemActionPerformed

  private void pageSetupMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_pageSetupMenuItemActionPerformed
    printerJob.pageDialog(printAttributes);
  }//GEN-LAST:event_pageSetupMenuItemActionPerformed

  private void printMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_printMenuItemActionPerformed
    ViewResult result = resultSettable.get();
    Preconditions.checkNotNull(result);
    if (printerJob.printDialog(printAttributes)) {
      ViewResultPrinter pr = new ViewResultPrinter(result, printAttributes, PRINT_SCALING, PRINT_FIT_WIDTH);
      // TODO[pd]: In the NetBeans-based implementation, name the printJob after the View.
      printerJob.setPrintable(pr);
      //FigureGeneration.configurePrinterJob(printerJob, pr);
      new SwingWorker<Void,Void>() {
        @Override
        protected Void doInBackground() throws PrinterException {
          printerJob.print(printAttributes);
          return null;
        }

        @Override
        protected void done() {
          try {
            get();
          } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
          } catch (ExecutionException e) {
            Throwable cause = e.getCause();
            if (cause instanceof RuntimeException) {
              throw (RuntimeException) cause;
            } else if (cause instanceof Error) {
              throw (Error) cause;
            }
            JOptionPane.showMessageDialog(WidgetFrame.this, e.getMessage(), "Print",
                JOptionPane.ERROR_MESSAGE);
          }
          JOptionPane.showMessageDialog(WidgetFrame.this, "Printing succeeded.", "Print",
              JOptionPane.INFORMATION_MESSAGE);
        }
      }.execute();
    }
  }//GEN-LAST:event_printMenuItemActionPerformed

  private void zoomInItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_zoomInItemActionPerformed
    cellWidget.setScaling(cellWidget.getScaling() * (1.0 + ZOOM_INCREMENT));
  }//GEN-LAST:event_zoomInItemActionPerformed

  private void zoomOutItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_zoomOutItemActionPerformed
    cellWidget.setScaling(cellWidget.getScaling() * (1.0 - ZOOM_INCREMENT));
  }//GEN-LAST:event_zoomOutItemActionPerformed

  private void zoomPlatformDefaultActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_zoomPlatformDefaultActionPerformed
    cellWidget.setScaling(1.0);
  }//GEN-LAST:event_zoomPlatformDefaultActionPerformed

  private void zoomOneToOneActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_zoomOneToOneActionPerformed
    cellWidget.setScaling(72.0 / UIUtil.getPixelsPerInch());
  }//GEN-LAST:event_zoomOneToOneActionPerformed

  private void zoomTwoToOneActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_zoomTwoToOneActionPerformed
    cellWidget.setScaling(2 * 72.0 / UIUtil.getPixelsPerInch());
  }//GEN-LAST:event_zoomTwoToOneActionPerformed

  private void lafPlatformActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lafPlatformActionPerformed
    setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
  }//GEN-LAST:event_lafPlatformActionPerformed

  private void lafMetalActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_lafMetalActionPerformed
    setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
  }//GEN-LAST:event_lafMetalActionPerformed

  // Variables declaration - do not modify//GEN-BEGIN:variables
  private javax.swing.JCheckBoxMenuItem alphaPermitted;
  private com.sieuferd.widget.CellWidget cellWidget;
  private javax.swing.JCheckBoxMenuItem drawCellDebugGraphics;
  private javax.swing.JMenuItem exitMenuItem;
  private javax.swing.JMenu fileMenu;
  private javax.swing.JPopupMenu.Separator fileMenuSeparator;
  private javax.swing.JButton focusTraversalTestButton;
  private javax.swing.JCheckBoxMenuItem forceFractional;
  private javax.swing.ButtonGroup lafButtonGroup;
  private javax.swing.JRadioButtonMenuItem lafMetal;
  private javax.swing.JRadioButtonMenuItem lafPlatform;
  private javax.swing.JMenuBar menuBar;
  private javax.swing.JMenuItem pageSetupMenuItem;
  private javax.swing.JMenuItem popupMenuItem1;
  private javax.swing.JMenuItem popupMenuItem2;
  private javax.swing.JMenuItem popupMenuItem3;
  private javax.swing.JMenuItem printMenuItem;
  private javax.swing.JCheckBoxMenuItem schemaOnly;
  private javax.swing.JScrollPane scrollPane;
  private javax.swing.JPopupMenu testPopupMenu;
  private javax.swing.JMenu viewMenu;
  private javax.swing.JPopupMenu.Separator viewMenuSeparator1;
  private javax.swing.JPopupMenu.Separator viewMenuSeparator2;
  private javax.swing.JPopupMenu.Separator viewMenuSeparator3;
  private javax.swing.JPopupMenu.Separator viewMenuSeparator4;
  private javax.swing.JMenuItem zoomInItem;
  private javax.swing.JMenuItem zoomOneToOne;
  private javax.swing.JMenuItem zoomOutItem;
  private javax.swing.JMenuItem zoomPlatformDefault;
  private javax.swing.JMenuItem zoomTwoToOne;
  private org.jdesktop.beansbinding.BindingGroup bindingGroup;
  // End of variables declaration//GEN-END:variables
}
