// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.harness;


import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLInserter;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.query.SQLTable;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.List;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

public final class XMLImporter {
  private static final int OUTPUT_BATCH_SIZE = 1000;
  private static final String DATABASE = "panama";
  private static final String JDBC_URL = "jdbc:postgresql://localhost:5432/" + DATABASE;
  private static final String JDBC_USER = "duser";
  private static final String JDBC_PASS = "";
  private static final String INPUT_FILE = "/Volumes/EXTSSD/enwiki-latest-abstract.xml";

  private static SAXParserFactory saxParserFactory;

  // From XLSXSpreadsheet.
  private static synchronized SAXParser createSAXParser() {
    /* "An application can use the same instance of the factory to obtain one or more instances of
    the SAXParser provided the instance of the factory isn't being used in more than one thread at a
    time." (SAXParserFactory Javadoc) */
    if (saxParserFactory == null) {
      saxParserFactory = SAXParserFactory.newInstance();
      saxParserFactory.setNamespaceAware(true);
    }
    try {
      return saxParserFactory.newSAXParser();
    } catch (ParserConfigurationException | SAXException e) {
      throw new RuntimeException("Unexpected exception while creating SAX parser", e);
    }
  }

  public static final void main(String args[]) throws SAXException, IOException, SQLException {
    try (SQLConnection output = new SQLConnection(DriverManager.getConnection(
          JDBC_URL, JDBC_USER, JDBC_PASS)))
    {
      final SQLTable outputTable = SQLTable.create(new SQLDatabase("public", ""), "wikipedia");
      final List<String> columnList = Lists.newArrayList("title", "abstract");
      try (SQLInserter inserter =
          new SQLInserter(output, outputTable, columnList, OUTPUT_BATCH_SIZE))
      {
        try (InputStream is = new BufferedInputStream(new FileInputStream(INPUT_FILE))) {
          final SAXParser saxParser = createSAXParser();
          final XMLReader xmlReader = saxParser.getXMLReader();
          xmlReader.setContentHandler(new MyContentHandler(inserter));
          xmlReader.parse(new InputSource(is));
        }
      }
    }
  }

  private static final class MyContentHandler extends DefaultHandler {
    private final SQLInserter inserter;
    private static final String REMOVE_TITLE_PREFIX = "Wikipedia: ";
    private static final int MAX_ABSTRACT_LENGTH = 256;
    final StringBuilder sb = new StringBuilder();
    boolean inTitle;
    boolean inAbstract;
    String titleText;
    String abstractText;

    public MyContentHandler(SQLInserter inserter) {
      Preconditions.checkNotNull(inserter);
      this.inserter = inserter;
    }

    private static String prettyTitle(String title) {
      if (!title.startsWith(REMOVE_TITLE_PREFIX))
        throw new AssertionError();
      title = title.substring(REMOVE_TITLE_PREFIX.length());
      return title;
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
      if (qName.equals("doc")) {
        // No action.
      } else if (qName.equals("title")) {
        inTitle = true;
      } else if (qName.equals("abstract")) {
        inAbstract = true;
      }
    }

    int i = 0;

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
      if (qName.equals("title")) {
        inTitle = false;
        titleText = sb.toString();
        sb.setLength(0);
      } else if (qName.equals("abstract")) {
        inAbstract = false;
        abstractText = sb.toString();
        sb.setLength(0);
      } else if (qName.equals("doc")) {
        if (titleText == null || abstractText == null)
          throw new AssertionError();
        titleText = prettyTitle(titleText);
        if (abstractText.length() > MAX_ABSTRACT_LENGTH)
          abstractText = abstractText.substring(0, MAX_ABSTRACT_LENGTH);
        //System.out.println(titleText + "\t" + abstractText);
        try {
          inserter.insert(new Object[] { titleText, abstractText });
          i++;
          if (i % 1000 == 0) {
            System.out.println("Inserted " + i + " rows");
          }
        } catch (SQLException e) {
          while (true) {
            SQLException ne = e.getNextException();
            if (ne == null)
              break;
            ne.printStackTrace();
          }
          throw new SAXException("SQL inserter problem", e);
        }
        titleText = null;
        abstractText = null;
      }
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
      if (inTitle || inAbstract)
        sb.append(ch, start, length);
    }
  }
}
