// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.legacy.driver;

import com.sieuferd.cell.support.Coordinates;
import com.sieuferd.cell.support.RenderContext;
import java.io.FileInputStream;
import java.io.IOException;
import java.sql.DriverManager;
import java.sql.SQLException;
import com.sieuferd.viewset.SieuferdProjectXML;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.Views;
import com.sieuferd.evaluator.CancelHandler;
import com.sieuferd.evaluator.DataProperties;
import com.sieuferd.evaluator.QueryEvaluator;
import com.sieuferd.harness.LoadedXMLDriver;
import com.sieuferd.harness.WidgetMain;
import com.sieuferd.layout.DefaultStyleSheets;
import com.sieuferd.layout.LAProp;
import com.sieuferd.layout.LabelFormatter;
import com.sieuferd.layout.LayoutGenerator;
import com.sieuferd.layout.LayoutMeasurer;
import com.sieuferd.query.QProp;
import com.sieuferd.util.SieuferdXMLException;
import org.apache.commons.math.stat.descriptive.SummaryStatistics;

// TODO[cc]: Move this into Harness.
public class ProfileDriver {
  private static final ViewDefaults VIEW_DEFAULTS = ViewDefaults.builder()
      .set(CProp.VISIBLE, true)
      .set(CProp.LABEL, "")
      .setAll(QProp.DEFAULTS)
      .setAll(LAProp.DEFAULTS)
      .setAll(DefaultStyleSheets.GRAYSCALE)
      .build();

  private static final long TIMING_RUNS = 5;
  private static final long TIMING_DRY_RUNS = 3;

  public static double getUnbiasedStandardDeviation(SummaryStatistics summaryStatistics) {
    long N = summaryStatistics.getN();
    return summaryStatistics.getStandardDeviation() * Math.sqrt(N / (N - 1.0));
  }

  public static double getStandardErrorOfTheMean(SummaryStatistics summaryStatistics) {
    long N = summaryStatistics.getN();
    return getUnbiasedStandardDeviation(summaryStatistics) / Math.sqrt(N);
  }

  public static void main(String args[])
      throws IOException, SieuferdXMLException, SQLException, InterruptedException
  {
    MutableView originalView = Views.newMutable(SieuferdProjectXML.readSieuferdViewXML(
        new FileInputStream(LoadedXMLDriver.QUERY_XML_FILE),
        LoadedXMLDriver.QUERY_XML_FILE.toURI().toString(), VIEW_DEFAULTS.getProperties()),
        VIEW_DEFAULTS);

    try (SQLConnection connection = new SQLConnection(DriverManager.getConnection(
        "jdbc:postgresql://127.0.0.1:5432/princeton", "duser", "dpass"))) {
      //System.out.println(data);
      //System.out.println(data.getView());

      final RenderContext rcx = DataCollection.getStandardizedRenderContext();
      final Coordinates pageSize = DataCollection.getStandardizedPageSize();
      SummaryStatistics queryPhaseStats        = new SummaryStatistics();
      SummaryStatistics measurePhaseStats      = new SummaryStatistics();
      SummaryStatistics autoStylePhaseStats    = new SummaryStatistics();
      SummaryStatistics createLayoutPhaseStats = new SummaryStatistics();

      long startTime;
      for (int i = 0; i < TIMING_RUNS + TIMING_DRY_RUNS; i++) {
        MutableView view = Views.newMutable(originalView);
        final ImmutableData data;
        startTime = System.nanoTime();
        final QueryEvaluator evaluator =
            QueryEvaluator.create(view, connection, null, new CancelHandler());
        DataProperties.configureDataProperties(view);
        try {
          data = WidgetMain.evaluate(view, evaluator, Integer.MAX_VALUE);
        } finally {
          evaluator.close();
        }
        if (i >= TIMING_DRY_RUNS)
          queryPhaseStats.addValue(System.nanoTime() - startTime);

        startTime = System.nanoTime();

        LayoutMeasurer.measureData(data, view, rcx);
        if (i >= TIMING_DRY_RUNS)
          measurePhaseStats.addValue(System.nanoTime() - startTime);

        startTime = System.nanoTime();
        LayoutGenerator.autoStyle(data, view, rcx, pageSize.getX(), LabelFormatter.SIMPLE);
        if (i >= TIMING_DRY_RUNS)
          autoStylePhaseStats.addValue(System.nanoTime() - startTime);

        startTime = System.nanoTime();
        LayoutGenerator.createLayout(data, ImmutableView.copyOf(view), rcx, pageSize.getX(),
            LabelFormatter.SIMPLE);
        if (i >= TIMING_DRY_RUNS)
          createLayoutPhaseStats.addValue(System.nanoTime() - startTime);
      }
      double total = queryPhaseStats.getMean()
                   + measurePhaseStats.getMean()
                   + autoStylePhaseStats.getMean()
                   + createLayoutPhaseStats.getMean();

      System.out.println(
          queryPhaseStats.getMean() / 1000000.0 + "\t" +
          getStandardErrorOfTheMean(queryPhaseStats) / 1000000.0 + "\t" +
          measurePhaseStats.getMean() / 1000000.0 + "\t" +
          getStandardErrorOfTheMean(measurePhaseStats) / 1000000.0 + "\t" +
          autoStylePhaseStats.getMean() / 1000000.0 + "\t" +
          getStandardErrorOfTheMean(autoStylePhaseStats) / 1000000.0 + "\t" +
          createLayoutPhaseStats.getMean() / 1000000.0 + "\t" +
          getStandardErrorOfTheMean(createLayoutPhaseStats) / 1000000.0);
      System.out.println("Query    : " + queryPhaseStats.getMean() / 1000000.0 + "ms / " +
          100 * queryPhaseStats.getMean() / total);
      System.out.println("Measure  : " + measurePhaseStats.getMean() / 1000000.0  + "ms / " +
          100 * measurePhaseStats.getMean() / total);
      System.out.println("AutoStyle: " + autoStylePhaseStats.getMean() / 1000000.0  + "ms / " +
          100 * autoStylePhaseStats.getMean() / total);
      System.out.println("Layout   : " + createLayoutPhaseStats.getMean() / 1000000.0  + "ms / " +
          100 * createLayoutPhaseStats.getMean() / total);
    }
  }
}
