// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.query.SQLTable;
import com.sieuferd.util.StringUtil;
import java.util.Objects;
import java.util.Set;
import javax.annotation.concurrent.Immutable;

@Immutable
public final class ForeignKeyConstraint {
  private final String name;
  private final SQLTable foreignKeyTable;
  private final SQLTable primaryKeyTable;
  private final ImmutableSet<ColumnConstraint> columnConstraints;

  /** @param name may be empty, but not null */
  public ForeignKeyConstraint(String name, SQLTable foreignKeyTable, SQLTable primaryKeyTable,
      Set<ColumnConstraint> columnConstraints)
  {
    StringUtil.checkArgumentTrimmed(name);
    Preconditions.checkNotNull(foreignKeyTable);
    Preconditions.checkNotNull(primaryKeyTable);
    this.name = name;
    this.foreignKeyTable = foreignKeyTable;
    this.primaryKeyTable = primaryKeyTable;
    this.columnConstraints = ImmutableSet.copyOf(columnConstraints);
  }

  /** @return may be empty, but not null */
  public String getName() {
    return name;
  }

  public SQLTable getForeignKeyTable() {
    return foreignKeyTable;
  }

  public SQLTable getPrimaryKeyTable() {
    return primaryKeyTable;
  }

  public Set<ColumnConstraint> getColumnConstraints() {
    return columnConstraints;
  }

  @Override
  public String toString() {
    return "ForeignKeyConstraint('" +
        name + "', " + foreignKeyTable + ", " + primaryKeyTable + ", " + columnConstraints + ")";
  }

  @Override
  public int hashCode() {
    return Objects.hash(name, foreignKeyTable, primaryKeyTable, columnConstraints);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof ForeignKeyConstraint))
      return false;
    final ForeignKeyConstraint other = (ForeignKeyConstraint) obj;
    return Objects.equals(this.name             , other.name)
        && Objects.equals(this.foreignKeyTable  , other.foreignKeyTable)
        && Objects.equals(this.primaryKeyTable  , other.primaryKeyTable)
        && Objects.equals(this.columnConstraints, other.columnConstraints);
  }

  public static final class ColumnConstraint {
    private final String foreignKeyColumnName;
    private final String primaryKeyColumnName;

    public ColumnConstraint(String foreignKeyColumnName, String primaryKeyColumnName) {
      StringUtil.checkArgumentTrimmed(foreignKeyColumnName);
      StringUtil.checkArgumentTrimmed(primaryKeyColumnName);
      this.foreignKeyColumnName = foreignKeyColumnName;
      this.primaryKeyColumnName = primaryKeyColumnName;
    }

    public String getForeignKeyColumnName() {
      return foreignKeyColumnName;
    }

    public String getPrimaryKeyColumnName() {
      return primaryKeyColumnName;
    }

    @Override
    public String toString() {
      return foreignKeyColumnName + "=" + primaryKeyColumnName;
    }

    @Override
    public int hashCode() {
      return Objects.hash(foreignKeyColumnName, primaryKeyColumnName);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof ColumnConstraint))
        return false;
      ColumnConstraint other = (ColumnConstraint) obj;
      return foreignKeyColumnName.equals(other.foreignKeyColumnName)
          && primaryKeyColumnName.equals(other.primaryKeyColumnName);
    }
  }
}
