// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.google.common.base.Preconditions;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.database.ForeignKeyConstraint.ColumnConstraint;
import com.sieuferd.query.SQLTable;
import com.sieuferd.util.StringUtil;
import java.util.Set;

final class ForeignKeyConstraintsBuilder {
  /**
   * Keys are ForeignKeyConstraint objects with no column constraints set. Nameless constraints are
   * gathered under one ForeignKeyConstraint key with an empty name.
   */
  private final Multimap<ForeignKeyConstraint,ColumnConstraint> constraints =
      ArrayListMultimap.create();
  /** Keys follow the same convention as for {@code constraints}. */
  private final Set<ForeignKeyConstraint> invalidConstraints = Sets.newLinkedHashSet();

  /**
   * @param foreignKeyName may be empty, but not null
   */
  public void addColumnConstraint(String foreignKeyName, int sequenceNumber,
      SQLTable fkTable, SQLTable pkTable, String fkTableColumnName, String pkTableColumnName)
  {
    Preconditions.checkNotNull(fkTable);
    Preconditions.checkNotNull(pkTable);
    StringUtil.checkArgumentTrimmed(foreignKeyName);
    StringUtil.checkArgumentTrimmed(fkTableColumnName);
    StringUtil.checkArgumentTrimmed(pkTableColumnName);
    final ForeignKeyConstraint id = new ForeignKeyConstraint(
        foreignKeyName, fkTable, pkTable, ImmutableSet.of());
    if (foreignKeyName.isEmpty() && sequenceNumber != 1 ||
        fkTableColumnName.isEmpty() || pkTableColumnName.isEmpty())
    {
      invalidConstraints.add(id);
    }
    constraints.put(id, new ColumnConstraint(fkTableColumnName, pkTableColumnName));
  }

  public Set<ForeignKeyConstraint> getRelationships() {
    final Set<ForeignKeyConstraint> ret = Sets.newLinkedHashSet();
    for (ForeignKeyConstraint id : constraints.keySet()) {
      if (invalidConstraints.contains(id))
        continue;
      if (id.getName().isEmpty()) {
        for (ColumnConstraint columnConstraint : constraints.get(id)) {
          ret.add(new ForeignKeyConstraint(id.getName(),
              id.getForeignKeyTable(), id.getPrimaryKeyTable(), ImmutableSet.of(columnConstraint)));
        }
      } else {
        ret.add(new ForeignKeyConstraint(
            id.getName(), id.getForeignKeyTable(), id.getPrimaryKeyTable(),
            ImmutableSet.copyOf(constraints.get(id))));
      }
    }
    return ret;
  }
}
