// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.google.common.base.Preconditions;
import com.sieuferd.query.SQLTable;
import com.sieuferd.util.StringUtil;
import java.sql.Types;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/**
 * Information about a table column. Reflects the metadata provided by
 * {@link java.sql.DatabaseMetaData#getColumns(String,String,String,String)}. Javadoc of this class
 * is based on that of the latter method.
 */
@Immutable
public final class SQLColumnInfo {
  private final SQLTable table;
  private final String name;
  private final SQLType type;
  private final int numPrecRadix;
  private final String remarks;
  private final String defaultValue;
  private final int ordinalPosition;
  private final @Nullable SQLTable referencedTable;
  private final SQLTypeCode sourceDataTypeCode;
  private final @Nullable Boolean autoIncrement;
  private final @Nullable Boolean generated;

  public SQLColumnInfo(SQLTable table, String name, SQLType type, int numPrecRadix, String remarks,
      String defaultValue, int ordinalPosition, @Nullable SQLTable referencedTable,
      SQLTypeCode sourceDataTypeCode, @Nullable Boolean autoIncrement, @Nullable Boolean generated)
  {
    Preconditions.checkNotNull(table);
    StringUtil.checkArgumentTrimmed(name);
    Preconditions.checkArgument(!name.isEmpty());
    Preconditions.checkNotNull(type);
    Preconditions.checkNotNull(sourceDataTypeCode);
    StringUtil.checkArgumentTrimmed(remarks);
    StringUtil.checkArgumentTrimmed(defaultValue);
    this.table = table;
    this.name = name;
    this.type = type;
    this.numPrecRadix = numPrecRadix;
    this.remarks = remarks;
    this.defaultValue = defaultValue;
    this.ordinalPosition = ordinalPosition;
    this.referencedTable = referencedTable;
    this.sourceDataTypeCode = sourceDataTypeCode;
    this.autoIncrement = autoIncrement;
    this.generated = generated;
  }

  public SQLTable getTable() {
    return table;
  }

  public String getName() {
    return name;
  }

  public SQLType getType() {
    return type;
  }

  /** Get the radix (typically either 10 or 2). */
  public int getNumPrecRadix() {
    return numPrecRadix;
  }

  /** Get any comment describing the column. */
  public String getRemarks() {
    return remarks;
  }

  /**
   * Get the default value for the column, which should be interpreted as a string when the value is
   * enclosed in single quotes.
   */
  public String getDefaultValue() {
    // Currently not populated; see SQLConnection.buildColumnInfo.
    throw new UnsupportedOperationException();
    // return defaultValue;
  }

  /** Get the index of column in table (starting at 1). */
  public int getOrdinalPosition() {
    return ordinalPosition;
  }

  /**
   * Get the table that is the scope of a reference attribute. Should be null if {@code dataType}
   * isn't {@link Types#REF}.
   */
  public @Nullable SQLTable getReferencedTable() {
    return referencedTable;
  }

  /**
   * Get the source type of a {@link Types#DISTINCT} type or user-generated {@link Types#REF} type.
   * JDBC type from {@link Types}. Should be {@link SQLTypeCode#NULL} if {@code dataType} isn't
   * {@code DISTINCT} or {@code REF}.
   */
  public SQLTypeCode getSourceDataTypeCode() {
    return sourceDataTypeCode;
  }

  /** Get whether this column is auto-incremented, or null if not known. */
  public @Nullable Boolean getAutoIncrement() {
    return autoIncrement;
  }

  /** Get whether this is a generated column, or null if not known. */
  public @Nullable Boolean getGenerated() {
    return generated;
  }

  @Override
  public String toString() {
    return "SQLColumnInfo(name=" + name +
        ", type=" + type +
        ", autoIncrement=" + autoIncrement +
        ", generated=" + generated +
        ")";
  }
}
