// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.sieuferd.query.SQLDatabase;
import com.sieuferd.query.SQLTable;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.database.dialect.BackendDialect.BeforeOrAfter;
import com.sieuferd.database.dialect.BackendDialectResolver;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.StringUtil;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.SQLRecoverableException;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;

// TODO[pd]: Move logic from SQLConnection to BackendDialect as appropriate.
/** Not thread-safe. However, each public method checks that it is called from the same thread as
the constructor. */
@NotThreadSafe
public final class SQLConnection implements AutoCloseable {
  /* Using the parameters from org.netbeans.modules.db.metadata.model.jdbc.JDBCSchema for
  DatabaseMetaData.getTables() here. */
  private static final ImmutableSet<String> TABLE_TYPE_NAMES =
      ImmutableSet.of("TABLE", "VIEW", "SYSTEM TABLE");
  private static final Logger LOG = Logger.getLogger(SQLConnection.class.getName());
  private final Connection jdbcConnection;
  private final Thread thread;
  private final boolean shieldClose;
  private @Nullable BackendDialect dialect;
  private boolean closed;
  /** True iff an error has invalidated the underlying jdbcConnection. */
  private boolean invalidated;

  public SQLConnection(Connection jdbcConnection, boolean shieldClose) {
    Preconditions.checkNotNull(jdbcConnection);
    this.shieldClose = shieldClose;
    this.jdbcConnection = jdbcConnection;
    thread = Thread.currentThread();
  }

  public SQLConnection(Connection jdbcConnection) {
    this(jdbcConnection, false);
  }

  /** Will fail cleanly if called on the wrong thread. */
  private void checkThread() {
    if (!Thread.currentThread().equals(thread))
      throw new IllegalStateException("SQL operation on incorrect thread");
  }

  /** Will fail cleanly if called on the wrong thread. */
  private void checkOpen() {
    checkThread();
    if (closed)
      throw new IllegalStateException("The SQLConnection was closed");
  }

  /** Will fail cleanly if called on the wrong thread. */
  private void checkValid() throws SQLException {
    checkOpen();
    if (invalidated)
      throw new SQLException("Connection invalid due to a previous error");
  }

  /** Will fail cleanly if called on the wrong thread. */
  public boolean isInvalidated() {
    checkThread();
    return invalidated;
  }

  // TODO[cc]: Rename to getDialect().
  public BackendDialect getBackendDialect() throws SQLException {
    checkOpen();
    if (dialect == null)
      dialect = BackendDialectResolver.INSTANCE.resolveDialect(jdbcConnection);
    return dialect;
  }

  /** Get the underlying JDBC connection. Clients should not attempt to close said connection. */
  public Connection getJDBCConnection() throws SQLException {
    checkValid();
    return jdbcConnection;
  }

  /** Additional error codes taken from com.zaxxer.hikari.proxy.ConnectionProxy. See
  https://github.com/brettwooldridge/HikariCP/blob/dev/hikaricp-common/src/main/java/com/zaxxer/hikari/proxy/ConnectionProxy.java . */
  private static final ImmutableSet<String> OTHER_DISCONNECT_SQLSTATES = ImmutableSet.of(
    "57P01", // PostgreSQL Admin shutdown
    "57P02", // PostgreSQL Crash shutdown
    "57P03", // PostgreSQL Cannot connect now
    "01002", // SQL92 disconnect error
    "JZ0C0", // Sybase disconnect error
    "JZ0C1"  // Sybase disconnect error
  );

  public void reportSQLException(SQLException e) {
    checkOpen();
    if (invalidated)
      return;
    /* This is based loosely on the logic in com.mchange.v2.c3p0.impl.DefaultConnectionTester, and
    similar logic in com.zaxxer.hikari.proxy.ConnectionProxy.checkException(). */
    /* Unlike in a connection pool, there is no benefit to actively attempt to validate the
    connection after every error, as it would hold up the only connection with work that might as
    well be a useful subsequent query instead. Furthermore, it is unnecessary to distinguish between
    problems that would invalidate all connections to the database (e.g. a database restart,
    signified in c3p0 by the status code DATABASE_IS_INVALID) and those that would invalidate only
    the connection that first raised an error (CONNECTION_IS_INVALID in c3p0). Instead, just
    disconnect on all connection-related SQLState codes, see
    http://stackoverflow.com/questions/14404866/how-to-detect-the-sql-error-state-of-sqlexception .
    Also disconnect on a selection of vendor-specific codes, as HikariCP does. Also disconnect on
    any SQLRecoverableException, since its Javadoc states that "the recovery operation must include
    closing the current connection". */
    final String state = e.getSQLState();
    // The getSQLState method has been observed to return null in some cases (e.g. on Oracle).
    if (e instanceof SQLRecoverableException ||
        (state != null && (state.startsWith("08") || OTHER_DISCONNECT_SQLSTATES.contains(state))))
    {
      /* This case is often (always?) seen on Oracle after a statement has been canceled. In this
      case the exception is a SQLRecoverableException with SQLState 08006 and message "IO Error:
      Connection reset" */
      LOG.log(Level.INFO, "Invalidating SQLConnection after {0} with SQLState {1}: {2}",
          new Object[]{ e.getClass(), e.getSQLState(), e.getMessage() });
      invalidated = true;
    } else {
      // See com.zaxxer.hikari.proxy.ConnectionProxy.checkException().
      SQLException nextException = e.getNextException();
      if (nextException != null)
        reportSQLException(nextException);
    }
  }

  private String getDatabaseNamePart(SQLDatabase database, boolean catalog) throws SQLException {
    final boolean isUsingCatalogs = getBackendDialect().isUsingCatalogs() ||
        (!database.getPrimaryNamePart().isEmpty() && !database.getSecondaryNamePart().isEmpty());
    final boolean primaryIsCatalog = isUsingCatalogs && getBackendDialect().isCatalogAtStart();
    return (catalog == primaryIsCatalog) ?
        database.getPrimaryNamePart() : database.getSecondaryNamePart();
  }

  private String toPattern(String name) throws SQLException {
    return getBackendDialect().quoteIdentifierForSearchString(name);
  }

  /**
   * @param catalogName may be empty or null
   * @param schemaName may be empty or null
   */
  private SQLDatabase createSQLDatabase(String catalogName, String schemaName) throws SQLException {
    final String catalogNameTrimmed = StringUtil.trimmedAndEmptyIfNull(catalogName);
    final String schemaNameTrimmed = StringUtil.trimmedAndEmptyIfNull(schemaName);
    if (catalogNameTrimmed.isEmpty()) {
      return new SQLDatabase(schemaNameTrimmed, "");
    } else if (schemaNameTrimmed.isEmpty()) {
      return new SQLDatabase(catalogNameTrimmed, "");
    } else if (getBackendDialect().isCatalogAtStart()) {
      return new SQLDatabase(catalogNameTrimmed, schemaNameTrimmed);
    } else {
      return new SQLDatabase(schemaNameTrimmed, catalogNameTrimmed);
    }
  }

  public Set<SQLTable> getTables(SQLDatabase database) throws SQLException {
    checkValid();
    return getTablesInternal(
        getDatabaseNamePart(database, true), toPattern(getDatabaseNamePart(database, false)), "%",
        true);
  }

  private Set<SQLTable> getTablesInternal(
      String catalogName, String schemaPattern, String tablePattern, boolean filterTableTypes)
      throws SQLException
  {
    final DatabaseMetaData md = jdbcConnection.getMetaData();
    try (ResultSet tableRS = md.getTables(catalogName, schemaPattern, tablePattern,
        filterTableTypes ? TABLE_TYPE_NAMES.toArray(new String[TABLE_TYPE_NAMES.size()]) : null))
    {
      Set<SQLTable> ret = Sets.newLinkedHashSet();
      while (tableRS.next()) {
        /* The Derby JDBC driver is known to return an empty (non-null) catalog name; others may
        return null. createSQLDatabase() accepts either. */
        ret.add(SQLTable.create(createSQLDatabase(
            tableRS.getString("TABLE_CAT"), tableRS.getString("TABLE_SCHEM")),
            tableRS.getString("TABLE_NAME")));
      }
      return Collections.unmodifiableSet(ret);
    }
  }

  public Map<SQLDatabase,Set<SQLTable>> getDatabasesAndTables() throws SQLException {
    checkValid();
    final Map<SQLDatabase,Set<SQLTable>> ret = Maps.newLinkedHashMap();
    final Set<String> catalogNames = Sets.newLinkedHashSet();
    for (SQLDatabase database : getDatabases()) {
      MapUtil.putChecked(ret, database, Sets.newLinkedHashSet());
      catalogNames.add(getDatabaseNamePart(database, true));
    }
    /* Iterate over catalogs, not over databases. In the MIT Data Warehouse example, there is one
    database per user (about 24000), the vast majority of which have no tables. */
    for (String catalogName : catalogNames) {
      LOG.log(Level.INFO, "Retrieving table list for catalog ''{0}''", catalogName);
      for (SQLTable table : getTablesInternal(catalogName, "%", "%", true)) {
        Set<SQLTable> set = ret.get(table.getDatabase());
        if (set == null) {
          /* Odd case, since maps were supposed to be constructed for each database already, but
          handle it anyway (e.g. in case metadata changes between the call to getDatabases and the
          call to getTablesInternal). */
          set = Sets.newLinkedHashSet();
          MapUtil.putChecked(ret, table.getDatabase(), set);
        }
        set.add(table);
      }
    }
    return Collections.unmodifiableMap(ret);
  }

  private static Set<String> getResultSetColumnNames(ResultSet rs) throws SQLException {
    ResultSetMetaData rsmd = rs.getMetaData();
    Set<String> ret = Sets.newLinkedHashSet();
    for (int i = 1; i <= rsmd.getColumnCount(); i++)
      ret.add(rsmd.getColumnLabel(i));
    return Collections.unmodifiableSet(ret);
  }

  public Set<SQLDatabase> getDatabases() throws SQLException {
    checkValid();
    final DatabaseMetaData md = jdbcConnection.getMetaData();
    /* See org.netbeans.modules.db.metadata.model.jdbc.JDBCCatalog/JDBCMetadata and subclasses.
    Said API is not public, so I roll my own simplified implementation here, without any loss of
    generality being intended. */
    Set<SQLDatabase> ret = Sets.newLinkedHashSet();
    // TODO[pd]: Consider consulting the BackendDialect instead here.
    if (md.supportsSchemasInTableDefinitions()) {
      try (ResultSet schemaRS = md.getSchemas()) {
        // The Postgres and Oracle JDBC drivers are known not to include a catalog here.
        final boolean hasCatalog = getResultSetColumnNames(schemaRS).contains("TABLE_CATALOG");
        while (schemaRS.next()) {
          ret.add(createSQLDatabase(
              hasCatalog ? schemaRS.getString("TABLE_CATALOG") : "",
              schemaRS.getString("TABLE_SCHEM")));
        }
      }
    } else {
      // This branch applies at least to the MySQL driver.
      /* JDBC drivers may use either the "catalog" or the "schema" JDBC concept, or a combination of
      them, to model what this API refers to as a SQLDatabase. */
      try (ResultSet catalogRS = md.getCatalogs()) {
        while (catalogRS.next())
          ret.add(createSQLDatabase(catalogRS.getString("TABLE_CAT"), ""));
      }
    }
    return Collections.unmodifiableSet(ret);
  }

  private boolean checkTableMatches(SQLTable requested, SQLTable actual) {
    final boolean ret = actual.equals(requested) ||
        (requested.getDatabase().isEmpty() &&
        requested.getTableName().equals(actual.getTableName()));
    if (!ret) {
      LOG.log(Level.WARNING,
          "Table mismatch when retrieving database metadata (expected {0}, got {1})",
          new Object[] { requested, actual });
    }
    return ret;
  }

  public Set<ForeignKeyConstraint> getForeignKeyConstraints(SQLTable baseTable, boolean forward)
      throws SQLException
  {
    checkValid();
    final String catalogName = getDatabaseNamePart(baseTable.getDatabase(), true);
    final String schemaName = getDatabaseNamePart(baseTable.getDatabase(), false);
    final String tableName = baseTable.getTableName();
    final DatabaseMetaData md = jdbcConnection.getMetaData();
    try (ResultSet rs = forward
        ? md.getImportedKeys(catalogName, schemaName, tableName)
        : md.getExportedKeys(catalogName, schemaName, tableName))
    {
      final ForeignKeyConstraintsBuilder ret = new ForeignKeyConstraintsBuilder();
      while (rs.next()) {
        final SQLTable fkTable = SQLTable.create(createSQLDatabase(
            rs.getString("FKTABLE_CAT"), rs.getString("FKTABLE_SCHEM")),
            rs.getString("FKTABLE_NAME"));
        final SQLTable pkTable = SQLTable.create(createSQLDatabase(
            rs.getString("PKTABLE_CAT"), rs.getString("PKTABLE_SCHEM")),
            rs.getString("PKTABLE_NAME"));
        if (!checkTableMatches(baseTable, forward ? fkTable : pkTable))
          continue;
        final String fkColumnName =
            StringUtil.trimmedAndEmptyIfNull(rs.getString("FKCOLUMN_NAME"));
        final String pkColumnName =
            StringUtil.trimmedAndEmptyIfNull(rs.getString("PKCOLUMN_NAME"));
        ret.addColumnConstraint(StringUtil.trimmedAndEmptyIfNull(rs.getString("FK_NAME")),
            rs.getShort("KEY_SEQ"), fkTable, pkTable, fkColumnName, pkColumnName);
      }
      return ret.getRelationships();
    }
  }

  public @Nonnull SQLTableInfo getTableInfo(SQLTable tab) throws SQLException {
    Preconditions.checkNotNull(tab);
    checkValid();
    final DatabaseMetaData md = jdbcConnection.getMetaData();
    final Map<String,SQLColumnInfo> columns = getColumnsWithInfo(tab);
    if (columns.isEmpty()) {
      final String catalogName = getDatabaseNamePart(tab.getDatabase(), true);
      final String schemaName = getDatabaseNamePart(tab.getDatabase(), false);
      final String tableName = tab.getTableName();
      // Be careful not to filter on specific table types only here.
      if (getTablesInternal(
          catalogName, toPattern(schemaName), toPattern(tableName), false).isEmpty())
      {
        throw new SQLException("Table " + tab + " not found");
      }
    }
    Set<String> primaryKeyColumnNames = getPrimaryKeyColumnNames(md, tab);
    /* The Postgres JDBC driver, as an example, returns an empty set if the primary key is not
    defined for a table. */
    if (primaryKeyColumnNames.isEmpty() && !columns.isEmpty())
      primaryKeyColumnNames = getBestRowIdentifier(md, tab);
    /* At least on Oracle, getBestRowIdentifier may suggest ROWID even for views in which it can not
    be legally used. */
    boolean pseudoColumnSuggested = false;
    for (String columnName : primaryKeyColumnNames) {
      if (getBackendDialect().isRowIdentifierPseudoColumn(columnName)) {
        pseudoColumnSuggested = true;
        break;
      }
    }
    if (primaryKeyColumnNames.isEmpty() && !columns.isEmpty() || pseudoColumnSuggested) {
      final String pseudoColumn = getBackendDialect().getRowIdentifierPseudoColumnName();
      if (pseudoColumn.isEmpty() ||
          (!pseudoColumn.isEmpty() && !isRowIdentifierLegal(tab, pseudoColumn)))
      {
        // Use all columns as a key.
        primaryKeyColumnNames = ImmutableSet.copyOf(columns.keySet());
      } else {
        // Use the pseudocolumn as a key.
        primaryKeyColumnNames = ImmutableSet.of(pseudoColumn);
      }
    }
    // Ensure that all primary key columns are also listed as columns, with attention to order.
    Set<String> extraColumns = Sets.difference(primaryKeyColumnNames, columns.keySet());
    List<SQLColumnInfo> finalColumns = Lists.newArrayList();
    // Add extra pseudokey columns at the beginning.
    finalColumns.addAll(getColumnInfoByTrialQuery(tab, extraColumns).values());
    finalColumns.addAll(columns.values());
    return new SQLTableInfo(tab, finalColumns, primaryKeyColumnNames);
  }

  private SQLFragment buildTrialQuery(SQLTable table, List<String> columnNames) throws SQLException
  {
    Preconditions.checkArgument(!columnNames.isEmpty());
    SQLFragmentBuilder ret = SQLFragment.builder();
    /* Don't use a DISTINCT in the trial query; I have actually observed it to cause a segfault on
    Oracle in some cases where the use of ROWID is not permitted, breaking the JDBC connection. */
    ret.appendJDBCString("SELECT ");
    List<SQLFragment> columnFragments = Lists.newArrayList();
    for (String columnName : columnNames)
      columnFragments.add(getBackendDialect().quoteIdentifier(columnName, true));
    ret.appendSQLFragment(SQLFragment.join(", ", columnFragments));
    ret.appendJDBCString(" FROM ");
    ret.appendSQLFragment(getBackendDialect().getTableIdentifier(table));
    return ret.build();
  }

  private Map<String,SQLColumnInfo> getColumnInfoByTrialQuery(
      SQLTable table, Set<String> columnNames) throws SQLException
  {
    Preconditions.checkNotNull(table);
    Preconditions.checkNotNull(columnNames);
    if (columnNames.isEmpty())
      return ImmutableMap.of();
    final List<String> columnNamesList = ImmutableList.copyOf(columnNames);
    final SQLFragment trialQuery = buildTrialQuery(table, columnNamesList);
    final String queryString = getBackendDialect().getJDBCStringWithLimit(trialQuery);
    final boolean oldAutoCommit;
    oldAutoCommit = jdbcConnection.getAutoCommit();
    try {
      jdbcConnection.setAutoCommit(false);
      try (PreparedStatement ps = jdbcConnection.prepareStatement(queryString)) {
        ps.setInt(1, 1);
        try (ResultSet rs = ps.executeQuery()) {
          rs.next();
          final Map<String,SQLColumnInfo> ret = Maps.newLinkedHashMap();
          final ResultSetMetaData rsmd = rs.getMetaData();
          int colNum = 1;
          for (String columnName : columnNamesList) {
            // TODO[pd]: Should generated actually be true here?
            MapUtil.putChecked(ret, columnName, new SQLColumnInfo(table, columnName,
                SQLType.createFromResultSetMetaData(rsmd, colNum),
                0, "", "", -1, null, SQLTypeCode.NULL, null, null));
            colNum++;
          }
          return Collections.unmodifiableMap(ret);
        }
      }
    } finally {
      jdbcConnection.rollback();
      jdbcConnection.setAutoCommit(oldAutoCommit);
    }
  }

  private boolean isRowIdentifierLegal(SQLTable table, String rowIdentifier) throws SQLException {
    Preconditions.checkNotNull(table);
    Preconditions.checkNotNull(rowIdentifier);
    Preconditions.checkArgument(!rowIdentifier.isEmpty());
    try {
      // Ignore result; just check for errors.
      getColumnInfoByTrialQuery(table, ImmutableSet.of(rowIdentifier));
      return true;
    } catch (SQLException e) {
      reportSQLException(e);
      if (invalidated)
        throw e;
      /* TODO[pd]: Be a little more specific in which kinds of errors are actually indications that
                   ROWID is forbidden vs. other kinds of errors that might happen. On Oracle, I know
                   of the following:

         ORA-01446: cannot select ROWID from, or sample, a view with DISTINCT, GROUP BY, etc.
         ORA-01445: cannot select ROWID from, or sample, a join view without a key-preserved table
         ORA-02031: no ROWID for fixed tables or for external-organized tables
      */
      return false;
    }
  }

  private @Nonnull Set<String> getBestRowIdentifier(DatabaseMetaData md, SQLTable tab)
      throws SQLException
  {
    if (!getBackendDialect().supportsGetBestRowIdentifier())
      return ImmutableSet.of();
    final String catalogName = getDatabaseNamePart(tab.getDatabase(), true);
    final String schemaName = getDatabaseNamePart(tab.getDatabase(), false);
    final String tableName = tab.getTableName();
    Set<String> ret = Sets.newLinkedHashSet();
    Integer foundScope = null;
    // TODO[pd]: Figure out what settings to use here.
    final boolean USE_BEST_SCOPE = true;
    final int SCOPE_OF_INTEREST = DatabaseMetaData.bestRowTransaction;
    try (ResultSet rs = md.getBestRowIdentifier(
        catalogName, schemaName, tableName, SCOPE_OF_INTEREST, true))
    {
      while (rs.next()) {
        final int actualScope = rs.getShort("SCOPE");
        final String columnName = rs.getString("COLUMN_NAME").trim();
        if (USE_BEST_SCOPE) {
          if (foundScope == null || actualScope > foundScope) {
            foundScope = actualScope;
            ret.clear();
          }
        } else {
          foundScope = SCOPE_OF_INTEREST;
        }
        if (actualScope == foundScope)
          ret.add(columnName);
      }
    }
    return Collections.unmodifiableSet(ret);
  }

  private @Nonnull Set<String> getPrimaryKeyColumnNames(DatabaseMetaData md, SQLTable tab)
      throws SQLException
  {
    final String catalogName = getDatabaseNamePart(tab.getDatabase(), true);
    final String schemaName = getDatabaseNamePart(tab.getDatabase(), false);
    final String tableName = tab.getTableName();
    Set<String> ret = Sets.newLinkedHashSet();
    try (ResultSet rs = md.getPrimaryKeys(catalogName, schemaName, tableName)) {
      while (rs.next()) {
        SQLTable actualTable = SQLTable.create(
            createSQLDatabase(rs.getString("TABLE_CAT"), rs.getString("TABLE_SCHEM")),
            rs.getString("TABLE_NAME"));
        if (!checkTableMatches(tab, actualTable))
          continue;
        ret.add(rs.getString("COLUMN_NAME").trim());
      }
    }
    return Collections.unmodifiableSet(ret);
  }

  private @Nonnull Map<String,SQLColumnInfo> getColumnsWithInfo(SQLTable tab) throws SQLException {
    final DatabaseMetaData md = jdbcConnection.getMetaData();
    final String catalogName = getDatabaseNamePart(tab.getDatabase(), true);
    final String schemaName = getDatabaseNamePart(tab.getDatabase(), false);
    final String tableName = tab.getTableName();
    Map<String,SQLColumnInfo> ret = Maps.newLinkedHashMap();
    try (ResultSet rs =
        md.getColumns(catalogName, toPattern(schemaName), toPattern(tableName), null))
    {
      final Set<String> resultSetColumnNames = getResultSetColumnNames(rs);
      final boolean hasGeneratedColumnField = resultSetColumnNames.contains("IS_GENERATEDCOLUMN");
      final boolean hasScopeTableField = resultSetColumnNames.contains("SCOPE_TABLE");
      while (rs.next()) {
        SQLColumnInfo columnInfo = buildColumnInfo(rs, hasGeneratedColumnField, hasScopeTableField);
        if (!checkTableMatches(tab, columnInfo.getTable()))
          continue;
        // Ignore duplicates.
        ret.put(columnInfo.getName(), columnInfo);
      }
    }
    return Collections.unmodifiableMap(ret);
  }

  public void markClosed() {
    checkThread();
    Preconditions.checkState(shieldClose);
    closed = true;
  }

  @Override
  public void close() throws SQLException {
    checkThread();
    if (!closed) {
      closed = true;
      if (!shieldClose)
        jdbcConnection.close();
    }
  }

  /* Technically, only YES and NO needs to be supported, but allow for possible variations in JDBC
  drivers. */
  private static final ImmutableMap<String,Boolean> JDBC_BOOL_MAP =
      ImmutableMap.<String,Boolean>builder()
      .put("yes", true)
      .put("no", false)
      .put("true", true)
      .put("false", false)
      .put("y", true)
      .put("n", false)
      .put("t", true)
      .put("f", false)
      .build();

  private static @Nullable Boolean getStringBoolean(ResultSet rs, String columnLabel)
      throws SQLException
  {
    return JDBC_BOOL_MAP.get(
        StringUtil.trimmedAndEmptyIfNull(rs.getString(columnLabel)).toLowerCase(Locale.US));
  }

  /**
   * @param rs a result set returned by
   *        {@link DatabaseMetaData#getColumns(String,String,String,String)}
   * @param hasGeneratedColumnField whether the result set includes the IS_GENERATEDCOLUMN field
   *        (introduced in JDBC 4.0)
   * @param hasScopeTableField whether the result set includes the SCOPE_TABLE field (once observed
   *        to be missing when using ojdbc6.jar, an old Oracle JDBC driver)
   */
  private SQLColumnInfo buildColumnInfo(
      ResultSet rs, boolean hasGeneratedColumnField, boolean hasScopeTableField)
      throws SQLException
  {
    Preconditions.checkNotNull(rs);
    final @Nullable Boolean nullable;
    {
      final int nullable1 = rs.getInt("NULLABLE");
      final @Nullable Boolean nullableIso = getStringBoolean(rs, "IS_NULLABLE");
      if (Boolean.TRUE.equals(nullableIso) || nullable1 == DatabaseMetaData.columnNullable) {
        // If either field indicates certain nullability, assume nullable.
        nullable = true;
      } else if (
          Boolean.FALSE.equals(nullableIso) || nullable1 == DatabaseMetaData.columnNoNulls)
      {
        /* If either field indicates certain non-nullability, and there is no inconsistency, assume
        non-nullable. */
        nullable = false;
      } else {
        // If both fields indicate unknown nullability, assume that nullability if unknown.
        nullable = null;
      }
    }
    final @Nullable SQLTable referencedTable;
    {
      final @Nullable String referencedTableName =
          hasScopeTableField ? rs.getString("SCOPE_TABLE") : null;
      if (referencedTableName == null) {
        referencedTable = null;
      } else {
        referencedTable = SQLTable.create(
            createSQLDatabase(rs.getString("SCOPE_CATALOG"), rs.getString("SCOPE_SCHEMA")),
            referencedTableName);
      }
    }
    final SQLType type = new SQLType(
        SQLTypeCode.create(rs.getInt("DATA_TYPE")),
        StringUtil.trimmedAndEmptyIfNull(rs.getString("TYPE_NAME")),
        rs.getInt("DECIMAL_DIGITS"),
        rs.getInt("COLUMN_SIZE"),
        rs.getInt("CHAR_OCTET_LENGTH"),
        nullable
    );
    /* Avoid fetching the COLUMN_DEF field, as it can contain arbitrarily long strings. On Oracle,
    I once observed a long string in the COLUMN_DEF field being retrieved using the deprecated LONG
    string type, which causes a "Stream has Already Been Closed" exception upon the call to
    ResultSet.getString unless the LONG column is the first one to be retrieved. */
    final boolean AVOID_FETCHING_COLUMN_DEF = true;
    return new SQLColumnInfo(
        SQLTable.create(
            createSQLDatabase(rs.getString("TABLE_CAT"), rs.getString("TABLE_SCHEM")),
            rs.getString("TABLE_NAME")),
        rs.getString("COLUMN_NAME").trim(),
        type,
        rs.getInt("NUM_PREC_RADIX"),
        StringUtil.trimmedAndEmptyIfNull(rs.getString("REMARKS")),
        AVOID_FETCHING_COLUMN_DEF ? "" :
            StringUtil.trimmedAndEmptyIfNull(rs.getString("COLUMN_DEF")),
        rs.getInt("ORDINAL_POSITION"),
        referencedTable,
        SQLTypeCode.create(rs.getShort("SOURCE_DATA_TYPE")),
        getStringBoolean(rs, "IS_AUTOINCREMENT"),
        hasGeneratedColumnField ? getStringBoolean(rs, "IS_GENERATEDCOLUMN") : null
    );
  }

  // TODO[cc]: Use a null primaryKeyColumns instead of an empty set to signify no primary key.
  // TODO[pd]: Disallow creation of a table without columns if the backend does not support it.
  /** @param primaryKeyColumns if empty, no primary key will be assigned */
  public void createTable(
      SQLTable table, Map<String,SQLType> columns, Set<String> primaryKeyColumns, boolean temporary)
      throws SQLException
  {
    Preconditions.checkNotNull(table);
    Preconditions.checkArgument(columns.keySet().containsAll(primaryKeyColumns));
    SQLFragmentBuilder fb = SQLFragment.builder();
    fb.appendJDBCString(getBackendDialect().getCreateTableStatement(temporary, false));
    fb.appendJDBCString(" ");
    fb.appendSQLFragment(getBackendDialect().getTableIdentifier(table));
    fb.appendJDBCString(" (");
    {
      final List<SQLFragment> columnFragments = Lists.newArrayList();
      for (Entry<String,SQLType> entry : columns.entrySet()) {
        final String columnName = entry.getKey();
        final SQLType type = entry.getValue();
        final String typeName = getBackendDialect().getTypeName(type);
        if (typeName.isEmpty())
          throw new SQLException("Data type " + type + " not supported by this backend");
        columnFragments.add(SQLFragment.builder()
            .appendSQLFragment(getBackendDialect().quoteIdentifier(columnName, true))
            .appendJDBCString(" ")
            .appendJDBCString(typeName)
            .build());
      }
      fb.appendSQLFragment(SQLFragment.join(", ", columnFragments));
    }
    {
      final List<SQLFragment> pkFragments = Lists.newArrayList();
      for (String columnName : primaryKeyColumns)
        pkFragments.add(getBackendDialect().quoteIdentifier(columnName, true));
      if (!pkFragments.isEmpty()) {
        if (!columns.isEmpty())
          fb.appendJDBCString(", ");
        fb.appendJDBCString("PRIMARY KEY(");
        fb.appendSQLFragment(SQLFragment.join(", ", pkFragments));
        fb.appendJDBCString(")");
      }
    }
    fb.appendJDBCString(")");
    executeUpdate(fb.build());
  }

  public void executeUpdate(SQLFragment sqlFragment) throws SQLException {
    try (PreparedStatement ps = getJDBCConnection().prepareStatement(sqlFragment.getJDBCString())) {
      int parameterIndex = 1;
      for (Object parameter : sqlFragment.getQueryParameters())
        getBackendDialect().bindParameterValue(ps, parameterIndex++, parameter);
      ps.execute();
    }
  }

  public void drop(SQLTable table, boolean view, boolean ifExists, boolean cascade)
      throws SQLException
  {
    SQLFragmentBuilder fb = SQLFragment.builder();
    fb.appendJDBCString("DROP ");
    fb.appendJDBCString(view ? "VIEW" : "TABLE");
    fb.appendJDBCString(" ");
    @Nullable BeforeOrAfter ifExistsForm = getBackendDialect().getIfExistsLocationForDropTable();
    if (ifExists && ifExistsForm == BeforeOrAfter.BEFORE)
      fb.appendJDBCString("IF EXISTS ");
    fb.appendSQLFragment(getBackendDialect().getTableIdentifier(table));
    if (ifExists && ifExistsForm == BeforeOrAfter.AFTER)
      fb.appendJDBCString(" IF EXISTS");
    if (cascade) {
      String cascadeString = getBackendDialect().getCascadeStringForDropTable();
      if (!cascadeString.isEmpty())
        fb.appendJDBCString(" " + cascadeString);
    }
    try {
      executeUpdate(fb.build());
    } catch (SQLException e) {
      if (ifExistsForm != null || !e.getMessage().contains("does not exist"))
        throw e;
    }
  }
}
