// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalTime;
import java.util.List;
import javax.annotation.concurrent.Immutable;

/**
 * A fragment of a SQL query, which may include constant query parameters. Identifiers should be
 * inserted as part of the regular JDBC string after first being quoted in backend-specific manner.
 * The latter design decision allows errors relating to unquoteable identifiers to be handled at the
 * point where the identifiers originate rather than when the entire query is later processed.
 */
@Immutable
public final class SQLFragment {
  private final String jdbcString;
  private final ImmutableList<Object> queryParameters;

  private SQLFragment(String jdbcString, List<Object> queryParameters) {
    /* Usually, the number of question marks in jdbcString should equal the size of queryParameters,
    but it's conceivable that some JDBC drivers might allow question marks in properly quoted
    identifiers, which would violate this assumption. So don't try to enforce it. */
    Preconditions.checkNotNull(jdbcString);
    for (Object queryParameter : queryParameters) {
      Preconditions.checkNotNull(queryParameter);
      if (queryParameter instanceof java.sql.Date ||
          queryParameter instanceof java.sql.Time ||
          queryParameter instanceof java.sql.Timestamp)
      {
        throw new AssertionError(
            "Use java.time.LocalDate/LocalTime/Instant rather than java.sql.Date/Time/Timestamp");
      }
    }
    this.jdbcString = jdbcString;
    this.queryParameters = ImmutableList.copyOf(queryParameters);
  }

  public static SQLFragmentBuilder builder() {
    return new SQLFragmentBuilder();
  }

  public static SQLFragment createParameter(Object queryParameter) {
    return new SQLFragment("?", ImmutableList.of(queryParameter));
  }

  public static SQLFragment create(String jdbcString, List<Object> queryParameters) {
    return new SQLFragment(jdbcString, queryParameters);
  }

  public static SQLFragment create(String jdbcString) {
    return new SQLFragment(jdbcString, ImmutableList.of());
  }

  public SQLFragment enclose(String before, String after) {
    return new SQLFragment(before + jdbcString + after, queryParameters);
  }

  public static SQLFragment join(String separator, List<SQLFragment> fragments) {
    SQLFragmentBuilder ret = builder();
    boolean first = true;
    for (SQLFragment fragment : fragments) {
      if (first) {
        first = false;
      } else {
        ret.appendJDBCString(separator);
      }
      ret.appendSQLFragment(fragment);
    }
    return ret.build();
  }

  public String getJDBCString() {
    return jdbcString;
  }

  public List<Object> getQueryParameters() {
    return queryParameters;
  }

  private static String stringLiteral(String s) {
    return "'" + s.replace("'", "''") + "'";
  }

  /**
   * Return a string representation for debugging purposes.
   */
  @Override
  public String toString() {
    StringBuilder ret = new StringBuilder();
    int paramIndex = 0;
    for (int i = 0; i < jdbcString.length(); i++){
      char c = jdbcString.charAt(i);
      if (c == '?' && paramIndex < queryParameters.size()) {
        Object parameter = queryParameters.get(paramIndex++);
        if (parameter instanceof String) {
          ret.append(stringLiteral((String) parameter));
        } else if (parameter instanceof LocalDate) {
          ret.append("(date ")
              .append(stringLiteral(parameter.toString())).append(")");
        } else if (parameter instanceof LocalTime) {
          ret.append("(time without time zone ")
              .append(stringLiteral(parameter.toString())).append(")");
        } else if (parameter instanceof Instant) {
          ret.append("(timestamp with time zone ")
              .append(stringLiteral(parameter.toString())).append(")");
        } else {
          ret.append(parameter);
        }
      } else {
        ret.append(c);
      }
    }
    return ret.toString();
  }
}
