// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.query.SQLTable;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.List;

public class SQLInserter implements AutoCloseable {
  private final int batchSize;
  private final SQLConnection connection;
  private final PreparedStatement insertStatement;
  /** Immutable after creation. */
  private final int[] columnTypes;
  private final boolean oldAutoCommit;
  private boolean aborted;
  private int currentBatchSize;

  public SQLInserter(
      SQLConnection connection, SQLTable table, List<String> columnNames, int batchSize)
      throws SQLException
  {
    Preconditions.checkNotNull(connection);
    Preconditions.checkNotNull(table);
    Preconditions.checkNotNull(columnNames);
    Preconditions.checkArgument(batchSize > 0);
    this.batchSize = batchSize;
    this.connection = connection;
    oldAutoCommit = connection.getJDBCConnection().getAutoCommit();
    connection.getJDBCConnection().setAutoCommit(false);
    final BackendDialect dialect = connection.getBackendDialect();
    final SQLTableInfo tableInfo = connection.getTableInfo(table);
    columnTypes = new int[columnNames.size()];
    {
      int i = 0;
      for (String columnName : columnNames) {
        SQLColumnInfo columnInfo = tableInfo.getColumnInfo(columnName);
        if (columnInfo == null)
          throw new SQLException("Column '" + columnName + "' not found");
        columnTypes[i++] = columnInfo.getType().getTypeCode().getCode();
      }
    }
    final SQLFragmentBuilder fb = SQLFragment.builder();
    fb.appendJDBCString("INSERT INTO ");
    fb.appendSQLFragment(dialect.getTableIdentifier(table));
    {
      final List<SQLFragment> columnNameFragments = Lists.newArrayList();
      final List<SQLFragment> columnParameterFragments = Lists.newArrayList();
      for (String columnName : columnNames) {
        columnNameFragments.add(dialect.quoteIdentifier(columnName, true));
        columnParameterFragments.add(SQLFragment.create("?"));
      }
      fb.appendJDBCString(" (");
      fb.appendSQLFragment(SQLFragment.join(", ", columnNameFragments));
      fb.appendJDBCString(") VALUES (");
      fb.appendSQLFragment(SQLFragment.join(", ", columnParameterFragments));
      fb.appendJDBCString(")");
    }
    insertStatement = connection.getJDBCConnection().prepareStatement(fb.build().getJDBCString());
  }

  // TODO[pd]: Consider setting row parameters one by one instead.
  public void insert(Object[] row) throws SQLException {
    Preconditions.checkState(!aborted);
    Preconditions.checkArgument(row.length == columnTypes.length);
    aborted = true;
    for (int i = 0; i < columnTypes.length; i++) {
      insertStatement.setObject(i + 1, row[i], columnTypes[i]);
    }
    insertStatement.addBatch();
    currentBatchSize++;
    if (currentBatchSize % batchSize == 0)
      flush();
    aborted = false;
  }

  private void flush() throws SQLException {
    insertStatement.executeBatch();
    currentBatchSize = 0;
  }

  @Override
  public void close() throws SQLException {
    try {
      flush();
    } finally {
      insertStatement.close();
      if (aborted) {
        connection.getJDBCConnection().rollback();
      } else {
        connection.getJDBCConnection().commit();
      }
      connection.getJDBCConnection().setAutoCommit(oldAutoCommit);
    }
  }
}
