// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database;

import com.google.common.base.Preconditions;
import com.sieuferd.util.StringUtil;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

// TODO[cc]: Rename this to avoid collision with java.sql.SQLType.
/**
 * Data type information for for a table column or result set column. Reflects the metadata provided
 * by {@link java.sql.ResultSetMetaData} or a subset of that provided by
 * {@link java.sql.DatabaseMetaData#getColumns(String,String,String,String)}. Javadoc of this class
 * is based on that of the above two classes.
 *
 * The scale/precision/length concepts are analogous to the {@code $s}/{@code $p}/{@code $l}
 * constants used in data type definitions in Hibernate dialects
 * ({@code org.hibernate.dialect.TypeNames}).
 */
@Immutable
public class SQLType {
  private final SQLTypeCode typeCode;
  private final String typeName;
  private final int scale;
  private final int precision;
  private final int length;
  private final @Nullable Boolean nullable;

  public SQLType(SQLTypeCode typeCode, String typeName, int scale, int precision, int length,
      @Nullable Boolean nullable)
  {
    Preconditions.checkNotNull(typeCode);
    StringUtil.checkArgumentTrimmed(typeName);
    this.typeCode = typeCode;
    this.typeName = typeName;
    this.scale = scale;
    this.precision = precision;
    this.length = length;
    this.nullable = nullable;
  }

  public SQLTypeCode getTypeCode() {
    return typeCode;
  }

  /**
   * Get the data source dependent type name. For a UDT, the type name is fully qualified.
   */
  public String getTypeName() {
    return typeName;
  }

  /**
   * Get the number of digits to right of the decimal point. Zero is returned for data types where
   * the scale is not applicable. Probably relevant for numeric data types only.
   */
  public int getScale() {
    return scale;
  }

  /**
   * Get the precision. For numeric data, this is the maximum precision. For character data, this
   * is the length in characters. For datetime datatypes, this is the length in characters of the
   * String representation (assuming the maximum allowed precision of the fractional seconds
   * component). For binary data, this is the length in bytes. For the {@code ROWID} datatype, this
   * is the length in bytes. Zero is returned for data types where the precision is not
   * applicable. Mostly relevant for numeric data types; for character types, {@link #getLength()}
   * would be preferred.
   */
  public int getPrecision() {
    return precision;
  }

  /**
   * Get the normal maximum width in bytes. Probably relevant for character and binary data types
   * only.
   */
  public int getLength() {
    return length;
  }

  /**
   * Return whether the field is nullable, or null if nullability is unknown.
   */
  public @Nullable Boolean getNullable() {
    return nullable;
  }

  public static SQLType createFromResultSetMetaData(ResultSetMetaData metaData, int columnIndex)
      throws SQLException
  {
    final @Nullable Boolean nullable;
    {
      final int nullableInt = metaData.isNullable(columnIndex);
      if (nullableInt == ResultSetMetaData.columnNullable) {
        nullable = true;
      } else if (nullableInt == ResultSetMetaData.columnNoNulls) {
        nullable = false;
      } else {
        nullable = null;
      }
    }
    return new SQLType(
        SQLTypeCode.create(metaData.getColumnType(columnIndex)),
        StringUtil.trimmedAndEmptyIfNull(metaData.getColumnTypeName(columnIndex)),
        metaData.getScale(columnIndex),
        metaData.getPrecision(columnIndex),
        metaData.getColumnDisplaySize(columnIndex),
        nullable);
  }

  @Override
  public String toString() {
    return "SQLType(typeCode=" + typeCode +
        ", typeName=" + typeName +
        ", s=" + scale +
        ", p=" + precision +
        ", l=" + length +
        ", nullable=" + nullable +
        ")";
  }
}
