// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import com.google.common.collect.Maps;
import com.sieuferd.database.SQLType;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLFragmentBuilder;
import com.sieuferd.query.SQLTable;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Map;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

/** Implementations should be immutable (and thus thread-safe) after construction. */
@Immutable
public abstract class BackendDialect {
  // Since this field is only modified during object construction, access should be thread-safe.
  private final Map<Function,FunctionImplementation> functionImplementations =
      Maps.newLinkedHashMap();

  protected BackendDialect() {
  }

  public enum BeforeOrAfter {
    BEFORE,
    AFTER
  }

  /* The API convention for this class is to not provide any default implementations; leave that to
  GenericJDBCBackendDialect. */

  public abstract String getDisplayName();
  // TODO[pd]: Allow different start and end quotes.
  /** @return may be empty, but not null */
  public abstract String getIdentifierQuoteString();
  /* TODO[cc]: Separate out settings that are used solely for the purposes of skipping tests. Or
               replace them with proper error handling in the relevant API instead, e.g. a "known to
               be unsupported" exception. */
  /** Used for unit testing only. */
  public abstract boolean supportsIdentifierQuoteInTableName();
  /** Used for unit testing only. */
  public abstract boolean supportsNanosecondTemporals();
  public abstract boolean allowsAliasInGroupBy();
  public abstract String getCatalogSeparator();
  /** @return may be empty, but not null */
  public abstract String getSearchStringEscape();
  public abstract boolean supportsNullsFirstLast();
  // TODO[cc]: Replace isCatalogAtStart and isUsingCatalogs with primaryDatabaseNamePartIsCatalog.
  public abstract boolean isCatalogAtStart();
  public abstract boolean isUsingCatalogs();
  public abstract boolean supportsGetBestRowIdentifier();
  /** @return may be empty, but not null */
  public abstract @Nonnull String getDatabaseNameTerm(boolean primaryPart);
  // TODO[pd]: Consider if I should be using ResultSet.getRowId for anything.
  /** @return may be empty, but not null */
  public abstract @Nonnull String getRowIdentifierPseudoColumnName();
  public abstract RenderedExpression getNullLiteral();
  public abstract RenderedExpression getBooleanLiteral(boolean value);
  public abstract RenderedExpression getLiteral(ConstantExpression exp);
  public abstract RenderVariant getStorageVariant();
  /**
   * Get a table expression that will evaluate to a table always containing a single row. Any fields
   * in the table will be ignored.
   */
  public abstract SQLFragment getSingleRowDummyTable();

  // TODO[pd]: Support dialects for which the limit parameter goes at the beginning.
  public abstract String getJDBCStringWithLimit(SQLFragment query);

  // TODO[cc]: Consider returning null or throwing an exception if the data type is not supported.
  // TODO[cc]: Consider renaming, or separating the nullability logic out.
  /**
   * @return type definition string supporting values of the given type, including nullability;
   *         empty string if the data type is not supported
   */
  public abstract String getTypeName(SQLType type);

  public abstract @Nullable String getCreateTableStatement(boolean temporary, boolean view);

  /** @return null if not supported */
  public abstract @Nullable BeforeOrAfter getIfExistsLocationForDropTable();
  /** @return empty string if not supported */
  public abstract String getCascadeStringForDropTable();

  // Analogous to org.hibernate.type.descriptor.ValueBinder.
  /**
   * Bind a parameter to a {@code PreparedStatement}.
   *
   * Temporal values will be supplied using the newer {@link java.time.LocalDate},
   * {@link java.time.LocalTime}, and {@link java.time.Instant} types rather than the older
   * {@link java.sql.Date}, {@link java.sql.Time}, {@link java.sql.Timestamp} ones.
   */
  public abstract void bindParameterValue(
      PreparedStatement statement, int parameterIndex, @Nonnull Object value)
      throws SQLException;

  // Analogous to org.hibernate.type.descriptor.ValueExtractor
  /* I considered always wrapping returned objects in a JDBCObjectWrapper unless their class was
  on a whitelist. However, this carries the risk of encountering collisions between string
  representations of distinct values, for instance with the default implementation of toString(),
  which is based on hashCode(). Such false positives would generate assertion errors in the query
  evaluator, while false negatives will not. */
  /**
   * Fetch a result value from a result set. This generally corresponds to a call to
   * {@code ResultSet#getObject(int)}; however, individual dialects may need to provide custom logic
   * to produce the right kind of object for some data types and {@link RenderVariant}s.
   *
   * <p>Any custom logic provided should take special care to ensure that null values remain null
   * for all data types and render variants. If the value is retrieved as a primitive type,
   * e.g. using {@link ResultSet#getInt(int)}, then {@link ResultSet#wasNull()} can be used to
   * distinguish a null value from a default one.
   *
   * <p>For temporal values, the newer {@link java.time.LocalDate}, {@link java.time.LocalTime},
   * and {@link java.time.Instant} types should be returned rather than the older
   * {@link java.sql.Date}, {@link java.sql.Time}, {@link java.sql.Timestamp} ones.
   */
  public abstract @Nullable Object fetchResultValue(
      ResultSet resultSet, int columnIndex, SQLType jdbcColumnType, RenderVariant renderVariant)
      throws SQLException;

  /** May only be called from subclass constructors */
  protected final void registerFunction(Function function, FunctionImplementation implementation) {
    Preconditions.checkArgument(implementation.getArgumentCount() == null ||
        function.isArgumentCountCorrect(implementation.getArgumentCount()));
    functionImplementations.put(function, implementation);
  }

  protected final void registerFunction(
      Function function, RenderVariant renderVariant, String template)
  {
    FunctionImplementation functionImplementation = new TemplateFunctionImplementation(
        function.getArgumentCount(), template, renderVariant);
    registerFunction(function, functionImplementation);
  }

  protected final void registerFunction(Function function, String template) {
    registerFunction(function, RenderVariant.DEFAULT, template);
  }

  /** May only be called from subclass constructors */
  protected final void unregisterFunction(Function function) {
    functionImplementations.remove(function);
  }

  public final Map<Function,FunctionImplementation> getFunctionImplementations() {
    return Collections.unmodifiableMap(functionImplementations);
  }

  public final String quoteIdentifierForSearchString(String identifier) throws SQLException {
    final String quoteString = getSearchStringEscape();
    if (quoteString.isEmpty()) {
      if (identifier.contains("%") || identifier.contains("_")) {
        throw new SQLException("Identifier '" + identifier + "' contains special characters, "+
            "but backend does not support metadata pattern quoting");
      }
      return identifier;
    } else {
      return identifier
          .replace(quoteString, quoteString + quoteString)
          .replace("%", quoteString + "%")
          .replace("_", quoteString + "_");
    }
  }

  public final boolean isRowIdentifierPseudoColumn(String columnName) {
    /* Since the column name may come from a DatabaseMetaData result rather than directly from
    getRowIdentifierPseudoColumnName(), do a case-insensitive comparison. */
    String pseudoColumnName = getRowIdentifierPseudoColumnName();
    return !pseudoColumnName.isEmpty() && columnName.equalsIgnoreCase(pseudoColumnName);
  }

  /* TODO[pd]: Allow quoteIdentifier and getTableIdentifier to be overridden by dialects, and make
               sure I do the right thing when a Hibernate dialect exists. */

  public final SQLFragment quoteIdentifier(String identifier, boolean column) {
    /* Some backends, e.g. Oracle, consider the row identifier pseudo-column to be a keyword rather
    than an identifier, and so it should not be quoted. */
    if (column && isRowIdentifierPseudoColumn(identifier))
      return SQLFragment.create(identifier);
    final String identifierQuoteString = getIdentifierQuoteString();
    if (identifierQuoteString.isEmpty()) {
      /* While it would be desirable to check that the identifier can be used safely without quoting
      in this case, the logic required for such a check would be excessive and not foolproof. For
      instance, it would be necessary to check against lists of possible SQL keywords, to take into
      account the case sensitivity policy of the driver, to take into account extra characters
      allowed in identifiers, etc. */
      return SQLFragment.create(identifier);
    } else {
      /* I think nothing special has to be done for identifiers with question marks; see
      http://forums.mysql.com/read.php?39,499385,499450#msg-499450 and
      http://bugs.mysql.com/bug.php?id=10155 for the MySQL case.
      QueryEvaluatorTest.testWeirdlyNamedTables should conver this case. */
      return SQLFragment.create(
          identifierQuoteString +
          identifier.replace(identifierQuoteString, identifierQuoteString + identifierQuoteString) +
          identifierQuoteString);
    }
  }

  public final SQLFragment getTableIdentifier(SQLTable sqlTable) {
    final SQLFragmentBuilder ret = SQLFragment.builder();
    final SQLDatabase database = sqlTable.getDatabase();
    if (!database.getPrimaryNamePart().isEmpty()) {
      ret.appendSQLFragment(quoteIdentifier(database.getPrimaryNamePart(), false));
      if (!database.getSecondaryNamePart().isEmpty()) {
        ret.appendJDBCString(getCatalogSeparator());
        ret.appendSQLFragment(quoteIdentifier(database.getSecondaryNamePart(), false));
      }
      ret.appendJDBCString(".");
    }
    ret.appendSQLFragment(quoteIdentifier(sqlTable.getTableName(), false));
    return ret.build();
  }
}
