// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.dialect.FunctionImplementation.RenderFunctionException.DynamicRenderFunctionException;
import com.sieuferd.database.dialect.FunctionImplementation.RenderFunctionException.WrongNumberOfArgumentsException;
import java.util.List;
import javax.annotation.Nullable;

// TODO[cc]: Go over anonymous subclasses of this and replace them with template string patterns.
// TODO[cc]: Store Function here, then avoid duplicating argumentCount.
public abstract class FunctionImplementation {
  private final @Nullable Integer argumentCount;

  /** @param argumentCount null for variadic functions */
  public FunctionImplementation(@Nullable Integer argumentCount) {
    this.argumentCount = argumentCount;
  }

  public final RenderedExpression render(List<RenderedExpression> arguments)
    throws WrongNumberOfArgumentsException, DynamicRenderFunctionException
  {
    if (argumentCount != null && arguments.size() != argumentCount)
      throw new WrongNumberOfArgumentsException();
    return renderImpl(arguments);
  }

  /* @param arguments is guaranteed to have the correct size */
  protected abstract RenderedExpression renderImpl(List<RenderedExpression> arguments)
      throws DynamicRenderFunctionException;

  /** @return null for variadic functions */
  public @Nullable Integer getArgumentCount() {
    return argumentCount;
  }

  public static abstract class RenderFunctionException extends Exception {
    public RenderFunctionException(String message) {
      super(message);
    }

    public RenderFunctionException() {
    }

    public static final class WrongNumberOfArgumentsException extends RenderFunctionException { }
    // TODO[cc]: Consider getting rid of this.
    public static final class DynamicRenderFunctionException extends RenderFunctionException {
      public DynamicRenderFunctionException(String message) {
        super(message);
      }
    }
  }

  // TODO[cc]: Consider removing.
  public static abstract class SimpleFunctionImplementation extends FunctionImplementation {
    private final RenderVariant outputVariant;

    public SimpleFunctionImplementation(
        RenderVariant outputVariant, @Nullable Integer argumentCount)
    {
      super(argumentCount);
      Preconditions.checkNotNull(outputVariant);
      this.outputVariant = outputVariant;
    }

    public SimpleFunctionImplementation(@Nullable Integer argumentCount) {
      this(RenderVariant.DEFAULT, argumentCount);
    }

    @Override
    protected final RenderedExpression renderImpl(List<RenderedExpression> arguments)
        throws DynamicRenderFunctionException
    {
      List<SQLFragment> preferredArguments = Lists.newArrayList();
      for (RenderedExpression argument : arguments)
        preferredArguments.add(argument.toPreferred().getFragment());
      return new RenderedExpression(renderFragment(preferredArguments), outputVariant);
    }

    /* @param arguments is guaranteed to have the correct size */
    protected abstract SQLFragment renderFragment(List<SQLFragment> arguments)
        throws DynamicRenderFunctionException;
  }
}
