// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLTypeCode;
import com.sieuferd.database.dialect.FunctionImplementation.RenderFunctionException.DynamicRenderFunctionException;
import com.sieuferd.database.dialect.RenderVariant.NullVariant;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.util.StringUtil;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.ZoneOffset;
import java.util.Calendar;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

public class GenericJDBCBackendDialect extends BackendDialect {
  private static final boolean USE_VALUE_BASED_CONDITIONAL_IMPLEMENTATION = true;
  private static final boolean ATTEMPT_CALENDAR_DEPENDENCE_DETECTION = false;
  private static final int CLOB_SAMPLE_CHARACTER_LIMIT = 4096;
  private static final ImmutableSet<Integer> NUMERIC_SQL_TYPES = ImmutableSet.<Integer>builder()
      .add(java.sql.Types.FLOAT)
      .add(java.sql.Types.DOUBLE)
      .add(java.sql.Types.REAL)
      .add(java.sql.Types.INTEGER)
      .add(java.sql.Types.TINYINT)
      .add(java.sql.Types.SMALLINT)
      .add(java.sql.Types.BIGINT)
      .add(java.sql.Types.DECIMAL)
      .add(java.sql.Types.NUMERIC)
      .build();

  private static final RenderedExpression LITERAL_NULL =
      RenderedExpression.simple(SQLFragment.create("NULL"));
  private static final RenderedExpression LITERAL_TRUE =
      RenderedExpression.simple(SQLFragment.create("true"));
  private static final RenderedExpression LITERAL_FALSE =
      RenderedExpression.simple(SQLFragment.create("false"));
  public static FunctionImplementation IDENTITY_FUNCTION = new FunctionImplementation(1) {
    @Override
    protected RenderedExpression renderImpl(List<RenderedExpression> arguments) {
      return arguments.get(0);
    }
  };
  private final String identifierQuoteString;
  private final String searchStringEscape;
  private final String catalogSeparator;
  private final boolean catalogAtStart;
  private final boolean usingCatalogs;
  private final String primaryDatabaseNamePartTerm;
  private final String secondaryDatabaseNamePartTerm;

  // Calendar objects are mutable, so don't try to recycle them.
  protected static final Calendar getUTCCalendar() {
    return getOffsetCalendar(ZoneOffset.UTC);
  }

  protected static final Calendar getOffsetCalendar(ZoneOffset offset) {
    Calendar ret = Calendar.getInstance(TimeZone.getTimeZone(offset), Locale.US);
    // Avoid having the Calendar's state depend on the system time.
    ret.setTimeInMillis(0);
    return ret;
  }

  /** @param connection will not be held on to, nor closed */
  public GenericJDBCBackendDialect(Connection connection) throws SQLException {
    Preconditions.checkNotNull(connection);

    {
      DatabaseMetaData metaData = connection.getMetaData();
      identifierQuoteString = StringUtil.trimmedAndEmptyIfNull(metaData.getIdentifierQuoteString());
      searchStringEscape = StringUtil.trimmedAndEmptyIfNull(metaData.getSearchStringEscape());
      String toCatalogSeparator = StringUtil.trimmedAndEmptyIfNull(metaData.getCatalogSeparator());
      catalogSeparator = toCatalogSeparator.isEmpty() ? "." : toCatalogSeparator;
      catalogAtStart = metaData.isCatalogAtStart();
      usingCatalogs = metaData.supportsCatalogsInTableDefinitions();
      final String catalogTerm = usingCatalogs
          ? StringUtil.trimmedAndEmptyIfNull(metaData.getCatalogTerm()) : "";
      final String schemaTerm = metaData.supportsSchemasInTableDefinitions()
          ? StringUtil.trimmedAndEmptyIfNull(metaData.getSchemaTerm()) : "";
      primaryDatabaseNamePartTerm   = usingCatalogs && catalogAtStart ? catalogTerm : schemaTerm;
      secondaryDatabaseNamePartTerm = usingCatalogs && catalogAtStart ? schemaTerm : catalogTerm;
    }

    registerFunction(Function.PARENTHESES, IDENTITY_FUNCTION);
    registerFunction(Function.TOTALLY_ORDERED_COLLATION, IDENTITY_FUNCTION);

    /* I considered using JDBC escapes to register certain functions, but decided against it. I
    don't think they're frequently used, so there's a good chance JDBC drivers will either not
    support them or have bugs relating to their use. See
    http://docs.oracle.com/cd/E13157_01/wlevs/docs30/jdbc_drivers/sqlescape.html . */

    registerFunction(Function.UNARY_PLUS, "+[1]");
    registerFunction(Function.UNARY_MINUS, "-[1]");

    registerFunction(Function.MULTIPLY, "[1] * [2]");
    /* I considered disabling ARITHABORT/ANSI_WARNINGS or a dialect-specific equivalent instead.
    These pages seem to discourage it, however:
      http://www.sommarskog.se/query-plan-mysteries.html
      http://msdn.microsoft.com/en-us/library/ms190306.aspx */
    /* See http://www.postgresql.org/docs/9.4/static/sql-expressions.html on safe ways to write
    expressions to avoid runtime errors. */
    registerFunction(Function.DIVIDE, "[1] / nullif([2], 0)");
    registerFunction(Function.PLUS, "[1] + [2]");
    registerFunction(Function.MINUS, "[1] - [2]");

    registerFunction(Function.POWER,
        "CASE sign([1]) " +
        "WHEN 1 THEN power([1], [2]) " +
        "WHEN -1 THEN (CASE WHEN [2] = floor([2]) THEN power([1], [2]) ELSE NULL END) " +
        "ELSE (CASE WHEN [2] >= 0 THEN power([1], [2]) ELSE NULL END) " +
        "END");
    registerFunction(Function.CONDITIONAL, new FunctionImplementation(3) {
      @Override
      protected RenderedExpression renderImpl(List<RenderedExpression> arguments) throws
          DynamicRenderFunctionException
      {
        /* The conditional function is a little tricky to implement due to the various possible NULL
        input cases and their interaction with backend data types. */

        RenderedExpression condition = arguments.get(0);
        RenderedExpression ifTrueSelectable = arguments.get(1).toSelectable();
        RenderedExpression ifFalseSelectable = arguments.get(2).toSelectable();
        boolean conditionIsNull = condition.getVariant() instanceof NullVariant;
        boolean ifTrueIsNull = ifTrueSelectable.getVariant() instanceof NullVariant;
        boolean ifFalseIsNull = ifFalseSelectable.getVariant() instanceof NullVariant;
        /* CASE might not allow every return argument to be null, so handle this case specially.
        E.g. "You cannot specify the literal NULL for every return_expr and the else_expr."
        ( https://docs.oracle.com/cd/B19306_01/server.102/b14200/expressions004.htm ) */
        /* Also avoid the CASE statement if the condition is null, to avoid problems with type
        mismatch between the condition and the boolean literals. */
        if (ifTrueIsNull && ifFalseIsNull || conditionIsNull)
          return getNullLiteral();
        final RenderVariant outputVariant;
        if (ifTrueIsNull) {
          outputVariant = ifFalseSelectable.getVariant();
        } else if (ifFalseIsNull) {
          outputVariant = ifTrueSelectable.getVariant();
        } else {
          outputVariant = ifTrueSelectable.getVariant();
          if (!ifFalseSelectable.getVariant().equals(outputVariant)) {
            /* These errors may also be caused by a type error in the Expression to be rendered; the
            wording chosen here assumes that such errors have been caught by a higher-level type
            system. */
            throw new DynamicRenderFunctionException("Inconsistent input render variants; " +
                "BackendDialect may need a custom implementation of the conditional function");
          }
        }
        /* If one of the alternatives are null, make sure to use an untyped NULL literal to make the
        case statement legal regardless of the type of the other non-null alternative. */
        SQLFragment untypedNullLiteral = getNullLiteral().toPreferred().getFragment();
        SQLFragment ifTrueFragment = ifTrueIsNull ? untypedNullLiteral :
            ifTrueSelectable.getFragment().enclose("(", ")");
        SQLFragment ifFalseFragment = ifFalseIsNull ? untypedNullLiteral :
            ifFalseSelectable.getFragment().enclose("(", ")");
        if (USE_VALUE_BASED_CONDITIONAL_IMPLEMENTATION) {
          /* This implementation avoids repeating the boolean condition in the CASE statement. On
          Oracle, however, the condition ends up being repeated by condition.toSelectable(). */
          if (ImmutableSet.of(
              getBooleanLiteral(true).toSelectable().getVariant(),
              getBooleanLiteral(false).toSelectable().getVariant(),
              condition.toSelectable().getVariant()).size() != 1)
          {
            throw new DynamicRenderFunctionException("Inconsistent boolean render variants; " +
                "BackendDialect may need a custom implementation of the conditional function");
          }
          SQLFragment conditionFragment = condition.toSelectable().getFragment().enclose("(", ")");
          return new RenderedExpression(SQLFragment.builder()
            .appendJDBCString("CASE ")
            .appendSQLFragment(conditionFragment)
            .appendJDBCString(" WHEN ")
            .appendSQLFragment(getBooleanLiteral(true).toSelectable().getFragment())
            .appendJDBCString(" THEN ")
            .appendSQLFragment(ifTrueFragment)
            .appendJDBCString(" WHEN ")
            .appendSQLFragment(getBooleanLiteral(false).toSelectable().getFragment())
            .appendJDBCString(" THEN ")
            .appendSQLFragment(ifFalseFragment)
            .appendJDBCString(" ELSE null END")
            .build(), outputVariant);
        } else {
          SQLFragment conditionFragment = condition.toConditional().getFragment().enclose("(", ")");
          return new RenderedExpression(SQLFragment.builder()
            .appendJDBCString("CASE WHEN ")
            .appendSQLFragment(conditionFragment)
            .appendJDBCString(" THEN ")
            .appendSQLFragment(ifTrueFragment)
            .appendJDBCString(" WHEN NOT ")
            .appendSQLFragment(conditionFragment)
            .appendJDBCString(" THEN ")
            .appendSQLFragment(ifFalseFragment)
            .appendJDBCString(" ELSE null END")
            .build(), outputVariant);
        }
      }
    });

    registerFunction(Function.LN, "CASE WHEN [1] <= 0 THEN null ELSE LN([1]) END");

    registerFunction(Function.GREATER_OR_EQUAL, "[1] >= [2]");
    registerFunction(Function.LESS_OR_EQUAL, "[1] <= [2]");
    registerFunction(Function.GREATER, "[1] > [2]");
    registerFunction(Function.LESS, "[1] < [2]");
    registerFunction(Function.IS, "[1] IS NOT DISTINCT FROM [2]");
    registerFunction(Function.IS_NOT, "[1] IS DISTINCT FROM [2]");

    registerFunction(Function.EQUAL, "[1] = [2]");
    // Some databases support either "!=" or "<>". The latter is the ANSI-compliant variant.
    registerFunction(Function.NOT_EQUAL, "[1] <> [2]");
    registerFunction(Function.LOGICAL_AND, "[1] AND [2]");
    registerFunction(Function.LOGICAL_OR, "[1] OR [2]");
    registerFunction(Function.LOGICAL_NOT, "NOT [1]");

    /* TODO[pd]: Consider using ISNULL instead of COALESCE on SQL Server. See
                 http://msdn.microsoft.com/en-us/library/ms190349.aspx . */
    registerFunction(Function.SUM, "COALESCE(SUM([1]), 0)");

    registerFunction(Function.AVERAGE, "AVG([1])");
    registerFunction(Function.STDEV, "STDDEV([1])");
    registerFunction(Function.MIN, "MIN([1])");
    registerFunction(Function.MAX, "MAX([1])");
    registerFunction(Function.COUNT, "COUNT([1])");

    /* TODO[pd]: Consider implementing CONCATENATE(a1 || ... || aN) at this level if it turns out to
                 be a standard SQL function. */
    registerFunction(Function.STRING_LOWER, "LOWER([1])");
    registerFunction(Function.STRING_UPPER, "UPPER([1])");
    registerFunction(Function.STRING_TRIM, "TRIM([1])");
    registerFunction(Function.STRING_LENGTH, "CHAR_LENGTH([1])");
    registerFunction(Function.STRING_CONTAINS, "POSITION([2] IN [1]) > 0");
    registerFunction(Function.STRING_STARTSWITH, "POSITION([2] IN [1]) = 1");
    registerFunction(Function.STRING_POSITION, "NULLIF(POSITION([2] in [1]), 0)");
    registerFunction(Function.STRING_LEFT, "LEFT([1], cast(floor([2]) as integer))");
  }

  @Override
  public String getDisplayName() {
    return "Generic JDBC Backend";
  }

  @Override
  public String toString() {
    return getDisplayName();
  }

  @Override
  public SQLFragment getSingleRowDummyTable() {
    return SQLFragment.create("(SELECT 0 AS dcol)");
  }

  @Override
  public boolean supportsNullsFirstLast() {
    return false;
  }

  @Override
  public boolean supportsNanosecondTemporals() {
    return true;
  }

  @Override
  public String getRowIdentifierPseudoColumnName() {
    return "";
  }

  @Override
  public String getIdentifierQuoteString() {
    return identifierQuoteString;
  }

  @Override
  public String getSearchStringEscape() {
    return searchStringEscape;
  }

  @Override
  public String getCatalogSeparator() {
    return catalogSeparator;
  }

  @Override
  public boolean isCatalogAtStart() {
    return catalogAtStart;
  }

  @Override
  public boolean isUsingCatalogs() {
    return usingCatalogs;
  }

  @Override
  public String getDatabaseNameTerm(boolean primaryPart) {
    return primaryPart ? primaryDatabaseNamePartTerm : secondaryDatabaseNamePartTerm;
  }

  @Override
  public RenderedExpression getNullLiteral() {
    return LITERAL_NULL;
  }

  @Override
  public RenderedExpression getBooleanLiteral(boolean value) {
    return value ? LITERAL_TRUE : LITERAL_FALSE;
  }

  /* Temporal literals seem to be treated as string-typed expressions until they are used in a
  context that require them to be treated as temporals, both on MySQL and PostgreSQL. See for
  instance https://dev.mysql.com/doc/refman/5.0/en/date-and-time-literals.html : "MySQL ignores the
  type keyword and each of the preceding constructions produces the string value 'str', with a type
  of VARCHAR." Force the correct types by means of a cast. */
  private final ImmutableMap<Class<?>,SQLType> CASTABLE_LITERALS =
      ImmutableMap.<Class<?>,SQLType>builder()
      .put(LocalDate.class, new SQLType(SQLTypeCode.create(Types.DATE), "", 0, 0, 0, true))
      .put(LocalTime.class, new SQLType(SQLTypeCode.create(Types.TIME), "", 0, 0, 0, true))
      .put(Instant.class, new SQLType(
          SQLTypeCode.create(Types.TIMESTAMP_WITH_TIMEZONE), "", 0, 0, 0, true))
      .build();

  @Override
  public RenderedExpression getLiteral(Expression.ConstantExpression exp) {
    final Object obj = exp.getValue();
    if (obj == null) {
      /* Render null literals directly rather than passing them as query parameters, as untyped
      null parameters are not supported by every JDBC driver. SQLFragment expressly forbids null
      parameters for the same reason. */
      return getNullLiteral();
    } else if (obj instanceof Boolean) {
      /* Render boolean constants directly rather than passing them as query parameters, as
      their value is likely to be important to backend query optimization logic that might be
      applied on the compiled or cached query prior to setting of parameters. */
      return getBooleanLiteral((Boolean) obj);
    }
    SQLFragment ret = SQLFragment.createParameter(obj);
    final SQLType castableType = CASTABLE_LITERALS.get(obj.getClass());
    if (castableType != null) {
      String castTypeName = getTypeName(castableType);
      if (!castTypeName.isEmpty())
        ret = ret.enclose("cast(", " as " + castTypeName + ")");
    }
    return RenderedExpression.simple(ret);
  }

  @Override
  public RenderVariant getStorageVariant() {
    return RenderVariant.DEFAULT;
  }

  /**
   * Implementation that handles the required conversion of temporal types specified in the
   * superclass.
   */
  @Override
  public void bindParameterValue(
      PreparedStatement statement, int parameterIndex, @Nonnull Object value)
      throws SQLException
  {
    if (value instanceof LocalDate) {
      final LocalDate localDate = (LocalDate) value;
      // Don't use java.sql.Date.valueOf(LocalDate); it depends on the system time zone setting.
      final Calendar cal = getUTCCalendar();
      cal.set(Calendar.YEAR, localDate.getYear());
      cal.set(Calendar.MONTH, localDate.getMonthValue() - 1);
      cal.set(Calendar.DAY_OF_MONTH, localDate.getDayOfMonth());
      final java.sql.Date sqlDate = new java.sql.Date(cal.getTimeInMillis());
      statement.setDate(parameterIndex, sqlDate, getUTCCalendar());
    } else if (value instanceof LocalTime) {
      // Use setTimestamp rather than setTime in order to properly encode sub-millisecond fractions.
      final LocalTime localTime = (LocalTime) value;
      final java.sql.Timestamp sqlTimestamp = java.sql.Timestamp.from(
          localTime.atOffset(ZoneOffset.UTC).atDate(LocalDate.ofEpochDay(0)).toInstant());
      statement.setTimestamp(parameterIndex, sqlTimestamp, getUTCCalendar());
    } else if (value instanceof Instant) {
      final java.sql.Timestamp sqlTimestamp;
      try {
        sqlTimestamp = java.sql.Timestamp.from((Instant) value);
      } catch (IllegalArgumentException e) {
        throw new SQLException(
            "Instant " + value + " out of bounds for representation as a java.sql.Timestamp", e);
      }
      statement.setTimestamp(parameterIndex, sqlTimestamp, getUTCCalendar());
    } else {
      statement.setObject(parameterIndex, value);
    }
  }

  /**
   * Implementation that handles the required conversion of temporal types specified in the
   * superclass, and proper retrieval of numeric types. Overriding methods can take advantage of
   * this logic by calling the super implementation.
   */
  @Override
  public @Nullable Object fetchResultValue(
      ResultSet resultSet, int columnIndex, SQLType jdbcColumnType, RenderVariant renderVariant)
      throws SQLException
  {
    final int jdbcType = jdbcColumnType.getTypeCode().getCode();
    if (NUMERIC_SQL_TYPES.contains(jdbcType)) {
      if (jdbcType == java.sql.Types.REAL ||
          jdbcType == java.sql.Types.FLOAT ||
          jdbcType == java.sql.Types.DOUBLE)
      {
        /* Handle non-finite floating-point values (NaN/Infinity) separately, since they cannot be
        represented by BigDecimal. */
        double doubleVal = resultSet.getDouble(columnIndex);
        if (!Double.isFinite(doubleVal))
          return resultSet.wasNull() ? null : doubleVal;
      }
      return resultSet.getBigDecimal(columnIndex);
    } else if (jdbcType == java.sql.Types.DATE) {
      final java.sql.Date ret = resultSet.getDate(columnIndex, getUTCCalendar());
      if (ret == null) {
        return null;
      } else {
        // Don't use java.sql.Date.toLocalDate(); it depends on the system time zone setting.
        final Calendar cal = getUTCCalendar();
        cal.setTime(ret);
        return LocalDate.of(
            cal.get(Calendar.YEAR), cal.get(Calendar.MONTH) + 1, cal.get(Calendar.DAY_OF_MONTH));
      }
    } else if (jdbcType == java.sql.Types.TIME) {
      // Use getTimestamp rather than getTime to get access to sub-millisecond fractions.
      final java.sql.Timestamp ret = resultSet.getTimestamp(columnIndex, getUTCCalendar());
      if (ret == null) {
        return null;
      } else {
        return ret.toInstant().atOffset(ZoneOffset.UTC).toLocalTime();
      }
    } else if (
        jdbcType == java.sql.Types.TIMESTAMP_WITH_TIMEZONE || jdbcType == java.sql.Types.TIMESTAMP)
    {
      final java.sql.Timestamp ret = resultSet.getTimestamp(columnIndex, getUTCCalendar());
      if (ret == null) {
        return ret;
      } else {
        /* It's fine to use java.sql.Timestamp.toInstant(); it does not depend on time zone
        settings. */
        final Instant instant = ret.toInstant();
        if (ATTEMPT_CALENDAR_DEPENDENCE_DETECTION) {
          /* If the fetched Timestamp depends on the Calendar parameter, assume the time zone is
          actually unknown, and construct a LocalDateTime instead of an Instant. */
          if (!ret.equals(resultSet.getTimestamp(columnIndex,
                  getOffsetCalendar(ZoneOffset.ofHours(+4)))) ||
              !ret.equals(resultSet.getTimestamp(columnIndex,
                  getOffsetCalendar(ZoneOffset.ofHours(-4)))))
          {
            return instant.atZone(ZoneOffset.UTC).toLocalDateTime();
          } else {
            return instant;
          }
        } else {
          return instant;
        }
      }
    } else if (jdbcType == java.sql.Types.CLOB) {
      final Clob ret = resultSet.getClob(columnIndex);
      if (ret == null)
        return ret;
      try {
        final long length = ret.length();
        if (length == 0)
          return "";
        // First character is position 1.
        final String s = ret.getSubString(1, (int) Math.min(length, CLOB_SAMPLE_CHARACTER_LIMIT));
        if (length > CLOB_SAMPLE_CHARACTER_LIMIT) {
          return new ClobSample(length, s);
        } else {
          return s;
        }
      } finally {
        ret.free();
      }
    } else {
      return resultSet.getObject(columnIndex);
    }
  }

  @Override
  public String getJDBCStringWithLimit(SQLFragment query) {
    return query.getJDBCString() + "\nLIMIT ?";
  }

  @Override
  public boolean supportsGetBestRowIdentifier() {
    return true;
  }

  @Override
  public boolean supportsIdentifierQuoteInTableName() {
    return true;
  }

  @Override
  public boolean allowsAliasInGroupBy() {
    /* Go with the conservative choice; setting this to true is mostly useful for keeping queries
    shorter. */
    return false;
  }

  @Override
  public String getTypeName(SQLType type) {
    return "";
  }

  @Override
  public @Nullable String getCreateTableStatement(boolean temporary, boolean view) {
    return "CREATE " +
        (temporary ? "TEMPORARY " : "") +
        (view ? "VIEW" : "TABLE");
  }

  @Override
  public @Nullable BeforeOrAfter getIfExistsLocationForDropTable() {
    return null;
  }

  @Override
  public String getCascadeStringForDropTable() {
    return "CASCADE";
  }
}
