// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.SQLTypeCode;
import com.sieuferd.util.StringUtil;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Types;
import javax.annotation.Nullable;
import org.hibernate.MappingException;
import org.hibernate.dialect.Dialect;

// TODO[pd]: Implement more specific versions of the methods defined by GenericJDBCBackendDialect.
public class HibernateBackendDialect extends GenericJDBCBackendDialect {
  private final Dialect hibernateDialect;

  public HibernateBackendDialect(Connection connection, Dialect hibernateDialect)
      throws SQLException
  {
    super(connection);
    Preconditions.checkNotNull(hibernateDialect);
    this.hibernateDialect = hibernateDialect;
  }

  @Override
  public String getDisplayName() {
    return "Hibernate/" + hibernateDialect.getClass().getSimpleName();
  }

  @Override
  public String getTypeName(SQLType type) {
    final String ret = getTypeNameInternal(type);
    if (ret.isEmpty()) {
      /* NUMERIC and DECIMAL are interchangeable; the latter simply allows the actual stored
      precision to be larger than the specified one. See
      http://stackoverflow.com/questions/759401 . */
      if (type.getTypeCode().getCode() == Types.DECIMAL) {
        return getTypeNameInternal(new SQLType(SQLTypeCode.create(Types.NUMERIC),
            "", type.getScale(), type.getPrecision(), type.getLength(), type.getNullable()));
      } else if (type.getTypeCode().getCode() == Types.NUMERIC) {
        return getTypeNameInternal(new SQLType(SQLTypeCode.create(Types.DECIMAL),
            "", type.getScale(), type.getPrecision(), type.getLength(), type.getNullable()));
      }
    }
    return ret;
  }

  private String getTypeNameInternal(SQLType type) {
    String typePart = null;
    if (type.getLength() > 0 || type.getPrecision() > 0) {
      try {
        /* Enum types originating from MySQL have Types.CHAR with a length of zero and a precision
        of 2. Account for this and possible similar cases here. */
        final int useLength = (type.getTypeCode().getCode() == Types.CHAR)
            ? Math.max(type.getLength(), type.getPrecision())
            : type.getLength();
        /* Deal with a SQLType(typeCode=DECIMAL, typeName=NUMBER, s=-127, p=0, l=22, nullable=true)
        case that was observed on Oracle. On Oracle, this corresponds to the "NUMBER" type name with
        no arguments, which means exact decimal floating-point storage. Use a default of 38 when
        converting to other backends, since that's the precision for which Oracle guarantees
        portability.

          http://docs.oracle.com/cd/B28359_01/server.111/b28318/datatype.htm#CNCPT313
        */
        final int usePrecision =
            type.getPrecision() <= 0 &&
            (type.getTypeCode().getCode() == Types.DECIMAL ||
            type.getTypeCode().getCode() == Types.NUMERIC)
            ? 38 : type.getPrecision();
        /* A majority of Hibernate dialects seem to register Types.CHAR only as a parameterless
        single-character type. */
        final int useTypeCode = (type.getTypeCode().getCode() == Types.CHAR && useLength > 1)
            ? Types.VARCHAR : type.getTypeCode().getCode();
        /* On Oracle, "If you specify a negative scale, Oracle Database rounds the actual data to
        the specified number of places to the left of the decimal point." Limiting to zero should be
        fine in such cases.
        http://docs.oracle.com/cd/B28359_01/server.111/b28318/datatype.htm#CNCPT313
        */
        final int useScale = Math.max(type.getScale(), 0);

        typePart = hibernateDialect.getTypeName(useTypeCode, useLength, usePrecision, useScale);
      } catch (MappingException e) {
        // No action.
      }
    }
    if (typePart == null) {
      try {
        typePart = hibernateDialect.getTypeName(type.getTypeCode().getCode());
      } catch (MappingException e) {
        // No action.
      }
    }
    if (typePart == null)
      return "";
    /* Leftover Hibernate type string template parameters could indicate some kind of problem, e.g.
    related to using the wrong variant of getTypeName. When I last looked at the Hibernate source
    code, the latter case should throw a MappingException, though. */
    if (typePart.contains("$l") || typePart.contains("$p") || typePart.contains("$s")) {
      return "";
    } else {
      String nullString = Boolean.FALSE.equals(type.getNullable())
          ? "NOT NULL"
          : hibernateDialect.getNullColumnString().trim();
      return typePart + (nullString.isEmpty() ? "" : (" " + nullString));
    }
  }

  @Override
  public @Nullable BeforeOrAfter getIfExistsLocationForDropTable() {
    if (hibernateDialect.supportsIfExistsBeforeTableName()) {
      return BeforeOrAfter.BEFORE;
    } else if (hibernateDialect.supportsIfExistsAfterTableName()) {
      return BeforeOrAfter.AFTER;
    } else {
      return null;
    }
  }

  @Override
  public String getCascadeStringForDropTable() {
    String cascadeConstraints =
        StringUtil.trimmedAndEmptyIfNull(hibernateDialect.getCascadeConstraintsString());
    if (!cascadeConstraints.isEmpty())
      return cascadeConstraints;
    if (!hibernateDialect.supportsCascadeDelete())
      return "";
    return super.getCascadeStringForDropTable();
  }
}
