// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import java.util.Objects;
import javax.annotation.concurrent.Immutable;

/**
 * Wrapper that can be used to ensure that special result objects returned by JDBC drivers can be
 * compared using {@code hashCode} and {@code equals}.
 */
@Immutable
public final class JDBCObjectWrapper {
  private final Class<?> delegateClass;
  private final String stringRepresentation;

  public JDBCObjectWrapper(Class<?> delegateClass, String stringRepresentation) {
    Preconditions.checkNotNull(delegateClass);
    Preconditions.checkNotNull(stringRepresentation);
    Preconditions.checkArgument(!delegateClass.equals(String.class));
    this.delegateClass = delegateClass;
    this.stringRepresentation = stringRepresentation;
  }

  public JDBCObjectWrapper(Object delegate) {
    this(delegate.getClass(), delegate.toString());
  }

  @Override
  public String toString() {
    return stringRepresentation;
  }

  public Class<?> getDelegateClass() {
    return delegateClass;
  }

  @Override
  public int hashCode() {
    return Objects.hash(delegateClass, stringRepresentation);
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof JDBCObjectWrapper))
      return false;
    final JDBCObjectWrapper other = (JDBCObjectWrapper) obj;
    return delegateClass.equals(other.delegateClass)
        && stringRepresentation.equals(other.stringRepresentation);
  }
}
