// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLFragmentBuilder;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.dialect.RenderVariant.NullVariant;
import com.sieuferd.datamodel.expressions.Function;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import org.hibernate.dialect.Dialect;

public class OracleBackendDialect extends HibernateBackendDialect {
  private static final RenderedExpression LITERAL_NULL = new NullVariant(
      SQLFragment.create("null"), SQLFragment.create("to_char(null)"),
      SQLFragment.create("1 = null"))
      .getNullLiteral();
  private static final RenderedExpression LITERAL_TRUE =
      new RenderedExpression(SQLFragment.create("1"), ValueBooleanVariant.INSTANCE);
  private static final RenderedExpression LITERAL_FALSE =
      new RenderedExpression(SQLFragment.create("0"), ValueBooleanVariant.INSTANCE);

  public OracleBackendDialect(Connection connection, Dialect hibernateDialect) throws SQLException {
    super(connection, hibernateDialect);
    registerFunction(Function.create("current_date"), "current_date");
    registerFunction(Function.create("current_timestamp"), "current_timestamp");
    registerFunction(Function.STRING_LENGTH, "LENGTH([1])");

    registerFunction(Function.STRING_CONCAT, new FunctionImplementation(null) {
      @Override
      protected RenderedExpression renderImpl(List<RenderedExpression> arguments) {
        /* The CONCAT function always takes exactly two arguments only, so use the || operator
        instead. See https://docs.oracle.com/cd/B19306_01/server.102/b14200/functions026.htm . The
        caveat mentioned at https://docs.oracle.com/cd/B19306_01/server.102/b14200/operators003.htm
        ("some IBM platforms use broken vertical bars...") is irrelevant here; they're talking
        about encoding of SQL scripts, not data. */
        SQLFragmentBuilder ret = SQLFragment.builder();
        long i = 0;
        /* Ensure a proper constant result in the no-arguments case, and the casting effect of the
        concatenation operator in the one-argument case. */
        if (arguments.size() < 2) {
          ret.appendJDBCString("''");
          i++;
        }
        for (RenderedExpression argument : arguments) {
          if (i++ > 0)
            ret.appendJDBCString(" || ");
          ret.appendSQLFragment(argument.toSelectable().getFragment());
        }
        return new RenderedExpression(ret.build(), RenderVariant.DEFAULT);
      }
    });

    final RenderVariant CB = ConditionalBooleanVariant.INSTANCE;
    registerFunction(Function.GREATER_OR_EQUAL, CB, "[1] >= [2]");
    registerFunction(Function.LESS_OR_EQUAL, CB, "[1] <= [2]");
    registerFunction(Function.GREATER, CB, "[1] > [2]");
    registerFunction(Function.LESS, CB, "[1] < [2]");
    registerFunction(Function.LOGICAL_AND, CB, "[c1] AND [c2]");
    registerFunction(Function.LOGICAL_OR, CB, "[c1] OR [c2]");
    registerFunction(Function.LOGICAL_NOT, CB, "NOT [c1]");

    registerFunction(Function.COUNT, "COUNT([s1])");
    registerFunction(Function.STRING_LEFT, "SUBSTR([1], 0, cast(floor([2]) as integer))");

    // For EQUAL, NOT_EQUAL, IS, and IS_NOT, arguments must be casted toSelectable().
    registerFunction(Function.EQUAL, new FunctionImplementation(2) {
      @Override
      protected RenderedExpression renderImpl(List<RenderedExpression> arguments) {
        return new RenderedExpression(SQLFragment.builder()
            .appendSQLFragment(arguments.get(0).toSelectable().getFragment().enclose("(", ")"))
            .appendJDBCString(" = ")
            .appendSQLFragment(arguments.get(1).toSelectable().getFragment().enclose("(", ")"))
            .build(),
            CB);
      }
    });

    registerFunction(Function.NOT_EQUAL, new FunctionImplementation(2) {
      @Override
      protected RenderedExpression renderImpl(List<RenderedExpression> arguments) {
        return new RenderedExpression(SQLFragment.builder()
            .appendSQLFragment(arguments.get(0).toSelectable().getFragment().enclose("(", ")"))
            .appendJDBCString(" <> ")
            .appendSQLFragment(arguments.get(1).toSelectable().getFragment().enclose("(", ")"))
            .build(),
            CB);
      }
    });

    registerFunction(Function.IS, new FunctionImplementation(2) {
      @Override
      protected RenderedExpression renderImpl(List<RenderedExpression> arguments) {
        // See https://docs.oracle.com/cd/B19306_01/server.102/b14200/functions040.htm
        return new RenderedExpression(SQLFragment.builder()
            .appendJDBCString("DECODE(")
            .appendSQLFragment(arguments.get(0).toSelectable().getFragment())
            .appendJDBCString(", ")
            .appendSQLFragment(arguments.get(1).toSelectable().getFragment())
            .appendJDBCString(", 1, 0)")
            .build(),
            ValueBooleanVariant.INSTANCE);
      }
    });
    registerFunction(Function.IS_NOT, new FunctionImplementation(2) {
      @Override
      protected RenderedExpression renderImpl(List<RenderedExpression> arguments) {
        // Like IS, but with the last two arguments inverted.
        return new RenderedExpression(SQLFragment.builder()
            .appendJDBCString("DECODE(")
            .appendSQLFragment(arguments.get(0).toSelectable().getFragment())
            .appendJDBCString(", ")
            .appendSQLFragment(arguments.get(1).toSelectable().getFragment())
            .appendJDBCString(", 0, 1)")
            .build(),
            ValueBooleanVariant.INSTANCE);
      }
    });
    registerFunction(
        Function.STRING_CONTAINS, ConditionalBooleanVariant.INSTANCE, "INSTR([1], [2]) > 0");
    registerFunction(
        Function.STRING_STARTSWITH, ConditionalBooleanVariant.INSTANCE, "INSTR([1], [2]) = 1");

    registerFunction(Function.STRING_POSITION, "NULLIF(INSTR([1], [2]), 0)");

    registerFunction(Function.STRING_NORMALIZE_WHITESPACE,
        "regexp_replace(trim([1]), '[[^[[:graph:]]]]+', ' ')");
  }

  @Override
  public String getDisplayName() {
    return "Oracle";
  }

  @Override
  public SQLFragment getSingleRowDummyTable() {
    return SQLFragment.create("dual");
  }

  @Override
  public String getRowIdentifierPseudoColumnName() {
    return "ROWID";
  }

  @Override
  public boolean supportsNullsFirstLast() {
    return true;
  }

  @Override
  public RenderedExpression getNullLiteral() {
    return LITERAL_NULL;
  }

  @Override
  public RenderedExpression getBooleanLiteral(boolean value) {
    return value ? LITERAL_TRUE : LITERAL_FALSE;
  }

  @Override
  public String getJDBCStringWithLimit(SQLFragment query) {
    // See org.hibernate.dialect.Oracle9iDialect.getLimitString.
    StringBuilder ret = new StringBuilder();
    ret.append("SELECT * FROM (");
    ret.append(query.getJDBCString());
    ret.append(") WHERE rownum <= ?");
    return ret.toString();
  }

  @Override
  public boolean supportsIdentifierQuoteInTableName() {
    return false;
  }

  @Override
  public Object fetchResultValue(
      ResultSet resultSet, int columnIndex, SQLType jdbcColumnType, RenderVariant renderVariant)
      throws SQLException
  {
    if (renderVariant instanceof ValueBooleanVariant) {
      final boolean ret = resultSet.getBoolean(columnIndex);
      return resultSet.wasNull() ? null : ret;
    } else {
      if ("timestamp with time zone".equalsIgnoreCase(jdbcColumnType.getTypeName())) {
        /* In Oracle, TIMESTAMP WITH TIME ZONE values include an actual stored time zone tag.
        Because java.sql.Timestamp cannot represent the entire value including the time zone tag,
        Oracle's JDBC driver encodes such timestamps as strings by default. We override this and
        retrieve the value as a java.sql.Timestamp anyway; the resulting time value will be correct,
        and will make more sense as a value to filter and do calculations on. The downside is that
        it is no longer possible to see what time zone the value was originally stored in. */
        java.sql.Timestamp ret = resultSet.getTimestamp(columnIndex, getUTCCalendar());
        return ret == null ? null : ret.toInstant();
      } else {
        final Object ret =
            super.fetchResultValue(resultSet, columnIndex, jdbcColumnType, renderVariant);
        if (ret != null && "ROWID".equals(jdbcColumnType.getTypeName())) {
          return new JDBCObjectWrapper(ret);
        } else {
          return ret;
        }
      }
    }
  }

  /** Boolean expressions represented as an Oracle "conditional expression". */
  private static final class ConditionalBooleanVariant extends RenderVariant {
    static final RenderVariant INSTANCE = new ConditionalBooleanVariant();

    @Override
    public RenderedExpression toPreferred(SQLFragment fragment) {
      return new RenderedExpression(fragment, INSTANCE);
    }

    @Override
    public RenderedExpression toSelectable(SQLFragment fragment) {
      return new RenderedExpression(SQLFragment.builder()
          .appendJDBCString("CASE WHEN ")
          .appendSQLFragment(fragment.enclose("(", ")"))
          .appendJDBCString(" THEN 1 WHEN NOT ")
          .appendSQLFragment(fragment.enclose("(", ")"))
          .appendJDBCString(" THEN 0 ELSE null END")
          .build(), ValueBooleanVariant.INSTANCE);
    }

    @Override
    public RenderedExpression toConditional(SQLFragment fragment) {
      return toPreferred(fragment);
    }

    @Override
    public RenderedExpression toOrderable(SQLFragment fragment) {
      return toSelectable(fragment);
    }
  }

  /** Boolean expressions represented by the integer values 1 or 0. */
  private static final class ValueBooleanVariant extends RenderVariant {
    static final RenderVariant INSTANCE = new ValueBooleanVariant();

    @Override
    public RenderedExpression toPreferred(SQLFragment fragment) {
      return new RenderedExpression(fragment.enclose("(", ") <> 0"),
          ConditionalBooleanVariant.INSTANCE);
    }

    @Override
    public RenderedExpression toSelectable(SQLFragment fragment) {
      return new RenderedExpression(fragment, INSTANCE);
    }

    @Override
    public RenderedExpression toConditional(SQLFragment fragment) {
      return toPreferred(fragment);
    }

    @Override
    public RenderedExpression toOrderable(SQLFragment fragment) {
      return toSelectable(fragment);
    }
  }
}
