// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.sieuferd.database.SQLType;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLFragmentBuilder;
import com.sieuferd.database.dialect.RenderVariant.NullVariant;
import com.sieuferd.database.dialect.FunctionImplementation.SimpleFunctionImplementation;
import com.sieuferd.datamodel.expressions.Function;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.List;
import javax.annotation.Nullable;
import org.hibernate.dialect.Dialect;

public class PostgreSQLBackendDialect extends HibernateBackendDialect {
  private static final RenderedExpression LITERAL_NULL =
      new NullVariant(SQLFragment.create("NULL"), SQLFragment.create("NULL::text"))
      .getNullLiteral();
  public PostgreSQLBackendDialect(Connection connection, Dialect hibernateDialect)
      throws SQLException
  {
    super(connection, hibernateDialect);
    // TODO[pd]: Don't support this function.
    registerFunction(Function.create("current_time"), "current_time");
    registerFunction(Function.create("current_date"), "current_date");
    registerFunction(Function.create("current_timestamp"), "current_timestamp");
    registerFunction(Function.create("epochmillis"), "extract(epoch from [1])");

    /* TODO[pd]: Do the whole COALESCE business here to handle the sum of an empty set. Can this
                 actually happen for window functions? */
    registerFunction(Function.CUML_SUM, "sum([1]) OVER (PARTITION BY [2] ORDER BY [3])");

    registerFunction(Function.create("timestamp_str"), "to_char([1] at time zone 'UTC', 'YYYY-MM-DD HH24:MI:FMSS.US000Z')");
    if (false) {
      registerFunction(Function.create("extract_second"), "extract(second from [1])");
      registerFunction(Function.create("extract_minute"), "extract(minute from [1])");
      registerFunction(Function.create("extract_hour"), "extract(hour from [1])");
      registerFunction(Function.create("extract_day"), "extract(day from [1])");
      registerFunction(Function.create("extract_month"), "extract(month from [1])");
      registerFunction(Function.create("extract_year"), "extract(year from [1])");
    }
    registerFunction(Function.TEMPORAL_YEAR, "extract(year from [1])");
    registerFunction(Function.TEMPORAL_MONTH, "extract(month from [1])");
    registerFunction(Function.TEMPORAL_DAY, "extract(day from [1])");
    registerFunction(Function.TEMPORAL_HOUR, "extract(hour from [1])");
    registerFunction(Function.TEMPORAL_MINUTE, "extract(minute from [1])");
    registerFunction(Function.TEMPORAL_SECOND, "extract(second from [1])");

    // Note: Opposite parameter order.
    registerFunction(Function.GROUP_CONCAT, "string_agg(CAST(([2]) AS text), [1])");
    registerFunction(Function.STRING_CONCAT, new SimpleFunctionImplementation(null) {
      @Override
      protected SQLFragment renderFragment(List<SQLFragment> arguments) {
        if (arguments.isEmpty())
          return SQLFragment.createParameter("");
        /* Use the concat() function rather than the || operator, since the latter would leave no
        operator at all in the single-argument case, ending up with inorrect NULL handling. */
        SQLFragmentBuilder ret = SQLFragment.builder();
        int i = 0;
        for (SQLFragment argument : arguments) {
          if (i++ > 0)
            ret.appendJDBCString(", ");
          ret.appendSQLFragment(argument.enclose("CAST((", ") AS text)"));
        }
        return ret.build().enclose("concat(", ")");
      }
    });
    // Note the escaped double brackets in the template string.
    registerFunction(Function.STRING_NORMALIZE_WHITESPACE,
        "regexp_replace(trim([1]), '[[^[[:graph:]]]]+', ' ', 'g')");
  }

  @Override
  public String getDisplayName() {
    return "PostgreSQL";
  }

  @Override
  public RenderedExpression getNullLiteral() {
    return LITERAL_NULL;
  }

  @Override
  public boolean supportsNullsFirstLast() {
    return true;
  }

  @Override
  public boolean supportsNanosecondTemporals() {
    return false;
  }

  @Override
  public boolean allowsAliasInGroupBy() {
    return true;
  }

  @Override
  public String getRowIdentifierPseudoColumnName() {
    return "ctid";
  }

  @Override
  public @Nullable Object fetchResultValue(
      ResultSet resultSet, int columnIndex, SQLType jdbcColumnType, RenderVariant renderVariant)
      throws SQLException
  {
    final Object ret =
        super.fetchResultValue(resultSet, columnIndex, jdbcColumnType, renderVariant);
    // TODO[te]: Add a test case that breaks if the "ret != null" condition below is omitted.
    if (ret != null && "tid".equals(jdbcColumnType.getTypeName())) {
      return new JDBCObjectWrapper(ret);
    } else {
      return ret;
    }
  }

  @Override
  public String getTypeName(SQLType type) {
    final int typeCode = type.getTypeCode().getCode();
    /* For the "time" type, the default "without time zone" is correct. The "date" type does not
    have a time zone modifier. */
    if (typeCode == Types.TIMESTAMP_WITH_TIMEZONE) {
      return "timestamp with time zone";
    } else {
      return super.getTypeName(type);
    }
  }
}
