// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import com.sieuferd.database.SQLFragment;
import javax.annotation.concurrent.Immutable;

/* TODO[cc]: This interface is awkward to use; consider redesigning it. Consider representing casts
             with a regular BackendFunction. */

/**
 * A method for representing a value in the backend. Dialects may provide custom implementations
 * that convert expressions into different forms depending on, for instance, which part of a SQL
 * query the expression needs to be used in. Different variants should never change the NULLness of
 * a value, as the nullability of rendered expressions needs to remain consistent with that of their
 * corresponding front-end {@link com.sieuferd.datamodel.expressions.Expression}s. In some dialects,
 * only the default variant may turn out to be necessary.
 */
@Immutable
public abstract class RenderVariant {
  public static final RenderVariant DEFAULT = new SimpleVariant();

  /**
   * Convert a fragment of this variant to the one conventionally used for parameters in this
   * backend's function definitions.
   */
  public abstract RenderedExpression toPreferred(SQLFragment fragment);

  // TODO[pd]: Consider making a separate JDBC retrieval form.
  /**
   * Convert a fragment of this variant to one suitable for use in a SELECT clause, including JDBC
   * retrieval.
   */
  public abstract RenderedExpression toSelectable(SQLFragment fragment);

  /**
   * Convert a fragment of this variant to one suitable for use as the conditional argument of a
   * WHERE, HAVING, CASE, or similar clause. It is the responsibility of the caller and a
   * higher-level type system to ensure that the expression represented by the fragment is actually
   * a boolean one.
   */
  public abstract RenderedExpression toConditional(SQLFragment fragment);

  /**
   * Convert a fragment of this type to a type suitable for use in an ORDER BY clause. This is
   * usually the same as either {@link #getPreferredForm()} or {@link #getSelectableForm()}.
   */
  public abstract RenderedExpression toOrderable(SQLFragment fragment);

  @Override
  public String toString() {
    return getClass().getSimpleName();
  }

  private static final class SimpleVariant extends RenderVariant {
    @Override
    public RenderedExpression toPreferred(SQLFragment fragment) {
      return new RenderedExpression(fragment, DEFAULT);
    }

    @Override
    public RenderedExpression toSelectable(SQLFragment fragment) {
      return new RenderedExpression(fragment, DEFAULT);
    }

    @Override
    public RenderedExpression toConditional(SQLFragment fragment) {
      return new RenderedExpression(fragment, DEFAULT);
    }

    @Override
    public RenderedExpression toOrderable(SQLFragment fragment) {
      return new RenderedExpression(fragment, DEFAULT);
    }
  }

  public static final class NullVariant extends RenderVariant {
    private final RenderedExpression untypedNullExpression;
    private final RenderedExpression typedNullExpression;
    private final RenderedExpression conditionalNullExpression;

    public NullVariant(SQLFragment untypedNullFragment, SQLFragment typedNullFragment) {
      this(untypedNullFragment, typedNullFragment, untypedNullFragment);
    }

    public NullVariant(SQLFragment untypedNullFragment, SQLFragment typedNullFragment,
        SQLFragment conditionalNullFragment)
    {
      Preconditions.checkNotNull(untypedNullFragment);
      Preconditions.checkNotNull(typedNullFragment);
      this.untypedNullExpression = new RenderedExpression(untypedNullFragment, this);
      this.typedNullExpression = new RenderedExpression(typedNullFragment, this);
      this.conditionalNullExpression = new RenderedExpression(conditionalNullFragment, this);
    }

    public RenderedExpression getNullLiteral() {
      return untypedNullExpression;
    }

    @Override
    public RenderedExpression toPreferred(SQLFragment fragment) {
      return untypedNullExpression;
    }

    @Override
    public RenderedExpression toSelectable(SQLFragment fragment) {
      return typedNullExpression;
    }

    @Override
    public RenderedExpression toConditional(SQLFragment fragment) {
      return conditionalNullExpression;
    }

    @Override
    public RenderedExpression toOrderable(SQLFragment fragment) {
      return typedNullExpression;
    }
  }
}
