// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.database.dialect;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLFragmentBuilder;
import com.sieuferd.database.dialect.FunctionImplementation.RenderFunctionException.DynamicRenderFunctionException;
import com.sieuferd.util.MapUtil;
import java.util.List;
import java.util.Map;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;

// TODO[cc]: Document the format.
@Immutable
public class TemplateFunctionImplementation extends FunctionImplementation {
  private static final char START_ARG_CHAR = '[';
  private static final char END_ARG_CHAR = ']';
  private static final ImmutableMap<Character,ArgumentContext> ARGUMENT_CONTEXTS =
      createArgumentContextsMap();

  private final RenderVariant outputVariant;
  private final ImmutableList<Part> parts;

  public TemplateFunctionImplementation(
      @Nullable Integer argumentCount, String template, RenderVariant outputVariant)
  {
    super(argumentCount);
    Preconditions.checkNotNull(template);
    Preconditions.checkNotNull(outputVariant);
    this.outputVariant = outputVariant;
    this.parts = ImmutableList.copyOf(parseTemplate(template, argumentCount));
  }

  @Override
  public String toString() {
    return "TemplateFunctionImplementation(" + outputVariant + ", " + parts + ")";
  }

  private static List<Part> parseTemplate(String template, @Nullable Integer argumentCount) {
    final List<Part> ret = Lists.newArrayList();
    final StringBuilder currentJDBCString = new StringBuilder();
    for (int i = 0; i < template.length(); i++) {
      // Quoted start and end characters.
      if (i + 1 < template.length()) {
        if (template.charAt(i) == START_ARG_CHAR && template.charAt(i + 1) == START_ARG_CHAR) {
          currentJDBCString.append(START_ARG_CHAR);
          i++;
          continue;
        }
        if (template.charAt(i) == END_ARG_CHAR && template.charAt(i + 1) == END_ARG_CHAR) {
          currentJDBCString.append(END_ARG_CHAR);
          i++;
          continue;
        }
      }
      if (template.charAt(i) == END_ARG_CHAR) {
        throw new IllegalArgumentException("Stray argument end character");
      } else if (template.charAt(i) == START_ARG_CHAR) {
        if (currentJDBCString.length() > 0) {
          ret.add(new StringPart(currentJDBCString.toString()));
          currentJDBCString.setLength(0);
        }
        i++;
        int start = i;
        for (; i < template.length(); i++) {
          if (template.charAt(i) == END_ARG_CHAR)
            break;
        }
        if (i >= template.length())
          throw new IllegalArgumentException("Unterminated argument in function template");
        final ArgumentContext argumentContext;
        if (!Character.isDigit(template.charAt(start))) {
          argumentContext = MapUtil.getChecked(ARGUMENT_CONTEXTS, template.charAt(start));
          start++;
        } else {
          argumentContext = ArgumentContext.PREFERRED;
        }
        final int oneBasedArgumentIndex;
        String argumentIndexString = template.substring(start, i);
        try {
          oneBasedArgumentIndex = Integer.parseInt(argumentIndexString);
          // Don't allow leading zeroes or other weird number representations.
          if (!argumentIndexString.equals(Integer.toString(oneBasedArgumentIndex)))
            throw new NumberFormatException();
          if (oneBasedArgumentIndex <= 0)
            throw new NumberFormatException();
        } catch (NumberFormatException e) {
          throw new IllegalArgumentException("Illegal argument index '" + argumentIndexString +
              "' (expected a numeric one-based index)");
        }
        if (argumentCount != null && oneBasedArgumentIndex > argumentCount) {
          throw new IllegalArgumentException(
              "Argument index out of range: " + oneBasedArgumentIndex);
        }
        ret.add(new ArgumentPart(argumentContext, oneBasedArgumentIndex - 1));
      } else {
        currentJDBCString.append(template.charAt(i));
      }
    }
    if (currentJDBCString.length() > 0)
      ret.add(new StringPart(currentJDBCString.toString()));
    return ret;
  }

  @Override
  protected RenderedExpression renderImpl(List<RenderedExpression> arguments)
      throws DynamicRenderFunctionException
  {
    final SQLFragmentBuilder ret = SQLFragment.builder();
    for (Part part : parts)
      part.render(ret, arguments);
    return new RenderedExpression(ret.build(), outputVariant);
  }

  private interface Part {
    public void render(SQLFragmentBuilder ret, List<RenderedExpression> arguments);
  }

  private static final class StringPart implements Part {
    private final String jdbcString;

    public StringPart(String jdbcString) {
      Preconditions.checkNotNull(jdbcString);
      Preconditions.checkArgument(!jdbcString.isEmpty());
      this.jdbcString = jdbcString;
    }

    @Override
    public String toString() {
      return "S{" + jdbcString + "}";
    }

    @Override
    public void render(SQLFragmentBuilder ret, List<RenderedExpression> arguments) {
      ret.appendJDBCString(jdbcString);
    }
  }

  private static final class ArgumentPart implements Part {
    private final ArgumentContext argumentContext;
    private final int argumentIndex;

    /** @param argumentIndex zero-based index */
    public ArgumentPart(ArgumentContext argumentContext, int argumentIndex) {
      Preconditions.checkNotNull(argumentContext);
      Preconditions.checkArgument(argumentIndex >= 0);
      this.argumentIndex = argumentIndex;
      this.argumentContext = argumentContext;
    }

    @Override
    public String toString() {
      return "A{" + argumentContext + (argumentIndex + 1) + "}";
    }

    @Override
    public void render(SQLFragmentBuilder ret, List<RenderedExpression> arguments) {
      final RenderedExpression argument = arguments.get(argumentIndex);
      final RenderedExpression casted;
      switch (argumentContext) {
        case PREFERRED:
          casted = argument.toPreferred();
        break; case SELECTABLE:
          casted = argument.toSelectable();
        break; case CONDITIONAL:
          casted = argument.toConditional();
        break; case ORDERABLE:
          casted = argument.toOrderable();
        break; default:
          throw new AssertionError();
      }
      /* TODO[pd]: Consider detecting situations where the extra pair of parentheses can be omitted
                   (if the argument is already enclosed by parentheses or commas). */
      ret.appendSQLFragment(casted.getFragment().enclose("(", ")"));
    }
  }

  private static ImmutableMap<Character,ArgumentContext> createArgumentContextsMap() {
    final Map<Character,ArgumentContext> ret = Maps.newLinkedHashMap();
    for (ArgumentContext variant : ArgumentContext.values())
      MapUtil.putChecked(ret, variant.code, variant);
    return ImmutableMap.copyOf(ret);
  }

  private static enum ArgumentContext {
    PREFERRED('p'),
    SELECTABLE('s'),
    CONDITIONAL('c'),
    ORDERABLE('o');

    private final char code;

    private ArgumentContext(char code) {
      this.code = code;
    }

    @Override
    public String toString() {
      return Character.toString(code);
    }
  }
}
