// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.LinkedHashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.evaluator.impl.InternalQueryUtil;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.FunctionExpression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.ExpressionVisitor.ExpressionVisitorAdapter;
import java.util.Collections;
import java.util.Set;
import javax.annotation.concurrent.Immutable;

@Immutable
public final class FunctionalDependencies {
  private static final boolean ADD_REVERSE_JOIN_DEPENDENCIES = false;
  /**
   * It should always be safe to leave this on, but turning it off may be useful for testing
   * purposes. When this is enabled, non-plural relations will end up having no tuple keys of their
   * own.
   */
  private static final boolean INHERIT_DEPENDENCIES_FROM_PARENT = true;
  /**
   * Each entry (K,(V1,...,Vn)) in the map represents a functional dependency
   * (V1,...,Vn) -> K, that is, each key path is fully identified by each set of paths stored as
   * values in the map.
   */
  private final ImmutableMultimap<PrimitivePath,ImmutableSet<PrimitivePath>> dependencies;

  public FunctionalDependencies(View sourceView) {
    Multimap<PrimitivePath,ImmutableSet<PrimitivePath>> toDependencies =
        LinkedHashMultimap.create();
    populate(sourceView, Path.ROOT, toDependencies);
    this.dependencies = ImmutableMultimap.copyOf(toDependencies);
  }

  /** Defines CProp.IN_TUPLE_KEY and CProp.PLURAL for all applicable paths. */
  public void configureTupleKeysAndPluralities(MutableView view) {
    configureTupleKeys(view, Path.ROOT, ImmutableSet.of());
    configurePluralities(view, Path.ROOT, ImmutableSet.of());
  }

  @Override
  public String toString() {
    return dependencies.toString();
  }

  private static ImmutableSet<PrimitivePath> getPrimaryKeyPaths(View view, RelationPath path) {
    Set<PrimitivePath> ret = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(path)) {
      if (!childPath.isRelation() &&
          Queries.isInPrimaryKey(view, childPath.getPrimitive()))
      {
        ret.add(childPath.getPrimitive());
      }
    }
    return ImmutableSet.copyOf(ret);
  }

  private static ImmutableSet<PrimitivePath> getExpressionDependencies(
      Expression exp, View view, final RelationPath contextRelation) {
    final Set<PrimitivePath> ret = Sets.newLinkedHashSet();
    collectExpressionDependencies(exp, view, contextRelation, ret);
    return ImmutableSet.copyOf(ret);
  }

  private static void collectExpressionDependencies(final Expression exp,
      final View view, final RelationPath contextRelation, final Set<PrimitivePath> ret)
  {
    exp.accept(new ExpressionVisitorAdapter<RuntimeException>() {
      @Override
      public void voidVisit(FunctionExpression exp) {
        if (Function.AGGREGATE_FUNCTIONS.contains(exp.getFunction())) {
          /* If an aggregate formula field were to be included in a tuple key, then it would not be
          possible to calculate the aggregate function, and so the aggregate function would return a
          constant null value. This means that the arguments are not part of its dependencies.
          However, the GROUP BY paths are. */
          ret.addAll(InternalQueryUtil.getNaiveGroupByPaths(view, contextRelation));
        } else {
          for (Expression argument : exp.getArguments())
            collectExpressionDependencies(argument, view, contextRelation, ret);
        }
      }

      @Override
      public void voidVisit(PathExpression exp) {
        final Path referencedPath = exp.getPath();
        /* Relation paths may only be referenced from within the COUNT aggregate function, so seeing
        a relation reference here would indicate an error case with an effective constant value of
        NULL. */
        if (!referencedPath.isRelation())
          ret.add(referencedPath.getPrimitive());
      }
    });
  }

  private static void populate(View view, Path path,
      Multimap<PrimitivePath,ImmutableSet<PrimitivePath>> toDependencies)
  {
    if (path.isRelation()) {
      final ImmutableSet<PrimitivePath> primaryKeyPaths =
          getPrimaryKeyPaths(view, path.getRelation());
      for (Path childPath : view.getChildPaths(path.getRelation())) {
        /* TODO[pd]: Maybe don't bother avoiding circular dependencies here, since they must be
                     caught later anyway. */
        if (!childPath.isRelation() &&
            !view.get(QProp.SQL_COLUMN_NAME, childPath).isEmpty() &&
            !primaryKeyPaths.contains(childPath.getPrimitive()))
        {
          toDependencies.put(childPath.getPrimitive(), primaryKeyPaths);
        }
        populate(view, childPath, toDependencies);
      }
    } else {
      PrimitivePath joinedOnPath = view.get(QProp.JOINED_ON, path).getNullable();
      if (joinedOnPath != null) {
        /* "For every tuple in existence at the level of path, the value at path is identified by
        the value at joinedOnPath." This is valid even when EXISTENTIAL_FILTER is off, since missing
        values are represented by empty sets rather than null values in this direction. */
        toDependencies.put(path.getPrimitive(), ImmutableSet.of(joinedOnPath));
        if (ADD_REVERSE_JOIN_DEPENDENCIES) {
          if (view.get(QProp.EXISTENTIAL_FILTER, path.getParent()))
            toDependencies.put(joinedOnPath, ImmutableSet.of(path.getPrimitive()));
        }
      }
      if (view.get(QProp.SQL_COLUMN_NAME, path).isEmpty()) {
        toDependencies.put(path.getPrimitive(), getExpressionDependencies(
            view.get(QProp.FORMULA_EXPRESSION, path), view, path.getParent()));
      }
    }
  }

  /** @param fullyIdentifiedPaths may have additional fully identified paths added to it */
  public boolean isIdentifiedBy(View view, RelationPath contextRelation, Expression expression,
      Set<PrimitivePath> fullyIdentifiedPaths)
  {
    for (PrimitivePath path : getExpressionDependencies(expression, view, contextRelation)) {
      if (!isIdentifiedBy(view, path, fullyIdentifiedPaths))
        return false;
    }
    return true;
  }

  /** @param fullyIdentifiedPaths may have additional fully identified paths added to it */
  public boolean isIdentifiedBy(
      View view, PrimitivePath path, Set<PrimitivePath> fullyIdentifiedPaths)
  {
    if (fullyIdentifiedPaths.contains(path) || !view.exists(path)) {
      return true;
    } else {
      for (ImmutableSet<PrimitivePath> candidateKey : dependencies.get(path)) {
        boolean allKeyPathsIdentified = true;
        for (PrimitivePath keyPath : candidateKey) {
          /* Pretend that path is fully identified for the duration of the recursive call to
          isIdentifiedBy. This will yield the right behavior in the case of a circular dependency,
          for which involved paths will attain constant NULL error values and can be considered as
          fully identified. */
          final boolean absentBefore = fullyIdentifiedPaths.add(path);
          final boolean keyIsIdentified = isIdentifiedBy(view, keyPath, fullyIdentifiedPaths);
          if (absentBefore)
            fullyIdentifiedPaths.remove(path);
          if (!keyIsIdentified) {
            allKeyPathsIdentified = false;
            break;
          }
        }
        if (allKeyPathsIdentified) {
          fullyIdentifiedPaths.add(path);
          return true;
        }
      }
      return false;
    }
  }

  private void configurePluralities(
      MutableView view, RelationPath path, ImmutableSet<PrimitivePath> fullyIdentifiedPaths)
  {
    final Set<PrimitivePath> useFullyIdentifiedPaths = Sets.newLinkedHashSet(fullyIdentifiedPaths);
    final Set<PrimitivePath> locallyProjectedPaths = Sets.newLinkedHashSet();
    boolean plural = false;
    for (Path childPath : view.getChildPaths(path)) {
      if (!childPath.isRelation() && view.get(CProp.PROJECTED, childPath)) {
        locallyProjectedPaths.add(childPath.getPrimitive());
        if (!plural && view.get(CProp.IN_TUPLE_KEY, childPath) &&
            !isIdentifiedBy(view, childPath.getPrimitive(), useFullyIdentifiedPaths))
        {
          plural = true;
        }
      }
    }
    view.set(CProp.PLURAL, path, plural);
    final ImmutableSet<PrimitivePath> newFullyIdentifiedPaths =
        ImmutableSet.copyOf(Sets.union(useFullyIdentifiedPaths, locallyProjectedPaths));
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation())
        configurePluralities(view, childPath.getRelation(), newFullyIdentifiedPaths);
    }
  }

  public Set<PrimitivePath> getKey(View view, ImmutableSet<PrimitivePath> paths) {
    return getKey(view, Sets.newLinkedHashSet(), paths);
  }

  // TODO[cc]: Consider making private again.
  /** @param fullyIdentifiedPaths will be augmented with {@code paths} and, possibly, additional
             fully identified paths */
  public Set<PrimitivePath> getKey(View view,
      Set<PrimitivePath> fullyIdentifiedPaths, ImmutableSet<PrimitivePath> paths)
  {
    final Set<PrimitivePath> primaryKeySQLColumnPaths = Sets.newLinkedHashSet();
    final Set<PrimitivePath> nonPrimaryKeySQLColumnPaths = Sets.newLinkedHashSet();
    final Set<PrimitivePath> nonAggregateFormulaPaths = Sets.newLinkedHashSet();
    final Set<PrimitivePath> aggregateFormulaPaths = Sets.newLinkedHashSet();
    for (PrimitivePath path : paths) {
      final Set<PrimitivePath> pathSet;
      if (view.get(QProp.SQL_COLUMN_NAME, path).isEmpty()) {
        pathSet = InternalQueryUtil.isAggregateFormula(view, path)
            ? aggregateFormulaPaths
            : nonAggregateFormulaPaths;
      } else if (view.get(QProp.IN_PRIMARY_KEY, path)) {
        pathSet = primaryKeySQLColumnPaths;
      } else {
        pathSet = nonPrimaryKeySQLColumnPaths;
      }
      pathSet.add(path.getPrimitive());
    }
    final Set<PrimitivePath> ret = Sets.newLinkedHashSet();
    /* Prefer primary key columns to serve as tuple keys, then non-primary key columns, then
    non-aggregate formulas, and then aggregate formulas last. An aggregate formula that is part of
    a key is considered an error, and evaluates to a constant null value; see
    QueryBlock.renderPathExpression. */
    for (PrimitivePath childPath :
        Iterables.concat(primaryKeySQLColumnPaths, nonPrimaryKeySQLColumnPaths,
            nonAggregateFormulaPaths, aggregateFormulaPaths))
    {
      if (!isIdentifiedBy(view, childPath, fullyIdentifiedPaths))
        ret.add(childPath);
      fullyIdentifiedPaths.add(childPath);
    }
    return Collections.unmodifiableSet(ret);
  }

  private void configureTupleKeys(
      MutableView view, RelationPath path, ImmutableSet<PrimitivePath> fullyIdentifiedPaths)
  {
    final Set<PrimitivePath> localIdentifiedPaths = Sets.newLinkedHashSet(fullyIdentifiedPaths);
    final Set<PrimitivePath> keys = getKey(view, localIdentifiedPaths,
        ImmutableSet.copyOf(Queries.getImmediateProjectedChildren(view, path)));
    final ImmutableSet<PrimitivePath> newFullyIdentifiedPaths = ImmutableSet.copyOf(
        INHERIT_DEPENDENCIES_FROM_PARENT ? localIdentifiedPaths : fullyIdentifiedPaths);
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation()) {
        configureTupleKeys(view, childPath.getRelation(), newFullyIdentifiedPaths);
      } else {
        view.set(CProp.IN_TUPLE_KEY, childPath, keys.contains(childPath.getPrimitive()));
      }
    }
  }
}
