// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.database.SQLColumnInfo;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLTableInfo;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.ViewProperty;
import com.sieuferd.datamodel.Views;
import com.sieuferd.evaluator.impl.InternalQueryUtil;
import com.sieuferd.evaluator.impl.ResultData;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.query.SQLTable;
import java.sql.SQLException;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

// TODO[cc]: Rename to EvaluationUtil, and possibly move parts to the Query module.
public final class QueryUtil {
  private QueryUtil() { }

  public static final ImmutableSet<ViewProperty<?>> QUERY_PROPERTIES =
      ImmutableSet.<ViewProperty<?>>builder()
      .add(SProp.CHILD_PATHS)
      .add(CProp.VISIBLE)
      .addAll(QProp.INSTANCE.getViewProperties())
      .build();
  // TODO[cc]: Get rid of this.
  public static final ViewDefaults QUERY_DEFAULTS = ViewDefaults.builder()
      .setAll(QProp.DEFAULTS)
      .setAll(CProp.DEFAULTS)
      .build();

  /* TODO[cc]: Consider moving these to the common-datamodel module. They reside here only to take
               advantage of the ResultData implementation of the Data interface. */
  public static final ImmutableView EMPTY_DATA_VIEW = createEmptyDataView();
  public static final ImmutableData EMPTY_INCOMPLETE_DATA =
      createEmptyIncompleteData(EMPTY_DATA_VIEW);

  private static ImmutableView createEmptyDataView() {
    MutableView view = Views.newMutable();
    view.set(CProp.PLURAL, Path.ROOT, true);
    view.set(CProp.PROJECTED, Path.ROOT, true);
    return ImmutableView.copyOf(view);
  }

  private static ImmutableData createEmptyIncompleteData(ImmutableView emptyView) {
    ResultData relationalData = new ResultData(emptyView);
    relationalData.markLimitReached(Path.ROOT, true);
    return ImmutableData.create(relationalData);
  }

  /* TODO[cc]: Consider getting rid of this, or moving it into Harness. Make sure not to lose the
               primary key column check below. */
  public static void configurePrimaryKeys(
      SQLConnection connection, MutableView view) throws SQLException
  {
    configurePrimaryKeys(connection, view, Path.ROOT);
  }

  private static void configurePrimaryKeys(
      SQLConnection connection, MutableView view, RelationPath path)
      throws SQLException
  {
    final SQLTable sqlTable = InternalQueryUtil.getInstantiatedTable(view, path.getRelation());
    final Map<String,Boolean> columns;
    if (sqlTable == null) {
      columns = ImmutableMap.of();
    } else {
      Map<String,Boolean> toColumns = Maps.newLinkedHashMap();
      SQLTableInfo tableInfo = connection.getTableInfo(sqlTable);
      for (SQLColumnInfo columnInfo : tableInfo.getColumns())
        toColumns.put(columnInfo.getName(), tableInfo.isPrimaryKey(columnInfo.getName()));
      columns = Collections.unmodifiableMap(toColumns);
    }
    final Set<String> seenColumnNames = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation()) {
        configurePrimaryKeys(connection, view, childPath.getRelation());
      } else {
        String columnName = view.get(QProp.SQL_COLUMN_NAME, childPath);
        view.set(QProp.IN_PRIMARY_KEY, childPath,
            columnName.isEmpty() ? false : Boolean.TRUE.equals(columns.get(columnName)));
        if (!columnName.isEmpty())
          seenColumnNames.add(columnName);
      }
    }
    for (Entry<String,Boolean> entry : columns.entrySet()) {
      if (entry.getValue() && !seenColumnNames.contains(entry.getKey())) {
        throw new AssertionError(
            "Missing primary key column " + entry.getKey() + " in set " + seenColumnNames);
      }
    }
  }

  public static List<OrderedPath> getOrderedPaths(
      final View view, RelationPath path, final FunctionalDependencies fd, boolean total)
  {
    final List<Path> sortedChildPaths = Lists.newArrayList();
    for (Path childPath : view.getChildPaths(path)) {
      if (view.get(QProp.ORDER_RANK, childPath).getPosition() != null)
        sortedChildPaths.add(childPath);
    }
    Collections.sort(sortedChildPaths, (Path o1, Path o2) ->
        view.get(QProp.ORDER_RANK, o1).compareTo(view.get(QProp.ORDER_RANK, o2)));
    List<OrderedPath> ret = Lists.newArrayList();
    for (Path childPath : sortedChildPaths) {
      final boolean descending = view.get(QProp.ORDER_DESCENDING, childPath);
      final boolean nullsFirst = view.get(QProp.ORDER_NULLS_FIRST, childPath);
      if (childPath.isRelation()) {
        for (OrderedPath orderedPath : getOrderedPaths(view, childPath.getRelation(), fd, true)) {
          ret.add(new OrderedPath(orderedPath.getPath(),
              descending ? !orderedPath.isDescending() : orderedPath.isDescending(),
              orderedPath.isNullsFirst()));
        }
      } else {
        ret.add(new OrderedPath(childPath.getPrimitive(), descending, nullsFirst));
      }
    }
    if (total) {
      /* I considered sorting fields with filters on them first if there were no other reasons to
      use a specific order (such as the presence of indices), as a UI tweak, but decided against
      it. It would suggest specific semantics which would not always be honored. Besides, such
      sorting would only be on filtered primitive fields, not relation fields, and so would be
      inconsistent. */
      /* Use FunctionalDependencies.getKey to take advantage of primary key indices for total
      orders. More general elimination of redundant sort terms, across all the sort terms generated
      for each SQL query, can be done at the query evaluator level. */
      for (PrimitivePath keyPath :
          fd.getKey(view, ImmutableSet.copyOf(Queries.getImmediateProjectedChildren(view, path))))
      {
        ret.add(new OrderedPath(keyPath, false, false));
      }
    }
    return ret;
  }
}
