// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.database.dialect.RenderVariant;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.evaluator.CancelHandler;
import com.sieuferd.evaluator.impl.EvaluationException.CanceledEvaluationException;
import com.sieuferd.evaluator.impl.RowFetcher.Row;
import com.sieuferd.util.MapUtil;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Map;
import java.util.Set;

final class EvaluationBranch implements AutoCloseable {
  // TODO[cc]: Get rid of this.
  private final boolean USE_LIMITS = true;
  private final SQLConnection connection;
  private final PreparedStatement ps;
  private final RelationPath branchPath;
  private final ImmutableMap<RelationPath,PrimitivePath> tuplePresencePaths;
  private final Map<PrimitivePath,RenderVariant> selectedPathsInOrder;
  private final SQLFragment query;
  private final ImmutableSet<PrimitivePath> keyPaths;
  private final BackendDialect dialect;

  EvaluationBranch(SQLConnection connection,
      ImmutableMap<RelationPath,PrimitivePath> tuplePresencePaths, RelationPath branchPath,
      Map<PrimitivePath,RenderVariant> selectedPathsInOrder, SQLFragment query,
      Set<PrimitivePath> keyPaths)
      throws SQLException
  {
    Preconditions.checkNotNull(connection);
    Preconditions.checkNotNull(tuplePresencePaths);
    Preconditions.checkNotNull(branchPath);
    Preconditions.checkNotNull(selectedPathsInOrder);
    Preconditions.checkNotNull(query);
    Preconditions.checkNotNull(keyPaths);
    this.connection = connection;
    this.selectedPathsInOrder = ImmutableMap.copyOf(selectedPathsInOrder);
    this.branchPath = branchPath;
    this.tuplePresencePaths = tuplePresencePaths;
    this.query = query;
    this.keyPaths = ImmutableSet.copyOf(keyPaths);
    this.ps = connection.getJDBCConnection().prepareStatement(
        connection.getBackendDialect().getJDBCStringWithLimit(query));
    this.dialect = connection.getBackendDialect();
    int parameterIndex = 1;
    for (Object parameter : query.getQueryParameters())
      this.dialect.bindParameterValue(ps, parameterIndex++, parameter);
  }

  public Set<PrimitivePath> getKeyPaths() {
    return keyPaths;
  }

  @Override
  public void close() throws SQLException {
    ps.close();
  }

  public void evaluate(
      ResultData toData, KeyMatcher parentKeyMatcher, KeyMatcher keyMatcher, int limit,
      CancelHandler cancelHandler) throws SQLException, EvaluationException
  {
    if (!USE_LIMITS)
      limit = Integer.MAX_VALUE;
    ps.setInt(query.getQueryParameters().size() + 1, limit);
    boolean beyondParentTuples = false;
    long tupleCount = 0;
    if (!cancelHandler.beginStatement(ps))
      throw new CanceledEvaluationException("Canceled before branch query was started");
    try {
      final PrimitivePath tuplePresencePath = MapUtil.getChecked(tuplePresencePaths, branchPath);
      try (RowFetcher rowFetcher = new RowFetcher(dialect, ps, selectedPathsInOrder)) {
        while (true) {
          if (cancelHandler.isCanceled())
            throw new CanceledEvaluationException("Canceled while retrieving result rows");
          final Row row = rowFetcher.fetchRow();
          if (row == null)
            break;
          tupleCount++;
          // Detect null values generated on the right-hand side of a left join.
          if (row.valueIsNull(tuplePresencePath))
            continue;
          final Tuple parentTuple = parentKeyMatcher.findTupleNullable(row);
          if (parentTuple == null) {
            if (!toData.isIncomplete())
              throw new EvaluationException("Failed to find parent for row " + row);
            /* TODO[pd]: See if I can avoid this case by formulating queries appropriately. Then
                         throw an exception if this case occurs. */
            beyondParentTuples = true;
          } else {
            if (beyondParentTuples) {
              throw new EvaluationException(
                  "Unexpected ordering discrepancy between branch queries");
            }
            final Tuple tuple = toData.addRelationTuple(branchPath, parentTuple, row);
            keyMatcher.addTuple(tuple, row);
          }
        }
      }
    } catch (SQLException e) {
      if (cancelHandler.isCanceled()) {
        connection.reportSQLException(e);
        throw new CanceledEvaluationException("Canceled during ongoing query", e);
      } else {
        throw e;
      }
    } finally {
      cancelHandler.endStatement();
    }
    if (tupleCount >= limit && !beyondParentTuples)
      toData.markLimitReached(branchPath, true);
  }

  @Override
  public String toString() {
    return query.toString();
  }
}
