// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.collect.Sets;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.query.InstanceDefinition.TableIdentifier;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.query.SQLTable;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.Expression.FunctionExpression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.ExpressionVisitor;
import java.util.Set;
import javax.annotation.Nullable;

// TODO[cc]: Consider removing or renaming.
public class InternalQueryUtil {
  private InternalQueryUtil() { }

  public static @Nullable SQLTable getInstantiatedTable(View view, RelationPath path) {
    TableIdentifier tableIdentifier = view.get(QProp.INSTANTIATED_TABLE, path).getTable();
    // TODO[pd]: Add a warning in this case.
    if (tableIdentifier != null && !(tableIdentifier instanceof SQLTable))
      throw new UnsupportedOperationException();
    return (SQLTable) tableIdentifier;
  }

  // TODO[cc]: Avoid duplication of logic with QueryBlock.
  /**
   * Determines if the specified path is an aggregate formula, that is, a column which may not
   * appear in a GROUP BY clause.
   */
  public static boolean isAggregateFormula(final View view, final PrimitivePath path) {
    return isAggregateFormula(view, path, Sets.newLinkedHashSet());
  }

  private static boolean isAggregateFormula(final View view, final PrimitivePath path,
      final Set<PrimitivePath> alreadyVisitedPaths)
  {
    if (!view.exists(path) || !alreadyVisitedPaths.add(path))
      return false;
    Expression exp = view.get(QProp.FORMULA_EXPRESSION, path);
    return exp.accept(new ExpressionVisitor<Boolean,RuntimeException>() {
      @Override
      public Boolean visit(FunctionExpression exp) {
        if (Function.AGGREGATE_FUNCTIONS.contains(exp.getFunction())) {
          return true;
        } else {
          for (Expression argument : exp.getArguments()) {
            if (argument.accept(this))
              return true;
          }
          return false;
        }
      }

      @Override
      public Boolean visit(ConstantExpression exp) {
        return false;
      }

      @Override
      public Boolean visit(PathExpression exp) {
        Path refPath = exp.getPath();
        /* Only sibling references to other aggregate formulas will make the current formula an
        aggregate formula itself. (And a relation reference is here equivalent to a reference to a
        constant.) */
        return !refPath.isRelation() && refPath.getParent().equals(path.getParent())
            ? isAggregateFormula(view, refPath.getPrimitive(), alreadyVisitedPaths)
            : false;
      }
    });
  }

  /**
   * @return the immediate child of basePath that is on the path to targetPath
   */
  public static @Nullable RelationPath findChildTowardsTargetPath(
      RelationPath basePath, RelationPath targetPath)
  {
    RelationPath ret = targetPath;
    while (!ret.isRoot()) {
      final RelationPath parentPath = ret.getParent();
      if (basePath.equals(parentPath))
        return ret;
      ret = parentPath;
    }
    return null;
  }

  public static boolean isNullable(View view, PrimitivePath path) {
    if (Queries.isInPrimaryKey(view, path)) {
      return false; // Not necessarily true on all backends, but use this assumption for now.
    } else if (view.get(QProp.SQL_COLUMN_NAME, path).isEmpty()) {
      Expression exp = view.get(QProp.FORMULA_EXPRESSION, path);
      if ((exp instanceof ConstantExpression) && ((ConstantExpression) exp).getValue() != null)
        return false;
    }
    return true;
  }

  public static Set<PrimitivePath> getNaiveGroupByPaths(View view, RelationPath path) {
    return Sets.filter(Queries.getProjectedToRoot(view, path),
        projectedPath -> !InternalQueryUtil.isAggregateFormula(view, projectedPath));
  }
}
