// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.impl.Evaluation.NWEvaluation;
import com.sieuferd.evaluator.impl.SQLIdentifier.InstanceID;
import com.sieuferd.util.Functional;
import com.sieuferd.util.MapUtil;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.concurrent.Immutable;

@Immutable
final class JoinPart {
  private final RelationPath warningPath;
  private final SQLFragment subquery;
  private final InstanceID instanceID;
  private final ImmutableSet<InstanceID> instanceDependencies;
  private final boolean leftJoin;
  // TODO[cc]: See if I can just store a RenderedExpression here.
  private final ImmutableList<NWEvaluation> onConditions;

  JoinPart(RelationPath warningPath, SQLFragment subquery, InstanceID instanceID,
      boolean leftJoin, List<NWEvaluation> onConditions, Set<InstanceID> instanceDependencies)
  {
    Preconditions.checkNotNull(warningPath);
    Preconditions.checkNotNull(subquery);
    Preconditions.checkNotNull(instanceID);
    Preconditions.checkNotNull(onConditions);
    this.warningPath = warningPath;
    this.subquery = subquery;
    this.instanceID = instanceID;
    this.leftJoin = leftJoin;
    this.onConditions = ImmutableList.copyOf(onConditions);
    this.instanceDependencies = ImmutableSet.copyOf(instanceDependencies);
  }

  public RelationPath getWarningPath() {
    return warningPath;
  }

  public SQLFragment getSubquery() {
    return subquery;
  }

  public InstanceID getInstanceID() {
    return instanceID;
  }

  public boolean isLeftJoin() {
    return leftJoin;
  }

  public ImmutableList<NWEvaluation> getOnConditions() {
    return onConditions;
  }

  public ImmutableSet<InstanceID> getInstanceDependencies() {
    return instanceDependencies;
  }

  @Override
  public String toString() {
    return instanceID + " " + (leftJoin ? "LEFT JOIN" : "JOIN") + " ON " + onConditions +
        " (depends on " + instanceDependencies + ")";
  }

  @Override
  public int hashCode() {
    // Any attempt to use this in a hash set or as a map key is an error; use InstanceID instead.
    throw new UnsupportedOperationException();
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!(obj instanceof JoinPart))
      return false;
    throw new UnsupportedOperationException();
  }

  private static void populateJoinList(QueryBuildingGlobals globals, JoinPart joinFragment,
      ImmutableMap<InstanceID,JoinPart> fragmentsByInstanceID,
      ImmutableSet<InstanceID> visitedInstances, Map<InstanceID,JoinPart> ret)
  {
    Preconditions.checkArgument(joinFragment.isLeftJoin());
    final InstanceID instanceID = joinFragment.getInstanceID();
    if (ret.containsKey(instanceID))
      return;
    if (visitedInstances.contains(instanceID)) {
      ret.put(instanceID, new JoinPart(joinFragment.getWarningPath(),
          joinFragment.getSubquery(), instanceID, joinFragment.isLeftJoin(),
          ImmutableList.of(NWEvaluation.error()), ImmutableSet.of()));
      // TODO[te]: Add a test case that triggers this error.
      globals.addWarning(joinFragment.getWarningPath(), new Warning(
          "Circular dependency in left join conditions"));
      return;
    }
    final ImmutableSet<InstanceID> newVisitedInstances =
        Functional.cons(visitedInstances, instanceID);
    for (InstanceID dependeeID : joinFragment.getInstanceDependencies()) {
      JoinPart dependeeFragment = MapUtil.getChecked(fragmentsByInstanceID, dependeeID);
      populateJoinList(globals, dependeeFragment, fragmentsByInstanceID, newVisitedInstances, ret);
    }
    if (!ret.containsKey(instanceID))
      ret.put(instanceID, joinFragment);
  }

  public static List<JoinPart> topologicalSort(
      QueryBuildingGlobals globals, Iterable<JoinPart> inputJoinFragments)
  {
    // For quick lookup only.
    final Map<InstanceID,JoinPart> fragmentsByInstanceID = Maps.newLinkedHashMap();
    for (JoinPart fragment : inputJoinFragments)
      MapUtil.putChecked(fragmentsByInstanceID, fragment.getInstanceID(), fragment);
    // Order is significant.
    final Map<InstanceID,JoinPart> ret = Maps.newLinkedHashMap();
    for (JoinPart fragment : inputJoinFragments) {
      populateJoinList(globals, fragment, ImmutableMap.copyOf(fragmentsByInstanceID),
          ImmutableSet.of(), ret);
    }
    // Go through ImmutableMap to ensure safe iteration order of values().
    return ImmutableList.copyOf(ImmutableMap.copyOf(ret).values());
  }
}