// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLFragmentBuilder;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.database.dialect.FunctionImplementation;
import com.sieuferd.database.dialect.FunctionImplementation.RenderFunctionException;
import com.sieuferd.database.dialect.RenderVariant;
import com.sieuferd.database.dialect.RenderedExpression;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.OrderedPath;
import com.sieuferd.evaluator.impl.Evaluation.NWEvaluation;
import com.sieuferd.evaluator.impl.EvaluationSettings.SettingsFlag;
import com.sieuferd.evaluator.impl.SQLIdentifier.InstanceID;
import com.sieuferd.query.Filter;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.query.SQLTable;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.Expression.FunctionExpression;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.expressions.ExpressionVisitor;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.util.Functional;
import com.sieuferd.util.MapUtil;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;

final class QueryBlock {
  private static final boolean GENERATE_QUERY_COMMENTS = true;
  // TODO[cc]: I can get rid of this now.
  private static final boolean JOINED_ON_MATCHES_NULLS = false;
  private static final boolean PREFER_DEEP_DISTINCTS = false;
  private final QueryBuildingGlobals globals;
  private final ImmutableView view;
  private final BackendDialect dialect;
  private final RelationPath basePath;
  private final boolean localCorrelated;
  private final boolean applyFilters;
  // TODO[cc]: Make this just a single (nullable) RelationPath instead.
  private final ImmutableSet<RelationPath> assumedExistingPaths;
  private final ImmutableList<OrderedPath> orderedPaths;
  /** For invariant checking only. */
  private final ImmutableSet<RelationPath> strictAncestors;
  /** Map iteration order is significant. */
  private final Map<PrimitivePath,SelectPart> selectParts = Maps.newLinkedHashMap();
  /** May not contain basePath, except if aggregationExpression is not null. */
  private final Map<RelationPath,RelationInstance> instances = Maps.newLinkedHashMap();
  private final Map<Expression,RelationInstance> aggregationInstances = Maps.newLinkedHashMap();
  private final Set<String> locallySelectedColumns = Sets.newLinkedHashSet();
  private final Set<PrimitivePath> localCorrelationPaths = Sets.newLinkedHashSet();
  private final InstanceID localInstanceID;
  /**
   * If not null, filters on values fully identified by the contained paths may be omitted. Used to
   * avoid circular dependencies between inputs and outputs of aggregates with filters on them.
   */
  private final @Nullable ImmutableSet<PrimitivePath> omittableFilterKeys;
  /* TODO[cc]: Collect these three in a single class, and make them null or not together. Except
               the initialization of the latter two actually depends on aggregationExpression being
               set; ugh. */
  private final @Nullable Expression aggregationExpression;
  private final @Nullable List<Warning> aggregationWarnings;
  private final @Nullable SelectPart aggregationSelectPart;
  private boolean frozen;

  QueryBlock(QueryBuildingGlobals globals, RelationPath basePath, boolean localCorrelated,
      boolean applyFilters, ImmutableSet<RelationPath> assumedExistingPaths,
      ImmutableList<OrderedPath> orderedPaths,
      @Nullable ImmutableSet<PrimitivePath> omittableFilterKeys,
      @Nullable Expression aggregationExpression)
  {
    Preconditions.checkNotNull(globals);
    Preconditions.checkNotNull(basePath);
    Preconditions.checkNotNull(assumedExistingPaths);
    Preconditions.checkNotNull(orderedPaths);
    /* localCorrelationPaths is never populated in the aggregate result block, since all paths are
    resolved in the aggregate input block. */
    Preconditions.checkArgument(aggregationExpression == null || !localCorrelated);
    Preconditions.checkArgument(aggregationExpression == null ||
        ((aggregationExpression instanceof FunctionExpression) &&
        Function.AGGREGATE_FUNCTIONS.contains(
            ((FunctionExpression) aggregationExpression).getFunction())));
    this.globals = globals;
    this.view = globals.getView();
    this.dialect = globals.getDialect();
    this.basePath = basePath;
    this.localCorrelated = localCorrelated;
    this.applyFilters = applyFilters;
    this.assumedExistingPaths = assumedExistingPaths;
    this.orderedPaths = orderedPaths;
    this.strictAncestors = ImmutableSet.copyOf(
        Views.getStrictAncestors(ImmutableSet.<Path>of(basePath)));
    /* The local instance should be created even if no reference is made to it, as it may affect the
    number of tuples in the result (e.g. the difference between an empty set of tuples and a set
    containing a single unit tuple). For relations not instantiating the table, the local instance
    is a single-tuple dummy table. */
    /* TODO[pd]: Consider only doing this when a table is instantiated. But I'd still need to have
                 something to instantiate during outward formula references. */
    this.localInstanceID = globals.IF_LINST.fresh();
    this.omittableFilterKeys = omittableFilterKeys;
    this.aggregationExpression = aggregationExpression;
    if (aggregationExpression == null) {
      aggregationSelectPart = null;
      aggregationWarnings = null;
    } else {
      Evaluation columnEvaluation = evaluateExpression(aggregationExpression, ImmutableSet.of());
      RenderedExpression selectableExpression =
          (columnEvaluation.isError() ? dialect.getNullLiteral() :
          columnEvaluation.getRenderedExpression()).toSelectable();
      aggregationSelectPart = new SelectPart(
          new Column(globals.IF_ACOL.fresh(), selectableExpression.getVariant()),
          selectableExpression.getFragment());
      aggregationWarnings = columnEvaluation.getWarnings();
    }
  }

  Column getAggregationColumn() {
    Preconditions.checkState(aggregationExpression != null);
    return aggregationSelectPart.getColumn();
  }

  List<Warning> getAggregationWarnings() {
    Preconditions.checkState(aggregationExpression != null);
    return aggregationWarnings;
  }

  private ImmutableList<SQLFragment> generateOrderByFragments(BackendDialect dialect,
      List<OrderedPath> orderedPaths, ImmutableSet<Path> alreadyVisitedPaths)
  {
    final List<SQLFragment> ret = Lists.newArrayList();
    for (OrderedPath orderedPath : orderedPaths) {
      final SQLFragmentBuilder builder = SQLFragment.builder();
      final Column column = requestFromSelectList(orderedPath.getPath(), alreadyVisitedPaths);
      builder.appendSQLFragment(column.getRenderedExpression().toOrderable().getFragment());
      if (orderedPath.isDescending())
        builder.appendJDBCString(" DESC");

      /* "[The SQL standard] only states that all NULLs must appear together after sorting, but it
      does not specify if they should appear before or after the other entries". See
      http://use-the-index-luke.com/sql/sorting-grouping/order-by-asc-desc-nulls-last . */
      if (isNullable(orderedPath.getPath()) && dialect.supportsNullsFirstLast())
        builder.appendJDBCString(" NULLS " + (orderedPath.isNullsFirst() ? "FIRST" : "LAST"));
      ret.add(builder.build());
    }
    return ImmutableList.copyOf(ret);
  }

  @Override
  public String toString() {
    return "QueryBlock(basePath=" + basePath +
        ", localCorrelated=" + localCorrelated +
        ", applyFilters=" + applyFilters +
        ", aggregationExpression=" + aggregationExpression + ")";
  }

  // TODO[pd]: See if I can replace more of these with functional equivalents based on Evaluation.
  public void addWarning(Path path, Warning warning) {
    globals.addWarning(path, warning);
  }

  // TODO[cc]: Get rid of this.
  public RelationPath getBasePath() {
    return basePath;
  }

  public Column requestFromSelectList(
      PrimitivePath requestedPath, ImmutableSet<Path> alreadyVisitedPaths)
  {
    SelectPart selectPart = selectParts.get(requestedPath);
    if (selectPart == null) {
      Preconditions.checkState(!frozen);
      NWEvaluation evaluation = evaluatePath(requestedPath, alreadyVisitedPaths);
      RenderedExpression selectableExpression =
          extractRenderedExpression(evaluation).toSelectable();
      selectPart = new SelectPart(
          new Column(globals.IF_COL.fresh(), selectableExpression.getVariant()),
          selectableExpression.getFragment());
      selectParts.put(requestedPath, selectPart);
    }
    return selectPart.getColumn();
  }

  public Map<PrimitivePath,RenderVariant> getSelectedPathsInOrder() {
    final Map<PrimitivePath,RenderVariant> ret = Maps.newLinkedHashMap();
    for (Entry<PrimitivePath,SelectPart> entry : selectParts.entrySet()) {
      MapUtil.putChecked(ret, entry.getKey(),
          entry.getValue().getColumn().getRenderedExpression().getVariant());
    }
    return Collections.unmodifiableMap(ret);
  }

  public GeneratedQuery generateQuery() {
    Preconditions.checkState(frozen);
    return generateQueryInternal(ImmutableSet.of(), "", true);
  }

  /**
   * Return true iff result tuples from this block would be distinct even if the DISTINCT keyword is
   * omitted.
   *
   * <p>If {@code PREFER_DEEP_DISTINCTS} is true, input tuples are assumed to be distinct. Otherwise
   * input tuples are assumed to be distinct only if their corresponding instance block can itself
   * omit the DISTINCT keyword and still produce distinct tuples.
   */
  boolean areResultTuplesDistinctWithoutDistinctKeyword() {
    if (aggregationExpression != null)
      return true;
    /* If input tuples are fully identified by output tuples, and the input tuples are themselves
    distinct, then output tuples will be distinct without a DISTINCT keyword. */
    final Set<PrimitivePath> inputPaths = Sets.newLinkedHashSet();
    for (Path childPath : view.getChildPaths(basePath)) {
      if (!childPath.isRelation() &&
          Queries.isInPrimaryKey(view, childPath.getPrimitive()))
      {
        inputPaths.add(childPath.getPrimitive());
      }
    }
    for (RelationInstance instance : instances.values()) {
      QueryBlock instanceBlock = instance.getQueryBlock();
      if (!PREFER_DEEP_DISTINCTS && !instanceBlock.areResultTuplesDistinctWithoutDistinctKeyword())
        return false;
      inputPaths.addAll(instanceBlock.getSelectedPathsInOrder().keySet());
    }
    final Set<PrimitivePath> identifiedByOutputPaths =
        Sets.newLinkedHashSet(getSelectedPathsInOrder().keySet());
    for (PrimitivePath inputPath : inputPaths) {
      if (!globals.getFunctionalDependencies().isIdentifiedBy(
          view, inputPath, identifiedByOutputPaths))
      {
        return false;
      }
    }
    return true;
  }

  private GeneratedQuery generateQueryInternal(
      ImmutableSet<Path> alreadyVisitedPaths, String indentation, boolean requireDistinct)
  {
    final Set<CorrelationParameter> correlationParameters = Sets.newLinkedHashSet();
    final Set<CorrelationParameter> joinCorrelationParameters = Sets.newLinkedHashSet();
    final List<NWEvaluation> whereConditions = applyFilters
        ? Lists.newArrayList(buildFilterConditions(joinCorrelationParameters, alreadyVisitedPaths))
        : Lists.newArrayList();
    final ImmutableList<SQLFragment> orderByFragments =
        generateOrderByFragments(dialect, orderedPaths, alreadyVisitedPaths);

    SQLFragmentBuilder ret = SQLFragment.builder();
    ret.appendJDBCString(indentation + "SELECT ");
    final boolean instanceRequireDistinct;
    if (aggregationExpression != null) {
      instanceRequireDistinct = true;
    } else {
      final boolean needsDistinct =
          requireDistinct && !areResultTuplesDistinctWithoutDistinctKeyword();
      final boolean useDistinct = needsDistinct ||
          !globals.querySettingsFlag(SettingsFlag.DISTINCT_ELIMINATION_OPTIMIZATION);
      if (useDistinct)
        ret.appendJDBCString("DISTINCT ");
      /* If we rely on a true value of areResultTuplesDistinctWithoutDistinctKeyword to avoid the
      DISTINCT keyword here, we must ensure that input tuples are distinct (per the contract of
      areResultTuplesDistinctWithoutDistinctKeyword). */
      instanceRequireDistinct = requireDistinct && !useDistinct;
    }

    final SQLFragment fromClause = buildFromClause(alreadyVisitedPaths,
        indentation, correlationParameters, whereConditions, instanceRequireDistinct);
    correlationParameters.addAll(joinCorrelationParameters);

    final List<SQLFragment> selectFragments = Lists.newArrayList();
    final List<SQLFragment> groupByFragments = Lists.newArrayList();
    for (SelectPart selectPart : selectParts.values()) {
      selectFragments.add(SQLFragment.builder()
          .appendSQLFragment(selectPart.getExpressionFragment())
          .appendJDBCString(" AS ")
          .appendSQLFragment(selectPart.getColumn().getIDFragment())
          .build());
      if (aggregationExpression != null) {
        groupByFragments.add(dialect.allowsAliasInGroupBy()
            ? selectPart.getColumn().getIDFragment()
            : selectPart.getExpressionFragment());
      }
    }
    if (aggregationExpression != null) {
      SelectPart selectPart = aggregationSelectPart;
      selectFragments.add(SQLFragment.builder()
          .appendSQLFragment(selectPart.getExpressionFragment())
          .appendJDBCString(" AS ")
          .appendSQLFragment(selectPart.getColumn().getIDFragment())
          .build());
    }
    ret.appendSQLFragment(selectFragments.isEmpty()
        ? SQLFragment.create("0")
        : SQLFragment.join(", ", selectFragments));

    ret.appendJDBCString("\n" + indentation + "FROM ");
    ret.appendSQLFragment(fromClause);
    if (!whereConditions.isEmpty()) {
      ret.appendJDBCString("\n" + indentation + "WHERE ");
      ret.appendSQLFragment(renderAndExpression(ImmutableList.copyOf(whereConditions))
          .toConditional().getFragment());
    }
    /* Beware: "[A]n ORDER BY clause applying to the result of a UNION, INTERSECT, or EXCEPT
    clause can only specify an output column name or number, not an expression." and "In case of
    ambiguity, a GROUP BY name will be interpreted as an input-column name rather than an output
    column name."

    http://www.postgresql.org/docs/9.0/static/sql-select.html

    So use ordinal numbers or globally unique column names only.
    */
    if (!groupByFragments.isEmpty()) {
      ret.appendJDBCString("\n" + indentation + "GROUP BY ");
      ret.appendSQLFragment(SQLFragment.join(", ", groupByFragments));
    }
    if (!orderByFragments.isEmpty()) {
      ret.appendJDBCString("\n" + indentation + "ORDER BY ");
      ret.appendSQLFragment(SQLFragment.join(", ", orderByFragments));
    }
    final SQLFragment retFragment;
    if (GENERATE_QUERY_COMMENTS) {
      retFragment = SQLFragment.builder()
          .appendJDBCString(
              indentation + "-- CP=" + correlationParameters + "\n")
          .appendSQLFragment(ret.build())
          .build();
    } else {
      retFragment = ret.build();
    }
    return new GeneratedQuery(retFragment, correlationParameters);
  }

  private boolean isNullable(PrimitivePath path) {
    return InternalQueryUtil.isNullable(view, path);
  }

  private RenderedExpression renderAndExpression(ImmutableList<? extends Evaluation> arguments) {
    Iterator<? extends Evaluation> iter = arguments.iterator();
    if (!iter.hasNext())
      return dialect.getBooleanLiteral(true);
    Evaluation ret = iter.next();
    while (iter.hasNext())
      ret = evaluateFunction(Function.LOGICAL_AND, ImmutableList.of(ret, iter.next()));
    return extractRenderedExpression(reportWarnings(basePath, ret));
  }

  private void addFilter(
      Path path, List<NWEvaluation> toWhereConditions, ImmutableSet<Path> alreadyVisitedPaths)
  {
    Filter filter = view.get(QProp.FILTER, path);
    /* The functional dependencies are valid whether or not the filter is omitted, since
    FunctionalDependencies currently ignores the FILTER property. */
    if (filter.isConstrained() && (omittableFilterKeys == null ||
        !globals.getFunctionalDependencies().isIdentifiedBy(view,
        basePath, filter.toExpression(), Sets.newLinkedHashSet(omittableFilterKeys))))
    {
      toWhereConditions.add(reportWarnings(path,
          evaluateExpression(filter.toExpression(), alreadyVisitedPaths)));
    }
  }

  private List<NWEvaluation> buildFilterConditions(
      Set<CorrelationParameter> toJoinCorrelationParameters, ImmutableSet<Path> alreadyVisitedPaths)
  {
    /* TODO[pd]: Consider doing constant folding and a proper conversion to conjunctive normal form
                 when processing WHERE clause conditions. Before this is done, any sibling formulas
                 referenced should be inlined. */
    final List<NWEvaluation> ret = Lists.newArrayList();
    addFilter(basePath, ret, alreadyVisitedPaths);
    for (Path childPath : view.getChildPaths(basePath)) {
      if (childPath.isRelation()) {
        /* Ensure that all child relations with existential filters are properly joined in, even if
        they are not referenced directly. */
        if (view.get(QProp.EXISTENTIAL_FILTER, childPath))
          requestInstance(childPath.getRelation());
      } else {
        addFilter(childPath, ret, alreadyVisitedPaths);
        final PrimitivePath joinedOnPath = view.get(QProp.JOINED_ON, childPath).getNullable();
        if (joinedOnPath != null) {
          if (strictAncestors.contains(joinedOnPath.getParent())) {
            if (!view.exists(joinedOnPath))
              addWarning(childPath, new Warning("Broken path reference in join condition"));
            final boolean useIs = JOINED_ON_MATCHES_NULLS &&
                isNullable(childPath.getPrimitive()) && isNullable(joinedOnPath);
            /* TODO[pd]: Check if the isStrictAncestorOf condition can be relaxed, for instance to
                         include lateral references. */
            /* Query CORRELATION_JOIN_OPTIMIZATION last in the short-circuiting condition so that
            unit tests can avoid trying both settings in cases where it doesn't make a
            difference. */
            if (joinedOnPath.getParent().isStrictAncestorOf(basePath) &&
                globals.querySettingsFlag(SettingsFlag.CORRELATION_JOIN_OPTIMIZATION))
            {
              toJoinCorrelationParameters.add(new CorrelationParameter(joinedOnPath,
                  requestFromSelectList(childPath.getPrimitive(), alreadyVisitedPaths), useIs));
            } else {
              final ImmutableList<Evaluation> operands = ImmutableList.of(
                  evaluatePath(joinedOnPath, alreadyVisitedPaths),
                  evaluatePath(childPath.getPrimitive(), alreadyVisitedPaths));
              ret.add(reportWarnings(childPath,
                  evaluateFunction(useIs ? Function.IS : Function.EQUAL, operands)));
            }
          } else {
            // TODO[pd]: Consider allowing arbitrary references, like in formulas.
            addWarning(childPath, new Warning("Can only join on ancestors' fields"));
          }
        }
      }
    }
    return Collections.unmodifiableList(ret);
  }

  private boolean isRowIdentifierPseudocolumnProjected() {
    for (String columnName : locallySelectedColumns) {
      if (dialect.isRowIdentifierPseudoColumn(columnName))
        return true;
    }
    return false;
  }

  // TODO[cc]: Remove the related dead code once I'm sure I won't be using it.
  private static final boolean WRAP_LOCAL_INSTANTIATIONS_IN_DISTINCT = false;

  private SQLFragment getLocalInstantiation(RelationPath relationPath, InstanceID instanceID) {
    SQLFragmentBuilder ret = SQLFragment.builder();
    final SQLTable table = InternalQueryUtil.getInstantiatedTable(view, relationPath);
    if (table == null || aggregationExpression != null) {
      ret.appendSQLFragment(dialect.getSingleRowDummyTable());
    } else if (!WRAP_LOCAL_INSTANTIATIONS_IN_DISTINCT ||
        /* To correctly wrap a projection including a ROWID or similar pseudocolumn in a SELECT
        DISTINCT, it would be necessary to assign the pseudocolumn a new name, since ROWID has
        special semantics beyond simply serving as an identifier. However, if ROWID is projected,
        then the SELECT DISTINCT is redundant, and so can be omitted altogether. See
        QueryEvaluatorTest.testAggregateFromPseudokeyViewParentRelation. */
        isRowIdentifierPseudocolumnProjected())
    {
      ret.appendSQLFragment(dialect.getTableIdentifier(table));
    } else {
      // TODO[pd]: Don't do this. It doesn't do the full job anyway.
      // TODO[cc]: Consider factoring out the SELECT clause syntax construction logic.
      // For aggregate expressions, wrap the local instance in a SELECT DISTINCT.
      ret.appendJDBCString("(SELECT DISTINCT ");
      List<SQLFragment> columnFragments = Lists.newArrayList();
      for (String columnName : locallySelectedColumns)
        columnFragments.add(dialect.quoteIdentifier(columnName, true));
      if (columnFragments.isEmpty()) {
        ret.appendJDBCString("0");
      } else {
        ret.appendSQLFragment(SQLFragment.join(", ", columnFragments));
      }
      ret.appendJDBCString(" FROM ");
      ret.appendSQLFragment(dialect.getTableIdentifier(table));
      ret.appendJDBCString(")");
    }
    ret.appendJDBCString(" ");
    ret.appendSQLFragment(instanceID.toSQLFragment());
    return ret.build();
  }

  private ImmutableSet<RelationInstance> getAllInstances() {
    return ImmutableSet.copyOf(Iterables.concat(instances.values(), aggregationInstances.values()));
  }

  /**
   * Build the FROM clause, not including the actual "FROM" keyword.
   *
   * @param toInnerJoinConditions will be populated
   */
  private SQLFragment buildFromClause(
      ImmutableSet<Path> alreadyVisitedPaths, String indentation,
      Set<CorrelationParameter> toCorrelationParameters, List<NWEvaluation> toInnerJoinConditions,
      boolean requireDistinct)
  {
    /* TODO[pd]: Split the calculation of correlation fragments from the generation of the actual
                 query, then require the QueryBlock to be frozen before the latter can be done. */
    Preconditions.checkArgument(toCorrelationParameters.isEmpty());

    Set<PrimitivePath> correlationPaths = Sets.newLinkedHashSet();

    if (localCorrelated) {
      for (PrimitivePath path : localCorrelationPaths)
        correlationPaths.add(path);
    }
    if (aggregationExpression != null) {
      if (!localCorrelationPaths.isEmpty())
        throw new AssertionError();

      /* Key reduction by functional dependencies is mandatory here, because inward references
      to the aggregate result may themselves be counted as part of the naive GROUP BY, and need to
      be removed from the key to avoid circular references. See
      testKeyInwardReferenceToAggregateFormula. */
      correlationPaths.addAll(globals.getFunctionalDependencies().getKey(view,
          ImmutableSet.copyOf(InternalQueryUtil.getNaiveGroupByPaths(view, basePath))));
    }

    ImmutableMap<RelationInstance,GeneratedQuery> processedInstances = ImmutableMap.of();
    while (true) {
      final ImmutableSet<RelationInstance> previousInstances = processedInstances.keySet();
      {
        final Map<RelationInstance,GeneratedQuery> toGeneratedQueries = Maps.newLinkedHashMap();
        for (RelationInstance instance : getAllInstances()) {
          MapUtil.putChecked(toGeneratedQueries, instance,
              instance.getQueryBlock().generateQueryInternal(
              alreadyVisitedPaths, indentation + "  ", requireDistinct));
        }
        processedInstances = ImmutableMap.copyOf(toGeneratedQueries);
      }
      for (Entry<RelationInstance,GeneratedQuery> entry : processedInstances.entrySet()) {
        for (CorrelationParameter parameter : entry.getValue().getCorrelationParameters())
          evaluatePath(parameter.getPath(), ImmutableSet.of());
      }
      if (processedInstances.keySet().equals(previousInstances))
        break;
    }
    // TODO[cc]: Make this a Map<InstanceID,SQLFragment> instead.
    final List<JoinPart> innerJoins = Lists.newArrayList();
    final List<JoinPart> leftJoins = Lists.newArrayList();
    for (Entry<RelationInstance,GeneratedQuery> entry : processedInstances.entrySet()) {
      final RelationInstance instance = entry.getKey();
      final GeneratedQuery generatedQuery = entry.getValue();
      final List<NWEvaluation> onConditions = Lists.newArrayList();
      final Set<InstanceID> instanceDependencies = Sets.newLinkedHashSet();
      for (CorrelationParameter parameter : generatedQuery.getCorrelationParameters()) {
        final PrimitivePath parameterPath = parameter.getPath();
        final RelationPath parameterRelationPath = parameterPath.getParent();
        final NWEvaluation parameterEvaluation = NWEvaluation.simple(
            parameter.getColumn().getRenderedExpression(instance.getInstanceID()));
        final RelationInstance otherInstance = requestInstance(parameterRelationPath);
        // No need to decorrelate an instance against itself.
        if (instance != otherInstance) {
          // TODO[cc]: Get rid of this.
          Set<RelationInstance> instanceDependenciesTemp = Sets.newLinkedHashSet();
          final NWEvaluation localColumnEvaluation = evaluatePath(
              parameterPath, ImmutableSet.of(), instanceDependenciesTemp);
          if (localColumnEvaluation.isError()) {
            onConditions.add(NWEvaluation.error());
          } else {
            for (RelationInstance instanceDependency : instanceDependenciesTemp) {
              if (instanceDependency.isLeftJoined())
                instanceDependencies.add(instanceDependency.getInstanceID());
            }
            final boolean isNullable = isNullable(parameterPath) && parameter.getUseIs();
            Evaluation onCondition = evaluateFunction(
                isNullable ? Function.IS : Function.EQUAL,
                ImmutableList.of(parameterEvaluation, localColumnEvaluation));
            onConditions.add(reportWarnings(basePath, onCondition));
          }
        }
        if (parameterRelationPath.isStrictAncestorOf(basePath))
          correlationPaths.add(parameterPath);
      }
      boolean leftJoin = instance.isLeftJoined();
      (leftJoin ? leftJoins : innerJoins).add(new JoinPart(
          instance.getQueryBlock().getBasePath(), generatedQuery.getQuery(),
          instance.getInstanceID(), leftJoin,
          leftJoin ? onConditions : ImmutableList.of(), instanceDependencies));
      if (!leftJoin)
        toInnerJoinConditions.addAll(onConditions);
    }
    if (!getAllInstances().equals(processedInstances.keySet()))
      throw new AssertionError();

    final SQLFragmentBuilder ret = SQLFragment.builder();
    /* While the LEFT JOIN syntax requires a complete explicit join order to be specified,
    PostgreSQL, as an example, is smart enough to realize that many left joins can be reordered for
    optimization purposes. See http://www.postgresql.org/docs/9.1/static/explicit-joins.html ("In
    other cases, the planner might be able to determine that more than one join order is
    safe..."). */
    ret.appendSQLFragment(getLocalInstantiation(basePath, localInstanceID));
    for (JoinPart joinFragment :
        Iterables.concat(innerJoins, JoinPart.topologicalSort(globals, leftJoins)))
    {
      ret.appendJDBCString(" ")
          .appendJDBCString(joinFragment.isLeftJoin() ? "LEFT JOIN" : "JOIN")
          .appendJDBCString(" (\n")
          .appendSQLFragment(joinFragment.getSubquery())
          .appendJDBCString("\n" + indentation + ") ")
          .appendSQLFragment(joinFragment.getInstanceID().toSQLFragment())
          .appendJDBCString(" ON ")
          .appendSQLFragment(
              renderAndExpression(joinFragment.getOnConditions()).toConditional().getFragment());
    }
    final Set<PrimitivePath> useCorrelationPaths;
    {
      Set<PrimitivePath> refinedCorrelationPaths = getRefinedCorrelationPaths(correlationPaths);
      if (!correlationPaths.equals(refinedCorrelationPaths) &&
          globals.querySettingsFlag(SettingsFlag.REFINE_CORRELATION_KEYS_OPTIMIZATION))
      {
        useCorrelationPaths = refinedCorrelationPaths;
      } else {
        useCorrelationPaths = correlationPaths;
      }
    }
    for (PrimitivePath correlationPath : useCorrelationPaths) {
      toCorrelationParameters.add(new CorrelationParameter(correlationPath,
          requestFromSelectList(correlationPath, ImmutableSet.of()),
          isNullable(correlationPath)));
    }
    return ret.build();
  }

  private Set<PrimitivePath> getRefinedCorrelationPaths(Set<PrimitivePath> originalPaths) {
    Set<PrimitivePath> allPaths = Sets.newLinkedHashSet(originalPaths);
    // TODO[pd]: Implement, or remove dead code.
    final boolean distinct = true;
    /* Add the primary key columns to the mix; they will likely serve as good identifiers for the
    original paths while yielding more efficient query plans. Only do this if it's OK to duplicate
    tuples of the original paths, however. */
    if (!allPaths.isEmpty() && aggregationExpression == null && !distinct) {
      for (Path childPath : view.getChildPaths(basePath)) {
        if (!childPath.isRelation() &&
            Queries.isInPrimaryKey(view, childPath.getPrimitive()))
        {
          allPaths.add(childPath.getPrimitive());
        }
      }
    }
    return Collections.unmodifiableSet(
        globals.getFunctionalDependencies().getKey(view, ImmutableSet.copyOf(allPaths)));
  }

  public void freeze() {
    generateQueryInternal(ImmutableSet.of(), "", true);
    for (RelationInstance instance : getAllInstances())
      instance.getQueryBlock().freeze();
    frozen = true;
  }

  /** @return null if the path should be handled in this QueryBlock */
  private @Nullable RelationPath getInstancePath(RelationPath targetPath) {
    if (aggregationExpression == null) {
      if (targetPath.equals(basePath))
        return null;
    } else {
      return Path.ROOT;
    }
    final RelationPath childTowardsReferencePath =
        InternalQueryUtil.findChildTowardsTargetPath(basePath, targetPath);
    if (childTowardsReferencePath != null) {
      return childTowardsReferencePath;
    } else {
      if (basePath.isRoot())
        throw new AssertionError();
      /* The relation will need to be handled by the lowest common ancestor of basePath and
      targetPath for lateral references to work correctly. Go up one parent at the time, however, to
      enable optimizations that might take advantage of joins that might exist between intermediate
      tables, and to allow the same instance to be shared by lateral references to different
      relations. */
      return basePath.getParent();
    }
  }

  /** @return null if the path should be handled in this QueryBlock */
  private @Nullable RelationInstance requestInstance(RelationPath targetPath) {
    Preconditions.checkNotNull(targetPath);
    final RelationPath instancePath = getInstancePath(targetPath);
    if (instancePath == null)
      return null;
    RelationInstance ret = instances.get(instancePath);
    if (ret == null) {
      Preconditions.checkState(!frozen);
      final boolean leftJoin =
          basePath.isParentOf(instancePath) && !view.get(QProp.EXISTENTIAL_FILTER, instancePath) &&
          !(assumedExistingPaths.contains(instancePath) &&
          globals.querySettingsFlag(SettingsFlag.LEFT_JOIN_ELIMINATION_OPTIMIZATION));
      // In the aggregate case, instancePath is always ROOT, but basePath may be any relation.
      final boolean outward = (aggregationExpression == null) && instancePath.isParentOf(basePath);

      /* TODO[te]: Add a unit test which fails if the outward reference case is not handled
                   specially when newAssumedExistingPaths is computed. */
      // Outward references escape the context in which tuples may be assumed to exist.
      final ImmutableSet<RelationPath> newAssumedExistingPaths = outward
          ? ImmutableSet.copyOf(Sets.difference(assumedExistingPaths, ImmutableSet.of(basePath)))
          : assumedExistingPaths;
      final ImmutableSet<PrimitivePath> newOmittableFilterKeys = (aggregationExpression != null)
          ? ImmutableSet.copyOf(InternalQueryUtil.getNaiveGroupByPaths(view, basePath))
          : (outward ? null : omittableFilterKeys);

      if (aggregationExpression != null && !instancePath.equals(Path.ROOT))
        throw new AssertionError();
      final boolean instanceLocalCorrelated = outward;
      /* Under earlier semantics, filters were always considered to be applied after aggregates in
      the same relation. This turns out to be a bit non-sensical in the context of hidden fields
      with filters on them; in these cases there would be no output to actually apply the filter on,
      and the filter would simply be dropped. This in turn had the potential to produce incorrect
      aggregate results; see testAggregateHiddenFilter. */
      /* Don't apply filters for correlated blocks; they will be decorrelated against the original
      block, which will have been filtered. */
      final boolean instanceApplyFilters = !instanceLocalCorrelated;
      ret = new RelationInstance((outward ? globals.IF_CINST : globals.IF_QINST).fresh(),
          leftJoin, new QueryBlock(globals, instancePath, instanceLocalCorrelated,
          instanceApplyFilters, newAssumedExistingPaths, ImmutableList.of(), newOmittableFilterKeys,
          null));
      MapUtil.putChecked(instances, instancePath, ret);
    }
    return ret;
  }

  private Evaluation evaluateAggregation(
      Expression aggregationExpression, ImmutableSet<Path> alreadyVisitedPaths)
  {
    Preconditions.checkNotNull(aggregationExpression);
    Preconditions.checkNotNull(alreadyVisitedPaths);
    RelationInstance instance = aggregationInstances.get(aggregationExpression);
    if (instance == null) {
      Preconditions.checkState(!frozen);
      /* Don't apply filters in aggregate result blocks; they are handled by the enclosing
      non-aggregate block. */
      instance = new RelationInstance(globals.IF_AINST.fresh(), false,
          new QueryBlock(globals, basePath, false, false,
          assumedExistingPaths, ImmutableList.of(), null, aggregationExpression));
      MapUtil.putChecked(aggregationInstances, aggregationExpression, instance);
    }
    return instance.requestAggregation();
  }

  private NWEvaluation evaluatePath(PrimitivePath path, ImmutableSet<Path> alreadyVisitedPaths) {
    return evaluatePath(path, alreadyVisitedPaths, Sets.newLinkedHashSet());
  }

  /* TODO[pd]: Instead of taking the mutable set toInstanceDependencies, return a result object that
               the instance dependencies. See if I can put the logic for adding to
               localCorrelationPaths in one place at the same time. */
  private NWEvaluation evaluatePath(PrimitivePath path, ImmutableSet<Path> alreadyVisitedPaths,
      Set<RelationInstance> toInstanceDependencies)
  {
    if (!view.exists(path))
      return NWEvaluation.error();
    final RelationInstance instance = requestInstance(path.getParent());
    if (instance == null) {
      String columnName = view.get(QProp.SQL_COLUMN_NAME, path);
      if (columnName.isEmpty()) {
        if (alreadyVisitedPaths.contains(path)) {
          for (Path alreadyVisitedPath : alreadyVisitedPaths)
            addWarning(alreadyVisitedPath, new Warning("Circular reference in formula"));
          return NWEvaluation.error();
        }
        if (InternalQueryUtil.isAggregateFormula(view, path) &&
            !globals.getFunctionalDependencies().isIdentifiedBy(view, path,
            Sets.newLinkedHashSet(InternalQueryUtil.getNaiveGroupByPaths(view, basePath))))
        {
          // TODO[pd]: Consider if this case could actually be allowed.
          /* This warning could alternatively be rendered at the referenced path actually containing
          the aggregate function (if different from this path), however the "blame" for the error is
          technically at this path, since the latter is the path that is supposed to serve as the
          tuple key. */
          /* The analogous error in Postgres is "column X must appear in the GROUP BY clause or be
          used in an aggregate function". The actual logical constraint involves calculation of
          functional dependencies; we do this in an earlier step, when the tuple key is
          determined. See
          http://www.postgresql.org/docs/9.3/static/sql-select.html#SQL-GROUPBY . */
          addWarning(path, new Warning("Cannot combine aggregate function with reference that " +
              "might potentially yield more than one value"));
          return NWEvaluation.error();
        }
        /* Canonicalize to centralize the handling of non-existent paths, and to take advantage of
        other basic simplifications. */
        Expression exp = Expressions.canonicalize(view, view.get(QProp.FORMULA_EXPRESSION, path));
        localCorrelationPaths.add(path);
        return reportWarnings(path, evaluateExpression(exp,
            Functional.cons(alreadyVisitedPaths, path), toInstanceDependencies));
      } else {
        /* This would also be a good place to check for non-existent columns, though this would
        require another round trip to the database. The primary cached source of metadata about
        relations is the view itself, so under this policy, no further check is necessary. */
        if (InternalQueryUtil.getInstantiatedTable(view, basePath) == null) {
          addWarning(path, new Warning("No table instantiated"));
          return NWEvaluation.error();
        }
        if (frozen && !locallySelectedColumns.contains(columnName))
          throw new IllegalStateException();
        locallySelectedColumns.add(columnName);
        localCorrelationPaths.add(path);
        return NWEvaluation.simple(
            new RenderedExpression(localInstanceID.createColumnReferenceFragment(
                dialect.quoteIdentifier(columnName, true)), dialect.getStorageVariant()));
      }
    } else {
      toInstanceDependencies.add(instance);
      return NWEvaluation.simple(instance.requestFromSelectList(path, alreadyVisitedPaths));
    }
  }

  private RenderedExpression extractRenderedExpression(NWEvaluation eval) {
    /* TODO[pd]: Mark the column as unretrievable instead. Still, evaluate it as a NULL rather
                 than as an unprojected column. */
    return eval.isError() ? dialect.getNullLiteral() : eval.getRenderedExpression();
  }

  private NWEvaluation reportWarnings(Path warningPath, Evaluation evaluation) {
    for (Warning warning : evaluation.getWarnings())
      addWarning(warningPath, warning);
    return evaluation.isError()
        ? NWEvaluation.error() : NWEvaluation.simple(evaluation.getRenderedExpression());
  }

  private Evaluation evaluateFunction(Function function, ImmutableList<Evaluation> arguments) {
    final List<Warning> warnings = Lists.newArrayList();
    final List<RenderedExpression> argumentRenderedExpressions = Lists.newArrayList();
    for (Evaluation argument : arguments) {
      warnings.addAll(argument.getWarnings());
      argumentRenderedExpressions.add(argument.isError() ? null : argument.getRenderedExpression());
    }
    if (argumentRenderedExpressions.contains(null)) {
      return Evaluation.warnings(warnings);
    } else if (function.equals(Function.SYNTAX_ERROR)) {
      warnings.add(new Warning("Encountered error literal in expression"));
      return Evaluation.warnings(warnings);
    } else if (function.equals(Function.BROKEN_REFERENCE)) {
      warnings.add(new Warning("Encountered broken reference in expression"));
      return Evaluation.warnings(warnings);
    }
    FunctionImplementation functionImplementation =
        dialect.getFunctionImplementations().get(function);
    if (functionImplementation == null) {
      warnings.add(new Warning("Function '" + function + "' not supported by database backend"));
      return Evaluation.warnings(warnings);
    }
    try {
      return new Evaluation(functionImplementation.render(argumentRenderedExpressions), warnings);
    } catch (RenderFunctionException.WrongNumberOfArgumentsException e) {
      warnings.add(new Warning("Wrong number of arguments for function '" + function + "'"));
      return Evaluation.warnings(warnings);
    } catch (RenderFunctionException.DynamicRenderFunctionException e) {
      warnings.add(new Warning("Error translating function '" + function + "'" +
          " (" + e.getMessage() + ")"));
      return Evaluation.warnings(warnings);
    }
  }

  // TODO[cc]: Consider getting rid of this.
  private Evaluation evaluateExpression(
      Expression expression, final ImmutableSet<Path> alreadyVisitedPaths)
  {
    return evaluateExpression(expression, alreadyVisitedPaths, Sets.newLinkedHashSet());
  }

  private FunctionExpression rewriteRelationCount(FunctionExpression exp) {
    if (exp.getFunction().equals(Function.COUNT) && exp.getArguments().size() == 1) {
      Expression arg = exp.getArguments().get(0);
      if (arg instanceof PathExpression) {
        PathExpression pexp = (PathExpression) arg;
        if (pexp.getPath().isRelation()) {
          /* Apply some syntactic sugar to allow the COUNT function to be used with a relation path
          as an argument. */
          PrimitivePath tuplePresencePath =
              MapUtil.getChecked(globals.getTuplePresencePaths(), pexp.getPath().getRelation());
          return new FunctionExpression(Function.COUNT,
              ImmutableList.<Expression>of(new PathExpression(tuplePresencePath)));
        }
      }
    }
    return exp;
  }

  private Evaluation evaluateExpression(
      Expression expression, final ImmutableSet<Path> alreadyVisitedPaths,
      final Set<RelationInstance> toInstanceDependencies)
  {
    return expression.accept(new ExpressionVisitor<Evaluation,RuntimeException>() {
      @Override
      public Evaluation visit(ConstantExpression exp) {
        return NWEvaluation.simple(dialect.getLiteral(exp));
      }

      @Override
      public Evaluation visit(FunctionExpression exp) {
        if (Function.AGGREGATE_FUNCTIONS.contains(exp.getFunction()) &&
            !exp.equals(aggregationExpression))
        {
          return evaluateAggregation(exp, alreadyVisitedPaths);
        } else {
          FunctionExpression useExp = rewriteRelationCount(exp);
          List<Evaluation> arguments = Lists.newArrayList();
          for (Expression argument : useExp.getArguments()) {
            arguments.add(evaluateExpression(
                argument, alreadyVisitedPaths, toInstanceDependencies));
          }
          return evaluateFunction(useExp.getFunction(), ImmutableList.copyOf(arguments));
        }
      }

      @Override
      public Evaluation visit(PathExpression exp) {
        final Path path = exp.getPath();
        if (path.isRelation()) {
          return Evaluation.warning(new Warning("Relation path reference not allowed here"));
        }
        if (aggregationExpression != null) {
          for (PrimitivePath projectedPath : globals.getFunctionalDependencies().getKey(view,
              ImmutableSet.copyOf(Queries.getProjectedToRoot(view, path.getParent()))))
          {
            if (basePath.isEqualToOrAncestorOf(projectedPath))
              evaluatePath(projectedPath, alreadyVisitedPaths);
          }
        }
        final NWEvaluation ret = evaluatePath(
            path.getPrimitive(), alreadyVisitedPaths, toInstanceDependencies);
        if (ret.isError()) {
          return Evaluation.warning(
              new Warning("Error in referenced field(s)", Warning.Type.ERROR_IN_REFERENCED_FIELD));
        } else {
          return NWEvaluation.simple(ret.getRenderedExpression());
        }
      }
    });
  }
}
