// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Multimap;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.FunctionalDependencies;
import com.sieuferd.evaluator.impl.EvaluationSettings.SettingsFlag;
import com.sieuferd.evaluator.impl.SQLIdentifier.ColumnID;
import com.sieuferd.evaluator.impl.SQLIdentifier.IdentifierFactory;
import com.sieuferd.evaluator.impl.SQLIdentifier.InstanceID;
import java.util.Map;
import java.util.Set;

// TODO[cc]: Rename to QueryBuildingParam.
final class QueryBuildingGlobals {
  public final IdentifierFactory<ColumnID> IF_COL = ColumnID.newFactory("col");
  public final IdentifierFactory<ColumnID> IF_ACOL = ColumnID.newFactory("acol");
  /* TODO[cc]: Go over terminology again now that "instances" has come to attain several different
               meanings. */
  /** Local instance. */
  public final IdentifierFactory<InstanceID> IF_LINST = InstanceID.newFactory("linst");
  /** Query instance. */
  public final IdentifierFactory<InstanceID> IF_QINST = InstanceID.newFactory("qinst");
  /** Correlation instance. */
  public final IdentifierFactory<InstanceID> IF_CINST = InstanceID.newFactory("cinst");
  /** Aggregation instance. */
  public final IdentifierFactory<InstanceID> IF_AINST = InstanceID.newFactory("ainst");

  private final BackendDialect dialect;
  private final ImmutableView view;
  private final ImmutableSet<Path> syntheticPaths;
  private final ImmutableMap<RelationPath,PrimitivePath> tuplePresencePaths;
  private final EvaluationSettings settings;
  private final FunctionalDependencies functionalDependencies;
  private final Multimap<Path,Warning> warnings;

  /**
   * @param syntheticPaths paths that have been added by the query evaluator itself, and for which
   *        warnings should be attached to a parent node instead
   * @param tuplePresencePaths paths, one for each relation, for which a NULL value signifies an
   *        empty set of tuples (such NULL values are created as a result of left joins)
   */
  QueryBuildingGlobals(
      BackendDialect dialect, ImmutableView view, Set<Path> syntheticPaths,
      Map<RelationPath,PrimitivePath> tuplePresencePaths, EvaluationSettings settings,
      Multimap<Path,Warning> toWarnings)
  {
    Preconditions.checkNotNull(dialect);
    Preconditions.checkNotNull(view);
    Preconditions.checkNotNull(toWarnings);
    Preconditions.checkNotNull(settings);
    Preconditions.checkArgument(!syntheticPaths.contains(Path.ROOT));
    this.dialect = dialect;
    this.view = view;
    this.syntheticPaths = ImmutableSet.copyOf(syntheticPaths);
    this.tuplePresencePaths = ImmutableMap.copyOf(tuplePresencePaths);
    this.warnings = toWarnings;
    this.settings = settings;
    this.functionalDependencies = new FunctionalDependencies(view);
  }

  public BackendDialect getDialect() {
    return dialect;
  }

  public ImmutableView getView() {
    return view;
  }

  public ImmutableMap<RelationPath, PrimitivePath> getTuplePresencePaths() {
    return tuplePresencePaths;
  }

  public boolean querySettingsFlag(SettingsFlag flag) {
    return settings.queryFlag(flag);
  }

  public EvaluationSettings getSettings() {
    return settings;
  }

  public FunctionalDependencies getFunctionalDependencies() {
    return functionalDependencies;
  }

  public void addWarning(Path path, Warning warning) {
    if (syntheticPaths.contains(path)) {
      addWarning(path.getParent(), warning);
    } else {
      warnings.put(path, warning);
    }
  }
}
