// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableMultimap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.common.collect.LinkedListMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableData;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.OptionalInteger;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.CancelHandler;
import com.sieuferd.evaluator.DataProperties;
import com.sieuferd.evaluator.OrderedPath;
import com.sieuferd.evaluator.QueryEvaluator;
import com.sieuferd.evaluator.QueryUtil;
import com.sieuferd.evaluator.impl.EvaluationException.CanceledEvaluationException;
import com.sieuferd.query.QProp;
import com.sieuferd.query.Queries;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.evaluator.FunctionalDependencies;
import com.sieuferd.util.MapUtil;
import com.sieuferd.util.StringUtil;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;

public class QueryEvaluatorImpl extends QueryEvaluator {
  private final ImmutableView view;
  private final SQLConnection connection;
  private final CancelHandler cancelHandler;
  private final ImmutableMultimap<Path,Warning> preEvaluationWarnings;
  private final ImmutableMap<RelationPath,EvaluationBranch> branches;
  private final @Nullable PrintWriter logWriter;

  /* TODO[cc]: Avoid the logWriter UI detail in here; maybe abstract with a general-purpose progress
               interface. */
  public QueryEvaluatorImpl(View inputView, SQLConnection connection,
      @Nullable PrintWriter logWriter, CancelHandler cancelHandler, EvaluationSettings settings)
  {
    Preconditions.checkNotNull(inputView);
    Preconditions.checkNotNull(connection);
    Preconditions.checkNotNull(cancelHandler);
    Preconditions.checkNotNull(settings);
    this.connection = connection;
    this.logWriter = logWriter;
    this.cancelHandler = cancelHandler;
    MutableView toView = Views.newMutable(QueryUtil.QUERY_DEFAULTS);
    toView.copyFrom(inputView, Path.ROOT, QueryUtil.QUERY_PROPERTIES, null);
    final Set<Path> syntheticPaths = Sets.newLinkedHashSet();
    final Map<RelationPath,PrimitivePath> tuplePresencePaths = Maps.newLinkedHashMap();
    populateTuplePresencePaths(Path.ROOT, toView, syntheticPaths, tuplePresencePaths);
    final Multimap<Path,Warning> toWarnings = LinkedListMultimap.create();
    DataProperties.configureDataProperties(toView);
    this.view = ImmutableView.copyOf(toView);
    {
      BackendDialect dialect;
      try {
        dialect = connection.getBackendDialect();
      } catch (SQLException e) {
        connection.reportSQLException(e);
        toWarnings.put(Path.ROOT,
            new Warning("Could not resolve database dialect: " + e.getMessage()));
        dialect = null;
      }
      Map<RelationPath,EvaluationBranch> toBranches = Maps.newLinkedHashMap();
      if (dialect != null) {
        QueryBuildingGlobals globals = new QueryBuildingGlobals(dialect,
            this.view, syntheticPaths, tuplePresencePaths, settings, toWarnings);
        populateBranches(globals, connection, toWarnings, this.view, Path.ROOT,
            ImmutableList.of(), ImmutableSet.of(), toBranches, logWriter);
      }
      branches = ImmutableMap.copyOf(toBranches);
    }
    this.preEvaluationWarnings = ImmutableMultimap.copyOf(toWarnings);
  }

  private static void populateTuplePresencePaths(
      RelationPath path, MutableView view, Set<Path> syntheticPaths,
      Map<RelationPath,PrimitivePath> tuplePresencePaths)
  {
    PrimitivePath tuplePresencePath = null;
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation()) {
        populateTuplePresencePaths(
            childPath.getRelation(), view, syntheticPaths, tuplePresencePaths);
      } else {
        if (tuplePresencePath == null &&
            Queries.shouldProject(view, childPath) &&
            !InternalQueryUtil.isNullable(view, childPath.getPrimitive()))
        {
          tuplePresencePath = childPath.getPrimitive();
        }
      }
    }
    if (tuplePresencePath == null) {
      tuplePresencePath = PrimitivePath.create(path);
      view.addPath(tuplePresencePath);
      view.set(CProp.VISIBLE, tuplePresencePath, true);
      /* Could use any non-null constant here. Using a boolean avoids generating an extra query
      parameter. */
      view.set(QProp.FORMULA_EXPRESSION, tuplePresencePath, ConstantExpression.create(true));
      view.set(CProp.LABEL, tuplePresencePath, "{TPP}");
      MapUtil.addChecked(syntheticPaths, tuplePresencePath);
    }
    MapUtil.putChecked(tuplePresencePaths, path, tuplePresencePath);
  }

  private static void populateBranches(QueryBuildingGlobals globals,
      SQLConnection connection, Multimap<Path,Warning> toWarnings, ImmutableView view,
      RelationPath path, ImmutableList<OrderedPath> inheritedOrderedPaths,
      ImmutableSet<PrimitivePath> inheritedKeyPaths,
      Map<RelationPath,EvaluationBranch> toEvaluationBranches, PrintWriter logWriter)
  {
    if (!view.get(CProp.VISIBLE, path))
      return;
    final ImmutableList<OrderedPath> orderedPaths;
    {
      final List<OrderedPath> toOrderedPaths = Lists.newArrayList();
      final Set<PrimitivePath> fullyIdentifiedSoFar = Sets.newLinkedHashSet();
      final FunctionalDependencies fds = globals.getFunctionalDependencies();
      // Must add inheritedOrderedPaths regardless of whether nested or not.
      for (OrderedPath orderedPath : Iterables.concat(inheritedOrderedPaths,
          QueryUtil.getOrderedPaths(view, path, fds, true)))
      {
        if (!fds.isIdentifiedBy(view, orderedPath.getPath(), fullyIdentifiedSoFar)) {
          if (orderedPath.isCollationTotallyOrdered())
            fullyIdentifiedSoFar.add(orderedPath.getPath());
          toOrderedPaths.add(orderedPath);
        }
      }
      orderedPaths = ImmutableList.copyOf(toOrderedPaths);
    }
    /* Try to find the optimal set of key paths without regards to what the ordered paths are, since
    relations may be ordered on fields that cannot be reliably used for equality checking. An
    actually observed example was a relation ordered on an aggregate SUM whose exact result
    depended, due to rounding, on the non-deterministic order of input tuples. It is not always
    possible to use a different key in such situations, but we should do it whenever we can. */
    final ImmutableSet<PrimitivePath> keyPaths = ImmutableSet.<PrimitivePath>builder()
        .addAll(inheritedKeyPaths)
        .addAll(globals.getFunctionalDependencies().getKey(view,
            Sets.newLinkedHashSet(inheritedKeyPaths),
            ImmutableSet.copyOf(Queries.getImmediateProjectedChildren(view, path))))
        .build();
    final ImmutableSet<RelationPath> assumedExistingPaths = ImmutableSet.copyOf(
        Views.getRelationPathsWithAncestors(ImmutableSet.of(path)));
    final QueryBlock queryBlock = new QueryBlock(globals,
        globals.getSettings().isUsingRootBasedBranches() ? Path.ROOT : path, false, true,
        assumedExistingPaths, orderedPaths, null, null);
    for (OrderedPath orderedPath : orderedPaths)
      queryBlock.requestFromSelectList(orderedPath.getPath(), ImmutableSet.of());
    for (PrimitivePath projectedPath : Queries.getImmediateProjectedChildren(view, path))
      queryBlock.requestFromSelectList(projectedPath, ImmutableSet.of());
    queryBlock.freeze();
    GeneratedQuery generatedQuery = queryBlock.generateQuery();
    SQLFragment query = generatedQuery.getQuery();
    if (logWriter != null) {
      logWriter.println(query);
      logWriter.println();
    }
    if (!generatedQuery.getCorrelationParameters().isEmpty())
      throw new AssertionError("Expected no correlation parameters in the root query block");
    final EvaluationBranch branch;
    try {
      branch = new EvaluationBranch(connection, globals.getTuplePresencePaths(), path,
          queryBlock.getSelectedPathsInOrder(), query, keyPaths);
    } catch (SQLException e) {
      connection.reportSQLException(e);
      toWarnings.put(path,
          new Warning("Problem preparing SQL query: " + e.getMessage()));
      return;
    }
    MapUtil.putChecked(toEvaluationBranches, path, branch);
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation()) {
        populateBranches(globals, connection, toWarnings, view, childPath.getRelation(),
            ImmutableList.copyOf(orderedPaths), ImmutableSet.copyOf(keyPaths),
            toEvaluationBranches, logWriter);
      }
    }
  }

  @Override
  public void close() throws SQLException {
    SQLException toThrow = null;
    for (EvaluationBranch branch : branches.values()) {
      try {
        branch.close();
      } catch (SQLException e) {
        connection.reportSQLException(e);
        if (toThrow == null)
          toThrow = new SQLException("Failed to close QueryEvaluator");
        toThrow.addSuppressed(e);
      }
    }
    if (toThrow != null)
      throw toThrow;
  }

  private void evaluateBranch(Multimap<Path,Warning> toWarnings,
      RelationPath path, ResultData ret, KeyMatcher parentKeyMatcher, int limit)
  {
    Preconditions.checkArgument(ret.isProjected(path));
    try {
      final KeyMatcher keyMatcher;
      final List<RelationPath> projectedChildren = Lists.newArrayList();
      for (Path childPath : view.getChildPaths(path)) {
        if (childPath.isRelation() && ret.isProjected(childPath.getRelation()))
          projectedChildren.add(childPath.getRelation());
      }
      final EvaluationBranch branch = branches.get(path);
      if (branch == null)
        return;
      OptionalInteger limitOverride = view.get(QProp.TUPLE_RESULT_LIMIT, path);
      /* An easy optimization at this point would be to not bother populating a KeyMatcher if
      projectedChildren is empty. However, KeyMatcher does provide a sanity check against duplicate
      key values in the query results, which has caught query evaluator bugs in the past. Such a
      sanity check is valuable even in production. For this reason, populate a KeyMatcher even for
      branches without children of their own. */
      keyMatcher = new KeyMatcher(ret, branch.getKeyPaths(), false);
      branch.evaluate(ret, parentKeyMatcher, keyMatcher,
          limitOverride.isPresent() ? Math.max(limitOverride.get(), 0) : limit, cancelHandler);
      for (RelationPath childPath : projectedChildren)
        evaluateBranch(toWarnings, childPath, ret, keyMatcher, limit);
    } catch (CanceledEvaluationException e) {
      ret.markLimitReached(Path.ROOT, true);
      Throwable cause = e.getCause();
      if (cause instanceof SQLException) {
        toWarnings.put(path, new Warning("The query was canceled (backend said '" +
            StringUtil.trimmedAndEmptyIfNull(e.getMessage()) + "')"));
      } else {
        toWarnings.put(path, new Warning("The query was canceled"));
      }
    } catch (SQLException e) {
      connection.reportSQLException(e);
      toWarnings.put(path, new Warning("SQL execution problem: '" +
          StringUtil.trimmedAndEmptyIfNull(e.getMessage()) + "'"));
    } catch (EvaluationException e) {
      toWarnings.put(path, new Warning("Evaluation problem: " +
          StringUtil.trimmedAndEmptyIfNull(e.getMessage())));
    }
  }

  @Override
  public ImmutableData evaluate(int limit, Multimap<Path,Warning> toWarnings) {
    Preconditions.checkNotNull(toWarnings);
    toWarnings.putAll(preEvaluationWarnings);
    /* TODO[pd]: Restart aborted transactions (I think at least Oracle can abort even read-only
                 transactions). */
    final Connection jdbcConnection;
    final boolean oldAutoCommit;
    final boolean oldReadOnly;
    final int oldTransactionIsolation;
    try {
      jdbcConnection = connection.getJDBCConnection();
      oldAutoCommit = jdbcConnection.getAutoCommit();
      oldReadOnly = jdbcConnection.isReadOnly();
      oldTransactionIsolation = jdbcConnection.getTransactionIsolation();
    } catch (SQLException e) {
      connection.reportSQLException(e);
      toWarnings.put(Path.ROOT,
          new Warning("Failed to retrieve initial connection settings: " + e.getMessage()));
      return QueryUtil.EMPTY_INCOMPLETE_DATA;
    }
    ResultData ret;
    try {
      try {
        jdbcConnection.setAutoCommit(false);
        jdbcConnection.setReadOnly(true);
        jdbcConnection.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
      } catch (SQLException e) {
        connection.reportSQLException(e);
        toWarnings.put(Path.ROOT,
            new Warning("Failed to prepare transaction state: " + e.getMessage()));
        return QueryUtil.EMPTY_INCOMPLETE_DATA;
      }
      ret = new ResultData(view);
      evaluateBranch(toWarnings,
          Path.ROOT, ret, new KeyMatcher(ret, ImmutableSet.of(), true), limit);
    } finally {
      try {
        jdbcConnection.rollback();
        jdbcConnection.setTransactionIsolation(oldTransactionIsolation);
        jdbcConnection.setReadOnly(oldReadOnly);
        jdbcConnection.setAutoCommit(oldAutoCommit);
      } catch (SQLException e) {
        connection.reportSQLException(e);
        toWarnings.put(Path.ROOT,
            new Warning("Failed to clean up after transaction: " + e.getMessage()));
      }
    }
    return ImmutableData.create(ret);
  }
}
