// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.evaluator.impl.RowFetcher.Row;
import com.sieuferd.util.MapUtil;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.concurrent.NotThreadSafe;

/** Mutable until freeze() is called. */
@NotThreadSafe
public final class ResultData extends Data {
  // TODO[te]: Consider making these actual unit tests.
  private static final boolean TEST_PARENT_TUPLE = true;
  private static final boolean TEST_COMPLETE_TUPLE_COUNT = true;
  private final ImmutableView view;
  private final ImmutableMap<RelationPath,DataBranch> branches;
  private boolean incomplete;

  public ResultData(View view) {
    // TODO[pd]: Consider excluding any path that's not projected.
    MutableView toView = Views.newMutable();
    copyProjectedPaths(Path.ROOT, view, toView);
    this.view = ImmutableView.copyOf(toView);
    Map<RelationPath,DataBranch> toBranches = Maps.newLinkedHashMap();
    populateBranches(toBranches, Path.ROOT);
    this.branches = ImmutableMap.copyOf(toBranches);
  }

  private void copyProjectedPaths(Path path, View src, MutableView dest) {
    if (path.isRelation()) {
      dest.set(CProp.PLURAL, path, src.get(CProp.PLURAL, path));
      for (Path childPath : src.getChildPaths(path.getRelation())) {
        if (src.get(CProp.PROJECTED, childPath)) {
          dest.addPath(childPath);
          copyProjectedPaths(childPath, src, dest);
        }
      }
    } else {
      dest.set(CProp.IN_TUPLE_KEY, path, src.get(CProp.IN_TUPLE_KEY, path));
    }
    dest.set(CProp.PROJECTED, path, src.get(CProp.PROJECTED, path));
  }

  private void populateBranches(Map<RelationPath,DataBranch> toBranches, RelationPath path) {
    MapUtil.putChecked(toBranches, path, new DataBranch(path));
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation())
        populateBranches(toBranches, childPath.getRelation());
    }
  }

  /* TODO[pd]: Have incomplete relations be the default until the opposite is indicated so that
               unprojected relations behave like incomplete empty ones. */
  public void markLimitReached(RelationPath path, boolean orderedLimit) {
    Preconditions.checkArgument(!isFrozen());
    if (!orderedLimit)
      throw new UnsupportedOperationException("Unordered limits not yet supported");
    incomplete = true;
    DataBranch branch = MapUtil.getChecked(branches, path);
    branch.markLimitReached(branch.getTotalTupleCount());
  }

  public Tuple addRelationTuple(RelationPath path, Tuple parentTuple, Row row) {
    Preconditions.checkNotNull(path);
    Preconditions.checkNotNull(parentTuple);
    Preconditions.checkNotNull(row);
    Preconditions.checkArgument(!isFrozen());
    return MapUtil.getChecked(branches, path).addTuple(parentTuple, row);
  }

  @Override
  public TupleInterval getRelationTuples(RelationPath path, Tuple parentTuple) {
    Preconditions.checkArgument(isProjected(path));
    return checkRelationTuples(MapUtil.getChecked(branches, path).getTuples(parentTuple), path);
  }

  /** @return an upper limit on the number of complete relations at the specified path */
  private int getCompleteRelationLimit(RelationPath path) {
    int tupleLimit = getCompleteTupleLimit(path.getRelation());
    /* A limit needs to be passed to the branch because in this implementation, we assume that
    branches do not know about their complete tuple limits (and that they should not make
    recursive calls themselves to find out). */
    return MapUtil.getChecked(branches, path.getRelation()).getCompleteRelationLimit(tupleLimit);
  }

  /** @return an upper limit on the number of complete tuples at the specified path */
  private int getCompleteTupleLimit(RelationPath path) {
    int ret = Integer.MAX_VALUE;
    for (Path childPath : view.getChildPaths(path)) {
      if (childPath.isRelation())
        ret = Math.min(ret, getCompleteRelationLimit(childPath.getRelation()));
    }
    return ret;
  }

  /* This alternative implementation and its associated methods is left for the record and for
  testing purposes. */
  private int getCompleteTupleCountAlternativeImplementation(RelationPath path) {
    int recRes = getCompleteTupleLimit(path);
    return Math.min(MapUtil.getChecked(branches, path).getTotalTupleCount(), recRes);
  }

  @Override
  public int getCompleteTupleCount(RelationPath path) {
    Preconditions.checkArgument(isProjected(path));
    int ret = MapUtil.getChecked(branches, path).getCompleteTupleCount();
    if (TEST_COMPLETE_TUPLE_COUNT && ret != getCompleteTupleCountAlternativeImplementation(path))
      throw new AssertionError();
    return ret;
  }

  @Override
  public Tuple getParentTuple(RelationPath childPath, Tuple childTuple) {
    return MapUtil.getChecked(branches, childPath).getParentTuple(childTuple);
  }

  @Override
  public boolean isIncomplete() {
    if (getCompleteRelationLimit(Path.ROOT) < Integer.MAX_VALUE != incomplete)
      throw new AssertionError();
    return incomplete;
  }

  @Override
  public PrimitiveValue getTupleField(Tuple tuple, PrimitivePath path) {
    Preconditions.checkArgument(isProjected(path));
    RelationPath parentPath = path.getParent();
    tuple.sanityCheck(this, parentPath);
    return MapUtil.getChecked(branches, parentPath).getTupleField(tuple, path);
  }

  @Override
  public boolean isProjected(Path path) {
    return view.exists(path);
  }

  // TODO[pd]: Consider if these checks are still relevant, and whether anything should be added.
  private TupleInterval checkRelationTuples(final TupleInterval tuples, final RelationPath path) {
    if (!view.get(CProp.PLURAL, path) && tuples.size() > 1)
      throw new AssertionError("Expected at most one tuple at non-plural path " + path);
    final ImmutableSet<PrimitivePath> expectedFields;
    {
      Set<PrimitivePath> toExpectedFields = Sets.newLinkedHashSet();
      for (Path childPath : view.getChildPaths(path)) {
        if (!childPath.isRelation())
          toExpectedFields.add(childPath.getPrimitive());
      }
      expectedFields = ImmutableSet.copyOf(toExpectedFields);
    }
    return new TupleInterval() {
      @Override
      public boolean isComplete() {
        return tuples.isComplete();
      }

      @Override
      public int size() {
        return tuples.size();
      }

      @Override
      public int getFirstTupleIndex() {
        return tuples.getFirstTupleIndex();
      }

      @Override
      public Tuple get(int index) {
        // TODO[te]: Consider checking for duplicate keys.
        final Tuple tuple = tuples.get(index);
        tuple.sanityCheck(ResultData.this, path);
        // Fail-fast.
        for (PrimitivePath expectedField : expectedFields)
          getTupleField(tuple, expectedField);
        return tuple;
      }
    };
  }

  private final class DataBranch {
    private final RelationPath branchPath;
    /** Indicies are parent tuple indicies, values are indices into dataByTupleIndex in ascending
    order, denoting the index of the first child tuple. */
    private final List<Integer> tupleIndexByParentTupleIndex = Lists.newArrayList();
    private final ImmutableMap<PrimitivePath,List<PrimitiveValue>> dataByTupleIndex;
    private int nextTupleIndex;
    private boolean reachedLimit;
    private int completeTupleLimit = Integer.MAX_VALUE;

    DataBranch(RelationPath branchPath) {
      Preconditions.checkNotNull(branchPath);
      this.branchPath = branchPath;
      Map<PrimitivePath,List<PrimitiveValue>> toDataByTupleIndex = Maps.newLinkedHashMap();
      for (Path path : view.getChildPaths(branchPath)) {
        if (!path.isRelation())
          MapUtil.putChecked(toDataByTupleIndex, path.getPrimitive(), Lists.newArrayList());
      }
      dataByTupleIndex = ImmutableMap.copyOf(toDataByTupleIndex);
    }

    /* TODO[pd]: Consider storing the parent tuple index of each child tuple to make this O(1), or
                 review all call sites to ensure this is not necessary. */
    private int getParentTupleIndex(int childTupleIndex) {
      Preconditions.checkArgument(childTupleIndex >= 0);
      Preconditions.checkArgument(childTupleIndex < getTotalTupleCount());
      int bsRes = Collections.binarySearch(tupleIndexByParentTupleIndex, childTupleIndex);
      if (bsRes >= 0) {
        /* In the case of an exact match, the child tuple is the first one in the parent relation,
        since tupleIndexByParentTupleIndex stores indicies of first child tuples. If there are
        multiple successive matches, the correct parent tuple must be the last matching one, since
        the preceding ones would have empty child relations. */
        while (bsRes + 1 < tupleIndexByParentTupleIndex.size() &&
               tupleIndexByParentTupleIndex.get(bsRes + 1) == childTupleIndex)
        {
          bsRes++;
        }
        return bsRes;
      } else {
        // For clarity, first reverse the binarySearch() return value convention.
        final int insertionPoint = -(bsRes + 1);
        /* The child tuples following the first one will get an insertion point right after the
        parent's index, so subtract one. */
        final int ret = insertionPoint - 1;
        // This error should have been caught by the argument check at the beginning of the method.
        if (ret < 0 || ret >= tupleIndexByParentTupleIndex.size())
          throw new AssertionError();
        return ret;
      }
    }

    private Tuple getParentTuple(Tuple childTuple) {
      childTuple.sanityCheck(ResultData.this, branchPath);
      if (branchPath.isRoot())
        return Tuple.createRootParent(ResultData.this);
      final RelationPath parentPath = branchPath.getParent();
      return Tuple.create(ResultData.this, parentPath,
          getParentTupleIndex(childTuple.getTupleIndex()));
    }

    /** @return an upper limit on the number of complete relations at this branch given the
                specified number of tuples */
    private int getCompleteRelationLimit(int completeTupleLimit) {
      int ret = Integer.MAX_VALUE;
      if (completeTupleLimit < getTotalTupleCount())
        ret = Math.min(ret, getParentTupleIndex(completeTupleLimit));
      if (reachedLimit)
        ret = Math.min(ret, Math.max(tupleIndexByParentTupleIndex.size() - 1, 0));
      return ret;
    }

    public int getTotalTupleCount() {
      return nextTupleIndex;
    }

    public int getCompleteTupleCount() {
      return Math.min(nextTupleIndex, completeTupleLimit);
    }

    public void markLimitReached(int atTupleCount) {
      this.reachedLimit = true;
      completeTupleLimit = Math.min(completeTupleLimit, atTupleCount);
      if (!branchPath.isRoot()) {
        MapUtil.getChecked(branches, branchPath.getParent())
            .markLimitReached(getCompleteRelationLimit(completeTupleLimit));
      }
    }

    public TupleInterval getTuples(Tuple parentTuple) {
      checkParentTuple(parentTuple);
      // Inclusive
      final int firstTupleIndex = getFirstChildTupleIndex(parentTuple.getTupleIndex());
      // Exclusive
      final int lastTupleIndex = getFirstChildTupleIndex(parentTuple.getTupleIndex() + 1);
      if (lastTupleIndex < firstTupleIndex)
        throw new AssertionError();
      final int size = lastTupleIndex - firstTupleIndex;
      // TODO[pd]: Deal with this eventually.
      final boolean orderedLimit = true;
      /* The result can be incomplete only if the number of tuples in the result reached the query
      limit. If the result is limited and the result is unordered, every relation is incomplete. In
      the case of an ordered limit, the last relation and every (empty) relation beyond is
      incomplete. */
      final boolean complete =
          !reachedLimit ||
          (orderedLimit && (firstTupleIndex + size) < nextTupleIndex) ||
          // For the non-plural case we know there can be no additional tuples beyond the first.
          size == 1 && !view.get(CProp.PLURAL, branchPath);
      final TupleInterval ret = new TupleInterval.PluralTupleInterval(
          ResultData.this, branchPath, firstTupleIndex, size, complete);

      if (TEST_PARENT_TUPLE && !branchPath.isRoot()) {
        for (Tuple childTuple : ret) {
          Tuple derivedParentTuple = getParentTuple(childTuple);
          if (!derivedParentTuple.equalsTuple(parentTuple))
            throw new AssertionError();
        }
      }

      return ret;
    }

    private int getFirstChildTupleIndex(int parentTupleIndex) {
      Preconditions.checkArgument(parentTupleIndex >= 0);
      if (parentTupleIndex >= tupleIndexByParentTupleIndex.size()) {
        /* This parent tuple doesn't have any child tuples, and there are no more child tuples in
        any of the tuples following the parent tuples either, thus our parent tuple list stops
        short. This is normal. */
        return nextTupleIndex;
      } else {
        return tupleIndexByParentTupleIndex.get(parentTupleIndex);
      }
    }

    public PrimitiveValue getTupleField(Tuple tuple, PrimitivePath path) {
      return MapUtil.getChecked(dataByTupleIndex, path).get(tuple.getTupleIndex());
    }

    public Tuple addTuple(Tuple parentTuple, Row row) {
      checkParentTuple(parentTuple);
      Preconditions.checkState(!reachedLimit);
      final int parentTupleIndex = parentTuple.getTupleIndex();
      if (nextTupleIndex == Integer.MAX_VALUE)
        throw new IndexOutOfBoundsException("Exceeded maximum number of tuples");
      if (parentTupleIndex < tupleIndexByParentTupleIndex.size() - 1)
        throw new IllegalArgumentException("Parent tuple indicies must increase monotonically");
      while (parentTupleIndex >= tupleIndexByParentTupleIndex.size())
        tupleIndexByParentTupleIndex.add(nextTupleIndex);
      Preconditions.checkState(tupleIndexByParentTupleIndex.size() - 1 == parentTupleIndex);
      final int tupleIndex = nextTupleIndex;
      Tuple tuple = Tuple.create(ResultData.this, branchPath, tupleIndex);
      nextTupleIndex++;
      for (Entry<PrimitivePath,List<PrimitiveValue>> entry : dataByTupleIndex.entrySet())
        entry.getValue().add(row.getValue(entry.getKey()));
      return tuple;
    }

    private void checkParentTuple(Tuple parentTuple) {
      parentTuple.sanityCheck(ResultData.this,
            branchPath.isRoot() ? null : branchPath.getParent());
    }
  }
}
