// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Joiner;
import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.database.dialect.RenderVariant;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.util.MapUtil;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;

// TODO[cc]: Consider implemeting Iterator.
/** Mutable. */
public final class RowFetcher implements AutoCloseable {
  private final BackendDialect dialect;
  private final ImmutableMap<PrimitivePath,Integer> pathIndices;
  /** Immutable. */
  private final ResultColumn[] columns;
  private final ResultSet resultSet;
  /** May contain nulls. */
  private final Object[] values;
  private @Nullable Row currentRow;

  public RowFetcher(BackendDialect dialect, PreparedStatement ps,
      Map<PrimitivePath,RenderVariant> selectedPathsInOrder)
      throws SQLException
  {
    Preconditions.checkNotNull(dialect);
    this.dialect = dialect;
    this.resultSet = ps.executeQuery();

    {
      final Map<PrimitivePath,Integer> toPathIndices = Maps.newLinkedHashMap();
      final List<ResultColumn> toColumns = Lists.newArrayList();
      int columnIndex = 1;
      int valuesArrayIndex = 0;
      final ResultSetMetaData metaData = resultSet.getMetaData();

      for (Entry<PrimitivePath,RenderVariant> entry : selectedPathsInOrder.entrySet()) {
        final PrimitivePath selectedPath = entry.getKey();
        toColumns.add(new ResultColumn(columnIndex,
            SQLType.createFromResultSetMetaData(metaData, columnIndex), entry.getValue()));
        MapUtil.putChecked(toPathIndices, selectedPath.getPrimitive(), valuesArrayIndex);
        valuesArrayIndex++;
        columnIndex++;
      }
      this.pathIndices = ImmutableMap.copyOf(toPathIndices);
      this.columns = toColumns.toArray(new ResultColumn[toColumns.size()]);
    }
    this.values = new Object[this.columns.length];
  }

  public Set<PrimitivePath> getPaths() {
    return pathIndices.keySet();
  }

  /** @return null if the end of the result set has been reached; the returned object is only valid
              until this method is called again */
  public @Nullable Row fetchRow() throws SQLException {
    currentRow = null;
    if (!resultSet.next())
      return null;
    /* According to the ResultSet Javadoc, it's most efficient to use column indicies rather than
    column names, and most portable to iterate over columns left-to-right. Calling fetchValue in the
    order used here ensures that both recommendations are followed. */
    for (int i = 0; i < columns.length; i++)
      values[i] = columns[i].fetchValue(dialect, resultSet);
    currentRow = new Row();
    return currentRow;
  }

  @Override
  public void close() throws SQLException {
    resultSet.close();
  }

  public final class Row {
    private Row() { }

    private @Nullable Object getValueObject(PrimitivePath path) {
      Preconditions.checkState(currentRow == this);
      return values[MapUtil.getChecked(pathIndices, path)];
    }

    public boolean valueIsNull(PrimitivePath path) {
      return getValueObject(path) == null;
    }

    // TODO[cc]: Consider getting rid of PrimitiveValue.
    public PrimitiveValue getValue(PrimitivePath path) {
      return new PrimitiveValue(getValueObject(path));
    }

    @Override
    public String toString() {
      List<String> ret = Lists.newArrayList();
      for (PrimitivePath path : getPaths())
        ret.add(path.toString() + "=" + getValue(path).toString());
      return "(" + Joiner.on(", ").join(ret) + ")";
    }
  }
}
