// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator.impl;

import com.google.common.base.Preconditions;
import com.sieuferd.database.SQLFragment;
import java.util.regex.Pattern;
import javax.annotation.concurrent.Immutable;

/* TODO[cc]: Consider storing a boolean parameter identifying whether the ID would need quoting or
             not, so that even table column names can be stored in a ColumnID, and code factored
             out. */
/**
 * Container for generated SQL identifiers. Identifiers held in this class should not need quoting.
 */
@Immutable
abstract class SQLIdentifier {
  private final Pattern SANITY_CHECK_PATTERN = Pattern.compile("^[A-Za-z][A-Za-z0-9_]*$");
  private final String identifier;

  protected SQLIdentifier(String identifier) {
    Preconditions.checkNotNull(identifier);
    /* Just a sanity check; there's no foolproof way of testing whether an arbitrary identifier will
    need quoting. Still, it's easy enough for the client to just pick some particular safe name or
    prefix. */
    Preconditions.checkArgument(SANITY_CHECK_PATTERN.matcher(identifier).matches());
    this.identifier = identifier;
  }

  @Override
  public String toString() {
    return identifier;
  }

  public SQLFragment toSQLFragment() {
    return SQLFragment.create(identifier);
  }

  public String getIdentifier() {
    return identifier;
  }

  @Override
  public final int hashCode() {
    return identifier.hashCode();
  }

  @Override
  public final boolean equals(Object obj) {
    if (obj == this)
      return true;
    if (!obj.getClass().equals(this.getClass()) || !(obj instanceof SQLIdentifier))
      return false;
    SQLIdentifier other = (SQLIdentifier) obj;
    return other.identifier.equals(this.identifier);
  }

  public static interface IdentifierFactory<T extends SQLIdentifier> {
    public T fresh();
  }

  final static class ColumnID extends SQLIdentifier {
    ColumnID(String identifier) {
      super(identifier);
    }

    public static IdentifierFactory<ColumnID> newFactory(final String prefix) {
      return new IdentifierFactory<ColumnID>() {
        long counter = 1;

        @Override
        public ColumnID fresh() {
          return new ColumnID(prefix + (counter++));
        }
      };
    }
  };

  final static class InstanceID extends SQLIdentifier {
    InstanceID(String identifier) {
      super(identifier);
    }

    public static IdentifierFactory<InstanceID> newFactory(final String prefix) {
      return new IdentifierFactory<InstanceID>() {
        long counter = 1;

        @Override
        public InstanceID fresh() {
          return new InstanceID(prefix + (counter++));
        }
      };
    }

    public SQLFragment createColumnReferenceFragment(SQLFragment columnFragment) {
      return SQLFragment.builder()
          .appendSQLFragment(toSQLFragment())
          .appendJDBCString(".")
          .appendSQLFragment(columnFragment)
          .build();
    }
  };
}
