// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.LinkedListMultimap;
import com.google.common.collect.Multimap;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.ViewDefaults;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.impl.EvaluationSettings;
import com.sieuferd.query.Filter;
import com.sieuferd.query.InstanceDefinition;
import com.sieuferd.query.QProp;
import com.sieuferd.datamodel.expressions.ExpressionErrorReport.ExpressionParseException;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.Expressions;
import java.sql.SQLException;
import java.util.Map;
import java.util.Set;
import javax.annotation.Nullable;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.BeforeClass;

public abstract class EvaluatorTestBase {
  public static final boolean SKIP_KNOWN_FAILING_TESTS = true;
  private static final boolean INCLUDE_OPTIMIZATION_TESTS = true;
  /* This makes the test suite take three times as long to complete, and is probably a bit overkill.
  Rely on specific tests being marked for wrapping instead. */
  private static final boolean RUN_ALL_TESTS_WRAPPED = false;
  private static final TestConfiguration CONFIGURATION = TestConfiguration.POSTGRES;
  // TODO[cc]: Make this private.
  protected static final EvaluatorTestState TEST_STATE = new EvaluatorTestState(CONFIGURATION);

  @BeforeClass
  public static void beforeClass() throws SQLException {
    TEST_STATE.setup();
  }

  @AfterClass
  public static void afterClass() throws SQLException {
    TEST_STATE.close();
  }

  @After
  public void after() throws SQLException {
    TEST_STATE.clean();
  }

  protected static MutableView createView() {
    MutableView ret = Views.newMutable(ViewDefaults.builder()
        .setAll(CProp.DEFAULTS)
        .setAll(QProp.DEFAULTS)
        .build());
    ret.set(CProp.LABEL, Path.ROOT, "root");
    return ret;
  }

  protected static void setInstantiatedTable(
      MutableView view, RelationPath path, @Nullable String tableName)
  {
    view.set(QProp.INSTANTIATED_TABLE, path, new InstanceDefinition(
        tableName == null ? null : TEST_STATE.getTableHandle(tableName)));
  }

  protected static void setFormula(MutableView view, PrimitivePath path, String formula) {
    try {
      view.set(QProp.FORMULA_EXPRESSION, path,
          Expressions.parse(Strepco.createNormalized(view), formula));
    } catch (ExpressionParseException e) {
      throw new RuntimeException(e);
    }
  }

  protected static void setFilter(MutableView view, Path path, String filter) {
    Strepco strepco = Strepco.createNormalized(view, Views.getContextPath(path));
    try {
      view.set(QProp.FILTER, path,
          Filter.fromExpression(Expressions.parse(strepco, filter)));
    } catch (ExpressionParseException e) {
      throw new RuntimeException(e);
    }
  }

  protected static PrimitivePath addSQLColumn(
      MutableView view, RelationPath parentPath, String name)
  {
    PrimitivePath ret = view.addPrimitivePath(parentPath);
    view.set(QProp.SQL_COLUMN_NAME, ret, name);
    view.set(CProp.LABEL, ret, name);
    return ret;
  }

  protected static PrimitivePath addFormulaColumn(
      MutableView view, RelationPath parentPath, String name, String formula)
  {
    PrimitivePath ret = view.addPrimitivePath(parentPath);
    view.set(QProp.SQL_COLUMN_NAME, ret, "");
    setFormula(view, ret, formula);
    view.set(CProp.LABEL, ret, name);
    return ret;
  }

  protected static RelationPath addRelation(MutableView view, RelationPath parentPath, String name) {
    RelationPath ret = view.addRelationPath(parentPath);
    view.set(CProp.LABEL, ret, name);
    return ret;
  }

  protected void expectColumnNames(String tableName, String ... columnNames) {
    try {
      Assert.assertEquals(ImmutableList.copyOf(columnNames),
          ImmutableList.copyOf(TEST_STATE.getColumnNames(tableName, false)));
    } catch (SQLException e) {
      throw new RuntimeException(e);
    }
  }

  protected void expectPKColumnNames(String tableName, String ... columnNames) {
    try {
      Assert.assertEquals(ImmutableSet.copyOf(columnNames),
          TEST_STATE.getColumnNames(tableName, true));
    } catch (SQLException e) {
      throw new RuntimeException(e);
    }
  }

  public static final View wrapViewInNewRoot(View input) {
    MutableView ret = createView();
    final RelationPath nestedRelationPath = ret.addRelationPath(Path.ROOT);
    final PathMap pathMap = ret.copyFrom(input, Path.ROOT, nestedRelationPath);
    Views.rewritePathReferences(ret, nestedRelationPath, pathMap);
    return ImmutableView.copyOf(ret);
  }

  protected void expectResultWrapped(
      View view, boolean sortTuples, String expectedResult)
  {
    final int wrapLevels = 2;
    Preconditions.checkArgument(wrapLevels >= 0);
    View curView = view;
    String curExpectedResult = expectedResult;
    for (int i = 0; i < wrapLevels; i++) {
      expectResult(curView, sortTuples, curExpectedResult, ImmutableSet.of());
      if (i == wrapLevels)
        break;
      curView = wrapViewInNewRoot(curView);
      curExpectedResult = "[(" + curExpectedResult + ")]";
    }
  }

  protected void expectResult(View view, boolean sortTuples, String expectedResult) {
    if (RUN_ALL_TESTS_WRAPPED) {
      expectResultWrapped(view, sortTuples, expectedResult);
    } else {
      expectResult(view, sortTuples, expectedResult, ImmutableSet.of());
    }
  }

  protected void expectResult(
      View view, boolean sortTuples, String expectedResult, Set<Path> expectedWarningPaths)
  {
    if (INCLUDE_OPTIMIZATION_TESTS) {
      for (EvaluationSettings settings : SettingsIterable.INSTANCE)
        expectResult(view, sortTuples, expectedResult, expectedWarningPaths, settings);
    } else {
      expectResult(view, sortTuples, expectedResult, expectedWarningPaths,
          EvaluationSettings.create(false));
    }
  }

  protected void expectResult(
      View view, boolean sortTuples, String expectedResult, Set<Path> expectedWarningPaths,
      EvaluationSettings settings)
  {
    Multimap<Path,Warning> actualWarnings = LinkedListMultimap.create();
    Data data;
    try {
      data = TEST_STATE.evaluateView(view, settings, actualWarnings);
    } catch (SQLException e) {
      throw new RuntimeException(e);
    }
    if (!expectedWarningPaths.equals(actualWarnings.keySet())) {
      if (actualWarnings.isEmpty()) {
        throw new RuntimeException("Expected warnings, got none");
      } else {
        StringBuilder sb = new StringBuilder();
        sb.append("Unexpected set of evaluation warnings (expected ")
            .append(expectedWarningPaths).append("):\n");
        for (Map.Entry<Path,Warning> entry : actualWarnings.entries()) {
          sb.append("  ")
              .append(entry.getKey().toString(view))
              .append(": ")
              .append(entry.getValue().getDisplayMessage())
              .append("\n");
        }
        throw new RuntimeException(sb.toString());
      }
    }
    // Do this after first checking the error list, otherwise failed tests may appear confusing.
    String actualResult = EvaluatorTestUtil.getDataString(view, data, sortTuples);
    if (!expectedResult.equals(actualResult)) {
      System.out.println("Expected: " + expectedResult);
      System.out.println("Actual  : " + actualResult);
    }
    Assert.assertEquals(expectedResult, actualResult);
  }

  public void expectResult(String expression, String result) {
    MutableView view = createView();
    PrimitivePath colA = addFormulaColumn(view, Path.ROOT, "formula", expression);
    view.set(CProp.VISIBLE, colA, true);
    expectResult(view, true, "[(" + result + ")]");
    TEST_STATE.fulfillFunctionTestRequirement(view.get(QProp.FORMULA_EXPRESSION, colA));
  }

  public void createTable(
      String name, String columnNames[], Class<?> columnTypes[], Object[][] data)
  {
    try {
      TEST_STATE.createTable(name, columnNames, columnTypes, data);
    } catch (SQLException e) {
      throw new RuntimeException(e);
    }
  }

  public void createTable(String name, Object data[]) {
    try {
      TEST_STATE.createTable(name, data);
    } catch (SQLException e) {
      throw new RuntimeException(e);
    }
  }

  public void createPassthroughView(String name, String targetTableName) {
    try {
      TEST_STATE.createPassthroughView(name, targetTableName);
    } catch (SQLException e) {
      throw new RuntimeException(e);
    }
  }
}
