// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLFragment;
import com.sieuferd.database.SQLInserter;
import com.sieuferd.database.SQLTableInfo;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.dialect.BackendDialect;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.datamodel.proptypes.Warning;
import com.sieuferd.evaluator.impl.EvaluationSettings;
import com.sieuferd.evaluator.impl.QueryEvaluatorImpl;
import com.sieuferd.datamodel.expressions.Function;
import com.sieuferd.query.SQLTable;
import com.sieuferd.datamodel.expressions.Expression;
import com.sieuferd.util.MapUtil;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;

final class EvaluatorTestState implements AutoCloseable {
  private static final boolean KEEP_TEST_TABLES_AFTER_TEST = false;
  private final TestConfiguration configuration;
  // Values are true if the table is in fact a view.
  private final Map<SQLTable,Boolean> createdTables = Maps.newLinkedHashMap();
  private final PrintWriter logWriter = new PrintWriter(System.out);
  private final FunctionTestRequirements functionTestRequirements = new FunctionTestRequirements();
  // Not currently used for much.
  private final Map<TestTableContents,String> uniqueTestTables = Maps.newLinkedHashMap();
  private int evaluationCount;
  private @Nullable SQLConnection connection;
  private @Nullable BackendDialect dialect;

  EvaluatorTestState(TestConfiguration configuration) {
    Preconditions.checkNotNull(configuration);
    this.configuration = configuration;
  }

  public SQLConnection getConnection() {
    return connection;
  }

  public void setup() throws SQLException {
    Preconditions.checkState(connection == null);
    String url = configuration.jdbcString;

    Connection jdbcConnection =
        DriverManager.getConnection(url, configuration.user, configuration.pass);
    jdbcConnection.setAutoCommit(true);
    connection = new SQLConnection(jdbcConnection);
    dialect = connection.getBackendDialect();
    for (Function function : getDialect().getFunctionImplementations().keySet())
      functionTestRequirements.addRequirements(function);
  }

  private void drop(SQLTable table, boolean view, boolean ifExists) throws SQLException {
    connection.drop(table, view, ifExists, true);
  }

  public void clean() throws SQLException {
    if (KEEP_TEST_TABLES_AFTER_TEST)
      return;
    for (Entry<SQLTable,Boolean> entry : ImmutableMap.copyOf(createdTables).entrySet()) {
      // Drop views conditionally, since they might have been dropped by the cascade.
      drop(entry.getKey(), entry.getValue(), entry.getValue());
      createdTables.remove(entry.getKey());
    }
  }

  public BackendDialect getDialect() {
    Preconditions.checkArgument(dialect != null, "setup() not called yet");
    return dialect;
  }

  /** @param data column names, followed by column types as Class instances, followed by data */
  public void createTable(String name, Object data[]) throws SQLException {
    int i = 0;
    List<String> columnNames = Lists.newArrayList();
    List<Class<?>> columnTypes = Lists.newArrayList();
    List<Object[]> rows = Lists.newArrayList();
    for (; i < data.length && data[i] instanceof String; i++)
      columnNames.add((String) data[i]);
    for (; i < data.length && data[i] instanceof Class<?>; i++)
      columnTypes.add((Class<?>) data[i]);
    Preconditions.checkArgument(columnNames.size() == columnTypes.size());
    final int columnCount = columnNames.size();
    while (i < data.length) {
      Object row[] = new Object[columnCount];
      for (int j = 0; j < columnCount; i++, j++)
        row[j] = data[i];
      rows.add(row);
    }
    createTable(name,
        columnNames.toArray(new String[columnNames.size()]),
        columnTypes.toArray(new Class<?>[columnTypes.size()]),
        rows.toArray(new Object[rows.size()][]));
  }

  public SQLTable getTableHandle(String tableName) {
    return SQLTable.create(configuration.database, tableName);
  }

  public void createPassthroughView(String name, String targetTableName) throws SQLException {
    SQLTable viewTable = getTableHandle(name);
    if (KEEP_TEST_TABLES_AFTER_TEST && createdTables.containsKey(viewTable)) {
      throw new UnsupportedOperationException("Can't create multiple tables with the same name in a"
          + " test run for which KEEP_TEST_TABLES_AFTER_TEST is enabled (createPassthroughView)");
    }
    SQLTable targetTable = getTableHandle(targetTableName);
    final Set<String> columnNames =
        Sets.newLinkedHashSet(connection.getTableInfo(targetTable).getColumnNames());
    String pseudoColumn = getDialect().getRowIdentifierPseudoColumnName();
    if (!pseudoColumn.isEmpty())
      columnNames.add(pseudoColumn);
    List<SQLFragment> columnFragments = Lists.newArrayList();
    for (String columnName : columnNames) {
      if (configuration.includePseudocolumnInPassthroughView ||
          !getDialect().isRowIdentifierPseudoColumn(columnName))
      {
        columnFragments.add(getDialect().quoteIdentifier(columnName, true));
      }
    }
    if (columnFragments.isEmpty())
      columnFragments.add(SQLFragment.create("0"));
    SQLFragment createView = SQLFragment.builder()
        .appendJDBCString(getDialect().getCreateTableStatement(
            configuration.temporaryTables && !KEEP_TEST_TABLES_AFTER_TEST, true))
        .appendJDBCString(" ")
        .appendSQLFragment(getDialect().getTableIdentifier(viewTable))
        .appendJDBCString(" AS ")
        .appendJDBCString("SELECT ")
        .appendSQLFragment(SQLFragment.join(", ", columnFragments))
        .appendJDBCString(" FROM ")
        .appendSQLFragment(getDialect().getTableIdentifier(targetTable))
        .build();
    drop(viewTable, false, true);
    drop(viewTable, true, true);
    connection.executeUpdate(createView);
    createdTables.put(viewTable, true);
  }

  /**
   * @param columnNames may include the prefix "pk" to indicate a primary key, and postfix "_n" to
   *        indicate nullable
   * @param data may include null values
   */
  public void createTable(
      String name, String columnNames[], Class<?> columnTypes[], Object[][] data)
      throws SQLException
  {
    final SQLTable table = getTableHandle(name);
    if (KEEP_TEST_TABLES_AFTER_TEST && createdTables.containsKey(table)) {
      throw new UnsupportedOperationException("Can't create multiple tables with the same name in a"
          + " test run for which KEEP_TEST_TABLES_AFTER_TEST is enabled");
    }
    /* Postgres supports tables with no columns, but there doesn't seem to be a syntactically legal
    way to insert an empty tuple into it. */
    Preconditions.checkArgument(columnNames.length > 0 || data.length == 0);
    Preconditions.checkArgument(columnNames.length == columnTypes.length);
    final Map<String,SQLType> columns = Maps.newLinkedHashMap();
    final Set<String> pkColumnNames = Sets.newLinkedHashSet();
    for (int i = 0; i < columnNames.length; i++) {
      final String columnName = columnNames[i];
      final SQLType columnType = MapUtil.getChecked(TestConfiguration.SQL_TYPES, columnTypes[i]);
      final boolean nullable = columnName.endsWith("_n");
      if (columnName.startsWith("pk"))
        pkColumnNames.add(columnName);
      MapUtil.putChecked(columns, columnName, new SQLType(columnType.getTypeCode(), "",
          columnType.getScale(), columnType.getPrecision(), columnType.getLength(),
          nullable));
    }
    /* Avoid shadowing non-temporary tables, which is permitted on Postgres. Not doing this once
    caused a problem where QueryUtil.configurePrimaryKeys would use the non-temporary table's
    metadata instead of that of the temporary one. */
    drop(table, false, true);
    drop(table, true, true);
    connection.createTable(table, columns, pkColumnNames,
        configuration.temporaryTables && !KEEP_TEST_TABLES_AFTER_TEST);
    createdTables.put(table, false);
    insertData(name, columnNames, data);
    uniqueTestTables.put(new TestTableContents(columnNames, columnTypes, data), name);
  }

  private void insertData(String tableName, String columnNames[], Object[][] data)
      throws SQLException
  {
    // TODO[te]: Fix the error that occurs in testCreateEmptyTable if this test is removed.
    if (data.length == 0)
      return;
    try (SQLInserter inserter =
        new SQLInserter(connection, getTableHandle(tableName), Arrays.asList(columnNames), 100))
    {
      for (Object[] row : data)
        inserter.insert(row);
    }
  }

  public void fulfillFunctionTestRequirement(Expression expression) {
    functionTestRequirements.fulfillRequirements(expression);
  }

  public void checkFunctionTestsCovered() {
    Set<String> remainingRequirements = functionTestRequirements.getRemainingRequirements();
    if (!remainingRequirements.isEmpty()) {
      logWriter.println("Known uncovered function tests:");
      for (String req : remainingRequirements)
        logWriter.println("  " + req);
      logWriter.flush();
      if (false)
        throw new RuntimeException("There were uncovered function tests");
    }
  }

  @Override
  public void close() throws SQLException {
    if (connection == null)
      return;
    clean();
    connection.close();
    connection = null;
    logWriter.println("Performed " + evaluationCount + " evaluations");
    logWriter.println("Unique test tables: " + uniqueTestTables.size());
    logWriter.flush();
  }

  /**
   * Create one copy of each unique test table that was encountered during tests.
   */
  private void createAllUniqueTestTables() throws SQLException {
    Set<String> takenNames = Sets.newLinkedHashSet();
    for (Entry<TestTableContents,String> entry : uniqueTestTables.entrySet()) {
      TestTableContents ttc = entry.getKey();
      String name = entry.getValue();
      // Skip the tables from testWeirdlyNamedTables.
      if (name.startsWith("tab_OD"))
        continue;
      String useName;
      for (int i = 1; ; i++) {
        useName = name + (i == 1 ? "" : i);
        if (takenNames.add(useName))
          break;
      }
      createTable(useName, ttc.getColumnNames(), ttc.getColumnTypes(), ttc.getData());
    }
  }

  public Data evaluateView(
      View view, EvaluationSettings settings, Multimap<Path,Warning> toWarnings) throws SQLException
  {
    Preconditions.checkNotNull(view);
    Preconditions.checkNotNull(toWarnings);
    MutableView useView = Views.newMutable(view);
    QueryUtil.configurePrimaryKeys(connection, useView);
    System.out.println(settings);
    System.out.println(EvaluatorTestUtil.generateQueryString(useView));
    evaluationCount++;
    // TODO[cc]: Use QueryEvaluator.create instead.
    try (QueryEvaluator evaluator = new QueryEvaluatorImpl(
        useView, connection, logWriter, new CancelHandler(), settings))
    {
      return evaluator.evaluate(Integer.MAX_VALUE, toWarnings);
    } finally {
      logWriter.flush();
    }
  }

  public Set<String> getColumnNames(String tableName, boolean primaryKeyOnly)
      throws SQLException
  {
    SQLTableInfo tableInfo = connection.getTableInfo(getTableHandle(tableName));
    Set<String> ret = Sets.newLinkedHashSet();
    for (String columnName : tableInfo.getColumnNames()) {
      if (!primaryKeyOnly || tableInfo.isPrimaryKey(columnName))
        ret.add(columnName);
    }
    return ret;
  }
}
