// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import com.google.common.collect.Lists;
import com.sieuferd.database.dialect.JDBCObjectWrapper;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.Data;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PrimitiveValue;
import com.sieuferd.datamodel.Tuple;
import com.sieuferd.datamodel.TupleInterval;
import com.sieuferd.datamodel.View;
import com.sieuferd.datamodel.Views;
import com.sieuferd.query.Filter;
import com.sieuferd.query.QProp;
import com.sieuferd.datamodel.expressions.Expression.PathExpression;
import com.sieuferd.datamodel.Strepco;
import com.sieuferd.datamodel.expressions.Expression.ConstantExpression;
import com.sieuferd.datamodel.expressions.Expressions;
import com.sieuferd.util.StringUtil;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.List;

public final class EvaluatorTestUtil {
  public static final String PSEUDOKEY_MARKER = "~";
  public static final String NON_PROJECTED_MARKER = ".";
  public static final String NULL_MARKER = "_";
  private EvaluatorTestUtil() { }

  private static String generateDataString(
      View view, Data data, Path path, Tuple parentTuple, boolean sortTuples)
  {
    if (!data.isProjected(path))
      return NON_PROJECTED_MARKER;
    if (path.isRelation()) {
      TupleInterval childTuples = data.getRelationTuples(path.getRelation(), parentTuple);
      if (sortTuples && !childTuples.isComplete()) {
        throw new IllegalArgumentException(
            "Can't sort tuples in incomplete relation " + path.toString(view));
      }
      List<String> tupleStrings = Lists.newArrayList();
      for (Tuple childTuple : childTuples) {
        StringBuilder tupleString = new StringBuilder();
        tupleString.append("(");
        boolean first = true;
        for (Path childPath : view.getChildPaths(path.getRelation())) {
          if (!first) {
            tupleString.append(",");
          } else {
            first = false;
          }
          tupleString.append(generateDataString(view, data, childPath, childTuple, sortTuples));
        }
        tupleString.append(")");
        tupleStrings.add(tupleString.toString());
      }
      if (sortTuples)
        Collections.sort(tupleStrings);
      StringBuilder ret = new StringBuilder();
      ret.append("[");
      ret.append(StringUtil.join(",", tupleStrings));
      if (!childTuples.isComplete())
        ret.append("...");
      ret.append("]");
      return ret.toString();
    } else {
      return getValueString(data.getTupleField(parentTuple, path.getPrimitive()));
    }
  }

  private static String getValueString(PrimitiveValue value) {
    final Object content = value.getContent();
    if (content == null) {
      return NULL_MARKER;
    } else if (content instanceof Number) {
      final DecimalFormat decimalFormat = new DecimalFormat("#.######");
      return decimalFormat.format(content);
    } else if (content instanceof JDBCObjectWrapper) {
      // TODO[pd]: Find a more reliable way to detect these than testing for JDBCObjectWrapper.
      return PSEUDOKEY_MARKER;
    } else {
      /* Use the default toString for all other types, including temporal types. None of the
      supported temporal types' string representations depend on the system time zone (LocalTime and
      LocalDate have only one interpretation, and Instant's toString always uses UTC). */
      final String escapedContentString;
      String contentString = content.toString();
      if (contentString.equals(NON_PROJECTED_MARKER) || contentString.equals(NULL_MARKER) ||
          contentString.equals(PSEUDOKEY_MARKER) || contentString.contains("'"))
      {
        escapedContentString = "'" + contentString.replace("'", "''") + "'";
      } else {
        escapedContentString = contentString;
      }
      // Add "???" to string representations of any unexpected data types.
      return ConstantExpression.fromObject(content) == null
          ? escapedContentString + "???"
          : escapedContentString;
    }
  }

  public static String getDataString(View view, Data data, boolean sortTuples) {
    return generateDataString(view, data, Path.ROOT, Tuple.createRootParent(data), sortTuples);
  }

  public static String generateQueryString(View view) {
    MutableView useView = Views.newMutable(view);
    DataProperties.configureDataProperties(useView);
    return generateQueryString(useView, Path.ROOT);
  }

  private static String generateQueryString(View view, RelationPath path) {
    List<String> fieldStrings = Lists.newArrayList();
    Strepco strepco = Strepco.createNormalized(view, path);
    for (Path childPath : view.getChildPaths(path)) {
      StringBuilder fs = new StringBuilder();
      if (!view.get(CProp.PROJECTED, childPath))
        fs.append(".");
      if (!childPath.isRelation() && view.get(CProp.IN_TUPLE_KEY, childPath))
        fs.append("*");
      fs.append(view.get(CProp.LABEL, childPath));
      if (childPath.isRelation()) {
        fs.append(generateQueryString(view, childPath.getRelation()));
      } else {
        final String columnName = view.get(QProp.SQL_COLUMN_NAME, childPath);
        if (columnName.isEmpty()) {
          fs.append(":");
          fs.append(Expressions.generateString(
              strepco, view.get(QProp.FORMULA_EXPRESSION, childPath)));
        } else {
          if (!columnName.equals(view.get(CProp.LABEL, childPath))) {
            fs.append(":");
            fs.append(columnName);
          }
        }
      }
      PrimitivePath joinedOn = childPath.isRelation() ? null :
          view.get(QProp.JOINED_ON, childPath).getNullable();
      if (joinedOn != null) {
        fs.append("=");
        fs.append(Expressions.generateString(strepco, new PathExpression(joinedOn)));
      }
      Filter filter = view.get(QProp.FILTER, childPath);
      if (filter.isConstrained())
        fs.append(" FILTER");
      fieldStrings.add(fs.toString());
    }
    return
        (view.get(QProp.EXISTENTIAL_FILTER, path) ? " E" : "") +
        "(" + StringUtil.join(", ", fieldStrings) + ")";
  }
}
