// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import java.sql.SQLException;
import org.junit.AfterClass;
import org.junit.Test;

/**
 * Tests of the behavior of individual frontend functions. These tests are intended primarily to
 * ensure that backend dialects correctly map frontend functions to appropriate backend renderings,
 * not to detect bugs in the external backend implementations.
 */
public class FunctionsTest extends EvaluatorTestBase {
  @AfterClass
  public static void afterClass() throws SQLException {
    EvaluatorTestBase.afterClass();
    TEST_STATE.checkFunctionTestsCovered();
  }

  @Test
  public void testBasic() {
    expectResult("2 + 3", "5");
    expectResult("2 + null", "_");
    expectResult("null", "_");
    expectResult("(null)", "_");
  }

  @Test
  public void testConditional() {
    expectResult("if(true, 1, 2)", "1");
    expectResult("if(false, 1, 2)", "2");
    expectResult("if(true, 'ja', 'nei')", "ja");
    expectResult("if(false, 'ja', 'nei')", "nei");
    expectResult("if(null, 1, 2)", "_");
    expectResult("if(true, null, 2)", "_");
    expectResult("if(false, null, 2)", "2");
    expectResult("if(true, 1, null)", "1");
    expectResult("if(false, 1, null)", "_");
    expectResult("if(false, null, null)", "_");
    expectResult("if(true, null, null)", "_");
    expectResult("if(null, null, null)", "_");
    expectResult("if(null, true, false)", "_");
    expectResult("2 + if(true, null, null)", "_");
    expectResult("len(if(true, null, null))", "_");
    // Expect null here, not 5.
    expectResult("if(null, 5, 5)", "_");

    expectResult("if(true, 1 = 0, if(true, true, true))", "false");
    expectResult("if(false, 1 = 0, if(true, true, true))", "true");
  }

  @Test
  public void testConditionalWithBooleanVariations() {
    expectResult("if(1 = 1, 1, 2)", "1");
    expectResult("if(1 = 0, 1, 2)", "2");
    expectResult("if(1 = 1, 1, 2)", "1");
    expectResult("if(1 = 0, 1, 2)", "2");
    expectResult("if(1 = 1, false, true)", "false");
    expectResult("if(1 = 0, false, true)", "true");
    expectResult("if(1 = 1, 1 = 0, 1 = 1)", "false");
    expectResult("if(true, 1 = 0, 1 = 1)", "false");
    expectResult("if(false, 1 = 0, 1 = 1)", "true");
    expectResult("if(1 = 0, 1 = 0, 1 = 1)", "true");
    expectResult("if(true, false, true)", "false");
    expectResult("if(false, false, true)", "true");
    expectResult("if(true, null, true)", "_");
    expectResult("if(false, null, true)", "true");
    expectResult("if(true, false, null)", "false");
    expectResult("if(false, false, null)", "_");
  }

  @Test
  public void testConstantsAndParentheses() {
    expectResult("5", "5");
    expectResult("null", "_");
    expectResult("true", "true");
    expectResult("false", "false");
    expectResult("'asdf'", "asdf");

    expectResult("(5)", "5");
    expectResult("(null)", "_");
    expectResult("(true)", "true");
    expectResult("(false)", "false");
    expectResult("('asdf')", "asdf");
  }

  @Test
  public void testArithmetic() {
    // Unary plus
    expectResult("+0", "0");
    expectResult("+73", "73");
    expectResult("+73.4", "73.4");
    expectResult("0 + (+73)", "73");
    // Unary minus
    expectResult("-0", "0");
    expectResult("-73", "-73");
    expectResult("-73.4", "-73.4");
    expectResult("0 + (-73)", "-73");
    // Binary plus
    expectResult("12 + 73", "85");
    expectResult("12.3 + 73.3", "85.6");

    expectResult("12 + null", "_");
    expectResult("null + 73", "_");
    // Binary minus
    expectResult("12 - 73", "-61");
    expectResult("12 - 73.3", "-61.3");
    // Multiply
    expectResult("2 * 9", "18");
    expectResult("2 * -9", "-18");
    expectResult("2.5 * 9", "22.5");
    expectResult("2.5 * 0.0000", "0");
    // Divide
    expectResult("2 / 2", "1");
    expectResult("1 / 1", "1");
    expectResult("(-1) / (-1)", "1");
    expectResult("(-1) / 1", "-1");
    expectResult("1 / (-1)", "-1");
    expectResult("0 / (-1)", "0");
    expectResult("0 / 1", "0");
    expectResult("(-0) / 1", "0");
    expectResult("1 / 0", "_");
    expectResult("1 / (-0)", "_");
    expectResult("0 / 0", "_");
    expectResult("(-0) / 0", "_");
    expectResult("0 / (-0)", "_");
    expectResult("0 / (-0)", "_");
    expectResult("(-1) / 0", "_");
    expectResult("(-1) / (-0)", "_");
    expectResult("5 / 2", "2.5");
    expectResult("5.000 / 2.000", "2.5");
  }

  @Test
  public void testPower() {
    // There should be no signed zeroes.
    expectResult("-0", "0");
    expectResult("-0 ** 1", "0");
    expectResult("2 ** 4", "16");
    expectResult("2 ** 0", "1");
    expectResult("(-2) ** 4", "16");
    expectResult("(-2) ** 5", "-32");
    expectResult("1024 ** (0.5)", "32");
    expectResult("(-2.7) ** 4", "53.1441");
    expectResult("0 ** 0", "1");
    expectResult("0 ** 0.1", "0");
    expectResult("0 ** 1", "0");
    expectResult("0 ** 2", "0");
    expectResult("0 ** 2.5", "0");

    // First argument negative, second argument not an integer.
    expectResult("(-2) ** 4.5", "_");
    expectResult("(-2.7) ** 4.5", "_");
    // First argument zero, second argument less than zero.
    expectResult("0 ** (-0.2)", "_");
    expectResult("0 ** (-1)", "_");
    expectResult("0 ** (-2)", "_");
    expectResult("0 ** (-2.3)", "_");
  }

  @Test
  public void testLogarithm() {
    expectResult("ln(5.2)", "1.648659");
    expectResult("ln(2.718281828459)", "1");
    expectResult("ln(0.5)", "-0.693147");
    expectResult("ln(1.0)", "0");
    expectResult("ln(1)", "0");
    expectResult("ln(0)", "_");
    expectResult("ln(0.000000143)", "-15.760421");
    expectResult("ln(0.0)", "_");
    expectResult("ln(-0.5)", "_");
    expectResult("ln(-1.0)", "_");
    expectResult("ln(-1.5)", "_");
    expectResult("ln(2)", "0.693147");
    expectResult("ln(2.0)", "0.693147");
    expectResult("ln(len('ab'))", "0.693147");
  }

  @Test
  public void testNumericOnlyRelationals() {
    expectResult("5 > 3", "true");
    expectResult("5 > 5", "false");
    expectResult("5 >= 3", "true");
    expectResult("5 >= 5", "true");
    expectResult("5 <= 3", "false");
    expectResult("5 <= 5", "true");
    expectResult("5 < 3", "false");
    expectResult("5 < 5", "false");
  }

  @Test
  public void testEqualsNotEquals() {
    expectResult("5 = 5", "true");
    expectResult("5 = 6", "false");
    expectResult("5 = null", "_");
    expectResult("'a' = 'a'", "true");
    expectResult("'a' = 'b'", "false");
    expectResult("'a' = 'bb'", "false");
    expectResult("'a' = null", "_");
    expectResult("true = true", "true");
    expectResult("false = false", "true");
    expectResult("true = null", "_");
    expectResult("false = null", "_");
    expectResult("true = false", "false");
    expectResult("false = true", "false");
    expectResult("null = null", "_");

    expectResult("'a' = 'A'", "false");
    expectResult("'a' != 'A'", "true");

    expectResult("5 != 5", "false");
    expectResult("5 != 6", "true");
    expectResult("5 != null", "_");
    expectResult("'a' != 'a'", "false");
    expectResult("'a' != 'b'", "true");
    expectResult("'a' != 'bb'", "true");
    expectResult("'a' != null", "_");
    expectResult("true != true", "false");
    expectResult("false != false", "false");
    expectResult("true != null", "_");
    expectResult("false != null", "_");
    expectResult("true != false", "true");
    expectResult("false != true", "true");
    expectResult("null != null", "_");
  }

  @Test
  public void testIsIsNot() {
    expectResult("5 is 5", "true");
    expectResult("5 is 6", "false");
    expectResult("5 is null", "false");
    expectResult("'a' is 'a'", "true");
    expectResult("'a' is 'b'", "false");
    expectResult("'a' is 'bb'", "false");
    expectResult("'a' is null", "false");
    expectResult("true is true", "true");
    expectResult("false is false", "true");
    expectResult("true is null", "false");
    expectResult("false is null", "false");
    expectResult("true is false", "false");
    expectResult("false is true", "false");
    expectResult("null is null", "true");

    expectResult("'a' is 'A'", "false");
    expectResult("'a' is not 'A'", "true");

    expectResult("5 is not 5", "false");
    expectResult("5 is not 6", "true");
    expectResult("5 is not null", "true");
    expectResult("'a' is not 'a'", "false");
    expectResult("'a' is not 'b'", "true");
    expectResult("'a' is not 'bb'", "true");
    expectResult("'a' is not null", "true");
    expectResult("true is not true", "false");
    expectResult("false is not false", "false");
    expectResult("true is not null", "true");
    expectResult("false is not null", "true");
    expectResult("true is not false", "true");
    expectResult("false is not true", "true");
    expectResult("null is not null", "false");

    /* The expressions "is not x" and "is (not x)" are semantically different, though the syntax may
    be a bit deceiving (see commentary in ANTLR_Expression.g4). */
    expectResult("true is (not null)", "false");
    expectResult("false is (not null)", "false");
  }

  @Test
  public void testLogical() {
    expectResult("not true", "false");
    expectResult("not false", "true");
    expectResult("not null", "_");
    expectResult("false and false", "false");
    expectResult("false and true", "false");
    expectResult("true and false", "false");
    expectResult("true and true", "true");
    expectResult("true and null", "_");
    expectResult("null and true", "_");
    expectResult("false and null", "false");
    expectResult("null and false", "false");
    expectResult("null and null", "_");
    expectResult("false or false", "false");
    expectResult("false or true", "true");
    expectResult("true or false", "true");
    expectResult("true or null", "true");
    expectResult("null or true", "true");
    expectResult("false or null", "_");
    expectResult("null or false", "_");
    expectResult("null or null", "_");
  }

  @Test
  public void testEmptyString() {
    expectResult("''", "");
    expectResult("concat()", "");
    expectResult("concat(null)", "");
    expectResult("concat(null, null)", "");
    expectResult("concat(null, '')", "");
    expectResult("concat(null, null, null)", "");
    expectResult("concat(null, null, '')", "");

    expectResult("trim('')", "");
    expectResult("trim(' ')", "");
    expectResult("trim('    ')", "");

    expectResult("pos('foobar', '')", "0");
    expectResult("pos('', '')", "0");
    expectResult("contains('foobar', '')", "true");
    expectResult("contains('', '')", "true");
    expectResult("startswith('foobar', '')", "true");
    expectResult("startswith('', '')", "true");

    expectResult("len('')", "0");

    expectResult("left('foobar', 0)", "");
    expectResult("left('f', 0)", "");
    expectResult("left('', 0)", "");
    expectResult("left('', 1)", "");
    expectResult("left('', 2)", "");
  }

  @Test
  public void testFloatingPointErrors() {
    /* TODO[te]: Add more of these. See for instance
                 http://en.wikipedia.org/wiki/Numeric_precision_in_Microsoft_Excel . */
    // See http://www.digi.no/931756/pussige-regnefeil-i-android-5-0
    expectResult("3.07 - 3.06", "0.01");
  }

  @Test
  public void testString() {
    // Concatentation
    expectResult("concat('abc')", "abc");
    expectResult("concat('abc', 'def')", "abcdef");
    expectResult("concat('abc', 'def', 'ghi')", "abcdefghi");
    expectResult("concat('abc', null, 'ghi')", "abcghi");
    expectResult("concat('abc', null, null)", "abc");
    expectResult("concat('abc', null, 5)", "abc5");
    /* Concatenation of non-string types only. See the part about "as at least one input is of a
    string type" at http://www.postgresql.org/docs/9.1/static/functions-string.html . */
    expectResult("concat(5)", "5");
    expectResult("concat(5, 6)", "56");

    // Trim
    expectResult("trim('  \t ')", "\t");
    expectResult("trim('\t')", "\t");
    expectResult("trim('  \\n ')", "\n");
    expectResult("trim('\\n')", "\n");
    expectResult("trim('asdf')", "asdf");
    expectResult("trim(' asdf')", "asdf");
    expectResult("trim('  asdf')", "asdf");
    expectResult("trim('  asdf ')", "asdf");
    expectResult("trim('  as df ')", "as df");
    expectResult("trim('  as \\ndf ')", "as \ndf");
    expectResult("trim('  as \\ndf\\n ')", "as \ndf\n");

    // Upper/lower
    expectResult("lower(' foobar')", " foobar");
    expectResult("lower(' fooBar')", " foobar");
    expectResult("lower(' FOOBAR')", " foobar");
    expectResult("lower('FooBar')", "foobar");
    expectResult("upper(' foobar')", " FOOBAR");
    expectResult("upper(' fooBar')", " FOOBAR");
    expectResult("upper(' FOOBAR')", " FOOBAR");
    expectResult("upper('FooBar')", "FOOBAR");

    // Position
    expectResult("pos('foobar', 'asdf')", "_");
    expectResult("pos('foobar', 'foobara')", "_");
    expectResult("pos('foobar', 'foo')", "0");
    expectResult("pos('foobar', 'oo')", "1");
    expectResult("pos('foooobar', 'oo')", "1");
    expectResult("pos('foooobar', 'ooo')", "1");
    expectResult("pos('foooobar', 'oooo')", "1");
    expectResult("pos('foooobar', 'oob')", "3");
    expectResult("pos('foooobar', 'bar')", "5");
    expectResult("pos('fooooBAR', 'BAR')", "5");
    expectResult("pos('foooobar', 'BAR')", "_");
    expectResult("pos('fooOObar', 'OO')", "3");
    expectResult("pos('fooOObar', 'oO')", "2");

    // Contains
    expectResult("contains('foobar', 'asdf')", "false");
    expectResult("contains('foobar', 'foobara')", "false");
    expectResult("contains('foobar', 'foo')", "true");
    expectResult("contains('foobar', 'oo')", "true");
    expectResult("contains('foooobar', 'oo')", "true");
    expectResult("contains('foooobar', 'ooo')", "true");
    expectResult("contains('foooobar', 'oooo')", "true");
    expectResult("contains('foooobar', 'oob')", "true");
    expectResult("contains('foooobar', 'bar')", "true");
    expectResult("contains('fooooBAR', 'BAR')", "true");
    expectResult("contains('foooobar', 'BAR')", "false");
    expectResult("contains('fooOObar', 'OO')", "true");
    expectResult("contains('fooOObar', 'oO')", "true");

    // Startswith
    expectResult("startswith('foobar', 'asdf')", "false");
    expectResult("startswith('foobar', 'foobara')", "false");
    expectResult("startswith('foobar', 'foo')", "true");
    expectResult("startswith('FOObar', 'foo')", "false");
    expectResult("startswith('foobar', 'oo')", "false");
    expectResult("startswith('foooobar', 'oo')", "false");
    expectResult("startswith('foooobar', 'ooo')", "false");
    expectResult("startswith('foooobar', 'oooo')", "false");
    expectResult("startswith('foooobar', 'oob')", "false");
    expectResult("startswith('foooobar', 'bar')", "false");
    expectResult("startswith('fooooBAR', 'BAR')", "false");
    expectResult("startswith('foooobar', 'BAR')", "false");
    expectResult("startswith('fooOObar', 'OO')", "false");
    expectResult("startswith('fooOObar', 'oO')", "false");

    // Length
    expectResult("len(' ')", "1");
    expectResult("len('a')", "1");
    expectResult("len('ab')", "2");
    expectResult("len('ab ')", "3");
    expectResult("len('ab c')", "4");
    expectResult("len('ab c\\n')", "5");
    expectResult("len(' ab c\\n')", "6");

    // Normalize whitespace.
    expectResult("normalize_whitespace('this is a test')", "this is a test");
    expectResult("normalize_whitespace('  this is a test  ')", "this is a test");
    expectResult("normalize_whitespace('  this is   a test  ')", "this is a test");
    expectResult("normalize_whitespace('  this is   a\\ntest  ')", "this is a test");
    expectResult("normalize_whitespace('  this IS   a\\ntest  ')", "this IS a test");
    expectResult("normalize_whitespace('  this\tIS   a\\ntest  ')", "this IS a test");
    expectResult("normalize_whitespace('  this\tIS   a\\nt   est  ')", "this IS a t est");
    // Detect a previous bug where regex character classes were not specified correctly.
    expectResult(
        "normalize_whitespace('  this\tIS   a\\ntestgraphblank  ')", "this IS a testgraphblank");

    // Left
    expectResult("left('foobar', 1)", "f");
    expectResult("left('foobar', 2)", "fo");
    expectResult("left('foobar', 3)", "foo");
    expectResult("left('foobar', 6)", "foobar");
    expectResult("left('foobar', 7)", "foobar");
    expectResult("left('foobar', 8)", "foobar");
    expectResult("left('foobar ', 6)", "foobar");
    expectResult("left('foobar ', 7)", "foobar ");
    expectResult("left('foobar k', 8)", "foobar k");
    expectResult("left('foobar k', 7)", "foobar ");
    expectResult("left('f', 1)", "f");
    expectResult("left('f', 2)", "f");
    expectResult("left('foobar', null)", "_");
    expectResult("left(null, null)", "_");
    expectResult("left(null, 0)", "_");
    expectResult("left(null, 1)", "_");
    expectResult("left(null, 2)", "_");

    expectResult("left('foobar', 2.00)", "fo");
    expectResult("left('foobar', 2.10)", "fo");
    expectResult("left('foobar', 2.90)", "fo");
    expectResult("left('foobar', 3.00)", "foo");
  }

  @Test
  public void testTemporal() {
    final boolean nanos = TEST_STATE.getDialect().supportsNanosecondTemporals();
    // Date literals.
    expectResult("{2015-07-28}", "2015-07-28");
    // Time literals of various precisions.
    expectResult("{15:34:10}", "15:34:10");
    expectResult("{15:34:10.0}", "15:34:10");
    expectResult("{15:34:10.123}", "15:34:10.123");
    expectResult("{15:34:10.123456}", "15:34:10.123456");
    if (nanos)
      expectResult("{15:34:10.123456789}", "15:34:10.123456789");
    // Timestamp literals.
    expectResult("{2015-07-28 15:34:10}", "2015-07-28T15:34:10Z");
    expectResult("{2015-07-28 15:34:10.123}", "2015-07-28T15:34:10.123Z");
    expectResult("{2015-07-28 15:34:10.123456}", "2015-07-28T15:34:10.123456Z");
    if (nanos)
      expectResult("{2015-07-28 15:34:10.123456789}", "2015-07-28T15:34:10.123456789Z");

    expectResult("concat({15:34:10})", "15:34:10");
    expectResult("concat({15:34:10.123})", "15:34:10.123");
    expectResult("concat({15:34:10.123456})", "15:34:10.123456");
    if (nanos)
      expectResult("concat({15:34:10.123456789})", "15:34:10.123456789");
    expectResult("concat({2015-07-28})", "2015-07-28");
    expectResult("timestamp_str({2015-07-28 15:34:10Z})", "2015-07-28 15:34:10.000000000Z");
    expectResult("timestamp_str({2015-07-28 15:34:10.123456})", "2015-07-28 15:34:10.123456000Z");

    // Now test the server-side interpretation:
    expectResult("concat({2015-07-28})", "2015-07-28");
    expectResult("concat({15:34:10})", "15:34:10");
    expectResult("concat({15:34:10.0})", "15:34:10");
    expectResult("concat({15:34:10.123})", "15:34:10.123");
    expectResult("concat({15:34:10.123456})", "15:34:10.123456");
    if (nanos)
      expectResult("concat({15:34:10.123456789})", "15:34:10.123456789");
    expectResult("timestamp_str({2015-07-28 15:34:10})", "2015-07-28 15:34:10.000000000Z");
    expectResult("timestamp_str({2015-07-28 15:34:10.123})", "2015-07-28 15:34:10.123000000Z");
    expectResult("timestamp_str({2015-07-28 15:34:10.123456})", "2015-07-28 15:34:10.123456000Z");
    if (nanos)
      expectResult("timestamp_str({2015-07-28 15:34:10.123456789})", "2015-07-28T15:34:10.123456789Z");
  }

  @Test
  public void testTemporalsInFunctions() {
    expectResult("if(true, {2015-07-28}, {2001-02-13})", "2015-07-28");
    expectResult("if(false, {2015-07-28}, {2001-02-13})", "2001-02-13");
    expectResult("if(true, {15:34:10.123}, {02:09:12.629})", "15:34:10.123");
    expectResult("if(false, {15:34:10.123}, {02:09:12.629})", "02:09:12.629");
    expectResult("if(true, {2015-07-28 15:34:10.123}, {1988-06-22 12:42:48.512532})", "2015-07-28T15:34:10.123Z");
    expectResult("if(false, {2015-07-28 15:34:10.123}, {1988-06-22 12:42:48.512532})", "1988-06-22T12:42:48.512532Z");
  }

  @Test
  public void testNullAggregates() {
    if (!SKIP_KNOWN_FAILING_TESTS) {
      expectResult("count(null)", "0");
      expectResult("sum(null)", "0");
      expectResult("avg(null)", "_");
      expectResult("min(null)", "_");
      expectResult("max(null)", "_");
      expectResult("stdev(null)", "_");
    }
  }
}
