// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer;

import com.google.common.base.Preconditions;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.healthmarketscience.jackcess.Column;
import com.healthmarketscience.jackcess.DataType;
import com.healthmarketscience.jackcess.Database;
import com.healthmarketscience.jackcess.Database.FileFormat;
import com.healthmarketscience.jackcess.DatabaseBuilder;
import com.healthmarketscience.jackcess.Index;
import com.healthmarketscience.jackcess.Row;
import com.healthmarketscience.jackcess.Table;
import com.healthmarketscience.jackcess.Table.ColumnOrder;
import com.sieuferd.database.SQLConnection;
import com.sieuferd.database.SQLInserter;
import com.sieuferd.database.SQLType;
import com.sieuferd.database.SQLTypeCode;
import com.sieuferd.query.SQLDatabase;
import com.sieuferd.query.SQLTable;
import com.sieuferd.util.MapUtil;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.sql.Types;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.ZoneOffset;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;
import javax.annotation.Nullable;

// TODO[te]: Make my Time Zone Test.accdb file into a unit test. Include a character encoding test.
public final class AccessImporter {
  /* In the Microsoft Access GUI, any date/time values with this date have their date portions
  hidden from the user; this date is used to encode all time values entered without dates. Thus we
  convert such dates to null values. Note that Access has no way to distinguish a missing date value
  from the actual date of 1899-12-30, even though earlier dates can be represented just fine. */
  private static final LocalDate NULL_DATE = LocalDate.parse("1899-12-30");
  private static final int OUTPUT_BATCH_SIZE = 1000;


  public static final void importFile(
      File file, SQLConnection output, SQLDatabase outputDatabase, PrintWriter pw)
      throws SQLException, IOException
  {
    DatabaseBuilder builder = new DatabaseBuilder(file)
        /* Access stores all datetimes in the local time zone, not in UTC. However, using UTC here
        allows us to retrieve raw, unadultered datetime values from Jackcess which can then be
        properly casted into a proper LocalDateTime object. */
        .setTimeZone(TimeZone.getTimeZone(ZoneOffset.UTC))
        .setReadOnly(true);
    try (Database input = builder.open()) {
      /* Access '97 (JET3) is the only Jackcess-supported format that does not use a standardized
      character set. For JET3, Jackcess will default to the platform character encoding, which is
      certain to be wrong on MacOS, and risky and non-portable on Windows. See
      com.healthmarketscience.jackcess.impl.JetFormat. */
      if (input.getFileFormat() == FileFormat.V1997)
        throw new IOException("Access '97 format not yet supported");
      pw.println("Got file format " + input.getFileFormat());
      input.setColumnOrder(ColumnOrder.DISPLAY);
      for (SQLTable outputTable : output.getTables(outputDatabase))
        output.drop(outputTable, false, true, true);
      importDatabase(input, output, outputDatabase, pw);
    }
  }

  private static void importDatabase(
      Database input, SQLConnection outputConn, SQLDatabase outputDatabase, PrintWriter pw)
      throws SQLException, IOException
  {
    for (Table table : input.newIterable().setIncludeLinkedTables(false)) {
      String tableName = ImportUtil.encodeHumanIdentifier(table.getName());
      pw.println("Found table " + table.getName());
      try {
        importTable(table, outputConn, SQLTable.create(outputDatabase, tableName), pw);
      } catch (IOException e) {
        e.printStackTrace(pw);
        // Continue.
      }
    }
  }

  private static class ValueExtractor {
    private final String inputColumnName;

    public ValueExtractor(String inputColumnName) {
      Preconditions.checkNotNull(inputColumnName);
      this.inputColumnName = inputColumnName;
    }

    public @Nullable Object extractValue(Row row) throws IOException {
      return row.get(inputColumnName);
    }
  }

  private static class TemporalValueExtractor extends ValueExtractor {
    private final boolean timePart;

    public TemporalValueExtractor(String inputColumnName, boolean timePart) {
      super(inputColumnName);
      this.timePart = timePart;
    }

    @Override
    public @Nullable Object extractValue(Row row) throws IOException {
      Object obj = super.extractValue(row);
      if (obj == null)
        return null;
      if (!(obj instanceof Date))
        throw new IOException("Expected Date object");

      /* By using UTC both here and and in DatabaseBuilder.setTimeZone, we get back a LocalDateTime
      reflecting the exact date that would be seen in the Microsoft Access GUI. */
      LocalDateTime ldt = LocalDateTime.ofInstant(((Date) obj).toInstant(), ZoneOffset.UTC);
      if (timePart) {
        return java.sql.Time.valueOf(LocalTime.from(ldt));
      } else {
        LocalDate ret = LocalDate.from(ldt);
        if (ret.equals(NULL_DATE))
          return null;
        return java.sql.Date.valueOf(ret);
      }
    }
  }

  private static final class CreatedColumns {
    // TODO[cc]: Get rid of all this redundancy.
    private final Set<String> primaryKeyColumnNames = Sets.newLinkedHashSet();
    private final List<String> outputColumnNames = Lists.newArrayList();
    private final Map<String,SQLType> namesAndTypes = Maps.newLinkedHashMap();
    private final List<ValueExtractor> extractors = Lists.newArrayList();

    public void add(String outputName, SQLType type, ValueExtractor extractor, boolean pk) {
      outputColumnNames.add(outputName);
      MapUtil.putChecked(namesAndTypes, outputName, type);
      if (pk)
        primaryKeyColumnNames.add(outputName);
      extractors.add(extractor);
    }

    public int size() {
      return outputColumnNames.size();
    }

    public List<ValueExtractor> getExtractors() {
      return Collections.unmodifiableList(extractors);
    }

    public Set<String> getPrimaryKeyColumnNames() {
      return Collections.unmodifiableSet(primaryKeyColumnNames);
    }

    public Map<String,SQLType> getNamesAndTypes() {
      return Collections.unmodifiableMap(namesAndTypes);
    }

    public List<String> getOutputColumnNames() {
      return Collections.unmodifiableList(outputColumnNames);
    }
  }

  private static void importTable(
      Table inputTable, SQLConnection outputConn, SQLTable outputTable, PrintWriter pw)
      throws SQLException, IOException
  {
    // Empty means no primary key.
    final Set<Integer> primaryKeyColumnIndices = Sets.newLinkedHashSet();
    for (Index index : inputTable.getIndexes()) {
      if (index.isPrimaryKey()) {
        for (Index.Column indexColumn : index.getColumns())
          primaryKeyColumnIndices.add(indexColumn.getColumn().getColumnIndex());
        break;
      }
    }
    final CreatedColumns createdColumns = new CreatedColumns();
    for (Column column : inputTable.getColumns()) {
      if (column.isCalculated()) {
        pw.println("  Skipping calculated field " + column.getName());
        continue;
      }
      final DataType dataType = column.getType();
      // TODO[pd]: Do I need to act on column.isVariableLength()?
      final boolean pk = primaryKeyColumnIndices.contains(column.getColumnIndex());
      // TODO[pd]: Does this hold?
      final boolean nullable = !pk;
      if (dataType == DataType.SHORT_DATE_TIME) {
        pw.println("  " + column.getName() + ": local timestamp");
        createdColumns.add(ImportUtil.encodeHumanIdentifier(column.getName() + "_date"),
            new SQLType(SQLTypeCode.create(Types.DATE), "", 0, 0, 0, true),
            new TemporalValueExtractor(column.getName(), false), pk);
        createdColumns.add(ImportUtil.encodeHumanIdentifier(column.getName() + "_time"),
            new SQLType(SQLTypeCode.create(Types.TIME), "", 0, 0, 0, true),
            new TemporalValueExtractor(column.getName(), true), pk);
      } else if (dataType == DataType.BINARY || dataType == DataType.OLE) {
        pw.println("  Skipping binary field '" + column.getName() + "'");
        continue;
      } else {
        String outputColumnName = ImportUtil.encodeHumanIdentifier(column.getName());
        final SQLTypeCode sqlTypeCode;
        try {
          sqlTypeCode = SQLTypeCode.create(dataType.getSQLType());
        } catch (SQLException e) {
          pw.println("  Skipping field '" + column.getName() + "' of unsupported type " + dataType);
          continue;
        }
        final SQLType sqlType = new SQLType(sqlTypeCode, "",
            column.getScale(), column.getPrecision(), column.getLengthInUnits(), nullable);
        createdColumns.add(outputColumnName, sqlType, new ValueExtractor(column.getName()), pk);
        pw.println("  " + column.getName() + ": type=" + sqlType +
            " out=" + outputConn.getBackendDialect().getTypeName(sqlType));
      }
    }
    pw.println("With PK " + createdColumns.getPrimaryKeyColumnNames());
    outputConn.createTable(outputTable, createdColumns.getNamesAndTypes(),
        createdColumns.getPrimaryKeyColumnNames(), false);

    final Object[] rowContents = new Object[createdColumns.size()];
    try (SQLInserter inserter = new SQLInserter(outputConn, outputTable,
        createdColumns.getOutputColumnNames(), OUTPUT_BATCH_SIZE))
    {
      for (Row row : inputTable) {
        int colIndex = 0;
        for (ValueExtractor extractor : createdColumns.getExtractors()) {
          Object valueObj = extractor.extractValue(row);
          rowContents[colIndex++] = valueObj;
        }
        //System.out.println(Arrays.asList(rowContents));
        inserter.insert(rowContents);
      }
    }
  }
}
