// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
/* An earlier version of this file was based in part
 *   org.apache.poi.ss.usermodel.DateUtil (version 3.12; latest stable as of 2015-09-26)
 * The complete license header for the original file can be found in the LICENSE_poiDateUtil.txt
 * file.
 */
package com.sieuferd.importer.excel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import java.util.AbstractMap.SimpleEntry;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import javax.annotation.Nullable;
import org.apache.poi.ss.usermodel.BuiltinFormats;

final class DateStyleParser {
  private DateStyleParser() { }

  public enum TemporalType {
    NOT_TEMPORAL,
    ELAPSED,
    DATE_AND_TIME,
    DATE_ONLY,
    TIME_ONLY;
  }

  enum TokenType {
    NON_TEMPORAL("E+", "E-", "e+", "e-", "General"),
    MINUTE_OR_MONTH("m", "mm"),
    YEAR("yy", "yyyy"),
    MONTH("mmm", "mmmm", "mmmmm"),
    DAY("d", "dd", "ddd", "dddd"),
    HOUR("h", "hh"),
    SECOND("s", "ss"),
    AMPM("AM/PM", "A/P"),
    ELAPSED("[h]", "[mm]", "[ss]"),
    // Include as it may impact the interpretation of MINUTE_OR_MONTH tokens.
    SEMICOLON(";");

    private final ImmutableSet<String> simpleTokens;

    private TokenType(String ... simpleTokens) {
      this.simpleTokens = ImmutableSet.copyOf(simpleTokens);
    }
  }

  /** Simple token sets by length, longest first. */
  private static final ImmutableMap<Integer,ImmutableMap<String,TokenType>> TOKEN_MAP =
      buildTokenMap();

  private static ImmutableMap<Integer,ImmutableMap<String,TokenType>> buildTokenMap() {
    final Map<Integer,Map<String,TokenType>> ret = Maps.newLinkedHashMap();
    int longestTokenLength = 0;
    for (TokenType tokenType : TokenType.values()) {
      for (String token : tokenType.simpleTokens) {
        final int length = token.length();
        longestTokenLength = Math.max(longestTokenLength, length);
        Map<String,TokenType> set = ret.get(length);
        if (set == null) {
          set = Maps.newLinkedHashMap();
          ret.put(length, set);
        }
        set.put(token, tokenType);
      }
    }
    final Map<Integer,ImmutableMap<String,TokenType>> retImm = Maps.newLinkedHashMap();
    // Go longest first.
    for (int i = longestTokenLength; i >= 0; i--) {
      final Map<String,TokenType> set = ret.get(i);
      if (set != null)
        retImm.put(i, ImmutableMap.copyOf(set));
    }
    return ImmutableMap.copyOf(retImm);
  }

  /** @return the longest simple token detected in formatString at the specified offset, or null if
              no simple token is found */
  private static @Nullable Entry<String,TokenType> findSimpleToken(String formatString, int offset)
  {
    for (Entry<Integer,ImmutableMap<String,TokenType>> entry : TOKEN_MAP.entrySet()) {
      final int length = entry.getKey();
      final Map<String,TokenType> tokenMap = entry.getValue();
      if (offset + length <= formatString.length()) {
        final String candidateToken = formatString.substring(offset, offset + length);
        TokenType type = tokenMap.get(candidateToken);
        if (type != null)
          return new SimpleEntry<>(candidateToken, type);
      }
    }
    return null;
  }

  private static int codePointCharCount(String s, int i) {
    return i >= s.length() ? 1 : Character.charCount(s.codePointAt(i));
  }

  /** Tokenize an Excel format string, returning a list of temporal tokens only. Package-visible for
  testing. */
  static List<TokenType> getTemporalTokens(final String formatString) {
    // See ECMA-376 page 1774.
    Preconditions.checkNotNull(formatString);
    final List<TokenType> ret = Lists.newArrayList();
    for (int i = 0; i < formatString.length(); ) {
      final char c = formatString.charAt(i);
      final Entry<String,TokenType> simpleToken = findSimpleToken(formatString, i);
      if (simpleToken != null) {
        if (simpleToken.getValue() != TokenType.NON_TEMPORAL)
          ret.add(simpleToken.getValue());
        i += simpleToken.getKey().length();
      } else if (c == '\\' || c == '_') {
        i++;
        // Uses the following character literally; skip.
        i += codePointCharCount(formatString, i);
      } else if (c == '*') {
        i++;
        // Skip additional '*' characters.
        while (i < formatString.length() && formatString.charAt(i) == '*')
          i++;
        // Uses the following character literally; skip.
        i += codePointCharCount(formatString, i);
      } else if (c == '\"' || c == '[') {
        final char endChar = (c == '\"') ? '\"' : ']';
        /* Quoted string to be used literally, or bracketed token (colors, conditionals) other than
        simple temporals; skip. */
        i++;
        for (; i < formatString.length(); i += codePointCharCount(formatString, i)) {
          if (formatString.charAt(i) == endChar) {
            i++;
            break;
          }
        }
      } else {
        // All other tokens are single-character, non-temporal tokens.
        i += codePointCharCount(formatString, i);
      }
    }
    return Collections.unmodifiableList(ret);
  }

  /**
   * Get the temporal type of a given format string.
   *
   * <p>This method was written from scratch as an alternative to
   * {@link org.apache.poi.ss.usermodel.DateUtil#isADateFormat}, with the ability to distinguish
   * between different kinds of temporal types (e.g. date and time vs. just date or just time).
   */
  public static TemporalType getTemporalType(final @Nullable String formatString) {
    if (formatString == null || formatString.isEmpty())
      return TemporalType.NOT_TEMPORAL;
    final List<TokenType> tokens = getTemporalTokens(formatString);
    final Set<TokenType>  tokenSet = tokens.isEmpty() ? ImmutableSet.of() : EnumSet.copyOf(tokens);
    if (tokenSet.contains(TokenType.ELAPSED))
      return TemporalType.ELAPSED;
    boolean usesMonth = tokenSet.contains(TokenType.MONTH);
    boolean usesMinutes = false;
    if (tokenSet.contains(TokenType.MINUTE_OR_MONTH)) {
      for (int i = 0; i < tokens.size(); i++) {
        if (tokens.get(i) == TokenType.MINUTE_OR_MONTH) {
          /* "If "m" or "mm" code is used immediately after the "h" or "hh" code (for hours) or
          immediately before the "ss" code (for seconds), the application shall display minutes
          instead of the month." (ECMA-376 p. 1780) Looking at Excel's behavior, it seems fine to
          substitute an "s" token instead as well. */
          if (i - 1 >= 0 && tokens.get(i - 1) == TokenType.HOUR ||
              i + 1 < tokens.size() && tokens.get(i + 1) == TokenType.SECOND)
          {
            usesMinutes = true;
          } else {
            usesMonth = true;
          }
        }
      }
    }
    final boolean usesDate =
        tokenSet.contains(TokenType.YEAR) || usesMonth || tokenSet.contains(TokenType.DAY);
    final boolean usesTime =
        tokenSet.contains(TokenType.HOUR) || usesMinutes || tokenSet.contains(TokenType.SECOND) ||
        tokenSet.contains(TokenType.AMPM);
    if (usesDate && usesTime) {
      return TemporalType.DATE_AND_TIME;
    } else if (usesDate) {
      return TemporalType.DATE_ONLY;
    } else if (usesTime) {
      if (tokenSet.contains(TokenType.HOUR)) {
        return TemporalType.TIME_ONLY;
      } else {
        return TemporalType.ELAPSED;
      }
    } else {
      return TemporalType.NOT_TEMPORAL;
    }
  }

  /** Indices are (non-extended) format codes. May contain nulls. */
  private static final List<TemporalType> INTERNAL_FORMAT_TEMPORAL_TYPES =
      createInternalFormatTemporalTypes();

  private static List<TemporalType> createInternalFormatTemporalTypes() {
    final List<TemporalType> ret = Lists.newArrayList();
    final String stringFormats[] = BuiltinFormats.getAll();
    for (int i = 0; i < stringFormats.length; i++)
      ret.add(i >= 0x17 && i <= 0x24 ? null : getTemporalType(stringFormats[i]));
    return Collections.unmodifiableList(ret);
  }

  /**
   * If the provided format code is one of Excel's internal formats, return the corresponding
   * temporal type.
   */
  public static @Nullable TemporalType getInternalTemporalType(int formatCode) {
    return (formatCode >= 0 && formatCode < INTERNAL_FORMAT_TEMPORAL_TYPES.size())
        ? INTERNAL_FORMAT_TEMPORAL_TYPES.get(formatCode)
        : null;
  }
}
