// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.sieuferd.database.SQLConnection;
import com.sieuferd.query.SQLDatabase;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.sql.SQLException;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.poifs.filesystem.DirectoryNode;
import org.apache.poi.poifs.filesystem.DocumentInputStream;
import org.apache.poi.poifs.filesystem.NPOIFSFileSystem;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

public abstract class ExcelImporter {
  public static final ExcelImporter MAC_EXCEL_CLIPBOARD_INSTANCE =
      new MacExcelClipboardImporter();
  public static final ExcelImporter WIN_EXCEL_CLIPBOARD_INSTANCE =
      new WindowsExcelClipboardImporter();

  public abstract void importClipboardStream(
      InputStream is, SQLConnection output, SQLDatabase outputDatabase,
      PrintWriter pw) throws SQLException, IOException;

  /**
   * Importer for the clipboard format used by Excel for Mac 2011. This format, identified on the
   * MacOS pasteboard by the mysterious letters "EMBS", is an XLSX workbook embedded within an OLE 2
   * Compound Document under the entry name "Package".
   */
  private static final class MacExcelClipboardImporter extends ExcelImporter {
    @Override
    public void importClipboardStream(InputStream is, SQLConnection output,
        SQLDatabase outputDatabase, PrintWriter pw) throws SQLException, IOException
    {
      try (NPOIFSFileSystem fs = new NPOIFSFileSystem(is)) {
        DirectoryNode dir = fs.getRoot();
        try (DocumentInputStream dis = dir.createDocumentInputStream(dir.getEntry("Package"))) {
          try (XSSFWorkbook wb = new XSSFWorkbook(dis)) {
            pw.println("Sheets: " + wb.getNumberOfSheets());
            pw.println("First sheet name: " + wb.getSheetName(0));
          }
        }
      }
    }
  }

  /**
   * Importer for the BIFF8 clipboard format used by Excel for Windows. This corresponds directly to
   * the XLS file format, and is known to be supported by Excel for Windows 2013.
   */
  private static final class WindowsExcelClipboardImporter extends ExcelImporter {
    @Override
    public void importClipboardStream(InputStream is, SQLConnection output,
        SQLDatabase outputDatabase, PrintWriter pw) throws SQLException, IOException
    {
      try (HSSFWorkbook wb = new HSSFWorkbook(is)) {
        pw.println("Sheets: " + wb.getNumberOfSheets());
        pw.println("First sheet name: " + wb.getSheetName(0));
      }
    }
  }
}
