// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.sieuferd.importer.ImportException;
import com.sieuferd.importer.excel.Spreadsheet.DataHandler;
import com.sieuferd.importer.excel.Spreadsheet.Region;
import java.io.IOException;
import java.util.List;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.NotThreadSafe;

/**
 * A spreadsheet-like data source. A spreadsheet may contain multiple regions that might each be
 * interpreted as a table. Spreadsheet values are typed per-cell rather than per-column; it is up to
 * the client to figure out how to take a region of cells and turn it into a proper relational
 * table.
 */
@NotThreadSafe
abstract class Spreadsheet implements AutoCloseable {
  /**
   * Get a list of regions. The list of regions should be in a suitable presentation order, as the
   * client may choose to ask the user to select a region to import from a list.
   */
  public abstract List<Region> getRegions();

  /**
   * Start handling the data in the specified region. This method will only return once all the data
   * has been handled or an exception is thrown. Note that this API design means that only one
   * region can be handled at a time, which should reduce the chance of bugs.
   */
  public abstract void handleData(Region region, DataHandler handler)
      throws ImportException, IOException;

  @Override
  public abstract void close() throws IOException;

  /**
   * Identifier for a region of cells in the spreadsheet. A {@code Region} object should remain
   * valid even when the spreadsheet is closed and then reopened, as long as the underlying region
   * remains intact. {@code Region} objects should be comparable using {@code Object.equals}.
   */
  static abstract class Region {
    /* TODO[pd]: If relevant, allow regions to specify whether they include a table row or not
                 (might be the case for regions using Excel's "table" feature). */

    /**
     * Get a suitable presentation name for this region, for instance to allow the user to select a
     * region to import. Clients should not assume that presentation names are necessarily unique.
     */
    public abstract String getPresentationName();
  }

  static interface DataHandler {
    /* TODO[pd]: Allow cell formatting information to be used to detect the presence of a header
                 row or not. */

    /**
     * Indicate that the current row is complete. Will be called at the end of every row, including
     * the last one.
     */
    public void rowComplete() throws ImportException;
    /**
     * Provide data for a cell in the current row. Cell data is guaranteed to be provided in order
     * of monotonically increasing column indices. Skipped cells are assumed to be null.
     *
     * @param columnIndex zero-based column index
     */
    public void cell(int columnIndex, @Nonnull Object data);

    public void reportWarning(ParseWarning warning);
  }

  // TODO[pd]: Make more exceptions into warnings.
  enum ParseWarning {
    ERRORS_CONVERTED_TO_NULL,
    INVALID_DATE_SERIAL_CONVERTED_TO_NULL,
    /**
     * A non-finite floating-point value (NaN or infinity) was converted to a null value. This
     * should never really happen with spreadsheets generated by Excel, but it's conceivable that
     * other tools might generate such values.
     */
    NON_FINITE_NUMBER_CONVERTED_TO_NULL
  }
}
