// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.sieuferd.importer.excel.DateStyleParser.TemporalType;
import com.sieuferd.importer.excel.Spreadsheet.DataHandler;
import com.sieuferd.importer.excel.Spreadsheet.ParseWarning;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.ZoneOffset;
import java.util.Calendar;
import java.util.TimeZone;
import javax.annotation.Nullable;
import org.apache.poi.ss.usermodel.DateUtil;

final class SpreadsheetUtil {
  private SpreadsheetUtil() { }

  private static final TimeZone TIME_ZONE_UTC = TimeZone.getTimeZone(ZoneOffset.UTC);

  /**
   * The maximum number of significant digits used by Excel. Numerical values parsed from the XLSX
   * or XLS formats must be rounded to this precision for numbers to appear identical,
   * digit-for-digit, to what they look like in Excel's formula bar.
   */
  private static final int SIGNIFICANT_DIGITS = 15;

  private static BigDecimal roundNumberToCorrectPrecision(final BigDecimal num) {
    /* It would seem best to keep the maximum precision of numbers available, but in order to get
    the exact same number as is shown in Excel's formula bar, we must truncate numbers to 15 digits.
    This is true both for the XLSX formats and the XLS format. Parsing XLSX numbers using
    Double.parseDouble is not enough to achieve the right rounding, even though Excel uses the same
    kinds of floating point numbers as Java (IEEE 754; see
    https://support.microsoft.com/en-us/kb/78113 and the deserialization code in
    org.apache.poi.hssf.record.NumberRecord, which boils down to Double.longBitsToDouble).

    For instance, the value entered and displayed as "3.51462293" in Excel's formula bar is
    stored as the string "3.5146229299999998" in the XLSX file format. Another example is the
    number 5.985:

      5.985               after typing into Excel
      5.9849999999999994  in XLSX file
      5.984999999999999   after doing BigDecimal.valueOf(Double.parseDouble(...))
      5.98499999999999    in Excel after TSV paste of the preceding
      5.985               after doing new BigDecimal(s) and rounding to 15 digits (see below)

    The unit tests include several numbers of different magnitudes and precisions (including
    both maximum and smaller precisions) to confirm correct digit-for-digit conversion. */
    BigDecimal ret = num;
    // See http://stackoverflow.com/questions/7572309/any-neat-way-to-limit-significant-figures-with-bigdecimal
    final int newScale = SIGNIFICANT_DIGITS - ret.precision() + ret.scale();
    ret = ret.setScale(newScale, RoundingMode.HALF_EVEN).stripTrailingZeros();
    /* Avoid having numbers like 30 show up as 3E+1 due to the preceding stripTrailingZeros.
    No rounding or other modification to the number happens at this step; this is mostly for
    BigDecimal.toString() to produce a nicer representation. */
    if (ret.scale() < 0)
      ret = ret.setScale(0);
    return ret;
  }

  public static @Nullable Object convertNumber(
      BigDecimal num, TemporalType temporalType, DataHandler dataHandler, boolean date1904)
  {
    final BigDecimal roundedNumber = roundNumberToCorrectPrecision(num);
    if (temporalType == TemporalType.NOT_TEMPORAL || temporalType == TemporalType.ELAPSED)
      return roundedNumber;
    final double dNum = roundedNumber.doubleValue();
    /* DateUtil.getJavaCalendar does not support sub-millisecond accuracy; this is probably OK,
    since Excel does not allow precision beyond milliseconds to be displayed in time formats,
    and since Excel does not display sub-second accuracy in the formula bar either. */
    /* DateUtil.getJavaDate throws a NPE where it should return null; use
    DateUtil.getJavaCalendar instead. */
    final Calendar calendarUTC = DateUtil.getJavaCalendar(dNum, date1904, TIME_ZONE_UTC);
    if (calendarUTC == null) {
      dataHandler.reportWarning(ParseWarning.INVALID_DATE_SERIAL_CONVERTED_TO_NULL);
      return null;
    }
    final LocalDateTime localDateTime =
        LocalDateTime.ofInstant(calendarUTC.toInstant(), ZoneOffset.UTC);
    if (temporalType == TemporalType.DATE_ONLY && (dNum % 1) == 0) {
      // The cell was formatted to show the date only, and has a time component of zero.
      return LocalDate.from(localDateTime);
    } else if (temporalType == TemporalType.TIME_ONLY && (dNum >= 0.0 && dNum < 1.0)) {
      // The cell was formatted to show the time only, and has a date component of zero.
      return LocalTime.from(localDateTime);
    } else {
      return localDateTime;
    }
  }
}
