// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.sieuferd.importer.ImportException;
import com.sieuferd.importer.excel.DateStyleParser.TemporalType;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import org.apache.poi.hssf.eventusermodel.AbortableHSSFListener;
import org.apache.poi.hssf.eventusermodel.HSSFEventFactory;
import org.apache.poi.hssf.eventusermodel.HSSFListener;
import org.apache.poi.hssf.eventusermodel.HSSFRequest;
import org.apache.poi.hssf.eventusermodel.HSSFUserException;
import org.apache.poi.hssf.record.BOFRecord;
import org.apache.poi.hssf.record.BoolErrRecord;
import org.apache.poi.hssf.record.BoundSheetRecord;
import org.apache.poi.hssf.record.CellValueRecordInterface;
import org.apache.poi.hssf.record.DateWindow1904Record;
import org.apache.poi.hssf.record.EOFRecord;
import org.apache.poi.hssf.record.ExtendedFormatRecord;
import org.apache.poi.hssf.record.FormatRecord;
import org.apache.poi.hssf.record.FormulaRecord;
import org.apache.poi.hssf.record.LabelRecord;
import org.apache.poi.hssf.record.LabelSSTRecord;
import org.apache.poi.hssf.record.NumberRecord;
import org.apache.poi.hssf.record.RKRecord;
import org.apache.poi.hssf.record.Record;
import org.apache.poi.hssf.record.SSTRecord;
import org.apache.poi.hssf.record.StringRecord;
import org.apache.poi.hssf.record.WSBoolRecord;
import org.apache.poi.hssf.record.common.UnicodeString;
import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.poifs.filesystem.DirectoryNode;
import org.apache.poi.poifs.filesystem.NPOIFSFileSystem;

final class XLSSpreadsheet extends Spreadsheet {
  private final NPOIFSFileSystem fs;
  private final Map<WorksheetRegion,Integer> worksheetBytePositions;
  private final boolean date1904;
  private final SSTRecord sharedStringTable;
  /** Indicies are extended format indices. */
  private final ImmutableList<ExtendedFormatRecord> extendedFormatRecords;
  /**
   * Keys are non-extended format codes. Format codes are not necessarily contiguous, so use a map
   * instead of a list. See the constructor for org.apache.poi.ss.usermodel.HSSFDataFormat. Don't
   * include internal format codes.
   */
  private final ImmutableMap<Integer,TemporalType> temporalTypesByFormatCode;

  private XLSSpreadsheet(NPOIFSFileSystem fs) throws IOException, ImportException {
    Preconditions.checkNotNull(fs);
    this.fs = fs;

    final InitialRecordListener initialRecordListener = new InitialRecordListener();
    processRecords(initialRecordListener.createHSSFRequest(), 0);
    date1904 = initialRecordListener.date1904;
    worksheetBytePositions = ImmutableMap.copyOf(initialRecordListener.worksheetBytePositions);
    sharedStringTable = (initialRecordListener.sstRecord == null)
        ? new SSTRecord() : initialRecordListener.sstRecord;
    extendedFormatRecords = ImmutableList.copyOf(initialRecordListener.extendedFormatRecords);
    temporalTypesByFormatCode =
        ImmutableMap.copyOf(initialRecordListener.temporalTypesByFormatCode);
  }

  private void processRecords(HSSFRequest hssfRequest, long offset)
      throws ImportException, IOException
  {
    try {
      new HSSFEventFactory().abortableProcessEvents(hssfRequest, getRecordInputStream(offset));
    } catch (NormalProcessEndException e) {
      // Ignore.
    } catch (HSSFUserException e) {
      // HSSFUserException uses getReason() instead of getCause(). Weird!
      final Throwable reason = e.getReason();
      if (reason instanceof IOException) {
        throw (IOException) reason;
      } else if (reason instanceof ImportException) {
        throw (ImportException) reason;
      } else {
        throw new ImportException(e.getMessage(), reason);
      }
    }
  }

  private InputStream getRecordInputStream(final long offset) throws ImportException, IOException {
    final DirectoryNode root = fs.getRoot();
    final String workbookEntryName;
    try {
      workbookEntryName = HSSFWorkbook.getWorkbookDirEntryName(root);
    } catch (IllegalArgumentException e) {
      throw new ImportException(e);
    }
    /* TODO[pd]: DocumentInputStream.skip(), or rather, its implementation in NDocumentInputStream,
                 does not do a proper seek on the underlying file. Consider caching the stream to
                 avoid quadratic behavior when reading all sheets in order. */
    final InputStream ret = root.createDocumentInputStream(workbookEntryName);
    long remainingSkip = offset;
    while (remainingSkip > 0) {
      final long skipped = ret.skip(remainingSkip);
      if (skipped <= 0)
        throw new IOException("Failed to seek in DocumentInputStream");
      remainingSkip -= skipped;
    }
    return ret;
  }

  public static XLSSpreadsheet create(InputStream is) throws ImportException, IOException {
    /* While this reads the entire content of the InputStream into memory, this is not as bad as in
    XLSXSpreadsheet's case, where the uncompressed end of a zip stream was dumped into memory. Thus
    there is no need for despearate measures such as writing the stream to disk and reopening it as
    a file. */
    return new XLSSpreadsheet(new NPOIFSFileSystem(is));
  }

  public static XLSSpreadsheet create(File file) throws ImportException, IOException {
    return new XLSSpreadsheet(new NPOIFSFileSystem(file, true));
  }

  @Override
  public void close() throws IOException {
    fs.close();
  }

  @Override
  public List<Region> getRegions() {
    return ImmutableList.copyOf(worksheetBytePositions.keySet());
  }

  @Override
  public void handleData(Region region, DataHandler handler) throws ImportException, IOException {
    @SuppressWarnings("element-type-mismatch")
    Integer worksheetBytePosition = worksheetBytePositions.get(region);
    if (worksheetBytePosition == null)
      throw new ImportException("Region " + region + " does not exist");
    WorksheetRecordListener worksheetRecordListener = new WorksheetRecordListener(handler);
    processRecords(worksheetRecordListener.createHSSFRequest(), worksheetBytePosition);
    worksheetRecordListener.signalPendingRowComplete();
  }

  private static boolean isWorksheetOrDialogSheet(BoundSheetRecord record) {
    // See StandardRecord.serialize(int, byte[])
    final int STANDARD_RECORD_HEADER_BYTES = 4;
    final byte[] bytes = record.serialize();
    // See the BoundSheet8 record definition in "MS-XLS: Excel Binary File Format Structure" p. 219.
    final byte dt = bytes[STANDARD_RECORD_HEADER_BYTES + 5];
    return dt == 0;
  }

  private static final class InitialRecordListener {
    int recordLevel = 0;
    final Map<WorksheetRegion,Integer> worksheetBytePositions = Maps.newLinkedHashMap();
    boolean date1904;
    @Nullable SSTRecord sstRecord;
    final List<ExtendedFormatRecord> extendedFormatRecords = Lists.newArrayList();
    final Map<Integer,TemporalType> temporalTypesByFormatCode = Maps.newLinkedHashMap();

    private HSSFRequest createHSSFRequest() {
      HSSFRequest ret = new HSSFRequest();
      ret.addListener((Record record) -> {
        date1904 = ((DateWindow1904Record) record).getWindowing() != 0;
      }, DateWindow1904Record.sid);
      ret.addListener((Record record) -> {
        BoundSheetRecord bsr = ((BoundSheetRecord) record);
        /* TODO[pd]: Consider excluding dialog sheets as well. Then throw an exception in the dialog
                     case in the WSBoolRecord handler below. */
        /* I would have liked to skip dialog sheets as well, but this would have required seeking to
        each individual worksheet just to read the WSBoolRecord. Since dialog sheets should be
        extremely rare, don't bother. */
        if (isWorksheetOrDialogSheet(bsr)) {
          worksheetBytePositions.put(new WorksheetRegion(
              worksheetBytePositions.size(), bsr.getSheetname()), bsr.getPositionOfBof());
        }
      }, BoundSheetRecord.sid);
      ret.addListener((Record record) -> {
        sstRecord = (SSTRecord) record;
      }, SSTRecord.sid);
      ret.addListener(new AbortableHSSFListener() {
        @Override
        public short abortableProcessRecord(Record record) throws HSSFUserException {
          recordLevel++;
          if (recordLevel > 1)
            throw new HSSFUserException("Unexpected nested BOFRecord");
          return 0;
        }
      }, BOFRecord.sid);
      ret.addListener(new AbortableHSSFListener() {
        @Override
        public short abortableProcessRecord(Record record) throws HSSFUserException {
          throw new NormalProcessEndException();
        }
      }, EOFRecord.sid);
      ret.addListener((Record record) -> {
        final FormatRecord formatRecord = (FormatRecord) record;
        final int formatCode = formatRecord.getIndexCode();
        TemporalType temporalType = DateStyleParser.getInternalTemporalType(formatCode);
        if (temporalType == null) {
          temporalType = DateStyleParser.getTemporalType(formatRecord.getFormatString());
          temporalTypesByFormatCode.put(formatCode, temporalType);
        }
      }, FormatRecord.sid);
      ret.addListener((Record record) -> {
        extendedFormatRecords.add((ExtendedFormatRecord) record);
      }, ExtendedFormatRecord.sid);
      return ret;
    }
  }

  private final class WorksheetRecordListener {
    final DataHandler dataHandler;
    int recordLevel = 0;
    int currentRow = 0;
    boolean pendingRowComplete;
    /**
     * Set when the next record is expected to be a StringRecord storing the cached string result
     * of a formula.
     */
    @Nullable FormulaRecord formulaRecordForCachedString;

    public WorksheetRecordListener(DataHandler handler) {
      Preconditions.checkNotNull(handler);
      this.dataHandler = handler;
    }

    public void signalPendingRowComplete() throws ImportException {
      if (pendingRowComplete) {
        pendingRowComplete = false;
        dataHandler.rowComplete();
      }
    }

    private HSSFRequest createHSSFRequest() {
      HSSFRequest ret = new HSSFRequest();
      ret.addListener(new AbortableHSSFListener() {
        @Override
        public short abortableProcessRecord(Record record) throws HSSFUserException {
          if (recordLevel == 0 && ((BOFRecord) record).getType() != BOFRecord.TYPE_WORKSHEET) {
            /* This case should never happen if the client uses only regions returned by getRegions,
            since isWorksheetOrDialogSheet excludes Macro, Chart, and VBA sheets. */
            throw new HSSFUserException("Not a plain worksheet");
          }
          recordLevel++;
          return 0;
        }
      }, BOFRecord.sid);
      ret.addListener(new AbortableHSSFListener() {
        @Override
        public short abortableProcessRecord(Record record) throws HSSFUserException {
          recordLevel--;
          if (recordLevel == 0)
            throw new NormalProcessEndException();
          if (recordLevel < 0)
            throw new HSSFUserException("Unexpected EOFRecord");
          return 0;
        }
      }, EOFRecord.sid);
      /* See org.apache.poi.hssf.model.RecordOrderer.isRowBlockRecord for a list of potentially
      relevant record types. */
      addListener(ret, (Record r) -> handleRecord((BoolErrRecord) r), BoolErrRecord.sid);
      addListener(ret, (Record r) -> handleRecord((FormulaRecord) r), FormulaRecord.sid);
      addListener(ret, (Record r) -> handleRecord((StringRecord) r), StringRecord.sid);
      addListener(ret, (Record r) -> handleRecord((LabelSSTRecord) r), LabelSSTRecord.sid);
      addListener(ret, (Record r) -> handleRecord((LabelRecord) r), LabelRecord.sid);
      addListener(ret, (Record r) -> handleRecord((NumberRecord) r), NumberRecord.sid);
      addListener(ret, (Record r) -> handleRecord((RKRecord) r), RKRecord.sid);
      addListener(ret, (Record r) -> handleRecord((WSBoolRecord) r), WSBoolRecord.sid);

      return ret;
    }

    // TODO[cc]: Consider using this for the BOF/EOF handlers as well.
    private void addListener(HSSFRequest req, HSSFListener listener, short sid) {
      req.addListener(new AbortableHSSFListener() {
        @Override
        public short abortableProcessRecord(Record record) throws HSSFUserException {
          if (recordLevel == 0)
            throw new HSSFUserException("Expected a BOF record");
          if (recordLevel > 1)
            return 0;
          try {
            listener.processRecord(record);
          } catch (WrapperException e) {
            Throwable cause = e.getCause();
            if (cause instanceof HSSFUserException) {
              throw (HSSFUserException) cause;
            } else {
              throw new HSSFUserException(cause);
            }
          }
          return 0;
        }
      }, sid);
    }

    private void signalCell(CellValueRecordInterface record, @Nullable Object value) {
      Preconditions.checkNotNull(record);
      if (value == null)
        return;
      try {
        final int row = record.getRow();
        if (row > currentRow) {
          signalPendingRowComplete();
          currentRow = row;
        } else if (row < currentRow) {
          throw new ImportException("Row numbers out of order");
        }
        dataHandler.cell(record.getColumn(), value);
        pendingRowComplete = true;
      } catch (ImportException e) {
        throw new WrapperException(e);
      }
    }

    private void handleRecord(WSBoolRecord record) {
      // See the comment about dialog sheets in the BoundSheetRecord handler.
      // WSBoolRecord.getDialog() does not seem to give the right result here.
      final boolean dialog = (record.getWSBool2() & 16) != 0;
      if (dialog)
        throw new WrapperException(new NormalProcessEndException());
    }

    private void handleRecord(BoolErrRecord record) {
      if (record.isBoolean()) {
        signalCell(record, record.getBooleanValue());
      } else {
        dataHandler.reportWarning(ParseWarning.ERRORS_CONVERTED_TO_NULL);
      }
    }

    private void handleRecord(FormulaRecord record) {
      final Object val;
      final int type = record.getCachedResultType();
      switch (type) {
        case HSSFCell.CELL_TYPE_BOOLEAN:
          val = record.getCachedBooleanValue();
        break; case HSSFCell.CELL_TYPE_NUMERIC:
          val = convertNumber(record, record.getValue());
        break; case HSSFCell.CELL_TYPE_ERROR:
          dataHandler.reportWarning(ParseWarning.ERRORS_CONVERTED_TO_NULL);
          val = null;
        break; case HSSFCell.CELL_TYPE_BLANK:
          val = null;
        break; case HSSFCell.CELL_TYPE_STRING:
          if (record.hasCachedResultString()) {
            formulaRecordForCachedString = record;
            // Wait for next record.
            return;
          } else {
            val = "";
          }
        break; case HSSFCell.CELL_TYPE_FORMULA:
        default:
          throw new WrapperException(new ImportException(
              "Unexpected formula value type code " + type));
      }
      signalCell(record, val);
    }

    private void handleRecord(StringRecord record) {
      if (formulaRecordForCachedString == null)
        throw new WrapperException(new ImportException(
            "Unexpected StringRecord (did not follow a FormulaRecord)"));
      try {
        signalCell(formulaRecordForCachedString, record.getString());
      } finally {
        formulaRecordForCachedString = null;
      }
    }

    private void handleRecord(LabelSSTRecord record) {
      final int idx = record.getSSTIndex();
      final UnicodeString s;
      try {
        s = sharedStringTable.getString(idx);
      } catch (IndexOutOfBoundsException e) {
        throw new WrapperException(new ImportException(
            "Shared string table index " + idx + " out of bounds", e));
      }
      signalCell(record, s.getString());
    }

    private void handleRecord(LabelRecord record) {
      signalCell(record, record.getValue());
    }

    private void handleRecord(NumberRecord record) {
      signalCell(record, convertNumber(record, record.getValue()));
    }

    private void handleRecord(RKRecord record) {
      signalCell(record, convertNumber(record, record.getRKNumber()));
    }

    private @Nullable Object convertNumber(CellValueRecordInterface record, double num) {
      if (!Double.isFinite(num)) {
        dataHandler.reportWarning(ParseWarning.NON_FINITE_NUMBER_CONVERTED_TO_NULL);
        return null;
      }
      // See HSSFCell.getCellStyle.
      final short extendedFormatIndex = record.getXFIndex();
      if (extendedFormatIndex < 0 || extendedFormatIndex >= extendedFormatRecords.size()) {
        throw new WrapperException(
            new ImportException("Invalid extended format index " + extendedFormatIndex));
      }
      final ExtendedFormatRecord extendedFormatRecord =
          extendedFormatRecords.get(extendedFormatIndex);
      final int formatCode = extendedFormatRecord.getFormatIndex();
      TemporalType temporalType = DateStyleParser.getInternalTemporalType(formatCode);
      if (temporalType == null) {
        temporalType = temporalTypesByFormatCode.get(formatCode);
        if (temporalType == null)
          throw new WrapperException(new ImportException("Invalid format code " + formatCode));
      }
      return SpreadsheetUtil.convertNumber(
          BigDecimal.valueOf(num), temporalType, dataHandler, date1904);
    }
  }

  @Immutable
  private static final class WorksheetRegion extends Region {
    private final int index;
    private final String worksheetName;

    public WorksheetRegion(int index, String worksheetName) {
      Preconditions.checkArgument(index >= 0);
      Preconditions.checkArgument(!worksheetName.isEmpty());
      this.index = index;
      this.worksheetName = worksheetName;
    }

    @Override
    public String getPresentationName() {
      return worksheetName;
    }

    @Override
    public String toString() {
      return index + "='" + worksheetName + "'";
    }

    @Override
    public int hashCode() {
      return Objects.hash(index, worksheetName);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof WorksheetRegion))
        return false;
      WorksheetRegion other = (WorksheetRegion) obj;
      return (this.index == other.index
          && this.worksheetName.equals(other.worksheetName));
    }
  }

  /**
   * Wrapper for exceptions thrown by handler methods registered using
   * WorksheetRecordListener.addListener. Exists just to allow the succinct lambda syntax to be used
   * to register all the record listeners.
   */
  private static final class WrapperException extends RuntimeException {
    @SuppressWarnings("ThrowableResultIgnored")
    public WrapperException(Throwable cause) {
      super(cause);
      Preconditions.checkNotNull(cause);
    }
  }

  /**
   * Exception thrown from an AbortableHSSFListener to end parsing mid-file under normal conditions.
   * Used for instance after encountering an EOF record for a worksheet that is being parsed.
   */
  private static final class NormalProcessEndException extends HSSFUserException { }
}
