// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.google.common.io.ByteStreams;
import com.sieuferd.importer.ImportException;
import com.sieuferd.importer.excel.DateStyleParser.TemporalType;
import com.sieuferd.importer.excel.Spreadsheet.DataHandler;
import com.sieuferd.importer.excel.Spreadsheet.Region;
import com.sieuferd.util.MyXMLUtil;
import com.sieuferd.util.StringUtil;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import javax.annotation.Nullable;
import javax.annotation.concurrent.Immutable;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.apache.poi.openxml4j.exceptions.InvalidFormatException;
import org.apache.poi.openxml4j.exceptions.OpenXML4JException;
import org.apache.poi.openxml4j.opc.OPCPackage;
import org.apache.poi.openxml4j.opc.PackageAccess;
import org.apache.poi.openxml4j.opc.PackagePart;
import org.apache.poi.openxml4j.opc.PackageRelationship;
import org.apache.poi.openxml4j.opc.PackageRelationshipTypes;
import org.apache.poi.ss.util.CellReference;
import org.apache.poi.xssf.eventusermodel.ReadOnlySharedStringsTable;
import org.apache.poi.xssf.eventusermodel.XSSFReader;
import org.apache.poi.xssf.model.StylesTable;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFRelation;
import org.openide.xml.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A SpreadsheetML (XLSX) spreadsheet, as defined by ECMA-376, 4th edition. See
 * {@link http://www.ecma-international.org/publications/standards/Ecma-376.htm}.
 */
final class XLSXSpreadsheet extends Spreadsheet {
  // TODO[cc]: Are these defined somewhere convenient?
  private static final String XMLNS = "http://schemas.openxmlformats.org/spreadsheetml/2006/main";
  private static final String XMLNS_R =
      "http://schemas.openxmlformats.org/officeDocument/2006/relationships";

  private static SAXParserFactory saxParserFactory;
  private final OPCPackage pack;
  private final ReadOnlySharedStringsTable sharedStringsTable;
  private final XSSFReader xssfReader;
  private final WorkbookData workbookData;
  private final StylesTable stylesTable;
  /** Index are style indices (also known as extended format indices in the XLS file format). */
  private final Map<Integer,TemporalType> temporalTypeCache = Maps.newHashMap();
  private @Nullable File tempFile;

  private static synchronized SAXParser createSAXParser() {
    /* "An application can use the same instance of the factory to obtain one or more instances of
    the SAXParser provided the instance of the factory isn't being used in more than one thread at a
    time." (SAXParserFactory Javadoc) */
    if (saxParserFactory == null) {
      saxParserFactory = SAXParserFactory.newInstance();
      saxParserFactory.setNamespaceAware(true);
    }
    try {
      return saxParserFactory.newSAXParser();
    } catch (ParserConfigurationException | SAXException e) {
      throw new RuntimeException("Unexpected exception while creating SAX parser", e);
    }
  }

  private XLSXSpreadsheet(OPCPackage pack, @Nullable File tempFile)
      throws IOException, ImportException
  {
    Preconditions.checkNotNull(pack);
    this.pack = pack;
    this.tempFile = tempFile;
    try {
      this.sharedStringsTable = new ReadOnlySharedStringsTable(pack);
      this.xssfReader = new XSSFReader(pack);
      // The styles table is global for the entire workbook, and is kept in a separate XML file.
      final StylesTable toStylesTable = xssfReader.getStylesTable();
      this.stylesTable = (toStylesTable == null) ? new StylesTable() : toStylesTable;
      this.workbookData = parseWorkbookData(pack);
    } catch (SAXException | OpenXML4JException e) {
      throw new ImportException(e);
    }
  }

  /** @param styleIndex may be -1 to indicate a default */
  private TemporalType getTemporalType(int styleIndex) throws SAXException {
    if (styleIndex < 0)
      return TemporalType.NOT_TEMPORAL;
    final XSSFCellStyle style;
    try {
      style = stylesTable.getStyleAt(styleIndex);
    } catch (IndexOutOfBoundsException e) {
      throw new SAXException("Style index " + styleIndex + " out of bounds");
    }
    return getTemporalType(style);
  }

  private static TemporalType getTemporalType(XSSFCellStyle style) {
    TemporalType ret = DateStyleParser.getInternalTemporalType(style.getDataFormat());
    return ret != null ? ret : DateStyleParser.getTemporalType(style.getDataFormatString());
  }

  /** @param styleIndex may be -1 to indicate a default */
  private TemporalType getTemporalTypeCached(int styleIndex) throws SAXException {
    TemporalType ret = temporalTypeCache.get(styleIndex);
    if (ret == null) {
      ret = getTemporalType(styleIndex);
      temporalTypeCache.put(styleIndex, ret);
    }
    return ret;
  }

  /**
   * Get the set of relationshipIDs that are actually worksheets, as opposed to other kinds of
   * content or sheets (e.g. chart sheets or dialog sheets).
   */
  private static Set<String> getWorksheetRelationshipIDs(PackagePart workbookPart)
      throws InvalidFormatException
  {
    final Set<String> ret = Sets.newLinkedHashSet();
    for(PackageRelationship rel : workbookPart.getRelationships()){
      if(rel.getRelationshipType().equals(XSSFRelation.WORKSHEET.getRelation()))
        ret.add(rel.getId());
    }
    return Collections.unmodifiableSet(ret);
  }

  @Immutable
  private static final class WorkbookData {
    final ImmutableList<WorksheetRegion> worksheetRegions;
    final boolean date1904;

    public WorkbookData(List<WorksheetRegion> worksheetRegions, boolean date1904) {
      this.worksheetRegions = ImmutableList.copyOf(worksheetRegions);
      this.date1904 = date1904;
    }
  }

  private static boolean isTrueString(String s) {
    /* Accept either "1" or "true" as a boolean true, as defined by the xsd:boolean type. Important
    attributes such as date1904 are explicitly stated to accept either (see ECMA-376 p. 3371). Be
    even a little more lenient than this, trimming first and being case-insensitive. */
    String st = s.trim();
    return st.equals("1") || st.toLowerCase(Locale.US).equals("true");
  }

  private static WorkbookData parseWorkbookData(OPCPackage pack)
      throws InvalidFormatException, SAXException, IOException
  {
    /* Do a manual parse of the worksheet list rather than relying on XSSFReader.getSheetsData(), as
    we'd like like to get at the underlying relID for later use, and as this avoids opening every
    worksheet's stream just to get a list of worksheets and their names. There is also a bug in
    xssfReader.getSheetsData() that causes a NPE for workbooks containing dialog sheets, and another
    bug where a sheet will be assigned the wrong name if it follows a dialog (or chart?) sheet. See
    XSSFReader.getSheetsData(). */
    final PackagePart workbookPart = pack.getPart(pack.getRelationshipsByType(
        PackageRelationshipTypes.CORE_DOCUMENT).getRelationship(0));
    final Set<String> worksheetRelationshipIDs = getWorksheetRelationshipIDs(workbookPart);

    final Document wbDoc = XMLUtil.parse(new InputSource(workbookPart.getInputStream()),
        false, true, XMLUtil.defaultErrorHandler(), null);
    final List<WorksheetRegion> worksheetRegions = Lists.newArrayList();
    boolean date1904 = false;
    final Element workbookPrElm =
        XMLUtil.findElement(wbDoc.getDocumentElement(), "workbookPr", XMLNS);
    if (workbookPrElm != null)
      date1904 = isTrueString(workbookPrElm.getAttribute("date1904"));
    final Element sheetsElm = XMLUtil.findElement(wbDoc.getDocumentElement(), "sheets", XMLNS);
    if (sheetsElm != null) {
      for (Element sheetElm : MyXMLUtil.findChildElementsByTagName(sheetsElm, "sheet", XMLNS)) {
        final String name = sheetElm.getAttribute("name");
        final String id = sheetElm.getAttributeNS(XMLNS_R, "id");
        if (!(name.isEmpty() || id.isEmpty()) && worksheetRelationshipIDs.contains(id))
          worksheetRegions.add(new WorksheetRegion(id, name));
      }
    }
    return new WorkbookData(worksheetRegions, date1904);
  }

  /**
   * Create an XLSXSpreadsheet from a File. This approach takes advantage of the random access
   * nature of files to avoid loading more data into memory than necessary. It is assumed that the
   * underlying file does not change while this XLSXSpreadsheet is open.
   */
  public static XLSXSpreadsheet create(File file) throws ImportException, IOException {
    return create(file, false);
  }

  private static XLSXSpreadsheet create(File file, boolean deleteOnClose)
      throws ImportException, IOException
  {
    if (!file.exists())
      throw new IOException("File does not exist: " + file);
    /* TODO[pd]: Consider if I should add an explicit file lock on opened input files (see
                 java.nio.channels.FileLock). And consider having a policy to only keep input files
                 open for a minimal amount of time (honored by clients of this class). */
    try {
      /* See the Javadoc for OPCPackage.open(InpustStream), which recommends using
      OPCPackage.open(String) to conserve memory. Reading the source code, I've confirmed that
      OPCPackage.open(File) works in the same way as the recommended OPCPackage.open(String). */
      return new XLSXSpreadsheet(
          OPCPackage.open(file, PackageAccess.READ), deleteOnClose ? file : null);
    } catch (InvalidFormatException e) {
      if (deleteOnClose)
        file.delete();
      throw new ImportException(e);
    }
  }

  /**
   * Create an XLSXSpreadsheet from an InputStream. The entire contents of the stream may be loaded
   * at once; for actual local files, {@link #create(File)} is prefered. The InputStream will be
   * closed at least by the time {@link #close()} is called.
   */
  public static XLSXSpreadsheet create(InputStream is) throws ImportException, IOException {
    /* Using OPCPackage.open(InputStream) here will force the entire _uncompressed_ contents of the
    InputStream to be dumped into memory, which is not good, because the XML-based format is
    extremely verbose. So write the compressed contents to disk and open as a file instead. */
    File tempFile = File.createTempFile("xlsximport", ".tmp");
    try (OutputStream os = new BufferedOutputStream(new FileOutputStream(tempFile))) {
      ByteStreams.copy(is, os);
      is.close();
      os.flush();
    } catch (IOException e) {
      tempFile.delete();
      throw e;
    }
    return create(tempFile, true);
  }

  @Override
  public List<Region> getRegions() {
    return ImmutableList.copyOf(workbookData.worksheetRegions);
  }

  @Override
  public void close() throws IOException {
    pack.close();
    if (tempFile != null) {
      tempFile.delete();
      tempFile = null;
    }
  }

  @Override
  public void handleData(Region region, DataHandler handler) throws IOException, ImportException {
    if (!getRegions().contains(region) || !(region instanceof WorksheetRegion))
      throw new ImportException("Region " + region + " does not exist");
    try (InputStream is = xssfReader.getSheet(((WorksheetRegion) region).rID)) {
      final SAXParser saxParser = createSAXParser();
      final XMLReader xmlReader = saxParser.getXMLReader();
      xmlReader.setContentHandler(new WorksheetXMLHandler(handler));
      xmlReader.parse(new InputSource(is));
    } catch (InvalidFormatException | SAXException e) {
      throw new ImportException(e);
    }
  }

  /* TODO[pd]: Add support for explicitly defined table regions, as defined by Excel's "table"
               feature. */
  @Immutable
  private static final class WorksheetRegion extends Region {
    private final String rID;
    private final String worksheetName;

    public WorksheetRegion(String rID, String worksheetName) {
      Preconditions.checkArgument(!rID.isEmpty());
      Preconditions.checkArgument(!worksheetName.isEmpty());
      this.rID = rID;
      this.worksheetName = worksheetName;
    }

    @Override
    public String getPresentationName() {
      return worksheetName;
    }

    @Override
    public String toString() {
      return rID + "='" + worksheetName + "'";
    }

    @Override
    public int hashCode() {
      return Objects.hash(rID, worksheetName);
    }

    @Override
    public boolean equals(Object obj) {
      if (obj == this)
        return true;
      if (!(obj instanceof WorksheetRegion))
        return false;
      WorksheetRegion other = (WorksheetRegion) obj;
      return (this.rID.equals(other.rID)
          && this.worksheetName.equals(other.worksheetName));
    }
  }

  // Package-visible for testing.
  final static int cellReferenceToColumnIndex(String cellReference) throws SAXException {
    /* "Regardless of the value of the refMode attribute, cell references shall be stored in XML in
    the A1 form." (ECMA-376 p. 2061) */
    for (int i = 0; i < cellReference.length(); i++) {
      if (Character.isDigit(cellReference.charAt(i)))
        return columnAddressToIndex(cellReference.substring(0, i));
    }
    throw new SAXException("Invalid cell reference '" + cellReference + "'");
  }

  // Package-visible for testing.
  final static int columnAddressToIndex(String columnAddress) throws SAXException {
    if (columnAddress.isEmpty())
      throw new SAXException("Empty column address");
    try {
      return CellReference.convertColStringToIndex(columnAddress);
    } catch (IllegalArgumentException e) {
      throw new SAXException("Illegal column reference '" + columnAddress + "'");
    }
  }

  /* ============================================================================================ */
  private static final ImmutableList<String> ROW_ELEMENT_STACK =
      ImmutableList.of("worksheet", "sheetData", "row");

  private static interface ValueParser {
    /**
     * @param stringValue may be empty
     * @param styleIndex may be -1 to indicate a default
     */
    public Object parseValue(String stringValue, int styleIndex) throws SAXException;
  }

  private static final class CellMeta {
    final int colIndex;
    final ValueParser parser;
    final int styleIndex;

    /** @param styleIndex may be -1 to indicate a default */
    public CellMeta(int colIndex, ValueParser parser, int styleIndex) {
      Preconditions.checkArgument(colIndex >= 0);
      Preconditions.checkNotNull(parser);
      this.colIndex = colIndex;
      this.parser = parser;
      this.styleIndex = styleIndex;
    }
  }

  private ImmutableMap<String,ValueParser> buildValueParsersMap(DataHandler dataHandler) {
    // See ECMA-376 p. 2433 for the applicable type codes.
    return ImmutableMap.<String,ValueParser>builder()
    .put("b", (s, styleIndex) -> { // Boolean.
      return isTrueString(s);
    })
    .put("d", (s, styleIndex) -> { // Date and time in ISO 8601 format. See ECMA-376 p. 2056 and 2064.
      // TODO[pd]: Make sure I support all the relevant formats here.
      try {
        return LocalDateTime.from(DateTimeFormatter.ISO_LOCAL_DATE_TIME.parse(s));
      } catch (DateTimeParseException e) {
        throw new SAXException("Unable to parse date/time value '" + s + "'", e);
      }
    })
    .put("e", (s, styleIndex) -> { // Error value.
      dataHandler.reportWarning(ParseWarning.ERRORS_CONVERTED_TO_NULL);
      return null;
    })
    .put("inlineStr", (s, styleIndex) -> { // Inline string.
      return applyEmptyStringPolicy(s);
    })
    .put("n", (s, styleIndex) -> { // Number.
      // Since this is the default type, we also need to handle the null case here.
      if (s.isEmpty())
        return null;
      final BigDecimal num;
      try {
        num = new BigDecimal(s);
      } catch (NumberFormatException e) {
        try {
          // Could it be a NaN or infinity value?
          if (!Double.isFinite(Double.parseDouble(s))) {
            dataHandler.reportWarning(ParseWarning.NON_FINITE_NUMBER_CONVERTED_TO_NULL);
            return null;
          }
        } catch (NumberFormatException e2) {
          // Ignore.
        }
        throw new SAXException("Unable to parse number '" + s + "'", e);
      }
      return SpreadsheetUtil.convertNumber(
          num, getTemporalTypeCached(styleIndex), dataHandler, workbookData.date1904);
    })
    .put("s", (s, styleIndex) -> { // Shared string.
      final int idx;
      try {
        idx = Integer.parseInt(s);
      } catch (NumberFormatException e) {
        throw new SAXException("Invalid shared string table index '" + s + "'", e);
      }
      try {
        return applyEmptyStringPolicy(sharedStringsTable.getEntryAt(idx));
      } catch (IndexOutOfBoundsException e) {
        throw new SAXException("Shared string table index " + idx + " out of bounds", e);
      }
    })
    .put("str", (s, styleIndex) -> { // Unformatted string. See ECMA-376 p. 2756.
      // Unformatted string resulting from a formula calculation.
      return applyEmptyStringPolicy(s);
    })
    .build();
  }

  private @Nullable String applyEmptyStringPolicy(String s) {
    /* Don't convert empty strings to nulls; Excel is perfectly capable of distinguishing the two.
    This method provides a centralized place to document this policy. */
    return s;
  }

  private final class WorksheetXMLHandler extends DefaultHandler {
    private static final String XMLNS = "http://schemas.openxmlformats.org/spreadsheetml/2006/main";
    private final DataHandler dataHandler;
    /** Contains empty strings for elements not in the expected namespace. */
    private final List<String> elementStack = Lists.newArrayList();
    private final StringBuilder valueBuilder = new StringBuilder();
    private boolean inRow;
    /** null when not in a column */
    private @Nullable CellMeta cellMeta;
    private boolean inV = false;
    private boolean inT = false;
    /** Keys are cell type codes. */
    private final ImmutableMap<String,ValueParser> valueParsers;

    WorksheetXMLHandler(DataHandler dataHandler) {
      Preconditions.checkNotNull(dataHandler);
      this.dataHandler = dataHandler;
      this.valueParsers = buildValueParsersMap(dataHandler);
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
      if (inV || inT)
        valueBuilder.append(ch, start, length);
    }

    private String elmName(String uri, String localName) {
      return uri.equals(XMLNS) ? localName : "";
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes)
        throws SAXException
    {
      final String elmName = elmName(uri, localName);
      elementStack.add(elmName);
      if (inRow && elmName.equals("c")) {
        if (valueBuilder.length() != 0)
          throw new SAXException("Unexpected parser state; is a column end tag missing?");
        int styleIndex = -1;
        final String styleStr = StringUtil.trimmedAndEmptyIfNull(attributes.getValue("s"));
        if (!styleStr.isEmpty()) {
          try {
            styleIndex = Integer.parseInt(styleStr);
          } catch (NumberFormatException e) {
            throw new SAXException("Invalid style index '" + styleStr + "'");
          }
        }
        String typeStr = StringUtil.trimmedAndEmptyIfNull(attributes.getValue("t"));
        if (typeStr.isEmpty())
          typeStr = "n"; // Number by default.
        ValueParser parser = valueParsers.get(typeStr);
        if (parser == null)
          throw new SAXException("Unexpected cell type code '" + typeStr + "'");
        cellMeta = new CellMeta(cellReferenceToColumnIndex(
            StringUtil.trimmedAndEmptyIfNull(attributes.getValue("r"))), parser, styleIndex);
      } else if (cellMeta != null && elmName.equals("v")) {
        if (valueBuilder.length() != 0)
          throw new SAXException("Expected a single cell value only");
        inV = true;
      } else if (cellMeta != null && elmName.equals("t")) {
        // There may be multiple <t> elements; just concatenate their contents.
        inT = true;
      } else if (elmName.equals("row") && elementStack.equals(ROW_ELEMENT_STACK)) {
        inRow = true;
      }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
      final String elmName = elmName(uri, localName);
      if (cellMeta != null && elmName.equals("c")) {
        String valStr = valueBuilder.toString();
        Object val = cellMeta.parser.parseValue(valStr, cellMeta.styleIndex);
        if (val != null)
          dataHandler.cell(cellMeta.colIndex, val);
        valueBuilder.setLength(0);
        cellMeta = null;
      } else if (cellMeta != null && elmName.equals("v")) {
        inV = false;
      } else if (cellMeta != null && elmName.equals("t")) {
        inT = false;
      } else if (elmName.equals("row") && elementStack.equals(ROW_ELEMENT_STACK)) {
        inRow = false;
        try {
          dataHandler.rowComplete();
        } catch (ImportException e) {
          throw new SAXException(e);
        }
      }
      if (!elementStack.remove(elementStack.size() - 1).equals(elmName))
        throw new SAXException("Unexpected end element </" + localName + ">");
    }
  }
}
