// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.google.common.base.Joiner;
import com.sieuferd.importer.excel.DateStyleParser.TemporalType;
import javax.annotation.Nullable;
import org.apache.poi.ss.usermodel.BuiltinFormats;
import org.junit.Assert;
import org.junit.Test;

public class DateStyleParserTest {
  private static void testFormatParse(
      String formatString, String expectedTokenization, TemporalType expectedType)
  {
    Assert.assertEquals(
        expectedTokenization, Joiner.on("|").join(DateStyleParser.getTemporalTokens(formatString)));
    Assert.assertEquals(
        expectedType, DateStyleParser.getTemporalType(formatString));
  }

  @Test
  public void testGetInternalTemporalType() {
    for (int i = 0; i <= BuiltinFormats.getAll().length + 500; i++) {
      if (i >= 0x17 && i <= 0x24)
        continue;
      String format = BuiltinFormats.getBuiltinFormat(i);
      if (format == null)
        continue;
      TemporalType internalTemporalType = DateStyleParser.getInternalTemporalType(i);
      TemporalType expected = getExpectedInternalTemporalType(i);
      Assert.assertEquals(expected, internalTemporalType);
    }
  }

  @Test
  public void testBuiltinFormats() {
    /* Test tokenization of all the built-in formats
    (see org.apache.poi.ss.usermodel.BuiltinFormats). */
    testFormatParse("General", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("0", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("0.00", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("#,##0", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("#,##0.00", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("\"$\"#,##0_);(\"$\"#,##0)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("\"$\"#,##0_);[Red](\"$\"#,##0)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("\"$\"#,##0.00_);(\"$\"#,##0.00)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("\"$\"#,##0.00_);[Red](\"$\"#,##0.00)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("0%", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("0.00%", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("0.00E+00", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("# ?/?", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("# ??/??", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("m/d/yy", "MINUTE_OR_MONTH|DAY|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("d-mmm-yy", "DAY|MONTH|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("d-mmm", "DAY|MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mmm-yy", "MONTH|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("h:mm AM/PM", "HOUR|MINUTE_OR_MONTH|AMPM", TemporalType.TIME_ONLY);
    testFormatParse("h:mm:ss AM/PM", "HOUR|MINUTE_OR_MONTH|SECOND|AMPM", TemporalType.TIME_ONLY);
    testFormatParse("h:mm", "HOUR|MINUTE_OR_MONTH", TemporalType.TIME_ONLY);
    testFormatParse("h:mm:ss", "HOUR|MINUTE_OR_MONTH|SECOND", TemporalType.TIME_ONLY);
    testFormatParse("m/d/yy h:mm", "MINUTE_OR_MONTH|DAY|YEAR|HOUR|MINUTE_OR_MONTH", TemporalType.DATE_AND_TIME);
    testFormatParse("#,##0_);(#,##0)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("#,##0_);[Red](#,##0)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("#,##0.00_);(#,##0.00)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("#,##0.00_);[Red](#,##0.00)", "SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("_(\"$\"* #,##0_);_(\"$\"* (#,##0);_(\"$\"* \"-\"_);_(@_)", "SEMICOLON|SEMICOLON|SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("_(* #,##0_);_(* (#,##0);_(* \"-\"_);_(@_)", "SEMICOLON|SEMICOLON|SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("_(* #,##0.00_);_(* (#,##0.00);_(* \"-\"??_);_(@_)", "SEMICOLON|SEMICOLON|SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("_(\"$\"* #,##0.00_);_(\"$\"* (#,##0.00);_(\"$\"* \"-\"??_);_(@_)", "SEMICOLON|SEMICOLON|SEMICOLON", TemporalType.NOT_TEMPORAL);
    testFormatParse("mm:ss", "MINUTE_OR_MONTH|SECOND", TemporalType.ELAPSED);
    testFormatParse("[h]:mm:ss", "ELAPSED|MINUTE_OR_MONTH|SECOND", TemporalType.ELAPSED);
    testFormatParse("mm:ss.0", "MINUTE_OR_MONTH|SECOND", TemporalType.ELAPSED);
    testFormatParse("##0.0E+0", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("@", "", TemporalType.NOT_TEMPORAL);
  }

  @Test
  public void testTricky() {
    testFormatParse("m/\\d/yy", "MINUTE_OR_MONTH|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("m/d\\/yy", "MINUTE_OR_MONTH|DAY|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("m/d/yy\\", "MINUTE_OR_MONTH|DAY|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("m/d/yy[", "MINUTE_OR_MONTH|DAY|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("m/d/yy\"", "MINUTE_OR_MONTH|DAY|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("m/d/yy\"h\"", "MINUTE_OR_MONTH|DAY|YEAR", TemporalType.DATE_ONLY);
    testFormatParse("m/\"d/yyh\"s", "MINUTE_OR_MONTH|SECOND", TemporalType.ELAPSED);
    testFormatParse("m/\"d/yyhs", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("[mmmm]", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("mmmm", "MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mmmm \"Hello\"", "MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mmmm \"Hello\" ss", "MONTH|SECOND", TemporalType.DATE_AND_TIME);
    testFormatParse("mmmm \"Hello\" ss;$$$$", "MONTH|SECOND|SEMICOLON", TemporalType.DATE_AND_TIME);
    testFormatParse("Emm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("E+mm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mmm", "MONTH", TemporalType.DATE_ONLY);
    testFormatParse("\\mmm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("_mmm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("*mmm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("**mmm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("***mmm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mm", "MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mmmm", "MONTH", TemporalType.DATE_ONLY);
    testFormatParse("yy mm", "YEAR|MINUTE_OR_MONTH", TemporalType.DATE_ONLY);
    testFormatParse("mm ss", "MINUTE_OR_MONTH|SECOND", TemporalType.ELAPSED);
    testFormatParse("hh mm ss", "HOUR|MINUTE_OR_MONTH|SECOND", TemporalType.TIME_ONLY);
    testFormatParse("yy mm ss", "YEAR|MINUTE_OR_MONTH|SECOND", TemporalType.DATE_AND_TIME);
  }

  @Test
  public void testMulticharCodepoints() {
    testFormatParse("\"\uD835\uDD0A\"", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("h", "HOUR", TemporalType.TIME_ONLY);
    testFormatParse("\\h", "", TemporalType.NOT_TEMPORAL);
    testFormatParse("\\\uD835\uDD0Ah", "HOUR", TemporalType.TIME_ONLY);
    // Does not actually eat the "h", since it's not a valid low surrogate.
    testFormatParse("\\\uD835h", "HOUR", TemporalType.TIME_ONLY);
    testFormatParse("\"\uD835\uDD0A\"h", "HOUR", TemporalType.TIME_ONLY);
    // Again, the quote is not a valid low surrogate.
    testFormatParse("\"\uD835\"h", "HOUR", TemporalType.TIME_ONLY);
    /* For now, I'm actually at loss as to how to generate a case that would not work unless code
    points are properly handled. Maybe such a case doesn't actually exist, and that the special
    code points handling wasn't necessary. */
  }

  private static @Nullable TemporalType getExpectedInternalTemporalType(int formatCode) {
    /* Internal Date Formats as described on page 427 in Microsoft Excel Dev's Kit. A newer
    reference is ECMA-376, 4th edition, p. 1767. See also
    org.apache.poi.ss.usermodel.BuiltinFormats. */
    switch (formatCode) {
      case 0:
      case 1:
      case 2:
      case 3:
      case 4:

      // These four are defined in BuiltinFormats but not in ECMA-376.
      case 5:
      case 6:
      case 7:
      case 8:

      case 9:
      case 10:
      case 11:
      case 12:
      case 13:
        return TemporalType.NOT_TEMPORAL;
      case 14:
      case 15:
      case 16:
      case 17:
        return TemporalType.DATE_ONLY;
      case 18:
      case 19:
      case 20:
      case 21:
        return TemporalType.TIME_ONLY;
      case 22:
        return TemporalType.DATE_AND_TIME;
      case 37:
      case 38:
      case 39:
      case 40:

      // These four are defined in BuiltinFormats but not in ECMA-376.
      case 41:
      case 42:
      case 43:
      case 44:
        return TemporalType.NOT_TEMPORAL;

      case 45: // "mm:ss"
      case 46: // "[h]:mm:ss"
      case 47: // "mmss.0"
        return TemporalType.ELAPSED;
      case 48:
      case 49:
        return TemporalType.NOT_TEMPORAL;
    }
    return null;
  }
}
