// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.importer.excel;

import com.google.common.base.Charsets;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import com.google.common.io.CharStreams;
import com.sieuferd.importer.ImportException;
import com.sieuferd.importer.excel.Spreadsheet.ParseWarning;
import com.sieuferd.importer.excel.Spreadsheet.Region;
import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.List;
import java.util.Set;
import javax.annotation.Nullable;
import org.junit.Assert;
import org.junit.Test;
import org.netbeans.api.annotations.common.StaticResource;
import org.xml.sax.SAXException;

public class ExcelImporterTest {
  private static final String SAMPLES_FOLDER = "com/sieuferd/importer/excel/";
  // TimeEntries example.

  // Generated with Microsoft Access 2013 (Windows).
  private static final @StaticResource String TIME_ENTRIES_ACCESS2013_XLS =
      SAMPLES_FOLDER + "TimeEntries_access2013.xls";
  // Generated with Microsoft Access 2013 (Windows).
  private static final @StaticResource String TIME_ENTRIES_ACCESS2013_XLSX =
      SAMPLES_FOLDER + "TimeEntries_access2013.xlsx";
  /* Generated with Microsoft Excel 2011 for Mac (paste values from TimeEntries_access2013.xls, then
  format the temporal column as date + time). */
  private static final @StaticResource String TIME_ENTRIES_EXCEL2011MAC_XLS =
      SAMPLES_FOLDER + "TimeEntries_excel2011mac.xls";
  /* Generated with Microsoft Excel 2011 for Mac (paste values from TimeEntries_access2013.xls, then
  format the temporal column as date + time). */
  private static final @StaticResource String TIME_ENTRIES_EXCEL2011MAC_XLSX =
      SAMPLES_FOLDER + "TimeEntries_excel2011mac.xlsx";
  /* Serialized clipboard contents when copying the table of data from TimeEntries_excel2011mac.xlsx
  on Microsoft Excel 2011 for Mac, using the native data flavor identifier
  "CorePasteboardFlavorType 0x454D4253". */
  private static final @StaticResource String TIME_ENTRIES_EXCEL2011MAC_CLIPBOARD =
      SAMPLES_FOLDER + "TimeEntries_excel2011mac_clipboard.embs";
  private static final @StaticResource String TIME_ENTRIES_EXPECTED =
    SAMPLES_FOLDER + "TimeEntries_expected.tsv";
  /* Since the Access-generated file formats temporal fields to only show the date, there are a few
  valid conversion differences vs. the Excel versions, where we show both time and date. */
  private static final @StaticResource String TIME_ENTRIES_ACCESS_EXPECTED =
    SAMPLES_FOLDER + "TimeEntries_access_expected.tsv";

  /* MultiSheet example. This sheet includes a chart sheet and a dialog sheet as well as a regular
  worksheet following the chart and dialog sheets. The chart and dialog sheets should be properly
  ignored by Spreadsheet.getRegions(). */

  private static final @StaticResource String MULTISHEET_EXCEL2011MAC_XLS =
    SAMPLES_FOLDER + "MultiSheet_excel2011mac.xls";
  private static final @StaticResource String MULTISHEET_EXCEL2011MAC_XLSX =
    SAMPLES_FOLDER + "MultiSheet_excel2011mac.xlsx";
  /* I have not gone through the contents of this one in detail, but this more elaborate example
  should at least guard against regressions. */
  private static final @StaticResource String MULTISHEET_EXPECTED =
    SAMPLES_FOLDER + "MultiSheet_expected.tsv";

  /* Succinct example. This one tries to cover a lot of different cases related to individual cell
  values in a minimal number of cells. */

  private static final @StaticResource String SUCCINCT_EXCEL2011MAC_XLS =
    SAMPLES_FOLDER + "Succinct_excel2011mac.xls";
  private static final @StaticResource String SUCCINCT_EXCEL2011MAC_XLSX =
    SAMPLES_FOLDER + "Succinct_excel2011mac.xlsx";
  private static final @StaticResource String SUCCINCT_EXPECTED =
    SAMPLES_FOLDER + "Succinct_expected.tsv";

  private static final @StaticResource String DATE1904_EXCEL2011MAC_XLS =
    SAMPLES_FOLDER + "Date1904_excel2011mac.xls";
  private static final @StaticResource String DATE1904_EXCEL2011MAC_XLSX =
    SAMPLES_FOLDER + "Date1904_excel2011mac.xlsx";
  private static final @StaticResource String DATE1904_EXPECTED =
    SAMPLES_FOLDER + "Date1904_expected.tsv";

  private static InputStream getResource(String resource) throws IOException {
    InputStream ret = ExcelImporterTest.class.getClassLoader().getResourceAsStream(resource);
    if (ret == null)
      throw new IOException("Resource not found");
    return ret;
  }

  private static String readExpected(String resource) throws IOException {
    // See http://stackoverflow.com/questions/309424/read-convert-an-inputstream-to-a-string .
    try (Reader r = new InputStreamReader(getResource(resource), Charsets.UTF_8)) {
      return CharStreams.toString(r);
    }
  }

  private void testImport(
      String expectedTSV, String resourceFile, String expectRegions, @Nullable Integer regionIndex,
      ImmutableSet<ParseWarning> expectedWarnings)
      throws IOException, ImportException
  {
    InputStream resourceStream = getResource(resourceFile);
    final boolean isXLSX = resourceFile.endsWith("xlsx");
    try (Spreadsheet spreadsheet = isXLSX
        ? XLSXSpreadsheet.create(resourceStream)
        : XLSSpreadsheet.create(resourceStream)) {
      List<Region> regions = spreadsheet.getRegions();
      Assert.assertEquals(expectRegions, regions.toString());
      final Set<ParseWarning> warnings = Sets.newLinkedHashSet();
      final StringBuilder tsvStringBuilder = new StringBuilder();
      final List<Region> useRegions = regionIndex == null ? regions :
          ImmutableList.of(regions.get(regionIndex));
      int i = 0;
      for (Region region : useRegions) {
        try (CharArrayWriter writer = new CharArrayWriter()) {
          try (PrintWriter pw = new PrintWriter(writer)) {
            TSVGenerator gen = new TSVGenerator(pw);
            spreadsheet.handleData(region, gen);
            warnings.addAll(gen.getWarnings());
            String sheetString = writer.toString();
            /* Don't add if empty; this is mostly here to skip the dialog sheet in Multisheet for
            XLS files. */
            if (!sheetString.isEmpty()) {
              if (i > 0)
                tsvStringBuilder.append("-----------------------------------------------------------------------\n");
              tsvStringBuilder.append(sheetString);
              i++;
            }
          }
        }
      }
      String tsvString = tsvStringBuilder.toString();
      if (!tsvString.equals(expectedTSV)) {
        System.out.println("=======================================================================");
        System.out.print(tsvString);
        System.out.println("=======================================================================");
        System.out.flush();
        throw new AssertionError(
            "Actual does not match expected output; actual output listed above");
      }
      Assert.assertEquals(expectedWarnings, warnings);
    }
  }

  @Test
  public void testSuccinct() throws IOException, ImportException {
    String expected = readExpected(SUCCINCT_EXPECTED);
    testImport(expected, SUCCINCT_EXCEL2011MAC_XLSX, "[rId1='Sheet1']", 0, ImmutableSet.of(
        ParseWarning.INVALID_DATE_SERIAL_CONVERTED_TO_NULL,
        ParseWarning.ERRORS_CONVERTED_TO_NULL));
    testImport(expected, SUCCINCT_EXCEL2011MAC_XLS, "[0='Sheet1']", 0, ImmutableSet.of(
        ParseWarning.INVALID_DATE_SERIAL_CONVERTED_TO_NULL,
        ParseWarning.ERRORS_CONVERTED_TO_NULL));
  }

  @Test
  public void testTimeEntries() throws IOException, ImportException {
    String expected = readExpected(TIME_ENTRIES_EXPECTED);
    ImmutableSet<ParseWarning> expectedWarnings = ImmutableSet.of();
    testImport(expected, TIME_ENTRIES_EXCEL2011MAC_XLSX, "[rId1='Sheet1']", 0, expectedWarnings);
    testImport(expected, TIME_ENTRIES_EXCEL2011MAC_XLS, "[0='Sheet1']", 0, expectedWarnings);
    String accessExpected = readExpected(TIME_ENTRIES_ACCESS_EXPECTED);
    testImport(accessExpected, TIME_ENTRIES_ACCESS2013_XLSX, "[rId1='TimeEntries']", 0, expectedWarnings);
    testImport(accessExpected, TIME_ENTRIES_ACCESS2013_XLS, "[0='TimeEntries']", 0, expectedWarnings);
  }

  @Test
  public void testDate1904() throws IOException, ImportException {
    String expected = readExpected(DATE1904_EXPECTED);
    testImport(expected, DATE1904_EXCEL2011MAC_XLSX, "[rId1='Sheet1']", 0, ImmutableSet.of());
    testImport(expected, DATE1904_EXCEL2011MAC_XLS, "[0='Sheet1']", 0, ImmutableSet.of());
  }


  @Test
  public void testMultiSheet() throws IOException, ImportException {
    String expected = readExpected(MULTISHEET_EXPECTED);
    // Both order, IDs, and presentation names matter here.
    String expectedRegions =
        "[rId1='sort-filter', rId2='wifi-hostnames_orig2', rId3='wifi-connections_orig1', " +
          "rId4='basic-interactions', rId7='following']";
    // The XLS reader does not skip dialog sheets.
    String expectedRegionsXLS =
        "[0='sort-filter', 1='wifi-hostnames_orig2', 2='wifi-connections_orig1', " +
          "3='basic-interactions', 4='testdialog', 5='following']";
    testImport(expected, MULTISHEET_EXCEL2011MAC_XLSX, expectedRegions, null, ImmutableSet.of());
    testImport(expected, MULTISHEET_EXCEL2011MAC_XLS, expectedRegionsXLS, null, ImmutableSet.of());
  }

  @Test
  public void testNameToColumn() throws SAXException {
    Assert.assertEquals(0, XLSXSpreadsheet.columnAddressToIndex("A"));
    Assert.assertEquals(1, XLSXSpreadsheet.columnAddressToIndex("B"));
    Assert.assertEquals(2, XLSXSpreadsheet.columnAddressToIndex("C"));
    Assert.assertEquals(24, XLSXSpreadsheet.columnAddressToIndex("Y"));
    Assert.assertEquals(25, XLSXSpreadsheet.columnAddressToIndex("Z"));
    Assert.assertEquals(26, XLSXSpreadsheet.columnAddressToIndex("AA"));
    Assert.assertEquals(27, XLSXSpreadsheet.columnAddressToIndex("AB"));
    Assert.assertEquals(28, XLSXSpreadsheet.columnAddressToIndex("AC"));
    Assert.assertEquals(103, XLSXSpreadsheet.columnAddressToIndex("CZ"));
    Assert.assertEquals(241, XLSXSpreadsheet.columnAddressToIndex("IH"));
    Assert.assertEquals(103, XLSXSpreadsheet.columnAddressToIndex("CZ"));
    Assert.assertEquals(700, XLSXSpreadsheet.columnAddressToIndex("ZY"));
    Assert.assertEquals(701, XLSXSpreadsheet.columnAddressToIndex("ZZ"));
    Assert.assertEquals(702, XLSXSpreadsheet.columnAddressToIndex("AAA"));
    Assert.assertEquals(703, XLSXSpreadsheet.columnAddressToIndex("AAB"));
    Assert.assertEquals(704, XLSXSpreadsheet.columnAddressToIndex("AAC"));
    Assert.assertEquals(746, XLSXSpreadsheet.columnAddressToIndex("ABS"));

    Assert.assertEquals(2, XLSXSpreadsheet.cellReferenceToColumnIndex("C4"));
    Assert.assertEquals(241, XLSXSpreadsheet.cellReferenceToColumnIndex("IH4"));
    Assert.assertEquals(746, XLSXSpreadsheet.cellReferenceToColumnIndex("ABS4"));
    Assert.assertEquals(2, XLSXSpreadsheet.cellReferenceToColumnIndex("C42"));
    Assert.assertEquals(241, XLSXSpreadsheet.cellReferenceToColumnIndex("IH42"));
    Assert.assertEquals(746, XLSXSpreadsheet.cellReferenceToColumnIndex("ABS42"));
  }
}
