// The SIEUFERD project by Eirik Bakke 2010-2016, ebakke@alum.mit.edu.
package com.sieuferd.evaluator;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import com.sieuferd.datamodel.CProp;
import com.sieuferd.datamodel.ImmutableView;
import com.sieuferd.datamodel.MutableView;
import com.sieuferd.datamodel.Path;
import com.sieuferd.datamodel.Path.PrimitivePath;
import com.sieuferd.datamodel.Path.RelationPath;
import com.sieuferd.datamodel.PathList;
import com.sieuferd.datamodel.PathMap;
import com.sieuferd.datamodel.SProp;
import com.sieuferd.datamodel.proptypes.Position;
import com.sieuferd.query.OptionalPrimitivePath;
import com.sieuferd.query.QProp;
import java.math.BigDecimal;
import java.time.LocalTime;
import java.util.List;
import java.util.Set;
import org.junit.Test;

public class QueryEvaluatorTest extends EvaluatorTestBase {
  @Test
  public void testCreateEmptyTable() {
    createTable("tab_empty",
      new String[] {},
      new Class<?>[] {},
      new Object[][] {
      }
    );
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_empty");
    expectResult(view, true, "[]");
  }

  @Test
  public void testCreateTableNoErrors() {
    createTable("tab_simple",
      new String[] {"pk_colA", "colB_n", "colC"},
      new Class<?>[] {String.class, String.class, Integer.class},
      new Object[][] {
        new Object[] {"A1", "B1", 1},
        new Object[] {"A2", "B2", 2},
        new Object[] {"A3", "B3", 3},
        new Object[] {"A4", "B4", 4},
        new Object[] {"A5", "B5", 5}
      }
    );
  }

  // TODO[te]: Test tables containing backticks, spaces, backslashes, and question marks.
  @Test
  public void testDetectColumnsAndKeys() {
    createTable("tab1", new String[] {"pk_colA", "colB_n", "colC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    // Different order.
    createTable("tab1B", new String[] {"pk_colA", "colC", "colB_n"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    createTable("tab2", new String[] {"pk_colA", "pk_colB", "colC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    // Different order.
    createTable("tab2B", new String[] {"pk_colB", "pk_colA", "colC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    createTable("tab3", new String[] {"colA", "pk_colB", "colC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    createTable("tab4", new String[] {"colA", "colB_n", "colC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    expectColumnNames("tab1", "pk_colA", "colB_n", "colC");
    expectColumnNames("tab1B", "pk_colA", "colC", "colB_n");
    expectColumnNames("tab2", "pk_colA", "pk_colB", "colC");
    expectColumnNames("tab2B", "pk_colB", "pk_colA", "colC");
    expectColumnNames("tab3", "colA", "pk_colB", "colC");
    expectPKColumnNames("tab1", "pk_colA");
    expectPKColumnNames("tab1B", "pk_colA");
    expectPKColumnNames("tab2", "pk_colA", "pk_colB");
    expectPKColumnNames("tab2B", "pk_colA", "pk_colB");
    expectPKColumnNames("tab3", "pk_colB");
    String pseudoColumn = TEST_STATE.getDialect().getRowIdentifierPseudoColumnName();
    if (!pseudoColumn.isEmpty()) {
      expectColumnNames("tab4", pseudoColumn, "colA", "colB_n", "colC");
      expectPKColumnNames("tab4", pseudoColumn);
    } else {
      expectColumnNames("tab4", "colA", "colB_n", "colC");
      expectPKColumnNames("tab4", "colA", "colB_n", "colC");
    }
  }

  @Test
  public void testWeirdlyNamedTables() {
    createTable("tab", new String[] {"pk_colTA", "colTB_n", "colTC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    createTable("tab1", new String[] {"pk_colA", "colB_n", "colC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    createTable("tab_", new String[] {"pk_colUA", "colUB_n", "colUC"},
        new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
    expectColumnNames("tab", "pk_colTA", "colTB_n", "colTC");
    expectPKColumnNames("tab", "pk_colTA");

    expectColumnNames("tab1", "pk_colA", "colB_n", "colC");
    expectPKColumnNames("tab1", "pk_colA");
    /* The underscore is a wildcard in JDBC metadata functions such as DatabaseMetaData.getColumns,
    and so will match both tab1 and tab_ unless special escaping is applied. */
    expectColumnNames("tab_", "pk_colUA", "colUB_n", "colUC");
    expectPKColumnNames("tab_", "pk_colUA");
    Set<String> oddCharacters = Sets.newLinkedHashSet(ImmutableSet.of(
        "`", "'", "\"", "\\", ",", ".", "_", " ", "%", "?", "[", "]", "\t", "\n"));
    {
      final String identifierQuoteString = TEST_STATE.getDialect().getIdentifierQuoteString();
      if (!identifierQuoteString.isEmpty()) {
        if (TEST_STATE.getDialect().supportsIdentifierQuoteInTableName()) {
          oddCharacters.add(identifierQuoteString);
        } else {
          oddCharacters.remove(identifierQuoteString);
        }
      }
      final String searchStringEscape = TEST_STATE.getDialect().getSearchStringEscape();
      if (!searchStringEscape.isEmpty())
        oddCharacters.add(searchStringEscape);
    }
    int i = 0;
    for (String oddCharacter : oddCharacters) {
      /* Don't let the odd character be at the end of the table name, as it may be stripped in the
      case of a whitespace character. */
      createTable("tab_OD" + oddCharacter + "D",
          new String[] {"pk_col" + i + "A", "col" + i + "B_n", "col" + i + "C"},
          new Class<?>[] {String.class, String.class, Integer.class}, new Object[][] { });
      i++;
    }
    i = 0;
    for (String oddCharacter : oddCharacters) {
      expectColumnNames("tab_OD" + oddCharacter + "D",
          "pk_col" + i + "A", "col" + i + "B_n", "col" + i + "C");
      expectPKColumnNames("tab_OD" + oddCharacter + "D", "pk_col" + i + "A");
      i++;
    }
  }

  @Test
  public void testSimpleSelect() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2,
      "A4", "B4", 4,
      "A3", "B3", 3, // Make sure the order doesn't matter.
      "A5", "B5", 5,
      "A6", null, 6,
      "A7", "B7", 7
    });
    MutableView view = createView();

    expectResult(view, true, "[()]");
    PrimitivePath simpleFormula = addFormulaColumn(view, Path.ROOT, "simpleFormula", "5 * 9");
    expectResult(view, true, "[(45)]");
    view.removePath(simpleFormula);

    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "colB_n");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC");

    expectColumnNames("tab_simple", "pk_colA", "colB_n", "colC");
    expectPKColumnNames("tab_simple", "pk_colA");

    expectResult(view, true,
        "[(A1,B1,1),(A2,B2,2),(A3,B3,3),(A4,B4,4),(A5,B5,5),(A6,_,6),(A7,B7,7)]");
    view.set(CProp.VISIBLE, colA, false);
    expectResult(view, true,
        "[(A1,B1,1),(A2,B2,2),(A3,B3,3),(A4,B4,4),(A5,B5,5),(A6,_,6),(A7,B7,7)]");
    view.set(CProp.VISIBLE, colB, false);
    expectResult(view, true,
        "[(A1,.,1),(A2,.,2),(A3,.,3),(A4,.,4),(A5,.,5),(A6,.,6),(A7,.,7)]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true,
        "[(.,.,1),(.,.,2),(.,.,3),(.,.,4),(.,.,5),(.,.,6),(.,.,7)]");
    view.set(CProp.VISIBLE, colC, false);
    expectResult(view, true, "[(.,.,.)]");
    // Reorder fields.
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.<Path>of(colB, colA, colC)));
    view.set(CProp.VISIBLE, colA, true);
    view.set(CProp.VISIBLE, colB, true);
    view.set(CProp.VISIBLE, colC, false);
    expectResult(view, true,
        "[(B1,A1,.),(B2,A2,.),(B3,A3,.),(B4,A4,.),(B5,A5,.),(B7,A7,.),(_,A6,.)]");
  }

  @Test
  public void testEmptyTable() {
    createTable("tab_empty", new Object[] {
      "pk_colA", "colB_n", "colC_n",
      String.class, String.class, Integer.class,
    });
    createTable("tab_onerow", new Object[] {
      "pk_colA", "colB_n", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1
    });
    createTable("tab_onerow_null", new Object[] {
      "pk_colA", "colB_n", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", null
    });
    createTable("tab_tworows", new Object[] {
      "pk_colA", "colB_n", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2
    });
    createTable("tab_tworows_null", new Object[] {
      "pk_colA", "colB_n", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", null
    });
    MutableView view = createView();
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "colB_n");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC_n");
    view.set(CProp.LABEL, colC, "colC");

    view.set(CProp.VISIBLE, colA, true);
    view.set(CProp.VISIBLE, colB, true);
    view.set(CProp.VISIBLE, colC, false);
    setInstantiatedTable(view, Path.ROOT, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, Path.ROOT, "tab_onerow");
    expectResult(view, true, "[(A1,B1,.)]");
    setInstantiatedTable(view, Path.ROOT, "tab_tworows");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.)]");

    view.set(CProp.VISIBLE, colA, false);
    view.set(CProp.VISIBLE, colB, false);
    view.set(CProp.VISIBLE, colC, false);
    setInstantiatedTable(view, Path.ROOT, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, Path.ROOT, "tab_onerow");
    expectResult(view, true, "[(.,.,.)]");
    setInstantiatedTable(view, Path.ROOT, "tab_tworows");
    expectResult(view, true, "[(.,.,.)]");

    /* Note: Just doing "select max('colC') from tab_empty" would return a single row in SQL; see
             http://infolab.stanford.edu/~ullman/fcdb/oracle/or-nulls.html . The correct behavior in
             our system is as shown here. */
    PrimitivePath sagg = addFormulaColumn(view, Path.ROOT, "fx", "max([colC])");
    setInstantiatedTable(view, Path.ROOT, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, Path.ROOT, "tab_onerow");
    expectResult(view, true, "[(.,.,.,1)]");
    setInstantiatedTable(view, Path.ROOT, "tab_onerow_null");
    expectResult(view, true, "[(.,.,.,_)]");
    setInstantiatedTable(view, Path.ROOT, "tab_tworows");
    expectResult(view, true, "[(.,.,.,2)]");
    setInstantiatedTable(view, Path.ROOT, "tab_tworows_null");
    expectResult(view, true, "[(.,.,.,1)]");
    view.removePath(sagg);


    MutableView view2 = createView();
    RelationPath childPath = view2.addRelationPath(Path.ROOT);
    PathMap view2map = view2.copyFrom(view, Path.ROOT, childPath);
    view2.set(CProp.LABEL, childPath, "child");
    addFormulaColumn(view2, Path.ROOT, "a01", "max([colC])");
    setInstantiatedTable(view2, view2map.getNewPath(Path.ROOT), "tab_empty");
    expectResult(view2, true, "[([],_)]");
    setInstantiatedTable(view2, view2map.getNewPath(Path.ROOT), "tab_onerow");
    expectResult(view2, true, "[([(.,.,.)],1)]");
    setInstantiatedTable(view2, view2map.getNewPath(Path.ROOT), "tab_onerow_null");
    expectResult(view2, true, "[([(.,.,.)],_)]");
    setInstantiatedTable(view2, view2map.getNewPath(Path.ROOT), "tab_tworows");
    expectResult(view2, true, "[([(.,.,.)],2)]");
    setInstantiatedTable(view2, view2map.getNewPath(Path.ROOT), "tab_tworows_null");
    expectResult(view2, true, "[([(.,.,.)],1)]");
  }

  @Test
  public void testUninstantiatedTable() {
    createTable("tab_empty", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
    });
    createTable("tab_onerow", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1
    });
    createTable("tab_tworows", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2
    });
    MutableView view = createView();
    PrimitivePath pk = addSQLColumn(view, Path.ROOT, "pk_colA");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, pk, false);
    PrimitivePath colA = addFormulaColumn(view, Path.ROOT, "formula", "5");
    view.set(CProp.VISIBLE, colA, true);
    setInstantiatedTable(view, Path.ROOT, null);
    expectResult(view, true, "[(.,5)]");
    // Constrast with the following cases, with otherwise the same view:
    setInstantiatedTable(view, Path.ROOT, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, Path.ROOT, "tab_onerow");
    expectResult(view, true, "[(.,5)]");
    setInstantiatedTable(view, Path.ROOT, "tab_tworows");
    expectResult(view, true, "[(.,5)]");
  }

  @Test
  public void testSimpleFormulas() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2,
      "A4", "B4", 4,
      "A3", "B33", 3,
      "A5", "B5", 5,
      "A6", null, 6,
      "A7", "B7", 7
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "colB_n");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC");
    PrimitivePath colD = addFormulaColumn(view, Path.ROOT, "colD", "5");
    PrimitivePath colE = addFormulaColumn(view, Path.ROOT, "colE", "[pk_colA]");
    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "colF", "[colB_n]");
    PrimitivePath colG = addFormulaColumn(view, Path.ROOT, "colG", "concat([colB_n], '$')");
    PrimitivePath colH = addFormulaColumn(view, Path.ROOT, "colH", "concat([colG], '#')");
    PrimitivePath colI = addFormulaColumn(view, Path.ROOT, "colI", "len([colB_n])");
    PrimitivePath colJ = addFormulaColumn(view, Path.ROOT, "colJ", "(2 + 3) * 10");
    PrimitivePath colK = addFormulaColumn(view, Path.ROOT, "colK", "15 / (len([colB_n]) - 2)");
    PrimitivePath colL = addFormulaColumn(view, Path.ROOT, "colK", "0 / 0");
    view.set(CProp.VISIBLE, colB, false);
    expectResult(view, true, "[" +
        "(A1,.,1,5,A1,B1,B1$,B1$#,2,50,_,_)," +
        "(A2,.,2,5,A2,B2,B2$,B2$#,2,50,_,_)," +
        "(A3,.,3,5,A3,B33,B33$,B33$#,3,50,15,_)," +
        "(A4,.,4,5,A4,B4,B4$,B4$#,2,50,_,_)," +
        "(A5,.,5,5,A5,B5,B5$,B5$#,2,50,_,_)," +
        "(A6,.,6,5,A6,_,$,$#,_,50,_,_)," +
        "(A7,.,7,5,A7,B7,B7$,B7$#,2,50,_,_)]");
  }

  @Test
  public void testSimpleErrors() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2,
      "A4", "B4", 4,
      "A3", "B3", 3,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "colB_n");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC");
    PrimitivePath colD = addFormulaColumn(view, Path.ROOT, "colD", "8 + #huh!");
    PrimitivePath colE = addFormulaColumn(view, Path.ROOT, "colE", "[colD]");
    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "colF", "if(false, #huh!, 2)");
    PrimitivePath colG = addFormulaColumn(view, Path.ROOT, "colG", "if(false, null, 2)");
    expectResult(view, true,
        "[(A1,B1,1,_,_,_,2),(A2,B2,2,_,_,_,2),(A3,B3,3,_,_,_,2),(A4,B4,4,_,_,_,2)]",
        ImmutableSet.<Path>of(colD, colE, colF));
  }

  @Test
  public void testCircularReference() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "colB_n");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC");
    PrimitivePath colD = addFormulaColumn(view, Path.ROOT, "colD", "6");
    PrimitivePath colE = addFormulaColumn(view, Path.ROOT, "colE", "[colD] + 2");
    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "colF", "3 + [colE]");
    // No error.
    expectResult(view, true, "[(A1,B1,1,6,8,11),(A2,B2,2,6,8,11)]");
    // Now set up the circular reference.
    setFormula(view, colD, "[colF]");
    expectResult(view, true, "[(A1,B1,1,_,_,_),(A2,B2,2,_,_,_)]",
        ImmutableSet.<Path>of(colD, colE, colF));
    // Also try a circular reference where colD points directly to itself.
    setFormula(view, colD, "[colD]");
    expectResult(view, true, "[(A1,B1,1,_,_,_),(A2,B2,2,_,_,_)]",
        ImmutableSet.<Path>of(colD, colE, colF));
  }

  @Test
  public void testUnboundedInwardReference() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "colB_n",
      String.class, String.class,
      "A1", "B1",
      "A2", "B2",
      "A3", "B3",
      "A4", "B4",
    });
    createTable("tab_empty", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
    });
    createTable("tab_onerow", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "cA1", "cB1", 11
    });
    createTable("tab_tworows", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "cA1", "cB1", 11,
      "cA2", "cB2", 22
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath rootColA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath rootColB = addSQLColumn(view, Path.ROOT, "colB_n");
    expectResult(view, true,
        "[(A1,B1),(A2,B2),(A3,B3),(A4,B4)]");
    RelationPath childPath = addRelation(view, Path.ROOT, "child");
    view.set(CProp.VISIBLE, childPath, false);
    PrimitivePath childColA = addSQLColumn(view, childPath, "pk_colA");
    PrimitivePath childColB = addSQLColumn(view, childPath, "colB_n");
    PrimitivePath childColC = addSQLColumn(view, childPath, "colC");

    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    view.set(QProp.EXISTENTIAL_FILTER, childPath, true);
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, childPath, "tab_onerow");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, "tab_tworows");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");

    view.set(QProp.EXISTENTIAL_FILTER, childPath, true);
    setInstantiatedTable(view, childPath, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, childPath, null);
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, "tab_onerow");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, "tab_tworows");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    view.set(QProp.EXISTENTIAL_FILTER, childPath, false);
    setInstantiatedTable(view, childPath, "tab_empty");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, null);
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, "tab_onerow");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");
    setInstantiatedTable(view, childPath, "tab_tworows");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A4,B4,.)]");

    addFormulaColumn(view, Path.ROOT, "fx1", "[child\\pk_colA]");
    view.set(QProp.EXISTENTIAL_FILTER, childPath, false);
    setInstantiatedTable(view, childPath, "tab_empty");
    expectResult(view, true,
        "[(A1,B1,.,_),(A2,B2,.,_),(A3,B3,.,_),(A4,B4,.,_)]");
    setInstantiatedTable(view, childPath, "tab_onerow");
    expectResult(view, true,
        "[(A1,B1,.,cA1),(A2,B2,.,cA1),(A3,B3,.,cA1),(A4,B4,.,cA1)]");
    setInstantiatedTable(view, childPath, "tab_tworows");
    expectResult(view, true,
        "[(A1,B1,.,cA1),(A1,B1,.,cA2),(A2,B2,.,cA1),(A2,B2,.,cA2),(A3,B3,.,cA1),(A3,B3,.,cA2),(A4,B4,.,cA1),(A4,B4,.,cA2)]");
    view.set(QProp.EXISTENTIAL_FILTER, childPath, true);
    setInstantiatedTable(view, childPath, "tab_empty");
    expectResult(view, true, "[]");
    setInstantiatedTable(view, childPath, "tab_onerow");
    expectResult(view, true,
        "[(A1,B1,.,cA1),(A2,B2,.,cA1),(A3,B3,.,cA1),(A4,B4,.,cA1)]");
    setInstantiatedTable(view, childPath, "tab_tworows");
    expectResult(view, true,
        "[(A1,B1,.,cA1),(A1,B1,.,cA2),(A2,B2,.,cA1),(A2,B2,.,cA2),(A3,B3,.,cA1),(A3,B3,.,cA2),(A4,B4,.,cA1),(A4,B4,.,cA2)]");

    RelationPath noinstChildPath = addRelation(view, Path.ROOT, "noinstchild");
    setInstantiatedTable(view, childPath, "tab_tworows");
    view.set(CProp.VISIBLE, noinstChildPath, false);
    setInstantiatedTable(view, noinstChildPath, null);
    view.set(QProp.EXISTENTIAL_FILTER, noinstChildPath, false);
    expectResult(view, true,
        "[(A1,B1,.,cA1,.),(A1,B1,.,cA2,.),(A2,B2,.,cA1,.),(A2,B2,.,cA2,.),(A3,B3,.,cA1,.),(A3,B3,.,cA2,.),(A4,B4,.,cA1,.),(A4,B4,.,cA2,.)]");
    view.set(QProp.EXISTENTIAL_FILTER, noinstChildPath, true);
    expectResult(view, true,
        "[(A1,B1,.,cA1,.),(A1,B1,.,cA2,.),(A2,B2,.,cA1,.),(A2,B2,.,cA2,.),(A3,B3,.,cA1,.),(A3,B3,.,cA2,.),(A4,B4,.,cA1,.),(A4,B4,.,cA2,.)]");
    PrimitivePath noinstFormula = addFormulaColumn(view, noinstChildPath, "nif", "4 * 3");
    PrimitivePath rootFormula = addFormulaColumn(view, Path.ROOT, "nifref", "[nif]");
    view.set(QProp.EXISTENTIAL_FILTER, noinstChildPath, false);
    expectResult(view, true,
        "[(A1,B1,.,cA1,.,12),(A1,B1,.,cA2,.,12),(A2,B2,.,cA1,.,12),(A2,B2,.,cA2,.,12),(A3,B3,.,cA1,.,12),(A3,B3,.,cA2,.,12),(A4,B4,.,cA1,.,12),(A4,B4,.,cA2,.,12)]");
    view.set(QProp.EXISTENTIAL_FILTER, noinstChildPath, true);
    expectResult(view, true,
        "[(A1,B1,.,cA1,.,12),(A1,B1,.,cA2,.,12),(A2,B2,.,cA1,.,12),(A2,B2,.,cA2,.,12),(A3,B3,.,cA1,.,12),(A3,B3,.,cA2,.,12),(A4,B4,.,cA1,.,12),(A4,B4,.,cA2,.,12)]");

    RelationPath inst2ChildPath = addRelation(view, Path.ROOT, "inst2child");
    view.set(CProp.VISIBLE, inst2ChildPath, false);
    setInstantiatedTable(view, inst2ChildPath, "tab_empty");
    PrimitivePath child2ColA = addSQLColumn(view, inst2ChildPath, "pk_colA");
    PrimitivePath child2ColB = addSQLColumn(view, inst2ChildPath, "colB_n");
    PrimitivePath child2ColC = addSQLColumn(view, inst2ChildPath, "colC");
    view.set(QProp.EXISTENTIAL_FILTER, inst2ChildPath, false);
    expectResult(view, true,
        "[(A1,B1,.,cA1,.,12,.),(A1,B1,.,cA2,.,12,.),(A2,B2,.,cA1,.,12,.),(A2,B2,.,cA2,.,12,.),(A3,B3,.,cA1,.,12,.),(A3,B3,.,cA2,.,12,.),(A4,B4,.,cA1,.,12,.),(A4,B4,.,cA2,.,12,.)]");
    view.set(QProp.EXISTENTIAL_FILTER, inst2ChildPath, true);
    expectResult(view, true, "[]");
    for (String tableName : ImmutableSet.of("tab_onerow", "tab_tworows")) {
      setInstantiatedTable(view, inst2ChildPath, tableName);
      for (Boolean existential : ImmutableSet.of(false, true)) {
        view.set(QProp.EXISTENTIAL_FILTER, inst2ChildPath, existential);
        expectResult(view, true,
            "[(A1,B1,.,cA1,.,12,.),(A1,B1,.,cA2,.,12,.),(A2,B2,.,cA1,.,12,.),(A2,B2,.,cA2,.,12,.),(A3,B3,.,cA1,.,12,.),(A3,B3,.,cA2,.,12,.),(A4,B4,.,cA1,.,12,.),(A4,B4,.,cA2,.,12,.)]");
      }
    }
    setFormula(view, rootFormula, "[nif] + [inst2child\\colC]");
    view.set(QProp.EXISTENTIAL_FILTER, childPath, false);
    view.set(QProp.EXISTENTIAL_FILTER, noinstChildPath, false);
    view.set(QProp.EXISTENTIAL_FILTER, inst2ChildPath, false);
    setInstantiatedTable(view, childPath, "tab_tworows");
    setInstantiatedTable(view, inst2ChildPath, "tab_onerow");
    expectResult(view, true,
        "[(A1,B1,.,cA1,.,23,.),(A1,B1,.,cA2,.,23,.),(A2,B2,.,cA1,.,23,.),(A2,B2,.,cA2,.,23,.),(A3,B3,.,cA1,.,23,.),(A3,B3,.,cA2,.,23,.),(A4,B4,.,cA1,.,23,.),(A4,B4,.,cA2,.,23,.)]");
    setInstantiatedTable(view, inst2ChildPath, "tab_tworows");
    expectResult(view, true, "[" +
        "(A1,B1,.,cA1,.,23,.),(A1,B1,.,cA1,.,34,.),(A1,B1,.,cA2,.,23,.),(A1,B1,.,cA2,.,34,.),(A2,B2,.,cA1,.,23,.),(A2,B2,.,cA1,.,34,.),(A2,B2,.,cA2,.,23,.),(A2,B2,.,cA2,.,34,.)," +
        "(A3,B3,.,cA1,.,23,.),(A3,B3,.,cA1,.,34,.),(A3,B3,.,cA2,.,23,.),(A3,B3,.,cA2,.,34,.),(A4,B4,.,cA1,.,23,.),(A4,B4,.,cA1,.,34,.),(A4,B4,.,cA2,.,23,.),(A4,B4,.,cA2,.,34,.)]");
    view.set(QProp.EXISTENTIAL_FILTER, inst2ChildPath, true);
    expectResult(view, true, "[" +
        "(A1,B1,.,cA1,.,23,.),(A1,B1,.,cA1,.,34,.),(A1,B1,.,cA2,.,23,.),(A1,B1,.,cA2,.,34,.),(A2,B2,.,cA1,.,23,.),(A2,B2,.,cA1,.,34,.),(A2,B2,.,cA2,.,23,.),(A2,B2,.,cA2,.,34,.)," +
        "(A3,B3,.,cA1,.,23,.),(A3,B3,.,cA1,.,34,.),(A3,B3,.,cA2,.,23,.),(A3,B3,.,cA2,.,34,.),(A4,B4,.,cA1,.,23,.),(A4,B4,.,cA1,.,34,.),(A4,B4,.,cA2,.,23,.),(A4,B4,.,cA2,.,34,.)]");
    RelationPath grandChildPath = addRelation(view, inst2ChildPath, "grandChild");
    PrimitivePath grandChildColA = addSQLColumn(view, grandChildPath, "pk_colA");
    PrimitivePath grandChildColB = addSQLColumn(view, grandChildPath, "colB_n");
    PrimitivePath grandChildColC = addSQLColumn(view, grandChildPath, "colC");
    view.set(CProp.VISIBLE, grandChildPath, false);
    view.set(QProp.EXISTENTIAL_FILTER, grandChildPath, true);
    setInstantiatedTable(view, grandChildPath, "tab_empty");
    expectResult(view, true, "[]");
  }

  @Test
  public void testLocalFilters() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "pk_colB", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 1,
      "A3", "B3", 1,
      "A4", "B4", 2,
      "A5", "B5", 2,
      "A6", "B6", null,
      "A7", "B7", 6
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "pk_colB");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC_n");
    expectResult(view, true,
        "[(A1,B1,1),(A2,B2,1),(A3,B3,1),(A4,B4,2),(A5,B5,2),(A6,B6,_),(A7,B7,6)]");
    for (Path fp : ImmutableList.of(colC, Path.ROOT)) {
      setFilter(view, colC, "true");
      setFilter(view, Path.ROOT, "true");

      setFilter(view, fp, "[colC_n] is 1");
      expectResult(view, true, "[(A1,B1,1),(A2,B2,1),(A3,B3,1)]");
      setFilter(view, fp, "[colC_n] is 1 or [colC_n] is 6");
      expectResult(view, true, "[(A1,B1,1),(A2,B2,1),(A3,B3,1),(A7,B7,6)]");
      setFilter(view, fp, "not ([colC_n] is 1)");
      expectResult(view, true, "[(A4,B4,2),(A5,B5,2),(A6,B6,_),(A7,B7,6)]");
      setFilter(view, fp, "not ([colC_n] is 1 or [colC_n] is null)");
      expectResult(view, true, "[(A4,B4,2),(A5,B5,2),(A7,B7,6)]");
      setFilter(view, fp, "not ([colC_n] is 1 or [colC_n] is 6)");
      expectResult(view, true, "[(A4,B4,2),(A5,B5,2),(A6,B6,_)]");
    }

    setFilter(view, colC, "true");
    setFilter(view, Path.ROOT, "true");
    view.set(CProp.VISIBLE, colC, false);
    setFilter(view, colC, "[colC_n] is 1");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.)]");
    setFilter(view, colC, "[colC_n] is 1 or [colC_n] is 6");
    expectResult(view, true, "[(A1,B1,.),(A2,B2,.),(A3,B3,.),(A7,B7,.)]");
    setFilter(view, colC, "not ([colC_n] is 1)");
    expectResult(view, true, "[(A4,B4,.),(A5,B5,.),(A6,B6,.),(A7,B7,.)]");
    setFilter(view, colC, "not ([colC_n] is 1 or [colC_n] is 6)");
    expectResult(view, true, "[(A4,B4,.),(A5,B5,.),(A6,B6,.)]");

    view.set(CProp.VISIBLE, colC, true);
    setFilter(view, colC, "true");
    setFilter(view, Path.ROOT, "[pk_colA] is 'A1' and [pk_colB] is 'B1'");
    expectResult(view, true, "[(A1,B1,1)]");
    setFilter(view, Path.ROOT,
        "[pk_colA] is 'A1' and [pk_colB] is 'B1' or [pk_colA] is 'A2' and [pk_colB] is 'B2'");
    expectResult(view, true, "[(A1,B1,1),(A2,B2,1)]");
    setFilter(view, Path.ROOT,
        "not ([pk_colA] is 'A1' and [pk_colB] is 'B1' or [pk_colA] is 'A2' and [pk_colB] is 'B2')");
    expectResult(view, true, "[(A3,B3,1),(A4,B4,2),(A5,B5,2),(A6,B6,_),(A7,B7,6)]");
    setFilter(view, colC, "[colC_n] is 2");
    expectResult(view, true, "[(A4,B4,2),(A5,B5,2)]");

    setFilter(view, colC, "true");
    setFilter(view, Path.ROOT, "true");

    createTable("tab_tworows", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "cA1", "cB1", 111,
      "cA2", "cB2", 222
    });

    RelationPath childPath = addRelation(view, Path.ROOT, "child");
    view.set(CProp.VISIBLE, childPath, false);
    setInstantiatedTable(view, childPath, "tab_tworows");
    PrimitivePath childColA = addSQLColumn(view, childPath, "pk_colA");
    PrimitivePath childColB = addSQLColumn(view, childPath, "colB_n");
    PrimitivePath childColC = addSQLColumn(view, childPath, "colC");
    expectResult(view, true,
        "[(A1,B1,1,.),(A2,B2,1,.),(A3,B3,1,.),(A4,B4,2,.),(A5,B5,2,.),(A6,B6,_,.),(A7,B7,6,.)]");
    setFilter(view, childPath,
        "[pk_colA] is 'cA1' and [colB_n] is 'cB1' or [pk_colA] is 'cA2' and [colB_n] is 'cB2'");
    expectResult(view, true,
        "[(A1,B1,1,.),(A2,B2,1,.),(A3,B3,1,.),(A4,B4,2,.),(A5,B5,2,.),(A6,B6,_,.),(A7,B7,6,.)]");
    setFilter(view, childPath,
        "[pk_colA] is 'cA1' and [colB_n] is 'cB1'");
    expectResult(view, true,
        "[(A1,B1,1,.),(A2,B2,1,.),(A3,B3,1,.),(A4,B4,2,.),(A5,B5,2,.),(A6,B6,_,.),(A7,B7,6,.)]");
    setFilter(view, childPath, "false");
    expectResult(view, true,
        "[(A1,B1,1,.),(A2,B2,1,.),(A3,B3,1,.),(A4,B4,2,.),(A5,B5,2,.),(A6,B6,_,.),(A7,B7,6,.)]");

    view.set(QProp.EXISTENTIAL_FILTER, childPath, true);
    setFilter(view, childPath, "false");
    expectResult(view, true, "[]");
    setFilter(view, childPath,
        "not ([pk_colA] is 'cA1' and [colB_n] is 'cB1' or [pk_colA] is 'cA2' and [colB_n] is 'cB2')");
    expectResult(view, true, "[]");
    setFilter(view, childPath,
        "[pk_colA] is 'cA1' and [colB_n] is 'cB1'");
    expectResult(view, true,
        "[(A1,B1,1,.),(A2,B2,1,.),(A3,B3,1,.),(A4,B4,2,.),(A5,B5,2,.),(A6,B6,_,.),(A7,B7,6,.)]");
    setFilter(view, childColC, "[colC] is 111");
    expectResult(view, true,
        "[(A1,B1,1,.),(A2,B2,1,.),(A3,B3,1,.),(A4,B4,2,.),(A5,B5,2,.),(A6,B6,_,.),(A7,B7,6,.)]");
    setFilter(view, childColC, "[colC] is 123");
    expectResult(view, true, "[]");

    PrimitivePath fxPath = addFormulaColumn(view, Path.ROOT, "fx", "[child\\pk_colA]");
    setFilter(view, childColC, "[colC] is 111");
    expectResult(view, true,
        "[(A1,B1,1,.,cA1),(A2,B2,1,.,cA1),(A3,B3,1,.,cA1),(A4,B4,2,.,cA1),(A5,B5,2,.,cA1),(A6,B6,_,.,cA1),(A7,B7,6,.,cA1)]");
    setFilter(view, childPath, "true");
    setFilter(view, childColC, "[colC] is 222");
    expectResult(view, true,
        "[(A1,B1,1,.,cA2),(A2,B2,1,.,cA2),(A3,B3,1,.,cA2),(A4,B4,2,.,cA2),(A5,B5,2,.,cA2),(A6,B6,_,.,cA2),(A7,B7,6,.,cA2)]");
    setFilter(view, childColC, "[colC] is 123");
    expectResult(view, true,
        "[]");
    view.set(QProp.EXISTENTIAL_FILTER, childPath, false);
    expectResult(view, true,
        "[(A1,B1,1,.,_),(A2,B2,1,.,_),(A3,B3,1,.,_),(A4,B4,2,.,_),(A5,B5,2,.,_),(A6,B6,_,.,_),(A7,B7,6,.,_)]");
  }

  @Test
  public void testHiddenFilterNotProjected() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "pk_colB", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 1,
      "A3", "B3", 1,
      "A4", "B4", 2,
      "A5", "B5", 2,
      "A6", "B6", null,
      "A7", "B7", 6
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "pk_colB");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC_n");
    expectResult(view, true,
        "[(A1,B1,1),(A2,B2,1),(A3,B3,1),(A4,B4,2),(A5,B5,2),(A6,B6,_),(A7,B7,6)]");
    view.set(CProp.VISIBLE, colA, false);
    view.set(CProp.VISIBLE, colB, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true, "[(.,.,1),(.,.,2),(.,.,6),(.,.,_)]");
    setFilter(view, colB, "not ([pk_colB] is 'B4')");
    expectResult(view, true, "[(.,.,1),(.,.,2),(.,.,6),(.,.,_)]");
    setFilter(view, colB, "not ([pk_colB] is 'B4' or [pk_colB] is 'B5')");
    expectResult(view, true, "[(.,.,1),(.,.,6),(.,.,_)]");
  }

  @Test
  public void testFilterCompoundConstants() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "pk_colB", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", -1,
      "A3", "B3", -1,
      "A4", "B4", 2,
      "A5", "B5", 2,
      "A6", "B6", null,
      "A7", "B7", 6
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colB = addSQLColumn(view, Path.ROOT, "pk_colB");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colC_n");
    expectResult(view, true,
        "[(A1,B1,1),(A2,B2,-1),(A3,B3,-1),(A4,B4,2),(A5,B5,2),(A6,B6,_),(A7,B7,6)]");
    setFilter(view, colC, "[colC_n] is -1");
    expectResult(view, true,
        "[(A2,B2,-1),(A3,B3,-1)]");
    setFilter(view, colC, "[colC_n] is -1 or [colC_n] is 2");
    expectResult(view, true,
        "[(A2,B2,-1),(A3,B3,-1),(A4,B4,2),(A5,B5,2)]");
    setFilter(view, colC, "not([colC_n] is -1 or [colC_n] is 2)");
    expectResult(view, true,
        "[(A1,B1,1),(A6,B6,_),(A7,B7,6)]");
  }

  @Test
  public void testSingleLevelOutwardReferences() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "Y1",
      "X2", "Y1",
      "X3", "Y1",
      "X4", "Y2",
      "X5", "Y2",
      "X6", null,
      "X7", "Y3",
      "X8", "Y4"
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n",
      String.class, String.class,
      "Y1", "Y1d",
      "Y2", "Y2d",
      "Y3", "Y3d",
      "Y5", "Y5d"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    expectResult(view, true,
        "[(X1,Y1),(X2,Y1),(X3,Y1),(X4,Y2),(X5,Y2),(X6,_),(X7,Y3),(X8,Y4)]");
    RelationPath tabYPath = addRelation(view, Path.ROOT, "tabY");
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    expectResult(view, true,
        "[(X1,Y1,.),(X2,Y1,.),(X3,Y1,.),(X4,Y2,.),(X5,Y2,.),(X6,_,.),(X7,Y3,.),(X8,Y4,.)]");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");
    // Cartesian product, before constraints are added.
    String joinResult1 = "[" +
        "(X1,Y1,.,Y1,Y1d),(X1,Y1,.,Y2,Y2d),(X1,Y1,.,Y3,Y3d),(X1,Y1,.,Y5,Y5d)," +
        "(X2,Y1,.,Y1,Y1d),(X2,Y1,.,Y2,Y2d),(X2,Y1,.,Y3,Y3d),(X2,Y1,.,Y5,Y5d)," +
        "(X3,Y1,.,Y1,Y1d),(X3,Y1,.,Y2,Y2d),(X3,Y1,.,Y3,Y3d),(X3,Y1,.,Y5,Y5d)," +
        "(X4,Y2,.,Y1,Y1d),(X4,Y2,.,Y2,Y2d),(X4,Y2,.,Y3,Y3d),(X4,Y2,.,Y5,Y5d)," +
        "(X5,Y2,.,Y1,Y1d),(X5,Y2,.,Y2,Y2d),(X5,Y2,.,Y3,Y3d),(X5,Y2,.,Y5,Y5d)," +
        "(X6,_,.,Y1,Y1d),(X6,_,.,Y2,Y2d),(X6,_,.,Y3,Y3d),(X6,_,.,Y5,Y5d)," +
        "(X7,Y3,.,Y1,Y1d),(X7,Y3,.,Y2,Y2d),(X7,Y3,.,Y3,Y3d),(X7,Y3,.,Y5,Y5d)," +
        "(X8,Y4,.,Y1,Y1d),(X8,Y4,.,Y2,Y2d),(X8,Y4,.,Y3,Y3d),(X8,Y4,.,Y5,Y5d)" +
        "]";
    expectResult(view, true, joinResult1);

    PrimitivePath colYfXA = addFormulaColumn(view, tabYPath, "YfXA", "[pk_colXA]");
    expectResult(view, true, joinResult1);

    view.set(QProp.JOINED_ON, colYA, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d)," +
        "(X2,Y1,.,Y1,Y1d)," +
        "(X3,Y1,.,Y1,Y1d)," +
        "(X4,Y2,.,Y2,Y2d)," +
        "(X5,Y2,.,Y2,Y2d)," +
        "(X6,_,.,_,_)," +
        "(X7,Y3,.,Y3,Y3d)," +
        "(X8,Y4,.,_,_)" +
        "]");

    // Now do an inward reference to the outward reference.
    PrimitivePath fYC = addFormulaColumn(view, Path.ROOT, "fYC", "concat('$', [YfXA])");
    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d,$X1)," +
        "(X2,Y1,.,Y1,Y1d,$X2)," +
        "(X3,Y1,.,Y1,Y1d,$X3)," +
        "(X4,Y2,.,Y2,Y2d,$X4)," +
        "(X5,Y2,.,Y2,Y2d,$X5)," +
        "(X6,_,.,_,_,$)," + // concat should produce a value even if an input is null
        "(X7,Y3,.,Y3,Y3d,$X7)," +
        "(X8,Y4,.,_,_,$)" + // See above.
        "]");

    setFormula(view, fYC, "[YfXA]");
    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d,X1)," +
        "(X2,Y1,.,Y1,Y1d,X2)," +
        "(X3,Y1,.,Y1,Y1d,X3)," +
        "(X4,Y2,.,Y2,Y2d,X4)," +
        "(X5,Y2,.,Y2,Y2d,X5)," +
        "(X6,_,.,_,_,_)," + // Note the null in the last column here; this is correct behavior.
        "(X7,Y3,.,Y3,Y3d,X7)," +
        "(X8,Y4,.,_,_,_)" + // As above.
        "]");

    view.set(CProp.VISIBLE, colfYA, false);
    view.set(CProp.VISIBLE, colfYB, false);

    String joinResult2 = "[" +
        "(X1,Y1,.,.,.,X1)," +
        "(X2,Y1,.,.,.,X2)," +
        "(X3,Y1,.,.,.,X3)," +
        "(X4,Y2,.,.,.,X4)," +
        "(X5,Y2,.,.,.,X5)," +
        "(X6,_,.,.,.,_)," +
        "(X7,Y3,.,.,.,X7)," +
        "(X8,Y4,.,.,.,_)" +
        "]";
    expectResult(view, true, joinResult2);
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabYPath, false);
    expectResult(view, true, joinResult2);

    view.set(CProp.VISIBLE, colXB, false);
    expectResult(view, true, "[" +
        "(X1,.,.,.,.,X1)," +
        "(X2,.,.,.,.,X2)," +
        "(X3,.,.,.,.,X3)," +
        "(X4,.,.,.,.,X4)," +
        "(X5,.,.,.,.,X5)," +
        "(X6,.,.,.,.,_)," +
        "(X7,.,.,.,.,X7)," +
        "(X8,.,.,.,.,_)" +
        "]");

    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, true);
    view.set(CProp.VISIBLE, colfYA, true);
    view.set(CProp.VISIBLE, colfYB, true);
    view.set(CProp.VISIBLE, colXB, true);

    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d,X1)," +
        "(X2,Y1,.,Y1,Y1d,X2)," +
        "(X3,Y1,.,Y1,Y1d,X3)," +
        "(X4,Y2,.,Y2,Y2d,X4)," +
        "(X5,Y2,.,Y2,Y2d,X5)," +
        "(X7,Y3,.,Y3,Y3d,X7)" +
        "]");

    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, false);
    setFilter(view, colYB, "[colYB_n] is 'Y1d' or [colYB_n] is 'Y3d'");
    String joinResult3 = "[" +
        "(X1,Y1,.,Y1,Y1d,X1)," +
        "(X2,Y1,.,Y1,Y1d,X2)," +
        "(X3,Y1,.,Y1,Y1d,X3)," +
        "(X4,Y2,.,_,_,_)," +
        "(X5,Y2,.,_,_,_)," +
        "(X6,_,.,_,_,_)," +
        "(X7,Y3,.,Y3,Y3d,X7)," +
        "(X8,Y4,.,_,_,_)" +
        "]";
    expectResult(view, true, joinResult3);

    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, true);
    String joinResult4 = "[" +
        "(X1,Y1,.,Y1,Y1d,X1)," +
        "(X2,Y1,.,Y1,Y1d,X2)," +
        "(X3,Y1,.,Y1,Y1d,X3)," +
        "(X7,Y3,.,Y3,Y3d,X7)" +
        "]";
    expectResult(view, true, joinResult4);

    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, false);
    // Equivalent join using a formula and a filter.
    PrimitivePath colYfXB = addFormulaColumn(view, tabYPath, "YfJoin", "[colXB_n] = [pk_colYA]");
    setFilter(view, colYfXB, "[YfJoin] is true");
    expectResult(view, true, joinResult3);
    view.set(QProp.JOINED_ON, colYA, OptionalPrimitivePath.absent());
    expectResult(view, true, joinResult3);
    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, true);
    expectResult(view, true, joinResult4);
  }

  @Test
  public void testInwardOutwardCircularReference() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "Y1",
      "X2", "Y1",
      "X3", "Y1",
      "X4", "Y2",
      "X5", "Y2",
      "X6", null,
      "X7", "Y3",
      "X8", "Y4"
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n",
      String.class, String.class,
      "Y1", "Y1d",
      "Y2", "Y2d",
      "Y3", "Y3d",
      "Y5", "Y5d"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabYPath = addRelation(view, Path.ROOT, "tabY");
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");
    PrimitivePath colYf = addFormulaColumn(view, tabYPath, "colYf", "5");
    PrimitivePath colXf = addFormulaColumn(view, Path.ROOT, "colXf", "[colYf]");

    view.set(QProp.JOINED_ON, colYA, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d,5)," +
        "(X2,Y1,.,Y1,Y1d,5)," +
        "(X3,Y1,.,Y1,Y1d,5)," +
        "(X4,Y2,.,Y2,Y2d,5)," +
        "(X5,Y2,.,Y2,Y2d,5)," +
        "(X6,_,.,_,_,_)," +
        "(X7,Y3,.,Y3,Y3d,5)," +
        "(X8,Y4,.,_,_,_)" +
        "]");

    setFormula(view, colYf, "[colXf]");
    // (*pk_colXA, colXB_n, .tabY(pk_colYA=[colXB_n], colYB_n, colYf:[colXf]), fYA:[pk_colYA], fYB:[colYB_n], colXf:[colYf])
    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d,_)," +
        "(X2,Y1,.,Y1,Y1d,_)," +
        "(X3,Y1,.,Y1,Y1d,_)," +
        "(X4,Y2,.,Y2,Y2d,_)," +
        "(X5,Y2,.,Y2,Y2d,_)," +
        "(X6,_,.,_,_,_)," +
        "(X7,Y3,.,Y3,Y3d,_)," +
        "(X8,Y4,.,_,_,_)" +
        "]", ImmutableSet.<Path>of(colXf, colYf));

    // Not errors.
    setFormula(view, colYf, "[colYB_n]");

    expectResult(view, true, "[" +
        "(X1,Y1,.,Y1,Y1d,Y1d)," +
        "(X2,Y1,.,Y1,Y1d,Y1d)," +
        "(X3,Y1,.,Y1,Y1d,Y1d)," +
        "(X4,Y2,.,Y2,Y2d,Y2d)," +
        "(X5,Y2,.,Y2,Y2d,Y2d)," +
        "(X6,_,.,_,_,_)," +
        "(X7,Y3,.,Y3,Y3d,Y3d)," +
        "(X8,Y4,.,_,_,_)" +
        "]");

    setFilter(view, colYf, "[colYf] is 'Y2d'");
    expectResult(view, true, "[" +
        "(X1,Y1,.,_,_,_)," +
        "(X2,Y1,.,_,_,_)," +
        "(X3,Y1,.,_,_,_)," +
        "(X4,Y2,.,Y2,Y2d,Y2d)," +
        "(X5,Y2,.,Y2,Y2d,Y2d)," +
        "(X6,_,.,_,_,_)," +
        "(X7,Y3,.,_,_,_)," +
        "(X8,Y4,.,_,_,_)" +
        "]");

    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, true);
    expectResult(view, true, "[" +
        "(X4,Y2,.,Y2,Y2d,Y2d)," +
        "(X5,Y2,.,Y2,Y2d,Y2d)" +
        "]");
  }

  @Test
  public void testManyToManyJoin() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "b",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });
    createTable("tabY2", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "c",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabYPath = addRelation(view, Path.ROOT, "tabY");
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colYC = addSQLColumn(view, tabYPath, "colYC_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");

    // root(pk_colXA, colXB_n, tabY(pk_colYA, colYB_n))

    view.set(QProp.JOINED_ON, colYC, OptionalPrimitivePath.of(colXB));
    // (pk_colXA, colXB_n, .tabY(pk_colYA, colYB_n, colYC_n=[colXB_n]), fYA:[pk_colYA], fYB:[colYB_n])
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d)," +
        "(X2,a,.,Y1,Y1d)," +
        "(X3,a,.,Y1,Y1d)," +
        "(X4,b,.,Y2,Y2d)," +
        "(X5,b,.,Y2,Y2d)," +
        "(X6,_,.,_,_)," +
        "(X7,c,.,Y3,Y3d)," +
        "(X8,d,.,_,_)," +
        "(X9,c,.,Y3,Y3d)" +
        "]");
    setInstantiatedTable(view, tabYPath, "tabY2");
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d)," +
        "(X2,a,.,Y1,Y1d)," +
        "(X3,a,.,Y1,Y1d)," +
        "(X4,b,.,_,_)," +
        "(X5,b,.,_,_)," +
        "(X6,_,.,_,_)," +
        "(X7,c,.,Y2,Y2d)," +
        "(X7,c,.,Y3,Y3d)," +
        "(X8,d,.,_,_)," +
        "(X9,c,.,Y2,Y2d)," +
        "(X9,c,.,Y3,Y3d)" +
        "]");

    /* Take the opportunity to test an antijoin-like query by filtering to include only null values
    produced by the left join. */
    setFilter(view, colfYB, "[fYB] is null");
    expectResult(view, true, "[" +
        "(X4,b,.,_,_)," +
        "(X5,b,.,_,_)," +
        "(X6,_,.,_,_)," +
        "(X8,d,.,_,_)" +
        "]");
    setFilter(view, colfYB, "true");

    /* This test also highlights the difference in placing an otherwise equal formula at two
    different levels. */
    addFormulaColumn(view, tabYPath, "tabYf", "concat([colYB_n], '$', [pk_colXA])");
    addFormulaColumn(view, Path.ROOT, "tabXf", "[tabYf]");
    addFormulaColumn(view, Path.ROOT, "tabXf2", "concat([colYB_n], '$', [pk_colXA])");
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d,Y1d$X1,Y1d$X1)," +
        "(X2,a,.,Y1,Y1d,Y1d$X2,Y1d$X2)," +
        "(X3,a,.,Y1,Y1d,Y1d$X3,Y1d$X3)," +
        "(X4,b,.,_,_,_,$X4)," +
        "(X5,b,.,_,_,_,$X5)," +
        "(X6,_,.,_,_,_,$X6)," +
        "(X7,c,.,Y2,Y2d,Y2d$X7,Y2d$X7)," +
        "(X7,c,.,Y3,Y3d,Y3d$X7,Y3d$X7)," +
        "(X8,d,.,_,_,_,$X8)," +
        "(X9,c,.,Y2,Y2d,Y2d$X9,Y2d$X9)," +
        "(X9,c,.,Y3,Y3d,Y3d$X9,Y3d$X9)" +
        "]");
  }

  @Test
  public void testDeepJoin() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabX2", new Object[] {
      "pk_colX2A", "colX2B_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "c",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });
    createTable("tabY2", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "y1", "y1d", "a",
      "y2", "y2d", "b",
      "y3", "y3d", "c",
      "y5", "y5d", "e"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath between = addRelation(view, Path.ROOT, "between");

    setInstantiatedTable(view, between, "tabX2");
    PrimitivePath colBetwA = addSQLColumn(view, between, "pk_colX2A");
    PrimitivePath colBetwB = addSQLColumn(view, between, "colX2B_n");
    view.set(QProp.JOINED_ON, colBetwA, OptionalPrimitivePath.of(colXA));

    RelationPath tabYPath = addRelation(view, between, "tabY");
    view.set(CProp.VISIBLE, between, false);
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colYC = addSQLColumn(view, tabYPath, "colYC_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");

    // First test regular child-to-parent joins on two levels.
    view.set(QProp.JOINED_ON, colYC, OptionalPrimitivePath.of(colBetwB));
    String joinResult1 = "[" +
        "(X1,a,.,Y1,Y1d)," +
        "(X2,a,.,Y1,Y1d)," +
        "(X3,a,.,Y1,Y1d)," +
        "(X4,b,.,_,_)," +
        "(X5,b,.,_,_)," +
        "(X6,_,.,_,_)," +
        "(X7,c,.,Y2,Y2d)," +
        "(X7,c,.,Y3,Y3d)," +
        "(X8,d,.,_,_)," +
        "(X9,c,.,Y2,Y2d)," +
        "(X9,c,.,Y3,Y3d)" +
        "]";
    expectResult(view, true, joinResult1);
    /* Now join on the outer relation instead, which should give the same result (since root and
    between both contain the same data and are joined on their primary keys). */
    view.set(QProp.JOINED_ON, colYC, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, joinResult1);

    // Now try with the middle relation not instantiating a table.
    setInstantiatedTable(view, between, null);
    /* This will give an error since we still have the SQL columns registered. And since there is a
    join condition on one of them, the inner relation ends up empty. */
    // This segfaults Oracle under the [CORRELATION_JOIN_OPTIMIZATION] setting.
    expectResult(view, true,
        "[(X1,a,.,_,_),(X2,a,.,_,_),(X3,a,.,_,_),(X4,b,.,_,_),(X5,b,.,_,_),(X6,_,.,_,_),(X7,c,.,_,_),(X8,d,.,_,_),(X9,c,.,_,_)]",
        ImmutableSet.<Path>of(colBetwA));
    view.removePath(colBetwA);
    view.removePath(colBetwB);
    // Now we should get the same result as before.
    expectResult(view, true, joinResult1);

    view.set(QProp.EXISTENTIAL_FILTER, between, true);
    // The result should be the same, since the "between" relation always has exactly one row.
    expectResult(view, true, joinResult1);
    view.set(QProp.EXISTENTIAL_FILTER, between, false);

    // Now add a second relation under [between], also joining with the root.
    RelationPath tabY2Path = addRelation(view, between, "tabY2");
    view.set(CProp.VISIBLE, tabY2Path, false);
    setInstantiatedTable(view, tabY2Path, "tabY");
    PrimitivePath colY2A = addSQLColumn(view, tabY2Path, "pk_colYA");
    PrimitivePath colY2B = addSQLColumn(view, tabY2Path, "colYB_n");
    PrimitivePath colY2C = addSQLColumn(view, tabY2Path, "colYC_n");
    PrimitivePath colfY2A = addFormulaColumn(view, Path.ROOT, "fY2A", "[tabY2\\pk_colYA]");
    PrimitivePath colfY2B = addFormulaColumn(view, Path.ROOT, "fY2B", "[tabY2\\colYB_n]");
    view.set(QProp.JOINED_ON, colY2C, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d,Y1,Y1d)," +
        "(X2,a,.,Y1,Y1d,Y1,Y1d)," +
        "(X3,a,.,Y1,Y1d,Y1,Y1d)," +
        "(X4,b,.,_,_,_,_)," +
        "(X5,b,.,_,_,_,_)," +
        "(X6,_,.,_,_,_,_)," +
        "(X7,c,.,Y2,Y2d,Y2,Y2d)," +
        "(X7,c,.,Y2,Y2d,Y3,Y3d)," +
        "(X7,c,.,Y3,Y3d,Y2,Y2d)," +
        "(X7,c,.,Y3,Y3d,Y3,Y3d)," +
        "(X8,d,.,_,_,_,_)," +
        "(X9,c,.,Y2,Y2d,Y2,Y2d)," +
        "(X9,c,.,Y2,Y2d,Y3,Y3d)," +
        "(X9,c,.,Y3,Y3d,Y2,Y2d)," +
        "(X9,c,.,Y3,Y3d,Y3,Y3d)" +
        "]");
    setInstantiatedTable(view, tabY2Path, "tabY2");
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d,y1,y1d)," +
        "(X2,a,.,Y1,Y1d,y1,y1d)," +
        "(X3,a,.,Y1,Y1d,y1,y1d)," +
        "(X4,b,.,_,_,y2,y2d)," + // Note: A result of (X4,b,.,_,_,_,_) here would indicate a bug.
        "(X5,b,.,_,_,y2,y2d)," + // Note: A result of (X5,b,.,_,_,_,_) here would indicate a bug.
        "(X6,_,.,_,_,_,_)," +
        "(X7,c,.,Y2,Y2d,y3,y3d)," +
        "(X7,c,.,Y3,Y3d,y3,y3d)," +
        "(X8,d,.,_,_,_,_)," +
        "(X9,c,.,Y2,Y2d,y3,y3d)," +
        "(X9,c,.,Y3,Y3d,y3,y3d)" +
        "]");

    view.set(QProp.EXISTENTIAL_FILTER, tabYPath, true);
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d,y1,y1d)," +
        "(X2,a,.,Y1,Y1d,y1,y1d)," +
        "(X3,a,.,Y1,Y1d,y1,y1d)," +
        "(X4,b,.,_,_,_,_)," + // Now with more nulls.
        "(X5,b,.,_,_,_,_)," + // Now with more nulls.
        "(X6,_,.,_,_,_,_)," +
        "(X7,c,.,Y2,Y2d,y3,y3d)," +
        "(X7,c,.,Y3,Y3d,y3,y3d)," +
        "(X8,d,.,_,_,_,_)," +
        "(X9,c,.,Y2,Y2d,y3,y3d)," +
        "(X9,c,.,Y3,Y3d,y3,y3d)" +
        "]");
  }

  @Test
  public void testLateralReferences() {
    Object tabXcontent[] = new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    };
    createTable("tabX", tabXcontent);
    createTable("tabX2", tabXcontent);
    /* For completeness, ensure that the join on tabX.pk_colXA = tabZ.pk_colZB includes matches with
    cardinality zero, one, and two in each direction:

      Z7 matches {}
      Z1 matches {X1}
      Z6 matches {X6,X9}

      X4 matches {}
      X1 matches {Z1}
      X3 matches {Z3,Z4}

    */
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB",
      String.class, String.class,
      "Z1", "X1",
      "Z2", "X2",
      "Z3", "X3",
      "Z4", "X3",
      "Z5", "X5",
      "Z6", "X6",
      "Z6", "X9",
      "Z7", "XX"
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "c",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });
    createTable("tabY2", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "y1", "y1d", "a",
      "y2", "y2d", "b",
      "y3", "y3d", "c",
      "y5", "y5d", "e"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath between = addRelation(view, Path.ROOT, "between");

    setInstantiatedTable(view, between, "tabX2");
    PrimitivePath colBetwA = addSQLColumn(view, between, "pk_colXA");
    PrimitivePath colBetwB = addSQLColumn(view, between, "colXB_n");
    view.set(QProp.JOINED_ON, colBetwA, OptionalPrimitivePath.of(colXA));

    RelationPath tabYPath = addRelation(view, between, "tabY");
    view.set(CProp.VISIBLE, between, false);
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colYC = addSQLColumn(view, tabYPath, "colYC_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");

    view.set(QProp.JOINED_ON, colYC, OptionalPrimitivePath.of(colXB));

    RelationPath tabY2Path = addRelation(view, between, "tabY2");
    view.set(CProp.VISIBLE, tabY2Path, false);
    setInstantiatedTable(view, tabY2Path, "tabY2");
    PrimitivePath colY2A = addSQLColumn(view, tabY2Path, "pk_colYA");
    PrimitivePath colY2B = addSQLColumn(view, tabY2Path, "colYB_n");
    PrimitivePath colY2C = addSQLColumn(view, tabY2Path, "colYC_n");
    PrimitivePath colfY2A = addFormulaColumn(view, Path.ROOT, "fY2A", "[tabY2\\pk_colYA]");
    PrimitivePath colfY2B = addFormulaColumn(view, Path.ROOT, "fY2B", "[tabY2\\colYB_n]");
    view.set(QProp.JOINED_ON, colY2C, OptionalPrimitivePath.of(colXB));
    // First, just the scenario from testDeepJoin.
    expectResult(view, true, "[" +
        "(X1,a,.,Y1,Y1d,y1,y1d)," +
        "(X2,a,.,Y1,Y1d,y1,y1d)," +
        "(X3,a,.,Y1,Y1d,y1,y1d)," +
        "(X4,b,.,_,_,y2,y2d)," +
        "(X5,b,.,_,_,y2,y2d)," +
        "(X6,_,.,_,_,_,_)," +
        "(X7,c,.,Y2,Y2d,y3,y3d)," +
        "(X7,c,.,Y3,Y3d,y3,y3d)," +
        "(X8,d,.,_,_,_,_)," +
        "(X9,c,.,Y2,Y2d,y3,y3d)," +
        "(X9,c,.,Y3,Y3d,y3,y3d)" +
        "]");

    ImmutableView originalView = ImmutableView.copyOf(view);

    /* Add another join at the root level, with an inward reference. Try both with an extra empty
    relation between the new join and the root and not. */
    for (final Boolean useBetweenZ : ImmutableList.of(false, true)) {
      view.copyFrom(originalView);
      RelationPath betweenZ = useBetweenZ ? addRelation(view, Path.ROOT, "betweenZ") : null;
      if (betweenZ != null)
        view.set(CProp.VISIBLE, betweenZ, false);

      RelationPath tabZ = addRelation(view, betweenZ == null ? Path.ROOT : betweenZ, "tabZ");
      view.set(CProp.VISIBLE, tabZ, false);
      setInstantiatedTable(view, tabZ, "tabZ");
      PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
      PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
      view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));
      addFormulaColumn(view, Path.ROOT, "RfZB", "[tabZ\\pk_colZA]");
      expectResult(view, true, "[" +
          "(X1,a,.,Y1,Y1d,y1,y1d,.,Z1)," +
          "(X2,a,.,Y1,Y1d,y1,y1d,.,Z2)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z3)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z4)," +
          "(X4,b,.,_,_,y2,y2d,.,_)," +
          "(X5,b,.,_,_,y2,y2d,.,Z5)," +
          "(X6,_,.,_,_,_,_,.,Z6)," +
          "(X7,c,.,Y2,Y2d,y3,y3d,.,_)," +
          "(X7,c,.,Y3,Y3d,y3,y3d,.,_)," +
          "(X8,d,.,_,_,_,_,.,_)," +
          "(X9,c,.,Y2,Y2d,y3,y3d,.,Z6)," +
          "(X9,c,.,Y3,Y3d,y3,y3d,.,Z6)" +
          "]");

      // Add a deep outward reference to the inward reference.
      PrimitivePath zout =
          addFormulaColumn(view, tabY2Path, "zout", "concat([tabY2\\colYB_n], '-', [RfZB])");
      PrimitivePath Rzout = addFormulaColumn(view, Path.ROOT, "Rzout", "[tabY2\\zout]");

      String lateralResult = "[" +
          "(X1,a,.,Y1,Y1d,y1,y1d,.,Z1,y1d-Z1)," +
          "(X2,a,.,Y1,Y1d,y1,y1d,.,Z2,y1d-Z2)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z3,y1d-Z3)," +
          // The two tuples below should probably be decorrelated away.
          //"(X3,a,.,Y1,Y1d,y1,y1d,.,Z3,y1d-Z4)," +
          //"(X3,a,.,Y1,Y1d,y1,y1d,.,Z4,y1d-Z3)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z4,y1d-Z4)," +
          "(X4,b,.,_,_,y2,y2d,.,_,y2d-)," +
          "(X5,b,.,_,_,y2,y2d,.,Z5,y2d-Z5)," +
          "(X6,_,.,_,_,_,_,.,Z6,_)," +
          "(X7,c,.,Y2,Y2d,y3,y3d,.,_,y3d-)," +
          "(X7,c,.,Y3,Y3d,y3,y3d,.,_,y3d-)," +
          "(X8,d,.,_,_,_,_,.,_,_)," +
          "(X9,c,.,Y2,Y2d,y3,y3d,.,Z6,y3d-Z6)," +
          "(X9,c,.,Y3,Y3d,y3,y3d,.,Z6,y3d-Z6)" +
          "]";
      expectResult(view, true, lateralResult);

      // This time, do a direct lateral reference, without going through an inward reference.
      setFormula(view, zout, "concat([tabY2\\colYB_n], '-', [tabZ\\pk_colZA])");
      expectResult(view, true, "[" +
          "(X1,a,.,Y1,Y1d,y1,y1d,.,Z1,y1d-Z1)," +
          "(X2,a,.,Y1,Y1d,y1,y1d,.,Z2,y1d-Z2)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z3,y1d-Z3)," +
          // TODO[pd]: Should these two tuples now appear, or not?
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z3,y1d-Z4)," + // Suspicious tuple.
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z4,y1d-Z3)," + // Suspicious tuple.
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z4,y1d-Z4)," +
          "(X4,b,.,_,_,y2,y2d,.,_,y2d-)," +
          "(X5,b,.,_,_,y2,y2d,.,Z5,y2d-Z5)," +
          "(X6,_,.,_,_,_,_,.,Z6,_)," +
          "(X7,c,.,Y2,Y2d,y3,y3d,.,_,y3d-)," +
          "(X7,c,.,Y3,Y3d,y3,y3d,.,_,y3d-)," +
          "(X8,d,.,_,_,_,_,.,_,_)," +
          "(X9,c,.,Y2,Y2d,y3,y3d,.,Z6,y3d-Z6)," +
          "(X9,c,.,Y3,Y3d,y3,y3d,.,Z6,y3d-Z6)" +
          "]");

      setFilter(view, tabZ, "false");
      expectResult(view, true, "[" +
          "(X1,a,.,Y1,Y1d,y1,y1d,.,_,y1d-)," +
          "(X2,a,.,Y1,Y1d,y1,y1d,.,_,y1d-)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,_,y1d-)," +
          "(X4,b,.,_,_,y2,y2d,.,_,y2d-)," +
          "(X5,b,.,_,_,y2,y2d,.,_,y2d-)," +
          "(X6,_,.,_,_,_,_,.,_,_)," +
          "(X7,c,.,Y2,Y2d,y3,y3d,.,_,y3d-)," +
          "(X7,c,.,Y3,Y3d,y3,y3d,.,_,y3d-)," +
          "(X8,d,.,_,_,_,_,.,_,_)," +
          "(X9,c,.,Y2,Y2d,y3,y3d,.,_,y3d-)," +
          "(X9,c,.,Y3,Y3d,y3,y3d,.,_,y3d-)" +
          "]");
      view.set(QProp.EXISTENTIAL_FILTER, tabZ, true);
      if (betweenZ != null)
        view.set(QProp.EXISTENTIAL_FILTER, betweenZ, true);
      expectResult(view, true, "[]");
      setFilter(view, tabZ, "[pk_colZA] is 'Z5'");
      expectResult(view, true, "[(X5,b,.,_,_,y2,y2d,.,Z5,y2d-Z5)]");
      view.set(QProp.EXISTENTIAL_FILTER, tabZ, false);
      if (betweenZ != null)
        view.set(QProp.EXISTENTIAL_FILTER, betweenZ, false);

      expectResult(view, true, "[" +
          "(X1,a,.,Y1,Y1d,y1,y1d,.,_,y1d-)," +
          "(X2,a,.,Y1,Y1d,y1,y1d,.,_,y1d-)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,_,y1d-)," +
          "(X4,b,.,_,_,y2,y2d,.,_,y2d-)," +
          "(X5,b,.,_,_,y2,y2d,.,Z5,y2d-Z5)," +
          "(X6,_,.,_,_,_,_,.,_,_)," +
          "(X7,c,.,Y2,Y2d,y3,y3d,.,_,y3d-)," +
          "(X7,c,.,Y3,Y3d,y3,y3d,.,_,y3d-)," +
          "(X8,d,.,_,_,_,_,.,_,_)," +
          "(X9,c,.,Y2,Y2d,y3,y3d,.,_,y3d-)," +
          "(X9,c,.,Y3,Y3d,y3,y3d,.,_,y3d-)" +
          "]");
      setFilter(view, tabZ, "true");

      PrimitivePath tabZcirc = addFormulaColumn(view, tabZ, "tabZcirc", "[zout]");
      setFormula(view, zout, "[tabZcirc]");
      expectResult(view, true, "[" +
          "(X1,a,.,Y1,Y1d,y1,y1d,.,Z1,_)," +
          "(X2,a,.,Y1,Y1d,y1,y1d,.,Z2,_)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z3,_)," +
          "(X3,a,.,Y1,Y1d,y1,y1d,.,Z4,_)," +
          "(X4,b,.,_,_,y2,y2d,.,_,_)," +
          "(X5,b,.,_,_,y2,y2d,.,Z5,_)," +
          "(X6,_,.,_,_,_,_,.,Z6,_)," +
          "(X7,c,.,Y2,Y2d,y3,y3d,.,_,_)," +
          "(X7,c,.,Y3,Y3d,y3,y3d,.,_,_)," +
          "(X8,d,.,_,_,_,_,.,_,_)," +
          "(X9,c,.,Y2,Y2d,y3,y3d,.,Z6,_)," +
          "(X9,c,.,Y3,Y3d,y3,y3d,.,Z6,_)" +
          "]", ImmutableSet.<Path>of(Rzout, zout, tabZcirc));
    }
  }

  @Test
  public void testInstancesChangedBug() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d","a",
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath between = addRelation(view, Path.ROOT, "between");

    RelationPath tabYPath = addRelation(view, between, "tabY");
    view.set(CProp.VISIBLE, between, false);
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colYC = addSQLColumn(view, tabYPath, "colYC_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");

    view.set(QProp.JOINED_ON, colYC, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, "[(X1,a,.,Y1,Y1d)]");
  }

  @Test
  public void testJoinedOnNonexistent() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "Y1",
      "X2", "Y1",
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n",
      String.class, String.class,
      "Y1", "Y1d",
      "Y2", "Y2d",
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabYPath = addRelation(view, Path.ROOT, "tabY");
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");
    PrimitivePath colYf = addFormulaColumn(view, tabYPath, "colYf", "5");
    PrimitivePath colXf = addFormulaColumn(view, Path.ROOT, "colXf", "[colYf]");

    view.set(QProp.JOINED_ON, colYA, OptionalPrimitivePath.of(PrimitivePath.create(Path.ROOT)));
    expectResult(view, true, "[" +
        "(X1,Y1,.,_,_,_)," +
        "(X2,Y1,.,_,_,_)" +
        "]", ImmutableSet.<Path>of(colYA));
  }

  @Test
  public void testSingleRelationAggregate() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB", "colXC", "colXD",
      String.class, String.class, Integer.class, Integer.class,
      "X1","A",   1,  1,
      "X2","A",   8,  8,
      "X3","A",  10,  8,
      "X4","A",   0,  8,
      "X5","B", 100,  8,
      "X6","B", 130, 10,
      "X7","C",5712,120,
      "X8","D",6721,140,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath pk_colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB");
    PrimitivePath colXC = addSQLColumn(view, Path.ROOT, "colXC");
    PrimitivePath colXD = addSQLColumn(view, Path.ROOT, "colXD");
    PrimitivePath colXfE = addFormulaColumn(view, Path.ROOT, "colXfE", "sum([colXC])");
    expectResult(view, true, "[" +
        "(X1,A,1,1,1)," +
        "(X2,A,8,8,8)," +
        "(X3,A,10,8,10)," +
        "(X4,A,0,8,0)," +
        "(X5,B,100,8,100)," +
        "(X6,B,130,10,130)," +
        "(X7,C,5712,120,5712)," +
        "(X8,D,6721,140,6721)" +
        "]");
    view.set(CProp.VISIBLE, pk_colXA, false);
    view.set(CProp.VISIBLE, colXC, false);
    view.set(CProp.VISIBLE, colXD, false);
    expectResult(view, true, "[" +
        "(X1,A,.,.,1)," +
        "(X2,A,.,.,8)," +
        "(X3,A,.,.,10)," +
        "(X4,A,.,.,0)," +
        "(X5,B,.,.,100)," +
        "(X6,B,.,.,130)," +
        "(X7,C,.,.,5712)," +
        "(X8,D,.,.,6721)" +
        "]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,19)," +
        "(.,B,.,.,230)," +
        "(.,C,.,.,5712)," +
        "(.,D,.,.,6721)" +
        "]");
    PrimitivePath colXfF = addFormulaColumn(view, Path.ROOT, "colXfF", "sum([colXD])");
    view.set(CProp.VISIBLE, colXfF, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,19,.)," +
        "(.,B,.,.,230,.)," +
        "(.,C,.,.,5712,.)," +
        "(.,D,.,.,6721,.)" +
        "]");
    view.set(CProp.VISIBLE, colXfE, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,.,.)," +
        "(.,B,.,.,.,.)," +
        "(.,C,.,.,.,.)," +
        "(.,D,.,.,.,.)" +
        "]");
    view.set(CProp.VISIBLE, colXfF, true);
    /* Since primary keys are not being projected, the 8's are going to get squashed into a single
    value. */
    expectResult(view, true, "[" +
        "(.,A,.,.,.,9)," +
        "(.,B,.,.,.,18)," +
        "(.,C,.,.,.,120)," +
        "(.,D,.,.,.,140)" +
        "]");
    view.set(CProp.VISIBLE, colXfE, true);
    expectResult(view, true, "[" +
        "(.,A,.,.,19,9)," +
        "(.,B,.,.,230,18)," +
        "(.,C,.,.,5712,120)," +
        "(.,D,.,.,6721,140)" +
        "]");
    PrimitivePath colXfG = addFormulaColumn(view, Path.ROOT, "colXfG", "sum([colXD])");
    expectResult(view, true, "[" +
        "(.,A,.,.,19,9,9)," +
        "(.,B,.,.,230,18,18)," +
        "(.,C,.,.,5712,120,120)," +
        "(.,D,.,.,6721,140,140)" +
        "]");
    PrimitivePath colXfH = addFormulaColumn(view, Path.ROOT, "colXfH", "sum([colXD]) - 1");
    expectResult(view, true, "[" +
        "(.,A,.,.,19,9,9,8)," +
        "(.,B,.,.,230,18,18,17)," +
        "(.,C,.,.,5712,120,120,119)," +
        "(.,D,.,.,6721,140,140,139)" +
        "]");
    PrimitivePath colXfI = addFormulaColumn(
        view, Path.ROOT, "colXfI", "sum([colXD]) - 1 + sum([colXC])");
    String colXfIAddedRes = "[" +
        "(.,A,.,.,19,9,9,8,27)," +
        "(.,B,.,.,230,18,18,17,247)," +
        "(.,C,.,.,5712,120,120,119,5831)," +
        "(.,D,.,.,6721,140,140,139,6860)" +
        "]";
    expectResult(view, true, colXfIAddedRes);
    setFormula(view, colXfI, "[colXfF] - 1 + sum([colXC])");
    expectResult(view, true, colXfIAddedRes);
    view.set(CProp.VISIBLE, colXfE, false);
    view.set(CProp.VISIBLE, colXfF, false);
    view.set(CProp.VISIBLE, colXfG, false);
    view.set(CProp.VISIBLE, colXfH, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,.,.,.,.,27)," +
        "(.,B,.,.,.,.,.,.,247)," +
        "(.,C,.,.,.,.,.,.,5831)," +
        "(.,D,.,.,.,.,.,.,6860)" +
        "]");
    setFilter(view, colXfI, "[colXfI] is 5831");
    expectResultWrapped(view, true, "[" +
        "(.,C,.,.,.,.,.,.,5831)" +
        "]");
    setFilter(view, colXfI, "true");
    PrimitivePath colXfJ = addFormulaColumn(view, Path.ROOT, "colXfJ", "avg([colXD])");
    expectResult(view, true, "[" +
        "(.,A,.,.,.,.,.,.,27,4.5)," +
        "(.,B,.,.,.,.,.,.,247,9)," +
        "(.,C,.,.,.,.,.,.,5831,120)," +
        "(.,D,.,.,.,.,.,.,6860,140)" +
        "]");

    view.set(CProp.VISIBLE, colXfE, true);
    setFilter(view, Path.ROOT, "not ([colXB] is 'B')");
    expectResult(view, true, "[" +
        "(.,A,.,.,19,.,.,.,27,4.5)," +
        "(.,C,.,.,5712,.,.,.,5831,120)," +
        "(.,D,.,.,6721,.,.,.,6860,140)" +
        "]");
    setFilter(view, colXfJ, "[colXfJ] is 4.5 or [colXfJ] is 120");
    expectResult(view, true, "[" +
        "(.,A,.,.,19,.,.,.,27,4.5)," +
        "(.,C,.,.,5712,.,.,.,5831,120)" +
        "]");
    setFilter(view, Path.ROOT, "not ([colXC] is 8)");
    setFilter(view, colXfJ, "true");
    expectResult(view, true, "[" +
        "(.,A,.,.,11,.,.,.,19,4.5)," + // Under old (bad) semantics: (.,A,.,.,19,.,.,.,27,4.5)
        "(.,B,.,.,230,.,.,.,247,9)," +
        "(.,C,.,.,5712,.,.,.,5831,120)," +
        "(.,D,.,.,6721,.,.,.,6860,140)" +
        "]");
    /* While not affecting aggregate inputs, these filters end up filtering out some result rows,
    since all tuples matching the removed groups are filtered out. */
    setFilter(view, Path.ROOT, "not ([colXC] is 5712)");
    expectResult(view, true, "[" +
        "(.,A,.,.,19,.,.,.,27,4.5)," +
        "(.,B,.,.,230,.,.,.,247,9)," +
        "(.,D,.,.,6721,.,.,.,6860,140)" +
        "]");
    setFilter(view, Path.ROOT, "[colXC] is 5712");
    expectResult(view, true, "[" +
        "(.,C,.,.,5712,.,.,.,5831,120)" +
        "]");
    setFilter(view, Path.ROOT, "not ([colXC] is 100 or [colXC] is 130)");
    expectResult(view, true, "[" +
        "(.,A,.,.,19,.,.,.,27,4.5)," +
        "(.,C,.,.,5712,.,.,.,5831,120)," +
        "(.,D,.,.,6721,.,.,.,6860,140)" +
        "]");

    view.set(CProp.VISIBLE, colXfE, false);
    setFilter(view, Path.ROOT, "true");
    setFilter(view, colXfJ, "true");

    view.set(CProp.VISIBLE, colXfJ, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,.,.,.,.,27,.)," +
        "(.,B,.,.,.,.,.,.,247,.)," +
        "(.,C,.,.,.,.,.,.,5831,.)," +
        "(.,D,.,.,.,.,.,.,6860,.)" +
        "]");
    PrimitivePath colXfK = addFormulaColumn(view, Path.ROOT, "colXfK", "1 / avg([colXD])");
    PrimitivePath colXfL = addFormulaColumn(view, Path.ROOT, "colXfL", "avg(1 / [colXD])");
    expectResult(view, true, "[" +
        "(.,A,.,.,.,.,.,.,27,.,0.222222,0.5625)," +
        "(.,B,.,.,.,.,.,.,247,.,0.111111,0.1125)," +
        "(.,C,.,.,.,.,.,.,5831,.,0.008333,0.008333)," +
        "(.,D,.,.,.,.,.,.,6860,.,0.007143,0.007143)" +
        "]");
  }

  @Test
  public void testSingleRelationAggregateWithFilter() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB", "colXC", "colXD",
      String.class, String.class, Integer.class, Integer.class,
      "X1","A",   1,  1,
      "X2","A",   8,  8,
      "X3","A",  10,  8,
      "X4","A",   0,  8,
      "X5","B", 100,  8,
      "X6","B", 130, 10,
      "X7","C",5712,120,
      "X8","D",6721,140,
    });


    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    view.set(CProp.VISIBLE, colX_XA, false);
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB");
    PrimitivePath colX_sumXC = addFormulaColumn(view, Path.ROOT, "sumXC", "null");
    PrimitivePath colX_sumXD = addFormulaColumn(view, Path.ROOT, "sumXD", "null");
    PrimitivePath colX_XC = addSQLColumn(view, Path.ROOT, "colXC");
    view.set(CProp.VISIBLE, colX_XC, false);
    PrimitivePath colX_XD = addSQLColumn(view, Path.ROOT, "colXD");
    view.set(CProp.VISIBLE, colX_XD, false);
    setFormula(view, colX_sumXC, "sum([colXC])");
    setFormula(view, colX_sumXD, "sum([colXD])");
    expectResult(view, true, "["
        + "(.,A,19,9,.,.),"
        + "(.,B,230,18,.,.),"
        + "(.,C,5712,120,.,.),"
        + "(.,D,6721,140,.,.)"
        + "]");
    setFilter(view, colX_XD, "not ([colXD] is 1)");
    expectResult(view, true, "["
        + "(.,A,18,8,.,.),"
        + "(.,B,230,18,.,.),"
        + "(.,C,5712,120,.,.),"
        + "(.,D,6721,140,.,.)"
        + "]");
    setFilter(view, colX_XD, "not ([colXD] is 1 or [colXD] is 8)");
    expectResult(view, true, "["
        + "(.,B,130,10,.,.),"
        + "(.,C,5712,120,.,.),"
        + "(.,D,6721,140,.,.)"
        + "]");
    setFilter(view, colX_XC, "not ([colXB] is 'C')");
    expectResult(view, true, "["
        + "(.,B,130,10,.,.),"
        + "(.,D,6721,140,.,.)"
        + "]");
  }

  @Test
  public void testAggregateSiblingReference() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB", "colXC", "colXD",
      String.class, String.class, Integer.class, Integer.class,
      "X1","A",   1,  1,
      "X2","A",   8,  8,
      "X3","A",  10,  8,
      "X4","A",   0,  8,
      "X5","B", 100,  8,
      "X6","B", 130, 10,
      "X7","C",5712,120,
      "X8","D",6721,140,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath pk_colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB");
    PrimitivePath colXC = addSQLColumn(view, Path.ROOT, "colXC");
    PrimitivePath colXD = addSQLColumn(view, Path.ROOT, "colXD");
    PrimitivePath colXfE = addFormulaColumn(view, Path.ROOT, "colXfE",
        "sum([colXC]) + [colXD]");
    expectResult(view, true, "[" +
        "(X1,A,1,1,2)," +
        "(X2,A,8,8,16)," +
        "(X3,A,10,8,18)," +
        "(X4,A,0,8,8)," +
        "(X5,B,100,8,108)," +
        "(X6,B,130,10,140)," +
        "(X7,C,5712,120,5832)," +
        "(X8,D,6721,140,6861)" +
        "]");
    view.set(CProp.VISIBLE, pk_colXA, false);
    view.set(CProp.VISIBLE, colXC, false);
    view.set(CProp.VISIBLE, colXD, true);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);

    expectResult(view, true, "[" +
        "(.,A,.,1,2)," +
        "(.,A,.,8,26)," +
        "(.,B,.,10,140)," +
        "(.,B,.,8,108)," +
        "(.,C,.,120,5832)," +
        "(.,D,.,140,6861)" +
        "]");

    /* Formula with reference to legal aggregate formula, where the original aggregate formula is
    either projected or not. */
    PrimitivePath refF1 = addFormulaColumn(view, Path.ROOT, "refF", "[colXfE]");
    expectResult(view, true, "[" +
        "(.,A,.,1,2,2)," +
        "(.,A,.,8,26,26)," +
        "(.,B,.,10,140,140)," +
        "(.,B,.,8,108,108)," +
        "(.,C,.,120,5832,5832)," +
        "(.,D,.,140,6861,6861)" +
        "]");
    view.set(CProp.VISIBLE, colXfE, false);
    expectResult(view, true, "[" +
        "(.,A,.,1,.,2)," +
        "(.,A,.,8,.,26)," +
        "(.,B,.,10,.,140)," +
        "(.,B,.,8,.,108)," +
        "(.,C,.,120,.,5832)," +
        "(.,D,.,140,.,6861)" +
        "]");
    view.removePath(refF1);
    view.set(CProp.VISIBLE, colXfE, true);

    /* Aggregate formula referenced colXD from outside the aggregate function; once colXD is
    projected out, the aggregate formula would become part of the key, which is not allowed. */
    view.set(CProp.VISIBLE, colXD, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,_)," +
        "(.,B,.,.,_)," +
        "(.,C,.,.,_)," +
        "(.,D,.,.,_)" +
        "]", ImmutableSet.<Path>of(colXfE));

    // Formula with reference to illegal aggregate formula should also yield warning.
    PrimitivePath refF2 = addFormulaColumn(view, Path.ROOT, "refF", "[colXfE]");
    expectResult(view, true, "[" +
        "(.,A,.,.,_,_)," +
        "(.,B,.,.,_,_)," +
        "(.,C,.,.,_,_)," +
        "(.,D,.,.,_,_)" +
        "]", ImmutableSet.<Path>of(colXfE, refF2));
    // Now hide the original aggregate formula, leaving only the reference to it.
    view.set(CProp.VISIBLE, colXfE, false);
    expectResult(view, true, "[" +
        "(.,A,.,.,.,_)," +
        "(.,B,.,.,.,_)," +
        "(.,C,.,.,.,_)," +
        "(.,D,.,.,.,_)" +
        "]", ImmutableSet.<Path>of(refF2));
  }

  @Test
  public void testInwardReferenceToFilteredRelation() {
    createTable("tab1", new Object[] {
      "pk_colX", "colY",
      String.class, String.class,
      "X1", "Y1",
      "X2", "Y2",
    });
    createTable("tab2", new Object[] {
      "pk_colA", "colB", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 1,
      "A3", "B3", 1,
      "A4", "B4", 2,
      "A5", "B5", 2,
      "A6", "B6", null,
      "A7", "B7", 6
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tab1");
    PrimitivePath col1X = addSQLColumn(view, Path.ROOT, "pk_colX");
    PrimitivePath col1Y = addSQLColumn(view, Path.ROOT, "colY");

    RelationPath tab2 = addRelation(view, Path.ROOT, "tab2");
    setInstantiatedTable(view, tab2, "tab2");
    PrimitivePath col2A = addSQLColumn(view, tab2, "pk_colA");
    PrimitivePath col2B = addSQLColumn(view, tab2, "colB");
    PrimitivePath col2C = addSQLColumn(view, tab2, "colC_n");

    PrimitivePath fA = addFormulaColumn(view, Path.ROOT, "fA", "[pk_colA]");
    view.set(CProp.VISIBLE, tab2, false);
    expectResult(view, true, "[" +
        "(X1,Y1,.,A1)," +
        "(X1,Y1,.,A2)," +
        "(X1,Y1,.,A3)," +
        "(X1,Y1,.,A4)," +
        "(X1,Y1,.,A5)," +
        "(X1,Y1,.,A6)," +
        "(X1,Y1,.,A7)," +
        "(X2,Y2,.,A1)," +
        "(X2,Y2,.,A2)," +
        "(X2,Y2,.,A3)," +
        "(X2,Y2,.,A4)," +
        "(X2,Y2,.,A5)," +
        "(X2,Y2,.,A6)," +
        "(X2,Y2,.,A7)" +
        "]");
    setFilter(view, col2C, "[colC_n] is 1");
    expectResult(view, true, "[" +
        "(X1,Y1,.,A1)," +
        "(X1,Y1,.,A2)," +
        "(X1,Y1,.,A3)," +
        "(X2,Y2,.,A1)," +
        "(X2,Y2,.,A2)," +
        "(X2,Y2,.,A3)" +
        "]");
  }

  @Test
  public void testOutwardReferenceToFilteredRelation() {
    createTable("tab2", new Object[] {
      "pk_colA", "colB", "colC_n",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 1,
      "A3", "B3", 1,
      "A4", "B4", 2,
      "A5", "B5", 2,
      "A6", "B6", null,
      "A7", "B7", 6
    });

    MutableView view = createView();

    RelationPath rel2 = addRelation(view, Path.ROOT, "rel2");
    setInstantiatedTable(view, rel2, "tab2");
    PrimitivePath col2A = addSQLColumn(view, rel2, "pk_colA");
    PrimitivePath col2B = addSQLColumn(view, rel2, "colB");
    PrimitivePath col2C = addSQLColumn(view, rel2, "colC_n");
    view.set(CProp.VISIBLE, rel2, false);

    RelationPath rel3 = addRelation(view, rel2, "rel3");
    addFormulaColumn(view, rel3, "rel3f", "[colB]");
    addFormulaColumn(view, Path.ROOT, "rootf", "[rel3f]");
    expectResult(view, true, "[(.,B1),(.,B2),(.,B3),(.,B4),(.,B5),(.,B6),(.,B7)]");
    setFilter(view, col2C, "[colC_n] is 1");
    expectResult(view, true, "[(.,B1),(.,B2),(.,B3)]");
    setFilter(view, col2C, "[colC_n] is null");
    expectResult(view, true, "[(.,B6)]");
  }

  @Test
  public void testAggregateOnJoinedChildRelation() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2, // Duplicate value "2"; needs to be properly identified by the full tuple key.
      "Z10", "X6", 0, // Zero should increase COUNT but not SUM.
      /* NULL should increase neither COUNT([colZC_n]) nor SUM([colZC_n]), but will increase
      COUNT([const1]). */
      "Z11", "X6", null,
      "Z12", "X7", 0, // Zero should increase COUNT but not SUM.
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");
    PrimitivePath const1 = addFormulaColumn(view, tabZ, "const1", "1");

    view.set(CProp.VISIBLE, tabZ, false);
    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "fs", "sum([colZC_n])");

    /* This test assumes the correct "modified" behavior of the SUM function, which is to return
    zero instead of NULL when doing SUM over an empty input. */
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,3)," +
        "(X3,a,.,28)," +
        "(X4,b,.,0)," +
        "(X5,b,.,100)," +
        "(X6,_,.,1000)," +
        "(X7,c,.,0)," +
        "(X8,d,.,0)," +
        "(X9,c,.,10000)" +
        "]");
    PrimitivePath colG = addFormulaColumn(view, Path.ROOT, "fc", "count([colZC_n])");
    PrimitivePath colH = addFormulaColumn(view, Path.ROOT, "fc2", "count([const1])");
    expectResult(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,28,4,4)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,1000,2,3)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "]");
    setFilter(view, colZB, "not ([pk_colZA] is 'Z6')");
    expectResult(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,28,4,4)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,0,1,2)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,0,0,0)" +
        "]");
    setFilter(view, colZB, "true");
    setFilter(view, colZC, "not ([colZC_n] is 14)");
    expectResultWrapped(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,14,3,3)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,1000,2,3)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "]");
    setFilter(view, colZC, "true");

    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colXA, false);
    expectResult(view, true, "[" +
        "(.,_,.,1000,2,3)," +
        "(.,a,.,32,6,6)," +
        "(.,b,.,100,1,1)," +
        "(.,c,.,10000,2,2)," +
        "(.,d,.,0,0,0)" +
        "]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, true);
    view.set(CProp.VISIBLE, colXA, true);

    view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);
    view.set(CProp.VISIBLE, colZA, false);
    view.set(CProp.VISIBLE, colZB, false);
    view.set(CProp.VISIBLE, colZC, true);
    expectResult(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,26,3,3)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +

        "(X6,_,.,1000,2,3)," +

        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "]");
    view.set(CProp.VISIBLE, colZC, false);
    expectResult(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,26,3,1)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,1000,2,1)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "]");
    view.set(CProp.VISIBLE, colZA, true);
    expectResult(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,28,4,4)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,1000,2,3)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "]");
    setFilter(view, colF, "[fs] is 100 or [fs] is 1000");
    expectResult(view, true, "[" +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,1000,2,3)" +
        "]");
  }

  @Test
  public void testSumAndCountOfMultipleNulls() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", null,
      "Z4", "X3", null,
      "Z5", "X5", 100,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", null,
      "Z9", "X3", null,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");
    PrimitivePath const1 = addFormulaColumn(view, tabZ, "const1", "1");

    view.set(CProp.VISIBLE, tabZ, false);
    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "fs", "sum([colZC_n])");
    PrimitivePath colG = addFormulaColumn(view, Path.ROOT, "fc", "count([colZC_n])");
    PrimitivePath colH = addFormulaColumn(view, Path.ROOT, "fc2", "count([const1])");
    expectResult(view, true, "[" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,0,0,4)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,0,0,1)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "]");
    view.set(CProp.VISIBLE, tabZ, true);
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,1,1)],1,1,1)," +
        "(X2,a,[(Z2,X2,3,1)],3,1,1)," +
        "(X3,a,[(Z3,X3,_,1),(Z4,X3,_,1),(Z8,X3,_,1),(Z9,X3,_,1)],0,0,4)," +
        "(X4,b,[],0,0,0)," +
        "(X5,b,[(Z5,X5,100,1)],100,1,1)," +
        "(X6,_,[(Z11,X6,_,1)],0,0,1)," +
        "(X7,c,[(Z12,X7,0,1)],0,1,1)," +
        "(X8,d,[],0,0,0)," +
        "(X9,c,[(Z6,X9,10000,1)],10000,1,1)" +
        "]");
  }

  @Test
  public void testCountOnRelationPath() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");

    view.set(CProp.VISIBLE, tabZ, false);
    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colG = addFormulaColumn(view, Path.ROOT, "fc", "sum([tabZ])");
    // Not valid; relation paths can be used only in the COUNT aggregate function.
    expectResult(view, true, "[" +
        "(X1,a,.,_)," +
        "(X2,a,.,_)," +
        "(X3,a,.,_)," +
        "(X4,b,.,_)," +
        "(X5,b,.,_)," +
        "(X6,_,.,_)," +
        "(X7,c,.,_)," +
        "(X8,d,.,_)," +
        "(X9,c,.,_)" +
        "]", ImmutableSet.<Path>of(colG));
    setFormula(view, colG, "count([tabZ])");
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,1)," +
        "(X3,a,.,4)," +
        "(X4,b,.,0)," +
        "(X5,b,.,1)," +
        "(X6,_,.,3)," +
        "(X7,c,.,1)," +
        "(X8,d,.,0)," +
        "(X9,c,.,1)" +
        "]");
    setFilter(view, colZB, "not ([pk_colZA] is 'Z6')");
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,1)," +
        "(X3,a,.,4)," +
        "(X4,b,.,0)," +
        "(X5,b,.,1)," +
        "(X6,_,.,2)," +
        "(X7,c,.,1)," +
        "(X8,d,.,0)," +
        "(X9,c,.,0)" +
        "]");
    setFilter(view, colZB, "true");

    view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);
    view.set(CProp.VISIBLE, colZA, false);
    view.set(CProp.VISIBLE, colZB, false);
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,1)," +
        "(X3,a,.,3)," + // Duplicate values collapsed.
        "(X4,b,.,0)," +
        "(X5,b,.,1)," +
        "(X6,_,.,3)," +
        "(X7,c,.,1)," +
        "(X8,d,.,0)," +
        "(X9,c,.,1)" +
        "]");
    view.set(CProp.VISIBLE, tabZ, true);
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,1)],1)," +
        "(X2,a,[(.,.,3)],1)," +
        "(X3,a,[(.,.,10),(.,.,14),(.,.,2)],3)," +
        "(X4,b,[],0)," +
        "(X5,b,[(.,.,100)],1)," +
        "(X6,_,[(.,.,0),(.,.,1000),(.,.,_)],3)," + // Correctly count the all-null tuple as well.
        "(X7,c,[(.,.,0)],1)," +
        "(X8,d,[],0)," +
        "(X9,c,[(.,.,10000)],1)" +
        "]");
    view.set(CProp.VISIBLE, colZC, false);
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.)],1)," +
        "(X2,a,[(.,.,.)],1)," +
        "(X3,a,[(.,.,.)],1)," +
        "(X4,b,[],0)," +
        "(X5,b,[(.,.,.)],1)," +
        "(X6,_,[(.,.,.)],1)," +
        "(X7,c,[(.,.,.)],1)," +
        "(X8,d,[],0)," +
        "(X9,c,[(.,.,.)],1)" +
        "]");
    view.set(CProp.VISIBLE, tabZ, false);
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,1)," +
        "(X3,a,.,1)," +
        "(X4,b,.,0)," +
        "(X5,b,.,1)," +
        "(X6,_,.,1)," +
        "(X7,c,.,1)," +
        "(X8,d,.,0)," +
        "(X9,c,.,1)" +
        "]");
  }

  @Test
  public void testAggregateOnDeepJoinedChildRelation() {
    /* Based on testAggregateOnJoinedChildRelation, but with an empty intermediary relation
    inserted. */
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath between = addRelation(view, Path.ROOT, "between");
    view.set(CProp.VISIBLE, between, false);

    RelationPath tabZ = addRelation(view, between, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "fx", "sum([colZC])");

    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,3)," +
        "(X3,a,.,26)," +
        "(X4,b,.,0)," +
        "(X5,b,.,100)," +
        "(X6,_,.,1000)," +
        "(X7,c,.,0)," +
        "(X8,d,.,0)," +
        "(X9,c,.,10000)" +
        "]");
    setFilter(view, colZB, "not ([pk_colZA] is 'Z6')");
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,3)," +
        "(X3,a,.,26)," +
        "(X4,b,.,0)," +
        "(X5,b,.,100)," +
        "(X6,_,.,0)," +
        "(X7,c,.,0)," +
        "(X8,d,.,0)," +
        "(X9,c,.,0)" +
        "]");
    setFilter(view, colZB, "true");
    setFilter(view, colZC, "not ([colZC] is 14)");
    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,3)," +
        "(X3,a,.,12)," +
        "(X4,b,.,0)," +
        "(X5,b,.,100)," +
        "(X6,_,.,1000)," +
        "(X7,c,.,0)," +
        "(X8,d,.,0)," +
        "(X9,c,.,10000)" +
        "]");
  }

  @Test
  public void testManyToManyNest() {
    // Based in part on testManyToManyJoin.

    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabY", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "b",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });
    createTable("tabY2", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "c",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabYPath = addRelation(view, Path.ROOT, "tabY");
    view.set(CProp.VISIBLE, tabYPath, false);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYA = addSQLColumn(view, tabYPath, "pk_colYA");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");
    PrimitivePath colYC = addSQLColumn(view, tabYPath, "colYC_n");
    PrimitivePath colfYA = addFormulaColumn(view, Path.ROOT, "fYA", "[pk_colYA]");
    PrimitivePath colfYB = addFormulaColumn(view, Path.ROOT, "fYB", "[colYB_n]");

    view.set(QProp.JOINED_ON, colYC, OptionalPrimitivePath.of(colXB));

    view.set(CProp.VISIBLE, tabYPath, true);

    expectResult(view, true, "[" +
        "(X1,a,[(Y1,Y1d,a)],Y1,Y1d)," +
        "(X2,a,[(Y1,Y1d,a)],Y1,Y1d)," +
        "(X3,a,[(Y1,Y1d,a)],Y1,Y1d)," +
        "(X4,b,[(Y2,Y2d,b)],Y2,Y2d)," +
        "(X5,b,[(Y2,Y2d,b)],Y2,Y2d)," +
        "(X6,_,[],_,_)," +
        "(X7,c,[(Y3,Y3d,c)],Y3,Y3d)," +
        "(X8,d,[],_,_)," +
        "(X9,c,[(Y3,Y3d,c)],Y3,Y3d)" +
        "]");

    view.set(CProp.VISIBLE, colfYA, false);
    view.set(CProp.VISIBLE, colfYB, false);
    expectResult(view, true, "[" +
        "(X1,a,[(Y1,Y1d,a)],.,.)," +
        "(X2,a,[(Y1,Y1d,a)],.,.)," +
        "(X3,a,[(Y1,Y1d,a)],.,.)," +
        "(X4,b,[(Y2,Y2d,b)],.,.)," +
        "(X5,b,[(Y2,Y2d,b)],.,.)," +
        "(X6,_,[],.,.)," +
        "(X7,c,[(Y3,Y3d,c)],.,.)," +
        "(X8,d,[],.,.)," +
        "(X9,c,[(Y3,Y3d,c)],.,.)" +
        "]");

    setInstantiatedTable(view, tabYPath, "tabY2");
    expectResult(view, true, "[" +
        "(X1,a,[(Y1,Y1d,a)],.,.)," +
        "(X2,a,[(Y1,Y1d,a)],.,.)," +
        "(X3,a,[(Y1,Y1d,a)],.,.)," +
        "(X4,b,[],.,.)," +
        "(X5,b,[],.,.)," +
        "(X6,_,[],.,.)," +
        "(X7,c,[(Y2,Y2d,c),(Y3,Y3d,c)],.,.)," +
        "(X8,d,[],.,.)," +
        "(X9,c,[(Y2,Y2d,c),(Y3,Y3d,c)],.,.)" +
        "]");

    view.set(CProp.VISIBLE, colfYA, true);
    view.set(CProp.VISIBLE, colfYB, true);
    expectResult(view, true, "[" +
        "(X1,a,[(Y1,Y1d,a)],Y1,Y1d)," +
        "(X2,a,[(Y1,Y1d,a)],Y1,Y1d)," +
        "(X3,a,[(Y1,Y1d,a)],Y1,Y1d)," +
        "(X4,b,[],_,_)," +
        "(X5,b,[],_,_)," +
        "(X6,_,[],_,_)," +
        "(X7,c,[(Y2,Y2d,c)],Y2,Y2d)," +
        "(X7,c,[(Y3,Y3d,c)],Y3,Y3d)," +
        "(X8,d,[],_,_)," +
        "(X9,c,[(Y2,Y2d,c)],Y2,Y2d)," +
        "(X9,c,[(Y3,Y3d,c)],Y3,Y3d)" +
        "]");
    view.set(CProp.VISIBLE, colfYA, false);
    view.set(CProp.VISIBLE, colfYB, false);
    expectResult(view, true, "[" +
        "(X1,a,[(Y1,Y1d,a)],.,.)," +
        "(X2,a,[(Y1,Y1d,a)],.,.)," +
        "(X3,a,[(Y1,Y1d,a)],.,.)," +
        "(X4,b,[],.,.)," +
        "(X5,b,[],.,.)," +
        "(X6,_,[],.,.)," +
        "(X7,c,[(Y2,Y2d,c),(Y3,Y3d,c)],.,.)," +
        "(X8,d,[],.,.)," +
        "(X9,c,[(Y2,Y2d,c),(Y3,Y3d,c)],.,.)" +
        "]");

    view.set(CProp.VISIBLE, colXA, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true, "[" +
        "(.,_,[],.,.)," +
        "(.,a,[(Y1,Y1d,a)],.,.)," +
        "(.,b,[],.,.)," +
        "(.,c,[(Y2,Y2d,c),(Y3,Y3d,c)],.,.)," +
        "(.,d,[],.,.)" +
        "]");
    view.set(CProp.VISIBLE, colXA, true);

    PrimitivePath fOut = addFormulaColumn(view, tabYPath, "fOut", "[pk_colXA]");
    expectResult(view, true, "[" +
        "(X1,a,[(Y1,Y1d,a,X1)],.,.)," +
        "(X2,a,[(Y1,Y1d,a,X2)],.,.)," +
        "(X3,a,[(Y1,Y1d,a,X3)],.,.)," +
        "(X4,b,[],.,.)," +
        "(X5,b,[],.,.)," +
        "(X6,_,[],.,.)," +
        "(X7,c,[(Y2,Y2d,c,X7),(Y3,Y3d,c,X7)],.,.)," +
        "(X8,d,[],.,.)," +
        "(X9,c,[(Y2,Y2d,c,X9),(Y3,Y3d,c,X9)],.,.)" +
        "]");
    view.set(CProp.VISIBLE, colXA, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true, "[" +
        "(.,_,[],.,.)," +
        "(.,a,[(Y1,Y1d,a,X1),(Y1,Y1d,a,X2),(Y1,Y1d,a,X3)],.,.)," +
        "(.,b,[],.,.)," +
        "(.,c,[(Y2,Y2d,c,X7),(Y2,Y2d,c,X9),(Y3,Y3d,c,X7),(Y3,Y3d,c,X9)],.,.)," +
        "(.,d,[],.,.)" +
        "]");
    view.set(CProp.VISIBLE, colXA, true);
    view.set(CProp.VISIBLE, colfYA, true);
    view.set(CProp.VISIBLE, colfYB, true);

    setFilter(view, colfYB, "[fYB] is null");
    expectResult(view, true, "[" +
        "(X4,b,[],_,_)," +
        "(X5,b,[],_,_)," +
        "(X6,_,[],_,_)," +
        "(X8,d,[],_,_)" +
        "]");
    setFilter(view, colfYB, "true");
  }

  @Test
  public void testManyToManyNest2() {
    /* Data from testLateralReferences (for which the join tabX.pk_colXA = tabZ.pk_colZB includes
    matches with all cardinalities). */
    Object tabXcontent[] = new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    };
    createTable("tabX", tabXcontent);
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB",
      String.class, String.class,
      "Z1", "X1",
      "Z2", "X2",
      "Z3", "X3",
      "Z4", "X3",
      "Z5", "X5",
      "Z6", "X6",
      "Z6", "X9",
      "Z7", "XX"
    });
    createTable("tabQ", new Object[] {
      "pk_colQA",
      Integer.class,
      1,
      2
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    view.set(CProp.VISIBLE, tabZ, true);
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view,tabZ, "pk_colZB");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));
    RelationPath tabQ = addRelation(view, Path.ROOT, "tabQ");
    setInstantiatedTable(view, tabQ, "tabQ");
    view.set(CProp.VISIBLE, tabQ, false);
    PrimitivePath colQA = addSQLColumn(view, tabQ, "pk_colQA");
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1)],.)," +
        "(X2,a,[(Z2,X2)],.)," +
        "(X3,a,[(Z3,X3),(Z4,X3)],.)," +
        "(X4,b,[],.)," +
        "(X5,b,[(Z5,X5)],.)," +
        "(X6,_,[(Z6,X6)],.)," +
        "(X7,c,[],.)," +
        "(X8,d,[],.)," +
        "(X9,c,[(Z6,X9)],.)" +
        "]");
    PrimitivePath fx = addFormulaColumn(view, Path.ROOT, "fx", "[pk_colQA]");
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1)],.,1)," +
        "(X1,a,[(Z1,X1)],.,2)," +
        "(X2,a,[(Z2,X2)],.,1)," +
        "(X2,a,[(Z2,X2)],.,2)," +
        "(X3,a,[(Z3,X3),(Z4,X3)],.,1)," +
        "(X3,a,[(Z3,X3),(Z4,X3)],.,2)," +
        "(X4,b,[],.,1)," +
        "(X4,b,[],.,2)," +
        "(X5,b,[(Z5,X5)],.,1)," +
        "(X5,b,[(Z5,X5)],.,2)," +
        "(X6,_,[(Z6,X6)],.,1)," +
        "(X6,_,[(Z6,X6)],.,2)," +
        "(X7,c,[],.,1)," +
        "(X7,c,[],.,2)," +
        "(X8,d,[],.,1)," +
        "(X8,d,[],.,2)," +
        "(X9,c,[(Z6,X9)],.,1)," +
        "(X9,c,[(Z6,X9)],.,2)" +
        "]");
    PrimitivePath fxI = addFormulaColumn(view, tabZ, "fxI", "[fx]");
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,1)],.,1)," +
        "(X1,a,[(Z1,X1,2)],.,2)," +
        "(X2,a,[(Z2,X2,1)],.,1)," +
        "(X2,a,[(Z2,X2,2)],.,2)," +
        "(X3,a,[(Z3,X3,1),(Z4,X3,1)],.,1)," +
        "(X3,a,[(Z3,X3,2),(Z4,X3,2)],.,2)," +
        "(X4,b,[],.,1)," +
        "(X4,b,[],.,2)," +
        "(X5,b,[(Z5,X5,1)],.,1)," +
        "(X5,b,[(Z5,X5,2)],.,2)," +
        "(X6,_,[(Z6,X6,1)],.,1)," +
        "(X6,_,[(Z6,X6,2)],.,2)," +
        "(X7,c,[],.,1)," +
        "(X7,c,[],.,2)," +
        "(X8,d,[],.,1)," +
        "(X8,d,[],.,2)," +
        "(X9,c,[(Z6,X9,1)],.,1)," +
        "(X9,c,[(Z6,X9,2)],.,2)" +
        "]");
    view.set(CProp.VISIBLE, fx, false);
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,1),(Z1,X1,2)],.,.)," +
        "(X2,a,[(Z2,X2,1),(Z2,X2,2)],.,.)," +
        "(X3,a,[(Z3,X3,1),(Z3,X3,2),(Z4,X3,1),(Z4,X3,2)],.,.)," +
        "(X4,b,[],.,.)," +
        "(X5,b,[(Z5,X5,1),(Z5,X5,2)],.,.)," +
        "(X6,_,[(Z6,X6,1),(Z6,X6,2)],.,.)," +
        "(X7,c,[],.,.)," +
        "(X8,d,[],.,.)," +
        "(X9,c,[(Z6,X9,1),(Z6,X9,2)],.,.)" +
        "]");
  }

  @Test
  public void testGroupByDeceptiveStrings() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "",
      "X2", " ",
      "X3", null,
      "X4", "A",
      "X5", "a",
      "X6", "A",
      "X7", "a",
      "X8", null,
      "X9", "a "
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabX2 = addRelation(view, Path.ROOT, "tabX2");
    setInstantiatedTable(view, tabX2, "tabX");
    PrimitivePath colX2A = addSQLColumn(view, tabX2, "pk_colXA");
    PrimitivePath colX2B = addSQLColumn(view, tabX2, "colXB_n");
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.of(colXA));
    expectResult(view, true, "[" +
        "(X1,,[(X1,)])," +
        "(X2, ,[(X2, )])," +
        "(X3,_,[(X3,_)])," +
        "(X4,A,[(X4,A)])," +
        "(X5,a,[(X5,a)])," +
        "(X6,A,[(X6,A)])," +
        "(X7,a,[(X7,a)])," +
        "(X8,_,[(X8,_)])," +
        "(X9,a ,[(X9,a )])" +
        "]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colXA, false);
    expectResult(view, true, "[" +
        "(., ,[(X2, )])," +
        "(.,,[(X1,)])," +
        "(.,A,[(X4,A),(X6,A)])," +
        "(.,_,[(X3,_),(X8,_)])," +
        "(.,a ,[(X9,a )])," +
        "(.,a,[(X5,a),(X7,a)])" +
        "]");
    view.set(CProp.VISIBLE, colX2B, false);
    expectResult(view, true, "[" +
        "(., ,[(X2,.)])," +
        "(.,,[(X1,.)])," +
        "(.,A,[(X4,.),(X6,.)])," +
        "(.,_,[(X3,.),(X8,.)])," +
        "(.,a ,[(X9,.)])," +
        "(.,a,[(X5,.),(X7,.)])" +
        "]");
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colX2B, OptionalPrimitivePath.of(colXB));
    // This test is verified to cover the CORRELATION_JOIN_OPTIMIZATION case.
    expectResult(view, true, "[" +
        "(., ,[(X2,.)])," +
        "(.,,[(X1,.)])," +
        "(.,A,[(X4,.),(X6,.)])," +
        "(.,_,[])," + // Empty subrelation here (join does not match on null values).
        "(.,a ,[(X9,.)])," +
        "(.,a,[(X5,.),(X7,.)])" +
        "]");
  }

  @Test
  public void testGroupByDeceptiveStringsWithPseudoKey() {
    String pseudoKey = TEST_STATE.getDialect().getRowIdentifierPseudoColumnName();
    if (pseudoKey.isEmpty())
      return;

    createTable("tabX", new Object[] {
      "colXA", "colXB_n",
      String.class, String.class,
      "X1", "",
      "X2", " ",
      "X3", null,
      "X4", "A",
      "X5", "a",
      "X6", "A",
      "X7", "a",
      "X8", null
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXKey = addSQLColumn(view, Path.ROOT, pseudoKey);
    view.set(CProp.LABEL, colXKey, "colXKey");
    view.set(CProp.VISIBLE, colXKey, false);
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabX2 = addRelation(view, Path.ROOT, "tabX2");
    setInstantiatedTable(view, tabX2, "tabX");
    PrimitivePath colX2Key = addSQLColumn(view, tabX2, pseudoKey);
    view.set(CProp.LABEL, colX2Key, "colX2Key");
    view.set(CProp.VISIBLE, colX2Key, false);
    PrimitivePath colX2A = addSQLColumn(view, tabX2, "colXA");
    PrimitivePath colX2B = addSQLColumn(view, tabX2, "colXB_n");
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.of(colXA));
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabX2, false);
    expectResult(view, true, "[" +
        "(.,X1,,[(.,X1,)])," +
        "(.,X2, ,[(.,X2, )])," +
        "(.,X3,_,[(.,X3,_)])," +
        "(.,X4,A,[(.,X4,A)])," +
        "(.,X5,a,[(.,X5,a)])," +
        "(.,X6,A,[(.,X6,A)])," +
        "(.,X7,a,[(.,X7,a)])," +
        "(.,X8,_,[(.,X8,_)])" +
        "]");
    view.set(CProp.VISIBLE, colXA, false);
    expectResult(view, true, "[" +
        "(.,., ,[(.,X2, )])," +
        "(.,.,,[(.,X1,)])," +
        "(.,.,A,[(.,X4,A),(.,X6,A)])," +
        "(.,.,_,[(.,X3,_),(.,X8,_)])," +
        "(.,.,a,[(.,X5,a),(.,X7,a)])" +
        "]");
    view.set(CProp.VISIBLE, colX2B, false);
    String res1 = "[" +
        "(.,., ,[(.,X2,.)])," +
        "(.,.,,[(.,X1,.)])," +
        "(.,.,A,[(.,X4,.),(.,X6,.)])," +
        "(.,.,_,[(.,X3,.),(.,X8,.)])," +
        "(.,.,a,[(.,X5,.),(.,X7,.)])" +
        "]";
    expectResult(view, true, res1);
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colX2B, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, "[" +
        "(.,., ,[(.,X2,.)])," +
        "(.,.,,[(.,X1,.)])," +
        "(.,.,A,[(.,X4,.),(.,X6,.)])," +
        "(.,.,_,[])," + // Empty subrelation here (join does not match on null values).
        "(.,.,a,[(.,X5,.),(.,X7,.)])" +
        "]");
    view.set(QProp.JOINED_ON, colX2B, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colX2Key, OptionalPrimitivePath.of(colXKey));
    expectResult(view, true, res1);

    addFormulaColumn(view, Path.ROOT, "fx", "count([tabX2])");
    expectResult(view, true, "[" +
        "(.,., ,[(.,X2,.)],1)," +
        "(.,.,,[(.,X1,.)],1)," +
        "(.,.,A,[(.,X4,.),(.,X6,.)],2)," +
        "(.,.,_,[(.,X3,.),(.,X8,.)],2)," +
        "(.,.,a,[(.,X5,.),(.,X7,.)],2)" +
        "]");
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colX2B, OptionalPrimitivePath.of(colXB));
    expectResult(view, true, "[" +
    "(.,., ,[(.,X2,.)],1)," +
    "(.,.,,[(.,X1,.)],1)," +
    "(.,.,A,[(.,X4,.),(.,X6,.)],2)," +
    /* Join does not match on null values, but the parent tuple should still appear with its
    aggregate result. */
    "(.,.,_,[],0)," +
    "(.,.,a,[(.,X5,.),(.,X7,.)],2)" +
    "]");
  }

  @Test
  public void testPseudoKeyNest() {
    String pseudoKey = TEST_STATE.getDialect().getRowIdentifierPseudoColumnName();
    if (pseudoKey.isEmpty())
      return;
    /* Data from testLateralReferences (for which the join tabX.pk_colXA = tabZ.pk_colZB includes
    matches with all cardinalities). */
    Object tabXcontent[] = new Object[] {
      "colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    };
    createTable("tabX", tabXcontent);
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB",
      String.class, String.class,
      "Z1", "X1",
      "Z2", "X2",
      "Z3", "X3",
      "Z4", "X3",
      "Z5", "X5",
      "Z6", "X6",
      "Z6", "X9",
      "Z7", "XX"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXKey = addSQLColumn(view, Path.ROOT, pseudoKey);
    view.set(CProp.LABEL, colXKey, "colXKey");
    view.set(CProp.VISIBLE, colXKey, false);
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    view.set(CProp.VISIBLE, tabZ, true);
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view,tabZ, "pk_colZB");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    expectResult(view, true, "[" +
        "(~,X1,a,[(Z1,X1)])," +
        "(~,X2,a,[(Z2,X2)])," +
        "(~,X3,a,[(Z3,X3),(Z4,X3)])," +
        "(~,X4,b,[])," +
        "(~,X5,b,[(Z5,X5)])," +
        "(~,X6,_,[(Z6,X6)])," +
        "(~,X7,c,[])," +
        "(~,X8,d,[])," +
        "(~,X9,c,[(Z6,X9)])" +
        "]");
  }

  @Test
  public void testInwardAndAggregate() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC",
      String.class, String.class, String.class,
      "Z1", "X1", "a",
      "Z2", "X2", "b",
      "Z3", "X3", "c",
      "Z4", "X3", "c",
      "Z5", "X5", "e",
      "Z6", "X6", "f",
      "Z6", "X9", "g",
      "Z7", "XX", "h",
      "Z8", "X3", "i",
      "Z9", "X3", "i"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    view.set(CProp.VISIBLE, tabZ, true);
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view,tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC");

    PrimitivePath inw = addFormulaColumn(view, Path.ROOT, "inw", "[colZC]");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,a)],a)," +
        "(X2,a,[(Z2,X2,b)],b)," +
        "(X3,a,[(Z3,X3,c),(Z4,X3,c)],c)," +
        "(X3,a,[(Z8,X3,i),(Z9,X3,i)],i)," +
        "(X4,b,[],_)," +
        "(X5,b,[(Z5,X5,e)],e)," +
        "(X6,_,[(Z6,X6,f)],f)," +
        "(X7,c,[],_)," +
        "(X8,d,[],_)," +
        "(X9,c,[(Z6,X9,g)],g)" +
        "]");

    PrimitivePath cnt = addFormulaColumn(view, Path.ROOT, "cnt", "count([colZC])");
    view.set(SProp.CHILD_PATHS, Path.ROOT,
        new PathList(ImmutableList.of(colXA, colXB, tabZ, inw, cnt)));
    String res1 = "[" +
        "(X1,a,[(Z1,X1,a)],a,1)," +
        "(X2,a,[(Z2,X2,b)],b,1)," +
        "(X3,a,[(Z3,X3,c),(Z4,X3,c)],c,2)," +
        "(X3,a,[(Z8,X3,i),(Z9,X3,i)],i,2)," +
        "(X4,b,[],_,0)," +
        "(X5,b,[(Z5,X5,e)],e,1)," +
        "(X6,_,[(Z6,X6,f)],f,1)," +
        "(X7,c,[],_,0)," +
        "(X8,d,[],_,0)," +
        "(X9,c,[(Z6,X9,g)],g,1)" +
        "]";
    expectResult(view, true, res1);
    setFormula(view, cnt, "count([pk_colZA])");
    expectResult(view, true, res1);
    /* Change the order of the aggregate function and the inward reference. This once erroneously
    triggered the "Cannot combine aggregate function with reference that may potentially yield more
    than one value" error. */
    view.set(SProp.CHILD_PATHS, Path.ROOT,
        new PathList(ImmutableList.of(colXA, colXB, tabZ, cnt, inw)));
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,a)],1,a)," +
        "(X2,a,[(Z2,X2,b)],1,b)," +
        "(X3,a,[(Z3,X3,c),(Z4,X3,c)],2,c)," +
        "(X3,a,[(Z8,X3,i),(Z9,X3,i)],2,i)," +
        "(X4,b,[],0,_)," +
        "(X5,b,[(Z5,X5,e)],1,e)," +
        "(X6,_,[(Z6,X6,f)],1,f)," +
        "(X7,c,[],0,_)," +
        "(X8,d,[],0,_)," +
        "(X9,c,[(Z6,X9,g)],1,g)" +
        "]");
  }

  @Test
  public void testAggregateSelectDistinctness() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC",
      String.class, String.class, String.class,
      "Z1", "X1", "a",
      "Z2", "X2", "bb",
      "Z3", "X3", "ccc",
      "Z4", "X3", "cZc",
      "Z5", "X5", "e",
      "Z6", "X6", "f",
      "Z6", "X9", "gg",
      "Z7", "XX", "h",
      "Z8", "X3", "i",
      "Z9", "X3", "i"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    view.set(CProp.VISIBLE, tabZ, true);
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view,tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);

    PrimitivePath len = addFormulaColumn(view, tabZ, "len", "len([colZC])");
    PrimitivePath cnt = addFormulaColumn(view, Path.ROOT, "cnt", "count([colZC])");
    /* Aggregate functions do not themselves collapse their input values, so the following two
    formulas will give different results. */
    PrimitivePath cntIn = addFormulaColumn(view, Path.ROOT, "cnt", "count([len])");
    PrimitivePath cntLen = addFormulaColumn(view, Path.ROOT, "cnt", "count(len([colZC]))");
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,a,1)],1,1,1)," +
        "(X2,a,[(Z2,X2,bb,2)],1,1,1)," +
        "(X3,a,[(Z3,X3,ccc,3),(Z4,X3,cZc,3),(Z8,X3,i,1),(Z9,X3,i,1)],4,4,4)," +
        "(X4,b,[],0,0,0)," +
        "(X5,b,[(Z5,X5,e,1)],1,1,1)," +
        "(X6,_,[(Z6,X6,f,1)],1,1,1)," +
        "(X7,c,[],0,0,0)," +
        "(X8,d,[],0,0,0)," +
        "(X9,c,[(Z6,X9,gg,2)],1,1,1)" +
        "]");
    view.set(CProp.VISIBLE, colZA, false);
    view.set(CProp.VISIBLE, colZB, false);
    view.set(CProp.VISIBLE, colZC, false);

    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.,1)],1,1,1)," +
        "(X2,a,[(.,.,.,2)],1,1,1)," +
        "(X3,a,[(.,.,.,1),(.,.,.,3)],3,2,3)," + // Last should be 3, not 2.
        "(X4,b,[],0,0,0)," +
        "(X5,b,[(.,.,.,1)],1,1,1)," +
        "(X6,_,[(.,.,.,1)],1,1,1)," +
        "(X7,c,[],0,0,0)," +
        "(X8,d,[],0,0,0)," +
        "(X9,c,[(.,.,.,2)],1,1,1)" +
        "]");
    view.removePath(cnt);
    view.removePath(cntIn);
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.,1)],1)," +
        "(X2,a,[(.,.,.,2)],1)," +
        "(X3,a,[(.,.,.,1),(.,.,.,3)],3)," + // Last should be 3, not 2.
        "(X4,b,[],0)," +
        "(X5,b,[(.,.,.,1)],1)," +
        "(X6,_,[(.,.,.,1)],1)," +
        "(X7,c,[],0)," +
        "(X8,d,[],0)," +
        "(X9,c,[(.,.,.,2)],1)" +
        "]");
  }

  @Test
  public void testOutwardReferenceToSelfAggregate() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC",
      String.class, String.class, String.class,
      "Z1", "X1", "a",
      "Z2", "X2", "b",
      "Z3", "X3", "c",
      "Z4", "X3", "c",
      "Z5", "X5", "e",
      "Z6", "X5", "f",
      "Z6", "X9", "g",
      "Z7", "XX", "h",
      "Z8", "X3", "i",
      "Z9", "X3", "i"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    view.set(CProp.VISIBLE, tabZ, true);
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view,tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath cnt = addFormulaColumn(view, Path.ROOT, "cnt", "count([colZC])");
    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,a)],1)," +
        "(X2,a,[(Z2,X2,b)],1)," +
        "(X3,a,[(Z3,X3,c),(Z4,X3,c),(Z8,X3,i),(Z9,X3,i)],4)," +
        "(X4,b,[],0)," +
        "(X5,b,[(Z5,X5,e),(Z6,X5,f)],2)," +
        "(X6,_,[],0)," +
        "(X7,c,[],0)," +
        "(X8,d,[],0)," +
        "(X9,c,[(Z6,X9,g)],1)" +
        "]");

    addFormulaColumn(view, tabZ, "outref", "[cnt]");
    String res1 = "[" +
        "(X1,a,[(Z1,X1,a,1)],1)," +
        "(X2,a,[(Z2,X2,b,1)],1)," +
        "(X3,a,[(Z3,X3,c,4),(Z4,X3,c,4),(Z8,X3,i,4),(Z9,X3,i,4)],4)," +
        "(X4,b,[],0)," +
        "(X5,b,[(Z5,X5,e,2),(Z6,X5,f,2)],2)," +
        "(X6,_,[],0)," +
        "(X7,c,[],0)," +
        "(X8,d,[],0)," +
        "(X9,c,[(Z6,X9,g,1)],1)" +
        "]";
    expectResult(view, true, res1);
    setFormula(view, cnt, "count([pk_colZA])");
    expectResult(view, true, res1);
    addFormulaColumn(view, Path.ROOT, "inagain", "[outref]");

    expectResult(view, true, "[" +
        "(X1,a,[(Z1,X1,a,1)],1,1)," +
        "(X2,a,[(Z2,X2,b,1)],1,1)," +
        "(X3,a,[(Z3,X3,c,4),(Z4,X3,c,4),(Z8,X3,i,4),(Z9,X3,i,4)],4,4)," +
        "(X4,b,[],0,_)," +
        "(X5,b,[(Z5,X5,e,2),(Z6,X5,f,2)],2,2)," +
        "(X6,_,[],0,_)," +
        "(X7,c,[],0,_)," +
        "(X8,d,[],0,_)," +
        "(X9,c,[(Z6,X9,g,1)],1,1)" +
        "]");
    addFormulaColumn(view, tabZ, "cnt", "count([colZC])");
    // (*pk_colXA, colXB_n, tabZ(*pk_colZA, pk_colZB=[pk_colXA], colZC, outref:[root\cnt], cnt:count([colZC])), cnt:count([pk_colZA]), inagain:[outref])
    expectResultWrapped(view, true, "[" +
        "(X1,a,[(Z1,X1,a,1,1)],1,1)," +
        "(X2,a,[(Z2,X2,b,1,1)],1,1)," +
        "(X3,a,[(Z3,X3,c,4,1),(Z4,X3,c,4,1),(Z8,X3,i,4,1),(Z9,X3,i,4,1)],4,4)," +
        "(X4,b,[],0,_)," +
        "(X5,b,[(Z5,X5,e,2,1),(Z6,X5,f,2,1)],2,2)," +
        "(X6,_,[],0,_)," +
        "(X7,c,[],0,_)," +
        "(X8,d,[],0,_)," +
        "(X9,c,[(Z6,X9,g,1,1)],1,1)" +
        "]");
  }

  @Test
  public void testAggregateFromPseudokeyViewParentRelation() {
    final String pseudoKey = TEST_STATE.getDialect().getRowIdentifierPseudoColumnName();
    if (pseudoKey.isEmpty())
      return;

    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createPassthroughView("tabX_view", "tabX");
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX_view");
    PrimitivePath colXk = addSQLColumn(view, Path.ROOT, pseudoKey);
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZk = addSQLColumn(view, tabZ, pseudoKey);
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");

    view.set(CProp.VISIBLE, tabZ, false);
    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "fs", "sum([colZC_n])");

    expectResult(view, true, "[" +
        "(~,X1,a,.,1)," +
        "(~,X2,a,.,3)," +
        "(~,X3,a,.,28)," +
        "(~,X4,b,.,0)," +
        "(~,X5,b,.,100)," +
        "(~,X6,_,.,1000)," +
        "(~,X7,c,.,0)," +
        "(~,X8,d,.,0)," +
        "(~,X9,c,.,10000)" +
        "]");
  }

  @Test
  public void testAggregateOnJoinedDistinctChildRelation() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "M",
      "X2", "a", "W",
      "X3", "a", "W",
      "X4", "a", "F",
      "X5", "b", "T",
      "X6", "b", "M",
      "X7", "b", "M",
      "X8", "b", "M",
      "X9", "b", "T"
    });
    createTable("tabX2", new Object[] {
      "pk_colX2A", "colX2B", "colX2C",
      String.class, String.class, String.class,
      "X1", "a", "M",
      "X2", "a", "W",
      "X3", "a", "W",
      "X4", "a", "F",
      "X5", "b", "T",
      "X6", "b", "M",
      "X7", "b", "M",
      "X8", "b", "M",
      "X9", "b", "T"
    });

    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB");
    PrimitivePath colXC = addSQLColumn(view, Path.ROOT, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colXA, false);
    view.set(CProp.VISIBLE, colXC, false);

    RelationPath tabX2 = addRelation(view, Path.ROOT, "tabX2");
    setInstantiatedTable(view, tabX2, "tabX2");
    PrimitivePath colX2A = addSQLColumn(view, tabX2, "pk_colX2A");
    PrimitivePath colX2B = addSQLColumn(view, tabX2, "colX2B");
    PrimitivePath colX2C = addSQLColumn(view, tabX2, "colX2C");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabX2, false);
    view.set(CProp.VISIBLE, colX2B, false);
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.of(colXA));

    PrimitivePath colFs = addFormulaColumn(view, Path.ROOT, "fs", "count([colX2C])");

    expectResult(view, true, "[" +
        "(.,a,.,[(X1,.,M),(X2,.,W),(X3,.,W),(X4,.,F)],4)," +
        "(.,b,.,[(X5,.,T),(X6,.,M),(X7,.,M),(X8,.,M),(X9,.,T)],5)" +
        "]");

    view.set(CProp.VISIBLE, tabX2, false);
    expectResult(view, true, "[" +
        "(.,a,.,.,4)," +
        "(.,b,.,.,5)" +
        "]");
    view.set(CProp.VISIBLE, colX2A, false);
    expectResult(view, true, "[" +
        "(.,a,.,.,3)," +
        "(.,b,.,.,2)" +
        "]");

    setFilter(view, colFs, "[fs] is 3");
    expectResult(view, true, "[" +
        "(.,a,.,.,3)" +
        "]");
  }

  @Test
  public void testDoubleGroupCount() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "M",
      "X2", "a", "W",
      "X3", "a", "W",
      "X4", "a", "F",
      "X5", "b", "T",
      "X6", "b", "M",
      "X7", "b", "M",
      "X8", "b", "M",
      "X9", "b", "T"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB");
    PrimitivePath colXC = addSQLColumn(view, Path.ROOT, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colXA, false);
    view.set(CProp.VISIBLE, colXC, false);
    expectResult(view, true, "[" +
        "(.,a,.)," +
        "(.,b,.)" +
        "]");

    RelationPath tabX2 = addRelation(view, Path.ROOT, "tabX2");
    setInstantiatedTable(view, tabX2, "tabX");
    PrimitivePath colX2A = addSQLColumn(view, tabX2, "pk_colXA");
    PrimitivePath colX2B = addSQLColumn(view, tabX2, "colXB");
    PrimitivePath colX2C = addSQLColumn(view, tabX2, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabX2, false);
    view.set(CProp.VISIBLE, colX2A, false);
    view.set(CProp.VISIBLE, colX2B, false);
    view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.of(colXA));
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F),(.,.,M),(.,.,W)])," +
        "(.,b,.,[(.,.,M),(.,.,T)])" +
        "]");

    RelationPath tabX3 = addRelation(view, tabX2, "tabX3");
    setInstantiatedTable(view, tabX3, "tabX");
    PrimitivePath colX3A = addSQLColumn(view, tabX3, "pk_colXA");
    PrimitivePath colX3B = addSQLColumn(view, tabX3, "colXB");
    PrimitivePath colX3C = addSQLColumn(view, tabX3, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabX3, false);
    view.set(CProp.VISIBLE, colX3A, true);
    view.set(CProp.VISIBLE, colX3B, false);
    view.set(CProp.VISIBLE, colX3C, false);
    view.set(QProp.JOINED_ON, colX3A, OptionalPrimitivePath.of(colX2A));
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,[(X4,.,.)]),(.,.,M,[(X1,.,.)]),(.,.,W,[(X2,.,.),(X3,.,.)])])," +
        "(.,b,.,[(.,.,M,[(X6,.,.),(X7,.,.),(X8,.,.)]),(.,.,T,[(X5,.,.),(X9,.,.)])])" +
        "]");

    view.set(CProp.VISIBLE, tabX3, false);
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.),(.,.,M,.),(.,.,W,.)])," +
        "(.,b,.,[(.,.,M,.),(.,.,T,.)])" +
        "]");

    addFormulaColumn(view, tabX2, "fx", "count([tabX3\\pk_colXA])");
    expectResultWrapped(view, true, "[" +
        "(.,a,.,[(.,.,F,.,1),(.,.,M,.,1),(.,.,W,.,2)])," +
        "(.,b,.,[(.,.,M,.,3),(.,.,T,.,2)])" +
        "]");

    setFilter(view, tabX2, "[fx] is 2");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,W,.,2)])," +
        "(.,b,.,[(.,.,T,.,2)])" +
        "]");
    // Now add some root-level filters.
    setFilter(view, colXB, "[colXB] is 'a'");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,W,.,2)])" +
        "]");
    setFilter(view, Path.ROOT, "not ([\\pk_colXA] is 'X2')");
    setFilter(view, tabX2, "[fx] is 1");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.,1),(.,.,M,.,1),(.,.,W,.,1)])" +
        "]");
    setFilter(view, Path.ROOT, "not ([\\pk_colXA] is 'X2' or [\\pk_colXA] is 'X3')");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.,1),(.,.,M,.,1)])" +
        "]");
  }

  @Test
  public void testDoubleGroupCountWithFormulaJoins() {
    // Based on testDoubleGroupCount.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "M",
      "X2", "a", "W",
      "X3", "a", "W",
      "X4", "a", "F",
      "X5", "b", "T",
      "X6", "b", "M",
      "X7", "b", "M",
      "X8", "b", "M",
      "X9", "b", "T"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB");
    PrimitivePath colXC = addSQLColumn(view, Path.ROOT, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colXA, false);
    view.set(CProp.VISIBLE, colXC, false);
    RelationPath tabX2 = addRelation(view, Path.ROOT, "tabX2");
    setInstantiatedTable(view, tabX2, "tabX");
    PrimitivePath colX2A = addSQLColumn(view, tabX2, "pk_colXA");
    PrimitivePath colX2B = addSQLColumn(view, tabX2, "colXB");
    PrimitivePath colX2C = addSQLColumn(view, tabX2, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabX2, false);
    view.set(CProp.VISIBLE, colX2A, false);
    view.set(CProp.VISIBLE, colX2B, false);
    //view.set(QProp.JOINED_ON, colX2A, OptionalPrimitivePath.of(colXA));
    PrimitivePath tabX2J = addFormulaColumn(view, tabX2, "tabX2J", "[tabX2\\pk_colXA] = [\\pk_colXA]");
    view.set(CProp.VISIBLE, tabX2J, false);
    setFilter(view, tabX2J, "[tabX2J] is true");;
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.),(.,.,M,.),(.,.,W,.)])," +
        "(.,b,.,[(.,.,M,.),(.,.,T,.)])" +
        "]");

    RelationPath tabX3 = addRelation(view, tabX2, "tabX3");
    setInstantiatedTable(view, tabX3, "tabX");
    PrimitivePath colX3A = addSQLColumn(view, tabX3, "pk_colXA");
    PrimitivePath colX3B = addSQLColumn(view, tabX3, "colXB");
    PrimitivePath colX3C = addSQLColumn(view, tabX3, "colXC");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabX3, false);
    view.set(CProp.VISIBLE, colX3A, true);
    view.set(CProp.VISIBLE, colX3B, false);
    view.set(CProp.VISIBLE, colX3C, false);
    //view.set(QProp.JOINED_ON, colX3A, OptionalPrimitivePath.of(colX2A));
    PrimitivePath tabX3J = addFormulaColumn(view, tabX3, "tabX3J", "[tabX3\\pk_colXA] = [tabX2\\pk_colXA]");
    view.set(CProp.VISIBLE, tabX3J, false);
    setFilter(view, tabX3J, "[tabX3J] is true");;
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.,[(X4,.,.,.)]),(.,.,M,.,[(X1,.,.,.)]),(.,.,W,.,[(X2,.,.,.),(X3,.,.,.)])])," +
        "(.,b,.,[(.,.,M,.,[(X6,.,.,.),(X7,.,.,.),(X8,.,.,.)]),(.,.,T,.,[(X5,.,.,.),(X9,.,.,.)])])" +
        "]");

    view.set(CProp.VISIBLE, tabX3, false);
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.,.),(.,.,M,.,.),(.,.,W,.,.)])," +
        "(.,b,.,[(.,.,M,.,.),(.,.,T,.,.)])" +
        "]");

    addFormulaColumn(view, tabX2, "fx", "count([tabX3\\pk_colXA])");
    expectResultWrapped(view, true, "[" +
        "(.,a,.,[(.,.,F,.,.,1),(.,.,M,.,.,1),(.,.,W,.,.,2)])," +
        "(.,b,.,[(.,.,M,.,.,3),(.,.,T,.,.,2)])" +
        "]");

    setFilter(view, tabX2, "[fx] is 2");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,W,.,.,2)])," +
        "(.,b,.,[(.,.,T,.,.,2)])" +
        "]");

    setFilter(view, colXB, "[colXB] is 'a'");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,W,.,.,2)])" +
        "]");
    setFilter(view, Path.ROOT, "not ([\\pk_colXA] is 'X2')");
    setFilter(view, tabX2, "[fx] is 1");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.,.,1),(.,.,M,.,.,1),(.,.,W,.,.,1)])" +
        "]");
    setFilter(view, Path.ROOT, "not ([\\pk_colXA] is 'X2' or [\\pk_colXA] is 'X3')");
    expectResult(view, true, "[" +
        "(.,a,.,[(.,.,F,.,.,1),(.,.,M,.,.,1)])" +
        "]");
  }

  @Test
  public void testKeyInwardReferenceToAggregateFormula() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    RelationPath tabX = addRelation(view, Path.ROOT, "tabX");

    setInstantiatedTable(view, tabX, "tabX");
    PrimitivePath colXA = addSQLColumn(view, tabX, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, tabX, "colXB_n");

    RelationPath tabZ = addRelation(view, tabX, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");
    PrimitivePath const1 = addFormulaColumn(view, tabZ, "const1", "1");

    view.set(CProp.VISIBLE, tabZ, false);
    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colF = addFormulaColumn(view, tabX, "fs", "sum([colZC_n])");
    PrimitivePath colG = addFormulaColumn(view, tabX, "fc", "count([colZC_n])");
    PrimitivePath colH = addFormulaColumn(view, tabX, "fc2", "count([const1])");

    expectResultWrapped(view, true, "[([" +
        "(X1,a,.,1,1,1)," +
        "(X2,a,.,3,1,1)," +
        "(X3,a,.,28,4,4)," +
        "(X4,b,.,0,0,0)," +
        "(X5,b,.,100,1,1)," +
        "(X6,_,.,1000,2,3)," +
        "(X7,c,.,0,1,1)," +
        "(X8,d,.,0,0,0)," +
        "(X9,c,.,10000,1,1)" +
        "])]");

    // fs, fc, fc2
    PrimitivePath inwFS = addFormulaColumn(view, Path.ROOT, "inwFS", "[fs]");
    // Put the formula first just to make it easier to read the result.
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(inwFS, tabX)));
    view.set(CProp.VISIBLE, colG, false);
    view.set(CProp.VISIBLE, colH, false);

    expectResult(view, true, "[" +
        "(0,[(X4,b,.,0,.,.),(X7,c,.,0,.,.),(X8,d,.,0,.,.)])," +
        "(1,[(X1,a,.,1,.,.)])," +
        "(100,[(X5,b,.,100,.,.)])," +
        "(1000,[(X6,_,.,1000,.,.)])," +
        "(10000,[(X9,c,.,10000,.,.)])," +
        "(28,[(X3,a,.,28,.,.)])," +
        "(3,[(X2,a,.,3,.,.)])" +
        "]");

    view.set(CProp.VISIBLE, colG, true);
    view.set(CProp.VISIBLE, colH, true);
    expectResult(view, true, "[" +
        "(0,[(X4,b,.,0,0,0),(X7,c,.,0,1,1),(X8,d,.,0,0,0)])," +
        "(1,[(X1,a,.,1,1,1)])," +
        "(100,[(X5,b,.,100,1,1)])," +
        "(1000,[(X6,_,.,1000,2,3)])," +
        "(10000,[(X9,c,.,10000,1,1)])," +
        "(28,[(X3,a,.,28,4,4)])," +
        "(3,[(X2,a,.,3,1,1)])" +
        "]");

    PrimitivePath inwFC = addFormulaColumn(view, Path.ROOT, "inwFC", "[fc]");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(inwFS, inwFC, tabX)));
    expectResult(view, true, "[" +
        "(0,0,[(X4,b,.,0,0,0),(X8,d,.,0,0,0)])," +
        "(0,1,[(X7,c,.,0,1,1)])," +
        "(1,1,[(X1,a,.,1,1,1)])," +
        "(100,1,[(X5,b,.,100,1,1)])," +
        "(1000,2,[(X6,_,.,1000,2,3)])," +
        "(10000,1,[(X9,c,.,10000,1,1)])," +
        "(28,4,[(X3,a,.,28,4,4)])," +
        "(3,1,[(X2,a,.,3,1,1)])" +
        "]");

    view.set(CProp.VISIBLE, inwFS, false);
    expectResult(view, true, "[" +
        "(.,0,[(X4,b,.,0,0,0),(X8,d,.,0,0,0)])," +
        "(.,1,[(X1,a,.,1,1,1),(X2,a,.,3,1,1),(X5,b,.,100,1,1),(X7,c,.,0,1,1),(X9,c,.,10000,1,1)])," +
        "(.,2,[(X6,_,.,1000,2,3)])," +
        "(.,4,[(X3,a,.,28,4,4)])" +
        "]");
  }

  @Test
  public void testTemporalFields() {
    createTable("tabX", new Object[] {
      "pk_colA", "colB_n",
      String.class, LocalTime.class,
      "A1", "10:23:00",
      "A2", "09:02:00",
      "A3", "00:00:00",
      "A4", "23:59:00",
      "A5", null,
      "A6", "12:00:00",
      "A7", "23:59:59.999999",
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath colC = addSQLColumn(view, Path.ROOT, "colB_n");
    if (false) {
      expectResult(view, true,
          "[(A1,B1,1),(A2,B2,1),(A3,B3,1),(A4,B4,2),(A5,B5,2),(A6,B6,_),(A7,B7,6)]");
    }
  }

  @Test
  public void testEmptySetVariations() {
    // Outer
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_Yid_n", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "foo",
      "X2", "a", "bar",
      "X3", "a", "baz",
      "X4", "b", "bar",
      "X5", "b", "bleh",
      "X6", null, "bloh",
      "X7", "c", "blih",
      "X8", "d", "blah",
      "X9", "e", "zaa",
    });
    // Inner
    createTable("tabY", new Object[] {
      "pk_colYid", "colYid2", "colYB_n",
      String.class, String.class, String.class,
      "a", "a", "AA",
      "b", "b", "BB",
      "c", "c", null,
      "d", "d", "D1D1",
      "d2", "d", "D2D2"
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB_Yid = addSQLColumn(view, Path.ROOT, "colXB_Yid_n");
    PrimitivePath colXC = addSQLColumn(view, Path.ROOT, "colXC");
    RelationPath tabYPath = addRelation(view, Path.ROOT, "tabY");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabYPath, false);
    view.set(CProp.VISIBLE, tabYPath, true);
    setInstantiatedTable(view, tabYPath, "tabY");
    PrimitivePath colYidK = addSQLColumn(view, tabYPath, "pk_colYid");
    PrimitivePath colYid2 = addSQLColumn(view, tabYPath, "colYid2");
    PrimitivePath colYB = addSQLColumn(view, tabYPath, "colYB_n");

    // Provably non-plural child relation.
    view.set(QProp.JOINED_ON, colYidK, OptionalPrimitivePath.of(colXB_Yid));
    view.set(QProp.JOINED_ON, colYid2, OptionalPrimitivePath.absent());
    expectResult(view, true, "[" +
        "(X1,a,foo,[(a,a,AA)])," +
        "(X2,a,bar,[(a,a,AA)])," +
        "(X3,a,baz,[(a,a,AA)])," +
        "(X4,b,bar,[(b,b,BB)])," +
        "(X5,b,bleh,[(b,b,BB)])," +
        "(X6,_,bloh,[])," +
        "(X7,c,blih,[(c,c,_)])," +
        "(X8,d,blah,[(d,d,D1D1)])," +
        "(X9,e,zaa,[])" +
        "]");

    // Plural child relation.
    view.set(QProp.JOINED_ON, colYidK, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colYid2, OptionalPrimitivePath.of(colXB_Yid));
    expectResult(view, true, "[" +
        "(X1,a,foo,[(a,a,AA)])," +
        "(X2,a,bar,[(a,a,AA)])," +
        "(X3,a,baz,[(a,a,AA)])," +
        "(X4,b,bar,[(b,b,BB)])," +
        "(X5,b,bleh,[(b,b,BB)])," +
        "(X6,_,bloh,[])," +
        "(X7,c,blih,[(c,c,_)])," +
        "(X8,d,blah,[(d,d,D1D1),(d2,d,D2D2)])," +
        "(X9,e,zaa,[])" +
        "]");

    view.set(CProp.VISIBLE, colYidK, false);
    view.set(CProp.VISIBLE, colYid2, false);


    // Provably non-plural child relation.
    view.set(QProp.JOINED_ON, colYidK, OptionalPrimitivePath.of(colXB_Yid));
    view.set(QProp.JOINED_ON, colYid2, OptionalPrimitivePath.absent());
    expectResult(view, true, "[" +
        "(X1,a,foo,[(.,.,AA)])," +
        "(X2,a,bar,[(.,.,AA)])," +
        "(X3,a,baz,[(.,.,AA)])," +
        "(X4,b,bar,[(.,.,BB)])," +
        "(X5,b,bleh,[(.,.,BB)])," +
        "(X6,_,bloh,[])," + // Empty child relation.
        "(X7,c,blih,[(.,.,_)])," + // Child relation with single all-null tuple.
        "(X8,d,blah,[(.,.,D1D1)])," +
        "(X9,e,zaa,[])" +
        "]");

    // Plural child relation.
    view.set(QProp.JOINED_ON, colYidK, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colYid2, OptionalPrimitivePath.of(colXB_Yid));
    expectResult(view, true, "[" +
        "(X1,a,foo,[(.,.,AA)])," +
        "(X2,a,bar,[(.,.,AA)])," +
        "(X3,a,baz,[(.,.,AA)])," +
        "(X4,b,bar,[(.,.,BB)])," +
        "(X5,b,bleh,[(.,.,BB)])," +
        "(X6,_,bloh,[])," + // Empty child relation.
        "(X7,c,blih,[(.,.,_)])," + // Child relation with single all-null tuple.
        "(X8,d,blah,[(.,.,D1D1),(.,.,D2D2)])," +
        "(X9,e,zaa,[])" +
        "]");

    view.set(CProp.VISIBLE, colYB, false);
    String res1 = "[" +
        "(X1,a,foo,[(.,.,.)])," +
        "(X2,a,bar,[(.,.,.)])," +
        "(X3,a,baz,[(.,.,.)])," +
        "(X4,b,bar,[(.,.,.)])," +
        "(X5,b,bleh,[(.,.,.)])," +
        "(X6,_,bloh,[])," +
        "(X7,c,blih,[(.,.,.)])," +
        "(X8,d,blah,[(.,.,.)])," + // Collapsed into a single tuple in the plural case.
        "(X9,e,zaa,[])" +
        "]";

    // Provably non-plural child relation.
    view.set(QProp.JOINED_ON, colYidK, OptionalPrimitivePath.of(colXB_Yid));
    view.set(QProp.JOINED_ON, colYid2, OptionalPrimitivePath.absent());
    expectResult(view, true, res1);
    // Plural child relation.
    view.set(QProp.JOINED_ON, colYidK, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colYid2, OptionalPrimitivePath.of(colXB_Yid));
    expectResult(view, true, res1);
  }

  @Test
  public void testAggregateHiddenFilter() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");

    view.set(CProp.VISIBLE, tabZ, false);
    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));

    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "fs", "sum([colZC_n])");

    expectResult(view, true, "[" +
        "(X1,a,.,1)," +
        "(X2,a,.,3)," +
        "(X3,a,.,28)," +
        "(X4,b,.,0)," +
        "(X5,b,.,100)," +
        "(X6,_,.,1000)," +
        "(X7,c,.,0)," +
        "(X8,d,.,0)," +
        "(X9,c,.,10000)" +
        "]");
    view.set(CProp.VISIBLE, colXA, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true, "[" +
        "(.,_,.,1000)," +
        "(.,a,.,32)," +
        "(.,b,.,100)," +
        "(.,c,.,10000)," +
        "(.,d,.,0)" +
        "]");
    setFilter(view, colXA, "not ([pk_colXA] is 'X3')");
    PrimitivePath colG = addFormulaColumn(view, Path.ROOT, "fg", "if([pk_colXA] is 'X3','y','n')");
    expectResult(view, true, "[" +
        "(.,_,.,1000,n)," +
        "(.,a,.,4,n)," +
        "(.,b,.,100,n)," +
        "(.,c,.,10000,n)," +
        "(.,d,.,0,n)" +
        "]");
    view.removePath(colG);
    expectResult(view, true, "[" +
        "(.,_,.,1000)," +
        /* Bug in the past: When filtered column is hidden and not referenced, aggregate was
        computed as if the filter was not active. */
        "(.,a,.,4)," +
        "(.,b,.,100)," +
        "(.,c,.,10000)," +
        "(.,d,.,0)" +
        "]");
  }

  @Test
  public void testLateralDecorrelation() {
    createTable("tabdep", new Object[] {
      "pk_did", "code",
      Integer.class, String.class,
      21, "COS",
      52, "JRN"
    });
    createTable("tabcc", new Object[] {
      "pk_ccid", "did", "cnum",
      Integer.class, Integer.class, String.class,
      1, 21, "c101",
      2, 21, "c126",
      3, 21, "c226",
      4, 21, "c320",
      5, 52, "j101",
      6, 52, "j240",
      7, 52, "j452",
      8, 52, "j453",
    });
    // TODO[pd]: Make some of these expectResultWrapped

    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabdep");
    PrimitivePath colDepPK = addSQLColumn(view, Path.ROOT, "pk_did");
    PrimitivePath colDepCode = addSQLColumn(view, Path.ROOT, "code");

    RelationPath tabCC = addRelation(view, Path.ROOT, "tabcc");
    setInstantiatedTable(view, tabCC, "tabcc");
    PrimitivePath colCCPK = addSQLColumn(view, tabCC, "pk_ccid");
    PrimitivePath colCCdid = addSQLColumn(view, tabCC, "did");
    PrimitivePath colCCcnum = addSQLColumn(view, tabCC, "cnum");
    view.set(QProp.JOINED_ON, colCCdid, OptionalPrimitivePath.of(colDepPK));

    expectResult(view, true, "[" +
          "(21,COS,[(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)])," +
          "(52,JRN,[(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)])" +
          "]");
    RelationPath tabDep2 = addRelation(view, Path.ROOT, "tabdep2");
    setInstantiatedTable(view, tabDep2, "tabdep");
    PrimitivePath colDep2PK = addSQLColumn(view, tabDep2, "pk_did");
    PrimitivePath colDep2Code = addSQLColumn(view, tabDep2, "code");
    view.set(QProp.JOINED_ON, colDep2PK, OptionalPrimitivePath.of(colDepPK));
    expectResult(view, true, "[" +
          "(21,COS,[(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS)])," +
          "(52,JRN,[(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN)])" +
          "]");
    // [sibling reference] = [lateral reference]
    PrimitivePath colFJ = addFormulaColumn(view, tabDep2, "fj", "[tabdep2\\pk_did] = [tabcc\\did]");
    expectResult(view, true, "[" +
          "(21,COS,[(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS,true)])," +
          "(52,JRN,[(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN,true)])" +
          "]");
    view.set(QProp.JOINED_ON, colCCdid, OptionalPrimitivePath.absent());
    view.set(QProp.JOINED_ON, colDep2PK, OptionalPrimitivePath.absent());
    view.removePath(colDepPK);
    view.removePath(colDepCode);
    setInstantiatedTable(view, Path.ROOT, null);
    setFilter(view, tabDep2, "[fj] is true");
    PrimitivePath colInw = addFormulaColumn(view, Path.ROOT, "inw", "[tabdep2\\pk_did]");
    setFormula(view, colFJ, "[inw] = [tabcc\\did]");
    if (!SKIP_KNOWN_FAILING_TESTS) {
      /* Currently causes a stack overflow due to a circular reference. Set the correct expected
      result here once I properly detect the error. */
      expectResult(view, true, "[" +
            "([(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS,true)],21)," +
            "([(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN,true)],52)" +
            "]");
    }
    setFormula(view, colInw, "[tabcc\\did]");
    // [sibling reference] = [outward reference]
    setFormula(view, colFJ, "[tabdep2\\pk_did] = [inw]");
    expectResult(view, true, "[" +
          "([(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS,true)],21)," +
          "([(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN,true)],52)" +
          "]");
    view.set(CProp.VISIBLE, colInw, false);
    if (!SKIP_KNOWN_FAILING_TESTS) {
      expectResult(view, true, "[" +
            "([(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS,true)],.)," +
            "([(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN,true)],.)" +
            "]");
      view.set(CProp.VISIBLE, colInw, true);
      setFormula(view, colFJ, "[tabdep2\\pk_did] = [tabcc\\did]");
      expectResult(view, true, "[" +
            "([(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS,true)],21)," +
            "([(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN,true)],52)" +
            "]");
      view.set(CProp.VISIBLE, colInw, false);
      expectResult(view, true, "[" +
            "([(1,21,c101),(2,21,c126),(3,21,c226),(4,21,c320)],[(21,COS,true)],.)," +
            "([(5,52,j101),(6,52,j240),(7,52,j452),(8,52,j453)],[(52,JRN,true)],.)" +
            "]");
    }
  }

  @Test
  public void testSortOnFloatingPointAggregateWithPotentialRoundingError() {
    createTable("associations", new Object[] {
      "pk_aid", "name",
      Integer.class, String.class,
      1, "Gevo"
    });
    // Make values unique to try to avoid dependency on DISTINCT elimination strategy.
    createTable("lobbying", new Object[] {
      "amount", "client", "pk_lid",
      Double.class, String.class, String.class,
      1000.000000000000000, "Gevo", "00",
      1000.100000000000000, "Gevo", "01",
      1029.240729114132591, "Gevo", "02",
      1029.340729114132591, "Gevo", "03",
      1029.440729114132591, "Gevo", "04",
      1029.540729114132591, "Gevo", "05",
      1046.014163031767339, "Gevo", "06",
      1046.114163031767339, "Gevo", "07",
      1046.214163031767339, "Gevo", "08",
      1046.314163031767339, "Gevo", "09",
      1073.502732064453085, "Gevo", "10",
      1073.602732064453085, "Gevo", "11",
      1073.702732064453085, "Gevo", "12",
         0.000000001581418, "Gevo", "13",
         0.000000000123455, "Gevo", "14",
         0.000000000075213, "Gevo", "15",
         0.000000000004127, "Gevo", "16",
         0.000000000000321, "Gevo", "17",
         0.000000000000031, "Gevo", "18",
         0.000000000000003, "Gevo", "19"
    });

    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "associations");
    PrimitivePath colAkey = addSQLColumn(view, Path.ROOT, "pk_aid");
    view.set(CProp.LABEL, colAkey, "pkey");
    view.set(CProp.VISIBLE, colAkey, false);
    PrimitivePath colAname = addSQLColumn(view, Path.ROOT, "name");

    expectResult(view, true, "[(1,Gevo)]");

    RelationPath tabL = view.addRelationPath(Path.ROOT);
    setInstantiatedTable(view, tabL, "lobbying");
    PrimitivePath colLkey = addSQLColumn(view, tabL, "pk_lid");
    PrimitivePath colLamount = addSQLColumn(view, tabL, "amount");
    PrimitivePath colLclient = addSQLColumn(view, tabL, "client");

    view.set(QProp.JOINED_ON, colLclient, OptionalPrimitivePath.of(colAname));
    view.set(CProp.VISIBLE, colLamount, false);
    view.set(CProp.VISIBLE, colLclient, false);

    expectResult(view, true, "[(1,Gevo,[" +
        "(00,.,.),(01,.,.),(02,.,.),(03,.,.),(04,.,.),(05,.,.),(06,.,.),(07,.,.),(08,.,.),(09,.,.),(10,.,.),(11,.,.),(12,.,.),(13,.,.),(14,.,.),(15,.,.),(16,.,.),(17,.,.),(18,.,.),(19,.,.)" +
        "])]");

    PrimitivePath colSum = addFormulaColumn(view, Path.ROOT, "total", "=sum([amount])");
    expectResult(view, true, "[(1,Gevo,[" +
        "(00,.,.),(01,.,.),(02,.,.),(03,.,.),(04,.,.),(05,.,.),(06,.,.),(07,.,.),(08,.,.),(09,.,.),(10,.,.),(11,.,.),(12,.,.),(13,.,.),(14,.,.),(15,.,.),(16,.,.),(17,.,.),(18,.,.),(19,.,.)" +
        "],13523.127765)]");
    view.set(QProp.ORDER_RANK, colSum, Position.create(new BigDecimal(1)));
    /* Confirmed to break on PostgreSQL 9.3.1.0 (18) in the commit at which the test was introduced.
    KeyMatcher...
      adds key tuple    [13523.127764778741, 1] but then attempts to
      look up key tuple [13523.127764778743, 1].
    Now fixed. */
    expectResult(view, true, "[(1,Gevo,[" +
        "(00,.,.),(01,.,.),(02,.,.),(03,.,.),(04,.,.),(05,.,.),(06,.,.),(07,.,.),(08,.,.),(09,.,.),(10,.,.),(11,.,.),(12,.,.),(13,.,.),(14,.,.),(15,.,.),(16,.,.),(17,.,.),(18,.,.),(19,.,.)" +
        "],13523.127765)]");
  }

  @Test
  public void testReferenceExampleFigure() {
    createTable("companies", new Object[] {
      "pk_cid", "name",
      Integer.class, String.class,
      1, "BFC Financial",
      2, "Light Options",
      3, "SG Blocks LLC",
      4, "EuroLink AS",
      5, "EFA Processing"
    });
    createTable("lobbying", new Object[] {
      "pk_lid", "company", "year", "type", "amount",
      String.class, String.class, String.class, String.class, Integer.class,
      "01", "Light Options" , 2008, "q1" , 10000,
      "02", "SG Blocks LLC" , 2010, "q2t", 10000,
      "03", "SG Blocks LLC" , 2010, "q1" , 10000,
      "04", "EFA Processing", 2010, "q4" , 45000,
      "05", "EFA Processing", 2011, "q1t", 15000,
      "06", "EFA Processing", 2010, "q3" , 45000,
      "07", "EFA Processing", 2010, "q2" , 15000,
      "08", "BFC Financial" , 2009, "q2t", 20000,
      "09", "BFC Financial" , 2009, "q1" , 10000
    });


    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "companies");
    PrimitivePath colCid = addSQLColumn(view, Path.ROOT, "pk_cid");
    PrimitivePath colCname = addSQLColumn(view, Path.ROOT, "name");

    RelationPath tabL = addRelation(view, Path.ROOT, "lobbying");
    setInstantiatedTable(view, tabL, "lobbying");
    PrimitivePath colLid = addSQLColumn(view, tabL, "pk_lid");
    PrimitivePath colLcomp = addSQLColumn(view, tabL, "company");
    PrimitivePath colLyear = addSQLColumn(view, tabL, "year");
    PrimitivePath colLtype = addSQLColumn(view, tabL, "type");
    PrimitivePath colLamount = addSQLColumn(view, tabL, "amount");

    view.set(QProp.JOINED_ON, colLcomp, OptionalPrimitivePath.of(colCname));
    view.set(CProp.VISIBLE, colLcomp, false);
    PrimitivePath colTotals = addFormulaColumn(view, Path.ROOT, "total", "sum([amount])");

    expectResult(view, true, "[" +
        "(1,BFC Financial,[(08,.,2009,q2t,20000),(09,.,2009,q1,10000)],30000)," +
        "(2,Light Options,[(01,.,2008,q1,10000)],10000)," +
        "(3,SG Blocks LLC,[(02,.,2010,q2t,10000),(03,.,2010,q1,10000)],20000)," +
        "(4,EuroLink AS,[],0)," +
        "(5,EFA Processing,[(04,.,2010,q4,45000),(05,.,2011,q1t,15000),(06,.,2010,q3,45000),(07,.,2010,q2,15000)],120000)" +
        "]");

    PrimitivePath colYearRef = addFormulaColumn(view, Path.ROOT, "Year Ref", "[year]");
    // Second part of reference example figure.
    expectResult(view, true, "[" +
        "(1,BFC Financial,[(08,.,2009,q2t,20000),(09,.,2009,q1,10000)],30000,2009)," +
        "(2,Light Options,[(01,.,2008,q1,10000)],10000,2008)," +
        "(3,SG Blocks LLC,[(02,.,2010,q2t,10000),(03,.,2010,q1,10000)],20000,2010)," +
        "(4,EuroLink AS,[],0,_)," +
        "(5,EFA Processing,[(04,.,2010,q4,45000),(06,.,2010,q3,45000),(07,.,2010,q2,15000)],105000,2010)," +
        "(5,EFA Processing,[(05,.,2011,q1t,15000)],15000,2011)" +
        "]");
    view.removePath(colYearRef);

    PrimitivePath colTotalsIf2010 = addFormulaColumn(view, Path.ROOT, "total 2010", "sum(if([year]='2010',[amount],null))");
    expectResult(view, true, "[" +
        "(1,BFC Financial,[(08,.,2009,q2t,20000),(09,.,2009,q1,10000)],30000,0)," +
        "(2,Light Options,[(01,.,2008,q1,10000)],10000,0)," +
        "(3,SG Blocks LLC,[(02,.,2010,q2t,10000),(03,.,2010,q1,10000)],20000,20000)," +
        "(4,EuroLink AS,[],0,0)," +
        "(5,EFA Processing,[(04,.,2010,q4,45000),(05,.,2011,q1t,15000),(06,.,2010,q3,45000),(07,.,2010,q2,15000)],120000,105000)" +
        "]");

    PrimitivePath colPerc = addFormulaColumn(view, tabL, "perc", "(100.0 * [amount]) / [total]");
    expectResult(view, true, "[" +
        "(1,BFC Financial,[(08,.,2009,q2t,20000,66.666667),(09,.,2009,q1,10000,33.333333)],30000,0)," +
        "(2,Light Options,[(01,.,2008,q1,10000,100)],10000,0)," +
        "(3,SG Blocks LLC,[(02,.,2010,q2t,10000,50),(03,.,2010,q1,10000,50)],20000,20000)," +
        "(4,EuroLink AS,[],0,0)," +
        "(5,EFA Processing,[(04,.,2010,q4,45000,37.5),(05,.,2011,q1t,15000,12.5),(06,.,2010,q3,45000,37.5),(07,.,2010,q2,15000,12.5)],120000,105000)" +
        "]");
    colYearRef = addFormulaColumn(view, Path.ROOT, "Year Ref", "[year]");
    if (!SKIP_KNOWN_FAILING_TESTS) {
      /* With all optimizations turned on, used to trigger a "duplicate key" error (though not
      anymore):
           Duplicate key [5, "2010", "04"] in map {[1, "2009", "08"]=lobbying#0, [1, "2009", "09"]=lobbying#1, [2, "2008", "01"]=lobbying#2, [3, "2010", "02"]=lobbying#3, [3, "2010", "03"]=lobbying#4, [5, "2010", "04"]=lobbying#5}
         With no optimizations turned on, triggers a "Circular dependency in left join conditions"
         error.
      */
      expectResult(view, true, "[" +
          "(1,BFC Financial,[(08,.,2009,q2t,20000,66.666667),(09,.,2009,q1,10000,33.333333)],30000,0,2009)," +
          "(2,Light Options,[(01,.,2008,q1,10000,100)],10000,0,2008)," +
          "(3,SG Blocks LLC,[(02,.,2010,q2t,10000,50),(03,.,2010,q1,10000,50)],20000,20000,2010)," +
          "(4,EuroLink AS,[],0,0,_)," +
          "(5,EFA Processing,[(04,.,2010,q4,45000,42.857143),(06,.,2010,q3,45000,42.857143),(07,.,2010,q2,15000,14.285714)],105000,105000,2010)," +
          "(5,EFA Processing,[(05,.,2011,q1t,15000,100)],15000,0,2011)" +
          "]");
      view.set(QProp.EXISTENTIAL_FILTER, tabL, true);
      expectResult(view, true, "[" +
          "(1,BFC Financial,[(08,.,2009,q2t,20000,66.666667),(09,.,2009,q1,10000,33.333333)],30000,0,2009)," +
          "(2,Light Options,[(01,.,2008,q1,10000,100)],10000,0,2008)," +
          "(3,SG Blocks LLC,[(02,.,2010,q2t,10000,50),(03,.,2010,q1,10000,50)],20000,20000,2010)," +
          "(5,EFA Processing,[(04,.,2010,q4,45000,42.857143),(06,.,2010,q3,45000,42.857143),(07,.,2010,q2,15000,14.285714)],105000,105000,2010)," +
          "(5,EFA Processing,[(05,.,2011,q1t,15000,100)],15000,0,2011)" +
          "]");
      view.set(QProp.EXISTENTIAL_FILTER, tabL, false);
    }
  }

  @Test
  public void testAggregateOnLateralReference() {
    createTable("tinstructors", new Object[] {
      "pk_colIid", "colIfirst", "colIlast",
      String.class, String.class, String.class,
      "I00", "Aaro", "Wil",
      "I01", "Aaro", "Dav",
      "I02", "Aaro", "Lou",
      "I03", "Abby", "Sar",
      "I04", "Abig", "Tan",
      "I05", "Abig", "Hug",
      "I06", "Aaro", "Dav", // Two Aaaro Dav, different ID.
    });
    createTable("tsections", new Object[] {
      "pk_colSid", "colSiid",
      String.class, String.class,
      "S00", "I00",
      "S01", "I00",
      "S02", "I01",
      "S03", "I01",
      "S04", "I01",
      "S05", "I03",
      "S06", "I03",
      "S07", "I04",
      "S08", "I04",
      "S09", "I05",
    });

    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tinstructors");
    PrimitivePath colIid = addSQLColumn(view, Path.ROOT, "pk_colIid");
    PrimitivePath colIfirst = addSQLColumn(view, Path.ROOT, "colIfirst");
    PrimitivePath colIlast = addSQLColumn(view, Path.ROOT, "colIlast");

    RelationPath tabS = addRelation(view, Path.ROOT, "sections");
    setInstantiatedTable(view, tabS, "tsections");
    PrimitivePath colSid = addSQLColumn(view, tabS, "pk_colSid");
    PrimitivePath colSiid = addSQLColumn(view, tabS, "colSiid");
    view.set(CProp.VISIBLE, colSiid, false);

    view.set(QProp.JOINED_ON, colSiid, OptionalPrimitivePath.of(colIid));

    expectResult(view, true, "[" +
        "(I00,Aaro,Wil,[(S00,.),(S01,.)])," +
        "(I01,Aaro,Dav,[(S02,.),(S03,.),(S04,.)])," +
        "(I02,Aaro,Lou,[])," +
        "(I03,Abby,Sar,[(S05,.),(S06,.)])," +
        "(I04,Abig,Tan,[(S07,.),(S08,.)])," +
        "(I05,Abig,Hug,[(S09,.)])," +
        "(I06,Aaro,Dav,[])" +
        "]");
    view.set(CProp.VISIBLE, colIid, false);
    view.set(CProp.VISIBLE, colIlast, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);

    RelationPath tabI2 = addRelation(view, Path.ROOT, "instructors2");
    setInstantiatedTable(view, tabI2, "tinstructors");
    PrimitivePath colI2id = addSQLColumn(view, tabI2, "pk_colIid");
    PrimitivePath colI2first = addSQLColumn(view, tabI2, "colIfirst");
    PrimitivePath colI2last = addSQLColumn(view, tabI2, "colIlast");
    view.set(CProp.VISIBLE, colI2first, false);
    view.set(QProp.JOINED_ON, colI2id, OptionalPrimitivePath.of(colIid));

    expectResult(view, true, "[" +
        "(.,Aaro,.,[(S00,.),(S01,.),(S02,.),(S03,.),(S04,.)],[(I00,.,Wil),(I01,.,Dav),(I02,.,Lou),(I06,.,Dav)])," +
        "(.,Abby,.,[(S05,.),(S06,.)],[(I03,.,Sar)])," +
        "(.,Abig,.,[(S07,.),(S08,.),(S09,.)],[(I04,.,Tan),(I05,.,Hug)])" +
        "]");

    addFormulaColumn(view, tabI2, "lateral section count", "count([pk_colSid])");
    if (!SKIP_KNOWN_FAILING_TESTS) {
      expectResult(view, true, "[" +
          "(.,Aaro,.,[(S00,.),(S01,.),(S02,.),(S03,.),(S04,.)],[(I00,.,Wil,5),(I01,.,Dav,5),(I02,.,Lou,5),(I06,.,Dav,5)])," +
          "(.,Abby,.,[(S05,.),(S06,.)],[(I03,.,Sar,2)])," +
          "(.,Abig,.,[(S07,.),(S08,.),(S09,.)],[(I04,.,Tan,3),(I05,.,Hug,3)])" +
          "]");
      // TODO[te]: Try doing the count of colSiid instead of pk_colSid.
      // TODO[te]: Try collecting would-be aggregate input tuples in an empty child relation instead.
    }
    if (!SKIP_KNOWN_FAILING_TESTS) {
      view.set(CProp.VISIBLE, colI2id, false);
      view.set(QProp.INCLUDE_PRIMARY_KEY, tabI2, false);
      expectResult(view, true, "[" +
          "(.,Aaro,.,[(S00,.),(S01,.),(S02,.),(S03,.),(S04,.)],[(.,.,Dav,5),(.,.,Lou,5),(.,.,Wil,5)])," +
          "(.,Abby,.,[(S05,.),(S06,.)],[(.,.,Sar,2)])," +
          "(.,Abig,.,[(S07,.),(S08,.),(S09,.)],[(.,.,Hug,3),(.,.,Tan,3)])" +
          "]");
    }
  }

  @Test
  public void testSortOnAggregate() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZC = addSQLColumn(view, tabZ, "colZC_n");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));
    view.set(CProp.VISIBLE, tabZ, true);

    PrimitivePath colF = addFormulaColumn(view, Path.ROOT, "fs", "sum([colZC_n])");

    // Expect a specific ordering here.
    expectResult(view, false, "[" +
        "(X1,a,[(Z1,X1,1)],1)," +
        "(X2,a,[(Z2,X2,3)],3)," +
        "(X3,a,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)],28)," +
        "(X4,b,[],0)," +
        "(X5,b,[(Z5,X5,100)],100)," +
        "(X6,_,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)],1000)," +
        "(X7,c,[(Z12,X7,0)],0)," +
        "(X8,d,[],0)," +
        "(X9,c,[(Z6,X9,10000)],10000)" +
        "]");
    view.set(QProp.ORDER_DESCENDING, colF, false);
    view.set(QProp.ORDER_RANK, colF, Position.create(BigDecimal.ONE));
    view.set(CProp.VISIBLE, tabZ, false);
    expectResult(view, false, "[" +
        "(X4,b,.,0)," +
        "(X7,c,.,0)," +
        "(X8,d,.,0)," +
        "(X1,a,.,1)," +
        "(X2,a,.,3)," +
        "(X3,a,.,28)," +
        "(X5,b,.,100)," +
        "(X6,_,.,1000)," +
        "(X9,c,.,10000)" +
        "]");
    view.set(CProp.VISIBLE, tabZ, true);
    expectResult(view, false, "[" +
        "(X4,b,[],0)," +
        "(X7,c,[(Z12,X7,0)],0)," +
        "(X8,d,[],0)," +
        "(X1,a,[(Z1,X1,1)],1)," +
        "(X2,a,[(Z2,X2,3)],3)," +
        "(X3,a,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)],28)," +
        "(X5,b,[(Z5,X5,100)],100)," +
        "(X6,_,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)],1000)," +
        "(X9,c,[(Z6,X9,10000)],10000)" +
        "]");
  }

  @Test
  public void testNullOrEmptyStringDistinctionInKey() {
    createTable("tworows", new Object[] {
      "pk_colA",
      Integer.class,
      1, 2
    });
    MutableView view = createView();

    setInstantiatedTable(view, Path.ROOT, "tworows");
    PrimitivePath colA = addSQLColumn(view, Path.ROOT, "pk_colA");
    addFormulaColumn(view, Path.ROOT, "fx", "if([pk_colA] = 1, '', null)");
    view.set(CProp.VISIBLE, colA, false);
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    expectResult(view, true, "[(.,),(.,_)]");
  }

  @Test
  public void testSynthesizedUnionUnpivot() {
    createTable("threerows", new Object[] {
      "pk_case",
      Integer.class,
      1, 2, 3
    });
    createTable("ucourses", new Object[] {
      "pk_id", "title", "dept1", "cnum1", "dept2_n", "cnum2_n", "dept3_n", "cnum3_n",
      Integer.class, String.class, String.class, String.class, String.class, String.class, String.class, String.class,
      1, "Huma", "ANT", "206", "EEB", "306", "GEO", "208",
      2, "Math", "APC", "199", "MAT", "199", null, null,
      3, "Home", "CLG", "108", null, null, null, null,
      4, "Peac", "WWS", "313", "POL", "387", null, null
    });
    MutableView view = createView();

    setInstantiatedTable(view, Path.ROOT, "ucourses");
    PrimitivePath colCid = addSQLColumn(view, Path.ROOT, "pk_id");
    PrimitivePath colCtitle = addSQLColumn(view, Path.ROOT, "title");
    PrimitivePath colCdept1 = addSQLColumn(view, Path.ROOT, "dept1");
    PrimitivePath colCcnum1 = addSQLColumn(view, Path.ROOT, "cnum1");
    PrimitivePath colCdept2 = addSQLColumn(view, Path.ROOT, "dept2_n");
    PrimitivePath colCcnum2 = addSQLColumn(view, Path.ROOT, "cnum2_n");
    PrimitivePath colCdept3 = addSQLColumn(view, Path.ROOT, "dept3_n");
    PrimitivePath colCcnum3 = addSQLColumn(view, Path.ROOT, "cnum3_n");
    view.set(CProp.VISIBLE, colCdept1, false);
    view.set(CProp.VISIBLE, colCcnum1, false);
    view.set(CProp.VISIBLE, colCdept2, false);
    view.set(CProp.VISIBLE, colCcnum2, false);
    view.set(CProp.VISIBLE, colCdept3, false);
    view.set(CProp.VISIBLE, colCcnum3, false);
    expectResult(view, true, "[" +
        "(1,Huma,.,.,.,.,.,.)," +
        "(2,Math,.,.,.,.,.,.)," +
        "(3,Home,.,.,.,.,.,.)," +
        "(4,Peac,.,.,.,.,.,.)" +
        "]");
    RelationPath tabCC = addRelation(view, Path.ROOT, "CC");
    RelationPath tabU = addRelation(view, tabCC, "threerows");
    PrimitivePath colUcase = addSQLColumn(view, tabU, "pk_case");
    setInstantiatedTable(view, tabU, "threerows");
    PrimitivePath colDept = addFormulaColumn(view, tabCC, "dept", "if([pk_case] = 1, [dept1], if([pk_case] = 2, [dept2_n], [dept3_n]))");
    PrimitivePath colCnum = addFormulaColumn(view, tabCC, "cnum", "if([pk_case] = 1, [cnum1], if([pk_case] = 2, [cnum2_n], [cnum3_n]))");
    view.set(CProp.VISIBLE, tabU, false);
    expectResult(view, true, "[" +
        "(1,Huma,.,.,.,.,.,.,[(.,ANT,206),(.,EEB,306),(.,GEO,208)])," +
        "(2,Math,.,.,.,.,.,.,[(.,APC,199),(.,MAT,199),(.,_,_)])," +
        "(3,Home,.,.,.,.,.,.,[(.,CLG,108),(.,_,_)])," +
        "(4,Peac,.,.,.,.,.,.,[(.,POL,387),(.,WWS,313),(.,_,_)])" +
        "]");
    setFilter(view, colDept, "not ([dept] is null)");
    expectResult(view, true, "[" +
        "(1,Huma,.,.,.,.,.,.,[(.,ANT,206),(.,EEB,306),(.,GEO,208)])," +
        "(2,Math,.,.,.,.,.,.,[(.,APC,199),(.,MAT,199)])," +
        "(3,Home,.,.,.,.,.,.,[(.,CLG,108)])," +
        "(4,Peac,.,.,.,.,.,.,[(.,POL,387),(.,WWS,313)])" +
        "]");
  }

  @Test
  public void testSynthesizedUnionTwoTab() {

    createTable("tworows", new Object[] {
      "pk_case",
      Integer.class,
      1, 2
    });
    // Like on testManyToManyNest2, but with tabZ split in two.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ_part1", new Object[] {
      "pk_colZA", "pk_colZB",
      String.class, String.class,
      "Z1", "X1",
      "Z2", "X2",
      "Z3", "X3"
    });
    createTable("tabZ_part2", new Object[] {
      "pk_colZA", "pk_colZB",
      String.class, String.class,
      "Z4", "X3",
      "Z5", "X5",
      "Z6", "X6",
      "Z6", "X9",
      "Z7", "XX"
    });
    createTable("tabQ", new Object[] {
      "pk_colQA",
      Integer.class,
      1,
      2
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colXA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colXB = addSQLColumn(view, Path.ROOT, "colXB_n");

    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");

    RelationPath tabZp1 = addRelation(view, tabZ, "tabZ_part1");
    RelationPath tabZp2 = addRelation(view, tabZ, "tabZ_part2");
    RelationPath tabU = addRelation(view, tabZ, "tabU");
    setInstantiatedTable(view, tabU, "tworows");
    PrimitivePath colUcase = addSQLColumn(view, tabU, "pk_case");
    view.set(CProp.VISIBLE, tabU, false);
    setInstantiatedTable(view, tabZp1, "tabZ_part1");
    PrimitivePath colZAp1 = addSQLColumn(view, tabZp1, "pk_colZA");
    PrimitivePath colZBp1 = addSQLColumn(view,tabZp1, "pk_colZB");
    view.set(CProp.VISIBLE, tabZp1, false);
    setInstantiatedTable(view, tabZp2, "tabZ_part2");
    PrimitivePath colZAp2 = addSQLColumn(view, tabZp2, "pk_colZA");
    PrimitivePath colZBp2 = addSQLColumn(view,tabZp2, "pk_colZB");
    view.set(CProp.VISIBLE, tabZp2, false);

    view.set(CProp.VISIBLE, tabZ, true);
    PrimitivePath colZA = addFormulaColumn(view, tabZ, "pk_colZA", "if([pk_case] = 1, [tabZ_part1\\pk_colZA],[tabZ_part2\\pk_colZA])");
    PrimitivePath colZB = addFormulaColumn(view, tabZ, "pk_colZB", "if([pk_case] = 1, [tabZ_part1\\pk_colZB],[tabZ_part2\\pk_colZB])");

    view.set(QProp.JOINED_ON, colZB, OptionalPrimitivePath.of(colXA));
    RelationPath tabQ = addRelation(view, Path.ROOT, "tabQ");
    setInstantiatedTable(view, tabQ, "tabQ");
    view.set(CProp.VISIBLE, tabQ, false);
    PrimitivePath colQA = addSQLColumn(view, tabQ, "pk_colQA");
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.,Z1,X1)],.)," +
        "(X2,a,[(.,.,.,Z2,X2)],.)," +
        "(X3,a,[(.,.,.,Z3,X3),(.,.,.,Z4,X3)],.)," +
        "(X4,b,[],.)," +
        "(X5,b,[(.,.,.,Z5,X5)],.)," +
        "(X6,_,[(.,.,.,Z6,X6)],.)," +
        "(X7,c,[],.)," +
        "(X8,d,[],.)," +
        "(X9,c,[(.,.,.,Z6,X9)],.)" +
        "]");
    PrimitivePath fx = addFormulaColumn(view, Path.ROOT, "fx", "[pk_colQA]");
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.,Z1,X1)],.,1)," +
        "(X1,a,[(.,.,.,Z1,X1)],.,2)," +
        "(X2,a,[(.,.,.,Z2,X2)],.,1)," +
        "(X2,a,[(.,.,.,Z2,X2)],.,2)," +
        "(X3,a,[(.,.,.,Z3,X3),(.,.,.,Z4,X3)],.,1)," +
        "(X3,a,[(.,.,.,Z3,X3),(.,.,.,Z4,X3)],.,2)," +
        "(X4,b,[],.,1)," +
        "(X4,b,[],.,2)," +
        "(X5,b,[(.,.,.,Z5,X5)],.,1)," +
        "(X5,b,[(.,.,.,Z5,X5)],.,2)," +
        "(X6,_,[(.,.,.,Z6,X6)],.,1)," +
        "(X6,_,[(.,.,.,Z6,X6)],.,2)," +
        "(X7,c,[],.,1)," +
        "(X7,c,[],.,2)," +
        "(X8,d,[],.,1)," +
        "(X8,d,[],.,2)," +
        "(X9,c,[(.,.,.,Z6,X9)],.,1)," +
        "(X9,c,[(.,.,.,Z6,X9)],.,2)" +
        "]");
    PrimitivePath fxI = addFormulaColumn(view, tabZ, "fxI", "[fx]");
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.,Z1,X1,1)],.,1)," +
        "(X1,a,[(.,.,.,Z1,X1,2)],.,2)," +
        "(X2,a,[(.,.,.,Z2,X2,1)],.,1)," +
        "(X2,a,[(.,.,.,Z2,X2,2)],.,2)," +
        "(X3,a,[(.,.,.,Z3,X3,1),(.,.,.,Z4,X3,1)],.,1)," +
        "(X3,a,[(.,.,.,Z3,X3,2),(.,.,.,Z4,X3,2)],.,2)," +
        "(X4,b,[],.,1)," +
        "(X4,b,[],.,2)," +
        "(X5,b,[(.,.,.,Z5,X5,1)],.,1)," +
        "(X5,b,[(.,.,.,Z5,X5,2)],.,2)," +
        "(X6,_,[(.,.,.,Z6,X6,1)],.,1)," +
        "(X6,_,[(.,.,.,Z6,X6,2)],.,2)," +
        "(X7,c,[],.,1)," +
        "(X7,c,[],.,2)," +
        "(X8,d,[],.,1)," +
        "(X8,d,[],.,2)," +
        "(X9,c,[(.,.,.,Z6,X9,1)],.,1)," +
        "(X9,c,[(.,.,.,Z6,X9,2)],.,2)" +
        "]");
    view.set(CProp.VISIBLE, fx, false);
    expectResult(view, true, "[" +
        "(X1,a,[(.,.,.,Z1,X1,1),(.,.,.,Z1,X1,2)],.,.)," +
        "(X2,a,[(.,.,.,Z2,X2,1),(.,.,.,Z2,X2,2)],.,.)," +
        "(X3,a,[(.,.,.,Z3,X3,1),(.,.,.,Z3,X3,2),(.,.,.,Z4,X3,1),(.,.,.,Z4,X3,2)],.,.)," +
        "(X4,b,[],.,.)," +
        "(X5,b,[(.,.,.,Z5,X5,1),(.,.,.,Z5,X5,2)],.,.)," +
        "(X6,_,[(.,.,.,Z6,X6,1),(.,.,.,Z6,X6,2)],.,.)," +
        "(X7,c,[],.,.)," +
        "(X8,d,[],.,.)," +
        "(X9,c,[(.,.,.,Z6,X9,1),(.,.,.,Z6,X9,2)],.,.)" +
        "]");
  }

  @Test
  public void testSelfJoinInwardOutwardFormula() {
    createTable("tab_simple", new Object[] {
      "pk_colA", "colB_n", "colC",
      String.class, String.class, Integer.class,
      "A1", "B1", 1,
      "A2", "B2", 2,
      "A3", "B3", 3,
      "A4", "B4", 4,
      "A5", "B5", 5,
      "A6", "B6", 6,
      "A7", "B7", 7
    });
    MutableView view = createView();

    view.set(CProp.LABEL, Path.ROOT, "t1");
    setInstantiatedTable(view, Path.ROOT, "tab_simple");
    PrimitivePath col1A = addSQLColumn(view, Path.ROOT, "pk_colA");
    PrimitivePath col1B = addSQLColumn(view, Path.ROOT, "colB_n");
    PrimitivePath col1C = addSQLColumn(view, Path.ROOT, "colC");
    RelationPath t2 = view.addRelationPath(Path.ROOT);
    view.set(CProp.LABEL, t2, "t2");
    setInstantiatedTable(view, t2, "tab_simple");
    PrimitivePath col2A = addSQLColumn(view, t2, "pk_colA");
    PrimitivePath col2B = addSQLColumn(view, t2, "colB_n");
    PrimitivePath col2C = addSQLColumn(view, t2, "colC");
    view.set(QProp.JOINED_ON, col2A, OptionalPrimitivePath.of(col1A));
    RelationPath t3 = view.addRelationPath(t2);
    view.set(CProp.LABEL, t3, "t3");
    setInstantiatedTable(view, t3, "tab_simple");
    PrimitivePath col3A = addSQLColumn(view, t3, "pk_colA");
    PrimitivePath col3B = addSQLColumn(view, t3, "colB_n");
    PrimitivePath col3C = addSQLColumn(view, t3, "colC");
    view.set(QProp.JOINED_ON, col3A, OptionalPrimitivePath.of(col2A));

    expectResult(view, true, "[" +
        "(A1,B1,1,[(A1,B1,1,[(A1,B1,1)])])," +
        "(A2,B2,2,[(A2,B2,2,[(A2,B2,2)])])," +
        "(A3,B3,3,[(A3,B3,3,[(A3,B3,3)])])," +
        "(A4,B4,4,[(A4,B4,4,[(A4,B4,4)])])," +
        "(A5,B5,5,[(A5,B5,5,[(A5,B5,5)])])," +
        "(A6,B6,6,[(A6,B6,6,[(A6,B6,6)])])," +
        "(A7,B7,7,[(A7,B7,7,[(A7,B7,7)])])" +
        "]");
    addFormulaColumn(view, t2, "fx", "concat([t1\\pk_colA],[t3\\pk_colA])");
    expectResult(view, true, "[" +
        "(A1,B1,1,[(A1,B1,1,[(A1,B1,1)],A1A1)])," +
        "(A2,B2,2,[(A2,B2,2,[(A2,B2,2)],A2A2)])," +
        "(A3,B3,3,[(A3,B3,3,[(A3,B3,3)],A3A3)])," +
        "(A4,B4,4,[(A4,B4,4,[(A4,B4,4)],A4A4)])," +
        "(A5,B5,5,[(A5,B5,5,[(A5,B5,5)],A5A5)])," +
        "(A6,B6,6,[(A6,B6,6,[(A6,B6,6)],A6A6)])," +
        "(A7,B7,7,[(A7,B7,7,[(A7,B7,7)],A7A7)])" +
        "]");
    addFormulaColumn(view, t2, "fx", "concat([t1\\pk_colA],'-',count([t3\\pk_colA]))");
    expectResult(view, true, "[" +
        "(A1,B1,1,[(A1,B1,1,[(A1,B1,1)],A1A1,A1-1)])," +
        "(A2,B2,2,[(A2,B2,2,[(A2,B2,2)],A2A2,A2-1)])," +
        "(A3,B3,3,[(A3,B3,3,[(A3,B3,3)],A3A3,A3-1)])," +
        "(A4,B4,4,[(A4,B4,4,[(A4,B4,4)],A4A4,A4-1)])," +
        "(A5,B5,5,[(A5,B5,5,[(A5,B5,5)],A5A5,A5-1)])," +
        "(A6,B6,6,[(A6,B6,6,[(A6,B6,6)],A6A6,A6-1)])," +
        "(A7,B7,7,[(A7,B7,7,[(A7,B7,7)],A7A7,A7-1)])" +
        "]");
  }

  @Test
  public void testAggregateReferencingSiblingAggregate() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });

    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    view.set(CProp.LABEL, Path.ROOT, "tabX");
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_sib = addFormulaColumn(view, Path.ROOT, "sib", "null");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_sib, "avg([agg])");
    setFormula(view, colX_agg, "sum([colZC_n])");
    expectResult(view, true, "["
        + "(X1,a,1,1,[(Z1,X1,1)]),"
        + "(X2,a,3,3,[(Z2,X2,3)]),"
        + "(X3,a,28,28,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,0,0,[]),"
        + "(X5,b,100,100,[(Z5,X5,100)]),"
        + "(X6,_,1000,1000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,0,0,[(Z12,X7,0)]),"
        + "(X8,d,0,0,[]),"
        + "(X9,c,10000,10000,[(Z6,X9,10000)])]");
    view.set(CProp.VISIBLE, colX_agg, false);
    expectResult(view, true, "["
        + "(X1,a,1,.,[(Z1,X1,1)]),"
        + "(X2,a,3,.,[(Z2,X2,3)]),"
        + "(X3,a,28,.,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,0,.,[]),"
        + "(X5,b,100,.,[(Z5,X5,100)]),"
        + "(X6,_,1000,.,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,0,.,[(Z12,X7,0)]),"
        + "(X8,d,0,.,[]),"
        + "(X9,c,10000,.,[(Z6,X9,10000)])]");

    // ----- Test Case ------------------------------------------------
    view.set(CProp.VISIBLE, colX_agg, true);
    view.set(CProp.VISIBLE, colX_sib, false);
    setFormula(view, colX_sib, "[colZC_n]");
    expectResult(view, true, "["
        + "(X1,a,.,1,[(Z1,X1,1)]),"
        + "(X2,a,.,3,[(Z2,X2,3)]),"
        + "(X3,a,.,28,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,.,0,[]),"
        + "(X5,b,.,100,[(Z5,X5,100)]),"
        + "(X6,_,.,1000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,.,0,[(Z12,X7,0)]),"
        + "(X8,d,.,0,[]),"
        + "(X9,c,.,10000,[(Z6,X9,10000)])]");

    setFormula(view, colX_agg, "sum([sib])");
    expectResult(view, true, "["
        + "(X1,a,.,1,[(Z1,X1,1)]),"
        + "(X2,a,.,3,[(Z2,X2,3)]),"
        + "(X3,a,.,26,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])," // 2's collapsed
        + "(X4,b,.,0,[]),"
        + "(X5,b,.,100,[(Z5,X5,100)]),"
        + "(X6,_,.,1000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,.,0,[(Z12,X7,0)]),"
        + "(X8,d,.,0,[]),"
        + "(X9,c,.,10000,[(Z6,X9,10000)])]");
  }

  @Test
  public void testAggregateReferencesToMultipleRelationLevels() {
    createTable("tabQ", new Object[] {
      "pk_colQA", "colQB_n",
      String.class, String.class,
      "Q1", "a",
      "Q2", "b",
      "Q3", null,
      "Q4", "c",
      "Q5", "d",
      "Q6", "e",
    });
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "L",
      "X2", "a", "S",
      "X3", "a", "L",
      "X4", "b", "L",
      "X5", "b", "L",
      "X6", null, "L",
      "X7", "c", "L",
      "X8", "d", "L",
      "X9", "c", "L"
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    createTable("tabP", new Object[] {
      "pk_colPA", "colPB",
      String.class, String.class,
      "P1", "Q1",
      "P2", "Q1",
      "P3", "Q5",
      "P4", "Q6",
      "P5", "Q6"
    });

    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    view.set(CProp.LABEL, Path.ROOT, "tabQ");
    setInstantiatedTable(view, Path.ROOT, "tabQ");
    PrimitivePath colQ_QA = addSQLColumn(view, Path.ROOT, "pk_colQA");
    PrimitivePath colQ_QB = addSQLColumn(view, Path.ROOT, "colQB_n");
    RelationPath tabX = addRelation(view, Path.ROOT, "tabX");
    setInstantiatedTable(view, tabX, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, tabX, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, tabX, "colXB_n");
    view.set(QProp.JOINED_ON, colX_XB, OptionalPrimitivePath.of(colQ_QB));
    PrimitivePath colX_XC = addSQLColumn(view, tabX, "colXC");
    RelationPath tabZ = addRelation(view, tabX, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    expectResult(view, true, "["
        + "(Q1,a,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])]),"
        + "(Q2,b,[(X4,b,L,[]),(X5,b,L,[(Z5,X5,100)])]),"
        + "(Q3,_,[]),"
        + "(Q4,c,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])]),"
        + "(Q5,d,[(X8,d,L,[])]),"
        + "(Q6,e,[])]");
    // ----- Test Case ------------------------------------------------
    PrimitivePath colQ_agg = addFormulaColumn(view, Path.ROOT, "agg", "sum(if([colXC] = 'L', [colZC_n], null))");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colQ_QA, colQ_QB, colQ_agg, tabX)));
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])]),"
        + "(Q2,b,100,[(X4,b,L,[]),(X5,b,L,[(Z5,X5,100)])]),"
        + "(Q3,_,0,[]),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])]),"
        + "(Q5,d,0,[(X8,d,L,[])]),"
        + "(Q6,e,0,[])]");
    // ----- Test Case ------------------------------------------------
    RelationPath tabP = addRelation(view, Path.ROOT, "tabP");
    setInstantiatedTable(view, tabP, "tabP");
    PrimitivePath colP_PA = addSQLColumn(view, tabP, "pk_colPA");
    PrimitivePath colP_PB = addSQLColumn(view, tabP, "colPB");
    view.set(CProp.VISIBLE, colP_PB, false);
    view.set(QProp.JOINED_ON, colP_PB, OptionalPrimitivePath.of(colQ_QA));
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)]),"
        + "(Q2,b,100,[(X4,b,L,[]),(X5,b,L,[(Z5,X5,100)])],[]),"
        + "(Q3,_,0,[],[]),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[]),"
        + "(Q5,d,0,[(X8,d,L,[])],[(P3,.)]),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    setFormula(view, colQ_agg, "sum(if([colXC] = 'L', if([colZC_n] is null, -1000000, [colZC_n]), null))");
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)]),"
        + "(Q2,b,-999900,[(X4,b,L,[]),(X5,b,L,[(Z5,X5,100)])],[]),"
        + "(Q3,_,0,[],[]),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[]),"
        + "(Q5,d,-1000000,[(X8,d,L,[])],[(P3,.)]),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)])"
        + "]");
    view.set(QProp.EXISTENTIAL_FILTER, tabZ, true);
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)]),"
        + "(Q2,b,100,[(X5,b,L,[(Z5,X5,100)])],[]),"
        + "(Q3,_,0,[],[]),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[]),"
        + "(Q5,d,0,[],[(P3,.)]),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)])"
        + "]");
    PrimitivePath colParagg = addFormulaColumn(view, Path.ROOT, "paragg", "count([pk_colPA])");
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)],2),"
        + "(Q2,b,100,[(X5,b,L,[(Z5,X5,100)])],[],0),"
        + "(Q3,_,0,[],[],0),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[],0),"
        + "(Q5,d,0,[],[(P3,.)],1),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)],2)"
        + "]");
    setFormula(view, colParagg, "count([pk_colPA]) + count([pk_colZA])");
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)],8),"
        + "(Q2,b,100,[(X5,b,L,[(Z5,X5,100)])],[],1),"
        + "(Q3,_,0,[],[],0),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[],2),"
        + "(Q5,d,0,[],[(P3,.)],1),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)],2)"
        + "]");
    setFormula(view, colParagg, "count([pk_colXA]) + count([pk_colZA])");
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)],9),"
        + "(Q2,b,100,[(X5,b,L,[(Z5,X5,100)])],[],2),"
        + "(Q3,_,0,[],[],0),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[],4),"
        + "(Q5,d,0,[],[(P3,.)],0),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)],0)"
        + "]");
    setFormula(view, colParagg, "count([pk_colZA]) + sum([colZC_n])");
    expectResult(view, true, "["
        + "(Q1,a,29,[(X1,a,L,[(Z1,X1,1)]),(X2,a,S,[(Z2,X2,3)]),(X3,a,L,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)])],[(P1,.),(P2,.)],38),"
        + "(Q2,b,100,[(X5,b,L,[(Z5,X5,100)])],[],101),"
        + "(Q3,_,0,[],[],0),"
        + "(Q4,c,10000,[(X7,c,L,[(Z12,X7,0)]),(X9,c,L,[(Z6,X9,10000)])],[],10002),"
        + "(Q5,d,0,[],[(P3,.)],0),"
        + "(Q6,e,0,[],[(P4,.),(P5,.)],0)"
        + "]");
  }

  @Test
  public void testJoinOnHiddenInwardKeyReference() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "L",
      "X2", "a", "S",
      "X3", "a", "L",
      "X4", "b", "L",
      "X5", "b", "L",
      "X6", null, "L",
      "X7", "c", "L",
      "X8", "d", "L",
      "X9", "c", "L"
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    PrimitivePath colU_inr = addFormulaColumn(view, Path.ROOT, "inr", "null");
    RelationPath tabX = addRelation(view, Path.ROOT, "tabX");
    setInstantiatedTable(view, tabX, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, tabX, "pk_colXA");
    view.set(CProp.VISIBLE, colX_XA, false);
    view.set(QProp.JOINED_ON, colX_XA, OptionalPrimitivePath.of(colU_inr));
    PrimitivePath colX_XB = addSQLColumn(view, tabX, "colXB_n");
    PrimitivePath colX_XC = addSQLColumn(view, tabX, "colXC");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    view.set(CProp.VISIBLE, colZ_ZC, false);
    setFormula(view, colU_inr, "[pk_colZB]");
    String res = "["
        + "(X1,[(X1,a,L)],[(Z1,X1,.)]),"
        + "(X2,[(X2,a,S)],[(Z2,X2,.)]),"
        + "(X3,[(X3,a,L)],[(Z3,X3,.),(Z4,X3,.),(Z8,X3,.),(Z9,X3,.)]),"
        + "(X5,[(X5,b,L)],[(Z5,X5,.)]),"
        + "(X6,[(X6,_,L)],[(Z10,X6,.),(Z11,X6,.),(Z6,X6,.)]),"
        + "(X7,[(X7,c,L)],[(Z12,X7,.)]),"
        + "(X9,[(X9,c,L)],[(Z6,X9,.)]),"
        + "(XX,[],[(Z7,XX,.)])"
        + "]";
    expectResult(view, true, res);
    if (!SKIP_KNOWN_FAILING_TESTS) {
      view.set(CProp.VISIBLE, colU_inr, false);
      expectResult(view, true, res);
      view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
      expectResult(view, true, res);
    }
  }

  @Test
  public void testAggregateInwardToUnprojected() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    PrimitivePath colX_inr = addFormulaColumn(view, Path.ROOT, "inr", "null");
    view.set(CProp.VISIBLE, colX_inr, false);
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    view.set(CProp.VISIBLE, colZ_ZA, false);
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_inr, "[pk_colZA]");


    String res1 = "["
        + "(X1,a,1,.,[(.,.,1)]),"
        + "(X2,a,3,.,[(.,.,3)]),"
        + "(X3,a,28,.,[(.,.,10),(.,.,14),(.,.,2)]),"
        + "(X4,b,0,.,[]),"
        + "(X5,b,100,.,[(.,.,100)]),"
        + "(X6,_,1000,.,[(.,.,0),(.,.,1000),(.,.,_)]),"
        + "(X7,c,0,.,[(.,.,0)]),"
        + "(X8,d,0,.,[]),"
        + "(X9,c,10000,.,[(.,.,10000)])"
        + "]";
    String res2 = "["
        + "(X1,a,1,.,[(.,.,1)]),"
        + "(X2,a,3,.,[(.,.,3)]),"
        + "(X3,a,26,.,[(.,.,10),(.,.,14),(.,.,2)])," // 2's collapsed
        + "(X4,b,0,.,[]),"
        + "(X5,b,100,.,[(.,.,100)]),"
        + "(X6,_,1000,.,[(.,.,0),(.,.,1000),(.,.,_)]),"
        + "(X7,c,0,.,[(.,.,0)]),"
        + "(X8,d,0,.,[]),"
        + "(X9,c,10000,.,[(.,.,10000)])"
        + "]";
    setFormula(view, colX_agg, "sum(if([pk_colZA] = 'Z3', [colZC_n], [colZC_n]))");
    expectResult(view, true, res1);
    setFormula(view, colX_agg, "sum(if([inr] = 'Z3', [colZC_n], [colZC_n]))");
    expectResult(view, true, res1);

    /* Adding a non-injective function (LEN) within the formula argument has no tuple-collapsing
    effect. */
    setFormula(view, colX_agg, "sum(if(len([pk_colZA]) = 2, [colZC_n], [colZC_n]))");
    expectResult(view, true, res1);
    setFormula(view, colX_agg, "sum(if(len([inr]) = 2, [colZC_n], [colZC_n]))");
    expectResult(view, true, res1);

    setFormula(view, colX_agg, "sum(if(true, [colZC_n], [colZC_n]))");
    expectResult(view, true, res2);

    /* If the non-injective function LEN is used in its own hidden field, collapsing will happen.
    Contrast with folding the len([pk_colZA]) expression into the aggregate above. */
    setFormula(view, colX_inr, "len([pk_colZA])");
    setFormula(view, colX_agg, "sum(if([inr] = 2, [colZC_n], [colZC_n]))");
    expectResult(view, true, res2);
  }

  @Test
  public void testNonInjectiveFunctionInAggregateKey() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();


    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_inr = addFormulaColumn(view, Path.ROOT, "inr", "null");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_inr, "len(concat([colZC_n]))");
    setFormula(view, colX_agg, "sum([colZC_n])");
    expectResult(view, true, "["
        + "(X1,a,1,1,[(Z1,X1,1)]),"
        + "(X2,a,1,3,[(Z2,X2,3)]),"
        + "(X3,a,1,4,[(Z8,X3,2),(Z9,X3,2)]),"
        + "(X3,a,2,24,[(Z3,X3,10),(Z4,X3,14)]),"
        + "(X4,b,0,0,[]),"
        + "(X5,b,3,100,[(Z5,X5,100)]),"
        + "(X6,_,0,0,[(Z11,X6,_)]),"
        + "(X6,_,1,0,[(Z10,X6,0)]),"
        + "(X6,_,4,1000,[(Z6,X6,1000)]),"
        + "(X7,c,1,0,[(Z12,X7,0)]),"
        + "(X8,d,0,0,[]),"
        + "(X9,c,5,10000,[(Z6,X9,10000)])"
        + "]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colX_XA, false);
    expectResult(view, true, "["
        + "(.,_,0,0,[(Z11,X6,_)]),"
        + "(.,_,1,0,[(Z10,X6,0)]),"
        + "(.,_,4,1000,[(Z6,X6,1000)]),"
        + "(.,a,1,8,[(Z1,X1,1),(Z2,X2,3),(Z8,X3,2),(Z9,X3,2)]),"
        + "(.,a,2,24,[(Z3,X3,10),(Z4,X3,14)]),"
        + "(.,b,0,0,[]),"
        + "(.,b,3,100,[(Z5,X5,100)]),"
        + "(.,c,1,0,[(Z12,X7,0)]),"
        + "(.,c,5,10000,[(Z6,X9,10000)]),"
        + "(.,d,0,0,[])"
        + "]");
    view.set(CProp.VISIBLE, colX_XB, false);
    expectResult(view, true, "["
        + "(.,.,0,0,[(Z11,X6,_)]),"
        + "(.,.,1,8,[(Z1,X1,1),(Z10,X6,0),(Z12,X7,0),(Z2,X2,3),(Z8,X3,2),(Z9,X3,2)]),"
        + "(.,.,2,24,[(Z3,X3,10),(Z4,X3,14)]),"
        + "(.,.,3,100,[(Z5,X5,100)]),"
        + "(.,.,4,1000,[(Z6,X6,1000)]),"
        + "(.,.,5,10000,[(Z6,X9,10000)])"
        + "]");
  }

  @Test
  public void testNonInjectiveFunctionToHiddenFieldInAggregateKey() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZH", "colZC_n",
      String.class, String.class, String.class, Integer.class,
      "Z1", "X1", "H01", 1,
      "Z2", "X2", "H02", 3,
      "Z3", "X3", "H03", 10,
      "Z4", "X3", "H04", 14,
      "Z5", "X5", "H05", 100,
      "Z6", "X6", "H06", 1000,
      "Z6", "X9", "H07", 10000,
      "Z7", "XX", "H08", 100000,
      "Z8", "X3", "H09", 2,
      "Z9", "X3", "H10", 2,
      "Z10", "X6", "H11", 0,
      "Z11", "X6", "H12", null,
      "Z12", "X7", "H13", 0,
      "Z13", "X3", "H14-", 2
    });
    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_inr = addFormulaColumn(view, Path.ROOT, "inr", "null");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    view.set(CProp.VISIBLE, colZ_ZA, false);
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZH = addSQLColumn(view, tabZ, "colZH");
    view.set(CProp.VISIBLE, colZ_ZH, false);
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_inr, "len([colZH])");
    setFormula(view, colX_agg, "sum([colZC_n])");
    expectResult(view, true, "["
        + "(X1,a,3,1,[(.,.,.,1)]),"
        + "(X2,a,3,3,[(.,.,.,3)]),"
        + "(X3,a,3,26,[(.,.,.,10),(.,.,.,14),(.,.,.,2)]),"
        + "(X3,a,4,2,[(.,.,.,2)]),"
        + "(X4,b,_,0,[]),"
        + "(X5,b,3,100,[(.,.,.,100)]),"
        + "(X6,_,3,1000,[(.,.,.,0),(.,.,.,1000),(.,.,.,_)]),"
        + "(X7,c,3,0,[(.,.,.,0)]),"
        + "(X8,d,_,0,[]),"
        + "(X9,c,3,10000,[(.,.,.,10000)])"
        + "]");
  }

  @Test
  public void testEmptyGroupBy() {
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    setInstantiatedTable(view, Path.ROOT, "tabZ");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    PrimitivePath colZ_ZA = addSQLColumn(view, Path.ROOT, "pk_colZA");
    view.set(CProp.VISIBLE, colZ_ZA, false);
    PrimitivePath colZ_ZB = addSQLColumn(view, Path.ROOT, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    PrimitivePath colZ_ZC = addSQLColumn(view, Path.ROOT, "colZC_n");
    view.set(CProp.VISIBLE, colZ_ZC, false);
    PrimitivePath colZ_agg = addFormulaColumn(view, Path.ROOT, "agg", "sum([colZC_n])");
    expectResult(view, true, "[(.,.,.,111130)]");
  }

  @Test
  public void testAggregateWithSiblingReferenceToNonIdentifiedHiddenField() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "L",
      "X2", "a", "S",
      "X3", "a", "L",
      "X4", "b", "L",
      "X5", "b", "L",
      "X6", null, "L",
      "X7", "c", "L",
      "X8", "d", "L",
      "X9", "c", "L"
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z01", "X1", 1,
      "Z02", "X2", 3,
      "Z03", "X3", 10,
      "Z04", "X3", 14,
      "Z05", "X5", 100,
      "Z06", "X6", 1000,
      "Z06", "X9", 10000,
      "Z07", "XX", 100000,
      "Z08", "X3", 2,
      "Z09", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
      "Z13-", "X2", 3,
      "Z14", "X3", 14
    });
    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_XC = addSQLColumn(view, Path.ROOT, "colXC");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_agg, "sum([colZC_n])");
    expectResult(view, true, "["
        + "(X1,a,L,1,[(Z01,.,1)]),"
        + "(X2,a,S,6,[(Z02,.,3),(Z13-,.,3)]),"
        + "(X3,a,L,42,[(Z03,.,10),(Z04,.,14),(Z08,.,2),(Z09,.,2),(Z14,.,14)]),"
        + "(X4,b,L,0,[]),"
        + "(X5,b,L,100,[(Z05,.,100)]),"
        + "(X6,_,L,1000,[(Z06,.,1000),(Z10,.,0),(Z11,.,_)]),"
        + "(X7,c,L,0,[(Z12,.,0)]),"
        + "(X8,d,L,0,[]),"
        + "(X9,c,L,10000,[(Z06,.,10000)])"
        + "]");
    PrimitivePath colX_inr = addFormulaColumn(view, Path.ROOT, "inr", "len([pk_colZA])");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_XA, colX_XB, colX_XC, colX_agg, colX_inr, tabZ)));
    expectResult(view, true, "["
        + "(X1,a,L,1,3,[(Z01,.,1)]),"
        + "(X2,a,S,3,3,[(Z02,.,3)]),"
        + "(X2,a,S,3,4,[(Z13-,.,3)]),"
        + "(X3,a,L,42,3,[(Z03,.,10),(Z04,.,14),(Z08,.,2),(Z09,.,2),(Z14,.,14)]),"
        + "(X4,b,L,0,_,[]),"
        + "(X5,b,L,100,3,[(Z05,.,100)]),"
        + "(X6,_,L,1000,3,[(Z06,.,1000),(Z10,.,0),(Z11,.,_)]),"
        + "(X7,c,L,0,3,[(Z12,.,0)]),"
        + "(X8,d,L,0,_,[]),"
        + "(X9,c,L,10000,3,[(Z06,.,10000)])"
        + "]");
    view.set(CProp.VISIBLE, colZ_ZA, false);
    expectResult(view, true, "["
        + "(X1,a,L,1,3,[(.,.,1)]),"
        + "(X2,a,S,3,3,[(.,.,3)]),"
        + "(X2,a,S,3,4,[(.,.,3)]),"
        + "(X3,a,L,26,3,[(.,.,10),(.,.,14),(.,.,2)]),"
        + "(X4,b,L,0,_,[]),"
        + "(X5,b,L,100,3,[(.,.,100)]),"
        + "(X6,_,L,1000,3,[(.,.,0),(.,.,1000),(.,.,_)]),"
        + "(X7,c,L,0,3,[(.,.,0)]),"
        + "(X8,d,L,0,_,[]),"
        + "(X9,c,L,10000,3,[(.,.,10000)])"
        + "]");
  }

  @Test
  public void testMultiLevelAggregateInwardReferenceWithUnidentifiedMiddleField() {
    createTable("tabQ", new Object[] {
      "pk_colQA", "colQB_n",
      String.class, String.class,
      "Q1", "a",
      "Q2", "b",
      "Q3", null,
      "Q4", "c",
      "Q5", "d",
      "Q6", "e",
    });
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC",
      String.class, String.class, String.class,
      "X1", "a", "L",
      "X2", "a", "S",
      "X3", "a", "L",
      "X4", "b", "L",
      "X5", "b", "L",
      "X6", null, "L",
      "X7", "c", "L",
      "X8", "d", "L",
      "X9", "c", "L"
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z01", "X1", 1,
      "Z02", "X2", 3,
      "Z03", "X3", 10,
      "Z04", "X3", 14,
      "Z05", "X5", 100,
      "Z06", "X6", 1000,
      "Z06", "X9", 10000,
      "Z07", "XX", 100000,
      "Z08", "X3", 2,
      "Z09", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
      "Z13-", "X2", 3,
      "Z14", "X3", 14
    });


    MutableView view = createView();
    view.set(CProp.LABEL, Path.ROOT, "tabQ");
    setInstantiatedTable(view, Path.ROOT, "tabQ");
    PrimitivePath colQ_QA = addSQLColumn(view, Path.ROOT, "pk_colQA");
    PrimitivePath colQ_QB = addSQLColumn(view, Path.ROOT, "colQB_n");
    PrimitivePath colQ_aggo = addFormulaColumn(view, Path.ROOT, "aggo", "null");
    RelationPath tabX = addRelation(view, Path.ROOT, "tabX");
    setInstantiatedTable(view, tabX, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, tabX, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, tabX, "colXB_n");
    view.set(CProp.VISIBLE, colX_XB, false);
    view.set(QProp.JOINED_ON, colX_XB, OptionalPrimitivePath.of(colQ_QB));
    PrimitivePath colX_XC = addSQLColumn(view, tabX, "colXC");
    PrimitivePath colX_aggm = addFormulaColumn(view, tabX, "aggm", "null");
    RelationPath tabZ = addRelation(view, tabX, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colQ_aggo, "sum([colZC_n])");
    setFormula(view, colX_aggm, "sum([colZC_n])");
    expectResult(view, true, "["
        + "(Q1,a,49,[(X1,.,L,1,[(Z01,X1,1)]),(X2,.,S,6,[(Z02,X2,3),(Z13-,X2,3)]),(X3,.,L,42,[(Z03,X3,10),(Z04,X3,14),(Z08,X3,2),(Z09,X3,2),(Z14,X3,14)])]),"
        + "(Q2,b,100,[(X4,.,L,0,[]),(X5,.,L,100,[(Z05,X5,100)])]),"
        + "(Q3,_,0,[]),"
        + "(Q4,c,10000,[(X7,.,L,0,[(Z12,X7,0)]),(X9,.,L,10000,[(Z06,X9,10000)])]),"
        + "(Q5,d,0,[(X8,.,L,0,[])]),"
        + "(Q6,e,0,[])]");
    if (!SKIP_KNOWN_FAILING_TESTS) {
      PrimitivePath colQ_inro = addFormulaColumn(view, Path.ROOT, "inro", "len([pk_colZA])");
      view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colQ_QA, colQ_QB, colQ_aggo, colQ_inro, tabX)));
      expectResult(view, true, "["
          + "(Q1,a,3,4,[(X2,.,S,3,[(Z13-,X2,3)])]),"
          + "(Q1,a,46,3,[(X1,.,L,1,[(Z01,X1,1)]),(X2,.,S,3,[(Z02,X2,3)]),(X3,.,L,42,[(Z03,X3,10),(Z04,X3,14),(Z08,X3,2),(Z09,X3,2),(Z14,X3,14)])]),"
          + "(Q2,b,0,_,[(X4,.,L,0,[])]),"
          + "(Q2,b,100,3,[(X5,.,L,100,[(Z05,X5,100)])]),"
          + "(Q3,_,0,_,[]),"
          + "(Q4,c,10000,3,[(X7,.,L,0,[(Z12,X7,0)]),(X9,.,L,10000,[(Z06,X9,10000)])]),"
          + "(Q5,d,0,_,[(X8,.,L,0,[])]),"
          + "(Q6,e,0,_,[])]");
      view.set(QProp.INCLUDE_PRIMARY_KEY, tabZ, false);
      view.set(CProp.VISIBLE, colZ_ZA, false);
      expectResult(view, true, "["
          + "(Q1,a,3,4,[(X2,.,S,3,[(.,.,3)])]),"
          + "(Q1,a,30,3,[(X1,.,L,1,[(.,.,1)]),(X2,.,S,3,[(.,.,3)]),(X3,.,L,26,[(.,.,10),(.,.,14),(.,.,2)])]),"
          + "(Q2,b,0,_,[(X4,.,L,0,[])]),"
          + "(Q2,b,100,3,[(X5,.,L,100,[(.,.,100)])]),"
          + "(Q3,_,0,_,[]),"
          + "(Q4,c,10000,3,[(X7,.,L,0,[(.,.,0)]),(X9,.,L,10000,[(.,.,10000)])]),"
          + "(Q5,d,0,_,[(X8,.,L,0,[])]),"
          + "(Q6,e,0,_,[])]");
    }
  }

  @Test
  public void testOutwardToAggregate() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });

    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_agg, "sum([colZC_n])");
    expectResult(view, true, "["
        + "(X1,a,1,[(Z1,X1,1)]),"
        + "(X2,a,3,[(Z2,X2,3)]),"
        + "(X3,a,28,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,0,[]),"
        + "(X5,b,100,[(Z5,X5,100)]),"
        + "(X6,_,1000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,0,[(Z12,X7,0)]),"
        + "(X8,d,0,[]),"
        + "(X9,c,10000,[(Z6,X9,10000)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    PrimitivePath colZ_aggs = addFormulaColumn(view, tabZ, "aggs", "sum([colZC_n])");
    view.set(SProp.CHILD_PATHS, tabZ, new PathList(ImmutableList.of(colZ_ZA, colZ_ZB, colZ_aggs, colZ_ZC)));
    expectResult(view, true, "["
        + "(X1,a,1,[(Z1,X1,1,1)]),"
        + "(X2,a,3,[(Z2,X2,3,3)]),"
        + "(X3,a,28,[(Z3,X3,10,10),(Z4,X3,14,14),(Z8,X3,2,2),(Z9,X3,2,2)]),"
        + "(X4,b,0,[]),"
        + "(X5,b,100,[(Z5,X5,100,100)]),"
        + "(X6,_,1000,[(Z10,X6,0,0),(Z11,X6,0,_),(Z6,X6,1000,1000)]),"
        + "(X7,c,0,[(Z12,X7,0,0)]),"
        + "(X8,d,0,[]),"
        + "(X9,c,10000,[(Z6,X9,10000,10000)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    PrimitivePath colZ_outr = addFormulaColumn(view, tabZ, "outr", "[agg]");
    view.set(SProp.CHILD_PATHS, tabZ, new PathList(ImmutableList.of(colZ_ZA, colZ_ZB, colZ_outr, colZ_aggs, colZ_ZC)));
    expectResult(view, true, "["
        + "(X1,a,1,[(Z1,X1,1,1,1)]),"
        + "(X2,a,3,[(Z2,X2,3,3,3)]),"
        + "(X3,a,28,[(Z3,X3,28,10,10),(Z4,X3,28,14,14),(Z8,X3,28,2,2),(Z9,X3,28,2,2)]),"
        + "(X4,b,0,[]),"
        + "(X5,b,100,[(Z5,X5,100,100,100)]),"
        + "(X6,_,1000,[(Z10,X6,1000,0,0),(Z11,X6,1000,0,_),(Z6,X6,1000,1000,1000)]),"
        + "(X7,c,0,[(Z12,X7,0,0,0)]),"
        + "(X8,d,0,[]),"
        + "(X9,c,10000,[(Z6,X9,10000,10000,10000)])"
        + "]");


    // ----- Test Case ------------------------------------------------
    PrimitivePath colX_agg2 = addFormulaColumn(view, Path.ROOT, "agg2", "sum([outr])");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_XA, colX_XB, colX_agg, colX_agg2, tabZ)));
    expectResult(view, true, "["
        + "(X1,a,1,1,[(Z1,X1,1,1,1)]),"
        + "(X2,a,3,3,[(Z2,X2,3,3,3)]),"
        + "(X3,a,28,112,[(Z3,X3,28,10,10),(Z4,X3,28,14,14),(Z8,X3,28,2,2),(Z9,X3,28,2,2)]),"
        + "(X4,b,0,0,[]),"
        + "(X5,b,100,100,[(Z5,X5,100,100,100)]),"
        + "(X6,_,1000,3000,[(Z10,X6,1000,0,0),(Z11,X6,1000,0,_),(Z6,X6,1000,1000,1000)]),"
        + "(X7,c,0,0,[(Z12,X7,0,0,0)]),"
        + "(X8,d,0,0,[]),"
        + "(X9,c,10000,10000,[(Z6,X9,10000,10000,10000)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    setFormula(view, colZ_outr, "if([colXB_n] = 'a', 1, 5)");
    view.removePath(colZ_aggs);
    view.removePath(colX_agg);
    expectResult(view, true, "["
        + "(X1,a,1,[(Z1,X1,1,1)]),"
        + "(X2,a,1,[(Z2,X2,1,3)]),"
        + "(X3,a,4,[(Z3,X3,1,10),(Z4,X3,1,14),(Z8,X3,1,2),(Z9,X3,1,2)]),"
        + "(X4,b,0,[]),"
        + "(X5,b,5,[(Z5,X5,5,100)]),"
        + "(X6,_,0,[(Z10,X6,_,0),(Z11,X6,_,_),(Z6,X6,_,1000)]),"
        + "(X7,c,5,[(Z12,X7,5,0)]),"
        + "(X8,d,0,[]),"
        + "(X9,c,5,[(Z6,X9,5,10000)])"
        + "]");
  }

  @Test
  public void testFormulaReferencingAggregateSibling() {
    // Based on testAggregateOnJoinedChildRelation.
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_agg, "sum([colZC_n]) + if([colXB_n] = 'a', 1, 5)");
    expectResult(view, true, "["
        + "(X1,a,2,[(Z1,X1,1)]),"
        + "(X2,a,4,[(Z2,X2,3)]),"
        + "(X3,a,29,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,5,[]),"
        + "(X5,b,105,[(Z5,X5,100)]),"
        + "(X6,_,_,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,5,[(Z12,X7,0)]),"
        + "(X8,d,5,[]),"
        + "(X9,c,10005,[(Z6,X9,10000)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    PrimitivePath colX_sref = addFormulaColumn(view, Path.ROOT, "sref", "[agg] * len([pk_colXA])");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_XA, colX_XB, colX_agg, colX_sref, tabZ)));
    expectResult(view, true, "["
        + "(X1,a,2,4,[(Z1,X1,1)]),"
        + "(X2,a,4,8,[(Z2,X2,3)]),"
        + "(X3,a,29,58,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,5,10,[]),"
        + "(X5,b,105,210,[(Z5,X5,100)]),"
        + "(X6,_,_,_,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,5,10,[(Z12,X7,0)]),"
        + "(X8,d,5,10,[]),"
        + "(X9,c,10005,20010,[(Z6,X9,10000)])"
        + "]");
  }

  @Test
  public void testOutwardReferenceToUnprojected() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC_n",
      String.class, String.class, Integer.class,
      "X1", "a", 53,
      "X2", "a", 262,
      "X3", "a", 21,
      "X4", "b", 87,
      "X5", "b", 87,
      "X6", null, 3,
      "X7", "c", null,
      "X8", "d", 421,
      "X9", "c", 12
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });


    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_XC = addSQLColumn(view, Path.ROOT, "colXC_n");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    expectResult(view, true, "["
        + "(X1,a,53,[(Z1,X1,1)]),"
        + "(X2,a,262,[(Z2,X2,3)]),"
        + "(X3,a,21,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,87,[]),"
        + "(X5,b,87,[(Z5,X5,100)]),"
        + "(X6,_,3,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,_,[(Z12,X7,0)]),"
        + "(X8,d,421,[]),"
        + "(X9,c,12,[(Z6,X9,10000)])"
        + "]");
    PrimitivePath colZ_agg = addFormulaColumn(view, tabZ, "agg", "[colXC_n]");
    view.set(SProp.CHILD_PATHS, tabZ, new PathList(ImmutableList.of(colZ_ZA, colZ_ZB, colZ_agg, colZ_ZC)));
    expectResult(view, true, "["
        + "(X1,a,53,[(Z1,X1,53,1)]),"
        + "(X2,a,262,[(Z2,X2,262,3)]),"
        + "(X3,a,21,[(Z3,X3,21,10),(Z4,X3,21,14),(Z8,X3,21,2),(Z9,X3,21,2)]),"
        + "(X4,b,87,[]),"
        + "(X5,b,87,[(Z5,X5,87,100)]),"
        + "(X6,_,3,[(Z10,X6,3,0),(Z11,X6,3,_),(Z6,X6,3,1000)]),"
        + "(X7,c,_,[(Z12,X7,_,0)]),"
        + "(X8,d,421,[]),"
        + "(X9,c,12,[(Z6,X9,12,10000)])"
        + "]");
    view.set(CProp.VISIBLE, colX_XC, false);
    setFormula(view, colZ_agg, "sum([colXC_n])");
    expectResult(view, true, "["
        + "(X1,a,.,[(Z1,X1,53,1)]),"
        + "(X2,a,.,[(Z2,X2,262,3)]),"
        + "(X3,a,.,[(Z3,X3,21,10),(Z4,X3,21,14),(Z8,X3,21,2),(Z9,X3,21,2)]),"
        + "(X4,b,.,[]),"
        + "(X5,b,.,[(Z5,X5,87,100)]),"
        + "(X6,_,.,[(Z10,X6,3,0),(Z11,X6,3,_),(Z6,X6,3,1000)]),"
        + "(X7,c,.,[(Z12,X7,0,0)]),"
        + "(X8,d,.,[]),"
        + "(X9,c,.,[(Z6,X9,12,10000)])"
        + "]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, false);
    view.set(CProp.VISIBLE, colX_XA, false);
    /* It might be reasonable for this behavior to change in the future, though the current behavior
    does make sense. */
    expectResult(view, true, "["
        + "(.,_,.,[(Z10,X6,3,0),(Z11,X6,3,_),(Z6,X6,3,1000)]),"
        + "(.,a,.,[(Z1,X1,53,1),(Z2,X2,262,3),(Z3,X3,21,10),(Z4,X3,21,14),(Z8,X3,21,2),(Z9,X3,21,2)]),"
        + "(.,b,.,[(Z5,X5,87,100)]),"
        + "(.,c,.,[(Z12,X7,0,0),(Z6,X9,12,10000)]),"
        + "(.,d,.,[])"
        + "]");
    view.set(QProp.INCLUDE_PRIMARY_KEY, Path.ROOT, true);
    view.set(CProp.VISIBLE, colX_XA, true);
    //view.set(CProp.VISIBLE, colX_XC, true);
    expectResult(view, true, "["
        + "(X1,a,.,[(Z1,X1,53,1)]),"
        + "(X2,a,.,[(Z2,X2,262,3)]),"
        + "(X3,a,.,[(Z3,X3,21,10),(Z4,X3,21,14),(Z8,X3,21,2),(Z9,X3,21,2)]),"
        + "(X4,b,.,[]),"
        + "(X5,b,.,[(Z5,X5,87,100)]),"
        + "(X6,_,.,[(Z10,X6,3,0),(Z11,X6,3,_),(Z6,X6,3,1000)]),"
        + "(X7,c,.,[(Z12,X7,0,0)]),"
        + "(X8,d,.,[]),"
        + "(X9,c,.,[(Z6,X9,12,10000)])"
        + "]");
    PrimitivePath colX_inr = addFormulaColumn(view, Path.ROOT, "inr", "[colZC_n]");
    view.set(CProp.VISIBLE, colX_inr, false);
    setFormula(view, colZ_agg, "sum([inr])");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_XA, colX_XB, colX_XC, colX_inr, tabZ)));
    expectResult(view, true, "["
        + "(X1,a,.,.,[(Z1,X1,1,1)]),"
        + "(X2,a,.,.,[(Z2,X2,3,3)]),"
        + "(X3,a,.,.,[(Z3,X3,10,10),(Z4,X3,14,14),(Z8,X3,2,2),(Z9,X3,2,2)]),"
        + "(X4,b,.,.,[]),"
        + "(X5,b,.,.,[(Z5,X5,100,100)]),"
        + "(X6,_,.,.,[(Z10,X6,0,0),(Z11,X6,0,_),(Z6,X6,1000,1000)]),"
        + "(X7,c,.,.,[(Z12,X7,0,0)]),"
        + "(X8,d,.,.,[]),"
        + "(X9,c,.,.,[(Z6,X9,10000,10000)])"
        + "]");
  }

  @Test
  public void testNestedAggregateAtSameLevel() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC_n",
      String.class, String.class, Integer.class,
      "X1", "a", 53,
      "X2", "a", 262,
      "X3", "a", 21,
      "X4", "b", 87,
      "X5", "b", 87,
      "X6", null, 3,
      "X7", "c", null,
      "X8", "d", 421,
      "X9", "c", 12
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", 1000,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", 0,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });
    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_XC = addSQLColumn(view, Path.ROOT, "colXC_n");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    expectResult(view, true, "["
        + "(X1,a,53,[(Z1,X1,1)]),"
        + "(X2,a,262,[(Z2,X2,3)]),"
        + "(X3,a,21,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,87,[]),"
        + "(X5,b,87,[(Z5,X5,100)]),"
        + "(X6,_,3,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,_,[(Z12,X7,0)]),"
        + "(X8,d,421,[]),"
        + "(X9,c,12,[(Z6,X9,10000)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    PrimitivePath colX_agg = addFormulaColumn(view, Path.ROOT, "agg", "sum([colZC_n] ** 2)");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_XA, colX_XB, colX_XC, colX_agg, tabZ)));
    expectResult(view, true, "["
        + "(X1,a,53,1,[(Z1,X1,1)]),"
        + "(X2,a,262,9,[(Z2,X2,3)]),"
        + "(X3,a,21,304,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,87,0,[]),"
        + "(X5,b,87,10000,[(Z5,X5,100)]),"
        + "(X6,_,3,1000000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,_,0,[(Z12,X7,0)]),"
        + "(X8,d,421,0,[]),"
        + "(X9,c,12,100000000,[(Z6,X9,10000)])"
        + "]");
    // ----- Test Case ------------------------------------------------
    setFormula(view, colX_agg, "sum([colZC_n]) ** 2");
    expectResult(view, true, "["
        + "(X1,a,53,1,[(Z1,X1,1)]),"
        + "(X2,a,262,9,[(Z2,X2,3)]),"
        + "(X3,a,21,784,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,87,0,[]),"
        + "(X5,b,87,10000,[(Z5,X5,100)]),"
        + "(X6,_,3,1000000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,_,0,[(Z12,X7,0)]),"
        + "(X8,d,421,0,[]),"
        + "(X9,c,12,100000000,[(Z6,X9,10000)])"
        + "]");
    PrimitivePath colX_sib = addFormulaColumn(view, Path.ROOT, "sib", "avg([agg])");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_XA, colX_XB, colX_XC, colX_agg, colX_sib, tabZ)));
    String res1 =  "["
        + "(X1,a,53,1,1,[(Z1,X1,1)]),"
        + "(X2,a,262,9,9,[(Z2,X2,3)]),"
        + "(X3,a,21,784,784,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,87,0,0,[]),"
        + "(X5,b,87,10000,10000,[(Z5,X5,100)]),"
        + "(X6,_,3,1000000,1000000,[(Z10,X6,0),(Z11,X6,_),(Z6,X6,1000)]),"
        + "(X7,c,_,0,0,[(Z12,X7,0)]),"
        + "(X8,d,421,0,0,[]),"
        + "(X9,c,12,100000000,100000000,[(Z6,X9,10000)])"
        + "]";
    expectResult(view, true, res1);
    setFormula(view, colX_agg, "avg(sum([colZC_n]) ** 2)");
    expectResult(view, true, res1);
  }

  @Test
  public void testVariousNullBehaviors() {
    // Based on the tests from http://www.sqlite.org/nulls.html .
    createTable("tabN", new Object[] {
      "pk_colNA", "colNB_n", "colNC_n",
      Integer.class, Integer.class, Integer.class,
      1,    0,    0,
      2,    0,    1,
      3,    1,    0,
      4,    1,    1,
      5, null,    0,
      6, null,    1,
      7, null, null,
    });


    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    view.set(CProp.LABEL, Path.ROOT, "tabN");
    setInstantiatedTable(view, Path.ROOT, "tabN");
    PrimitivePath colN_NA = addSQLColumn(view, Path.ROOT, "pk_colNA");
    PrimitivePath colN_NB = addSQLColumn(view, Path.ROOT, "colNB_n");
    PrimitivePath colN_NC = addSQLColumn(view, Path.ROOT, "colNC_n");
    view.set(CProp.LABEL, colN_NB, "b");
    view.set(CProp.LABEL, colN_NC, "c");
    List<PrimitivePath> formulas = Lists.newArrayList();
    formulas.add(addFormulaColumn(view, Path.ROOT, "f01", "if([b] <> 0, 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f02", "if(not ([b] <> 0), 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f03", "if([b] <> 0 and [c] <> 0, 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f04", "if(not ([b] <> 0 and [c] <> 0), 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f05", "if([b] <> 0 or [c] <> 0, 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f06", "if(not ([b] <> 0 or [c] <> 0), 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f07", "if([b] = [c], 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f08", "if([c] = [b], 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f09", "if([b] is [c], 1, 0)"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f10", "if([c] is [b], 1, 0)"));
    expectResult(view, true, "["
        //    b,c|1,2,3,4,5,6,7,8,9,10
        + "(1,0,0,0,1,0,1,0,1,1,1,1,1),"
        + "(2,0,1,0,1,0,1,1,0,0,0,0,0),"
        + "(3,1,0,1,0,0,1,1,0,0,0,0,0),"
        + "(4,1,1,1,0,1,0,1,0,1,1,1,1),"
        + "(5,_,0,_,_,0,1,_,_,_,_,0,0),"
        + "(6,_,1,_,_,_,_,1,0,_,_,0,0),"
        + "(7,_,_,_,_,_,_,_,_,_,_,1,1)"
        + "]");
    for (PrimitivePath formulaPath : formulas)
      view.removePath(formulaPath);
    formulas.clear();

    formulas.add(addFormulaColumn(view, Path.ROOT, "f11", "[b] * 0"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f12", "[b] * [c]"));
    formulas.add(addFormulaColumn(view, Path.ROOT, "f13", "[b] + [c]"));
    expectResult(view, true, "["
        + "(1,0,0,0,0,0),"
        + "(2,0,1,0,0,1),"
        + "(3,1,0,0,0,1),"
        + "(4,1,1,0,1,2),"
        + "(5,_,0,_,_,_),"
        + "(6,_,1,_,_,_),"
        + "(7,_,_,_,_,_)"
        + "]");
    for (PrimitivePath formulaPath : formulas)
      view.removePath(formulaPath);
    formulas.clear();

    MutableView view2 = createView();
    RelationPath childPath = view2.addRelationPath(Path.ROOT);
    PathMap view2map = view2.copyFrom(view, Path.ROOT, childPath);
    view2.set(CProp.LABEL, childPath, "child");
    addFormulaColumn(view2, Path.ROOT, "a01", "count([pk_colNA])");
    addFormulaColumn(view2, Path.ROOT, "a02", "count([b])");
    addFormulaColumn(view2, Path.ROOT, "a03", "sum([b])");
    addFormulaColumn(view2, Path.ROOT, "a04", "avg([b])");
    addFormulaColumn(view2, Path.ROOT, "a05", "min([b])");
    addFormulaColumn(view2, Path.ROOT, "a06", "max([b])");
    view2.set(CProp.VISIBLE, childPath, false);
    expectResult(view2, true, "[(.,7,4,2,0.5,0,1)]");
    view2.set(QProp.INCLUDE_PRIMARY_KEY, childPath, false);
    view2.set(CProp.VISIBLE, childPath, true);
    view2.set(CProp.VISIBLE, view2map.getNewPath(colN_NA), false);
    view2.set(CProp.VISIBLE, view2map.getNewPath(colN_NB), true);
    view2.set(CProp.VISIBLE, view2map.getNewPath(colN_NC), false);
    // Distinctness of nulls.
    expectResult(view2, true, "[("
        + "[(.,0,.),(.,1,.),(.,_,.)],7,2,1,0.5,0,1"
        + ")]");

    PrimitivePath filterCol = addFormulaColumn(view, Path.ROOT, "ffx", "true");
    setFilter(view, filterCol, "[ffx] is true");
    view.set(CProp.VISIBLE, filterCol, false);
    expectResult(view, true, "["
        + "(1,0,0,.),"
        + "(2,0,1,.),"
        + "(3,1,0,.),"
        + "(4,1,1,.),"
        + "(5,_,0,.),"
        + "(6,_,1,.),"
        + "(7,_,_,.)"
        + "]");
    setFormula(view, filterCol, "[b] < 10");
    expectResult(view, true, "["
        + "(1,0,0,.),"
        + "(2,0,1,.),"
        + "(3,1,0,.),"
        + "(4,1,1,.)"
        + "]");
    setFormula(view, filterCol, "not ([b] > 10)");
    expectResult(view, true, "["
        + "(1,0,0,.),"
        + "(2,0,1,.),"
        + "(3,1,0,.),"
        + "(4,1,1,.)"
        + "]");
    setFormula(view, filterCol, "[b] < 10 OR [c] = 1");
    expectResult(view, true, "["
        + "(1,0,0,.),"
        + "(2,0,1,.),"
        + "(3,1,0,.),"
        + "(4,1,1,.),"
        + "(6,_,1,.)"
        + "]");
    setFormula(view, filterCol, "[b] < 10 AND [c] = 1");
    expectResult(view, true, "["
        + "(2,0,1,.),"
        + "(4,1,1,.)"
        + "]");
    setFormula(view, filterCol, "not ([b] < 10 AND [c] = 1)");
    expectResult(view, true, "["
        + "(1,0,0,.),"
        + "(3,1,0,.),"
        + "(5,_,0,.)"
        + "]");
    setFormula(view, filterCol, "not ([c] = 1 AND [b] < 10)");
    expectResult(view, true, "["
        + "(1,0,0,.),"
        + "(3,1,0,.),"
        + "(5,_,0,.)"
        + "]");
  }

  @Test
  public void testNullInSynthesizedUnion() {
    // See union test from http://www.sqlite.org/nulls.html .
    createTable("tabN", new Object[] {
      "pk_colNA", "colNB_n", "colNC_n",
      Integer.class, Integer.class, Integer.class,
      1,    0,    0,
      2,    0,    1,
      3,    1,    0,
      4,    1,    1,
      5, null,    0,
      6, null,    1,
      7, null, null,
    });
    createTable("tworows", new Object[] {
      "pk_case",
      Integer.class,
      1, 2
    });


    MutableView view = createView();
    // ----- Test Case ------------------------------------------------
    PrimitivePath colU_inrCase = addFormulaColumn(view, Path.ROOT, "inrCase", "null");
    PrimitivePath colU_inrA = addFormulaColumn(view, Path.ROOT, "inrA", "null");
    PrimitivePath colU_inrB = addFormulaColumn(view, Path.ROOT, "inrB", "null");
    RelationPath tabTworows = addRelation(view, Path.ROOT, "tworows");
    setInstantiatedTable(view, tabTworows, "tworows");
    PrimitivePath colTworows_case = addSQLColumn(view, tabTworows, "pk_case");
    RelationPath tabN = addRelation(view, Path.ROOT, "tabN");
    setInstantiatedTable(view, tabN, "tabN");
    PrimitivePath colN_NA = addSQLColumn(view, tabN, "pk_colNA");
    PrimitivePath colN_NB = addSQLColumn(view, tabN, "colNB_n");
    PrimitivePath colN_NC = addSQLColumn(view, tabN, "colNC_n");
    setFormula(view, colU_inrCase, "if([pk_case] = 1, \"one\", \"two\")");
    setFormula(view, colU_inrA, "[pk_colNA]");
    setFormula(view, colU_inrB, "[colNB_n]");
    expectResult(view, true, "["
        + "(one,1,0,[(1)],[(1,0,0)]),"
        + "(one,2,0,[(1)],[(2,0,1)]),"
        + "(one,3,1,[(1)],[(3,1,0)]),"
        + "(one,4,1,[(1)],[(4,1,1)]),"
        + "(one,5,_,[(1)],[(5,_,0)]),"
        + "(one,6,_,[(1)],[(6,_,1)]),"
        + "(one,7,_,[(1)],[(7,_,_)]),"
        + "(two,1,0,[(2)],[(1,0,0)]),"
        + "(two,2,0,[(2)],[(2,0,1)]),"
        + "(two,3,1,[(2)],[(3,1,0)]),"
        + "(two,4,1,[(2)],[(4,1,1)]),"
        + "(two,5,_,[(2)],[(5,_,0)]),"
        + "(two,6,_,[(2)],[(6,_,1)]),"
        + "(two,7,_,[(2)],[(7,_,_)])"
        + "]");
    view.set(CProp.VISIBLE, colU_inrA, false);
    expectResult(view, true, "["
        + "(one,.,0,[(1)],[(1,0,0),(2,0,1)]),"
        + "(one,.,1,[(1)],[(3,1,0),(4,1,1)]),"
        + "(one,.,_,[(1)],[(5,_,0),(6,_,1),(7,_,_)]),"
        + "(two,.,0,[(2)],[(1,0,0),(2,0,1)]),"
        + "(two,.,1,[(2)],[(3,1,0),(4,1,1)]),"
        + "(two,.,_,[(2)],[(5,_,0),(6,_,1),(7,_,_)])"
        + "]");
    view.set(CProp.VISIBLE, colU_inrCase, false);
    expectResult(view, true, "["
        + "(.,.,0,[(1),(2)],[(1,0,0),(2,0,1)]),"
        + "(.,.,1,[(1),(2)],[(3,1,0),(4,1,1)]),"
        + "(.,.,_,[(1),(2)],[(5,_,0),(6,_,1),(7,_,_)])"
        + "]");
  }

  @Test
  public void testFilterOnNullAggregateOutput() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z1", "X1", 1,
      "Z2", "X2", 3,
      "Z3", "X3", 10,
      "Z4", "X3", 14,
      "Z5", "X5", 100,
      "Z6", "X6", null,
      "Z6", "X9", 10000,
      "Z7", "XX", 100000,
      "Z8", "X3", 2,
      "Z9", "X3", 2,
      "Z10", "X6", null,
      "Z11", "X6", null,
      "Z12", "X7", 0,
    });


    MutableView view = createView();
    view.set(CProp.LABEL, Path.ROOT, "tabX");
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_aggmax = addFormulaColumn(view, Path.ROOT, "aggmax", "null");
    PrimitivePath colX_aggstdev = addFormulaColumn(view, Path.ROOT, "aggstdev", "null");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    setFormula(view, colX_aggmax, "max([colZC_n])");
    setFormula(view, colX_aggstdev, "stdev([colZC_n])");
    expectResult(view, true, "["
        + "(X1,a,1,_,[(Z1,X1,1)]),"
        + "(X2,a,3,_,[(Z2,X2,3)]),"
        + "(X3,a,14,6,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,_,_,[]),"
        + "(X5,b,100,_,[(Z5,X5,100)]),"
        + "(X6,_,_,_,[(Z10,X6,_),(Z11,X6,_),(Z6,X6,_)]),"
        + "(X7,c,0,_,[(Z12,X7,0)]),"
        + "(X8,d,_,_,[]),"
        + "(X9,c,10000,_,[(Z6,X9,10000)])"
        + "]");
    setFilter(view, colX_aggmax, "[aggmax] is null");
    expectResult(view, true, "["
        + "(X4,b,_,_,[]),"
        + "(X6,_,_,_,[(Z10,X6,_),(Z11,X6,_),(Z6,X6,_)]),"
        + "(X8,d,_,_,[])"
        + "]");
    setFilter(view, colX_aggmax, "not ([aggmax] is null)");
    expectResult(view, true, "["
        + "(X1,a,1,_,[(Z1,X1,1)]),"
        + "(X2,a,3,_,[(Z2,X2,3)]),"
        + "(X3,a,14,6,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X5,b,100,_,[(Z5,X5,100)]),"
        + "(X7,c,0,_,[(Z12,X7,0)]),"
        + "(X9,c,10000,_,[(Z6,X9,10000)])"
        + "]");
    setFilter(view, colX_aggmax, "not ([aggmax] is 0)");
    expectResult(view, true, "["
        + "(X1,a,1,_,[(Z1,X1,1)]),"
        + "(X2,a,3,_,[(Z2,X2,3)]),"
        + "(X3,a,14,6,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,_,_,[]),"
        + "(X5,b,100,_,[(Z5,X5,100)]),"
        + "(X6,_,_,_,[(Z10,X6,_),(Z11,X6,_),(Z6,X6,_)]),"
        + "(X8,d,_,_,[]),"
        + "(X9,c,10000,_,[(Z6,X9,10000)])"
        + "]");
    setFilter(view, colX_aggmax, "[aggmax] is 14 or [aggmax] is 0");
    expectResult(view, true, "["
        + "(X3,a,14,6,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X7,c,0,_,[(Z12,X7,0)])"
        + "]");
    setFilter(view, colX_aggmax, "[aggmax] is null or [aggmax] is 14");
    expectResult(view, true, "["
        + "(X3,a,14,6,[(Z3,X3,10),(Z4,X3,14),(Z8,X3,2),(Z9,X3,2)]),"
        + "(X4,b,_,_,[]),"
        + "(X6,_,_,_,[(Z10,X6,_),(Z11,X6,_),(Z6,X6,_)]),"
        + "(X8,d,_,_,[])"
        + "]");
    setFilter(view, colX_aggmax, "not ([aggmax] is null or [aggmax] is 14)");
    expectResult(view, true, "["
        + "(X1,a,1,_,[(Z1,X1,1)]),"
        + "(X2,a,3,_,[(Z2,X2,3)]),"
        + "(X5,b,100,_,[(Z5,X5,100)]),"
        + "(X7,c,0,_,[(Z12,X7,0)]),"
        + "(X9,c,10000,_,[(Z6,X9,10000)])"
        + "]");
  }

  @Test
  public void testNoInstAggregate() {
    createTable("tab_empty", new Object[] {
    });


    MutableView view = createView();
    PrimitivePath colU_acount = addFormulaColumn(view, Path.ROOT, "acount", "count(1)");
    PrimitivePath colU_acountr = addFormulaColumn(view, Path.ROOT, "acountr", "count([root])");
    PrimitivePath colU_amax = addFormulaColumn(view, Path.ROOT, "amax", "max(1)");
    PrimitivePath colU_asum = addFormulaColumn(view, Path.ROOT, "asum", "sum(1)");
    expectResult(view, true, "[(1,1,1,1)]");
    PrimitivePath colU_filt = addFormulaColumn(view, Path.ROOT, "filt", "'a'");
    setFilter(view, colU_filt, "[filt] is 'a'");
    expectResult(view, true, "[(1,1,1,1,a)]");
    setFilter(view, colU_filt, "[filt] is 'b'");
    expectResult(view, true, "[]");
    setFilter(view, colU_filt, "[filt] is 'a'");
    setFormula(view, colU_amax, "max([acountr])");
    setFormula(view, colU_asum, "sum([amax])");
    expectResult(view, true, "[(1,1,1,1,a)]");
    setFilter(view, colU_filt, "[filt] is 'b'");
    expectResult(view, true, "[]");
    setFilter(view, colU_filt, "[filt] is 'a'");
    setInstantiatedTable(view, Path.ROOT, "tab_empty");
    expectResult(view, true, "[]");
    view.removePath(colU_acountr);
    view.removePath(colU_amax);
    view.removePath(colU_asum);
    view.removePath(colU_filt);
    expectResult(view, true, "[]");

    setInstantiatedTable(view, Path.ROOT, null);
    expectResult(view, true, "[(1)]");
    PrimitivePath colU_sql = addSQLColumn(view, Path.ROOT, "nosuch");
    expectResult(view, true, "[(1,_)]", ImmutableSet.of(colU_sql));
  }

  @Test
  public void testLateralBidirectional() {
    createTable("tabQ", new Object[] {
      "pk_colQA", "colQB_n",
      String.class, String.class,
      "Q01", "a",
      "Q02", "b",
      "Q03", null,
      "Q04", "c",
      "Q05", "d",
      "Q06", "e",
      "Q07", "e",
      "Q08", "e",
      "Q09-", "b",
      "Q10", "k",
    });
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC",
      String.class, String.class, String.class,
      "X01", "a", "L",
      "X02", "a", "S",
      "X03", "a", "L",
      "X04", "b", "L",
      "X05", "b", "L",
      "X06", null, "L",
      "X07", "c", "L",
      "X08", "d", "L",
      "X09", "c", "L",
      "X10", "e", "L"
    });
    createTable("tabZ", new Object[] {
      "pk_colZA", "pk_colZB", "colZC_n",
      String.class, String.class, Integer.class,
      "Z01", "X01", 1,
      "Z02", "X02", 3,
      "Z03", "X03", 10,
      "Z04", "X03", 14,
      "Z05", "X05", 100,
      "Z06", "X06", 1000,
      "Z06", "X09", 10000,
      "Z07", "X0X", 100000,
      "Z08", "X03", 2,
      "Z09", "X03", 2,
      "Z10", "X06", 0,
      "Z11", "X06", null,
      "Z12", "X07", 0,
      "Z13", "X02", 3,
      "Z14", "X03", 14,
      "Z15", "X05", 101,
    });

    MutableView view = createView();
    view.set(CProp.LABEL, Path.ROOT, "tabX");
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_XC = addSQLColumn(view, Path.ROOT, "colXC");
    RelationPath tabZ = addRelation(view, Path.ROOT, "tabZ");
    setInstantiatedTable(view, tabZ, "tabZ");
    PrimitivePath colZ_ZA = addSQLColumn(view, tabZ, "pk_colZA");
    PrimitivePath colZ_ZB = addSQLColumn(view, tabZ, "pk_colZB");
    view.set(CProp.VISIBLE, colZ_ZB, false);
    view.set(QProp.JOINED_ON, colZ_ZB, OptionalPrimitivePath.of(colX_XA));
    PrimitivePath colZ_ZC = addSQLColumn(view, tabZ, "colZC_n");
    RelationPath tabQ = addRelation(view, Path.ROOT, "tabQ");
    setInstantiatedTable(view, tabQ, "tabQ");
    PrimitivePath colQ_QA = addSQLColumn(view, tabQ, "pk_colQA");
    PrimitivePath colQ_QB = addSQLColumn(view, tabQ, "colQB_n");
    view.set(QProp.JOINED_ON, colQ_QB, OptionalPrimitivePath.of(colX_XB));


    expectResult(view, true, "["
        + "(X01,a,L,[(Z01,X01,1)],[(Q01,a)]),"
        + "(X02,a,S,[(Z02,X02,3),(Z13,X02,3)],[(Q01,a)]),"
        + "(X03,a,L,[(Z03,X03,10),(Z04,X03,14),(Z08,X03,2),(Z09,X03,2),(Z14,X03,14)],[(Q01,a)]),"
        + "(X04,b,L,[],[(Q02,b),(Q09-,b)]),"
        + "(X05,b,L,[(Z05,X05,100),(Z15,X05,101)],[(Q02,b),(Q09-,b)]),"
        + "(X06,_,L,[(Z06,X06,1000),(Z10,X06,0),(Z11,X06,_)],[]),"
        + "(X07,c,L,[(Z12,X07,0)],[(Q04,c)]),"
        + "(X08,d,L,[],[(Q05,d)]),"
        + "(X09,c,L,[(Z06,X09,10000)],[(Q04,c)]),"
        + "(X10,e,L,[],[(Q06,e),(Q07,e),(Q08,e)])"
        + "]");
    PrimitivePath colZ_fcq = addFormulaColumn(view, tabZ, "fcq", "count([pk_colQA])");
    PrimitivePath colQ_fcz = addFormulaColumn(view, tabQ, "fcz", "count([tabZ])");
    String res1 = "["
        + "(X01,a,L,[(Z01,X01,1,1)],[(Q01,a,1)]),"
        + "(X02,a,S,[(Z02,X02,3,1),(Z13,X02,3,1)],[(Q01,a,2)]),"
        + "(X03,a,L,[(Z03,X03,10,1),(Z04,X03,14,1),(Z08,X03,2,1),(Z09,X03,2,1),(Z14,X03,14,1)],[(Q01,a,5)]),"
        + "(X04,b,L,[],[(Q02,b,0),(Q09-,b,0)]),"
        + "(X05,b,L,[(Z05,X05,100,2),(Z15,X05,101,2)],[(Q02,b,2),(Q09-,b,2)]),"
        + "(X06,_,L,[(Z06,X06,1000,0),(Z10,X06,0,0),(Z11,X06,_,0)],[]),"
        + "(X07,c,L,[(Z12,X07,0,1)],[(Q04,c,1)]),"
        + "(X08,d,L,[],[(Q05,d,0)]),"
        + "(X09,c,L,[(Z06,X09,10000,1)],[(Q04,c,1)]),"
        + "(X10,e,L,[],[(Q06,e,0),(Q07,e,0),(Q08,e,0)])"
        + "]";
    expectResult(view, true, res1);
    setFormula(view, colZ_fcq, "count([fcz])");
    if (!SKIP_KNOWN_FAILING_TESTS) {
      expectResult(view, true, res1);
      setFormula(view, colQ_fcz, "8");
      expectResult(view, true, "["
        + "(X01,a,L,[(Z01,X01,1,1)],[(Q01,a,8)]),"
        + "(X02,a,S,[(Z02,X02,3,1),(Z13,X02,3,1)],[(Q01,a,8)]),"
        + "(X03,a,L,[(Z03,X03,10,1),(Z04,X03,14,1),(Z08,X03,2,1),(Z09,X03,2,1),(Z14,X03,14,1)],[(Q01,a,8)]),"
        + "(X04,b,L,[],[(Q02,b,8),(Q09-,b,8)]),"
        + "(X05,b,L,[(Z05,X05,100,2),(Z15,X05,101,2)],[(Q02,b,8),(Q09-,b,8)]),"
        + "(X06,_,L,[(Z06,X06,1000,0),(Z10,X06,0,0),(Z11,X06,_,0)],[]),"
        + "(X07,c,L,[(Z12,X07,0,1)],[(Q04,c,8)]),"
        + "(X08,d,L,[],[(Q05,d,8)]),"
        + "(X09,c,L,[(Z06,X09,10000,1)],[(Q04,c,8)]),"
        + "(X10,e,L,[],[(Q06,e,8),(Q07,e,8),(Q08,e,8)])"
        + "]");
    }
    setFormula(view, colQ_fcz, "len([pk_colQA])");
    expectResult(view, true, "["
      + "(X01,a,L,[(Z01,X01,1,1)],[(Q01,a,3)]),"
      + "(X02,a,S,[(Z02,X02,3,1),(Z13,X02,3,1)],[(Q01,a,3)]),"
      + "(X03,a,L,[(Z03,X03,10,1),(Z04,X03,14,1),(Z08,X03,2,1),(Z09,X03,2,1),(Z14,X03,14,1)],[(Q01,a,3)]),"
      + "(X04,b,L,[],[(Q02,b,3),(Q09-,b,4)]),"
      + "(X05,b,L,[(Z05,X05,100,2),(Z15,X05,101,2)],[(Q02,b,3),(Q09-,b,4)]),"
      + "(X06,_,L,[(Z06,X06,1000,0),(Z10,X06,0,0),(Z11,X06,_,0)],[]),"
      + "(X07,c,L,[(Z12,X07,0,1)],[(Q04,c,3)]),"
      + "(X08,d,L,[],[(Q05,d,3)]),"
      + "(X09,c,L,[(Z06,X09,10000,1)],[(Q04,c,3)]),"
      + "(X10,e,L,[],[(Q06,e,3),(Q07,e,3),(Q08,e,3)])"
      + "]");
    setFormula(view, colZ_fcq, "[fcz]");
    setFormula(view, colQ_fcz, "[fcq]");
    String resErr1 = "["
      + "(X01,a,L,[(Z01,X01,1,_)],[(Q01,a,_)]),"
      + "(X02,a,S,[(Z02,X02,3,_),(Z13,X02,3,_)],[(Q01,a,_)]),"
      + "(X03,a,L,[(Z03,X03,10,_),(Z04,X03,14,_),(Z08,X03,2,_),(Z09,X03,2,_),(Z14,X03,14,_)],[(Q01,a,_)]),"
      + "(X04,b,L,[],[(Q02,b,_),(Q09-,b,_)]),"
      + "(X05,b,L,[(Z05,X05,100,_),(Z15,X05,101,_)],[(Q02,b,_),(Q09-,b,_)]),"
      + "(X06,_,L,[(Z06,X06,1000,_),(Z10,X06,0,_),(Z11,X06,_,_)],[]),"
      + "(X07,c,L,[(Z12,X07,0,_)],[(Q04,c,_)]),"
      + "(X08,d,L,[],[(Q05,d,_)]),"
      + "(X09,c,L,[(Z06,X09,10000,_)],[(Q04,c,_)]),"
      + "(X10,e,L,[],[(Q06,e,_),(Q07,e,_),(Q08,e,_)])"
      + "]";

    expectResult(view, true, resErr1, ImmutableSet.of(colZ_fcq, colQ_fcz));
    setFormula(view, colZ_fcq, "count([fcz])");
    if (!SKIP_KNOWN_FAILING_TESTS) {
      expectResult(view, true, resErr1, ImmutableSet.of(colZ_fcq, colQ_fcz));
    }
  }

  @Test
  public void testSameExpressionTwice() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n", "colXC",
      String.class, String.class, String.class,
      "X01", "a", "L",
      "X02", "a", "S",
      "X03", "a", "L",
      "X04", "b", "L",
      "X05", "b", "L",
      "X06", null, "L",
      "X07", "c", "L"
    });
    MutableView view = createView();


    view.set(CProp.LABEL, Path.ROOT, "tabX");
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    PrimitivePath colX_XC = addSQLColumn(view, Path.ROOT, "colXC");
    expectResult(view, true, "["
        + "(X01,a,L),"
        + "(X02,a,S),"
        + "(X03,a,L),"
        + "(X04,b,L),"
        + "(X05,b,L),"
        + "(X06,_,L),"
        + "(X07,c,L)"
        + "]");
    PrimitivePath colX_ref = addFormulaColumn(view, Path.ROOT, "ref", "[pk_colXA]");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_ref, colX_XA, colX_XB, colX_XC)));
    String res1 = "["
        + "(X01,X01,a,L),"
        + "(X02,X02,a,S),"
        + "(X03,X03,a,L),"
        + "(X04,X04,b,L),"
        + "(X05,X05,b,L),"
        + "(X06,X06,_,L),"
        + "(X07,X07,c,L)"
        + "]";
    expectResult(view, true, res1);
    view.set(CProp.VISIBLE, colX_XA, false);
    expectResult(view, true, res1);
    PrimitivePath colX_ref2 = addFormulaColumn(view, Path.ROOT, "ref2", "[pk_colXA]");
    view.set(SProp.CHILD_PATHS, Path.ROOT, new PathList(ImmutableList.of(colX_ref, colX_ref2, colX_XA, colX_XB, colX_XC)));
    expectResult(view, true, "["
        + "(X01,X01,X01,a,L),"
        + "(X02,X02,X02,a,S),"
        + "(X03,X03,X03,a,L),"
        + "(X04,X04,X04,b,L),"
        + "(X05,X05,X05,b,L),"
        + "(X06,X06,X06,_,L),"
        + "(X07,X07,X07,c,L)"
        + "]");
    setFormula(view, colX_ref, "[colXB_n]");
    setFormula(view, colX_ref2, "[colXB_n]");
    expectResult(view, true, "["
        + "(_,_,X06,_,L),"
        + "(a,a,X01,a,L),"
        + "(a,a,X02,a,S),"
        + "(a,a,X03,a,L),"
        + "(b,b,X04,b,L),"
        + "(b,b,X05,b,L),"
        + "(c,c,X07,c,L)"
        + "]");
  }


  @Test
  public void testManyToManyJoinOutwardAggregate() {
    createTable("tabX", new Object[] {
      "pk_colXA", "colXB_n",
      String.class, String.class,
      "X1", "a",
      "X2-", "a",
      "X3", "a",
      "X4", "b",
      "X5", "b",
      "X6", null,
      "X7-------", "c",
      "X8", "d",
      "X9", "c",
    });
    createTable("tabY2", new Object[] {
      "pk_colYA", "colYB_n", "colYC_n",
      String.class, String.class, String.class,
      "Y1", "Y1d", "a",
      "Y2", "Y2d", "c",
      "Y3", "Y3d", "c",
      "Y5", "Y5d", "e"
    });


    MutableView view = createView();
    view.set(CProp.LABEL, Path.ROOT, "tabX");
    setInstantiatedTable(view, Path.ROOT, "tabX");
    PrimitivePath colX_XA = addSQLColumn(view, Path.ROOT, "pk_colXA");
    PrimitivePath colX_XB = addSQLColumn(view, Path.ROOT, "colXB_n");
    RelationPath tabY2 = addRelation(view, Path.ROOT, "tabY2");
    setInstantiatedTable(view, tabY2, "tabY2");
    PrimitivePath colY2_YA = addSQLColumn(view, tabY2, "pk_colYA");
    PrimitivePath colY2_YB = addSQLColumn(view, tabY2, "colYB_n");
    PrimitivePath colY2_YC = addSQLColumn(view, tabY2, "colYC_n");
    view.set(QProp.JOINED_ON, colY2_YC, OptionalPrimitivePath.of(colX_XB));
    RelationPath tabX2 = addRelation(view, tabY2, "tabX");
    view.set(CProp.LABEL, tabX2, "tabXinner");
    setInstantiatedTable(view, tabX2, "tabX");
    PrimitivePath colX2_XA = addSQLColumn(view, tabX2, "pk_colXA");
    PrimitivePath colX2_XB = addSQLColumn(view, tabX2, "colXB_n");
    view.set(CProp.VISIBLE, colX2_XB, false);
    view.set(QProp.JOINED_ON, colX2_XB, OptionalPrimitivePath.of(colY2_YC));
    expectResult(view, true, "["
        + "(X1,a,[(Y1,Y1d,a,[(X1,.),(X2-,.),(X3,.)])]),"
        + "(X2-,a,[(Y1,Y1d,a,[(X1,.),(X2-,.),(X3,.)])]),"
        + "(X3,a,[(Y1,Y1d,a,[(X1,.),(X2-,.),(X3,.)])]),"
        + "(X4,b,[]),"
        + "(X5,b,[]),"
        + "(X6,_,[]),"
        + "(X7-------,c,[(Y2,Y2d,c,[(X7-------,.),(X9,.)]),(Y3,Y3d,c,[(X7-------,.),(X9,.)])]),"
        + "(X8,d,[]),"
        + "(X9,c,[(Y2,Y2d,c,[(X7-------,.),(X9,.)]),(Y3,Y3d,c,[(X7-------,.),(X9,.)])])"
        + "]");
    PrimitivePath colY2_fout = addFormulaColumn(view, tabY2, "fout", "sum(len([pk_colXA]))");
    PrimitivePath colY2_fin = addFormulaColumn(view, tabY2, "fin", "sum(len([tabXinner\\pk_colXA]))");
    view.set(SProp.CHILD_PATHS, tabY2, new PathList(ImmutableList.of(colY2_YA, colY2_YB, colY2_YC, colY2_fout, colY2_fin, tabX2)));
    expectResult(view, true, "["
        + "(X1,a,[(Y1,Y1d,a,2,7,[(X1,.),(X2-,.),(X3,.)])]),"
        + "(X2-,a,[(Y1,Y1d,a,3,7,[(X1,.),(X2-,.),(X3,.)])]),"
        + "(X3,a,[(Y1,Y1d,a,2,7,[(X1,.),(X2-,.),(X3,.)])]),"
        + "(X4,b,[]),"
        + "(X5,b,[]),"
        + "(X6,_,[]),"
        + "(X7-------,c,[(Y2,Y2d,c,9,11,[(X7-------,.),(X9,.)]),(Y3,Y3d,c,9,11,[(X7-------,.),(X9,.)])]),"
        + "(X8,d,[]),"
        + "(X9,c,[(Y2,Y2d,c,2,11,[(X7-------,.),(X9,.)]),(Y3,Y3d,c,2,11,[(X7-------,.),(X9,.)])])"
        + "]");
    PrimitivePath colX2_outerXA = addFormulaColumn(view, tabX2, "outerXA", "[pk_colXA]");
    PrimitivePath colX2_outerXB = addFormulaColumn(view, tabX2, "outerXB", "[colXB_n]");
    PrimitivePath colX2_midfout = addFormulaColumn(view, tabX2, "midfout", "[fout]");
    PrimitivePath colX2_midfin = addFormulaColumn(view, tabX2, "midfin", "[fin]");
    PrimitivePath colX2_midYA = addFormulaColumn(view, tabX2, "midYA", "[pk_colYA]");
    view.set(SProp.CHILD_PATHS, tabX2, new PathList(ImmutableList.of(colX2_XA, colX2_XB, colX2_outerXA, colX2_outerXB, colX2_midfout, colX2_midfin, colX2_midYA)));
    expectResult(view, true, "["
        + "(X1,a,[(Y1,Y1d,a,2,7,[(X1,.,X1,a,2,7,Y1),(X2-,.,X1,a,2,7,Y1),(X3,.,X1,a,2,7,Y1)])]),"
        + "(X2-,a,[(Y1,Y1d,a,3,7,[(X1,.,X2-,a,3,7,Y1),(X2-,.,X2-,a,3,7,Y1),(X3,.,X2-,a,3,7,Y1)])]),"
        + "(X3,a,[(Y1,Y1d,a,2,7,[(X1,.,X3,a,2,7,Y1),(X2-,.,X3,a,2,7,Y1),(X3,.,X3,a,2,7,Y1)])]),"
        + "(X4,b,[]),"
        + "(X5,b,[]),"
        + "(X6,_,[]),"
        + "(X7-------,c,[(Y2,Y2d,c,9,11,[(X7-------,.,X7-------,c,9,11,Y2),(X9,.,X7-------,c,9,11,Y2)]),(Y3,Y3d,c,9,11,[(X7-------,.,X7-------,c,9,11,Y3),(X9,.,X7-------,c,9,11,Y3)])]),"
        + "(X8,d,[]),"
        + "(X9,c,[(Y2,Y2d,c,2,11,[(X7-------,.,X9,c,2,11,Y2),(X9,.,X9,c,2,11,Y2)]),(Y3,Y3d,c,2,11,[(X7-------,.,X9,c,2,11,Y3),(X9,.,X9,c,2,11,Y3)])])"
        + "]");
  }
}
